// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   texture.h
 * Author: gleseach
 *
 * Created on 26 mai 2016, 12:52
 */

#ifndef TEXTURE_OGL3_H
#define TEXTURE_OGL3_H

#include "3d/texture.h"

namespace OGL3
{

class TextureGL3 final: public OGL::Texture
{
public:
    TextureGL3(OGL::TextureType type);
    ~TextureGL3() override;

    void create(OGL::Context* context) override;
    void destroy() override;

    void bind() const override;
    void unbind() const override;

    void allocate1D(int width, int channels, DataType data_type, void * buffer) override;
    void allocate2D(int width, int height, int channels, DataType data_type, void * buffer) override;
    void allocate2D(const Image& image) override;
    void allocate2D(const std::vector<Image>& mipmaps) override;

    /*!
     * \brief Set the wrapping and the filtering modes for the current object.
     * \warning After the call of this function the object is in unbound state !
     */
    void setWrapAndFilterMode(OGL::WrapMode wrap, OGL::FilterMode filter) override;

    /*! 
     * \brief Set the wrapping and the filtering modes for the current object.
     * \warning The current object have to be bound before calling this function.
     */
    void setWrapAndFilterModeNoBind(OGL::WrapMode wrap, OGL::FilterMode filter) override;

    /*! 
     * \brief Set the wrapping mode for the current object.
     * \warning After the call of this function the object is in unbound state !
     */
    void setWrapMode(OGL::WrapMode wrap) override;

    /*! 
     * \brief Set the filtering mode for the current object.
     * \warning After the call of this function the object is in unbound state !
     */
    void setFilterMode(OGL::FilterMode filter) override;

    /*!
     * \brief Set the filtering mode for the current object.
     * \warning After the call of this function the object is in unbound state !
     */
    void setFilterMode(OGL::FilterMode minfilter, OGL::FilterMode magfilter) override;

    /*!
     * \brief Set the anisotropic filter value for the current object. 1.f disable the filter, > 1.f enable the filter.
     * \warning After the call of this function the object is in unbound state !
     */
    void setAnisotropicFilter(float val) override;

    /*!
     * \brief Set the anisotropic filter value for the current object. 1.f disable the filter, > 1.f enable the filter.
     * \warning The current object have to be bound before calling this function.
     */
    void setAnisotropicFilterNoBind(float val) override;

    /*!
     * \brief Set the wrapping mode for the current object without change the binding state.
     * \warning The current object have to be bound before calling this function.
     */
    void setWrapModeNoBind(OGL::WrapMode wrap) override;

    /*! 
     * \brief Set the filtering mode for the current object without change the binding state.
     * \warning The current object have to be bound before calling this function.
     */
    void setFilterModeNoBind(OGL::FilterMode filter) override;

    /*!
     * \brief Set the filtering mode for the current object without change the binding state.
     * \warning The current object have to be bound before calling this function.
     * \note magfilter can only be set to linear or nearest.
     */
    void setFilterModeNoBind(OGL::FilterMode minfilter, OGL::FilterMode magfilter) override;

    /*! 
     * \brief Resize the buffer an already created object.
     * \warning Don't call if the object doesn't exist.
     */
    void resize(int width, int height) override;

    /*!
     * \brief Retrieve the current bound texture ID and compare with the object ID.
     * \return true if the current object is bound, false if not.
     */
    bool isCurrentBinding() const override;

    /*!
     * \brief Update the entire texture data with the array passed in parameter.
     * \warning The array passed in parameter has to be the same size than the texture array.
     */
    void setData(void * data) override;

    /*!
     * \brief Update a part of the texture data.
     * \param[in] x Column of the origin.
     * \param[in] y Row of the origin.
     * \param[in] width Width of the part to write.
     * \param[in] height Height of the part to write.
     * \param[in] data Pointer to the data to write.
     */
    void setData(unsigned int x, unsigned int y, unsigned int width, unsigned int height, void * data) override;

    void dumpToImage(Image& img) override;

    void generateMipmaps() override;

private:
    void allocate(int * dimensions, int size, int channels, DataType data_type, void * buffer) override;
};

}

#endif /* TEXTURE_OGL3_H */

