// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   context.h
 * Author: gleseach
 *
 * Created on 26 mai 2016, 12:39
 */

#ifndef CONTEXT_H
#define CONTEXT_H

#include "3d/resources.h"
#include "3d/opengl_enum.h"

class TypedBuffer;
struct PipelineParameters;

namespace OGL
{

class Texture;
class Buffer;
class FrameBuffer;
class Program;
class Shader;
class RenderBuffer;
class VertexArrayObject;
class VertexAttribArray;
class DrawMode;
class Utils2;
class ProgramBindHelper;
class ShaderManager;
class Query;
class Sync;
class VertexAttribDivisor;

class Context
{
public:
    Context();
    virtual ~Context();

    virtual bool                      init(const SEAScope::RendererResources*) = 0;
    virtual OGL::Texture*             createTexture(OGL::TextureType type) = 0;
    virtual OGL::Buffer*              createBuffer(OGL::BufferType type) = 0;
    virtual OGL::FrameBuffer*         createFramebuffer() = 0;
    virtual OGL::Program*             createProgram() = 0;
    virtual OGL::Shader*              createShader(OGL::ShaderType type, const std::string& name) = 0;
    virtual OGL::RenderBuffer*        createRenderbuffer(OGL::RenderBufferType type) = 0;
    virtual OGL::VertexArrayObject*   createVertexArrayObject() = 0;
    virtual OGL::VertexAttribArray*   createVertexAttribArray(unsigned int id, unsigned int component_count, Buffer * buffer, DataType dataType = DataType::float32, unsigned int stride = 0, unsigned int offset = 0) = 0;
    virtual OGL::VertexAttribArray*   createVertexAttribArray(unsigned int id, const TypedBuffer& typedBuffer, unsigned int stride = 0, unsigned int offset = 0) = 0;
    virtual OGL::DrawMode*            createDrawMode() = 0;
    virtual OGL::Utils2*              createUtils() = 0;
    virtual OGL::ProgramBindHelper*   createProgramHelper() = 0;
    virtual OGL::Query*               createQuery() = 0;
    virtual OGL::Sync*                createSync() = 0;
    virtual OGL::VertexAttribDivisor* createVertexAttribDivisor(OGL::VertexAttribArray* vaa, unsigned int divisor) = 0;

    virtual std::string getShaderPath() const = 0;
    virtual void deactiveTextureUnit(unsigned int id) = 0;
    virtual void activeTextureUnit(unsigned int id, OGL::Texture* object) = 0;
    virtual void setPolygonMode(OGL::Face face, OGL::RasterizationMode mode) = 0;
    virtual void setFlag(OGL::Capability cap, bool status) = 0;
    virtual void setViewport(const glm::ivec4& size) = 0;
    virtual glm::ivec4 getViewportSize() = 0;
    virtual void setClearColor(const glm::vec4& color) = 0;
    virtual void clear(int bits) = 0;
    virtual void setDepthMask(bool status) = 0;
    virtual void setColorMask(bool red, bool green, bool blue, bool alpha) = 0;
    virtual void setBlendFunc(OGL::BlendingFactor sourceFactor, OGL::BlendingFactor destinationFactor) = 0;
    virtual void setStencilFunc(OGL::Func func, int ref, unsigned int mask) = 0;
    virtual void setStencilMask(bool status) = 0;
    virtual void setStencilOp(OGL::StencilOp sfail, OGL::StencilOp dpfail, OGL::StencilOp dppass) = 0;
	virtual const SEAScope::RendererResources* getResources() const = 0;
    virtual void setLineWidth(float val) = 0;

    OGL::Texture* getCurrentTextureBind(unsigned int textureUnit);
    void setShaderManager(OGL::ShaderManager* shaderMngr);
    OGL::ShaderManager* getShaderManager() const;

    PipelineParameters* pipelineParameters();

    virtual int getGPUInfoInteger(GPUInfoInteger val) const = 0;
    virtual float getGPUInfoFloat(GPUInfoFloat val) const = 0;

    void setPixelSize(float size) {_pixelSize = size;}
    float getPixelSize() {return _pixelSize;}
    void setPixelSizeUnit(float val) { _pixelSizeUnit = val; }
    float getPixelSizeUnit() { return _pixelSizeUnit; }

protected:
    virtual void activeTextureUnit(unsigned int id) = 0;

    std::array<std::pair<bool, OGL::Texture*>, 8> _textureUnit;

protected:
    OGL::ShaderManager* _shaderMngr;
    PipelineParameters* _pipelineParams;
    float _pixelSize;
    float _pixelSizeUnit;
    unsigned int _msaa;
	const SEAScope::RendererResources* _resources;
};

}

#endif /* CONTEXT_H */

