// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/*
 * File:   data_set.h
 * Author: gleseach
 *
 * Created on 26 août 2015, 15:17
 */

#ifndef DATA_SET_H
#define	DATA_SET_H

#include "core/structs.h"
#include "3d/pickable.h"
#include "3d/opengl_enum.h"

namespace OGL
{
    class StaticMesh;
}

class Image;
class StreamlinesData;
class ParticlesData;

class DataSet : public Pickable
{
protected:
    std::uint32_t _width;
    std::uint32_t _height;
    float _resolution;

    unsigned int _dimensionCount;
    Image* _data;
    Image* _dataMask;
    float* _offset;
    float* _scaleFactor;
    float _scaleFactorGL;        // 1.0 -> 255.0 for unsigned byte raw data

    const float* lons;
    const float* lats;
    unsigned int lat_size;
    unsigned int lon_size;
    float* min;
    float* max;

    glm::vec4 corners[5];
    glm::vec3 normal;
    glm::vec4 centerPlane;

    const std::uint32_t* gcpx;
    const std::uint32_t* gcpy;

    OGL::StaticMesh* mesh_spherical;
    OGL::StaticMesh* mesh_planar;

    StreamlinesData* streamlinesData;
    ParticlesData* _particlesData;

    std::vector<glm::vec3> _gcps;

    glm::bvec2 _seamLess;
    std::vector<glm::ivec4> _areas;
    bool _areaChanged;

public:
    DataSet(const SEAScope::RenderableId& rId);
    ~DataSet() override;

    // Prevent copies
    DataSet(const DataSet&) = delete;
    DataSet& operator=(const DataSet&) = delete;

    // Allocate buffers and send data to the gpu
    virtual void generateGPUData(OGL::Context* ctx);
    virtual bool compute(int);

    //! Return the offset and scale factor to uncompress the physical data.
    glm::vec4 uncompressVector() const;
    glm::vec3 uncompressVector3() const;

    float* offsetArray() const;
    float* scaleFactorArray() const;

    virtual void draw(OGL::Context* ctx, const glm::mat4& mvp) override;
    virtual void drawPicking(OGL::Context* ctx, const glm::mat4& mvp) override;

    // Accessors
    void setStreamlinesData(StreamlinesData* streamlinesData);
    StreamlinesData* getStreamlinesData() const;
    void setMesh_planar(OGL::StaticMesh* mesh_planar);
    OGL::StaticMesh* getMesh_planar() const;
    void setMesh_spherical(OGL::StaticMesh* mesh_spherical);
    OGL::StaticMesh* getMesh_spherical() const;
    void setGcpy(const std::uint32_t* gcpy);
    const std::uint32_t* getGcpy() const;
    void setGcpx(const std::uint32_t* gcpx);
    const std::uint32_t* getGcpx() const;
    void setCenterPlane(glm::vec4 centerPlane);
    glm::vec4 getCenterPlane() const;
    void setNormal(glm::vec3 normal);
    glm::vec3 getNormal() const;
    void setCorner(int i, const glm::vec4& cornersIn);
    const glm::vec4& getCorner(int i) const;
    void setMax(float* max);
    float* getMax() const;
    void setMin(float* min);
    float* getMin() const;
    void setLon_size(unsigned int lon_size);
    const unsigned int& getLon_size() const;
    void setLat_size(unsigned int lat_size);
    const unsigned int& getLat_size() const;
    void setLats(const float* lats);
    const float* getLats() const;
    void setLons(const float* lons);
    const float* getLons() const;
    void setScaleFactorGL(float _scaleFactorGL);
    float getScaleFactorGL() const;
    void setScaleFactor(float* _scaleFactor);
    float* getScaleFactor() const;
    void setOffset(float* _offset);
    float* getOffset() const;
    void setDataMask(unsigned int width, unsigned int height, unsigned int componentCount, DataType dataType, void* data);
    Image* getDataMask() const;
    void setData(unsigned int width, unsigned int height, unsigned int componentCount, DataType dataType, void* data);
    Image* getData() const;
    void setDimensionCount(unsigned int _dimensionCount);
    const unsigned int& getDimensionCount() const;
    void setResolution(float _resolution);
    float getResolution() const;
    void setHeight(std::uint32_t _height);
    const std::uint32_t& getHeight() const;
    void setWidth(std::uint32_t _width);
    const std::uint32_t& getWidth() const;
    virtual void setFilterMode(OGL::FilterMode filterMode);
    const std::vector<glm::vec3>& getGcps() const;
    void setSeamLess(bool x, bool y);
    const glm::bvec2& getSeamLess() const;
    void setVisibleAreas(std::vector<glm::ivec4>&& areas);
    const std::vector<glm::ivec4>& getVisibleArea() const;
    ParticlesData* getParticlesData() const { return _particlesData; }
    void setParticlesData(ParticlesData* val) { _particlesData = val; }
    bool getAreaChanged() const { return _areaChanged; }
    void setAreaChanged(bool val) { _areaChanged = val; }
};

#endif	/* DATA_SET_H */
