// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef FRUSTUM_H
#define FRUSTUM_H

#include "3d/plane.h"

// fix nanogui problem with windows.h
#ifdef _WIN32
    #ifdef near
        #undef near
    #endif

    #ifdef far
        #undef far
    #endif
#endif

class Frustum
{
public:
    enum CornerPosition
    {
        topLeft = 0,
        topRight,
        bottomLeft,
        bottomRight
    };

    enum MiddlePosition
    {
        top = 0,
        left,
        right,
        bottom,
        near,
        far
    };

    Frustum();

	void setPerspective(float angle, float ratio, float near, float far);
	void lookAt(const glm::vec3 &p, const glm::vec3 &l, const glm::vec3 &u);
	void lookAt(const glm::mat4& view);

    const glm::vec3& getNearTopLeft() const {return ntl;}
    const glm::vec3& getNearTopRight() const {return ntr;}
    const glm::vec3& getNearBottomLeft() const {return nbl;}
    const glm::vec3& getNearBottomRight() const {return nbr;}

    const glm::vec3& getFarTopLeft() const {return ftl;}
    const glm::vec3& getFarTopRight() const {return ftr;}
    const glm::vec3& getFarBottomLeft() const {return fbl;}
    const glm::vec3& getFarBottomRight() const {return fbr;}

    float getNearWidth() const {return nw;}
    float getNearHeight() const {return nh;}
    float getFarWidth() const {return fw;}
    float getFarHeight() const {return fh;}

    glm::vec3 getTopLeftRay() const;
    glm::vec3 getTopRightRay() const;
    glm::vec3 getBottomLeftRay() const;
    glm::vec3 getBottomRightRay() const;

    glm::vec3 getCornerRay(CornerPosition pos);
    glm::vec3 getMiddleRay(MiddlePosition pos);

    const glm::vec3& getXAxis() const {return X;}
    const glm::vec3& getYAxis() const {return Y;}
    const glm::vec3& getZAxis() const {return Z;}

    void updatePlanes(const glm::mat4& projView);

    bool contain(const glm::vec3& point) const;

    Plane getPlane(MiddlePosition pos) {return _planes[pos];}

private:
    glm::vec3 ntl;
    glm::vec3 ntr;
    glm::vec3 nbl;
    glm::vec3 nbr;
    glm::vec3 ftl;
    glm::vec3 ftr;
    glm::vec3 fbl;
    glm::vec3 fbr;
	float nearD;
    float farD;
    float ratio;
    float angle;
    float tang;
	float nw;
    float nh;
    float fw;
    float fh;
    glm::vec3 X,Y,Z;
    Plane _planes[6];
};

#endif /* FRUSTUM_H */

