// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   granule_tree.h
 * Author: gleseach
 *
 * Created on 19 décembre 2016, 16:24
 */

#ifndef GRANULE_TREE_H
#define GRANULE_TREE_H

#include "core/structs.h"

template<typename T>
class GranuleTree
{
public:
    GranuleTree();
    ~GranuleTree();

    void add(const SEAScope::RenderableId& rId, T object);
    bool remove(const SEAScope::RenderableId& rId, T& obj);
    bool exist(const SEAScope::RenderableId& rId) const;
    bool exist(const SEAScope::RenderableId& rId, T& obj) const;
    T get(const SEAScope::RenderableId& rId);
    void populate(const std::set<SEAScope::RenderableId>& rIds, std::list<SEAScope::RenderableId>& list);
    void clear();

    typedef std::list<std::pair<SEAScope::RenderableId, T>> ListPairT;
    void populatePair(const std::set<SEAScope::RenderableId>& rIds, ListPairT& existing, ListPairT& notFound);

private:
    typedef std::map<SEAScope::VariableId, T> VariableMap;
    typedef std::map<SEAScope::GranuleId, VariableMap> GranuleMap;
    typedef std::map<SEAScope::CollectionId, GranuleMap> CollectionMap;


private:
    CollectionMap _tree;
};

template<typename T>
GranuleTree<T>::GranuleTree()
{
}

template<typename T>
GranuleTree<T>::~GranuleTree()
{
    this->clear();
}

template<typename T>
bool GranuleTree<T>::exist(const SEAScope::RenderableId& rId) const
{
    bool result = false;
    auto collectionIt = _tree.find(rId.collection);

    if (_tree.end() !=  collectionIt)
    {
        auto granuleIt = collectionIt->second.find(rId.granule);

        if (collectionIt->second.end() !=  granuleIt)
        {
            auto variableIt = granuleIt->second.find(rId.variable);

            if (granuleIt->second.end() !=  variableIt)
            {
                result = true;
            }
        }
    }

    return result;
}

template<typename T>
bool GranuleTree<T>::exist(const SEAScope::RenderableId& rId, T& obj) const
{
    bool result = false;
    auto collectionIt = _tree.find(rId.collection);

    if (_tree.end() !=  collectionIt)
    {
        auto granuleIt = collectionIt->second.find(rId.granule);

        if (collectionIt->second.end() !=  granuleIt)
        {
            auto variableIt = granuleIt->second.find(rId.variable);

            if (granuleIt->second.end() !=  variableIt)
            {
                result = true;
                obj = static_cast<T>(variableIt->second);
            }
        }
    }

    return result;
}

template<typename T>
void GranuleTree<T>::add(const SEAScope::RenderableId& rId, T object)
{
    _tree[rId.collection][rId.granule][rId.variable] = object;
}

template<typename T>
bool GranuleTree<T>::remove(const SEAScope::RenderableId& rId, T& obj)
{
    auto collectionIt = _tree.find(rId.collection);

    if (_tree.end() !=  collectionIt)
    {
        auto granuleIt = collectionIt->second.find(rId.granule);

        if (collectionIt->second.end() !=  granuleIt)
        {
            auto variableIt = granuleIt->second.find(rId.variable);

            if (granuleIt->second.end() !=  variableIt)
            {
                obj = static_cast<T>(variableIt->second);
                granuleIt->second.erase(variableIt);
                return true;
            }
        }
    }

    return false;
}

template<typename T>
T GranuleTree<T>::get(const SEAScope::RenderableId& rId)
{
    return _tree[rId.collection][rId.granule][rId.variable];
}

template<typename T>
void GranuleTree<T>::populate(const std::set<SEAScope::RenderableId>& rIds, std::list<SEAScope::RenderableId>& list)
{
    SEAScope::RenderableId rId;

    for (auto itCollection = _tree.begin(); itCollection != _tree.end(); ++itCollection)
    {
        rId.collection = itCollection->first;
        GranuleMap& granuleMap = itCollection->second;

        for (auto itGranule = granuleMap.begin(); itGranule != granuleMap.end(); ++itGranule)
        {
            rId.granule = itGranule->first;
            VariableMap& variableMap = itGranule->second;

            for (auto itVariable = variableMap.begin(); itVariable != variableMap.end(); ++itVariable)
            {
                rId.variable = itVariable->first;

                bool found = false;

                for (auto itRId : rIds)
                {
                    if (    itRId.collection == rId.collection
                        &&  itRId.granule == rId.granule
                        &&  itRId.variable == rId.variable)
                    {
                        found = true;
                        break;
                    }
                }

                if (false == found) // Not found
                {
                    // Add to the output list
                    list.push_back(rId);
                }
            }
        }
    }
}

template<typename T>
void GranuleTree<T>::populatePair(const std::set<SEAScope::RenderableId>& rIds, ListPairT& existing, ListPairT& notFound)
{
    SEAScope::RenderableId rId;

    for (auto itCollection = _tree.begin(); itCollection != _tree.end(); ++itCollection)
    {
        rId.collection = itCollection->first;
        GranuleMap& granuleMap = itCollection->second;

        for (auto itGranule = granuleMap.begin(); itGranule != granuleMap.end(); ++itGranule)
        {
            rId.granule = itGranule->first;
            VariableMap& variableMap = itGranule->second;

            for (auto itVariable = variableMap.begin(); itVariable != variableMap.end(); ++itVariable)
            {
                rId.variable = itVariable->first;

                bool found = false;

                for (auto itRId : rIds)
                {
                    if (    itRId.collection == rId.collection
                        &&  itRId.granule == rId.granule
                        &&  itRId.variable == rId.variable)
                    {
                        found = true;
                        break;
                    }
                }

                if (false == found) // Not found
                {
                    // Add to the output list
                    notFound.push_back(std::make_pair(rId, itVariable->second));
                }
                else
                {
                    existing.push_back(std::make_pair(rId, itVariable->second));
                }
            }
        }
    }
}

template<typename T>
void GranuleTree<T>::clear()
{
    for (auto itCollection = _tree.begin(); itCollection != _tree.end(); ++itCollection)
    {
        GranuleMap& granuleMap = itCollection->second;

        for (auto itGranule = granuleMap.begin(); itGranule != granuleMap.end(); ++itGranule)
        {
            VariableMap& variableMap = itGranule->second;

            for (auto itVariable = variableMap.begin(); itVariable != variableMap.end(); ++itVariable)
            {
                itVariable->second = nullptr;
            }
            variableMap.clear();
        }
        granuleMap.clear();
    }
    _tree.clear();
}

#endif /* GRANULE_TREE_H */
