// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef ISCENE_RENDERER_H
#define ISCENE_RENDERER_H

#include <string>
#include <vector>
#include <set>

#include "core/rendering_cfg.h"
#include "core/structs.h"
#include "3d/geodetic.h"
#include "3d/scene_graph_node.h"

// Forward declarations
class IScene;
class DataSet;
class Camera;
class SceneView;
class IPreRenderer;
class IPostRenderer;

namespace SEAScope
{
    struct RendererResources;
    class ColormapManager;
}

namespace OGL
{
    class Context;
    class ColormapManagerGPU;
}

class ISceneRenderer
{
    public:
        virtual ~ISceneRenderer();

        /*!
         * \brief Initialize the renderer.
         * \param[in] renderer Name of the renderer to load (OpenGL3, OpenGLES2). 
         * \param[in] colormapMngr Pointer to a ColormapManager object.
         */
        virtual bool initialize(const std::string& renderer,
                                SEAScope::ColormapManager* colormapMngr,
                                const SEAScope::RendererResources* resources
                                /*,float scaleFactor*/) = 0;

        /*!
         * \brief Change the rendering configuration of a granule
         */
        virtual void setDisplayCfg(IScene*, const SEAScope::RenderingCfg& cfg) = 0;

        virtual const std::set<SEAScope::RenderableId>& getToRender() const = 0;
        virtual void setToRender(const std::set<SEAScope::RenderableId>&) = 0;

        virtual bool updateStarted(IScene*) = 0;
        virtual bool updateFinished(IScene*) = 0;

        virtual OGL::Context* getContext() const = 0;

        virtual void resize(int, int) = 0;

        virtual int getMaxTextureSize() const = 0;

        virtual bool renderSceneView(SceneView*) = 0;

        virtual void setRasterizationMode(OGL::RasterizationMode) = 0;

        virtual OGL::ColormapManagerGPU* getColormapManager() = 0;

        // Draw all objects with their configuration
        virtual void draw(IScene*, const Camera*) = 0;
        virtual void drawRaw(IScene*, const Camera*) = 0;

        // Draw only one granule
        virtual void drawOneTrueColors(IScene*, const Camera*, const SEAScope::RenderableId&) = 0;
        virtual void drawOneRaw(IScene*, const Camera*, const SEAScope::RenderableId&) = 0;

        virtual void drawWorld(IScene*, const Camera*, const glm::mat4& mvp, FillType type) = 0;

        virtual void registerPreRenderer(IPreRenderer*) = 0;
        virtual void unregisterPreRenderer(IPreRenderer*) = 0;

        virtual void registerPostRenderer(IPostRenderer*) = 0;
        virtual void unregisterPostRenderer(IPostRenderer*) = 0;

};

#endif
