// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   math_func.h
 * Author: gleseach
 *
 * Created on 25 août 2015, 10:56
 */

#ifndef MATH_FUNC_H
#define	MATH_FUNC_H

#define TO_RADIANS(val) (val * M_PI / 180.0)
#define TO_DEGREES(val) (val * 180.0 / M_PI)

#include "num_utils.h"

inline bool isPowerOfTwo(unsigned int x)
{
  return ((x != 0) && ((x & (~x + 1)) == x));
}

template <typename T> int sign(T val)
{
    return (T(0) < val) - (val < T(0));
}

inline unsigned int nextPowerOfTwo(unsigned int n)
{
    double y = floor(log(n) / log(2));
    return static_cast<unsigned int>(pow(2, y + 1));
}

inline double AngleBetween(const glm::vec3& a, const glm::vec3& b)
{
    return static_cast<double>(acos(glm::dot(glm::normalize(a), glm::normalize(b))));
}

inline double AngleBetweenNormalized(const glm::vec3& a, const glm::vec3& b)
{
    return static_cast<double>(acos(glm::dot(a, b)));
}

inline float angleSigned2d(const glm::vec2& v1, const glm::vec2& v2)
{
    return atan2(v2.y, v2.x) - atan2(v1.y, v1.x);
}

inline glm::vec3 MultiplyComponents(const glm::vec3& v, const glm::vec3& scale)
{
    return glm::vec3(v.x * scale.x, v.y * scale.y, v.z * scale.z);
}

inline glm::vec3 RotateAroundAxis(const glm::vec3& vector, const glm::vec3& axis, double theta)
{
    double u = static_cast<double>(axis.x);
    double v = static_cast<double>(axis.y);
    double w = static_cast<double>(axis.z);

    double cosTheta = cos(theta);
    double sinTheta = sin(theta);

    double ms = static_cast<double>(glm::length2(axis));
    double m = sqrt(ms);

    return glm::vec3(
        ((u * (u * static_cast<double>(vector.x) + v * static_cast<double>(vector.y) + w * static_cast<double>(vector.z))) + 
        (((static_cast<double>(vector.x) * (v * v + w * w)) - (u * (v * static_cast<double>(vector.y) + w * static_cast<double>(vector.z)))) * cosTheta) + 
        (m * ((-w * static_cast<double>(vector.y)) + (v * static_cast<double>(vector.z))) * sinTheta)) / ms,

        ((v * (u * static_cast<double>(vector.x) + v * static_cast<double>(vector.y) + w * static_cast<double>(vector.z))) + 
        (((static_cast<double>(vector.y) * (u * u + w * w)) - (v * (u * static_cast<double>(vector.x) + w * static_cast<double>(vector.z)))) * cosTheta) + 
        (m * ((w * static_cast<double>(vector.x)) - (u * static_cast<double>(vector.z))) * sinTheta)) / ms,

        ((w * (u * static_cast<double>(vector.x) + v * static_cast<double>(vector.y) + w * static_cast<double>(vector.z))) + 
        (((static_cast<double>(vector.z) * (u * u + v * v)) - (w * (u * static_cast<double>(vector.x) + v * static_cast<double>(vector.y)))) * cosTheta) + 
        (m * (-(v * static_cast<double>(vector.x)) + (u * static_cast<double>(vector.y))) * sinTheta)) / ms);
}

inline glm::quat rotationBetweenVectors(glm::vec3 start, glm::vec3 dest)
{
    start = glm::normalize(start);
    dest = glm::normalize(dest);

    float cosTheta = glm::dot(start, dest);
    glm::vec3 rotationAxis;

    rotationAxis = glm::cross(start, dest);

    float s = sqrt( (1+cosTheta)*2 );
    float invs = 1 / s;

    return glm::quat(s * 0.5f, rotationAxis.x * invs, rotationAxis.y * invs, rotationAxis.z * invs);
}

inline void rotationBetweenVectors2(glm::vec3 start, glm::vec3 dest, glm::quat& xy, glm::quat& xz)
{
    glm::vec3 xyPlaneStart = glm::normalize(glm::vec3(start.x, start.y, 0.f));
    glm::vec3 xzPlaneStart = glm::normalize(glm::vec3(start.x, 0.f, start.z));

    glm::vec3 xyPlaneDest = glm::normalize(glm::vec3(dest.x, dest.y, 0.f));
    glm::vec3 xzPlaneDest = glm::normalize(glm::vec3(dest.x, 0.f, dest.z));

    xy = rotationBetweenVectors(xyPlaneStart, xyPlaneDest);
    xz = rotationBetweenVectors(xzPlaneStart, xzPlaneDest);
}

inline double CubicInterpolate(double y0,double y1, double y2,double y3, double mu)
{
   double a0,a1,a2,a3,mu2;

   mu2 = mu*mu;
   a0 = y3 - y2 - y0 + y1;
   a1 = y0 - y1 - a0;
   a2 = y2 - y0;
   a3 = y1;

   return(a0*mu*mu2+a1*mu2+a2*mu+a3);
}

inline double HermiteInterpolate(
   double y0,double y1,
   double y2,double y3,
   double mu,
   double tension,
   double bias)
{
   double m0,m1,mu2,mu3;
   double a0,a1,a2,a3;

	mu2 = mu * mu;
	mu3 = mu2 * mu;
    m0  = (y1-y0)*(1+bias)*(1-tension)/2;
    m0 += (y2-y1)*(1-bias)*(1-tension)/2;
    m1  = (y2-y1)*(1+bias)*(1-tension)/2;
    m1 += (y3-y2)*(1-bias)*(1-tension)/2;
    a0 =  2*mu3 - 3*mu2 + 1;
    a1 =    mu3 - 2*mu2 + mu;
    a2 =    mu3 -   mu2;
    a3 = -2*mu3 + 3*mu2;

    return(a0*y1+a1*m0+a2*m1+a3*y2);
}

inline double CatmullRomInterpolate(
   double y0,double y1,
   double y2,double y3,
   double mu)
{
   double a0,a1,a2,a3,mu2;

   mu2 = mu*mu;
    a0 = -0.5*y0 + 1.5*y1 - 1.5*y2 + 0.5*y3;
   a1 = y0 - 2.5*y1 + 2*y2 - 0.5*y3;
   a2 = -0.5*y0 + 0.5*y2;
   a3 = y1;

   return(a0*mu*mu2+a1*mu2+a2*mu+a3);
}

// Intersection test between a ray and a sphere (with all informations at intersection point)
inline bool intersect(const glm::vec3& raydir, const glm::vec3& rayorig,const glm::vec3& pos,
    const float& rad, glm::vec3& hitpoint,float& distance, glm::vec3& normal)
{
    float a = glm::dot(raydir, raydir);
    float b = glm::dot(raydir , (2.0f * ( rayorig - pos)));
    float c = glm::dot(pos,pos) + glm::dot(rayorig,rayorig) -2.0f*glm::dot(rayorig,pos) - rad*rad;
    float D = b*b + (-4.0f)*a*c;

    // If ray can not intersect then stop
    if (D < 0)
    {
        return false;
    } 

    D = sqrt(D);

    // Ray can intersect the sphere, solve the closer hitpoint
    float t = (-0.5f) * (b+D) / a;

    if (t > 0.0f)
    {
        distance=sqrt(a)*t;
        hitpoint=rayorig + t*raydir;
        normal=(hitpoint - pos) / rad;
    }
    else
    {
        return false;
    }

    return true;
}

// Intersection test between a ray and a sphere (no distance, no normal at intersection)
inline bool intersect(const glm::vec3& raydir, const glm::vec3& rayorig,const glm::vec3& pos,
    const float& rad, glm::vec3& hitpoint)
{
    float a = glm::dot(raydir, raydir);
    float b = glm::dot(raydir , (2.0f * ( rayorig - pos)));
    float c = glm::dot(pos,pos) + glm::dot(rayorig,rayorig) -2.0f*glm::dot(rayorig,pos) - rad*rad;
    float D = b*b + (-4.0f)*a*c;

    // If ray can not intersect then stop
    if (D < 0)
    {
        return false;
    } 

    D = sqrt(D);

    // Ray can intersect the sphere, solve the closer hitpoint
    float t = (-0.5f) * (b+D) / a;

    if (t > 0.0f)
    {
        hitpoint = rayorig + t*raydir;
    }
    else
    {
        return false;
    }

    return true;
}

inline bool intersectUnitarySphereWorldCentered(const glm::vec3& raydir, const glm::vec3& rayorig, glm::vec3& hitpoint)
{
    return intersect(raydir, rayorig, glm::vec3(0.f,0.f,0.f), 1.f, hitpoint);
}

/* \brief Launch a ray from the screen position (x, y) and return the result of the intersection with the globe.
 */
inline bool getPickingPosition(glm::ivec2 viewportSize, int x, int y, glm::vec3& hitPoint, const glm::mat4& modelView, const glm::mat4& proj)
{
    glm::vec4 viewport = glm::vec4(0, 0, viewportSize.x, viewportSize.y);
    glm::vec3 wincoord = glm::vec3(x, viewportSize.y - y, -1.f);
    glm::vec3 wincoord2 = glm::vec3(x, viewportSize.y - y, 1.f);

    glm::vec3 objcoord = glm::unProject(wincoord, modelView, proj, viewport);
    glm::vec3 objcoord2 = glm::unProject(wincoord2, modelView, proj, viewport);
    glm::vec3 objcoord3 = glm::normalize(objcoord2 - objcoord);

    float distance;
    glm::vec3 normal;

    return intersect(objcoord3, objcoord, glm::vec3(0.f), 1.f, hitPoint, distance, normal);
}

/* \brief Launch a ray from the screen position (x, y) and return the result of the intersection with the globe.
 */
inline bool getPickingPosition(glm::ivec2 viewportSize, int x, int y, glm::vec3& hitPoint, const glm::quat& q, const glm::mat4& view, const glm::mat4& proj)
{
    return getPickingPosition(viewportSize, x, y, hitPoint, view * glm::mat4_cast(q), proj);
}

inline bool isInFrustum(const glm::vec4& point, const glm::mat4& projView)
{
    // World space to clip space
    glm::vec4 p = projView * point;

    // Coordinates components have to be between -w and w.
    return (p.x > -p.w && p.x < p.w &&
            p.y > -p.w && p.y < p.w &&
            p.z > -p.w && p.z < p.w);
}

inline bool isInFrustum(const glm::vec3& point, const glm::mat4& projView)
{
    return isInFrustum(glm::vec4(point, 1.f), projView);
}

typedef int OutCode;

const int INSIDE = 0; // 0000
const int LEFT = 1;   // 0001
const int RIGHT = 2;  // 0010
const int BOTTOM = 4; // 0100
const int TOP = 8;    // 1000

// Compute the bit code for a point (x, y) using the clip rectangle
// bounded diagonally by (xmin, ymin), and (xmax, ymax)

// ASSUME THAT xmax, xmin, ymax and ymin are global constants.

inline OutCode ComputeOutCode(const glm::vec2& p, const glm::vec4& bbox)
{
	OutCode code;

	code = INSIDE;          // initialised as being inside of clip window

	if (p.x < bbox.x)           // to the left of clip window
		code |= LEFT;
	else if (p.x > bbox.z)      // to the right of clip window
		code |= RIGHT;
	if (p.y < bbox.y)           // below the clip window
		code |= BOTTOM;
	else if (p.y > bbox.w)      // above the clip window
		code |= TOP;

	return code;
}

// Cohen–Sutherland clipping algorithm clips a line from
// P0 = (x0, y0) to P1 = (x1, y1) against a rectangle with 
// diagonal from (xmin, ymin) to (xmax, ymax).
inline bool isSegmentIntersectRectangle(const glm::vec2& p0, const glm::vec2& p1, const glm::vec4& rect)
{
	// compute out codes for P0, P1, and whatever point lies outside the clip rectangle
	OutCode outcode0 = ComputeOutCode(p0, rect);
	OutCode outcode1 = ComputeOutCode(p1, rect);
	bool accept = false;

    float x0 = p0.x;
    float y0 = p0.y;

    float x1 = p1.x;
    float y1 = p1.y;

    float x(0.f), y(0.f);

	while (true)
    {
        // Bitwise OR is 0. Trivially accept and get out of loop
		if (!(outcode0 | outcode1))
        {
			accept = true;
			break;
		}
        else if (outcode0 & outcode1) // Bitwise AND is not 0. Trivially reject and get out of loop
        {
			break;
		}
        else
        {
			// failed both tests, so calculate the line segment to clip
			// from an outside point to an intersection with clip edge

			// At least one endpoint is outside the clip rectangle; pick it.
			OutCode outcodeOut = outcode0 ? outcode0 : outcode1;

			// Now find the intersection point;
			// use formulas y = y0 + slope * (x - x0), x = x0 + (1 / slope) * (y - y0)
			if (outcodeOut & TOP) {           // point is above the clip rectangle
				x = x0 + (x1 - x0) * (rect.w - y0) / (y1 - y0);
				y = rect.w;
			} else if (outcodeOut & BOTTOM) { // point is below the clip rectangle
				x = x0 + (x1 - x0) * (rect.y - y0) / (y1 - y0);
				y = rect.y;
			} else if (outcodeOut & RIGHT) {  // point is to the right of clip rectangle
				y = y0 + (y1 - y0) * (rect.z - x0) / (x1 - x0);
				x = rect.z;
			} else if (outcodeOut & LEFT) {   // point is to the left of clip rectangle
				y = y0 + (y1 - y0) * (rect.x - x0) / (x1 - x0);
				x = rect.x;
			}

			// Now we move outside point to intersection point to clip
			// and get ready for next pass.
			if (outcodeOut == outcode0)
            {
				x0 = x;
				y0 = y;
				outcode0 = ComputeOutCode(glm::vec2(x0, y0), rect);
			}
            else
            {
				x1 = x;
				y1 = y;
				outcode1 = ComputeOutCode(glm::vec2(x1, y1), rect);
			}
		}
	}

    return accept;
}

// a1 is line1 start, a2 is line1 end, b1 is line2 start, b2 is line2 end
inline bool lineIntersect(const glm::vec2& a1, const glm::vec2& a2, const glm::vec2& b1, const glm::vec2& b2)
{
    glm::vec2 b = a2 - a1;
    glm::vec2 d = b2 - b1;

    float bDotDPerp = b.x * d.y - b.y * d.x;

    // if b dot d == 0, it means the lines are parallel so have infinite intersection points
    if (bDotDPerp == 0.f)
        return false;

    bDotDPerp = 1.f / bDotDPerp;

    glm::vec2 c = b1 - a1;

    float t = (c.x * d.y - c.y * d.x) * bDotDPerp;

    if (t < 0.f || t > 1.f)
        return false;

    t = (c.x * b.y - c.y * b.x) * bDotDPerp;

    if (t < 0.f || t > 1.f)
        return false;

    return true;
}

#define	DONT_INTERSECT    0
#define	DO_INTERSECT      1
#define COLLINEAR         2

/**************************************************************
 *                                                            *
 *    NOTE:  The following macro to determine if two numbers  *
 *    have the same sign, is for 2's complement number        *
 *    representation.  It will need to be modified for other  *
 *    number systems.                                         *
 *                                                            *
 **************************************************************/

#define SAME_SIGNS( a, b )	\
        (a*b >= 0.f)
//		(((long) ((unsigned long) a ^ (unsigned long) b)) >= 0 )

inline bool lines_intersect(const glm::vec2& p0, const glm::vec2& p1, const glm::vec2& p2, const glm::vec2& p3)
{
    float x1, y1, x2, y2, x3, y3, x4, y4;
    float a1, a2, b1, b2, c1, c2; /* Coefficients of line eqns. */
    float r1, r2, r3, r4;         /* 'Sign' values */

    x1 = p0.x;
    y1 = p0.y;
    x2 = p1.x;
    y2 = p1.y;
    x3 = p2.x;
    y3 = p2.y;
    x4 = p3.x;
    y4 = p3.y;

    /* Compute a1, b1, c1, where line joining points 1 and 2
     * is "a1 x  +  b1 y  +  c1  =  0".
     */

    a1 = y2 - y1;
    b1 = x1 - x2;
    c1 = x2 * y1 - x1 * y2;

    /* Compute r3 and r4.
     */

    r3 = a1 * x3 + b1 * y3 + c1;
    r4 = a1 * x4 + b1 * y4 + c1;

    /* Check signs of r3 and r4.  If both point 3 and point 4 lie on
     * same side of line 1, the line segments do not intersect.
     */

    if ( false == SEAScope::floating_point_equality(r3, 0.0f) &&
         false == SEAScope::floating_point_equality(r4, 0.0f) &&
         SAME_SIGNS( r3, r4 ))
    {
        return ( false );
    }

    /* Compute a2, b2, c2 */

    a2 = y4 - y3;
    b2 = x3 - x4;
    c2 = x4 * y3 - x3 * y4;

    /* Compute r1 and r2 */

    r1 = a2 * x1 + b2 * y1 + c2;
    r2 = a2 * x2 + b2 * y2 + c2;

    /* Check signs of r1 and r2.  If both point 1 and point 2 lie
     * on same side of second line segment, the line segments do
     * not intersect.
     */

    if ( false == SEAScope::floating_point_equality(r1, 0.0f) &&
         false == SEAScope::floating_point_equality(r2, 0.0f) &&
         SAME_SIGNS( r1, r2 ))
    {
        return ( false );
    }

    return ( true );
} /* lines_intersect */

inline void bresenham(glm::ivec2 p0, glm::ivec2 p1, std::vector<glm::ivec2>& pixels)
{
    // Bresenham's line algorithm
    const bool steep = (abs(p1.y - p0.y) > abs(p1.x - p0.x));

    if(steep)
    {
        std::swap(p0.x, p0.y);
        std::swap(p1.x, p1.y);
    }

    if(p0.x > p1.x)
    {
        std::swap(p0.x, p1.x);
        std::swap(p0.y, p1.y);
    }

    const int dx = p1.x - p0.x;
    const int dy = abs(p1.y - p0.y);

    float error = static_cast<float>(dx) / 2.0f;
    const int ystep = (p0.y < p1.y) ? 1 : -1;
    int y = p0.y;

    const int maxX = p1.x;

    for(int x=p0.x; x < maxX; ++x)
    {
        if(steep)
        {
            pixels.push_back(glm::ivec2(y, x));
        }
        else
        {
            pixels.push_back(glm::ivec2(x, y));
        }

        error -= static_cast<float>(dy);

        if(error < 0)
        {
            y += ystep;
            error += static_cast<float>(dx);
        }
    }
}

float shortestSphericalDistance(float a, float b, bool& change);
bool containNorthPole(const glm::vec2 corners[4]);
int containPole(const glm::vec2 corners[5]);



#endif	/* MATH_FUNC_H */
