// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   meta.h
 * Author: gleseach
 *
 * Created on 15 juin 2016, 12:32
 */

#ifndef META_H
#define META_H

#include "3d/opengl_enum.h"
#include "3d/array.h"
#include "3d/image_wrapper.h"
#include "3d/buffer_desc.h"

// Meta prog functions
template<typename T>  inline DataType getDataType()                  {return DataType::none;}
template<>            inline DataType getDataType<unsigned char>()   {return DataType::unsigned_byte;}
template<>            inline DataType getDataType<unsigned int>()    {return DataType::unsigned_int;}
template<>            inline DataType getDataType<unsigned short>()  {return DataType::unsigned_short;}
template<>            inline DataType getDataType<float>()           {return DataType::float32;}

// Array
template<typename T> inline Array makeArray(const std::vector<T>& v) {return Array(BufferDesc(static_cast<unsigned int>(v.size()), 1, getDataType<T>()), reinterpret_cast<void*>(const_cast<T*>(v.data())));}

template<>inline Array makeArray<glm::vec2>(const std::vector<glm::vec2>& v) {return Array(BufferDesc(static_cast<unsigned int>(v.size()), 2, DataType::float32), reinterpret_cast<void*>(const_cast<glm::vec2*>(v.data())));}
template<>inline Array makeArray<glm::ivec2>(const std::vector<glm::ivec2>& v) {return Array(BufferDesc(static_cast<unsigned int>(v.size()), 2, DataType::unsigned_int), reinterpret_cast<void*>(const_cast<glm::ivec2*>(v.data())));}

template<>inline Array makeArray<glm::vec3>(const std::vector<glm::vec3>& v) {return Array(BufferDesc(static_cast<unsigned int>(v.size()), 3, DataType::float32), reinterpret_cast<void*>(const_cast<glm::vec3*>(v.data())));}
template<>inline Array makeArray<glm::ivec3>(const std::vector<glm::ivec3>& v) {return Array(BufferDesc(static_cast<unsigned int>(v.size()), 3, DataType::unsigned_int), reinterpret_cast<void*>(const_cast<glm::ivec3*>(v.data())));}

template<>inline Array makeArray<glm::vec4>(const std::vector<glm::vec4>& v) {return Array(BufferDesc(static_cast<unsigned int>(v.size()), 4, DataType::float32), reinterpret_cast<void*>(const_cast<glm::vec4*>(v.data())));}
template<>inline Array makeArray<glm::ivec4>(const std::vector<glm::ivec4>& v) {return Array(BufferDesc(static_cast<unsigned int>(v.size()), 4, DataType::unsigned_int), reinterpret_cast<void*>(const_cast<glm::ivec4*>(v.data())));}

// Description
template<typename T> inline BufferDesc makeDesc(const std::vector<T>& v) {return BufferDesc(static_cast<unsigned int>(v.size()), 1, getDataType<T>());}

template<>inline BufferDesc makeDesc<glm::vec2>(const std::vector<glm::vec2>& v) {return BufferDesc(static_cast<unsigned int>(v.size()),2, DataType::float32);}
template<>inline BufferDesc makeDesc<glm::ivec2>(const std::vector<glm::ivec2>& v) {return BufferDesc(static_cast<unsigned int>(v.size()),2, DataType::unsigned_int);}

template<>inline BufferDesc makeDesc<glm::vec3>(const std::vector<glm::vec3>& v) {return BufferDesc(static_cast<unsigned int>(v.size()),3, DataType::float32);}
template<>inline BufferDesc makeDesc<glm::ivec3>(const std::vector<glm::ivec3>& v) {return BufferDesc(static_cast<unsigned int>(v.size()),3, DataType::unsigned_int);}

template<>inline BufferDesc makeDesc<glm::vec4>(const std::vector<glm::vec4>& v) {return BufferDesc(static_cast<unsigned int>(v.size()),4, DataType::float32);}
template<>inline BufferDesc makeDesc<glm::ivec4>(const std::vector<glm::ivec4>& v) {return BufferDesc(static_cast<unsigned int>(v.size()),4, DataType::unsigned_int);}


// ImageWrapper
template<typename T> inline ImageWrapper* makeImageWrapper(const glm::ivec2& size, const std::vector<T>& v) {return new ImageWrapper(size, 1, sizeof(T), getDataType<T>(), reinterpret_cast<void*>(const_cast<T*>(v.data())));}

template<>inline ImageWrapper* makeImageWrapper<glm::vec3>(const glm::ivec2& size, const std::vector<glm::vec3>& v) {return new ImageWrapper(size, 3, sizeof(float), DataType::float32, reinterpret_cast<void*>(const_cast<glm::vec3*>(v.data())));}
template<>inline ImageWrapper* makeImageWrapper<glm::ivec3>(const glm::ivec2& size, const std::vector<glm::ivec3>& v) {return new ImageWrapper(size, 3, sizeof(int), DataType::unsigned_int, reinterpret_cast<void*>(const_cast<glm::ivec3*>(v.data())));}

template<>inline ImageWrapper* makeImageWrapper<glm::vec4>(const glm::ivec2& size, const std::vector<glm::vec4>& v) {return new ImageWrapper(size, 4, sizeof(float), DataType::float32, reinterpret_cast<void*>(const_cast<glm::vec4*>(v.data())));}
template<>inline ImageWrapper* makeImageWrapper<glm::ivec4>(const glm::ivec2& size, const std::vector<glm::ivec4>& v) {return new ImageWrapper(size, 4, sizeof(int), DataType::unsigned_int, reinterpret_cast<void*>(const_cast<glm::ivec4*>(v.data())));}


#endif /* META_H */

