// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   Program.h
 * Author: gleseach
 *
 * Created on 6 août 2015, 15:23
 */

#ifndef PROGRAM_H
#define	PROGRAM_H

#include <iostream>

#include "3d/bindable_object.h"
#include "3d/opengl_enum.h"

namespace OGL
{

class Shader;

class Program : public BindableObject
{
public:
    Program();
    virtual ~Program() override;

    virtual void attachShaders(const std::vector<Shader*>& shaders) = 0;
    virtual void attachShaders(const Shader* shaders, unsigned int size) = 0;
    virtual bool link() = 0;

    virtual void bindAttribLocation(unsigned int pos, const char *name) = 0;
    virtual int getUniformLocation(const char *name) = 0;
    virtual int getAttribLocation(const char *name) = 0;

    virtual void setTransformFeedbackVaryings(const std::vector<std::string>& varyings, OGL::TransformFeedbackBufferMode bufferMode) = 0;

    //! Return the number of uniform used by the program.
    virtual int getUniformCount() const = 0;

    //! Return the name of the uniform with the id passed in parameter.
    virtual std::string getUniformName(int id) const  = 0;

    //! Return the type of the uniform with the id passed in parameter.
    virtual OGL::UniformType getUniformType(int id) const = 0;

    //! Set a uniform by his name.
    template<typename T> void setUniformByName(const char* /*name*/, const T& /*value*/) {std::cout << "not implemented";}

    //! Set a uniform by his shader id.
    template<typename T> void setUniform(int /*location*/, const T& /*value*/) {std::cout << "not implemented";}

    virtual void uniform1i(int location, const int& value) = 0;
    virtual void uniform1u(int location, const std::uint32_t& value) = 0;
    virtual void uniform1f(int location, const float& value) = 0;
    virtual void uniform2fv(int location, const glm::vec2& value) = 0;
    virtual void uniform3fv(int location, const glm::vec3& value) = 0;
    virtual void uniform4fv(int location, const glm::vec4& value) = 0;
    virtual void uniform2bv(int location, const glm::bvec2& value) = 0;
    virtual void uniformMatrix4fv(int location, const glm::mat4& value) = 0;

protected:
    std::vector<Shader*> shaderList;
};

// Name family
template<> inline void Program::setUniformByName<int>(const char* name, const int& value)             {uniform1i(getUniformLocation(name), value);}
template<> inline void Program::setUniformByName<std::uint32_t>(const char* name, const std::uint32_t& value) {uniform1u(getUniformLocation(name), value);}
template<> inline void Program::setUniformByName<float>(const char* name, const float& value)         {uniform1f(getUniformLocation(name), value);}
template<> inline void Program::setUniformByName<glm::vec2>(const char* name, const glm::vec2& value) {uniform2fv(getUniformLocation(name), value);}
template<> inline void Program::setUniformByName<glm::vec3>(const char* name, const glm::vec3& value) {uniform3fv(getUniformLocation(name), value);}
template<> inline void Program::setUniformByName<glm::vec4>(const char* name, const glm::vec4& value) {uniform4fv(getUniformLocation(name), value);}
template<> inline void Program::setUniformByName<glm::mat4>(const char* name, const glm::mat4& value) {uniformMatrix4fv(getUniformLocation(name), value);}

// Location family
template<> inline void Program::setUniform<int>(int location, const int& value)             {uniform1i(location, value);}
template<> inline void Program::setUniform<std::uint32_t>(int location, const std::uint32_t& value) {uniform1u(location, value);}
template<> inline void Program::setUniform<float>(int location, const float& value)         {uniform1f(location, value);}
template<> inline void Program::setUniform<glm::vec2>(int location, const glm::vec2& value) {uniform2fv(location, value);}
template<> inline void Program::setUniform<glm::vec3>(int location, const glm::vec3& value) {uniform3fv(location, value);}
template<> inline void Program::setUniform<glm::vec4>(int location, const glm::vec4& value) {uniform4fv(location, value);}
template<> inline void Program::setUniform<glm::mat4>(int location, const glm::mat4& value) {uniformMatrix4fv(location, value);}

// Boolean version
template<> inline void Program::setUniform<glm::bvec2>(int location, const glm::bvec2& value) {uniform2bv(location, value);}

} // namespace OGL

#endif	/* PROGRAM_H */

