// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   scene_graph.h
 * Author: gleseach
 *
 * Created on 2 novembre 2016, 11:58
 */

#ifndef SCENE_GRAPH_H
#define SCENE_GRAPH_H

#include "core/structs.h"
#include "3d/granule_tree.h"
#include "3d/scene_graph_node.h"

class Renderable2;
class Pickable;

namespace SEAScope
{
    struct RenderableId;
}

namespace OGL
{
    class Context;
}

class Node;
class IRenderer2;
class IPickingRenderer;
class IRawRenderer;
struct DataRenderingParam;
class IWorldVisibility;

class SceneGraph
{
public:
    SceneGraph();
    ~SceneGraph();

    void create(OGL::Context* ctx, IWorldVisibility* visibility/*, RendererFactory* rendererFactory*/);

    void addObject(Renderable2* object);
    void removeObject(Renderable2* object);

    Pickable* pickable(unsigned int id);

    //! Draw all objects in the scene graph in "true colors" mode
    void draw(OGL::Context* ctx, const glm::mat4& mvp);

    //! Draw all objects in the scene graph in picking mode
    void drawPicking(OGL::Context* ctx, const glm::mat4& mvp);

    void drawRaw(OGL::Context* ctx, const glm::mat4& mvp);

    //! Draw only the node with the required RenderableId.
    void drawOne(OGL::Context* ctx, const SEAScope::RenderableId& rId, const glm::mat4& mvp);
    void drawOneRaw(OGL::Context* ctx, const SEAScope::RenderableId& rId, const glm::mat4& mvp);

    void setOnTop(Renderable2* obj);

    Node* getNode(const SEAScope::RenderableId& rId);
    void getNodes(RendererType type, std::vector<Node*>& array);

    void getNodesRecurse(RendererType type, std::vector<Node*>& array, Node* node);
    void addTo(Node* parent, Node* node, bool updateRenderer = true);

    void addTo(Node* parent, Node* node, const SEAScope::RenderableId& rId);
    void addTo(Node* parent, Node* node, const SEAScope::RenderableId& rId, bool ownRenderer);
    bool remove(const SEAScope::RenderableId& rId, Node*& node);
    void setVisible(const SEAScope::RenderableId& rId, bool status);
    void updateRenderer(Node* node);
    void zChanged(Node* node);
    Node* getRoot();

private:
    std::list<Renderable2*> _objects;

    std::vector<std::vector<IRenderer2*>> _renderers;
    std::vector<IPickingRenderer*> _pickingRenderers;
    std::vector<IRawRenderer*> _rawRenderers;
    std::vector<unsigned int> _rendererToPicking;
    std::vector<unsigned int> _rendererToRaw;

    GranuleTree<Node*> _nodes;
    Node _root;
};

#endif /* SCENE_GRAPH_H */
