// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   node_factory.h
 * Author: gleseach
 *
 * Created on 1 décembre 2016, 15:18
 */

#ifndef SCENEGRAPH_CONTROLLER_H
#define SCENEGRAPH_CONTROLLER_H

#include "core/types.h"
#include "core/structs.h"

#include "3d/geodetic.h"
#include "3d/scene_graph_node.h"
#include "3d/scene_dictionnary.h"
#include "3d/data_rendering_param.h"
#include "core/interfaces/iapp_events_subscriber.h"

class SceneGraph;
class DataSet;
class Pickable;
struct DataRenderingParam;
class PickingManager;
class Globe;
class Polyline;
class ShapeEditor;
class BillboardPool;
class Icons;;

namespace OGL
{
    class ColormapManagerGPU;
    class Context;
}

namespace SEAScope
{
    struct TrajectoryMarkers;
    class AppEvent;
    class RenderingCfg;
    class AppEventsDispatcher;
}

class SceneGraphController: public virtual SEAScope::IAppEventsSubscriber
{
public:
    SceneGraphController(SEAScope::AppEventsDispatcher&);
    virtual ~SceneGraphController() override;

    bool handleAppEvent(std::type_index, const SEAScope::AppEvent&) override;

    void setSceneGraph(OGL::Context*, SceneGraph* sg);
    void setColormapManager(OGL::ColormapManagerGPU* cm);

    /*
     * return true if the granule doesn't exist and is added in the scene, false 
     *        if it is already in the scene 
     */
    bool addGranule(OGL::Context* ctx, DataSet* dataSet, const SEAScope::RenderingCfg& rCfg);
    void addGlobe(Globe* globe);
    bool addTrajectory( OGL::Context* ctx,
                        std::vector<float>& relativeTime,
                        std::vector<Geodetic2D>& points,
                        std::vector<float>& fData,
                        std::vector<bool>& mask,
                        const SEAScope::RenderingCfg& rCfg,
                        const SEAScope::RenderableId& rId);

    bool addShape(OGL::Context* ctx, Pickable* shape, const SEAScope::RenderingCfg&);
    bool setTrajectoryMarkers(OGL::Context* ctx, SEAScope::TrajectoryMarkers&&);

    void updateDisplayParameters(OGL::Context* ctx, const SEAScope::RenderingCfg& renderingCfg);

    /*!
     * \brief Test if an object with this id exist in the scene. If the object is a dataset, check the resolution.
     * \param[in] rId The id of the object to find.
     * \param[in] resolution The resolution of the dataSet to find.
     * \return true if the object with the couple id | resolution exist.
     */
    bool exist(const SEAScope::RenderableId& rId, float resolution);

    /*!
     * \brief Test if an object with this id exist in the scene
     */
    bool exist(const SEAScope::RenderableId& rId);

    void setVisibleData(const std::set<SEAScope::RenderableId>& granuleIds);

    SceneGraph* getSceneGraph() const;
    const SceneDictionnary* getSceneDictionnary() const;

    void drawWorld(OGL::Context* ctx, const glm::mat4& mvp, FillType type);

    void remove(const SEAScope::RenderableId& rId);
	void setVisible(const SEAScope::RenderableId& rId, bool status);

    void setIcons(Icons* icons);
private:
    DataRenderingParam* createRenderingParam(OGL::Context* ctx, const SEAScope::RenderingCfg& renderingCfg);
    void updateRenderingCfg(OGL::Context* ctx, DataRenderingParam* param, const SEAScope::RenderingCfg& renderingCfg);
    FillType fillTypeFromRenderingCfg(DataRenderingParam* param);
    void addToSceneGraph(OGL::Context* ctx, const SEAScope::RenderingCfg& rCfg, RendererType rendererType, Renderable2* renderable, bool ownRenderer);
    Node* getCollectionNode(OGL::Context* ctx, const SEAScope::CollectionId& colId, RendererType rendererType);

private:
    SceneGraph* _sceneGraph;
    SceneDictionnary _dictionnary;
    OGL::ColormapManagerGPU* _colormapMngr;
    GranuleNode* _world;
    std::list<DataSet*> _cameraMovedDataSets;
    std::map<SEAScope::CollectionId, BillboardPool*> _collectionBillboardPools;
    Icons* _icons;

protected:
    SEAScope::AppEventsDispatcher& _appEventsDispatcher;
};

#endif /* SCENEGRAPH_CONTROLLER_H */

