// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   scene_graph_node.h
 * Author: gleseach
 *
 * Created on 2 décembre 2016, 11:08
 */

#ifndef SCENE_GRAPH_NODE_H
#define SCENE_GRAPH_NODE_H

#include <list>
#include <string>

#include "3d/data_rendering_param.h"
#include "3d/app_enum.h"

enum RendererType
{
    renderer_raster = 0,
    renderer_barb,
    renderer_arrow,
    renderer_uv_streamline,
    renderer_dot_cloud,
    renderer_trajectory,
    renderer_polyline,
    renderer_polygon,
    renderer_globe,
    renderer_rawrgb,
    renderer_quadscreen,
    renderer_quadscreen_textured,
    renderer_quadscreen_textured_instanced,
    renderer_null,
    renderer_count
};

enum PickingRendererType
{
    picking_renderer_polygon = 0,
    picking_renderer_trajectory,
    picking_renderer_polyline,
    picking_renderer_raster,
    picking_renderer_globe,
    picking_renderer_screen_quad,
    picking_renderer_null,
    picking_renderer_count
};

enum RawRendererType
{
    raw_renderer_raster = 0,
    raw_renderer_trajectory,
    raw_renderer_rgb,
    raw_renderer_globe,
    raw_renderer_null,
    raw_renderer_count
};

enum FillType
{
    fill_false_colors = 0,
    fill_raw,
    fill_color,
    fill_count,
    fill_null
};

class Renderable2;
class IRenderer2;
class IPickingRenderer;
class IRawRenderer;

namespace OGL
{
    class Context;
}

struct RenderingParameters
{
    DataRenderingParam* renderingParam;
    RendererType rendererType;
    FillType fillType;

    RenderingParameters()
     :  renderingParam(nullptr),
        rendererType(RendererType::renderer_null),
        fillType(FillType::fill_null)
    {
    }

    ~RenderingParameters()
    {
        delete renderingParam;
        renderingParam = nullptr;
    }
};

class Node
{
protected:
    Node* _parent;
    std::list<Node*>* _children;
    RenderingParameters* _rCfg;
    IRenderer2* _renderer;
    IPickingRenderer* _rendererPicking;
    IRawRenderer* _rendererRaw;
    bool _visible;
    Node* _topChild;

public:
    Node();
    virtual ~Node();

    Node(const Node&) = delete;
    Node& operator=(const Node&) = delete;

    virtual void draw(OGL::Context* ctx, const glm::mat4& mvp);
    virtual void drawPicking(OGL::Context* ctx, const glm::mat4& mvp);
    virtual void drawRaw(OGL::Context* ctx, const glm::mat4& mvp);

    void addChild(Node* node);
    bool removeChild(Node* node);
    std::list<Node*>* children() const;
    void addTopChild(Node* node);
    Node* getTopChild();
    void setRenderingParam(DataRenderingParam* param);
    void setRendererType(RendererType type);
    void setFillType(FillType type);
    void setParent(Node* parent);
    void setRenderer(IRenderer2* renderer);
    void setPickingRenderer(IPickingRenderer* renderer);
    void setRawRenderer(IRawRenderer* renderer);

    RendererType rendererType() const;
    FillType fillType() const;
    DataRenderingParam* dataRenderingParam() const;
    Node* parent() const;
    IRenderer2* renderer() const;
    IPickingRenderer* pickingRenderer() const;
    IRawRenderer* rawRenderer() const;
    void setVisible(bool status);
    bool visible() const;
};

class GranuleNode : public Node
{
public:
    GranuleNode();
    ~GranuleNode() override;

    GranuleNode(const GranuleNode&) = delete;
    GranuleNode& operator=(const GranuleNode&) = delete;

    void draw(OGL::Context* ctx, const glm::mat4& mvp) override;
    void drawPicking(OGL::Context* ctx, const glm::mat4& mvp) override;
    void drawRaw(OGL::Context* ctx, const glm::mat4& mvp) override;

    void setRenderable(Renderable2* obj);
    Renderable2* renderable() const;

private:
    Renderable2* _renderable;
};

#endif /* SCENE_GRAPH_NODE_H */
