// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef SHAPES_ALLOCATOR_H
#define SHAPES_ALLOCATOR_H

#include <mutex>
#include <vector>
#include <algorithm>

#include "3d/shape.h"
#include "3d/shape_mesh.h"
#include "3d/singleton.h"

class ShapesAllocator final: public Singleton<ShapesAllocator>
{
    public:
        ShapesAllocator();
        ~ShapesAllocator() override;

        template<typename T>
        static T* getShape();

        template<typename T>
        static T* getMesh();

        static bool release(OGL::Shape*);
        static bool release(ShapeMesh*);

    protected:
        template<typename T>
        void createShape() {
            std::lock_guard<std::mutex> scopeLock { this->_shapesMutex };
            this->_shapes.emplace_back(new T);
        }

        bool deleteShape(OGL::Shape*);

        template<typename T>
        void createMesh() {
            std::lock_guard<std::mutex> scopeLock { this->_meshesMutex };
            this->_meshes.emplace_back(new T);
        }

        bool deleteMesh(ShapeMesh*);

        std::mutex _shapesMutex;
        std::vector<OGL::Shape*> _shapes;

        std::mutex _meshesMutex;
        std::vector<ShapeMesh*> _meshes;
};

template<typename T>
T* ShapesAllocator::getShape()
{
    ShapesAllocator::getInstance()->createShape<T>();
    return static_cast<T*>(ShapesAllocator::getInstance()->_shapes.back());
}

template<typename T>
T* ShapesAllocator::getMesh()
{
    ShapesAllocator::getInstance()->createMesh<T>();
    return static_cast<T*>(ShapesAllocator::getInstance()->_meshes.back());
}
#endif
