// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   sphere.h
 * Author: gleseach
 *
 * Created on 7 août 2015, 10:02
 */

#ifndef SPHERE_H
#define	SPHERE_H

// icosahedron

static glm::vec3 icosahedronVertices[] =
{
    glm::vec3(0,            -0.525731,  0.850651),  // 0
    glm::vec3(0.850651,     0,          0.525731),  // 1
    glm::vec3(0.850651,     0,          -0.525731), // 2
    glm::vec3(-0.850651,    0,          -0.525731), // 3
    glm::vec3(-0.850651,    0,          0.525731),  // 4
    glm::vec3(-0.525731,    0.850651,   0),         // 5
    glm::vec3(0.525731,     0.850651,   0),         // 6
    glm::vec3(0.525731,     -0.850651,  0),         // 7
    glm::vec3(-0.525731,    -0.850651,  0),         // 8
    glm::vec3(0,            -0.525731,  -0.850651), // 9
    glm::vec3(0,            0.525731,   -0.850651), // 10
    glm::vec3(0,            0.525731,   0.850651)   // 11
};

static unsigned char icosahedronTriangles[] =
{
	1, 2, 6,
	1, 7, 2,
	3, 4, 5,
	4, 3, 8,
	6, 5, 11,
	5, 6, 10,
	9, 10, 2,
	10, 9, 3,
	7, 8, 9,
	8, 7, 0,
	11, 0, 1,
	0, 11, 4,
	6, 2, 10,
	1, 6, 11,
	3, 5, 10,
	5, 4, 11,
	2, 7, 9,
	7, 1, 0,
	3, 9, 8,
	4, 8, 0
};

struct Triangle
{
    unsigned short id;
    unsigned short a, b, c;
    Triangle* children[4];
    bool isLeaf;
    bool isRejected;

    Triangle()
    : id{0u}, a{0u}, b{0u}, c{0u},
      children{nullptr, nullptr, nullptr, nullptr}, isLeaf{false},
      isRejected{false}
    {
    }

    ~Triangle()
    {
        for (unsigned int i = 0; i < 4; ++i)
        {
            delete children[i];
            children[i] = nullptr;
        }
    }
};

struct Octahedron
{
    int trianglesY;
    int pointsY;

    // Different offsets and sizes to navigate in the vertices array
    std::vector<int> offsetPerSlice;
    std::vector<int> vertexPerSlice;
    std::vector<int> vertexPerTriangleSlice;
    std::vector<int> trianglePerSlice;
};

void createCubeSphere(std::vector<glm::vec3>& vertices, float radius, int subdivision);

void createUVSphere(float radius, int latitude, int longitude, std::vector<glm::vec3>& vertices, std::vector<glm::vec2>& texcoords);

void createOctahedronSphere(int subdivisions, float radius, std::vector<glm::vec3>& vertices, std::vector<int>& triangles, std::vector<glm::vec3>& uv, std::vector<glm::vec3>& normals);
void createOctahedronSphere(int subdivisions, float radius, std::vector<glm::vec3>& vertices, std::vector<unsigned int>& triangles);
void createOctahedronSphere(int subdivision, std::vector<glm::vec3>& vertices, std::vector<Triangle*>& triangles);

void createIcosahedronSphere(int subdivision, std::vector<glm::vec3>& vertices);

// Get the distance between two points for a given subdivision parameter.
float getIcosahedronDistance(int subdivision);
float getIcosahedronLonLat(int subdivision);

float getOctahedronLonLat(int subdivision);

void subdivTriangle(
        const glm::vec3& a, const glm::vec3& b, const glm::vec3& c,
        int subdiv,
        std::vector<glm::vec3>& vertices);

void subdivBarycentric(
        const glm::vec3& a, const glm::vec3& b, const glm::vec3& c,
        int subdiv,
        std::vector<glm::vec3>& barycentric);


#endif	/* SPHERE_H */

