// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   spherical_math.h
 * Author: gleseach
 *
 * Created on 14 octobre 2016, 17:45
 */

#ifndef SPHERICAL_MATH_H
#define SPHERICAL_MATH_H

class Camera;

float arcLength(const glm::vec3& p0, const glm::vec3& p1);

/*!
 * \brief Compute the shortest distance between two points on a unit sphere.
 */
float arcLengthUnitSphere(const glm::vec3& p0, const glm::vec3& p1);

float arcLengthOnEarth(const glm::vec3& p0, const glm::vec3& p1);

/*!
 * \brief Compute the intersection between two arcs on a unit sphere.
 */
bool arcIntersectUnitSphere(const glm::vec3& arc0p0, const glm::vec3& arc0p1,
                            const glm::vec3& arc1p0, const glm::vec3& arc1p1,
                            glm::vec3& intersectionPoint);

/*!
 * \brief Compute if a point is in an infinite tetrahedron.
 * \param[in] p Point to test.
 * \param[in] apex Tetrahedron summit.
 * \param[in] base0 1st point of the tetrahedron basis.
 * \param[in] base1 2nd point of the tetrahedron basis.
 * \param[in] base2 3rd point of the tetrahedron basis.
 * \return true if p is in the tetrahedron, false otherwise.
 * \warning Basis points must be in clockwise order.
 */
bool isPointInInfiniteTetrahedron(
    const glm::vec3& p,
    const glm::vec3& apex,
    const glm::vec3& base0, const glm::vec3& base1, const glm::vec3& base2);

/*!
 * \brief Compute if a point is in an infinite pyramid.
 * \param[in] p Point to test.
 * \param[in] apex pyramid summit.
 * \param[in] base0 1st point of the pyramid basis.
 * \param[in] base1 2nd point of the pyramid basis.
 * \param[in] base2 3rd point of the pyramid basis.
 * \param[in] base2 4th point of the pyramid basis.
 * \return true if p is in the pyramid, false otherwise.
 * \warning Basis points must be in clockwise order.
 */
bool isPointInInfinitePyramid(
    const glm::vec3& p,
    const glm::vec3& apex,
    const glm::vec3& base0,
    const glm::vec3& base1,
    const glm::vec3& base2,
    const glm::vec3& base3);

/*!
 * \brief Compute if a point is in an infinite pyramid with his summit at the world origin.
 * \param[in] p Point to test.
 * \param[in] base0 1st point of the pyramid basis.
 * \param[in] base1 2nd point of the pyramid basis.
 * \param[in] base2 3rd point of the pyramid basis.
 * \param[in] base2 4th point of the pyramid basis.
 * \return true if p is in the pyramid, false otherwise.
 * \warning Basis points must be in clockwise order.
 */
bool isPointInInfiniteUnitPyramidCentered(
    const glm::vec3& p,
    const glm::vec3& base0,
    const glm::vec3& base1,
    const glm::vec3& base2,
    const glm::vec3& base3);


/*!
 * \brief Spherical linear interpolation.
 */
glm::vec3 slerp(const glm::vec3& p0, const glm::vec3 p1, float t);

/*!
 * \brief Subdivise a line to follow the globe curvature.
 * \param[in] p0 First point of the line.
 * \param[in] p1 Second point of the line.
 * \param[in] minAngleDegrees Angle between two points on the line.
 * \param[out] innerPoints New points created between p0 and p1.
 */
void subdiviseLine(const glm::vec3& p0, const glm::vec3 p1, double minAngleDegrees, std::vector<glm::vec3>& innerPoints);
void subdiviseLine2(const glm::vec3& p0, const glm::vec3 p1, float pixelRes, std::vector<glm::vec3>& innerPoints);

/*!
 * \brief Compute the distance on the globe of a pixel at the equator (in meter).
 */
float equatorPixelResolution(const Camera* camera, const glm::ivec2& viewport);

float equatorUnitPixelResolution(const Camera* camera, const glm::ivec2& viewport);

/*!
 * \brief Signed angle from a to b in longitude
 */
float shortestLongitudeAngle(float a, float b);

/*!
 * \brief Center value angle in radians between [-PI, PI] range.
 */
float setToRangeMinusPiPlusPi(float rad);

/*!
 * \brief Compute if the triangle build with the three vertices a, b, c is built with the right orientation.
 */
bool isTrigoOnGlobe(const glm::vec3& a, const glm::vec3& b, const glm::vec3& c);

#endif /* SPHERICAL_MATH_H */
