// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/*
 * File:   streamlines_data.h
 * Author: gleseach
 *
 * Created on 9 janvier 2017, 12:32
 */

#ifndef STREAMLINES_DATA_H
#define STREAMLINES_DATA_H

#include "3d/chrono.h"
#include "data/image.h"
#include "core/log.h"


struct StreamlinesChunk
{
    float time;
    size_t head;
    size_t tail;
};


class StreamlinesBuffer
{
public:
    StreamlinesBuffer()
        : _currentId(0), _chunkSize(0), _ttl(0)
    {
    }
    ~StreamlinesBuffer()
    {
    }

    void init(size_t streamlinesCount, size_t fps, float ttl)
    {
        _ttl = ttl;

        // streamlines per frame
        _chunkSize = static_cast<size_t>(ceilf(static_cast<float>(streamlinesCount) / ttl / static_cast<float>(fps)));
        const size_t bufferSize = static_cast<size_t>(ceil(static_cast<double>(streamlinesCount) / static_cast<double>(_chunkSize)));
        _buffer.resize(bufferSize);

        float inter = _ttl / static_cast<float>((_buffer.size() - 1ull));

        size_t counter = 0;
        float timeCounter = -_ttl;

        for
            (StreamlinesChunk& chunk : _buffer)
        {
            chunk.time = timeCounter;
            chunk.head = counter;
            chunk.tail = std::min(counter + _chunkSize - 1, streamlinesCount - 1);
            
            counter += _chunkSize;
            timeCounter += inter;

            if
                (   chunk.head > streamlinesCount - 1
                 || chunk.tail > streamlinesCount - 1)
            {
                LOG(SEAScope::LogLevel::Error, "StreamlinesBuffer", "indice overflow");
            }
        }
    }

    bool getChunk(float t, StreamlinesChunk& chunk)
    {
        StreamlinesChunk& current = _buffer[_currentId];

        if
            (t >= current.time + _ttl)
        {
            chunk = current;
            current.time = t;
            _currentId = ++_currentId % _buffer.size();
            return true;
        }

        return false;
    }

    size_t getChunkSize() const { return _chunkSize; }
    const std::vector<StreamlinesChunk>& getBuffer() const { return _buffer; }

private:
    std::vector<StreamlinesChunk> _buffer;
    size_t _currentId;
    size_t _chunkSize;
    float _ttl;
};


namespace OGL
{
    class VertexArrayObject;
    class FrameBuffer;
    class Buffer;
    class Context;
    class Texture;
}

class DataSet;
class StreamlinesBuffer;

class StreamlinesData
{
public:
    StreamlinesData();
    virtual ~StreamlinesData();

    void create(OGL::Context* ctx,
        unsigned int streamlineCount,
        unsigned int streamlineLength,
        float speed,
        DataSet* _dataSet);

    OGL::Texture* getGcpTexture() const { return _gcp; }
    OGL::Texture* getGcpNorthEastTexture() const { return _gcpNorthEast; }
    OGL::Texture* getGcpAR() const { return _gcpAR; }
    glm::vec2  getGcpUVMin() const { return _uvMin; }
    glm::vec2  getGcpUVRange() const { return _uvRange; }

    unsigned int getIndiceCount() const { return _indiceCount; }
    void setIndiceCount(unsigned int val) { _indiceCount = val; }
    unsigned int getStreamlineLength() const { return _streamlineLength; }
    void setStreamlineLength(unsigned int val);
    OGL::VertexArrayObject* getTargetVao() const { return this->_vaoRendering[_targetId]; }
    unsigned int getParticleCount() const { return _particleCount; }
    unsigned int getCurrentStreamline() const { return _currentStreamline; }
    void setCurrentStreamline(unsigned int val) { _currentStreamline = val; }
    float getArea() const { return _area; }
    void setArea(float val) { _area = val; }
    unsigned int getStreamlineCount() const { return _streamlineCount; }
    void setStreamlineCount(unsigned int val);
    float getSpeed() const { return _speed; }
    void setSpeed(float val) { _speed = val; }

    OGL::Buffer* getVboSrc() const { return _vbo[_sourceId]; }
    OGL::Buffer* getVboDst() const { return _vbo[_targetId]; }

    OGL::VertexArrayObject* getVaoPhysicSrc() const { return _vaoPhysic[_sourceId]; }
    OGL::VertexArrayObject* getVaoPhysicDst() const { return _vaoPhysic[_targetId]; }

    // Chrono
    Chrono& getLastTime() { return _lastTime; }
    Chrono& getLastUpdateTime() { return _lastUpdateTime; }
    Chrono& getOriginTime() { return _origin; }

    void swap();
    void update(OGL::Context* ctx);

    std::chrono::time_point<std::chrono::system_clock> _lastUpdate;

    float getStreamlinesPerMs() const { return _streamlinesPerMs; }

    void setRemaining(float f);
    float getRemaining() {return _remaining;}

    unsigned int getStreamlinesGenerated() const { return _streamlinesGenerated; }
    void setStreamlinesGenerated(unsigned int val) { _streamlinesGenerated = val; }

    float getStreamlinesPerFrameTarget() const { return _streamlinesPerFrameTarget; }
    void setStreamlinesPerFrameTarget(float val) { _streamlinesPerFrameTarget = val; }
    float getTTL() const { 
        return _ttl; 
    }
    float getFadeOut() const { 
        return _fadeOut; 
    }

    float getFadeIn() const {
        return _fadeIn;
    }

    StreamlinesBuffer* getSB() { return _sb; }

private:
    void clearBuffers();
    void initBuffers(OGL::Context* ctx);
    void setParticleCount(unsigned int val) { _particleCount = val; }

private:
    unsigned int _currentStreamline; //< Streamline id to update
    unsigned int _particleCount; //< Number of particles
    unsigned int _streamlineCount;
    unsigned int _streamlineCountTmp;
    bool _update;
    unsigned int _streamlineLength;
    unsigned int _streamlineLengthTmp;
    float _streamlinePerSec;
    float _speed;

    unsigned int _indiceCount;
    OGL::Buffer* _vertices_local_id_buffer;
    OGL::Buffer* _indices_buffer;
    OGL::Buffer* _vbo[2];
    OGL::VertexArrayObject* _vaoPhysic[2];
    OGL::VertexArrayObject* _vaoRendering[2];
    unsigned int _sourceId;
    unsigned int _targetId;
    Chrono _lastTime;
    Chrono _lastUpdateTime;
    Chrono _origin;

    OGL::Texture* _gcp;
    OGL::Texture* _gcpNorthEast;
    OGL::Texture* _gcpAR; // aspect ratio

    float _area;
    glm::vec2 _uvMin;
    glm::vec2 _uvRange;

    float _framerate;
    float _ttl;
    float _fadeIn;
    float _fadeOut;
    float _streamlinesPerMs;
    float _remaining;
    float _streamlinesPerFrameTarget;
    unsigned int _streamlinesGenerated;
    StreamlinesBuffer* _sb;
};

#endif /* STREAMLINES_DATA_H */

