// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   texture.h
 * Author: gleseach
 *
 * Created on 3 septembre 2015, 10:54
 */

#ifndef TEXTURE_H
#define	TEXTURE_H

#include "3d/bindable_object.h"
#include "3d/opengl_enum.h"

class Image;

namespace OGL
{

class Texture : public BindableObject
{
public:
    Texture(OGL::TextureType type);
    virtual ~Texture() override;

    virtual void allocate1D(int width, int channels, DataType data_type, void * buffer) = 0;
    virtual void allocate2D(int width, int height, int channels, DataType data_type, void * buffer) = 0;
    virtual void allocate2D(const Image& image) = 0;

    /*
     * \brief allocate a texture with its mipmaps. Mipmap Id == array Id.
     */
    virtual void allocate2D(const std::vector<Image>& mipmaps) = 0;

    /*! 
     * \brief Set the wrapping and the filtering modes for the current object.
     * \warning After the call of this function the object is in unbound state !
     */
    virtual void setWrapAndFilterMode(OGL::WrapMode wrap, OGL::FilterMode filter) = 0;

    /*! 
     * \brief Set the wrapping and the filtering modes for the current object.
     * \warning The current object have to be bound before calling this function.
     */
    virtual void setWrapAndFilterModeNoBind(OGL::WrapMode wrap, OGL::FilterMode filter) = 0;

    /*! 
     * \brief Set the wrapping mode for the current object.
     * \warning After the call of this function the object is in unbound state !
     */
    virtual void setWrapMode(OGL::WrapMode wrap) = 0;

    /*! 
     * \brief Set the filtering mode (min and mag) for the current object.
     * \warning After the call of this function the object is in unbound state !
     */
    virtual void setFilterMode(OGL::FilterMode filter) = 0;

    /*!
     * \brief Set the filtering mode for the current object. magfilter can be only nearest or linear.
     * \warning After the call of this function the object is in unbound state !
     */
    virtual void setFilterMode(OGL::FilterMode minfilter, OGL::FilterMode magfilter) = 0;

    /*!
     * \brief Set the anisotropic filter value for the current object. 1.f disable the filter, > 1.f enable the filter.
     * \warning After the call of this function the object is in unbound state !
     */
    virtual void setAnisotropicFilter(float val) = 0;

    /*!
     * \brief Set the anisotropic filter value for the current object. 1.f disable the filter, > 1.f enable the filter.
     * \warning The current object have to be bound before calling this function.
     */
    virtual void setAnisotropicFilterNoBind(float val) = 0;

    /*! 
     * \brief Set the wrapping mode for the current object without change the binding state.
     * \warning The current object have to be bound before calling this function.
     */
    virtual void setWrapModeNoBind(OGL::WrapMode wrap) = 0;

    /*! 
     * \brief Set the filtering mode (min and mag) for the current object without change the binding state. 
     * \warning The current object have to be bound before calling this function.
     */
    virtual void setFilterModeNoBind(OGL::FilterMode filter) = 0;

    /*!
     * \brief Set the filtering mode for the current object without change the binding state.
     * \warning The current object have to be bound before calling this function.
     */
    virtual void setFilterModeNoBind(OGL::FilterMode minfilter, OGL::FilterMode magfilter) = 0;

    /*! 
     * \brief Resize the buffer an already created object.
     * \warning Don't call if the object doesn't exist.
     */
    virtual void resize(int width, int height) = 0;

    /*!
     * \brief Retrieve the current bound texture ID and compare with the object ID.
     * \return true if the current object is bound, false if not.
     */
    virtual bool isCurrentBinding() const = 0;

    /*!
     * \brief Update the entire texture data with the array passed in parameter.
     * \warning The array passed in parameter has to be the same size than the texture array.
     */
    virtual void setData(void * data) = 0;

    /*!
     * \brief Update a part of the texture data.
     * \param[in] x Column of the origin.
     * \param[in] y Row of the origin.
     * \param[in] width Width of the part to write.
     * \param[in] height Height of the part to write.
     * \param[in] data Pointer to the data to write.
     */
    virtual void setData(unsigned int x, unsigned int y, unsigned int width, unsigned int height, void * data) = 0;

    virtual void dumpToImage(Image& img) = 0;

    OGL::FilterMode getFilterModeMin() const {return _minFilter;}
    OGL::FilterMode getFilterModeMag() const {return _magFilter;}
    OGL::WrapMode getWrapMode() const {return _wrap;}

    virtual void generateMipmaps() = 0;

    unsigned int width() const { return _width; }
    unsigned int height() const { return _height; }

protected:
    virtual void allocate(int * dimensions, int size, int channels, DataType data_type, void * buffer) = 0;


protected:
    unsigned int gl_type;
    unsigned int gl_binding;
    unsigned int gl_dataType;
    unsigned int gl_format;
    unsigned int gl_internal_format;
    TextureType type;
    int _channels;
    DataType _dataType;
    unsigned int _width;
    unsigned int _height;
    OGL::WrapMode _wrap;
    OGL::FilterMode _minFilter;
    OGL::FilterMode _magFilter;
};

} // namespace GL

#endif	/* TEXTURE_H */

