// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   transect.h
 * Author: gleseach
 *
 * Created on 8 novembre 2016, 11:25
 */

#ifndef TRANSECT_H
#define TRANSECT_H

#include "core/structs.h"

namespace OGL
{
    class Buffer;
    class Program;
    class Context;
    class FrameBuffer;
}

class ShapeMesh;
class Image;
class IScene;
class SceneView;
struct PointInfo;

/*!
 * \brief Extracts physical data from granules following the polyline path.
 */
class Transect
{
public:
    //! Constructor
    Transect();

    //! Destructor
    ~Transect();

    /*!
     * \brief Initialize GPU stuff
     */
    void create(OGL::Context* ctx, int width, int height);

    bool getPoints(ShapeMesh*, std::vector<glm::vec3>&) const;
    /*!
     * \brief Process the extraction
     */
    bool extract(OGL::Context* ctx, ShapeMesh*, SceneView& sceneView, std::map<SEAScope::RenderableId, SEAScope::Extraction>& result, bool discardExtractedData);

    void resize(int width, int height);

    OGL::FrameBuffer* getFBO() const;

private:
    /*!
     * \brief Convert a cartesian world position in a GCP with lat,lon globe position and i,j image position.
     * \param[in] point The input data, a point in world position.
     * \param[in] mvp The model view projection matrix to transform the input point.
     * \param[in] viewport The viewport size.
     * \param[in] ok A boolean to know if the return object is valid or not.
     * \return A GCP object.
     * \note The projected point has to be in the viewport to be valid.
     */
    SEAScope::GCP createGCP(const glm::vec3& point, const glm::mat4& mvp, const glm::ivec4& viewport, bool& ok) const;

    /*!
     * \brief Render each granule in a separate framebuffer object.
     * \param[in] ctx OpenGL context wrapper.
     * \param[in] polylineMesh The polyline mesh object.
     * \param[in] result Map of all extraction data.
     * \param[in] rIdMask Renderable id of the mask.
     */
    bool renderToFbo(
        OGL::Context* ctx,
        const std::vector<glm::vec3>&,
        SceneView& sceneView, std::map<SEAScope::RenderableId, SEAScope::Extraction>& result,
        const SEAScope::RenderableId& rIdMask);

    /*!
     * \brief Create a mask depending of the visibility of the polyline.
     * \param[in] result Map of all extraction data.
     * \param[in] points Points of the polyline.
     * \param[in] tmpExtraction Extraction object with the gcp of the polyline.
     * \param[in] rIdMask Renderable id of the mask.
     */
    void createMask(
        std::map<SEAScope::RenderableId, SEAScope::Extraction>& result,
        std::vector<PointInfo>& points,
        SEAScope::Extraction& tmpExtraction,
        SEAScope::RenderableId& rIdMask);

    /*!
     * \brief Render each granule in a separate framebuffer object.
	 * \param[in] result Map of all extraction data.
	 * \param[in] points Points of the polyline.
	 * \param[in] rIdMask Renderable id of the mask.
     */
    void checkData(
        std::map<SEAScope::RenderableId, SEAScope::Extraction>& result,
        std::vector<PointInfo>& points,
        SEAScope::RenderableId& rIdMask);

    /*!
     * \brief Render each granule in a separate framebuffer object.
	 * \param[in] result Map of all extraction data.
	 * \param[in] tmpExtraction Extraction object with the gcp of the polyline.
	 * \param[in] rIdMask Renderable id of the mask.
	 * \param[in] points Points of the polyline.
     */
    void extractData(
        std::map<SEAScope::RenderableId, SEAScope::Extraction>& result,
        SEAScope::Extraction& tmpExtraction,
        SEAScope::RenderableId& rIdMask,
        std::vector<PointInfo>& points);

private:
    OGL::FrameBuffer* _fbo;
    std::map<SEAScope::RenderableId, Image*> _extractedData;
    std::vector<glm::vec3> _allPoints;
};

#endif /* TRANSECT_H */
