// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef APP_EVENTS_DISPATCHER_H
#define APP_EVENTS_DISPATCHER_H

#include <typeinfo>
#include <typeindex>
#include <algorithm>
#include <unordered_map>
#include <vector>

#include "core/interfaces/iapp_events_subscriber.h"

namespace SEAScope
{
    // Forward declarations
    class AppEvent;

    class AppEventsDispatcher
    {
        public:
            AppEventsDispatcher();
            virtual ~AppEventsDispatcher();

            template<class T>
            void subscribe(IAppEventsSubscriber* subscriber)
            {
                this->_subscribers[std::type_index(typeid(T))].push_back(subscriber);
            }

            template<class T>
            void unsubscribe(IAppEventsSubscriber* subscriber)
            {
                std::type_index eventType = std::type_index(typeid(T));
                auto eventTypeFound = this->_subscribers.find(eventType);
                if
                    (this->_subscribers.end() == eventTypeFound)
                {
                    // not subscribers vector for the event type
                    return;
                }

                auto subscriberFound = std::find(eventTypeFound->second.begin(),
                                                 eventTypeFound->second.end(),
                                                 subscriber);
                if
                    (eventTypeFound->second.end() != subscriberFound)
                {
                    eventTypeFound->second.erase(subscriberFound);
                }

                if
                    (0 >= eventTypeFound->second.size())
                {
                    this->_subscribers.erase(eventTypeFound);
                }
            }

            template<class T>
            void dispatch(const T& event)
            {
                for
                    (auto& subscriber : this->_subscribers[std::type_index(typeid(T))])
                {
                    subscriber->handleAppEvent(std::type_index(typeid(T)), event);
                }
            }

        protected:
            std::unordered_map<std::type_index, std::vector<IAppEventsSubscriber*> > _subscribers;
    };
}

#endif
