// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef DATA_NODE_H
#define DATA_NODE_H

#include <cstdint>
#include <vector>
#include <map>


#include <iostream>

namespace SEAScope
{
    enum class NodeType: std::uint8_t {
        empty,
        boolean,
        number,
        text,
        image,
        list,
        dict
    };

    class DataNode
    {
        public:
            virtual ~DataNode();

            virtual NodeType getType() const = 0;
    };

    class EmptyNode final : public DataNode
    {
        public:
            EmptyNode()
            {}
            EmptyNode(EmptyNode&&)
            {}
            ~EmptyNode() override;

            NodeType getType() const override
            {
                return NodeType::empty;
            }
    };

    class DictNode final : public DataNode
    {
        public:
            DictNode()
            {}
            DictNode(DictNode&& src)
            {
                for
                    (const auto& it : src.children)
                {
                    this->children.insert(std::make_pair(it.first, std::move(it.second)));
                }

                src.children.clear();
            }
            ~DictNode() override;

            NodeType getType() const override
            {
                return NodeType::dict;
            }

            DictNode& operator= (DictNode&& src)
            {
                for
                    (auto& it : this->children)
                {
                    delete it.second;
                }
                this->children.clear();

                for
                    (const auto& it : src.children)
                {
                    this->children.insert(std::make_pair(it.first, std::move(it.second)));
                }

                src.children.clear();

                return (*this);
            }

            std::map<std::string, DataNode*> children;
    };

    class ListNode final : public DataNode
    {
        public:
            ListNode()
            {}
            ListNode(ListNode&& src)
            {
                for
                    (auto& it : src.children)
                {
                    this->children.push_back(std::move(it));
                }

                src.children.clear();
            }
            ~ListNode() override;

            NodeType getType() const override
            {
                return NodeType::list;
            }

            std::vector<DataNode*> children;
    };

    class BoolNode final : public DataNode
    {
        public:
            BoolNode()
            {}
            BoolNode(BoolNode&& src): value(src.value)
            {}
            BoolNode(bool v): value(v)
            {}
            ~BoolNode() override;

            NodeType getType() const override
            {
                return NodeType::boolean;
            }

            bool value;
    };

    class NumberNode final : public DataNode
    {
        public:
            NumberNode()
            {}
            NumberNode(NumberNode&& src): value(src.value)
            {}
            NumberNode(double v): value(v)
            {}
            ~NumberNode() override;

            NodeType getType() const override
            {
                return NodeType::number;
            }

            double value;
    };

    class TextNode final : public DataNode
    {
        public:
            TextNode()
            {}
            TextNode(TextNode&& src): value(std::move(src.value))
            {}
            TextNode(const std::string& v): value(v)
            {}
            ~TextNode() override;

            NodeType getType() const override
            {
                return NodeType::text;
            }

            std::string value;
    };

    class ImageNode final : public DataNode
    {
        public:
            ImageNode()
            {}
            ImageNode(ImageNode&& src): value(std::move(src.value))
            {}
            ImageNode(const std::string& v): value(v)
            {}
            ~ImageNode() override;

            NodeType getType() const override
            {
                return NodeType::image;
            }

            std::string value;
    };

    bool testDataNode(DictNode&);
    void printDataNode(const DataNode&);
}

#endif
