// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef DYNAMIC_H
#define DYNAMIC_H

#include <string>
#include <cstdint>

#include "core/log.h"
#include "structs.h"
#include "plugins_manager.h"

namespace SEAScope
{

	template<class T>
	class dynamic
	{
		private:
			T* object;	// current implementation
			std::string api;
			std::string name;
			std::string version;

		public:
			dynamic(const std::string& inName, const std::string& inVersion): object(nullptr), api(T::api()), name(inName), version(inVersion)
			{
				PluginMeta* meta(nullptr);	//
				bool ok(false);	//

				meta = dynamic<T>::manager->getPlugin(this->api, this->name, this->version, ok);
				if
					(false == ok)
				{
					LOG( SEAScope::LogLevel::Error, "dynamic<T>"
									  , "Could not find plugin {}/{}@{}"
									  , api, name, version);
					return;
				}

                void* opaque {nullptr};
                opaque = meta->create();
				this->object = reinterpret_cast<T*>(opaque);
                opaque = nullptr;
                meta = nullptr;

				dynamic<T>::manager->pluginCreated(api, this->name, this->version);
			}

			~dynamic()
			{
                if
                    (nullptr == this->object)
                {
                    return;
                }

				PluginMeta* meta(nullptr);	//
				bool ok(false);	//

				meta = dynamic<T>::manager->getPlugin(this->api, this->name, this->version, ok);
				if
					(false == ok)
				{
					LOG( SEAScope::LogLevel::Error, "dynamic<T>"
									  , "Could not find plugin {}/{}@{}"
									  , this->api, this->name, this->version);
					return;
				}

                void* opaque {nullptr};
                opaque = reinterpret_cast<void*>(this->object);
                meta->destroy(opaque);
                this->object = nullptr;
                opaque = nullptr;
                meta = nullptr;

                /*Notify manager so that it can release the plugin once it is not required anymore */
                dynamic<T>::manager->pluginDestroyed( this->api
                                                    , this->name
                                                    , this->version);
			}

			/**
			 *
			 */
			T* operator-> (void)
			{
				return (object);
			}

			bool valid() const
			{
				return nullptr != this->object;
			}

			static PluginsManager* manager;
	};
};

template<class T>
SEAScope::PluginsManager* SEAScope::dynamic<T>::manager = nullptr;

#endif
