// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef ENGINE_H
#define ENGINE_H

#include <map>
#include <set>
#include <vector>
#include <string>
#include <functional>
#include <unordered_map>

#include "core/types.h"
#include "core/structs.h"
#include "core/interfaces/imetadata_store.h"
#include "core/interfaces/idata_store.h"
#include "core/interfaces/iplugins_loader.h"
#include "core/plugins_manager.h"
#include "core/metadata_mediator.h"
#include "core/data_mediator.h"
#include "core/display_queue.h"
#include "core/rendering_control.h"
#include "core/rendering_cfg.h"
#include "core/sources_control.h"
#include "core/collections_control.h"
#include "core/timeline.h"
#include "core/colormap_manager.h"
#include "core/sockets_manager.h"
#include "core/settings.h"
#include "core/state.h"
#include "data/events_coverage.h"

#include "ui/iadvanced_rendering_control_view.h"
#include "ui/icollections_control_view.h"
#include "ui/irendering_control_view.h"
#include "ui/isources_control_view.h"
#include "ui/itimeline_view.h"

#include "core/engine_updates_manager.h"

#include "core/interfaces/igranules_index.h"
#include "core/interfaces/iapp_events_subscriber.h"
#include "local/user_generated.h"

#include "tools/window_controller.h"
#include "tools/scene_view_controller_original.h"
#include "tools/debug_controller.h"
#include "tools/selection_controller.h"
#include "tools/extraction_controller.h"
#include "tools/timeline_controller.h"
#include "tools/annotations_controller.h"

#include "3d/icons.h"

#include "core/rendering_mediator.h"

namespace SEAScope
{
    // Forward declarations
    class AppEvent;
    class AppEventsDispatcher;
    class GranuleInfo;

	class Engine: public virtual IAppEventsSubscriber
	{
		public:
			Engine(AppEventsDispatcher&, IGranulesIndex&, IAnnotationsStore&, IGranulesIndex&);
			virtual ~Engine() override;

            bool handleAppEvent(std::type_index, const AppEvent&) override;

			bool init(SceneView*, IMetadataStore*, IDataStore*, IPluginsLoader*);
			bool initColormaps();

			bool setCollectionsUI(ICollectionsControlView*);
			bool setRenderingUI(IRenderingControlView*);
			bool setAdvancedRenderingUI(IAdvancedRenderingControlView*);
			bool setTimelineUI(ITimelineView*);
			bool setSourcesUI(ISourcesControlView*);

			bool shutdown();

			bool setRenderables(std::map<CollectionId, std::vector<GranuleMetadata> >&&);

			void updateRendering(const RenderingCfg*, const RenderingParam&);
			void renderablesChanged();
			bool getRenderableLabel(const RenderableId&, std::string&);
			bool getRenderableLabel(const RenderableId&, std::string&, const std::string&);
			bool getRenderableUnits(const RenderableId&, std::string&);

            void update(timestamp_t, timestamp_t, timestamp_t);
            void computeTimeHints(const std::map<CollectionId, std::vector<GranuleMetadata>>&, timestamp_t, timestamp_t, timestamp_t);

			bool getRenderableInfo(const RenderableId&, GranuleInfo&);

            bool sendProcessingOpcode(std::uint64_t);

            UserGenerated& getUserGenerated();

            ExtractionController& getExtractionController();
            WindowController& getWindowController();
            SceneViewControllerBase& getSceneViewController();
            DebugController& getDebugController();
            SelectionController& getSelectionController();
            TimelineController& getTimelineController();
            AnnotationsController& getAnnotationsController();

            const ExtractionController& getExtractionController() const;
            const WindowController& getWindowController() const;
            const SceneViewControllerBase& getSceneViewController() const;
            const DebugController& getDebugController() const;
            const SelectionController& getSelectionController() const;
            const TimelineController& getTimelineController() const;
            const AnnotationsController& getAnnotationsController() const;

            Icons& getIcons();

            EngineUpdatesManager& getUpdatesManager();

            void setSettingsChangedCallback(const std::function<void()>&);

            bool getRenderedGranuleMetadata(const RenderableId&, GranuleMetadata&);

            const std::set<CollectionId>& getActiveCollections() const { return this->_activeCollections; }

            void variableAddedByAPI(const RenderableId&);

            AppEventsDispatcher& _appEventsDispatcher;
			MetadataMediator metadataMediator;
			DataMediator dataMediator;
            RenderingMediator renderingMediator;
			Timeline timeline;
			RenderingControl renderingControl;
			SourcesControl sourcesControl;
			CollectionsControl collectionsControl;
			PluginsManager plugins;
			ColormapManager cmaps;
			Settings settings;
			State state;
			SocketsManager sockets;
			DisplayQueue displayQueue;

        protected:
            bool updateCollections(const std::set<CollectionId>& toAdd, const std::set<CollectionId>& toDel, bool forceUpdate);
            bool updateCollections(const std::set<CollectionId>& toAdd, const std::set<CollectionId>& toDel, const std::set<CollectionId>& toUpdate, bool forceUpdate);
            void updateActiveVariables(const CollectionVariable&, bool, bool);

            UserGenerated _userGenerated;

            WindowController _windowController;
            SceneViewControllerOriginal _sceneViewController;
            DebugController _debugController;
            SelectionController _selectionController;
            ExtractionController _extractionController;
            TimelineController _timelineController;
            AnnotationsController _annotationsController;

            std::set<CollectionId> _activeCollections;
            Icons _icons;

            EngineUpdatesManager _updatesManager;

            std::function<void()> _settingsChangedCallback;
	};
};

#endif
