// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef ENGINE_UPDATES_MANAGER_H
#define ENGINE_UPDATES_MANAGER_H

#include <map>
#include <set>
#include <mutex>
#include <chrono>
#include <condition_variable>

#include "core/types.h"

namespace OGL
{
    class Context;
}

// Forward declarations
class SceneView;

namespace SEAScope
{
    // Forward declarations
    class Engine;
    struct RenderableId;
    class GranuleMetadata;
    struct TrajectoryMarkers;
    class IUserInterface;

    class EngineUpdatesManager final
    {
        public:
            EngineUpdatesManager();
            ~EngineUpdatesManager();

            bool run(Engine&);
            void terminate();
            void requestUpdate();
            void triggerPendingUpdate(Engine&);

            /* UI updates */
            void updateUI(Engine&, IUserInterface&, SceneView&);
            void setGranuleCounters(std::map<CollectionId, std::uint32_t>&&);
            void sceneViewChanged();
            void addVariable(const RenderableId&);

            /* Time hints */
            void updateTimeHints(OGL::Context* ctx, SceneView&);
            void setTimeHints(TrajectoryMarkers&&);

            /* renderables */
            void setRenderables(std::map<CollectionId, std::vector<GranuleMetadata> >&&);
            void getRenderables(std::map<CollectionId, std::vector<GranuleMetadata> >&);
            bool findRenderable(const RenderableId&, GranuleMetadata&);

        protected:
            bool _stopEngineUpdates;
            bool _engineUpdateRequested;
			std::mutex _engineUpdateMutex;
			std::condition_variable _engineUpdateCondition;
            std::chrono::high_resolution_clock::time_point _lastUpdate;

            std::mutex _uiUpdateMutex;
            std::map<CollectionId, std::uint32_t> _granuleCounters;
            std::set<RenderableId> _addedVariables;
            bool _granuleCountersChanged;
            bool _sceneViewChanged;

            std::mutex _timeHintsMutex;
            std::map<CollectionId, TrajectoryMarkers> _timeHints;
            std::set<CollectionId> _timeHintsChanged;

            std::mutex _renderablesMutex;
			std::map<CollectionId, std::vector<GranuleMetadata> > _renderables;
    };
}

#endif
