// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef JOB_H
#define JOB_H

#include <cstdint>
#include <string>
#include <functional>
#include <set>

#include "core/types.h"

namespace SEAScope
{
	enum class JobStatus: std::uint8_t {CREATED, PENDING, RUNNING, COMPLETE, CANCELLED, SUPERSEEDED};
	class Job
	{
		public:
			Job();
			virtual ~Job();

            Job(const Job& src)
            : task(src.task), id(src.id), name(src.name), errorMsg(src.errorMsg),
              beginDatetime(src.beginDatetime), endDatetime(src.endDatetime),
              status(src.status)
            {
            }
            Job(Job&& src)
            {
                *this = std::forward<Job&&>(src);  // call move assignment operator
            }


            Job& operator=(const Job& src)
            {
                this->task = src.task;
                this->id = src.id;
                this->name = src.name;
                this->errorMsg = src.errorMsg;
                this->beginDatetime = src.beginDatetime;
                this->endDatetime = src.endDatetime;
                this->status = src.status;

                return *this;
            }

            Job& operator=(Job&& src)
            {
                this->task = std::move(src.task);
                this->id = std::move(src.id);
                this->name = std::move(src.name);
                this->errorMsg = std::move(src.errorMsg);
                this->beginDatetime = src.beginDatetime;
                this->endDatetime = src.endDatetime;
                this->status = src.status;

                return *this;
            }

			void start();
			void cancel();
			void done();

			std::function<bool(std::string&)> task;
			std::string id;
			std::string name;
			std::string errorMsg;
			timestamp_t beginDatetime;
			timestamp_t endDatetime;
			JobStatus status;
	};

	class DataLoadingJob final: public Job
	{
		public:
            DataLoadingJob()
            :Job()
            {
            }

            DataLoadingJob(DataLoadingJob&& src)
            {
                *this = std::forward<DataLoadingJob&&>(src);  // call move assignment operator
            }
            ~DataLoadingJob() override;

            DataLoadingJob& operator=(DataLoadingJob&& src)
            {
                this->task = std::move(src.task);
                this->id = std::move(src.id);
                this->name = std::move(src.name);
                this->errorMsg = std::move(src.errorMsg);
                this->beginDatetime = src.beginDatetime;
                this->endDatetime = src.endDatetime;
                this->status = src.status;

                this->uri = std::move(src.uri);
                this->fields = std::move(src.fields);

                return *this;
            }

			std::string uri;
			std::set<std::string> fields;
	};
}

#endif
