// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef LOG_H
#define LOG_H

#include "vendors/_fmt.h"
#include "vendors/_string_view.h"

#include <string>
#include "core/interfaces/ilog_handler.h"

#ifndef _WIN32
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wgnu-zero-variadic-macro-arguments"
#endif

/*
* Use this macro to avoid complex object construction, string allocation for nothing.
*/
#define LOG(level, logger, message, ...) \
if (true == SEAScope::Log::isValid(level)) \
{ \
	SEAScope::Log::log(level, logger, message, ##__VA_ARGS__); \
} (void)0

#ifndef _WIN32
#pragma clang diagnostic pop
#endif

namespace SEAScope
{
    class Log
    {
    public:
        static bool log(const LogLevel&, const char*, const string_view&);

        template<typename ...Details>
        static bool log(const LogLevel& level, const char* logger, const string_view& msgTemplate, const Details&... details)
        {
            if
                (nullptr == Log::implementation)
            {
                return false;
            }

            if
                (false == Log::implementation->isMessageValid(level, logger))
            {
                return false;
            }

            std::string msg = fmt::format(msgTemplate, details...);
            return Log::log(level, logger, msg);
        }

        static LogLevel levelFromString(const string_view& logLevelStr)
        {
            if
                ("debug" == logLevelStr)
            {
                return LogLevel::Debug;
            }
            else if
                ("info" == logLevelStr)
            {
                return LogLevel::Info;
            }
            else if
                ("warn" == logLevelStr)
            {
                return LogLevel::Warn;
            }
            else if
                ("error" == logLevelStr)
            {
                return LogLevel::Error;
            }
            else if
                ("fatal" == logLevelStr)
            {
                return LogLevel::Fatal;
            }

            return LogLevel::Info;
        }

        static bool isValid(LogLevel level)
        {
            if
                (nullptr == Log::implementation)
            {
                return false;
            }

            return level >= implementation->getLevel();
        }

        static ILogHandler* implementation;
    };
};

#endif
