// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef RENDERING_MEDIATOR_H
#define RENDERING_MEDIATOR_H

#include <map>
#include <vector>
#include <unordered_map>

#include "core/types.h"
#include "core/data_bucket.h"
#include "core/raw_to_renderable_queue.h"
#include "data/gpu_data_builder.h"

// Forward declarations
class SceneView;
class SceneGraphController;

namespace OGL
{
    class Shape;
};

namespace SEAScope
{
    // Forward declarations
    class AppEventsDispatcher;
    class RenderingControlModel;
    class MetadataMediator;
    class DataMediator;
    class RenderingCfg;
    class GranuleMetadata;
    class Settings;
    class ThreadsPool;
    class DisplayQueue;

    struct RenderableIdHash
    {
        // https://stackoverflow.com/questions/17016175/c-unordered-map-using-a-custom-class-type-as-the-key
        std::size_t operator()(const RenderableId& t) const
        {
            // Compute individual hash values for first, second and third
            // http://stackoverflow.com/a/1646913/126995
            size_t res = 17;
            res = res * 31 + std::hash<SourceId>()( t.granule.first );
            res = res * 31 + std::hash<CollectionLocalId>()( t.collection.second );
            res = res * 31 + std::hash<GranuleLocalId>()( t.granule.second );
            res = res * 31 + std::hash<VariableId>()( t.variable );
            return res;
        }
    };

    struct RenderableIdURIHash
    {
        // https://stackoverflow.com/questions/17016175/c-unordered-map-using-a-custom-class-type-as-the-key
        std::size_t operator()(const std::pair<RenderableId, std::string>& t) const
        {
            // Compute individual hash values for first, second and third
            // http://stackoverflow.com/a/1646913/126995
            size_t res = 17;
            res = res * 31 + std::hash<SourceId>()(t.first.granule.first);
            res = res * 31 + std::hash<CollectionLocalId>()(t.first.collection.second);
            res = res * 31 + std::hash<GranuleLocalId>()(t.first.granule.second);
            res = res * 31 + std::hash<VariableId>()(t.first.variable);
            res = res * 31 + std::hash<std::string>()(t.second);
            return res;
        }
    };

    struct RenderingExpectations
    {
        std::unordered_map<SourceId, std::unordered_map<CollectionLocalId, std::unordered_map<VariableId, std::set<std::string> > > > fields;
        std::set<RenderableId> ids;
    };

    struct RenderedGranuleInfo
    {
        IDFDescriptor idf;
        DataModel dataModel;
    };

    class RenderingMediator
    {
        public:
            RenderingMediator(AppEventsDispatcher&, RenderingControlModel&, MetadataMediator&, DataMediator&, Settings&, DisplayQueue&);
            virtual ~RenderingMediator();

            bool show(const std::map<CollectionId, std::vector<GranuleMetadata> >& renderables);
            bool addVariables(const std::set<CollectionVariable>&);
            bool removeVariables(const std::set<CollectionVariable>&);
            bool removeCollections(const std::set<CollectionId>&);
            void updateRendering(const RenderingCfg* rCfg, const RenderingParam& param);

            bool getRenderableURI(const RenderableId&, std::string&);
            void setSceneView(SceneView*);

            bool setWorkers(ThreadsPool*);
            void shutdown();

        protected:
            bool buildExpectations(const std::map<CollectionId, std::vector<GranuleMetadata> >& renderables,
                                   RenderingExpectations&);

            void beginUpdate(RenderingExpectations&&);
            void updateComplete();

            void getViewportInfo(std::uint32_t&, float&) const;
            bool isTextureSizeSupported(const RenderableId&, std::uint32_t, std::uint32_t);
            bool alreadyHandled(const RenderableId&, float);
            bool isGPUDataNeeded(const RenderableId&, const std::string&, float);

            bool processLoadedData(const CollectionId&, const GranuleId&, const std::string&, float, const std::set<std::string>&);
            bool formatRenderableData(const std::string&, const std::set<RenderableId>& rId, const std::string& uri, const std::set<std::string>& fields, const CollectionCfg& colCfg, float resolution);
            bool sendToDisplayQueue(const RenderableId& rId, const Image* img, OGL::Shape* geometry, const CollectionCfg& colCfg, const DataInfo& dataInfo, const std::string& uri, float resolution);

            AppEventsDispatcher& _appEventsDispatcher;
            RenderingControlModel& _model;
            MetadataMediator& _metadataMediator;
            DataMediator& _dataMediator;
            Settings& _settings;
            DisplayQueue& _displayQueue;

			GPUDataBuilder _gpuData;
            RawToRenderableQueue _rawToRenderableQueue;

            SceneView* _sceneView;
            SceneGraphController* _sceneGraphController;
            ThreadsPool* _workers;

            std::uint32_t _maxTextureSize;

            std::mutex _granulesInfoMutex;
            std::unordered_map<SourceId, std::unordered_map<GranuleLocalId, RenderedGranuleInfo> > _granulesInfo;

            std::mutex _beingProcessedMutex;
            std::unordered_map<std::string, bool> _beingProcessed;

            std::mutex _compositeMutex;
            std::unordered_map<std::string, DataBucket> _compositeCache;

            std::mutex _expectedMutex;
            RenderingExpectations _expected;

            std::mutex _toGPUMutex;
            std::unordered_map<std::pair<RenderableId, std::string>, bool, RenderableIdURIHash> _toGPU;
    };
}

#endif
