// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef DATA_INFO_H
#define DATA_INFO_H

#include <cstdint>
#include <string>
#include <vector>
#include <bitset>
#include <array>
#include <map>
#include <set>
#include "core/types.h"
#include "data/events_coverage.h"
#include "data/image.h"

namespace SEAScope
{
    // Forward declarations
    class GranuleMetadata;

	struct GCP
	{
        std::uint64_t time; /*Milliseconds since EPOCH*/
		float lat;
		float lon;
		std::uint32_t i;
		std::uint32_t j;

		GCP(std::uint64_t _time, float _lat, float _lon, std::uint32_t _i, std::uint32_t _j)
        : time{_time}, lat{_lat}, lon{_lon}, i{_i}, j{_j}
        {}
	};

	struct Shape
	{
		std::uint32_t xArity; /*Minor axis*/
		std::uint32_t yArity; /*Major axis*/
        std::vector<std::uint64_t> gcpsTime;
        std::vector<float> gcpsLat;
        std::vector<float> gcpsLon;
        std::vector<std::uint32_t> gcpsMajorAxisIndex;
        std::vector<std::uint32_t> gcpsMinorAxisIndex;

        Shape()
        : xArity{0u}, yArity{0u}
        {}
	};

	struct DataInfo
	{
		std::uint8_t channels;
		std::uint32_t xArity;
		std::uint32_t yArity;
		DataType dataType;
		std::vector<float> offsets;
		std::vector<float> scaleFactors;
		std::vector<float> fillValues;

        DataInfo()
        : channels{0u}, xArity{0u}, yArity{0u}, dataType{DataType::none}
        {
        }
	};

	struct IDFDescriptor
	{
		std::string uri;
		std::uint32_t xArity;
		std::uint32_t yArity;
		Shape shape;

		/* Mandatory resolution-dependent attributes */
		float resolution; /* idf_spatial_resolution converted in meters */
		std::uint32_t subsamplingFactor; /* idf_subsampling_factor */

		IDFDescriptor()
        : xArity{0u}, yArity{0u}, resolution{0.f}, subsamplingFactor{0}
		{
		}
	};

	struct VariableCfg
	{
		std::string label;
		std::string units;
		std::set<std::string> fields;
		std::map<std::string, std::string> tags;
	};

	struct CollectionCfg
	{
		std::string label;
		bool xSeamless;
		bool ySeamless;
		bool mustBeCurrent;
		std::map<std::string, std::string> tags;
		std::map<VariableId, VariableCfg> variables;
	};

	struct RenderableId
	{
		bool granuleLevel;
		CollectionId collection;
		GranuleId granule;
		VariableId variable;

		RenderableId()
        : granuleLevel{true}, collection{0u, 0u}, granule{0u, 0u}, variable{0u}
        {}

		RenderableId(bool gLevel, CollectionId cId, GranuleId gId, VariableId vId)
        : granuleLevel{gLevel}, collection{cId}, granule{gId}, variable{vId}
		{}

        bool operator==(const RenderableId& other) const
        {
            return      granuleLevel == other.granuleLevel
                    &&  collection == other.collection
                    &&  granule == other.granule
                    &&  variable == other.variable;
        }

        bool operator!=(const RenderableId& other) const
        {
			return !(*this == other);
        }

		bool operator < (const RenderableId& other) const
		{
			if
				(granuleLevel)
			{
				if
					(false == other.granuleLevel)
				{
					return true;
				}

				if
					(granule.first == other.granule.first)
				{
					/* Same source, compare granule ids */
					if
						(granule.second == other.granule.second)
					{
						/* Same granules, comapre variables */
						return variable < other.variable;
					}
					return granule.second < other.granule.second;
				}
				return granule.first < other.granule.first;
			}
			else
			{
				if
					(true == other.granuleLevel)
				{
					return false;
				}

				if
					(collection.first == other.collection.first)
				{
					/* Same source, compare collection ids */
					if
						(collection.second == other.collection.second)
					{
						/* Same granules, comapre variables */
						return variable < other.variable;
					}
					return collection.second < other.collection.second;
				}
				return collection.first < other.collection.first;
			}
		}
	};

	struct CatalogueEntry
	{
		std::string collectionLabel;
		std::string variableLabel;
		std::map<std::string, std::string> collectionTags;
		std::map<std::string, std::string> variableTags;
		bool selected;
	};

	struct SourceInfo
	{
		SourceId id;
		std::string title;
		std::string description;
        std::map<std::string, std::string> tags;
		/*logo ?*/
	};

    struct PluginMeta
	{
        std::string path = "";
        std::string api = "";
        std::string name = "";
        std::string version = "";
        std::string description = "";
        std::uint64_t refCount = 0;
        void* handle = nullptr;
        PluginInstanciator create = nullptr;
        PluginDestroyer destroy = nullptr;
        bool loaded = false;
    };

	struct SearchCriteria
	{
		timestamp_t current = 0;
		timestamp_t start = 0;
		timestamp_t stop = 0;
		float resolution = 0.f;
		std::map<std::uint32_t, bool> collections;
	};

	struct Extraction
	{
		std::vector<float> fillValues;
		Image buffer;
		std::vector<GCP> shape;
		std::vector<GCP> gcps;
		std::string uri;
		std::vector<std::string> fields;
		timestamp_t start;
		timestamp_t stop;
	};

    struct TrajectoryMarkers
    {
        CollectionId collectionId;
        std::vector<glm::vec3> position;
        std::vector<glm::vec3> direction;
    };

    struct SceneViewState
    {
        bool lock;
        glm::quat q;
        float zoom;
    };

    struct TimelineState
    {
        std::string timespan;
        std::string timestep;
        std::string datetime;
        bool nearestMode;
        bool extendedView;
    };
};

#endif
