// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef TIMELINE_CFG_H
#define TIMELINE_CFG_H

#include <cstdint>
#include <string>
#include <array>
#include <vector>
#include <map>

#include "core/types.h"


namespace SEAScope
{
	class Timeframe final
	{
		public:
			Timeframe(const std::string& label, const std::string& description, timestamp_t pastOffset, timestamp_t futureOffset);
			Timeframe();
            Timeframe(const Timeframe& src)
            : label(src.label), description(src.description), pastOffset(src.pastOffset), futureOffset(src.futureOffset)
            {
            }
            Timeframe& operator=(const Timeframe& src)
            {
                this->label = src.label;
                this->description = src.description;
                this->pastOffset = src.pastOffset;
                this->futureOffset = src.futureOffset;

                return *this;
            }
			~Timeframe();

			std::string label;
			std::string description;
			timestamp_t pastOffset;
			timestamp_t futureOffset;
	};

	class Timestep final
	{
		public:
			Timestep(const std::string& label, const std::string& description, timestamp_t value);
			Timestep();
            Timestep(const Timestep& src)
            : label(src.label), description(src.description), value(src.value)
            {
            }
            Timestep& operator=(const Timestep& src)
            {
                this->label = src.label;
                this->description = src.description;
                this->value = src.value;

                return *this;
            }
			~Timestep();

			std::string label;
			std::string description;
			timestamp_t value;

            static const timestamp_t ONE_THIRD_TIMESPAN_VALUE;
            static const char ONE_THIRD_TIMESPAN_LABEL[];
	};

	class TimelineCfg final
	{
		public:
			TimelineCfg();
            TimelineCfg(const TimelineCfg& src)
			: defaultTimespan(src.defaultTimespan), defaultTimestep(src.defaultTimestep),
              useNearestMode(src.useNearestMode), advancedScrolling(src.advancedScrolling),
              _initialDatetime(src._initialDatetime), _initialDatetimeStr(src._initialDatetimeStr),
              _minYear(src._minYear), _maxYear(src._maxYear)
            {
			    this->timeframes.assign(src.timeframes.begin(), src.timeframes.end());
			    this->timesteps.assign(src.timesteps.begin(), src.timesteps.end());
            }
            TimelineCfg& operator=(const TimelineCfg& src)
            {
			    this->timeframes.assign(src.timeframes.begin(), src.timeframes.end());
			    this->timesteps.assign(src.timesteps.begin(), src.timesteps.end());
                this->_initialDatetime = src._initialDatetime;
                this->_initialDatetimeStr = src._initialDatetimeStr;
			    this->_minYear = src._minYear;
			    this->_maxYear = src._maxYear;
			    this->defaultTimespan = src.defaultTimespan;
			    this->defaultTimestep = src.defaultTimestep;
                this->useNearestMode = src.useNearestMode;
                this->advancedScrolling = src.advancedScrolling;

                return *this;
            }
			~TimelineCfg();

			bool addTimeframe(const std::string&, const std::string&, timestamp_t, timestamp_t);
			bool getTimeframe(const std::string& label, Timeframe& result) const;
			bool getTimeframe(timestamp_t pastOffset, timestamp_t futureOffset, Timeframe& result) const;
			bool getTimeframes( std::vector<std::string>& labels
							  , std::vector<std::string>& descriptions) const;
			bool getTimeframes(std::map<std::string, std::array<timestamp_t, 2>>& result) const;
            bool removeTimeframe(const std::string&, std::string&);

			bool addTimestep(const std::string&, const std::string&, timestamp_t);
			bool getTimestep(const std::string& label, Timestep& result) const;
			bool getTimestep(timestamp_t value, Timestep& result) const;
			bool getTimesteps( std::vector<std::string>& labels
							 , std::vector<std::string>& descriptions) const;
			bool getTimesteps( std::map<std::string, timestamp_t>& result) const;
            bool removeTimestep(const std::string&, std::string&);

            bool setInitialDatetime(timestamp_t);
            timestamp_t getInitialDatetime() const;

            bool setInitialDatetimeString(const std::string&);
            const std::string& getInitialDatetimeString() const;

            bool getUseAutoMaxYear() const;
            void setUseAutoMaxYear(bool);

            std::uint16_t getMinYear() const;
            void setMinYear(std::uint16_t);

            std::uint16_t getMaxYear() const;
            void setMaxYear(std::uint16_t);

			std::vector<Timeframe> timeframes;
			std::vector<Timestep> timesteps;
			std::string defaultTimespan;
			std::string defaultTimestep;
            bool useNearestMode;
            bool advancedScrolling;

            const static std::uint16_t MIN_VALID_YEAR;
            const static std::uint16_t MAX_VALID_YEAR;

        protected:
            timestamp_t _initialDatetime;
            std::string _initialDatetimeStr;
            bool _useAutoMaxYear;
			std::uint16_t _minYear;
			std::uint16_t _maxYear;
	};
};

#endif
