// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef TIMELINE_MODEL_H
#define TIMELINE_MODEL_H

#include <cstdint>
#include <functional>
#include <map>
#include "core/types.h"
#include "data/events_coverage.h"

namespace SEAScope
{
	class TimelineModel
	{
		public:
			TimelineModel();
			virtual ~TimelineModel();

			bool getCoverage(const CollectionId&, EventsCoverage&) const;
			bool getCoverage(const std::string&, EventsCoverage&) const;
			bool setCoverage(const CollectionId&, const EventsCoverage&, bool);
			bool removeCoverage(const CollectionId&, bool);
            bool updateCoverages();

			bool goToNext();
			bool goToPrev();

			bool setCurrent(timestamp_t);
			bool setBounds(timestamp_t, timestamp_t);
			bool setTimeframeOffsets(timestamp_t, timestamp_t);
			bool setTimestep(timestamp_t);
			bool setPrevNext(timestamp_t, timestamp_t);

			bool setCurrentCallback(const std::function<void(timestamp_t)>&);
			bool setBoundsCallback(const std::function<void(timestamp_t, timestamp_t)>&);
			bool setTimeframeCallback(const std::function<void(timestamp_t, timestamp_t)>&);
			bool setTimestepCallback(const std::function<void(timestamp_t)>&);
			bool setCoverageCallback(const std::function<void()>&);

			bool isDataAvailable(std::int16_t) const;
			bool isDataAvailable(std::int16_t, std::uint8_t) const;
			bool isDataAvailable(std::int16_t, std::uint8_t, std::uint8_t) const;

			bool setNearestMode(bool);
			bool applyNearestMode(timestamp_t&, bool, bool, bool);
            bool shiftByYears(std::int16_t);
            bool shiftByMonths(std::int32_t);
            bool shiftByDays(std::int32_t);
            bool shiftByTimeframeLength(float, float);

            bool setNearestDataBoundCallback(const std::function<bool(timestamp_t&, const timestamp_t, const timestamp_t)>&);

			std::function<void(timestamp_t)> currentCallback;
			std::function<void(timestamp_t, timestamp_t)> boundsCallback;
			std::function<void(timestamp_t, timestamp_t)> timeframeCallback;
			std::function<void(timestamp_t)> timestepCallback;
			std::function<void()> coverageCallback;

			timestamp_t current;
			timestamp_t lowerBound;
			timestamp_t upperBound;
			timestamp_t timeframePastOffset;
			timestamp_t timeframeFutureOffset;
			timestamp_t timestep;
			timestamp_t prev;
			timestamp_t next;
			bool nearestMode;
			bool forceNearestMode;

			std::map<CollectionId, EventsCoverage> coverages;

        protected:
            std::function<bool(timestamp_t&, const timestamp_t, const timestamp_t)> _nearestDataBoundCallback;
	};
};

#endif
