// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef IMAGE_H
#define IMAGE_H

#include "core/data_type.h"
#include "vendors/_glm.h"
#include "data/image_desc.h"

enum ERotation
{
    clockwise,
    counterclockwise
};

class Image
{
public:
    Image();
    virtual ~Image();

	// Disable copy constructor and assignment
	Image(const Image&) = delete;
	Image& operator=(const Image&) = delete;

    //! Dupplicate the current object
    Image* clone() const;

    //! Clear image buffer and parameters
	void deallocate();

    //! Allocate the image buffer in a raw way (data type = none)
    //! Copy the data from data pointer, do nothing if null.
    void allocate(unsigned int width, unsigned int height, unsigned int componentCount, unsigned int componentSize, void * data);

    //! Allocate the image buffer in a type safe way
    //! Copy the data from data pointer, do nothing if null.
    void allocate(unsigned int width, unsigned int height, unsigned int componentCount, DataType dataType, void * data);

    //! Copy the data from data pointer, do nothing if null.
    void allocate(const ImageDesc& desc, void * data);

    //! Replace the Image pointer data by the pointer in parameter.
    //! Initialize Image attributes with ImageDesc variables.
    //! Can be use to wrap a pointer allocate by a loading library to avoid a copy.
    //! Pointer will be deleted by the destructor as usual.
    void setData(const ImageDesc& desc, void* data);

    /*!
     * \brief Wrapper around generical copyTo. The destination image will have the size of the image chunk.
     * \param[in] srcOrigin The origin of the copy in the source image.
     * \param[in] region The size of the image chunk to copy.
     * \param[out] dstImage The destination image, will be allocated if the size is not correct.
     */
    void copyTo(const glm::ivec2& srcOrigin, const glm::ivec2& region, Image& dstImage);

    /*!
     * \brief Copy a part of the source image in the destination image. The destination have to be allocated.
     * \param[in] srcOrigin The origin of the copy in the source image.
     * \param[inout] dstImage The destination image, have to be allocated to receive the requested amount of data.
     * \param[in] dstOrigin The origin to copy in the destination image.
     * \param[in] region The size of the image chunk to copy.
     */
    void copyTo(const glm::ivec2& srcOrigin, const Image& dstImage, const glm::ivec2& dstOrigin, const glm::ivec2& region);

    //! Return a raw pointer to the image data buffer
    void * data() const {return this->_rawData;}

    //! Return the pointer of the pixel located at x,y coordinates.
    void * pixelData(unsigned int x, unsigned int y);

	//! Set the pixel data located at x,y coordinates.
	void setPixelData(unsigned int x, unsigned int y, void* data);

    //! Extract one channel from the image source and save it in the destination image.
    void extractChannel(unsigned int channelID, Image& out) const;

    //! Return the number of pixels in the image
    unsigned int pixelCount() const {return _width * _height;}

    //! Return the size in bytes of the image
    unsigned long long dataSize() const;

    // Return the image status. Valid mean that the image is initialized and allocated.
    inline bool isValid() const {return this->_valid;}

    //! Return the number of atomic values in the image.
    inline std::uint64_t valuesCount() const {return static_cast<std::uint64_t>(this->_width) * static_cast<std::uint64_t>(this->_height) * static_cast<std::uint64_t>(this->_componentCount);}

    //! Return the image width
    unsigned int width() const {return _width;}

    //! Return the image height
    unsigned int height() const {return _height;}

    //! Return the number of channels
    unsigned int channels() const {return _componentCount;}

    //! Return the type of an atomic value
    DataType dataType() const {return _dataType;}

    //! Return the size in byte of one pixel.
    unsigned int pixelMemSize() const;

    //! Flip the image using the vertical axis.
    void flipVerticaly();

    //! Flip the image using the horizontal axis.
    void flipHorizontaly();

    //! Apply a rotation of 90 degrees to the image.
    void rotate90(ERotation rotation);

    ImageDesc getDesc() const;

protected:
    unsigned int _width; ///< Image width
    unsigned int _height; ///< Image height
    unsigned int _componentCount; ///< Components per pixel
    unsigned int _componentSize;  ///< Size in byte of one component
    DataType _dataType; ///< The basic type of an atomic value
    void * _rawData; ///< Pointer to the image buffer
    bool _valid; ///< Allocated or not flag
};

#endif /* IMAGE_H */

