// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef SQLITE_GRANULES_INDEX_H
#define SQLITE_GRANULES_INDEX_H

#include <mutex>

#include "core/interfaces/igranules_index.h"

struct sqlite3;
struct sqlite3_stmt;

namespace SEAScope
{
	class SQLiteGranulesIndex: public virtual IGranulesIndex
	{
		public:
			SQLiteGranulesIndex();
			virtual ~SQLiteGranulesIndex() override;

			bool hasGranule(const std::string&) const override;
			bool hasGranule(std::uint64_t) const override;
			bool add(GranuleMetadata&, bool) override;
            bool updateTimeCoverage(std::uint64_t, const timestamp_t, const timestamp_t) override;
            bool remove(const std::string&, bool) override;
            bool remove(std::uint64_t, bool) override;
			std::array<bool, 2> adjacentGranules( const timestamp_t
												, const timestamp_t
												, const std::set<std::uint32_t>&
												, GranuleMetadata&
												, GranuleMetadata&) const override;
			bool closestGranuleBound( const timestamp_t
                                    , const timestamp_t
                                    , const timestamp_t
									, const std::set<std::uint32_t>&
									, GranuleMetadata&) const override;
			bool getCoverage(const std::set<std::uint32_t>&, EventsCoverage&) const override;
			bool getCoverage(std::uint32_t, EventsCoverage&) const override;
			bool findGranules(const SearchCriteria&, std::vector<GranuleMetadata>&) const override;
            std::uint64_t generateGranuleId() override;
            bool setGranuleAddedCallback(const std::function<void(std::uint64_t, std::uint32_t)>&) override;
            std::tuple<bool, bool> load(const std::string& path) override;
            bool save(const std::string& path) override;
            bool setGranuleIdGenerator(std::uint64_t) override;
            bool startBulkOperation() override;
            bool stopBulkOperation() override;
            bool cancelBulkOperation() override;
            bool increaseBulkOpsCount() override;
            bool countGranules(std::size_t&) const override;

            bool setCollectionIdGenerator(std::uint32_t) override;
            std::uint32_t generateCollectionId() override;
            bool addCollection(const std::string&, const std::uint32_t) override;
            bool removeCollection(const std::string&) override;
            bool removeCollection(std::uint32_t) override;
            bool listCollections(std::vector<std::pair<std::uint32_t, std::string> >&) const override;
            bool countCollections(std::size_t&) const override;

            bool clear() override;

        protected:
            bool prepareStatements();
            bool readNextGranuleId();
            bool readNextCollectionId();
            bool insertGranule(GranuleMetadata&);
            bool updateGranule(GranuleMetadata&);
            bool getGranuleById(GranuleMetadata&, std::uint64_t);
            bool getGranuleByDataId(GranuleMetadata&, const std::string&);
            bool getCoverageUsingStatement(sqlite3_stmt*, EventsCoverage&) const;
            bool bindFindCurrentGranules(std::uint32_t, timestamp_t) const;
            bool bindFindInTimeRangeGranules(std::uint32_t, timestamp_t, timestamp_t) const;

            sqlite3* _db;
            sqlite3_stmt* _getNextGranuleIdStmt;
            sqlite3_stmt* _insertGranuleStmt;
            sqlite3_stmt* _updateGranuleStmt;
            sqlite3_stmt* _getGranuleByDataIdStmt;
            sqlite3_stmt* _getGranuleByIdStmt;
            sqlite3_stmt* _hasGranuleByDataIdStmt;
            sqlite3_stmt* _hasGranuleByIdStmt;
            sqlite3_stmt* _removeGranuleStmt;
            sqlite3_stmt* _removeGranuleByIdStmt;
            sqlite3_stmt* _getTimeCoverageByCollectionIdStmt;
            sqlite3_stmt* _findCurrentGranulesStmt;
            sqlite3_stmt* _findInTimeRangeGranulesStmt;
            sqlite3_stmt* _beginTransactionStmt;
            sqlite3_stmt* _commitTransactionStmt;
            sqlite3_stmt* _rollbackTransactionStmt;

            sqlite3_stmt* _getNextCollectionIdStmt;
            sqlite3_stmt* _insertCollectionStmt;
            sqlite3_stmt* _removeCollectionStmt;
            sqlite3_stmt* _removeCollectionByURIStmt;
            sqlite3_stmt* _listCollectionsStmt;

            sqlite3_stmt* _countGranulesStmt;
            sqlite3_stmt* _countCollectionsStmt;
            sqlite3_stmt* _deleteAllCollectionsStmt;

            std::mutex _idGeneratorMutex;
			std::uint64_t _idGenerator;
            std::uint64_t _uncommittedOpsCount;
            std::uint64_t _maxUncommittedBulkOps;
			std::function<void(std::uint64_t, std::uint32_t)> _granuleAddedCallback;

            std::uint32_t _collectionIdGenerator;

	};
};

#endif
