// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef ANNOTATIONS_CONTROLLER_H
#define ANNOTATIONS_CONTROLLER_H

#include <set>
#include <string>
#include <unordered_map>

#include "core/interfaces/iapp_events_subscriber.h"
#include "core/rendering_cfg.h"
#include "local/user_generated.h"

namespace OGL
{
    // Forward declarations
    class Shape;
}

struct ShapeTypeHash
{
    template<typename T>
    std::size_t operator()(T t) const
    {
        return static_cast<std::size_t>(t);
    }
};

namespace SEAScope
{
    // Forward declarations
    class AppEventsDispatcher;
    class UserGenerated;
    class IAnnotationsStore;
    class IGranulesIndex;
    class Annotation;
    struct RenderableId;
    class DictNode;

    class AnnotationsController final : public virtual IAppEventsSubscriber
    {
        public:
            AnnotationsController(AppEventsDispatcher&, IAnnotationsStore&, IGranulesIndex&);
            ~AnnotationsController() override;

            bool handleAppEvent(std::type_index, const AppEvent&) override;

			bool initUserShapes();
            bool loadFromFile(const std::string&, const std::map<ShapeType, CollectionId>&, const AnnotationFormat&, double, double, const std::set<CollectionId>&);
            bool saveToFile(const std::string&, const std::set<RenderableId>&, const std::map<ShapeType, CollectionId>&, const AnnotationFormat&);

            bool init(const std::string&);
            void initAnnotation(std::uint64_t, std::uint32_t, const Annotation&);
            bool addAnnotation(const RenderableId&, OGL::Shape*, double, double);
            bool updateAnnotation(const RenderableId&, OGL::Shape*);
            bool saveAnnotation(const RenderableId&, const std::string&, const AnnotationFormat&, DictNode*);

            void createAnnotationsCollection(const ShapeType&, const std::string&, const std::string&, bool);
            void createAnnotationsCollection(const ShapeType&, const std::string&, const std::string&, bool, const CollectionId&);
            bool handleDefaultCollectionVariable(const CollectionVariable&, CollectionVariable&, ShapeType, bool);

            bool getAnnotation(const RenderableId&, Annotation&) const;

            UserGenerated* getUserGenerated();
            const IAnnotationsStore* getStore() const;

        protected:
            bool applyMigrations();

            AppEventsDispatcher& _appEventsDispatcher;
            UserGenerated _userGenerated;
            IAnnotationsStore& _store;

            std::map<CollectionId, RenderingCfg> _renderingCfgs;

            std::unordered_map<ShapeType, CollectionVariable, ShapeTypeHash> _defaultColvars;
    };
};

#endif
