// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef SCENE_VIEW_CONTROLLER_ORIGINAL_H
#define SCENE_VIEW_CONTROLLER_ORIGINAL_H

#include "tools/scene_view_controller_base.h"

namespace SEAScope
{

enum class SceneViewControllerOriginalActions : ActionType {
	CURSOR_MOVE,
	RESET,
	MOVE_START,
	MOVE,
	MOVE_END,
	ZOOM_IN_TO,
	ZOOM_IN,
	ZOOM_OUT_FROM,
	ZOOM_OUT,
	INCREASE_ZOOM_SPEED,
	DECREASE_ZOOM_SPEED,
	TOGGLE_POLES_AXIS_LOCKING,
	RESIZE_VIEWPORT,
	USE_FULL_MODE,
	USE_WIREFRAME_MODE,
	USE_DOT_MODE,
};

class SceneViewControllerOriginal final : public virtual SceneViewControllerBase
{
public:
	SceneViewControllerOriginal(AppEventsDispatcher&);
	~SceneViewControllerOriginal() override;

	bool isActive() const override;
	bool handleAppEvent(std::type_index, const AppEvent&) override;

	/**
	 * Reset the position and altitude of the scene's camera to their defaults.
	 */
	void resetCamera() override;

	/**
	 * Move the scene's camera one step closer to the surface of the globe.
	 */
	void zoomIn() override;

	/**
	 * Move the scene's camera one step farther from the surface of the globe.
	 */
	void zoomOut() override;

	/**
	 * Move the scene's camera to look at the given location, while maintaining
	 * its altitude.
	 *
	 * @param lon The longitude of the location to look at.
	 * @param lat The latitude of the location to look at.
	 */
	void lookAt(float lon, float lat) override;

	/**
	 * Set the altitude of the scene's camera to the given one.
	 *
	 * @param altitude The altitude at which to place the scene's camera. Where
	 *                 `0` is at the center of the globe and `1` is at its surface.
	 */
	void setZoom(float altitude) override;
	void setState(const SceneViewState&) override;
	void getState(SceneViewState&) const override;

private:
	// Events
	void initEvents();
	void cleanEvents();
	bool performAction(const ActionType& actionType, const InputEvent& event, const InputDevicesState& devicesState) override;
	void updatePossibleActions(const std::set<SceneViewControllerOriginalActions>& toEnable, const std::set<SceneViewControllerOriginalActions>& toDisable);

	// Camera actions
	bool startCameraMovement();
	bool moveCamera(const InputEvent& event);
	bool stopCameraMovement();
    bool setZoomImpl(float);
    bool zoomInImpl();
	void zoomInTo(); // zoom on cursor

	void updateCameraRotation();
	void updateCursorPosition(const InputDevicesState& devicesState);
	bool getGlobePosition(glm::vec3& globePosition, const glm::vec2& mousePosition, const glm::quat q);
	bool getGlobePosition(glm::vec3& globePosition);
	void setLockPolesAxis(bool lockPolesAxis);
	void emitCameraMovedEvent();

private:
	float _zoomSpeed;
	glm::ivec2 _lastMousePosition;
	glm::vec2 _angles;
	float _zoom;
	glm::quat _q;

	bool _lockPolesAxis;
    bool _arcBallWasOnGlobe;
	glm::quat _arcBallQ;
	glm::vec3 _arcBallGlobePos;

	bool _viewLocked;
    glm::vec3 _zoomGlobePos;
};

} // namespace SEAScope

#endif // SCENE_VIEW_CONTROLLER_ORIGINAL_H
