// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef SELECTION_CONTROLLER_H
#define SELECTION_CONTROLLER_H

#include <set>
#include <map>
#include <cstdint>
#include <string>
#include <functional>
#include <unordered_map>

#include "vendors/_glm.h"
#include "core/base_input_events_subscriber.h"
#include "core/interfaces/iapp_events_subscriber.h"
#include "3d/ipre_renderer.h"
#include "3d/ipost_renderer.h"
#include "3d/selection_rectangle.h"

namespace OGL
{
    class Context;
    class FrameBuffer;
    class Buffer;
    class Sync;
}

class Camera;
class SceneView;
class Pickable;

namespace SEAScope
{
    // Forward declarations
    class AppEventsDispatcher;

    enum class SelectionControllerActions: ActionType {
        RESIZE_FRAMEBUFFER,
        RECTANGLE_START,
        RECTANGLE_RESIZE,
        RECTANGLE_END,
        SELECT_ALL,
        UPDATE_SELECTION,
        EXCLUSIVE_SELECTION,
        TOGGLE_PICKING_RENDERING,
    };

    class SelectionController final : public virtual BaseInputEventsSubscriber,
                                      public virtual IAppEventsSubscriber,
                                      public virtual IPreRenderer,
                                      public virtual IPostRenderer
    {
        public:
            SelectionController(AppEventsDispatcher&);
            ~SelectionController() override;

            bool performAction(const ActionType&, const InputEvent&, const InputDevicesState&) override;

            bool preRender(OGL::Context*, SceneView*) override;
            bool postRender(OGL::Context*, SceneView*) override;

            void updatePossibleActions(const std::set<SelectionControllerActions>& toEnable,
                                       const std::set<SelectionControllerActions>& toDisable);

            void getPickingObjectAsync(const glm::ivec2&);
            bool readPickingObjectAsync(Pickable*&);

            bool handleAppEvent(std::type_index, const AppEvent&) override;

            void setHovered(Pickable*);

            void updateSelection();
            void exclusiveSelection();

            void pickableNotRendered(Pickable*);
            void pickableRemoved(Pickable*);
            void lockSelection(bool);

        protected:
            AppEventsDispatcher& _appEventsDispatcher;

            bool _showPicking;
            bool _pickingUpdate;
            bool _refreshPickedObject;
            bool _asyncRunning;
            bool _selectionLocked;

            SelectionRectangle _rectangle;

            OGL::FrameBuffer* _pickingFBO;
            OGL::Buffer* _pickingPBO;
            OGL::Sync* _sync;

            std::uint32_t _uniqueId;
            std::unordered_map<std::uint32_t, Pickable*> _pickables;

            bool _hasQueuedScreenPosition;
            glm::ivec2 _queuedScreenPosition;
            glm::ivec2 _latestPickedPosition;
            std::int32_t _screenHeight;
            std::vector<Pickable*> _hovered;
            std::vector<Pickable*> _selected;
    };
};

#endif
