// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iostream>
#include <Foundation/Foundation.h>

#include "platform/app_launcher_impl.h"
#include "platform/nsapplication_delegate.h"
#include "ini_settings.h"
#include "core/settings.h"
#include "core/path.h"
#include "vendors/_glfw3.h"
#include "vendors/_mini.h"

using namespace SEAScope;

bool getBundleIdentifier(std::string&);
bool getBundleResourcePath(std::string&);
bool getPathForDirectory(NSSearchPathDirectory, std::string&);
bool createSymbolicLink(const std::string&, const std::string&);
bool isPathReachable(const std::string&);
bool realPath(const std::string&, std::string&);
bool copyItemAtPathTo(const std::string&, const std::string&);
void pathParentDir(const std::string&, std::string&);
bool getApplicationSupportDir(std::string&);
bool getLogDir(std::string&);
bool isLaunchedAsBin();
bool redirectLogs();
void copyMissingBundledResources(const std::string&, const std::string&);

bool getBundleIdentifier(std::string& identifier)
{
    @autoreleasepool
    {
        NSString *appBundleID = [[NSBundle mainBundle] bundleIdentifier];
        if
            (nil == appBundleID)
        {
            return false;
        }
        identifier.assign([appBundleID UTF8String]);
        return true;
    }
}

bool getBundleResourcePath(std::string& path)
{
    @autoreleasepool
    {
        NSString *appBundleID = [[NSBundle mainBundle] bundleIdentifier];
        if
            (nil == appBundleID)
        {
            // Not launched from a bundle
            return false;
        }

        NSURL *resourceURL = [[NSBundle mainBundle] resourceURL];
        if
            (nil == resourceURL)
        {
            return false;
        }
        path.assign(resourceURL.fileSystemRepresentation);
        return true;
    }
}

bool getPathForDirectory(NSSearchPathDirectory directory, std::string& path)
{
    @autoreleasepool
    {
        NSFileManager *fileManager = [NSFileManager defaultManager];
        NSArray *possibleURLs = (
            [fileManager URLsForDirectory:directory inDomains:NSUserDomainMask]
        );
        if
            (0 == [possibleURLs count])
        {
            return false;
        }

        // Use the first directory (if multiple are returned)
        NSURL *dirURL = possibleURLs[0];
        if
            (!dirURL)
        {
            return false;
        }

        path.assign(dirURL.fileSystemRepresentation);
        return true;
    }
}

bool createSymbolicLink(const std::string& fromPath, const std::string& toPath)
{
    @autoreleasepool
    {
        NSURL *fromURL = [NSURL fileURLWithFileSystemRepresentation:fromPath.c_str()
                                                        isDirectory:NO
                                                      relativeToURL:nil];
        NSURL *toURL = [NSURL fileURLWithFileSystemRepresentation:toPath.c_str()
                                                      isDirectory:NO
                                                    relativeToURL:nil];

        NSURL *fromDirURL = [fromURL URLByDeletingLastPathComponent];
        NSFileManager *fileManager = [NSFileManager defaultManager];
        NSError *error = nil;
        BOOL ok = [fileManager createDirectoryAtURL:fromDirURL
                        withIntermediateDirectories:YES
                                         attributes:nil
                                              error:&error];
        if
            (false == ok)
        {
            NSLog(@"createSymbolicLink %@ -> %@ failed: \n%@", fromURL, toURL, error);
            return false;
        }

        ok = [fileManager createSymbolicLinkAtURL:fromURL
                               withDestinationURL:toURL
                                            error:&error];
        if
            (false == ok)
        {
            NSLog(@"createSymbolicLink %@ -> %@ failed: \n%@", fromURL, toURL, error);
            return false;
        }
        return true;
    }
}

bool isPathReachable(const std::string& path)
{
    @autoreleasepool
    {
        NSURL *url = [NSURL fileURLWithFileSystemRepresentation:path.c_str()
                                                    isDirectory:NO
                                                  relativeToURL:nil];
        return [url checkResourceIsReachableAndReturnError:nullptr];
    }
}

bool realPath(const std::string& path, std::string& realPath)
{
    @autoreleasepool
    {
        NSURL *url = [NSURL fileURLWithFileSystemRepresentation:path.c_str()
                                                    isDirectory:NO
                                                  relativeToURL:nil];
        url = [url URLByStandardizingPath];
        url = [url URLByResolvingSymlinksInPath];

        NSNumber *isSymlink = nil;
        [url getResourceValue:&isSymlink forKey:NSURLIsSymbolicLinkKey error:nullptr];
        if
            (true == [isSymlink boolValue])
        {
            NSString *symlinkPath = url.path;
            if
                (nil == url.path)
            {
                return false;
            }

            NSFileManager *fileManager = [NSFileManager defaultManager];
            NSString *symlinkedPath = (
                [fileManager destinationOfSymbolicLinkAtPath:symlinkPath
                                                       error:nullptr]
            );
            if
                (nil == symlinkedPath)
            {
                return false;
            }

            realPath.assign([symlinkedPath UTF8String]);
            return true;
        }

        realPath.assign(url.fileSystemRepresentation);
        return true;
    }
}

bool copyItemAtPathTo(const std::string& srcPath, const std::string& dstPath)
{
    @autoreleasepool
    {
        NSString *src = [NSString stringWithUTF8String:srcPath.c_str()];
        NSString *dst = [NSString stringWithUTF8String:dstPath.c_str()];
        NSError *error = nil;
        NSFileManager *fileManager = [NSFileManager defaultManager];
        BOOL ok = [fileManager copyItemAtPath:src toPath:dst error:&error];
        if
            (false == ok)
        {
            NSLog(@"copyItemAtPathTo %@ -> %@ failed: \n%@", src, dst, error);
            return false;
        }
        return true;
    }
}

void pathParentDir(const std::string& path, std::string& parentDir)
{
    @autoreleasepool
    {
        NSURL *url = [NSURL fileURLWithFileSystemRepresentation:path.c_str()
                                                    isDirectory:NO
                                                  relativeToURL:nil];
        NSURL *parentURL = [url URLByDeletingLastPathComponent];
        parentDir.assign(parentURL.fileSystemRepresentation);
    }
}

bool getApplicationSupportDir(std::string& path)
{
    std::string appSupportPath;
    if
        (false == getPathForDirectory(NSApplicationSupportDirectory, appSupportPath))
    {
        return false;
    }

    std::string appIdentifier;
    if
        (false == getBundleIdentifier(appIdentifier))
    {
        return false;
    }

    Path::joinPaths({appSupportPath, appIdentifier}, path);
    return true;
}

bool getLogDir(std::string& path)
{
    std::string libraryPath;
    if
        (false == getPathForDirectory(NSLibraryDirectory, libraryPath))
    {
        return false;
    }

    std::string appIdentifier;
    if
        (false == getBundleIdentifier(appIdentifier))
    {
        return false;
    }

    Path::joinPaths({libraryPath, "Logs", appIdentifier}, path);
    return true;
}

bool isLaunchedAsBin()
{
    struct stat filestat;
    dev_t null_dev;
    ino_t null_ino;
    dev_t null_rdev;

    stat("/dev/null", &filestat);
    null_dev = filestat.st_dev;
    null_ino = filestat.st_ino;
    null_rdev = filestat.st_rdev;

    fstat(STDOUT_FILENO, &filestat);
    if
        (
            filestat.st_dev != null_dev ||
            filestat.st_ino != null_ino ||
            filestat.st_rdev != null_rdev
        )
    {
        return true;
    }

    fstat(STDERR_FILENO, &filestat);
    if
        (
            filestat.st_dev != null_dev ||
            filestat.st_ino != null_ino ||
            filestat.st_rdev != null_rdev
        )
    {
        return true;
    }

    fstat(STDIN_FILENO, &filestat);
    if
        (
            filestat.st_dev != null_dev ||
            filestat.st_ino != null_ino ||
            filestat.st_rdev != null_rdev
        )
    {
        return true;
    }

    std::string cwd;
    Path::currentWorkingDir(cwd);
    if
        (cwd != "/")
    {
        return true;
    }

    return false;
}

bool redirectLogs()
{
    std::string logDir;
    if
        (false == getLogDir(logDir))
    {
        @autoreleasepool
        {
            NSLog(
                @"[ERROR] Could not get log dir %@",
                [NSString stringWithUTF8String:logDir.c_str()]
            );
        }
        return false;
    }
    if
        (false == Path::createDirs(logDir, 0755))
    {
        @autoreleasepool
        {
            NSLog(
                @"[ERROR] Could not create log dir %@",
                [NSString stringWithUTF8String:logDir.c_str()]
            );
        }
        return false;
    }

    std::string logPath = logDir + Path::separator() + "seascope.log";

    int logFd = open(logPath.c_str(), O_WRONLY | O_CREAT | O_TRUNC, 0755);
    if
        (-1 == logFd)
    {
        @autoreleasepool
        {
            NSLog(
                @"[ERROR] Could not open output log file %@",
                [NSString stringWithUTF8String:logPath.c_str()]
            );
        }
        return false;
    }
    int r = dup2(logFd, STDOUT_FILENO);
    if
        (-1 == r)
    {
        NSLog(@"[ERROR] Could not redirect stdout to file.");
        return false;
    }
    @autoreleasepool
    {
        NSLog(
            @"Redirected stdout to %@",
            [NSString stringWithUTF8String:logPath.c_str()]
        );
    }
    r = dup2(logFd, STDERR_FILENO);
    if
        (-1 == r)
    {
        NSLog(@"[ERROR] Could not redirect stderr to file.");
        return false;
    }
    close(logFd);
    @autoreleasepool
    {
        NSLog(
            @"Redirected stderr to %@",
            [NSString stringWithUTF8String:logPath.c_str()]
        );
    }

    return true;
}

void copyMissingBundledResources(const std::string& bundledPath, const std::string& customPath)
{
    if
        (true == Path::dirExists(customPath))
    {
        std::forward_list<std::string> bundledFilePaths;

        if
            (false == Path::listFiles(bundledPath, bundledFilePaths))
        {
            std::cerr << "Cannot list bundled resources in: " << bundledPath << std::endl;
            return;
        }

        std::string delimiter(1, Path::separator());
        for
            (const auto& bundledFilePath : bundledFilePaths)
        {
            std::forward_list<std::string> pathComponents;
            std::string fileName;
            Path::splitPaths(bundledFilePath, delimiter, pathComponents);
            pathComponents.reverse();
            fileName = pathComponents.front();

            if
                (fileName == "." || fileName == "..")
            {
                continue;
            }

            std::string path;
            Path::joinPaths({customPath, fileName}, path);
            copyMissingBundledResources(bundledFilePath, path);
        }
        return;
    }

    if
        (true == Path::fileExists(customPath))
    {
        return;
    }

    std::string delimiter(1, Path::separator());
    std::forward_list<std::string> pathComponents;
    std::string fileName;
    Path::splitPaths(bundledPath, delimiter, pathComponents);
    pathComponents.reverse();
    fileName = pathComponents.front();

    if
        (true == copyItemAtPathTo(bundledPath, customPath))
    {
        std::cerr << "Copied bundled resource " << fileName << " to configured path: " << customPath << std::endl;
    }
    else
    {
        std::cerr << "Cannot copy bundled resource " << fileName << " to configured path: " << customPath << std::endl;
    }
}

/**
 *
 */
AppLauncherImpl::AppLauncherImpl()
{
}

/**
 *
 */
AppLauncherImpl::~AppLauncherImpl()
{
}

/**
 *
 */
bool AppLauncherImpl::initWorkspace(const std::string& /*programPath*/,
                                    bool hasWorkspace, std::string& workspacePath,
                                    bool hasConfig, std::string& configPath,
                                    bool hasListen, std::string& listenURL,
                                    bool hasLogFilters, std::string& logFilters) const
{
    // Prevent GLFW from setting the current directory to the application to
    // the Contents/Resources subdirectory of the application's bundle
    glfwInitHint(GLFW_COCOA_CHDIR_RESOURCES, GLFW_FALSE);

    bool launchedAsBin = isLaunchedAsBin();
    if
        (false == launchedAsBin)
    {
        if
            (false == redirectLogs())
        {
            NSLog(@"[WARNING] Cannot redirect logs to file.");
        }

        std::string homeDir;
        if
            (  true == Path::homeDir(homeDir)
            && 0 == chdir(homeDir.c_str()))
        {
            std::cerr << "Changed current directory to user's home directory: " << homeDir << std::endl;
        }
    }

    bool workspacePointsToResources {false};
    if
        (false == hasWorkspace)
    {
        if
            (true == getBundleResourcePath(workspacePath))
        {
            std::cerr << "Using bundle resource directory as workspace." << std::endl;
            workspacePointsToResources = true;
        }
        else
        {
            std::cerr << "[WARNING] Cannot get bundle resource directory." << std::endl;

            if
                (true == launchedAsBin)
            {
                Path::currentWorkingDir(workspacePath);
                std::cerr << "Using current directory as workspace." << std::endl;
            }
            else
            {
                std::cerr << "No workspace specified and cannot get bundle resource directory." << std::endl;
                return false;
            }
        }
    }
    else
    {
        std::string workspacePathTmp = workspacePath;
        Path::absolutePath(workspacePathTmp, workspacePath);
    }
    Path::setPathValid(workspacePath);

    if
        (false == hasConfig)
    {
        if
            (true == workspacePointsToResources)
        {
            std::string appSupportDir;
            if
                (false == getApplicationSupportDir(appSupportDir))
            {
                std::cerr << "Cannot get application support directory." << std::endl;
                return false;
            }

            configPath = appSupportDir + Path::separator() + "config.ini";
            Path::setPathValid(configPath);

            if
                (false == isPathReachable(configPath))
            {
                std::string homeDir;
                if
                    (false == Path::homeDir(homeDir))
                {
                    std::cerr << "Cannot get user's home directory." << std::endl;
                    return false;
                }

                std::string workspaceConfigPath = (
                    homeDir
                    + Path::separator() + "SEAScope-workspace"
                    + Path::separator() + "config.ini"
                );
                if
                    (false == createSymbolicLink(configPath, workspaceConfigPath))
                {
                    std::cerr << "Cannot symlink the config to ~/SEAScope-workspace/config.ini" << std::endl;
                    return false;
                }
            }
        }
        else
        {
            configPath = workspacePath + Path::separator() + "config.ini";
        }
    }
    else
    {
        std::string configPathTmp = configPath;
        Path::absolutePath(configPathTmp, configPath);
    }
    Path::setPathValid(configPath);
    if
        (false == realPath(configPath, configPath))
    {
        std::cerr << "Cannot get real path of config file: " << configPath << std::endl;
        return false;
    }

    std::string configDir;
    pathParentDir(configPath, configDir);

    if
        (false == Path::fileExists(configPath))
    {
        bool canBeFixed {false};
        canBeFixed = Path::writeFile(configPath);
        if
            (false == canBeFixed)
        {
            std::cerr << "Cannot write to file '" << configPath << "', please check permissions" << std::endl;
            return false;
        }

        // Default config
        mINI::INIStructure root;
        root["paths"];
        root["paths"]["colormaps"] = configDir + Path::separator() + "colormaps";
        root["paths"]["data"] = configDir + Path::separator() + "data";
        root["paths"]["state"] = configDir + Path::separator() + "state.db";
        root["paths"]["index"] = configDir + Path::separator() + "index.fb";
        root["paths"]["annotations"] = configDir + Path::separator() + "annotations.db";
        root["paths"]["custom"] = configDir + Path::separator() + "custom";

        mINI::INIFile parser(configPath);
        bool ok = parser.write(root, true /*pretty print*/);
        if
            (false == ok)
        {
            std::cerr << "Could not write default configuration to '" << configPath << "'" << std::endl;
            return false;
        }
    }

    std::string resourcePath;
    if
        (true == getBundleResourcePath(resourcePath))
    {
        Path::setPathValid(resourcePath);

        Settings settings;
        INISettings settingsStore;
        if
            (true == settingsStore.load(settings, configPath))
        {
            settingsDidLoad(configPath, settings);

            bool changedSettings {false};

            if
                (0 >= settings.customPath.length())
            {
                Path::joinPaths({configDir, "custom"}, settings.customPath);
                changedSettings = true;
            }

            std::string bundledColormapsPath = resourcePath + Path::separator() + "colormaps";
            copyMissingBundledResources(bundledColormapsPath, settings.colormapsPath);

            std::string bundledImagesPath = resourcePath + Path::separator() + "images";
            std::string userImagesPath = configDir + Path::separator() + "worldmaps";
            if
                (  true == Path::dirExists(userImagesPath)
                || true == Path::createDirs(userImagesPath, 0755))
            {
                std::array<std::string, 3> imageFileNames{{
                    "earth_highres.png",
                    "earth_lowres.png",
                    "black.png",
                }};

                for
                    (const auto& fileName : imageFileNames)
                {
                    std::string bundledImagePath = bundledImagesPath + Path::separator() + fileName;
                    std::string userImagePath = userImagesPath + Path::separator() + fileName;

                    if
                        (  false == Path::fileExists(bundledImagePath)
                        || true == Path::fileExists(userImagePath))
                    {
                        continue;
                    }

                    if
                        (true == copyItemAtPathTo(bundledImagePath, userImagePath))
                    {
                        std::cerr << "Copied bundled resource " << fileName << " to configured path: " << userImagesPath << std::endl;
                    }
                    else
                    {
                        std::cerr << "Cannot copy bundled resource " << fileName << " to configured path: " << userImagesPath << std::endl;
                    }
                }

                if
                    ("highres" == settings.worldMap)
                {
                    Path::joinPaths({userImagesPath, "earth_highres.png"}, settings.worldMap);
                    changedSettings = true;
                }
                else if
                    ("lowres" == settings.worldMap)
                {
                    Path::joinPaths({userImagesPath, "earth_lowres.png"}, settings.worldMap);
                    changedSettings = true;
                }
                else if
                    ("black" == settings.worldMap)
                {
                    Path::joinPaths({userImagesPath, "black.png"}, settings.worldMap);
                    changedSettings = true;
                }
            }
            else
            {
                std::cerr << "Could not create world map images directory at: " << userImagesPath << std::endl;
            }

            if
                (true == changedSettings)
            {
                settingsStore.save(settings, configPath);
            }
        }
        else
        {
            std::cerr << "Cannot parse config to determine the paths." << std::endl;
        }
    }
    else
    {
        std::cerr << "Cannot get bundle resources directory to copy to configured path." << std::endl;
    }

    if
        (false == hasListen)
    {
        listenURL = "127.0.0.1:11155";
    }

    if
        (false == hasLogFilters)
    {
        logFilters = "info";
    }

    return true;
}
