// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <AppKit/AppKit.h>
#include "platform/dialog_impl.h"

#include "core/path.h"

using namespace SEAScope;

/**
 *
 */
DialogImpl::DialogImpl()
{
}

/**
 *
 */
DialogImpl::~DialogImpl()
{
}

bool DialogImpl::available()
{
    return true;
}

/**
 *
 */
bool DialogImpl::askYesNoQuestion(const std::string& prompt, const std::string& yesLabel, const std::string& noLabel)
{
    CFStringRef message = CFStringCreateWithCString(
                                                    nullptr,
                                                    prompt.c_str(),
                                                    kCFStringEncodingUTF8);
    CFStringRef defaultButtonTitle = CFStringCreateWithCString(
                                                    nullptr,
                                                    yesLabel.c_str(),
                                                    kCFStringEncodingUTF8);
    CFStringRef alternateButtonTitle = CFStringCreateWithCString(
                                                    nullptr,
                                                    noLabel.c_str(),
                                                    kCFStringEncodingUTF8);
    CFOptionFlags answer;
    CFUserNotificationDisplayAlert(
                                   0,
                                   kCFUserNotificationNoteAlertLevel,
                                   nullptr, nullptr, nullptr,
                                   CFSTR("SEAScope"),
                                   message,
                                   defaultButtonTitle,
                                   alternateButtonTitle,
                                   nullptr,
                                   &answer);
    CFRelease(message);
    return answer == kCFUserNotificationDefaultResponse;
}

std::string DialogImpl::fileDialog(const std::string& currentPath, bool save)
{
    @autoreleasepool
    {
        NSURL *currentURL = [NSURL fileURLWithFileSystemRepresentation:currentPath.c_str()
                                                           isDirectory:NO
                                                         relativeToURL:nil];

        if
            (false == save)
        {
            NSOpenPanel *panel = [NSOpenPanel openPanel];
            panel.title = @"Choose File";
            panel.prompt = @"Choose";
            panel.message = @"Choose file to open";
            panel.directoryURL = currentURL;
            panel.canCreateDirectories = YES;
            panel.canSelectHiddenExtension = YES;
            panel.treatsFilePackagesAsDirectories = NO;
            panel.canChooseFiles = YES;
            panel.canChooseDirectories = NO;
            panel.resolvesAliases = YES;
            panel.allowsMultipleSelection = NO;
            if
                (@available(macOS 10.10, *))
            {
                panel.canDownloadUbiquitousContents = YES;
                panel.canResolveUbiquitousConflicts = YES;
            }

            NSModalResponse result = [panel runModal];
            if
                (NSModalResponseOK == result && YES == panel.URL.fileURL)
            {
                std::string path;
                path.assign(panel.URL.fileSystemRepresentation);
                return path;
            }
            return "";
        }
        else
        {
            NSSavePanel *panel = [NSSavePanel savePanel];
            panel.title = @"Choose Save Location";
            panel.prompt = @"Choose";
            panel.message = @"Choose where to save";
            panel.directoryURL = currentURL;
            panel.canCreateDirectories = YES;
            panel.canSelectHiddenExtension = YES;
            panel.treatsFilePackagesAsDirectories = NO;

            BOOL isDirectory = (
                currentPath.rfind(Path::separator()) == currentPath.size() - 1
            );
            if
                (NO == isDirectory)
            {
                NSFileManager *fileManager = [NSFileManager defaultManager];
                [fileManager fileExistsAtPath:static_cast<NSString *>(currentURL.path)
                                  isDirectory:&isDirectory];
            }
            if
                (NO == isDirectory)
            {
                NSString *fileName = currentURL.lastPathComponent;
                if
                    (nil != fileName)
                {
                    panel.nameFieldStringValue = fileName;
                }
            }

            NSModalResponse result = [panel runModal];
            if
                (NSModalResponseOK == result && YES == panel.URL.fileURL)
            {
                std::string path;
                path.assign(panel.URL.fileSystemRepresentation);
                return path;
            }
            return "";
        }
    }
}

std::string DialogImpl::dirDialog(const std::string& currentPath)
{
    @autoreleasepool
    {
        NSURL *currentURL = [NSURL fileURLWithFileSystemRepresentation:currentPath.c_str()
                                                           isDirectory:NO
                                                         relativeToURL:nil];

        NSOpenPanel *panel = [NSOpenPanel openPanel];
        panel.title = @"Choose Folder";
        panel.prompt = @"Choose";
        panel.message = @"Choose folder";
        panel.directoryURL = currentURL;
        panel.canCreateDirectories = YES;
        panel.canSelectHiddenExtension = YES;
        panel.treatsFilePackagesAsDirectories = NO;
        panel.canChooseFiles = NO;
        panel.canChooseDirectories = YES;
        panel.resolvesAliases = YES;
        panel.allowsMultipleSelection = NO;
        if
            (@available(macOS 10.10, *))
        {
            panel.canDownloadUbiquitousContents = YES;
            panel.canResolveUbiquitousConflicts = YES;
        }

        NSModalResponse result = [panel runModal];
        if
            (NSModalResponseOK == result && YES == panel.URL.fileURL)
        {
            std::string path;
            path.assign(panel.URL.fileSystemRepresentation);
            return path;
        }
        return "";
    }
}
