// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <Foundation/Foundation.h>
#include <Cocoa/Cocoa.h>

#include "platform/nsapplication_delegate.h"
#include "core/log.h"

using namespace SEAScope;

static NSURL *_configURL = nil;
static NSURL *_configDataURL = nil;
static NSURL *_configCustomURL = nil;

std::string objToString(NSObject *);
void openApplicationAtURL(NSURL *, NSDictionary<NSString *, NSString *> * __nullable, void (^ __nullable)(NSRunningApplication *, NSError *));
void openURLsWithApplicationAtURL(NSArray<NSURL *> *, NSURL *, NSDictionary<NSString *, NSString *> * __nullable, void (^ __nullable)(NSRunningApplication *, NSError *));
void launchImportDroplet(NSArray<NSURL *> * __nullable, void (^ __nullable)(NSRunningApplication *, NSError *));

void SEAScope::settingsDidLoad(std::string &path, Settings &settings)
{
    if
        (nil != _configURL)
    {
        [_configURL release];
    }
    _configURL = (
        [[NSURL alloc] initFileURLWithFileSystemRepresentation:path.c_str()
                                                   isDirectory:NO
                                                 relativeToURL:nil]
    );

    if
        (nil != _configDataURL)
    {
        [_configDataURL release];
    }
    _configDataURL = (
        [[NSURL alloc] initFileURLWithFileSystemRepresentation:settings.dataPath.c_str()
                                                   isDirectory:YES
                                                 relativeToURL:nil]
    );

    if
        (nil != _configCustomURL)
    {
        [_configCustomURL release];
    }
    _configCustomURL = (
        [[NSURL alloc] initFileURLWithFileSystemRepresentation:settings.customPath.c_str()
                                                   isDirectory:YES
                                                 relativeToURL:nil]
    );
}

std::string objToString(NSObject *obj)
{
    if
        (nil == obj)
    {
        return "nil";
    }
    NSString *description;
    if
        ([obj isKindOfClass:NSString.class])
    {
        description = static_cast<NSString *>(obj);
    }
    else
    {
        description = obj.description;
    }
    return std::string([description UTF8String]);
}

void openApplicationAtURL(NSURL *url, NSDictionary<NSString *, NSString *> *env, void (^completionHandler)(NSRunningApplication *, NSError *))
{
    if
        (nil == url)
    {
        NSError *error = [NSError errorWithDomain:@"SEAScope"
                                             code:-10
                                         userInfo:@{
            NSLocalizedDescriptionKey: @"Application not found"
        }];
        completionHandler(nil, error);
        return;
    }

    if
        (@available(macOS 11, *))
    {
        NSWorkspaceOpenConfiguration *config = [NSWorkspaceOpenConfiguration configuration];
        config.allowsRunningApplicationSubstitution = NO;
        config.createsNewApplicationInstance = YES;
        if
            (nil != env)
        {
            config.environment = static_cast<NSDictionary<NSString *, NSString *> *>(env);
        }

        [[NSWorkspace sharedWorkspace]
                      openURL:url
                configuration:config
            completionHandler:completionHandler];
    }
    else
    {
        #pragma clang diagnostic push
        #pragma clang diagnostic ignored "-Wdeprecated-declarations"
        NSError *error = nil;
        NSDictionary *config;
        if
            (nil == env)
        {
            config = @{};
        }
        else
        {
            config = @{
                NSWorkspaceLaunchConfigurationEnvironment: static_cast<NSDictionary<NSString *, NSString *> *>(env),
            };
        }

        NSRunningApplication *app = [[NSWorkspace sharedWorkspace]
                  openURL:url
                  options:NSWorkspaceLaunchNewInstance
            configuration:config
                    error:&error
        ];
        completionHandler(app, error);
        #pragma clang diagnostic pop
    }
}

void openURLsWithApplicationAtURL(NSArray<NSURL *> *urls, NSURL *appURL, NSDictionary<NSString *, NSString *> *env, void (^completionHandler)(NSRunningApplication *, NSError *))
{
    if
        (nil == appURL)
    {
        NSError *error = [NSError errorWithDomain:@"SEAScope"
                                             code:-10
                                         userInfo:@{
            NSLocalizedDescriptionKey: @"Application not found"
        }];
        completionHandler(nil, error);
        return;
    }

    if
        (@available(macOS 11, *))
    {
        NSWorkspaceOpenConfiguration *config = [NSWorkspaceOpenConfiguration configuration];
        config.allowsRunningApplicationSubstitution = NO;
        config.createsNewApplicationInstance = YES;
        if
            (nil != env)
        {
            config.environment = static_cast<NSDictionary<NSString *, NSString *> *>(env);
        }

        [[NSWorkspace sharedWorkspace]
                        openURLs:urls
            withApplicationAtURL:appURL
                   configuration:config
               completionHandler:completionHandler];
    }
    else
    {
        #pragma clang diagnostic push
        #pragma clang diagnostic ignored "-Wdeprecated-declarations"
        NSError *error = nil;
        NSDictionary *config;
        if
            (nil == env)
        {
            config = @{};
        }
        else
        {
            config = @{
                NSWorkspaceLaunchConfigurationEnvironment: static_cast<NSDictionary<NSString *, NSString *> *>(env),
            };
        }

        NSRunningApplication *app = [[NSWorkspace sharedWorkspace]
                        openURLs:urls
            withApplicationAtURL:appURL
                         options:NSWorkspaceLaunchNewInstance
                   configuration:config
                           error:&error
        ];
        completionHandler(app, error);
        #pragma clang diagnostic pop
    }
}

void launchImportDroplet(NSArray<NSURL *> *urls, void (^completionHandler)(NSRunningApplication *, NSError *))
{
    NSURL *url = [[NSBundle mainBundle] URLForAuxiliaryExecutable:@"Import to SEAScope.app"];
    NSDictionary *env = @{
        @"SEASCOPE_APP_PATH": [NSBundle mainBundle].bundlePath,
        @"SEASCOPE_DATA_PATH": (_configDataURL == nil ? @"" : _configDataURL.path),
    };

    if
        (nil == urls)
    {
        openApplicationAtURL(url, env, completionHandler);
    }
    else
    {
        openURLsWithApplicationAtURL(urls, url, env, completionHandler);
    }
}

@interface SEAScopeKVOObserver : NSObject

+ (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(NSObject *)object options:(NSKeyValueObservingOptions)options handler:(BOOL (^)(NSDictionary *))handler;

@end

@implementation SEAScopeKVOObserver

+ (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(NSObject *)object options:(NSKeyValueObservingOptions)options handler:(BOOL (^)(NSDictionary *))handler
{
    [object retain];
    [object addObserver:self.self forKeyPath:keyPath options:options context:[handler copy]];
}

+ (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(NSObject *)object change:(NSDictionary *)change context:(void *)context
{
    BOOL (^handler)(NSDictionary *) = static_cast<id>(context);
    BOOL done = handler(change);
    if
        (YES == done)
    {
        [object removeObserver:self.self forKeyPath:keyPath context:context];
        [object release];
        [handler release];
    }
}

@end

@interface GLFWApplicationDelegate : NSObject <NSApplicationDelegate>
@end

@implementation GLFWApplicationDelegate (SEAScopeExtension)

- (void)seascopeShowConfigFile:(NSMenuItem *)sender
{
    if
        (nil == _configURL)
    {
        return;
    }
    Log::log(
        LogLevel::Debug,
        "NSApplicationDelegate",
        "seascopeShowConfigFile: " + objToString(_configURL)
    );
    [[NSWorkspace sharedWorkspace] activateFileViewerSelectingURLs:@[_configURL]];
}

- (void)seascopeOpenDataDir:(NSMenuItem *)sender
{
    if
        (nil == _configDataURL)
    {
        return;
    }
    Log::log(
        LogLevel::Debug,
        "NSApplicationDelegate",
        "seascopeOpenDataDir: " + objToString(_configDataURL)
    );
    [[NSWorkspace sharedWorkspace] openURL:_configDataURL];
}

- (void)seascopeOpenCustomDir:(NSMenuItem *)sender
{
    if
        (nil == _configCustomURL)
    {
        return;
    }
    Log::log(
        LogLevel::Debug,
        "NSApplicationDelegate",
        "seascopeOpenCustomDir: " + objToString(_configCustomURL)
    );
    [[NSWorkspace sharedWorkspace] openURL:_configCustomURL];
}

- (void)seascopeOpenImportPreferences:(NSMenuItem *)sender
{
    launchImportDroplet(nil, ^(NSRunningApplication *app, NSError *error) {
        if
            (nil == app || nil != error)
        {
            Log::log(
                LogLevel::Debug,
                "NSApplicationDelegate",
                "Cannot launch import droplet: " + objToString(error)
            );
        }
    });
}

- (NSMenu *)applicationDockMenu:(NSApplication *)sender
{
    NSMenu *menu = [[NSMenu alloc] init];
    if
        (nil != [[NSBundle mainBundle] URLForAuxiliaryExecutable:@"Import to SEAScope.app"])
    {
        [menu addItemWithTitle:@"Data Import Preferences"
                        action:@selector(seascopeOpenImportPreferences:)
                 keyEquivalent:@""].target = self;
    }
    if
        (nil != _configURL)
    {
        [menu addItemWithTitle:@"Show Config File"
                        action:@selector(seascopeShowConfigFile:)
                 keyEquivalent:@""].target = self;
    }
    if
        (nil != _configDataURL)
    {
        [menu addItemWithTitle:@"Open Data Folder"
                        action:@selector(seascopeOpenDataDir:)
                 keyEquivalent:@""].target = self;
    }
    if
        (nil != _configCustomURL)
    {
        [menu addItemWithTitle:@"Open Customization Folder"
                        action:@selector(seascopeOpenCustomDir:)
                 keyEquivalent:@""].target = self;
    }
    if
        (0 == menu.numberOfItems)
    {
        [menu release];
        return nil;
    }
    return [menu autorelease];
}

- (void)application:(NSApplication *)sender openFiles:(NSArray<NSString *> *)filenames
{
    Log::log(
        LogLevel::Debug,
        "NSApplicationDelegate",
        "-application:openFiles: " + objToString(filenames)
    );

    NSMutableArray *urls = [NSMutableArray new];
    for
        (NSString *filename in filenames)
    {
        [urls addObject:[NSURL fileURLWithPath:filename]];
    }

    launchImportDroplet(urls, ^(NSRunningApplication *app, NSError *error) {
        if
            (nil == app || nil != error)
        {
            Log::log(
                LogLevel::Debug,
                "NSApplicationDelegate",
                "Cannot launch import droplet: " + objToString(error)
            );
            [sender replyToOpenOrPrint:NSApplicationDelegateReplyFailure];

            CFUserNotificationDisplayNotice(
                0,
                kCFUserNotificationNoteAlertLevel,
                nullptr,
                nullptr,
                nullptr,
                CFSTR("Import Failed"),
                CFSTR("Cannot launch import droplet."),
                nullptr
            );
            return;
        }

        [app activateWithOptions:( NSApplicationActivateAllWindows
                                 | NSApplicationActivateIgnoringOtherApps)];

        [SEAScopeKVOObserver
            observeValueForKeyPath:@"terminated"
                        ofObject:app
                        options:(NSKeyValueObservingOptionNew
                                | NSKeyValueObservingOptionInitial)
                        handler:^(NSDictionary *changes) {
            NSNumber *newValue = changes[NSKeyValueChangeNewKey];
            BOOL terminated = newValue.boolValue;
            if
                (YES == terminated)
            {
                [sender replyToOpenOrPrint:NSApplicationDelegateReplySuccess];
            }
            return terminated;
        }];
    });
}

@end
