// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <string>
#include <sstream>
#include <vector>
#include <mutex>
#include <array>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/event.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>

#include "platform/server.h"
#include "core/perfs.h"
#include "core/control_protocol_handler.h"
#include <core/perfs.h>
#include <core/sequential_buffer.h>
#include "core/sockets_manager.h"

#include "core/log.h"

class SocketData
{
public:
    struct kevent event;
    int socket;
    SEAScope::IProtocolHandler* protocol;
    SEAScope::SequentialBuffer<SOCKET_CHUNKSIZE> buffer;

    SocketData()
    {
        socket = 0;
        protocol = nullptr;
    }

    ~SocketData()
    {
    }
};

Server::Server()
:   _listener(nullptr)
{
}

Server::~Server()
{
    delete _listener;
    _listener = nullptr;
}

void Server::initialize(const std::string& address, std::int64_t port, SEAScope::SocketsManager* socketManager)
{
	_address = address;
	_port = port;
	_socketManager = socketManager;

    if (nullptr == _listener)
        _listener = new SocketData;

    _kq = -1;
}

bool Server::run()
{
	{
		std::unique_lock<std::mutex> lock(_mutex);
		_stopThread = false;
	}

	struct sockaddr a;	//
	struct sockaddr_in b;	//
	std::ostringstream o;	//
	socklen_t l;	//
	int rc;	//

	memset(&b, 0, sizeof(struct sockaddr_in));
	b.sin_family = AF_INET;
	b.sin_addr.s_addr = inet_addr(_address.c_str());
	b.sin_port = htons(_port);
	l = sizeof(b);
	//a = (struct sockaddr*)(&b);
	std::memcpy(&a, &b, l);

	_listener->socket = socket(AF_INET, SOCK_STREAM, 0);

	int flag {1};
	rc = setsockopt(_listener->socket, SOL_SOCKET, SO_REUSEPORT, &flag, sizeof(flag));
	if (0 != rc)
	{
        LOG(SEAScope::LogLevel::Error, "Server", "Failed to set the SO_REUSEPORT flag on the socket");
        return false;
	}

	rc = bind(_listener->socket, &a, l);
	if (0 != rc)
	{
        LOG(SEAScope::LogLevel::Error, "Server", "Bind failed");
        return false;
	}

    rc = listen(_listener->socket, SOMAXCONN);
    if (0 != rc)
    {
        LOG(SEAScope::LogLevel::Error, "Server", "Listen failed");
        return false;
    }

    _kq = kqueue();

    EV_SET(&_listener->event, _listener->socket, EVFILT_READ, EV_ADD, 0, 0, nullptr);

    if (-1 == kevent(_kq, &_listener->event, 1, nullptr, 0, nullptr))
    {
        LOG(SEAScope::LogLevel::Error, "Server", "kevent failed");
        return false;
    }

    bool done = false;
    const unsigned int evCount = 32;
    struct kevent evList[evCount];
    int nev;
    struct timespec timeout;

    // Set timeout to 100ms
    timeout.tv_sec = 0;
    timeout.tv_nsec = 100000000; // 100 ms

    while(false == done)
    {
        nev = kevent(_kq, nullptr, 0, evList, evCount, &timeout);

        if (0 == nev)
        {
            std::unique_lock<std::mutex> lock(_mutex);

            if (true == _stopThread)
            {
                done = true;
            }
        }
        else
        {
            for (int i = 0; i < nev; ++i)
            {
                if (evList[i].flags & EV_EOF) // Close event
                {
                    SocketData* obj = static_cast<SocketData*>(evList[i].udata);

                    LOG(SEAScope::LogLevel::Debug, "Server"
                                      , "Closing connection after receiving EOF");
                    closeSocket(*obj);

                    delete obj;
                }
                else if (static_cast<int>(evList[i].ident) == _listener->socket) // Accept event
                {
                    acceptConnection(*_listener);
                }
                else if (evList[i].filter == EVFILT_READ) // Read event
                {
                    SocketData* obj = static_cast<SocketData*>(evList[i].udata);

                    if (false == readSocket(*obj))
                    {
                        LOG(SEAScope::LogLevel::Debug, "Server"
                                          , "Closing connection after receiving messages");
                        closeSocket(*obj);

                        delete obj;
                    }
                }
            }
        }
    }

    rc = close(_kq);

    if (-1 == rc)
    {
        LOG(SEAScope::LogLevel::Error, "Server", "Close kqueue failed");
    }

    _kq = -1;

	return true;
}

void Server::stop()
{
	std::unique_lock<std::mutex> lock(_mutex);
	_stopThread = true;
}

bool Server::acceptConnection(SocketData& s)
{
	SocketData* newClient = new SocketData;
	newClient->socket = accept(s.socket, nullptr, nullptr);

    struct kevent evSet;

    EV_SET(&evSet, newClient->socket, EVFILT_READ, EV_ADD, 0, 0, newClient);

    if (-1 == kevent(_kq, &evSet, 1, nullptr, 0, nullptr))
    {
        LOG(SEAScope::LogLevel::Error, "Server", "kevent failed");
        return false;
    }

	bool ok;
	newClient->protocol = _socketManager->getProtocolHandler("SEASCOPE_CONTROL", ok);

	return true;
}

bool Server::readSocket(SocketData& s)
{
	std::uint64_t left;
	std::uint8_t* res = s.buffer.getPointer(left);
	std::uint64_t toRead = (SOCKET_CHUNKSIZE > left) ? left : SOCKET_CHUNKSIZE;
	std::int64_t loaded = ::recv(s.socket, static_cast<unsigned char*>(res), toRead, 0);

	std::uint64_t consumed;
	bool msgComplete;

	while
		(0 < loaded)
	{
		std::uint64_t cumulatedRead;	//
		cumulatedRead = 0;
		msgComplete = true;
		while
			(cumulatedRead < static_cast<std::uint64_t>(loaded))
		{
			msgComplete = false;
			msgComplete = s.protocol->parseStream(s.socket, res, cumulatedRead, static_cast<std::uint64_t>(loaded) - cumulatedRead, consumed);
			cumulatedRead += consumed;
		}

		left = 0;
		res = s.buffer.getPointer(left);
		toRead = (SOCKET_CHUNKSIZE > left) ? left : SOCKET_CHUNKSIZE;
		loaded = recv(s.socket, static_cast<unsigned char*>(res), toRead, 0);

		if
			((true == msgComplete) && (0 == loaded))
		{
			return false;
		}
	}

	return true;
}

bool Server::closeSocket(SocketData& s)
{
    int res = ::close(s.socket);
	if (0 != res)
	{
        LOG(SEAScope::LogLevel::Error, "Server", "Close failed");
        return false;
	}

	return true;
}
