// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#import <Foundation/Foundation.h>
#import <Cocoa/Cocoa.h>
#import <objc/runtime.h>

#import "platform/imprecise_scroll_fix.h"

/// A workaround for the problem described in GLFW's issue #2219 on GitHub.
/// @link: https://github.com/glfw/glfw/issues/2219

#if __clang_major__ > 13
    #pragma clang diagnostic ignored "-Wobjc-load-method"
#endif

@interface GLFWWindowDelegate : NSObject
@end

@interface GLFWWindowDelegate (WorkaroundIssue2219)
@property (nonatomic) BOOL _wi2219_windowWasFocused;
@end

@implementation GLFWWindowDelegate (WorkaroundIssue2219)

+ (void)load {
    if (self == GLFWWindowDelegate.self) {
        static dispatch_once_t onceToken;
        dispatch_once(&onceToken, ^{
            Class cls = [self class];

            SEL swizzlePairs[] = {
                @selector(windowDidBecomeKey:), @selector(swizzle_windowDidBecomeKey:),
                @selector(windowDidResignKey:), @selector(swizzle_windowDidResignKey:),
            };
            for
                (uint64_t i = 0; i < sizeof(swizzlePairs) / sizeof(SEL); i += 2)
            {
                Method originalMethod = class_getInstanceMethod(cls, swizzlePairs[i]);
                Method swizzledMethod = class_getInstanceMethod(cls, swizzlePairs[i + 1]);

                method_exchangeImplementations(originalMethod, swizzledMethod);
            }
        });
    }
}

static char GLFWContentView_WorkaroundIssue2219_windowWasFocused_key;

@dynamic _wi2219_windowWasFocused;

-(void)set_wi2219_windowWasFocused:(BOOL)windowWasFocused
{
  objc_setAssociatedObject(self,
                           &GLFWContentView_WorkaroundIssue2219_windowWasFocused_key,
                           windowWasFocused ? @"" : nil,
                           OBJC_ASSOCIATION_RETAIN_NONATOMIC);
}

-(BOOL)_wi2219_windowWasFocused
{
   return nil != objc_getAssociatedObject(self, &GLFWContentView_WorkaroundIssue2219_windowWasFocused_key);
}

- (void)swizzle_windowDidBecomeKey:(NSNotification *)notification
{
    self._wi2219_windowWasFocused = NO;
    [self swizzle_windowDidBecomeKey:notification];
}

- (void)swizzle_windowDidResignKey:(NSNotification *)notification
{
    self._wi2219_windowWasFocused = NO;
    [self swizzle_windowDidResignKey:notification];
}

@end

@interface GLFWContentView : NSView <NSTextInputClient>
@end

@implementation GLFWContentView (WorkaroundIssue2219)

+ (void)load {
    if (self == GLFWContentView.self) {
        static dispatch_once_t onceToken;
        dispatch_once(&onceToken, ^{
            Class cls = [self class];

            SEL swizzlePairs[] = {
                @selector(mouseDown:), @selector(swizzle_mouseDown:),
                @selector(mouseUp:), @selector(swizzle_mouseUp:),
                @selector(rightMouseDown:), @selector(swizzle_rightMouseDown:),
                @selector(rightMouseUp:), @selector(swizzle_rightMouseUp:),
                @selector(otherMouseDown:), @selector(swizzle_otherMouseDown:),
                @selector(otherMouseUp:), @selector(swizzle_otherMouseUp:),
                @selector(scrollWheel:), @selector(swizzle_scrollWheel:),
            };
            for
                (uint64_t i = 0; i < sizeof(swizzlePairs) / sizeof(SEL); i += 2)
            {
                Method originalMethod = class_getInstanceMethod(cls, swizzlePairs[i]);
                Method swizzledMethod = class_getInstanceMethod(cls, swizzlePairs[i + 1]);

                method_exchangeImplementations(originalMethod, swizzledMethod);
            }
        });
    }
}

- (void)_workaroundIssue2219_callMouseMovedIfNeeded:(NSEvent *)event
{
    if
        (self.window.keyWindow)
    {
        GLFWWindowDelegate *glfwWindowDelegate = static_cast<GLFWWindowDelegate *>(self.window.delegate);
        if
            (glfwWindowDelegate._wi2219_windowWasFocused)
        {
            return;
        }
        glfwWindowDelegate._wi2219_windowWasFocused = YES;
    }

    [self mouseMoved:event];
}

- (void)swizzle_mouseDown:(NSEvent *)event
{
    [self _workaroundIssue2219_callMouseMovedIfNeeded:event];
    [self swizzle_mouseDown:event];
}

- (void)swizzle_mouseUp:(NSEvent *)event
{
    [self _workaroundIssue2219_callMouseMovedIfNeeded:event];
    [self swizzle_mouseUp:event];
}

- (void)swizzle_rightMouseDown:(NSEvent *)event
{
    [self _workaroundIssue2219_callMouseMovedIfNeeded:event];
    [self swizzle_rightMouseDown:event];
}

- (void)swizzle_rightMouseUp:(NSEvent *)event
{
    [self _workaroundIssue2219_callMouseMovedIfNeeded:event];
    [self swizzle_rightMouseUp:event];
}

- (void)swizzle_otherMouseDown:(NSEvent *)event
{
    [self _workaroundIssue2219_callMouseMovedIfNeeded:event];
    [self swizzle_otherMouseDown:event];
}

- (void)swizzle_otherMouseUp:(NSEvent *)event
{
    [self _workaroundIssue2219_callMouseMovedIfNeeded:event];
    [self swizzle_otherMouseUp:event];
}

- (void)swizzle_scrollWheel:(NSEvent *)event
{
    [self _workaroundIssue2219_callMouseMovedIfNeeded:event];

    // Fix imprecise scrolling deltas
    NSEvent *eventProxy = static_cast<NSEvent *>([[FixScrollNSEventProxy alloc] initWithEvent:event]);
    [self swizzle_scrollWheel:eventProxy];
    [eventProxy release];
}

@end
