// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   buffer.cpp
 * Author: gleseach
 * 
 * Created on 3 septembre 2015, 11:07
 */

#include "3d/stdafx.h"

#include "3d/OGL3/opengl.h"

#include "3d/OGL3/bufferGL3.h"
#include "3d/OGL3/opengl_error.h"
#include "3d/OGL3/enum_translation.h"
#include "3d/memory_manager.h"
#include "3d/array.h"

#include "core/log.h"

namespace OGL3
{

BufferGL3::BufferGL3(OGL::BufferType type)
: OGL::Buffer(type)
{
}

BufferGL3::~BufferGL3()
{
    destroy();
}

void BufferGL3::create(OGL::Context* /*context*/)
{
    GL_CHECK(glGenBuffers(1, &this->handle));
}

void BufferGL3::destroy()
{
    if(0 != this->handle)
    {
        GL_CHECK(glDeleteBuffers(1, &this->handle));
        this->handle = 0;
        MemoryManager::getInstance()->release(buffer, _size);
    }
}

void BufferGL3::bind() const
{
    GLenum gl_type = bufferTargetEnumToGLenum(this->_type);
    GL_CHECK(glBindBuffer(gl_type, this->handle));
}

void BufferGL3::unbind() const
{
    GLenum gl_type = bufferTargetEnumToGLenum(_type);
    GL_CHECK(glBindBuffer(gl_type, 0));
}

void BufferGL3::allocate(const Array& array)
{
    allocate(array.getDescription().getArraySize(), array.getDataPtr());
}

void BufferGL3::allocate(unsigned int size, void * data, OGL::BufferUsage usage)
{
    if (0 == this->handle)
    {
        GL_CHECK(glGenBuffers(1, &this->handle));
    }

    if (0 != _size)
    {
        MemoryManager::getInstance()->release(buffer, _size);
    }

    _size = size;

    bind();

    GLenum gl_type = bufferTargetEnumToGLenum(_type);
    GLenum gl_usage = bufferUsageEnumToGLenum(usage);
    
    GL_CHECK(glBufferData(gl_type, size, data, gl_usage));

    unbind();
    
    MemoryManager::getInstance()->consume(buffer, _size);
}

void BufferGL3::copy(void * data)
{
    bind();

    GLenum gl_type = bufferTargetEnumToGLenum(_type);
    GL_CHECK(glBufferSubData(gl_type, 0, _size, data));
}

void BufferGL3::bindTo(int index)
{
    glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, static_cast<GLuint>(index), this->handle);
}

void BufferGL3::copyTo(OGL::Buffer* buffer, unsigned int readOffset, unsigned int writeOffset, unsigned int size)
{
    // Bind source buffer
    bind();
    
    // Bind target buffer
    buffer->bind();
    
    GLenum sourceGLType = bufferTargetEnumToGLenum(this->_type);
    GLenum targetGLType = bufferTargetEnumToGLenum(buffer->getBufferType());

    // Copy
    glCopyBufferSubData(sourceGLType, targetGLType, readOffset, writeOffset, size);
    
    buffer->unbind();
    unbind();
}

void BufferGL3::copy(unsigned int offset, unsigned int size, void* data)
{
    if (offset + size > _size)
    {
		LOG(SEAScope::LogLevel::Warn, "BufferGL3", "Copy canceled, write operation will cause a buffer overflow");
        return;
    }
    
    // Bind source buffer
    bind();
    
    GLenum gl_type = bufferTargetEnumToGLenum(this->_type);

    // Copy
    glBufferSubData(gl_type, offset, size, data);
    
    unbind();
}

void* BufferGL3::map(OGL::BufferAccess access)
{
    GLenum bufferTarget = bufferTargetEnumToGLenum(_type);
    GLenum glEnum = bufferAccessEnumToGLenum(access);
    
    return glMapBuffer(bufferTarget, glEnum);
}

void BufferGL3::unmap()
{
    GLenum bufferTarget = bufferTargetEnumToGLenum(_type);

    glUnmapBuffer(bufferTarget);
}

} // namespace OGL3

