// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include "3d/stdafx.h"
#include "3d/OGL3/opengl.h"

#include "3d/OGL3/contextGL3.h"

#include "3d/OGL3/bufferGL3.h"
#include "3d/OGL3/textureGL3.h"
#include "3d/OGL3/framebufferGL3.h"
#include "3d/OGL3/programGL3.h"
#include "3d/OGL3/renderbufferGL3.h"
#include "3d/OGL3/shaderGL3.h"
#include "3d/OGL3/vertex_array_objectGL3.h"
#include "3d/OGL3/vertex_attrib_arrayGL3.h"
#include "3d/OGL3/draw_modeGL3.h"
#include "3d/OGL3/utilsGL3.h"
#include "3d/OGL3/opengl_error.h"
#include "3d/OGL3/queryGL3.h"
#include "3d/OGL3/syncGL3.h"
#include "3d/OGL3/shader_list.h"
#include "3d/OGL3/vertex_attrib_divisorGL3.h"

#include "3d/program_bind_helper.h"
#include "3d/typed_buffer.h"
#include "core/path.h"
#include "3d/opengl_enum.h"
#include "core/log.h"

#include "3d/shader_manager.h"


namespace
{
    GLenum faceToGLEnum(OGL::Face face)
    {
        GLenum glFace = GL_FRONT_AND_BACK;
        
        switch(face)
        {
        case OGL::front:
            glFace = GL_FRONT;
            break;
        case OGL::back:
            glFace = GL_BACK;
            break;
        case OGL::front_and_back:
            glFace = GL_FRONT_AND_BACK;
            break;
        }
        
        return glFace;
    }
    
    GLenum modeToGLEnum(OGL::RasterizationMode mode)
    {
        GLenum glMode = GL_FILL;
        
        switch(mode)
        {
        case OGL::fill:
            glMode = GL_FILL;
            break;
        case OGL::line:
            glMode = GL_LINE;
            break;
        case OGL::point:
            glMode = GL_POINT;
            break;
        }
        
        return glMode;
    }
    
    GLenum blendingFactorToGLEnum(OGL::BlendingFactor factor)
    {
        GLenum val = GL_ONE_MINUS_SRC_ALPHA;
        
        switch(factor)
        {
        case OGL::one:
            val = GL_ONE;
            break;
        case OGL::one_minus_src_color:
            val = GL_ONE_MINUS_SRC_COLOR;
            break;
        case OGL::src_alpha:
            val = GL_SRC_ALPHA;
            break;
        case OGL::one_minus_src_alpha:
            val = GL_ONE_MINUS_SRC_ALPHA;
            break;
        }
        
        return val;
    }
    
    GLenum funcToGLEnum(OGL::Func func)
    {
        GLenum val = GL_ALWAYS;
        
        switch(func)
        {
        case OGL::always:
            val = GL_ALWAYS;
            break;
        case OGL::less:
            val = GL_LESS;
            break;
        case OGL::less_or_equal:
            val = GL_LEQUAL;
            break;
        case OGL::equal:
            val = GL_EQUAL;
            break;
        case OGL::not_equal:
            val = GL_NOTEQUAL;
            break;
        case OGL::greater:
            val = GL_GREATER;
            break;
        case OGL::greater_or_equal:
            val = GL_GEQUAL;
            break;
        case OGL::never:
            val = GL_NEVER;
            break;
        }
        
        return val;
    }
    
    GLenum stencilOpToGLEnum(OGL::StencilOp op)
    {
        GLenum val = GL_REPLACE;
        
        switch(op)
        {
        case OGL::zero:
            val = GL_ZERO;
            break;
        case OGL::decr:
            val = GL_DECR;
            break;
        case OGL::decr_wrap:
            val = GL_DECR_WRAP;
            break;
        case OGL::incr:
            val = GL_INCR;
            break;
        case OGL::incr_wrap:
            val = GL_INCR_WRAP;
            break;
        case OGL::invert:
            val = GL_INVERT;
            break;
        case OGL::keep:
            val = GL_KEEP;
            break;
        case OGL::replace:
            val = GL_REPLACE;
            break;
        }
        
        return val;
    }
}

namespace OGL3
{

ContextGL3::ContextGL3()
:   OGL::Context()
{
    LOG(SEAScope::LogLevel::Debug, "OGL3::Context" , "Create OpenGL 3.3 renderer.");
}
    
bool ContextGL3::init(const SEAScope::RendererResources* resources)
{
	this->_resources = resources;
    GLFWwindow* window = glfwGetCurrentContext();
    
//    _msaa = glfwGetWindowFrameSize(window, GLFW_SAMPLES);
    
    if (nullptr == window)
    {
        LOG(SEAScope::LogLevel::Error, "OGL3::Context" , "No context bound, initialization failed");
        return false;
    }
	
    GLADloadproc glfwLoader = [] (const char * value) -> void* {
        return reinterpret_cast<void*>(glfwGetProcAddress(value));
    };
	if (!gladLoadGLLoader(glfwLoader))
	{
    	LOG(SEAScope::LogLevel::Error, "OGL3::Context", "Failed to initialize GLAD");
    	return false;
	}
	
    int major, minor;
    
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "GL context info:");
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "\tVendor: {}", reinterpret_cast<const char*>(glGetString(GL_VENDOR)));
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "\tRenderer: {}", reinterpret_cast<const char*>(glGetString(GL_RENDERER)));
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "\tOpenGL: {}", reinterpret_cast<const char*>(glGetString(GL_VERSION)));
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "\tGLSL: {}", reinterpret_cast<const char*>(glGetString(GL_SHADING_LANGUAGE_VERSION)));

    int textureSizeMax = getGPUInfoInteger(OGL::max_texture_2d_size);
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "\tTexture max size: {}", textureSizeMax);

    glGetIntegerv(GL_MAJOR_VERSION, &major);
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "\tVersion major: {}", major);

    glGetIntegerv(GL_MINOR_VERSION, &minor);
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "\tVersion minor: {}", minor);

    float f[2];
    glGetFloatv(GL_ALIASED_LINE_WIDTH_RANGE, f);
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "\tAliased line width range: {}, {}", f[0], f[1]);

    glGetFloatv(GL_SMOOTH_LINE_WIDTH_RANGE, f);
    LOG(SEAScope::LogLevel::Info, "OGL3::Context", "\tSmooth line width range: {}, {}", f[0], f[1]);

    // Check if OpenGL version is at least 3.3
    if (3 > major || (3 == major && 3 > minor))
    {
        LOG(SEAScope::LogLevel::Error, "OGL3::Context" , "The version of current OpenGL context is too low, the minimal required version is OpenGL 3.3.");
        LOG(SEAScope::LogLevel::Error, "OGL3::Context" , "Please update your graphic card driver or use a graphic card that supports OpenGL 3.3.");
        return false;
    }
    
    // OpenGL Debug function, need a context > 4.3

    //glEnable(GL_DEBUG_OUTPUT);
    //glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);

    // Callback for opengl message, replace glGetError
    //glDebugMessageCallback(OGL3::DebugCallBack, NULL);

    glClearStencil(0);
    
    loadShaders();
    
    return true;
}

void ContextGL3::loadShaders()
{
    std::string shaderFolder = this->_resources->shadersDir + "OGL" + SEAScope::Path::separator();
    _shaderMngr->addShaderFromFiles(shaderFolder, OGL3::ShaderList::getShaderList());
}
    
OGL::Buffer* ContextGL3::createBuffer(OGL::BufferType type)
{
    return new OGL3::BufferGL3(type);
}

OGL::Texture* ContextGL3::createTexture(OGL::TextureType type)
{
    return new OGL3::TextureGL3(type);
}

OGL::FrameBuffer* ContextGL3::createFramebuffer()
{
    return new OGL3::FrameBufferGL3;
}
   
OGL::Program* ContextGL3::createProgram()
{
    return new OGL3::ProgramGL3;
}
    
OGL::RenderBuffer* ContextGL3::createRenderbuffer(OGL::RenderBufferType type)
{
    return new OGL3::RenderBufferGL3(type);
}

OGL::Shader* ContextGL3::createShader(OGL::ShaderType type, const std::string& name)
{
    return new OGL3::ShaderGL3(type, name);
}

OGL::VertexArrayObject* ContextGL3::createVertexArrayObject()
{
    return new OGL3::VertexArrayObjectGL3;
}

OGL::VertexAttribArray* ContextGL3::createVertexAttribArray(unsigned int id, unsigned int component_count, OGL::Buffer* buffer, DataType dataType, unsigned int stride, unsigned int offset)
{
    return new OGL3::VertexAttribArrayGL3(id, component_count, buffer, dataType, stride, offset);
}

OGL::VertexAttribArray* ContextGL3::createVertexAttribArray(unsigned int id, const TypedBuffer& typedBuffer, unsigned int stride, unsigned int offset)
{
    auto& desc = typedBuffer.getDescription();
    return new OGL3::VertexAttribArrayGL3(id, desc.getComponentCount(), typedBuffer.getBuffer(), desc.getDataType(), stride, offset);
}

OGL::DrawMode* ContextGL3::createDrawMode()
{
    return new OGL3::DrawModeGL3;
}

OGL::Utils2* ContextGL3::createUtils()
{
    return new OGL3::UtilsGL3;
}

OGL::ProgramBindHelper* ContextGL3::createProgramHelper()
{
    return new OGL::ProgramBindHelper;
}

OGL::Query* ContextGL3::createQuery()
{
    return new OGL3::Query;
}

OGL::Sync* ContextGL3::createSync()
{
    return new OGL3::Sync;
}

OGL::VertexAttribDivisor* ContextGL3::createVertexAttribDivisor(OGL::VertexAttribArray* vaa, unsigned int divisor)
{
    return new OGL3::VertexAttribDivisorGL3(vaa->getIndex(), divisor);
}

void ContextGL3::activeTextureUnit(unsigned int id)
{
    glActiveTexture(GL_TEXTURE0 + id);
}

void ContextGL3::activeTextureUnit(unsigned int id, OGL::Texture* object)
{
    activeTextureUnit(id);
    object->bind();
    
    _textureUnit[id].first = true;
    _textureUnit[id].second = object;
}

void ContextGL3::deactiveTextureUnit(unsigned int id)
{
    _textureUnit[id].second->unbind();   
    _textureUnit[id].first = false;
    _textureUnit[id].second = nullptr;
}

void ContextGL3::setPolygonMode(OGL::Face face, OGL::RasterizationMode mode)
{
    GLenum glFace = faceToGLEnum(face);
    GLenum glMode = modeToGLEnum(mode);
    
    glPolygonMode(glFace, glMode);
}

void ContextGL3::setFlag(OGL::Capability cap, bool status)
{
    GLenum glCap = GL_CULL_FACE;
    
    switch(cap)
    {
    case OGL::blending:
        glCap = GL_BLEND;
        break;
    case OGL::depth_test:
        glCap = GL_DEPTH_TEST;
        break;
    case OGL::cull_face:
        glCap = GL_CULL_FACE;
        break;
    case OGL::rasterizer_discard:
        glCap = GL_RASTERIZER_DISCARD;
        break;
    case OGL::stencil_test:
        glCap = GL_STENCIL_TEST;
        break;
    }
    
    if (status)
    {
        glEnable(glCap);
    }
    else
    {
        glDisable(glCap);
    }
}
  
void ContextGL3::setViewport(const glm::ivec4& size)
{
    glViewport(size.x, size.y, size.z, size.w);
}

void ContextGL3::setClearColor(const glm::vec4& color)
{
    glClearColor(color.r, color.g, color.b, color.a);
}

void ContextGL3::clear(int bits)
{
    GLenum glBits = 0;
    
    if (OGL::color_buffer & bits)
    {
        glBits |= GL_COLOR_BUFFER_BIT;
    }
    
    if (OGL::depth_buffer & bits)
    {
        glBits |= GL_DEPTH_BUFFER_BIT;
    }
    
    if (OGL::stencil_buffer & bits)
    {
        glBits |= GL_STENCIL_BUFFER_BIT;
    }
    
    glClear(glBits);
}

void ContextGL3::setDepthMask(bool status)
{
    glDepthMask(static_cast<GLboolean>(status ? GL_TRUE : GL_FALSE));
}

void ContextGL3::setBlendFunc(OGL::BlendingFactor sourceFactor, OGL::BlendingFactor destinationFactor)
{
    GLenum src = blendingFactorToGLEnum(sourceFactor);
    GLenum dst = blendingFactorToGLEnum(destinationFactor);
    
    GL_CHECK(glBlendFunc(src, dst));
}

glm::ivec4 ContextGL3::getViewportSize()
{
    glm::ivec4 size;
    glGetIntegerv(GL_VIEWPORT, &size.x);
    return size;
}

std::string ContextGL3::getShaderPath() const
{
	std::ostringstream buffer("");	//

	buffer << this->_resources->shadersDir << std::string("OGL");
    return buffer.str();
}

void ContextGL3::setStencilMask(bool status)
{
    unsigned int mask = status ? 0xFFu : 0x00u;
    glStencilMask(mask);
}

void ContextGL3::setStencilFunc(OGL::Func func, int ref, unsigned int mask)
{
    GLenum glFunc = funcToGLEnum(func);
    
    glStencilFunc(glFunc, ref, mask);
}

void ContextGL3::setStencilOp(OGL::StencilOp sfail, OGL::StencilOp dpfail, OGL::StencilOp dppass)
{
    GLenum glFail = stencilOpToGLEnum(sfail);
    GLenum glZFail= stencilOpToGLEnum(dpfail);
    GLenum glZPass = stencilOpToGLEnum(dppass);
    
    glStencilOp(glFail, glZFail, glZPass);
}

void ContextGL3::setColorMask(bool red, bool green, bool blue, bool alpha)
{
    glColorMask(static_cast<GLboolean>(red),
                static_cast<GLboolean>(green),
                static_cast<GLboolean>(blue),
                static_cast<GLboolean>(alpha));
}

int ContextGL3::getGPUInfoInteger(OGL::GPUInfoInteger type) const
{
    GLenum key = 0;
    int val = -1;
    
    switch(type)
    {
    case OGL::max_texture_2d_size:
        key = GL_MAX_TEXTURE_SIZE;
        break;
        
    case OGL::msaa_samples:
        key = GL_SAMPLES;
        break;
    }
    
    if (key != 0)
    {
        glGetIntegerv(key, &val);
    }
    
    return val;
}

float ContextGL3::getGPUInfoFloat(OGL::GPUInfoFloat type) const
{
    GLenum key = 0;
    float val = -1.f;

    switch (type)
    {
    case OGL::GPUInfoFloat::max_anisotropy:
        key = GL_MAX_TEXTURE_MAX_ANISOTROPY;
        break;
    }

    if (key != 0)
    {
        glGetFloatv(key, &val);
    }

    return val;
}

const SEAScope::RendererResources* ContextGL3::getResources() const
{
	return this->_resources;
}

void ContextGL3::setLineWidth(float val)
{
    glLineWidth(val);
}

} // namespace OGL3
