// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/OGL3/enum_translation.h"
#include "core/log.h"

namespace OGL3
{

GLenum primitiveToGLType(OGL::Primitive primitive)
{
    GLenum gl_type = GL_TRIANGLES;
    
    switch(primitive)
    {
    case OGL::points:
        gl_type = GL_POINTS;
        break;
    case OGL::lines:
        gl_type = GL_LINES;
        break;
    case OGL::line_strip:
        gl_type = GL_LINE_STRIP;
        break;
    case OGL::triangle_strip:
        gl_type = GL_TRIANGLE_STRIP;
        break;
    case OGL::triangle_fan:
        gl_type = GL_TRIANGLE_FAN;
        break;
    case OGL::triangles:
    case OGL::PrimitiveCount:
        gl_type = GL_TRIANGLES;
        break;
    }
    
    return gl_type;
}

GLenum dataTypeToGLDataType(DataType type)
{
    GLenum gl_type = GL_UNSIGNED_BYTE;

    switch(type)
    {
    case DataType::float32:
        gl_type = GL_FLOAT;
        break;
    case DataType::unsigned_short:
        gl_type = GL_UNSIGNED_SHORT;
        break;
    case DataType::unsigned_int:
        gl_type = GL_UNSIGNED_INT;
        break;
    case DataType::unsigned_byte:
    case DataType::none:
    case DataType::DataTypeCount:
        gl_type = GL_UNSIGNED_BYTE;
        break;
    }

    return gl_type;
}

GLenum bufferTargetEnumToGLenum(OGL::BufferType type)
{
    GLenum gl_type = GL_ARRAY_BUFFER;

    switch (type)
    {
    case OGL::element_array_buffer:
        gl_type = GL_ELEMENT_ARRAY_BUFFER;
        break;
    case OGL::copy_read_buffer:
        gl_type = GL_COPY_READ_BUFFER;
        break;
    case OGL::copy_write_buffer:
        gl_type = GL_COPY_WRITE_BUFFER;
        break;
    case OGL::pixel_pack_buffer:
        gl_type = GL_PIXEL_PACK_BUFFER;
        break;
    case OGL::pixel_unpack_buffer:
        gl_type = GL_PIXEL_UNPACK_BUFFER;
        break;
    case OGL::array_buffer:
    case OGL::BufferTypeCount:
        gl_type = GL_ARRAY_BUFFER;
        break;
    }
    return gl_type;
}

GLenum bufferUsageEnumToGLenum(OGL::BufferUsage usage)
{
    GLenum gl_type = GL_STATIC_READ;

    switch (usage)
    {
    case OGL::BufferUsage::dynamic_copy:
        gl_type = GL_DYNAMIC_COPY;
        break;
    case OGL::BufferUsage::dynamic_draw:
        gl_type = GL_DYNAMIC_DRAW;
        break;
    case OGL::BufferUsage::dynamic_read:
        gl_type = GL_DYNAMIC_READ;
        break;
    case OGL::BufferUsage::static_copy:
        gl_type = GL_STATIC_COPY;
        break;
    case OGL::BufferUsage::static_draw:
        gl_type = GL_STATIC_DRAW;
        break;
    case OGL::BufferUsage::static_read:
        gl_type = GL_STATIC_READ;
        break;
    case OGL::BufferUsage::stream_copy:
        gl_type = GL_STREAM_COPY;
        break;
    case OGL::BufferUsage::stream_draw:
        gl_type = GL_STREAM_DRAW;
        break;
    case OGL::BufferUsage::stream_read:
        gl_type = GL_STREAM_READ;
        break;
    case OGL::BufferUsage::BufferUsageCount:
		LOG(SEAScope::LogLevel::Error, "EnumTranslation",
						  "Provided BufferUsage enum value not supported");
		gl_type = GL_STATIC_READ; /* should not be used since the input is invalid */
		break;
    }
    return gl_type;
}

GLenum bufferAccessEnumToGLenum(OGL::BufferAccess access)
{
    GLenum gl_type = GL_READ_WRITE;

    switch (access)
    {
    case OGL::read_only:
        gl_type = GL_READ_ONLY;
        break;
    case OGL::write_only:
        gl_type = GL_WRITE_ONLY;
        break;
    case OGL::read_write:
        gl_type = GL_READ_WRITE;
        break;
    }
    return gl_type;
}

} // namespace OGL3
