// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"

#include "3d/OGL3/opengl.h"

#include "3d/OGL3/framebufferGL3.h"
#include "3d/OGL3/opengl_error.h"
#include "3d/OGL3/textureGL3.h"
#include "3d/OGL3/renderbufferGL3.h"
#include "3d/OGL3/contextGL3.h"
#include "3d/OGL3/enum_translation.h"

#include "core/log.h"
#include "3d/chrono.h"


namespace OGL3
{

FrameBufferGL3::FrameBufferGL3()
:   OGL::FrameBuffer()
{
}

FrameBufferGL3::FrameBufferGL3(unsigned int renderTargetCount, bool useDepth)
:   OGL::FrameBuffer(renderTargetCount, useDepth)
{
}

FrameBufferGL3::~FrameBufferGL3()
{
    destroy();
}

void FrameBufferGL3::create(OGL::Context* context)
{
    if (0 == this->handle)
    {
        GL_CHECK(glGenFramebuffers(1, &this->handle));
        
        _texture = new OGL::Texture*[_renderTargetCount];
        _colorAttachement = new unsigned int[_renderTargetCount];
            
        for (unsigned int i = 0; i < _renderTargetCount; ++i)
        {
            _texture[i] = context->createTexture(OGL::texture2D);
            _texture[i]->create(context);
            
            _colorAttachement[i] = GL_COLOR_ATTACHMENT0 + i;
        }
        
        _renderBuffer = context->createRenderbuffer(OGL::render_buffer_depth_stencil);
        _renderBuffer->create(context);
    }
}

void FrameBufferGL3::destroy()
{
    if (0 != this->handle)
    {
        for (unsigned int i = 0; i < _renderTargetCount; ++i)
        {
            delete _texture[i];
			_texture[i] = nullptr;
        }
        
        delete [] _texture; // call destroy
        _texture = nullptr;
        
        if (_useDepth)
        {
            delete _renderBuffer; // call destroy
            _renderBuffer = nullptr;    
        }
        
        delete [] _colorAttachement;
        _colorAttachement = nullptr;
        
        GL_CHECK(glDeleteFramebuffers(1, &this->handle));
        this->handle = 0;
    }
}

void FrameBufferGL3::bind() const
{
    GL_CHECK(glBindFramebuffer(GL_FRAMEBUFFER, this->handle));
            
    if (_renderTargetCount > 1)
    {
        GL_CHECK(glDrawBuffers(static_cast<GLsizei>(_renderTargetCount), _colorAttachement));
    }
}

void FrameBufferGL3::unbind() const
{
    GL_CHECK(glBindFramebuffer(GL_FRAMEBUFFER, 0));
}

bool FrameBufferGL3::allocate(unsigned int width, unsigned int height, DataType dataType)
{
    _dataType = dataType;
    _componentCount = 4;
    
    // Bind the current fbo object
    bind();
    
    for (unsigned int i = 0; i < _renderTargetCount; ++i)
    {
        // Create 2D texture to store rendering pass data.
        _texture[i]->bind();
        _texture[i]->allocate2D(static_cast<int>(width), static_cast<int>(height), 4, dataType, nullptr);
        _texture[i]->setWrapAndFilterMode(OGL::clamp_to_edge, OGL::nearest);
        _texture[i]->unbind();
        
        // Set "renderedTexture" as our colour attachement #0 + i
        GL_CHECK(glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + i, _texture[i]->getHandle(), 0));
    }
    
    if (_useDepth)
    {
        // Create a renderBuffer
        _renderBuffer->bind();
        _renderBuffer->allocate(width, height);
        _renderBuffer->unbind();

        // Use the renderBuffer as a depth buffer
        GL_CHECK(glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_RENDERBUFFER, _renderBuffer->getHandle()));
    }

    // Always check that our framebuffer is ok
    GLenum err;
    GL_CHECK(err = glCheckFramebufferStatus(GL_FRAMEBUFFER));
    
    if(err != GL_FRAMEBUFFER_COMPLETE)
    {
        switch(err)
        {
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
            LOG(SEAScope::LogLevel::Error, "OGL3::FrameBuffer", "FBO ERROR: GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT");
            break;
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
            LOG(SEAScope::LogLevel::Error, "OGL3::FrameBuffer", "FBO ERROR: GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT");
            break;
        case GL_FRAMEBUFFER_UNSUPPORTED:
            LOG(SEAScope::LogLevel::Error, "OGL3::FrameBuffer", "FBO ERROR: GL_FRAMEBUFFER_UNSUPPORTED");
            break;
        }
        
        unbind();
        return false;
    }
    else
    {
        LOG(SEAScope::LogLevel::Debug, "OGL3::FrameBuffer", "Fbo creation complete");                    
    }
    
    unbind();

    return true;
}

void FrameBufferGL3::resize(unsigned int width, unsigned int height)
{
    for (unsigned int i = 0; i < _renderTargetCount; ++i)
    {
        _texture[i]->bind();
        _texture[i]->resize(static_cast<int>(width), static_cast<int>(height));
        _texture[i]->unbind();
    }
    
    if (_useDepth)
    {
        _renderBuffer->bind();
        _renderBuffer->allocate(width, height);
        _renderBuffer->unbind();
    }
}

glm::vec3 FrameBufferGL3::readPixel(int x, int y) const
{
    // Set the framebuffer current
    bind();
    
    glm::vec3 color;
    
    GLenum gl_type = dataTypeToGLDataType(_dataType);
    
    // Read 1 RGB pixel
    glReadPixels(x, y, 1, 1, GL_RGB, gl_type, &color.r);
    
    // Disable the framebuffer
    unbind();
    
    // Return the read pixel
    return color;
}

void FrameBufferGL3::readPixelAsync(int x, int y) const
{
    GLenum gl_type = dataTypeToGLDataType(_dataType);
    
    glReadPixels(x, y, 1, 1, GL_RGB, gl_type, nullptr);
}

void FrameBufferGL3::blit(unsigned int width, unsigned int height)
{
    glBindFramebuffer(GL_READ_FRAMEBUFFER, this->handle);
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
    
    glBlitFramebuffer(0, 0, static_cast<GLint>(width), static_cast<GLint>(height), 0, 0, static_cast<GLint>(width), static_cast<GLint>(height), GL_COLOR_BUFFER_BIT, GL_NEAREST);
    
    glBindFramebuffer(GL_READ_FRAMEBUFFER, 0);
}

void FrameBufferGL3::dumpToImage(Image& img)
{
    _texture[0]->dumpToImage(img);
}

} // namespace OGL3
