// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"

#include "3d/OGL3/opengl.h"
#include "3d/OGL3/programGL3.h"
#include "3d/OGL3/shaderGL3.h"
#include "3d/OGL3/opengl_error.h"

#include "core/log.h"


namespace
{
    
OGL::UniformType glUniformTypeToEnum(GLenum val)
{
    if (val == GL_FLOAT) return OGL::unary_float;
    else if (val == GL_FLOAT_VEC2) return OGL::vec2_float;
    else if (val == GL_FLOAT_VEC3) return OGL::vec3_float;
    else if (val == GL_FLOAT_VEC4) return OGL::vec4_float;
    else if (val == GL_INT ) return OGL::unary_int;
    else if (val == GL_INT_VEC2) return OGL::vec2_int;
    else if (val == GL_INT_VEC3) return OGL::vec3_int;
    else if (val == GL_INT_VEC4) return OGL::vec4_int;
    else if (val == GL_UNSIGNED_INT) return OGL::unary_uint;
    else if (val == GL_UNSIGNED_INT_VEC2) return OGL::vec2_uint;
    else if (val == GL_UNSIGNED_INT_VEC3) return OGL::vec3_uint;
    else if (val == GL_UNSIGNED_INT_VEC4) return OGL::vec4_uint;
    else if (val == GL_FLOAT_MAT3) return OGL::matrix3x3_float;
    else if (val == GL_FLOAT_MAT4) return OGL::matrix4x4_float;
    else if (val == GL_SAMPLER_1D) return OGL::sampler1D;
    else if (val == GL_SAMPLER_2D) return OGL::sampler2D;
    else if (val == GL_SAMPLER_3D) return OGL::sampler3D;
    
    return OGL::unknown_uniform_type;
}

GLenum enumToGLTransformFeedbackEnum(OGL::TransformFeedbackBufferMode bufferMode)
{
    GLenum val = 0;
    
    switch(bufferMode)
    {
    case OGL::interleaved_attribs:
        val = GL_INTERLEAVED_ATTRIBS;
        break;
    case OGL::separate_attribs:
        val = GL_SEPARATE_ATTRIBS;
        break;
    }
    
    return val;
}

} // namespace

namespace OGL3
{

ProgramGL3::ProgramGL3()
:   OGL::Program()
{
}
    
ProgramGL3::~ProgramGL3()
{
    destroy();
}
    
void ProgramGL3::create(OGL::Context* /*context*/)
{
    GL_CHECK(this->handle = glCreateProgram());
}

void ProgramGL3::destroy()
{
    if (0 != this->handle)
    {
        GL_CHECK(glDeleteProgram(this->handle));
        this->handle = 0;
    }
}

void ProgramGL3::attachShaders(const std::vector<OGL::Shader*>& shaders)
{
    for (const OGL::Shader* shader : shaders)
    {
        shaderList.push_back(static_cast<ShaderGL3*>(const_cast<OGL::Shader*>(shader)));
        GL_CHECK(glAttachShader(this->handle, shader->getHandle()));
    }
}

void ProgramGL3::attachShaders(const OGL::Shader* shaders, unsigned int size)
{
    for (unsigned int i = 0; i < size; ++i)
    {
        shaderList.push_back(const_cast<OGL::Shader*>(&shaders[i]));
        GL_CHECK(glAttachShader(this->handle, shaders[i].getHandle()));
    }
}

bool ProgramGL3::link()
{
    GLint status; //
    GLint log_length; //
    
    GL_CHECK(glLinkProgram(this->handle));
    
    // Check the link status
    GL_CHECK(glGetProgramiv(this->handle, GL_LINK_STATUS, &status));
    
    // Get log length
    GL_CHECK(glGetProgramiv(this->handle, GL_INFO_LOG_LENGTH, &log_length));

    // Check if there are something to print (warning, error).
    // Test > 1 because with some drivers empty string have a length of 1 : "\0".
    if (log_length > 1)
    {
        std::string log_data;

        // Allocate a buffer
        log_data.resize(static_cast<std::size_t>(log_length));

        // Get the log
        GL_CHECK(glGetProgramInfoLog(this->handle, log_length, &log_length, static_cast<GLchar*>(const_cast<char*>(log_data.c_str()))));

        // Check link status.
        if (GL_TRUE == status)
        {
            // Link ok, print warning log
            LOG(SEAScope::LogLevel::Debug, "OGL3::Program", "Link success. Log:\n {}", log_data);
        }
        else
        {
            // Link ko, print error log
            std::string errString;
            
            for (OGL::Shader* shader : this->shaderList)
            {
                errString += shader->getName() + " ";
            }
            
            std::stringstream ss;
            ss << "Link fail between " << errString << "\nLog: " << log_data.c_str();
            
            LOG(SEAScope::LogLevel::Error, "OGL3::Program", ss.str());
        }
    }
    
    return GL_TRUE == status;
}

void ProgramGL3::bind() const
{
    GL_CHECK(glUseProgram(this->handle));
}

void ProgramGL3::unbind() const
{
    GL_CHECK(glUseProgram(0));
}

void ProgramGL3::bindAttribLocation(unsigned int pos, const char *name)
{
    GL_CHECK(glBindAttribLocation(this->handle, pos, name));
}

int ProgramGL3::getUniformLocation(const char *name)
{
    GLint result;
    GL_CHECK(result = glGetUniformLocation(this->handle, name));
    return result;
}

int ProgramGL3::getAttribLocation(const char *name)
{
    GLint result;
    GL_CHECK(result = glGetAttribLocation(this->handle, name));
    return result;
}

int ProgramGL3::getUniformCount() const
{
    int count;
    glGetProgramiv(handle, GL_ACTIVE_UNIFORMS, &count);
    return count;
}

std::string ProgramGL3::getUniformName(int id) const
{
    int length; // string length
    
    // Get the maximum length of a uniform name.
    glGetProgramiv(handle, GL_ACTIVE_UNIFORM_MAX_LENGTH, &length);
    
    // Allocate string buffer
    std::string name(static_cast<std::size_t>(length) + 1, '\0');
    // Hey ! I'm an useless int :p
    unsigned int dummy;
    GLsizei size;
        
    // Ask to OpenGL the name of the uniform
    glGetActiveUniform(handle, static_cast<GLuint>(id), length,  nullptr,  &size,  &dummy,  const_cast<char*>(name.data()));
    
    // Return the name
    return name;
}

OGL::UniformType ProgramGL3::getUniformType(int id) const
{
    int dummy;
    GLenum type;
    glGetActiveUniform(handle, static_cast<GLuint>(id), 0,  nullptr,  &dummy,  &type,  nullptr);
    return glUniformTypeToEnum(type);
}

void ProgramGL3::uniform1i(int location, const int& value)
{
    glUniform1i(location, value);
}

void ProgramGL3::uniform1u(int location, const std::uint32_t& value)
{
    glUniform1ui(location, value);
}

void ProgramGL3::uniform1f(int location, const float& value)
{
    glUniform1f(location, value);
}

void ProgramGL3::uniform2fv(int location, const glm::vec2& value)
{
    glUniform2fv(location, 1, &value.x);
}

void ProgramGL3::uniform3fv(int location, const glm::vec3& value)
{
    glUniform3fv(location, 1, &value.x);
}

void ProgramGL3::uniform4fv(int location, const glm::vec4& value)
{
    glUniform4fv(location, 1, &value.x);
}

void ProgramGL3::uniformMatrix4fv(int location, const glm::mat4& value)
{
    glUniformMatrix4fv(location, 1, GL_FALSE, glm::value_ptr(value));
}

void ProgramGL3::uniform2bv(int location, const glm::bvec2& value)
{
    glUniform2i(location, static_cast<GLint>(value.x), static_cast<GLint>(value.y));
}

void ProgramGL3::setTransformFeedbackVaryings(const std::vector<std::string>& varyings, OGL::TransformFeedbackBufferMode bufferMode)
{
    const char** v = new const char*[varyings.size()];
    
    for (size_t i = 0; i < varyings.size(); ++i)
    {
        v[i] = varyings[i].c_str();
    }
    
    GLenum glBufferMode = enumToGLTransformFeedbackEnum(bufferMode);
    
    glTransformFeedbackVaryings(this->handle, static_cast<GLsizei>(varyings.size()), v, glBufferMode);
    
    delete [] v;
	v = nullptr;
}

} // namespace OGL3
