// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"

#include "3d/OGL3/opengl.h"
#include "3d/OGL3/shaderGL3.h"
#include "3d/OGL3/opengl_error.h"

#include "3d/file.h"

#include "core/log.h"


namespace OGL3
{

ShaderGL3::ShaderGL3(OGL::ShaderType typeIn, const std::string& nameIn)
:   OGL::Shader(typeIn, nameIn)
{
}

ShaderGL3::~ShaderGL3()
{
    destroy();
}

void ShaderGL3::create(OGL::Context* /*context*/)
{
    GLenum type(GL_VERTEX_SHADER); // shader type
    
    switch(this->_type)
    {
    case OGL::vertex_shader: 
        type = GL_VERTEX_SHADER;
        break;
        
    case OGL::geometry_shader:
        type = GL_GEOMETRY_SHADER;
        break;
        
    case OGL::fragment_shader: 
        type = GL_FRAGMENT_SHADER;
        break;
    case OGL::tess_control_shader:
    case OGL::tess_evaluation_shader:
    case OGL::compute_shader:
    case OGL::ShaderTypeCount:
        type = GL_VERTEX_SHADER;
        break;
    }
    
    // Create a new shader object
    GL_CHECK(this->handle = glCreateShader(type));
}

void ShaderGL3::destroy()
{
    if (0 != this->handle)
    {
        GL_CHECK(glDeleteShader(this->handle));
        this->handle = 0;
    }
}

bool ShaderGL3::loadFromFile(OGL::Context* context, const std::string& file_name)
{
    _filename = file_name;
    std::string text_source;
    
    LOG(SEAScope::LogLevel::Debug, "OGL3::Shader", "Read {}", file_name);
    
    bool result = File::readAll(file_name, text_source);
    
    if (0 == this->handle)
    {
        create(context);
    }
    
    if (true == result)
    {
        result = compile(text_source);
    }
    
    return result;
}

bool ShaderGL3::loadFromString(OGL::Context* context, const std::string& source)
{
    // Check string length
    if (0 == source.size())
    {
        LOG(SEAScope::LogLevel::Error, "OGL3::Shader", "Text source is empty.");
        return false;
    }
    
    if (0 == this->handle)
    {
        create(context);
    }
    
    return compile(source);
}

bool ShaderGL3::compile(const std::string& source)
{
    GLint status;
    GLint log_length; // int variable use many time

    // Get the string buffer
    const GLchar * string = static_cast<const GLchar*>(source.c_str());
    
    // Set the shader source
    GL_CHECK(glShaderSource(this->handle, 1, static_cast<const GLchar**>(&string), nullptr));
    
    // Compile the shader source
    GL_CHECK(glCompileShader(this->handle));
    
    // Check compilation status
    GL_CHECK(glGetShaderiv(this->handle, GL_COMPILE_STATUS, &status));
    
    // Get the log length
    GL_CHECK(glGetShaderiv(this->handle, GL_INFO_LOG_LENGTH, &log_length));

    // Check log length. Maybe some warnings ?
    // Check > 1 because with some drivers, empty string length equal '\0'
    if (log_length > 1)
    {
        std::string log_data;
        
        // Allocate a buffer
        log_data.resize(static_cast<std::size_t>(log_length));
        
        // Get the log
        GL_CHECK(glGetShaderInfoLog(this->handle, log_length, &log_length, static_cast<GLchar*>(const_cast<char*>(log_data.c_str()))));
        
        // Print the log
        if (GL_TRUE == status)
        {
            LOG(SEAScope::LogLevel::Error, "OGL3::Shader", "Shader compilation is a success . Log:\n {}", log_data);
        }
        else
        {
            LOG(SEAScope::LogLevel::Error, "OGL3::Shader", "Shader compilation has failed.\n {}\n Log:\n {}", _filename, log_data);
        }
    }
    
    return status == GL_TRUE;
}

} // namespace OGL3
