// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/OGL3/textureGL3.h"

#include "3d/OGL3/opengl.h"
#include "3d/OGL3/opengl_error.h"
#include "data/image.h"
#include "3d/memory_manager.h"

#include "3d/ppm.h"

#include "core/log.h"

namespace OGL3
{
    struct TextureDesc
    {
        GLenum internal_format;
        GLenum format;
        GLenum data_type;
    };
}

namespace
{
    
GLenum enumToGLenum(OGL::TextureType type)
{
    GLenum gl_type = GL_TEXTURE_2D;

    switch(type)
    {
    case OGL::texture1D:
        gl_type = GL_TEXTURE_1D;
        break;
    case OGL::texture2D:
    case OGL::TextureTypeCount:
        gl_type = GL_TEXTURE_2D;
        break;
    }
    
    return gl_type;
}

GLenum enumToGLBinding(OGL::TextureType type)
{
    GLenum gl_result = GL_TEXTURE_BINDING_2D;

    switch(type)
    {
    case OGL::texture1D:
        gl_result = GL_TEXTURE_BINDING_1D;
        break;
    case OGL::texture2D:
    case OGL::TextureTypeCount:
        gl_result = GL_TEXTURE_BINDING_2D;
        break;
    }
    
    return gl_result;
}


GLenum wrapModeToGLenum(OGL::WrapMode wrap)
{
    switch(wrap)
    {
    case OGL::clamp_to_edge:
        return GL_CLAMP_TO_EDGE;
    case OGL::repeat:
        return GL_REPEAT;
    case OGL::WrapModeCount:
        return GL_CLAMP_TO_EDGE;
    }

    return GL_CLAMP_TO_EDGE;
}

GLenum filterModeToGLenum(OGL::FilterMode filter)
{
    switch(filter)
    {
    case OGL::nearest:
        return GL_NEAREST;
    case OGL::linear:
        return GL_LINEAR;
    case OGL::nearest_mipmap_nearest:
        return GL_NEAREST_MIPMAP_NEAREST;
    case OGL::nearest_mipmap_linear:
        return GL_NEAREST_MIPMAP_LINEAR;
    case OGL::linear_mipmap_nearest:
        return GL_LINEAR_MIPMAP_LINEAR;
    case OGL::linear_mipmap_linear:
        return GL_LINEAR_MIPMAP_NEAREST;
    case OGL::FilterModeCount:
        return GL_LINEAR;
    }

    return GL_LINEAR;
}


OGL3::TextureDesc fillTextureDesc(DataType dataType, int channels)
{
    OGL3::TextureDesc desc = { 0, 0, 0 };

    if (DataType::float32 == dataType)
    {
        desc.data_type = GL_FLOAT;

        if (1 == channels)
        {
            desc.internal_format = GL_R32F;
            desc.format = GL_RED;
        }
        else if (2 == channels)
        {
            desc.internal_format = GL_RG32F;
            desc.format = GL_RG;
        }
        else if (3 == channels)
        {
            desc.internal_format = GL_RGB32F;
            desc.format = GL_RGB;
        }
        else if (4 == channels)
        {
            desc.internal_format = GL_RGBA32F;
            desc.format = GL_RGBA;
        }
    }
    else if (DataType::unsigned_byte == dataType)
    {
        desc.data_type = GL_UNSIGNED_BYTE;

        if (1 == channels)
        {
            desc.internal_format = GL_R8;
            desc.format = GL_RED;
        }
        else if (2 == channels)
        {
            desc.internal_format = GL_RG;
            desc.format = GL_RG;
        }
        else if (3 == channels)
        {
            desc.internal_format = GL_RGB;
            desc.format = GL_RGB;
        }
        else if (4 == channels)
        {
            desc.internal_format = GL_RGBA;
            desc.format = GL_RGBA;
        }
    }

    return desc;
}

} // namespace

namespace OGL3
{
    
TextureGL3::TextureGL3(OGL::TextureType typeIn)
:   OGL::Texture(typeIn)
{
    this->gl_type = enumToGLenum(typeIn);
    this->gl_binding = enumToGLBinding(typeIn);
    this->type = typeIn;
    _width = 0;
    _height = 0;
    _channels = 0;
}

TextureGL3::~TextureGL3()
{
    destroy();
}

void TextureGL3::allocate1D(int width, int channels, DataType data_type, void* buffer)
{
    allocate(&width, 1, channels, data_type, buffer);
}

void TextureGL3::allocate2D(int width, int height, int channels, DataType data_type, void* buffer)
{
    int dimensions[] = { width , height };
    allocate(dimensions, 2, channels, data_type, buffer);
}

void TextureGL3::allocate2D(const Image& image)
{
    int dimensions[] = { static_cast<int>(image.width()), static_cast<int>(image.height()) };
    allocate(dimensions, 2, static_cast<int>(image.channels()), image.dataType(), image.data());
}

void TextureGL3::allocate2D(const std::vector<Image>& mipmaps)
{
    _channels = static_cast<int>(mipmaps[0].channels());
    _dataType = mipmaps[0].dataType();

    bind();

    setWrapModeNoBind(OGL::clamp_to_edge);
    setFilterModeNoBind(OGL::linear_mipmap_linear, OGL::linear);
    //setFilterModeNoBind(OGL::nearest_mipmap_linear, OGL::linear);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, static_cast<GLint>(mipmaps.size() - 1));

    float val;
    glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY, &val);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY, val);

    TextureDesc desc = fillTextureDesc(_dataType, _channels);

    int alignment = _channels == 3 ? 1 : _channels;
    glPixelStorei(GL_UNPACK_ALIGNMENT, alignment);

    gl_dataType = desc.data_type;
    gl_format = desc.format;

    _width = static_cast<unsigned int>(mipmaps[0].width());
    _height = static_cast<unsigned int>(mipmaps[1].width());
    
    for (size_t i = 0; i < mipmaps.size(); ++i)
    {
        int dimensions[2] = 
        { 
            static_cast<int>(mipmaps[i].width()), 
            static_cast<int>(mipmaps[i].height()) 
        };

        glTexImage2D(
            this->gl_type, 
            static_cast<GLint>(i), 
            static_cast<GLint>(desc.internal_format), 
            dimensions[0], 
            dimensions[1], 
            0, 
            desc.format, 
            desc.data_type, 
            mipmaps[i].data());
    }

    unbind();
}

void TextureGL3::setWrapModeNoBind(OGL::WrapMode wrap)
{
    _wrap = wrap;
    
    GLenum gl_wrap = wrapModeToGLenum(wrap);
        
    // Wrapping mode
    GL_CHECK(glTexParameteri(this->gl_type, GL_TEXTURE_WRAP_S, static_cast<GLint>(gl_wrap)));
    
    if (GL_TEXTURE_2D == this->gl_type)
    {
        GL_CHECK(glTexParameteri(this->gl_type, GL_TEXTURE_WRAP_T, static_cast<GLint>(gl_wrap)));
    }
}

void TextureGL3::setData(void* data)
{
    bind();
    
    if (OGL::texture1D == this->type)
    {
        GL_CHECK(glTexSubImage1D(this->gl_type, 0, 0, static_cast<GLsizei>(_width), this->gl_format, this->gl_dataType, data));
    }
    else if (OGL::texture2D == this->type)
    {
        GL_CHECK(glTexSubImage2D(this->gl_type, 0, 0, 0, static_cast<GLsizei>(_width), static_cast<GLsizei>(_height), this->gl_format, this->gl_dataType, data));
    }
    
    unbind();
}

void TextureGL3::create(OGL::Context* /*context*/)
{
    if (0 == this->handle)
    {
        GL_CHECK(glGenTextures(1, &this->handle));
    }
}

void TextureGL3::destroy()
{
    if (0 != this->handle)
    {
        GL_CHECK(glDeleteTextures(1, &this->handle));
        this->handle = 0;
        
        unsigned long long size = _width * _height * static_cast<unsigned int>(_channels) * (_dataType == DataType::unsigned_byte ? 1 : 4);
        MemoryManager::getInstance()->release(MemoryCategory::texture, size);
        
        _width = 0;
        _height = 0;
        _channels = 0;
        _dataType = DataType::none;
    }
}

void TextureGL3::bind() const
{
    GL_CHECK(glBindTexture(this->gl_type, this->handle));
}

void TextureGL3::unbind() const
{
    GLenum glType = enumToGLenum(type);
    GL_CHECK(glBindTexture(glType, 0));
}

void TextureGL3::allocate(int * dimensions, int size, int channels, DataType dataType, void * buffer)
{
    if (channels < 1 || channels > 4)
    {
        LOG(SEAScope::LogLevel::Error, "OGL3::Texture", "Invalid number of channels: {}", channels);
        return;
    }
    
    if (0 == this->handle)
    {
        LOG(SEAScope::LogLevel::Error, "OGL3::Texture", "Texture object is not created. Please call create function before allocate.", channels);
        return;
    }
    
    if (0 != _width)
    {
        unsigned long long textureSize = _width * (_height == 0 ? 1 : _height) * static_cast<unsigned int>(_channels) * (_dataType == DataType::unsigned_byte ? 1 : 4);
        MemoryManager::getInstance()->release(MemoryCategory::texture, textureSize);
    }
    
    _channels = channels;
    _dataType = dataType;
    
    bind();
    
    setWrapAndFilterModeNoBind(OGL::repeat, OGL::linear);
        
    TextureDesc desc = fillTextureDesc(_dataType, _channels);

    int alignment = _channels == 3 ? 1 : _channels;

    // 1 -> 1, 2 -> 2, 3 -> 1, 4 -> 4
    glPixelStorei(GL_UNPACK_ALIGNMENT, alignment);
    
    gl_dataType = desc.data_type;
    gl_format = desc.format;
    
    if (1 == size)
    {
        _width = static_cast<unsigned int>(dimensions[0]);

        GL_CHECK(glTexImage1D(this->gl_type, 0, static_cast<GLint>(desc.internal_format), dimensions[0], 0, desc.format, desc.data_type, buffer));
    }
    else if (2 == size)
    {
        _width = static_cast<unsigned int>(dimensions[0]);
        _height = static_cast<unsigned int>(dimensions[1]);

        GL_CHECK(glTexImage2D(this->gl_type, 0, static_cast<GLint>(desc.internal_format), dimensions[0], dimensions[1], 0, desc.format, desc.data_type, buffer));
    }
    
    unsigned long long textureSize = _width * (_height == 0 ? 1 : _height) * static_cast<unsigned int>(_channels) * (_dataType == DataType::unsigned_byte ? 1 : 4);
    MemoryManager::getInstance()->consume(MemoryCategory::texture, textureSize);
    
    LOG(SEAScope::LogLevel::Debug, "OGL3::Texture", "texture allocation : {} MB", static_cast<float>(textureSize) / 1024.f / 1024.f);

    unbind();
}

void TextureGL3::setWrapAndFilterMode(OGL::WrapMode wrap, OGL::FilterMode filter)
{
    bind();
    setWrapModeNoBind(wrap);
    setFilterModeNoBind(filter);
    unbind();
}

void TextureGL3::setWrapAndFilterModeNoBind(OGL::WrapMode wrap, OGL::FilterMode filter)
{
    setWrapModeNoBind(wrap);
    setFilterModeNoBind(filter);
}

void TextureGL3::setWrapMode(OGL::WrapMode wrap)
{
    bind();
    
    setWrapModeNoBind(wrap);

    unbind();
}

bool TextureGL3::isCurrentBinding() const
{
    GLint whichID;
    
    // Get the current texture object ID bound.
    glGetIntegerv(this->gl_binding, &whichID);
    
    // Compare with the current object handle.
    return static_cast<GLint>(this->handle) == whichID;
}

void TextureGL3::setFilterMode(OGL::FilterMode filter)
{
    bind();
    
    setFilterModeNoBind(filter);

    unbind();
}

void TextureGL3::setFilterMode(OGL::FilterMode minfilter, OGL::FilterMode magfilter)
{
    bind();

    setFilterModeNoBind(minfilter, magfilter);

    unbind();
}

void TextureGL3::setAnisotropicFilter(float val)
{
    bind();

    setAnisotropicFilterNoBind(val);

    unbind();
}

void TextureGL3::setAnisotropicFilterNoBind(float val)
{
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY, val);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_LOD_BIAS, -0.2f); // magic value
}

void TextureGL3::setFilterModeNoBind(OGL::FilterMode filter)
{
    _minFilter = filter;
    _magFilter = filter;
    
    GLenum gl_filter = filterModeToGLenum(filter);
    
    GL_CHECK(glTexParameteri(this->gl_type, GL_TEXTURE_MIN_FILTER, static_cast<GLint>(gl_filter)));
    GL_CHECK(glTexParameteri(this->gl_type, GL_TEXTURE_MAG_FILTER, static_cast<GLint>(gl_filter)));
}

void TextureGL3::setFilterModeNoBind(OGL::FilterMode minfilter, OGL::FilterMode magfilter)
{
    _minFilter = minfilter;
    _magFilter = magfilter;

    GLenum gl_minFilter = filterModeToGLenum(minfilter);
    GLenum gl_magFilter = filterModeToGLenum(magfilter);

    GL_CHECK(glTexParameteri(this->gl_type, GL_TEXTURE_MIN_FILTER, static_cast<GLint>(gl_minFilter)));
    GL_CHECK(glTexParameteri(this->gl_type, GL_TEXTURE_MAG_FILTER, static_cast<GLint>(gl_magFilter)));
}

void TextureGL3::resize(int width, int height)
{
    allocate2D(width, height, _channels, _dataType, nullptr);
}

void TextureGL3::setData(unsigned int x, unsigned int y, unsigned int width, unsigned int height, void * data)
{
    bind();
    
    GL_CHECK(glTexSubImage2D(this->gl_type, 0, static_cast<GLint>(x), static_cast<GLint>(y), static_cast<GLsizei>(width), static_cast<GLsizei>(height), this->gl_format, this->gl_dataType, data));
    
    unbind();
}

void TextureGL3::dumpToImage(Image& img)
{
    bind();
    
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    
    img.allocate(_width, _height, static_cast<unsigned int>(_channels), _dataType, nullptr);
    
    glGetTexImage(this->gl_type, 0, this->gl_format, this->gl_dataType, img.data());
    
    glPixelStorei(GL_PACK_ALIGNMENT, 4);

    unbind();
}

void TextureGL3::generateMipmaps()
{
    bind();

    glGenerateMipmap(GL_TEXTURE_2D);

    unbind();
}

} // namespace OGL3
