// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/anchor_pool_node.h"
#include "3d/screen_quad_color.h"
#include "3d/shape_anchor.h"
#include "3d/shape_editor.h"
#include "3d/shape_anchor.h"
#include "3d/texture_loader.h"
#include "3d/context.h"
#include "3d/icons.h"
#include "3d/static_mesh.h"
#include "3d/screen_quad.h"
#include "3d/ipicking_renderer.h"
#include "core/path.h"
#include "core/app_events_dispatcher.h"
#include "events/anchor_added_event.h"
#include "events/anchor_removed_event.h"

#include <iostream>

AnchorPoolNode::AnchorPoolNode(SEAScope::AppEventsDispatcher& appEventsDispatcher)
: _appEventsDispatcher(appEventsDispatcher),
  _screenQuadMesh(nullptr)
{
    this->_anchors.resize(AnchorType::AnchorTypeCount);

    this->_anchorDecoration.resize(AnchorType::AnchorTypeCount);
    for (auto& deco : this->_anchorDecoration)
    {
        deco.resize(ObjectCursorState::ObjectStateCount);
    }
}

AnchorPoolNode::~AnchorPoolNode()
{
    this->clear();

    delete this->_screenQuadMesh;
    this->_screenQuadMesh = nullptr;
}

bool AnchorPoolNode::initialize(OGL::Context* ctx, const Icons& icons)
{
    bool ok(false);
    OGL::Texture* texture(nullptr);

    texture = icons.getTexture(IconType::SMALL_CIRCLE_OUTLINE_32, ok);
    if
        (false == ok)
    {
        return false;
    }
    this->setDecoration(AnchorType::PointMoveable,
                        ObjectCursorState::ObjectNormal,
                        texture);

    texture = icons.getTexture(IconType::SMALL_CIRCLE_MOVE_32, ok);
    if
        (false == ok)
    {
        return false;
    }
    this->setDecoration(AnchorType::PointMoveable,
                        ObjectCursorState::ObjectHighlight,
                        texture);

    texture = icons.getTexture(IconType::SMALL_CIRCLE_FILLED_32, ok);
    if
        (false == ok)
    {
        return false;
    }
    this->setDecoration(AnchorType::PointMoveable,
                        ObjectCursorState::ObjectSelected,
                        texture);

    texture = icons.getTexture(IconType::SMALL_CIRCLE_PLUS_32, ok);
    if
        (false == ok)
    {
        return false;
    }
	this->setDecoration(AnchorType::PointAdd,
                        ObjectCursorState::ObjectNormal,
                        texture);

    texture = icons.getTexture(IconType::CIRCLE_PLUS_32, ok);
    if
        (false == ok)
    {
        return false;
    }
	setDecoration(AnchorType::PointAdd,
                  ObjectCursorState::ObjectHighlight,
                  texture);

    this->_screenQuadMesh = ScreenQuad::createMesh(ctx);

    return true;
}

void AnchorPoolNode::setDecoration(AnchorType type, ObjectCursorState state, OGL::Texture* texture)
{
    this->_anchorDecoration[type][state] = texture;
}

void AnchorPoolNode::removeAnchor(AnchorType type, unsigned int id)
{
    ShapeAnchor* sa(nullptr);
    auto it = this->_anchors[type].begin();
    for
        (; it != this->_anchors[type].end(); ++it)
    {
        if
            (id == (*it)->getId())
        {
            sa = *it;
            break;
        }
    }

    if
        (this->_anchors[type].end() != it)
    {
        SEAScope::AnchorRemovedEvent event(sa);
        this->_appEventsDispatcher.dispatch<SEAScope::AnchorRemovedEvent>(event);

        this->_anchors[type].erase(it);

        delete sa;
        sa = nullptr;
    }
}

void AnchorPoolNode::addAnchor(AnchorController* controller, AnchorType type, unsigned int id)
{
    ShapeAnchor* anchor = new ShapeAnchor();
    anchor->initialize(type, controller, this->_screenQuadMesh, id);
    // source id is set to 0, the default value, because actual data sources
    // have identifiers > 0.
    // collection id is the value of anchor type (+1 to distinguish from default value)
    // granule id is built using the anchor id (+ 1 to distinguish from default value)
    // with an offset based on the anchor type to mitigate the risk of having
    // the same granule id for two different anchors (each anchor type has a 10^18 range of granule ids)
    anchor->setRenderableId(SEAScope::RenderableId(
                             true,
							 {0u, type + 1u},
							 {0u, (1u + type) * 1000000000000000000u + id + 1u},
							 0u));

    SEAScope::AnchorAddedEvent event(anchor);
    this->_appEventsDispatcher.dispatch<SEAScope::AnchorAddedEvent>(event);

    this->_anchors[type].push_back(anchor);
}

void AnchorPoolNode::clear()
{
    for (size_t i = 0; i < this->_anchors.size(); ++i)
    {
		for (ShapeAnchor* sa : this->_anchors[i])
		{
            SEAScope::AnchorRemovedEvent event(sa);
            this->_appEventsDispatcher.dispatch<SEAScope::AnchorRemovedEvent>(event);

			delete sa;
			sa = nullptr;
		}

		this->_anchors[i].clear();
    }
}

void AnchorPoolNode::draw(OGL::Context* ctx, const glm::mat4& mvp)
{
    _renderer->bind(ctx, mvp, nullptr);

    ShapeAnchor* anchor = nullptr;

    OGL::Texture* current = nullptr;

    for (size_t i = 0; i < this->_anchors.size(); ++i)
    {
        for (size_t j = 0; j < this->_anchors[i].size(); ++j)
        {
            anchor = this->_anchors[i][j];

            Style& style = anchor->getStyle();

            OGL::Texture* texture = nullptr;

            if (true == style.isHighlighted())
            {
                texture = this->_anchorDecoration[i][ObjectCursorState::ObjectHighlight];
            }
            else if (StyleState::selected == style.getState())
            {
                texture = this->_anchorDecoration[i][ObjectCursorState::ObjectSelected];
            }
            else // normal
            {
                texture = this->_anchorDecoration[i][ObjectCursorState::ObjectNormal];
            }

            if (texture != current)
            {
                ctx->activeTextureUnit(0, texture);
                current = texture;
            }

            _renderer->draw(ctx, anchor);
        }
    }
}

void AnchorPoolNode::drawPicking(OGL::Context* ctx, const glm::mat4& mvp)
{
    _rendererPicking->bind(ctx, mvp, nullptr);

    for (size_t i = 0; i < this->_anchors.size(); ++i)
    {
        for (size_t j = 0; j < this->_anchors[i].size(); ++j)
        {
            _rendererPicking->draw(ctx, this->_anchors[i][j]);
        }
    }
}

/**
 *
 */
void AnchorPoolNode::showAll()
{
    for (size_t i = 0; i < this->_anchors.size(); ++i)
    {
        for (size_t j = 0; j < this->_anchors[i].size(); ++j)
        {
            this->_anchors[i][j]->setVisible(true);
        }
    }
}

/**
 *
 */
void AnchorPoolNode::showOnly(AnchorType type, unsigned int id)
{
    for (size_t i = 0; i < this->_anchors.size(); ++i)
    {
        for (size_t j = 0; j < this->_anchors[i].size(); ++j)
        {
            this->_anchors[i][j]->setVisible((i == type) && (j == id));
        }
    }
}
