// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   centripetal_catmull_rom.cpp
 * Author: gleseach
 * 
 * Created on 24 septembre 2015, 15:50
 */

#include "3d/stdafx.h"
#include "3d/centripetal_catmull_rom.h"

struct CubicPoly
{
    float c0, c1, c2, c3;

    float eval(float t)
    {
        float t2 = t*t;
        float t3 = t2 * t;
        return c0 + c1*t + c2*t2 + c3*t3;
    }
};

/*
 * Compute coefficients for a cubic polynomial
 *   p(s) = c0 + c1*s + c2*s^2 + c3*s^3
 * such that
 *   p(0) = x0, p(1) = x1
 *  and
 *   p'(0) = t0, p'(1) = t1.
 */
static void InitCubicPoly(float x0, float x1, float t0, float t1, CubicPoly &p)
{
    p.c0 = x0;
    p.c1 = t0;
    p.c2 = -3*x0 + 3*x1 - 2*t0 - t1;
    p.c3 = 2*x0 - 2*x1 + t0 + t1;
}

// standard Catmull-Rom spline: interpolate between x1 and x2 with previous/following points x1/x4
// (we don't need this here, but it's for illustration)
/*
static void InitCatmullRom(float x0, float x1, float x2, float x3, CubicPoly &p)
{
    // Catmull-Rom with tension 0.5
    InitCubicPoly(x1, x2, 0.5f*(x2-x0), 0.5f*(x3-x1), p);
}
*/

// compute coefficients for a nonuniform Catmull-Rom spline
static void InitNonuniformCatmullRom(float x0, float x1, float x2, float x3, float dt0, float dt1, float dt2, CubicPoly &p)
{
    // compute tangents when parameterized in [t1,t2]
    float t1 = (x1 - x0) / dt0 - (x2 - x0) / (dt0 + dt1) + (x2 - x1) / dt1;
    float t2 = (x2 - x1) / dt1 - (x3 - x1) / (dt1 + dt2) + (x3 - x2) / dt2;

    // rescale tangents for parametrization in [0,1]
    t1 *= dt1;
    t2 *= dt1;

    InitCubicPoly(x1, x2, t1, t2, p);
}

typedef glm::vec2 Vec2D;

//struct Vec2D
//{
//    Vec2D(float _x, float _y) : x(_x), y(_y) {}
//    float x, y;
//};

static float VecDistSquared(const Vec2D& p, const Vec2D& q)
{
    float dx = q.x - p.x;
    float dy = q.y - p.y;
    return dx*dx + dy*dy;
}

static void InitCentripetalCR(const Vec2D& p0, const Vec2D& p1, const Vec2D& p2, const Vec2D& p3,
    CubicPoly &px, CubicPoly &py)
{
    float dt0 = std::pow(VecDistSquared(p0, p1), 0.25f);
    float dt1 = std::pow(VecDistSquared(p1, p2), 0.25f);
    float dt2 = std::pow(VecDistSquared(p2, p3), 0.25f);

    // safety check for repeated points
    if (dt1 < 1e-4f)    dt1 = 1.0f;
    if (dt0 < 1e-4f)    dt0 = dt1;
    if (dt2 < 1e-4f)    dt2 = dt1;

    InitNonuniformCatmullRom(p0.x, p1.x, p2.x, p3.x, dt0, dt1, dt2, px);
    InitNonuniformCatmullRom(p0.y, p1.y, p2.y, p3.y, dt0, dt1, dt2, py);
}

static void interpolate(Vec2D p0, Vec2D p1, Vec2D p2, Vec2D p3, std::vector<glm::vec3>& out)
{
    CubicPoly px, py;
    InitCentripetalCR(p0, p1, p2, p3, px, py);
    
    glm::vec3 p;
    
    int iteration = 30;
    float step = 1.f / static_cast<float>(iteration);
    for (int i = 0; i <= 30; ++i)
    {
        p.x = px.eval(step * static_cast<float>(i));
        p.y = py.eval(step * static_cast<float>(i));
        p.z = 0.f;
        out.push_back(p);
    }
}

void createCentripetalCatmullRomSpline(const std::vector<glm::vec2>& points, std::vector<glm::vec3>& out)
{
    if (points.size() < 4)
        return;
    
    interpolate(points[0], points[0], points[1], points[2], out);
    
    for (unsigned int i = 0; i < points.size()-3; ++i)
    {
        interpolate(points[i], points[i+1], points[i+2], points[i+3], out);
    }
    
    std::size_t id = points.size() - 3;
    interpolate(points[id], points[id+1], points[id+2], points[id+2], out);
}
