// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/colormap_manager_gpu.h"
#include "3d/texture.h"
#include "3d/context.h"

#include "core/types.h"
#include "core/colormap_manager.h"

namespace OGL
{
    
ColormapManagerGPU::ColormapManagerGPU()
:   _mngr(nullptr)
{
}

ColormapManagerGPU::~ColormapManagerGPU()
{
    // Clear colormap cache
    clear();
}

void ColormapManagerGPU::init(OGL::Context* ctx, SEAScope::ColormapManager* mngr)
{
    // Save manager pointer object
    _mngr = mngr;
    
    // Loop on all colormap and create texture object
    for(auto& pair : _mngr->store)
    {
        _colormapCache[pair.first] = loadColormap(ctx, pair.second);
    }
}

void ColormapManagerGPU::clear()
{
    // Delete all texture objects
    for(auto& pair : _colormapCache)
    {
        delete pair.second;
		pair.second = nullptr;
    }
    
    _colormapCache.clear();
}

OGL::Texture* ColormapManagerGPU::getColormap(OGL::Context* ctx, const std::string& colormapName)
{
    // search in existing colormap
    auto it = _colormapCache.find(colormapName);
    
    // If it doesn't exist, load it
    if (it == _colormapCache.end())
    {
        SEAScope::Colormap result;
        
        // Use the colormapmanager to load the resquested colormap
        if (_mngr->getColormap(colormapName, result))
        {
            // Creature an OpenGL texture with it
            OGL::Texture* tex = loadColormap(ctx, result);
            
            // Add to the colormap cache
            _colormapCache[colormapName] = tex;
            
            return tex;
        }
        else if (0 < _colormapCache.size()) // check if it's not empty
        {
            // It doesn't exist in the colormap list, 
            // return the pointer of the first colormap by default.
            return _colormapCache.begin()->second;
        }
        else
        {
            // No colormap found and no colormap in cache, nothing to return.
            return nullptr;
        }
    }
    else
    {
        // It already exist, return the pointer of the texture object.
        return it->second;
    }
}

OGL::Texture * ColormapManagerGPU::loadColormap(OGL::Context* ctx, const std::vector<std::array<uint8_t, 3>>& rgb)
{
    // Create texture object
    OGL::Texture* tex = ctx->createTexture(OGL::texture1D);
    // Create the object with the current rendering API
    tex->create(ctx);
    // Allocate a buffer in gpu memory
    tex->allocate1D(static_cast<int>(rgb.size()), 3, DataType::unsigned_byte, reinterpret_cast<void*>(const_cast<uint8_t*>(&rgb[0][0])));
    // Set texture wrap parameter
    tex->setWrapMode(OGL::WrapMode::clamp_to_edge);
    return tex;
}

} // namespace OGL
