// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/data_set_tiled.h"

#include "core/copy.h"
#include "3d/mesh_creator.h"
#include "3d/grid.h"
#include "3d/meta.h"
#include "3d/helper.h"
#include "3d/texture.h"
#include "3d/static_mesh.h"
#include "3d/context.h"

struct Tile
{
    std::uint32_t* gcpx;
    std::uint32_t* gcpy;
    glm::ivec2 gcpSize;
    glm::ivec2 gcpOffset;
    glm::ivec2 dataOffset;
    float* lons;
    float* lats;
    Image data;
    Image mask;
    OGL::StaticMesh* mesh;
    OGL::Texture* dataTex;
    OGL::Texture* maskTex;
    unsigned int borders;

    Tile()
        : gcpx(nullptr), gcpy(nullptr), lons(nullptr), lats(nullptr), 
          mesh(nullptr), dataTex(nullptr), maskTex(nullptr), borders(0u)
    {}

    ~Tile()
    {
        delete[] gcpx;
        delete[] gcpy;
        delete[] lons;
        delete[] lats;
        delete mesh;
        delete dataTex;
        delete maskTex;
    }
};

DataSetTiled::DataSetTiled(const SEAScope::RenderableId& rId)
    : DataSet(rId)
{
}

DataSetTiled::~DataSetTiled()
{
    for
        (Tile* t : _tiles)
    {
        delete t;
    }
}

bool DataSetTiled::compute(int maxTextureSize)
{
    // Check pixels count between two gcps vs the size max of a texture.
    if 
        (this->gcpx[1] - this->gcpx[0] > static_cast<std::uint32_t>(maxTextureSize)
      || this->gcpy[this->lon_size] - this->gcpy[0] > static_cast<std::uint32_t>(maxTextureSize))
    {
        LOG(SEAScope::LogLevel::Error, "DataSetTiled::compute", "the max size of the texture ({}) is less than the number of pixels between two gcps ({}, {})", 
            maxTextureSize,
            this->gcpx[1] - this->gcpx[0], this->gcpy[this->lon_size] - this->gcpy[0]);
        return false;
    }

    if
        (0 < this->_tiles.size())
    {
        LOG(SEAScope::LogLevel::Error, "DataSetTiled::compute", "Already initialized");
        return false;
    }

    glm::ivec2 tileCount;
    tileCount.x = static_cast<int>(std::ceil(static_cast<float>(_width) / static_cast<float>(maxTextureSize)));
    tileCount.y = static_cast<int>(std::ceil(static_cast<float>(_height) / static_cast<float>(maxTextureSize)));

    const unsigned int lonSize = this->lon_size;
    const unsigned int latSize = this->lat_size;

    unsigned int tileLonSize = static_cast<unsigned int>(std::floor(static_cast<float>(lonSize) / static_cast<float>(tileCount.x)));
    unsigned int tileLatSize = static_cast<unsigned int>(std::floor(static_cast<float>(latSize) / static_cast<float>(tileCount.y)));

    glm::ivec2 tileShit(tileLonSize - 1, tileLatSize - 1);
    glm::ivec2 tileShitCounter(0, 0);

    tileCount.x = 0;
    unsigned int counter = 0;

    while (true)
    {
        if
            (counter >= lonSize - 1)
        {
            break;
        }

        counter += static_cast<unsigned int>(tileShit.x);
        ++tileCount.x;
    }

    tileCount.y = 0;
    counter = 0;

    while (true)
    {
        if
            (counter >= latSize - 1)
        {
            break;
        }

        counter += static_cast<unsigned int>(tileShit.y);
        ++tileCount.y;
    }

    const unsigned int count = static_cast<unsigned int>(tileCount.x * tileCount.y);
    _tiles.resize(count);

    unsigned int tileYBorders;
    unsigned int tileXBorders;

    constexpr unsigned int topBorder = 1<<0;
    constexpr unsigned int bottomBorder = 1<<1;
    constexpr unsigned int leftBorder = 1<<2;
    constexpr unsigned int rightBorder = 1<<3;

    for
        (int y = 0; y < tileCount.y; ++y)
    {
        tileShitCounter.x = 0;
        tileYBorders = 0;

        if
            (y > 0)
        {
            tileYBorders |= topBorder;
        }

        if
            (y < tileCount.y - 1)
        {
            tileYBorders |= bottomBorder;
        }

        for
            (int x = 0; x < tileCount.x; ++x)
        {
            tileXBorders = 0;

            if
                (x > 0)
            {
                tileXBorders |= leftBorder;
            }

            if
                (x < tileCount.x - 1)
            {
                tileXBorders |= rightBorder;
            }

            Tile* tile = new Tile;
            _tiles[static_cast<unsigned int>(y * tileCount.x + x)] = tile;

            tile->borders = tileXBorders | tileYBorders;

            tile->gcpSize.x = static_cast<int>(tileShitCounter.x + tileShit.x > static_cast<int>(lonSize) - 1 ? lonSize - static_cast<unsigned int>(tileShitCounter.x) : tileLonSize);
            tile->gcpSize.y = static_cast<int>(tileShitCounter.y + tileShit.y > static_cast<int>(latSize) - 1 ? latSize - static_cast<unsigned int>(tileShitCounter.y) : tileLatSize);

            tile->gcpOffset = tileShitCounter;

            const unsigned int gcpCount = static_cast<unsigned int>(tile->gcpSize.x * tile->gcpSize.y);
            tile->gcpx = new std::uint32_t[gcpCount];
            tile->gcpy = new std::uint32_t[gcpCount];

            memcpy2D_t<std::uint32_t>(
                const_cast<std::uint32_t*>(this->gcpx), tile->gcpOffset, glm::ivec2(lonSize, latSize), 1u,
                tile->gcpx, glm::ivec2(0, 0), tile->gcpSize,
                tile->gcpSize);

            memcpy2D_t<std::uint32_t>(
                const_cast<std::uint32_t*>(this->gcpy), tile->gcpOffset, glm::ivec2(lonSize, latSize), 1u,
                tile->gcpy, glm::ivec2(0, 0), tile->gcpSize,
                tile->gcpSize);

            // Matrix data
            glm::ivec2 origin(tile->gcpx[0], tile->gcpy[0]);
            glm::ivec2 region(tile->gcpx[tile->gcpSize.x - 1] - tile->gcpx[0], 
                              tile->gcpy[gcpCount - 1] - tile->gcpy[0]);

            // Add borders start
            if
                (tile->borders & leftBorder)
            {
                origin.x = origin.x - 1;
                region.x++;
            }

            if
                (tile->borders & rightBorder)
            {
                region.x++;
            }

            if
                (tile->borders & topBorder)
            {
                origin.y = origin.y - 1;
                region.y++;
            }

            if
                (tile->borders & bottomBorder)
            {
                region.y++;
            }
            // Add borders end


            this->_data->copyTo(origin, region, tile->data);
            this->_dataMask->copyTo(origin, region, tile->mask);

            std::uint32_t offset = tile->gcpx[0];
            std::uint32_t* ptr = tile->gcpx;

            // Add borders start
            if
                (tile->borders & leftBorder)
            {
                offset--;
            }
            // Add borders end

            for
                (size_t i = 0; i < gcpCount; ++i)
            {
                *ptr = *ptr - offset;
                ++ptr;
            }
            
            offset = tile->gcpy[0];
            ptr = tile->gcpy;

            // Add borders start
            if
                (tile->borders & topBorder)
            {
                offset--;
            }
            // Add borders end

            for
                (size_t i = 0; i < gcpCount; ++i)
            {
                *ptr = *ptr - offset;
                ++ptr;
            }

            tile->lons = new float[gcpCount];
            tile->lats = new float[gcpCount];

            memcpy2D_t<float>(
                const_cast<float*>(this->lons), tile->gcpOffset, glm::ivec2(lonSize, latSize), 1u,
                tile->lons, glm::ivec2(0, 0), tile->gcpSize,
                tile->gcpSize);

            memcpy2D_t<float>(
                const_cast<float*>(this->lats), tile->gcpOffset, glm::ivec2(lonSize, latSize), 1u,
                tile->lats, glm::ivec2(0, 0), tile->gcpSize,
                tile->gcpSize);

            tileShitCounter.x += tileShit.x;
        }

        tileShitCounter.y += tileShit.y;
    }

    return true;
}

void DataSetTiled::generateGPUData(OGL::Context* ctx)
{
    if
        (   0 < _tiles.size()
         && nullptr != _tiles[0]->dataTex)
    {
        LOG(SEAScope::LogLevel::Error, "DataSetTiled::generateGPUData", "Already initialized");
        return;
    }

    for
        (Tile* tile : _tiles)
    {
        std::vector<glm::vec2> uv;
        std::vector<glm::vec3> gcps;
        std::vector<unsigned int> indices;

        MeshCreator mesh_creator_globe;

        createGridFromGCP(
			tile->lons, 
			static_cast<unsigned int>(tile->gcpSize.x), 
			tile->lats, 
			static_cast<unsigned int>(tile->gcpSize.y),
			tile->gcpx, 
			tile->gcpy, 
			tile->data.width(), 
			tile->data.height(), 
			gcps, 
			uv, 
			indices);

        mesh_creator_globe.setDrawType(OGL::draw_elements, OGL::triangles);
        mesh_creator_globe.setVertices(makeArray(gcps));
        mesh_creator_globe.setTextureCoordinates(makeArray(uv));
        mesh_creator_globe.setIndices(makeArray(indices));

        Image& finalImage = tile->data;
        Image& finalImageMask = tile->mask;

        OGL::Texture* colors = OGL::Helper::createTexture(ctx, finalImage);
        colors->setWrapAndFilterMode(OGL::clamp_to_edge, OGL::linear);
        mesh_creator_globe.setTexture(colors, 0);

        OGL::Texture* mask = OGL::Helper::createTexture(ctx, finalImageMask);
        mask->setWrapAndFilterMode(OGL::clamp_to_edge, OGL::nearest);
        mesh_creator_globe.setTexture(mask, 1);

        tile->dataTex = colors;
        tile->maskTex = mask;

        tile->mesh = mesh_creator_globe.createStaticMesh(ctx);

        tile->data.deallocate();
        tile->mask.deallocate();
    }
}

void DataSetTiled::draw(OGL::Context* ctx, const glm::mat4&)
{
    for
        (Tile* tile : _tiles)
    {
        tile->mesh->bindTextures(ctx);
        tile->mesh->draw();
    }
}

void DataSetTiled::drawPicking(OGL::Context* ctx, const glm::mat4&)
{
    for
        (Tile* tile : _tiles)
    {
        // Bind mask data to texture unit 0
        ctx->activeTextureUnit(0, tile->mesh->getTexture(1));

        tile->mesh->draw();
    }

}

void DataSetTiled::setFilterMode(OGL::FilterMode filterMode)
{
    for
        (Tile* tile : _tiles)
    {
        // Get texture which store physical data
        OGL::Texture* texture = tile->mesh->getTexture(0);

        if (nullptr != texture)
        {
            texture->setFilterMode(filterMode);
        }
    }
}
