// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/dataset_visibility_globe.h"

#include "3d/scene_graph_controller.h"
#include "3d/scene_graph.h"
#include "3d/scene_graph_node.h"

#include "3d/data_set.h"
#include "3d/camera.h"
#include "3d/globe_visibility.h"
#include "3d/ellipsoid.h"
#include "3d/spherical_math.h"
#include "3d/spherical_quad.h"

DataSetVisibilityGlobe::DataSetVisibilityGlobe()
:   DataSetVisibility()
{
}

DataSetVisibilityGlobe::~DataSetVisibilityGlobe()
{
}

void DataSetVisibilityGlobe::update()
{
    SceneGraph* sg = _sgc->getSceneGraph();
    
    std::vector<Node*> nodes;
    sg->getNodes(renderer_uv_streamline, nodes);
    sg->getNodes(renderer_barb, nodes);
    sg->getNodes(renderer_arrow, nodes);
    
    // Collection loop
    for (Node* node : nodes)
    {
        auto list = node->children();
        
        // DataSet loop
        for (Node* child : *list)
        {
            GranuleNode* gnode = dynamic_cast<GranuleNode*>(child);
            
            if (gnode)
            {
                DataSet* dataSet = dynamic_cast<DataSet*>(gnode->renderable());
                
                if (dataSet)
                {
                    computeDataSetVisibility(dataSet);
                }
            }
        }
    }
}

void DataSetVisibilityGlobe::computeDataSetVisibility(DataSet* dataSet)
{
    // Cast to globe visibility
    GlobeVisibility* gv = static_cast<GlobeVisibility*>(_world);
    
    // Get the visible quad of the globe
    SphericalQuad* sq = gv->getVisibleQuad();
    
    // Get gcps grid information
    unsigned int latSize = dataSet->getLat_size();
    unsigned int lonSize = dataSet->getLon_size();
    
    /*!
     * 0 = top left
     * 1 = bottom right
     * 2 = top right
     * 3 = bottom left
     */
    const glm::vec3& p0 = sq->getCornerPosition(0);
    const glm::vec3& p1 = sq->getCornerPosition(1);
    const glm::vec3& p2 = sq->getCornerPosition(2);
    const glm::vec3& p3 = sq->getCornerPosition(3);

    // Variables
    unsigned int id;
    bool oneInRowVisible;
    std::vector<glm::ivec4> areas;
    bool createNewArea = true;
    bool areaInCreation = false;
    glm::ivec4 area;
    
    auto& gcps = dataSet->getGcps();
    
    // Loop on all gcps, brute force mode
    for (unsigned int lat = 0; lat < latSize; ++lat)
    {
        // Boolean to know if a row of gcps is visible or not
        oneInRowVisible = false;
        
        for (unsigned int lon = 0; lon < lonSize; ++lon)
        {
            // Compute current id 2D to 1D
            id = lat * lonSize + lon;
            
            // Compute visibility for the current gcp
            if (isPointInInfiniteUnitPyramidCentered(gcps[id], p0, p2, p1, p3))
            {
                // The gcp is visible, update the boolean for the row
                if (false == oneInRowVisible)
                {
                    oneInRowVisible = true;
                }
                
                // Check if its a new area. First gcp visible or we are leaving a not visible area.
                if (true == createNewArea)
                {
                    // Create new area with min == max
                    area = glm::ivec4(lon, lat, lon, lat);
                    // update booleans
                    areaInCreation = true;
                    createNewArea = false;
                }
                else
                {
                    // update the current area
                    
                    if (static_cast<int>(lon) < area.x)
                    {
                        area.x = static_cast<int>(lon);
                    }
                    
                    if (static_cast<int>(lat) < area.y)
                    {
                        area.y = static_cast<int>(lat);
                    }
                    
                    if (static_cast<int>(lon) > area.z)
                        area.z = static_cast<int>(lon);
                    
                    if (static_cast<int>(lat) > area.w)
                        area.w = static_cast<int>(lat);
                }
            }
        }
        
        // Check if at least one gcp is visible in the last row
        if (false == oneInRowVisible)
        {
            // We are in an hidden area, set the boolean to true to create a new area if a gcp is visible again.
            createNewArea = true;
            
            // Check if an area was in creation
            if (true == areaInCreation)
            {
                // Add a border to include parts between visible and not visible gcps.
                area.x = std::max(0, area.x - 1);
                area.y = std::max(0, area.y - 1);
                area.z = std::min(static_cast<int>(dataSet->getLon_size()) - 1, area.z + 1);
                area.w = std::min(static_cast<int>(dataSet->getLat_size()) - 1, area.w + 1);
        
                // Yes, so push the current area
                areas.push_back(area);
                
                // Clear the boolean
                areaInCreation = false;
            }
        }
    }
    
    // Check if an area was in creation
    if (true == areaInCreation)
    {
        // Add a border to include parts between visible and not visible gcps.
        area.x = std::max(0, area.x - 1);
        area.y = std::max(0, area.y - 1);
        area.z = std::min(static_cast<int>(dataSet->getLon_size()) - 1, area.z + 1);
        area.w = std::min(static_cast<int>(dataSet->getLat_size()) - 1, area.w + 1);
        
        // Yes, so push the current area
        areas.push_back(area);
    }

    auto oldAreas = dataSet->getVisibleArea();
    bool diff = false;

    if
        (oldAreas.size() != areas.size())
    {
        diff = true;
    }
    else
    {
		for 
            (size_t i = 0; i < oldAreas.size(); ++i)
		{
            if
                (oldAreas[i] != areas[i])
            {
                diff = true;
            }
		}
    }

    if
        (true == diff)
    {
		// Update the visible areas of the current dataset
		dataSet->setVisibleAreas(std::move(areas));
    }
}
