// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/ear_clipping.h"

#include "3d/spherical_math.h"

#include "num_utils.h"

namespace
{
	bool vec3_equals(const glm::vec3& a, const glm::vec3& b)
	{
		return SEAScope::floating_point_equality(a.x, b.x)
			&& SEAScope::floating_point_equality(a.y, b.y)
			&& SEAScope::floating_point_equality(a.z, b.z);
	}
}
	

void EarClipping::run(const std::vector<glm::vec3>& points0, std::vector<unsigned int>& outIndices)
{
	std::vector<glm::vec3> points;

	bool exists;

	// Remove duplicate points
	for (size_t j = 0; j < points0.size(); ++j)
	{
		exists = false;

		for (size_t i = 0; i < points.size(); ++i)
		{
			if (true == vec3_equals(points0[j], points[i]))
			{
				exists = true;
				break;
			}
		}

		if (false == exists)
		{
			points.push_back(points0[j]);
		}
	}
	

	// Find the orientation used during the creation.
	int orientation = computeOrientation(points);

	outIndices.reserve((points.size() - 2) * 3);

	if (orientation > 0)
	{
		// Fill list id
		for (unsigned int i = 0; i < points.size(); ++i)
		{
			_indices.push_back(i);
		}
	}
	else
	{
		for (int i = static_cast<int>(points.size()) - 1; i >= 0; --i)
		{
			_indices.push_back(static_cast<unsigned int>(i));
		}
	}

	for (unsigned int i = 0; i < _indices.size(); ++i)
	{
		p3 p;
		p.ear = false;
		p.curr = i;

		if (i == 0)
		{
			p.prev = static_cast<unsigned int>(_indices.size()) - 1;
		}
		else
		{
			p.prev = i - 1;
		}

		p.next = (i + 1) > (_indices.size() - 1) ? 0 : (i + 1);

		_pointsArray.push_back(p);
	}

	size_t pointCount = _indices.size();

	for (size_t i = 0; i < pointCount; ++i)
	{
		updatePoint(points, _pointsArray[i]);
	}

	std::vector<p3> out;

	while (false == _ears.empty())
	{
		auto itEar = _ears.begin();

		p3& p = *itEar;
		p3& pPrev = _pointsArray[p.prev];
		p3& pNext = _pointsArray[p.next];

		pPrev.next = pNext.curr;
		pNext.prev = pPrev.curr;

		if (pPrev.next != pNext.prev)
		{
			out.push_back(p);

			updatePoint(points, pPrev);
			updatePoint(points, pNext);
		}

		_ears.erase(itEar);
	}

	for (p3& p : out)
	{
		outIndices.push_back(_indices[p.prev]);
		outIndices.push_back(_indices[p.next]);
		outIndices.push_back(_indices[p.curr]);
	}
}

void EarClipping::updatePoint(const std::vector<glm::vec3>& points, p3& p)
{
	const glm::vec3& a = points[_indices[p.prev]];
	const glm::vec3& b = points[_indices[p.curr]];
	const glm::vec3& c = points[_indices[p.next]];

	if (true == isConvex(a, b, c))
	{
		p.convex = true;

		bool isEar = true;

		for (size_t j = 0; j < points.size(); ++j)
		{
			if (j != _indices[p.prev] &&
				j != _indices[p.curr] &&
				j != _indices[p.next])
			{
				if (true == isPointInInfiniteTetrahedron(points[j], glm::vec3(0.f, 0.f, 0.f), a, b, c))
				{
					isEar = false;
					break;
				}
			}
		}

		if (true == _pointsArray[p.prev].ear ||
			true == _pointsArray[p.next].ear)
		{
			isEar = false;
		}

		if (true == isEar)
		{
			p.ear = true;
			_ears.push_back(p);
		}
	}
	else
	{
		p.convex = false;
	}
}

/*!
 *  \brief Count the convex and concave angles.
 *  \return 1 if the polygon is clockwise and -1 if it is counterclockwise.
 */
int EarClipping::computeOrientation(const std::vector<glm::vec3>& points)
{
	size_t pointCount = points.size();

	glm::vec3 sum(0.f, 0.f, 0.f);

	for (size_t i = 0; i < pointCount; ++i)
	{
		sum += glm::cross(points[i], points[(i + 1) % pointCount]);
	}

	sum *= 0.5f;

	float d = glm::dot(glm::normalize(points[0]), sum);

	return d > 0.f ? -1 : 1;
}

/*!
 *  \brief Compute if three points on a globe form a convex angle in a clockwise orientation.
 */
bool EarClipping::isConvex(const glm::vec3& p0, const glm::vec3& p1, const glm::vec3& p2)
{
	glm::vec3 u = p1 - p0;
	glm::vec3 v = p2 - p1;

	return glm::dot(glm::cross(v, u), p1) >= 0.f;
}
