// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/ellipse.h"

#include "core/types.h"

#include "3d/spherical_math.h"
#include "3d/shapes_allocator.h"

Ellipse::Ellipse()
: RectangleBase{SEAScope::ShapeType::ELLIPSE},
  _center{0, 0, 0}, _betaPoint{0, 0, 0}, _alphaPoint{0, 0, 0},
  _alphaAngle{0.f}, _betaAngle{0.f}
{
}

Ellipse::~Ellipse()
{
}

Ellipse* Ellipse::clone() const
{
	Ellipse* e = ShapesAllocator::getShape<Ellipse>();

	e->_center = this->_center;
	e->_alphaAngle = this->_alphaAngle;
	e->_betaAngle = this->_betaAngle;
	e->_alphaPoint = this->_alphaPoint;
	e->_betaPoint = this->_betaPoint;
	e->_rotationGlobe = this->_rotationGlobe;
	e->_rotationZ = this->_rotationZ;

	return e;
}

void Ellipse::initialize(const glm::vec3& center, float alphaAngle, float betaAngle)
{
	/*
		The ellipse formula compute ellipse's points centered around north pole ((0, 0, 1) in seascope coordinate scene).
		First, We will compute the rotation to move the points around another position on the globe.
		Then we will compute another rotation to always have the PI / 2 ellipse's point targeting the north pole.
	*/

	this->_center = center;
	this->_alphaAngle = alphaAngle;
	this->_betaAngle = betaAngle;

	this->_rotationGlobe = glm::rotation(glm::vec3(0.f, 0.f, 1.f), this->_center);
	
	const glm::vec3 northPole(0.f, 0.f, 1.f);

	glm::vec3 vCenterToNP = glm::normalize(northPole - this->_center);

	// Compute coordinates system at north pole
	glm::vec3 vPoleX = glm::normalize(glm::cross(vCenterToNP, northPole));

	// Original coordinates system of the ellipse
	const glm::vec3 vSourceX(1.f, 0.f, 0.f);

	this->_rotationZ = glm::rotation(vSourceX, vPoleX);

	this->_alphaPoint = computePoint(0.f);
	this->_betaPoint = computePoint(static_cast<float>(M_PI_2));
}

void Ellipse::initialize(const glm::vec3& center, const glm::vec3& alphaPoint, const glm::vec3& betaPoint)
{
	this->_center = center;
	this->_alphaPoint = alphaPoint;
	this->_betaPoint = betaPoint;

	this->_rotationGlobe = glm::rotation(glm::vec3(0.f, 0.f, 1.f), this->_center);

	this->_alphaAngle = glm::angle(glm::rotation(this->_center, alphaPoint));
	this->_betaAngle = glm::angle(glm::rotation(this->_center, betaPoint));

	const glm::vec3 northPole(0.f, 0.f, 1.f);

	glm::quat qCenterToNPole = glm::rotation(this->_center, northPole);
	glm::vec3 vXPole = qCenterToNPole * this->_alphaPoint;
	glm::vec3 vYPole = glm::normalize(glm::cross(northPole, vXPole));

	// Original coordinates system of the ellipse
	const glm::vec3 vSourceY(0.f, 1.f, 0.f);

	this->_rotationZ = glm::rotation(vSourceY, vYPole);
}

void Ellipse::setCenter(const glm::vec3& val)
{
	glm::vec3 previousCenter = this->_center;
	this->_center = val;

	this->_rotationGlobe = glm::rotation(glm::vec3(0.f, 0.f, 1.f), this->_center);

	glm::quat q = glm::rotation(previousCenter, this->_center);

	this->_alphaPoint = q * this->_alphaPoint;
	this->_betaPoint = q * this->_betaPoint;
}

void Ellipse::setAlphaPoint(const glm::vec3& val)
{
	this->_alphaPoint = val;

	const glm::vec3 northPole(0.f, 0.f, 1.f);

	glm::quat qCenterToNPole = glm::rotation(this->_center, northPole);
	glm::vec3 vXPole = qCenterToNPole * this->_alphaPoint;
	glm::vec3 vYPole = glm::normalize(glm::cross(northPole, vXPole));

	// Original coordinates system of the ellipse
	const glm::vec3 vSourceY(0.f, 1.f, 0.f);

	this->_rotationZ = glm::rotation(vSourceY, vYPole);
	this->_betaPoint = computePoint(static_cast<float>(M_PI_2));

	this->_alphaAngle = glm::angle(glm::rotation(this->_center, this->_alphaPoint));
}

void Ellipse::setBetaPoint(const glm::vec3& val)
{
	this->_betaPoint = val;

	const glm::vec3 northPole(0.f, 0.f, 1.f);

	glm::quat qCenterToNPole = glm::rotation(this->_center, northPole);
	glm::vec3 vYPole = qCenterToNPole * this->_betaPoint;
	glm::vec3 vXPole = glm::normalize(glm::cross(vYPole, northPole));

	// Original coordinates system of the ellipse
	const glm::vec3 vSourceX(1.f, 0.f, 0.f);

	this->_rotationZ = glm::rotation(vSourceX, vXPole);
	this->_alphaPoint = computePoint(0.f);

	this->_betaAngle = glm::angle(glm::rotation(this->_center, this->_betaPoint));
}

glm::vec3 Ellipse::computePoint(float a, float alpha, float beta)
{
	glm::vec3 p;

	p.x = sinf(alpha) * cosf(a);
	p.y = sinf(beta) * sinf(a);
	p.z = sqrtf(1.f - (powf(sinf(alpha), 2.f) * powf(cosf(a), 2.f)) - (powf(sinf(beta), 2.f) * powf(sinf(a), 2.f)));

	return p;
}

glm::vec3 Ellipse::computePoint(float a)
{
	glm::vec3 p = computePoint(a, this->_alphaAngle, this->_betaAngle);

	p = p - glm::vec3(0.f, 0.f, 1.f);
	p = this->_rotationZ * p;
	p += glm::vec3(0.f, 0.f, 1.f);

	p = this->_rotationGlobe * p;

	return p;
}
