// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/ellipse_editor.h"

#include "3d/ellipse.h"
#include "3d/ellipse_builder.h"

#include "3d/core_objects.h"
#include "3d/geodetic.h"
#include "3d/ellipsoid.h"
#include "3d/geodetic.h"
#include "3d/spherical_math.h"
#include "3d/math_func.h"
#include "3d/geometry.h"
#include "3d/pickable.h"

#define CENTER	0
#define X		1
#define Y		2

EllipseEditor::EllipseEditor()
: ShapeEditor{},
  _program{nullptr},
  _buffer{nullptr},
  _vao{nullptr},
  _drawMode{nullptr},
  _pointCount{0u},
  _currentPos{0, 0, 0}
{
}

EllipseEditor::~EllipseEditor()
{
}

void EllipseEditor::initialize(OGL::Context* ctx, Pickable* obj)
{
	_shapeToEdit = obj;

	_program = OGL::Helper::createProgram(ctx, vertex_line_tesselate, fragment_uni_color);

	_program->bind();
	_program->setUniformByName("color", glm::vec4(1.f, 0.f, 0.f, 1.f));
	_program->unbind();

	const int segmentCount = 20;
	std::vector<float> percents;
	percents.reserve(segmentCount);

	for (int i = 0; i < segmentCount; ++i)
	{
		percents.push_back(static_cast<float>(i) / (segmentCount - 1));
	}

	_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(percents));
	OGL::VertexAttribArray* vaa = ctx->createVertexAttribArray(0, 1, _buffer, DataType::float32);

	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(vaa);
	_vao->fill();

	_drawMode = OGL::Helper::createDrawModeArrays(ctx, OGL::line_strip, segmentCount);

	_pointCount = 50;
	_points.resize(_pointCount + 1);

	Ellipse* p = static_cast<Ellipse*>(_shapeToEdit->getShape());

	_anchorsPosSource[CENTER] = p->getCenter();
	_anchorsPosSource[X] = p->getAlphaPoint();
	_anchorsPosSource[Y] = p->getBetaPoint();

	_anchorsPos[CENTER] = p->getCenter();
	_anchorsPos[X] = p->getAlphaPoint();
	_anchorsPos[Y] = p->getBetaPoint();
}

void EllipseEditor::computePoints()
{
	Ellipse* e = static_cast<Ellipse*>(_shapeToEdit->getShape());

	float interval = static_cast<float>(M_PI * 2 / (_pointCount - 1));
	float a;

	for (size_t i = 0; i < _pointCount; ++i)
	{
		a = interval * static_cast<float>(i);
		_points[i] = e->computePoint(a);
	}
}

void EllipseEditor::draw(OGL::Context* /*ctx*/, const glm::mat4& mvp)
{
	_program->bind();

	_program->setUniformByName("mvp", mvp);
	_program->setUniformByName("color", green4f);

	_vao->bind();

	for (unsigned int i = 0; i < _pointCount; ++i)
	{
		_program->setUniformByName("p0", _points[i]);
		_program->setUniformByName("p1", _points[i + 1]);

		_drawMode->draw();
	}

	_vao->unbind();
}

OGL::Shape* EllipseEditor::createShape()
{
	return this->_shapeToEdit->getShape()->clone();
}

void EllipseEditor::setBuilder(ShapeBuilder* builder)
{
	// takes ownership
	this->_builder = builder;
}

unsigned int EllipseEditor::getAnchorCount() const
{
	return 3;
}

glm::vec3 EllipseEditor::getAnchorPos(AnchorType, unsigned int vertexId) const
{
	return _anchorsPos[vertexId];
}

void EllipseEditor::move(AnchorType /*type*/, unsigned int vertexId, const glm::vec3& pos)
{
	//glm::vec3 lastPos = _anchorsPos[vertexId];
	_anchorsPos[vertexId] = pos;

	if
		(0 == vertexId)
	{
		Ellipse* e = static_cast<Ellipse*>(_shapeToEdit->getShape());

		e->setCenter(pos);

		_anchorsPos[X] = e->getAlphaPoint();
		_anchorsPos[Y] = e->getBetaPoint();
	}
	else if
		(1 == vertexId)
	{
		Ellipse* e = static_cast<Ellipse*>(_shapeToEdit->getShape());

		e->setAlphaPoint(pos);

		_anchorsPos[X] = e->getAlphaPoint();
		_anchorsPos[Y] = e->getBetaPoint();
	}
	else if
		(2 == vertexId)
	{
		Ellipse* e = static_cast<Ellipse*>(_shapeToEdit->getShape());

		e->setBetaPoint(pos);

		_anchorsPos[X] = e->getAlphaPoint();
		_anchorsPos[Y] = e->getBetaPoint();
	}

	computePoints();
}

/**
 *
 */
void EllipseEditor::move(AnchorType type, unsigned int vertexId)
{
    this->move(type, vertexId, this->_currentPos);
}

/**
 *
 */
void EllipseEditor::updateCursorPosition(const glm::vec3& position)
{
    this->_currentPos = position;
}

/**
 *
 */
void EllipseEditor::startMove(unsigned int)
{

}

/**
 *
 */
void EllipseEditor::stopMove()
{

}

/**
 *
 */
void EllipseEditor::pressed(AnchorType, unsigned int)
{

}

/**
 *
 */
void EllipseEditor::remove(AnchorType /*type*/, unsigned int /*vertexId*/)
{

}

/**
 *
 */
bool EllipseEditor::shapeIsValid() const
{
    return true;
}
