// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/ellipse_mesh.h"

#include "3d/core_objects.h"
#include "3d/ellipse.h"
#include "3d/math_func.h"
#include "3d/line.h"
#include "3d/trajectory_color.h"
#include "3d/data_rendering_param.h"
#include "3d/shapes_allocator.h"


EllipseMesh::EllipseMesh()
: ShapeMesh{SEAScope::RenderableId()},
  _verticeBuffer{nullptr},
  _directionBuffer{nullptr},
  _indicesBuffer{nullptr},
  _vao{nullptr},
  _drawMode{nullptr}
{
	_flags.set(RenderableFlags::editable, true);
	_flags.set(RenderableFlags::erasable, false);
}

EllipseMesh::~EllipseMesh()
{
	delete _verticeBuffer;
	_verticeBuffer = nullptr;
	delete _directionBuffer;
	_directionBuffer = nullptr;
	delete _indicesBuffer;
	_indicesBuffer = nullptr;
	delete _vao;
	_vao = nullptr;
}

void EllipseMesh::create(OGL::Context* ctx, Ellipse* ellipse)
{
	_shape = ellipse;

	std::vector<glm::vec3> points;

	computePoints(ellipse, points);

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Compute mesh
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////

	std::vector<glm::vec3> vertices;
	std::vector<glm::vec2> direction;
	std::vector<unsigned int> indices;

	polylineOnSphereIndices2(points, vertices, direction, indices);

	// GPU objects
	_verticeBuffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(vertices));
	_directionBuffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(direction));
	_indicesBuffer = OGL::Helper::createBuffer(ctx, OGL::element_array_buffer, makeArray(indices));

	const int positionLoc = 0;
	const int previousLoc = 1;
	const int nextLoc = 2;
	const int directionLoc = 3;

	OGL::VertexAttribArray* verticesAttrib0 = ctx->createVertexAttribArray(positionLoc, 3, _verticeBuffer, DataType::float32, 0, sizeof(glm::vec3) * 4);
	OGL::VertexAttribArray* verticesAttrib1 = ctx->createVertexAttribArray(previousLoc, 3, _verticeBuffer, DataType::float32, 0, 0);
	OGL::VertexAttribArray* verticesAttrib2 = ctx->createVertexAttribArray(nextLoc, 3, _verticeBuffer, DataType::float32, 0, sizeof(glm::vec3) * 8);
	OGL::VertexAttribArray* directionAttrib = ctx->createVertexAttribArray(directionLoc, 2, _directionBuffer);

	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(verticesAttrib0);
	_vao->addObject(verticesAttrib1);
	_vao->addObject(verticesAttrib2);
	_vao->addObject(directionAttrib);
	_vao->addObject(_indicesBuffer);
	_vao->fill();

	unsigned int vertice_count = static_cast<unsigned int>(indices.size());

	_drawMode = OGL::Helper::createDrawModeElements(ctx, OGL::triangles, DataType::unsigned_int, vertice_count);
}

void EllipseMesh::computePoints(Ellipse* e, std::vector<glm::vec3>& points)
{
	const size_t pointCount = 50;  // Number of points in the ellipse
	float intervalAngle = static_cast<float>((M_PI * 2.0) / (pointCount - 1)); // Interval angle between two points

	if 
		(points.size() != pointCount)
	{
		points.resize(pointCount);
	}

	// Compute ellipse points centered around Z axis
	for (size_t i = 0; i < pointCount; ++i)
	{
		float a = static_cast<float>(i) * intervalAngle;
		points[i] = e->computePoint(a);
	}
}

void EllipseMesh::update(Ellipse* ellipse)
{
	if 
		(nullptr != ellipse)
	{
		ShapesAllocator::release(_shape);

		_shape = ellipse;

		std::vector<glm::vec3> points;

		computePoints(ellipse, points);

		std::vector<glm::vec3> vertices;

		polylineOnSphereIndices2Update(points, vertices);

		this->_verticeBuffer->copy(vertices.data());
	}
}

void EllipseMesh::draw(OGL::Context* /*ctx*/, const glm::mat4& /*mvp*/)
{
	_vao->bind();

	_drawMode->draw();

	_vao->unbind();
}

void EllipseMesh::drawPicking(OGL::Context* /*ctx*/, const glm::mat4& /*mvp*/)
{
	_vao->bind();

	_drawMode->draw();

	_vao->unbind();
}
