// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/ellipse_solid_mesh.h"

#include "3d/core_objects.h"
#include "3d/ellipse.h"
#include "3d/geometry.h"
#include "3d/shapes_allocator.h"

EllipseSolidMesh::EllipseSolidMesh()
	: ShapeMesh(SEAScope::RenderableId())
{
	this->_buffer = nullptr;
	this->_indicesBuffer = nullptr;
	this->_vao = nullptr;
	this->_drawMode = nullptr;
	this->_shape = nullptr;

	this->_flags.set(RenderableFlags::editable, true);
	this->_flags.set(RenderableFlags::erasable, false);
}

EllipseSolidMesh::~EllipseSolidMesh()
{
	delete this->_buffer;
	this->_buffer = nullptr;

	delete this->_indicesBuffer;
	this->_indicesBuffer = nullptr;

	delete this->_vao;
	this->_vao = nullptr;

	delete this->_drawMode;
	this->_drawMode = nullptr;

	this->_shape = nullptr;   // deleted elsewhere
}

void EllipseSolidMesh::create(OGL::Context* ctx, Ellipse* ellipse)
{
	this->_shape = ellipse;

	std::vector<glm::vec3> pointsBase;
	std::vector<unsigned int> indicesbase;

	// Compute ellipse perimeter points
	computePoints(ellipse, pointsBase);

	// Add center point
	pointsBase.push_back(ellipse->getCenter());
	size_t centerId = pointsBase.size() - 1;

	// Compute indices array to create triangles around the center points.
	for (size_t i = 0; i < pointsBase.size(); ++i)
	{
		indicesbase.push_back(static_cast<unsigned int>(i));
		indicesbase.push_back(static_cast<unsigned int>(i+1u));
		indicesbase.push_back(static_cast<unsigned int>(centerId));
	}

	std::vector<glm::vec3> points;
	std::vector<unsigned int> indices;

	// Split the triangles into smaller triangles
	subdivise(pointsBase, indicesbase, 5.f, points, indices);

	if
		(nullptr != this->_buffer)
	{
		delete this->_buffer;
		this->_buffer = nullptr;
	}

	this->_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(points));
	OGL::VertexAttribArray* vaa = ctx->createVertexAttribArray(0, 3, this->_buffer, DataType::float32);

	if
		(nullptr != this->_indicesBuffer)
	{
		delete this->_indicesBuffer;
		this->_indicesBuffer = nullptr;
	}
	this->_indicesBuffer = OGL::Helper::createBuffer(ctx, OGL::element_array_buffer, makeArray(indices));

	if
		(nullptr != this->_vao)
	{
		delete this->_vao;
		this->_vao = nullptr;
	}
	this->_vao = ctx->createVertexArrayObject();
	this->_vao->create(ctx);
	this->_vao->addObject(vaa);
	this->_vao->addObject(this->_indicesBuffer);
	this->_vao->fill();

	if
		(nullptr != this->_drawMode)
	{
		delete this->_drawMode;
		this->_drawMode = nullptr;
	}
	this->_drawMode = OGL::Helper::createDrawModeElements(ctx, OGL::triangles, DataType::unsigned_int, static_cast<unsigned int>(indices.size()));

	setStyle(normal);
}

void EllipseSolidMesh::computePoints(Ellipse* e, std::vector<glm::vec3>& points)
{
	const size_t pointCount = 50;  // Number of points in the ellipse
	float intervalAngle = static_cast<float>((M_PI * 2.0) / (pointCount - 1)); // Interval angle between two points

	if
		(points.size() != pointCount)
	{
		points.resize(pointCount);
	}

	// Compute ellipse points centered around Z axis
	for (size_t i = 0; i < pointCount; ++i)
	{
		float a = static_cast<float>(i) * intervalAngle;
		points[i] = e->computePoint(a);
	}
}

void EllipseSolidMesh::update(Ellipse* ellipse)
{
	if
		(nullptr != ellipse)
	{
		ShapesAllocator::release(this->_shape);

		this->_shape = ellipse;

		std::vector<glm::vec3> pointsBase;
		std::vector<unsigned int> indicesbase;

		// Compute ellipse perimeter points
		computePoints(ellipse, pointsBase);

		// Add center point
		pointsBase.push_back(ellipse->getCenter());
		size_t centerId = pointsBase.size() - 1;

		// Compute indices array to create triangles around the center points.
		for (size_t i = 0; i < pointsBase.size(); ++i)
		{
			indicesbase.push_back(static_cast<unsigned int>(i));
			indicesbase.push_back(static_cast<unsigned int>(i + 1u));
			indicesbase.push_back(static_cast<unsigned int>(centerId));
		}

		std::vector<glm::vec3> points;
		std::vector<unsigned int> indices;

		// Split the triangles into smaller triangles
		subdivise(pointsBase, indicesbase, 5.f, points, indices);

		this->_buffer->allocate(makeArray(points));
		this->_indicesBuffer->allocate(makeArray(indices));

		this->_drawMode->setElementCount(static_cast<unsigned int>(indices.size()));
	}
}

void EllipseSolidMesh::draw(OGL::Context* ctx, const glm::mat4& /*mvp*/)
{
	ctx->setFlag(OGL::cull_face, false);

	this->_vao->bind();

	this->_drawMode->draw();

	this->_vao->unbind();
}

void EllipseSolidMesh::drawPicking(OGL::Context* ctx, const glm::mat4& /*mvp*/)
{
	ctx->setDepthMask(false);

	this->_vao->bind();

	this->_drawMode->draw();

	this->_vao->unbind();

	ctx->setDepthMask(true);
}
