// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   Ellipsoid.cpp
 * Author: gleseach
 * 
 * Created on 25 août 2015, 09:55
 */

#include "3d/stdafx.h"
#include "3d/ellipsoid.h"

#include "3d/math_func.h"

typedef glm::vec3 Vector3D;

//Ellipsoid Ellipsoid::Wgs84         = Ellipsoid(6378137.0, 6378137.0, 6356752.314245);
//Ellipsoid Ellipsoid::ScaledWgs84   = Ellipsoid(1.0, 1.0, 6356752.314245 / 6378137.0);
//Ellipsoid Ellipsoid::UnitSphere    = Ellipsoid(1.0, 1.0, 1.0);

Ellipsoid::Ellipsoid(double x, double y, double z)
    : Ellipsoid(Vector3D(x, y, z))
{
}

Ellipsoid::Ellipsoid(const Vector3D& radii)
{
    if ((radii.x <= 0.0f) || (radii.y <= 0.0f) || (radii.z <= 0.0f))
    {
        throw "radii";
    }

    _radii = radii;
    _radiiSquared = Vector3D(
        radii.x * radii.x,
        radii.y * radii.y,
        radii.z * radii.z);
    _radiiToTheFourth = Vector3D(
        _radiiSquared.x * _radiiSquared.x,
        _radiiSquared.y * _radiiSquared.y,
        _radiiSquared.z * _radiiSquared.z);
    _oneOverRadiiSquared = Vector3D(
        1.f / (radii.x * radii.x),
        1.f / (radii.y * radii.y),
        1.f / (radii.z * radii.z));
}

Vector3D Ellipsoid::CentricSurfaceNormal(const Vector3D& positionOnEllipsoid)
{
    return glm::normalize(positionOnEllipsoid);
}

Vector3D Ellipsoid::GeodeticSurfaceNormal(const Vector3D& positionOnEllipsoid)
{
    Vector3D tmp;
    tmp =  MultiplyComponents(positionOnEllipsoid, _oneOverRadiiSquared);
    return glm::normalize(tmp);
}

Vector3D Ellipsoid::GeodeticSurfaceNormal(const Geodetic3D& geodetic)
{
    double cosLatitude = cos(geodetic.Latitude());

    return Vector3D(
        cosLatitude * cos(geodetic.Longitude()),
        cosLatitude * sin(geodetic.Longitude()),
        sin(geodetic.Latitude()));
}

const Vector3D& Ellipsoid::Radii()
{
    return _radii;
}

const Vector3D& Ellipsoid::RadiiSquared()
{
    return _radiiSquared;
}

const Vector3D& Ellipsoid::OneOverRadiiSquared()
{
    return _oneOverRadiiSquared;
}

double Ellipsoid::MinimumRadius()
{
    float result = std::min(_radii.x, std::min(_radii.y, _radii.z));
    return static_cast<double>(result);
}

double Ellipsoid::MaximumRadius()
{
    float result = std::max(_radii.x, std::max(_radii.y, _radii.z));
    return static_cast<double>(result);
}

int Ellipsoid::Intersections(const Vector3D& origin, const Vector3D& direction0, double* result)
{
    Vector3D direction = glm::normalize(direction0);

    // By laborious algebraic manipulation....
    float a = direction.x * direction.x * _oneOverRadiiSquared.x +
               direction.y * direction.y * _oneOverRadiiSquared.y +
               direction.z * direction.z * _oneOverRadiiSquared.z;
    float b = 2.0f *
               (origin.x * direction.x * _oneOverRadiiSquared.x +
                origin.y * direction.y * _oneOverRadiiSquared.y +
                origin.z * direction.z * _oneOverRadiiSquared.z);
    float c = origin.x * origin.x * _oneOverRadiiSquared.x +
               origin.y * origin.y * _oneOverRadiiSquared.y +
               origin.z * origin.z * _oneOverRadiiSquared.z - 1.0f;

    // Solve the quadratic equation: ax^2 + bx + c = 0.
    // Algorithm is from Wikipedia's "Quadratic equation" topic, and Wikipedia credits
    // Numerical Recipes in C, section 5.6: "Quadratic and Cubic Equations"
    float discriminant = b * b - 4.0f * a * c;
    if (discriminant < 0.0f)
    {
        // no intersections
        return 0; 
    }
    else if (discriminant == 0.0f)
    {
        // one intersection at a tangent point
        result[0] =  static_cast<double>(-0.5f * b / a);
        return 1;
    }

    float t = -0.5f * (b + (b > 0.0f ? 1.0f : -1.0f) * sqrt(discriminant));
    double root1 = static_cast<double>(t / a);
    double root2 = static_cast<double>(c / t);

    // Two intersections - return the smallest first.
    if (root1 < root2)
    {
        result[0] = root1;
        result[1] = root2;
        return 2;
    }
    else
    {
        result[0] = root2;
        result[1] = root1;
        return 2;
    }
}

Vector3D Ellipsoid::ToVector3D(const Geodetic2D& geodetic)
{
    return ToVector3D(Geodetic3D(geodetic.Longitude(), geodetic.Latitude(), 0.0));
}

Vector3D Ellipsoid::ToVector3D(const Geodetic3D& geodetic)
{
    Vector3D n = GeodeticSurfaceNormal(geodetic);
    Vector3D k = MultiplyComponents(_radiiSquared, n);
    float gamma = sqrt(
        (k.x * n.x) +
        (k.y * n.y) +
        (k.z * n.z));

    Vector3D rSurface = k / gamma;
    return rSurface + n * static_cast<float>(geodetic.Height());
}

void Ellipsoid::ToGeodetic3D(const std::vector<Vector3D>& positions, std::vector<Geodetic3D>& geodetics)
{
    for(const Vector3D& position : positions)
    {
        geodetics.push_back(ToGeodetic3D(position));
    }
}

void Ellipsoid::ToGeodetic2D(const std::vector<Vector3D>& positions, std::vector<Geodetic2D>& geodetics)
{
    for (const Vector3D& position : positions)
    {
        geodetics.push_back(ToGeodetic2D(position));
    }
}

Geodetic2D Ellipsoid::ToGeodetic2D(const Vector3D& positionOnEllipsoid)
{
    Vector3D n = GeodeticSurfaceNormal(positionOnEllipsoid);
    float azimuthAngle = atan2(n.y, n.x);
    float elevationAngle = asin(n.z / glm::length(n));
    return Geodetic2D(static_cast<double>(azimuthAngle), static_cast<double>(elevationAngle));
}

Geodetic3D Ellipsoid::ToGeodetic3D(const Vector3D& position)
{
    Vector3D p = ScaleToGeodeticSurface(position);
    Vector3D h = position - p;
    float heightSign = static_cast<float>(sign(glm::dot(h, position)));
    double height = static_cast<double>(heightSign * glm::length(h));
    return Geodetic3D(ToGeodetic2D(p), height);
}

Vector3D Ellipsoid::ScaleToGeodeticSurface(const Vector3D& position)
{
    float beta = 1.0f / sqrt(
        (position.x * position.x) * _oneOverRadiiSquared.x +
        (position.y * position.y) * _oneOverRadiiSquared.y +
        (position.z * position.z) * _oneOverRadiiSquared.z);

    float n = glm::length(Vector3D(
        beta * position.x * _oneOverRadiiSquared.x,
        beta * position.y * _oneOverRadiiSquared.y,
        beta * position.z * _oneOverRadiiSquared.z));

    float alpha = (1.0f - beta) * (glm::length(position) / n);

    float x2 = position.x * position.x;
    float y2 = position.y * position.y;
    float z2 = position.z * position.z;

    float da = 0.0f;
    float db = 0.0f;
    float dc = 0.0f;

    float s = 0.0f;
    float dSdA = 1.0f;

    do
    {
        alpha -= (s / dSdA);

        da = 1.0f + (alpha * _oneOverRadiiSquared.x);
        db = 1.0f + (alpha * _oneOverRadiiSquared.y);
        dc = 1.0f + (alpha * _oneOverRadiiSquared.z);

        float da2 = da * da;
        float db2 = db * db;
        float dc2 = dc * dc;

        float da3 = da * da2;
        float db3 = db * db2;
        float dc3 = dc * dc2;

        s = x2 / (_radiiSquared.x * da2) +
            y2 / (_radiiSquared.y * db2) +
            z2 / (_radiiSquared.z * dc2) - 1.0f;

        dSdA = -2.0f *
            (x2 / (_radiiToTheFourth.x * da3) +
             y2 / (_radiiToTheFourth.y * db3) +
             z2 / (_radiiToTheFourth.z * dc3));
    }
    while (std::abs(s) > 1e-10f);

    return Vector3D(
        position.x / da,
        position.y / db,
        position.z / dc);
}

Vector3D Ellipsoid::ScaleToGeocentricSurface(const Vector3D& position)
{
    float beta = 1.0f / sqrt(
        (position.x * position.x) * _oneOverRadiiSquared.x +
        (position.y * position.y) * _oneOverRadiiSquared.y +
        (position.z * position.z) * _oneOverRadiiSquared.z);

    return beta * position;
}

std::list<Vector3D> Ellipsoid::ComputeCurve(
    const Vector3D& start, 
    const Vector3D& stop, 
    double granularity)
{
    if (granularity <= 0.0)
    {
        throw "Granularity must be greater than zero.";
    }

    Vector3D normal = glm::normalize(glm::cross(start, stop));
    double theta = AngleBetween(start, stop);
    int n = std::max(static_cast<int>(theta / granularity) - 1, 0);

    std::list<Vector3D> positions;

    positions.push_back(start);

    for (int i = 1; i <= n; ++i)
    {
        double phi = (i * granularity);

        positions.push_back(ScaleToGeocentricSurface(RotateAroundAxis(start, normal, phi)));
    }

    positions.push_back(stop);

    return positions;
}
