// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/extractor.h"

#include "3d/opengl_enum.h"
#include "3d/core_objects.h"
#include "3d/polygon_mesh.h"
#include "3d/polygon.h"

#include "3d/camera.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "3d/scene_view.h"
#include "3d/query.h"

#include "3d/ppm.h"
#include "3d/iscene_renderer.h"

#include "core/structs.h"
#include "core/log.h"


Extractor::Extractor()
: _fbo{nullptr}
{
}

Extractor::~Extractor()
{
    delete _fbo;
    _fbo = nullptr;
}

void Extractor::create(OGL::Context* ctx, int width, int height)
{
    _fbo = OGL::Helper::createFramebuffer(ctx, static_cast<unsigned int>(width), static_cast<unsigned int>(height), DataType::float32);
}

/**
 *
 */

static bool getPolygonBBox(glm::ivec4& bbox, Image& polygonMask)
{
    bool hMinUpdated(false);    //
    bool vMinUpdated(false);    //
    std::vector<float> lineCounters(polygonMask.height());   //
    std::vector<float> columnCounters(polygonMask.width());    //

    std::fill(lineCounters.begin(), lineCounters.end(), 0.f);
    std::fill(columnCounters.begin(), columnCounters.end(), 0.f);

    glm::vec4* dataPtr(nullptr);    //
    unsigned int i(0);  //
    unsigned int j(0);  //
    for
        (; i < polygonMask.width(); ++i)
    {
        for
            (j = 0; j < polygonMask.height(); ++j)
        {
            dataPtr = static_cast<glm::vec4*>(polygonMask.pixelData(i, j));
            lineCounters[j] += dataPtr->w;
            columnCounters[i] += dataPtr->w;
        }
    }

    dataPtr = nullptr;

    j = 0;
    for
        (; j < polygonMask.height(); ++j)
    {
        if
            (false == vMinUpdated && 0.f < lineCounters[j])
        {
            bbox.w = static_cast<int>(polygonMask.height() - j);
            vMinUpdated = true;
            continue;
        }

        if
            (true == vMinUpdated && 0.f == lineCounters[j])
        {
            bbox.y = static_cast<int>(polygonMask.height() - j);
            break;
        }
    }

    i = 0;
    for
        (; i < polygonMask.width(); ++i)
    {
        if
            (false == hMinUpdated && 0.f < columnCounters[i])
        {
            bbox.x = static_cast<int>(i);
            hMinUpdated = true;
            continue;
        }

        if
            (true == hMinUpdated && 0.f == columnCounters[i])
        {
            bbox.z = static_cast<int>(i);
            break;
        }
    }

    if
        ((bbox.x >= bbox.z) || (bbox.y >= bbox.w))
    {
        return false;
    }
    return true;
}

/**
 *
 */
static bool computeExtractionGCPs(SEAScope::Extraction& ex, const glm::ivec4 viewport, const glm::ivec4& bbox, Image& globePosition)
{
    int width = bbox.z - bbox.x;
    int height = bbox.w - bbox.y;

    LOG(SEAScope::LogLevel::Debug, "Extractor", "width: {}, height: {}, {} {} {} {}", width, height, bbox.x, bbox.z, bbox.y, bbox.w);
    // Extract the positions for the bounding box area of the polygon.
    Image userShapePosition;
    globePosition.copyTo(glm::ivec2(bbox.x, viewport.w - (bbox.y + height)), glm::ivec2(width, height), userShapePosition);

    // Todo: compute the gcp density for a given zoom
    const float iteration = 8.0f;
    const float intervalW = static_cast<float>(userShapePosition.width() - 1) / (iteration - 1.f);
    const float intervalH = static_cast<float>(userShapePosition.height() - 1) / (iteration - 1.f);

    for
        (unsigned int y = 0; y < static_cast<unsigned int>(iteration); ++y)
    {
        for
            (unsigned int x = 0; x < static_cast<unsigned int>(iteration); ++x)
        {
            int xPos = static_cast<int>(static_cast<float>(x) * intervalW);
            int yPos = static_cast<int>(static_cast<float>(y) * intervalH);

            glm::vec4* pixelPtr = static_cast<glm::vec4*>(userShapePosition.pixelData(static_cast<unsigned int>(xPos), static_cast<unsigned int>(yPos)));

            if (pixelPtr->w != 0.f)
            {
                Geodetic2D geo = UnitSphere::ToGeodetic2D(glm::vec3(*pixelPtr));
                ex.gcps.push_back(
                    SEAScope::GCP(0u, /*placeholder for time*/
                                  static_cast<float>(glm::degrees(geo.Latitude())),
                                  static_cast<float>(glm::degrees(geo.Longitude())),
                                  static_cast<std::uint32_t>(xPos),
                                  static_cast<std::uint32_t>(yPos)));
            }
            else
            {
                LOG(SEAScope::LogLevel::Error, "Extractor",
                                   "Error no longitude/latitude found for an extraction gcp");
                return false;
            }
        }
    }

    userShapePosition.deallocate();

    return true;
}

/**
 *
 */
bool Extractor::extract(OGL::Context* ctx, Renderable2* polygon, SceneView& sceneView, std::map<SEAScope::RenderableId, SEAScope::Extraction>& result, bool discardExtractedData)
{
    if
        (nullptr == polygon)
    {
        return false;
    }

    Camera* cam = sceneView.getCamera();
    glm::mat4 mvp = cam->getProjectionMatrix() * cam->getViewMatrix() * cam->getModelMatrix();
    const glm::ivec4 viewport = ctx->getViewportSize();

    _fbo->bind();

    ctx->setFlag(OGL::cull_face, false);
    ctx->setFlag(OGL::blending, false);
    ctx->setViewport(viewport);

    // Enable depth test and make depth buffer writable
    ctx->setFlag(OGL::depth_test, true);
    ctx->setDepthMask(true);

    // Reset color, depth and stencil buffers
    ctx->setClearColor(glm::vec4(0.f));
    ctx->clear(OGL::color_buffer | OGL::depth_buffer | OGL::stencil_buffer);

    // Draw world everywhere to get locations for every pixel of the screen
    // -- Ignore stencil test (we want the whole screen)
    ctx->setFlag(OGL::stencil_test, false);
    // -- Disable write to stencil buffer
    ctx->setStencilMask(false);
    // -- Draw world to fill depth and color buffers
    ctx->setDepthMask(true);
    ctx->setColorMask(true, true, true, true);
    sceneView.getRenderer()->drawWorld(sceneView.getScene(),
                                        sceneView.getConstCamera(),
                                        mvp, fill_raw);
    ctx->setDepthMask(false);
    ctx->setColorMask(false, false, false, false);

    // Export pixel locations from FBO to Image
    Image globePosition;
    _fbo->dumpToImage(globePosition);

    // Fill the stencil buffer with the polygon
    // -- Disable depth test
    ctx->setFlag(OGL::depth_test, false);
    // -- Enable stencil test
    ctx->setFlag(OGL::stencil_test, true);
    // -- Configure stencil behavior
    ctx->setStencilFunc(OGL::always, 1, 0xFF);
    ctx->setStencilOp(OGL::keep, OGL::keep, OGL::replace);
    // -- Enable write while filling stencil buffer with polygon 
    ctx->setStencilMask(true);
    polygon->draw(ctx, mvp);
    ctx->setStencilMask(false);

    // Configure stencil test to use the buffer values set with the polygon
    ctx->setStencilFunc(OGL::equal, 1, 0xFF);

    // Restore the writing to the color buffer
    ctx->setColorMask(true, true, true, true);

    // Get mask for the polygon
    // -- Reset content of color buffers
    ctx->clear(OGL::color_buffer);
    // -- Draw world with stencil test on in order to get the bbox of the
    //    extraction
    Image polygonMask;
    sceneView.getRenderer()->drawWorld(sceneView.getScene(),
                                        sceneView.getConstCamera(),
                                        mvp, fill_raw);
    _fbo->dumpToImage(polygonMask);

    // Compute bounding box of the polygon in screen coordinates
    bool ok(false); //
    glm::ivec4 bbox({0,
                     0,
                     static_cast<int>(viewport.z - 1),
                     static_cast<int>(viewport.w - 1)});
    ok = getPolygonBBox(bbox, polygonMask);
    if
        (false == ok)
    {
        LOG(SEAScope::LogLevel::Error, "Extractor", "Cannot extract data because no shape vertex is visible in the viewport");
        return false;
    }

    int width = bbox.z - bbox.x;
    int height = bbox.w - bbox.y;

    // Compute GCPs for the extraction
    SEAScope::Extraction ex;
    ok = computeExtractionGCPs(ex, viewport, bbox, globePosition);
    globePosition.deallocate();
    if
        (false == ok)
    {
        LOG(SEAScope::LogLevel::Error, "Extractor",
                           "Could not geolocate all points within extraction bounding box");
        LOG(SEAScope::LogLevel::Info, "Extractor",
                           "Please try to keep the bounding box of the extraction footprint close to the center of the screen and make sure it does not go over the globe boundary");
        return false;
    }

    // Create an OpenGL occlusion query wrapper object
    OGL::Query* query = ctx->createQuery();

    // Initialize the opengl object
    query->create(ctx);

    glm::vec4 backgroundColor(0.f, 0.f, 0.f, 1.f);

    // Make sure depth test is enabled as it will be used to detect empty
    // granules
    ctx->setFlag(OGL::depth_test, true);

    // Loop on all granules
    Image tmpImg;
    std::list<SEAScope::RenderableId> toRemove;
    for
        (auto& it : result)
    {
        // Copy shape/gcps
        it.second.shape = ex.shape;
        it.second.gcps = ex.gcps;

        backgroundColor = glm::vec4(0.f, 0.f, 0.f, 1.f);

        // Configure the background color with fill values
        switch
            (it.second.fillValues.size())
        {
            case 1:
                backgroundColor = glm::vec4(it.second.fillValues[0],
                                             it.second.fillValues[0],
                                             it.second.fillValues[0],
                                             1.f);
                break;
            case 2:
                backgroundColor = glm::vec4(it.second.fillValues[0],
                                             it.second.fillValues[1],
                                             it.second.fillValues[0],
                                             1.f);
                break;
            case 3:
                backgroundColor = glm::vec4(it.second.fillValues[0],
                                             it.second.fillValues[1],
                                             it.second.fillValues[2],
                                             1.f);
                break;
            default:
                break;
        }

        // Reset color buffer with a color that corresponds to granule fill values
        ctx->setClearColor(backgroundColor);
        ctx->clear(OGL::color_buffer);

        // Start to count the samples which pass the depth test
        query->begin(OGL::samples_passed);

        // Draw the granule
        sceneView.getRenderer()->drawOneRaw(sceneView.getScene(),
                                             sceneView.getConstCamera(),
                                             it.first);

        // Stop the counter
        query->end();

        // Get the result
        unsigned int queryResult = query->getResult();

        // Check the number of pixels which are valid data in the polygon area.
        if
            (0 < queryResult)
        {
            if
                (false == discardExtractedData)
            {
                // The extraction request specifies that extracted data should be
                // included in the results and there is at least one pixel
                // intersecting the polygon mask => copy extracted data to the
                // result
                _fbo->dumpToImage(tmpImg);
                tmpImg.copyTo(glm::ivec2(bbox.x, viewport.w - (bbox.y + height)), glm::ivec2(width, height), it.second.buffer);
            }
        }
        else
        {
            // No valid data in the requested area for this granule, remove from the list
            toRemove.push_back(it.first);
        }
    }

    // Destroy opengl object and delete its wrapper
    query->destroy();
    delete query;
    query = nullptr;

    // Remove granules that did not pass the depth test (not visible)
    for
        (auto& rId : toRemove)
    {
        result.erase(rId);
    }

    // Disable stencil test to prevent unwanted cropping in future draw calls
    ctx->setFlag(OGL::stencil_test, false);

    _fbo->unbind();

    return true;
}

void Extractor::resize(int width, int height)
{
    if
        (nullptr != _fbo)
    {
        _fbo->resize(static_cast<unsigned int>(width), static_cast<unsigned int>(height));
    }
}

SEAScope::GCP Extractor::createGCP(const glm::vec3& point, const glm::mat4& mvp, const glm::ivec4& viewport) const
{
    Geodetic2D geo = UnitSphere::ToGeodetic2D(point);

    glm::vec3 pScreen = glm::project(point, glm::mat4(1.f), mvp, viewport);

    return SEAScope::GCP(0u, /*placeholder for time*/
                         static_cast<float>(glm::degrees(geo.Latitude())),
                         static_cast<float>(glm::degrees(geo.Longitude())),
                         static_cast<uint32_t>(pScreen.x),
                         static_cast<uint32_t>(static_cast<float>(viewport.w) - pScreen.y));
}

OGL::FrameBuffer* Extractor::getFBO() const
{
    return this->_fbo;
}
