// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/frustum.h"

/* This is ok, but only because 3d/stdafx has been included before (otherwise
   redefining the GLM options would be required) */
#include <glm/gtc/matrix_access.hpp>

#define ANG2RAD 3.14159265358979323846/180.0

Frustum::Frustum()
: ntl{0, 0, 0}, ntr{0, 0, 0}, nbl{0, 0, 0}, nbr{0, 0, 0},
  ftl{0, 0, 0}, ftr{0, 0, 0}, fbl{0, 0, 0}, fbr{0, 0, 0},
  nearD{0.f}, farD{0.f}, ratio{0.f}, angle{0.f}, tang{0.f},
  nw{0.f}, nh{0.f}, fw{0.f}, fh{0.f},
  X{0, 0, 0}, Y{0, 0, 0}, Z{0, 0, 0}
{
}

void Frustum::setPerspective(float angleIn, float ratioIn, float near, float far)
{
    this->ratio = ratioIn;
    this->angle = angleIn;
    this->nearD = near;
    this->farD = far;

    this->tang = static_cast<float>(tan((static_cast<double>(this->angle) * ANG2RAD) * 0.5));
    this->nh = this->nearD * tang;
    this->nw = this->nh * this->ratio;
    this->fh = this->farD  * this->tang;
    this->fw = this->fh * this->ratio;
}

void Frustum::lookAt(const glm::vec3 &p, const glm::vec3 &l, const glm::vec3 &u)
{
    glm::vec3 nc,fc;

	Z = p - l;
	Z = glm::normalize(Z);

	X = glm::cross(u, Z);
	X = glm::normalize(X);

	Y = glm::cross(Z, X);

	nc = p - Z * nearD;
	fc = p - Z * farD;

	ntl = nc + Y * nh - X * nw;
	ntr = nc + Y * nh + X * nw;
	nbl = nc - Y * nh - X * nw;
	nbr = nc - Y * nh + X * nw;

	ftl = fc + Y * fh - X * fw;
	ftr = fc + Y * fh + X * fw;
	fbl = fc - Y * fh - X * fw;
	fbr = fc - Y * fh + X * fw;
}

void Frustum::lookAt(const glm::mat4& view)
{
    Y = glm::vec3(view[0].x, view[0].y, view[0].z);
    Z = glm::vec3(view[1].x, view[1].y, view[1].z);
    X = glm::vec3(view[2].x, view[2].y, view[2].z);

    glm::mat3 rotMat(view);
    glm::vec3 p(view[3]);

    p = -p * rotMat;

    glm::vec3 nc = p - Z * nearD;
	glm::vec3 fc = p - Z * farD;

    ntl = nc + Y * nh - X * nw;
	ntr = nc + Y * nh + X * nw;
	nbl = nc - Y * nh - X * nw;
	nbr = nc - Y * nh + X * nw;

	ftl = fc + Y * fh - X * fw;
	ftr = fc + Y * fh + X * fw;
	fbl = fc - Y * fh - X * fw;
	fbr = fc - Y * fh + X * fw;
}

glm::vec3 Frustum::getTopLeftRay() const
{
    return glm::normalize(ftl - ntl);
}

glm::vec3 Frustum::getTopRightRay() const
{
    return glm::normalize(ftr - ntr);
}

glm::vec3 Frustum::getBottomLeftRay() const
{
    return glm::normalize(fbl - nbl);
}

glm::vec3 Frustum::getBottomRightRay() const
{
    return glm::normalize(fbr - nbr);
}

glm::vec3 Frustum::getCornerRay(CornerPosition pos)
{
    if (topLeft == pos)
    {
        return getTopLeftRay();
    }
    else if (topRight == pos)
    {
        return getTopRightRay();
    }
    else if (bottomLeft == pos)
    {
        return getBottomLeftRay();
    }
    else
    {
        return getBottomRightRay();
    }
}

glm::vec3 Frustum::getMiddleRay(MiddlePosition pos)
{
    glm::vec3 middleFar;
    glm::vec3 middleNear;

    if (top == pos)
    {
        middleFar = (ftl + ((ftr - ftl) * 0.5f));
        middleNear = (ntl + ((ntr - ntl) * 0.5f));
    }
    else if (bottom == pos)
    {
        middleFar = (fbl + ((fbr - fbl) * 0.5f));
        middleNear = (nbl + ((nbr - nbl) * 0.5f));
    }
    else if (right == pos)
    {
        middleFar = (ftr + ((fbr - ftr) * 0.5f));
        middleNear = (ntr + ((nbr - ntr) * 0.5f));
    }
    else
    {
        middleFar = (ftl + ((fbl - ftl) * 0.5f));
        middleNear = (ntl + ((nbl - ntl) * 0.5f));
    }

    return glm::normalize(middleFar - middleNear);
}

static void extractPlanesGL(Plane * p_planes, const glm::mat4& comboMatrix, bool /*normalize*/)
{
    glm::vec4 e;
    
    // Left clipping plane
    e.x = comboMatrix[0][3] + comboMatrix[0][0];
    e.y = comboMatrix[1][3] + comboMatrix[1][0];
    e.z = comboMatrix[2][3] + comboMatrix[2][0];
    e.w = comboMatrix[3][3] + comboMatrix[3][0];
    p_planes[0].setEquation(e);
    
    // Right clipping plane
    e.x = comboMatrix[0][3] - comboMatrix[0][0];
    e.y = comboMatrix[1][3] - comboMatrix[1][0];
    e.z = comboMatrix[2][3] - comboMatrix[2][0];
    e.w = comboMatrix[3][3] - comboMatrix[3][0];
    p_planes[1].setEquation(e);

    // Top clipping plane
    e.x = comboMatrix[0][3] - comboMatrix[0][1];
    e.y = comboMatrix[1][3] - comboMatrix[1][1];
    e.z = comboMatrix[2][3] - comboMatrix[2][1];
    e.w = comboMatrix[3][3] - comboMatrix[3][1];
    p_planes[2].setEquation(e);

    // Bottom clipping plane
    e.x = comboMatrix[0][3] + comboMatrix[0][1];
    e.y = comboMatrix[1][3] + comboMatrix[1][1];
    e.z = comboMatrix[2][3] + comboMatrix[2][1];
    e.w = comboMatrix[3][3] + comboMatrix[3][1];
    p_planes[3].setEquation(e);

    // Near clipping plane
    e.x = comboMatrix[0][3] + comboMatrix[0][2];
    e.y = comboMatrix[1][3] + comboMatrix[1][2];
    e.z = comboMatrix[2][3] + comboMatrix[2][2];
    e.w = comboMatrix[3][3] + comboMatrix[3][2];
    p_planes[4].setEquation(e);

    // Far clipping plane
    e.x = comboMatrix[0][3] - comboMatrix[0][2];
    e.y = comboMatrix[1][3] - comboMatrix[1][2];
    e.z = comboMatrix[2][3] - comboMatrix[2][2];
    e.w = comboMatrix[3][3] - comboMatrix[3][2];
    p_planes[5].setEquation(e);
}

void Frustum::updatePlanes(const glm::mat4& projView)
{
    extractPlanesGL(_planes, projView, false);
}

bool Frustum::contain(const glm::vec3& point) const
{
    for (int i = 0; i < 6; ++i)
    {
        if (NEGATIVE == _planes[i].classifyPoint(point))
        {
            return false;
        }
    }
    
    return true;
}
