// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/gcp_utils.h"

#include "3d/data_set.h"
#include "data/image.h"
#include "3d/core_objects.h"


void computeGcps(OGL::Context* ctx, DataSet* dataSet, OGL::Texture*& gcps, OGL::Texture*& gcpNorthEast, OGL::Texture*& gcpsAspectRatioTex, glm::vec2& uvMin, glm::vec2& uvRange)
{
	Image gcpData;
	gcpData.allocate(static_cast<unsigned int>(dataSet->getWidth() + 1), static_cast<unsigned int>(dataSet->getHeight() + 1), 3, DataType::float32, nullptr);

	computeGcpPerPixel(dataSet, gcpData);

	Image gcpAspectRatio;
	gcpAspectRatio.allocate(dataSet->getLon_size(), dataSet->getLat_size(), 2, DataType::float32, nullptr);

	computeGcpAspectRatio(dataSet, gcpAspectRatio);

	Image gcpNorthEastData;
	gcpNorthEastData.allocate(gcpData.width(), gcpData.height(), 4, DataType::float32, nullptr);

	computeUVNorth(gcpData, gcpNorthEastData, dataSet);

	uvMin = glm::vec2(0.5f / static_cast<float>(gcpData.width()), 0.5f / static_cast<float>(gcpData.height()));
	glm::vec2 uvMax = glm::vec2(
		(-0.5f + static_cast<float>(gcpData.width())) / static_cast<float>(gcpData.width()),
		(-0.5f + static_cast<float>(gcpData.height())) / static_cast<float>(gcpData.height())
	);
	uvRange = uvMax - uvMin;

	gcpNorthEast = ctx->createTexture(OGL::texture2D);
	gcpNorthEast->create(ctx);
	gcpNorthEast->allocate2D(gcpNorthEastData);
	gcpNorthEast->setWrapAndFilterMode(OGL::clamp_to_edge, OGL::linear);

	// Create textures to store gcps position and up vector

	gcps = OGL::Helper::createTexture(ctx, gcpData);
	gcps->setWrapAndFilterMode(OGL::clamp_to_edge, OGL::linear);

	gcpsAspectRatioTex = OGL::Helper::createTexture(ctx, gcpAspectRatio);
	gcpsAspectRatioTex->setWrapAndFilterMode(OGL::clamp_to_edge, OGL::linear);
}

void computeGcps(OGL::Context* ctx, DataSet* dataSet, Image& gcpData, OGL::Texture*& gcps, glm::vec2& uvMin, glm::vec2& uvRange)
{
    gcpData.allocate(static_cast<unsigned int>(dataSet->getWidth() + 1), static_cast<unsigned int>(dataSet->getHeight() + 1), 3, DataType::float32, nullptr);

    computeGcpPerPixel(dataSet, gcpData);

    uvMin = glm::vec2(0.5f / static_cast<float>(gcpData.width()), 0.5f / static_cast<float>(gcpData.height()));
    glm::vec2 uvMax = glm::vec2(
        (-0.5f + static_cast<float>(gcpData.width())) / static_cast<float>(gcpData.width()),
        (-0.5f + static_cast<float>(gcpData.height())) / static_cast<float>(gcpData.height())
    );
    uvRange = uvMax - uvMin;

    // Create texture to store gcps position
    gcps = OGL::Helper::createTexture(ctx, gcpData);
    gcps->setWrapAndFilterMode(OGL::clamp_to_edge, OGL::linear);
}

void computeGcpPerPixel(DataSet* dataSet, Image& img)
{
	const std::vector<glm::vec3>& gcps = dataSet->getGcps();
	unsigned int lonSize = dataSet->getLon_size();
	unsigned int latSize = dataSet->getLat_size();
	const std::uint32_t* gcpx = dataSet->getGcpx();
	const std::uint32_t* gcpy = dataSet->getGcpy();
	unsigned int width = img.width();

	unsigned int id[4];
	glm::u32vec2 ij[4];
	glm::vec3 p[4];

	glm::vec3* pos = static_cast<glm::vec3*>(img.data());

	for (unsigned int y = 0; y < latSize - 1; ++y)
	{
		for (unsigned int x = 0; x < lonSize - 1; ++x)
		{
			id[0] = y * lonSize + x;
			id[1] = y * lonSize + (x + 1);
			id[2] = (y + 1) * lonSize + (x + 1);
			id[3] = (y + 1) * lonSize + x;

			for (unsigned int i = 0; i < 4; ++i)
			{
				ij[i] = glm::u32vec2(gcpx[id[i]], gcpy[id[i]]);
				p[i] = gcps[id[i]];
			}

			// Number of pixels between two gcps
			glm::ivec2 pixCount(ij[1].x - ij[0].x, ij[3].y - ij[0].y);

			// Number of gcps
			glm::u32vec2 gcpCount(pixCount.x + 1, pixCount.y + 1);

			glm::vec3 pX0;
			glm::vec3 pX1;
			glm::vec3 pXY;
			float aX, aY;

			for (std::uint32_t pixY = 0; pixY < gcpCount.y; ++pixY)
			{
				aY = static_cast<float>(pixY) / static_cast<float>(gcpCount.y - 1);

				for (std::uint32_t pixX = 0; pixX < gcpCount.x; ++pixX)
				{
					aX = static_cast<float>(pixX) / static_cast<float>(gcpCount.x - 1);

					pX0 = (glm::mix(p[0], p[1], aX));
					pX1 = (glm::mix(p[3], p[2], aX));
					pXY = glm::normalize(glm::mix(pX0, pX1, aY));

					glm::vec2 pixCoord(ij[0].x + pixX, ij[0].y + pixY);
					unsigned int pixId = (static_cast<unsigned int>(pixCoord.y) * width + static_cast<unsigned int>(pixCoord.x));

					pos[pixId] = pXY;
				}
			}
		}
	}

}

void computeGcpAspectRatio(DataSet* dataSet, Image& img)
{
    const std::vector<glm::vec3>& gcps = dataSet->getGcps();
    unsigned int lonSize = dataSet->getLon_size();
    unsigned int latSize = dataSet->getLat_size();

    glm::vec2* pos = static_cast<glm::vec2*>(img.data());

    glm::vec3 pCurrent, pU, pV;

	// Compute the aspect ratio of the quad formed by the gcps.
	// In the texture texels are square but on the globe the gcps formed rectangles.
    for (unsigned int y = 0; y < latSize; ++y)
    {
        for (unsigned int x = 0; x < lonSize; ++x)
        {
            unsigned int currentId = y * lonSize + x;

            pCurrent = gcps[currentId];

            if
                (lonSize - 1 == x)
            {
                pU = gcps[currentId - 1];
            }
            else
            {
                pU = gcps[currentId + 1];
            }

            if
                (latSize - 1 == y)
            {
                pV = gcps[currentId - lonSize];
            }
            else
            {
                pV = gcps[currentId + lonSize];
            }

            float distU = glm::distance(pCurrent, pU);
            float distV = glm::distance(pCurrent, pV);

			float aspect = distU / distV;

			// Save the result in the first channel of the texture.
			pos[currentId].x = aspect;
        }
    }

	glm::vec3 pNextX;

	std::vector<float> lensX(latSize);
	std::vector<float> lensTotalX(latSize);
	float maxX = 0.f;

	// Now compute a ratio to apply to the streamlines to have homogeneous 
	// length everywhere on the globe

	// Compute the distance between the first gcp and the second on each row.
	// Then compute the length of the full row assuming each gcp is at equal distance.
    for (unsigned int y = 0; y < latSize; ++y)
    {
        unsigned int currentId = y * lonSize;
        unsigned int nextXId = y * lonSize + 1;

        pCurrent = gcps[currentId];
		pNextX = gcps[nextXId];

        float dist = glm::distance(pCurrent, pNextX);

		lensX[y] = dist;
		lensTotalX[y] = dist * static_cast<float>((lonSize));
		maxX = std::max(maxX, lensTotalX[y]);
    }

	// For each gcp compute a factor to apply to the streamline movement.
	for (unsigned int y = 0; y < latSize; ++y)
	{
		for (unsigned int x = 0; x < lonSize; ++x)
		{
            unsigned int currentId = y * lonSize + x;

			float f = (maxX / lensTotalX[y]);

			// Save the result in the second channel of the texture.
			pos[currentId].y = f;
		}
	}
}

void computeUVNorth(Image& gcps, Image& gcpNorthEastData, DataSet*)
{
	const unsigned int lonSize = gcps.width();
	const unsigned int latSize = gcps.height();

	unsigned int idOrigin, idX, idY;
	glm::vec3 pOrigin, vY, vX, vNorth;
	float inverseX, inverseY;

	for (unsigned int lat = 0; lat < latSize; ++lat)
	{
		for (unsigned int lon = 0; lon < lonSize; ++lon)
		{
			// ID of the current gcp in the gcp array
			idOrigin = lat * lonSize + lon;

			// Checks longitude border
			if (lon == lonSize - 1)
			{
				// It's a border, take the previous neighbour 
				idX = lat * lonSize + (lon - 1);
				// We will have to inverse the vector because it's the previous and not the next neighbour
				inverseX = -1.f;
			}
			else
			{
				// Normal case, take the next neighbour
				idX = lat * lonSize + (lon + 1);
				// Normal case, nothing to do
				inverseX = 1.f;
			}

			// Checks latitude border
			if (lat == latSize - 1)
			{
				// It's a border, take the neighbour in the previous row
				idY = (lat - 1) * lonSize + lon;
				// We will have to inverse the vector because it's the previous and not the next row
				inverseY = -1.f;
			}
			else
			{
				// Normal case, take the neighbour in the next row
				idY = (lat + 1) * lonSize + lon;
				// Normal case, nothing to do
				inverseY = 1.f;
			}

			// gcp position in world position (x,y,z)
			pOrigin = static_cast<glm::vec3*>(gcps.data())[idOrigin];

			// Computes directions vectors with neighbours
			vX = glm::normalize((static_cast<glm::vec3*>(gcps.data())[idX] - pOrigin) * inverseX);
			vY = glm::normalize((static_cast<glm::vec3*>(gcps.data())[idY] - pOrigin) * inverseY);

			// North pole is in the Z axis
			const glm::vec3 northPole(0.f, 0.f, 1.f);

			// Compute direction vector to the north pole (temporary vector)
			glm::vec3 tmp = glm::normalize(northPole - pOrigin);

			// Compute the perpendicular vector (X vector)
			glm::vec3 vEast = glm::cross(tmp, pOrigin);

			// Compute the up vector (Y vector)
			vNorth = glm::normalize(glm::cross(pOrigin, vEast));

			// Computes the north vector in texture coordinates
			glm::vec2 north = glm::normalize(glm::vec2(glm::dot(vX, vNorth), glm::dot(vY, vNorth)));
			glm::vec2 east = glm::normalize(glm::vec2(glm::dot(vX, vEast), glm::dot(vY, vEast)));

			static_cast<glm::vec4*>(gcpNorthEastData.data())[idOrigin] = glm::vec4(north, east);
		}
	}
}
