// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   Geodetic.cpp
 * Author: gleseach
 * 
 * Created on 25 août 2015, 10:58
 */

#include "3d/stdafx.h"
#include "3d/geodetic.h"

#include "num_utils.h"

Geodetic3D::Geodetic3D(double longitude, double latitude, double height)
{
    _longitude = longitude;
    _latitude = latitude;
    _height = height;
}

Geodetic3D::Geodetic3D(double longitude, double latitude)
{
    _longitude = longitude;
    _latitude = latitude;
    _height = 0;
}

Geodetic3D::Geodetic3D(const Geodetic2D& geodetic2D)
{
    _longitude = geodetic2D.Longitude();
    _latitude = geodetic2D.Latitude();
    _height = 0;
}

Geodetic3D::Geodetic3D(const Geodetic2D& geodetic2D, double height)
{
    _longitude = geodetic2D.Longitude();
    _latitude = geodetic2D.Latitude();
    _height = height;
}

double Geodetic3D::Longitude() const
{
    return _longitude;
}

double Geodetic3D::Latitude() const
{
    return _latitude;
}

double Geodetic3D::Height() const
{
    return _height;
}

bool Geodetic3D::Equals(const Geodetic3D& other) const
{
    return SEAScope::floating_point_equality(_longitude, other._longitude) && SEAScope::floating_point_equality(_latitude, other._latitude) && SEAScope::floating_point_equality(_height, other._height);
}

bool operator ==(const Geodetic3D& left, const Geodetic3D& right)
{
    return left.Equals(right);
}

bool operator !=(const Geodetic3D& left, const Geodetic3D& right)
{
    return !left.Equals(right);
}

//////////////////////////

Geodetic2D::Geodetic2D(double longitude, double latitude)
{
    _longitude = longitude;
    _latitude = latitude;
}

Geodetic2D::Geodetic2D(const glm::vec2& lonlat)
{
    _longitude = static_cast<double>(lonlat.x);
    _latitude = static_cast<double>(lonlat.y);
}

Geodetic2D::Geodetic2D(const Geodetic3D& geodetic3D)
{
    _longitude = geodetic3D.Longitude();
    _latitude = geodetic3D.Latitude();
}

double Geodetic2D::Longitude() const
{
    return _longitude;
}

double Geodetic2D::Latitude() const
{
    return _latitude;
}

bool Geodetic2D::EqualsEpsilon(const Geodetic2D& other, double epsilon) const
{
    return (std::abs(_longitude - other._longitude) <= epsilon) &&
           (std::abs(_latitude - other._latitude) <= epsilon);
}

bool Geodetic2D::Equals(const Geodetic2D& other) const
{
    return SEAScope::floating_point_equality(_longitude, other._longitude) && SEAScope::floating_point_equality(_latitude, other._latitude);
}

bool operator ==(const Geodetic2D& left, const Geodetic2D& right)
{
    return left.Equals(right);
}

bool operator !=(const Geodetic2D& left, const Geodetic2D& right)
{
    return !left.Equals(right);
}
