// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/geometry.h"

#include "3d/spherical_math.h"
#include "3d/math_func.h"
#include "3d/ear_clipping.h"

#include "core/log.h"

class TriangleIndicesUnsignedInt
{
public:
    TriangleIndicesUnsignedInt(unsigned int a, unsigned int b, unsigned int c)
    {
        I0 = a;
        I1 = b;
        I2 = c;
    }

    unsigned int I0, I1, I2;
};

class Edge
{
public:
    Edge(unsigned int a, unsigned int b)
    {
        _id = (a & 0x0000FFFF) | ((b & 0x0000FFFF) << 16);
    }

    bool operator < (const Edge& e) const
    {
        return _id < e._id;
    }

    unsigned int _id;
};

bool isPolygonValid(const std::vector<glm::vec3>& points)
{
    const std::size_t count = points.size();

    if (count < 3)
    {
        return false;
    }

    bool isValid = true;
    glm::vec3 intersection;

    // Loop on all points minus the last one.
    for (std::size_t i = 0; i < count - 1; ++i)
    {
        // Check the current segment with all remaining segments but without the direct neighbour.
        // Due to floating point precision error.
        for (std::size_t j = i+2; j < count; ++j)
        {
            if (arcIntersectUnitSphere(points[i], points[i+1], points[j], points[(j+1) % count], intersection))
            {
                isValid = false;
                break;
            }
        }

        if (!isValid)
        {
            break;
        }
    }

    return isValid;
}

bool triangulateEarClipping(const std::vector<glm::vec3>& points, std::vector<unsigned int>& outIndices)
{
    EarClipping ec;
    ec.run(points, outIndices);

    return true;
}

void subdivise(const std::vector<glm::vec3>& positions, const std::vector<unsigned int>& indices, float granularity, std::vector<glm::vec3>& outPositions, std::vector<unsigned int>& outIndices)
{
    float radGranularity = glm::radians(granularity);
    //
    // Use two queues:  one for triangles that need (or might need) to be 
    // subdivided and other for triangles that are fully subdivided.
    //
    std::queue<TriangleIndicesUnsignedInt> triangles;
    std::vector<TriangleIndicesUnsignedInt> done;

    std::vector<unsigned int> indicesValues;
    std::copy(indices.begin(), indices.end(), std::back_inserter(indicesValues));

    for (std::size_t i = 0; i < indicesValues.size(); i += 3)
    {
        triangles.push(TriangleIndicesUnsignedInt(indicesValues[i], indicesValues[i + 1], indicesValues[i + 2]));
    }

    //
    // New positions due to edge splits are appended to the positions list.
    //
    outPositions = positions;

    //
    // Used to make sure shared edges are not split more than once.
    //
    std::map<Edge, int> edges;

    //
    // Subdivide triangles until we run out
    //
    while (triangles.size() != 0)
    {
        TriangleIndicesUnsignedInt triangle = triangles.front();
        triangles.pop();

        glm::vec3 v0 = outPositions[triangle.I0];
        glm::vec3 v1 = outPositions[triangle.I1];
        glm::vec3 v2 = outPositions[triangle.I2];

        float g0 = static_cast<float>(AngleBetween(v0, v1));
        float g1 = static_cast<float>(AngleBetween(v1, v2));
        float g2 = static_cast<float>(AngleBetween(v2, v0));

        float max = std::max(g0, std::max(g1, g2));

        if (max > radGranularity)
        {
            if (SEAScope::floating_point_equality(g0, max))
            {
                Edge edge(std::min(triangle.I0, triangle.I1), std::max(triangle.I0, triangle.I1));
                int i;

                auto it = edges.find(edge);

                if (it == edges.end())
                {
                    outPositions.push_back(glm::normalize(slerp(v0, v1, 0.5f)));
                    i = static_cast<int>(outPositions.size()) - 1;
                    edges[edge] = i;
                }
                else
                {
                    i = it->second;
                }

                triangles.push(TriangleIndicesUnsignedInt(triangle.I0, static_cast<unsigned int>(i), triangle.I2));
                triangles.push(TriangleIndicesUnsignedInt(static_cast<unsigned int>(i), triangle.I1, triangle.I2));
            }
            else if (SEAScope::floating_point_equality(g1, max))
            {
                Edge edge(std::min(triangle.I1, triangle.I2), std::max(triangle.I1, triangle.I2));
                int i;

                auto it = edges.find(edge);

                if (it == edges.end())
                {
                    outPositions.push_back(glm::normalize(slerp(v1, v2, 0.5f)));
                    i = static_cast<int>(outPositions.size()) - 1;
                    edges[edge] = i;
                }
                else
                {
                    i = it->second;
                }
                triangles.push(TriangleIndicesUnsignedInt(triangle.I1, static_cast<unsigned int>(i), triangle.I0));
                triangles.push(TriangleIndicesUnsignedInt(static_cast<unsigned int>(i), triangle.I2, triangle.I0));
            }
            else if (SEAScope::floating_point_equality(g2, max))
            {
                Edge edge(std::min(triangle.I2, triangle.I0), std::max(triangle.I2, triangle.I0));
                int i;

                auto it = edges.find(edge);

                if (it == edges.end())
                {
                    outPositions.push_back(glm::normalize(slerp(v2, v0, 0.5f)));
                    i = static_cast<int>(outPositions.size()) - 1;
                    edges[edge] = i;
                }
                else
                {
                    i = it->second;
                }

                triangles.push(TriangleIndicesUnsignedInt(triangle.I2, static_cast<unsigned int>(i), triangle.I1));
                triangles.push(TriangleIndicesUnsignedInt(static_cast<unsigned int>(i), triangle.I0, triangle.I1));
            }
        }
        else
        {
            done.push_back(triangle);
        }
    }

    //
    // New indices
    //
    outIndices.reserve(done.size() * 3);

    for(TriangleIndicesUnsignedInt t : done)
    {
        outIndices.push_back(t.I0);
        outIndices.push_back(t.I1);
        outIndices.push_back(t.I2);
    }
}

void subdiviseContour(const std::vector<glm::vec3>& points, float angleMax, std::vector<glm::vec3>& outPositions)
{
    if (points.size() > 1)
    {
        for (size_t i = 0; i < points.size() - 1; ++i)
        {
            const glm::vec3& a = points[i];
            const glm::vec3& b = points[i+1];

            float angle = static_cast<float>(AngleBetween(a, b));

            outPositions.push_back(a);

            if (angle > angleMax)
            {
                int count = static_cast<int>(std::floor(angle / angleMax));

                float percentItr = 1.f / static_cast<float>(count + 1);

                for (int j = 1; j < (count + 1); ++j)
                {
                    outPositions.push_back(glm::normalize(glm::mix(a, b, percentItr * static_cast<float>(j))));
                }
            }

			outPositions.push_back(b);
        }
    }
}
