// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/globe_renderer_factory.h"

// List of all renderers include
#include "3d/raster_false_colors.h"
#include "3d/raster_color.h"
#include "3d/raster_raw.h"

// Trajectory
#include "3d/trajectory_false_colors.h"
#include "3d/trajectory_color.h"

// Barbs
#include "3d/barbs_false_colors.h"
#include "3d/barbs_color.h"

// Arrows
#include "3d/arrows_false_colors.h"
#include "3d/arrows_color.h"

// Streamlines
#include "3d/uv_streamlines_color.h"
#include "3d/uv_streamlines_false_colors.h"

// Polyline
#include "3d/polyline_color.h"
#include "3d/polyline_custom_color.h"
#include "3d/polygon_color.h"

#include "3d/screen_quad_color.h"
#include "3d/screen_quad_textured.h"

#include "3d/globe_renderer.h"

#include "3d/texture_renderer.h"

// List of all picking renderers include
#include "3d/polygon_picking.h"
#include "3d/polyline_picking.h"
#include "3d/trajectory_picking.h"
#include "3d/raster_picking.h"
#include "3d/null_picking.h"
#include "3d/globe_picking.h"
#include "3d/screen_quad_picking.h"

// List of all raw renderers include
#include "3d/raster_raw.h"
#include "3d/raster_rgb.h"
#include "3d/trajectory_raw.h"
#include "3d/rgb_raw.h"
#include "3d/globe_raw.h"
#include "3d/null_raw.h"

#include "3d/null_renderer.h"
#include "3d/billboard_pool_renderer.h"

GlobeRendererFactory::GlobeRendererFactory()
: RendererFactory{}
{
    this->_visibility = nullptr;  // RendererFactory::_visibility
}

GlobeRendererFactory::~GlobeRendererFactory()
{
}

IRenderer2* GlobeRendererFactory::createRenderer(OGL::Context* ctx, RendererType renderer, FillType fill)
{
    IRenderer2* r = nullptr;
    
    switch(renderer)
    {
    case renderer_raster:
    {
        switch(fill)
        {
        case fill_false_colors:
            r = new RasterFalseColors;
            break;
        case fill_color:
            r = new RasterColor;
            break;
        case fill_raw:
        case fill_count:
        case fill_null:
            break;
        }
    }
        break;
        
    case renderer_trajectory:
    {
        switch(fill)
        {
        case fill_false_colors:
            r = new TrajectoryFalseColors;
            break;
        case fill_color:
            r = new TrajectoryColor;
            break;
        case fill_raw:
        case fill_count:
        case fill_null:
            break;
        }
    }
        break;
        
    case renderer_barb:
    {
        switch(fill)
        {
        case fill_false_colors:
            r = new BarbsFalseColors;
            break;
        case fill_color:
            r = new BarbsColor;
            break;
        case fill_raw:
        case fill_count:
        case fill_null:
            break;
        }
    }
    break;
    
    case renderer_arrow:
    {
        switch(fill)
        {
        case fill_false_colors:
            r = new ArrowsFalseColors;
            break;
        case fill_color:
            r = new ArrowsColor;
            break;
        case fill_raw:
        case fill_count:
        case fill_null:
            break;
        }
    }
    break;
    
    case renderer_uv_streamline:
    {
        switch(fill)
        {
        case fill_false_colors:
            r = new UVStreamlinesFalseColors();
            break;
        case fill_color:
            r = new UVStreamlinesColor();
            break;
        case fill_raw:
        case fill_count:
        case fill_null:
            break;
        }
    }
    break;
    
    case renderer_polyline:
    {
        switch(fill)
        {
        case fill_false_colors:
        case fill_color:
            r = new PolylineCustomColor;
            break;
        case fill_raw:
        case fill_count:
        case fill_null:
            break;
        }
    }
    break;
    
    case renderer_polygon:
    {
        switch(fill)
        {
        case fill_false_colors:
        case fill_color:
            r = new PolygonColor;
            break;
        case fill_raw:
        case fill_count:
        case fill_null:
            break;
        }
    }
    break;
    
    case renderer_globe:
    {
        r = new GlobeRenderer;
    }
    break;
    
    case renderer_rawrgb:
    {
        r = new RasterRGB;
    }
    break;
    
    case renderer_quadscreen:
    {
        r = new ScreenQuadColor();
    }
    break;

	case renderer_quadscreen_textured:
	{
        r = new ScreenQuadTextured();
	}
	break;

    case renderer_quadscreen_textured_instanced:
    {
        switch (fill)
        {
        case fill_color:
            r = new BillboardPoolRenderer();
            break;
        case fill_false_colors:
        case fill_raw:
        case fill_count:
        case fill_null:
            break;
        }
    }
    break;
        
    case renderer_dot_cloud:
    case renderer_count:
        break;
    case renderer_null:
        r = new NullRenderer();
        break;
    }
    
    if (nullptr != r)
    {
        // Initialize the renderer
        r->create(ctx);
    }
    
    return r;
}

IPickingRenderer* GlobeRendererFactory::createPickingRenderer(OGL::Context* ctx, PickingRendererType renderer)
{
    IPickingRenderer* r = nullptr;
    
    switch(renderer)
    {
    case picking_renderer_trajectory:
        r = new TrajectoryPicking;
        break;
    
    case picking_renderer_polyline:
        r = new TrajectoryPicking;
        break;
    
    case picking_renderer_polygon:
        r = new PolygonPicking;
        break;
        
    case picking_renderer_raster:
        r = new RasterPicking;
        break;
        
    case picking_renderer_globe:
        r = new GlobePicking;
        break;

	case picking_renderer_screen_quad:
		r = new ScreenQuadPicking;
		break;

    case picking_renderer_null:
    case picking_renderer_count:
        r = new NullPicking;
        break;
    }
    
    if (nullptr != r)
    {
        // Initialize the renderer
        r->create(ctx);
    }
    
    return r;
}

IRawRenderer* GlobeRendererFactory::createRawRenderer(OGL::Context* ctx, RawRendererType renderer)
{
    IRawRenderer* r = nullptr;
    
    switch(renderer)
    {
    case raw_renderer_raster:
        r = new RasterRaw;
        break;
        
    case raw_renderer_trajectory:
        r = new TrajectoryRaw;
        break;
            
    case raw_renderer_rgb:
        r = new RGBRaw;
        break;
      
    case raw_renderer_globe:
        r = new GlobeRaw;
        break;
        
    case raw_renderer_null:
    case raw_renderer_count:
        r = new NullRaw;
        break;
    }
    
    if (nullptr != r)
    {
        // Initialize the renderer
        r->create(ctx);
    }
    
    return r;
}
