// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/globe_scene.h"

//
#include "3d/texture_loader.h"
#include "3d/scene_graph.h"
#include "3d/scene_graph_controller.h"

// Camera
#include "3d/camera.h"
#include "3d/globe_visibility.h"
#include "3d/dataset_visibility_globe.h"

#include "data/image.h"

// OpenGL
#include "3d/context.h"
#include "3d/query.h"
#include "3d/program_bind_helper.h"
#include "3d/program.h"

#include "3d/default_value.h"
#include "3d/math_func.h"
#include "3d/spherical_math.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "3d/helper.h"
#include "3d/basic_mesh.h"
#include "3d/camera.h"

#include "3d/texture.h"
#include "3d/texture_loader.h"

#include "3d/shape.h"
#include "3d/billboard.h"
#include "3d/static_mesh.h"
#include "3d/screen_quad.h"
#include "3d/shape_gpu_creator.h"
#include "3d/data_set.h"
#include "3d/particles_data.h"

#include "core/rendering_cfg.h"

#include "core/log.h"
#include "3d/icons.h"

constexpr static float EARTH_RADIUS_IN_METERS = 6371008.8f;

GlobeScene::GlobeScene(SEAScope::AppEventsDispatcher& appEventsDispatcher)
: IScene(),
  _nodeFactory(appEventsDispatcher),
  _globe(SEAScope::RenderableId(true, {0u, 0u}, {0u, 0u}, 0u)),
  _pointBillboardStaticMesh(nullptr),
  _pointBillboardTexture(nullptr)
{
}

GlobeScene::~GlobeScene()
{
    delete this->_pointBillboardStaticMesh;
    this->_pointBillboardStaticMesh = nullptr;

    this->_pointBillboardTexture = nullptr; // not owning
}

/**
 *
 */
void GlobeScene::create(OGL::Context* ctx, IWorldVisibility* visibility)
{
    this->_sceneGraph.create(ctx, visibility);
    this->_nodeFactory.setSceneGraph(ctx, &(this->_sceneGraph));
}

/**
 *
 */
void GlobeScene::addWorldToSceneGraph(OGL::Context* ctx, OGL::Texture* worldMapTexture)
{
    // Check the object creation
    if (nullptr == worldMapTexture)
    {
        // Creation failed, use fall back
        Image img;

        // Wonderful color
        unsigned char pixels[] =
        {
            255, 0, 255,
            255, 0, 255,
            255, 0, 255,
            255, 0, 255
        };

        img.allocate(2, 2, 3, DataType::unsigned_byte, pixels);

        worldMapTexture = OGL::Helper::createTexture(ctx, img);
    }

    this->_globe.create(ctx, worldMapTexture);

    this->_nodeFactory.addGlobe(&this->_globe);
}

/**
 *
 */
glm::vec2 GlobeScene::worldPositionToLonLat(const glm::vec3& position) const
{
    Geodetic2D geo = UnitSphere::ToGeodetic2D(position);
    return glm::vec2(glm::degrees(geo.Longitude()), glm::degrees(geo.Latitude()));
}

/**
 *
 */
float GlobeScene::getAltitude(glm::vec3 cameraPosition) const
{
    return EARTH_RADIUS_IN_METERS * (cameraPosition.x - 1.f);
}

/**
 *
 */
float GlobeScene::getZoom(float altitudeInMeters) const
{
    return 1.f + altitudeInMeters / EARTH_RADIUS_IN_METERS;
}

/**
 *
 */
SceneGraphController* GlobeScene::getNodeFactory()
{
    return &(this->_nodeFactory);
}

/**
 *
 */
SceneGraph* GlobeScene::getSceneGraph()
{
    return &(this->_sceneGraph);
}

/**
 *
 */
bool GlobeScene::addShape(OGL::Context* ctx, OGL::Shape* shape, const SEAScope::RenderableId& rId, const SEAScope::RenderingCfg& rCfg, const Icons& icons)
{
    bool ok(false);

	if (nullptr == shape)
	{
        return false;
    }

	Pickable* pickable = ShapeGPUCreator::create(ctx, shape, ok);
    if
        (false == ok)
    {
        return false;
    }

    if
        (SEAScope::ShapeType::POINT == shape->getType())
    {
        if
            (nullptr == this->_pointBillboardStaticMesh)
        {
            this->_pointBillboardStaticMesh = ScreenQuad::createMesh(ctx);
        }

        if
            (nullptr == this->_pointBillboardTexture)
        {
            this->_pointBillboardTexture = icons.getTexture(IconType::SMALL_CIRCLE_FILLED_32, ok);
        }

        Billboard* billboard = static_cast<Billboard*>(pickable);
        billboard->setMesh(this->_pointBillboardStaticMesh);
        billboard->setTexture(this->_pointBillboardTexture);
    }

    pickable->setRenderableId(rId);

    this->_nodeFactory.addShape(ctx, pickable, rCfg);

    pickable = nullptr;

    return true;
}

/**
 *
 */
bool GlobeScene::addGranule(OGL::Context* ctx, DataSet* rawData, const SEAScope::RenderingCfg& renderingCfg, DataSetVisibility* datasetVisibility)
{
    bool ok(false);

    ok = this->_nodeFactory.addGranule(ctx, rawData, renderingCfg);
    if
        (true == ok)
    {
        if
            (SEAScope::RenderMethod::STREAMLINES == renderingCfg.renderMethod()
                || SEAScope::RenderMethod::ARROWS == renderingCfg.renderMethod()
                || SEAScope::RenderMethod::BARBS == renderingCfg.renderMethod())
        {
            static_cast<DataSetVisibilityGlobe*>(datasetVisibility)->computeDataSetVisibility(rawData);

            if
                (SEAScope::RenderMethod::ARROWS == renderingCfg.renderMethod()
                 || SEAScope::RenderMethod::BARBS == renderingCfg.renderMethod())
            {
                rawData->getParticlesData()->update(ctx);
            }
        }
    }

    return ok;
}

/**
 *
 */
bool GlobeScene::addTrajectory(OGL::Context* ctx, std::vector<float>& relativeTime, std::vector<Geodetic2D>& points, std::vector<float>& fData, std::vector<bool>& mask, const SEAScope::RenderingCfg& renderingCfg, const SEAScope::RenderableId& rId)
{
    bool ok(false);

    ok = this->_nodeFactory.addTrajectory(ctx, relativeTime, points, fData, mask, renderingCfg, rId);

    return ok;
}
