// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   grid.cpp
 * Author: gleseach
 * 
 * Created on 26 août 2015, 11:23
 */

#include "3d/stdafx.h"
#include "3d/grid.h"
#include "3d/unit_sphere.h"
#include "3d/math_func.h"
#include "3d/spherical_math.h"

void createGrid(const glm::vec3 points[4], int subdivisionX, int subdivisionY,std::vector<glm::vec3>& vertices, std::vector<glm::vec2>& uv, std::vector<std::uint32_t>& indices)
{
    subdivisionX = std::max(1, subdivisionX);
    subdivisionY = std::max(1, subdivisionY);
    
    glm::vec3 vX = (points[1] - points[0]) / static_cast<float>(subdivisionX);
    glm::vec3 vY = (points[3] - points[0]) / static_cast<float>(subdivisionY);
    
    float uv_decalX = 1.f / static_cast<float>(subdivisionX);
    float uv_decalY = 1.f / static_cast<float>(subdivisionY);
    
    unsigned int vertices_lengthX = static_cast<unsigned int>(subdivisionX + 1);
    unsigned int vertices_lengthY = static_cast<unsigned int>(subdivisionY + 1);

    for (unsigned int y = 0; y < vertices_lengthY; ++y)
    {
        for (unsigned int x = 0; x < vertices_lengthX; ++x)
        {
            vertices.push_back(points[0] + vX * static_cast<float>(x) + vY * static_cast<float>(y));
            uv.push_back(glm::vec2(uv_decalX * static_cast<float>(x), uv_decalY * static_cast<float>(y)));
        }
    }
    
    for (unsigned int y = 1; y < vertices_lengthY; ++y)
    {
        for (unsigned int x = 1; x < vertices_lengthX; ++x)
        {
            indices.push_back(static_cast<std::uint32_t>((y-1) * vertices_lengthX + (x-1)));
            indices.push_back(static_cast<std::uint32_t>(y * vertices_lengthX + (x-1)));
            indices.push_back(static_cast<std::uint32_t>((y-1) * vertices_lengthX + x));
   
            indices.push_back(static_cast<std::uint32_t>((y-1) * vertices_lengthX + x));
            indices.push_back(static_cast<std::uint32_t>(y * vertices_lengthX + (x-1)));
            indices.push_back(static_cast<std::uint32_t>(y * vertices_lengthX + x));
        }
    }
}

void createQuad(std::vector<glm::vec3>& vertices, std::vector<glm::vec2>& uv, float size)
{
    float half_size = size / 2.f;
    float quart_size = size / 4.f;
    
    glm::vec3 a(-half_size, -quart_size, 0);
    glm::vec3 b(-half_size,  quart_size, 0);
    glm::vec3 c( half_size,  quart_size, 0);
    glm::vec3 d( half_size, -quart_size, 0);
    
    glm::vec2 uva(0.f, 1.f);
    glm::vec2 uvb(0.f, 0.f);
    glm::vec2 uvc(1.f, 0.f);
    glm::vec2 uvd(1.f, 1.f);
    
    vertices.push_back(a);
    vertices.push_back(c);
    vertices.push_back(b);
    
    vertices.push_back(a);
    vertices.push_back(d);
    vertices.push_back(c);
    
    uv.push_back(uva);
    uv.push_back(uvc);
    uv.push_back(uvb);
    
    uv.push_back(uva);
    uv.push_back(uvd);
    uv.push_back(uvc);
}

/*
static void createGrid(float * lons, unsigned int lon_size, float * lats, unsigned int lat_size, unsigned int precision, std::vector<glm::vec3>& vertices, std::vector<glm::vec2>& uv, std::vector<unsigned int>& indices, std::function<glm::vec3(const glm::vec2&)>& func)
{
    // Number of subdivision for the grid
    unsigned int lat_count = lat_size / precision;
    unsigned int lon_count = lon_size / precision;
    
    // Number of subdivision for the grid in float
    float lat_countf = static_cast<float>(lat_size) / static_cast<float>(precision);
    float lon_countf = static_cast<float>(lon_size) / static_cast<float>(precision);
    
    // lat and long UV mapping coord for the last integer position of the grid.
    float lat_max = lat_countf * static_cast<float>(precision) / static_cast<float>(lat_size);
    float lon_max = lon_countf * static_cast<float>(precision) / static_cast<float>(lon_size);
    
    glm::vec3 c;
    glm::vec2 uvc;
    
    // lonlat coords are for the top left of the pixel
    // compute the lonlat size of one pixel and add it at the end of the grid
    float lon_step = lons[1] - lons[0];
    float lat_step = lats[1] - lats[0];
    
    // Vertices & UV for drawelements
    for (unsigned int lat = 0; lat <= lat_count; ++lat)
    {
        for (unsigned int lon = 0; lon <= lon_count; ++lon)
        {
            c = func(glm::vec2(lons[std::min((lon) * precision, lon_size-1)] + (lon_step * (lon == lon_count)),
                               lats[std::min((lat) * precision, lat_size-1)] + (lat_step * (lat == lat_count))));
            vertices.push_back(c);
            
            uvc = glm::vec2(static_cast<float>(lon) / lon_countf  *  lon_max ,  static_cast<float>(lat) / (lat_countf) * lat_max);
            uv.push_back(uvc);
        }
    }
    
    unsigned int ia, ib;
    
    // triangles strip with degenerate triangles
    int count = 0;
    for (unsigned int lat = 1; lat <= lat_count; ++lat)
    {
        if (count % 2)
        {
            for (unsigned int lon = 0; lon <= lon_count; ++lon)
            {
                ia = (lat-1) * static_cast<unsigned int>(lon_count+1) + lon;
                ib = (lat)   * static_cast<unsigned int>(lon_count+1) + lon;

                // Insert in this order to make clock wise triangles
                indices.push_back(ib);
                indices.push_back(ia);
            }
        }
        else
        {
            for (int lon = static_cast<int>(lon_count); lon >= 0 ; --lon)
            {
                ia = (lat - 1) * (lon_count + 1) + static_cast<unsigned int>(lon);
                ib = (lat) * (lon_count + 1) + static_cast<unsigned int>(lon);

                // Insert in this order to make clock wise triangles
                indices.push_back(ia);
                indices.push_back(ib);
            }
        }
        
        ++count;
    }
}
*/

void fromLonlatToSphereUnitVector(const std::vector<glm::vec2>& lonlats, std::vector<glm::vec3>& xyz)
{
    xyz.reserve(lonlats.size());

    for (const glm::vec2& lonlat : lonlats)
    {
        xyz.push_back(UnitSphere::ToVector3D(Geodetic3D(Geodetic2D(TO_RADIANS(static_cast<double>(lonlat.x)), TO_RADIANS(static_cast<double>(lonlat.y))))));
    }
}

static glm::vec3 fromLonlatToSphereUnit(const glm::vec2& lonlat)
{
    return UnitSphere::ToVector3D(Geodetic3D(Geodetic2D(TO_RADIANS(static_cast<double>(lonlat.x)), TO_RADIANS(static_cast<double>(lonlat.y)))));
}

/*
static void createGridOnGlobe(float * lons, unsigned int lon_size, float * lats, unsigned int lat_size, unsigned int precision, std::vector<glm::vec3>& vertices, std::vector<glm::vec2>& uv, std::vector<unsigned int>& indices)
{
    std::function<glm::vec3(const glm::vec2&)> func = fromLonlatToSphereUnit;
    createGrid(lons, lon_size, lats, lat_size, precision, vertices, uv, indices, func);
}
*/

static glm::vec3 fromLonlatToMercatorPlan(const glm::vec2& latlon)
{
    float longitude = (latlon.x + 180.f) / 360.f;
    float latitude = (latlon.y + 90.f) / 180.f;
    
    float size = 0.5f;
    float half_size = size / 2.f;
    float quart_size = size / 4.f;
    
    glm::vec3 start(-half_size, -quart_size, 0);
    glm::vec3 dirLon(size,0,0);
    glm::vec3 dirLat(0,half_size,0);
    
    return start + dirLon * longitude + dirLat * latitude;
}

/*
static void createGridOnPlan(float * lons, unsigned int lon_size, float * lats, unsigned int lat_size, unsigned int precision, std::vector<glm::vec3>& vertices, std::vector<glm::vec2>& uv, std::vector<unsigned int>& indices)
{
    // Number of subdivision for the grid
    unsigned int lat_count = lat_size / precision;
    unsigned int lon_count = lon_size / precision;
    
    // Number of subdivision for the grid in float
    float lat_countf = static_cast<float>(lat_size) / static_cast<float>(precision);
    float lon_countf = static_cast<float>(lon_size) / static_cast<float>(precision);
    
    // lat and long UV mapping coord for the last integer position of the grid.
    float lat_max = lat_countf * static_cast<float>(precision) / static_cast<float>(lat_size);
    float lon_max = lon_countf * static_cast<float>(precision) / static_cast<float>(lon_size);
    
    glm::vec3 c;
    glm::vec2 uvc;
           
    glm::vec2 latlon;
    
    // Vertices & UV for drawelements
    for (unsigned int lat = 0; lat <= precision; ++lat)
    {
        for (unsigned int lon = 0; lon <= precision; ++lon)
        {
            latlon.x = lons[std::min((lon) * lon_count, lon_size-1)];
            latlon.y = lats[std::min((lat) * lat_count, lat_size-1)];
            
            c = fromLonlatToMercatorPlan(latlon);
            vertices.push_back(c);
            
            uvc = glm::vec2(static_cast<float>(lon) / static_cast<float>(precision)  *  lon_max ,  static_cast<float>(lat) / static_cast<float>(precision) * lat_max);
            uv.push_back(uvc);
        }
    }
    
    indices.push_back(0);
    indices.push_back(1);
    indices.push_back(2);
    indices.push_back(3);
}
*/

void createGridFromGCP(const float * lons,
                       unsigned int lon_size,
                       const float * lats,
                       unsigned int lat_size,
                       const std::uint32_t* gcpx,
                       const std::uint32_t* gcpy,
                       unsigned int width,
                       unsigned int height,
                       std::vector<glm::vec3>& vertices,
                       std::vector<glm::vec2>& uv,
                       std::vector<unsigned int>& indices)
{
    unsigned int id;
    
    for (unsigned int lat = 0; lat < lat_size; ++lat)
    {
        for (unsigned int lon = 0; lon < lon_size; ++lon)
        {
            id = lat * lon_size + lon;
            
            vertices.push_back(fromLonlatToSphereUnit(glm::vec2(lons[id], lats[id])));
            uv.push_back(glm::vec2(static_cast<float>(gcpx[id]) / static_cast<float>(width),
                                   static_cast<float>(gcpy[id]) / static_cast<float>(height)));
        }
    }
    
    unsigned int id0, id1, id2, id3;
    
    id0 = 0;
    id1 = 1;
    id2 = lon_size + 1;
    id3 = lon_size;
    
    int ok = 0;
    int ko = 0;

    float lonSize = static_cast<float>(lon_size);
    float latSize = static_cast<float>(lat_size);

    for (unsigned int i = 0; i < 20; ++i)
    {
        // [0, 1]
        float result = static_cast<float>(i) / 20.f;
        
        unsigned int lonTmp = static_cast<unsigned int>(std::round((lonSize - 2.0f) * result));
        unsigned int latTmp = static_cast<unsigned int>(std::round((latSize - 2.0f) * result));
        
        unsigned int idTmp = lon_size * latTmp + lonTmp;
        
        // Orientation for this triangle
        bool reverseTmp = isTrigoOnGlobe(vertices[idTmp], vertices[idTmp + 1], vertices[idTmp + lon_size]);
        
        // Increment counter
        if (true == reverseTmp)
        {
            ++ok;
        }
        else
        {
            ++ko;
        }
    }
    
    // Compare counters, ok vs ko triangles.
    bool reverse = ko < ok;
    
    for (unsigned int lat = 0; lat < lat_size-1; ++lat)
    {
        for (unsigned int lon = 0; lon < lon_size-1; ++lon)
        {
            id0 = lat * lon_size + lon;
            id1 = lat * lon_size + lon + 1;
            id2 = (lat+1) * lon_size + lon + 1;
            id3 = (lat+1) * lon_size + lon;
            
            if (reverse)
            {
                indices.push_back(id0);
                indices.push_back(id2);
                indices.push_back(id1);

                indices.push_back(id2);
                indices.push_back(id0);
                indices.push_back(id3);
            }
            else
            {
                indices.push_back(id0);
                indices.push_back(id1);
                indices.push_back(id2);

                indices.push_back(id2);
                indices.push_back(id3);
                indices.push_back(id0);
            }
        }
    }
}

void createGridOnPlanFromGCP(const float * lons,
                       unsigned int lon_size,
                       const float * lats,
                       unsigned int lat_size,
                       const std::uint32_t* gcpx,
                       const std::uint32_t* gcpy,
                       unsigned int width,
                       unsigned int height,
                       std::vector<glm::vec3>& vertices,
                       std::vector<glm::vec2>& uv,
                       std::vector<unsigned int>& indices)
{
    unsigned int id;
    
    for (unsigned int lat = 0; lat < lat_size; ++lat)
    {
        for (unsigned int lon = 0; lon < lon_size; ++lon)
        {
            id = lat * lon_size + lon;
            
            vertices.push_back(fromLonlatToMercatorPlan(glm::vec2(lons[id], lats[id])));
            uv.push_back(glm::vec2(static_cast<float>(gcpx[id]) / static_cast<float>(width - 1),
                                   static_cast<float>(gcpy[id]) / static_cast<float>(height - 1)));
        }
    }
    
    unsigned int id0, id1, id2, id3;
    
    for (unsigned int lat = 0; lat < lat_size-1; ++lat)
    {
        for (unsigned int lon = 0; lon < lon_size-1; ++lon)
        {
            id0 = lat * lon_size + lon;
            id1 = lat * lon_size + lon + 1;
            id2 = (lat+1) * lon_size + lon + 1;
            id3 = (lat+1) * lon_size + lon;
            
            indices.push_back(id0);
            indices.push_back(id2);
            indices.push_back(id1);
            
            indices.push_back(id2);
            indices.push_back(id0);
            indices.push_back(id3);
        }
    }
}
