// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/icons.h"

#include <vector>

#include "3d/texture.h"
#include "3d/texture_loader.h"
#include "core/path_tools.h"
#include "3d/context.h"
#include "data/image.h"

Icons::Icons()
{
}

Icons::~Icons()
{
    for
        (auto& it: this->_textures)
    {
        delete it.second;
        it.second = nullptr;
    }
    this->_textures.clear();
}

void Icons::setTexture(IconType iconType, OGL::Texture* texture, const std::string& key)
{
    setTexture(static_cast<std::uint16_t>(iconType), texture, key);
}

std::string Icons::getName(const std::uint16_t& id) const
{
    auto it = _idToName.find(id);

    if
        (it != _idToName.end())
    {
        return it->second;
    }

    return std::string("");
}

bool Icons::getId(const std::string& name, std::uint16_t& result) const
{
    auto found = this->_nameToId.find(name);
    if
        (this->_nameToId.end() == found)
    {
        return false;
    }

    result = found->second;
    return true;
}

void Icons::setTexture(IconType iconType, OGL::Context* ctx, const std::string& texturePath, const std::string& key)
{
    OGL::Texture* texture = TextureLoader::loadFromFile(ctx, texturePath);

    texture->setWrapMode(OGL::WrapMode::clamp_to_edge);

    this->setTexture(iconType, texture, key);
}

void Icons::setTexture(OGL::Context* ctx, std::uint16_t iconType, const Image& img, const std::string& name)
{
    OGL::Texture* texture = ctx->createTexture(OGL::texture2D);
    texture->create(ctx);
    texture->allocate2D(img);
    texture->setWrapMode(OGL::clamp_to_edge);

    auto found = this->_textures.find(iconType);
    if
        (this->_textures.end() != found)
    {
        if
            (found->second != texture)
        {
            delete found->second;
            found->second = nullptr;
        }
    }

    if
        (nullptr != texture)
    {
        this->_textures[iconType] = texture;
        this->_nameToId[name] = iconType;
        this->_idToName[iconType] = name;
    }
}

void Icons::setTexture(std::uint16_t iconType, OGL::Texture* texture, const std::string& key)
{
    auto found = this->_textures.find(iconType);
    if
        (this->_textures.end() != found)
    {
        if
            (found->second != texture)
        {
            delete found->second;
            found->second = nullptr;
        }
    }

    if
        (nullptr != texture)
    {
        this->_textures[iconType] = texture;
        this->_nameToId[key] = iconType;
        this->_idToName[iconType] = key;
    }
}

OGL::Texture* Icons::getTexture(IconType iconType, bool& ok) const
{
    return getTexture(static_cast<std::uint16_t>(iconType), ok);
}

OGL::Texture* Icons::getTexture(std::uint16_t iconType, bool& ok) const
{
    auto found = this->_textures.find(iconType);
    if
        (this->_textures.end() == found)
    {
        ok = false;
        return  nullptr;
    }

    ok = true;
    return found->second;
}

OGL::Texture* Icons::getTexture(const std::string& name, bool& ok) const
{
    auto found = this->_nameToId.find(name);
    if
        (this->_nameToId.end() == found)
    {
        ok = false;
        return  nullptr;
    }

    return getTexture(found->second, ok);
}
