// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/line.h"

void line(const glm::vec3& p0, const glm::vec3& p1, //coordinates of the line
        float w, //width/thickness of the line in pixel
        const glm::vec3& c0, //RGB color components
        const glm::vec3& c1, //color of background when alphablend=false,
        //  Br=alpha of color when alphablend=true
        bool alphablend, //use alpha blend or not
        std::vector<glm::vec3>& vertices,
        std::vector<glm::vec4>& colors)
{
    line(p0.x, p0.y, p1.x, p1.y, w, c0.r, c0.g, c0.b, c1.r, c1.g, c1.b, alphablend, vertices, colors);
}

void line(float x1, float y1, float x2, float y2, //coordinates of the line
        float w, //width/thickness of the line in pixel
        float Cr, float Cg, float Cb, //RGB color components
        float Br, float Bg, float Bb, //color of background when alphablend=false,
        //  Br=alpha of color when alphablend=true
        bool alphablend, //use alpha blend or not
        std::vector<glm::vec3>& vertices,
        std::vector<glm::vec4>& colors)
{
    float t(0.f);
    float R(0.f);
    float f = w - std::floor(w);  // w - static_cast<int> (w);
    float A(1.f);

    //determine parameters t,R
    /*   */
    if (w >= 0.0f && w < 1.0f)
    {
        t = 0.05f;
        R = 0.48f + 0.32f * f;
        if (!alphablend)
        {
            Cr += 0.88f * (1.0f - f);
            Cg += 0.88f * (1.0f - f);
            Cb += 0.88f * (1.0f - f);
            if (Cr > 1.0f) Cr = 1.0f;
            if (Cg > 1.0f) Cg = 1.0f;
            if (Cb > 1.0f) Cb = 1.0f;
        }
        else
        {
            A *= f;
        }
    }
    else if (w >= 1.0f && w < 2.0f)
    {
        t = 0.05f + f * 0.33f;
        R = 0.768f + 0.312f * f;
    }
    else if (w >= 2.0f && w < 3.0f)
    {
        t = 0.38f + f * 0.58f;
        R = 1.08f;
    }
    else if (w >= 3.0f && w < 4.0f)
    {
        t = 0.96f + f * 0.48f;
        R = 1.08f;
    }
    else if (w >= 4.0f && w < 5.0f)
    {
        t = 1.44f + f * 0.46f;
        R = 1.08f;
    }
    else if (w >= 5.0f && w < 6.0f)
    {
        t = 1.9f + f * 0.6f;
        R = 1.08f;
    }
    else if (w >= 6.0f)
    {
        float ff = w - 6.0f;
        t = 2.5f + ff * 0.50f;
        R = 1.08f;
    }
    //printf( "w=%f, f=%f, C=%.4f\n", w,f,C);

    //determine angle of the line to horizontal
    float tx = 0.0f, ty = 0.0f; //core thinkness of a line
    float Rx = 0.0f, Ry = 0.0f; //fading edge of a line
    //float cx = 0.0f, cy = 0.0f; //cap of a line
    float ALW = 0.01f;
    float dx = x2 - x1;
    float dy = y2 - y1;
    if (std::abs(dx) < ALW)
    {
        //vertical
        tx = t;
        ty = 0;
        Rx = R;
        Ry = 0;
        if (w > 0.0f && w < 1.0f)
            tx *= 8;
        else if (w == 1.0f)
            tx *= 10;
    }
    else if (std::abs(dy) < ALW)
    {
        //horizontal
        tx = 0;
        ty = t;
        Rx = 0;
        Ry = R;
        if (w > 0.0f && w < 1.0f)
            ty *= 8;
        else if (w == 1.0f)
            ty *= 10;
    }
    else
    {
        if (w < 3)
        { //approximate to make things even faster
            float m = dy / dx;
            //and calculate tx,ty,Rx,Ry
            if (m>-0.4142f && m <= 0.4142f)
            {
                // -22.5< angle <= 22.5, approximate to 0 (degree)
                tx = t * 0.1f;
                ty = t;
                Rx = R * 0.6f;
                Ry = R;
            }
            else if (m > 0.4142f && m <= 2.4142f)
            {
                // 22.5< angle <= 67.5, approximate to 45 (degree)
                tx = t*-0.7071f;
                ty = t * 0.7071f;
                Rx = R*-0.7071f;
                Ry = R * 0.7071f;
            }
            else if (m > 2.4142f || m <= -2.4142f)
            {
                // 67.5 < angle <=112.5, approximate to 90 (degree)
                tx = t;
                ty = t * 0.1f;
                Rx = R;
                Ry = R * 0.6f;
            }
            else if (m>-2.4142f && m<-0.4142f)
            {
                // 112.5 < angle < 157.5, approximate to 135 (degree)
                tx = t * 0.7071f;
                ty = t * 0.7071f;
                Rx = R * 0.7071f;
                Ry = R * 0.7071f;
            }
            else
            {
                // error in determining angle
            }
        }
        else
        { //calculate to exact
            dx = y1 - y2;
            dy = x2 - x1;
            float L = sqrt(dx * dx + dy * dy);
            dx /= L;
            dy /= L;
            //cx = -0.6f * dy;
            //cy = 0.6f * dx;
            tx = t*dx;
            ty = t*dy;
            Rx = R*dx;
            Ry = R*dy;
        }
    }

    glm::vec3 line_vertex[] =
    {
        glm::vec3(x1 - tx - Rx, y1 - ty - Ry,   0.f), //fading edge1
        glm::vec3(x2 - tx - Rx, y2 - ty - Ry,   0.f),
        glm::vec3(x1 - tx,      y1 - ty,        0.f), //core

        glm::vec3(x2 - tx - Rx, y2 - ty - Ry,   0.f),
        glm::vec3(x1 - tx,      y1 - ty,        0.f), //core
        glm::vec3(x2 - tx,      y2 - ty,        0.f),

        glm::vec3(x1 - tx,      y1 - ty,        0.f), //core
        glm::vec3(x2 - tx,      y2 - ty,        0.f),
        glm::vec3(x1 + tx,      y1 + ty,        0.f),

        glm::vec3(x2 - tx,      y2 - ty,        0.f),
        glm::vec3(x1 + tx,      y1 + ty,        0.f),
        glm::vec3(x2 + tx,      y2 + ty,        0.f),

        glm::vec3(x1 + tx,      y1 + ty,        0.f),
        glm::vec3(x2 + tx,      y2 + ty,        0.f),
        glm::vec3(x1 + tx + Rx, y1 + ty + Ry,   0.f), //fading edge2

        glm::vec3(x2 + tx,      y2 + ty,        0.f),
        glm::vec3(x1 + tx + Rx, y1 + ty + Ry,   0.f), //fading edge2
        glm::vec3(x2 + tx + Rx, y2 + ty + Ry,   0.f)
    };

    std::size_t elements_count = sizeof(line_vertex) / sizeof(glm::vec3);

    vertices.reserve(elements_count);

    std::copy(line_vertex, line_vertex + elements_count, std::back_inserter(vertices));

    if (!alphablend)
    {
        glm::vec4 line_color[] =
        {
            glm::vec4(Br, Bg, Bb, 1.f),
            glm::vec4(Br, Bg, Bb, 1.f),
            glm::vec4(Cr, Cg, Cb, 1.f),
            glm::vec4(Cr, Cg, Cb, 1.f),
            glm::vec4(Cr, Cg, Cb, 1.f),
            glm::vec4(Cr, Cg, Cb, 1.f),
            glm::vec4(Br, Bg, Bb, 1.f),
            glm::vec4(Br, Bg, Bb, 1.f)
        };

        elements_count = sizeof(line_color) / sizeof(glm::vec4);
        colors.reserve(elements_count);
        std::copy(line_color, line_color + elements_count, std::back_inserter(colors));
    }
    else
    {
        glm::vec4 line_color[] =
        {
            glm::vec4(Cr, Cg, Cb, 0.f),
            glm::vec4(Cr, Cg, Cb, 0.f),
            glm::vec4(Cr, Cg, Cb, A),

            glm::vec4(Cr, Cg, Cb, 0.f),
            glm::vec4(Cr, Cg, Cb, A),
            glm::vec4(Cr, Cg, Cb, A),

            glm::vec4(Cr, Cg, Cb, A),
            glm::vec4(Cr, Cg, Cb, A),
            glm::vec4(Cr, Cg, Cb, A),

            glm::vec4(Cr, Cg, Cb, A),
            glm::vec4(Cr, Cg, Cb, A),
            glm::vec4(Cr, Cg, Cb, A),

            glm::vec4(Cr, Cg, Cb, A),
            glm::vec4(Cr, Cg, Cb, A),
            glm::vec4(Cr, Cg, Cb, 0.f),

            glm::vec4(Cr, Cg, Cb, A),
            glm::vec4(Cr, Cg, Cb, 0.f),
            glm::vec4(Cr, Cg, Cb, 0.f)
        };

        elements_count = sizeof(line_color) / sizeof(glm::vec4);
        colors.reserve(elements_count);
        std::copy(line_color, line_color + elements_count, std::back_inserter(colors));
    }
}

static void lineParams(const glm::vec3& a, const glm::vec3& b, //coordinates of the line
        float w, //width/thickness of the line in pixel
        glm::vec4 /*color*/,
        float& coreSize,
        float& edgeSize,
        float& alpha)
{
    float t(0.f);
    float R(0.f);
    float f = w - std::floor(w); //w - static_cast<int> (w);
    float A = 1.0f;
    float x1(0.f), y1(0.f), x2(0.f), y2(0.f);
    //float Cr(0.f), Cg(0.f), Cb(0.f);

    x1 = a.x;
    y1 = a.y;
    x2 = b.x;
    y2 = b.y;

    //Cr = color.x;
    //Cg = color.y;
    //Cb = color.z;

    if (w >= 0.0f && w < 1.0f)
    {
        t = 0.05f;
        R = 0.48f + 0.32f * f;
        A *= f;
    }
    else if (w >= 1.0f && w < 2.0f)
    {
        t = 0.05f + f * 0.33f;
        R = 0.768f + 0.312f * f;
    }
    else if (w >= 2.0f && w < 3.0f)
    {
        t = 0.38f + f * 0.58f;
        R = 1.08f;
    }
    else if (w >= 3.0f && w < 4.0f)
    {
        t = 0.96f + f * 0.48f;
        R = 1.08f;
    }
    else if (w >= 4.0f && w < 5.0f)
    {
        t = 1.44f + f * 0.46f;
        R = 1.08f;
    }
    else if (w >= 5.0f && w < 6.0f)
    {
        t = 1.9f + f * 0.6f;
        R = 1.08f;
    }
    else if (w >= 6.0f)
    {
        float ff = w - 6.0f;
        t = 2.5f + ff * 0.50f;
        R = 1.08f;
    }

    //determine angle of the line to horizontal
    float tx = 0.0f, ty = 0.0f; //core thickness of a line
    float Rx = 0.0f, Ry = 0.0f; //fading edge of a line
    //float cx = 0.0f, cy = 0.0f; //cap of a line
    float ALW = 0.01f;
    float dx = x2 - x1;
    float dy = y2 - y1;
    if (std::abs(dx) < ALW)
    {
        //vertical
        tx = t;
        ty = 0.0f;
        Rx = R;
        Ry = 0.0f;
        if (w > 0.0f && w < 1.0f)
            tx *= 8;
        else if (w == 1.0f)
            tx *= 10;
    }
    else if (std::abs(dy) < ALW)
    {
        //horizontal
        tx = 0.0f;
        ty = t;
        Rx = 0.0f;
        Ry = R;
        if (w > 0.0f && w < 1.0f)
            ty *= 8;
        else if (w == 1.0f)
            ty *= 10;
    }
    else
    {
        if (w < 3)
        { //approximate to make things even faster
            float m = dy / dx;
            //and calculate tx,ty,Rx,Ry
            if (m>-0.4142f && m <= 0.4142f)
            {
                // -22.5< angle <= 22.5, approximate to 0 (degree)
                tx = t * 0.1f;
                ty = t;
                Rx = R * 0.6f;
                Ry = R;
            }
            else if (m > 0.4142f && m <= 2.4142f)
            {
                // 22.5< angle <= 67.5, approximate to 45 (degree)
                tx = t*-0.7071f;
                ty = t * 0.7071f;
                Rx = R*-0.7071f;
                Ry = R * 0.7071f;
            }
            else if (m > 2.4142f || m <= -2.4142f)
            {
                // 67.5 < angle <=112.5, approximate to 90 (degree)
                tx = t;
                ty = t * 0.1f;
                Rx = R;
                Ry = R * 0.6f;
            }
            else if (m>-2.4142f && m<-0.4142f)
            {
                // 112.5 < angle < 157.5, approximate to 135 (degree)
                tx = t * 0.7071f;
                ty = t * 0.7071f;
                Rx = R * 0.7071f;
                Ry = R * 0.7071f;
            }
            else
            {
                // error in determining angle
                //printf( "error in determining angle: m=%.4f\n",m);
            }
        }
        else
        { //calculate to exact
            dx = y1 - y2;
            dy = x2 - x1;
            float L = sqrt(dx * dx + dy * dy);
            dx /= L;
            dy /= L;
            //cx = -0.6f * dy;
            //cy = 0.6f * dx;
            tx = t*dx;
            ty = t*dy;
            Rx = R*dx;
            Ry = R*dy;
        }
    }

    coreSize = glm::length(glm::vec2(tx, ty));
    edgeSize = glm::length(glm::vec2(Rx, Ry));
    alpha = A;
}

struct Segment
{
    glm::vec3 n;
    float angle;
    glm::vec3 dir;
    int sign;
    float coreSize, edgeSize;
    glm::vec3 bisector;
    float miter;

    glm::vec3 left[4];
    glm::vec3 right[4];

    glm::vec4 color[2];
};

/*
// signed angle between two 2D vectors
static float angleBetween(glm::vec2& a, glm::vec2& b)
{
    float angle = atan2(b.y, b.x) - atan2(a.y, a.x);
    
    // Set the angle between [-PI, PI]
    if (static_cast<double>(angle) < -M_PI)
    {
        angle += static_cast<float>(M_PI * 2.0);
    }
    else if (static_cast<double>(angle) > M_PI)
    {
        angle -= static_cast<float>(M_PI * 2.0);
    }
    
    return angle;
}
*/

// Signed angle between two 3D vectors in the (0,0,-1) plan
static float angleBetween(glm::vec3& a, glm::vec3& b)
{
    // Compute angle
    float dotVal = glm::dot(glm::normalize(a), glm::normalize(b));

    if (dotVal > 1.f)
    {
        dotVal = 1.f;
    }

    float angle = acos(dotVal);

    // Compute the normal of the a, b couple
    glm::vec3 n = glm::cross(a, b);

    // Compare the sign of the dot product between the normal of the plan and 
    // the normal of the plan.
    float s = glm::dot(glm::vec3(0.f, 0.f, -1.f), n);

    if (s < 0.f)
    {
        angle = -angle;
    }

    // Set the angle between [-PI, PI]
    if (static_cast<double>(angle) < -M_PI)
    {
        angle += static_cast<float>(M_PI * 2.0);
    }
    else if (static_cast<double>(angle) > M_PI)
    {
        angle -= static_cast<float>(M_PI * 2.0);
    }

    return angle;
}

static int sign(glm::vec3& a, glm::vec3& b)
{
    float f = angleBetween(a, b);
    return (f > 0.0f) ? 1 : ((f < 0.0f) ? -1 : 0);
}

static void addTriangle(
                 std::vector<glm::vec3>& vertices,
                 std::vector<glm::vec4>& colors,
                 const glm::vec3& a, const glm::vec3& b, const glm::vec3& c,
                 const glm::vec4& ca, const glm::vec4& cb, const glm::vec4& cc)
{
    vertices.push_back(a);
    vertices.push_back(b);
    vertices.push_back(c);

    colors.push_back(ca);
    colors.push_back(cb);
    colors.push_back(cc);
}

void polyline(
        const std::vector<glm::vec3>& points,
        float w,
        const glm::vec3& color,
        std::vector<glm::vec3>& vertices,
        std::vector<glm::vec4>& colors,
        int type)
{
    if (points.size() < 3)
    {
        return;
    }

    w /= 2.0f;

    std::vector<Segment> seg(points.size()-1);

    glm::vec3 n;
    glm::vec3 dir;
    float alpha;

    // Compute normal of each line's segment
    for (std::size_t i = 0; i < seg.size(); ++i)
    {
        dir = points[i] - points[i+1];
        dir = glm::normalize(dir);

        n = glm::normalize(glm::cross(dir, glm::vec3(0.f, 0.f, 1.f)));

        seg[i].dir = dir;

        seg[i].n = n;
        seg[i].sign = 0;

        lineParams(points[i], points[i+1], w, glm::vec4(1.f, 1.f, 1.f, 1.f), seg[i].coreSize, seg[i].edgeSize, alpha);

        seg[i].left[0] = points[i] + n * seg[i].coreSize;
        seg[i].left[1] = points[i] - n * seg[i].coreSize;
        seg[i].left[2] = points[i] + n * (seg[i].coreSize + seg[i].edgeSize);
        seg[i].left[3] = points[i] - n * (seg[i].coreSize + seg[i].edgeSize);

        seg[i].right[0] = points[i+1] + n * seg[i].coreSize;
        seg[i].right[1] = points[i+1] - n * seg[i].coreSize;
        seg[i].right[2] = points[i+1] + n * (seg[i].coreSize + seg[i].edgeSize);
        seg[i].right[3] = points[i+1] - n * (seg[i].coreSize + seg[i].edgeSize);

        seg[i].color[0] = glm::vec4(color, alpha);
        seg[i].color[1] = glm::vec4(color, 0.f);
    }

    for (std::size_t i = 0; i < seg.size()-1; ++i)
    {
        glm::vec3 bisector = glm::normalize(seg[i].n + seg[i+1].n);

        float angle = angleBetween(bisector, seg[i].n);

        float lengthCore = seg[i].coreSize / cos(angle);
        float lengthEdge = (seg[i].coreSize + seg[i].edgeSize) / cos(angle);

        // Miter
        if (type == 0)
        {
            seg[i].right[0] = seg[i+1].left[0] = points[i+1] + bisector * lengthCore;
            seg[i].right[1] = seg[i+1].left[1] = points[i+1] - bisector * lengthCore;
            seg[i].right[2] = seg[i+1].left[2] = points[i+1] + bisector * lengthEdge;
            seg[i].right[3] = seg[i+1].left[3] = points[i+1] - bisector * lengthEdge;
        }
        else if (type == 1 || type == 2)
        {
            // Bevel
            seg[i].angle = angleBetween(seg[i].dir, seg[i+1].dir);
            seg[i].sign = sign(seg[i].dir, seg[i+1].dir);

            if (seg[i].sign == 1)
            {
                seg[i].right[1] = seg[i+1].left[1] = points[i+1] - bisector * lengthCore;
                seg[i].right[3] = seg[i+1].left[3] = points[i+1] - bisector * lengthEdge;
            }
            else if (seg[i].sign == -1)
            {
                seg[i].right[0] = seg[i+1].left[0] = points[i+1] + bisector * lengthCore;
                seg[i].right[2] = seg[i+1].left[2] = points[i+1] + bisector * lengthEdge;
            }
        }
    }

    // Generate vertices from segment data
    for (std::size_t i = 0; i < seg.size(); ++i)
    {
        // Segment core
        addTriangle(vertices, colors, seg[i].left[0], seg[i].right[0], seg[i].left[1],  seg[i].color[0], seg[i].color[0], seg[i].color[0]);
        addTriangle(vertices, colors, seg[i].left[1], seg[i].right[1], seg[i].right[0], seg[i].color[0], seg[i].color[0], seg[i].color[0]);

        addTriangle(vertices, colors, seg[i].left[2], seg[i].right[2], seg[i].left[0],  seg[i].color[1], seg[i].color[1], seg[i].color[0]);
        addTriangle(vertices, colors, seg[i].left[0], seg[i].right[2], seg[i].right[0], seg[i].color[0], seg[i].color[1], seg[i].color[0]);

        addTriangle(vertices, colors, seg[i].left[1], seg[i].right[1], seg[i].left[3],  seg[i].color[0], seg[i].color[0], seg[i].color[1]);
        addTriangle(vertices, colors, seg[i].left[3], seg[i].right[1], seg[i].right[3], seg[i].color[1], seg[i].color[0], seg[i].color[1]);

        // Bevel
        if (type == 1 && i < seg.size()-1)
        {
            if (seg[i].sign == 1)
            {
                addTriangle(vertices, colors, seg[i].right[0], seg[i+1].left[0], seg[i].right[1], seg[i].color[0], seg[i].color[0], seg[i].color[0]);
                addTriangle(vertices, colors, seg[i].right[0], seg[i].right[2], seg[i+1].left[2], seg[i].color[0], seg[i].color[1], seg[i].color[1]);
                addTriangle(vertices, colors, seg[i].right[0], seg[i+1].left[2], seg[i+1].left[0], seg[i].color[0], seg[i].color[1], seg[i].color[0]);
            }
            else if (seg[i].sign == -1)
            {
                addTriangle(vertices, colors, seg[i].right[1], seg[i+1].left[0], seg[i+1].left[1], seg[i].color[0], seg[i].color[0], seg[i].color[0]);
                addTriangle(vertices, colors, seg[i].right[3], seg[i].right[1], seg[i+1].left[1], seg[i].color[1], seg[i].color[0], seg[i].color[0]);
                addTriangle(vertices, colors, seg[i].right[3], seg[i+1].left[1], seg[i+1].left[3], seg[i].color[1], seg[i].color[0], seg[i].color[1]);
            }
        }
        else if (type == 2 && i < seg.size()-1) // Round
        {
            const int precision = 5;
            glm::vec3 v0, v1;
            glm::vec3 segN;
            glm::vec3 a;

            float step = seg[i].angle / precision;

            if (seg[i].sign == 1)
            {
                v0 = segN = seg[i].n;
                a = seg[i].right[1];
            }
            else if (seg[i].sign == -1)
            {
                v0 = segN = -seg[i].n;
                a = seg[i].right[0];
            }

            for (int p = 0; p <= precision; ++p)
            {
                glm::mat4 matrix(1.f);
                matrix = glm::rotate(matrix, static_cast<float>(p) * step, glm::vec3(0.f,0.f,-1.f));

                glm::vec4 tmp = matrix * glm::vec4(n, 0.f);

                v1 = glm::vec3(tmp.x, tmp.y, tmp.z);

                glm::vec3 b = points[i+1] + v0 * w;
                glm::vec3 c = points[i+1] + v1 * w;

                addTriangle(vertices, colors, a, b, c, seg[i].color[0], seg[i].color[0], seg[i].color[1]);

                v0 = v1;
            }
        }
    }
}

void polyline2(const std::vector<glm::vec3>& points, std::vector<glm::vec3>& vertices, std::vector<glm::vec3>& normals, std::vector<float>& miter)
{
    glm::vec3 dir;

    std::vector<Segment> seg(points.size()-1);

    // Compute normal of each line's segment
    for (std::size_t i = 0; i < seg.size(); ++i)
    {
        dir = points[i] - points[i+1];
        dir = glm::normalize(dir);

        seg[i].n = glm::normalize(glm::cross(dir, glm::vec3(0.f, 0.f, 1.f)));

        seg[i].dir = dir;
        seg[i].left[0] = points[i];
        seg[i].left[1] = points[i];
        seg[i].right[0] = points[i+1];
        seg[i].right[1] = points[i+1];
    }

    for (std::size_t i = 0; i < seg.size()-1; ++i)
    {
        seg[i].bisector = glm::normalize(seg[i].n + seg[i+1].n);

        float angle = angleBetween(seg[i].bisector, seg[i].n);
        seg[i].miter = 1.f / cos(angle);
    }

    vertices.push_back(points[0]);
    vertices.push_back(points[0]);

    miter.push_back(1.f);
    miter.push_back(-1.f);

    normals.push_back(seg[0].n);
    normals.push_back(seg[0].n);

    for (std::size_t i = 0; i < seg.size()-1; ++i)
    {
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);

        normals.push_back(seg[i].bisector);
        normals.push_back(seg[i].bisector);

        miter.push_back(seg[i].miter);
        miter.push_back(-seg[i].miter);
    }

    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);

    normals.push_back(seg[seg.size()-1].n);
    normals.push_back(seg[seg.size()-1].n);

    miter.push_back(1.f);
    miter.push_back(-1.f);
}

/*
static glm::mat4 MakeMatrix( const glm::vec3& X, const glm::vec3& Y )  
{  
    glm::mat4 M(1.f);  
    M[0] = glm::vec4(glm::normalize( X ), 0.f);  
    M[2] = glm::vec4(glm::normalize( glm::cross(X,Y) ), 0.f);  
    M[1] = glm::vec4(glm::normalize( glm::cross(glm::vec3(M[2].x, M[2].y, M[2].z),X) ), 0.f);  
    return M;
}
*/

void polylineOnSphere(const std::vector<glm::vec3>& points, std::vector<glm::vec3>& vertices, std::vector<glm::vec3>& normals, std::vector<float>& miter)
{
    glm::vec3 dir;
    glm::vec3 middle;

    std::vector<Segment> seg(points.size()-1);

    // Compute normal of each line's segment
    for (std::size_t i = 0; i < seg.size(); ++i)
    {
        dir = points[i] - points[i+1];
        middle = points[i] + dir / 2.f;
        dir = glm::normalize(dir);

        seg[i].n = glm::normalize(glm::cross(dir, middle));

        seg[i].dir = dir;
        seg[i].left[0] = points[i];
        seg[i].left[1] = points[i];
        seg[i].right[0] = points[i+1];
        seg[i].right[1] = points[i+1];
    }

    for (std::size_t i = 0; i < seg.size()-1; ++i)
    {
        glm::vec3 bisector = seg[i].n + seg[i+1].n;
        seg[i].miter = glm::length(bisector);

        if (seg[i].miter >= 2.f)
        {
            seg[i].miter = 1.f;
        }

        seg[i].bisector = glm::normalize(bisector);

        float angle = angleBetween(seg[i].bisector, seg[i].n);
        seg[i].miter = 1.f / cos(angle);
    }

    vertices.push_back(points[0]);
    vertices.push_back(points[0]);

    miter.push_back(1.f);
    miter.push_back(-1.f);

    normals.push_back(seg[0].n);
    normals.push_back(seg[0].n);

    for (std::size_t i = 0; i < seg.size()-1; ++i)
    {
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);

        normals.push_back(seg[i].bisector);
        normals.push_back(seg[i].bisector);

        miter.push_back(seg[i].miter);
        miter.push_back(-seg[i].miter);
    }

    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);

    normals.push_back(seg[seg.size()-1].n);
    normals.push_back(seg[seg.size()-1].n);

    miter.push_back(1.f);
    miter.push_back(-1.f);
}

void polylineOnSphere2(
        const std::vector<glm::vec3>& points,
        std::vector<glm::vec3>& vertices,
        std::vector<glm::vec3>& previous,
        std::vector<glm::vec3>& next,
        std::vector<float>& direction)
{
    glm::vec3 dir;
    glm::vec3 middle;

    std::vector<Segment> seg(points.size()-1);

    // Compute normal of each line's segment
    for (std::size_t i = 0; i < seg.size(); ++i)
    {
        dir = points[i] - points[i+1];
        middle = points[i] + dir / 2.f;
        dir = glm::normalize(dir);

        seg[i].n = glm::normalize(glm::cross(dir, middle));

        seg[i].dir = dir;
        seg[i].left[0] = points[i];
        seg[i].left[1] = points[i];
        seg[i].right[0] = points[i+1];
        seg[i].right[1] = points[i+1];
    }

    for (std::size_t i = 0; i < seg.size()-1; ++i)
    {
        glm::vec3 bisector = seg[i].n + seg[i+1].n;
        seg[i].miter = glm::length(bisector);

        if (seg[i].miter >= 2.f)
        {
            seg[i].miter = 1.f;
        }

        seg[i].bisector = glm::normalize(bisector);

        float angle = angleBetween(seg[i].bisector, seg[i].n);
        seg[i].miter = 1.f / cos(angle);
    }

    vertices.push_back(points[0]);
    vertices.push_back(points[0]);

    previous.push_back(points[0]);
    previous.push_back(points[0]);

    next.push_back(points[1]);
    next.push_back(points[1]);

    direction.push_back(1.f);
    direction.push_back(-1.f);


    for (std::size_t i = 0; i < seg.size()-1; ++i)
    {
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);

        previous.push_back(points[i]);
        previous.push_back(points[i]);

        next.push_back(points[std::min(i+2, points.size()-1)]);
        next.push_back(points[std::min(i+2, points.size()-1)]);

        direction.push_back(2.f);
        direction.push_back(-2.f);
    }

    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);

    previous.push_back(points[points.size()-2]);
    previous.push_back(points[points.size()-2]);

    direction.push_back(3.f);
    direction.push_back(-3.f);
}

void polylineOnSphereIndices(
        const std::vector<glm::vec3>& points,
        std::vector<glm::vec3>& vertices,
        std::vector<glm::vec3>& previous,
        std::vector<glm::vec3>& next,
        std::vector<float>& direction,
        std::vector<unsigned int>& indices)
{
    vertices.push_back(points[0]);
    vertices.push_back(points[0]);

    previous.push_back(points[0]);
    previous.push_back(points[0]);

    next.push_back(points[1]);
    next.push_back(points[1]);

    direction.push_back(1.f);
    direction.push_back(-1.f);

    for (std::size_t i = 0; i < points.size()-2; ++i)
    {
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);

        previous.push_back(points[i]);
        previous.push_back(points[i]);

        next.push_back(points[std::min(i+2, points.size()-1)]);
        next.push_back(points[std::min(i+2, points.size()-1)]);

        direction.push_back(2.f);
        direction.push_back(-2.f);
    }

    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);

    previous.push_back(points[points.size()-2]);
    previous.push_back(points[points.size()-2]);

    direction.push_back(3.f);
    direction.push_back(-3.f);

    unsigned int offset;
    for (unsigned int i = 0; i < points.size()-1; ++i)
    {
        offset = i * 2;
        indices.push_back(offset + 0);
        indices.push_back(offset + 1);
        indices.push_back(offset + 2);

        indices.push_back(offset + 2);
        indices.push_back(offset + 3);
        indices.push_back(offset + 1);
    }
}

void polylineOnSphereIndices2(
        const std::vector<glm::vec3>& points,
        std::vector<glm::vec3>& vertices,
        std::vector<glm::vec2>& direction,
        std::vector<unsigned int>& indices)
{
    vertices.reserve((points.size()-1) * 4 + 12);
    direction.reserve((points.size()-1) * 4 + 4);
    indices.reserve((points.size()-1) * 6);

    vertices.push_back(points[0]);
    vertices.push_back(points[0]);
    vertices.push_back(points[0]);
    vertices.push_back(points[0]);
    vertices.push_back(points[0]);
    vertices.push_back(points[0]);

    // 1st param: orientation; 2nd param: flag 0 = begining, 1 = end of the segment
    direction.push_back(glm::vec2(-1.f,  1.f));
    direction.push_back(glm::vec2(-1.f, -1.f));


    for (std::size_t i = 0; i < points.size()-1; ++i)
    {
        // 4 times the same points to have 4 differents parameters
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);

        // end of the segment
        // Top
        direction.push_back(glm::vec2(1.f,  2.f));
        // Bottom 
        direction.push_back(glm::vec2(1.f, -2.f));

        // begining of the segment
        // Top
        direction.push_back(glm::vec2(-1.f, 2.f));
        // Bottom
        direction.push_back(glm::vec2(-1.f,-2.f));
    }

    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);

    direction.push_back(glm::vec2(1.f, 3.f));
    // Bottom
    direction.push_back(glm::vec2(1.f,-3.f));

    unsigned int offset;
    for (unsigned int i = 0; i < points.size()-1; ++i)
    {
        offset = i * 4;
        indices.push_back(offset + 0);
        indices.push_back(offset + 1);
        indices.push_back(offset + 2);

        indices.push_back(offset + 2);
        indices.push_back(offset + 1);
        indices.push_back(offset + 3);
    }
}

void polylineOnSphereIndices2Update(
    const std::vector<glm::vec3>& points,
    std::vector<glm::vec3>& vertices)
{
    const size_t arraySize = (points.size() - 1) * 4 + 12;

    vertices.clear();
    vertices.reserve(arraySize);

	vertices.push_back(points[0]);
	vertices.push_back(points[0]);
	vertices.push_back(points[0]);
	vertices.push_back(points[0]);
	vertices.push_back(points[0]);
	vertices.push_back(points[0]);

	for (std::size_t i = 0; i < points.size() - 1; ++i)
	{
		// 4 times the same points to have 4 differents parameters
		vertices.push_back(points[i + 1]);
		vertices.push_back(points[i + 1]);
		vertices.push_back(points[i + 1]);
		vertices.push_back(points[i + 1]);
	}

	vertices.push_back(points[points.size() - 1]);
	vertices.push_back(points[points.size() - 1]);
	vertices.push_back(points[points.size() - 1]);
	vertices.push_back(points[points.size() - 1]);
	vertices.push_back(points[points.size() - 1]);
	vertices.push_back(points[points.size() - 1]);
}

void maskedPolylineOnSphereIndices2(
        const std::vector<glm::vec3>& points,
        std::vector<glm::vec3>& vertices,
        std::vector<glm::vec2>& direction,
        std::vector<unsigned int>& indices,
		std::vector<bool>& mask)
{
    vertices.reserve((points.size()-1) * 4 + 12);
    direction.reserve((points.size()-1) * 4 + 4);
    indices.reserve((points.size()-1) * 6);

    vertices.push_back(points[0]);
    vertices.push_back(points[0]);
    vertices.push_back(points[0]);
    vertices.push_back(points[0]);
    vertices.push_back(points[0]);
    vertices.push_back(points[0]);

    // 1st param: orientation; 2nd param: flag 0 = begining, 1 = end of the segment
    direction.push_back(glm::vec2(-1.f,  1.f));
    direction.push_back(glm::vec2(-1.f, -1.f));

    for (std::size_t i = 0; i < points.size()-1; ++i)
    {
        // 4 times the same points to have 4 differents parameters
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);
        vertices.push_back(points[i+1]);

        // end of the segment
        // Top
        direction.push_back(glm::vec2(1.f,  2.f));
        // Bottom 
        direction.push_back(glm::vec2(1.f, -2.f));

        // begining of the segment
        // Top
        direction.push_back(glm::vec2(-1.f, 2.f));
        // Bottom
        direction.push_back(glm::vec2(-1.f,-2.f));
    }

    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);
    vertices.push_back(points[points.size()-1]);

    direction.push_back(glm::vec2(1.f, 3.f));
    // Bottom
    direction.push_back(glm::vec2(1.f,-3.f));

    unsigned int offset;
	unsigned int pointsCount = static_cast<unsigned int>(points.size());
    for (unsigned int i = 0; i < pointsCount-1; ++i)
    {
        offset = i * 4;

		if
			(  false == mask[i]
			|| ((i < pointsCount - 2) && (false == mask[i+1])))
		{
			continue;
		}
        indices.push_back(offset + 0);
        indices.push_back(offset + 1);
        indices.push_back(offset + 2);

        indices.push_back(offset + 2);
        indices.push_back(offset + 1);
        indices.push_back(offset + 3);
    }
}
