// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/math_func.h"

#include "3d/unit_sphere.h"

float shortestSphericalDistance(float a, float b, bool& change)
{
    float dist0 = std::abs(a - b);
    
    if (sign(a) == sign(b))
    {
        change = false;
        return dist0;
    }
    else
    {
        float dist1 = static_cast<float>(std::abs(sign(a) * 180.0 - static_cast<double>(a)) + std::abs(sign(b) * 180.0 - static_cast<double>(b)));
        
        if (dist0 < dist1)
        {
            change = false;
            return dist0;
        }
        else
        {
            change = true;
            return dist1;
        }
    }
}

bool containNorthPole(const glm::vec2 corners[4])
{
    glm::vec3 corners3[4];
    float xDots[4];
    float yDots[4];
    
    const glm::vec3 northAxisY(0.f, 1.f, 0.f);
    const glm::vec3 northAxisX(1.f, 0.f, 0.f);
    
    for (int i=0; i < 4; ++i)
    {
        corners3[i] = UnitSphere::ToVector3D(Geodetic2D(glm::radians(corners[i])));
        corners3[i] = glm::normalize(corners3[i]);

        yDots[i] = glm::dot(northAxisY, corners3[i]);
        xDots[i] = glm::dot(northAxisX, corners3[i]);
    }
    
    bool xMinus = false;
    bool yPlus = false;
    bool yMinus = false;
    
    for (int i=0; i < 4; ++i)
    {
        if (yDots[i] < 0)
        {
            yMinus = true;
        }
        else if (yDots[i] > 0)
        {
            yPlus = true;
        }
        
        if (xDots[i] < 0)
        {
            xMinus = true;
        }
        /*
        else if (xDots[i] > 0)
        {
            xPlus = true;
        }
        */
    }
    
    return xMinus && yMinus && yMinus && yPlus;
}

int containPole(const glm::vec2 corners[5])
{
    glm::vec3 corners3[5];
    float xDots[5];
    float yDots[5];
    
    const glm::vec3 northAxisY(0.f, 1.f, 0.f);
    const glm::vec3 northAxisX(1.f, 0.f, 0.f);
    const glm::vec3 northAxisZ(0.f, 0.f, 1.f);
    
    for (int i=0; i < 5; ++i)
    {
        corners3[i] =UnitSphere::ToVector3D(Geodetic2D(glm::radians(corners[i])));
        corners3[i] = glm::normalize(corners3[i]);

        yDots[i] = glm::dot(northAxisY, corners3[i]);
        xDots[i] = glm::dot(northAxisX, corners3[i]);
    }
        
    bool xMinus = false;
    bool yPlus = false;
    bool yMinus = false;
    
    for (int i=0; i < 4; ++i)
    {
        if (yDots[i] < 0)
        {
            yMinus = true;
        }
        else if (yDots[i] > 0)
        {
            yPlus = true;
        }
        
        if (xDots[i] < 0)
        {
            xMinus = true;
        }
        /*
        else if (xDots[i] > 0)
        {
            xPlus = true;
        }
        */
    }
    
    if (xMinus && yMinus && yMinus && yPlus)
    {
        float zCenterDot = glm::dot(northAxisZ, corners3[4]);
        
        if (zCenterDot > 0)
        {
            return 1; // North pole
        }
        else
        {
            return -1; // South pole
        }
    }
    
    return 0;
}
