// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   mesh_creator.cpp
 * Author: gleseach
 * 
 * Created on 14 septembre 2015, 11:51
 */

#include "3d/stdafx.h"

#include "3d/array.h"

#include "3d/mesh_creator.h"
#include "3d/vertex_array_object.h"
#include "3d/vertex_attrib_array.h"
#include "3d/buffer.h"
#include "3d/static_mesh.h"
#include "3d/program_bind_helper.h"
#include "3d/context.h"
#include "3d/helper.h"
#include "3d/meta.h"

#include "core/log.h"

MeshCreator::MeshCreator()
: draw_type{OGL::DrawType::DrawTypeCount},
  object_type{OGL::Primitive::PrimitiveCount}
{
}

MeshCreator::~MeshCreator()
{
}

OGL::StaticMesh * MeshCreator::createStaticMesh(OGL::Context* context)
{
    return this->createStaticMesh(context, nullptr);
}

OGL::StaticMesh * MeshCreator::createStaticMesh(OGL::Context* context, OGL::ProgramBindHelper* program)
{
    if ((nullptr == this->vertices.getDataPtr()))
    {
        return nullptr;
    }
    
    if (nullptr != program)
    {
        // check data compatibility with program
        if (program->hasTextureCoordinates() == false && 
            (nullptr != texcoords.getDataPtr()))
        {
            LOG(SEAScope::LogLevel::Warn, "StaticMesh",
                               "Texture coordinates data found but not used by the shader");
        }

        if (program->hasTextureCoordinates() == true && 
            (nullptr == texcoords.getDataPtr()))
        {
            LOG(SEAScope::LogLevel::Warn, "StaticMesh",
                               "The program need texture coordinates but no data found");
        }

        if (program->hasTextureCoordinates() == true &&
            this->textures.empty() == true)
        {
            LOG(SEAScope::LogLevel::Warn, "StaticMesh",
                               "The program have texture coordinates but no texture found");
        }
    }

    OGL::StaticMesh * mesh = new OGL::StaticMesh;
    mesh->create(context);

    OGL::VertexArrayObject * vao = context->createVertexArrayObject();

    vao->create(context);

    // Vertices data
    OGL::Buffer * vertices_buffer = OGL::Helper::createBuffer(context, OGL::array_buffer, vertices);
    
    mesh->addBuffer(vertices_buffer);
    
    int vertices_location = 0;
    
    if (nullptr != program)
    {
        vertices_location = program->getVerticesLocation();
    }

    OGL::VertexAttribArray* vertices_attrib = context->createVertexAttribArray(static_cast<unsigned int>(vertices_location), this->vertices.getDescription().getComponentCount(), vertices_buffer);
    
    vao->addObject(vertices_attrib);

    if (nullptr != texcoords.getDataPtr())
    {
        OGL::Buffer * texcoords_buffer = OGL::Helper::createBuffer(context, OGL::array_buffer, this->texcoords);
        
        mesh->addBuffer(texcoords_buffer);

        int texCoords_location = 1;
        
        if (nullptr != program)
        {
            texCoords_location = program->getTextureCoordinatesLocation();
        }
        
        OGL::VertexAttribArray* texcoords_attrib = context->createVertexAttribArray(static_cast<unsigned int>(texCoords_location), this->texcoords.getDescription().getComponentCount(), texcoords_buffer);
        vao->addObject(texcoords_attrib);
    }
    
    if (!this->attributes.empty())
    {
        for (Attribute& attrib : this->attributes)
        {
            OGL::Buffer * buffer = OGL::Helper::createBuffer(context, OGL::array_buffer, attrib.array);
            
            mesh->addBuffer(buffer);
            
            OGL::VertexAttribArray* vertex_attrib = context->createVertexAttribArray(attrib.binding, attrib.array.getDescription().getComponentCount(), buffer);
            vao->addObject(vertex_attrib);
        }
    }
    
    if (nullptr != this->indices.getDataPtr())
    {
        OGL::Buffer * indices_buffer = OGL::Helper::createBuffer(context, OGL::element_array_buffer, indices);
        
        mesh->addBuffer(indices_buffer);

        vao->addObject(indices_buffer);
    }
    
    vao->fill();

    mesh->setVao(vao);
    
    if (OGL::draw_arrays == this->draw_type)
    {
        mesh->setDrawDescription(this->draw_type, this->object_type, DataType::none, this->vertices.getDescription().getLength());
    }
    else
    {
        mesh->setDrawDescription(this->draw_type, this->object_type, indices.getDescription().getDataType(), this->indices.getDescription().getLength());
    }

    if (false == textures.empty())
    {
        mesh->setTextureMap(textures);
    }

    if (nullptr != program)
    {
        mesh->setProgramHelper(program);
    }
    
    return mesh;
}

void MeshCreator::setVertices(const Array& verticesIn)
{
    this->vertices = verticesIn;
}

void MeshCreator::setTextureCoordinates(const Array& texcoordsIn)
{
    this->texcoords = texcoordsIn;
}

void MeshCreator::setIndices(const Array& indicesIn)
{
    this->indices = indicesIn;
}

void MeshCreator::setDrawType(OGL::DrawType draw_typeIn, OGL::Primitive object_typeIn)
{
    this->draw_type = draw_typeIn;
    this->object_type = object_typeIn;
}

void MeshCreator::setTexture(OGL::Texture * texture, unsigned int texture_channel)
{
    this->textures[texture_channel] = texture; 
}

void MeshCreator::addAttributeArray(int binding, const Array& array)
{
    Attribute attribute;
    attribute.binding        = static_cast<unsigned int>(binding);
    attribute.array          = array;
    
    this->attributes.push_back(attribute);
}
