// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/data_rendering_param.h"
#include "3d/particle_grid.h"

#include "3d/context.h"
#include "3d/shader_manager.h"
#include "3d/shader_list.h"
#include "3d/program_bind_helper.h"
#include "3d/program.h"
#include "3d/static_mesh.h"
#include "3d/math_func.h"
#include "3d/texture.h"
#include "data/image.h"
#include "3d/ppm.h"
#include "3d/data_set.h"
#include "3d/meta.h"
#include "3d/helper.h"
#include "3d/pipeline_parameters.h"
#include "3d/particles_data.h"
#include "core/copy.h"

#include "core/structs.h"

ParticleGrid::ParticleGrid()
{
    _indirectionTexture = nullptr;
    _atlasTexture = nullptr;
    _currentProgram = nullptr;
    _mode = barb;
    _particleLocalBarbsProgram = nullptr;
    _particleLocalBarbsFalseColorsProgram = nullptr;
}

ParticleGrid::~ParticleGrid()
{
    delete _indirectionTexture;
    _indirectionTexture = nullptr;

    delete _atlasTexture;
    _atlasTexture = nullptr;

    delete _particleLocalBarbsProgram;
    _particleLocalBarbsProgram = nullptr;
    delete _particleLocalBarbsFalseColorsProgram;
    _particleLocalBarbsFalseColorsProgram = nullptr;
}

//void ParticleGrid::create(OGL::Context* ctx, ParticleMode particleMode)
//{
//    particleMode = ParticleMode::arrow;
//
//    _particleLocalBarbsProgram = ctx->getShaderManager()->createProgramAutoBind(vertex_uv_particle, fragment_particle_arrow);
//    _particleLocalBarbsFalseColorsProgram = ctx->getShaderManager()->createProgramAutoBind(vertex_uv_particle, fragment_particle_arrow_false_colors);
//
//    int width, height;
//    char* pixels = nullptr;
//
//    //Image tileImg;
//    //tileImg.allocate(static_cast<unsigned int>(tileSize), static_cast<unsigned int>(tileSize), 1, DataType::unsigned_byte, nullptr);
//
//    Image atlasImg;
//    //atlasImg.allocate(static_cast<unsigned int>(128), static_cast<unsigned int>(128), 1, DataType::unsigned_byte, nullptr);
//    atlasImg.allocate(static_cast<unsigned int>(128), static_cast<unsigned int>(128), 3, DataType::unsigned_byte, nullptr);
//
//
//    void * imgData = atlasImg.data();
//
//    glm::u8vec3 patterns[5] = 
//    {
//        glm::u8vec3(255, 0, 0),
//        glm::u8vec3(0, 255, 0),
//        glm::u8vec3(0, 0, 255),
//        glm::u8vec3(255, 255, 0),
//        glm::u8vec3(255, 0, 255)
//    };
//
//    //std::vector<unsigned int> sizes = {64, 32, 16};
//    std::vector<unsigned int> sizes = {64, 32};
//
//    _atlasTexture = OGL::Helper::createTexture(ctx, atlasImg);
//
//    _atlasTexture = ctx->createTexture(OGL::texture2D);
//    _atlasTexture->create(ctx);
//
//    std::vector<Image> mipmaps(sizes.size());
//
//    std::string pgmPath;
//
//    for (unsigned int i = 0; i < sizes.size(); ++i)
//    {
//        std::stringstream ss;
//        //ss << "arrow_" << sizes[i] << ".pgm";
//        ss << "barbs_65_" << sizes[i] << ".pgm";
//        //pgmPath = ctx->getResources()->arrowsDir + ss.str();
//        pgmPath = ctx->getResources()->windBarbsDir + ss.str();
//
//        int width, height;
//        char* pixels = nullptr;
//
//        if
//            (true == loadPGM(pgmPath.c_str(), width, height, pixels))
//        {
//
//        }
//
//        //mipmaps[i].allocate(sizes[i], sizes[i], 3, DataType::unsigned_byte, nullptr);
//        mipmaps[i].allocate(sizes[i], sizes[i], 1, DataType::unsigned_byte, pixels);
//
//        //size_t count = fillBuffer(mipmaps[i].data(), mipmaps[i].dataSize(), &patterns[i].r, sizeof(glm::u8vec3));
//    }
//
//    _atlasTexture->allocate2D(mipmaps);
//}

void ParticleGrid::create(OGL::Context* ctx, ParticleMode particleMode)
{
    _particleLocalBarbsProgram = ctx->getShaderManager()->createProgramAutoBind(vertex_uv_particle, fragment_particle_barb);
    _particleLocalBarbsFalseColorsProgram = ctx->getShaderManager()->createProgramAutoBind(vertex_uv_particle, fragment_particle_barb_false_colors);

    int tileSize = 128; // size + 1px border around.
    int tileCount = 0;

    if (barb == particleMode)
    {
        tileCount = 21;
    }
    else
    {
        tileCount = 1;
    }

    // Number of tiles
    int atlasSize = tileSize * tileCount;

    // Compute the size of the atlas. A power of two, thanks OpenGL ES 2 !
    if (false == isPowerOfTwo(static_cast<unsigned int>(atlasSize)))
    {
        atlasSize = static_cast<int>(nextPowerOfTwo(static_cast<unsigned int>(atlasSize)));
    }

    // Divide by 4 to have the size of the texture edge.
    atlasSize = std::max(static_cast<int>(std::ceil(static_cast<float>(atlasSize) / 4.f)), tileSize);

    // Compute the number of tiles in a row
    int tilePerRow = atlasSize / tileSize;
    // Compute the number of tiles in a column
    int tilePerColumn = static_cast<int>(std::ceil(static_cast<float>(tileCount) / static_cast<float>(tilePerRow)));

    // Array for the origin of all tiles
    std::vector<glm::vec4> atlasCoord;
    std::vector<glm::ivec2> atlasCoordInt;

    for (int row = 0; row < tilePerColumn; ++row)
    {
        for (int col = 0; col < tilePerRow; ++col)
        {
            if (static_cast<int>(atlasCoord.size()) < tileCount)
            {
                // +1 for the border used for linear filtering to avoid conflict between tiles.

                float r = (static_cast<float>(col * tileSize) + 1.0f) / (static_cast<float>(atlasSize) - 1.0f);
                float g = (static_cast<float>(row * tileSize) + 1.0f) / (static_cast<float>(atlasSize) - 1.0f);
                float b = (static_cast<float>((col + 1) * tileSize) - 1.0f) / (static_cast<float>(atlasSize) - 1.0f);
                float a = (static_cast<float>((row + 1) * tileSize) - 1.0f) / (static_cast<float>(atlasSize) - 1.0f);

                atlasCoord.push_back(glm::vec4(r,g,b,a));

                atlasCoordInt.push_back(glm::vec2(col * tileSize, row * tileSize));
            }
            else
            {
                break;
            }
        }
    }

    // Indirection texture to retrieve the coordinates of the tile for a given wind speed value.
    _indirectionTexture = ctx->createTexture(OGL::texture1D);
    _indirectionTexture->create(ctx);

    _indirectionTexture->allocate1D(tileCount, 4, DataType::float32, &atlasCoord[0].x);

    _indirectionTexture->setFilterMode(OGL::nearest);
    _indirectionTexture->setWrapMode(OGL::clamp_to_edge);

    int width, height;
    char * pixels = nullptr;

    Image tileImg;
    tileImg.allocate(static_cast<unsigned int>(tileSize), static_cast<unsigned int>(tileSize), 1, DataType::unsigned_byte, nullptr);

    Image atlasImg;
    atlasImg.allocate(static_cast<unsigned int>(atlasSize), static_cast<unsigned int>(atlasSize), 1, DataType::unsigned_byte, nullptr);

    char buffer[10];

    for (int i = 0; i < static_cast<int>(tileCount); ++i)
    {
		std::string pgmPath;
        if (particleMode == barb)
        {
            snprintf(buffer, 10, "%d.pgm", i * 5);
			pgmPath = ctx->getResources()->windBarbsDir + std::string(buffer);
        }
        else
        {
            snprintf(buffer, 10, "%d.pgm", i * 5);
			pgmPath = ctx->getResources()->arrowsDir + std::string(buffer);
        }

        if (loadPGM(pgmPath.c_str(), width, height, pixels))
        {
            tileImg.allocate(static_cast<unsigned int>(width), static_cast<unsigned int>(height), 1, 1, pixels);
            tileImg.copyTo(glm::ivec2(0,0), atlasImg, atlasCoordInt[static_cast<std::size_t>(i)], glm::ivec2(tileSize, tileSize));
        }

        if (pixels)
        {
            free(pixels);
            pixels = nullptr;
        }
    }

    _atlasTexture = OGL::Helper::createTexture(ctx, atlasImg);
    _atlasTexture->generateMipmaps();
    _atlasTexture->setFilterMode(OGL::linear_mipmap_linear, OGL::linear);
    _atlasTexture->setAnisotropicFilter(ctx->getGPUInfoFloat(OGL::GPUInfoFloat::max_anisotropy));
}

void ParticleGrid::bind(OGL::Context* ctx, DataRenderingParam* dataParam)
{
    PipelineParameters* pipeline = ctx->pipelineParameters();

    bool useColormap = false;
    OGL::ProgramBindHelper* pbh = nullptr;

	if (dataParam->colormap)
	{
		useColormap = true;
		pbh = _particleLocalBarbsFalseColorsProgram;
	}
	else
	{
	   pbh = _particleLocalBarbsProgram;
	}

    _currentProgram = pbh->getProgram();

    ctx->setDepthMask(false);
    ctx->setFlag(OGL::cull_face, false);
    ctx->setFlag(OGL::blending, true);
    ctx->setBlendFunc(OGL::src_alpha, OGL::one_minus_src_alpha);

    ctx->activeTextureUnit(1, _indirectionTexture);
    ctx->activeTextureUnit(2, _atlasTexture);

    if (useColormap)
    {
        ctx->activeTextureUnit(4, dataParam->colormap);
    }

    pbh->bind(ctx);

    glm::ivec4 viewport = ctx->getViewportSize();

    _currentProgram->setUniformByName("projViewModel", pipeline->viewProjection);
    _currentProgram->setUniformByName("invProjViewModel", pipeline->inverseViewProjection);
    _currentProgram->setUniformByName("viewport", glm::vec2(static_cast<float>(viewport.z), static_cast<float>(viewport.w)));


    _currentProgram->setUniformByName("alpha0", dataParam->alpha);

    if (useColormap)
    {
        _currentProgram->setUniformByName("range", dataParam->range);
        _currentProgram->setUniformByName("logscale", dataParam->log ? 1.f : 0.f);
    }
    else
    {
        _currentProgram->setUniformByName("color0", dataParam->color);
    }
}

void ParticleGrid::draw(OGL::Context* ctx, DataSet* dataSet)
{
	ParticlesData* particlesData = dataSet->getParticlesData();

    OGL::Texture* texData = dataSet->getMesh_spherical()->getTexture(0);
    OGL::Texture* texMask = dataSet->getMesh_spherical()->getTexture(1);

    ctx->activeTextureUnit(0, texMask);
    ctx->activeTextureUnit(3, texData);


	float squareEdgeSize = particlesData->getParticleSize() * static_cast<float>(M_SQRT2) / 2.f;

	_currentProgram->setUniformByName("arrowTileSize", squareEdgeSize);

	_currentProgram->setUniformByName("offsetUV", particlesData->getOffsetUV());
	_currentProgram->setUniformByName("scissorUV", particlesData->getScissorUV());

    glm::vec4 scaleFactors;
    glm::vec4 offsets;
    std::uint32_t channels;
    float glFactor;

    channels = static_cast<std::uint32_t>(dataSet->getDimensionCount());
    glFactor = dataSet->getScaleFactorGL();

    auto _scaleFactors = dataSet->getScaleFactor();
    auto _offsets = dataSet->getOffset();
    for
        (std::uint32_t i=0; i < channels; ++i)
    {
        scaleFactors[static_cast<int>(i)] = _scaleFactors[i];
        offsets[static_cast<int>(i)] = _offsets[i];
    }
    _currentProgram->setUniformByName("scaleFactor", scaleFactors);
    _currentProgram->setUniformByName("offset", offsets);
    _currentProgram->setUniformByName("glFactor", glFactor);

	std::uint32_t meteo;
	if
		(this->_mode == barb)
	{
		meteo = 0;
	}
	else
	{
		meteo = 1;
	}
    
    _currentProgram->setUniformByName("meteoConvention", meteo);

    ParticlesData* particles = dataSet->getParticlesData();

    particles->bind(ctx);

	_currentProgram->setUniformByName("uvMin", particles->getUVMin());
	_currentProgram->setUniformByName("uvRange", particles->getUVRange());
	_currentProgram->setUniformByName("dist", particles->getParticleSizePixels());

    particles->draw(ctx);
}

void ParticleGrid::generateParticlesData(OGL::Context* ctx, DataSet* dataSet)
{
	ParticlesData* particlesData = new ParticlesData;
    particlesData->create(ctx, dataSet);
	dataSet->setParticlesData(particlesData);
}
