// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/particles_data.h"

#include "3d/data_set.h"
#include "3d/core_objects.h"
#include "3d/program_bind_helper.h"
#include "3d/unit_sphere.h"
#include "3d/geodetic.h"
#include "3d/random.h"
#include "3d/spherical_math.h"
#include "3d/default_value.h"
#include "3d/gcp_utils.h"
#include "core/colormap_manager.h"
#include "3d/ppm.h"
#include "core/structs.h"

ParticlesData::ParticlesData()
:	_offsetUV(0.f, 0.f),
	_scissorUV(0.f, 0.f, 1.f, 1.f),
	_dataSet(nullptr),
	_dist(1.f)
{
	_vertices_buffer = nullptr;
	_texcoords_buffer = nullptr;
	_displacement_buffer = nullptr;
	_vao = nullptr;
	_drawMode = nullptr;
	_gcp = nullptr;
	_particleSize = 40.f;
	_density = 0.5f;
	_update = false;
}

ParticlesData::~ParticlesData()
{
	delete _vertices_buffer;
	delete _texcoords_buffer;
	delete _displacement_buffer;
	delete _vao;
	delete _drawMode;
	delete _gcp;
}

void ParticlesData::create(OGL::Context* ctx, DataSet* dataSet)
{
	_dataSet = dataSet;

    computeGcps(ctx, dataSet, gcpData, _gcp, _uvMin, _uvRange);

	_distData.allocate(gcpData.width(), gcpData.height(), 2, DataType::float32, nullptr);

	constexpr float earthRadius = 6371009.f;

	_distX.resize(_distData.width(), 0.f);
	_distY.resize(_distData.height(), 0.f);

	for (unsigned int y = 0; y < _distData.height() - 1; ++y)
	{
		for (unsigned int x = 0; x < _distData.width() - 1; ++x)
		{
			glm::vec3* pixelCenter = static_cast<glm::vec3*>(gcpData.pixelData(x, y));
			glm::vec3* pixelU = static_cast<glm::vec3*>(gcpData.pixelData(x+1, y));
			glm::vec3* pixelV = static_cast<glm::vec3*>(gcpData.pixelData(x, y+1));

			glm::vec2 distUV(glm::distance(*pixelCenter, *pixelU) * earthRadius, glm::distance(*pixelCenter, *pixelV) * earthRadius);

            _distY[y] += distUV.x;
            _distX[x] += distUV.y;

			_distData.setPixelData(x, y, &distUV.x);

			if (x == _distData.width() - 2)
			{
                _distY[y] += distUV.x;
				_distData.setPixelData(x+1, y, &distUV.x);
			}

			if (y == _distData.height() - 2)
			{
                _distX[x] += distUV.y;
				_distData.setPixelData(x, y+1, &distUV.x);
			}
		}
	}

	_distData.setPixelData(_distData.width() - 1, _distData.height()-1, _distData.pixelData(_distData.width() - 2, _distData.height() - 1));

	_vertices_buffer = ctx->createBuffer(OGL::array_buffer);
	_vertices_buffer->create(ctx);

	_displacement_buffer = ctx->createBuffer(OGL::array_buffer);
	_displacement_buffer->create(ctx);

	OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(0, 2, _vertices_buffer);
	OGL::VertexAttribArray* displacements_attrib = ctx->createVertexAttribArray(2, 2, _displacement_buffer);

	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(vertices_attrib);
	_vao->addObject(displacements_attrib);

	_vao->fill();

	_drawMode = OGL::Helper::createDrawModeArrays(ctx, OGL::triangles, static_cast<unsigned int>(0));

	updateSize(ctx);
	update(ctx);
}

void ParticlesData::update(OGL::Context* ctx)
{
	auto areas = _dataSet->getVisibleArea();

	if (0u == areas.size())
	{
		_drawMode->setElementCount(0u);
	}
	else
	{
        const std::uint32_t* gcpX = _dataSet->getGcpx();
        const std::uint32_t* gcpY = _dataSet->getGcpy();
        unsigned int lonSize = _dataSet->getLon_size();

		float pixScreenSize = ctx->getPixelSize();
		float billbScreenSize = _particleSize * pixScreenSize;

		glm::vec3* pixel = nullptr;

		std::vector<glm::vec2> vertices;
		std::vector<glm::vec2> displacement;

		// Inverse density. In the UI 1.0 = max density, 0.0 = min density. 
		// Here we need the inverse to compute the size.
		const float density = 1.f - _density;
		const float limitSize = billbScreenSize + billbScreenSize * density;

        glm::vec2 limit(limitSize, limitSize);
        glm::vec2 accum(limit.x + 1.f, limit.y + 1.f);

		for (size_t areaId = 0; areaId < areas.size(); ++areaId)
		{
			glm::ivec4& bbox = areas[areaId];

			glm::ivec2 minBB(bbox.x, bbox.y);
			glm::ivec2 maxBB(bbox.z, bbox.w);

			unsigned int id = static_cast<unsigned int>(minBB.y) * lonSize + static_cast<unsigned int>(minBB.x);

			glm::ivec2 dataCoordMin(gcpX[id], gcpY[id]);

			id = static_cast<unsigned int>(maxBB.y) * lonSize + static_cast<unsigned int>(maxBB.x);

			glm::ivec2 dataCoordMax(gcpX[id], gcpY[id]);

			for (unsigned int y = 0; y < _distData.height(); ++y)
			{
				if
					(accum.y >= limit.y)
				{
					accum.x = limit.x + 1.f;
					accum.y = 0.f;

					unsigned int particleCount = static_cast<unsigned int>(std::floor(_distY[y] / limit.y));

					if
						(true == _dataSet->getSeamLess().x)
					{
						float inter = 1.f / static_cast<float>(particleCount + 1u);

						for (unsigned int x = 0; x <= particleCount; ++x)
						{
							float u = static_cast<float>(x) * inter;
							float v = static_cast<float>(y) / static_cast<float>(_distData.height() - 1u);
							addPoint(glm::vec2(u, v),
								vertices,
								displacement);
						}
					}
					else
					{
						for (unsigned int x = 0; x < _distData.width(); ++x)
						{
							pixel = static_cast<glm::vec3*>(_distData.pixelData(x, y));

							if
								(accum.x >= limit.x)
							{
								// Generate particles only in a visible area
								if
								    (static_cast<int>(x) >= dataCoordMin.x
								        && static_cast<int>(x) < dataCoordMax.x
								        && static_cast<int>(y) >= dataCoordMin.y
								        && static_cast<int>(y) < dataCoordMax.y)
								{
									float u = static_cast<float>(x) / static_cast<float>(_distData.width() - 1u);
									float v = static_cast<float>(y) / static_cast<float>(_distData.height() - 1u);

									addPoint(glm::vec2(u, v),
										vertices,
										displacement);
								}

								accum.x = 0.f;
							}
							else
							{
								accum.x += pixel->x;
							}
						}
					}
				}
				else
				{
					pixel = static_cast<glm::vec3*>(_distData.pixelData(0, y));
					accum.y += pixel->y;
				}
			}
		}

		_vertices_buffer->allocate(makeArray(vertices));
		_displacement_buffer->allocate(makeArray(displacement));

		_drawMode->setElementCount(static_cast<unsigned int>(vertices.size()));
	}
}

void ParticlesData::addPoint(const glm::vec2& point, std::vector<glm::vec2>& vertices, std::vector<glm::vec2>& displacement)
{
	vertices.push_back(point);
	vertices.push_back(point);
	vertices.push_back(point);
	vertices.push_back(point);
	vertices.push_back(point);
	vertices.push_back(point);

	displacement.push_back(glm::vec2(-1.f, 1.f));
	displacement.push_back(glm::vec2(1.f, 1.f));
	displacement.push_back(glm::vec2(1.f, -1.f));
	displacement.push_back(glm::vec2(1.f, -1.f));
	displacement.push_back(glm::vec2(-1.f, -1.f));
	displacement.push_back(glm::vec2(-1.f, 1.f));
}

void ParticlesData::updateSize(OGL::Context* ctx)
{
    float pixelSize = ctx->getPixelSize();
    constexpr float oneDegreeDistanceEquator = 111322.222222f;
    float size = oneDegreeDistanceEquator / pixelSize;
    constexpr float oneDegreeUnit = (static_cast<float>(M_PI) * 2.f) / 360.f;
    float pixelSizeUnit = oneDegreeUnit / size;

    _dist = pixelSizeUnit * _particleSize;
}

void ParticlesData::needUpdate()
{
	this->_update = true;
}

void ParticlesData::bind(OGL::Context* ctx)
{
	ctx->activeTextureUnit(5, _gcp);
}

void ParticlesData::draw(OGL::Context* ctx)
{
	updateSize(ctx);

	if
		(true == this->_update)
	{
		update(ctx);
		this->_update = false;
	}

	ctx->setFlag(OGL::cull_face, false);

	_vao->bind();
	_drawMode->draw();
	_vao->unbind();

	ctx->setFlag(OGL::cull_face, true);
}

void ParticlesData::setParticleSize(float val)
{
	this->_particleSize = val;
	this->_update = true;
}

glm::vec2 ParticlesData::getOffsetUV()
{
	return _offsetUV;
}

void ParticlesData::setOffsetUV(const glm::vec2& val)
{
	_offsetUV = val;
}

void ParticlesData::initBuffers(OGL::Context* ctx, DataSet* dataSet)
{
	glm::vec2 texcoord;

	unsigned int lonSize = dataSet->getLon_size()*2;
	unsigned int latSize = dataSet->getLat_size()*2;

	std::vector<glm::vec2> vertices;
	std::vector<glm::vec2> displacement;

	for (unsigned int lon = 0; lon < lonSize; ++lon)
	{
		for (unsigned int lat = 0; lat < latSize; ++lat)
		{
			texcoord.x = static_cast<float>(lon) / static_cast<float>(lonSize);
			texcoord.y = static_cast<float>(lat) / static_cast<float>(latSize);

			// Duplicate to make a square with two triangles
			vertices.push_back(texcoord);
			vertices.push_back(texcoord);
			vertices.push_back(texcoord);
			vertices.push_back(texcoord);
			vertices.push_back(texcoord);
			vertices.push_back(texcoord);

			displacement.push_back(glm::vec2(-1.f, 1.f));
			displacement.push_back(glm::vec2(1.f, 1.f));
			displacement.push_back(glm::vec2(1.f, -1.f));
			displacement.push_back(glm::vec2(1.f, -1.f));
			displacement.push_back(glm::vec2(-1.f, -1.f));
			displacement.push_back(glm::vec2(-1.f, 1.f));
		}
	}

	_vertices_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(vertices));
	_displacement_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(displacement));

	OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(0, 2, _vertices_buffer);
	OGL::VertexAttribArray* displacements_attrib = ctx->createVertexAttribArray(2, 2, _displacement_buffer);

	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(vertices_attrib);
	_vao->addObject(displacements_attrib);

	_vao->fill();

	_drawMode = OGL::Helper::createDrawModeArrays(ctx, OGL::triangles, static_cast<unsigned int>(vertices.size()));
}

void ParticlesData::setParticlesDensity(float value)
{
	this->_density = value;
    this->_update = true;
}
