// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/shape.h"
#include "3d/pickable.h"

#include "3d/app_enum.h"
#include "3d/shapes_allocator.h"

Pickable::Pickable(const SEAScope::RenderableId& rId)
: Renderable2{rId},
  _pickingId{0u},
  _shape{nullptr}
{
    _flags.set(pickable, true);
    _flags.set(erasable, true);
}

Pickable::Pickable(const Pickable& src) : Renderable2(src), _pickingColor(src._pickingColor),
_pickingId(src._pickingId), _style(src._style), _shape(src._shape)
{
}

Pickable::~Pickable()
{
    if (nullptr != _shape)
    {
		ShapesAllocator::release(_shape);
        _shape = nullptr;
    }
}

void Pickable::setPickingId(unsigned int id, glm::vec3 pickingColor)
{
    _pickingColor = pickingColor;
    _pickingId = id;
    //std::cout << "Pickable got ID: " << id << " and color (" << pickingColor.x << ", " << pickingColor.y << ", " << pickingColor.z << ")" << std::endl;
}

unsigned int Pickable::pickingId() const
{
    return _pickingId;
}

const glm::vec3& Pickable::pickingColor() const
{
    return _pickingColor;
}

glm::vec4 Pickable::getCurrentStyleColor() const
{
    glm::vec4 color;
    
    switch(_style.getState())
    {
    case StyleState::normal:
    {
        if (_style.isHighlighted())
        {
            color = glm::vec4(0.f, 1.f, 1.f, 1.f);
        }
        else
        {
            color = glm::vec4(0.f, 1.f, 0.f, 1.f);
        }
    }
        break;
        
    case StyleState::selected:
    {
        if (_style.isHighlighted())
        {
            color = glm::vec4(1.f, 0.5f, 0.f, 1.f);
        }
        else
        {
            color = glm::vec4(1.f, 1.f, 0.f, 1.f);
        }
    }
        break;
        
    case StyleState::edition:
        color = glm::vec4(1.f, 0.f, 1.f, 1.f);
        break;
        
    case StyleState::styleCount:
        color = glm::vec4(.5f, 0.5f, .5f, 1.f);
        break;
    }
    
    return color;
}

Style& Pickable::getStyle()
{
    return _style;
}

void Pickable::setStyle(StyleState style)
{
    _style.setState(style);
}

bool Pickable::isMoveable() const
{
    return _flags.test(moveable);
}

bool Pickable::isAnchor() const
{
    return _flags.test(anchor);
}
