// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/point_creator.h"

#include "3d/point_builder.h"
#include "3d/shapes_allocator.h"

#include "3d/context.h"
#include "3d/pipeline_parameters.h"
#include "3d/anchor_pool_node.h"

PointCreator::PointCreator()
: _builder(nullptr),
  _positionSet(false),
  _position({0, 0, 0}),
  _currentPos({0, 0, 0})
{
}

PointCreator::~PointCreator()
{
    delete this->_builder;
    this->_builder = nullptr;
}

void PointCreator::initialize(OGL::Context* /*ctx*/)
{
}

void PointCreator::setBuilder(ShapeBuilder* builder)
{
    this->_builder = dynamic_cast<PointBuilder*>(builder);
}

void PointCreator::draw(OGL::Context* /*ctx*/, const glm::mat4& /*mvp*/)
{
}

OGL::Shape* PointCreator::createShape()
{
    OGL::Shape* shape(nullptr);

    if
        (  (nullptr != this->_builder)
        && (true == this->_positionSet))
    {
        this->_builder->setPosition(this->_position);
        shape = this->_builder->build();
    }

    this->reset();

    return shape;
}

void PointCreator::reset()
{
    this->_positionSet = false;
    this->_anchorPool->clear();
}

OGL::Shape* PointCreator::getTemporaryShape()
{
    return nullptr;
}

SEAScope::ShapeType PointCreator::getShapeType()
{
    return this->_builder->getShapeType();
}

void PointCreator::addVertexUnderCursor()
{
    if
        (false == this->_positionSet)
    {
        this->_anchorPool->addAnchor(this, AnchorType::PointMoveable, 0);
    }

    this->move(AnchorType::PointMoveable, 0, this->_currentPos);

    this->_position = this->_currentPos;
    this->_positionSet = true;
}

void PointCreator::setAnchorPool(AnchorPoolNode* val)
{
    ShapeCreator::setAnchorPool(val);

    this->_anchorPool->clear();
}

void PointCreator::updateCursorPosition(const glm::vec3& position, std::uint32_t, std::uint32_t)
{
    this->_currentPos = position;
}


/**
 *
 */
unsigned int PointCreator::getAnchorCount() const
{
    if
        (true == this->_positionSet)
    {
        return 1;
    }
    return 0;
}

glm::vec3 PointCreator::getAnchorPos(AnchorType, unsigned int) const
{
    return this->_position;
}

void PointCreator::move(AnchorType, unsigned int, const glm::vec3& pos)
{
    this->_position = pos;
}

void PointCreator::move(AnchorType type, unsigned int vertexId)
{
    this->move(type, vertexId, this->_currentPos);
}

void PointCreator::pressed(AnchorType, unsigned int)
{

}

void PointCreator::remove(AnchorType, unsigned int)
{
    if
        (true == this->_positionSet)
    {
        this->_anchorPool->clear();
        this->_positionSet = false;
    }
}
