// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/polygon_mesh.h"

#include "3d/core_objects.h"
#include "3d/polygon.h"

PolygonMesh::PolygonMesh()
: ShapeMesh(SEAScope::RenderableId())
{
	_buffer = nullptr;
	_indicesBuffer = nullptr;
	_vao = nullptr;
	_drawMode = nullptr;
	_shape = nullptr;

	_flags.set(RenderableFlags::editable, true);
	_flags.set(RenderableFlags::erasable, true);
}

PolygonMesh::~PolygonMesh()
{
	_points.clear();

	delete _buffer;
	_buffer = nullptr;

	delete _indicesBuffer;
	_indicesBuffer = nullptr;

	delete _vao;
	_vao = nullptr;

	delete _drawMode;
	_drawMode = nullptr;

    _shape = nullptr;   // deleted elsewhere
}

void PolygonMesh::create(OGL::Context* ctx, const std::vector<glm::vec3>& points, const std::vector<unsigned int>& indices, Polygon* shape)
{
	_shape = shape;
	_points = points;

    if
        (nullptr != _buffer)
    {
        delete _buffer;
        _buffer = nullptr;
    }
	_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(points));
	OGL::VertexAttribArray* vaa = ctx->createVertexAttribArray(0, 3, _buffer, DataType::float32);

    if
        (nullptr != _indicesBuffer)
    {
        delete _indicesBuffer;
        _indicesBuffer = nullptr;
    }
	_indicesBuffer = OGL::Helper::createBuffer(ctx, OGL::element_array_buffer, makeArray(indices));

    if
        (nullptr != _vao)
    {
        delete _vao;
        _vao = nullptr;
    }
	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(vaa);
	_vao->addObject(_indicesBuffer);
	_vao->fill();

    if
        (nullptr != _drawMode)
    {
        delete _drawMode;
        _drawMode = nullptr;
    }
	_drawMode = OGL::Helper::createDrawModeElements(ctx, OGL::triangles, DataType::unsigned_int, static_cast<unsigned int>(indices.size()));

	setStyle(normal);
}

void PolygonMesh::draw(OGL::Context* /*ctx*/, const glm::mat4& /*mvp*/)
{
	_vao->bind();

	_drawMode->draw();

	_vao->unbind();
}

void PolygonMesh::drawPicking(OGL::Context* ctx, const glm::mat4& /*mvp*/)
{
	ctx->setDepthMask(false);

	_vao->bind();

	_drawMode->draw();

	_vao->unbind();

	ctx->setDepthMask(true);
}
