// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/polyline_creator.h"

#include "3d/shape.h"
#include "3d/core_objects.h"
#include "3d/polyline_mesh.h"
#include "3d/geodetic.h"
#include "3d/ellipsoid.h"
#include "3d/geodetic.h"
#include "3d/spherical_math.h"
#include "3d/math_func.h"
#include "3d/geometry.h"

#include "3d/shapes_allocator.h"
#include "3d/polyline_based_builder.h"
#include "3d/polyline.h"

#include "core/log.h"

PolylineCreator::PolylineCreator()
: _builder(nullptr),
  _program(nullptr),
  _buffer(nullptr),
  _vao(nullptr),
  _drawMode(nullptr),
  _lastPos({0, 0, 0}),
  _currentPos({0, 0, 0}),
  _tmpShape(nullptr),
  _p0Loc(0),
  _p1Loc(0),
  _mvpLoc(0),
  _colorLoc(0),
  _allowIntersection(false),
  _currentPointValid(false),
  _closed(false),
  _closedValid(true)
{
}

PolylineCreator::~PolylineCreator()
{
	delete _buffer;
	_buffer = nullptr;
	delete _program;
	_program = nullptr;
	delete _vao;
	_vao = nullptr;
	delete _drawMode;
	_drawMode = nullptr;
    delete _builder;
    _builder = nullptr;

	ShapesAllocator::release(_tmpShape);
    _tmpShape = nullptr;
}

void PolylineCreator::initialize(OGL::Context* ctx)
{
	_program = OGL::Helper::createProgram(ctx, vertex_line_tesselate, fragment_uni_color);

	_program->bind();
	_p0Loc = _program->getUniformLocation("p0");
	_p1Loc = _program->getUniformLocation("p1");
	_mvpLoc = _program->getUniformLocation("mvp");
	_colorLoc = _program->getUniformLocation("color");
	_program->unbind();

	const int segmentCount = 20;
	std::vector<float> percents;
	percents.reserve(segmentCount);

	for (int i = 0; i < segmentCount; ++i)
	{
		percents.push_back(static_cast<float>(i) / (segmentCount - 1));
	}

	_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(percents));
	OGL::VertexAttribArray* vaa = ctx->createVertexAttribArray(0, 1, _buffer, DataType::float32);

	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(vaa);
	_vao->fill();

	_drawMode = OGL::Helper::createDrawModeArrays(ctx, OGL::line_strip, segmentCount);

	_tmpShape = ShapesAllocator::getShape<Polyline>();
	_tmpShape->addPoint(_currentPos);
}

void PolylineCreator::setBuilder(ShapeBuilder* builder)
{
	_builder = dynamic_cast<PolylineBasedBuilder*>(builder);
}

void PolylineCreator::draw(OGL::Context* /*ctx*/, const glm::mat4& mvp)
{
	if (_points.size() > 0)
	{
		_program->bind();

		_program->setUniform(_mvpLoc, mvp);

		_vao->bind();

		for (unsigned int i = 0; i < _points.size() - 1; ++i)
		{
			_program->setUniform(_colorLoc, _colors[i]);
			_program->setUniform(_p0Loc, _points[i]);
			_program->setUniform(_p1Loc, _points[i + 1]);

			_drawMode->draw();
		}

		_program->setUniform(_colorLoc, green4f);
		_program->setUniform(_p0Loc, *_points.rbegin());
		_program->setUniform(_p1Loc, _currentPos);

		_drawMode->draw();

		if (true == _closed && _points.size() > 2)
		{
			if (_closedValid)
				_program->setUniform(_colorLoc, green4f);
			else
				_program->setUniform(_colorLoc, red4f);

			_program->setUniform(_p0Loc, _points[0]);
			_program->setUniform(_p1Loc, _currentPos);

			_drawMode->draw();
		}

		_vao->unbind();
	}
}

OGL::Shape* PolylineCreator::createShape()
{
	OGL::Shape* shape = nullptr;

	if (_builder)
	{
		_builder->setPoints(std::move(_points));
		shape = _builder->build();
	}

	return shape;
}

void PolylineCreator::reset()
{
	_points.clear();

    this->_tmpShape->clear();
	this->_tmpShape->addPoint(_currentPos);
}

OGL::Shape* PolylineCreator::getTemporaryShape()
{
	return _tmpShape;
}

SEAScope::ShapeType PolylineCreator::getShapeType()
{
	return _builder->getShapeType();
}

void PolylineCreator::allowSegmentIntersection(bool status)
{
	_allowIntersection = status;
}

void PolylineCreator::setClosed(bool status)
{
	_closed = status;
}

void PolylineCreator::addPoint(const glm::vec3& geo)
{
    if
        ((0 < _points.size()) && (geo == _points.back()))
    {
        LOG(SEAScope::LogLevel::Debug, "PolylineCreator",
                           "Vertex not added because it is at the same location as the trailing vertex");
        return;
    }

	_points.push_back(geo);
	_colors.push_back(green4f);

	_tmpShape->setLastPoint(geo);
	_tmpShape->addPoint(geo); // new point for current cursor position
}

void PolylineCreator::updateArcIntersection()
{
	if (_points.size() > 2)
	{
		glm::vec3 resultPos;

		for (unsigned int i = 0; i < _colors.size(); ++i)
		{
			_colors[i] = green4f;
		}

		_currentPointValid = true;

		for (unsigned int i = 0; i < _points.size() - 2; ++i)
		{
			if (arcIntersectUnitSphere(_currentPos, _points[_points.size() - 1], _points[i], _points[i + 1], resultPos))
			{
				_colors[i] = red4f;

				if (true == _currentPointValid)
				{
					_currentPointValid = false;
				}

				break;
			}
		}

		if (true == _closed)
		{
			_closedValid = true;

			for (unsigned int i = 1; i < _points.size() - 1; ++i)
			{
				if (arcIntersectUnitSphere(_currentPos, _points[0], _points[i], _points[i + 1], resultPos))
				{
					_closedValid = false;

					if (true == _currentPointValid)
					{
						//_currentPointValid = false;
					}

					break;
				}
			}
		}
	}
}

/**
 *
 */
void PolylineCreator::addVertexUnderCursor()
{
    if
        (true == this->_currentPointValid)
    {
        this->addPoint(this->_currentPos);
    }
}

/**
 *
 */
void PolylineCreator::updateCursorPosition(const glm::vec3& position, std::uint32_t /*screenX*/, std::uint32_t /*screenY*/)
{
    this->_currentPos = position;
    this->_tmpShape->setLastPoint(position);
	this->_currentPointValid = true;

    if
        (false == this->_allowIntersection)
    {
        this->updateArcIntersection();
    }
}

/**
 *
 */
unsigned int PolylineCreator::getAnchorCount() const
{
    return 0;
}

/**
 *
 */
glm::vec3 PolylineCreator::getAnchorPos(AnchorType, unsigned int) const
{
    return {0, 0, 0};
}

/**
 *
 */
void PolylineCreator::move(AnchorType, unsigned int, const glm::vec3&)
{
}

/**
 *
 */
void PolylineCreator::move(AnchorType, unsigned int)
{
}

/**
 *
 */
void PolylineCreator::pressed(AnchorType, unsigned int)
{
}

/**
 *
 */
void PolylineCreator::remove(AnchorType, unsigned int)
{
}
