// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/polyline_editor.h"

#include "3d/polyline_builder.h"

#include "3d/core_objects.h"
#include "3d/polyline_mesh.h"
#include "3d/geodetic.h"
#include "3d/ellipsoid.h"
#include "3d/geodetic.h"
#include "3d/spherical_math.h"
#include "3d/math_func.h"
#include "3d/geometry.h"
#include "3d/polyline_base.h"
#include "3d/scene_graph_node.h"
#include "3d/shape_anchor.h"
#include "3d/anchor_pool_node.h"


static bool hasIntersection(const PolylineBase* p, const unsigned int vertexId, const bool asLinearRing)
{
    if
        (nullptr == p)
    {
        return false;
    }

	unsigned int count = p->getPointCount();
    if
        (vertexId >= count)
    {
        // The modified vertex does not seem to be in the polyline
        return false;
    }

    if
        (3 > count)
    {
        // Intersections are impossible with less than 3 points
        return false;
    }



    /*
    for (unsigned int i = 0; i < _colors.size(); ++i)
    {
        _colors[i] = green4f;
    }
    */

    // Configure check for segment leading to ref point
    unsigned int leadingPoint1Id {0u};
    unsigned int leadingPoint2Id {0u};
    bool checkLead {true};

    leadingPoint2Id = vertexId;
    if
        (0 == vertexId)
    {
        if
            (true == asLinearRing)
        {
            // this is a linear ring, so the segment leading to the ref vertex
            // comes from the last vertex
            leadingPoint1Id = count - 1u;
        }
        else
        {
            // Not a closed shape, so if ref point is the first one then there
            // is no segment leading to it
            checkLead = false;
        }
    }
    else
    {
        leadingPoint1Id = vertexId - 1u;
    }

    // Configure check for segment leaving ref point
    unsigned int tailingPoint1Id {0u}; //
    unsigned int tailingPoint2Id {0u}; //
    bool checkTail {true};  //

    tailingPoint1Id = vertexId;
    if
        (vertexId == count - 1u)
    {
        if
            (true == asLinearRing)
        {
            // This is a liner ring, so the segment leaving the ref vertex goes
            // to the first vertex
            tailingPoint2Id = 0u;
        }
        else
        {
            // Not a closed shape, so if ref point is the last one then there
            // is no segment leaving from it
            checkTail = false;
        }
    }
    else
    {
        tailingPoint2Id = vertexId + 1u;
    }

    glm::vec3 refSegmentPoint1;
    glm::vec3 refSegmentPoint2;
    glm::vec3 point1;
    glm::vec3 point2;
    glm::vec3 resultPos;
    bool intersectFound {false};    //

    // Actually check segment leading to reference point
    if
        (true == checkLead)
    {
        refSegmentPoint1 = p->getPoint(leadingPoint1Id);
        refSegmentPoint2 = p->getPoint(leadingPoint2Id);
        for
            (unsigned int i = 0; i < count - 1u; ++i)
        {
            if
                (  i == leadingPoint1Id
                || i + 1u == leadingPoint1Id
                || i == leadingPoint2Id
                || i + 1u == leadingPoint2Id)
            {
                // Skip
                // TODO: check colinear
                continue;
            }

            point1 = p->getPoint(i);
            point2 = p->getPoint(i + 1u);

            intersectFound = arcIntersectUnitSphere(refSegmentPoint1, refSegmentPoint2,
                                                    point1, point2, resultPos);
            if
                (true == intersectFound)
            {
                return true;
            }
        }

        if
            (true == asLinearRing)
        {
            if
                (  count - 1u == leadingPoint1Id
                || count - 1u == leadingPoint2Id
                || 0 == leadingPoint1Id
                || 0 == leadingPoint2Id)
            {
                // Skip
                // TODO: check colinear
            }
            else
            {
                point1 = p->getPoint(count - 1u);
                point2 = p->getPoint(0);

                intersectFound = arcIntersectUnitSphere(refSegmentPoint1, refSegmentPoint2,
                                                        point1, point2, resultPos);
                if
                    (true == intersectFound)
                {
                    return true;
                }

            }
        }
    }

    // Actually check segment leaving from reference point
    if
        (true == checkTail)
    {
        refSegmentPoint1 = p->getPoint(tailingPoint1Id);
        refSegmentPoint2 = p->getPoint(tailingPoint2Id);
        for
            (unsigned int i = 0; i < count - 1u; ++i)
        {
            if
                (  i == tailingPoint1Id
                || i + 1u == tailingPoint1Id
                || i == tailingPoint2Id
                || i + 1u == tailingPoint2Id)
            {
                // Skip
                // TODO: check colinear
                continue;
            }

            point1 = p->getPoint(i);
            point2 = p->getPoint(i + 1u);

            intersectFound = arcIntersectUnitSphere(refSegmentPoint1, refSegmentPoint2,
                                                    point1, point2, resultPos);
            if
                (true == intersectFound)
            {
                return true;
            }
        }

        if
            (true == asLinearRing)
        {
            if
                (  count - 1u == tailingPoint1Id
                || count - 1u == tailingPoint2Id
                || 0 == tailingPoint1Id
                || 0 == tailingPoint2Id)
            {
                // Skip
                // TODO: check colinear
            }
            else
            {
                point1 = p->getPoint(count - 1u);
                point2 = p->getPoint(0);

                intersectFound = arcIntersectUnitSphere(refSegmentPoint1, refSegmentPoint2,
                                                        point1, point2, resultPos);
                if
                    (true == intersectFound)
                {
                    return true;
                }

            }
        }
    }

    return false;
}

PolylineEditor::PolylineEditor()
: ShapeEditor(), _program(nullptr), _buffer(nullptr), _vao(nullptr), _drawMode(nullptr), _asLinearRing(false), _detectSelfIntersections(false), _hasSelfIntersections(false)
{
}

PolylineEditor::~PolylineEditor()
{
	delete _buffer;
	_buffer = nullptr;
	delete _program;
	_program = nullptr;
	delete _vao;
	_vao = nullptr;
	delete _drawMode;
	_drawMode = nullptr;
}

void PolylineEditor::initialize(OGL::Context* ctx, Pickable* obj)
{
	_shapeToEdit = obj;

	_program = OGL::Helper::createProgram(ctx, vertex_line_tesselate, fragment_uni_color);

	_program->bind();
	_program->setUniformByName("color", glm::vec4(1.f, 0.f, 0.f, 1.f));
	_program->unbind();

	const int segmentCount = 20;
	std::vector<float> percents;
	percents.reserve(segmentCount);

	for (int i = 0; i < segmentCount; ++i)
	{
		percents.push_back(static_cast<float>(i) / (segmentCount - 1));
	}

	_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(percents));
	OGL::VertexAttribArray* vaa = ctx->createVertexAttribArray(0, 1, _buffer, DataType::float32);

	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(vaa);
	_vao->fill();

	_drawMode = OGL::Helper::createDrawModeArrays(ctx, OGL::line_strip, segmentCount);
}

void PolylineEditor::draw(OGL::Context* /*ctx*/, const glm::mat4& mvp)
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());

	if (p->getPointCount() > 0)
	{
		_program->bind();

		_program->setUniformByName("mvp", mvp);
        if
            (true == this->_hasSelfIntersections)
        {
            _program->setUniformByName("color", red4f);
        }
        else
        {
		    _program->setUniformByName("color", green4f);
        }

		_vao->bind();

		for (unsigned int i = 0; i < p->getPointCount() - 1; ++i)
		{
			_program->setUniformByName("p0", p->getPoint(i));
			_program->setUniformByName("p1", p->getPoint(i + 1));

			_drawMode->draw();
		}

        if
            ((true == this->_asLinearRing)
             && 2 < p->getPointCount())
        {
            _program->setUniformByName("p0", p->getPoint(p->getPointCount() - 1));
            _program->setUniformByName("p1", p->getPoint(0));
            _drawMode->draw();
        }

		_vao->unbind();
	}
}

OGL::Shape* PolylineEditor::createShape()
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());

	std::vector<glm::vec3> points = p->getPoints();
	static_cast<PolylineBuilder*>(_builder)->setPoints(std::move(points));
	return _builder->build();
}

void PolylineEditor::setBuilder(ShapeBuilder* builder)
{
    // takes ownership
	this->_builder = builder;
}

unsigned int PolylineEditor::getAnchorCount() const
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());
	return p->getPointCount();
}

glm::vec3 PolylineEditor::getAnchorPos(AnchorType type, unsigned int vertexId) const
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());

	glm::vec3 pos;

	if (AnchorType::PointMoveable == type)
	{
		pos = p->getPoint(vertexId);
	}
	else if (AnchorType::PointAdd == type)
	{
		pos = getMiddlePoint(vertexId);
	}

	return pos;
}

glm::vec3 PolylineEditor::getMiddlePoint(unsigned int id) const
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());

	glm::vec3 point;

	if ((p->getPointCount() - 1u) > id)
	{
		point = glm::normalize(glm::mix(p->getPoint(id), p->getPoint(id + 1), 0.5f));
	}

    if
        (true == this->_asLinearRing && (p->getPointCount() - 1u == id))
    {
        point = glm::normalize(glm::mix(p->getPoint(id), p->getPoint(0), 0.5f));
    }

	return point;
}

void PolylineEditor::move(AnchorType type, unsigned int vertexId, const glm::vec3& pos)
{
	if
        (AnchorType::PointMoveable == type)
	{
		PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());
		p->setPoint(vertexId, pos);

        if
            (true == this->_detectSelfIntersections)
        {
            this->_hasSelfIntersections = hasIntersection(p, vertexId, this->_asLinearRing);
        }
	}
}

void PolylineEditor::move(AnchorType type, unsigned int vertexId)
{
    this->move(type, vertexId, this->_currentPos);
}

void PolylineEditor::pressed(AnchorType type, unsigned int id)
{
	if (AnchorType::PointAdd == type)
	{
        glm::vec3 addedPos = this->getAnchorPos(AnchorType::PointAdd, id);

        unsigned int i = this->getAnchorCount();

		PolylineBase* p = static_cast<PolylineBase*>(this->_shapeToEdit->getShape());
        p->addPoint(addedPos);
        // now i is the index of the added point

        this->_anchorPool->addAnchor(this, AnchorType::PointMoveable, i);
        if
            (true == this->_asLinearRing)
        {
            this->_anchorPool->addAnchor(this, AnchorType::PointAdd, i);
        }
        else
        {
            this->_anchorPool->addAnchor(this, AnchorType::PointAdd, i - 1);
        }

        while
            (i > id + 1)
        {
            p->setPoint(i, p->getPoint(i - 1));
            i = i - 1;
        }

        p->setPoint(id + 1, addedPos);

	}
}

void PolylineEditor::setAnchorPool(AnchorPoolNode* val)
{
	ShapeEditor::setAnchorPool(val);

	this->_anchorPool->clear();

	unsigned int count = getAnchorCount();

	for (unsigned int i = 0; i < count; ++i)
	{
		this->_anchorPool->addAnchor(this, AnchorType::PointMoveable, i);
	}

	for (unsigned int i = 0; i + 1 < count; ++i)
	{
		this->_anchorPool->addAnchor(this, AnchorType::PointAdd, i);
	}

    if
        (true == this->_asLinearRing && (0u < count))
    {
        // Add an anchor between the last and the first vertices.
		this->_anchorPool->addAnchor(this, AnchorType::PointAdd, count - 1u);
    }
}

/**
 *
 */
void PolylineEditor::updateCursorPosition(const glm::vec3& position)
{
    this->_currentPos = position;
}

/**
 *
 */
void PolylineEditor::startMove(unsigned int vertexId)
{
    this->_anchorPool->showOnly(AnchorType::PointMoveable, vertexId);
}

/**
 *
 */
void PolylineEditor::stopMove()
{
    this->_anchorPool->showAll();
}

/**
 *
 */
void PolylineEditor::remove(AnchorType type, unsigned int id)
{
    if
        (AnchorType::PointMoveable != type)
    {
        return;
    }

    PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());
    if
        (  ((true == this->_asLinearRing) && 3 >= p->getPointCount())
        || ((false == this->_asLinearRing) &&  2 >= p->getPointCount()))
    {
        // it cannot be a polyline unless there are at least two vertices, or
        // a linear ring unless there are at least three vertices
        return;
    }

    // Remove one moveable anchor and one add anchor
    if
        (false == this->_asLinearRing)
    {
        // For polylines:
        // - the last moveable anchor has ID: count - 1
        // - the last add anchor has ID: count - 2
        this->_anchorPool->removeAnchor(AnchorType::PointMoveable, p->getPointCount() - 1);
        this->_anchorPool->removeAnchor(AnchorType::PointAdd, p->getPointCount() - 2);
    }
    else
    {
        // For linear rings there are as many moveable and add anchors, so the
        // last anchor of each type has ID: count - 1
        this->_anchorPool->removeAnchor(AnchorType::PointMoveable, p->getPointCount() - 1);
        this->_anchorPool->removeAnchor(AnchorType::PointAdd, p->getPointCount() - 1);
    }

    p->deletePoint(id);

    if
        (true == this->_detectSelfIntersections)
    {
        unsigned int count {0u};
        bool hasSelfIntersections {false};

        count = p->getPointCount();
        for
            (unsigned int i = 0; i < count; i = i + 2)
        {
            hasSelfIntersections = hasSelfIntersections || hasIntersection(p, i, this->_asLinearRing);
        }
        this->_hasSelfIntersections = hasSelfIntersections;
    }
}

/**
 *
 */
bool PolylineEditor::asLinearRing() const
{
    return this->_asLinearRing;
}

/**
 *
 */
bool PolylineEditor::asLinearRing(bool value)
{
    if
        (value == this->_asLinearRing)
    {
        return false;
    }

    this->_asLinearRing = value;
    return true;
}

/**
 *
 */
bool PolylineEditor::detectSelfIntersections() const
{
    return this->_detectSelfIntersections;
}

/**
 *
 */
bool PolylineEditor::detectSelfIntersections(bool value)
{
    if
        (value == this->_detectSelfIntersections)
    {
        return false;
    }

    this->_detectSelfIntersections = value;
    this->_hasSelfIntersections = false;
    return true;
}

/**
 *
 */
bool PolylineEditor::shapeIsValid() const
{
    return (false == (this->_detectSelfIntersections && this->_hasSelfIntersections));
}
