// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   polyline.cpp
 * Author: gleseach
 * 
 * Created on 30 septembre 2015, 17:45
 */

#include "3d/stdafx.h"
#include "3d/data_rendering_param.h"
#include "3d/polyline_mesh.h"

#include "3d/unit_sphere.h"
#include "3d/line.h"
#include "3d/program.h"
#include "3d/math_func.h"
#include "3d/polyline.h"

#include "3d/shapes_allocator.h"
#include "3d/buffer.h"
#include "3d/vertex_array_object.h"
#include "3d/vertex_attrib_array.h"
#include "3d/context.h"
#include "3d/draw_mode.h"
#include "3d/array.h"
#include "3d/meta.h"

#include "core/log.h"

struct Segment0
{
    bool subdiv;
    int id;
    int subdiv_count;
};

PolylineMesh::PolylineMesh()
:   ShapeMesh(SEAScope::RenderableId()),
    _verticeCount(0),
    _attributeArray(nullptr),
    _program(nullptr),
    _verticeBuffer(nullptr),
    _directionBuffer(nullptr),
    _indicesBuffer(nullptr),
    _attribBuffer(nullptr),
    _vertexArrObj(nullptr),
    _highlighted(false),
    _ctx(nullptr),
    _drawMode(nullptr)
{
	_flags.set(RenderableFlags::editable, true);
	_flags.set(RenderableFlags::erasable, true);
}

PolylineMesh::~PolylineMesh()
{
    delete this->_verticeBuffer;
    this->_verticeBuffer = nullptr;
    delete this->_directionBuffer;
    this->_directionBuffer = nullptr;
    delete this->_attribBuffer;
    this->_attribBuffer = nullptr;
    delete this->_indicesBuffer;
    this->_indicesBuffer = nullptr;
    delete this->_vertexArrObj;
    this->_vertexArrObj = nullptr;
    delete this->_attributeArray;
    this->_attributeArray = nullptr;
    delete _drawMode;
    _drawMode = nullptr;
}

void PolylineMesh::initialize(OGL::Context* ctx, const SEAScope::RenderableId& rId, std::vector<Geodetic2D>&& points, OGL::Shape* shape)
{
    this->initialize(ctx, rId, std::move(points), shape, nullptr);
}

void PolylineMesh::initialize(OGL::Context* ctx, const SEAScope::RenderableId& rId, std::vector<Geodetic2D>&& points, OGL::Shape* shape, Array* attrib)
{
    this->_ctx = ctx;
    this->_shape = shape;
    this->_verticeBuffer = this->_ctx->createBuffer(OGL::array_buffer);
    this->_directionBuffer = this->_ctx->createBuffer(OGL::array_buffer);
    this->_attribBuffer = this->_ctx->createBuffer(OGL::array_buffer);
    this->_indicesBuffer = this->_ctx->createBuffer(OGL::element_array_buffer);
    this->setRenderableId(rId);
    this->_points = std::move(points);
    this->_attributeArray = attrib;
    updateData();
}

void PolylineMesh::setPolyline(Polyline* p)
{
    if
        ((nullptr != this->_shape) && (p != this->_shape))
    {
        ShapesAllocator::release(this->_shape);
        this->_shape = nullptr;
    }
	this->_shape = p;
}

Polyline* PolylineMesh::getPolyline() const
{
	return static_cast<Polyline*>(_shape);
}

void PolylineMesh::setPoints(std::vector<Geodetic2D>&& points)
{
    this->_points = std::move(points);
    if
        (nullptr != this->_vertexArrObj)
    {
        delete this->_vertexArrObj;
        this->_vertexArrObj = nullptr;
    }
    updateData();
}

void PolylineMesh::setMask(std::vector<bool>&& mask)
{
	this->_mask = std::move(mask);
}

void PolylineMesh::setAttribute(int id, std::vector<float>&& attrib)
{
    this->_attributeMap[id] = std::move(attrib);
}

Array PolylineMesh::duplicate(const Array& data, bool& ok) const
{
    const BufferDesc& desc = data.getDescription();
    const unsigned int elementSize = desc.getElementSize();
    const unsigned int arrayLengthOut = (desc.getLength() - 1) * 4;
    const unsigned int arraySizeOut = arrayLengthOut * 4;

    ok = false;

    void* buffer = malloc(arraySizeOut);
    if
        (nullptr == buffer)
    {
        LOG(SEAScope::LogLevel::Error, "PolylineMesh",
            "Could not allocate memory for copying the polyline mesh");
        ok = false;
        return Array();
    }
    void* element = malloc(elementSize);
    if
        (nullptr == element)
    {
        LOG(SEAScope::LogLevel::Error, "PolylineMesh",
            "Could not allocate memory for copying the polyline mesh");
        ok = false;
        return Array();
    }

    char* ptr = static_cast<char*>(data.getDataPtr());
    char* ptrOut = static_cast<char*>(buffer);

    int duplicateCount;

    for (int i = 0; i < static_cast<int>(desc.getLength()); ++i)
    {
        memcpy(element, ptr, elementSize);

        if (i == 0 || i == static_cast<int>(desc.getLength() - 1))
        {
            duplicateCount = 2;
        }
        else
        {
            duplicateCount = 4;
        }

        for (int j = 0; j < duplicateCount; ++j)
        {
            memcpy(ptrOut, element, elementSize);
            ptrOut += elementSize;
        }

        ptr += elementSize;
    }

	free(element);
    ok = true;
    return Array(BufferDesc(arrayLengthOut, desc.getComponentCount(), desc.getDataType()), buffer);
}

void PolylineMesh::updateData()
{
    std::vector<glm::vec3> pointsTmp;
    std::vector<glm::vec3> vertices;
    std::vector<glm::vec2> direction;
    std::vector<float> attrib0;
    std::vector<unsigned int> indices;
	std::vector<bool> maskTmp;

    if
        (nullptr == this->_shape)
    {
        LOG(SEAScope::LogLevel::Error, "PolylineMesh",
                           "Polyline mesh not updated because it has no shape");
        return;
    }

    CreateBufferForSphere(pointsTmp, attrib0, maskTmp);

	if
		(0 < maskTmp.size())
	{
		maskedPolylineOnSphereIndices2(pointsTmp, vertices, direction, indices, maskTmp);
	}
	else
	{
		polylineOnSphereIndices2(pointsTmp, vertices, direction, indices);
	}

    if (false == indices.empty())
    {

        if (false == attrib0.empty())
        {
            Array att(BufferDesc(static_cast<unsigned int>(attrib0.size()), 1, DataType::float32), attrib0.data());
            bool duplicateOk {false};
            Array attrib1 = this->duplicate(att, duplicateOk);
            if
                (false == duplicateOk)
            {
                return;
            }
            this->_attribBuffer->allocate(attrib1);
            free(attrib1.getDataPtr());
        }

        if (this->_vertexArrObj == nullptr)
        {
            this->_verticeBuffer->allocate(static_cast<unsigned int>(sizeof(glm::vec3) * vertices.size()),
                                          reinterpret_cast<void*>(&vertices[0].x));
            this->_directionBuffer->allocate(static_cast<unsigned int>(sizeof(glm::vec2) * direction.size()),
                                            reinterpret_cast<void*>(&direction[0].x));
            this->_indicesBuffer->allocate(static_cast<unsigned int>(sizeof(unsigned int) * indices.size()),
                                          reinterpret_cast<void*>(&indices[0]));

            const int positionLoc = 0;
            const int previousLoc = 1;
            const int nextLoc = 2;
            const int directionLoc = 3;

            OGL::VertexAttribArray* verticesAttrib0 = _ctx->createVertexAttribArray(positionLoc, 3, this->_verticeBuffer, DataType::float32, 0, sizeof(glm::vec3) * 4);
            OGL::VertexAttribArray* verticesAttrib1 = _ctx->createVertexAttribArray(previousLoc, 3, this->_verticeBuffer, DataType::float32, 0, 0);
            OGL::VertexAttribArray* verticesAttrib2 = _ctx->createVertexAttribArray(nextLoc, 3, this->_verticeBuffer, DataType::float32, 0, sizeof(glm::vec3) * 8);
            OGL::VertexAttribArray* directionAttrib = _ctx->createVertexAttribArray(directionLoc, 2, this->_directionBuffer);

            OGL::VertexAttribArray* dataAttrib = nullptr;

            if (false == attrib0.empty())
            {
                const int dataLoc = 4;
                dataAttrib = this->_ctx->createVertexAttribArray(dataLoc, 1, this->_attribBuffer, DataType::float32);
            }

            // vertexArrObj is null, allocate it
            this->_vertexArrObj = this->_ctx->createVertexArrayObject();

            this->_vertexArrObj->addObject(verticesAttrib0);
            this->_vertexArrObj->addObject(verticesAttrib1);
            this->_vertexArrObj->addObject(verticesAttrib2);
            this->_vertexArrObj->addObject(directionAttrib);

            if (false == attrib0.empty())
            {
                this->_vertexArrObj->addObject(dataAttrib);
            }

            this->_vertexArrObj->addObject(this->_indicesBuffer);

            this->_vertexArrObj->create(this->_ctx);
            this->_vertexArrObj->bind();

            verticesAttrib0->bind();
            verticesAttrib1->bind();
            verticesAttrib2->bind();
            directionAttrib->bind();

            if (false == attrib0.empty() && nullptr != dataAttrib)
            {
                dataAttrib->bind();
            }

            this->_indicesBuffer->bind();

            this->_vertexArrObj->unbind();
        }
    }

    this->_verticeCount = static_cast<unsigned int>(indices.size());

    if (0 < this->_verticeCount)
    {
        if
            (nullptr == this->_drawMode)
        {
		    this->_drawMode = this->_ctx->createDrawMode();
        }
		this->_drawMode->setDataType(DataType::unsigned_int);
		this->_drawMode->setPrimitive(OGL::triangles);
		this->_drawMode->setDrawType(OGL::draw_elements);
		this->_drawMode->setElementCount(this->_verticeCount);
    }

    // It is necessary to keep points in memory in case the polyline mesh is
    // used as a pattern for extraction, so do not clear this->_points
	//this->_points.clear();
}

void PolylineMesh::draw()
{
    if (this->_verticeCount < 2)
    {
        return;
    }

    this->_vertexArrObj->bind();

    this->_drawMode->draw();

    this->_vertexArrObj->unbind();
}

void PolylineMesh::setProgram(OGL::Program* program)
{
    if
        (nullptr != this->_program)
    {
        delete this->_program;
        this->_program = nullptr;
    }
    this->_program = program;

    this->_program->bind();
    this->_program->setUniformByName("color", glm::vec3(1.f, 0.f,0.f));

    this->_program->setUniformByName("aspect", 1.f);
    this->_program->setUniformByName("thickness", 1.f);
    this->_program->setUniformByName("miter", static_cast<int>(1));

    this->_program->unbind();
}

void PolylineMesh::setThickness(float value)
{
    this->_program->bind();
    this->_program->setUniformByName("thickness", value);
    this->_program->unbind();
}

void PolylineMesh::CreateBufferForSphere(std::vector<glm::vec3>& pointsXYZ, std::vector<float>& attrib0, std::vector<bool>& mask)
{
    double doubleTmp(0.0);  //

    pointsXYZ.clear();
    pointsXYZ.reserve(this->_points.size());
    attrib0.clear();
    attrib0.reserve(this->_points.size());
    mask.clear();
    mask.reserve(this->_points.size());

    double precision = 5.0 * M_PI / 180.0; // degrees
    std::vector<Segment0> segments;

    for (std::size_t i = 0; i < this->_points.size()-1; ++i)
    {
        double lat = this->_points[i+1].Latitude() - this->_points[i].Latitude();
        double lon = this->_points[i+1].Longitude() - this->_points[i].Longitude();

        Segment0 seg;
        seg.subdiv = std::abs(lat) > precision || std::abs(lon) > precision;

        if (seg.subdiv)
        {
            doubleTmp = std::max(std::abs(lat) / precision, std::abs(lon) / precision);
            seg.subdiv_count = static_cast<int>(doubleTmp);
        }
        else
        {
            seg.subdiv_count = 1;
        }

        seg.id = static_cast<int>(i);
        segments.push_back(seg);
    }

	glm::quat q0, q1, q2;
	glm::vec3 p0, p1, p2;

	p0 = UnitSphere::ToVector3D(this->_points[0]);

    // Add first point
	pointsXYZ.push_back(p0);

	std::vector<float> toto;
	std::vector<float>& attribute0 = toto;
	if (false == this->_attributeMap.empty())
	{
		attribute0 = this->_attributeMap.begin()->second;
		attrib0.push_back(attribute0[0]);
	}

	if
		(0 < this->_mask.size())
	{
		mask.push_back(this->_mask[0]);
	}

    // Add intermediary points
	float a(0.f), step(0.f);
	float attMix(0.f);
	bool maskMix(false);
	for (const auto& seg : segments)
	{
        std::size_t segId = static_cast<std::size_t>(seg.id);
        float subdivCount = static_cast<float>(seg.subdiv_count);
		p0 = UnitSphere::ToVector3D(this->_points[segId]);
		p1 = UnitSphere::ToVector3D(this->_points[segId + 1]);
		q1 = rotationBetweenVectors(p0, p1);

		step = 1.f / subdivCount;

		for (int j = 1; j < seg.subdiv_count; ++j)
		{
			a = step * static_cast<float>(j);

			q2 = glm::slerp(q0, q1, a);

			p2 = q2 * p0;

			pointsXYZ.push_back(p2);

			if (false == this->_attributeMap.empty())
			{
				attMix = glm::mix(attribute0[segId],
                                  attribute0[segId + 1],
                                  a);
				attrib0.push_back(attMix);
			}

			if
				(0 < this->_mask.size())
			{
				maskMix = this->_mask[segId] && this->_mask[segId + 1];
				mask.push_back(maskMix);
			}
		}

        // Add last point for the segment
		pointsXYZ.push_back(p1);

		if (false == this->_attributeMap.empty())
		{
			attrib0.push_back(attribute0[segId + 1]);
		}

		if
			(0 < this->_mask.size())
		{
			mask.push_back(this->_mask[segId + 1]);
		}
	}
}

void PolylineMesh::draw(OGL::Context* /*ctx*/, const glm::mat4& /*mvp*/)
{
    this->draw();
}

void PolylineMesh::drawPicking(OGL::Context* /*ctx*/, const glm::mat4& /*mvp*/)
{
    this->draw();
}
