// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/program_bind_helper.h"

#include "3d/program.h"
#include "3d/context.h"


#define TEXTURE_KEYWORD     "texture"
#define TEXTURE_MAX         8
#define ATTRIB_POSITION     "position"
#define ATTRIB_TEXCOORD     "texCoord"
#define UNIFORM_MATRIX_MVP  "mvp"
//#define TEXTURE_RATIO       "textureRatio"

// Tuple define
#define TEXTURE_UNIT    0
#define TEXTURE_OBJECT  1

namespace OGL
{

ProgramBindHelper::ProgramBindHelper()
:   _program(nullptr)
{
    this->hasPosition = false;
    this->position_location = -1;
    this->hasTexCoord = false;
    this->texCoord_location = -1;
    this->hasMVP = false;
    this->matrix_mvp_location = -1;
}

ProgramBindHelper::~ProgramBindHelper()
{
    delete _program;
    _program = nullptr;
}

void ProgramBindHelper::setProgram(OGL::Program* program)
{
    if
        (nullptr != _program)
    {
        delete _program;
        _program = nullptr;
    }
    _program = program;
    
    parse();
}

void ProgramBindHelper::parse()
{
    char buffer[50];
    
    int result;
        
    _program->bind();
    
    for (int i = 0; i < TEXTURE_MAX; ++i)
    {
        // Generate uniform texture name : textureX
        snprintf(buffer, 50, "%s%d", TEXTURE_KEYWORD, i);
        
        // Ask OpenGL the location of textureX uniform
        result = _program->getUniformLocation(buffer);
        
        // If textureX doesn't exist, location equal -1
        if (result != -1)
        {
            // Set the texture unit equal to the X of textureX
            _program->uniform1i(result, i);
            
            // Save the id and the location
            _textures.push_back(std::make_tuple(i, nullptr));
        }
    }
    
    result = _program->getAttribLocation(ATTRIB_POSITION);
    if (-1 != result)
    {
        this->hasPosition = true;
        this->position_location = result;
        _program->bindAttribLocation(0, ATTRIB_POSITION);
    }
    
    result = _program->getAttribLocation(ATTRIB_TEXCOORD);
    if (-1 != result)
    {
        this->hasTexCoord = true;
        this->texCoord_location = result;
        _program->bindAttribLocation(1, ATTRIB_TEXCOORD);
    }
    
    result = _program->getUniformLocation(UNIFORM_MATRIX_MVP);
    if (-1 != result)
    {
        this->hasMVP = true;
        this->matrix_mvp_location = result;
    }
    
    parseSpecialized();
    
    _program->unbind();
}

void ProgramBindHelper::setTexture(int textureUnit, OGL::Texture* texturePtr)
{
    // Loop on all textures
    for (size_t i = 0; i < _textures.size(); ++i)
    {
        // Search the texture bind to the texture unit with the right id.
        if (textureUnit == std::get<TEXTURE_UNIT>(_textures[i]))
        {
            // Update texture object.
            std::get<TEXTURE_OBJECT>(_textures[i]) = texturePtr;
            break;
        }
    }
}

void ProgramBindHelper::bind(OGL::Context* ctx)
{
    OGL::Texture* texture = nullptr;
    
    _program->bind();
    
    // Bind texture object(s)
    for (auto& tuple : _textures)
    {
        texture = std::get<TEXTURE_OBJECT>(tuple);
        
        if (texture)
        {
            ctx->activeTextureUnit(static_cast<unsigned int>(std::get<TEXTURE_UNIT>(tuple)), texture);
        }
    }
    
    bindSpecialized(ctx);
}

void ProgramBindHelper::unbind()
{
    _program->unbind();
}

int ProgramBindHelper::getVerticesLocation()
{
    return this->position_location;
}
    
bool ProgramBindHelper::hasTextureCoordinates()
{
    return this->hasTexCoord;
}

int ProgramBindHelper::getTextureCoordinatesLocation()
{
    return this->texCoord_location;
}

bool ProgramBindHelper::hasModelViewProjectionMatrix()
{
    return this->hasMVP;
}

int ProgramBindHelper::getModelViewProjectionMatrixLocation()
{
    return this->matrix_mvp_location;
}
    
} // namespace OGL
