// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/rectangle_creator.h"

#include "3d/shape.h"
#include "3d/rectangle_based_builder.h"

#include "3d/core_objects.h"
//#include "3d/polyline_mesh.h"
#include "3d/geodetic.h"
#include "3d/ellipsoid.h"
#include "3d/geodetic.h"
#include "3d/spherical_math.h"
#include "3d/math_func.h"
#include "3d/geometry.h"

#include "3d/shapes_allocator.h"
#include "3d/rectangle.h"
#include "3d/unit_sphere.h"
#include "3d/ellipse_mesh.h"
#include "3d/ellipse.h"

RectangleCreator::RectangleCreator()
: _builder{nullptr},
  _program{nullptr},
  _buffer{nullptr},
  _vao{nullptr},
  _drawMode{nullptr},
  _lastPos{ 0, 0, 0 },
  _currentPos{ 0, 0, 0 },
  _tmpShape{nullptr},
  _p0Loc{-1},
  _p1Loc{-1},
  _mvpLoc{-1},
  _colorLoc{-1},
  _initialized{false},
  _alpha{1.f},
  _beta{1.f},
  /*_alphaLocked(false),
  _betaLocked(false)*/
  _ellipseMesh{nullptr}
{
}

RectangleCreator::~RectangleCreator()
{
	delete _buffer;
	_buffer = nullptr;
	delete _program;
	_program = nullptr;
	delete _vao;
	_vao = nullptr;
	delete _drawMode;
	_drawMode = nullptr;
	delete _builder;
	_builder = nullptr;

	//ShapesAllocator::getInstance()->release(_tmpShape);
	_tmpShape = nullptr;
}

void RectangleCreator::initialize(OGL::Context* ctx)
{
	_program = OGL::Helper::createProgram(ctx, vertex_line_tesselate, fragment_uni_color);

	_program->bind();
	_p0Loc = _program->getUniformLocation("p0");
	_p1Loc = _program->getUniformLocation("p1");
	_mvpLoc = _program->getUniformLocation("mvp");
	_colorLoc = _program->getUniformLocation("color");
	_program->unbind();

	const int segmentCount = 100;
	std::vector<float> percents;
	percents.reserve(segmentCount);

	for (int i = 0; i < segmentCount; ++i)
	{
		percents.push_back(static_cast<float>(i) / (segmentCount - 1));
	}

	_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(percents));
	OGL::VertexAttribArray* vaa = ctx->createVertexAttribArray(0, 1, _buffer, DataType::float32);

	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(vaa);
	_vao->fill();

	_drawMode = OGL::Helper::createDrawModeArrays(ctx, OGL::line_strip, segmentCount);

	//_tmpShape = new Rectangle;

	_initialized = false;
}

void RectangleCreator::setBuilder(ShapeBuilder* builder)
{
	_builder = dynamic_cast<RectangleBasedBuilder*>(builder);
}

void RectangleCreator::draw(OGL::Context* ctx, const glm::mat4& mvp)
{
	//// Draw bounding box
	//auto points = _corners;

	//_program->bind();

	////glm::quat q = rotationBetweenVectors(glm::vec3(0.f, 0.f, 1.f), glm::vec3(1.f, 0.f, 0.f));

	////glm::mat4 model = glm::mat4(q);

	//_program->setUniform(_mvpLoc, mvp * _model);
	//_program->setUniform(_colorLoc, green4f);

	//_program->setUniformByName("angles", glm::vec2(_alpha, _beta));

	//_vao->bind();

	//for (size_t i = 0; i < 1; ++i)
	////for (size_t i = 0; i < points.size(); ++i)
	//{
	//	_program->setUniform(_p0Loc, points[i % points.size()]);
	//	_program->setUniform(_p1Loc, points[(i + 1ull) % points.size()]);

	//	_drawMode->draw();
	//}

	//_vao->unbind();

    if
        (nullptr != this->_ellipseMesh)
    {
	    this->_ellipseMesh->draw(ctx, mvp);
    }
}

OGL::Shape* RectangleCreator::createShape()
{
	OGL::Shape* shape = nullptr;

	if (_builder)
	{
		//_builder->setCorners();
		shape = _builder->build();
	}

	return shape;
}

void RectangleCreator::reset()
{
}

OGL::Shape* RectangleCreator::getTemporaryShape()
{
	return nullptr;
}

SEAScope::ShapeType RectangleCreator::getShapeType()
{
	return _builder->getShapeType();
}

float RectangleCreator::getAlpha() const
{
    return this->_alpha;
}

void RectangleCreator::setAlpha(float alpha)
{
    this->_alpha = alpha;
}

float RectangleCreator::getBeta() const
{
    return this->_beta;
}

void RectangleCreator::setBeta(float beta)
{
    this->_beta = beta;
}

void RectangleCreator::addVertexUnderCursor()
{
}

void RectangleCreator::updateCursorPosition(const glm::vec3&, std::uint32_t, std::uint32_t)
{
}

//bool RectangleCreator::handleMouseButtonEvent(int button, int state, int /*mods*/, const glm::vec3& geo)
//{
//	const glm::vec3 zAxis(0.f, 0.f, 1.f);
//
//	if (GLFW_MOUSE_BUTTON_LEFT == button && GLFW_PRESS == state)
//	{
//		// First state, first click to define starting point.
//		if (false == _initialized)
//		{
//			_corners[0] = geo;
//			_corners[1] = geo;
//			_corners[2] = geo;
//			_corners[3] = geo;
//
//			glm::vec2 lonlat = UnitSphere::ToGeodetic2D(geo).toVec2();
//			_cornersLonLat[0] = lonlat;
//			_cornersLonLat[1] = lonlat;
//			_cornersLonLat[2] = lonlat;
//			_cornersLonLat[3] = lonlat;
//
//			_middles[0] = geo;
//
//			_initialized = true;
//		}
//		else if (false == _betaLocked)
//		{
//			_middles[2] = geo;
//			_beta = static_cast<float>(AngleBetween(_middles[0], _middles[2]) / 2.0);
//			_alpha = _beta / 2.f;
//			_betaLocked = true;
//
//			glm::vec3 center = slerp(_middles[0], _middles[2], 0.5f);
//
//			glm::quat q = rotationBetweenVectors(zAxis, center);
//			_model = glm::mat4(q);
//
//
//			//glm::vec3 localZAxis = center;
//			//glm::vec3 localXAxis = glm::cross(localZAxis, zAxis);
//			//glm::vec3 localYAxis = glm::cross(localZAxis, localXAxis);
//
//			//printf("X:%f,%f,%f Y:%f,%f,%f Z:%f,%f,%f\n", localXAxis.x, localXAxis.y, localXAxis.z, localYAxis.x, localYAxis.y, localYAxis.z, localZAxis.x, localZAxis.y, localZAxis.z);
//		}
//		else if (false == _alphaLocked)
//		{
//
//		}
//	}
//
//	return false;
//}

//bool RectangleCreator::handleMouseMoveEvent(int /*x*/, int /*y*/, const glm::vec3& /*geo*/)
//{
    /*
	const glm::vec3 zAxis(0.f, 0.f, 1.f);

	glm::vec3 localZAxis = geo;
	glm::vec3 localXAxis = glm::cross(localZAxis, zAxis);
	glm::vec3 localYAxis = glm::cross(localZAxis, localXAxis);

	printf("X:%f,%f,%f Y:%f,%f,%f Z:%f,%f,%f\n", localXAxis.x, localXAxis.y, localXAxis.z, localYAxis.x, localYAxis.y, localYAxis.z, localZAxis.x, localZAxis.y, localZAxis.z);
    */

	//_ellipseMesh->setCenter();

	//if (true == _initialized)
	//{
	//	_currentPos = geo;
	//	_currentLonLat = UnitSphere::ToGeodetic2D(geo).toVec2();

	//	_cornersLonLat[2] = _currentLonLat;
	//	_cornersLonLat[1] = glm::vec2(_cornersLonLat[2].x, _cornersLonLat[0].y);
	//	_cornersLonLat[3] = glm::vec2(_cornersLonLat[0].x, _cornersLonLat[2].y);

	//	_corners[2] = _currentPos;

	//	_corners[1] = UnitSphere::ToVector3D(_cornersLonLat[1]);
	//	_corners[3] = UnitSphere::ToVector3D(_cornersLonLat[3]);

	//	float angle = AngleBetween(_corners[0], _currentPos);
	//	_beta = angle / 2.f;

	//	//glm::quat q = rotationBetweenVectors(glm::vec3(0.f, 0.f, 1.f), glm::vec3(1.f, 0.f, 0.f));

	//	printf("%f, pos: %f %f %f\n", angle, _currentPos.x, _currentPos.y, _currentPos.z);
	//}

//	return false;
//}

/**
 *
 */
unsigned int RectangleCreator::getAnchorCount() const
{
    return 0;
}

/**
 *
 */
glm::vec3 RectangleCreator::getAnchorPos(AnchorType, unsigned int) const
{
    return {0, 0, 0};
}

/**
 *
 */
void RectangleCreator::move(AnchorType, unsigned int, const glm::vec3&)
{
}

/**
 *
 */
void RectangleCreator::move(AnchorType, unsigned int)
{
}

/**
 *
 */
void RectangleCreator::pressed(AnchorType, unsigned int)
{
}

/**
 *
 */
void RectangleCreator::remove(AnchorType, unsigned int)
{
}
