// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/scene_graph.h"
#include "3d/pickable.h"
#include "3d/scene_graph_node.h"
#include "3d/3d_irenderer.h"
#include "3d/ipicking_renderer.h"
#include "3d/iraw_renderer.h"
#include "3d/renderer_factory.h"
#include "3d/context.h"
#include "3d/iworld_visibility.h"
#include "3d/globe_renderer_factory.h"

#include "core/log.h"

SceneGraph::SceneGraph()
:   _renderers(renderer_count),
    _pickingRenderers(picking_renderer_count),
    _rawRenderers(raw_renderer_count)
{
    for (unsigned int i = 0; i < picking_renderer_count; ++i)
    {
        _pickingRenderers[i] = nullptr;
    }
    
    for (unsigned int i = 0; i < raw_renderer_count; ++i)
    {
        _rawRenderers[i] = nullptr;
    }
}

SceneGraph::~SceneGraph()
{
    // Clear all renderers
    for (unsigned int i = 0; i < _renderers.size(); ++i)
    {       
        for (unsigned int j = 0; j < _renderers[i].size(); ++j)
        {
            delete _renderers[i][j];
            _renderers[i][j] = nullptr;
        }
    }
    
    for (unsigned int i = 0; i < _pickingRenderers.size(); ++i)
    {
        delete _pickingRenderers[i];
        _pickingRenderers[i] = nullptr;
    }
    
    for (unsigned int i = 0; i < _rawRenderers.size(); ++i)
    {
        delete _rawRenderers[i];
        _rawRenderers[i] = nullptr;
    }
}

void SceneGraph::create(OGL::Context* ctx, IWorldVisibility* visibility/*, RendererFactory* rendererFactory*/)
{
    GlobeRendererFactory rendererFactory;
    rendererFactory.initialize(visibility);
    
    // Allocate all renderers
    for (unsigned int i = 0; i < renderer_count; ++i)
    {
        _renderers[i].resize(fill_count);
        
        for (unsigned int j = 0; j < fill_count; ++j)
        {
            _renderers[i][j] = rendererFactory.createRenderer(ctx, static_cast<RendererType>(i), static_cast<FillType>(j));
        }
    }
    
    // Allocate all picking renderers
    for (unsigned int i = 0; i < picking_renderer_count; ++i)
    {
        _pickingRenderers[i] = rendererFactory.createPickingRenderer(ctx, static_cast<PickingRendererType>(i));
    }
    
    // Allocate all raw renderers
    for (unsigned int i = 0; i < raw_renderer_count; ++i)
    {
        _rawRenderers[i] = rendererFactory.createRawRenderer(ctx, static_cast<RawRendererType>(i));
    }
    
    _rendererToPicking.resize(renderer_count);
    
    // Indirection array between color renderers and picking renderers
    _rendererToPicking[renderer_arrow] = picking_renderer_null;
    _rendererToPicking[renderer_barb] = picking_renderer_null;
    _rendererToPicking[renderer_dot_cloud] = picking_renderer_null;
    _rendererToPicking[renderer_polygon] = picking_renderer_polygon;
    _rendererToPicking[renderer_polyline] = picking_renderer_polyline;
    _rendererToPicking[renderer_raster] = picking_renderer_raster;
    _rendererToPicking[renderer_trajectory] = picking_renderer_trajectory;
    _rendererToPicking[renderer_globe] = picking_renderer_globe;
    _rendererToPicking[renderer_rawrgb] = picking_renderer_raster;
    _rendererToPicking[renderer_quadscreen] = picking_renderer_screen_quad;
    _rendererToPicking[renderer_quadscreen_textured] = picking_renderer_screen_quad;
        
    _rendererToRaw.resize(renderer_count);
    
    // Indirection array between color renderers and raw renderers
    _rendererToRaw[renderer_arrow] = raw_renderer_null;
    _rendererToRaw[renderer_barb] = raw_renderer_null;
    _rendererToRaw[renderer_dot_cloud] = raw_renderer_null;
    _rendererToRaw[renderer_polygon] = raw_renderer_null;
    _rendererToRaw[renderer_polyline] = raw_renderer_null;
    _rendererToRaw[renderer_raster] = raw_renderer_raster;
    _rendererToRaw[renderer_trajectory] = raw_renderer_trajectory;
    _rendererToRaw[renderer_globe] = raw_renderer_globe;
    _rendererToRaw[renderer_rawrgb] = raw_renderer_rgb;
    _rendererToRaw[renderer_quadscreen] = raw_renderer_null;
    _rendererToRaw[renderer_quadscreen_textured] = raw_renderer_null;
}

void SceneGraph::addObject(Renderable2* /*object*/)
{
    LOG(SEAScope::LogLevel::Warn, "SceneGraph",
                       "addObject deprecated");
}

void SceneGraph::removeObject(Renderable2* /*object*/)
{
    LOG(SEAScope::LogLevel::Warn, "SceneGraph",
                       "removeObject deprecated");
}

Pickable* SceneGraph::pickable(unsigned int id)
{
    Pickable* obj = nullptr;
    
    auto it = _objects.begin();
    
    for (; it != _objects.end(); ++it)
    {
        if ((*it)->flag(RenderableFlags::pickable))
        {
            Pickable* p = static_cast<Pickable*>(*it);
            
            if (p->pickingId() == id)
            {
                obj = p;
                break;
            }
        }
    }
    
    return obj;
}

void SceneGraph::draw(OGL::Context* ctx, const glm::mat4& mvp)
{   
    // Don't modify the z-buffer, draw by layer.
    ctx->setDepthMask(false);
    
    _root.draw(ctx, mvp);
}

void SceneGraph::drawOne(OGL::Context* ctx, const SEAScope::RenderableId& rId, const glm::mat4& mvp)
{
    // Don't modify the z-buffer
    ctx->setDepthMask(false);
    
    Node* node = _nodes.get(rId);
    
    if (nullptr != node)
    {
        node->draw(ctx, mvp);
    }
    else
    {
        LOG(SEAScope::LogLevel::Error, "SceneGraph", "RenderableId not found, nothing to draw");
    }
}

void SceneGraph::drawOneRaw(OGL::Context* ctx, const SEAScope::RenderableId& rId, const glm::mat4& mvp)
{
    // Don't modify the z-buffer
    ctx->setDepthMask(false);
    
    Node* node = _nodes.get(rId);
    
    if (nullptr != node)
    {
        node->drawRaw(ctx, mvp);
    }
    else
    {
        LOG(SEAScope::LogLevel::Error, "SceneGraph", "RenderableId not found, nothing to draw");
    }
}

void SceneGraph::drawPicking(OGL::Context* ctx, const glm::mat4& mvp)
{
    // Don't modify the z-buffer, draw by layer.
    ctx->setDepthMask(false);
    
    _root.drawPicking(ctx, mvp);
}

void SceneGraph::drawRaw(OGL::Context* ctx, const glm::mat4& mvp)
{
    // Don't modify the z-buffer, draw by layer.
    ctx->setDepthMask(false);
    
    _root.drawRaw(ctx, mvp);
}

void SceneGraph::setOnTop(Renderable2* obj)
{
    auto it = std::find(_objects.begin(), _objects.end(), obj);
    
    if (it != _objects.end())
    {
		_objects.splice(_objects.end(), _objects, it);
    }
}

void SceneGraph::updateRenderer(Node* node)
{
    node->setRenderer(_renderers[node->rendererType()][node->fillType()]);
    node->setPickingRenderer(_pickingRenderers[_rendererToPicking[node->rendererType()]]);
    node->setRawRenderer(_rawRenderers[_rendererToRaw[node->rendererType()]]);
}

void SceneGraph::addTo(Node* parent, Node* node, bool update)
{
    if (nullptr == parent)
    {
        parent = &_root;
        if
            (true == update)
        {
            updateRenderer(node);
        }
    }
    
    parent->addChild(node);
}

void SceneGraph::addTo(Node* parent, Node* node, const SEAScope::RenderableId& rId)
{
    if (nullptr == parent)
    {
        parent = &_root;
        updateRenderer(node);
    }
    
    parent->addChild(node);
    
    _nodes.add(rId, node);
}

void SceneGraph::addTo(Node* parent, Node* node, const SEAScope::RenderableId& rId, bool ownRenderer)
{
    if (nullptr == parent)
    {
        parent = &_root;
        updateRenderer(node);
    }
    
    if (ownRenderer)
    {
        updateRenderer(node);
    }
    
    parent->addChild(node);
    
    _nodes.add(rId, node);
}

bool SceneGraph::remove(const SEAScope::RenderableId& rId, Node*& node)
{
    bool found = false;
    
    if (_nodes.exist(rId))
    {
        node = nullptr;
        
        _nodes.remove(rId, node);

        if (nullptr != node)
        {
            found = true;

            if (node->parent())
            {
                node->parent()->removeChild(node);   
            }
        }
    }
    else
    {
        LOG(SEAScope::LogLevel::Warn, "SceneGraph", "Node not found");
    }
    
    return found;
}

void SceneGraph::setVisible(const SEAScope::RenderableId& rId, bool status)
{
    if (_nodes.exist(rId))
    {
        Node* node = _nodes.get(rId);

        if (nullptr != node)
        {
            node->setVisible(status);
        }
    }
    else
    {
        LOG(SEAScope::LogLevel::Warn, "SceneGraph", "Node not found");
    }
}

Node* SceneGraph::getNode(const SEAScope::RenderableId& rId)
{
    Node* node = nullptr;
    
    if (_nodes.exist(rId))
    {
        node = _nodes.get(rId);
    }
    
    return node;
}

void SceneGraph::zChanged(Node* node)
{
    Node* parent = node->parent();
    
    if (nullptr != parent)
    {
        // Sub-nodes are sorted by z-index when they are added to their parent
        // so removing the node and inserting it back will place it at the
        // right location
        if (parent->removeChild(node))
        {
            parent->addChild(node);
        }
    }
}

Node* SceneGraph::getRoot()
{
    return &this->_root;
}

void SceneGraph::getNodes(RendererType type, std::vector<Node*>& array)
{
    getNodesRecurse(type, array, &_root);
}

void SceneGraph::getNodesRecurse(RendererType type, std::vector<Node*>& array, Node* node)
{
    if
        (nullptr != node)
    {
        std::list<Node*>* children = node->children();

        if
            (nullptr != children)
        {
            for (auto it : *children)
            {
                getNodesRecurse(type, array, it);
            }
        }

        if (type == node->rendererType())
        {
            array.push_back(node);
        }
    }
}
