// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/data_rendering_param.h"
#include "3d/scene_graph_node.h"

#include "3d/3d_irenderer.h"
#include "3d/ipicking_renderer.h"
#include "3d/iraw_renderer.h"

#include "3d/renderable.h"

Node::Node()
:   _parent(nullptr),
    _children(nullptr),
    _rCfg(nullptr),
    _renderer(nullptr),
    _rendererPicking(nullptr),
    _rendererRaw(nullptr),
    _visible(true),
    _topChild(nullptr)
{
}

Node::~Node()
{
    if
        (nullptr != this->_children)
    {
        for
            (auto child : *this->_children)
        {
            delete child;
            child = nullptr;
        }

        delete _children;
	    _children = nullptr;
    }

    delete this->_topChild;
    this->_topChild = nullptr;

    delete _rCfg;
	_rCfg = nullptr;

    _parent = nullptr;
    _renderer = nullptr;
    _rendererPicking = nullptr;
    _rendererRaw = nullptr;
}

void Node::addChild(Node* node)
{
    if
        (this == node)
    {
        return;
    }

    if (nullptr == _children)
    {
        _children = new std::list<Node*>();
    }
    
    node->_parent = this;
    bool inserted = false;

    if (    nullptr != node->_rCfg
        &&  nullptr != node->_rCfg->renderingParam
        &&  false == _children->empty())
    {
        float& nodeZ = node->_rCfg->renderingParam->zIndex;
        
        DataRenderingParam* param = nullptr;

        for (auto it = _children->begin(); it != _children->end(); ++it)
        {
            if
                (nullptr != (*it)->_rCfg)
            {
                param = (*it)->_rCfg->renderingParam;

                if (nullptr != param
                    && param->zIndex > nodeZ)
                {
                    _children->insert(it, node);
                    inserted = true;
                    break;
                }
            }
        }
    }

    if (false == inserted)
    {
        _children->push_back(node);
    }

    if
        (nullptr != _topChild
      && false == _topChild->visible())
    {
        _topChild->setVisible(true);
    }
}

void Node::setFillType(FillType type)
{
    if (nullptr == _rCfg)
    {
        _rCfg = new RenderingParameters;
    }
    
    _rCfg->fillType = type;
}

void Node::setRendererType(RendererType type)
{
    if (nullptr == _rCfg)
    {
        _rCfg = new RenderingParameters;
    }
    
    _rCfg->rendererType = type;
}

void Node::setRenderingParam(DataRenderingParam* param)
{
    if (nullptr == _rCfg)
    {
        _rCfg = new RenderingParameters;
    }
    
    if
        (nullptr != _rCfg->renderingParam)
    {
        delete _rCfg->renderingParam;
        _rCfg->renderingParam = nullptr;
    }
    _rCfg->renderingParam = param;
}

void Node::setParent(Node* node)
{
    _parent = node;
}

DataRenderingParam* Node::dataRenderingParam() const
{
    if (nullptr == _rCfg)
    {
        return nullptr;
    }
    
    return _rCfg->renderingParam;
}

Node* Node::parent() const
{
    return _parent;
}

void Node::draw(OGL::Context* ctx, const glm::mat4& mvp)
{
    if (nullptr != _children && true == _visible)
    {
        if (_renderer && _rCfg)
        {
            _renderer->bind(ctx, mvp, _rCfg->renderingParam);
        }
        
        auto it = _children->cbegin();
        
        for (; it != _children->cend(); ++it)
        {
            (*it)->draw(ctx, mvp);
        }

        if
            (nullptr != _topChild)
        {
            _topChild->draw(ctx, mvp);
        }
    }
}

void Node::drawPicking(OGL::Context* ctx, const glm::mat4& mvp)
{
    if (nullptr != _children && true == _visible)
    {
        if (_rendererPicking && _rCfg)
        {
            _rendererPicking->bind(ctx, mvp, _rCfg->renderingParam);
        }
        
        auto it = _children->cbegin();
        
        for (; it != _children->cend(); ++it)
        {
            (*it)->drawPicking(ctx, mvp);
        }
    }
}

void Node::drawRaw(OGL::Context* ctx, const glm::mat4& mvp)
{
    if (nullptr != _children && true == _visible)
    {
        if (_rendererRaw && _rCfg)
        {
            _rendererRaw->bind(ctx, mvp, _rCfg->renderingParam);
        }
        
        auto it = _children->cbegin();
        
        for (; it != _children->cend(); ++it)
        {
            (*it)->drawRaw(ctx, mvp);
        }
    }
}

void Node::setRenderer(IRenderer2* renderer)
{
    _renderer = renderer;
}

void Node::setPickingRenderer(IPickingRenderer* renderer)
{
    _rendererPicking = renderer;
}

void Node::setRawRenderer(IRawRenderer* renderer)
{
    _rendererRaw = renderer;
}

IRenderer2* Node::renderer() const
{
    return _renderer;
}

IPickingRenderer* Node::pickingRenderer() const
{
    return _rendererPicking;
}

IRawRenderer* Node::rawRenderer() const
{
    return _rendererRaw;
}

FillType Node::fillType() const
{
    if
        (nullptr != _rCfg)
    {
        return _rCfg->fillType;
    }
    
    return FillType::fill_null;
}

RendererType Node::rendererType() const
{
    if
        (nullptr != _rCfg)
    {
        return _rCfg->rendererType;
    }

    return RendererType::renderer_null;
}

bool Node::removeChild(Node* node)
{
    if
        (nullptr != _children)
    {
        auto it = std::find(_children->begin(), _children->end(), node);
        if
            (_children->end() != it)
        {
            _children->erase(it);
            node->_parent = nullptr;

            if
                (true == _children->empty())
            {
                if
                    (nullptr != _topChild)
                {
                    _topChild->setVisible(false);
                    //delete _topChild;
                    //_topChild = nullptr;
                }
            }

            return true;
        }
    }

    return false;
}

std::list<Node*>* Node::children() const
{
    return _children;
}

void Node::setVisible(bool status)
{
    _visible = status;
}

bool Node::visible() const
{
    return this->_visible;
}

Node* Node::getTopChild()
{
    return _topChild;
}

void Node::addTopChild(Node* node)
{
    node->_parent = this;
    _topChild = node;
}

////////////////////////////////////////////////////////////////////////////////

GranuleNode::GranuleNode()
: _renderable(nullptr)
{
}

GranuleNode::~GranuleNode()
{
    _renderable = nullptr;
}

void GranuleNode::setRenderable(Renderable2* obj)
{
    _renderable = obj;
}

Renderable2* GranuleNode::renderable() const
{
    return _renderable;
}

void GranuleNode::draw(OGL::Context* ctx, const glm::mat4& mvp)
{
    if (_visible)
    {
        IRenderer2* renderer = nullptr;
        DataRenderingParam* param = nullptr;
        
        if (_rCfg && _rCfg->renderingParam)
        {
            param = _rCfg->renderingParam;
        }
        else
        {
            param = _parent->dataRenderingParam();
        }
    
        if (_renderer && _rCfg)
        {
            renderer = _renderer;
            renderer->bind(ctx, mvp, param);
        }
        else
        {
            renderer = _parent->renderer();
        }

        renderer->draw(ctx, _renderable);
    }
}

void GranuleNode::drawPicking(OGL::Context* ctx, const glm::mat4& mvp)
{
    if (_visible)
    {
        IPickingRenderer* renderer = nullptr;

        DataRenderingParam* param = nullptr;

        if (_rCfg && _rCfg->renderingParam)
        {
            param = _rCfg->renderingParam;
        }
        else
        {
            param = _parent->dataRenderingParam();
        }

        if (_rendererPicking)
        {
            renderer = _rendererPicking;
            renderer->bind(ctx, mvp, param);
        }
        else
        {
            renderer = _parent->pickingRenderer();
        }

        renderer->draw(ctx, _renderable);
    }
}

void GranuleNode::drawRaw(OGL::Context* ctx, const glm::mat4& mvp)
{
    if (_visible)
    {
        IRawRenderer* renderer = nullptr;

        DataRenderingParam* param = nullptr;

        if (_rCfg && _rCfg->renderingParam)
        {
            param = _rCfg->renderingParam;
        }
        else
        {
            param = _parent->dataRenderingParam();
        }

        if (_rendererRaw)
        {
            renderer = _rendererRaw;
            renderer->bind(ctx, mvp, param);
        }
        else
        {
            renderer = _parent->rawRenderer();
            renderer->bind(ctx, mvp, param);
        }

        renderer->draw(ctx, _renderable);
    }
}

