// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>

#include "3d/scene_renderer.h"
#include "3d/buffer.h"
#include "3d/context.h"
#include "3d/framebuffer.h"
#include "3d/helper.h"
#include "core/log.h"

#include "3d/context_factory.h"
#include "3d/scene_graph.h"
#include "3d/scene_graph_controller.h"
#include "3d/iscene.h"
#include "3d/camera.h"
#include "3d/pipeline_parameters.h"
#include "3d/scene_view.h"
#include "3d/ipre_renderer.h"
#include "3d/ipost_renderer.h"

SceneRenderer::SceneRenderer()
: _ctx(nullptr),
  _rasterizationMode(OGL::fill)
{

}

SceneRenderer::~SceneRenderer()
{
    delete this->_ctx;
    this->_ctx = nullptr;
}


/**
 *
 */
static void fillPipelineParameters(OGL::Context* ctx, const Camera* camera)
{
    PipelineParameters* pp = ctx->pipelineParameters();

    pp->model = camera->getModelMatrix();
    pp->view = camera->getViewMatrix();
    pp->projection = camera->getProjectionMatrix();
    pp->inverseView = glm::inverse(pp->view);
    pp->inverseProjection = glm::inverse(pp->projection);
    pp->viewProjection = pp->projection * pp->view * pp->model;
    pp->inverseViewProjection = glm::inverse(pp->viewProjection);
    pp->eyePosition = camera->getWorldPosition();
}


/**
 *
 */
void SceneRenderer::registerPreRenderer(IPreRenderer* renderer)
{
    this->_preRenderers.push_back(renderer);
}


/**
 *
 */
void SceneRenderer::unregisterPreRenderer(IPreRenderer* renderer)
{
    auto found = std::find(this->_preRenderers.begin(), this->_preRenderers.end(), renderer);
    if
        (this->_preRenderers.end() != found)
    {
        this->_preRenderers.erase(found);
    }
}


/**
 *
 */
void SceneRenderer::registerPostRenderer(IPostRenderer* renderer)
{
    this->_postRenderers.push_back(renderer);
}


/**
 *
 */
void SceneRenderer::unregisterPostRenderer(IPostRenderer* renderer)
{
    auto found = std::find(this->_postRenderers.begin(), this->_postRenderers.end(), renderer);
    if
        (this->_postRenderers.end() != found)
    {
        this->_postRenderers.erase(found);
    }
}


/**
 *
 */
bool SceneRenderer::renderSceneView(SceneView* sceneView)
{
    bool preventDraw(false);
    bool pluginPreventsDraw(false);

    this->_ctx->setFlag(OGL::cull_face, true);

    if
        (OGL::fill != this->_rasterizationMode)
    {
        this->_ctx->setPolygonMode(OGL::front_and_back, this->_rasterizationMode);
    }

    for
        (const auto& preRenderer : this->_preRenderers)
    {
        pluginPreventsDraw = preRenderer->preRender(this->_ctx, sceneView);
        preventDraw = (preventDraw || pluginPreventsDraw);
    }

    // Set preventDraw to true in order to prevent nominal rendering
    // (useful in debug mode to render picking colors)
    if
        (false == preventDraw)
    {
        this->_ctx->setFlag(OGL::cull_face, true);

        // Initialize rendering
        glm::vec4 clearColor(0.1f, 0.1f, 0.1f, 0.f);
        this->_ctx->setBlendFunc(OGL::src_alpha, OGL::one_minus_src_alpha);
        this->_ctx->setFlag(OGL::blending, false);
        this->_ctx->setFlag(OGL::depth_test, true);
        this->_ctx->setDepthMask(true);
        this->_ctx->setClearColor(clearColor);
        this->_ctx->clear(OGL::color_buffer | OGL::depth_buffer);

        this->draw(sceneView->getScene(), sceneView->getConstCamera());
    }

    for
        (const auto& postRenderer : this->_postRenderers)
    {
        postRenderer->postRender(this->_ctx, sceneView);
    }

    if
        (OGL::fill != this->_rasterizationMode)
    {
        this->_ctx->setPolygonMode(OGL::front_and_back, OGL::fill);
    }

    return true;
}

void SceneRenderer::draw(IScene* scene, Camera const* camera)
{
    fillPipelineParameters(this->_ctx, camera);

    PipelineParameters* pp = this->_ctx->pipelineParameters();

    scene->getNodeFactory()->getSceneGraph()->draw(this->_ctx, pp->viewProjection);
}

void SceneRenderer::drawOneTrueColors(IScene* scene, const Camera* camera, const SEAScope::RenderableId& rId)
{
    fillPipelineParameters(this->_ctx, camera);

    PipelineParameters* pp = this->_ctx->pipelineParameters();

    scene->getNodeFactory()->getSceneGraph()->drawOne(this->_ctx, rId, pp->viewProjection);
}

void SceneRenderer::drawOneRaw(IScene* scene, const Camera* camera, const SEAScope::RenderableId& rId)
{
    fillPipelineParameters(this->_ctx, camera);

    PipelineParameters* pp = this->_ctx->pipelineParameters();

    scene->getNodeFactory()->getSceneGraph()->drawOneRaw(this->_ctx, rId, pp->viewProjection);
}

void SceneRenderer::drawRaw(IScene* scene, const Camera* camera)
{
    fillPipelineParameters(this->_ctx, camera);

    PipelineParameters* pp = this->_ctx->pipelineParameters();

    scene->getNodeFactory()->getSceneGraph()->drawRaw(this->_ctx, pp->viewProjection);
}

void SceneRenderer::drawWorld(IScene* scene, const Camera* camera, const glm::mat4& /*mvp*/, FillType type)
{
    fillPipelineParameters(this->_ctx, camera);

    PipelineParameters* pp = this->_ctx->pipelineParameters();

    scene->getNodeFactory()->drawWorld(this->_ctx, pp->viewProjection, type);
}

// ISceneRenderer
bool SceneRenderer::initialize(const std::string& renderer, SEAScope::ColormapManager* colormapMngr, const SEAScope::RendererResources* resources)
{
    OGL::ContextType contextType;

    // Convert renderer string name to enum    
    if ("OpenGLES2" == renderer)
    {
        contextType = OGL::OpenGLES2;
    }
    else if ("OpenGL3" == renderer)
    {
        contextType = OGL::OpenGL3;
    }
    else
    {
        LOG(SEAScope::LogLevel::Info, "SceneRenderer", "Unknown renderer name:  {}, default renderer used.", renderer);

        contextType = OGL::OpenGL3;
    }

    this->_ctx = OGL::ContextFactory::createContext(contextType);

    // Initialize the context (load GL extensions)
    this->_ctx->init(resources);


    this->_colormapMngr.init(this->_ctx, colormapMngr);

    return true;
}


void SceneRenderer::setDisplayCfg(IScene* scene, const SEAScope::RenderingCfg& rCfg)
{
    scene->getNodeFactory()->updateDisplayParameters(this->_ctx, rCfg);
}

const std::set<SEAScope::RenderableId>& SceneRenderer::getToRender() const
{
    return this->_toRender;
}

void SceneRenderer::setToRender(const std::set<SEAScope::RenderableId>& toRender)
{
    // @Todo Merge 3d and model to avoid a copy
    this->_toRender = toRender;
}

bool SceneRenderer::updateStarted(IScene* /*scene*/)
{
	this->_toRender.clear();
	return true;
}

bool SceneRenderer::updateFinished(IScene* scene)
{
    // @Todo Remove that from loop rendering
    scene->getNodeFactory()->setVisibleData(this->_toRender); // SLOW
	return true;
}


OGL::Context* SceneRenderer::getContext() const
{
    return this->_ctx;
}

void SceneRenderer::resize(int width, int height)
{
    LOG(SEAScope::LogLevel::Debug, "SceneRenderer",
                       "Resize to {}x{}", width, height);

    this->_ctx->setViewport(glm::ivec4(0, 0, width, height));
}

int SceneRenderer::getMaxTextureSize() const
{
    return this->_ctx->getGPUInfoInteger(OGL::max_texture_2d_size);
}

OGL::ColormapManagerGPU* SceneRenderer::getColormapManager()
{
    return &(this->_colormapMngr);
}

/**
 *
 */
void SceneRenderer::setRasterizationMode(OGL::RasterizationMode mode)
{
    this->_rasterizationMode = mode;
}
