// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/scene_view.h"
#include "3d/iscene.h"
#include "num_utils.h"

#include <vector>
#include "3d/spherical_math.h"

#include "vendors/_glm.h"

#include "3d/default_value.h"
#include "3d/context.h"
#include "core/log.h"

#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "3d/math_func.h"
#include "3d/iscene_renderer.h"
#include "3d/scene_graph_controller.h"

#include "core/colormap_manager.h"

#include "core/app_events_dispatcher.h"
#include "events/camera_moved_event.h"
#include "events/scene_view_updated_event.h"
#include "events/pixel_size_changed_event.h"

/**
 *
 */
SceneView::SceneView(SEAScope::AppEventsDispatcher& appEventsDispatcher)
: _appEventsDispatcher(appEventsDispatcher), _camera(),
  _scene(nullptr), _renderer(nullptr), _pixelSize(1.f)
{
    this->_appEventsDispatcher.subscribe<SEAScope::CameraMovedEvent>(this);
}

/**
 *
 */
SceneView::~SceneView()
{
    this->_appEventsDispatcher.unsubscribe<SEAScope::CameraMovedEvent>(this);

    this->_scene = nullptr; // not owning
    this->_renderer = nullptr; // not owning
}

/**
 *
 */
bool SceneView::setWindow(SEAScope::IWindow&)
{
    return true;
};

/**
 *
 */
bool SceneView::renderFrame()
{
    bool ok(false);
    ok = this->_renderer->renderSceneView(this);
    return ok;
}

/**
 *
 */
void SceneView::init(IScene* scene, ISceneRenderer* renderer, OGL::Texture* worldMapTexture)
{
    this->_scene = scene;
    this->_renderer = renderer;

    scene->create(renderer->getContext(), &(this->_visibility));
    scene->getNodeFactory()->setColormapManager(renderer->getColormapManager());
    scene->addWorldToSceneGraph(renderer->getContext(), worldMapTexture);

    this->_dataSetVisibility.initialize(scene->getNodeFactory(), &(this->_visibility));

    this->_visibility.setCamera(&(this->_camera));
    this->_visibility.update();
}

/**
 *
 */
IScene* SceneView::getScene() const
{
    return this->_scene;
}

/**
 *
 */
ISceneRenderer* SceneView::getRenderer() const
{
    return this->_renderer;
}

/**
 *
 */
void SceneView::setViewport(const glm::ivec2& viewport)
{
    if
        (viewport != this->_viewport)
    {
        this->_viewport = viewport;
        this->_camera.perspective(CAMERA_PERSPECTIVE_FOVY,
                                  static_cast<float>(viewport.x) / static_cast<float>(viewport.y),
                                  FRUSTUM_NEAR_PLANE,
                                  FRUSTUM_FAR_PLANE);
        this->updatePixelSize();
        this->_renderer->resize(viewport.x, viewport.y);
    }
}

/**
 *
 */
glm::ivec2 SceneView::getViewport() const
{
    return this->_viewport;
}

/**
 *
 */
Camera* SceneView::getCamera()
{
    return &(this->_camera);
}

/**
 *
 */
const Camera* SceneView::getConstCamera() const
{
    return &(this->_camera);
}

/**
 *
 */
float SceneView::computePixelSize(const glm::ivec2& viewport) const
{
    float pixelSize(0.f);

    pixelSize = equatorPixelResolution(this->getConstCamera(), viewport);
    return pixelSize;
}

/**
 *
 */
void SceneView::updatePixelSize()
{
    float pixelSize(0.f);

    pixelSize = this->computePixelSize(this->getViewport());
    if
        (true == SEAScope::floating_point_equality(pixelSize, this->_pixelSize))
    {
        // No change
        return;
    }

    // pixelSize in meters
    this->_pixelSize = pixelSize;
    this->_renderer->getContext()->setPixelSize(pixelSize);

    // pixelSize in radians (notice the "Unit" in "equatorUnitPixelResolution")
    pixelSize = equatorUnitPixelResolution(this->getConstCamera(), this->getViewport());
    this->_renderer->getContext()->setPixelSizeUnit(pixelSize);

    SEAScope::PixelSizeChangedEvent event(this->_pixelSize);
    this->_appEventsDispatcher.dispatch<SEAScope::PixelSizeChangedEvent>(event);
}

/**
 *
 */
float SceneView::getPixelSize() const
{
    return this->_pixelSize;
}

/**
 *
 */
float SceneView::getAltitude() const
{
    return this->_scene->getAltitude(this->_camera.getWorldPosition());
}

/**
 *
 */
bool SceneView::getLonLat(double x, double y, glm::vec2& result, glm::vec3& hitpoint) const
{
    if
        ((0 > x) || (x > static_cast<double>(this->_viewport.x))
         || (0 > y) || (y > static_cast<double>(this->_viewport.y)))
    {
        return false;
    }


    bool onWorld(false);
    onWorld = getPickingPosition(this->_viewport,
                                 static_cast<int>(x),
                                 static_cast<int>(y),
                                 hitpoint,
                                 glm::quat_cast(this->_camera.getModelMatrix()),
                                 this->_camera.getViewMatrix(),
                                 this->_camera.getProjectionMatrix());

    if
        (true == onWorld)
    {
        result = this->_scene->worldPositionToLonLat(hitpoint);
    }

    return onWorld;
}

/**
 *
 */
bool SceneView::handleAppEvent(std::type_index eventType, const SEAScope::AppEvent& /*event*/)
{
    if
        (std::type_index(typeid(SEAScope::CameraMovedEvent)) == eventType)
    {
        this->updatePixelSize();
        this->_visibility.update();
        this->_dataSetVisibility.update();  // uses world visibility so update world visibility first

        SEAScope::SceneViewUpdatedEvent event(this);
        this->_appEventsDispatcher.dispatch<SEAScope::SceneViewUpdatedEvent>(event);

        return true;
    }

    return false;
}
