// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/screen_quad.h"

#include "3d/core_objects.h"
#include "3d/typed_buffer.h"
#include "3d/static_mesh.h"
#include "3d/mesh_creator.h"

ScreenQuad::ScreenQuad()
: _program{nullptr},
  _displacementBuffer{nullptr},
  _vao{nullptr},
  _drawMode{nullptr},
  _size{0u}
{
}

ScreenQuad::~ScreenQuad()
{
    delete _displacementBuffer;
    _displacementBuffer = nullptr;
    delete _drawMode;
    _drawMode = nullptr;
    delete _program;
    _program = nullptr;
    delete _vao;
    _vao = nullptr;
}

void ScreenQuad::create(OGL::Context* ctx)
{
    std::vector<glm::vec2> displacement =
    {
        glm::vec2(-1.f,  1.f),
        glm::vec2( 1.f, -1.f),
        glm::vec2( 1.f,  1.f),
        glm::vec2( 1.f, -1.f),
        glm::vec2(-1.f,  1.f),
        glm::vec2(-1.f, -1.f)
    };

    _displacementBuffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(displacement));
    OGL::VertexAttribArray* vaa = OGL::Helper::createVertexAttribArray(ctx, 0, TypedBuffer(_displacementBuffer, makeDesc(displacement)));

    _vao = ctx->createVertexArrayObject();
    _vao->create(ctx);
    _vao->addObject(vaa);
    _vao->fill();

    _drawMode = OGL::Helper::createDrawModeArrays(ctx, OGL::triangles, 6);

    _program = OGL::Helper::createProgram(ctx, vertex_screenspace_quad, fragment_uni_color);

    _program->bind();
    _program->setUniformByName("arrowTileSize", 10.f);
    _program->setUniformByName("color", glm::vec4(1.f, 0.f, 0.f, 1.f));
    _program->unbind();
}

OGL::StaticMesh* ScreenQuad::createMesh(OGL::Context* ctx)
{
	std::vector<glm::vec2> displacement =
	{
		glm::vec2(-1.f,  1.f),
		glm::vec2(1.f, -1.f),
		glm::vec2(1.f,  1.f),
		glm::vec2(1.f, -1.f),
		glm::vec2(-1.f,  1.f),
		glm::vec2(-1.f, -1.f)
	};

	std::vector<glm::vec2> texCoords =
	{
		glm::vec2(0.f, 0.f),
		glm::vec2(1.f, 1.f),
		glm::vec2(1.f, 0.f),
		glm::vec2(1.f, 1.f),
		glm::vec2(0.f, 0.f),
		glm::vec2(0.f, 1.f)
	};

    MeshCreator mc;
    mc.setVertices(makeArray(displacement));
    mc.setDrawType(OGL::draw_arrays, OGL::triangles);
    mc.setTextureCoordinates(makeArray(texCoords));

    return mc.createStaticMesh(ctx);
}

void ScreenQuad::setSize(unsigned int size)
{
    _size = size;

    _program->bind();
    _program->setUniformByName("arrowTileSize", _size);
    _program->unbind();
}

void ScreenQuad::bind(OGL::Context* ctx, const glm::mat4& mvp)
{
    ctx->setDepthMask(false);

    _program->bind();

    glm::ivec4 viewport = ctx->getViewportSize();

    _program->setUniformByName("mvp", mvp);
    _program->setUniformByName("viewport", glm::vec2(static_cast<float>(viewport.z), static_cast<float>(viewport.w)));
}

void ScreenQuad::draw(OGL::Context* ctx, const glm::mat4& mvp, const glm::vec3& pos, const glm::vec3& color)
{
    ctx->setDepthMask(false);

    _program->bind();

    glm::ivec4 viewport = ctx->getViewportSize();

    _program->setUniformByName("mvp", mvp);
    _program->setUniformByName("position", pos);
    _program->setUniformByName("viewport", glm::vec2(static_cast<float>(viewport.z), static_cast<float>(viewport.w)));
    _program->setUniformByName("color", glm::vec4(color, 1.f));

    _vao->bind();
    _drawMode->draw();
    _vao->unbind();

    _program->unbind();

    ctx->setDepthMask(true);
}

void ScreenQuad::draw2(const glm::vec3& pos, const glm::vec4& color)
{
    _program->setUniformByName("position", pos);
    _program->setUniformByName("color", color);

    _vao->bind();
    _drawMode->draw();
    _vao->unbind();
}
