// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/screen_quad_textured.h"

#include "3d/core_objects.h"
#include "3d/program_bind_helper.h"

#include "3d/billboard.h"
#include "3d/math_values.h"
#include "3d/data_rendering_param.h"

#include <iostream>

ScreenQuadTextured::ScreenQuadTextured()
: TextureRenderer{},
  _sizeLoc{-1},
  _viewportLoc{-1},
  _positionLoc{-1},
  _colorLoc{-1},
  _param{nullptr}
{

}

ScreenQuadTextured::~ScreenQuadTextured()
{

}

void ScreenQuadTextured::create(OGL::Context* ctx)
{
	createProgram(ctx, vertex_screenspace_quad_tex, fragment_tex_color_alpha);

	OGL::Program* p = _program->getProgram();

	this->_sizeLoc = p->getUniformLocation("arrowTileSize");
	this->_viewportLoc = p->getUniformLocation("viewport");
	this->_positionLoc = p->getUniformLocation("position");
	this->_colorLoc = p->getUniformLocation("color");
}

void ScreenQuadTextured::bind(OGL::Context* ctx, const glm::mat4& mvp, DataRenderingParam* param)
{
	TextureRenderer::bind(ctx, mvp, param);

	OGL::Program* p = _program->getProgram();
	glm::ivec4 viewport = ctx->getViewportSize();

	constexpr float squareEdgeSize = 32.f; // px
	constexpr float halfDiagonal = (squareEdgeSize * SEAScope::Math::fsqrt2) * 0.5f;

	p->setUniform(this->_sizeLoc, halfDiagonal);
	p->setUniform(this->_viewportLoc, glm::vec2(static_cast<float>(viewport.z), static_cast<float>(viewport.w)));
    if
        (nullptr != param)
    {
        p->setUniform(this->_colorLoc, param->color);
    }
    else
    {
	    p->setUniform(this->_colorLoc, glm::vec3(0.f, 1.f, 0.f));
    }
	p->setUniform(this->_alphaLoc, 1.f);

	_param = param;
}

void ScreenQuadTextured::draw(OGL::Context* ctx, Renderable2* object)
{
	OGL::Program* p = _program->getProgram();

    Billboard* billboard = static_cast<Billboard*>(object);
	glm::vec3 billboardPos = billboard->getPosition();

    if
        (nullptr != this->_param)
    {
        glm::vec3 color;
        Style& style = billboard->getStyle();

        if (style.getState() == StyleState::normal && false == style.isHighlighted())
        {
            color = this->_param->color;
        }
        else
        {
            color = billboard->getCurrentStyleColor();
        }

        p->setUniform(this->_colorLoc, color);
        p->setUniform(this->_alphaLoc, this->_param->alpha);
    }

	p->setUniform(this->_positionLoc, billboardPos);

	TextureRenderer::draw(ctx, object);
}
