// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/screen_quad_textured_instanced.h"

#include "3d/core_objects.h"
#include "3d/typed_buffer.h"
#include "3d/static_mesh.h"
#include "3d/mesh_creator.h"
#include "3d/program_bind_helper.h"
#include "3d/data_rendering_param.h"
#include "3d/vertex_attrib_divisor.h"

ScreenQuadTexturedInstanced::ScreenQuadTexturedInstanced()
{
    _displacementBuffer = nullptr;
    _texcoordBuffer = nullptr;
    _drawMode = nullptr;
    _vao = nullptr;
    _positionsBuffer = nullptr;
    _directionsBuffer = nullptr;
}

ScreenQuadTexturedInstanced::~ScreenQuadTexturedInstanced()
{
    delete _displacementBuffer;
    _displacementBuffer = nullptr;
    delete _texcoordBuffer;
    _texcoordBuffer = nullptr;
    delete _drawMode;
    _drawMode = nullptr;
    delete _vao;
    _vao = nullptr;
    delete _positionsBuffer;
    _positionsBuffer = nullptr;
    delete _directionsBuffer;
    _directionsBuffer = nullptr;
}

void ScreenQuadTexturedInstanced::updatePositionDirection(const std::vector<glm::vec3>& positions, const std::vector<glm::vec3>& directions)
{
    _drawMode->setInstanceCount(static_cast<unsigned int>(positions.size()));

    const unsigned int posSize = static_cast<unsigned int>(positions.size()) * sizeof(glm::vec3);

    if
        (_positionsBuffer->getSize() < posSize)
    {
        _positionsBuffer->allocate(posSize, const_cast<glm::vec3*>(positions.data()));
    }
    else
    {
        _positionsBuffer->copy(0, posSize, const_cast<glm::vec3*>(positions.data()));
    }

    const unsigned int dirSize = static_cast<unsigned int>(directions.size()) * sizeof(glm::vec3);

    if
        (_directionsBuffer->getSize() < dirSize)
    {
        _directionsBuffer->allocate(dirSize, const_cast<glm::vec3*>(directions.data()));
    }
    else
    {
        _directionsBuffer->copy(0, posSize, const_cast<glm::vec3*>(directions.data()));
    }
}

void ScreenQuadTexturedInstanced::clear()
{
    _drawMode->setInstanceCount(0);
}

void ScreenQuadTexturedInstanced::create(OGL::Context* ctx)
{
    std::vector<glm::vec2> displacement =
    {
        glm::vec2(-1.f,  1.f),
        glm::vec2(1.f, -1.f),
        glm::vec2(1.f,  1.f),
        glm::vec2(1.f, -1.f),
        glm::vec2(-1.f,  1.f),
        glm::vec2(-1.f, -1.f)
    };

    std::vector<glm::vec2> texCoords =
    {
        glm::vec2(0.f, 0.f),
        glm::vec2(1.f, 1.f),
        glm::vec2(1.f, 0.f),
        glm::vec2(1.f, 1.f),
        glm::vec2(0.f, 0.f),
        glm::vec2(0.f, 1.f)
    };

    _displacementBuffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(displacement));
    OGL::VertexAttribArray* vaa = OGL::Helper::createVertexAttribArray(ctx, 0, TypedBuffer(_displacementBuffer, makeDesc(displacement)));

    _texcoordBuffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(texCoords));
    OGL::VertexAttribArray* vaa2 = OGL::Helper::createVertexAttribArray(ctx, 1, TypedBuffer(_texcoordBuffer, makeDesc(texCoords)));

    _positionsBuffer = ctx->createBuffer(OGL::array_buffer);
    _positionsBuffer->create(ctx);
    OGL::VertexAttribArray* vaaPosition = OGL::Helper::createVertexAttribArray(ctx, 2, TypedBuffer(_positionsBuffer, BufferDesc(0, 3, DataType::float32)));

    _directionsBuffer = ctx->createBuffer(OGL::array_buffer);
    _directionsBuffer->create(ctx);
    OGL::VertexAttribArray* vaaDirection = OGL::Helper::createVertexAttribArray(ctx, 3, TypedBuffer(_directionsBuffer, BufferDesc(0, 3, DataType::float32)));

    OGL::VertexAttribDivisor* vadPosition = ctx->createVertexAttribDivisor(vaaPosition, 1);
    OGL::VertexAttribDivisor* vadDirection = ctx->createVertexAttribDivisor(vaaDirection, 1);

    _vao = ctx->createVertexArrayObject();
    _vao->create(ctx);
    // Attrib
    _vao->addObject(vaa);
    _vao->addObject(vaa2);
    _vao->addObject(vaaPosition);
    _vao->addObject(vaaDirection);
    // Divisor
    _vao->addObject(vadPosition);
    _vao->addObject(vadDirection);
    _vao->fill();

    _drawMode = ctx->createDrawMode();
    _drawMode->setDrawType(OGL::draw_arrays_instanced);
    _drawMode->setDataType(DataType::float32);
    _drawMode->setPrimitive(OGL::triangles);
    _drawMode->setElementCount(6);
    _drawMode->setInstanceCount(0);
}

void ScreenQuadTexturedInstanced::draw()
{
    _vao->bind();
    _drawMode->drawInstanced();
    _vao->unbind();
}
