// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/selection_rectangle.h"

#include "3d/mesh_creator.h"
#include "3d/grid.h"
#include "3d/meta.h"
#include "3d/context.h"
#include "3d/shader_list.h"
#include "3d/helper.h"
#include "3d/buffer.h"
#include "3d/vertex_array_object.h"
#include "3d/program_bind_helper.h"
#include "3d/program.h"
#include "3d/draw_mode.h"
#include "3d/vertex_attrib_array.h"


SelectionRectangle::SelectionRectangle()
: _width{0},
  _height{0},
  _visible{false},
  _program{nullptr},
  _vao{nullptr},
  _drawMode{nullptr},
  _vertices{nullptr},
  _indices{nullptr},
  _colorLoc{-1},
  _color{0.69921875f, 0.80859375f, 0.91796875f, 0.5f},
  _colorChanged{false}
{
}

SelectionRectangle::~SelectionRectangle()
{
	delete _program;
	delete _vao;
	delete _drawMode;
	delete _vertices;
	delete _indices;
}

void SelectionRectangle::initialize(OGL::Context* ctx, int width, int height)
{
	_width = width;
	_height = height;

	std::vector<unsigned short> indices =
	{
		0, 1, 2,
		0, 2, 3
	};

	_program = OGL::Helper::createProgramAutoBind(ctx, vertex_uv_grid, fragment_uni_color);

	_colorLoc = _program->getProgram()->getUniformLocation("color");

	_program->bind(ctx);
	_program->getProgram()->setUniform(_colorLoc, this->_color);
	_program->unbind();

	_vertices = ctx->createBuffer(OGL::array_buffer);
	_vertices->create(ctx);
	_vertices->allocate(sizeof(glm::vec2) * 4, nullptr, OGL::BufferUsage::dynamic_draw);

	_indices = ctx->createBuffer(OGL::element_array_buffer);
	_indices->create(ctx);
	_indices->allocate(sizeof(unsigned short) * 6, indices.data());

	OGL::VertexAttribArray* vaa = ctx->createVertexAttribArray(0, 2, _vertices);

	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(vaa);
	_vao->addObject(_indices);
	_vao->fill();

	_drawMode = ctx->createDrawMode();
	_drawMode->setDataType(DataType::unsigned_short);
	_drawMode->setDrawType(OGL::draw_elements);
	_drawMode->setPrimitive(OGL::triangles);
	_drawMode->setElementCount(6);
}

bool SelectionRectangle::initialized() const
{
    return (nullptr != this->_vao);
}

void SelectionRectangle::setVisible(bool status)
{
	_visible = status;
}

void SelectionRectangle::setStartCorner(const glm::ivec2& p)
{
	_start = glm::ivec2(p.x, _height - p.y);
	_end = _start;

	update();
}

void SelectionRectangle::setEndCorner(const glm::ivec2& p)
{
	_end = glm::ivec2(p.x, _height - p.y);

	update();
}

void SelectionRectangle::setWindowSize(int w, int h)
{
	_width = w;
	_height = h;
    this->update();
}

glm::ivec2 SelectionRectangle::getWindowSize() const
{
    return glm::ivec2(this->_width, this->_height);
}

void SelectionRectangle::setColor(const glm::vec4& color)
{
    if
        (color != this->_color)
    {
        this->_color = color;
        this->_colorChanged = true;
    }
}

void SelectionRectangle::draw(OGL::Context* ctx, const glm::mat4& /*mvp*/)
{
	if (true == _visible)
	{
        if
            (true == this->_colorChanged)
        {
            _program->bind(ctx);
            _program->getProgram()->setUniform(_colorLoc, this->_color);
            _program->unbind();
            this->_colorChanged = false;
        }

		ctx->setFlag(OGL::blending, true);

		_program->bind(ctx);
		_vao->bind();
		_drawMode->draw();
		_vao->unbind();
		_program->unbind();

		ctx->setFlag(OGL::blending, false);
	}
}

void SelectionRectangle::update()
{
    float widthTmp(0.f);    //
    float heightTmp(0.f);   //

	glm::ivec2 min = glm::min(_start, _end);
	glm::ivec2 max = glm::max(_start, _end);

    widthTmp = static_cast<float>(this->_width);
    heightTmp = static_cast<float>(this->_height);
	float w2 = widthTmp / 2.f;
	float h2 = heightTmp / 2.f;

    float minX = static_cast<float>(min.x);
    float minY = static_cast<float>(min.y);
    float maxX = static_cast<float>(max.x);
    float maxY = static_cast<float>(max.y);

	glm::vec2 minf((minX / w2) - 1.f, (minY / h2) - 1.f);
	glm::vec2 maxf((maxX / w2) - 1.f, (maxY / h2) - 1.f);

	const std::vector<glm::vec2> vertices =
	{
		glm::vec2(minf.x, minf.y),
		glm::vec2(maxf.x, minf.y),
		glm::vec2(maxf.x, maxf.y),
		glm::vec2(minf.x, maxf.y)
	};

	_vertices->bind();
	void* ptr = _vertices->map(OGL::write_only);
	memcpy(ptr, vertices.data(), sizeof(glm::vec2) * 4);
	_vertices->unmap();
	_vertices->unbind();
}
