// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/shader_manager.h"

#include "3d/shader.h"
#include "3d/program.h"
#include "3d/context.h"
#include "3d/program_bind_helper.h"

#include "core/log.h"

#include "core/path.h"
#include "core/path_tools.h"

namespace OGL
{

ShaderManager::ShaderManager(OGL::Context* ctx)
:   _ctx(ctx)
{
}

ShaderManager::~ShaderManager()
{
    clear();
}

bool ShaderManager::addShaderFromFile(const std::string& file_name, const std::string& name, OGL::ShaderType type)
{
    OGL::Shader * shader = _ctx->createShader(type, name);
    
    bool result = shader->loadFromFile(_ctx, file_name);
    
    if (true == result && 
        this->shaders_id.find(name) == this->shaders_id.end())
    {
        this->shaders_id[name] = shader;
    }
    else
    {
        delete shader;
		shader = nullptr;
    }
    
    return result;
}

OGL::Shader * ShaderManager::getShader(const std::string& shader_name)
{
    auto it = shaders_id.find(shader_name);
    
    OGL::Shader * shader = nullptr;
    
    if (shaders_id.end() != it)
    {
        shader = it->second;
    }
    
    return shader;
}

void ShaderManager::clear()
{
    for (auto it = this->shaders_id.begin(); it != this->shaders_id.end(); ++it)
    {
        it->second->destroy();
        delete it->second;
		it->second = nullptr;
    }
    
    this->shaders_id.clear();
}

bool ShaderManager::addShaderFromFiles(const std::string& shaderFolder, const std::vector<std::string>& files)
{
    std::string ext;
    OGL::ShaderType type;
    bool result = true;
    
    for (const std::string& str : files)
    {
        ext = extension(str);
        
        if ("vert" == ext)
        {
            type = OGL::vertex_shader;
        }
        else if ("tesc" == ext)
        {
            type = OGL::tess_control_shader;
        }
        else if ("tese" == ext)
        {
            type = OGL::tess_evaluation_shader;
        }
        else if ("geom" == ext)
        {
            type = OGL::geometry_shader;
        }
        else if ("frag" == ext)
        {
            type = OGL::fragment_shader;
        }
        else if ("comp" == ext)
        {
            type = OGL::compute_shader;
        }
        else
        {
            continue;
        }
        
        result &= addShaderFromFile(shaderFolder + str, str, type);
    }
    
    return result;
}

OGL::Program * ShaderManager::createProgramTransformFeedback(const std::string& vertex, const std::string& output)
{
    LOG(SEAScope::LogLevel::Debug, "ShaderManager", "Create program with {}", vertex);

    OGL::Program * program = _ctx->createProgram();
    
    std::vector<OGL::Shader*> shaders;
    
    shaders.push_back(this->shaders_id[vertex]);
    
    program->create(_ctx);
    
    program->attachShaders(shaders);
    
    std::vector<std::string> varyings;
    varyings.push_back(output);
    
    program->setTransformFeedbackVaryings(varyings, interleaved_attribs);
    
    if (false == program->link())
    {
        program->destroy();
        delete program;
        program = nullptr;
    }
    
    return program;
}

OGL::Program * ShaderManager::createProgramTransformFeedback(const std::string& vertex, const std::vector<std::string>& outputs, TransformFeedbackBufferMode bufferMode)
{
    LOG(SEAScope::LogLevel::Debug, "ShaderManager", "Create program with {}", vertex);

    OGL::Program * program = _ctx->createProgram();
    
    std::vector<OGL::Shader*> shaders;
    
    shaders.push_back(this->shaders_id[vertex]);
    
    program->create(_ctx);
    
    program->attachShaders(shaders);
    
    program->setTransformFeedbackVaryings(outputs, bufferMode);
    
    if (false == program->link())
    {
        program->destroy();
        delete program;
        program = nullptr;
    }
    
    return program;
}

OGL::Program * ShaderManager::createProgram(const std::string& vertex, const std::string& fragment)
{
    LOG(SEAScope::LogLevel::Debug, "ShaderManager", "Create program with {} and {}", vertex, fragment);

    OGL::Program * program = _ctx->createProgram();
    
    std::vector<OGL::Shader*> shaders;
    
    shaders.push_back(this->shaders_id[vertex]);
    shaders.push_back(this->shaders_id[fragment]);
    
    program->create(_ctx);
    
    program->attachShaders(shaders);
    
    if (false == program->link())
    {
        program->destroy();
        delete program;
        program = nullptr;
    }
    
    return program;
}

OGL::ProgramBindHelper * ShaderManager::createProgramTransformFeedbackAutoBind(const std::string& vertex_shader, const std::string& output)
{
    OGL::ProgramBindHelper* helper = nullptr;
    
    OGL::Program* program = createProgramTransformFeedback(vertex_shader, output);
    
    if (program)
    {
        helper = _ctx->createProgramHelper();
        helper->setProgram(program);
    }
    
    return helper;
}

OGL::ProgramBindHelper * ShaderManager::createProgramAutoBind(const std::string& vertex_shader, const std::string& fragment_shader)
{
    OGL::ProgramBindHelper* helper = nullptr;
    
    OGL::Program* program = createProgram(vertex_shader, fragment_shader);
    
    if (program)
    {
        helper = _ctx->createProgramHelper();
        helper->setProgram(program);
    }
    
    return helper;
}
    
} // namespace OGL
