// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/shape_creator_factory.h"

#include "3d/context.h"

// Shape creator
#include "3d/point_creator.h"
#include "3d/polyline_creator.h"
#include "3d/rectangle_creator.h"
#include "3d/rectangle_creator_one_click.h"

// Builders
#include "3d/point_builder.h"
//// Polyline based
#include "3d/polyline_builder.h"
#include "3d/polygon_builder.h"
//// Rectangle based
#include "3d/rectangle_builder.h"
#include "3d/circle_builder.h"
#include "3d/ellipse_builder.h"

#include "core/log.h"

ShapeCreator* ShapeCreatorFactory::create(OGL::Context* ctx, SEAScope::ShapeType shape)
{
	ShapeCreator* sc = nullptr;

	switch (shape)
	{
	case SEAScope::ShapeType::POLYLINE:
	{
		PolylineCreator* pc = new PolylineCreator();
		pc->initialize(ctx);
		pc->setBuilder(new PolylineBuilder);
		pc->allowSegmentIntersection(true);
		sc = pc;
	}
	break;

	case SEAScope::ShapeType::POLYGON:
	{
		PolylineCreator* pc = new PolylineCreator();
		pc->initialize(ctx);
		pc->setBuilder(new PolygonBuilder);
		pc->allowSegmentIntersection(false);
		pc->setClosed(true);
		sc = pc;
	}
	break;

	case SEAScope::ShapeType::RECTANGLE:
	{
		sc = new RectangleCreator();
		sc->initialize(ctx);
		sc->setBuilder(new RectangleBuilder);
	}
	break;

	case SEAScope::ShapeType::CIRCLE:
	{
		sc = new RectangleCreator();
		sc->initialize(ctx);
		sc->setBuilder(new CircleBuilder);
	}
	break;

	case SEAScope::ShapeType::ELLIPSE:
	{
		sc = new RectangleCreator();
		sc->initialize(ctx);
		sc->setBuilder(new EllipseBuilder);
	}
	break;

	case SEAScope::ShapeType::ELLIPSE2:
	{
		sc = new RectangleCreatorOneClick();
		sc->initialize(ctx);
		sc->setBuilder(new EllipseBuilder);
	}
	break;

    case SEAScope::ShapeType::POINT:
    {
        sc = new PointCreator();
        sc->initialize(ctx);
        sc->setBuilder(new PointBuilder());
    }
    break;

    case SEAScope::ShapeType::UNKNOWN_SHAPE_TYPE:
    {
        LOG(SEAScope::LogLevel::Error, "ShapeCreatorFactory",
                           "Create method called for a shape type that is not supported");
    }
    break;

	}

	return sc;
}

