// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/shape_gpu_creator.h"

// Shapes
#include "3d/shape.h"
#include "3d/point.h"
#include "3d/polyline.h"
#include "3d/polygon.h"
#include "3d/unit_sphere.h"
#include "3d/ellipse.h"

// GPU shapes
#include "3d/polyline_mesh.h"
#include "3d/polygon_mesh.h"
#include "3d/ellipse_mesh.h"
#include "3d/ellipse_solid_mesh.h"
#include "3d/billboard.h"


#include "3d/geometry.h"

#include "core/log.h"
#include "3d/shapes_allocator.h"
#include "3d/math_func.h"

static void cartesianToLatlon(const std::vector<glm::vec3>& xyz, std::vector<Geodetic2D>& geos)
{
    geos.clear();
	geos.reserve(xyz.size());

	for (const glm::vec3& p : xyz)
	{
		geos.push_back(UnitSphere::ToGeodetic2D(p));
	}
}


Pickable*  ShapeGPUCreator::create(OGL::Context* ctx, OGL::Shape* shape, bool& ok)
{
	Pickable* pickable = nullptr;
    ok = false;

	switch (shape->getType())
	{
	case SEAScope::ShapeType::POLYLINE:
	{
		Polyline* polyline = static_cast<Polyline*>(shape);

        PolylineMesh* p = ShapesAllocator::getMesh<PolylineMesh>();
        std::vector<Geodetic2D> points;
        cartesianToLatlon(polyline->getPoints(), points);
        p->initialize(ctx, SEAScope::RenderableId(), std::move(points), polyline);
        polyline = nullptr;
        pickable = p;
	}
	break;


	case SEAScope::ShapeType::POLYGON:
	{
		Polygon* polygon = static_cast<Polygon*>(shape);

        std::vector<unsigned int> indices;
        std::vector<glm::vec3> points;

		auto pPoints = polygon->getPoints();

        ok = triangulateEarClipping(polygon->getPoints(), indices);
        if
            (false == ok)
        {
            return nullptr;
        }

        std::vector<glm::vec3> newPoints;
        std::vector<unsigned int> newIndices;

        // Now subdivise triangles to follow the globe curvature.
        subdivise(polygon->getPoints(), indices, 4.f, newPoints, newIndices);

        PolygonMesh* p = static_cast<PolygonMesh*>(ShapesAllocator::getMesh<PolygonMesh>());
        p->create(ctx, newPoints, newIndices, polygon);
        pickable = p;
	}
	break;


	case SEAScope::ShapeType::CIRCLE:
	{
		/* Not implemented yet */
	}
	break;


	case SEAScope::ShapeType::RECTANGLE:
	{
		/* Not implemented yet */
	}
	break;

    case SEAScope::ShapeType::ELLIPSE:
    case SEAScope::ShapeType::ELLIPSE2:
    {
        Ellipse* ellipse = static_cast<Ellipse*>(shape);
        EllipseMesh* em = ShapesAllocator::getMesh<EllipseMesh>();
        em->create(ctx, ellipse);
        pickable = em;
    }
    break;

    case SEAScope::ShapeType::POINT:
    {
		Point* point = static_cast<Point*>(shape);

        Billboard* billboard = ShapesAllocator::getMesh<Billboard>();
        billboard->initialize(point);
        pickable = billboard;
    }
    break;

    case SEAScope::ShapeType::UNKNOWN_SHAPE_TYPE:
    {
        LOG(SEAScope::LogLevel::Error, "ShapeGPUCreator",
                           "Create method called for a shape type that is not supported");
    }
    break;

	}

    ok = nullptr != pickable;
    return pickable;
}

void ShapeGPUCreator::update(OGL::Context* ctx, Pickable* obj, OGL::Shape* shape)
{
	switch (shape->getType())
	{
	case SEAScope::ShapeType::POLYLINE:
	{
		PolylineMesh* mesh = static_cast<PolylineMesh*>(obj);
		Polyline* p = static_cast<Polyline*>(shape);

		if (nullptr != mesh && nullptr != p)
		{
			std::vector<Geodetic2D> points;
			cartesianToLatlon(p->getPoints(), points);
			mesh->setPolyline(p);
            mesh->setPoints(std::move(points));
		}
	}
	break;


	case SEAScope::ShapeType::POLYGON:
	{
		PolygonMesh* mesh = static_cast<PolygonMesh*>(obj);
		Polygon* p = static_cast<Polygon*>(shape);

		if (nullptr != mesh && nullptr != p)
		{
			std::vector<unsigned int> indices;

			bool res;

			res = triangulateEarClipping(p->getPoints(), indices);

			if (res)
			{
				std::vector<glm::vec3> newPoints;
				std::vector<unsigned int> newIndices;

				// Now subdivise triangles to follow the globe curvature.
				subdivise(p->getPoints(), indices, 4.f, newPoints, newIndices);

				mesh->create(ctx, newPoints, newIndices, p);
			}
		}
	}
	break;


	case SEAScope::ShapeType::CIRCLE:
	{
		/* Not implemented yet */
	}
	break;


	case SEAScope::ShapeType::RECTANGLE:
	{
		/* Not implemented yet */
	}
	break;

    case SEAScope::ShapeType::ELLIPSE:
    case SEAScope::ShapeType::ELLIPSE2:
    {
        EllipseMesh* mesh = static_cast<EllipseMesh*>(obj);
        Ellipse* e = static_cast<Ellipse*>(shape);
        if (nullptr != mesh && nullptr != e)
        {

            mesh->update(e);
        }
    }
    break;

    case SEAScope::ShapeType::POINT:
    {
        if
            (nullptr != obj && nullptr != shape)
        {
            Billboard* mesh = static_cast<Billboard*>(obj);
            Point* point = static_cast<Point*>(shape);

            mesh->setPosition(point->getPosition());
        }
    }
    break;

    case SEAScope::ShapeType::UNKNOWN_SHAPE_TYPE:
    {
        LOG(SEAScope::LogLevel::Error, "ShapeGPUCreator",
                           "Update method called for a shape type that is not supported");
    }
    break;

	}
}
