// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/sphere.h"

#include "3d/geodetic.h"
#include "3d/unit_sphere.h"

#include "num_utils.h"

void cube_subdiv(float radius, int max_depth, std::vector<glm::vec3>& vertices);
void subdivQuad(
        const glm::vec3& a, const glm::vec3& b, const glm::vec3& c, const glm::vec3& d,
        int depth, int max_depth, std::vector<glm::vec3>& vertices);

void createCubeSphere(std::vector<glm::vec3>& vertices, float radius, int subdivision)
{
    cube_subdiv(radius, subdivision, vertices);
}

void cube_subdiv(float radius, int max_depth, std::vector<glm::vec3>& vertices)
{
    float s = radius;
    
    const glm::vec3 cube[] =
    {
        // back
        glm::vec3(-s,-s,-s),  
        glm::vec3(-s, s,-s),  
        glm::vec3( s, s,-s),  
        glm::vec3( s,-s,-s),  
        
        // front
        glm::vec3(-s,-s, s),  
        glm::vec3( s,-s, s),  
        glm::vec3( s, s, s),  
        glm::vec3(-s, s, s), 

        // right
        glm::vec3(s, -s,-s),  
        glm::vec3(s,  s,-s),  
        glm::vec3(s,  s, s),  
        glm::vec3(s, -s, s), 
         
        // left
        glm::vec3(-s, -s,-s), 
        glm::vec3(-s, -s, s), 
        glm::vec3(-s,  s, s),  
        glm::vec3(-s,  s,-s),  
        
        // bottom
        glm::vec3(-s, -s, -s),  
        glm::vec3( s, -s, -s),  
        glm::vec3( s, -s,  s),  
        glm::vec3(-s, -s,  s),
        
        // top
        glm::vec3(-s,  s, -s),  
        glm::vec3(-s,  s,  s), 
        glm::vec3( s,  s,  s),
        glm::vec3( s,  s, -s)
    };
    
    for (int i = 0; i < 6; ++i)
    {
        subdivQuad( cube[0 + i * 4], cube[1 + i * 4], cube[2 + i * 4], cube[3 + i * 4],
                0, 
                max_depth, 
                vertices);
    }
}


// Subdivision of a quad
void subdivQuad(
        const glm::vec3& a, const glm::vec3& b, const glm::vec3& c, const glm::vec3& d, 
        int depth, 
        int max_depth, 
        std::vector<glm::vec3>& vertices)
{
    if (depth == max_depth)
    {
        // vertices coords
        vertices.push_back(glm::normalize(a));
        vertices.push_back(glm::normalize(b));
        vertices.push_back(glm::normalize(c));
        vertices.push_back(glm::normalize(c));
        vertices.push_back(glm::normalize(d));
        vertices.push_back(glm::normalize(a));
    }
    else
    {
        // Compute new vertices positions
        glm::vec3 ab = a + ((b - a) * .5f);
        glm::vec3 cd = c + ((d - c) * .5f);
        glm::vec3 bc = b + ((c - b) * .5f);
        glm::vec3 ad = a + ((d - a) * .5f);
        glm::vec3 o =  ab + ((cd - ab) * .5f);
    
        // recursive function call
        subdivQuad(ab, b, bc, o,    depth + 1, max_depth, vertices);
        subdivQuad(o, bc, c, cd,    depth + 1, max_depth, vertices);
        subdivQuad(ad, o, cd, d,    depth + 1, max_depth, vertices);
        subdivQuad(a, ab, o, ad,    depth + 1, max_depth, vertices);
    }
}

void createUVSphere(float radius, int latitude, int longitude, std::vector<glm::vec3>& vertices, std::vector<glm::vec2>& texcoords)
{
    double tmpValue(0.0);    //
    double theta, phi;
    glm::vec3 p0, p1, p2, p3;    
    glm::vec2 uv0, uv1, uv2, uv3;
    
    double lat = static_cast<double>(latitude);
    double lon = static_cast<double>(longitude);

    double latitude_step = (1.0 / (lat - 1.0));
    double longitude_step = (1.0 / (lon - 1.0));
    
    for(int i = 0; i < longitude - 1; ++i)
    {
        phi = M_PI_2;
        theta = -M_PI + (i * longitude_step * M_PI * 2.0);

        tmpValue = cos(phi) * cos(theta);
        p0.x =  static_cast<float>(tmpValue);
        tmpValue = sin(phi);
        p0.y = static_cast<float>(tmpValue);
        tmpValue = cos(phi) * sin(theta);
        p0.z = static_cast<float>(tmpValue);
        
        tmpValue = (M_PI - theta) / (M_PI * 2.0);
        uv0.x = static_cast<float>(tmpValue);
        tmpValue = (M_PI_2 - phi) / M_PI;
        uv0.y = static_cast<float>(tmpValue);
        
        phi = M_PI_2 - latitude_step * M_PI;
        theta = -M_PI + (i * longitude_step * M_PI * 2.0);

        tmpValue = cos(phi) * cos(theta);
        p1.x = static_cast<float>(tmpValue);
        tmpValue = sin(phi);
        p1.y = static_cast<float>(tmpValue);
        tmpValue = cos(phi) * sin(theta);
        p1.z = static_cast<float>(tmpValue);
        
        tmpValue = (M_PI - theta) / (M_PI * 2.0);
        uv1.x = static_cast<float>(tmpValue);
        tmpValue = (M_PI_2 - phi) / M_PI;
        uv1.y = static_cast<float>(tmpValue);
        
        theta = -M_PI + ((i+1) * longitude_step * M_PI * 2.0);

        tmpValue = cos(phi) * cos(theta);
        p2.x =  static_cast<float>(tmpValue);
        tmpValue = sin(phi);
        p2.y = static_cast<float>(tmpValue);
        tmpValue = cos(phi) * sin(theta);
        p2.z = static_cast<float>(tmpValue);
        
        tmpValue = (M_PI - theta) / (M_PI * 2.0);
        uv2.x = static_cast<float>(tmpValue);
        tmpValue = (M_PI_2 - phi) / M_PI;
        uv2.y = static_cast<float>(tmpValue);
        
        vertices.push_back(p0 * radius);
        vertices.push_back(p2 * radius);
        vertices.push_back(p1 * radius);
        
        texcoords.push_back(uv0);
        texcoords.push_back(uv2);
        texcoords.push_back(uv1);
    }
    
    for(int j = 1; j < latitude-2; ++j)
    {
        for(int i = 0; i < longitude-1; ++i)
        {
            phi = M_PI_2 - (j * latitude_step * M_PI);
            theta = -M_PI + (i * longitude_step * M_PI * 2.0);

            tmpValue = cos(phi) * cos(theta);
            p0.x = static_cast<float>(tmpValue);
            tmpValue = sin(phi);
            p0.y = static_cast<float>(tmpValue);
            tmpValue = cos(phi) * sin(theta);
            p0.z = static_cast<float>(tmpValue);

            tmpValue = (M_PI - theta) / (M_PI * 2.0);
            uv0.x = static_cast<float>(tmpValue);
            tmpValue = (M_PI_2 - phi) / M_PI;
            uv0.y = static_cast<float>(tmpValue);
            
            phi = M_PI_2 - (j * latitude_step * M_PI);
            theta = -M_PI + ((i+1) * longitude_step * M_PI * 2.0);

            tmpValue = cos(phi) * cos(theta);
            p1.x = static_cast<float>(tmpValue);
            tmpValue = sin(phi);
            p1.y = static_cast<float>(tmpValue);
            tmpValue = cos(phi) * sin(theta);
            p1.z = static_cast<float>(tmpValue);

            tmpValue = (M_PI - theta) / (M_PI * 2.0);
            uv1.x = static_cast<float>(tmpValue);
            tmpValue = (M_PI_2 - phi) / M_PI;
            uv1.y = static_cast<float>(tmpValue);
            
            phi = M_PI_2 - ((j+1) * latitude_step * M_PI);
            theta = -M_PI + (i * longitude_step * M_PI * 2.0);

            tmpValue = cos(phi) * cos(theta);
            p2.x = static_cast<float>(tmpValue);
            tmpValue = sin(phi);
            p2.y = static_cast<float>(tmpValue);
            tmpValue = cos(phi) * sin(theta);
            p2.z = static_cast<float>(tmpValue);

            tmpValue = (M_PI - theta) / (M_PI * 2.0);
            uv2.x = static_cast<float>(tmpValue);
            tmpValue = (M_PI_2 - phi) / M_PI;
            uv2.y = static_cast<float>(tmpValue);

            phi = M_PI_2 - ((j+1) * latitude_step * M_PI);
            theta = -M_PI + ((i+1) * longitude_step * M_PI * 2.0);

            tmpValue = cos(phi) * cos(theta);
            p3.x = static_cast<float>(tmpValue);
            tmpValue = sin(phi);
            p3.y = static_cast<float>(tmpValue);
            tmpValue = cos(phi) * sin(theta);
            p3.z = static_cast<float>(tmpValue);

            tmpValue = (M_PI - theta) / (M_PI * 2.0);
            uv3.x = static_cast<float>(tmpValue);
            tmpValue = (M_PI_2 - phi) / M_PI;
            uv3.y = static_cast<float>(tmpValue);

            vertices.push_back(p0 * radius);
            vertices.push_back(p1 * radius);
            vertices.push_back(p2 * radius);
            
            vertices.push_back(p1 * radius);
            vertices.push_back(p3 * radius);
            vertices.push_back(p2 * radius);

            texcoords.push_back(uv0);
            texcoords.push_back(uv1);
            texcoords.push_back(uv2);
            
            texcoords.push_back(uv1);
            texcoords.push_back(uv3);
            texcoords.push_back(uv2);
        }
    }
    
    
    for(int i = 0; i < longitude-1; ++i)
    {
        phi = -M_PI_2;
        theta = -M_PI + (i * longitude_step * M_PI * 2.0);

        tmpValue = cos(phi) * cos(theta);
        p0.x = static_cast<float>(tmpValue);
        tmpValue = sin(phi);
        p0.y = static_cast<float>(tmpValue);
        tmpValue = cos(phi) * sin(theta);
        p0.z = static_cast<float>(tmpValue);
        
        tmpValue = (M_PI - theta) / (M_PI * 2.0);
        uv0.x = static_cast<float>(tmpValue);
        tmpValue = (M_PI_2 - phi) / M_PI;
        uv0.y = static_cast<float>(tmpValue);
        
        phi = M_PI_2 - ((latitude-2) * latitude_step * M_PI);
        theta = -M_PI + (i * longitude_step * M_PI * 2.0);

        tmpValue = cos(phi) * cos(theta);
        p1.x = static_cast<float>(tmpValue);
        tmpValue = sin(phi);
        p1.y = static_cast<float>(tmpValue);
        tmpValue = cos(phi) * sin(theta);
        p1.z = static_cast<float>(tmpValue);
        
        tmpValue = (M_PI - theta) / (M_PI * 2.0);
        uv1.x = static_cast<float>(tmpValue);
        tmpValue = (M_PI_2 - phi) / M_PI;
        uv1.y = static_cast<float>(tmpValue);
        
        theta = -M_PI + ((i+1) * longitude_step * M_PI * 2.0);

        tmpValue = cos(phi) * cos(theta);
        p2.x = static_cast<float>(tmpValue);
        tmpValue = sin(phi);
        p2.y = static_cast<float>(tmpValue);
        tmpValue = cos(phi) * sin(theta);
        p2.z = static_cast<float>(tmpValue);
        
        tmpValue = (M_PI - theta) / (M_PI * 2.0);
        uv2.x = static_cast<float>(tmpValue);
        tmpValue = (M_PI_2 - phi) / M_PI;
        uv2.y = static_cast<float>(tmpValue);
        
        vertices.push_back(p0 * radius);
        vertices.push_back(p1 * radius);
        vertices.push_back(p2 * radius);
        
        texcoords.push_back(uv0);
        texcoords.push_back(uv1);
        texcoords.push_back(uv2);
    }
}

/*
static void generateSphere(int width, int height)
{
    double theta, phi, tmpValue, heightTmp, widthTmp, iTmp, jTmp;
    float *dat, *tex;
    int i, j, t, nvec, *idx;

    nvec = (height-2)* width+2;
//    ntri = (height-2)*(width-1)*2;

    std::size_t nvecTmp = static_cast<std::size_t>(nvec);
    dat = static_cast<float*>(malloc(nvecTmp * 3 * sizeof(float)));
    tex = static_cast<float*>(malloc(nvecTmp * 2 * sizeof(float)));
    idx = static_cast<int*>(malloc( 3 * sizeof(int)));

    widthTmp = static_cast<double>(width);
    heightTmp = static_cast<double>(height);

    for(t=0, j=1; j<height-1; j++)
    {
        jTmp = static_cast<double>(j);

        for(i=0; i <width; i++)
        {
            iTmp = static_cast<double>(i);

            theta = jTmp / (heightTmp - 1.0) * M_PI;
            phi   = iTmp / (widthTmp - 1.0 ) * M_PI * 2.0;

            tmpValue = sin(theta) * cos(phi);
            dat[t++] = static_cast<float>(tmpValue);
            tmpValue = cos(theta);
            dat[t++] = static_cast<float>(tmpValue);
            tmpValue = -sin(theta) * sin(phi);
            dat[t++] = static_cast<float>(tmpValue);
        }
    }

    dat[t++]=0; dat[t++]= 1; dat[t++]=0;
    dat[t++]=0; dat[t++]=-1; dat[t++]=0;

    for(t=0, j=0; j<height-3; j++ )
    {
        for(i=0; i<width-1; i++ )
        {
            idx[t++] = (j  )*width + i  ;
            idx[t++] = (j+1)*width + i+1;
            idx[t++] = (j  )*width + i+1;
            idx[t++] = (j  )*width + i  ;
            idx[t++] = (j+1)*width + i  ;
            idx[t++] = (j+1)*width + i+1;
        }
    }
    
    for( i=0; i<width-1; i++ )
    {
        idx[t++] = (height-2)*width;
        idx[t++] = i;
        idx[t++] = i+1;
        idx[t++] = (height-2)*width+1;
        idx[t++] = (height-3)*width + i+1;
        idx[t++] = (height-3)*width + i;
    }

    free(idx);
    free(dat);
}
*/


//////////////////////////////////////////////////////////////////////////////////////////////////
// Octahedron

using namespace glm;


/*
static void CreateUV (std::vector<vec3>& vertices, std::vector<vec3>& uv) 
{
    float previousX = 1.f;  //
    double tmpValue(0.0);   //
    
    for (std::size_t i = 0; i < vertices.size(); i++) 
    {
        vec3& v = vertices[i];
        
        if (SEAScope::floating_point_equality(v.x, previousX)) 
        {
            uv[i - 1].x = 1.f;
        }
        
        previousX = v.x;
        
        vec3 textureCoordinates;
        tmpValue = static_cast<double>(atan2(v.x, v.z)) / (-2.0 * M_PI);
        textureCoordinates.x = static_cast<float>(tmpValue);
        
        if (textureCoordinates.x < 0.f)
        {
            textureCoordinates.x += 1.f;
        }
        tmpValue = static_cast<double>(asin(v.y)) / M_PI + 0.5;
        textureCoordinates.y = static_cast<float>(tmpValue);
        uv[i] = textureCoordinates;
    }
    uv[vertices.size() - 4].x = uv[0].x = 0.125f;
    uv[vertices.size() - 3].x = uv[1].x = 0.375f;
    uv[vertices.size() - 2].x = uv[2].x = 0.625f;
    uv[vertices.size() - 1].x = uv[3].x = 0.875f;
}
*/
static int CreateVertexLine(vec3& from, vec3& to, int steps, int vIndex, std::vector<vec3>& vertices) 
{
    float stepsTmp(0.f);    //
    std::size_t v = static_cast<std::size_t>(vIndex);

    stepsTmp = static_cast<float>(steps);

    for (int i = 1; i <= steps; i++) 
    {
        vertices[v++] = mix(from, to, static_cast<float>(i) / stepsTmp);
    }
    return static_cast<int>(v);
}

static int CreateLowerStrip (int steps, int vTop, int vBottom, int tIndex, std::vector<unsigned int>& triangles) 
{
    std::size_t t = static_cast<std::size_t>(tIndex);

    for (int i = 1; i < steps; i++)
    {
        triangles[t++] = static_cast<unsigned int>(vBottom);
        triangles[t++] = static_cast<unsigned int>(vTop - 1);
        triangles[t++] = static_cast<unsigned int>(vTop);

        triangles[t++] = static_cast<unsigned int>(vBottom);
        triangles[t++] = static_cast<unsigned int>(vTop);
        triangles[t++] = static_cast<unsigned int>(vBottom);

        vBottom++;
        vTop++;
    }
    
    triangles[t++] = static_cast<unsigned int>(vBottom);
    triangles[t++] = static_cast<unsigned int>(vTop - 1);
    triangles[t++] = static_cast<unsigned int>(vTop);
    
    return static_cast<int>(t);
}

static int CreateUpperStrip (int steps, int vTop, int vBottom, int tIndex, std::vector<unsigned int>& triangles) 
{
    std::size_t t = static_cast<std::size_t>(tIndex);

    triangles[t++] = static_cast<unsigned int>(vBottom);
    triangles[t++] = static_cast<unsigned int>(vTop - 1);
    triangles[t++] = static_cast<unsigned int>(vBottom + 1);

    vBottom++;
    
    for (int i = 1; i <= steps; i++) 
    {
        triangles[t++] = static_cast<unsigned int>(vTop - 1);
        triangles[t++] = static_cast<unsigned int>(vTop);
        triangles[t++] = static_cast<unsigned int>(vBottom);

        triangles[t++] = static_cast<unsigned int>(vBottom);
        triangles[t++] = static_cast<unsigned int>(vTop);
        triangles[t++] = static_cast<unsigned int>(vBottom + 1);

        vBottom++;
        vTop++;
    }
    
    return static_cast<int>(t);
}

/*
static void Normalize (std::vector<vec3>& vertices, std::vector<vec3>& normals) 
{
    for (std::size_t i = 0; i < vertices.size(); i++) 
    {
        normals[i] = vertices[i] = normalize(vertices[i]);
    }
}
*/

static void CreateOctahedron (std::vector<vec3>& vertices, std::vector<unsigned int>& triangles, int resolution) 
{
    std::size_t v(0);   //
    std::size_t t(0);   //
    int tTmp(0);    //
    int vTmp(0);    //
    int vBottom(0); //
    vec3 from, to;
    float resolutionTmp(0.f);   //

    constexpr vec3 down(0,-1,0);
    constexpr vec3 up(0,1,0);
    constexpr vec3 forward(0,0,1);
    constexpr vec3 left(-1,0,0);
    constexpr vec3 back(0,0,-1);
    constexpr vec3 right(1,0,0);

    constexpr vec3 directions[] =
    {
        left,
        back,
        right,
        forward
    };

    resolutionTmp = static_cast<float>(resolution);
    
    for (std::size_t i = 0; i < 4; i++) 
    {
        vertices[v++] = down;
    }

    for (int i = 1; i <= resolution; i++) 
    {
        float progress = static_cast<float>(i) / resolutionTmp;
        
        to = mix(down, forward, progress);
        vertices[v++] = to;
        
        for (int d = 0; d < 4; d++) 
        {
            from = to;
            to = mix(down, directions[d], progress);

            tTmp = static_cast<int>(t);
            vTmp = static_cast<int>(v);

            tTmp = CreateLowerStrip(i, vTmp, vBottom, tTmp, triangles);
            vTmp = CreateVertexLine(from, to, i, vTmp, vertices);

            t = static_cast<std::size_t>(tTmp);
            v = static_cast<std::size_t>(vTmp);

            vBottom += i > 1 ? (i - 1) : 1;
        }
        vBottom = vTmp - 1 - i * 4;
    }

    for (int i = resolution - 1; i >= 1; i--) 
    {
        float progress = static_cast<float>(i) / resolutionTmp;
        
        vertices[v++] = to = mix(up, forward, progress);
        
        for (int d = 0; d < 4; d++) 
        {
            from = to;
            to = mix(up, directions[d], progress);

            tTmp = static_cast<int>(t);
            vTmp = static_cast<int>(v);

            tTmp = CreateUpperStrip(i, vTmp, vBottom, tTmp, triangles);
            vTmp = CreateVertexLine(from, to, i, vTmp, vertices);

            t = static_cast<std::size_t>(tTmp);
            v = static_cast<std::size_t>(vTmp);

            vBottom += i + 1;
        }
        vBottom = vTmp - 1 - i * 4;
    }

    for (int i = 0; i < 4; i++) 
    {
        triangles[t++] = static_cast<unsigned int>(vBottom);
        triangles[t++] = static_cast<unsigned int>(v);
        triangles[t++] = static_cast<unsigned int>(vBottom + 1);
        
        vertices[v++] = up;

        vBottom++;
    }
}

/*
static void Create (int subdivisions, float radius, std::vector<vec3>& vertices, std::vector<unsigned int>& triangles, std::vector<vec3>& uv, std::vector<vec3>& normals) 
{
    std::size_t resolution = 1 << subdivisions;
    
    vertices.resize((resolution + 1) * (resolution + 1) * 4 - (resolution * 2 - 1) * 3);
    triangles.resize((1 << (subdivisions * 2 + 3)) * 3);
        
    CreateOctahedron(vertices, triangles, subdivisions);
    
    normals.resize(vertices.size());
    
    Normalize(vertices, normals);

    uv.resize(vertices.size());
    
    CreateUV(vertices, uv);

    if (radius != 1.f) 
    {
        for (std::size_t i = 0; i < vertices.size(); i++) 
        {
            vertices[i] *= radius;
        }
    }
}
*/

/*
static void createOctahedronSphere(int subdivisions, float radius, std::vector<vec3>& vertices, std::vector<unsigned int>& triangles, std::vector<vec3>& uv, std::vector<vec3>& normals)
{
    Create(subdivisions, radius, vertices, triangles, uv, normals);
}
*/

void createOctahedronSphere(int subdivisions, float radius, std::vector<vec3>& vertices, std::vector<unsigned int>& triangles)
{
    std::size_t resolution = 1ull << subdivisions;
    
    vertices.resize((resolution + 1) * (resolution + 1) * 4 - (resolution * 2 - 1) * 3);
    triangles.resize((1ull << (subdivisions * 2 + 3)) * 3);
        
    CreateOctahedron(vertices, triangles, subdivisions);

    if (radius != 1.f) 
    {
        for (std::size_t i = 0; i < vertices.size(); i++) 
        {
            vertices[i] *= radius;
        }
    }
}

float getIcosahedronDistance(int subdivision)
{
    int divisor = 1 << subdivision;
    float divisorTmp = static_cast<float>(divisor);
    glm::vec3 p0 = glm::mix(icosahedronVertices[2], icosahedronVertices[3], 1.f / divisorTmp);
    p0 = glm::normalize(p0);
    return glm::distance(icosahedronVertices[2], p0);
}

float getIcosahedronLonLat(int subdivision)
{
    int divisor = 1 << subdivision;
    float divisorTmp = static_cast<float>(divisor);
    glm::vec3 p0 = glm::mix(icosahedronVertices[2], icosahedronVertices[3], 1.f / divisorTmp);
    p0 = glm::normalize(p0);
    
    Geodetic2D g0 = UnitSphere::ToGeodetic2D(icosahedronVertices[2]);
    Geodetic2D g1 = UnitSphere::ToGeodetic2D(p0);
    
    glm::vec2 vg0(glm::degrees(g0.Longitude()), glm::degrees(g0.Latitude()));
    glm::vec2 vg1(glm::degrees(g1.Longitude()), glm::degrees(g1.Latitude()));
    
    return glm::distance(vg0, vg1);
}

float getOctahedronLonLat(int subdivision)
{
    int divisor = 1 << subdivision;
    float divisorTmp = static_cast<float>(divisor);
    glm::vec3 p0 = glm::mix(glm::vec3(1.f, 0.f, 0.f), glm::vec3(0.f, 1.f, 0.f), 1.f / divisorTmp);
    p0 = glm::normalize(p0);
    
    Geodetic2D g0 = UnitSphere::ToGeodetic2D(glm::vec3(1.f, 0.f, 0.f));
    Geodetic2D g1 = UnitSphere::ToGeodetic2D(p0);
    
    glm::vec2 vg0(glm::degrees(g0.Longitude()), glm::degrees(g0.Latitude()));
    glm::vec2 vg1(glm::degrees(g1.Longitude()), glm::degrees(g1.Latitude()));
    
    return glm::distance(vg0, vg1);
}

void subdivTriangle(
        const glm::vec3& a, const glm::vec3& b, const glm::vec3& c, 
        int subdiv,
        std::vector<glm::vec3>& vertices)
{
    if (subdiv > 0)
    {
        // Compute new vertices positions
        glm::vec3 ab = a + ((b - a) * .5f);
        glm::vec3 bc = b + ((c - b) * .5f);
        glm::vec3 ca = c + ((a - c) * .5f);
        
        vertices.push_back(glm::normalize(ab));
        vertices.push_back(glm::normalize(bc));
        vertices.push_back(glm::normalize(ca));
        
        --subdiv;
        
        // recursive function call
        subdivTriangle(a, ab, ca, subdiv, vertices);
        subdivTriangle(ab, b, bc, subdiv, vertices);
        subdivTriangle(ca, bc, c, subdiv, vertices);
        subdivTriangle(ca, ab, bc, subdiv, vertices);
    }
}

void subdivBarycentric(
        const glm::vec3& a, const glm::vec3& b, const glm::vec3& c, 
        int subdiv,
        std::vector<glm::vec3>& barycentric)
{
    if (subdiv > 0)
    {
        // Compute new vertices positions
        glm::vec3 ab = (b + a) * .5f;
        glm::vec3 bc = (c + b) * .5f;
        glm::vec3 ca = (a + c) * .5f;
        
        barycentric.push_back(ab);
        barycentric.push_back(bc);
        barycentric.push_back(ca);
        
        --subdiv;
        
        // recursive function call
        subdivTriangle(a, ab, ca, subdiv, barycentric);
        subdivTriangle(ab, b, bc, subdiv, barycentric);
        subdivTriangle(ca, bc, c, subdiv, barycentric);
        subdivTriangle(ca, ab, bc, subdiv, barycentric);
    }
}

void createIcosahedronSphere(int subdivision, std::vector<glm::vec3>& vertices)
{
    int triangleCount = sizeof(icosahedronTriangles) / 3;
    
    for (int i = 0; i < triangleCount; ++i)
    {
        glm::vec3& a = icosahedronVertices[icosahedronTriangles[i*3+0]];
        glm::vec3& b = icosahedronVertices[icosahedronTriangles[i*3+1]];
        glm::vec3& c = icosahedronVertices[icosahedronTriangles[i*3+2]];
        
        vertices.push_back(glm::normalize(a));
        vertices.push_back(glm::normalize(b));
        vertices.push_back(glm::normalize(c));
        
        subdivTriangle(
            a, b, c,
            subdivision,
            vertices);
    }
}

static void addVertex(const glm::vec3& vertex, std::vector<glm::vec3>& vertices)
{
    glm::vec3 p = UnitSphere::ToVector3D(Geodetic2D(glm::radians(glm::vec2(vertex.x, vertex.y))));
    vertices.push_back(p);
}

/*
static void toto(const glm::vec3& pole, const glm::vec3* middle, int i, int subdivY, std::vector<glm::vec3>& vertices)
{
    float subdivYTmp(0.f);  //

    subdivYTmp = static_cast<float>(subdivY);

    // position in % between top and equator.
    float a = static_cast<float>(i) / subdivYTmp;

    // number of vertices for this slice.
    int subdivX = i * 4;

    // number of inner vertices for this slice per octahedron triangle.
    int interX = (subdivX - 4) / 4;
    float interXTmp = static_cast<float>(interX);

    // Compute left point on the triangle edge
    glm::vec3 p0 = glm::mix(pole, middle[0], a);
    // Copy the position 
    glm::vec3 p1 = p0;

    // Get points pointers.
    glm::vec3 * pL = &p0;
    glm::vec3 * pR = &p1;

    // Loop on 4 equators octahedron points.
    for (int j = 0; j < 4; ++j)
    {   
        // Swap left and right pointers.
        // We are turning around the octahedron, so each loop the right point
        // become the left point. Less computation.
        std::swap(pL, pR);

        // Compute right point
        *pR = glm::mix(pole, middle[(j+1)%4], a);

        // Add the left point the vertices array
        addVertex(*pL, vertices);
        
        // Loop on inners points
        for (int k = 0; k < interX; ++k)
        {
            // Compute the % between left and right points
            float interMix = static_cast<float>(k + 1) / (interXTmp + 1.0f);
            // Compute the position
            glm::vec3 pIn = glm::mix(*pL, *pR, interMix);
            // Add to the vertices array
            addVertex(pIn, vertices);
        }
    }
}
*/

static void totoLonLat(const glm::vec3* pole, const glm::vec3* middle, int i, int subdivY, std::vector<glm::vec3>& vertices)
{
    float subdivYTmp(0.f);  //

    subdivYTmp = static_cast<float>(subdivY);

    // position in % between top and equator.
    float a = static_cast<float>(i) / subdivYTmp;

    // number of vertices for this slice.
    int subdivX = i * 4;

    // number of inner vertices for this slice per octahedron triangle.
    int interX = (subdivX - 4) / 4;
    float interXTmp = static_cast<float>(interX);

    // Compute left point on the triangle edge
    glm::vec3 p0 = glm::mix(pole[0], middle[0], a);
    // Copy the position 
    glm::vec3 p1 = p0;

    // Get points pointers.
    glm::vec3 * pL = &p0;
    glm::vec3 * pR = &p1;

    // Loop on 4 equators octahedron points.
    for (int j = 0; j < 4; ++j)
    {   
        // Swap left and right pointers.
        // We are turning around the octahedron, so each loop the right point
        // become the left point. Less computation.
        std::swap(pL, pR);

        // Compute right point
        *pR = glm::mix(pole[(j+1)], middle[(j+1)], a);

        // Add the left point the vertices array
        addVertex(*pL, vertices);
        
        // Loop on inners points
        for (int k = 0; k < interX; ++k)
        {
            // Compute the % between left and right points
            float interMix = static_cast<float>(k+1) / (interXTmp + 1.f);
            // Compute the position
            glm::vec3 pIn = glm::mix(*pL, *pR, interMix);
            // Add to the vertices array
            addVertex(pIn, vertices);
        }
    }
}

static void fillCoord(Octahedron* o, Triangle* t, const glm::ivec2& a, const glm::ivec2& b, const glm::ivec2& c)
{
    std::size_t index(0);   //
    int tmpValue(0);  //

    index = static_cast<std::size_t>(a.y);
    tmpValue = o->offsetPerSlice[index] + (t->id * o->vertexPerTriangleSlice[index] + a.x) % o->vertexPerSlice[index];
    t->a = static_cast<std::uint16_t>(tmpValue);

    index = static_cast<std::size_t>(b.y);
    tmpValue = o->offsetPerSlice[index] + (t->id * o->vertexPerTriangleSlice[index] + b.x) % o->vertexPerSlice[index];
    t->b = static_cast<std::uint16_t>(tmpValue);

    index = static_cast<std::size_t>(c.y);
    tmpValue = o->offsetPerSlice[index] + (t->id * o->vertexPerTriangleSlice[index] + c.x) % o->vertexPerSlice[index];
    t->c = static_cast<std::uint16_t>(tmpValue);
}

static void subdivTriangleQuadtree(Octahedron* o, const glm::ivec2& a, const glm::ivec2& b, const glm::ivec2& c, Triangle* t, int depth, int maxDepth)
{
    fillCoord(o, t, a, b, c);

    if (depth == maxDepth)
    {
        return;
    }

    glm::ivec2 ab;
    glm::ivec2 bc;
    glm::ivec2 ca;

    ab = a + (b - a) / 2;
    bc = b + (c - b) / 2;
    ca = c + (a - c) / 2;
    
    for (int i = 0; i < 4; ++i)
    {
        t->children[i] = new Triangle;
        t->children[i]->id = t->id;
    }
    
    subdivTriangleQuadtree(o, a, ca, ab, t->children[0], depth + 1, maxDepth);
    subdivTriangleQuadtree(o, ab, bc, b, t->children[1], depth + 1, maxDepth);
    subdivTriangleQuadtree(o, ca, c, bc, t->children[2], depth + 1, maxDepth);
    subdivTriangleQuadtree(o, ca, bc, ab, t->children[3], depth + 1, maxDepth);
}

void createOctahedronSphere(int subdivision, std::vector<glm::vec3>& vertices, std::vector<Triangle*>& triangles)
{
    glm::vec3 top(0.f, 90.f, 0.f);
    glm::vec3 middle[5] = 
    {
        glm::vec3(360.f, 0.f, 0.f),
        glm::vec3(270.f, 0.f, 0.f),
        glm::vec3(180.f, 0.f, 0.f),
        glm::vec3(90.f, 0.f, 0.f),
        glm::vec3(0.f, 0.f, 0.f)
    };
    glm::vec3 bottom(0.f, -90.f, 0.f);
    
    glm::vec3 poleNorth[5] = 
    {
        glm::vec3(360.f, 90.f, 0.f),
        glm::vec3(270.f, 90.f, 0.f),
        glm::vec3(180.f, 90.f, 0.f),
        glm::vec3(90.f, 90.f, 0.f),
        glm::vec3(0.f, 90.f, 0.f)
    };
    
    glm::vec3 poleSouth[5] = 
    {
        glm::vec3(360.f, -90.f, 0.f),
        glm::vec3(270.f, -90.f, 0.f),
        glm::vec3(180.f, -90.f, 0.f),
        glm::vec3(90.f, -90.f, 0.f),
        glm::vec3(0.f, -90.f, 0.f)
    };
    
    // Add top vertex
    addVertex(top, vertices);

    // Compute number of slices in the Y axis.
    int subdivY = 1 << subdivision;
    std::size_t subdivYTmp = static_cast<std::size_t>(subdivY);
    
    // Generate vertices
    for (int i = 1; i <= subdivY; ++i)
    {
        totoLonLat(poleNorth, middle, i, subdivY, vertices);
    }
    
    for (int i = subdivY-1; i >= 1; --i)
    {
        totoLonLat(poleSouth, middle, i, subdivY, vertices);
    }
    
    addVertex(bottom, vertices);

    Octahedron o;
    
    o.trianglesY = subdivY * 2;
    o.pointsY = o.trianglesY + 1;
    
    // Differents offsets and sizes to navigate in the vertices array
    std::size_t pointsYSize = static_cast<std::size_t>(o.pointsY);
    std::size_t trianglesYSize = static_cast<std::size_t>(o.trianglesY);

    o.offsetPerSlice.resize(pointsYSize);
    o.vertexPerSlice.resize(pointsYSize);
    o.trianglePerSlice.resize(trianglesYSize);
    o.vertexPerTriangleSlice.resize(pointsYSize);
    
    int count = 0;
    
    for (std::size_t i = 0; i <= subdivYTmp; ++i)
    {
        o.vertexPerTriangleSlice[i] = static_cast<int>(i);
        o.vertexPerTriangleSlice[o.vertexPerSlice.size() - 1 - i] = o.vertexPerTriangleSlice[i];
        
        o.vertexPerSlice[i] = std::max(1, count * 4);
        o.vertexPerSlice[o.vertexPerSlice.size() - 1 - i] = o.vertexPerSlice[i];
        ++count;
    }
    
    o.offsetPerSlice[0] = 0;

    for (std::size_t i = 1; i < pointsYSize; ++i)
    {
        o.offsetPerSlice[i] = o.offsetPerSlice[i-1] + o.vertexPerSlice[i-1];
    }
    
    count = 0;
    
    for (std::size_t i = 0; i < subdivYTmp; ++i)
    {
        o.trianglePerSlice[i] = ((count * 2) + 1);
        o.trianglePerSlice[o.trianglePerSlice.size() - 1 - i] = o.trianglePerSlice[i];
        ++count;
    }
    
    Triangle* t;
    
    for (int k = 0; k < 4; ++k)
    {
        t = new Triangle;
        t->id = static_cast<std::uint16_t>(k);
        subdivTriangleQuadtree(&o, glm::ivec2(0, 0), glm::ivec2(0, o.pointsY / 2), glm::ivec2(o.pointsY / 2, o.pointsY / 2), t, 0, subdivision);
        
        triangles.push_back(t);
    }
    
    for (int k = 0; k < 4; ++k)
    {
        t = new Triangle;
        t->id = static_cast<std::uint16_t>(k);
        subdivTriangleQuadtree(&o, glm::ivec2(0, o.pointsY-1), glm::ivec2(o.pointsY / 2, o.pointsY / 2), glm::ivec2(0, o.pointsY / 2), t, 0, subdivision);
        
        triangles.push_back(t);
    }
}
