// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/spherical_math.h"

#include "3d/math_func.h"
#include "3d/camera.h"


float arcLength(const glm::vec3& p0, const glm::vec3& p1)
{
    return acos(glm::dot(p0, p1));
}

float arcLengthUnitSphere(const glm::vec3& p0, const glm::vec3& p1)
{
    return acos(glm::dot(glm::normalize(p0), glm::normalize(p1)));
}

float arcLengthOnEarth(const glm::vec3& p0, const glm::vec3& p1)
{
    const float earthRadius = 6371009.f;
    return arcLengthUnitSphere(p0, p1) * earthRadius;
}

bool arcIntersectUnitSphere(const glm::vec3& arc0p0, const glm::vec3& arc0p1, const glm::vec3& arc1p0, const glm::vec3& arc1p1, glm::vec3& /*intersectionPoint*/)
{
    // Compute the length of the two arcs
    float d1 = arcLengthUnitSphere(arc0p0, arc0p1);
    float d2 = arcLengthUnitSphere(arc1p0, arc1p1);

    // Compute the normal of each great circle
    glm::vec3 n0 = glm::cross(arc0p0, arc0p1);
    glm::vec3 n1 = glm::cross(arc1p0, arc1p1);

    // Compute the cross product of the two normals, it's a possible solution.
    glm::vec3 r0 = glm::cross(n0, n1);
    // Compute the opposite
    glm::vec3 r1 = -r0;

    // Compute all distances between all points.
    float r0a0p0 = arcLengthUnitSphere(r0, arc0p0);
    float r0a0p1 = arcLengthUnitSphere(r0, arc0p1);
    float r0a1p0 = arcLengthUnitSphere(r0, arc1p0);
    float r0a1p1 = arcLengthUnitSphere(r0, arc1p1);

    float r1a0p0 = arcLengthUnitSphere(r1, arc0p0);
    float r1a0p1 = arcLengthUnitSphere(r1, arc0p1);
    float r1a1p0 = arcLengthUnitSphere(r1, arc1p0);
    float r1a1p1 = arcLengthUnitSphere(r1, arc1p1);

    bool isInside = false;

    // Check the first solution
    if ((r0a0p0 < d1) && (r0a0p1 < d1) &&
        (r0a1p0 < d2) && (r0a1p1 < d2))
    {
        isInside = true;
    }
    else if ((r1a0p0 < d1) && (r1a0p1 < d1) &&
             (r1a1p0 < d2) && (r1a1p1 < d2))
    {
        // Check the second solution
        isInside = true;
    }

    return isInside;
}

bool isPointInInfiniteTetrahedron(const glm::vec3& p, const glm::vec3& apex, const glm::vec3& base0, const glm::vec3& base1, const glm::vec3& base2)
{
    glm::vec3 v0 = base0 - apex;
    glm::vec3 v1 = base1 - apex;
    glm::vec3 v2 = base2 - apex;

    glm::vec3 n0 = glm::cross(v0, v1);
    glm::vec3 n1 = glm::cross(v1, v2);
    glm::vec3 n2 = glm::cross(v2, v0);

    glm::vec3 vp = p - apex;

    return  (glm::dot(vp, n0) < 0.f) &&
            (glm::dot(vp, n1) < 0.f) &&
            (glm::dot(vp, n2) < 0.f);
}

bool isPointInInfinitePyramid(const glm::vec3& p, const glm::vec3& apex, const glm::vec3& base0, const glm::vec3& base1, const glm::vec3& base2, const glm::vec3& base3)
{
    glm::vec3 v0 = base0 - apex;
    glm::vec3 v1 = base1 - apex;
    glm::vec3 v2 = base2 - apex;
    glm::vec3 v3 = base3 - apex;

    glm::vec3 n0 = glm::cross(v0, v1);
    glm::vec3 n1 = glm::cross(v1, v2);
    glm::vec3 n2 = glm::cross(v2, v3);
    glm::vec3 n3 = glm::cross(v3, v0);

    glm::vec3 vp = p - apex;

    return  (glm::dot(vp, n0) < 0.f) &&
            (glm::dot(vp, n1) < 0.f) &&
            (glm::dot(vp, n2) < 0.f) &&
            (glm::dot(vp, n3) < 0.f);
}

bool isPointInInfiniteUnitPyramidCentered(const glm::vec3& p, const glm::vec3& v0, const glm::vec3& v1, const glm::vec3& v2, const glm::vec3& v3)
{
    glm::vec3 n0 = glm::cross(v0, v1);
    glm::vec3 n1 = glm::cross(v1, v2);
    glm::vec3 n2 = glm::cross(v2, v3);
    glm::vec3 n3 = glm::cross(v3, v0);

    return  (glm::dot(p, n0) < 0.f) &&
            (glm::dot(p, n1) < 0.f) &&
            (glm::dot(p, n2) < 0.f) &&
            (glm::dot(p, n3) < 0.f);
}

glm::vec3 slerp(const glm::vec3& p0, const glm::vec3 p1, float t)
{
    const float CLOSEST_TO_1_FROM_0 = std::nextafter(1.f, 0.f);
    const float CLOSEST_TO_MINUS_1_FROM_0 = std::nextafter(-1.f, 0.f);

    float dotp = glm::dot(glm::normalize(p0), glm::normalize(p1));

    if ((dotp > CLOSEST_TO_1_FROM_0) || (dotp < CLOSEST_TO_MINUS_1_FROM_0))
    {
        if
            (t <= 0.5f)
        {
            return p0;
        }
        return p1;
    }

    float theta = acos(dotp * 3.14159f/180.0f);
    return ((p0*sinf((1.f-t)*theta) + p1*sinf(t*theta)) / sinf(theta));
}

void subdiviseLine(const glm::vec3& p0, const glm::vec3 p1, double minAngleDegrees, std::vector<glm::vec3>& innerPoints)
{
    double angle = glm::degrees(AngleBetween(p0, p1));

    int subdivision = static_cast<int>(std::ceil(angle / minAngleDegrees));
    float subdivisionTmp = static_cast<float>(subdivision);
    float step = 1.f / subdivisionTmp;

    glm::vec3 p;

    for (int i = 1; i < subdivision; ++i)
    {
        float subIndex = static_cast<float>(i);
        p = glm::normalize(slerp(p0, p1, step * subIndex));

        innerPoints.push_back(p);
    }
}

void subdiviseLine2(const glm::vec3& p0, const glm::vec3 p1, float pixelRes, std::vector<glm::vec3>& innerPoints)
{
    float dist = arcLengthOnEarth(p0, p1);
    float pixelCountTmp(0.f);   //
    int pixelCount(0); //

    pixelCountTmp = dist / pixelRes;
    pixelCount = static_cast<int>(pixelCountTmp);

    float step = 1.f / pixelCountTmp;

    glm::vec3 p;

    for (int i = 1; i < pixelCount; ++i)
    {
        float subIndex = static_cast<float>(i);
        p = glm::normalize(slerp(p0, p1, step * subIndex));

        innerPoints.push_back(p);
    }
}

float equatorPixelResolution(const Camera* camera, const glm::ivec2& viewport)
{
    const glm::vec3 p0(0.999961913f, -0.00872653536f, 0.f);
    const glm::vec3 p1(0.999961913f, 0.00872653536f, 0.f);

    // Project points from world space to screen space.
    const glm::vec3 p2 = glm::project(p0, camera->getViewMatrix(), camera->getProjectionMatrix(), glm::ivec4(0, 0, viewport.x, viewport.y));
    const glm::vec3 p3 = glm::project(p1, camera->getViewMatrix(), camera->getProjectionMatrix(), glm::ivec4(0, 0, viewport.x, viewport.y));


    // One degree distance on the globe, in meters.
    // Perimeter at latitude 0 = 40 076 km
    const float oneDegreeDistanceEquator = 111322.222222f; // 40076000.f / 360.f

    // Compute the number of pixels between the two projected points.
    const float pixelDistance = glm::distance(p2, p3);

    // Compute the size of one pixel at the equator.
    const float distPerPixel = oneDegreeDistanceEquator / pixelDistance;


    // Return the size of one pixel at the equator in meter.
    return distPerPixel;
}

float equatorUnitPixelResolution(const Camera* camera, const glm::ivec2& viewport)
{
    const glm::vec3 p0(0.999961913f, -0.00872653536f, 0.f);
    const glm::vec3 p1(0.999961913f, 0.00872653536f, 0.f);

    // Project points from world space to screen space.
    const glm::vec3 p2 = glm::project(p0, camera->getViewMatrix(), camera->getProjectionMatrix(), glm::ivec4(0, 0, viewport.x, viewport.y));
    const glm::vec3 p3 = glm::project(p1, camera->getViewMatrix(), camera->getProjectionMatrix(), glm::ivec4(0, 0, viewport.x, viewport.y));

    // Compute the number of pixels between the two projected points.
    const float pixelDistance = glm::distance(p2, p3);

    float oneDegreeUnitSphere = static_cast<float>((2.0 * M_PI) / 360.0);
    const float distPerPixelUnit = oneDegreeUnitSphere / pixelDistance;

    // Return the size of one pixel at the equator in meter.
    return distPerPixelUnit;
}

/*!
 * \brief Center value angle in radians between [-PI, PI] range.
 */
float setToRangeMinusPiPlusPi(float rad)
{
    double sign = ((rad > 0.f) ? 1.0 : -1.0);
    double tmpVal = std::abs(static_cast<double>(rad));
    tmpVal = std::fmod(tmpVal + M_PI, 2.0 * M_PI) - M_PI;
    return static_cast<float>(sign * tmpVal);
}

/*!
 * \brief Signed angle from a to b in longitude
 */
float shortestLongitudeAngle(float a, float b)
{
    // Set a and b in the range [-Pi, Pi]
    a = setToRangeMinusPiPlusPi(a);
    b = setToRangeMinusPiPlusPi(b);

    // Compte difference
    double result = static_cast<double>(b - a);
    // Check difference sign
    double sign = (result > 0) ? 1.0 : -1.0;
    // Compte the absolute value
    result = std::abs(result);

    // If the difference is greater than PI it's not the shortest path
    if (result > M_PI)
    {
        result = (2.0 * M_PI) - result;
        sign *= -1.0;
    }

    // Apply the sign to the result
    return static_cast<float>(sign * result);
}

bool isTrigoOnGlobe(const glm::vec3& a, const glm::vec3& b, const glm::vec3& c)
{
	glm::vec3 ab, ac;

	ab = a - b;
	ac = a - c;

	// Compute triangle's normal
	glm::vec3 n = glm::normalize(glm::cross(ab, ac));

	// Compute the distance between the vertice A and the globe center
	float lenA = glm::length(a);

	// Compute the distance between the sum of A + N and the globe center
	float lenAN = glm::length(a + n);

	// Compare the distances, if lenAN is smaller than lenA, the normal is in the wrong direction
	// it's mean that the triangle is build with the wrong orientation.
	return lenA > lenAN;
}
