// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   static_mesh.cpp
 * Author: gleseach
 * 
 * Created on 4 septembre 2015, 10:21
 */

#include "3d/stdafx.h"
#include "3d/static_mesh.h"

#include "3d/vertex_array_object.h"
#include "3d/vertex_attrib_array.h"
#include "3d/buffer.h"
#include "3d/opengl_enum.h"
#include "3d/texture.h"
#include "3d/program_bind_helper.h"
#include "3d/context.h"
#include "3d/draw_mode.h"
#include "3d/meta.h"
#include "3d/helper.h"

namespace OGL
{

StaticMesh::StaticMesh()
{
    this->vao = nullptr;
    this->program = nullptr;
    this->_drawMode = nullptr;
}

StaticMesh::~StaticMesh()
{
    delete this->vao;
	this->vao = nullptr;

    // Textures might be shared with another mesh, so we cannot deallocate them
    // automatically. Make sure you do delete them before all meshes that use
    // them are destroyed.
    /*
    for
        (auto itTexture : textures)
    {
        delete itTexture.second;
        itTexture.second = nullptr;
    }
    */
    textures.clear();

    for (OGL::Buffer * buffer : buffers)
    {
        delete buffer;
		buffer = nullptr;
    }

    buffers.clear();

    delete this->_drawMode;
    this->_drawMode = nullptr;

    delete program;
    program = nullptr;
}

void StaticMesh::create(OGL::Context* context)
{
    this->_drawMode = context->createDrawMode();
}

void StaticMesh::create(OGL::Context* context, const std::vector<glm::vec3>& vertices)
{
    this->vao = context->createVertexArrayObject();
    this->_drawMode = context->createDrawMode();
    _drawMode->setDataType(DataType::unsigned_int);
    _drawMode->setDrawType(OGL::draw_arrays);
    _drawMode->setPrimitive(OGL::triangles);
    _drawMode->setElementCount(static_cast<unsigned int>(vertices.size()));

    Buffer* buffer = OGL::Helper::createBuffer(context, OGL::array_buffer, makeArray(vertices));

    buffers.push_back(buffer);

    VertexAttribArray * vaa = context->createVertexAttribArray(0, 3, buffer);

    this->vao->addObject(vaa);

    this->vao->create(context);
    this->vao->bind();
    vaa->bind();
    vaa->unbind();
    this->vao->unbind();
}

void StaticMesh::create(OGL::Context* context, const std::vector<glm::vec3>& vertices, const std::vector<glm::vec2>& texCoords)
{
    this->vao = context->createVertexArrayObject();

    this->_drawMode = context->createDrawMode();
    _drawMode->setDataType(DataType::unsigned_int);
    _drawMode->setDrawType(OGL::draw_arrays);
    _drawMode->setPrimitive(OGL::triangles);
    _drawMode->setElementCount(static_cast<unsigned int>(vertices.size()));

    Buffer* buffer = OGL::Helper::createBuffer(context, OGL::array_buffer, makeArray(vertices));
    buffers.push_back(buffer);

    VertexAttribArray * vaaVertices = context->createVertexAttribArray(0, 3, buffer);

    buffer = OGL::Helper::createBuffer(context, OGL::array_buffer, makeArray(texCoords));
    buffers.push_back(buffer);

    VertexAttribArray * vaaTexCoords = context->createVertexAttribArray(1, 2, buffer);

    this->vao->addObject(vaaVertices);
    this->vao->addObject(vaaTexCoords);

    this->vao->create(context);
    this->vao->bind();
    vaaVertices->bind();
    vaaTexCoords->bind();
    vaaVertices->unbind();
    vaaTexCoords->unbind();
    this->vao->unbind();
}

void StaticMesh::addBuffer(OGL::Buffer * buffer)
{
    buffers.push_back(buffer);
}

void StaticMesh::bindTextures(OGL::Context* context)
{
    for (auto it : this->textures)
    {
        context->activeTextureUnit(it.first, it.second);
    }
}

void StaticMesh::draw()
{
    this->vao->bind();

    _drawMode->draw();

    this->vao->unbind();
}

void StaticMesh::setVao(OGL::VertexArrayObject* vaoIn)
{
    this->vao = vaoIn;
}

void StaticMesh::setDrawDescription(OGL::DrawType draw_type, OGL::Primitive object_type, DataType dataType, unsigned int vertice_count)
{
    _drawMode->setPrimitive(object_type);
    _drawMode->setDrawType(draw_type);
    _drawMode->setElementCount(vertice_count);
    _drawMode->setDataType(dataType);
}

void StaticMesh::setTextureMap(const std::map<unsigned int, OGL::Texture*>& texturesIn)
{
    this->textures = texturesIn;
}

void StaticMesh::setProgramHelper(OGL::ProgramBindHelper * programIn)
{
    this->program = programIn;
}

} // namespace OGL
