// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/streamlines_data.h"

#include "3d/data_set.h"
#include "3d/core_objects.h"
#include "3d/program_bind_helper.h"
#include "3d/unit_sphere.h"
#include "3d/geodetic.h"
#include "3d/random.h"
#include "3d/spherical_math.h"
#include "3d/default_value.h"
#include "3d/gcp_utils.h"

StreamlinesData::StreamlinesData()
: _currentStreamline{0u},
  _particleCount{0u},
  _streamlineCount{0u},
  _streamlineCountTmp(0),
  _update{false},
  _streamlineLength{0u},
  _streamlineLengthTmp{0u},
  _streamlinePerSec{0.f},
  _speed{1.f},
  _indiceCount{0u},
  _vertices_local_id_buffer{nullptr},
  _indices_buffer{nullptr},
  _vbo{nullptr, nullptr},
  _vaoPhysic{nullptr, nullptr},
  _vaoRendering{nullptr, nullptr},
  _sourceId{0u},
  _targetId{1u},
  _gcp{nullptr},
  _gcpNorthEast{nullptr},
  _gcpAR(nullptr),
  _area{1.f},
  _framerate{60.f},
  _ttl{3.f},
  _fadeIn{1.f},
  _fadeOut{1.f},
  _streamlinesPerMs{0.f},
  _remaining(0.f),
  _streamlinesPerFrameTarget{0.f},
  _streamlinesGenerated{0u},
  _sb(nullptr)
{
}

StreamlinesData::~StreamlinesData()
{
	clearBuffers();

	delete _gcp;
    _gcp = nullptr;
	delete _gcpNorthEast;
    _gcpNorthEast = nullptr;
    delete _gcpAR;
    _gcpAR = nullptr;
    delete _sb;
    _sb = nullptr;
}

void StreamlinesData::create(OGL::Context* ctx, unsigned int streamlineCount, unsigned int streamlineLength, float speed, DataSet* dataSet)
{
    _streamlineCount = streamlineCount;
    _streamlineLength = streamlineLength;
    setSpeed(speed);
    setCurrentStreamline(0);

	_ttl = 3.0f; // sec
    _fadeIn = 1.f;
    _fadeOut = 1.f;
	_framerate = 60.f;
    _streamlinesPerMs = static_cast<float>(_streamlineCount) / static_cast<float>(getTTL() * 1000.f);

	setStreamlinesPerFrameTarget(std::ceil((static_cast<float>(streamlineCount) / getTTL()) / _framerate));

    float ttl = 5.f;

    _streamlinePerSec = static_cast<float>(getStreamlineCount()) / ttl;

    computeGcps(ctx, dataSet, _gcp, _gcpNorthEast, _gcpAR, _uvMin, _uvRange);

    setParticleCount(getStreamlineCount() * getStreamlineLength());

    _sb = new StreamlinesBuffer;

    initBuffers(ctx);

    const std::vector<glm::vec3>& gcps = dataSet->getGcps();
    const unsigned int lonSize = dataSet->getLon_size();

    const glm::vec3& p0 = gcps[lonSize + 1];
    const glm::vec3& p1 = gcps[lonSize + 2];
    const glm::vec3& p2 = gcps[lonSize * 2 + 1];

    double doubleTmp = static_cast<double>(glm::distance(p0, p1)) * GLOBE_RADIUS_EQUATOR_KM;
    float lonDist = static_cast<float>(doubleTmp);
    doubleTmp = static_cast<double>(glm::distance(p0, p2)) * GLOBE_RADIUS_EQUATOR_KM;
    float latDist = static_cast<float>(doubleTmp);

    setArea(lonDist * latDist);

    // Start all chronos
    _origin.start();
    getLastUpdateTime().start();
    getLastTime().start();

    _lastUpdate = std::chrono::system_clock::now();
}

void StreamlinesData::setStreamlineLength(unsigned int val)
{
	this->_streamlineLengthTmp = val;
	this->_update = true;
}

void StreamlinesData::setStreamlineCount(unsigned int val)
{
	this->_streamlineCountTmp = val;
	this->_update = true;
}

void StreamlinesData::swap()
{
	std::swap(this->_sourceId, this->_targetId);
}

void StreamlinesData::update(OGL::Context* ctx)
{
	if
		(false == this->_update)
	{
        return;
    }

    if
        (0 != this->_streamlineCountTmp &&
            this->_streamlineCount != this->_streamlineCountTmp)
    {
        this->_streamlineCount = this->_streamlineCountTmp;
        this->_streamlineCountTmp = 0;
    }

    if
        (0 != this->_streamlineLengthTmp &&
            this->_streamlineLength != this->_streamlineLengthTmp)
    {
        this->_streamlineLength = this->_streamlineLengthTmp;
        this->_streamlineLengthTmp = 0;
    }

    setParticleCount(this->_streamlineCount * this->_streamlineLength);
    setCurrentStreamline(0);

    // Destroy all buffers
    clearBuffers();

    // Create new buffers according to the new size
    initBuffers(ctx);

    this->_update = false;
}

void StreamlinesData::setRemaining(float f)
{
	_remaining = f;
}

void StreamlinesData::clearBuffers()
{
	for (int i = 0; i < 2; ++i)
	{
		delete _vbo[i];
        _vbo[i] = nullptr;
	}

	delete _vertices_local_id_buffer;
    _vertices_local_id_buffer = nullptr;

	delete _indices_buffer;
    _indices_buffer = nullptr;

	for (int i = 0; i < 2; ++i)
	{
		delete _vaoRendering[i];
        _vaoRendering[i] = nullptr;
	}

	for (int i = 0; i < 2; ++i)
	{
		delete _vaoPhysic[i];
        _vaoPhysic[i] = nullptr;
	}
}

void StreamlinesData::initBuffers(OGL::Context* ctx)
{
    for (int i = 0; i < 2; ++i)
    {
        _vbo[i] = OGL::Helper::createBuffer(ctx, OGL::array_buffer, Array(BufferDesc(getParticleCount(), 3, DataType::float32), nullptr));
    }

    glm::vec3* particlesPosition = new glm::vec3[getParticleCount()];

    glm::vec3 uv;

    _sb->init(_streamlineCount, 60, _fadeIn + _ttl + _fadeOut);

    auto& buf = _sb->getBuffer();

    for (const StreamlinesChunk& sc : buf)
    {
        size_t size = sc.tail - sc.head + 1;

        for (size_t i = 0; i < size; ++i)
        {
            uv = glm::vec3(randomOne(), randomOne(), sc.time);
            size_t id = sc.head + i;

            for (size_t j = 0; j < getStreamlineLength(); ++j)
            {
                particlesPosition[id * getStreamlineLength() + j] = uv;
            }
        }
    }

    //Fill the source vbo
    _vbo[_sourceId]->copy(particlesPosition);

    delete[] particlesPosition;
    particlesPosition = nullptr;

    std::vector<unsigned int> indices;
    indices.reserve(getParticleCount());
    std::vector<float> vertex_local_id(getParticleCount() + 1);

    unsigned int offset;

    for (unsigned int i = 0; i < getStreamlineCount(); ++i)
    {
        offset = i * getStreamlineLength();

        for (unsigned int j = 1; j < getStreamlineLength(); ++j)
        {
            indices.push_back(offset + j - 1);
            indices.push_back(offset + j);
        }

        for (unsigned int k = 0; k < getStreamlineLength(); ++k)
        {
            vertex_local_id[offset + k] = static_cast<float>(k);
        }
    }
    setIndiceCount(static_cast<unsigned int>(indices.size()));

    _vertices_local_id_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(vertex_local_id));
    _indices_buffer = OGL::Helper::createBuffer(ctx, OGL::element_array_buffer, makeArray(indices));

    unsigned int vertices_location = 0;

    // Rendering
    for (int i = 0; i < 2; ++i)
    {
        OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(vertices_location, 3, _vbo[i]);
        unsigned int localID = 1;

        // localId to compute alpha
        OGL::VertexAttribArray* vertices_attrib2 = ctx->createVertexAttribArray(localID, 1, _vertices_local_id_buffer);

        _vaoRendering[i] = ctx->createVertexArrayObject();
        _vaoRendering[i]->create(ctx);

        _vaoRendering[i]->addObject(vertices_attrib);
        _vaoRendering[i]->addObject(vertices_attrib2);
        _vaoRendering[i]->addObject(_indices_buffer);

        _vaoRendering[i]->fill();
    }

    // Physic
    for (int i = 0; i < 2; ++i)
    {
        // Longitude / latitude array
        OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(vertices_location, 3, _vbo[i]);

        // "Position in the streamline" array
        unsigned int localID = 2;
        OGL::VertexAttribArray* vertices_attrib2 = ctx->createVertexAttribArray(localID, 1, _vertices_local_id_buffer);

        // Longitude / latitude array with an offset. The current vertex knows the position of his neighbor.
        unsigned int neighbourId = 1;
        OGL::VertexAttribArray* vertices_attrib_next = ctx->createVertexAttribArray(neighbourId, 3, _vbo[i], DataType::float32, 0, sizeof(float) * 3);

        // Vao creation
        _vaoPhysic[i] = ctx->createVertexArrayObject();
        _vaoPhysic[i]->create(ctx);

        // Fill the vao, cpu side
        _vaoPhysic[i]->addObject(vertices_attrib);
        _vaoPhysic[i]->addObject(vertices_attrib_next);
        _vaoPhysic[i]->addObject(vertices_attrib2);
        _vaoPhysic[i]->addObject(_indices_buffer);

        // bind all objects to fill vao OpenGL side.
        _vaoPhysic[i]->fill();
    }

    setStreamlinesGenerated(0);

    // Start all chronos
    _origin.reset();
    getLastUpdateTime().reset();
    getLastTime().reset();
}
