// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/transect.h"

#include "3d/shape_mesh.h"
#include "3d/polyline_mesh.h"
#include "3d/polyline.h"
#include "3d/billboard.h"
#include "3d/opengl_enum.h"
#include "3d/core_objects.h"

#include "3d/camera.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "3d/iscene.h"
#include "3d/spherical_math.h"
#include "3d/iworld_visibility.h"
#include "3d/math_func.h"
#include "3d/ppm.h"
#include "3d/scene_view.h"
#include "3d/iscene_renderer.h"

#include "core/structs.h"
#include "core/log.h"

struct PointInfo
{
	bool globeVisible;
	bool viewportVisible;
	glm::ivec2 viewportCoord;
	glm::vec3 worldPos;
	Geodetic2D lonlat;
};

namespace
{
	PointInfo createPointInfo(IWorldVisibility* visibility, const glm::mat4& mvp, const glm::vec4& viewport, const glm::vec3& p)
	{
		PointInfo pi;

		pi.worldPos = p;
		pi.lonlat = UnitSphere::ToGeodetic2D(p);
		pi.globeVisible = visibility->isVisible(p);

        glm::vec3 pScreen = glm::project(p, glm::mat4(1.f), mvp, viewport);

        pi.viewportCoord = glm::ivec2(static_cast<int>(pScreen.x), static_cast<int>(pScreen.y));

        // Test if the gcp is in the viewport [0, width[, [0, height[
        pi.viewportVisible = (0 <= pScreen.x && pScreen.x < viewport.z && 0 <= pScreen.y && pScreen.y < viewport.w);

        return pi;
    }
}

Transect::Transect()
:   _fbo(nullptr)
{
}

Transect::~Transect()
{
    delete _fbo;
	_fbo = nullptr;
}

void Transect::create(OGL::Context* ctx, int width, int height)
{
    _fbo = OGL::Helper::createFramebuffer(ctx, static_cast<unsigned int>(width), static_cast<unsigned int>(height), DataType::float32);
}

template<typename T, typename N>
void pixelsModifier(Image& in, Image& out, int multiplier)
{
    T* pIn = static_cast<T*>(in.data());
    N* pOut = static_cast<N*>(out.data());

    for (unsigned int x = 0; x < in.width(); ++x)
    {
        for (unsigned int y = 0; y < in.height(); ++y)
        {
            for (unsigned int z = 0; z < in.channels(); ++z)
            {
                int id = static_cast<int>((y * in.width() + x) * in.channels() + z);

                pOut[id] = static_cast<N>(pIn[id] * multiplier);
            }
        }
    }
}

/**
 *
 */
bool Transect::getPoints(ShapeMesh* shapeMesh, std::vector<glm::vec3>& result) const
{
    PolylineMesh* polylineMesh(nullptr);

    polylineMesh = dynamic_cast<PolylineMesh*>(shapeMesh);
    if
        (nullptr != polylineMesh)
    {
        Polyline* polyline = polylineMesh->getPolyline();
        if
            (nullptr == polyline
            || 0 >= polyline->getPoints().size())
        {
            // Use polyline mesh points instead of polyline points
            for
                (const auto& point : polylineMesh->getPoints())
            {
                result.emplace_back(UnitSphere::ToVector3D(point));
            }
        }
        else
        {
            result = polylineMesh->getPolyline()->getPoints();
        }

        return true;
    }
    polylineMesh = nullptr;

    Billboard* billboard(nullptr);
    billboard = dynamic_cast<Billboard*>(shapeMesh);
    if
        (nullptr != billboard)
    {
        result.emplace_back(billboard->getPosition());

        return true;
    }

    return false;
}

bool Transect::extract(OGL::Context* ctx, ShapeMesh* shapeMesh, SceneView& sceneView, std::map<SEAScope::RenderableId, SEAScope::Extraction>& result, bool discardExtractedData)
{
	bool ok;	//
	ok = false;
    if (nullptr == shapeMesh)
    {
        return false;
    }

    // Create a renderableId for the mask object
    SEAScope::RenderableId rIdMask = shapeMesh->renderableId();

    // Search mask object
    auto itMask = result.find(rIdMask);

    if (result.end() == itMask)
    {
        // Not found create one
        result[rIdMask];
    }

    std::vector<glm::vec3> pointsBuffer;
    ok = this->getPoints(shapeMesh, pointsBuffer);
    if
        (false == ok)
    {
        LOG(SEAScope::LogLevel::Error, "Transect", "Could not get points from footprint");
        return false;
    }

    ok = renderToFbo(ctx, pointsBuffer, sceneView, result, rIdMask);
    if
        (false == ok)
    {
        LOG(SEAScope::LogLevel::Error, "Transect", "Extraction failed");
        return false;
    }

	////////////////////////////////////////////////////////////////////////////////////
	// Transect

	std::vector<PointInfo> points;

	// Get the object which compute the visibility for the current scene
	IWorldVisibility* visibility = sceneView.getVisibility();
	Camera* cam = sceneView.getCamera();
	glm::mat4 mvp = cam->getProjectionMatrix() * cam->getViewMatrix() * cam->getModelMatrix();
	const glm::ivec4 viewport = ctx->getViewportSize();

	// Compute visibility informations and screen coordinates for all points
    for (glm::vec3& p : _allPoints)
    {
        points.push_back(createPointInfo(visibility, mvp, viewport, p));
    }

    SEAScope::Extraction tmpExtraction;

	// Create GCP from all points
    for (std::size_t i = 0; i < pointsBuffer.size(); ++i)
    {
        auto gcp = createGCP(pointsBuffer[static_cast<unsigned int>(i)], mvp, viewport, ok);
        tmpExtraction.shape.push_back(gcp);
    }

	createMask(result, points, tmpExtraction, rIdMask);

	checkData(result, points, rIdMask);

	if (false == result.empty())
	{
        if
            (false == discardExtractedData)
        {
		    extractData(result, tmpExtraction, rIdMask, points);
        }
	}
	else
	{
		LOG(SEAScope::LogLevel::Warn, "Transect", "No data to extract");
		ok = false;
	}

    return ok;
}

void Transect::resize(int width, int height)
{
    if (_fbo)
    {
        _fbo->resize(static_cast<unsigned int>(width), static_cast<unsigned int>(height));
    }
}

SEAScope::GCP Transect::createGCP(const glm::vec3& point, const glm::mat4& mvp, const glm::ivec4& viewport, bool& ok) const
{
    Geodetic2D geo = UnitSphere::ToGeodetic2D(point);

    glm::vec3 pScreen = glm::project(point, glm::mat4(1.f), mvp, viewport);

    // Test if the gcp is in the viewport [0, width[, [0, height[
	ok = (0 <= pScreen.x && pScreen.x < static_cast<float>(viewport.z) && 0 <= pScreen.y && pScreen.y < static_cast<float>(viewport.w));

    return SEAScope::GCP(0u, /*placeholder for time*/
                         static_cast<float>(glm::degrees(geo.Latitude())),
                         static_cast<float>(glm::degrees(geo.Longitude())),
                         static_cast<uint32_t>(pScreen.x),
                         static_cast<uint32_t>(static_cast<float>(viewport.w) - pScreen.y));
}

/**
 *
 */
bool Transect::renderToFbo(OGL::Context* ctx, const std::vector<glm::vec3>& polyPoints, SceneView& sceneView, std::map<SEAScope::RenderableId, SEAScope::Extraction>& result, const SEAScope::RenderableId& rIdMask)
{
	ctx->setFlag(OGL::cull_face, true);

	_fbo->bind();

	const glm::ivec4 viewport = ctx->getViewportSize();

	ctx->setFlag(OGL::blending, true);
	ctx->setViewport(viewport);
	ctx->setFlag(OGL::depth_test, true);
	ctx->setDepthMask(true);
	ctx->clear(OGL::depth_buffer | OGL::stencil_buffer);
	ctx->setColorMask(true, true, true, true);
	ctx->setFlag(OGL::stencil_test, false);
	ctx->setDepthMask(false);

	std::vector<glm::vec3> innerPoints;

	_allPoints.clear();

	// Fill the array with all points
	_allPoints.push_back(polyPoints[0]);

	float pixelRes = sceneView.computePixelSize(glm::ivec2(viewport.z, viewport.w));

	for (unsigned int i = 0; i < polyPoints.size() - 1; ++i)
	{
		innerPoints.clear();

		subdiviseLine2(polyPoints[i], polyPoints[i + 1], pixelRes, innerPoints);

		std::copy(innerPoints.begin(), innerPoints.end(), std::back_inserter(_allPoints));

		_allPoints.push_back(polyPoints[i + 1]);
	}

	// Loop on all granules
	for (auto& it : result)
	{
		if (it.first == rIdMask)
			continue;

		switch
			(it.second.fillValues.size())
		{
		case 0:
			ctx->setClearColor(glm::vec4(0.f, 0.f, 0.f, 1.f));
			break;
		case 1:
			ctx->setClearColor(glm::vec4(it.second.fillValues[0],
				it.second.fillValues[0],
				it.second.fillValues[0],
				1.f));
			break;
		case 2:
			ctx->setClearColor(glm::vec4(it.second.fillValues[0],
				it.second.fillValues[1],
				it.second.fillValues[0],
				1.f));
			break;
		default:
			ctx->setClearColor(glm::vec4(it.second.fillValues[0],
				it.second.fillValues[1],
				it.second.fillValues[2],
				1.f));
		}

		ctx->clear(OGL::color_buffer);

		sceneView.getRenderer()->drawOneRaw(sceneView.getScene(),
                                            sceneView.getConstCamera(),
                                            it.first);

		Image* img = new Image;

		_extractedData[it.first] = img;

		_fbo->dumpToImage(*img);
	}

	ctx->setFlag(OGL::stencil_test, false);

	_fbo->unbind();

    return true;
}

void Transect::createMask(std::map<SEAScope::RenderableId, SEAScope::Extraction>& result, std::vector<PointInfo>& points, SEAScope::Extraction& tmpExtraction, SEAScope::RenderableId& rIdMask)
{
	// Create a mask for visible/hidden data
	SEAScope::Extraction& extractionMask = result[rIdMask];
	extractionMask.buffer.allocate(static_cast<unsigned int>(points.size()), 1, 4, DataType::float32, nullptr);
	extractionMask.shape = tmpExtraction.shape;
	extractionMask.uri = "mask";
	extractionMask.fillValues.push_back(-1.f);

	float* maskDataPtr = static_cast<float*>(extractionMask.buffer.data());

	SEAScope::GCP gcp(0u, 0.f, 0.f, 0u, 0u);

	// Generate a mask along the polyline
	for (int i = 0; i < static_cast<int>(points.size()); ++i)
	{
		PointInfo& pi = points[static_cast<std::size_t>(i)];
		maskDataPtr[i * 4] = (pi.viewportVisible && pi.globeVisible) ? 1.f : 0.f;
		maskDataPtr[i * 4 + 1] = maskDataPtr[i * 4];
		maskDataPtr[i * 4 + 2] = maskDataPtr[i * 4];
		maskDataPtr[i * 4 + 3] = maskDataPtr[i * 4];

		gcp.lon = static_cast<float>(glm::degrees(pi.lonlat.Longitude()));
		gcp.lat = static_cast<float>(glm::degrees(pi.lonlat.Latitude()));

		if (pi.viewportVisible && pi.globeVisible)
		{
			gcp.i = static_cast<std::uint32_t>(pi.viewportCoord.x);
			gcp.j = static_cast<std::uint32_t>(pi.viewportCoord.y);
		}
		else
		{
			gcp.i = 0;
			gcp.j = 0;
		}

		tmpExtraction.gcps.push_back(gcp);
	}

	extractionMask.gcps = tmpExtraction.gcps;
}

void Transect::checkData(std::map<SEAScope::RenderableId, SEAScope::Extraction>& result, std::vector<PointInfo>& points, SEAScope::RenderableId& rIdMask)
{
	// Check if there are data for all granules along the polyline.
	std::list<SEAScope::RenderableId> toRemove;

	for (auto& it : result)
	{
		if (it.first == rIdMask)
			continue;

		Image* img = _extractedData[it.first];
		glm::vec4* inpData = static_cast<glm::vec4*>(img->data());

		bool hasData = false;

		for (PointInfo& pi : points)
		{
			if (pi.viewportVisible && pi.globeVisible)
			{
				float r = inpData[(static_cast<unsigned int>(pi.viewportCoord.y) * img->width() + static_cast<unsigned int>(pi.viewportCoord.x))].r;

				if (false == SEAScope::floating_point_equality(r, it.second.fillValues[0]))
				{
					// One valid data found, stop the search
					hasData = true;
					break;
				}
			}
		}

		// Data found ?
		if (false == hasData)
		{
			// Nop, add to the remove list
			toRemove.push_back(it.first);
		}
	}

	// Loop on all granule to remove before extraction
	for (auto& rId : toRemove)
	{
		// Clear fbo clone
		Image* img = _extractedData[rId];
		delete img;
		img = nullptr;
		_extractedData.erase(rId);

		// remove the granule from the extraction list
		result.erase(rId);
	}
}

void Transect::extractData(std::map<SEAScope::RenderableId, SEAScope::Extraction>& result, SEAScope::Extraction& tmpExtraction, SEAScope::RenderableId& rIdMask, std::vector<PointInfo>& points)
{
	// Extraction loop for remaining granule(s)
	for (auto& it : result)
	{
		// Check if it's the mask, nothing to do
		if (it.first == rIdMask)
			continue;

		SEAScope::Extraction& ex = it.second;

		ex.buffer.allocate(static_cast<unsigned int>(points.size()), 1, 4, DataType::float32, nullptr);
		ex.shape = tmpExtraction.shape;
		ex.gcps = tmpExtraction.gcps;

		Image* img = _extractedData[it.first];
		glm::vec4* outData = static_cast<glm::vec4*>(ex.buffer.data());
		glm::vec4* inpData = static_cast<glm::vec4*>(img->data());

		glm::vec4 fillValues;
		for (std::size_t i = 0; i < it.second.fillValues.size(); ++i)
		{
			fillValues[static_cast<int>(i)] = it.second.fillValues[i];
		}

		int pixelX = 0;

		//int ptVisible = 0;
		//int ptNotVisible = 0;

		for (PointInfo& pi : points)
		{
			if (pi.viewportVisible && pi.globeVisible)
			{
				outData[pixelX] = inpData[(static_cast<unsigned int>(pi.viewportCoord.y) * img->width() + static_cast<unsigned int>(pi.viewportCoord.x))];
				//ptVisible++;

			}
			else
			{
				outData[pixelX] = fillValues;
				//ptNotVisible++;
			}

			++pixelX;
		}

		delete img;
		_extractedData.erase(it.first);
	}

	_extractedData.clear();
}

/**
 *
 */
OGL::FrameBuffer* Transect::getFBO() const
{
    return this->_fbo;
}
