// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/uv_streamlines.h"

#include "3d/core_objects.h"
#include "3d/program_bind_helper.h"
#include "3d/shader_manager.h"
#include "3d/streamlines_data.h"
#include "3d/data_set.h"
#include "3d/random.h"
#include "3d/static_mesh.h"
#include "3d/data_set.h"

#include "core/log.h"

UVStreamlines::UVStreamlines()
:   _physicProgram(nullptr),
    _drawPhysic(nullptr)
{
    _scaleFactorsLoc = 0;
    _offsetLoc = 0;
    _glFactorLoc = 0;
    _speedLoc = 0;
    _headIDLoc = 0;
    _uvTexelStepLoc = 0;
    _rangeMaxLoc = 0;
    _seamlessLoc = 0;
    _uvMinLoc = 0;
    _uvRangeLoc = 0;
}

UVStreamlines::~UVStreamlines()
{
    delete _physicProgram;
    _physicProgram = nullptr;
    delete _drawPhysic;
    _drawPhysic = nullptr;
}

void UVStreamlines::create(OGL::Context* ctx)
{
    OGL::ShaderManager* mngr = ctx->getShaderManager();

    _physicProgram = mngr->createProgramTransformFeedbackAutoBind(vertex_physic_uv_streamline, "position0");

    OGL::Program* program = _physicProgram->getProgram();

    _scaleFactorsLoc = program->getUniformLocation("scaleFactor");
    _offsetLoc = program->getUniformLocation("offset");
    _glFactorLoc = program->getUniformLocation("glFactor");
    _speedLoc = program->getUniformLocation("speed");
    _headIDLoc = program->getUniformLocation("headID");
    _uvTexelStepLoc = program->getUniformLocation("uvTexelStep");
    _rangeMaxLoc = program->getUniformLocation("rangeMax");
    _seamlessLoc = program->getUniformLocation("seamless");
    _uvMinLoc = program->getUniformLocation("uvMin");
    _uvRangeLoc = program->getUniformLocation("uvRange");

    // Physic
    _drawPhysic = ctx->createDrawMode();
    _drawPhysic->setPrimitive(OGL::points);
    _drawPhysic->setDrawType(OGL::draw_arrays);
    _drawPhysic->setElementCount(0);

    _particlesBuffer.resize(10000);
}

void UVStreamlines::update(OGL::Context* ctx, StreamlinesData* sd, DataSet* dataSet)
{
    sd->update(ctx);

    if
        (0 >= sd->getParticleCount())
    {
        return;
    }

    const float currentTime = sd->getOriginTime().elapsedf(second);

    StreamlinesChunk sc;
    bool res = sd->getSB()->getChunk(currentTime, sc);

    if
        (true == res)
    {
        auto& areas = dataSet->getVisibleArea();

        // Check if a part of the dataset is visible
        if (false == areas.empty())
        {
            const unsigned int streamlineToGenerate = static_cast<unsigned int>(sc.tail - sc.head + 1);

            unsigned int particleBufferSizeMin = streamlineToGenerate * sd->getStreamlineLength();

            if
                (_particlesBuffer.size() < particleBufferSizeMin)
            {
                _particlesBuffer.resize(particleBufferSizeMin);
            }

            // Compute the number of streamline per area
            unsigned int counter = 0;

            std::uint32_t const* const gcpx = dataSet->getGcpx();
            std::uint32_t const* const gcpy = dataSet->getGcpy();

            float width = static_cast<float>(dataSet->getWidth());
            float height = static_cast<float>(dataSet->getHeight());
            float tmp = 0.f;

            const unsigned int lonSize = dataSet->getLon_size();

            glm::vec2 gcpUVMin, gcpUVMax;
            glm::vec3 gcpUV;
            for (unsigned int areaId = 0; areaId < areas.size(); ++areaId)
            {
                const glm::ivec4& area = areas[areaId];

                unsigned int id0 = static_cast<unsigned int>(area.y) * lonSize + static_cast<unsigned int>(area.x);
                unsigned int id1 = static_cast<unsigned int>(area.w) * lonSize + static_cast<unsigned int>(area.z);

                unsigned int currentStreamlinesToGenerate = streamlineToGenerate / static_cast<unsigned int>(areas.size());

                tmp = static_cast<float>(gcpx[id0]);
                gcpUVMin.x = tmp / width;

                tmp = static_cast<float>(gcpy[id0]);
                gcpUVMin.y = tmp / height;

                tmp = static_cast<float>(gcpx[id1]);
                gcpUVMax.x = std::min(1.f, tmp / width);

                tmp = static_cast<float>(gcpy[id1]);
                gcpUVMax.y = std::min(1.f, tmp / height);

                for (unsigned int i = 0; i < currentStreamlinesToGenerate; ++i)
                {
                    gcpUV.x = gcpUVMin.x + (gcpUVMax.x - gcpUVMin.x) * randomOne();
                    gcpUV.y = gcpUVMin.y + (gcpUVMax.y - gcpUVMin.y) * randomOne();
                    gcpUV.z = currentTime;

                    for (unsigned int j = 0; j < sd->getStreamlineLength(); ++j)
                    {
                        _particlesBuffer[counter++] = gcpUV;
                    }
                }
            }

            constexpr unsigned int elementSize = sizeof(glm::vec3);

            // Copy the new data in the position buffer
            sd->getVboSrc()->copy(
                static_cast<unsigned int>(sc.head * sd->getStreamlineLength() * elementSize),
                static_cast<unsigned int>(streamlineToGenerate * sd->getStreamlineLength() * elementSize),
                _particlesBuffer.data());
        }
    }
}

void UVStreamlines::physicPass(OGL::Context* ctx, StreamlinesData* sd, DataSet* dataSet, float rangeMax)
{
    if
        (0 >= sd->getParticleCount())
    {
        return;
    }

    _drawPhysic->setElementCount(sd->getParticleCount());

    ctx->setFlag(OGL::blending, false);

    _physicProgram->bind(ctx);

    // physical data
    ctx->activeTextureUnit(0, dataSet->getMesh_spherical()->getTexture(0));
    dataSet->getMesh_spherical()->getTexture(0)->setFilterModeNoBind(OGL::linear);

    // mask
    ctx->activeTextureUnit(1, dataSet->getMesh_spherical()->getTexture(1));

    ctx->activeTextureUnit(2, sd->getGcpNorthEastTexture());

    ctx->activeTextureUnit(3, sd->getGcpAR());

    float elapsed_seconds = sd->getLastTime().elapsedf(second);

    sd->getLastTime().reset();

    auto p = _physicProgram->getProgram();

    float step = sd->getSpeed();
    float speed = elapsed_seconds * step;

    // Compute step between a texel in UV coordinates [0,1]
    float width = static_cast<float>(dataSet->getWidth());
    float height = static_cast<float>(dataSet->getHeight());

    glm::vec2 uvTexelStep = glm::vec2(1.f / width,
									  1.f / height);

    p->setUniform(_speedLoc, speed);
    p->setUniform(_headIDLoc, static_cast<float>(sd->getStreamlineLength() - 1));
    p->setUniform(_uvTexelStepLoc, uvTexelStep);
    p->setUniform(_rangeMaxLoc, rangeMax);
    p->setUniform(_seamlessLoc, dataSet->getSeamLess());

    glm::vec4 scaleFactors;
    glm::vec4 offsets;
    std::uint32_t channels;
    float glFactor;

    channels = static_cast<std::uint32_t>(dataSet->getDimensionCount());
    glFactor = dataSet->getScaleFactorGL();

    auto _scaleFactors = dataSet->getScaleFactor();
    auto _offsets = dataSet->getOffset();
    for
        (std::uint32_t i=0; i < channels; ++i)
    {
        scaleFactors[static_cast<int>(i)] = _scaleFactors[i];
        offsets[static_cast<int>(i)] = _offsets[i];
    }

    p->setUniform(_scaleFactorsLoc, scaleFactors);
    p->setUniform(_offsetLoc, offsets);
    p->setUniform(_glFactorLoc, glFactor);
    p->setUniform(_uvMinLoc, sd->getGcpUVMin());
    p->setUniform(_uvRangeLoc, sd->getGcpUVRange());

    ctx->setFlag(OGL::rasterizer_discard, true);

    sd->getVaoPhysicSrc()->bind();
    sd->getVboDst()->bindTo(0);

    _drawPhysic->drawTransformedFeedback();

    sd->getVaoPhysicSrc()->unbind();
    _physicProgram->unbind();

    ctx->setFlag(OGL::rasterizer_discard, false);
}

void UVStreamlines::swapVBO(StreamlinesData* sd)
{
    sd->swap();
}
