// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/uv_streamlines_color.h"


#include "3d/data_rendering_param.h"
#include "3d/core_objects.h"
#include "3d/program_bind_helper.h"
#include "3d/shader_manager.h"
#include "3d/streamlines_data.h"
#include "3d/data_set.h"
#include "3d/static_mesh.h"
#include "3d/uv_streamlines.h"

UVStreamlinesColor::UVStreamlinesColor()
: ColorRenderer(), _streamlines{nullptr}, _param{nullptr},
  _drawRender{nullptr}, _alphaLoc{-1}, _streamlineLengthLoc{-1}, _sizeLoc{-1},
  _texelSizeLoc{-1}, _ttlLoc(-1), _fadeOutLoc(-1), _fadeInLoc(-1),
  _currentTimeLoc(-1), _currentTime(0.f)
{
}

UVStreamlinesColor::~UVStreamlinesColor()
{
	delete _streamlines;
	_streamlines = nullptr;
	delete _drawRender;
	_drawRender = nullptr;
}

void UVStreamlinesColor::create(OGL::Context* ctx)
{
	OGL::ShaderManager* mngr = ctx->getShaderManager();

	_program = mngr->createProgramAutoBind(vertex_rendering_spherical_uv_streamline, fragment_color_alpha);

	OGL::Program* program = _program->getProgram();

	_mvpLoc = program->getUniformLocation("mvp");
	_alphaLoc = program->getUniformLocation("alpha0");
	_streamlineLengthLoc = program->getUniformLocation("streamlineLength");
	_colorLoc = program->getUniformLocation("fillColor");
	_sizeLoc = program->getUniformLocation("size");
	_texelSizeLoc = program->getUniformLocation("texelSize");
	_ttlLoc = program->getUniformLocation("ttl");
	_fadeOutLoc = program->getUniformLocation("fadeOut");
	_fadeInLoc = program->getUniformLocation("fadeIn");
	_currentTimeLoc = program->getUniformLocation("currentTime");

	_streamlines = new UVStreamlines;
	_streamlines->create(ctx);

	// Rendering
	_drawRender = ctx->createDrawMode();
	_drawRender->setPrimitive(OGL::lines);
	_drawRender->setDrawType(OGL::draw_elements);
	_drawRender->setDataType(DataType::unsigned_int);
	_drawRender->setElementCount(0);
}

void UVStreamlinesColor::bind(OGL::Context* ctx, const glm::mat4& mvp, DataRenderingParam* param)
{
	ctx->setDepthMask(false);

	_param = param;

	OGL::Program* p = _program->getProgram();

	p->bind();
	p->setUniform(_mvpLoc, mvp);
	p->unbind();
}

void UVStreamlinesColor::draw(OGL::Context* ctx, Renderable2* object)
{
	DataSet* dataSet = static_cast<DataSet*>(object);

	if (false == dataSet->getVisibleArea().empty())
	{
		StreamlinesData* sd = dataSet->getStreamlinesData();

		std::chrono::time_point<std::chrono::system_clock> now = std::chrono::system_clock::now();

		long long nsec = std::chrono::duration_cast<std::chrono::nanoseconds>(now - sd->_lastUpdate).count();
		_currentTime = sd->getOriginTime().elapsedf(TimeUnit::second);

		if (nsec >= 16666666) // 1 / 60 fps in nanoseconds
		{
            _streamlines->update(ctx, sd, dataSet);

            _streamlines->physicPass(ctx, sd, dataSet, _param->range.y);

            renderPass(ctx, sd, dataSet);

            _streamlines->swapVBO(sd);
		}
		else
		{
            renderPass(ctx, sd, dataSet);
		}
	}
}

void UVStreamlinesColor::renderPass(OGL::Context* ctx, StreamlinesData* sd, DataSet* dataSet)
{
	_drawRender->setElementCount(sd->getIndiceCount());

	ctx->setBlendFunc(OGL::src_alpha, OGL::one_minus_src_alpha);
	ctx->setFlag(OGL::blending, true);

	OGL::ProgramBindHelper* pbh = _program;

	pbh->bind(ctx);
	{
		auto p = pbh->getProgram();

		p->setUniform(_alphaLoc, _param->alpha);
		p->setUniform(_streamlineLengthLoc, static_cast<float>(sd->getStreamlineLength()) - 1.0f);
		p->setUniform(_colorLoc, _param->color);
		p->setUniform(_sizeLoc, glm::vec2(dataSet->getWidth() + 1, dataSet->getHeight() + 1));
		p->setUniform(_texelSizeLoc, glm::vec2(1.f / static_cast<float>((dataSet->getWidth() + 2)), 1.f / static_cast<float>((dataSet->getHeight() + 2))));
		p->setUniform(_ttlLoc, sd->getTTL());
		p->setUniform(_fadeInLoc, sd->getFadeIn());
		p->setUniform(_fadeOutLoc, sd->getFadeOut());
		p->setUniform(_currentTimeLoc, _currentTime);

		ctx->activeTextureUnit(0, sd->getGcpTexture());

		sd->getTargetVao()->bind();
		_drawRender->draw();
		sd->getTargetVao()->unbind();
	}
	pbh->unbind();

	ctx->setFlag(OGL::blending, false);
}
