// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/visibility_grid.h"

#include "3d/shader_list.h"
#include "3d/shader_manager.h"
#include "3d/mesh_creator.h"
#include "3d/program.h"
#include "3d/static_mesh.h"
#include "3d/context.h"
#include "3d/meta.h"

VisibilityGrid::VisibilityGrid()
: _visible{false}, _mesh{nullptr}, _program{nullptr}
{
}

void VisibilityGrid::enable(bool flag)
{
    _visible = flag;
}

void VisibilityGrid::create(OGL::Context* ctx, OGL::ShaderManager* shader_mngr, unsigned subdivision)
{
    float subdivisionTmp(0.0f); //
    float step(0.0f);   //
    float iTmp(0.0f);   //
    float jTmp(0.0f);   //

    subdivisionTmp = static_cast<float>(subdivision);

    step = 1.0f / (subdivisionTmp - 1.0f);

    std::vector<glm::vec2> vertices;

    for (unsigned int i = 0; i < subdivision; ++i)
    {
        iTmp = static_cast<float>(i);
        for (unsigned int j = 0; j < subdivision; ++j)
        {
            jTmp = static_cast<float>(j);
            vertices.push_back(glm::vec2(iTmp * step, jTmp * step));
        }
    }

    unsigned int id0, id1, id2, id3;
    IndicesArray indices;

    for (unsigned int lat = 0; lat < subdivision-1; ++lat)
    {
        for (unsigned int lon = 0; lon < subdivision-1; ++lon)
        {
            id0 = lat * subdivision + lon;
            id1 = lat * subdivision + lon + 1;
            id2 = (lat+1) * subdivision + lon + 1;
            id3 = (lat+1) * subdivision + lon;

            indices.push_back(id0);
            indices.push_back(id2);
            indices.push_back(id1);

            indices.push_back(id2);
            indices.push_back(id0);
            indices.push_back(id3);
        }
    }

    MeshCreator mc;
    mc.setIndices(makeArray(indices));
    mc.setDrawType(OGL::draw_elements, OGL::triangles);
    mc.setVertices(makeArray(vertices));
    _mesh = mc.createStaticMesh(ctx);

    _program = shader_mngr->createProgram(vertex_lonlat_globe_mvp, fragment_uni_color);

    setColor(glm::vec4(1.0f, 0.2f, 0.2f, 0.5));
}

void VisibilityGrid::setBBox(const glm::vec4& top, const glm::vec4& bottom)
{
    _program->bind();
    _program->setUniformByName("top", top);
    _program->setUniformByName("bottom", bottom);
    _program->unbind();
}

void VisibilityGrid::draw(OGL::Context* ctx, const glm::mat4& mvp)
{
    if (_visible)
    {
        ctx->setFlag(OGL::cull_face, false);

        _program->bind();
        _program->setUniformByName("mvp", mvp);

        _mesh->draw();

        _program->unbind();
    }
}

void VisibilityGrid::setColor(const glm::vec4& color)
{
    _program->bind();
    _program->setUniformByName("color", color);
    _program->unbind();
}

