// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/advanced_rendering_control.h"

using namespace SEAScope;

/**
 *
 */
AdvancedRenderingControl::AdvancedRenderingControl(): current(nullptr), view(nullptr), resetCallback(nullptr)
{
};

/**
 *
 */
AdvancedRenderingControl::~AdvancedRenderingControl()
{
};

/**
 *
 */
bool AdvancedRenderingControl::init(IAdvancedRenderingControlView* viewIn)
{
	this->view = viewIn;
	this->view->hide();

	this->view->setResetCallback([this] () {
		if
			(nullptr != this->resetCallback)
		{
			this->resetCallback(this->current);
		}
	});

	this->view->setSaveConfigCallback([this]() {
		if
			(nullptr != this->saveConfigCallback)
		{
			this->saveConfigCallback(this->current);
		}
		});

	this->view->setScaleChangedCallback([this](float min, float max, bool log) {
		if
			(nullptr == this->current)
		{
			return;
		}

		/* make sure log scale is not used with negative values */
		log = log && (0.f < min) && (min < max);

		this->current->setScale(min, max, log);
	});

	this->view->setColormapChangedCallback([this](const std::string& cmapLabel) {
		if
			(nullptr == this->current)
		{
			return;
		}
		this->current->setColormap(cmapLabel);
	});

	this->view->setColorChangedCallback([this](const std::array<std::uint8_t, 3>& color) {
		if
			(nullptr == this->current)
		{
			return;
		}
		this->current->setColor(color);
	});

    this->view->setFilterModeCallback([this](const std::string& mode) {
        if
            (nullptr == this->current)
        {
            return;
        }
        if
            ("nearest" == mode)
        {
            this->current->setFilterMode(FilterMode::NEAREST);
        }
        else if
            ("bilinear" == mode)
        {
            this->current->setFilterMode(FilterMode::BILINEAR);
        }
    });

    this->view->setStreamlinesCallback([this](uint32_t count, uint32_t length, float speed) {
        if
            (nullptr == this->current)
        {
            return;
        }

        this->current->setParticlesCount(count);
        this->current->setStreamlineLength(length);
        this->current->setStreamlineSpeed(speed);
    });

	this->view->setBillboardsSizeCallback([this](float size) {
		if
			(nullptr == this->current)
		{
			return;
		}

		this->current->setBillboardsSize(size);
	});

	this->view->setBillboardsDensityCallback([this](float value) {
		if
			(nullptr == this->current)
		{
			return;
		}

		this->current->setBillboardsDensity(value);
	});

    this->view->setMarkerChangedCallback([this](const std::string& name)
		{
            if
				(nullptr == this->current)
            {
                return;
            }

			 this->current->setCurrentTimeTrajectoryIconName(name);
        });

    this->view->setLineThicknessCallback([this](float value) {
        if
            (nullptr == this->current)
        {
            return;
        }

        this->current->setLineThickness(value);
    });

	return true;
};

/**
 *
 */
bool AdvancedRenderingControl::setTarget(RenderingCfg* cfg, const std::string& label)
{
	LOG( LogLevel::Debug, "AdvancedRenderingControl"
			, "setTarget: hide control? {}"
			, cfg == this->current);

	if
		(cfg == this->current)
	{
		/* Toggle view */
		this->view->hide();
		this->current = nullptr;
		return false;
	}

	this->current = cfg;
	this->view->update(cfg, label);
	this->view->show();

	return true;
};

/**
 *
 */
bool AdvancedRenderingControl::update()
{
	if
		(nullptr != this->current)
	{
		this->view->update(this->current);
	}
	return true;
};
