// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/api_out.h"
#include "core/log.h"
#include "core/granule_metadata.h"

using namespace SEAScope;

flatbuffers::Offset<API::Variable> APIOut::variable( flatbuffers::FlatBufferBuilder& builder
									 , const VariableId& vId
									 , const VariableCfg& input
									 , bool& ok)
{
	auto _label = builder.CreateString(input.label);
	auto _units = builder.CreateString(input.units);
	std::vector<flatbuffers::Offset<flatbuffers::String> > _fields;
	for
		(const auto& itField : input.fields)
	{
		_fields.push_back(builder.CreateString(itField));
	}

	std::vector<flatbuffers::Offset<SEAScope::API::Tag> > tagsVector;
	for
		(const auto& itTag : input.tags)
	{
		auto tagName = builder.CreateString(itTag.first);
		auto tagValue = builder.CreateString(itTag.second);
		auto tag = SEAScope::API::CreateTag( builder
										   , tagName
										   , tagValue);
		tagsVector.push_back(tag);
	}

    // Note: defaultRendering is set to RenderingMethod::RASTER to avoid a
    // breaking change in the flatbuffers schema, but this field is actually
    // deprecated and ignored
	auto variable = SEAScope::API::CreateVariable( builder
												 , vId
												 , _label
												 , _units
												 , builder.CreateVector(_fields)
												 , builder.CreateVector(tagsVector)
												 , API::RenderingMethod_RASTER);
	ok = true;
	return variable;
};

/**
 *
 */
flatbuffers::Offset<API::Collection> APIOut::collection( flatbuffers::FlatBufferBuilder& builder
													   , const CollectionId& colId
													   , const CollectionCfg& colCfg
													   , bool& ok)
{
	SEAScope::API::RenderableId _rId(false, colId.first, colId.second, 0, 0);

	auto _label = builder.CreateString(colCfg.label);

	std::vector<flatbuffers::Offset<SEAScope::API::Tag> > tagsVector;
	for
		(const auto& itTag : colCfg.tags)
	{
		auto tagName = builder.CreateString(itTag.first);
		auto tagValue = builder.CreateString(itTag.second);
		auto tag = SEAScope::API::CreateTag( builder
										   , tagName
										   , tagValue);
		tagsVector.push_back(tag);
	}

	auto tags = builder.CreateVector(tagsVector);

	std::vector<flatbuffers::Offset<SEAScope::API::Variable> > variablesVector;
	for
		(const auto& it: colCfg.variables)
	{
		auto variable = APIOut::variable(builder, it.first, it.second, ok);
		if
			(false == ok)
		{
			return 0;
		}
		variablesVector.push_back(variable);
	}
	auto variables = builder.CreateVector(variablesVector);

	auto collection = SEAScope::API::CreateCollection( builder
													 , &_rId
													 , colCfg.mustBeCurrent
													 , colCfg.xSeamless
													 , colCfg.ySeamless
													 , false /*NEWSAligned - deprecated */
													 , _label
													 , tags
													 , variables
													 , 0 /*defaultVariable - deprecated*/);
	ok = true;
	return collection;
};

/**
 *
 */
flatbuffers::Offset<API::Source> APIOut::source( flatbuffers::FlatBufferBuilder& builder
											   , std::uint32_t srcId
											   , const SourceInfo& srcInfo
											   , bool& ok)
{
	auto _label = builder.CreateString(srcInfo.title);
	auto _descr = builder.CreateString(srcInfo.description);

	std::vector<flatbuffers::Offset<SEAScope::API::Tag> > tagsVector;
	for
		(const auto& itTag : srcInfo.tags)
	{
		auto tagName = builder.CreateString(itTag.first);
		auto tagValue = builder.CreateString(itTag.second);
		auto tag = SEAScope::API::CreateTag( builder
										   , tagName
										   , tagValue);
		tagsVector.push_back(tag);
	}

	auto tags = builder.CreateVector(tagsVector);
	auto src = SEAScope::API::CreateSource( builder
										  , srcId
										  , _label
										  , _descr
										  , tags);
	ok = true;
	return src;
};

/**
 *
 */
flatbuffers::Offset<API::Extraction> APIOut::extraction( flatbuffers::FlatBufferBuilder& builder
													   , const RenderableId& rId
													   , const Extraction& input
													   , bool& ok)
{
	std::uint64_t xArity;	//
	std::uint64_t yArity;	//
	std::uint64_t channels;	//

	xArity = input.buffer.width();
	yArity = input.buffer.height();
	channels = input.buffer.channels();

    LOG(LogLevel::Debug, "APIout", "meta extract: {}x{}x{}", xArity, yArity, channels);

	/* Build shape */
	std::vector<API::GCP> shapeVector;	//
	for
		(const auto& shapeElement : input.shape)
	{
		auto g = API::GCP(shapeElement.time, shapeElement.lon, shapeElement.lat,
						  shapeElement.i, shapeElement.j);
		shapeVector.push_back(g);
	}
	SEAScope::API::GCP* _gcpsShape = shapeVector.data();
	auto shape = builder.CreateVectorOfStructs(_gcpsShape, shapeVector.size());

    LOG(LogLevel::Debug, "APIout", "Shape built");

	/* Build internal GCPs */
	std::vector<SEAScope::API::GCP> gcpsVector;	//
	for
		(const auto& gcpElement : input.gcps)
	{
		auto g = SEAScope::API::GCP(gcpElement.time, gcpElement.lon, gcpElement.lat,
								    gcpElement.i, gcpElement.j);
		gcpsVector.push_back(g);
	}
	SEAScope::API::GCP* _gcps = gcpsVector.data();
	auto gcps = builder.CreateVectorOfStructs(_gcps, gcpsVector.size());

    LOG(LogLevel::Debug, "APIout", "GCPs built");

	SEAScope::API::RenderableId _rId( 0 < rId.granule.second, rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);

    LOG(LogLevel::Debug, "APIout", "rId built");

	auto uri = builder.CreateString(input.uri);

    LOG(LogLevel::Debug, "APIout", "URI built");
	LOG(LogLevel::Debug, "APIout", "Data size / type size: {} =?= Values count: {}", (input.buffer.dataSize() / sizeof(float)), input.buffer.valuesCount());
	LOG(LogLevel::Debug, "APIout", "Valid buffer {} [{}]", input.buffer.isValid(), input.buffer.dataSize());
	LOG(LogLevel::Debug, "APIout", "Time range: [{}, {}]", input.start, input.stop);
	float* data = reinterpret_cast<float*>(input.buffer.data());

	auto fieldNames = builder.CreateVectorOfStrings(input.fields);
	auto extraction = SEAScope::API::CreateExtraction( builder
		, &_rId
		, static_cast<std::uint32_t>(xArity)
		, static_cast<std::uint32_t>(yArity)
		, static_cast<std::uint32_t>(channels)
		, builder.CreateVector(data, xArity * yArity * channels)
		, shape
		, gcps
		, builder.CreateVector(input.fillValues)
		, uri
		, fieldNames
		, input.start
		, input.stop
	);

    LOG(LogLevel::Debug, "APIout", "Extraction built");
	data = nullptr;

	ok = true;
	return extraction;
};

/**
 *
 */
flatbuffers::Offset<API::GranuleMetadata> APIOut::granuleMetadata( flatbuffers::FlatBufferBuilder& builder
																 , const GranuleMetadata& input
																 , bool& ok)
{
	std::vector<flatbuffers::Offset<SEAScope::API::IDFDescriptor> > uris;
	for
		(std::uint32_t i = 0; i < input.viewsCount(); ++i)
	{
        IDFDescriptor idf;
        input.getView(i, idf);

		std::vector<SEAScope::API::GCP> gcpsVector;	//
        const std::size_t gcpsCount {idf.shape.gcpsLat.size()};
		for
			(std::size_t index = 0u; index < gcpsCount; ++index)
		{
			auto g = SEAScope::API::GCP(idf.shape.gcpsTime[index],
                                        idf.shape.gcpsLon[index],
                                        idf.shape.gcpsLat[index],
                                        idf.shape.gcpsMinorAxisIndex[index],
                                        idf.shape.gcpsMajorAxisIndex[index]);
			gcpsVector.push_back(g);
		}

		SEAScope::API::GCP* _gcps = gcpsVector.data();
		auto gcps = builder.CreateVectorOfStructs(_gcps, gcpsVector.size());
		auto shape = SEAScope::API::CreateShape(builder, idf.shape.xArity, idf.shape.yArity, gcps);
		auto uri = builder.CreateString(idf.uri);
		auto idfDescriptor = SEAScope::API::CreateIDFDescriptor(builder, uri, idf.resolution, idf.xArity, idf.yArity, shape, static_cast<std::uint32_t>(idf.subsamplingFactor));
		uris.push_back(idfDescriptor);
	}

	SEAScope::API::DataModel dataModel;	//
	dataModel = SEAScope::API::DataModel_LAT_LON;	// default value
    DataModel _dataModel = input.dataModel();
	if
		(DataModel::LAT_LON == _dataModel)
	{
		dataModel = SEAScope::API::DataModel_LAT_LON;
	}
	else if
		(DataModel::Y_X == _dataModel)
	{
		dataModel = SEAScope::API::DataModel_Y_X;
	}
	else if
		(DataModel::NJ_NI == _dataModel)
	{
		dataModel = SEAScope::API::DataModel_NJ_NI;
	}
	else if
		(DataModel::ROW_CELL == _dataModel)
	{
		dataModel = SEAScope::API::DataModel_ROW_CELL;
	}
	else if
		(DataModel::TIME_STATION == _dataModel)
	{
		dataModel = SEAScope::API::DataModel_TIME_STATION;
	}
	else if
		(DataModel::TIME == _dataModel)
	{
		dataModel = SEAScope::API::DataModel_TIME;
	}

	auto urisVector = builder.CreateVector(uris);
	auto dataId = builder.CreateString(input.dataId());
	auto granule = SEAScope::API::CreateGranuleMetadata( builder
		, input.sourceId()
		, input.collectionId()
		, input.granuleId()
		, dataId
		, dataModel
		, input.timeCoverageStart()
		, input.timeCoverageEnd()
		, urisVector
        , input.has(IDFAttributes::TITLE)
        , input.has(IDFAttributes::INSTITUTION)
        , input.has(IDFAttributes::COMMENT)
        , input.has(IDFAttributes::FILE_ID)
        , input.has(IDFAttributes::PRODUCT_VERSION)
        , input.has(IDFAttributes::LAT_MIN)
        , input.has(IDFAttributes::LAT_MAX)
        , input.has(IDFAttributes::LON_MIN)
        , input.has(IDFAttributes::LON_MAX)
        , input.has(IDFAttributes::CREATOR_EMAIL)
        , input.has(IDFAttributes::STATION_ID)
        , input.has(IDFAttributes::PLATFORM)
        , input.has(IDFAttributes::SENSOR)
        , builder.CreateString(input.title())
        , builder.CreateString(input.institution())
        , builder.CreateString(input.comment())
        , builder.CreateString(input.fileId())
        , builder.CreateString(input.productVersion())
        , input.latMin()
        , input.latMax()
        , input.lonMin()
        , input.lonMax()
        , builder.CreateString(input.creatorEmail())
        , builder.CreateString(input.stationId())
        , builder.CreateString(input.platform())
        , builder.CreateString(input.sensor())
    );

	ok = true;
	return granule;
};

/**
 *
 */
flatbuffers::Offset<API::RenderingCfg> APIOut::renderingCfg( flatbuffers::FlatBufferBuilder& builder, const RenderingCfg& input, bool& ok)
{
	SEAScope::API::RenderableId _rId( 0 < input.target.granule.second,
                                    input.target.collection.first,
                                    input.target.collection.second,
                                    input.target.granule.second,
                                    input.target.variable);
    API::Color _color(input.color()[0], input.color()[1], input.color()[2]);
	API::RenderingMethod _rMethod;
    API::FilterMode _filterMode;

	_rMethod = SEAScope::API::RenderingMethod_RASTER;
	if
		(SEAScope::RenderMethod::RASTER == input.renderMethod())
	{
		_rMethod = API::RenderingMethod_RASTER;
	}
	else if
		(RenderMethod::ARROWS == input.renderMethod())
	{
		_rMethod = API::RenderingMethod_ARROWS;
	}
	else if
		(RenderMethod::BARBS == input.renderMethod())
	{
		_rMethod = API::RenderingMethod_BARBS;
	}
	else if
		(RenderMethod::TRAJECTORIES == input.renderMethod())
	{
		_rMethod = API::RenderingMethod_TRAJECTORIES;
	}
	else if
		(RenderMethod::DOTSCLOUD == input.renderMethod())
	{
		_rMethod = API::RenderingMethod_DOTSCLOUD;
	}
	else if
		(RenderMethod::STREAMLINES == input.renderMethod())
	{
		_rMethod = API::RenderingMethod_STREAMLINES;
	}
	else if
		(RenderMethod::RAWRGB == input.renderMethod())
	{
		_rMethod = API::RenderingMethod_RAWRGB;
	}

    _filterMode = API::FilterMode_NEAREST;
    if
        (FilterMode::NEAREST == input.filterMode())
    {
        _filterMode = API::FilterMode_NEAREST;
    }
    else if
        (FilterMode::BILINEAR == input.filterMode())
    {
        _filterMode = API::FilterMode_BILINEAR;
    }

    auto rcfg = SEAScope::API::CreateRenderingCfg( builder,
        input.rendered(),
        input.logscale(),
        input.min(),
        input.max(),
        input.opacity(),
        input.zindex(),
        &_color,
        builder.CreateString(input.colormap()),
        _rMethod,
        input.particlesCount(),
        input.particleTTL(),
        input.streamlineLength(),
        input.streamlineSpeed(),
        &_rId,
        _filterMode,
        input.billboardsSize(),
        input.billboardsDensity(),
        input.lineThickness());

    ok = true;
    return rcfg;
};

/**
 *
 */
flatbuffers::Offset<API::Timestep> APIOut::timestep( flatbuffers::FlatBufferBuilder& builder
											       , const std::string& label
											       , const timestamp_t value
											       , bool& ok)
{
	auto _label = builder.CreateString(label);
	auto result = SEAScope::API::CreateTimestep(builder, _label, value);
	ok = true;
	return result;
};

/**
 *
 */
flatbuffers::Offset<API::Timespan> APIOut::timespan( flatbuffers::FlatBufferBuilder& builder
											       , const std::string& label
											       , const timestamp_t pastOffset
                                                   , const timestamp_t futureOffset
											       , bool& ok)
{
	auto _label = builder.CreateString(label);
	auto result = SEAScope::API::CreateTimespan(builder, _label, pastOffset, futureOffset);
	ok = true;
	return result;
};
