// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <array>
#include <sstream>

#include "core/application.h"
#include "core/path.h"
#include "core/log.h"

#include "core/interfaces/iprotocol_handler.h"
#include "core/interfaces/iuser_interface.h"
#include "core/control_protocol_handler.h"

#include "3d/data_set.h"
#include "3d/datasets_allocator.h"
#include "3d/scene_view.h"
#include "3d/iscene.h"
#include "3d/iscene_renderer.h"
#include "3d/scene_renderer.h"
#include "3d/icons.h"
#include "core/icon_loader.h"

#include "core/app_events_dispatcher.h"

using namespace SEAScope;

/**
 * Constructor.
 */
Application::Application(AppEventsDispatcher& appEventsDispatcher, IGranulesIndex& granulesIndex, IAnnotationsStore& annotationsStore, IGranulesIndex& annotationsIndex, IUserInterface& ui)
: settingsStore(nullptr),
  stateStore(nullptr),
  window(nullptr),
  engine(appEventsDispatcher, granulesIndex, annotationsStore, annotationsIndex),
  _ui{ui}
{
    this->_grunt.engine = &(this->engine);
};

/**
 * Destructor.
 */
Application::~Application()
{
};

/**
 *
 */
bool Application::init( const std::string& installDir
					  , const std::string& configPath
                      , const std::string& pluginsPath
					  , ISettingsStore& settingsStoreIn
					  , IStateStore& stateStoreIn
					  , IWindow& windowIn
					  , RendererResources& resources
					  , const std::string& logFilter)
{
	bool ok;	//

    this->settingsStore = &settingsStoreIn;
	this->stateStore = &stateStoreIn;
	this->window = &windowIn;

	ok = this->loadSettings(configPath, logFilter);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Application"
				, "Could not load settings from {}", this->settingsPath);
		return false;
	}

    if
        (0 < pluginsPath.size())
    {
        this->engine.settings.pluginsPath = pluginsPath;
    }

	ok = this->initWorkspace(installDir);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Application", "Could not initialize workspace");
		return false;
	}

	ok = this->initWorkers();
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Application"
				, "Could not initialize workers");
		return false;
	}

	ok = this->initRendererResources(installDir, resources);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Application"
				, "Could not initialize renderer resources");
		return false;
	}

    // initRendererResources may have computed a path for the world map, update
    // settings accordingly
    this->engine.settings.worldMap = resources.worldmapPath;

	ok = this->initWindow();
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Application"
				, "Could not initialize window");
		return false;
	}

    ok = this->initUI(installDir, resources);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Application", "Could not initialize User Interface");
        return false;
    }

	return ok;
};

/**
 *
 */
bool Application::initRendererResources(const std::string& installDir,
										RendererResources& resources) const
{
	std::vector<std::string> shadersDirElems;	//
	std::vector<std::string> windBarbsDirElems;	//
	std::vector<std::string> arrowsDirElems;	//
	std::vector<std::string> worldmapPathElems;	//

	shadersDirElems.push_back(installDir);
	shadersDirElems.push_back(std::string("GLSL"));
	shadersDirElems.push_back(std::string("")); // required to keep the trailing slash
	Path::joinPaths(shadersDirElems, resources.shadersDir);

	windBarbsDirElems.push_back(installDir);
	windBarbsDirElems.push_back(std::string("images"));
	windBarbsDirElems.push_back(std::string("windBarbs"));
	windBarbsDirElems.push_back(std::string("pgm"));
	windBarbsDirElems.push_back(std::string("")); // required to keep the trailing slash
	Path::joinPaths(windBarbsDirElems, resources.windBarbsDir);

	arrowsDirElems.push_back(installDir);
	arrowsDirElems.push_back(std::string("images"));
	arrowsDirElems.push_back(std::string("arrows"));
	arrowsDirElems.push_back(std::string("pgm"));
	arrowsDirElems.push_back(std::string("")); // required to keep the trailing slash
	Path::joinPaths(arrowsDirElems, resources.arrowsDir);

	if
		("highres" == this->engine.settings.worldMap)
	{
		worldmapPathElems.push_back(installDir);
		worldmapPathElems.push_back(std::string("images"));
		worldmapPathElems.push_back(std::string("earth_highres.png"));
		Path::joinPaths(worldmapPathElems, resources.worldmapPath);
	}
	else if
		("lowres" == this->engine.settings.worldMap)
	{
		worldmapPathElems.push_back(installDir);
		worldmapPathElems.push_back(std::string("images"));
		worldmapPathElems.push_back(std::string("earth_lowres.png"));
		Path::joinPaths(worldmapPathElems, resources.worldmapPath);
	}
	else if
		("black" == this->engine.settings.worldMap)
	{
		worldmapPathElems.push_back(installDir);
		worldmapPathElems.push_back(std::string("images"));
		worldmapPathElems.push_back(std::string("black.png"));
		Path::joinPaths(worldmapPathElems, resources.worldmapPath);
	}
	else
	{
		resources.worldmapPath = this->engine.settings.worldMap;
	}

	LOG(LogLevel::Debug, "Application", "SHADERS DIR: {}~", resources.shadersDir);
	LOG(LogLevel::Debug, "Application", "ARROWS DIR: {}#", resources.arrowsDir);
	LOG(LogLevel::Debug, "Application", "WIND BARBS DIR: {}#", resources.windBarbsDir);
	return true;
};

/**
 *
 */
bool Application::loadSettings(const std::string& configPath, const std::string& logFilter)
{
	bool ok;

	this->settingsPath = configPath;
	ok = this->settingsStore->load(this->engine.settings, this->settingsPath);

	this->engine.settings.logFilter = logFilter;

	return ok;
};

/**
 *
 */
bool Application::initWorkspace(const std::string& installDir)
{
	bool ok;
	std::vector<std::string> pathElems;
	std::array<std::string, 6> appDirs;

	/* Set default values for paths */
	if
		(0 >= this->engine.settings.pluginsPath.length())
	{
		Path::joinPaths({installDir, "plugins"},
						this->engine.settings.pluginsPath);
	}

	if
		(0 >= this->engine.settings.colormapsPath.length())
	{
		Path::joinPaths({installDir, "colormaps"},
						this->engine.settings.colormapsPath);
	}

	if
		(0 >= this->engine.settings.operatorsPath.length())
	{
		Path::joinPaths({installDir, "operators"},
						this->engine.settings.operatorsPath);
	}

    if
        (0 >= this->engine.settings.dataPath.length())
    {
        Path::joinPaths({ installDir, "data" },
            this->engine.settings.dataPath);
    }

    if
        (0 >= this->engine.settings.customPath.length())
    {
        Path::joinPaths({ installDir, "custom" },
            this->engine.settings.customPath);
    }


	/* Normalize paths to avoid issues on some platforms */
	Path::setPathValid(this->engine.settings.pluginsPath);
	Path::setPathValid(this->engine.settings.colormapsPath);
	Path::setPathValid(this->engine.settings.operatorsPath);
	Path::setPathValid(this->engine.settings.dataPath);
	Path::setPathValid(this->engine.settings.indexPath);
	Path::setPathValid(this->engine.settings.statePath);
    Path::setPathValid(this->engine.settings.annotationsPath);
    Path::setPathValid(this->engine.settings.customPath);

	/* Create application directories if need be */
	appDirs[0] = this->engine.settings.pluginsPath;
	appDirs[1] = this->engine.settings.colormapsPath;
	appDirs[2] = this->engine.settings.operatorsPath;
	appDirs[3] = this->engine.settings.dataPath;
	appDirs[4] = this->engine.settings.customPath;
	Path::joinPaths({ this->engine.settings.customPath, "icons" }, appDirs[5]);

	for
		(const auto& path: appDirs)
	{
		ok = Path::dirExists(path);
		if
			(false == ok)
		{
			ok = Path::createDirs(path, 0755);
			if
				(false == ok)
			{
				LOG( LogLevel::Error, "Application"
                        , "Could not create application directory '{}'"
						, path);
				return false;
			}
		}
	}

	return true;
}

/**
 *
 */
bool Application::saveSettings()
{
	bool ok;	//

	ok = this->settingsStore->save(this->engine.settings, this->settingsPath);
	return ok;
};

/**
 *
 */
bool Application::loadState()
{
	bool ok;	//

	ok = false;

	ok = this->stateStore->setStatePath(this->engine.settings.statePath);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Application", "Could not set state path");
		return false;
	}

    SceneViewState viewState;   //
    TimelineState timelineState;    //

    // Lock poles axis by default
    viewState.lock = true;

    timelineState.timespan = this->engine.settings.timelineCfg.defaultTimespan;
    timelineState.timestep = this->engine.settings.timelineCfg.defaultTimestep;
    timelineState.nearestMode = this->engine.settings.timelineCfg.useNearestMode;
    timelineState.extendedView = true;

	ok = this->stateStore->load(this->engine.state, viewState, timelineState);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Application", "Could not load settings from {}", this->engine.settings.statePath);
        return false;
    }

    SceneViewControllerBase& sceneViewController = this->engine.getSceneViewController();
    sceneViewController.setState(viewState);

    this->engine.timeline.setState(timelineState);
	return true;
};

/**
 *
 */
bool Application::saveState()
{
	bool ok {false};	//

    SceneViewState viewState;   //
    SceneViewControllerBase& sceneViewController = this->engine.getSceneViewController();
    sceneViewController.getState(viewState);

    TimelineState timelineState;    //
    ok = this->engine.timeline.getState(timelineState);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Application", "Failed to get timeline state");
        return false;
    }

    LOG(LogLevel::Debug, "Application", "Timeline state: {} {} {} {} {}",
             timelineState.datetime, timelineState.timespan, timelineState.timestep, timelineState.nearestMode, timelineState.extendedView);

	ok = this->stateStore->save(this->engine.state, viewState, timelineState);
	return ok;
};

/**
 *
 */
bool Application::initWindow()
{
	bool ok	//
	;

	ok = this->window->init(this->engine.settings);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Application"
				, "Could not initialize window");
		return false;
	}

    ok = this->window->create( this->engine.settings.windowWidth
							 , this->engine.settings.windowHeight);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Application"
				, "Could not create window");
		return false;
	}

	return true;
}

/**
 *
 */
bool Application::startEngineUpdates()
{
    bool ok{false}; //
    Job engineUpdateJob;    //
    JobStatus status {JobStatus::CREATED};   //
    std::string engineUpdateJobId{""};  //

    engineUpdateJob.name = "engineUpdate";
    engineUpdateJob.task = [this] (std::string& /*errMsg*/) -> bool {
        return this->engine.getUpdatesManager().run(this->engine);
    };

    ok = this->internalWorkers.submitJob(std::move(engineUpdateJob), engineUpdateJobId, status);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Application",
                 "Could not submit engine update job to task scheduler: {}", engineUpdateJob.errorMsg);
    }
    return ok;
}

/**
 *
 */
bool Application::run(SceneView& sceneView)
{
    bool ok{false}; //
	bool pursue{true};	//
    ISceneRenderer* renderer {nullptr};   //

    Icons& icons = this->engine.getIcons();

    ok = this->startEngineUpdates();
    if
        (false == ok)
    {
        return false;
    }

    renderer = sceneView.getRenderer();
    EngineUpdatesManager& updatesManager = this->engine.getUpdatesManager();

	while
		(true == pursue)
	{
        updatesManager.triggerPendingUpdate(this->engine);
        updatesManager.updateUI(this->engine, this->_ui, sceneView);
        updatesManager.updateTimeHints(renderer->getContext(), sceneView);

		while
			(false == this->engine.displayQueue.queue.empty())
		{
			this->engine.displayQueue.stageFirst();
            DisplayJob& j = this->engine.displayQueue.jobOnStage;

			switch
				(j.jobType)
			{
				case DisplayJobType::UPDATE_STARTED:
                    LOG(LogLevel::Debug, "Application", "Update started");
					renderer->updateStarted(sceneView.getScene());
					renderer->setToRender(j.toRender);
					break;
				case DisplayJobType::LOAD_DATA:
                    if
                        (nullptr != j.shape)
                    {
                        ok = sceneView.getScene()->addShape(renderer->getContext(), j.shape, j.rId, j.renderingCfg, icons);
                    }
                    else if
                        (nullptr != j.dataset)
                    {
                        ok = sceneView.getScene()->addGranule(renderer->getContext(), j.dataset, j.renderingCfg, sceneView.getDataSetVisibility());
                    }
                    else
                    {
                        ok = sceneView.getScene()->addTrajectory(renderer->getContext(), j.relativeTime, j.points, j.fData, j.mask, j.renderingCfg, j.rId);
                    }

                    if
                        (false == ok)
                    {
                        // Dataset was already in the renderer and this duplicate will not be used, delete it to free memory
                        LOG(LogLevel::Debug, "Application", "Dataset already in the renderer");
                        DatasetsAllocator::release(j.dataset);
                    }

                    // Avoid keeping a ref on the dataset that has either been destroyed or passed down to the renderer
                    j.dataset = nullptr;

					break;
				case DisplayJobType::UPDATE_RENDERING:
					renderer->setDisplayCfg(sceneView.getScene(), j.renderingCfg);
					break;
				case DisplayJobType::UPDATE_COMPLETE:
                    LOG(LogLevel::Debug, "Application", "Update complete");
					renderer->updateFinished(sceneView.getScene());
					break;
                case DisplayJobType::NOOP:
                    break;
			}

			this->engine.displayQueue.unstage();
		}
		pursue = this->window->renderFrame();
	}

    this->engine.getUpdatesManager().terminate();

	return true;
}

/**
 *
 */
bool Application::initWorkers()
{
	bool ok;	//

	ok = this->dataWorkers.init(1, &this->dataScheduler);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Application", "Failed to initialize data workers");
		LOG(LogLevel::Debug, "Application", "Init parameters: {}", 1);
		return false;
	}

	ok = this->processWorkers.init(4, &this->processScheduler);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Application", "Failed to initialize processing workers");
		LOG(LogLevel::Debug, "Application", "Init parameters: {}", 4);
		return false;
	}

	ok = this->internalWorkers.init(3, &this->internalScheduler);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Application", "Failed to initialize internal workers");
		LOG(LogLevel::Debug, "Application", "Init parameters: {}", 3);
		return false;
	}

	return true;
}

/**
 *
 */
bool Application::initUI(const std::string& installDir, const RendererResources& resources)
{
    bool ok {false};    //

	ok = this->_ui.init(*this->window, this->engine.settings, &(this->engine.state));
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Application", "Could not initialize main user interface");
		return false;
	}

    // Assemble the Scene MVC
    this->_sceneRenderer.initialize(this->engine.settings.glRenderer, &(this->engine.cmaps), &resources);

    // Load pictograms used for drawing tools
    Icons& icons = this->engine.getIcons();
    std::string iconsDir = installDir + Path::separator() + "images" + Path::separator() + "icons";

    IconLoader il;
	// Loads SEAScope icons
    il.load(this->_sceneRenderer.getContext(), iconsDir, icons);

	std::string userIconsDir = this->engine.settings.customPath + Path::separator() + "icons";
	Path::setPathValid(userIconsDir);

	ok = Path::dirExists(userIconsDir);
	if
		(true == ok)
	{
		// Parses the user icons folder and load all compatibles PNG
		il.parse(this->_sceneRenderer.getContext(), userIconsDir, icons);
	}

    this->_ui.setIcons(icons);

    return true;
}

/**
 *
 */
bool Application::initEngine(SceneView* sceneView, IMetadataStore* metadataStore, IDataStore* dataStore, IPluginsLoader* pluginsLoader)
{
    bool ok {false};    //

    ok = this->engine.init(sceneView, metadataStore, dataStore, pluginsLoader);
	this->engine.setCollectionsUI(this->_ui.getCollectionsUI());
	this->engine.setRenderingUI(this->_ui.getRenderingUI());
	this->engine.setAdvancedRenderingUI(this->_ui.getAdvancedRenderingUI());
	this->engine.setTimelineUI(this->_ui.getTimelineUI());
	this->engine.setSourcesUI(this->_ui.getSourcesUI());
	this->engine.dataMediator.setWorkers(&(this->dataWorkers));
    this->engine.renderingMediator.setWorkers(&(this->processWorkers));

    return ok;
}

/**
 *
 */
bool Application::shutdown()
{
	bool ok;	//

    ok = this->saveState();

    this->engine.renderingMediator.shutdown();

	ok = this->dataWorkers.shutdown();
	ok = this->processWorkers.shutdown();
	ok = this->internalWorkers.shutdown();

	ok = this->engine.shutdown();

    if
        (nullptr != this->window)
    {
        this->window->clean();
        this->window = nullptr;
    }

	return ok;
}

/**
 *
 */
bool Application::openControlSocket(const std::string& interface, std::uint64_t port)
{
	bool ok;	//

	ok = this->engine.sockets.registerProtocol( "SEASCOPE_CONTROL",
		[this] () -> IProtocolHandler* {
			ControlProtocolHandler* handler {nullptr};	//

			handler = new ControlProtocolHandler(&(this->internalWorkers),
                                                 &(this->_grunt));
			return handler;
		}
	);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Application", "Failed to register the SEASCOPE_CONTROL protocol");
		return false;
	}

	ok = this->engine.sockets.createServer(interface, static_cast<std::int64_t>(port), "SEASCOPE_CONTROL");
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Application", "Failed to create sockets server on {}:{}", interface, port);
		return false;
	}

	return true;
}

/**
 *
 */
SceneRenderer& Application::getSceneRenderer()
{
    return this->_sceneRenderer;
}
