// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/collections_control.h"

using namespace SEAScope;

/**
 *
 */
CollectionsControl::CollectionsControl()
: view(nullptr), collectionCallback(nullptr), variableCallback(nullptr),
exclusiveCallback(nullptr)
{
};

/**
 *
 */
CollectionsControl::~CollectionsControl()
{
};

/**
 *
 */
bool CollectionsControl::init(MetadataMediator* metadata, ICollectionsControlView* viewIn)
{
	bool ok;	//

	this->view = viewIn;
	ok = this->model.init(metadata);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Catalogue control"
				, "Model initialization failed.");
		return false;
	}

	ok = this->model.setAddedCallback([this](const std::vector<std::pair<CollectionVariable, CatalogueEntry> >& vars) {
		this->view->addVariables(vars);
	});
	ok = this->model.setRemovedCallback([this](const std::vector<CollectionVariable>& vars) {
		this->view->removeVariables(vars);
	});
	ok = this->model.setSelectionCallback([this](const CollectionVariable& colvar, bool selected) {
		size_t count;	//
		this->view->selectVariable(colvar, selected);
		if
			(nullptr != this->variableCallback)
		{
			this->variableCallback(colvar, selected);
		}

		if
			(nullptr == this->collectionCallback)
		{
			return;
		}

		count = this->model.countSelectedVars(colvar.first);
		if
			(  false == selected
			&& 0 == count)
		{
			/* Collection is no longer active since it was the last selected variable */
			this->collectionCallback(colvar.first, false);
		}
		else if
			(  true == selected
			&& 1 == count)
		{
			/* Collection becomes active since this is the first selected variable */
			this->collectionCallback(colvar.first, true);
		}
	});

	this->model.exclusiveSelectionCallback = [this](const CollectionVariable& colvar) {
		for
			(const auto& it: this->model.store)
		{
			this->view->selectVariable(it.first, it.first == colvar);
		}

		if
			(nullptr != this->exclusiveCallback)
		{
			this->exclusiveCallback(colvar);
		}
	};

	/* Set view callbacks */
	ok = this->view->setSelectionCallback([this](const CollectionVariable& colvar, bool selected, bool shiftPressed) {
		LOG(LogLevel::Debug, "Collections control", "*** *** *** selection callback *** *** ***");
		LOG(LogLevel::Debug, "Collections control", "Selected? {}\nShift pressed? {}", selected, shiftPressed);
		this->model.select(colvar, selected||shiftPressed,  shiftPressed);
	});

	return ok;
};

/**
 *
 */
bool CollectionsControl::setCollectionStateCallback(const std::function<void(const CollectionId&, bool)>& callback)
{
	this->collectionCallback = callback;
	return true;
};

/**
 *
 */
bool CollectionsControl::setVariableStateCallback(const std::function<void(const CollectionVariable&, bool)>& callback)
{
	this->variableCallback = callback;
	return true;
};
