// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/collections_control_model.h"
#include "debug.h"

using namespace SEAScope;

/**
 *
 */
CollectionsControlModel::CollectionsControlModel()
: metadata{nullptr}, addedCallback{nullptr}, removedCallback{nullptr},
  selectionCallback{nullptr}, exclusiveSelectionCallback{nullptr}
{
};

/**
 *
 */
CollectionsControlModel::~CollectionsControlModel()
{
};

/**
 *
 */
bool CollectionsControlModel::init(MetadataMediator* inMetadata)
{
	this->metadata = inMetadata;
	return nullptr != inMetadata;
};

/**
 *
 */
bool CollectionsControlModel::addCollections(const std::set<CollectionId>& collectionIds)
{
	bool ok;	//
	CollectionCfg colCfg;	//
	std::vector<std::pair<CollectionVariable, CatalogueEntry> > added;	//

	for
		(auto& collectionId : collectionIds)
	{
		ok = this->metadata->getCollectionCfg(collectionId, colCfg);
		if
			(false == ok)
		{
			LOG( LogLevel::Error, "Catalogue control model"
					, "Could not retrieve configuration for collection {}/{}"
					, collectionId.first, collectionId.second);
			return false;
		}

		Debug::displayCollection(colCfg);

		for
			(auto& vCfg : colCfg.variables)
		{
			LOG( LogLevel::Debug, "CatalogueControlModel"
					, "Creating entry for {}:{}"
					, colCfg.label, vCfg.second.label);
			CollectionVariable colvar(collectionId, vCfg.first);	//
			auto found = this->store.find(colvar);
			if
				(this->store.end() == found)
			{
				CatalogueEntry entry;
				entry.collectionLabel = colCfg.label;
				entry.variableLabel = vCfg.second.label;
				entry.collectionTags = colCfg.tags;
				entry.variableTags = vCfg.second.tags;
				entry.selected = false;
				std::pair<CollectionVariable, CatalogueEntry> newCol(colvar, entry);
				this->store.insert(newCol);
				added.push_back(newCol);
			}
		}
	}

	if
		(  0 < added.size()
		&& nullptr != this->addedCallback)
	{
		this->addedCallback(added);
	}

	return true;
};

/**
 *
 */
bool CollectionsControlModel::removeCollections(const std::set<CollectionId>& collectionIds)
{
	std::vector<CollectionVariable> removed;	//
	CollectionCfg colCfg;	//
	size_t erased;	//
	bool ok;	//

	for
		(auto& collectionId : collectionIds)
	{
		ok = this->metadata->getCollectionCfg(collectionId, colCfg);
		if
			(false == ok)
		{
			LOG( LogLevel::Error, "Catalogue control model"
					, "Could not retrieve configuration for collection {}/{}"
					, collectionId.first, collectionId.second);
			return false;
		}

		for
			(auto& vCfg : colCfg.variables)
		{
			CollectionVariable colvar(collectionId, vCfg.first);
			erased = this->store.erase(colvar);
			if
				(0 < erased)
			{
				removed.push_back(colvar);
			}
		}
	}

	if
		(  0 < removed.size()
		&& nullptr != this->removedCallback)
	{
		this->removedCallback(removed);
	}
	return 0 < removed.size();
};

/**
 *
 */
bool CollectionsControlModel::select(const CollectionVariable& colvar, bool selected)
{
    return this->select(colvar, selected, false);
};

/**
 *
 */
bool CollectionsControlModel::select(const CollectionVariable& colvar, bool selected, bool exclusive)
{
    bool prevSelection; //

	auto found = this->store.find(colvar);
	if
		(this->store.end() == found)
	{
		LOG( LogLevel::Error, "Catalogue control model"
				, "Cannot change selection state of a variable that is not present in the store: {}/{}:{} selected? {}"
				, colvar.first.first, colvar.first.second, colvar.second, selected);
		return false;
	}

	prevSelection = found->second.selected;

	if
		(!exclusive && selected == prevSelection)
	{
		// Nothing changed
		return true;
	}
    else if
        (exclusive && selected)
    {
        for
            (auto& it: this->store)
        {
            prevSelection = it.second.selected;
            it.second.selected = (it.first == colvar);
        }

		if
			(nullptr != this->exclusiveSelectionCallback)
		{
			this->exclusiveSelectionCallback(colvar);
		}
    }
    else
    {
		found->second.selected = selected;
        if
            (nullptr != this->selectionCallback)
        {
            this->selectionCallback(colvar, selected);
        }
    }

	return true;
};

/**
 *
 */
bool CollectionsControlModel::getSelectedCollections(std::set<CollectionId>& collections) const
{
	for
		(auto& storeEntry : this->store)
	{
		if
			(true == storeEntry.second.selected)
		{
			collections.insert(storeEntry.first.first);
		}
	}

	return false == collections.empty();
};

/**
 *
 */
bool CollectionsControlModel::getSelectedVariables(std::set<CollectionVariable>& variables) const
{
	for
		(auto& storeEntry : this->store)
	{
		if
			(true == storeEntry.second.selected)
		{
			variables.insert(storeEntry.first);
		}
	}

	return false == variables.empty();
};

/**
 *
 */
bool CollectionsControlModel::setAddedCallback(const CatalogueControlAddedCallback& callback)
{
	this->addedCallback = callback;
	return true;
};

/**
 *
 */
bool CollectionsControlModel::setRemovedCallback(const CatalogueControlRemovedCallback& callback)
{
	this->removedCallback = callback;
	return true;
};

/**
 *
 */
bool CollectionsControlModel::setSelectionCallback(const CatalogueControlSelectionCallback& callback)
{
	this->selectionCallback = callback;
	return true;
};

/**
 *
 */
size_t CollectionsControlModel::countSelectedVars(const CollectionId& collectionId) const
{
	size_t result;	//

	result = 0;
	for
		(auto& storeEntry : this->store)
	{
		if
			(  true == storeEntry.second.selected
			&& collectionId == storeEntry.first.first)
		{
			result = result + 1;
		}
	}

	return result;
};

/**
 *
 */
bool CollectionsControlModel::getSelectedVars(const CollectionId& collectionId, std::set<VariableId>& result) const
{
	for
		(auto& storeEntry : this->store)
	{
		if
			(  true == storeEntry.second.selected
			&& collectionId == storeEntry.first.first)
		{
			result.insert(storeEntry.first.second);
		}
	}

	return 0 < result.size();
};

/**
 *
 */
bool CollectionsControlModel::getCollectionVariableByLabel(const std::string& colLabel, const std::string& varLabel, CollectionVariable& result) const
{
	for
		(const auto& it : this->store)
	{
		if
			(  it.second.collectionLabel == colLabel
			&& it.second.variableLabel == varLabel)
		{
			result = it.first;
			return true;
		}
	}
	return false;
}
