// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <memory>
#include <cstring>

#include "core/log.h"
#include "core/job.h"
#include "core/control_protocol_handler.h"
#include "core/engine.h"
#include "core/threads_pool.h"
#include "core/command_grunt.h"

#include "flatbuffers/flatbuffers.h"

using namespace SEAScope;

/**
 *
 */
ControlProtocolHandler::ControlProtocolHandler(ThreadsPool* workers, CommandGrunt* grunt)
: _msgSize(0), _cumulatedRead(0), _buffer(nullptr), _workers(workers), _grunt(grunt)
{
};

/**
 *
 */
ControlProtocolHandler::~ControlProtocolHandler()
{
    this->_workers = nullptr;   // not owned
    this->_grunt = nullptr; // not owned
};

/**
 *
 */
static bool sendMessage(socket_t sock, std::uint8_t* inBuffer, std::uint64_t dataSize, CommandGrunt* grunt)
{
	std::uint8_t dataSizeBytes[8];	//
    bool ok;    //

	dataSizeBytes[0] = static_cast<std::uint8_t>(dataSize & 0x00000000000000ff);
	dataSizeBytes[1] = static_cast<std::uint8_t>((dataSize & 0x000000000000ff00) >> 8);
	dataSizeBytes[2] = static_cast<std::uint8_t>((dataSize & 0x0000000000ff0000) >> 16);
	dataSizeBytes[3] = static_cast<std::uint8_t>((dataSize & 0x00000000ff000000) >> 24);
	dataSizeBytes[4] = static_cast<std::uint8_t>((dataSize & 0x000000ff00000000) >> 32);
	dataSizeBytes[5] = static_cast<std::uint8_t>((dataSize & 0x0000ff0000000000) >> 40);
	dataSizeBytes[6] = static_cast<std::uint8_t>((dataSize & 0x00ff000000000000) >> 48);
	dataSizeBytes[7] = static_cast<std::uint8_t>((dataSize & 0xff00000000000000) >> 56);

    ok = grunt->engine->sockets.monitor->sendData(sock, &(dataSizeBytes[0]), 8);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Control protocol handler", "Failed to send data size through the socket");
        return false;
    }

    ok = grunt->engine->sockets.monitor->sendData(sock, inBuffer, dataSize);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Control protocol handler", "Failed to send data through the socket");
        return false;
    }

    return true;
};


/**
 *
 */
bool ControlProtocolHandler::parseStream(socket_t sock, std::uint8_t* stream, std::uint64_t offset, std::uint64_t available, std::uint64_t& consumed)
{
	std::allocator<std::uint8_t> allocator;	//
	std::string jobId;	//
    JobStatus status {JobStatus::CREATED};  //
	Job job;	//
	bool ok;	//

	consumed = 0;
	if
		(nullptr == this->_buffer)
	{
		LOG(LogLevel::Debug, "Control protocol handler", "Create new message: +{} /{}", offset, available);

		/* Read message size: needs at least 8 bytes to read a uint64_t */
		if
			(8 > available)
		{
			consumed = 0;
			return false;
		}

		this->_msgSize = static_cast<std::uint64_t>(stream[offset+7]) | \
						 static_cast<std::uint64_t>(stream[offset+6]) << 8 | \
						 static_cast<std::uint64_t>(stream[offset+5]) << 16 | \
						 static_cast<std::uint64_t>(stream[offset+4]) << 24 | \
						 static_cast<std::uint64_t>(stream[offset+3]) << 32 | \
						 static_cast<std::uint64_t>(stream[offset+2]) << 40 | \
						 static_cast<std::uint64_t>(stream[offset+1]) << 48 | \
						 static_cast<std::uint64_t>(stream[offset+0]) << 56;
		this->_cumulatedRead = 0;
		this->_buffer = allocator.allocate(this->_msgSize);
		consumed += 8;
		LOG(LogLevel::Debug, "Control protocol handler", "Message size: {}", this->_msgSize);
	}

	if
		(available - consumed < this->_msgSize - this->_cumulatedRead)
	{
		/* Read all remaining bytes in the buffer */
		memcpy(&(this->_buffer[this->_cumulatedRead]), &(stream[offset + consumed]), available - consumed);
		this->_cumulatedRead += available - consumed;
		consumed = available;
	}
	else
	{
		memcpy(&(this->_buffer[this->_cumulatedRead]), &(stream[offset + consumed]), this->_msgSize - this->_cumulatedRead);
		consumed += this->_msgSize - this->_cumulatedRead;

		LOG(LogLevel::Debug, "Control protocol handler", "Message complete: {}", reinterpret_cast<const char*>(this->_buffer));
		LOG(LogLevel::Debug, "Control protocol handler", "Workers? {}", nullptr!=this->_workers);

		/* TODO: send message to job queue (let workers perform heavy tasks) */
		if
			(nullptr != this->_workers)
		{
			std::uint8_t* bfr;	//
			std::uint64_t bfrSize;	//
            CommandGrunt* grunt {nullptr};  //

			bfr = this->_buffer;
			bfrSize = this->_msgSize;
            grunt = this->_grunt;


			LOG(LogLevel::Debug, "Control protocol handler", "Creating job...");
			job.name = "parseControlCommand";
			job.task = [bfr, bfrSize, sock, grunt] (std::string& /*s*/) -> bool {
				std::allocator<std::uint8_t> localAllocator;	//
				std::uint8_t* result;	//
				std::uint64_t resultSize;	//
				bool success;	//

				LOG(LogLevel::Debug, "Control protocol handler: worker", "Message complete from {}: [{}] {}", sock, bfrSize, reinterpret_cast<const char*>(bfr));

				/* Execute command */
				flatbuffers::FlatBufferBuilder builder;
				result = grunt->execute(builder, bfr, resultSize, success);
				if
					(false == success)
				{
					LOG(LogLevel::Error, "Control protocol handler: worker", "Execution failed");
					// TODO: cleanup
					return false;
				}

				/* Send data back [TEST] */
				if
					(0 < resultSize)
				{
					LOG(LogLevel::Debug, "Control protocol handler", "Result ready to be sent: {} bytes [{}]", resultSize, nullptr == result);

					success = sendMessage(sock, result, resultSize, grunt);
					if
						(false == success)
					{
						LOG(LogLevel::Error, "Control protocol handler", "Failed to send data");
						// TODO: cleanup memory
						return false;
					}
					// TODO: cleanup memory
				}

				/* Free the buffer */
				localAllocator.deallocate(bfr, bfrSize);
				return true;
			};

			LOG(LogLevel::Debug, "Control protocol handler", "Submitting job...");
			ok = this->_workers->submitJob(std::move(job), jobId, status);
			if
				(false == ok)
			{
				LOG( LogLevel::Error, "Control protocol handler"
						, "Failed to submit message parsing job");
				/*TODO: find a way to handle this error: discarding the message
				is not an option since it could have side-effects */
				return false;
			}
			job.id = jobId;
			LOG(LogLevel::Debug, "Control protocol handler", "Job submitted: {}", jobId);
		}
		else
		{
			LOG(LogLevel::Debug, "Control protocol handler", "!!!! WORKERS IS NULL !!!!");
		}


		this->_buffer = nullptr; // Stream content will feed another message buffer

		return true;
	}

	return false;
};

/**
 *
 */
bool ControlProtocolHandler::sendData(socket_t sock, std::uint8_t* inBuffer, std::uint64_t dataSize)
{
    return sendMessage(sock, inBuffer, dataSize, this->_grunt);
}

/**
 *
 */
bool ControlProtocolHandler::reset()
{
	return true;
};
