// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>

#include "core/log.h"
#include "core/source.h"
#include "core/data_bucket.h"
#include "core/data_mediator.h"
#include "core/dynamic.h"
#include "core/job.h"
#include "core/utils.h"
#include "core/raw_to_renderable_queue.h"
#include "core/interfaces/idatastream_service_connector.h"
#include "core/interfaces/iannotations_store.h"
#include "core/threads_pool.h"
#include "core/sources_control_model.h"
#include "data/annotation.h"
#include "3d/shape.h"

using namespace SEAScope;

/**
 *
 */
DataMediator::DataMediator()
: _sources{nullptr},
  _workers{nullptr},
  _store{nullptr},
  _annotations{nullptr},
  _parallelizeFieldsLoading{true}
{
};

/**
 *
 */
DataMediator::~DataMediator()
{
    this->_store = nullptr; // not owned
    this->_sources = nullptr;   // not owned
};

/**
 *
 */
bool DataMediator::init(IDataStore* store, const IAnnotationsStore* annotations, SourcesControlModel* sources)
{
	this->_store = store;
	this->_sources = sources;
    this->_annotations = annotations;
	return true;
};

/**
 *
 */
void DataMediator::setWorkers(ThreadsPool* workers)
{
    this->_workers = workers;
}

/**
 *
 */
bool DataMediator::loadData(const CollectionId& collectionId,
                            const GranuleId& granuleId,
						    const std::string& uri,
                            float resolution,
						    const std::set<std::string>& requestedFields,
						    DataModel dataModel,
                            RawToRenderableQueue& outputQueue)
{
	bool ok {false};	//
    if
        (   1 < requestedFields.size()
         && true == this->_parallelizeFieldsLoading)
    {
        // Recursive call
        for
            (const auto& requestedField : requestedFields)
        {
            ok = this->loadData(collectionId, granuleId, uri, resolution, {requestedField}, dataModel, outputQueue);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "DataMediator", "Failed to load field '{}' for '{}'", requestedField, uri);
                return false;
            }
        }

        return ok;
    }

    if
        (SEAScope::Source::ANNOTATIONS == collectionId.first)
    {
        std::set<std::string> allFields;
        allFields = requestedFields;

        outputQueue.addJob(collectionId,
                           granuleId,
                           uri,
                           resolution,
                           std::move(allFields));
        return true;
    }

    // If a data bucket already exists, then it should be handled by the
    // component which initiated the bucket creation, so we only get the
    // buckets that were missing until now.
    std::unordered_map<std::string, DataBucket*> dataBuckets;
    ok = this->_store->getMissingDataBuckets(uri, requestedFields, dataBuckets);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "DataMediator",
            "Failed to get exclusive access to fields from '{}'", uri);
        return false;
    }

    if
        (0 >= dataBuckets.size())
    {
        // No data bucket was created, i.e. data loading is handled by another
        // thread / call for all the requested fields
        // It does not mean that everything is ready, data loading may still be
        // ongoing for some fields so we need to ask the store which ones are
        // fully loaded.
        std::set<std::string> allFields;   //
        this->_store->listFields(uri, allFields);

        LOG(LogLevel::Debug, "DataMediator",
            "skip data loading because all fields are already available");

        if
            (0 < allFields.size())
        {
            outputQueue.addJob(collectionId,
                               granuleId,
                               uri,
                               resolution,
                               std::move(allFields));
        }

        return true;
    }

    std::string fieldsStr;
    SEAScope::implodeStringSet(requestedFields, ",", fieldsStr);
    LOG(LogLevel::Debug, "DataMediator",
        "load data requested for {}:{}", uri, fieldsStr);


	/* Create a job and delegate the download task to a worker */
	DataLoadingJob job;	//
    job.uri = uri;
    job.name = "downloadData";

    // List fields that remain unhandled
    for
        (const auto& it : dataBuckets)
    {
        job.fields.insert(it.first);
    }

	job.task = [this, collectionId, granuleId, uri, dataModel, resolution, dataBuckets, &outputQueue] (std::string& /*s*/) -> bool {
		bool success {false};	//

        LOG(LogLevel::Debug, "DataMediator", "Load data from plugin: granule = {}/{}", granuleId.first, granuleId.second);
        LOG(LogLevel::Debug, "DataMediator", "Load data from plugin: uri = {}", uri);
        //LOG(LogLevel::Debug, "DataMediator", "Load data from plugin: missing fields = {}", missingFields.size());
        LOG(LogLevel::Debug, "DataMediator", "Load data from plugin: model = {}", static_cast<int>(dataModel));

        dynamic<IDatastreamServiceConnector>* datastream;	//
        datastream = this->_sources->getDatastream(granuleId.first, success);
        if
            (false == success)
        {
            LOG( LogLevel::Error, "DataMediator"
                    , "Could not find a suitable datastream plugin for source {}"
                    , granuleId.first);
            return false;
        }

        success = (*datastream)->loadData(granuleId, uri, dataModel, dataBuckets);
        LOG(LogLevel::Debug, "DataMediator", "Data loading complete? {}", success);

        // Set flag to indicate that data buckets are loaded
        for
            (auto& it: dataBuckets)
        {
            if
                (false == it.second->isMissing())
            {
                it.second->setLoaded(true);
            }
        }

        if
            (false == success)
        {
            // Data loading failed for at least some fields, so remove unused
            // data buckets (or the whole data entry if it does not contain any
            // field with data)
            this->_store->removeData(uri);
        }

        std::set<std::string> allFields;   //
        this->_store->listFields(uri, allFields);

        if
            (0 < allFields.size())
        {
            outputQueue.addJob(collectionId,
                               granuleId,
                               uri,
                               resolution,
                               std::move(allFields));
        }

		return true;
	};

    LOG(LogLevel::Debug, "DataMediator", "Submiting job for {}:{}", uri, fieldsStr);

	std::string jobId;	//
    JobStatus status {JobStatus::CREATED};  //
	ok = this->_workers->submitJob(std::move(job), jobId, status);
    if
        (false == ok)
    {
        LOG(LogLevel::Debug, "DataMediator", "Workers refused the job");
    }

	job.id = jobId;

	return ok;
};

/**
 *
 */
GranuleData& DataMediator::getURIDataRef(const std::string& uri)
{
    return this->_store->getURIDataRef(uri);
}

/**
 *
 */
bool DataMediator::getField(const GranuleId granuleId,
                            const std::string& uri,
                            const std::string& field,
                            const Image*& buffer,
                            OGL::Shape*& geometry,
                            const DataInfo*& info) const
{
    if
        (  (nullptr == this->_store)
        || (nullptr == this->_annotations))
    {
        return false;
    }

    if
        (SEAScope::Source::ANNOTATIONS == granuleId.first)
    {
        bool ok {false};    //
        Annotation annotation;  //
        CollectionLocalId collectionId {0u};   //
        ok = this->_annotations->get(granuleId.second, collectionId, annotation);
        if
            (false == ok)
        {
            return false;
        }

        // The Annotation object does not own the OGL::Shape and will not
        // destroy it (it has to be done by the ShapesAllocator)
        geometry = annotation.getShape();
    }
    else
    {
        geometry = nullptr;
        return this->_store->getField(uri, field, buffer, info);
    }
    return true;
}

/**
 *
 */
bool DataMediator::getFields(const std::string& uri,
                             const std::set<std::string>& fields,
                             std::vector<const Image*>& buffers,
                             std::vector<const DataInfo*>& infos) const
{
    if
        (nullptr == this->_store)
    {
        return false;
    }
    return this->_store->getFields(uri, fields, buffers, infos);
}
