// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>
#include <sstream>
#include <iostream>
#include "core/log.h"
#include "core/types.h"
#include "core/guid.h"
#include "core/utils.h"
#include "core/data_scheduler.h"

using namespace SEAScope;

/**
 *
 */
DataScheduler::DataScheduler()
{
};

/**
 *
 */
DataScheduler::~DataScheduler()
{
};

/**
 *
 */
static void buildJobId(const std::string& uri, const std::set<std::string>& fields, std::string& result)
{
    std::ostringstream buffer {};
    buffer << uri << ":";
    for
        (const auto& field : fields)
    {
        buffer << field << "+";
    }
    result = buffer.str();
}

/**
 *
 */
static bool parseJobId(const std::string& jobId, OngoingDataLoadingTask& result)
{
    std::vector<std::string> idTokens;  //
    SEAScope::explodeString(jobId, ":", idTokens);
    if
        (2 != idTokens.size())
    {
        return false;
    }
    result.uri = idTokens[0];

    std::vector<std::string> fieldsVector;  //
    SEAScope::explodeString(idTokens[1], "+", fieldsVector);
    fieldsVector.erase(std::remove_if(fieldsVector.begin(), fieldsVector.end(),
                                    [](const std::string& item) -> bool { return 0 >= item.size(); }),
                       fieldsVector.end());

    result.fields.insert(fieldsVector.begin(), fieldsVector.end());

    return true;
}

/**
 *
 */
void DataScheduler::getOngoingFieldsSet(const std::string& uri, std::set<std::string>& fieldsUnion) const
{
    for
        (const auto& ongoingJob : this->_ongoingJobs)
    {
        if
            (ongoingJob.second.uri == uri)
        {
            fieldsUnion.insert(ongoingJob.second.fields.begin(),
                               ongoingJob.second.fields.end());
        }
    }
}

/**
 *
 */
void DataScheduler::getPendingFieldsSet(const std::string& uri, std::set<std::string>& fieldsUnion) const
{
    for
        (const auto& pendingJob : this->_pendingJobs)
    {
        if
            (pendingJob.uri == uri)
        {
            fieldsUnion.insert(pendingJob.fields.begin(),
                               pendingJob.fields.end());
        }
    }
}

/**
 *
 */
bool DataScheduler::addJob(Job&& job, std::string& jobId, JobStatus& status)
{
    std::set<std::string> fieldsUnion;  //
    std::set<std::string> exclusiveFields;  //
	DataLoadingJob* j {nullptr};	//

	j = dynamic_cast<DataLoadingJob*>(&job);
	if
		(nullptr == j)
	{
		LOG(LogLevel::Error, "DataScheduler", "The job is not a DataLoadingJob");
		return false;
	}

    status = JobStatus::CANCELLED;

    // Only keep fields that are not already targetted by an ongoing job
    this->getOngoingFieldsSet(j->uri, fieldsUnion);
    this->getPendingFieldsSet(j->uri, fieldsUnion);
    if
        (0 < fieldsUnion.size())
    {
        // At least one job already exists for this URI
        std::set_difference(j->fields.begin(), j->fields.end(),
                            fieldsUnion.begin(), fieldsUnion.end(),
                            std::inserter(exclusiveFields, exclusiveFields.begin()));

        if
            (0 >= exclusiveFields.size())
        {
            LOG(LogLevel::Debug, "DataScheduler", "Job superseeded (already loading fields with ongoing or pending jobs)");

            // All the requested fields are already being loaded by ongoing
            // jobs
            status = JobStatus::SUPERSEEDED;

            return true;
        }

        j->fields = exclusiveFields;
    }

    // Generate an identifier
    buildJobId(j->uri, j->fields, j->id);

    // Pass the generated identifier to the caller
    jobId = j->id;

    j->status = JobStatus::PENDING;
    status = j->status;

    this->_pendingJobs.emplace_back(std::move(*j));

	return true;
};

/**
 *
 */
bool DataScheduler::removeJob(const std::string& jobId)
{
	auto it = std::remove_if( this->_pendingJobs.begin(),
				              this->_pendingJobs.end(),
				              [&jobId] (const DataLoadingJob& j) -> bool { return j.id == jobId; });
    this->_pendingJobs.erase(it, this->_pendingJobs.end());

	return true;
};

/**
 *
 */
bool DataScheduler::selectJob(Job& job)
{
	if
		(this->_pendingJobs.empty())
	{
		return false;
	}

    job = std::move(*this->_pendingJobs.begin());
    this->_pendingJobs.erase(this->_pendingJobs.begin());

    LOG(LogLevel::Debug, "DataScheduler", "Select job {}", job.id);
	return true;
};

/**
 *
 */
bool DataScheduler::updateJobStatus(const std::string& jobId, const JobStatus& status)
{
    if
        (JobStatus::RUNNING == status)
    {
        this->_ongoingJobs[jobId];
        parseJobId(jobId, this->_ongoingJobs[jobId]);
        return true;
    }
    else if
        (  (JobStatus::COMPLETE == status)
        || (JobStatus::CANCELLED == status))
    {
        auto foundOngoing = this->_ongoingJobs.find(jobId);
        if
            (this->_ongoingJobs.end() != foundOngoing)
        {
            this->_ongoingJobs.erase(foundOngoing);
            return true;
        }
    }
	auto found = std::find_if( this->_pendingJobs.begin()
							 , this->_pendingJobs.end()
							 , [jobId] (const DataLoadingJob& j) { return j.id == jobId; });

	if
		(  this->_pendingJobs.end() == found
		|| found->status == status)
	{
		return false;
	}

	found->status = status;
	return true;
};

/**
 *
 */
std::size_t DataScheduler::countPendingJobs() const
{
	std::int64_t result;	//

	result = std::count_if( this->_pendingJobs.begin()
						  , this->_pendingJobs.end()
						  , [](const DataLoadingJob& j) { return JobStatus::PENDING == j.status; });
	return static_cast<std::size_t>(result);
};
