// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <chrono>

#include "core/log.h"
#include "core/path.h"
#include "core/utils.h"
#include "core/engine.h"
#include "core/app_events_dispatcher.h"
#include "core/measurements.h"
#include "3d/data_set.h"
#include "3d/unit_sphere.h"
#include "3d/geodetic.h"

#include "3d/datasets_allocator.h"
#include "3d/shapes_allocator.h"
#include "3d/iscene.h"
#include "3d/iscene_renderer.h"
#include "3d/scene_view.h"
#include "3d/scene_graph.h"
#include "3d/scene_graph_controller.h"
#include "3d/polyline_mesh.h"

#include "events/selected_pickables_changed_event.h"
#include "events/selected_pickable_with_info_changed_event.h"
#include "events/center_time_on_granule_request.h"
#include "events/display_granule_info_request.h"
#include "events/granule_info_ready_for_display_event.h"
#include "events/remote_pickle_export_request.h"
#include "events/remote_png_export_request.h"
#include "events/extract_from_renderable_id_request.h"
#include "events/collections_update_requested_event.h"
#include "events/drawing_complete_event.h"
#include "events/add_prepared_annotation_request.h"
#include "events/collection_registration_request.h"
#include "events/granule_added_event.h"
#include "events/rendering_cfg_update_event.h"
#include "events/renderables_changed_event.h"
#include "events/collection_removal_request.h"
#include "events/collection_removed_event.h"
#include "events/variables_added_event.h"
#include "events/variable_selection_request.h"
#include "events/drawing_started_event.h"
#include "events/drawing_complete_event.h"
#include "events/pixel_size_changed_event.h"
#include "events/annotation_time_coverage_updated_event.h"
#include "events/display_extraction_info_request.h"
#include "events/list_overlayed_granules_before_saving_one_annotation_to_file_request.h"
#include "events/save_one_annotation_to_file_request.h"
#include "events/add_default_time_to_load_annotations_from_file_request.h"
#include "events/load_annotations_from_file_request.h"
#include "events/shift_to_prev_time_request.h"
#include "events/shift_to_next_time_request.h"
#include "events/save_annotations_to_file_request.h"
#include "events/save_annotations_from_renderable_ids_to_file_request.h"
#include "events/app_settings_changed_event.h"
#include "events/upcoming_collections_update_event.h"
#include "events/move_variable_to_top_request.h"
#include "events/scene_view_updated_event.h"
#include "events/change_year_month_request.h"
#include "events/change_year_request.h"
#include "core/source.h"
#include "core/granule_metadata.h"


#include "data/annotation.h"


using namespace SEAScope;

/**
 *
 */
Engine::Engine(AppEventsDispatcher& appEventsDispatcher, IGranulesIndex& granulesIndex, IAnnotationsStore& annotationsStore, IGranulesIndex& annotationsIndex)
: _appEventsDispatcher(appEventsDispatcher),
  renderingMediator(appEventsDispatcher, this->renderingControl.model, this->metadataMediator, this->dataMediator, this->settings, this->displayQueue),
  _userGenerated(SEAScope::Source::API, appEventsDispatcher, granulesIndex),
  _windowController(appEventsDispatcher),
  _sceneViewController(appEventsDispatcher),
  _selectionController(appEventsDispatcher),
  _timelineController(appEventsDispatcher),
  _annotationsController(appEventsDispatcher, annotationsStore, annotationsIndex),
  _settingsChangedCallback(nullptr)
{
    this->_windowController.setId(0);
    this->_sceneViewController.setId(1);
    this->_debugController.setId(2);
    this->_debugController.setSceneViewController(&(this->_sceneViewController));
    this->_extractionController.setId(3);
    this->_selectionController.setId(4);
    this->_timelineController.setId(5);

    this->_appEventsDispatcher.subscribe<SelectedPickablesChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<CenterTimeOnGranuleRequest>(this);
    this->_appEventsDispatcher.subscribe<DisplayGranuleInfoRequest>(this);
    this->_appEventsDispatcher.subscribe<ExtractFromRenderableIdRequest>(this);
    this->_appEventsDispatcher.subscribe<RemotePickleExportRequest>(this);
    this->_appEventsDispatcher.subscribe<RemotePNGExportRequest>(this);
    this->_appEventsDispatcher.subscribe<DrawingStartedEvent>(this);
    this->_appEventsDispatcher.subscribe<DrawingCompleteEvent>(this);
    this->_appEventsDispatcher.subscribe<CollectionRegistrationRequest>(this);
    this->_appEventsDispatcher.subscribe<GranuleAddedEvent>(this);
    this->_appEventsDispatcher.subscribe<CollectionRemovalRequest>(this);
    this->_appEventsDispatcher.subscribe<CollectionsUpdateRequestedEvent>(this);
    this->_appEventsDispatcher.subscribe<VariableSelectionRequest>(this);
    this->_appEventsDispatcher.subscribe<PixelSizeChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<AnnotationTimeCoverageUpdatedEvent>(this);
    this->_appEventsDispatcher.subscribe<ListOverlayedGranulesBeforeSavingOneAnnotationToFileRequest>(this);
    this->_appEventsDispatcher.subscribe<AddDefaultTimeToLoadAnnotationsFromFileRequest>(this);
    this->_appEventsDispatcher.subscribe<ShiftToPrevTimeRequest>(this);
    this->_appEventsDispatcher.subscribe<ShiftToNextTimeRequest>(this);
    this->_appEventsDispatcher.subscribe<SaveAnnotationsToFileRequest>(this);
    this->_appEventsDispatcher.subscribe<AppSettingsChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<MoveVariableToTopRequest>(this);
    this->_appEventsDispatcher.subscribe<SceneViewUpdatedEvent>(this);
    this->_appEventsDispatcher.subscribe<ChangeYearMonthRequest>(this);
    this->_appEventsDispatcher.subscribe<ChangeYearRequest>(this);
};

/**
 *
 */
Engine::~Engine()
{
    this->_appEventsDispatcher.unsubscribe<SelectedPickablesChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<CenterTimeOnGranuleRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DisplayGranuleInfoRequest>(this);
    this->_appEventsDispatcher.unsubscribe<ExtractFromRenderableIdRequest>(this);
    this->_appEventsDispatcher.unsubscribe<RemotePickleExportRequest>(this);
    this->_appEventsDispatcher.unsubscribe<RemotePNGExportRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DrawingStartedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<DrawingCompleteEvent>(this);
    this->_appEventsDispatcher.unsubscribe<CollectionRegistrationRequest>(this);
    this->_appEventsDispatcher.unsubscribe<GranuleAddedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<CollectionRemovalRequest>(this);
    this->_appEventsDispatcher.unsubscribe<CollectionsUpdateRequestedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<VariableSelectionRequest>(this);
    this->_appEventsDispatcher.unsubscribe<PixelSizeChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<AnnotationTimeCoverageUpdatedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<ListOverlayedGranulesBeforeSavingOneAnnotationToFileRequest>(this);
    this->_appEventsDispatcher.unsubscribe<AddDefaultTimeToLoadAnnotationsFromFileRequest>(this);
    this->_appEventsDispatcher.unsubscribe<ShiftToPrevTimeRequest>(this);
    this->_appEventsDispatcher.unsubscribe<ShiftToNextTimeRequest>(this);
    this->_appEventsDispatcher.unsubscribe<SaveAnnotationsToFileRequest>(this);
    this->_appEventsDispatcher.unsubscribe<AppSettingsChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<MoveVariableToTopRequest>(this);
    this->_appEventsDispatcher.unsubscribe<SceneViewUpdatedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<ChangeYearMonthRequest>(this);
    this->_appEventsDispatcher.unsubscribe<ChangeYearRequest>(this);
};


/**
 *
 */
bool Engine::init(SceneView* sceneView, IMetadataStore* metaStore, IDataStore* dataStore, IPluginsLoader* pluginsLoader)
{
    std::set<std::string> individualPaths;  //
    bool ok;    //

	if
		(nullptr == sceneView)
	{
		LOG( LogLevel::Error, "Engine"
				, "Rendering is not possible with a null scene view.");
		return false;
	}

    this->_sceneViewController.setSceneView(sceneView);
    this->renderingMediator.setSceneView(sceneView);

    /* Initialize metadata mediator */
    ok = this->metadataMediator.init(metaStore, &(this->sourcesControl.model));
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine", "Failed to initialize metadata mediator");
        return false;
    }

    /* Initialize data mediator */
    ok = this->dataMediator.init(dataStore, this->_annotationsController.getStore(), &(this->sourcesControl.model));
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine", "Failed to initialize data mediator");
        return false;
    }

    /* Initialize plugins */
    this->plugins.init(pluginsLoader);
    ok = this->plugins.discoverPlugins( this->settings.pluginsPath
                                      , individualPaths);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine", "Plugins discovery failed");
        return false;
    }
    dynamic<IIndexServiceConnector>::manager = &this->plugins;
    dynamic<IDatastreamServiceConnector>::manager = &this->plugins;

    /* Initialize colormaps */
    ok = this->initColormaps();
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine", "Colormaps initialization failed");
        return false;
    }

    /* Initialize sources control */
    ok = this->sourcesControl.model.init(this->plugins, this->settings);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine", "Sources control initialization failed");
        return false;
    }

	/* Metadata callbacks */
	this->metadataMediator.variableCfgAdded = [this] (RenderingCfg* rCfg) -> bool{
        bool success;    //

		LOG( LogLevel::Debug, "Engine"
				, "Variable configuration added: {}/{}:{}"
				, rCfg->target.collection.first
				, rCfg->target.collection.second
				, rCfg->target.variable);
		success = this->renderingControl.model.registerCfg(rCfg);
        return success;
	};


	/* Set rendering callbacks */
	this->renderingControl.callback = [this] ( const RenderingCfg* rCfg
                                             , const RenderingParam& param) {
        if
            (  RenderingParam::RENDERED == param
            && false == rCfg->target.granuleLevel)
        {
            CollectionVariable colvar {rCfg->target.collection, rCfg->target.variable};
            this->updateActiveVariables(colvar, true /*assume it was selected*/, rCfg->rendered());
        }
		this->updateRendering(rCfg, param);
	};


	this->renderingControl.toLabel = [this] ( const CollectionVariable& colvar
                                            , std::string& label) -> bool {
        bool success;    //
        RenderableId rId;   //

        rId.collection = colvar.first;
        rId.variable = colvar.second;
        success = this->getRenderableLabel(rId, label);
        return success;
	};

    this->renderingControl.getUnits = [this] ( const CollectionVariable& colvar
                                             , std::string& units) -> bool {
        bool success;    //
        RenderableId rId;   //

        rId.collection = colvar.first;
        rId.variable = colvar.second;
        success = this->getRenderableUnits(rId, units);
        LOG(LogLevel::Debug, "Engine", "-------> get renderable units {} {}",
                 success, units);
        return success;
    };

	this->renderingControl.advanced.saveConfigCallback = [this](SEAScope::RenderingCfg* cfg) {
        bool success {false};

        // Write to config.ini
        success = this->metadataMediator.saveVariableRenderingCfg(cfg);
        if
            (false == success)
        {
            LOG(LogLevel::Warn, "Engine",
                     "Error while saving rendering settings to the collection configuration file, changes will be lost when you close SEAScope.");
        }

        // Update the default RenderingCfg
        renderingControl.model.updateDefaultCfg(cfg);
	};

	return true;
};

/**
 *
 */
bool Engine::initColormaps()
{
    bool ok;    //
    std::forward_list<std::string> colormaps;   //

    ok = Path::listFiles(this->settings.colormapsPath, colormaps);
    if
        (false == ok)
    {
        return false;
    }

    for
        (const auto& it : colormaps)
    {
        if
            (  4 >= it.size()
            || ".rgb" != it.substr(it.size() -4, 4))
        {
            /* Not an RGB file */
            continue;
        }
        LOG(LogLevel::Debug, "Engine", "Load colormap: {}", it);
        ok = this->cmaps.loadColormap(it);
        if
            (false == ok)
        {
            return false;
        }
    }

    return true;
};

bool Engine::getRenderableUnits(const RenderableId& rId, std::string& units)
{
    std::ostringstream buffer; //
    VariableCfg varCfg; //
    bool ok;    //

    ok = this->metadataMediator.getVariableCfg( rId.collection
                                              , rId.variable
                                              , varCfg);
    if
        (false == ok)
    {
        LOG( LogLevel::Error, "Engine"
                , "Could not retrieve variable cfg for {}/{}:{}"
                , rId.collection.first, rId.collection.second, rId.variable);
        return false;
    }

    units = varCfg.units;

    LOG(LogLevel::Debug, "Engine", "getUnits for {}/{}:{}  => {}",
            rId.collection.first, rId.collection.second, rId.variable, units);

    return true;
};

/**
 *
 */
bool Engine::getRenderableLabel(const RenderableId& rId, std::string& label, const std::string& separator)
{
    std::ostringstream buffer;	//
    CollectionCfg colCfg;	//
    VariableCfg varCfg;	//
    bool ok;	//

    ok = this->metadataMediator.getCollectionCfg(rId.collection, colCfg);
    if
        (false == ok)
    {
        LOG( LogLevel::Error, "Main"
                , "Could not retrieve collection cfg for {}/{}"
                , rId.collection.first, rId.collection.second);
        return false;
    }

    ok = this->metadataMediator.getVariableCfg( rId.collection
                                              , rId.variable
                                              , varCfg);
    if
        (false == ok)
    {
        LOG( LogLevel::Error, "Main"
                , "Could not retrieve variable cfg for {}/{}:{}"
                , rId.collection.first, rId.collection.second, rId.variable);
        return false;
    }
    buffer << colCfg.label << separator << varCfg.label;
    label = buffer.str();

    return true;
}

/**
 *
 */
bool Engine::getRenderableLabel(const RenderableId& rId, std::string& label)
{
    return this->getRenderableLabel(rId, label, ": ");
};

/**
 *
 */
bool Engine::setRenderables(std::map<CollectionId, std::vector<GranuleMetadata> >&& renderables)
{
    this->_updatesManager.setRenderables(std::forward<std::map<CollectionId, std::vector<GranuleMetadata> > >(renderables));

    std::map<CollectionId, std::vector<GranuleMetadata> > renderablesCopy;
    this->_updatesManager.getRenderables(renderablesCopy);

    this->renderingMediator.show(renderablesCopy);

    RenderablesChangedEvent event(std::move(renderablesCopy));
    this->_appEventsDispatcher.dispatch<RenderablesChangedEvent>(event);

	return true;
};

/**
 *
 */
void Engine::updateRendering(const RenderingCfg* rCfg, const RenderingParam& param)
{
    this->renderingMediator.updateRendering(rCfg, param);
};


/**
 *
 */
bool Engine::setCollectionsUI(ICollectionsControlView* view)
{
    this->collectionsControl.init(&this->metadataMediator, view);
    this->collectionsControl.setVariableStateCallback([this] (const CollectionVariable& colvar, bool selected) {
		if
			(selected)
		{
			this->renderingControl.addRendering(colvar);
		}
		else
		{
			this->renderingControl.removeRendering(colvar);
		}

        bool ok {false};    //
        RenderingCfg* rCfg {nullptr};   //
        rCfg = this->renderingControl.model.getCollectionCfg(colvar.first, colvar.second, ok);
        if
            ((true == ok) && (nullptr != rCfg))
        {
            this->updateActiveVariables(colvar, selected, rCfg->rendered());
        }
        rCfg = nullptr;
	});

	this->collectionsControl.exclusiveCallback = [this] (const CollectionVariable& colvar) {
        this->renderingControl.exclusive(colvar);

        LOG(LogLevel::Debug, "Engine", "Exclusive selection in rendering control: done");

        auto found = this->_activeCollections.find(colvar.first);
        if
            (this->_activeCollections.end() == found)
        {
            std::set<CollectionId> previouslyActiveCollections;
            previouslyActiveCollections = this->_activeCollections; // copy
            this->updateCollections({colvar.first}, previouslyActiveCollections, false /* i.e. search only if necessary */);
        }
        else
        {
            // The collection owning the selected variable was already active,
            // so remove all items form the active collections set except this
            // one
            std::set<CollectionId> collectionsToDisable;
            std::copy_if(this->_activeCollections.begin(),
                         this->_activeCollections.end(),
                         std::inserter(collectionsToDisable, collectionsToDisable.end()),
                         [colvar](const CollectionId& item) -> bool { return (item.first != colvar.first.first) || (item.second != colvar.first.second); });
            if
                (0 < collectionsToDisable.size())
            {
                // updateCollections will result in an update of the renderables
                this->updateCollections({}, collectionsToDisable, false /* i.e. search only if necessary */);
            }
            else
            {
                // Pass previous search results to the rendering mediator so that
                // it can do whatever is necessary to show only the renderables
                // associated with the variable
                std::map<CollectionId, std::vector<GranuleMetadata> > renderables;
                this->_updatesManager.getRenderables(renderables);

                this->renderingMediator.show(renderables);
            }
        }
	};

    return true;
};

/**
 *
 */
bool Engine::setRenderingUI(IRenderingControlView* view)
{
    this->renderingControl.init(view, &this->cmaps);
    return true;
};

/**
 *
 */
bool Engine::setAdvancedRenderingUI(IAdvancedRenderingControlView* view)
{
    view->init(&this->cmaps);
    this->renderingControl.advanced.init(view);
    return true;
};

/**
 *
 */
bool Engine::setTimelineUI(ITimelineView* view)
{
    this->timeline.init(view, this->settings.timelineCfg);
    return true;
};

/**
 *
 */
bool Engine::setSourcesUI(ISourcesControlView* view)
{
    this->sourcesControl.init(view);
    return true;
};

/**
 *
 */
bool Engine::shutdown()
{
	DisplayJob j;	//
    bool ok;    //

    this->sourcesControl.model.reset();

	ok = this->plugins.releaseAll();
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Engine", "Could not release all plugins");
	}

    dynamic<IIndexServiceConnector>::manager = nullptr;
    dynamic<IDatastreamServiceConnector>::manager = nullptr;

    ok =  ok && this->sockets.shutdown();
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Engine", "Could not stop the sockets manager properly");
	}

    while
		(false == this->displayQueue.queue.empty())
	{
		this->displayQueue.pop(j);

        // Make sure all resources are released
        DatasetsAllocator::release(j.dataset);
        j.dataset = nullptr;
        ShapesAllocator::release(j.shape);
        j.shape = nullptr;
	}


    return ok;
};

/**
 *
 */
bool Engine::getRenderableInfo(const RenderableId& rId, GranuleInfo& info)
{
	bool ok {false};	//

    info.resetMetadata();

    ok = this->renderingMediator.getRenderableURI(rId, info.uri);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine", "Could not find URI for renderable object");
        return false;
    }

	ok = this->metadataMediator.getCollectionCfg( rId.collection
												, info.collection);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Engine"
				, "Could not retrieve collection configuration for renderable object");
		return false;
	}

	ok = this->metadataMediator.getVariableCfg( rId.collection, rId.variable
											  , info.variable);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Engine"
				, "Could not retrieve variable configuration for renderable object");
		return false;
	}

	ok = this->metadataMediator.getDataCoverage(rId.collection, info.coverage);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Engine"
				, "Could not retrieve data coverage for the collection of renderable object");
		return false;
	}

    GranuleMetadata gMeta;
    ok = this->_updatesManager.findRenderable(rId, gMeta);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine",
                 "Could not find granule metadata for picket object");
        return false;
    }
    info.setMetadata(std::move(gMeta));

	if
		(info.collection.xSeamless && info.collection.ySeamless)
	{
		info.area = 510072000000000.f; /* Earth surface */
		info.perimeter = std::numeric_limits<float>::quiet_NaN();
	}
	else
	{
		for
			(std::uint32_t i = 0; i < info.metadata.viewsCount(); ++i)
		{
            IDFDescriptor idf;
            info.metadata.getView(i, idf);
			if
				(idf.uri == info.uri)
			{
				if
					(info.collection.xSeamless || info.collection.ySeamless)
				{
					/* ring band around the Earth => no perimeter */
					info.perimeter = std::numeric_limits<float>::quiet_NaN();
				}
				else
				{
					info.perimeter = static_cast<float>(SEAScope::perimeterFromShape(idf.shape));
				}

				info.area = static_cast<float>(SEAScope::areaFromShape(idf.shape));
				break;
			}
		}
	}
	return true;
};

/**
 *
 */
bool Engine::sendProcessingOpcode(std::uint64_t opcode)
{
	std::uint8_t opcodeBytes[8];	//
    socket_t sock;  //
    bool ok;    //

    ok = false;

	sock = this->sockets.factory->connect(this->settings.processorAddress,
										  this->settings.processorPort, ok);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Engine", "Could not connect to server");
		return false;
	}

	opcodeBytes[0] = static_cast<std::uint8_t>(opcode & 0x00000000000000ff);
    opcodeBytes[1] = static_cast<std::uint8_t>((opcode & 0x000000000000ff00) >> 8);
	opcodeBytes[2] = static_cast<std::uint8_t>((opcode & 0x0000000000ff0000) >> 16);
	opcodeBytes[3] = static_cast<std::uint8_t>((opcode & 0x00000000ff000000) >> 24);
	opcodeBytes[4] = static_cast<std::uint8_t>((opcode & 0x000000ff00000000) >> 32);
	opcodeBytes[5] = static_cast<std::uint8_t>((opcode & 0x0000ff0000000000) >> 40);
	opcodeBytes[6] = static_cast<std::uint8_t>((opcode & 0x00ff000000000000) >> 48);
	opcodeBytes[7] = static_cast<std::uint8_t>((opcode & 0xff00000000000000) >> 56);

    ok = this->sockets.monitor->sendData(sock, &(opcodeBytes[0]), 8);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine", "Failed to send opcode through the socket");
        return false;
    }
    return true;
};

/**
 *
 */
void Engine::update(timestamp_t current, timestamp_t begin, timestamp_t end)
{
	bool ok;	//
	float zoom;	//
	std::vector<GranuleId> granules;	//
	std::map<CollectionId, std::vector<GranuleMetadata> > searchResults;	//
	std::vector<std::pair<CollectionId, bool> > collectionMustBeCurrent;	//
	CollectionCfg colCfg;	//

    SceneView* sceneView = this->_sceneViewController.getSceneView();
	zoom = sceneView->getPixelSize();
    sceneView = nullptr;

	/* Retrieve match criteria for each rendered collection */
	for
		(const auto& collection : this->_activeCollections)
	{
		ok = this->metadataMediator.getCollectionCfg(collection, colCfg);
		if
			(false == ok)
		{
			LOG( LogLevel::Error, "Engine"
					, "Could not retrieve configuration for collection {}/{}"
					, collection.first, collection.second);
			return;
		}
		collectionMustBeCurrent.push_back(std::make_pair(collection, colCfg.mustBeCurrent));
	}

    LOG(LogLevel::Debug, "Engine", "Search granules...");
	ok = this->metadataMediator.searchGranules( current
											   , begin
											   , end
											   , zoom
											   , collectionMustBeCurrent
											   , searchResults);

    /*
	LOG(LogLevel::Debug, "Engine", "{} granules found", searchResults.size());
	for
		(const auto& it : searchResults)
	{
		for
			(const auto& itGranule : it.second)
		{
			LOG( LogLevel::Debug, "Main"
            , "  ==|> {}/{}/{} {}"
					, itGranule.id.first, itGranule.collectionId, itGranule.id.second, itGranule.uris[0].uri);
		}
	}
    */

    this->computeTimeHints(searchResults, current, begin, end);

	this->setRenderables(std::move(searchResults));
};

/**
 *
 */
void Engine::computeTimeHints(const std::map<CollectionId, std::vector<GranuleMetadata>>& results, timestamp_t current, timestamp_t /*timeRangeStart*/, timestamp_t /*timeRangeEnd*/)
{
    for
        (const auto& collectionResults : results)
    {
        if
            (  (0 < collectionResults.second.size())
            && (DataModel::TIME != collectionResults.second[0].dataModel()))
        {
            // If a granule of the collection is not a trajectory, then
            // none is and the collection does not support time hints
            LOG(LogLevel::Debug, "Engine", "Time hints not computed for collection {}.{}",
                     collectionResults.first.first, collectionResults.first.second);
            continue;
        }

        TrajectoryMarkers markers;
        markers.collectionId = collectionResults.first;
        markers.position.reserve(collectionResults.second.size());
        markers.direction.reserve(collectionResults.second.size());

        for
            (const auto& granule : collectionResults.second)
        {
            // Assumptions:
            //  - there is no multi-resolution support for trajectories, so
            //    only one IDFDescriptor should be available (uris[0])
            //  - GCP time values are sorted in ascending order
            //  - granules in results have a temporal coverage which intersects
            //    the range of the timeline

            IDFDescriptor idf;
            granule.getView(0, idf);
            const std::vector<std::uint64_t>& t = idf.shape.gcpsTime;

            if
                (2 > t.size())
            {
                continue;
            }

            // Search iterator for the GCP time at or immediately after current
            // datetime
            const auto itCurrent = std::lower_bound(t.cbegin(), t.cend(), 1000 * current);
            if
                (itCurrent == t.cend())
            {
                // All GCP time values are below current datetime, so there should be no time hint for the current time
                LOG(LogLevel::Debug, "Engine", "Current out of time range (max(time) < current) for granule {}.{} {})",
                         granule.sourceId(), granule.granuleId(), granule.dataId());
                continue;
            }

            if
                (t[0] > 1000 * current)
            {
                // All GCP time values are above current datetime, so there should be no time hint for the current time
                LOG(LogLevel::Debug, "Engine", "Current out of time range (min(time) > current) for granule {}.{} ({})",
                         granule.sourceId(), granule.granuleId(), granule.dataId());
                continue;
            }

            // std::distance can return negative values but casting to unsigned type is safe here because itCurrent >= t.cbegin()
            std::size_t nextIndex = static_cast<std::size_t>(std::distance(t.cbegin(), itCurrent));

            if
                (t[nextIndex] == 1000 * current)
            {
                if
                    (nextIndex < t.size() - 1)
                {
                    // the time of the GCP is exactly equal to current time
                    // if nextIndex == prevIndex, we can compute the position but not the direction
                    // => we shift everything by 1
                    nextIndex = nextIndex + 1;
                }
                // otherwise we are exactly on the last GCP, so we can keep the
                // value of nextIndex computed by lower_bound.
            }

            std::size_t prevIndex = nextIndex - 1;  // at this point we know that nextIndex > 0, so it is safe to subtract 1
            double interpolationParam {0.0};
            if
                (t[nextIndex] > t[prevIndex])
            {
                // time values are sorted in ascending order, but it is safer to check for strict inequality to make sure the same time value has not been used twice, leading to a divide by zero error.
                interpolationParam = static_cast<double>(1000 * current - t[prevIndex]) / static_cast<double>(t[nextIndex] - t[prevIndex]);
            }

            const std::vector<float>& latitude = idf.shape.gcpsLat;
            const std::vector<float>& longitude = idf.shape.gcpsLon;

            glm::vec3 gcpPosBefore = UnitSphere::ToVector3D(
                Geodetic2D{static_cast<double>(glm::radians(longitude[prevIndex])), static_cast<double>(glm::radians(latitude[prevIndex]))});
            glm::vec3 gcpPosAfter = UnitSphere::ToVector3D(
                Geodetic2D{static_cast<double>(glm::radians(longitude[nextIndex])), static_cast<double>(glm::radians(latitude[nextIndex]))});

            glm::vec3 pos = glm::normalize(glm::mix(gcpPosBefore, gcpPosAfter, interpolationParam));
            glm::vec3 dir = glm::normalize(gcpPosAfter - gcpPosBefore);

            markers.position.push_back(pos);
            markers.direction.push_back(dir);

            LOG(LogLevel::Debug, "Engine", "Current in time range for granule {}.{} ({}): {} [{}] < {} < {} [{}]",
                     granule.sourceId(), granule.granuleId(), granule.dataId(), t[prevIndex], prevIndex, current, t[nextIndex], nextIndex);
        }

        this->_updatesManager.setTimeHints(std::move(markers));
    }
}

/**
 *
 */
bool Engine::updateCollections(const std::set<CollectionId>& toAdd, const std::set<CollectionId>& toDel, bool forceUpdate)
{
    return this->updateCollections(toAdd, toDel, {} /*toUpdate*/, forceUpdate);
}

/**
 *
 */
bool Engine::updateCollections(const std::set<CollectionId>& toAdd, const std::set<CollectionId>& toDel, const std::set<CollectionId>& toUpdate, bool forceUpdate)
{
    // Signal to other components that an update is about to happen
    UpcomingCollectionsUpdateEvent outEvent(toAdd, toDel);
    this->_appEventsDispatcher.dispatch<UpcomingCollectionsUpdateEvent>(outEvent);

	bool ok;	//
	timestamp_t newPrev;	//
	timestamp_t newNext;	//
	EventsCoverage coverage;	//
	timestamp_t minDelta;	//

	minDelta = 1;

	LOG(LogLevel::Debug, "Engine", "Rendered collections changed: -{} +{}", toDel.size(), toAdd.size());

	// Remove unselected collections from data coverage
    std::set<CollectionId> actuallyRemoved;
	auto itDel = toDel.begin();
	for
		(;itDel != toDel.end(); itDel++)
	{
        auto found = this->_activeCollections.find(*itDel);
        if
            (this->_activeCollections.end() == found)
        {
            continue;
        }

        this->_activeCollections.erase(found);
        actuallyRemoved.insert(*itDel);

        LOG(LogLevel::Debug, "Engine", "Remove coverage from collection {}/{}", (*itDel).first, (*itDel).second);
		this->timeline.model.removeCoverage(*itDel, true); // silent
	}

    //
    auto itUpdate = toUpdate.begin();
    for
        (;itUpdate != toUpdate.end(); itUpdate++)
    {
        ok = this->metadataMediator.getDataCoverage(*itUpdate, coverage);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Engine", "Could not retrieve data coverage");
            return false;
        }
        this->timeline.model.setCoverage(*itUpdate, coverage, true); // silent
    }

	// Add newly selected collections to data coverage
    std::set<CollectionId> actuallyAdded;
	auto itAdd = toAdd.begin();
	for
		(;itAdd != toAdd.end(); itAdd++)
	{
        auto insertResult = this->_activeCollections.insert(*itAdd);
        if
            (false == insertResult.second)
        {
            continue;
        }

        actuallyAdded.insert(*itAdd);

        LOG(LogLevel::Debug, "Engine", "Add coverage from collection {}/{}", (*itAdd).first, (*itAdd).second);

        if
            (0 < toUpdate.count(*itAdd))
        {
            // Coverage has already been updated, no need to do it twice
            continue;
        }

        ok = this->metadataMediator.getDataCoverage(*itAdd, coverage);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Engine", "Could not retrieve data coverage");
            return false;
        }
        this->timeline.model.setCoverage(*itAdd, coverage, true); // silent
	}

    if
        (  (false == forceUpdate)
        && (0 >= actuallyAdded.size() + actuallyRemoved.size()))
    {
        // No actual change
        return true;
    }

    ok = this->timeline.model.updateCoverages();
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine", "Failed to update timeline coverage");
        return false;
    }

    if
        (  true == this->timeline.model.nearestMode
        && 0 < this->_activeCollections.size())
    {
        timestamp_t currentDatetime {0};    //

        currentDatetime = this->timeline.model.current;
        ok = this->timeline.model.applyNearestMode(currentDatetime, false, false, false);
        if
            (true == ok)
        {
            this->timeline.model.setCurrent(currentDatetime);
        }
    }

    // Update previous/next in the timeline
    this->metadataMediator.searchPrevNext( this->timeline.model.current
                                          , this->_activeCollections
                                          , newPrev
                                          , newNext
                                          , minDelta);
    this->timeline.model.setPrevNext(newPrev, newNext);

    if
        (  (true == forceUpdate)
        || (0 < actuallyAdded.size()))
    {
        // Search for granules in the new configuration
        LOG(LogLevel::Debug, "Engine", "UPDATE REQUESTED");
        this->_updatesManager.requestUpdate();
    }
    else
    {
        LOG(LogLevel::Debug, "Engine", "REMOVE GRANULES FROM DESELECTED COLLECTIONS");

        this->renderingMediator.removeCollections(toDel);

        // Erase granules that belong to the removed collections from the
        // results
        std::map<CollectionId, std::vector<GranuleMetadata> > renderables;
        this->_updatesManager.getRenderables(renderables);

        auto it = actuallyRemoved.begin();
        for
            (; it != actuallyRemoved.end(); ++ it)
        {
            renderables.erase(*it);
        }

        this->_updatesManager.setRenderables(std::move(renderables));
    }

	return true;
};

/**
 *
 */
void Engine::updateActiveVariables(const CollectionVariable& colvar, bool selected, bool rendered)
{
    auto wasActiveCollection {false};    //

    wasActiveCollection = (this->_activeCollections.end() != this->_activeCollections.find(colvar.first));

    if
        (selected && rendered)
    {
        if
            (true == wasActiveCollection)
        {
            LOG(LogLevel::Debug, "Engine", "The collection was already active, simply add renderables for the new variable");

            // Pass previous search results to the rendering mediator so that
            // it can do whatever is necessary to display the new variable
            std::map<CollectionId, std::vector<GranuleMetadata> > renderables;
            this->_updatesManager.getRenderables(renderables);

            this->renderingMediator.addVariables({colvar});
        }
        else
        {
            LOG(LogLevel::Debug, "Engine", "The collection will become active, triggering a full update");
            this->updateCollections({colvar.first}, {}, false /* i.e. search only if necessary */);
        }
    }
    else if
        (false == wasActiveCollection)
    {
        // Collection was not active, and variable is either not selected or
        // not rendered => nothing to do.
    }
    else
    {
        std::size_t count {0u}; //
        count = this->renderingControl.model.countRenderedVariables(colvar.first);

        if
            (0 >= count)
        {
            // There is no more active variable attached to the collection, so
            // remove it
            this->updateCollections({}, {colvar.first}, false /* i.e. search only if necessary */);
        }
        else
        {
            // Pass previous search results to the rendering mediator so that
            // it can do whatever is necessary to remove the renderables
            // associated with the variable
            std::map<CollectionId, std::vector<GranuleMetadata> > renderables;
            this->_updatesManager.getRenderables(renderables);

            this->renderingMediator.removeVariables({colvar});
        }
    }
}

/**
 *
 */
EngineUpdatesManager& Engine::getUpdatesManager()
{
    return this->_updatesManager;
}

/**
 *
 */
bool Engine::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(SelectedPickablesChangedEvent)) == eventType)
    {
        GranuleInfo info;

        const std::vector<Pickable*>& selected = static_cast<const SelectedPickablesChangedEvent&>(event).getSelected();
        if
            (1 == selected.size())
        {
            const RenderableId& rId= selected[0]->renderableId();

            if
                (SEAScope::Source::ANNOTATIONS == rId.collection.first)
            {
                return true;  // event has been processed
            }


            bool ok(false);
            ok = this->getRenderableInfo(rId, info);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "Engine",
                         "Could not retrieve picked object info");
                return true;  // event has been processed, even if it failed
            }

            SelectedPickableWithInfoChangedEvent outEvent(selected[0], info, RenderableType::GRANULE);
            this->_appEventsDispatcher.dispatch<SelectedPickableWithInfoChangedEvent>(outEvent);
        }

        return true;
    }
    else if
        (std::type_index(typeid(CenterTimeOnGranuleRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const CenterTimeOnGranuleRequest&>(event).getRenderableId();
		bool success(false);    //
		GranuleInfo info;   //
		timestamp_t center; //
        double centerTmp(0.0);  //
        double timeExtent(0.0); //

		LOG(LogLevel::Debug, "Engine", "center time callback");

		success = this->getRenderableInfo(rId, info);
		if
			(false == success)
		{
			LOG( LogLevel::Error, "Engine"
					, "Could not retrieve picked object info");
			return true;  // event has been processed, even if it failed
		}

        timeExtent = static_cast<double>(info.metadata.timeCoverageStart() + info.metadata.timeCoverageEnd());
		centerTmp = std::floor(0.0005 * timeExtent);
        center = static_cast<timestamp_t>(centerTmp);
		LOG(LogLevel::Debug, "Main", "Center on {}", center);
		this->timeline.model.setCurrent(center);

        return true;
    }
    else if
        (std::type_index(typeid(DisplayGranuleInfoRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const DisplayGranuleInfoRequest&>(event).getRenderableId();
		bool success(false);	//
		GranuleInfo info;	//

		success = this->getRenderableInfo(rId, info);
		if
			(false == success)
		{
			LOG( LogLevel::Error, "Engine"
					, "Could not retrieve picked object info");
			return true;    // event has been processed, even if it failed
		}

        GranuleInfoReadyForDisplayEvent outEvent(info);
        this->_appEventsDispatcher.dispatch<GranuleInfoReadyForDisplayEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(ExtractFromRenderableIdRequest)) == eventType)
    {
        bool ok(false);
        const RenderableId& rId = static_cast<const ExtractFromRenderableIdRequest&>(event).getRenderableId();

        ok = this->_extractionController.makePersistentExtraction(rId, this);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Engine", "Extraction failed");
            return true;    // event has been processed, even if it failed
        }

        DisplayExtractionInfoRequest outEvent(this->_extractionController.getExtractedData(),
                                              this->_extractionController.getExtractedLabels());
        this->_appEventsDispatcher.dispatch<DisplayExtractionInfoRequest>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(RemotePickleExportRequest)) == eventType)
    {
		bool success(false);    //

		success = this->sendProcessingOpcode(0 /* noop */);
		if
			(false == success)
		{
			LOG(LogLevel::Info, "Engine", "Remote processing server is not available");
            return true;    // event has been processed, even if it failed
		}

		success = this->sendProcessingOpcode(3 /* serialize extraction */);
        return true;    // event has been processed, regardless of success
    }
    else if
        (std::type_index(typeid(RemotePNGExportRequest)) == eventType)
    {
		bool success(false);    //

		success = this->sendProcessingOpcode(0 /* noop */);
		if
			(false == success)
		{
			LOG(LogLevel::Info, "Engine", "Remote processing server is not available");
            return true;    // event has been processed, even if it failed
		}

		success = this->sendProcessingOpcode(1 /* plot extraction */);
        return true;    // event has been processed, regardless of success
    }
    else if
        (std::type_index(typeid(CollectionsUpdateRequestedEvent)) == eventType)
    {
        const std::set<CollectionId>& toAdd = static_cast<const CollectionsUpdateRequestedEvent&>(event).getToAdd();
        const std::set<CollectionId>& toRemove = static_cast<const CollectionsUpdateRequestedEvent&>(event).getToRemove();
        const std::set<CollectionId>& toUpdate = static_cast<const CollectionsUpdateRequestedEvent&>(event).getToUpdate();

        LOG(LogLevel::Debug, "Engine", "CollectionsUpdateRequestedEvent");

        // This event is sent from the AnnotationsController after the creation
        // or the removal of granules, so a new search is necessary
        this->updateCollections(toAdd, toRemove, toUpdate, true /* i.e. force new search */);

        return true;
    }
    else if
        (std::type_index(typeid(DrawingStartedEvent)) == eventType)
    {
        const CollectionVariable& collectionVariable = static_cast<const DrawingStartedEvent&>(event).getCollectionVariable();

        this->collectionsControl.view->allowDeletion(collectionVariable, false);

        return true;
    }
    else if
        (std::type_index(typeid(DrawingCompleteEvent)) == eventType)
    {
        OGL::Shape* shape = static_cast<const DrawingCompleteEvent&>(event).getShape();
        const CollectionVariable& collectionVariable = static_cast<const DrawingCompleteEvent&>(event).getCollectionVariable();

        this->collectionsControl.view->allowDeletion(collectionVariable, true);

        if
            (nullptr == shape)
        {
            // User did not draw anything and switched the drawing tool off
            return true;
        }
        // Default time coverage: current timespan
        timestamp_t start = this->timeline.model.current - this->timeline.model.timeframePastOffset;
        timestamp_t stop = this->timeline.model.current + this->timeline.model.timeframeFutureOffset;
        double defaultStart = 1000.0 * static_cast<double>(start);
        double defaultStop = 1000.0 * static_cast<double>(stop);

        AddPreparedAnnotationRequest outEvent(collectionVariable, shape, defaultStart, defaultStop);
        this->_appEventsDispatcher.dispatch<AddPreparedAnnotationRequest>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(CollectionRegistrationRequest)) == eventType)
    {
        const CollectionId& collectionId = static_cast<const CollectionRegistrationRequest&>(event).getCollectionId();
        const CollectionCfg& collectionCfg = static_cast<const CollectionRegistrationRequest&>(event).getCollectionCfg();
        const std::vector<RenderingCfg*>& collectionRenderingCfgs = static_cast<const CollectionRegistrationRequest&>(event).getCollectionRenderingCfgs();

        this->metadataMediator.collectionCfgLoaded(collectionId,
                                                   collectionCfg,
                                                   collectionRenderingCfgs);

        this->collectionsControl.model.addCollections({collectionId});

        std::map<RenderingCfg*, std::string> addedVariables;
        for
            (const auto rCfg : collectionRenderingCfgs)
        {
            std::string colvarStr("");
            SEAScope::getColvarLabel(collectionCfg.label,
                                     collectionCfg.variables.at(rCfg->target.variable).label,
                                     colvarStr);
            addedVariables[rCfg] = colvarStr;
        }
        VariablesAddedEvent outEvent(addedVariables);
        this->_appEventsDispatcher.dispatch<VariablesAddedEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(CollectionRemovalRequest)) == eventType)
    {
        const CollectionId& colId = static_cast<const CollectionRemovalRequest&>(event).getCollectionId();

        // We will need the collection configuration to notify other components
        // of the variables that have been deleted
        CollectionCfg colCfg;
        this->metadataMediator.getCollectionCfg(colId, colCfg);
        std::vector<VariableId> variables;
        for
            (const auto& variable : colCfg.variables)
        {
            variables.push_back(variable.first);
        }

        // Remove RenderingCfg objects for variables related to the removed collection
        this->renderingControl.model.removeCollection(colId);

        this->collectionsControl.model.removeCollections({colId});

        // Remove VariableCfg and CollectionCfg objects
        this->metadataMediator.store->deleteCollectionCfg(colId);

        CollectionRemovedEvent outEvent(colId, variables);
        this->_appEventsDispatcher.dispatch<CollectionRemovedEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(GranuleAddedEvent)) == eventType)
    {
        std::uint32_t sourceId = static_cast<const GranuleAddedEvent&>(event).getSourceId();
        std::uint32_t collectionId = static_cast<const GranuleAddedEvent&>(event).getCollectionId();
        //std::uint64_t granuleId = static_cast<const GranuleAddedEvent&>(event).getGranuleId();

		std::set<CollectionId> selected;	//
		std::set<CollectionId> toAdd;	//
		std::set<CollectionId> toDel;	//
		CollectionId userCollection;	//
		bool success(false);	//

		success = this->collectionsControl.model.getSelectedCollections(selected);
		if
			(false == success)
		{
			return true;  // event has been processed, regardless of success
		}

		userCollection.first = sourceId;
		userCollection.second = collectionId;
		auto found = selected.find(userCollection);
		if
			(selected.end() == found)
		{
			return true;  // event has been processed, regardless of success
		}

		// Update data availability
		toAdd.insert(userCollection);
        LOG(LogLevel::Debug, "Engine", "GranuleAddedEvent");
		this->updateCollections(toAdd, toDel, true /* i.e. force new search */);

        return true;
    }
    else if
        (std::type_index(typeid(VariableSelectionRequest)) == eventType)
    {
        const CollectionVariable& collectionVariable = static_cast<const VariableSelectionRequest&>(event).getCollectionVariable();
        bool selected = static_cast<const VariableSelectionRequest&>(event).getSelected();

        bool success {false};    //
        RenderingCfg* cfg {nullptr};    //
        cfg = this->renderingControl.model.getCollectionCfg(collectionVariable.first, collectionVariable.second, success);
        if
            (false == success)
        {
            LOG(LogLevel::Error, "Engine", "Could not get rendering configuration for {}.{}:{}",
                     collectionVariable.first.first, collectionVariable.first.second, collectionVariable.second);
            return true;    // event has been processed, regardless of success
        }

        // if a variable is selected using this event, then also make sure that
        // it will be rendered
        if
            (true == selected)
        {
            cfg->setRendered(true, false);
        }

        this->collectionsControl.model.select(collectionVariable, selected, false);

        return true;
    }
    else if
        (std::type_index(typeid(PixelSizeChangedEvent)) == eventType)
    {
        if
            (0 < this->_activeCollections.size())
        {
            // Reuse results from previous search and pass them to the
            // rendering mediator: it will select the most suitable version for
            // each granule based on resolution (required after each zoom-in
            // and zoom-out)
            std::map<CollectionId, std::vector<GranuleMetadata> > renderables;
            this->_updatesManager.getRenderables(renderables);

            this->renderingMediator.show(renderables);
        }
        return true;
    }
    else if
        (std::type_index(typeid(AnnotationTimeCoverageUpdatedEvent)) == eventType)
    {
        const RenderableId& rId = static_cast<const AnnotationTimeCoverageUpdatedEvent&>(event).getRenderableId();
        LOG(LogLevel::Debug, "Engine", "AnnotationTimeCoverageUpdatedEvent");
        this->updateCollections({rId.collection}, {}, {rId.collection}, true /* i.e. force new search in case the modified annotation is not in the current time frame anymore */);
        return true;
    }
    else if
        (std::type_index(typeid(ListOverlayedGranulesBeforeSavingOneAnnotationToFileRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const ListOverlayedGranulesBeforeSavingOneAnnotationToFileRequest&>(event).getRenderableId();
        const std::string& path = static_cast<const ListOverlayedGranulesBeforeSavingOneAnnotationToFileRequest&>(event).getPath();
        const AnnotationFormat& exportFormat = static_cast<const  ListOverlayedGranulesBeforeSavingOneAnnotationToFileRequest&>(event).getExportFormat();

        std::vector<std::string> intersectedURIs;  //
        bool ok(false); //
        ok = this->_extractionController.listIntersectedGranules(rId, this, intersectedURIs);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Engine", "Listing of intersected granules failed");
            return true;    // event has been processed, even if it failed
        }

        DictNode* extraProperties(nullptr);
        if
            (0 < intersectedURIs.size())
        {
            extraProperties = new DictNode();
            ListNode* listNode = new ListNode();
            for
                (const auto& it : intersectedURIs)
            {
                if
                    ("mask" == it)
                {
                    // Skip the mask pseudo-granule used when extraction is
                    // performed on line-type data (poylines, trajectories...)
                    continue;
                }

                if
                    ("[all trajectories]" == it)
                {
                    // Skip the generic URI used for collections of
                    // trajectories as it does not provide any information on
                    // which granule has actually been intersected
                    continue;
                }
                listNode->children.emplace_back(new TextNode(it));
            }

            extraProperties->children.emplace("intersectedGranules", listNode);
            listNode = nullptr;
        }

        SaveOneAnnotationToFileRequest outEvent(rId, path, exportFormat, extraProperties);
        this->_appEventsDispatcher.dispatch<SaveOneAnnotationToFileRequest>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(AddDefaultTimeToLoadAnnotationsFromFileRequest)) == eventType)
    {
        const std::string& path = static_cast<const AddDefaultTimeToLoadAnnotationsFromFileRequest&>(event).getPath();
        const std::map<ShapeType, CollectionId>& collectionsMapping = static_cast<const AddDefaultTimeToLoadAnnotationsFromFileRequest&>(event).getCollectionsMapping();
        const AnnotationFormat& importFormat = static_cast<const AddDefaultTimeToLoadAnnotationsFromFileRequest&>(event).getImportFormat();

        // Default time coverage: 24h centred on current time
        double currentTimeInSeconds = static_cast<double>(this->timeline.model.current);
        double defaultStart = 1000.0 * (currentTimeInSeconds - 43200.0);
        double defaultStop = 1000.0 * (currentTimeInSeconds + 43200.0);

        std::set<CollectionId> selectedCollections;
        this->collectionsControl.model.getSelectedCollections(selectedCollections);

        LoadAnnotationsFromFileRequest outEvent(path, collectionsMapping, importFormat, defaultStart, defaultStop, selectedCollections);
        this->_appEventsDispatcher.dispatch<LoadAnnotationsFromFileRequest>(outEvent);
        return true;
    }
    else if
        (std::type_index(typeid(ShiftToPrevTimeRequest)) == eventType)
    {
        this->timeline.model.goToPrev();
        return true;
    }
    else if
        (std::type_index(typeid(ShiftToNextTimeRequest)) == eventType)
    {
        this->timeline.model.goToNext();
        return true;
    }
    else if
        (std::type_index(typeid(SaveAnnotationsToFileRequest)) == eventType)
    {
        const std::string& path = static_cast<const SaveAnnotationsToFileRequest&>(event).getPath();
        const std::map<ShapeType, CollectionId>& collectionsMapping = static_cast<const SaveAnnotationsToFileRequest&>(event).getCollectionsMapping();
        const AnnotationFormat& exportFormat = static_cast<const SaveAnnotationsToFileRequest&>(event).getExportFormat();

        SceneView* sceneView = this->_sceneViewController.getSceneView();
        const std::set<RenderableId>& rIds = sceneView->getRenderer()->getToRender();
        sceneView = nullptr;

        SaveAnnotationsFromRenderableIdsToFileRequest outEvent(path, rIds, collectionsMapping, exportFormat);
        this->_appEventsDispatcher.dispatch<SaveAnnotationsFromRenderableIdsToFileRequest>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(AppSettingsChangedEvent)) == eventType)
    {
        const Settings& _settings = static_cast<const AppSettingsChangedEvent&>(event).getSettings();

        /*
        std::cout << "Settings changed\n";
        std::cout << "GL Renderer: " << _settings.glRenderer << "\n";
        std::cout << "GL context: " << std::to_string(_settings.glContextMajor) << "." << std::to_string(_settings.glContextMinor) << "\n";
        std::cout << "Colormaps path: " << _settings.colormapsPath << "\n";
        std::cout << "Data path: " << _settings.dataPath << "\n";
        std::cout << "Annotations path: " << _settings.annotationsPath << "\n";
        std::cout << "Window width: " << _settings.windowWidth << "\n";
        std::cout << "Window height: " << _settings.windowHeight << "\n";
        std::cout << "VSync: " << std::to_string(_settings.vsync) << "\n";
        std::cout << "Antialiasing: " << std::to_string(_settings.antialiasing) << "\n";
        std::cout << "Antialiasing type: " << _settings.antialiasingType << "\n";
        std::cout << "Antialising factor: " << std::to_string(_settings.antialiasingFactor) << "\n";
        std::cout << "Fullscreen: " << std::to_string(_settings.fullscreen) << "\n";
        std::cout << "World map: " << _settings.worldMap << "\n";
        std::cout << "Min granule screen estate: " << std::to_string(_settings.minGranuleScreenEstate) << "\n";
        std::cout << "Log filter: " << _settings.logFilter << "\n";
        std::cout << "State path: " << _settings.statePath << "\n";
        std::cout << "Index path: " << _settings.indexPath << "\n";
        std::cout << "Processor enabled: " << std::to_string(_settings.processorEnabled) << "\n";
        std::cout << "Processor address: " << _settings.processorAddress << "\n";
        std::cout << "Processor port: " << std::to_string(_settings.processorPort) << "\n";
        std::cout << "Skip index check: " << std::to_string(_settings.skipIndexCheck) << "\n";
        std::cout << "Search request min delay: " << std::to_string(_settings.searchRequestMinDelay) << "\n";
        std::cout << "Explicit drawing mode exit: " << std::to_string(_settings.explicitDrawingModeExit) << "\n";
        std::cout << "Click pixels tolerance: " << std::to_string(_settings.clickPixelsTolerance) << "\n";
        std::cout << "Min year: " << std::to_string(_settings.timelineCfg.minYear) << "\n";
        std::cout << "Max year: " << std::to_string(_settings.timelineCfg.maxYear) << "\n";
        std::cout << "Auto max year: " << std::to_string(_settings.timelineCfg.getUseAutoMaxYear()) << "\n";
        std::cout << "Initial datetime: " << std::to_string(_settings.timelineCfg.getInitialDatetime()) << "\n";
        std::cout << "Initial datetime string: " << _settings.timelineCfg.getInitialDatetimeString() << "\n";
        std::cout << "Use nearest mode: " << std::to_string(_settings.timelineCfg.useNearestMode) << "\n";
        std::cout << "Use nearest in day mode: " << std::to_string(_settings.timelineCfg.useNearestInDayMode) << "\n";
        std::cout << "Advanced scrolling: " << std::to_string(_settings.timelineCfg.advancedScrolling) << "\n";
        std::cout << "Default timestep: " << _settings.timelineCfg.defaultTimestep << "\n";
        for
            (const auto& timestep : _settings.timelineCfg.timesteps)
        {
            std::cout << "\t" << timestep.label << " | " << timestep.description << " | " << std::to_string(timestep.value) << "\n";
        }

        std::cout << "Default timespan: " << _settings.timelineCfg.defaultTimespan << "\n";
        for
            (const auto& timespan : _settings.timelineCfg.timeframes)
        {
            std::cout << "\t" << timespan.label << " | " << timespan.description << " | " << std::to_string(timespan.pastOffset) << " | " << std::to_string(timespan.futureOffset) << "\n";
        }
        */

        this->settings = _settings;
        if
            (nullptr != this->_settingsChangedCallback)
        {
            this->_settingsChangedCallback();
        }

        return true;
    }
    else if
        (std::type_index(typeid(MoveVariableToTopRequest)) == eventType)
    {
        const CollectionVariable& colvar = static_cast<const MoveVariableToTopRequest&>(event).getCollectionVariable();
        this->renderingControl.displayOnTop(colvar);
        return true;
    }
    else if
        (std::type_index(typeid(SceneViewUpdatedEvent)) == eventType)
    {
        this->_updatesManager.sceneViewChanged();

        return true;
    }
    else if
        (std::type_index(typeid(ChangeYearMonthRequest)) == eventType)
    {
        const uint16_t year = static_cast<const ChangeYearMonthRequest&>(event).getYear();
        const uint8_t month = static_cast<const ChangeYearMonthRequest&>(event).getMonth();

        this->timeline.view->setMonth(year, month);

        return true;
    }
    else if
        (std::type_index(typeid(ChangeYearRequest)) == eventType)
    {
        const uint16_t year = static_cast<const ChangeYearRequest&>(event).getYear();

        this->timeline.view->setYear(year);

        return true;
    }

    return false;
}

UserGenerated& Engine::getUserGenerated()
{
    return this->_userGenerated;
}

/**
 *
 */
ExtractionController& Engine::getExtractionController()
{
    return this->_extractionController;
}

/**
 *
 */
WindowController& Engine::getWindowController()
{
    return this->_windowController;
}

/**
 *
 */
SceneViewControllerBase& Engine::getSceneViewController()
{
    return this->_sceneViewController;
}

/**
 *
 */
DebugController& Engine::getDebugController()
{
    return this->_debugController;
}

/**
 *
 */
SelectionController& Engine::getSelectionController()
{
    return this->_selectionController;
}

/**
 *
 */
TimelineController& Engine::getTimelineController()
{
    return this->_timelineController;
}

/**
 *
 */
AnnotationsController& Engine::getAnnotationsController()
{
    return this->_annotationsController;
}

/**
 *
 */
const ExtractionController& Engine::getExtractionController() const
{
    return this->_extractionController;
}

/**
 *
 */
const WindowController& Engine::getWindowController() const
{
    return this->_windowController;
}

/**
 *
 */
const SceneViewControllerBase& Engine::getSceneViewController() const
{
    return this->_sceneViewController;
}

/**
 *
 */
const DebugController& Engine::getDebugController() const
{
    return this->_debugController;
}

/**
 *
 */
const SelectionController& Engine::getSelectionController() const
{
    return this->_selectionController;
}

/**
 *
 */
const TimelineController& Engine::getTimelineController() const
{
    return this->_timelineController;
}

/**
 *
 */
const AnnotationsController& Engine::getAnnotationsController() const
{
    return this->_annotationsController;
}

/**
 *
 */
Icons& Engine::getIcons()
{
    return this->_icons;
}

/**
 *
 */
void Engine::setSettingsChangedCallback(const std::function<void()>& callback)
{
    this->_settingsChangedCallback = callback;
}

/**
 *
 */
bool Engine::getRenderedGranuleMetadata(const RenderableId& rId, GranuleMetadata& result)
{
    return this->_updatesManager.findRenderable(rId, result);
}

/**
 *
 */
void Engine::variableAddedByAPI(const RenderableId& rId)
{
    this->_updatesManager.addVariable(rId);
}
