// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/engine_updates_manager.h"
#include "core/log.h"
#include "core/structs.h"
#include "core/engine.h"
#include "core/granule_metadata.h"
#include "3d/scene_view.h"
#include "3d/context.h"
#include "core/interfaces/iuser_interface.h"
#include "ui/icamera_control_view.h"


using namespace SEAScope;

/**
 *
 */
EngineUpdatesManager::EngineUpdatesManager()
: _stopEngineUpdates{false},
  _engineUpdateRequested{false},
  _granuleCountersChanged{false},
  _sceneViewChanged{false}
{
    this->_lastUpdate = std::chrono::high_resolution_clock::now();
}

/**
 *
 */
EngineUpdatesManager::~EngineUpdatesManager()
{
}

/**
 *
 */
void EngineUpdatesManager::requestUpdate()
{
    // Scope-based locking
    {
        // Put unique lock on mutex.
        std::unique_lock<std::mutex> lock(this->_engineUpdateMutex);

        this->_engineUpdateRequested = true;
    }

    this->_engineUpdateCondition.notify_one();
}

/**
 *
 */
void EngineUpdatesManager::triggerPendingUpdate(Engine& engine)
{
    bool hasPendingUpdate {false};
    std::chrono::high_resolution_clock::time_point lastUpdate {};

    // Scope-based locking
    {
        // Put unique lock on mutex.
        std::unique_lock<std::mutex> lock(this->_engineUpdateMutex);
        hasPendingUpdate = this->_engineUpdateRequested;
        lastUpdate = this->_lastUpdate;
    }

    if
        (false == hasPendingUpdate)
    {
        return;
    }

    std::chrono::high_resolution_clock::time_point now = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> timeDiff = now - lastUpdate;

    if
        (timeDiff.count() > engine.settings.searchRequestMinDelay)
    {
        this->_engineUpdateCondition.notify_one();
    }
}

/**
 *
 */
void EngineUpdatesManager::terminate()
{
    // Scope-based locking
    {
        // Put unique lock on mutex.
        std::unique_lock<std::mutex> lock(this->_engineUpdateMutex);

        this->_stopEngineUpdates = true;
    }

    this->_engineUpdateCondition.notify_one();
}

/**
 *
 */
bool EngineUpdatesManager::run(Engine& engine)
{
    while
        (true)
    {
        // Scope-based locking
        {
            // Put unique lock on mutex.
            std::unique_lock<std::mutex> lock(this->_engineUpdateMutex);

            // Wait until either an engine update or app termination is requested
            this->_engineUpdateCondition.wait(lock, [this]{ return this->_engineUpdateRequested || this->_stopEngineUpdates; });

            if
                (true == this->_stopEngineUpdates)
            {
                return true;
            }

            this->_engineUpdateRequested = false;
        }

        std::chrono::high_resolution_clock::time_point now = std::chrono::high_resolution_clock::now();
        std::chrono::duration<double> timeDiff = now - this->_lastUpdate;

        LOG(LogLevel::Debug, "EngineUpdatesManager", "Search request delay: {}",
                 engine.settings.searchRequestMinDelay);
        if
            (timeDiff.count() < engine.settings.searchRequestMinDelay)
        {
            LOG(LogLevel::Debug, "EngineUpdatesManager", "Update delayed due to throttling");
            this->_engineUpdateRequested = true;
            continue;
        }

        timestamp_t current;	//
        timestamp_t begin;	//
        timestamp_t end;	//
	    engine.timeline.getBounds(current, begin, end);

        engine.update(current, begin, end);

        // Scope-based locking
        {
            // Put unique lock on mutex.
            std::unique_lock<std::mutex> lock(this->_engineUpdateMutex);

	        this->_lastUpdate = now;
        }
    }
}

/**
 *
 */
void EngineUpdatesManager::updateUI(Engine& engine, IUserInterface& ui, SceneView& sceneView)
{
    const std::lock_guard<std::mutex> lock{this->_uiUpdateMutex};
    if
        (true == this->_granuleCountersChanged)
    {
        engine.renderingControl.model.setCounters(this->_granuleCounters);
        this->_granuleCountersChanged = false;
    }

    if
        (true == this->_sceneViewChanged)
    {
        ui.getCameraUI()->update(&sceneView);
        this->_sceneViewChanged = false;
    }

    if
        (0 < this->_addedVariables.size())
    {
	    std::set<CollectionId> collections; //
        for
            (const RenderableId& rId: this->_addedVariables)
        {
	        collections.insert(rId.collection);
        }

        // trigger update of the view
	    engine.collectionsControl.model.addCollections(collections);
        this->_addedVariables = {};
    }
}

/**
 *
 */
void EngineUpdatesManager::addVariable(const RenderableId& rId)
{
    const std::lock_guard<std::mutex> lock{this->_uiUpdateMutex};
    this->_addedVariables.insert(rId);
}

/**
 *
 */
void EngineUpdatesManager::sceneViewChanged()
{
    const std::lock_guard<std::mutex> lock{this->_uiUpdateMutex};
    this->_sceneViewChanged = true;
}

/**
 *
 */
void EngineUpdatesManager::setGranuleCounters(std::map<CollectionId, std::uint32_t>&& value)
{
    const std::lock_guard<std::mutex> lock{this->_uiUpdateMutex};
    this->_granuleCounters = value;
    this->_granuleCountersChanged = true;
}

/**
 *
 */
void EngineUpdatesManager::updateTimeHints(OGL::Context* ctx, SceneView& sceneView)
{
    std::lock_guard<std::mutex> lock(this->_timeHintsMutex);
    if
        (0 >= this->_timeHintsChanged.size())
    {
        return;
    }

    SceneGraphController* sgc {nullptr};
    sgc = sceneView.getScene()->getNodeFactory();
    if
        (nullptr == sgc)
    {
        LOG(LogLevel::Error, "EngineUpdatesManager",
                 "Cannot update time hints because scene graph controller is null");
        return;
    }

    for
        (auto& hints : this->_timeHints)
    {
        sgc->setTrajectoryMarkers(ctx, std::move(hints.second));
    }

    this->_timeHintsChanged.clear();
}

/**
 *
 */
void EngineUpdatesManager::setTimeHints(TrajectoryMarkers&& hints)
{
    std::lock_guard<std::mutex> lock(this->_timeHintsMutex);
    this->_timeHints[hints.collectionId] = std::forward<TrajectoryMarkers>(hints);
    this->_timeHintsChanged.insert(hints.collectionId);
}

/**
 *
 */
void EngineUpdatesManager::setRenderables(std::map<CollectionId, std::vector<GranuleMetadata> >&& value)
{
    std::lock_guard<std::mutex> lock(this->_renderablesMutex);
    this->_renderables = std::forward<std::map<CollectionId, std::vector<GranuleMetadata> > >(value);

    /* Update granules count in rendering control */
    std::map<CollectionId, std::uint32_t> gCounters;	//

    for
        (auto& renderable : this->_renderables)
    {
        gCounters[renderable.first] = static_cast<unsigned int>(renderable.second.size());
    }

    this->setGranuleCounters(std::move(gCounters));
}

/**
 *
 */
bool EngineUpdatesManager::findRenderable(const RenderableId& rId, GranuleMetadata& result)
{
    std::lock_guard<std::mutex> lock(this->_renderablesMutex);

    auto cFound = this->_renderables.find({rId.collection});
    if
        (this->_renderables.end() == cFound)
    {
        LOG(LogLevel::Error, "EngineUpdatesManager",
                 "Collection not found: {}/{}", rId.collection.first, rId.collection.second);
        return false;
    }

    for
        (const auto& gMeta : cFound->second)
    {
        if
            (gMeta.sourceId() == rId.granule.first && gMeta.granuleId() == rId.granule.second)
        {
            result = gMeta; // Copy
            return true;
        }
    }

    return false;
}

/**
 *
 */
void EngineUpdatesManager::getRenderables(std::map<CollectionId, std::vector<GranuleMetadata> >& result)
{
    std::lock_guard<std::mutex> lock(this->_renderablesMutex);

    // Copy
    result = this->_renderables;
}
