// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>
#include "core/log.h"
#include "core/types.h"
#include "core/guid.h"
#include "core/fifo_scheduler.h"

using namespace SEAScope;

/**
 *
 */
FIFOScheduler::FIFOScheduler()
{
};

/**
 *
 */
FIFOScheduler::~FIFOScheduler()
{
};

/**
 *
 */
bool FIFOScheduler::addJob(Job&& job, std::string& jobId, JobStatus& status)
{
	bool ok;	//

	ok = GUID::generate(job.id);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "FIFOScheduler", "Could not generate GUID for new job");
		return false;
	}
	job.status = JobStatus::PENDING;
	this->jobs.push_back(job);
    jobId = job.id;
    status = job.status;
	return true;
};

/**
 *
 */
bool FIFOScheduler::removeJob(const std::string& jobId)
{
	auto it = std::remove_if( this->jobs.begin()
				  , this->jobs.end()
				  , [&jobId] (const Job& j) -> bool { return j.id == jobId; });

    this->jobs.erase(it, this->jobs.end());

	return true;
};

/**
 *
 */
bool FIFOScheduler::selectJob(Job& job)
{
	if
		(this->jobs.empty())
	{
		return false;
	}

	job = std::move(*this->jobs.begin());
	this->jobs.erase(this->jobs.begin());
	return true;
};

/**
 *
 */
bool FIFOScheduler::updateJobStatus(const std::string& jobId, const JobStatus& status)
{
	auto found = std::find_if( this->jobs.begin()
						  	 , this->jobs.end()
							 , [jobId] (const Job& j) { return j.id == jobId; });

	if
		(  this->jobs.end() == found
		|| found->status == status)
	{
		return false;
	}

	found->status = status;
	return true;
};

/**
 *
 */
std::size_t FIFOScheduler::countPendingJobs() const
{
	std::int64_t result;	//

	result = std::count_if( this->jobs.begin()
						  , this->jobs.end()
						  , [](const Job& j) { return JobStatus::PENDING == j.status; });
	return static_cast<std::size_t>(result);
};
