// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/granule_metadata.h"
#include "core/log.h"
#include "core/structs.h"

#include <iostream>

using namespace SEAScope;

constexpr std::uint32_t SourceIdIndex = 0;
constexpr std::uint32_t CollectionIdIndex = SourceIdIndex + 4;
constexpr std::uint32_t GranuleIdIndex = CollectionIdIndex + 4;
constexpr std::uint32_t DataModelIndex = GranuleIdIndex + 8;
constexpr std::uint32_t AvailableAttributesIndex = DataModelIndex + 1;
constexpr std::uint32_t TimeCoverageStartIndex = AvailableAttributesIndex + 4;
constexpr std::uint32_t TimeCoverageEndIndex = TimeCoverageStartIndex + 8;
constexpr std::uint32_t LatMinIndex = TimeCoverageEndIndex + 8;
constexpr std::uint32_t LatMaxIndex = LatMinIndex + 4;
constexpr std::uint32_t LonMinIndex = LatMaxIndex + 4;
constexpr std::uint32_t LonMaxIndex = LonMinIndex + 4;
constexpr std::uint32_t StringsOffsetIndex = LonMaxIndex + 4;
constexpr std::uint32_t DataIdIndex = StringsOffsetIndex + 4;
constexpr std::uint32_t TitleIndex = DataIdIndex + 4;
constexpr std::uint32_t InstitutionIndex = TitleIndex + 4;
constexpr std::uint32_t CommentIndex = InstitutionIndex + 4;
constexpr std::uint32_t FileIdIndex = CommentIndex + 4;
constexpr std::uint32_t ProductVersionIndex = FileIdIndex + 4;
constexpr std::uint32_t CreatorEmailIndex = ProductVersionIndex + 4;
constexpr std::uint32_t StationIdIndex = CreatorEmailIndex + 4;
constexpr std::uint32_t PlatformIndex = StationIdIndex + 4;
constexpr std::uint32_t SensorIndex = PlatformIndex + 4;
constexpr std::uint32_t GcpsOffsetIndex = SensorIndex + 4;
constexpr std::uint32_t ViewsOffsetIndex = GcpsOffsetIndex + 4;

namespace
{
    ///////////////////////////////////////////////////////////////////////////////////////////////////////
    // Copy

    template<typename T>
    void copyBlob(std::vector<std::uint8_t>& blob, std::uint32_t pos, T value)
    {
        memcpy(blob.data() + pos, &value, sizeof(T));
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////////////
    // Emplace

    template<typename T>
    void emplaceBlob(std::vector<std::uint8_t>& blob, T value)
    {
        size_t pos = blob.size();
        blob.resize(pos + sizeof(T));
        memcpy(blob.data() + pos, &value, sizeof(T));
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////////////
    // Insert

    template<typename T>
    void insertBlob(std::vector<std::uint8_t>& blob, std::uint32_t pos, T value)
    {
        blob.insert(blob.begin() + pos, sizeof(T), 0);
        memcpy(blob.data() + pos, &value, sizeof(T));
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////////////
    // Read

    template<typename T>
    T readBlob(const std::vector<std::uint8_t>& blob, std::uint32_t pos)
    {
        T val;
        memcpy(&val, blob.data() + pos, sizeof(T));
        return val;
    }

    std::string readString(const std::vector<std::uint8_t>& blob, std::uint32_t start, std::uint32_t end)
    {
        std::uint32_t stringsOffset = readBlob<std::uint32_t>(blob, StringsOffsetIndex);
        std::uint32_t startOffset = readBlob<std::uint32_t>(blob, start);
        std::uint32_t endOffset = readBlob<std::uint32_t>(blob, end);

        return std::string{ reinterpret_cast<const char*>(blob.data()) + stringsOffset + startOffset, endOffset - startOffset };
    }
}


/**
 *
 */
GranuleMetadata::GranuleMetadata()
{
}

/**
 * Assumptions:
 * - the cumulative length of strings will not exceed the max value for
 *   std::uint32_t
 * - the cumulative length of gcps will not exceed the max value for
 *   std::uint32_t
 */
void GranuleMetadata::build(std::uint32_t sourceId,
    std::uint32_t collectionId,
    std::uint64_t granuleId,
    DataModel dataModel,
    timestamp_t timeCoverageStart,
    timestamp_t timeCoverageEnd,
    bool hasLatMin,
    float latMin,
    bool hasLatMax,
    float latMax,
    bool hasLonMin,
    float lonMin,
    bool hasLonMax,
    float lonMax,
    const std::string& dataId,
    const std::string& title,
    const std::string& institution,
    const std::string& comment,
    const std::string& fileId,
    const std::string& productVersion,
    const std::string& creatorEmail,
    const std::string& stationId,
    const std::string& platform,
    const std::string& sensor,
    const IDFDescriptor& descriptor)
{
    this->_blob.clear();

    // Reserve memory for all the fields that are not owned by the
    // IDFDescriptor (it could be included later on if optimization is needed)
    this->_blob.reserve(3 * sizeof(std::uint64_t) /*granuleId + timeCoverageStart + timeCoverageEnd*/
        + 15 * sizeof(std::uint32_t) /*sourceId, collectionId + stringsOffset + dataIdEnd + titleEnd + institutionEnd + commentEnd + fileIdEnd + productVersionEnd + creatorEmailEnd + stationIdEnd + platformEnd + sensorEnd + gcpsOffset* + availableAttributes*/
        + 1 * sizeof(std::uint8_t) /*dataModel*/
        + 4 * sizeof(float) /*latMin + latMax + lonMin + lonMax*/
        + dataId.length()
        + title.length()
        + institution.length()
        + comment.length()
        + fileId.length()
        + productVersion.length()
        + creatorEmail.length()
        + stationId.length()
        + platform.length()
        + sensor.length());


    // Granule ID - source
    emplaceBlob<std::uint32_t>(this->_blob, sourceId);

    // Collection ID
    emplaceBlob<std::uint32_t>(this->_blob, collectionId);

    // Granule ID - granule
    emplaceBlob<std::uint64_t>(this->_blob, granuleId);

    // Data model
    this->_blob.emplace_back(static_cast<std::uint8_t>(dataModel));

    // Attributes availability
    std::uint32_t availableAttributes {0u};
    if
        (0 < title.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::TITLE));
    }
    if
        (0 < institution.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::INSTITUTION));
    }
    if
        (0 < comment.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::COMMENT));
    }
    if
        (0 < fileId.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::FILE_ID));
    }
    if
        (0 < productVersion.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::PRODUCT_VERSION));
    }
    if
        (true == hasLatMin)
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::LAT_MIN));
    }
    if
        (true == hasLatMax)
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::LAT_MAX));
    }
    if
        (true == hasLonMin)
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::LON_MIN));
    }
    if
        (true == hasLonMax)
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::LON_MAX));
    }
    if
        (0 < creatorEmail.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::CREATOR_EMAIL));
    }
    if
        (0 < stationId.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::STATION_ID));
    }
    if
        (0 < platform.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::PLATFORM));
    }
    if
        (0 < sensor.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::SENSOR));
    }

    emplaceBlob<std::uint32_t>(this->_blob, availableAttributes);

    // Time coverage start
    emplaceBlob<std::uint64_t>(this->_blob, timeCoverageStart);

    // Time coverage end
    emplaceBlob<std::uint64_t>(this->_blob, timeCoverageEnd);

    // Lat min
    emplaceBlob<float>(this->_blob, latMin);

    // Lat max
    emplaceBlob<float>(this->_blob, latMax);

    // Lon min
    emplaceBlob<float>(this->_blob, lonMin);

    // Lon Max
    emplaceBlob<float>(this->_blob, lonMax);

    // Strings offset
    std::uint32_t stringsOffset {GcpsOffsetIndex + 4u};
    emplaceBlob<std::uint32_t>(this->_blob, stringsOffset);

    // Data ID
    std::uint32_t dataIdEnd {static_cast<std::uint32_t>(dataId.length())};
    emplaceBlob<std::uint32_t>(this->_blob, dataIdEnd);

    // Title
    std::uint32_t titleEnd {dataIdEnd + static_cast<std::uint32_t>(title.length())};
    emplaceBlob<std::uint32_t>(this->_blob, titleEnd);

    // Institution
    std::uint32_t institutionEnd {titleEnd + static_cast<std::uint32_t>(institution.length())};
    emplaceBlob<std::uint32_t>(this->_blob, institutionEnd);

    // Comment
    std::uint32_t commentEnd {institutionEnd + static_cast<std::uint32_t>(comment.length())};
    emplaceBlob<std::uint32_t>(this->_blob, commentEnd);

    // File ID
    std::uint32_t fileIdEnd {commentEnd + static_cast<std::uint32_t>(fileId.length())};
    emplaceBlob<std::uint32_t>(this->_blob, fileIdEnd);

    // Product version
    std::uint32_t productVersionEnd {fileIdEnd + static_cast<std::uint32_t>(productVersion.length())};
    emplaceBlob<std::uint32_t>(this->_blob, productVersionEnd);

    // Creator email
    std::uint32_t creatorEmailEnd {productVersionEnd + static_cast<std::uint32_t>(creatorEmail.length())};
    emplaceBlob<std::uint32_t>(this->_blob, creatorEmailEnd);

    // Station ID
    std::uint32_t stationIdEnd {creatorEmailEnd + static_cast<std::uint32_t>(stationId.length())};
    emplaceBlob<std::uint32_t>(this->_blob, stationIdEnd);

    // Platform
    std::uint32_t platformEnd {stationIdEnd + static_cast<std::uint32_t>(platform.length())};
    emplaceBlob<std::uint32_t>(this->_blob, platformEnd);

    // Sensor
    std::uint32_t sensorEnd {platformEnd + static_cast<std::uint32_t>(sensor.length())};
    emplaceBlob<std::uint32_t>(this->_blob, sensorEnd);

    // GCPs offset
    std::uint32_t gcpsOffset {stringsOffset + sensorEnd};
    emplaceBlob<std::uint32_t>(this->_blob, gcpsOffset);

    // Serialize strings
    if
        (0 < dataId.length())
    {
        this->_blob.insert(this->_blob.end(), dataId.begin(), dataId.end());
    }

    if
        (0 < title.length())
    {
        this->_blob.insert(this->_blob.end(), title.begin(), title.end());
    }

    if
        (0 < institution.length())
    {
        this->_blob.insert(this->_blob.end(), institution.begin(), institution.end());
    }

    if
        (0 < comment.length())
    {
        this->_blob.insert(this->_blob.end(), comment.begin(), comment.end());
    }

    if
        (0 < fileId.length())
    {
        this->_blob.insert(this->_blob.end(), fileId.begin(), fileId.end());
    }

    if
        (0 < productVersion.length())
    {
        this->_blob.insert(this->_blob.end(), productVersion.begin(), productVersion.end());
    }

    if
        (0 < creatorEmail.length())
    {
        this->_blob.insert(this->_blob.end(), creatorEmail.begin(), creatorEmail.end());
    }

    if
        (0 < stationId.length())
    {
        this->_blob.insert(this->_blob.end(), stationId.begin(), stationId.end());
    }

    if
        (0 < platform.length())
    {
        this->_blob.insert(this->_blob.end(), platform.begin(), platform.end());
    }

    if
        (0 < sensor.length())
    {
        this->_blob.insert(this->_blob.end(), sensor.begin(), sensor.end());
    }

    // IDFDescriptor
    this->addIDFDescriptor(descriptor);
}

/**
 *
 */
void GranuleMetadata::addIDFDescriptor(const IDFDescriptor& descriptor)
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);
    constexpr std::uint32_t gcpSize = sizeof(std::uint64_t) + 2 * (sizeof(std::uint32_t) + sizeof(float));
    const std::uint32_t uriSize {static_cast<std::uint32_t>(descriptor.uri.length())};
    const std::uint32_t originalSize {static_cast<std::uint32_t>(this->_blob.size())};

    // Get current stringsOffset
    std::uint32_t stringsOffset = readBlob<std::uint32_t>(this->_blob, StringsOffsetIndex);

    // Get current gcpsOffset
    std::uint32_t gcpsOffset = readBlob<std::uint32_t>(this->_blob, GcpsOffsetIndex);

    // IDFDescriptors must be stored in ascending order of spacing (i.e.
    // the IDF with the best resolution first)
    float existingSpacing{ 0.f };
    std::uint32_t viewOffset = ViewsOffsetIndex;

    while
        (viewOffset < stringsOffset)
    {
        existingSpacing = readBlob<float>(this->_blob, viewOffset + 8);

        if
            (existingSpacing > descriptor.resolution)
        {
            break;
        }
        viewOffset = viewOffset + descriptorSize;
        //std::cout << "viewOffset = " << viewOffset << " < stringsOffset = " << stringsOffset << std::endl;
    }

    // xArity
    insertBlob<std::uint32_t>(this->_blob, viewOffset, descriptor.xArity);

    //tmp = readBlob<std::uint32_t>(_blob, viewOffset);
    // yArity
    insertBlob<std::uint32_t>(this->_blob, viewOffset + 4, descriptor.yArity);

    // spacing in meters
    insertBlob<float>(this->_blob, viewOffset + 8, descriptor.resolution);

    // subsamplingFactor
    insertBlob<std::uint32_t>(this->_blob, viewOffset + 12, descriptor.subsamplingFactor);

    // uriStart
    std::uint32_t uriStartOffset = gcpsOffset - stringsOffset;
    insertBlob<std::uint32_t>(this->_blob, viewOffset + 16, uriStartOffset);

    // uriEnd
    std::uint32_t uriEndOffset {uriStartOffset + uriSize};
    insertBlob<std::uint32_t>(this->_blob, viewOffset + 20, uriEndOffset);

    //std::cout << "Inserting URI start = " << std::to_string(uriStartOffset) << ", end = " << std::to_string(uriEndOffset) << std::endl;

    // shape xArity
    insertBlob<std::uint32_t>(this->_blob, viewOffset + 24, descriptor.shape.xArity);

    // shape yArity
    insertBlob<std::uint32_t>(this->_blob, viewOffset + 28, descriptor.shape.yArity);

    // gcps start
    std::uint32_t gcpsStart = originalSize - gcpsOffset;
    insertBlob<std::uint32_t>(this->_blob, viewOffset + 32, gcpsStart);

    // gcps end
    std::uint32_t gcpsEnd = gcpsStart + static_cast<std::uint32_t>(descriptor.shape.gcpsLat.size()) * gcpSize;
    insertBlob<std::uint32_t>(this->_blob, viewOffset + 36, gcpsEnd);

    // Store uri string
    this->_blob.insert(this->_blob.begin() + gcpsOffset + descriptorSize,
        descriptor.uri.begin(), descriptor.uri.end());

    // Store GCPs
    const std::size_t gcpsCount {descriptor.shape.gcpsLat.size()};
    std::size_t oldSize {this->_blob.size()};
    this->_blob.resize(oldSize + gcpSize * gcpsCount);
    std::uint8_t* timePtr = this->_blob.data() + oldSize;
    std::uint8_t* latPtr = timePtr + gcpsCount * sizeof(std::uint64_t);
    std::uint8_t* lonPtr = latPtr + gcpsCount * sizeof(float);
    std::uint8_t* minorAxisPtr = lonPtr + gcpsCount * sizeof(float);
    std::uint8_t* majorAxisPtr = minorAxisPtr + gcpsCount * sizeof(std::uint32_t);

    std::memcpy(timePtr, descriptor.shape.gcpsTime.data(), gcpsCount * sizeof(std::uint64_t));
    std::memcpy(latPtr, descriptor.shape.gcpsLat.data(), gcpsCount * sizeof(float));
    std::memcpy(lonPtr, descriptor.shape.gcpsLon.data(), gcpsCount * sizeof(float));
    std::memcpy(minorAxisPtr, descriptor.shape.gcpsMinorAxisIndex.data(), gcpsCount * sizeof(std::uint32_t));
    std::memcpy(majorAxisPtr, descriptor.shape.gcpsMajorAxisIndex.data(), gcpsCount * sizeof(std::uint32_t));

    // Compute new offsets and update their values in the blob
    stringsOffset = stringsOffset + descriptorSize;
    copyBlob<std::uint32_t>(this->_blob, StringsOffsetIndex, stringsOffset);

    gcpsOffset = gcpsOffset + descriptorSize + uriSize;
    copyBlob<std::uint32_t>(this->_blob, GcpsOffsetIndex, gcpsOffset);
}

/**
 *
 */
std::uint32_t GranuleMetadata::sourceId() const
{
    return readBlob<std::uint32_t>(this->_blob, SourceIdIndex);
}

/**
 *
 */
std::uint32_t GranuleMetadata::collectionId() const
{
    return readBlob<std::uint32_t>(this->_blob, CollectionIdIndex);
}

/**
 *
 */
std::uint64_t GranuleMetadata::granuleId() const
{
    return readBlob<std::uint64_t>(this->_blob, GranuleIdIndex);
}

/**
 *
 */
DataModel GranuleMetadata::dataModel() const
{
    return static_cast<DataModel>(this->_blob[DataModelIndex]);
}

/**
 *
 */
timestamp_t GranuleMetadata::timeCoverageStart() const
{
    return readBlob<std::uint64_t>(this->_blob, TimeCoverageStartIndex);
}

/**
 *
 */
timestamp_t GranuleMetadata::timeCoverageEnd() const
{
    return readBlob<std::uint64_t>(this->_blob, TimeCoverageEndIndex);
}

/**
 *
 */
float GranuleMetadata::latMin() const
{
    return readBlob<float>(_blob, LatMinIndex);
}

/**
 *
 */
float GranuleMetadata::latMax() const
{
    return readBlob<float>(_blob, LatMaxIndex);
}

/**
 *
 */
float GranuleMetadata::lonMin() const
{
    return readBlob<float>(_blob, LonMinIndex);
}

/**
 *
 */
float GranuleMetadata::lonMax() const
{
    return readBlob<float>(_blob, LonMaxIndex);
}

/**
 *
 */
std::string GranuleMetadata::dataId() const
{
    std::uint32_t stringsOffset = readBlob<std::uint32_t>(this->_blob, StringsOffsetIndex);
    std::uint32_t dataIdEnd = readBlob<std::uint32_t>(this->_blob, DataIdIndex);

    //std::cout << "dataIdEnd" << dataIdEnd << std::endl;
    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset, dataIdEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::title() const
{
    return readString(this->_blob, DataIdIndex, TitleIndex);
}

/**
 *
 */
std::string GranuleMetadata::institution() const
{
    return readString(this->_blob, TitleIndex, InstitutionIndex);
}

/**
 *
 */
std::string GranuleMetadata::comment() const
{
    return readString(this->_blob, InstitutionIndex, CommentIndex);
}

/**
 *
 */
std::string GranuleMetadata::fileId() const
{
    return readString(this->_blob, CommentIndex, FileIdIndex);
}

/**
 *
 */
std::string GranuleMetadata::productVersion() const
{
    return readString(this->_blob, FileIdIndex, ProductVersionIndex);
}


/**
 *
 */
std::string GranuleMetadata::creatorEmail() const
{
    return readString(this->_blob, ProductVersionIndex, CreatorEmailIndex);
}

/**
 *
 */
std::string GranuleMetadata::stationId() const
{
    return readString(this->_blob, CreatorEmailIndex, StationIdIndex);
}

/**
 *
 */
std::string GranuleMetadata::platform() const
{
    return readString(this->_blob, StationIdIndex, PlatformIndex);
}

/**
 *
 */
std::string GranuleMetadata::sensor() const
{
    return readString(this->_blob, PlatformIndex, SensorIndex);
}

/**
 *
 */
std::uint32_t GranuleMetadata::viewsCount() const
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);

    // Get current stringsOffset
    std::uint32_t stringsOffset = readBlob<std::uint32_t>(this->_blob, StringsOffsetIndex);

    std::uint32_t count = (stringsOffset - ViewsOffsetIndex) / descriptorSize;
    return count;
}

/**
 *
 */
bool GranuleMetadata::getView(std::uint32_t index, IDFDescriptor& idf) const
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);
    constexpr std::uint32_t gcpSize = sizeof(std::uint64_t) + 2 * sizeof(std::uint32_t) + 2 * sizeof(float);
    const std::uint32_t descriptorOffset {ViewsOffsetIndex + index * descriptorSize};
    std::uint32_t startOffset {0u};    //
    std::uint32_t endOffset {0u};   //

    // Get current stringsOffset
    std::uint32_t stringsOffset = readBlob<std::uint32_t>(this->_blob, StringsOffsetIndex);

    // Get current gcpsOffset
    std::uint32_t gcpsOffset = readBlob<std::uint32_t>(this->_blob, GcpsOffsetIndex);

    idf.xArity = readBlob<std::uint32_t>(this->_blob, descriptorOffset);
    idf.yArity = readBlob<std::uint32_t>(this->_blob, descriptorOffset + 4);
    idf.resolution = readBlob<float>(this->_blob, descriptorOffset + 8);
    idf.subsamplingFactor = readBlob<std::uint32_t>(this->_blob, descriptorOffset + 12);

    startOffset = readBlob<std::uint32_t>(this->_blob, descriptorOffset + 16);
    endOffset = readBlob<std::uint32_t>(this->_blob, descriptorOffset + 20);

    //std::cout << "getting " << std::to_string(index) << "th view" << std::endl;
    //std::cout << "stringsOffset = " << std::to_string(stringsOffset) << ", startOffset = " << std::to_string(startOffset) << ", endOffset = " << std::to_string(endOffset) << std::endl;
    idf.uri = std::string { reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + startOffset, endOffset - startOffset };

    idf.shape.xArity = readBlob<std::uint32_t>(this->_blob, descriptorOffset + 24);
    idf.shape.yArity = readBlob<std::uint32_t>(this->_blob, descriptorOffset + 28);

    startOffset = readBlob<std::uint32_t>(this->_blob, descriptorOffset + 32);
    endOffset = readBlob<std::uint32_t>(this->_blob, descriptorOffset + 36);

    //std::cout << "shape.xArity = " << std::to_string(idf.shape.xArity) << ", shape.yArity = " << std::to_string(idf.shape.yArity) << ", uri = " << idf.uri << std::endl;
    //std::cout << "gcpsOffset = " << std::to_string(gcpsOffset) << ", startOffset = " << std::to_string(startOffset) << ", endOffset = " << std::to_string(endOffset) <<std::endl;
    std::size_t gcpsCount {0u}; //
    gcpsCount = static_cast<std::size_t>((endOffset - startOffset) / gcpSize);
    idf.shape.gcpsTime.resize(gcpsCount);
    idf.shape.gcpsLat.resize(gcpsCount);
    idf.shape.gcpsLon.resize(gcpsCount);
    idf.shape.gcpsMinorAxisIndex.resize(gcpsCount);
    idf.shape.gcpsMajorAxisIndex.resize(gcpsCount);

    const std::uint8_t* timePtr = this->_blob.data() + gcpsOffset + startOffset;
    const std::uint8_t* latPtr = timePtr + gcpsCount * sizeof(std::uint64_t);
    const std::uint8_t* lonPtr = latPtr + gcpsCount * sizeof(float);
    const std::uint8_t* minorAxisPtr = lonPtr + gcpsCount * sizeof(float);
    const std::uint8_t* majorAxisPtr = minorAxisPtr + gcpsCount * sizeof(std::uint32_t);

    std::memcpy(idf.shape.gcpsTime.data(), timePtr, gcpsCount * sizeof(std::uint64_t));
    std::memcpy(idf.shape.gcpsLat.data(), latPtr, gcpsCount * sizeof(float));
    std::memcpy(idf.shape.gcpsLon.data(), lonPtr, gcpsCount * sizeof(float));
    std::memcpy(idf.shape.gcpsMinorAxisIndex.data(), minorAxisPtr, gcpsCount * sizeof(std::uint32_t));
    std::memcpy(idf.shape.gcpsMajorAxisIndex.data(), majorAxisPtr, gcpsCount * sizeof(std::uint32_t));

    return true;
}

/**
 *
 */
bool GranuleMetadata::getViewURI(std::uint32_t index, std::string& result) const
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);
    const std::uint32_t descriptorOffset {ViewsOffsetIndex + index * descriptorSize};

    // Get current stringsOffset
    result = readString(this->_blob, descriptorOffset + 16, descriptorOffset + 20);
    return true;
}

/**
 *
 */
bool GranuleMetadata::selectResolution(std::uint32_t viewportPixelArea,
    float viewportSpacingInMeters,
    float minScreenEstate,
    std::uint32_t& xArity,
    std::uint32_t& yArity,
    IDFDescriptor& idf) const
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);
    std::int64_t j {0};	//
    float deltaRes{ 0.f };	//
    float viewportSize{ 0.f };   //
    float granuleSize{ 0.f };	//
    unsigned int defaultIndex{ 0u }; //
    float spacingInMeters{ 0.f };    //

    std::size_t viewsCount {this->viewsCount()};    //

    viewportSize = static_cast<float>(static_cast<double>(viewportSpacingInMeters) * std::sqrt(viewportPixelArea));
    j = static_cast<std::int64_t>(viewsCount) - 1;
    for
        (; j > -1; --j)
    {
        std::uint32_t k = static_cast<std::uint32_t>(j);
        std::uint32_t descriptorOffset = ViewsOffsetIndex + k * descriptorSize;

        xArity = readBlob<std::uint32_t>(this->_blob, descriptorOffset);
        yArity = readBlob<std::uint32_t>(this->_blob, descriptorOffset + 4);

        spacingInMeters = readBlob<float>(this->_blob, descriptorOffset + 8);

        deltaRes = viewportSpacingInMeters - spacingInMeters;
        LOG(LogLevel::Debug, "GranuleMetadata", "Check res {}: 0 < {} ? {}",
            spacingInMeters, deltaRes, 0.f < deltaRes);
        if
            (0.f < deltaRes)
        {
            granuleSize = static_cast<float>(static_cast<double>(spacingInMeters) * std::sqrt(xArity * xArity + yArity * yArity));
            LOG(LogLevel::Debug, "GranuleMetadata", "Screen estate: {}m / {}m => {}"
                , granuleSize, viewportSize, minScreenEstate);
            if
                (minScreenEstate > granuleSize / viewportSize)
            {
                // The portion of the viewport occupied by the granule is below
                // the threshold
                if
                    (k < viewsCount - 1)
                {
                    // Select lower resolution since the higher resolution
                    // is too far away from the viewport res
                    return this->getView(k + 1, idf);
                }

                // no lower resolution, higher resolution too far
                return false;
            }

            // found a suitable higher res
            return this->getView(k, idf);
        }

        // Ignore lower res data
    }

    // No suitable higher res data found, use closest low res
    LOG(LogLevel::Debug, "GranuleMetadata", "Use first resolution by default");
    if
        (defaultIndex >= viewsCount)
    {
        LOG(LogLevel::Error, "GranuleMetadata",
            "Even first resolution is not suitable (probably a dimension exceeding GPU memory capacity)");
        return false;
    }
    return this->getView(defaultIndex, idf);
};


/**
 *
 */
bool GranuleMetadata::has(IDFAttributes attr) const
{
    std::uint32_t val = readBlob<std::uint32_t>(this->_blob, AvailableAttributesIndex);
    return (0 < (val & (1u << static_cast<std::size_t>(attr))));
}

/**
 *
 */
void GranuleMetadata::sourceId(std::uint32_t value)
{
    copyBlob<std::uint32_t>(this->_blob, SourceIdIndex, value);
}

/**
 *
 */
void GranuleMetadata::collectionId(std::uint32_t value)
{
    copyBlob<std::uint32_t>(this->_blob, CollectionIdIndex, value);
}

/**
 *
 */
void GranuleMetadata::granuleId(std::uint64_t value)
{
    copyBlob<std::uint64_t>(this->_blob, GranuleIdIndex, value);
}

/**
 *
 */
void GranuleMetadata::timeCoverageStart(timestamp_t value)
{
    copyBlob<std::uint64_t>(this->_blob, TimeCoverageStartIndex, value);
}

/**
 *
 */
void GranuleMetadata::timeCoverageEnd(timestamp_t value)
{
    copyBlob<std::uint64_t>(this->_blob, TimeCoverageEndIndex, value);
}

/**
 *
 */
const std::uint8_t* GranuleMetadata::getBlob(std::uint64_t& size) const
{
    size = this->_blob.size();
    return this->_blob.data();
}

/**
 *
 */
void GranuleMetadata::setBlob(std::uint8_t* value, std::uint64_t size)
{
    this->_blob.assign(value, value + size);

    //std::cout << "setBlob(0x???, " << std::to_string(size) << "):" << std::endl;
    //std::cout << "\tsourceId: " << std::to_string(this->sourceId()) << std::endl;
    //std::cout << "\tcollectionId: " << std::to_string(this->collectionId()) << std::endl;
    //std::cout << "\tgranuleId: " << std::to_string(this->granuleId()) << std::endl;
    //std::cout << "\tdataModel: " << std::to_string(static_cast<int>(this->dataModel())) << std::endl;
    //std::cout << "\ttimeCoverageStart: " << std::to_string(this->timeCoverageStart()) << std::endl;
    //std::cout << "\ttimeCoverageEnd: " << std::to_string(this->timeCoverageEnd()) << std::endl;
    //std::cout << "\tlatMin: " << std::to_string(this->latMin()) << std::endl;
    //std::cout << "\tlatMax: " << std::to_string(this->latMax()) << std::endl;
    //std::cout << "\tlonMin: " << std::to_string(this->lonMin()) << std::endl;
    //std::cout << "\tlonMax: " << std::to_string(this->lonMax()) << std::endl;
    //std::cout << "\tdataId: " << this->dataId() << std::endl;
    //std::cout << "\ttitle: " << this->title() << std::endl;
    //std::cout << "\tinstitution: " << this->institution() << std::endl;
    //std::cout << "\tcomment: " << this->comment() << std::endl;
    //std::cout << "\tfileId: " << this->fileId() << std::endl;
    //std::cout << "\tproductVersion: " << this->productVersion() << std::endl;
    //std::cout << "\tstationId: " << this->stationId() << std::endl;
    //std::cout << "\tplatform: " << this->platform() << std::endl;
    //std::cout << "\tsensor: " << this->sensor() << std::endl;
    //std::cout << "\tviewsCount: " << std::to_string(this->viewsCount()) << std::endl;
}
