// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "core/icon_loader.h"

#include "core/path.h"
#include "core/log.h"
#include "data/image.h"
#include "3d/context.h"
#include "3d/texture.h"
#include "core/path_tools.h"
#include <stb_image.h>


IconLoader::IconLoader()
{
}

IconLoader::~IconLoader()
{
}

void IconLoader::parse(OGL::Context* ctx, const std::string& path, Icons& icons)
{
    std::forward_list<std::string> fileList;
    constexpr int maxSize = 256;

    bool ok = SEAScope::Path::listFiles(path, fileList);

    if
        (true == ok)
    {
        Image img;

        auto it = fileList.cbegin(); 
        auto itEnd = fileList.cend();

        std::uint16_t i = 0;
        const std::string ext("png");

        for
            (; it != itEnd; ++it)
        {
            // Check if the current file has the .png extension.
            if
                (ext != extension(*it))
            {
                continue;
            }

            int width, height, channel;
            int ret = stbi_info((*it).c_str(), &width, &height, &channel);

            if
                (0 == ret)
            {
                continue;
            }

            if 
                (maxSize < width
              && maxSize < height)
            {
                LOG(SEAScope::LogLevel::Error, "IconLoader::parse",
                    "Icon is too big, max size is {}x{}", maxSize, maxSize);
                continue;
            }

            if
                (4 != channel)
            {
                LOG(SEAScope::LogLevel::Error, "IconLoader::parse",
                    "Icon need to be RGBA");
                continue;
            }

            unsigned char* data = stbi_load((*it).c_str(), &width, &height, &channel, 0);

            if
                (nullptr == data)
            {
                LOG(SEAScope::LogLevel::Error, "IconLoader",
                    "An error occurred during icon loading");
                return;
            }

            ImageDesc desc;
            desc.width = static_cast<unsigned int>(width);
            desc.height = static_cast<unsigned int>(height);
            desc.channels = static_cast<unsigned int>(channel);
            desc.dataType = DataType::unsigned_byte;

            img.setData(desc, data);

            std::uint16_t iconId = static_cast<std::uint16_t>(static_cast<std::uint16_t>(IconType::ICON_USER_0) + i);

            icons.setTexture(ctx, iconId, img, filename(*it));

            ++i;
        }
    }
}

void IconLoader::load(OGL::Context* ctx, const std::string& iconsDir, Icons& icons)
{
    std::string path = iconsDir;
    addEndingSeparator(path);

    // Add an underscore at the beginning to identify internal icons.
    icons.setTexture(IconType::CIRCLE_OUTLINE_32,       ctx, path + "circle_outline_32.png",        "_circle_outline_32.png");
    icons.setTexture(IconType::CIRCLE_DOT_32,           ctx, path + "circle_dot_32.png",            "_circle_dot_32.png");
    icons.setTexture(IconType::CIRCLE_FILLED_32,        ctx, path + "circle_filled_32.png",         "_circle_filled_32.png");
    icons.setTexture(IconType::LOSANGE_OUTLINE_32,      ctx, path + "losange_outline_32.png",       "_losange_outline_32.png");
    icons.setTexture(IconType::CIRCLE_PLUS_32,          ctx, path + "circle_plus_32.png",           "_circle_plus_32.png");
    icons.setTexture(IconType::SMALL_CIRCLE_FILLED_32,  ctx, path + "small_circle_filled_32.png",   "_small_circle_filled_32.png");
    icons.setTexture(IconType::SMALL_CIRCLE_MOVE_32,    ctx, path + "small_circle_move_32.png",     "_small_circle_move_32.png");
    icons.setTexture(IconType::SMALL_CIRCLE_PLUS_32,    ctx, path + "small_circle_plus_32.png",     "_small_circle_plus_32.png");
    icons.setTexture(IconType::SMALL_CIRCLE_OUTLINE_32, ctx, path + "small_circle_outline_32.png",  "_small_circle_outline_32.png");
    icons.setTexture(IconType::SMALL_CROSSHAIR_32,      ctx, path + "small_crosshair_32.png",       "_small_crosshair_32.png");
    icons.setTexture(IconType::CURRENT_TIME_MARKER,     ctx, path + "diamond.png",                  "_diamond.png");
}
