// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include "core/input_bindings.h"
#include "core/input_event.h"
#include "core/input_devices_state.h"
#include "core/utils.h"
#include "core/log.h"

using namespace SEAScope;

/**
 *
 */
InputBindings::InputBindings()
{

}

/**
 *
 */
InputBindings::~InputBindings()
{

}
/**
 *
 */
bool InputBindings::checkAction(const ActionId& actionId,
                                const InputEvent& event,
                                const InputDevicesState& state) const
{
    auto found = this->_bindings.find(actionId);
    if
        (this->_bindings.end() == found)
    {
        // No binding registered for this action
        return false;
    }

    // Check level to avoid string allocation
    if
        (Log::isValid(LogLevel::Debug))
    {
        LOG(LogLevel::Debug, "InputBindings", " event {}", event.str());
    }

    bool involvesPressedKeys(false);
    bool involvesPressedButtons(false);
    bool involvesResizeEvent(false);
    bool involvesAnythingEvent(false);

    //bool showDebug = (1 == actionId.getActorId() && 1 == actionId.getActionType());
    for
        (const auto& condition : found->second.getEvents())
    {
        if
            (EventType::RESIZE == condition.getType())
        {
            involvesResizeEvent = true;
        }

        if
            (EventType::ANYTHING == condition.getType())
        {
            involvesAnythingEvent = true;
        }

        if
            (  (EventType::KEY_PRESS == condition.getType())
            || (EventType::KEY_STATE_PRESSED == condition.getType()))
        {
            involvesPressedKeys = true;
        }

        if
            (  (EventType::MOUSE_BUTTON_PRESS == condition.getType())
            || (EventType::MOUSE_BUTTON_STATE_PRESSED == condition.getType()))
        {
            involvesPressedButtons = true;
        }
    }

    if
        ((false == involvesResizeEvent) && (false == involvesAnythingEvent))
    {
        if
            ((false == involvesPressedKeys) && (0 < state.getPressedKeysCount()))
        {
            return false;
        }

        if
            ((false == involvesPressedButtons) && (0 < state.getPressedButtonsCount()))
        {
            return false;
        }
    }

    for
        (const auto& condition : found->second.getEvents())
    {
        // Check level to avoid string allocation
        if
            (Log::isValid(LogLevel::Debug))
        {
            LOG(LogLevel::Debug, "InputBindings", " check {}", condition.str());
        }

        if
            (EventType::ANYTHING == condition.getType())
        {
            /* this condition is always met*/
            continue;
        }
        else if
            (EventType::KEY_RELEASE == condition.getType())
        {
            if
                ((condition.getType() == event.getType())
                && (condition.getKeyScanCode() == event.getKeyScanCode()))
            {
                // Conditions met by new event
                continue;
            }
            return false;
        }
        else if
            (EventType::MOUSE_BUTTON_RELEASE == condition.getType())
        {
            if
                ((condition.getType() == event.getType())
                 && condition.getButton() == event.getButton())
            {
                // Conditions met by new event
                continue;
            }
            return false;
        }
        else if
            (EventType::KEY_PRESS == condition.getType())
        {
            if
                ((condition.getType() == event.getType())
                 && (condition.getKeyScanCode() == event.getKeyScanCode()))
            {
                // Condition met by new event
                continue;
            }
            return false;
        }
        else if
            (EventType::KEY_STATE_PRESSED == condition.getType())
        {
            if
                ((EventType::KEY_PRESS == event.getType())
                 && (condition.getKeyScanCode() == event.getKeyScanCode()))
            {
                // Condition met by new event
                continue;
            }
            else if
                (true == state.isKeyPressed(condition.getKeyScanCode()))
            {
                // Condition met by state
                continue;
            }
            else
            {
                // Condition not met
                return false;
            }
        }
        else if
            (EventType::KEY_STATE_RELEASED == condition.getType())
        {
            if
                ((EventType::KEY_RELEASE == event.getType())
                 && (condition.getKeyScanCode() == event.getKeyScanCode()))
            {
                // Condition met by new event
                continue;
            }
            else if
                (false == state.isKeyPressed(condition.getKeyScanCode()))
            {
                // Condition met by state
                continue;
            }
            else
            {
                // Condition not met
                return false;
            }
        }
        else if
            (EventType::MOUSE_BUTTON_STATE_PRESSED == condition.getType())
        {
            if
                ((EventType::MOUSE_BUTTON_PRESS == event.getType())
                 && (condition.getButton() == event.getButton()))
            {
                // Condition met by new event
                continue;
            }
            else if
                (true == state.isButtonPressed(condition.getButton()))
            {
                // Condition met by state
                continue;
            }
            else
            {
                // Condition not met
                return false;
            }
        }
        else if
            (EventType::MOUSE_BUTTON_STATE_RELEASED == condition.getType())
        {
            if
                ((EventType::MOUSE_BUTTON_RELEASE == event.getType())
                 && (condition.getButton() == event.getButton()))
            {
                // Condition met by new event
                continue;
            }
            else if
                (false == state.isButtonPressed(condition.getButton()))
            {
                // Condition met by state
                continue;
            }
            else
            {
                // Condition not met
                return false;
            }
        }
        else if
            (EventType::MOUSE_BUTTON_PRESS == condition.getType())
        {
            if
                ((condition.getType() == event.getType())
                 && (condition.getButton() == event.getButton()))
            {
                // Condition met by new event
                continue;
            }
            return false;
        }
        else if
            (EventType::MOUSE_BUTTON_CLICK == condition.getType())
        {
            if
                ((condition.getType() == event.getType())
                 && (condition.getButton() == event.getButton()))
            {
                // Condition met by new event
                continue;
            }
            return false;
        }
        else if
            ((EventType::MOUSE_WHEEL_CHANGE == condition.getType())
             && (EventType::MOUSE_WHEEL_CHANGE == event.getType()))
        {
            double conditionDx(0.0);
            double conditionDy(0.0);
            double dx(0.0);
            double dy(0.0);

            conditionDx = condition.getMouseWheelDx();
            conditionDy = condition.getMouseWheelDy();

            if
                (0.0 != conditionDx)
            {
                LOG(LogLevel::Debug, "InputBindings",
                         "Wheel DX {} {}", conditionDx, dx);

                dx = event.getMouseWheelDx();
                if
                    (std::isnan(conditionDx) && (0.0 == dx))
                {
                    // Condition not met: change along X axis
                    return false;
                }
                else if
                    (0.0 >= dx * conditionDx)
                {
                    // There has been no change along the X axis or the directions do not match
                    return false;
                }
            }

            if
                (0.0 != conditionDy)
            {
                LOG(LogLevel::Debug, "InputBindings",
                         "Wheel DY {} {}", conditionDy, dy);

                dy = event.getMouseWheelDy();
                if
                    (std::isnan(conditionDy) && (0.0 == dy))
                {
                    // Condition not met: change along the Y axis
                    return false;
                }
                else if
                    (0.0 >= dy * conditionDy)
                {
                    // There has been no change along the Y axis or the directions do not match
                    return false;
                }
            }

            // Condition met by new event
            continue;
        }
        else if
            ((EventType::MOUSE_MOVE == condition.getType())
             && (EventType::MOUSE_MOVE == event.getType()))
        {
            double conditionDx(0.0);
            double conditionDy(0.0);
            double dx(0.0);
            double dy(0.0);

            conditionDx = condition.getMouseDx();
            conditionDy = condition.getMouseDy();

            // Handle the "any move" case
            if
                (std::isnan(conditionDx) && std::isnan(conditionDy))
            {
                // The system triggered a MOUSE_MOVE event, so there has been
                // a move, no need to check any further
                return true;
            }

            if
                (0.0 != conditionDx)
            {
                LOG(LogLevel::Debug, "InputBindings",
                         "Move DX {} {}", conditionDx, dx);

                dx = event.getMouseDx();
                if
                    (std::isnan(conditionDx) && (0.0 == dx))
                {
                    // Condition not met: change along X axis
                    return false;
                }
                else if
                    (0.0 >= dx * conditionDx)
                {
                    // There has been no change along the X axis or the directions do not match
                    return false;
                }
            }

            if
                (0.0 != conditionDy)
            {
                LOG(LogLevel::Debug, "InputBindings",
                         "Move DY {} {}", conditionDy, dy);

                dy = event.getMouseDy();
                if
                    (std::isnan(conditionDy) && (0.0 == dy))
                {
                    // Condition not met: change along the Y axis
                    return false;
                }
                else if
                    (0.0 >= dy * conditionDy)
                {
                    // There has been no change along the Y axis or the directions do not match
                    return false;
                }
            }

            // Condition met by new event
            continue;

        }
        else if
            ((EventType::RESIZE == condition.getType())
             && (EventType::RESIZE == event.getType()))
        {
            LOG(LogLevel::Debug, "InputBindings", "RESIZE");
            return true;
        }
        else
        {
            return false;
        }
    }

    return true;
}

/**
 *
 */
bool InputBindings::registerAction(const ActionId& actionId,
                                   const std::string& defaultBindingsDefinition,
                                   Synchronicity sync)
{
    /* If an entry exists in the _synchronicities map, then the action has
       already been registered. Checking with the _bindings map would not
       provide relevant results because the entry could have been created by
       loading user-defined bindings even though the actor does not exist.*/
    auto found = this->_synchronicities.find(actionId);
    if
        (this->_synchronicities.end() != found)
    {
        // already registered
        return false;
    }

    this->_bindings.emplace(actionId, InputBinding());
    this->_synchronicities.emplace(actionId, sync);

    this->translateBindingDefinition(defaultBindingsDefinition,
                                     this->_bindings[actionId]);

    return true;
}

/**
 *
 */
void InputBindings::setBinding(const ActionId& /*actionId*/, InputBinding&& /*binding*/)
{
    /**this->_bindings[actionId] = std::move(binding);*/
}

/**
 *
 */
bool InputBindings::hasValidBinding(const ActionId& actionId) const
{
    auto found = this->_bindings.find(actionId);
    if
        (this->_bindings.end() == found)
    {
        return false;
    }

    return found->second.isValid();
}


/**
 *
 */
bool InputBindings::translateBindingDefinition(const std::string& definition, InputBinding& result) const
{
    std::vector<std::string> tokens;    //

    result.clear();

    SEAScope::explodeString(definition, " ", tokens);
    for
        (const auto& token : tokens)
    {
        if
            (0 >= token.size())
        {
            LOG(LogLevel::Warn, "InputBindings",
                     "Binding definition contains empty tokens (extra spaces)");
            continue;
        }

        if
            (1 >= token.size())
        {
            LOG(LogLevel::Error, "InputBindings",
                     "Binding definition contains an incomplete token: {}",
                     token);
            continue;
        }

        std::string actionStr;
        std::string targetStr;

        actionStr = token.front();
        targetStr = token.substr(1);

        LOG(LogLevel::Debug, "InputBindings",
                 "Extracted token '{} {} {}'", token, actionStr, targetStr);

        InputEvent event;   //

        if
            (">" == actionStr)
        {
            if
                (0 == targetStr.find("BUTTON"))
            {
                std::string buttonNumStr;
                int buttonNum(-1);

                buttonNumStr = targetStr.substr(6);
                buttonNum = std::stoi(buttonNumStr, nullptr);

                event.setType(EventType::MOUSE_BUTTON_PRESS);
                event.setButton(buttonNum);
            }
            else
            {
                auto found = this->_keyCodes.find(targetStr);
                int scanCode(-1);

                if
                    (this->_keyCodes.end() == found)
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Binding definition uses an unknown key: {}",
                             targetStr);
                    return false;
                }

                scanCode = found->second;

                if
                    (-1 >= scanCode)
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Binding definition uses a key which has no scancode on this platform: {}",
                             targetStr);
                    return false;
                }
                event.setType(EventType::KEY_PRESS);
                event.setKeyScanCode(scanCode);
            }
        }
        else if
            ("<" == actionStr)
        {
            if
                (0 == targetStr.find("BUTTON"))
            {
                std::string buttonNumStr;
                int buttonNum(-1);

                buttonNumStr = targetStr.substr(6);
                buttonNum = std::stoi(buttonNumStr, nullptr);

                event.setType(EventType::MOUSE_BUTTON_RELEASE);
                event.setButton(buttonNum);
            }
            else
            {
                auto found = this->_keyCodes.find(targetStr);
                int scanCode(-1);

                if
                    (this->_keyCodes.end() == found)
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Binding definition uses an unknown key: {}",
                             targetStr);
                    return false;
                }

                scanCode = found->second;

                if
                    (-1 >= scanCode)
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Binding definition uses a key which has no scancode on this platform: {}",
                             targetStr);
                    return false;
                }
                event.setType(EventType::KEY_RELEASE);
                event.setKeyScanCode(scanCode);
            }
        }
        else if
            ("}" == actionStr)
        {
            if
                (0 == targetStr.find("BUTTON"))
            {
                std::string buttonNumStr;
                int buttonNum(-1);

                buttonNumStr = targetStr.substr(6);
                buttonNum = std::stoi(buttonNumStr, nullptr);

                event.setType(EventType::MOUSE_BUTTON_STATE_PRESSED);
                event.setButton(buttonNum);
            }
            else
            {
                auto found = this->_keyCodes.find(targetStr);
                int scanCode(-1);

                if
                    (this->_keyCodes.end() == found)
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Binding definition uses an unknown key: {}",
                             targetStr);
                    return false;
                }

                scanCode = found->second;

                if
                    (-1 >= scanCode)
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Binding definition uses a key which has no scancode on this platform: {}",
                             targetStr);
                    return false;
                }
                event.setType(EventType::KEY_STATE_PRESSED);
                event.setKeyScanCode(scanCode);
            }
        }
        else if
            ("{" == actionStr)
        {
            if
                (0 == targetStr.find("BUTTON"))
            {
                std::string buttonNumStr;
                int buttonNum(-1);

                buttonNumStr = targetStr.substr(6);
                buttonNum = std::stoi(buttonNumStr, nullptr);

                event.setType(EventType::MOUSE_BUTTON_STATE_RELEASED);
                event.setButton(buttonNum);
            }
            else
            {
                auto found = this->_keyCodes.find(targetStr);
                int scanCode(-1);

                if
                    (this->_keyCodes.end() == found)
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Binding definition uses an unknown key: {}",
                             targetStr);
                    return false;
                }

                scanCode = found->second;

                if
                    (-1 >= scanCode)
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Binding definition uses a key which has no scancode on this platform: {}",
                             targetStr);
                    return false;
                }
                event.setType(EventType::KEY_STATE_RELEASED);
                event.setKeyScanCode(scanCode);
            }
        }
        else if
            ("!" == actionStr)
        {
            if
                (0 == targetStr.find("BUTTON"))
            {
                std::string buttonNumStr;
                int buttonNum(-1);

                buttonNumStr = targetStr.substr(6);
                buttonNum = std::stoi(buttonNumStr, nullptr);

                event.setType(EventType::MOUSE_BUTTON_CLICK);
                event.setButton(buttonNum);
            }
            else
            {
                LOG(LogLevel::Error, "InputBindings",
                         "Bindings for click events require a BUTTON");
                return false;
            }
        }
        else if
            ("*" == actionStr)
        {
            event.setType(EventType::MOUSE_WHEEL_CHANGE);
            std::string axisStr;
            std::string remainingStr;
            std::string directionStr;
            double direction(std::numeric_limits<double>::quiet_NaN());

            axisStr = targetStr.front();
            remainingStr = targetStr.substr(1);
            if
                (0 < remainingStr.size())
            {
                directionStr = remainingStr.front();
                if
                    ("+" == directionStr)
                {
                    direction = 1.0;
                }
                else if
                    ("-" == directionStr)
                {
                    direction = -1.0;
                }
                else
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Unsupported direction in binding definiton: '{}' ('{}')",
                             directionStr, token);
                    return false;
                }
            }

            if
                ("X" == axisStr)
            {
                event.setMouseWheelDx(direction);
            }
            else if
                ("Y" == axisStr)
            {
                event.setMouseWheelDy(direction);
            }
            else
            {
                LOG(LogLevel::Error, "InputBindings",
                         "Unsupported axis in binding definition: '{}' ('{}')",
                         axisStr, token);
                return false;
            }
        }
        else if
            ("~" == actionStr)
        {
            event.setType(EventType::MOUSE_MOVE);
            std::size_t offset(0);
            while
                (offset < targetStr.size())
            {
                std::string axisStr;
                std::string remainingStr;
                std::string directionStr;
                double direction(std::numeric_limits<double>::quiet_NaN());

                axisStr = targetStr.substr(offset, 1);
                remainingStr = targetStr.substr(offset + 1);
                if
                    (0 < remainingStr.size())
                {
                    directionStr = remainingStr.front();
                    if
                        ("+" == directionStr)
                    {
                        direction = 1.0;
                    }
                    else if
                        ("-" == directionStr)
                    {
                        direction = -1.0;
                    }
                    else if
                        ("|" != directionStr)
                    {
                        LOG(LogLevel::Error, "InputBindings",
                                 "Unsupported direction in binding definiton: '{}' ('{}')",
                                 directionStr, token);
                        return false;
                    }

                    offset = offset + 2;  // axis + direction
                }
                else
                {
                    offset = offset + 1;  // only axis
                }

                if
                    ("X" == axisStr)
                {
                    event.setMouseDx(direction);
                }
                else if
                    ("Y" == axisStr)
                {
                    event.setMouseDy(direction);
                }
                else
                {
                    LOG(LogLevel::Error, "InputBindings",
                             "Unsupported axis in binding definition: '{}' ('{}')",
                             axisStr, token);
                    return false;
                }
            }
        }
        else if
            ("#" == actionStr)
        {
            event.setType(EventType::RESIZE);
        }
        else if
            ("_" == actionStr)
        {
            event.setType(EventType::ANYTHING);
        }
        else
        {
            LOG(LogLevel::Error, "InputBindings",
                    "Unknown action type: {}", actionStr);
            return false;
        }

        result.addEvent(std::move(event));
    }

    return true;
}

/**
 *
 */
bool InputBindings::getBindingDescription(const ActionId& actionId, std::string& result) const
{
    std::ostringstream buffer("");

    auto found = this->_bindings.find(actionId);
    if
        (this->_bindings.end() == found)
    {
        return false;
    }

    for
        (const auto& condition : found->second.getEvents())
    {
        if
            (0 < buffer.str().size())
        {
            buffer << ", and ";
        }

        if
            (EventType::KEY_RELEASE == condition.getType())
        {
            std::string key("UNKNOWN KEY");
            this->getKeyName(condition.getKeyScanCode(), key);
            buffer << "release the '" << key << "' key";
        }
        else if
            (EventType::MOUSE_BUTTON_RELEASE == condition.getType())
        {
            std::string buttonName("");
            this->getMouseButtonName(condition.getButton(), buttonName);
            buffer << "release the " << buttonName << " mouse button";
        }
        else if
            (EventType::KEY_PRESS == condition.getType())
        {
            std::string key("UNKNOWN KEY");
            this->getKeyName(condition.getKeyScanCode(), key);
            buffer << "press the '" << key << "' key";
        }
        else if
            (EventType::KEY_STATE_PRESSED == condition.getType())
        {
            std::string key("UNKNOWN KEY");
            this->getKeyName(condition.getKeyScanCode(), key);
            buffer << "hold the '" << key << "' key pressed";
        }
        else if
            (EventType::KEY_STATE_RELEASED == condition.getType())
        {
            std::string key("UNKNOWN KEY");
            this->getKeyName(condition.getKeyScanCode(), key);
            buffer << "do not press nor hold the '" << key << "' key";
        }
        else if
            (EventType::MOUSE_BUTTON_STATE_PRESSED == condition.getType())
        {
            std::string buttonName("");
            this->getMouseButtonName(condition.getButton(), buttonName);
            buffer << "hold the " << buttonName << " mouse button pressed";
        }
        else if
            (EventType::MOUSE_BUTTON_STATE_RELEASED == condition.getType())
        {
            std::string buttonName("");
            this->getMouseButtonName(condition.getButton(), buttonName);
            buffer << "do not press nor hold the " << buttonName << " mouse button";
        }
        else if
            (EventType::MOUSE_BUTTON_PRESS == condition.getType())
        {
            std::string buttonName("");
            this->getMouseButtonName(condition.getButton(), buttonName);
            buffer << "press the " << buttonName << " mouse button";
        }
        else if
            (EventType::MOUSE_BUTTON_CLICK == condition.getType())
        {
            std::string buttonName("");
            this->getMouseButtonName(condition.getButton(), buttonName);
            buffer << "click with the " << buttonName << " mouse button";
        }
        else if
            (EventType::MOUSE_WHEEL_CHANGE == condition.getType())
        {
            double conditionDx(0.0);
            double conditionDy(0.0);

            conditionDx = condition.getMouseWheelDx();
            conditionDy = condition.getMouseWheelDy();

            if
                (std::isnan(conditionDx))
            {
                buffer << "scroll along the X axis using the mouse wheel";
            }
            else if
                (0.0 < conditionDx)
            {
                buffer << "scroll up along the X axis using the mouse wheel";
            }
            else if
                (0.0 > conditionDx)
            {
                buffer << "scroll down along the X axis using the mouse wheel";
            }

            if
                (std::isnan(conditionDy))
            {
                buffer << "scroll along the Y axis using the mouse wheel";
            }
            else if
                (0.0 < conditionDy)
            {
                buffer << "scroll up along the Y axis using the mouse wheel";
            }
            else if
                (0.0 > conditionDy)
            {
                buffer << "scroll down along the Y axis using the mouse wheel";
            }
        }
        else if
            (EventType::MOUSE_MOVE == condition.getType())
        {
            double conditionDx(0.0);
            double conditionDy(0.0);

            conditionDx = condition.getMouseDx();
            conditionDy = condition.getMouseDy();

            // Handle the "any move" case
            if
                (std::isnan(conditionDx) && std::isnan(conditionDy))
            {
                buffer << "move the mouse cursor";
                continue;
            }

            std::ostringstream direction("");

            if
                (0.0 < conditionDy)
            {
                direction << "top";
            }
            else if
                (0.0 > conditionDy)
            {
                direction << "bottom";
            }

            if
                (0.0 < conditionDx)
            {
                if
                    (0 < direction.str().size())
                {
                    direction << "-";
                }
                direction << "right";
            }
            else if
                (0.0 > conditionDx)
            {
                if
                    (0 < direction.str().size())
                {
                    direction << "-";
                }
                direction << "left";
            }

            buffer << "move the mouse cursor to the " << direction.str();
        }
    }

    result = buffer.str();

    return true;
}

/**
 *
 */
void InputBindings::getKeyName(int keyScanCode, std::string& result) const
{
    for
        (const auto& it: this->_keyCodes)
    {
        if
            (it.second == keyScanCode)
        {
            result = it.first;
            break;
        }
    }
}

/**
 *
 */
void InputBindings::getMouseButtonName(int button, std::string& result) const
{
    if
        (0 == button)
    {
        result = "left";
    }
    else if
        (1 == button)
    {
        result = "right";
    }
    else if
        (2 == button)
    {
        result = "middle";
    }
    else
    {
        std::ostringstream buffer("");
        buffer << "'BUTTON_" << button << "'";
        result = buffer.str();
    }
}
