// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/input_devices_state.h"

#include <array>
#include <cstdlib>

using namespace SEAScope;

/**
 *
 */
InputDevicesState::InputDevicesState()
: _mouseX(0.0), _mouseY(0.0), _pressedKeysCount(0), _pressedButtonsCount(0)
{

}

/**
 *
 */
InputDevicesState::~InputDevicesState()
{

}

/**
 *
 */
double InputDevicesState::getMouseX() const
{
    return this->_mouseX;
}

/**
 *
 */
void InputDevicesState::setMouseX(double x)
{
    this->_mouseX = x;
}

/**
 *
 */
double InputDevicesState::getMouseY() const
{
    return this->_mouseY;
}

/**
 *
 */
void InputDevicesState::setMouseY(double y)
{
    this->_mouseY = y;
}

/**
 *
 */
bool InputDevicesState::isKeyPressed(int scancode) const
{
    auto found = this->_keys.find(scancode);
    if
        (this->_keys.end() == found)
    {
        return false;
    }
    return found->second;
}

/**
 *
 */
void InputDevicesState::setKeyPressed(int scancode, bool pressed)
{
    if
        (pressed == this->_keys[scancode])
    {
        return;
    }

    this->_keys[scancode] = pressed;

    if
        ((false == pressed) && (0 < this->_pressedKeysCount))
    {
        this->_pressedKeysCount = this->_pressedKeysCount - 1u;
    }
    else
    {
        this->_pressedKeysCount = this->_pressedKeysCount + 1u;
    }
}

/**
 *
 */
bool InputDevicesState::isButtonPressed(int button) const
{
    auto found = this->_buttons.find(button);
    if
        (this->_buttons.end() == found)
    {
        return false;
    }
    return found->second;
}

/**
 *
 */
void InputDevicesState::setButtonPressed(int button, bool pressed)
{
    if
        (pressed == this->_buttons[button])
    {
        return;
    }

    this->_buttons[button] = pressed;
    if
        ((false == pressed) && (0 < this->_pressedButtonsCount))
    {
        this->_pressedButtonsCount = this->_pressedButtonsCount - 1u;
    }
    else
    {
        this->_buttonsPressedPosition[button] = { {static_cast<int>(this->_mouseX), static_cast<int>(this->_mouseY)} };
        //std::cout << "Setting position for mouse button " << button << " pressed to " << this->_buttonsPressedPosition[button][0] << ", " << this->_buttonsPressedPosition[button][1] << std::endl;

        this->_pressedButtonsCount  =this->_pressedButtonsCount + 1u;
    }
}

/**
 *
 */
std::uint16_t InputDevicesState::getPressedKeysCount() const
{
    return this->_pressedKeysCount;
}

/**
 *
 */
std::uint16_t InputDevicesState::getPressedButtonsCount() const
{
    return this->_pressedButtonsCount;
}

/**
 *
 */
bool InputDevicesState::hasMouseMovedSinceButtonPress(int button, std::uint32_t pixelsTolerance) const
{
    bool xChanged(false);
    bool yChanged(false);
    int xDiff(0);
    int yDiff(0);

    auto found = this->_buttonsPressedPosition.find(button);
    if
        (this->_buttonsPressedPosition.end() != found)
    {
        xDiff = static_cast<int>(this->_mouseX) - found->second[0];
        yDiff = static_cast<int>(this->_mouseY) - found->second[1];
        xChanged = pixelsTolerance < static_cast<std::uint32_t>(abs(xDiff));
        yChanged = pixelsTolerance < static_cast<std::uint32_t>(abs(yDiff));
    }

    return (xChanged || yChanged);
}

/**
 *
 */
void InputDevicesState::reset()
{
    this->_buttonsPressedPosition.clear();
    this->_pressedKeysCount = 0;
    this->_pressedButtonsCount = 0;
    for
        (auto& button : this->_buttons)
    {
        button.second = false;
    }
    for
        (auto& key : this->_keys)
    {
        key.second = false;
    }
}

