// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include "core/input_event.h"

#include <iostream>

using namespace SEAScope;

/**
 *
 */
InputEvent::InputEvent()
: _type(EventType::NONE),
  _mouseDx(0.0),
  _mouseDy(0.0),
  _mouseWheelDx(0.0),
  _mouseWheelDy(0.0),
  _button(-1),
  _keyScanCode(-1),
  _width(0),
  _height(0)
{

}

/**
 *
 */
InputEvent::InputEvent(InputEvent&& src) noexcept
: _type(src._type),
  _mouseDx(src._mouseDx),
  _mouseDy(src._mouseDy),
  _mouseWheelDx(src._mouseWheelDx),
  _mouseWheelDy(src._mouseWheelDy),
  _button(src._button),
  _keyScanCode(src._keyScanCode),
  _width(src._width),
  _height(src._height)
{

}

/**
 *
 */
InputEvent::~InputEvent()
{

}

/**
 *
 */
std::string InputEvent::str() const
{
    std::ostringstream serialized("");
    switch
        (this->_type)
    {
        case EventType::NONE:
            serialized << "NONE";
            break;
        case EventType::MOUSE_MOVE:
            serialized << "MOUSE MOVE " << this->_mouseDx << " " << this->_mouseDy;
            break;
        case EventType::MOUSE_WHEEL_CHANGE:
            serialized << "MOUSE WHEEL CHANGE " << this->_mouseWheelDx << " " << this->_mouseWheelDy;
            break;
        case EventType::MOUSE_BUTTON_PRESS:
            serialized << "MOUSE BUTTON PRESS " << this->_button;
            break;
        case EventType::MOUSE_BUTTON_REPEAT:
            serialized << "MOUSE BUTTON REPEAT " << this->_button;
            break;
        case EventType::MOUSE_BUTTON_RELEASE:
            serialized << "MOUSE BUTTON RELEASE " << this->_button;
            break;
        case EventType::MOUSE_BUTTON_CLICK:
            serialized << "MOUSE BUTTON CLICK " << this->_button;
            break;
        case EventType::KEY_PRESS:
            serialized << "KEY PRESS " << this->_keyScanCode;
            break;
        case EventType::KEY_REPEAT:
            serialized << "KEY REPEAT " << this->_keyScanCode;
            break;
        case EventType::KEY_RELEASE:
            serialized << "KEY RELEASE " << this->_keyScanCode;
            break;
        case EventType::MOUSE_BUTTON_STATE_PRESSED:
            serialized << "MOUSE BUTTON STATE PRESSED " << this->_button;
            break;
        case EventType::MOUSE_BUTTON_STATE_RELEASED:
            serialized << "MOUSE BUTTON STATE RELEASED " << this->_button;
            break;
        case EventType::KEY_STATE_PRESSED:
            serialized << "KEY STATE PRESSED " << this->_keyScanCode;
            break;
        case EventType::KEY_STATE_RELEASED:
            serialized << "KEY STATE RELEASED " << this->_keyScanCode;
            break;
        case EventType::RESIZE:
            serialized << "RESIZE " << this->_width << "x" << this->_height;
            break;
		case EventType::CLOSE:
            serialized << "CLOSE ";
            break;
        case EventType::ANYTHING:
            serialized << "ANYTHING";
            break;
    }

    return serialized.str();
}

/**
 *
 */
EventType InputEvent::getType() const
{
    return this->_type;
}

/**
 *
 */
void InputEvent::setType(EventType type)
{
    this->_type = type;
}

/**
 *
 */
double InputEvent::getMouseDx() const
{
    return this->_mouseDx;
}

/**
 *
 */
void InputEvent::setMouseDx(double dx)
{
    this->_mouseDx = dx;
}

/**
 *
 */
double InputEvent::getMouseDy() const
{
    return this->_mouseDy;
}

/**
 *
 */
void InputEvent::setMouseDy(double dy)
{

    this->_mouseDy = dy;
}

/**
 *
 */
double InputEvent::getMouseWheelDx() const
{
    return this->_mouseWheelDx;
}

/**
 *
 */
void InputEvent::setMouseWheelDx(double value)
{
    this->_mouseWheelDx = value;
}

/**
 *
 */
double InputEvent::getMouseWheelDy() const
{
    return this->_mouseWheelDy;
}

/**
 *
 */
void InputEvent::setMouseWheelDy(double value)
{
    this->_mouseWheelDy = value;
}

/**
 *
 */
int InputEvent::getButton() const
{
    return this->_button;
}

/**
 *
 */
void InputEvent::setButton(int value)
{
    this->_button = value;
}

/**
 *
 */
int InputEvent::getKeyScanCode() const
{
    return this->_keyScanCode;
}

/**
 *
 */
void InputEvent::setKeyScanCode(int value)
{
    this->_keyScanCode = value;
}

/**
 *
 */
int InputEvent::getWidth() const
{
    return this->_width;
}

/**
 *
 */
void InputEvent::setWidth(int width)
{
    this->_width = width;
}

/**
 *
 */
int InputEvent::getHeight() const
{
    return this->_height;
}

/**
 *
 */
void InputEvent::setHeight(int height)
{
    this->_height = height;
}
