// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/input_events_dispatcher.h"

#include "core/action_id.h"
#include "core/input_bindings.h"
#include "core/interfaces/iinput_events_subscriber.h"
#include "core/log.h"


using namespace SEAScope;

/**
 *
 */
InputEventsDispatcher::InputEventsDispatcher()
: _clickPixelsTolerance(0)
{
}

/**
 *
 */
InputEventsDispatcher::~InputEventsDispatcher()
{
}


/**
 *
 */
const InputDevicesState& InputEventsDispatcher::getInputDevicesState() const
{
    return this->_inputDevicesState;
}

/**
 *
 */
bool InputEventsDispatcher::addSubscriber(IInputEventsSubscriber* subscriber)
{
    ActorId actorId = subscriber->getId();

    /* For now all actions are synchronous but the subscriber should be able
       to define some actions as asynchronous in the future */
    for
        (const auto& actionType : subscriber->getAllActions())
    {
        LOG(LogLevel::Debug, "InputEventsDispatcher",
                 "Registering action {}", actionType);

        bool ok(false); //
        std::string definition; //

        ok = subscriber->getDefaultBindingDefinition(actionType, definition);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "InputEventsDispatcher",
                     "Default bindings not found for action {} on actor {}",
                     actionType, actorId);
            return false;
        }

        auto actionId = ActionId(actorId, actionType);
        this->getInputBindings().registerAction(actionId, definition,
                                                Synchronicity::SYNC);
    }

    this->_subscribers.push_back(subscriber);
    return true;
}

/**
 *
 */
const std::vector<IInputEventsSubscriber*>& InputEventsDispatcher::getSubscribers() const
{
    return this->_subscribers;
}

/**
 *
 */
bool InputEventsDispatcher::dispatchEvent(const InputEvent& event)
{
    bool preemptEvents(false);

    for
        (auto& subscriber : this->_subscribers)
    {
        if
            (false == subscriber->isActive())
        {
            LOG(LogLevel::Debug, "InputEventsDispatcher", "Skipped inactive subscriber {}", subscriber->getId());
            continue;
        }

        for
            (const auto& actionType : subscriber->getPossibleActions())
        {
            bool isActionValid(false);
            auto actionId = ActionId(subscriber->getId(), actionType);
            isActionValid = this->getInputBindings().checkAction(actionId, event,
                                                                 this->_inputDevicesState);

            if
                (Log::isValid(LogLevel::Debug))
            {
                std::string buffer("");
                subscriber->getActionLabel(actionType, buffer);
                LOG(LogLevel::Debug, "InputEventsDispatcher",
                    "Dispatch action '{}'? {}", buffer, isActionValid);
            }

            if
                (false == isActionValid)
            {
                continue;
            }

            //LOG(LogLevel::Error, "InputEventsDispatcher", "Conditions met for action {}:{}", actionId.getActorId(), actionId.getActionType());

            subscriber->performAction(actionType, event, this->_inputDevicesState);

            /* TODO: handle asynchronous actions too */
        }

        if
            (subscriber->preemptEvents())
        {
            std::string actorLabel;
            subscriber->getActorLabel(actorLabel);
        }
        preemptEvents = preemptEvents || subscriber->preemptEvents();
    }

    return preemptEvents;
}

/**
 *
 */
void InputEventsDispatcher::loadBindings(const std::string& /*path*/)
{
    // DEBUG
}

/**
 *
 */
void InputEventsDispatcher::saveBindings(const std::string& /*path*/) const
{

}

/**
 *
 */
std::uint32_t InputEventsDispatcher::getClickPixelsTolerance() const
{
    return this->_clickPixelsTolerance;
}

/**
 *
 */
bool InputEventsDispatcher::setClickPixelsTolerance(std::uint32_t value)
{
    if
        (value == this->_clickPixelsTolerance)
    {
        return false;
    }

    this->_clickPixelsTolerance = value;
    return true;
}
