// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cmath>
#include <sstream>

#include "core/measurements.h"
#include "3d/polyline_base.h"
#include "3d/point.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"

#include "core/log.h"
#include "num_utils.h"


/**
 *
 */
bool SEAScope::getOuterRing(const Shape& shape, std::vector<GCP>& result)
{
	std::size_t i = 0;	//
	std::size_t j = 0;	//

    /*
	LOG(SEAScope::LogLevel::Warn,
                       "Measurements", "Outer ring size: {}x{}",
                       shape.xArity, shape.yArity);
    */
	if
		((1 >= shape.xArity) || (1 >= shape.yArity))
	{
		/* Not a polygon */
		return false;
	}

	// Bottom
	j = 0;
	for
		(i = 0; i < shape.xArity; ++i)
	{
        std::size_t index {j * shape.xArity + i};
		result.emplace_back(shape.gcpsTime[index],
                            shape.gcpsLat[index],
                            shape.gcpsLon[index],
                            shape.gcpsMinorAxisIndex[index],
                            shape.gcpsMajorAxisIndex[index]);
	}

	// Right
	//LOG(SEAScope::LogLevel::Warn, "Measurements", "RIGHT");
	i = static_cast<size_t>(shape.xArity - 1);
	for
		(j = 1; j < shape.yArity; ++j)  // start from 1 since index 0 has already been computed in previous loop
	{
        std::size_t index {j * shape.xArity + i};
		result.emplace_back(shape.gcpsTime[index],
                            shape.gcpsLat[index],
                            shape.gcpsLon[index],
                            shape.gcpsMinorAxisIndex[index],
                            shape.gcpsMajorAxisIndex[index]);
	}

	// Top
	//LOG(SEAScope::LogLevel::Warn, "Measurements", "TOP");
	j = static_cast<size_t>(shape.yArity - 1);
	for
		(i = static_cast<size_t>(shape.xArity - 1); i-- > 0;)  // start from arity - 2 (-1 then --) since index arity - 1 has already been computed in previous loop
	{
        std::size_t index {j * shape.xArity + i};
		result.emplace_back(shape.gcpsTime[index],
                            shape.gcpsLat[index],
                            shape.gcpsLon[index],
                            shape.gcpsMinorAxisIndex[index],
                            shape.gcpsMajorAxisIndex[index]);
	}

	// Left
	//LOG(SEAScope::LogLevel::Warn, "Measurements", "LEFT");
	i = 0;
	for
		(j = static_cast<size_t>(shape.yArity - 1); j-- > 1;) // start from arity - 2 (-1 then --) since index arity - 1 has already been computed in previous loop
	{
        std::size_t index {j * shape.xArity + i};
		result.emplace_back(shape.gcpsTime[index],
                            shape.gcpsLat[index],
                            shape.gcpsLon[index],
                            shape.gcpsMinorAxisIndex[index],
                            shape.gcpsMajorAxisIndex[index]);
	}

	return true;
}

/**
 *
 */
bool SEAScope::haversineDist(double lat1, double lon1, double lat2, double lon2,
						     double& dist)
{
	double a;	//
	double c;	//
	double lonDist = lon2 - lon1;	// Longitude distance (radians)
	double latDist = lat2 - lat1;	// Latitude distance (radians)
	a = std::sin(0.5 * latDist) * std::sin(0.5 * latDist) + std::cos(lat1) * std::cos(lat2) * std::sin(0.5 * lonDist) * std::sin(0.5 * lonDist);
	c = 2 * std::atan2(std::sqrt(a), std::sqrt(1 - a));
	dist = SEAScope::EARTH_MEAN_RADIUS_IN_METER * c; // Distance in m
	return true;
}

/**
 * https://en.wikipedia.org/wiki/Vincenty%27s_formulae
 * https://www.movable-type.co.uk/scripts/latlong-vincenty.html
 */
bool SEAScope::vincentyDist(double lat1, double lon1, double lat2, double lon2,
                            double& dist, double threshold)
{
	// WGS84 ellipsoid
	constexpr double a = 6378137.0; /* semi-major axis */
	constexpr double f = 1.0 / 298.257223563; /* flattening: (a - b) / a */
	constexpr double b = (1.0 - f) * a; /* semi-minor axis */

	double dlon = lon2 - lon1; // L = difference in longitude (radians)
	//U = reduced latitude, defined by tan U = (1-f)·tanφ.
	double tanU1 = (1.0 - f) * std::tan(lat1);
	double cosU1 = 1.0 / std::sqrt((1.0 + tanU1 * tanU1));
	double sinU1 = tanU1 * cosU1;
	double tanU2 = (1 - f) * std::tan(lat2);
	double cosU2 = 1.0 / std::sqrt((1.0 + tanU2 * tanU2));
	double sinU2 = tanU2 * cosU2;

	bool antipodal = std::abs(dlon) > 0.5 * SEAScope::PI || std::abs(lat2 - lat1) > 0.5 * SEAScope::PI;

	double dlonAux = dlon; // λ = difference in longitude on an auxiliary sphere
	double sinDlonAux = 0.0;
	double cosDlonAux = 0.0;
	double angularDist = antipodal ? SEAScope::PI : 0.0; // σ = angular distance P₁ P₂ on the sphere
	double sinAngularDist = 0.0;
	double cosAngularDist = antipodal ? -1.0 : 1.0;
	double sin2AngularDist = 0.0;
	double cosAngularDistMid = 1.0;                      // σM = angular distance on the sphere from the equator to the midpoint of the line
	double sinAzimuth = 0.0;
	double cos2Azimuth = 1;         // α = azimuth of the geodesic at the equator
	double C = 0.0;
	double epsilon = std::numeric_limits<double>::epsilon();

	double oldDlonAux = 0.0;
	int iterations = 0;
	do {
		sinDlonAux = std::sin(dlonAux);
		cosDlonAux = std::cos(dlonAux);
		sin2AngularDist= (cosU2 * sinDlonAux) * (cosU2 * sinDlonAux) + (cosU1 * sinU2 - sinU1 * cosU2 * cosDlonAux) * (cosU1 * sinU2 - sinU1 * cosU2 * cosDlonAux);
		if
			(std::abs(sin2AngularDist) < epsilon)
		{
			break;  // co-incident/antipodal points (falls back on λ/σ = L)
		}
		sinAngularDist = std::sqrt(sin2AngularDist);
		cosAngularDist = sinU1 * sinU2 + cosU1 * cosU2 * cosDlonAux;
		angularDist = std::atan2(sinAngularDist, cosAngularDist);
		sinAzimuth = cosU1 * cosU2 * sinDlonAux / sinAngularDist;
		cos2Azimuth = 1.0 - sinAzimuth * sinAzimuth;
		cosAngularDistMid = (false == floating_point_equality(cos2Azimuth, 0.0)) ? (cosAngularDist - 2 * sinU1 * sinU2 / cos2Azimuth) : 0.0; // on equatorial line cos²α = 0 (§6)
		C = f / 16 * cos2Azimuth * (4 + f * (4 - 3 * cos2Azimuth));
		oldDlonAux = dlonAux;
		dlonAux = dlon + (1 - C) * f * sinAzimuth * (angularDist + C * sinAngularDist * (cosAngularDistMid + C * cosAngularDist * (-1 + 2 * cosAngularDistMid * cosAngularDistMid)));
		double iterationCheck = antipodal ? std::abs(dlonAux) - SEAScope::PI : std::abs(dlonAux);
		if
			(iterationCheck > SEAScope::PI)
		{
			// 'λ > π'
			return false;
		}
	} while (std::abs(dlonAux - oldDlonAux) > threshold && ++iterations < 1000);


	if
		(iterations >= 1000)
	{
		// Vincenty formula failed to converge
		return false;
	}

	double uSq = cos2Azimuth * (a * a - b * b) / (b * b);
	double A = 1.0 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)));
	double B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)));
	double angularDistDiff = B * sinAngularDist * (cosAngularDistMid + B / 4 * (cosAngularDist * (-1 + 2 * cosAngularDistMid * cosAngularDistMid) - B / 6 * cosAngularDistMid * (-3 + 4 * sinAngularDist * sinAngularDist) * (-3 + 4 * cosAngularDistMid * cosAngularDistMid)));

	dist = b * A * (angularDist - angularDistDiff); // s = length of the geodesic
	return true;
}

/**
 *
 */
double SEAScope::distance(double lat1, double lon1, double lat2, double lon2)
{
	bool ok;	//
	double result = 0.0;	//

	ok  = SEAScope::vincentyDist(SEAScope::PI * lat1 / 180.0,
								 SEAScope::PI * lon1 / 180.0,
								 SEAScope::PI * lat2 / 180.0,
								 SEAScope::PI * lon2 / 180.0,
								 result, 1E-12);
	if
		(false == ok)
	{
		LOG(SEAScope::LogLevel::Warn, "Measurements",
						   "Vincenty inverse formula failed, using haversine as fallback");

		// Fallback on haversine
		ok = SEAScope::haversineDist(SEAScope::PI * lat2 / 180.0,
									 SEAScope::PI * lon2 / 180.0,
									 SEAScope::PI * lat2 / 180.0,
									 SEAScope::PI * lon2 / 180.0,
									 result);
		// cannot fail
	}

	/*
	ok = SEAScope::vincentiDist(SEAScope::PI * 50.06632 / 180.0,
                                SEAScope::PI * -5.71475 / 180.0,
                                SEAScope::PI * 58.64402 / 180.0,
                                SEAScope::PI * -3.07009 / 180.0, dist, 1E-12);
	LOG(SEAScope::LogLevel::Error, "Measurements", "Vincenty test: {} {}", ok, dist);
	*/
	return result;
}

/**
 *
 */
double SEAScope::polarTriangleArea(double tan1, double lon1, double tan2, double lon2)
{
	double deltaLon = lon1 - lon2;
	double t = tan1 * tan2;
	return 2 * std::atan2(t * std::sin(deltaLon), 1 + t * std::cos(deltaLon));
}

/**
 *
 */
double SEAScope::computeSignedArea(const std::vector<double>& lats, const std::vector<double>& lons)
{
	double partialResult = 0.0;	//
	double result = 0.0;	//
	std::size_t size = lats.size();	//
	if
		(3 > size)
	{
		/* Not a polygon */
		return std::numeric_limits<double>::quiet_NaN();
	}

	double prevTanLat = std::tan(0.5 * SEAScope::PI * (0.5 - lats[size -1] / 180.0));
	double prevLon = SEAScope::PI * lons[size - 1] / 180.0;
	for
		(std::size_t i = 0; i < size; ++i)
	{
		double tanLat = std::tan(0.5 * SEAScope::PI * (0.5 - lats[i] / 180.0));
		double lon = SEAScope::PI * lons[i] / 180.0;
		partialResult = SEAScope::polarTriangleArea(tanLat, lon, prevTanLat, prevLon);
        //LOG(SEAScope::LogLevel::Warn, "Measurements",
        //                   "point: {} {} -> {} {} = {}", 180*prevLon/SEAScope::PI, -180 * (((std::atan(prevTanLat)) / (0.5 * SEAScope::PI)) - 0.5), lons[i], lats[i], partialResult);
		prevTanLat = tanLat;
		prevLon = lon;
		//LOG(SEAScope::LogLevel::Warn, "Measurements", "PolarTriangle: {} {} -> {}", lats[i], lons[i], partialResult);
		result = result + partialResult;
	}

	return result * SEAScope::EARTH_MEAN_RADIUS_IN_METER * SEAScope::EARTH_MEAN_RADIUS_IN_METER;
}

double SEAScope::perimeterFromShape(const OGL::Shape* shape, bool linearRing)
{
    std::vector<Geodetic2D> geos;   //
    const PolylineBase* polyline = nullptr; //
	double dist = 0.0;	//
	double currentDist = 0.0;	//
	double lastLon = 0.0;	//
	double lastLat = 0.0;	//
	std::uint32_t point_ind = 0;	//

    polyline = dynamic_cast<const PolylineBase*>(shape);
    if
        (nullptr == polyline)
    {
        return std::numeric_limits<double>::quiet_NaN();
    }

    const std::vector<glm::vec3>& xyz = polyline->getPoints();
    geos.reserve(xyz.size());

    for (const glm::vec3& p : xyz)
    {
            geos.push_back(UnitSphere::ToGeodetic2D(p));
    }

    if
        (  true == linearRing
        && 0 < xyz.size()
        && geos[0] != geos[xyz.size() - 1])
    {
        geos.push_back(UnitSphere::ToGeodetic2D(xyz[0]));
    }

    //LOG(SEAScope::LogLevel::Warn, "Measurements", "geos: {} {} -> {} {}", geos[0].Longitude(), geos[0].Latitude(), geos[geos.size() - 1].Longitude(), geos[geos.size() - 1].Latitude());
    for
        (const auto& point : geos)
    {
        if
            (0 < point_ind)
        {
            currentDist = SEAScope::distance(lastLat, lastLon,
                                             180.0 * point.Latitude() / SEAScope::PI,
                                             180.0 * point.Longitude() / SEAScope::PI);
            //LOG(SEAScope::LogLevel::Warn, "Measurements",
            //                   "point: {} {} -> {} {} = {}", lastLon, lastLat, 180.0 * point.Longitude() / SEAScope::PI, 180.0 * point.Latitude() / SEAScope::PI, currentDist);
            dist += currentDist;
        }

        lastLon = 180.0 * point.Longitude() / SEAScope::PI;
        lastLat = 180.0 * point.Latitude() / SEAScope::PI;
        point_ind += 1;
    }
    polyline = nullptr;

    return dist;
}

double SEAScope::areaFromShape(const OGL::Shape* shape)
{
    std::vector<Geodetic2D> geos;   //
    const PolylineBase* polyline = nullptr; //
	std::vector<double> lats;	//
	std::vector<double> lons;	//
	double signedArea = 0.0;	//

    polyline = dynamic_cast<const PolylineBase*>(shape);
    if
        (nullptr == polyline)
    {
        return std::numeric_limits<double>::quiet_NaN();
    }

    const std::vector<glm::vec3>& xyz = polyline->getPoints();
    geos.reserve(xyz.size());

    for (const glm::vec3& p : xyz)
    {
            geos.push_back(UnitSphere::ToGeodetic2D(p));
    }

    for
        (const auto& point : geos)
    {
        lons.push_back(180.0 * point.Longitude() / SEAScope::PI);
        lats.push_back(180.0 * point.Latitude() / SEAScope::PI);
    }
    polyline = nullptr;

	signedArea = SEAScope::computeSignedArea(lats, lons);
	if
		(std::isnan(signedArea))
	{
		return std::numeric_limits<double>::quiet_NaN();
	}
	return std::abs(signedArea);
}

/**
 *
 */
double SEAScope::areaFromShape(const SEAScope::Shape& shape)
{
	std::vector<double> lats;	//
	std::vector<double> lons;	//
	std::vector<GCP> outerGCPs; //
	double signedArea;	//
	bool ok;

	ok = SEAScope::getOuterRing(shape, outerGCPs);
	if
		(false == ok)
	{
		// Not a polygon
		return std::numeric_limits<double>::quiet_NaN();
	}

	for
		(const auto& gcp : outerGCPs)
	{
		lats.push_back(static_cast<double>(gcp.lat));
		lons.push_back(static_cast<double>(gcp.lon));
	}

	signedArea = SEAScope::computeSignedArea(lats, lons);
	if
		(std::isnan(signedArea))
	{
		return std::numeric_limits<double>::quiet_NaN();
	}
	return std::abs(signedArea);
}

/**
 *
 */
double SEAScope::perimeterFromShape(const SEAScope::Shape& shape)
{
	double dist = 0.0;	//
	double currentDist = 0.0;	//
	double lastLon = 0.0;	//
	double lastLat = 0.0;	//
	std::uint32_t gcp_ind = 0;	//
	std::vector<GCP> outerGCPs; //
	bool ok;

	// For polygons, only take the outer GCPs
	ok = SEAScope::getOuterRing(shape, outerGCPs);
	if
		(true == ok)
	{
		for
			(const auto& gcp: outerGCPs)
		{
			if
				(0 < gcp_ind)
			{
				currentDist = SEAScope::distance(lastLat, lastLon, static_cast<double>(gcp.lat), static_cast<double>(gcp.lon));
				dist += currentDist;
			}

			lastLon = static_cast<double>(gcp.lon);
			lastLat = static_cast<double>(gcp.lat);
			gcp_ind += 1;
		}
	}
	else
	{
        const std::size_t maxIndex = shape.gcpsLat.size();
		for
			(std::size_t index = 0; index < maxIndex; ++index)
		{
			if
				(0 < gcp_ind)
			{
				currentDist = SEAScope::distance(lastLat, lastLon, static_cast<double>(shape.gcpsLat[index]), static_cast<double>(shape.gcpsLon[index]));
				dist += currentDist;
			}

			lastLon = static_cast<double>(shape.gcpsLon[index]);
			lastLat = static_cast<double>(shape.gcpsLat[index]);
			gcp_ind += 1;
		}
	}

	return dist;
}

/**
 *
 */
bool SEAScope::formatDistance(double meters, std::string& result)
{
    std::ostringstream buffer;

    if
        (std::isnan(meters))
    {
        return false;
    }

    buffer.precision(3);
    buffer << std::fixed;

    if
        (1000 < meters)
    {
        buffer << meters / 1000 << "km";
    }
    else
    {
        buffer << meters << "m";
    }

    result = buffer.str();
    return true;
}

/**
 *
 */
bool SEAScope::formatArea(double squareMeters, std::string& result)
{
    std::ostringstream buffer;

    if
        (std::isnan(squareMeters))
    {
        return false;
    }

    buffer.precision(3);
    buffer << std::fixed;

    if
        (1000000 < squareMeters)
    {
        buffer << squareMeters / 1000000 << "km²";
    }
    else
    {
        buffer << squareMeters << "m²";
    }

    result = buffer.str();
    return true;
}

/**
 *
 */
bool SEAScope::formatLongitude(double longitude, std::string& result)
{
    std::ostringstream buffer;

    if
        (std::isnan(longitude))
    {
        return false;
    }

    buffer.precision(5);
    buffer << std::fixed;

    buffer << longitude ;
    buffer << " ( ";

    if
        (0 > longitude)
    {
        buffer << "-";
        longitude = -longitude;
    }

    double degrees;
    double minutes;
    double seconds;
    double remainder;
    remainder = 60.0 * modf(longitude, &degrees);
    remainder = 60.0 * modf(remainder, &minutes);
    remainder = 60.0 * modf(remainder, &seconds);

    buffer.precision(0);
    buffer << degrees << "° " << minutes << "' " << seconds << "\" )";

    result = buffer.str();
    return true;
}

/**
 *
 */
bool SEAScope::formatLatitude(double latitude, std::string& result)
{
    std::ostringstream buffer;

    if
        (std::isnan(latitude))
    {
        return false;
    }

    buffer.precision(5);
    buffer << std::fixed;

    buffer << latitude;
    buffer << " ( ";

    if
        (0 > latitude)
    {
        buffer << "-";
        latitude = -latitude;
    }

    double degrees;
    double minutes;
    double seconds;
    double remainder;
    remainder = 60.0 * modf(latitude, &degrees);
    remainder = 60.0 * modf(remainder, &minutes);
    remainder = 60.0 * modf(remainder, &seconds);

    buffer.precision(0);
    buffer << degrees << "° " << minutes << "' " << seconds << "\" )";


    result = buffer.str();
    return true;
}

/**
 *
 */
void SEAScope::getShapeMeasurements(OGL::Shape* shape, double& area, double& perimeter, double& lon, double& lat)
{
    perimeter = std::numeric_limits<double>::quiet_NaN();
    area = std::numeric_limits<double>::quiet_NaN();
    lon = std::numeric_limits<double>::quiet_NaN();
    lat = std::numeric_limits<double>::quiet_NaN();

    if
        (nullptr == shape)
    {
        return;
    }

    if
        (SEAScope::ShapeType::POLYLINE == shape->getType())
    {
        perimeter = SEAScope::perimeterFromShape(shape, false /* do not join last and first vertices */);

        const PolylineBase* polyline = dynamic_cast<const PolylineBase*>(shape);
        const std::vector<glm::vec3>& xyz = polyline->getPoints();
        if
            (   (2 < xyz.size())
             && (xyz[0] == xyz[xyz.size() - 1]))
        {
            area = SEAScope::areaFromShape(shape);
        }

        polyline = nullptr;
    }
    else if
        (SEAScope::ShapeType::POLYGON == shape->getType())
    {
        area = SEAScope::areaFromShape(shape);
        perimeter = SEAScope::perimeterFromShape(shape, true /* close shape */);
    }
    else if
        (SEAScope::ShapeType::ELLIPSE == shape->getType())
    {
        area = SEAScope::areaFromShape(shape);
        perimeter = SEAScope::perimeterFromShape(shape, true /* close shape */);
    }
    else if
        (SEAScope::ShapeType::POINT == shape->getType())
    {
        glm::vec3 position = static_cast<Point*>(shape)->getPosition();
        Geodetic2D geo = UnitSphere::ToGeodetic2D(position);
        lon = 180.0 * geo.Longitude() / SEAScope::PI;
        lat = 180.0 * geo.Latitude() / SEAScope::PI;
    }
}
