// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>

#include "core/log.h"
#include "core/metadata_mediator.h"
#include "core/dynamic.h"
#include "core/interfaces/iindex_service_connector.h"
#include "core/granule_metadata.h"
#include "debug.h"

using namespace SEAScope;

/**
 *
 */
MetadataMediator::MetadataMediator(): store(nullptr), variableCfgAdded(nullptr), sources(nullptr)
{
};

/**
 *
 */
MetadataMediator::~MetadataMediator()
{
};

/**
 *
 */
bool MetadataMediator::init(IMetadataStore* storeIn, SourcesControlModel* sourcesIn)
{
	this->store = storeIn;
	this->sources = sourcesIn;
	return true;
};

/**
 *
 */
bool MetadataMediator::registerUserGeneratedIndex(const SourceId& sourceId, IIndexServiceConnector* index)
{
    auto found = this->_userGeneratedIndexes.find(sourceId);
    if
        (this->_userGeneratedIndexes.end() != found)
    {
        return false;
    }

    this->_userGeneratedIndexes[sourceId] = index;

    return true;
}

IIndexServiceConnector* MetadataMediator::getUserGeneratedIndex(const SourceId& sourceId, bool& ok) const
{
    auto userGeneratedFound = this->_userGeneratedIndexes.find(sourceId);
    if
        (this->_userGeneratedIndexes.end() != userGeneratedFound)
    {
        ok = true;
        return userGeneratedFound->second;
    }

    ok = false;
    return nullptr;
}

/**
 *
 */
bool MetadataMediator::collectionCfgLoaded(const CollectionId& collectionId, const CollectionCfg& cfg,  const std::vector<RenderingCfg*>& rCfgs)
{
	bool ok(false);	//

	ok = this->store->setCollectionCfg(collectionId, cfg);
    if
        (false == ok)
    {
        return false;
    }
	if
		(nullptr != this->variableCfgAdded)
	{
        const float keyFactor {10000.f};  //
        const std::uint32_t binSize {100}; //
		auto it = rCfgs.begin();
		for
			(; it != rCfgs.end(); it++)
		{
            float key {0.f};    //
            key = (*it)->zindex() * keyFactor;

            std::uint32_t baseIndex {0u};   //
            baseIndex = static_cast<std::uint32_t>(std::floor(key));

            auto baseIndexFound = this->_zindexHistogram.find(baseIndex);
            if
                (this->_zindexHistogram.end() == baseIndexFound)
            {
                this->_zindexHistogram[baseIndex] = binSize - 1u;
            }
            else if
                (0 < this->_zindexHistogram[baseIndex])
            {
                this->_zindexHistogram[baseIndex] -= 1u;
            }
            else
            {
                LOG(LogLevel::Warn, "Metadata mediator",
                    "Too many collections with zindex {}, ordering collections may produce unpredictable results", (*it)->zindex());
            }

            // Compute a unique value based on the original zindex value
            std::uint32_t uniqueIndex {0u}; //
            uniqueIndex = baseIndex * binSize + this->_zindexHistogram[baseIndex];

            // Switch back to the original values range for zindex
            float newZindex {0.f};  //
            newZindex = static_cast<float>(uniqueIndex) / (keyFactor * static_cast<float>(binSize));

            // Update rendering configuration with new zindex value
            (*it)->setZindex(newZindex);

			LOG( LogLevel::Debug, "Metadata mediator"
					, "Registering collection rendering cfg {}/{}:{} {}"
					, collectionId.first, collectionId.second, (*it)->target.variable, (*it)->colormap());

			(*it)->target.collection.first = collectionId.first; // transform into a global id
			ok = this->variableCfgAdded(*it);
			if
				(false == ok)
			{
				LOG( LogLevel::Error, "Metadata mediator"
						, "Error while registering collection rendering cfg");
				return false;
			}
		}
	}
	return true;
};

/**
 *
 */
bool MetadataMediator::getCollectionCfg(const CollectionId& collectionId, CollectionCfg& cfg)
{
	dynamic<IIndexServiceConnector>* catalogue(nullptr);	//
	std::vector<RenderingCfg*> rCfgs;	//
	bool ok(false);	//

	ok = this->store->getCollectionCfg(collectionId, cfg);
	if
		(true == ok)
	{
		return true;
	}

	if
		(1 == collectionId.first)
	{
		/* The source with ID 1 corresponds to the user-generated data and can
			only exist in store/cache */
		return false;
	}

	/* TODO: find a way to handle async calls */

	catalogue = this->sources->getIndex(collectionId.first, ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "MetadataMediator"
				, "Could not find a suitable catalogue plugin for source {}"
				, collectionId.first);
		return false;
	}

	ok = (*catalogue)->getCollectionCfg(collectionId.second, cfg, rCfgs);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "MetadataMediator"
				, "Could not get collection {} configuration from catalogue"
				, collectionId.second);
		return false;
	}

	LOG( LogLevel::Debug, "MetadataMediator"
			, "Collection configuration retrieved");
	Debug::displayCollection(cfg);

	this->collectionCfgLoaded(collectionId, cfg, rCfgs);
	return ok;
};

/**
 *
 */
bool MetadataMediator::getVariableCfg(const CollectionId& collectionId, const VariableId& variableId, VariableCfg& result)
{
	bool ok(false);	//
	CollectionCfg colCfg;	//

	ok = this->getCollectionCfg(collectionId, colCfg);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "MetadataMediator"
				, "Could not retrieve collection cfg associated with the variable");
		return false;
	}

	auto found = colCfg.variables.find(variableId);
	if
		(colCfg.variables.end() == found)
	{
		LOG( LogLevel::Error, "MetadataMediator"
				, "No configuration for variable {} in {}/{}"
				, variableId, collectionId.first, collectionId.second);
		return false;
	}

	result = found->second;
	return true;
};

/**
 *
 */
bool MetadataMediator::getFields( const CollectionId& collectionId
								, const std::set<VariableId>& vars
								, std::set<std::string>& fields)
{
	bool ok(false);	//
	CollectionCfg colCfg;	//

	ok = this->getCollectionCfg(collectionId, colCfg);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "MetadataMediator"
				, "Could not retrieve collection information for {}/{}"
				, collectionId.first, collectionId.second);
		return false;
	}

	for
		(auto& variable : vars)
	{
		auto found = colCfg.variables.find(variable);
		if
			(colCfg.variables.end() == found)
		{
			LOG( LogLevel::Error, "MetadataMediator"
					, "Variable {} not found for collection {}/{}"
					, variable, collectionId.first, collectionId.second);
			return false;
		}

		for
			(auto& field : found->second.fields)
		{
			fields.insert(field);
		}
	}

	return true;
};

/**
 *
 */
bool MetadataMediator::getFields( const CollectionVariable& colvar, std::set<std::string>& fields)
{
	bool ok(false);	//
	CollectionCfg colCfg;	//

	ok = this->getCollectionCfg(colvar.first, colCfg);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "MetadataMediator"
				, "Could not retrieve collection information for {}/{}"
				, colvar.first.first, colvar.first.second);
		return false;
	}

	auto found = colCfg.variables.find(colvar.second);
	if
		(colCfg.variables.end() == found)
	{
		LOG( LogLevel::Error, "MetadataMediator"
				, "Variable {} not found for collection {}/{}"
				, colvar.second, colvar.first.first, colvar.first.second);
		return false;
	}

	for
		(auto& field : found->second.fields)
	{
		fields.insert(field);
	}

	return true;
};

/**
 *
 */
bool MetadataMediator::searchPrevNext( timestamp_t current
									 , const std::set<CollectionId>& collections
									 , timestamp_t& prev
									 , timestamp_t& next
									 , timestamp_t minDelta)
{
	std::map<SourceId, std::set<CollectionId> > localCollections;	//
	dynamic<IIndexServiceConnector>* index(nullptr);	//
	std::vector<timestamp_t> prevs;	//
	std::vector<timestamp_t> nexts;	//
	std::array<bool, 2> oks;	//
	bool ok(false);	//

	prev = current;
	next = current;

	LOG( LogLevel::Debug, "MetadataMediator"
			, "searchPrevNext: {}"
			, current);

	ok = this->collectionsBySource(collections, localCollections);
	LOG( LogLevel::Debug, "MetadataMediator"
			, "Local collections list built");

	for
		(const auto& itSrc : localCollections)
	{
		timestamp_t srcPrev;	//
		timestamp_t srcNext;	//

		LOG(LogLevel::Debug, "MetadataMediator", "add source to search: {}", itSrc.first);

        IIndexServiceConnector* userGeneratedIndex(nullptr);
        userGeneratedIndex = this->getUserGeneratedIndex(itSrc.first, ok);
        if
            (true == ok)
        {
			/* user-generated data are stored are not handled by plugins */
			oks = userGeneratedIndex->searchPrevNext(current,
												     itSrc.second,
												     srcPrev,
													 srcNext,
													 minDelta);
		}
		else
		{

			index = this->sources->getIndex(itSrc.first, ok);
			if
				(false == ok)
			{
				LOG( LogLevel::Error, "MetadataMediator"
						, "Could not find a suitable index plugin for source {}"
						, itSrc.first);
				return false;
			}

			LOG( LogLevel::Debug, "MetadataMediator"
					, "Index found for source {}"
					, itSrc.first);

			oks = (*index)->searchPrevNext( current
										 , itSrc.second
										 , srcPrev
										 , srcNext
										 , minDelta);
		}

		LOG( LogLevel::Debug, "MetadataMediator"
				, "Plugin responded: {} [{}] {} [{}]"
				, srcPrev, oks[0], srcNext, oks[1]);
		if
			(true == oks[0])
		{
			LOG(LogLevel::Debug, "MetadataMediator", "Stack prev: {}", current - srcPrev);
			prevs.push_back(current - srcPrev);
		}

		if
			(true == oks[1])
		{
			LOG(LogLevel::Debug, "MetadataMediator", "Stack next: {}", srcNext - current);
			nexts.push_back(srcNext - current);
		}
	}

	timestamp_t minPrev = std::numeric_limits<timestamp_t>::max();
	for
		(const auto& it : prevs)
	{
		if
			(it < minPrev)
		{
			minPrev = it;
			prev = current - minPrev;
		}
	}

	timestamp_t minNext = std::numeric_limits<timestamp_t>::max();
	for
		(const auto& it : nexts)
	{
		if
			(it < minNext)
		{
			minNext = it;
			next = current + minNext;
		}
	}

	LOG( LogLevel::Debug, "MetadataMediator"
			, "prev/next: {} {}"
			, prev, next);
	return true;
};

/**
 *
 */
bool MetadataMediator::searchClosestBound( timestamp_t current
									 , const std::set<CollectionId>& collections
									 , timestamp_t& closest
                                     , const timestamp_t searchLowerBound
                                     , const timestamp_t searchUpperBound)
{
	std::map<SourceId, std::set<CollectionId> > localCollections;	//
	dynamic<IIndexServiceConnector>* index(nullptr);	//
	std::vector<timestamp_t> closests;	//
	timestamp_t minDiff;
	bool ok(false);	//

	closest = current;

	LOG( LogLevel::Debug, "MetadataMediator"
			, "searchPrevNext: {}"
			, current);

	ok = this->collectionsBySource(collections, localCollections);
	LOG( LogLevel::Debug, "MetadataMediator"
			, "Local collections list built");

	for
		(const auto& itSrc : localCollections)
	{
		timestamp_t srcClosest;	//

		LOG(LogLevel::Debug, "MetadataMediator", "add source to search: {}", itSrc.first);

        IIndexServiceConnector* userGeneratedIndex(nullptr);
        userGeneratedIndex = this->getUserGeneratedIndex(itSrc.first, ok);
        if
            (true == ok)
        {
			ok = userGeneratedIndex->searchClosestBound(current,
														itSrc.second,
														srcClosest,
                                                        searchLowerBound,
                                                        searchUpperBound);

		}
		else
		{
			index = this->sources->getIndex(itSrc.first, ok);
			if
				(false == ok)
			{
				LOG( LogLevel::Error, "MetadataMediator"
						, "Could not find a suitable index plugin for source {}"
						, itSrc.first);
				return false;
			}

			LOG( LogLevel::Debug, "MetadataMediator"
					, "Index found for source {}"
					, itSrc.first);

			ok = (*index)->searchClosestBound( current
											 , itSrc.second
											 , srcClosest
                                             , searchLowerBound
                                             , searchUpperBound);
		}

		LOG( LogLevel::Debug, "MetadataMediator"
				, "Plugin responded: {} [{}]"
				, srcClosest, ok);
		if
			(true == ok)
		{
			closests.push_back(srcClosest);
		}
	}

	minDiff = std::numeric_limits<timestamp_t>::max();
	for
		(const auto& it: closests)
	{
		LOG(LogLevel::Debug, "MetadataMediator", "\tCandidate for closest: {}", it);
        if
            (it == current)
        {
            closest = current;
            break;
        }
		if
			(it > current
			&& it - current < minDiff)
		{
			minDiff = it - current;
			closest = it;
		}
		else if
			(  it < current
			&& current - it < minDiff)
		{
			minDiff = current - it;
			closest = it;
		}
	}

	LOG( LogLevel::Debug, "MetadataMediator"
			, "closest bound: {}"
			, closest);
	return true;
};

/**
 *
 */
bool MetadataMediator::searchGranules( timestamp_t current
									 , timestamp_t start
									 , timestamp_t stop
									 , float zoom
									 , const std::vector<std::pair<CollectionId, bool> >& collections
									 , std::map<CollectionId, std::vector<GranuleMetadata> >& result)
{
	std::map<SourceId, std::map<std::uint32_t, bool> > localCollections;	//
	dynamic<IIndexServiceConnector>* index(nullptr);	//
	bool ok(false);	//

    for
        (const auto& it : collections)
    {
        result[it.first];   // Create entry
    }

	/* Sort collections by source */
	for
		(const auto& it: collections)
	{
		auto found = localCollections.find(it.first.first);
		if
			(localCollections.end() == found)
		{
			std::map<std::uint32_t, bool> localCfg;
			localCfg[it.first.second] = it.second;
			localCollections[it.first.first] = localCfg;
		}
		else
		{
			found->second[it.first.second] = it.second;
		}
	}

	/* Request granules for each source */
	for
		(const auto& itSrc : localCollections)
	{
		std::vector<GranuleMetadata> localResults;
		SearchCriteria criteria;

		criteria.current = current;
		criteria.start = start;
		criteria.stop = stop;
		criteria.resolution = zoom;
		criteria.collections = itSrc.second;

        IIndexServiceConnector* userGeneratedIndex(nullptr);
        userGeneratedIndex = this->getUserGeneratedIndex(itSrc.first, ok);
        if
            (true == ok)
        {
			ok = userGeneratedIndex->searchGranules(criteria, localResults);
		}
		else
		{
			index = this->sources->getIndex(itSrc.first, ok);
			if
				(false == ok)
			{
				LOG( LogLevel::Error, "MetadataMediator"
						, "Could not find a suitable index plugin for source {}"
						, itSrc.first);
				return false;
			}

			ok = (*index)->searchGranules(criteria, localResults);
		}

		for
			(auto& it: localResults)
		{
            /*
			LOG( LogLevel::Debug, "MetadataMediator"
					, " *> Granule found: {}"
					, it.uris[0].uri);
            */

			CollectionId colId(itSrc.first, it.collectionId());
			it.sourceId(itSrc.first); // Make global identifier
			auto found = result.find(colId);
			if
				(result.end() == found)
			{
				std::vector<GranuleMetadata> colGranules;
				colGranules.push_back(it);
				result[colId] = colGranules;
			}
			else
			{
				result[colId].push_back(it);
			}
		}
	}

	return 0 < result.size();
};

/**
 *
 */
bool MetadataMediator::getDataCoverage( const std::set<CollectionId>& collections
									  , EventsCoverage& result)
{
	std::map<SourceId, std::set<CollectionId> > localCollections;	//
	dynamic<IIndexServiceConnector>* index(nullptr);	//
	bool ok(false);	//

    result.reset();
    /*
	for
		(const auto& it: collections)
	{
		LOG(LogLevel::Debug, "MetadataMediator", "getCoverage for collection {}/{}", it.first, it.second);
	}
    */

	ok = this->collectionsBySource(collections, localCollections);

	for
		(const auto& itSrc : localCollections)
	{
		EventsCoverage localResults;

        IIndexServiceConnector* userGeneratedIndex(nullptr);
        userGeneratedIndex = this->getUserGeneratedIndex(itSrc.first, ok);
        if
            (true == ok)
        {
			ok = userGeneratedIndex->getDataCoverage(itSrc.second, localResults);
		}
		else
		{
			index = this->sources->getIndex(itSrc.first, ok);
			if
				(false == ok)
			{
				LOG( LogLevel::Error, "MetadataMediator"
						, "Could not find a suitable index plugin for source {}"
						, itSrc.first);
				return false;
			}

			ok = (*index)->getDataCoverage(itSrc.second, localResults);
		}

		if
			(false == ok)
		{
			LOG( LogLevel::Error, "MetadataMediator"
					, "Could not retrieve data coverage from source {}"
					, itSrc.first);
			return false;
		}

		localResults.display();

		ok = result.merge(localResults);
	}

	return true;
};

/**
 *
 */
bool MetadataMediator::getDataCoverage( const CollectionId& collectionId
									  , EventsCoverage& result)
{
	dynamic<IIndexServiceConnector>* index(nullptr);	//
	bool ok(false);	//

    result.reset();

    IIndexServiceConnector* userGeneratedIndex(nullptr);
    userGeneratedIndex = this->getUserGeneratedIndex(collectionId.first, ok);
    if
        (true == ok)
    {
		ok = userGeneratedIndex->getDataCoverage(collectionId, result);
	}
	else
	{
		index = this->sources->getIndex(collectionId.first, ok);
		if
			(false == ok)
		{
			LOG( LogLevel::Error, "MetadataMediator"
					, "Could not find a suitable index plugin for source {}"
					, collectionId.first);
			return false;
		}

		ok = (*index)->getDataCoverage(collectionId, result);
	}

	if
		(false == ok)
	{
		LOG( LogLevel::Error, "MetadataMediator"
				, "Could not retrieve data coverage from source {}"
				, collectionId.first);
		return false;
	}

	return true;
};

/**
 *
 */
bool MetadataMediator::collectionsBySource( const std::set<CollectionId>& collections
										  , std::map<SourceId, std::set<CollectionId> >& result) const
{
	/* Build collections set for each source */
	for
		(auto& it: collections)
	{
		auto found = result.find(it.first);
		if
			(result.end() == found)
		{
			std::set<CollectionId> srcCollections;
			srcCollections.insert(it);
			result[it.first] = srcCollections;
		}
		else
		{
			found->second.insert(it);
		}
	}
	return true;
};

/**
 *
 */
bool MetadataMediator::saveVariableRenderingCfg(const RenderingCfg* cfg)
{
    bool ok {false};    //
	dynamic<IIndexServiceConnector>* catalogue(nullptr);	//

	catalogue = this->sources->getIndex(cfg->target.collection.first, ok);
    if
        (false == ok)
    {
		LOG( LogLevel::Error, "MetadataMediator"
				, "Could not find a suitable catalogue plugin for source {}"
				, cfg->target.collection.first);
        return false;
    }

    ok = (*catalogue)->saveVariableRenderingCfg(cfg);
    return ok;
}
