// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/plugins_manager.h"

using namespace SEAScope;

/**
 *
 */
PluginsManager::PluginsManager(): loader(nullptr)
{
};

/**
 *
 */
PluginsManager::~PluginsManager()
{
	this->releaseAll();
};

/**
 *
 */
bool PluginsManager::init(IPluginsLoader* loaderIn)
{
	this->loader = loaderIn;
	return true;
};

/**
 *
 */
bool PluginsManager::releaseAll()
{
	for
		(auto& plugin : this->metadata)
	{
		if
			(plugin.second.loaded)
		{
			this->loader->releasePlugin(&plugin.second);
		}
	}
	return true;
};

/**
 *
 */
PluginMeta& PluginsManager::getMeta(const std::string& path)
{
	PluginMeta meta;	//

	auto found = this->metadata.find(path);
	if
		(this->metadata.end() == found)
	{
		meta.path = path;
		this->metadata[path] = meta;
	}

	return this->metadata[path];
};

/**
 *
 */
bool PluginsManager::registerPlugin(const std::string& pluginPath)
{
	bool ok;	//

	PluginMeta& m = this->getMeta(pluginPath);

	ok = this->loader->loadPlugin(&m);
	if
		(false == ok)
	{
		return false;
	}

	ok = this->loader->getPluginInfo(&m);
	if
		(false == ok)
	{
		LOG( LogLevel::Warn, "Plugins manager"
				, "Plugin not registred, impossible to retrieve information");
		return false;
	}

	ok = this->loader->releasePlugin(&m);
	if
		(false == ok)
	{
		LOG( LogLevel::Warn, "Plugins manager"
				, "Failed to release a plugin, possible memory leak: {}"
				, pluginPath);
	}

	return true;
};

/**
 *
 */
bool PluginsManager::discoverPlugins(const std::string& root, std::set<std::string>& results)
{
	std::set<std::string> paths;	//
	bool ok;	//

	ok = this->loader->findPlugins(root, paths);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "PluginsManager"
				, "Failed to find plugins in {}"
				, root);
		return false;
	}

	// Retrieve information about the detected plugins
	for
		(const auto& pluginPath: paths)
	{
		ok = this->registerPlugin(pluginPath);
		if
			(false == ok)
		{
			continue;
		}

		results.insert(pluginPath);
	}
	return true;
};

/**
 *
 */
PluginMeta* PluginsManager::getPluginMetadata(const std::string& api, const std::string& name, const std::string& version, bool& ok)
{
	ok = false;
	for
		(auto& it : this->metadata)
	{
		if
			(  api == it.second.api
			&& name == it.second.name
			&& version == it.second.version)
		{
			ok = true;
			return &it.second;
		}
	}

	return nullptr;
};

/**
 *
 */
PluginMeta* PluginsManager::getPlugin(const std::string& api, const std::string& name, const std::string& version, bool& ok)
{
	PluginMeta* meta;	//

	meta = this->getPluginMetadata(api, name, version, ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Plugins manager"
				, "Plugin not found in metadata");
		return nullptr;
	}

	if
		(true == meta->loaded)
	{
		// Plugin already activated
		return meta;
	}

	ok = this->loader->loadPlugin(meta);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Plugins manager"
				, "Plugin could not be loaded: {}", meta->path);
		return nullptr;
	}

	ok = this->loader->getPluginInfo(meta);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Plugins manager"
				, "Could not retrieve plugin info");
		return nullptr;
	}
	meta->loaded = true;

	return meta;
};

/**
 *
 */
bool PluginsManager::pluginCreated(const std::string& api, const std::string& name, const std::string& version)
{
	PluginMeta* meta;	//
	bool ok;	//

	meta = this->getPluginMetadata(api, name, version, ok);
	if
		(false == ok)
	{
		return false;
	}

	meta->refCount += 1;
	return true;
};

/**
 *
 */
bool PluginsManager::pluginDestroyed(const std::string& api, const std::string& name, const std::string& version)
{
	PluginMeta* meta;	//
	bool ok;	//

	meta = this->getPluginMetadata(api, name, version, ok);
	if
		(false == ok)
	{
		return false;
	}

	meta->refCount -= 1;

	ok = this->loader->releasePlugin(meta);
	return ok;
};
