// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <algorithm>

#include "core/log.h"
#include "core/types.h"
#include "core/rendering_control.h"

using namespace SEAScope;

/**
 *
 */
RenderingControl::RenderingControl()
: view(nullptr), cmaps(nullptr), callback(nullptr), toLabel(nullptr),
  getUnits(nullptr)
{
};

/**
 *
 */
RenderingControl::~RenderingControl()
{
};

/**
 *
 */
bool RenderingControl::getCollectionVariable(const IRenderingWidget* widget, CollectionVariable& result) const
{
	bool ok;	//

	auto found = this->colvars.find(const_cast<IRenderingWidget*>(widget));
	if
		(this->colvars.end() == found)
	{
		ok = false;
	}
	else
	{
		result = found->second;
		ok = true;
	}

	return ok;
};

/**
 *
 */
RenderingCfg* RenderingControl::getRenderingCfg(const IRenderingWidget* widget, CollectionVariable& colvar) const
{
	RenderingCfg* cfg;	//
	bool ok;	//

	cfg = nullptr;

	ok = this->getCollectionVariable(widget, colvar);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "RenderingControl"
				, "Could not retrieve collection variable for target widget");
		return nullptr;
	}

	cfg = this->model.getCollectionCfg(colvar.first, colvar.second, ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Rendering control"
				, "Could not retrieve rendering configuration for {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return nullptr;
	}

	return cfg;
};

/**
 *
 */
bool RenderingControl::init(IRenderingControlView* viewIn, const ColormapManager* cmapsIn)
{
	if
		(nullptr == viewIn)
	{
		LOG(LogLevel::Error, "RenderingControl", "Cannot use a null view");
		return false;
	}

	this->view = viewIn;
	this->cmaps = cmapsIn;

	/* Set model callbacks */
	this->model.callback = [this] (const RenderingCfg* cfg, const RenderingParam& param) {
        CollectionVariable colvar;
        colvar.first = cfg->target.collection;
        colvar.second = cfg->target.variable;

		LOG( LogLevel::Debug, "RenderingControl", "Configuration changed for {}/{}:{}"
				, cfg->target.collection.first, cfg->target.collection.second
				, cfg->target.variable);

		if
			(false == cfg->target.granuleLevel)
		{
			auto found = this->widgets.find(colvar);
			if
				(this->widgets.end() != found)
			{
				found->second->update(const_cast<RenderingCfg*>(cfg));

				if
					(RenderingParam::RENDERED == param)
                {
                    this->updateAdvancedRenderingAfterRemoval(cfg, found->second);
                }
			}
		}

		this->advanced.update();

		if
			(nullptr != this->callback)
		{
			this->callback(cfg, param);
		}
	};

	this->model.countersCallback = [this] () {
		this->updateCounters();
	};

	/* Set view callbacks */
	this->view->setOrderCallback([this] (IRenderingWidget* target, IRenderingWidget* below, IRenderingWidget* above) -> bool {
		bool ok;	//
		ok = this->updateCollectionZindex(target, below, above);
		return ok;
	});

	this->view->setOpacityCallback([this] (IRenderingWidget* target, float opacity) -> bool {
		bool ok;	//
		ok = this->updateCollectionOpacity(target, opacity);
		return ok;
	});

	this->view->setVisibilityCallback([this] (IRenderingWidget* target, bool rendered, bool exclusive) -> bool {
		bool ok;	//
		ok = this->updateCollectionVisibility(target, rendered, exclusive);
		return ok;
	});

	this->view->setMinCallback([this] (IRenderingWidget* target, float rel) {
		this->updateCollectionExtrema(target, rel, true);
	});

	this->view->setMaxCallback([this] (IRenderingWidget* target, float rel) {
		this->updateCollectionExtrema(target, rel, false);
	});

	this->advanced.resetCallback = [this] (const RenderingCfg* cfg) {
		this->resetRenderingCfg(cfg->target);
	};

	return true;
};

/**
 *
 */
bool RenderingControl::updateCollectionZindex( IRenderingWidget* target
											 , IRenderingWidget* below
											 , IRenderingWidget* above)
{
	std::ostringstream buffer;	//
	CollectionVariable colvar;	//
	float targetZindex;	//
	float belowZindex;	//
	float aboveZindex;	//
	float newZindex;	//
	RenderingCfg* cfg;	//
	std::string label;	//
    std::string units;  //
	bool ok;	//

	/* Retrieve below info */
	belowZindex = 0.f;
	if
		(nullptr != below)
	{
		cfg = this->getRenderingCfg(below, colvar);
		if
			(nullptr == cfg)
		{
			LOG( LogLevel::Error, "Rendering control"
					, "Could not retrieve rendering configuration associated with \"below\" widget");
			return false;
		}
		belowZindex = cfg->zindex();
	}

	/* Retrieve above info */
	aboveZindex = 0.5f * std::numeric_limits<float>::max();
	if
		(nullptr != above)
	{
		cfg = this->getRenderingCfg(above, colvar);
		if
			(nullptr == cfg)
		{
			LOG( LogLevel::Error, "Rendering control"
					, "Could not retrieve rendering configuration associated with \"above\" widget");
			return false;
		}
		aboveZindex = cfg->zindex();
	}

	/* Retrieve target info */
	cfg = this->getRenderingCfg(target, colvar);
	if
		(nullptr == cfg)
	{
		LOG( LogLevel::Error, "Rendering control"
				, "Could not retrieve rendering configuration associated with \"target\" widget");
		return false;
	}
	targetZindex = cfg->zindex();

	newZindex = 0.5f * (belowZindex + aboveZindex);
	LOG( LogLevel::Debug, "RenderingControl", "Change zindex from {} to {} [{}, {}]"
			, targetZindex, newZindex, belowZindex, aboveZindex);

	ok = cfg->setZindex(newZindex);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Rendering control", "Could not set zindex for {}/{}:{} {} -> {}"
				, colvar.first.first, colvar.first.second, colvar.second, targetZindex, newZindex);
		return false;
	}

	ok = this->getVariableLabel(colvar, label);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "RenderingControl"
				, "Could not determine label for variable {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}

    units = "";
    LOG(LogLevel::Debug, "RenderingControl", "Getting units... {}", nullptr == this->getUnits);
    if
        (nullptr != this->getUnits)
    {
        ok = this->getUnits(colvar, units);
        if
            (false == ok)
        {
            LOG( LogLevel::Error, "RenderingControl"
                    , "Could not determine units for variable {}/{}:{}"
                    , colvar.first.first, colvar.first.second, colvar.second);
            return false;
        }
    }

	target->update(label, units, cfg);
	this->view->update();

	return true;
};

/**
 *
 */
bool RenderingControl::updateCollectionVisibility(IRenderingWidget* w, bool rendered, bool exclusive)
{
	RenderingCfg* cfg {nullptr};	//
	CollectionVariable colvar;	//

    cfg = this->getRenderingCfg(w, colvar);
    if
        (nullptr == cfg)
    {
        LOG( LogLevel::Error, "RenderingControl"
                , "Could not retrieve rendering configuration associated with the widget that emitted the visibility change event");
        return false;
    }
    cfg->setRendered(rendered, false);
    LOG( LogLevel::Debug, "RenderingControl", "Rendered updated for {}/{}:{}"
            , colvar.first.first, colvar.first.second, colvar.second);

    if
        (true == exclusive)
    {
        for
            (auto it : this->colvars)
        {
            bool ok {false};
            cfg = this->model.getCollectionCfg(it.second.first, it.second.second, ok);
            if
                (false == ok)
            {
                LOG( LogLevel::Error, "Rendering control"
                        , "Could not retrieve rendering configuration for {}/{}:{}"
                        , colvar.first.first, colvar.first.second, colvar.second);
                return false;
            }

            if
                (it.first != w)
            {
                cfg->setRendered(false, false);
            }
        }
    }
	return true;
};

/**
 *
 */
bool RenderingControl::updateCollectionOpacity(IRenderingWidget* target, float opacity)
{
	bool ok;	//
	RenderingCfg* cfg;	//
	CollectionVariable colvar;	//

	if
		(nullptr == target)
	{
		LOG( LogLevel::Error, "RenderingControl"
				, "updateCollectionOpacity received a null target");
		return false;
	}

	ok = this->getCollectionVariable(target, colvar);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "RenderingControl"
				, "Could not retrieve collection variable for target widget");
		return false;
	}

	cfg = this->model.getCollectionCfg(colvar.first, colvar.second, ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Rendering control"
				, "Could not retrieve rendering configuration for {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}

	ok = cfg->setOpacity(opacity);

	return ok;
};

/**
 *
 */
bool RenderingControl::updateCollectionExtrema(IRenderingWidget* target, float rel, bool targetIsMin)
{
	bool ok;	//
	RenderingCfg* cfg;	//
	CollectionVariable colvar;	//

	if
		(nullptr == target)
	{
		LOG( LogLevel::Error, "RenderingControl"
				, "updateCollectionOpacity received a null target");
		return false;
	}

	ok = this->getCollectionVariable(target, colvar);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "RenderingControl"
				, "Could not retrieve collection variable for target widget");
		return false;
	}

	cfg = this->model.getCollectionCfg(colvar.first, colvar.second, ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Rendering control"
				, "Could not retrieve rendering configuration for {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}

	ok = cfg->scrollExtrema(rel, targetIsMin);

	return ok;
}

/**
 *
 */
bool RenderingControl::addRendering(const CollectionVariable& colvar)
{
	std::string label;	//
    std::string units;  //
	IRenderingWidget* rw;	//
	RenderingCfg* cfg;	//
	bool ok;	//

	cfg = this->model.getCollectionCfg( colvar.first, colvar.second
										   , ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Rendering control"
				, "Could not find rendering configuration for {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}
	/* Add item in the view */
	rw = this->view->createRendering();
    rw->setOpacityScrollStep(0.05f);
	rw->setColormapGetter([this] (const std::string& cmapName, Colormap& result) -> bool {
		bool success;	//

		if
			(nullptr == this->cmaps)
		{
			LOG( LogLevel::Error, "RenderingControl"
					, "Colormaps manager not set for the rendering control");
			success = false;
		}
		else
		{
			success = this->cmaps->getColormap(cmapName, result);
		}
		return success;
	});

	rw->setSettingsCallback([this](IRenderingWidget* w) -> bool {
		bool success;	//
		success = this->toggleAdvancedControl(w);
		return success;
	});

	ok = this->getVariableLabel(colvar, label);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "RenderingControl"
				, "Could not determine label for variable {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}

	ok = this->getVariableUnits(colvar, units);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "RenderingControl"
				, "Could not determine units for variable {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}

	cfg->setRendered(true, true);
	rw->update(label, units, cfg);

	ok = this->view->addRendering(rw);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Rendering control"
				, "Could not add widget to view for {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}
	this->widgets.insert(std::make_pair(colvar, rw));
	this->colvars.insert(std::make_pair(rw, colvar));

	this->updateCounters();

	return ok;
};

/**
 *
 */
bool RenderingControl::getVariableLabel(const CollectionVariable& colvar, std::string& label) const
{
	std::ostringstream buffer;	//
	bool ok;	//

	ok = false;

	if
		(nullptr != this->toLabel)
	{
		ok = this->toLabel(colvar, label);
	}
	else
	{
		buffer << colvar.first.first << "/" << colvar.first.second << ":" \
		   << colvar.second;
		label = buffer.str();
		ok = true;
	}

	return ok;
};

/**
 *
 */
bool RenderingControl::getVariableUnits(const CollectionVariable& colvar, std::string& units) const
{
    bool ok;    //

    ok = true;
    units = "";
    if
        (nullptr != this->getUnits)
    {
        ok = this->getUnits(colvar, units);
    }
    return ok;
}

/**
 *
 */
void RenderingControl::updateAdvancedRenderingAfterRemoval(const RenderingCfg* cfg, IRenderingWidget* widget)
{
    if
        (  false == cfg->rendered()
        && this->advanced.current == cfg)
    {
        this->advanced.view->hide();
        this->advanced.current = nullptr;
        widget->setSettingsState(false);
    }
}

/**
 *
 */
bool RenderingControl::removeRendering(const CollectionVariable& colvar)
{
	bool ok;	//
	RenderingCfg* cfg;	//

	cfg = this->model.getCollectionCfg(colvar.first, colvar.second, ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "RenderingControl"
				, "Could not retrieve rendering configuration for {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}

	cfg->setRendered(false, true);

	/* Remove item from view */
	auto foundWidget = this->widgets.find(colvar);
	if
		(this->widgets.end() == foundWidget)
	{
		LOG( LogLevel::Error, "Rendering control", "Could not find widget for {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}

    this->updateAdvancedRenderingAfterRemoval(cfg, foundWidget->second);

	auto foundColvar = this->colvars.find(foundWidget->second);
	if
		(this->colvars.end() == foundColvar)
	{
		LOG( LogLevel::Error, "Rendering control", "Could not find  {}/{}:{} in widgets map"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}

	ok = this->view->removeRendering(foundWidget->second);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "Rendering control"
				, "Could not remove widget from view for {}/{}:{}"
				, colvar.first.first, colvar.first.second, colvar.second);
		return false;
	}

	this->widgets.erase(foundWidget);
	this->colvars.erase(foundColvar);


	return ok;
};


/**
 *
 */
bool RenderingControl::renderedVariables( const CollectionId& collectionId
										, std::set<VariableId>& vars) const
{
	RenderingCfg* cfg;	//
	bool ok;	//

	auto it = this->widgets.begin();
	for
		(; it != this->widgets.end(); it++)
	{
		if
			(  it->first.first.first != collectionId.first
			|| it->first.first.second != collectionId.second)
		{
			continue;
		}

		cfg = this->model.getCollectionCfg(it->first.first, it->first.second, ok);
		if
			(false == ok)
		{
			LOG( LogLevel::Error, "Rendering control"
					, "Could not find rendering configuration for {}/{}:{}"
					, it->first.first.first, it->first.first.second, it->first.second);
			return false;
		}
		if
			(true == cfg->rendered())
		{
			vars.insert(it->first.second);
		}
	}
	return true;
};

/**
 *
 */
bool RenderingControl::updateCounters()
{
	LOG( LogLevel::Debug, "RenderingControl", "updateCounters");
	for
		(auto& it : this->widgets)
	{
		LOG( LogLevel::Debug, "RenderingControl", "update counter for {}/{}:{}"
				, it.first.first.first, it.first.first.second, it.first.second);
		auto found = this->model.counters.find(it.first.first);
		if
			(this->model.counters.end() == found)
		{
			it.second->setCounter(0);
		}
		else
		{
			it.second->setCounter(found->second);
		}
	}
	return true;
};

/**
 *
 */
bool RenderingControl::toggleAdvancedControl(IRenderingWidget* w)
{
	CollectionVariable colvar;	//
	RenderingCfg* cfg;	//
	std::string label;	//
    std::string units;  //
	bool ok;	//

	for
		(auto& it : this->widgets)
	{
		it.second->setSettingsState(false);
	}

	cfg = this->getRenderingCfg(w, colvar);
	this->getVariableLabel(colvar, label);
	ok = this->advanced.setTarget(cfg, label);
	w->setSettingsState(ok);
	return false;
};

/**
 *
 */
bool RenderingControl::resetRenderingCfg(const RenderableId& rId)
{
	bool ok;	//
	ok = this->model.resetCollectionCfg(rId.collection, rId.variable);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "RenderingControl", "Could not reset rendering configuration");
		return false;
	}
	LOG(LogLevel::Debug, "RenderingControl", "Configuration has been reset");
	return true;
};

/**
 *
 */
IRenderingWidget* RenderingControl::getTopWidget() const
{
    IRenderingWidget* topWidget {nullptr};    //
    RenderingCfg* cfg {nullptr};    //
    float zIndex {0.0}; //
    float maxZIndex {0.0};  //
    bool ok {false};    //

    for
        (auto it : this->widgets)
    {
	    cfg = this->model.getCollectionCfg(it.first.first, it.first.second, ok);
        if
            (false == ok)
        {
            LOG( LogLevel::Error, "Rendering control"
                    , "Could not retrieve rendering configuration for {}/{}:{}"
                    , it.first.first.first, it.first.first.second, it.first.second);
            return nullptr;
        }

        zIndex = cfg->zindex();
        if
            (zIndex > maxZIndex)
        {
            topWidget = it.second;
            maxZIndex = zIndex;
        }
    }

    return topWidget;
}

/**
 *
 */
bool RenderingControl::displayOnTop(const CollectionVariable& colvar)
{
    auto found = this->widgets.find(colvar);
    if
        (this->widgets.end() == found)
    {
        LOG(LogLevel::Error, "RenderingControl", "Could not find the widget associated with collection variable {}.{}:{}",
                 colvar.first.first, colvar.first.second, colvar.second);
        return false;
    }

    IRenderingWidget* currentTop {nullptr};
    currentTop = this->getTopWidget();
    if
        (nullptr == currentTop)
    {
        LOG(LogLevel::Error, "RenderingControl", "Could not identify the variable currently displayed on top");
        return false;
    }

    if
        (found->second == currentTop)
    {
        // Already on top
        return true;
    }

    return this->updateCollectionZindex(found->second, currentTop, nullptr);
}

/**
 *
 */
void RenderingControl::exclusive(const CollectionVariable& target)
{
    std::vector<CollectionVariable> variables;  //
    bool found {false}; //

    // List all collection-variables with a widget that are not the target
    for
        (const auto& it : this->widgets)
    {
        if
            (it.first == target)
        {
            found = true;
        }
        else
        {
            variables.push_back(it.first);
        }
    }

    // Remove widgets associated with collection-variables that are not
    // rendered anymore
    for
        (const auto& it : variables)
    {
        this->removeRendering(it);
    }

    if
        (false == found)
    {
        this->addRendering(target);
    }
}
