// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/rendering_control_model.h"

using namespace SEAScope;

/**
 *
 */
RenderingControlModel::RenderingControlModel(): callback(nullptr), countersCallback(nullptr)
{
};

/**
 *
 */
RenderingControlModel::~RenderingControlModel()
{
    for
        (auto& it : this->defaultCfgs)
    {
        delete it.second;
        it.second = nullptr;
    }

    this->defaultCfgs.clear();

    for
        (auto& it : this->_volatile)
    {
        delete it;
        it = nullptr;
    }

    this->_volatile.clear();
};

/**
 *
 */
RenderingCfg* RenderingControlModel::getCfg( const CollectionId& collectionId
						  , const GranuleId& granuleId
						  , const VariableId& variableId
						  , bool& found) const
{
	RenderingCfg* result = nullptr;

	result = this->getGranuleCfg(granuleId, variableId, found);
	if
		(true == found)
	{
		return result;
	}

	result = this->getCollectionCfg(collectionId, variableId, found);
	return result;
};

/**
 *
 */
RenderingCfg* RenderingControlModel::getGranuleCfg( const GranuleId& granuleId
								 , const VariableId& variableId
								 , bool& ok) const
{
	std::pair<GranuleId, VariableId> gKey(granuleId, variableId);

	auto granuleFound = this->granules.find(gKey);
	if
		(this->granules.end() != granuleFound)
	{
		ok = true;

		return granuleFound->second;
	}

	ok = false;
	return nullptr;
};

/**
 *
 */
RenderingCfg* RenderingControlModel::getCollectionCfg( const CollectionId& collectionId
									, const VariableId& variableId
									, bool& ok) const
{
	std::pair<CollectionId, VariableId> cKey(collectionId, variableId);

	auto collectionFound = this->collections.find(cKey);
	if
		(this->collections.end() != collectionFound)
	{
		ok = true;

		return collectionFound->second;
	}

	ok = false;
	return nullptr;
};

/**
 *
 */
bool RenderingControlModel::registerCfg(RenderingCfg* cfg)
{
	bool ok;	//

	if
		(nullptr == cfg)
	{
		LOG( LogLevel::Error, "Rendering configs"
				, "Tried to register a non existant config");
		return false;
	}

    cfg->setRendered(false, true /*silent*/);

	ok = false;
	if
		(cfg->target.granuleLevel)
	{
		ok = this->registerGranuleCfg(cfg);
	}
	else
	{
		ok = this->registerCollectionCfg(cfg);
	}

	if
		(true == ok)
	{
		cfg->callback = [this](const RenderingCfg* cfgTmp, const RenderingParam& param) {
			this->cfgChanged(cfgTmp, param);
		};
	}

	return ok;
};

/**
 *
 */
bool RenderingControlModel::registerGranuleCfg(RenderingCfg* cfg)
{
	bool ok;	//

	ok = false;
	std::pair<GranuleId, VariableId> key( cfg->target.granule
										, cfg->target.variable);
	auto found = this->granules.find(key);
	if
		(this->granules.end() == found)
	{
		this->granules[key] = cfg;
		ok = true;
	}

	return ok;
};

bool RenderingControlModel::resetCollectionCfg( const CollectionId& collectionId
											  , const VariableId& variableId)
{
	auto key = std::make_pair(collectionId, variableId);
	auto found = this->defaultCfgs.find(key);
	if
		(this->defaultCfgs.end() == found)
	{
		return false;
	}

	found->second->applyTo(this->collections[key]);
	return true;
};

/**
 *
 */
bool RenderingControlModel::registerCollectionCfg(RenderingCfg* cfg)
{
	bool ok;	//
	RenderingCfg* defaultCfg(nullptr);	//

	ok = false;
	std::pair<CollectionId, VariableId> key( cfg->target.collection
										   , cfg->target.variable);
	auto found = this->collections.find(key);
	if
		(this->collections.end() == found)
	{
		this->collections[key] = cfg;
		ok = true;

		defaultCfg = new RenderingCfg();
		cfg->applyTo(defaultCfg);
		this->defaultCfgs[key] = defaultCfg;
        defaultCfg = nullptr;
	}

	return ok;
};

/**
 *
 */
const RenderingCfg* RenderingControlModel::getDefaultCfg( const CollectionId& collectionId
														, const VariableId& variableId) const
{
	auto key = std::make_pair(collectionId, variableId);
	auto found = this->defaultCfgs.find(key);
	if
		(this->defaultCfgs.end() != found)
	{
		return found->second;
	}
	return nullptr;
};

/**
 *
 */
void RenderingControlModel::cfgChanged(const RenderingCfg* cfg, const RenderingParam& param)
{
	if
		(nullptr != this->callback)
	{
		this->callback(cfg, param);
	}
};

/**
 *
 */
bool RenderingControlModel::setCounters(const std::map<CollectionId, uint32_t>& countersIn)
{
	this->counters = countersIn;
	if
		(nullptr != this->countersCallback)
	{
		this->countersCallback();
	}

	return true;
};

/**
 *
 */
bool RenderingControlModel::removeVariable(const CollectionId& collectionId, const VariableId& variableId)
{
    std::pair<CollectionId, VariableId> key = std::make_pair(collectionId, variableId);
    auto collectionLevelFound = this->collections.find(key);
    if
        (this->collections.end() == collectionLevelFound)
    {
        LOG(LogLevel::Error, "RenderingControlModel",
                 "Rendering config not found at collection level for {}.{}:{}, cannot delete",
                collectionId.first, collectionId.second, variableId);
        return false;
    }

    auto defaultFound = this->defaultCfgs.find(key);
    if
        (this->defaultCfgs.end() == defaultFound)
    {
        LOG(LogLevel::Error, " RenderingControlModel",
                 "Default rendering not found for {}.{}:{}, cannot delete",
                 collectionId.first, collectionId.second, variableId);
        return false;
    }

    this->collections.erase(collectionLevelFound);
    this->defaultCfgs.erase(defaultFound);

    return true;
}

/**
 *
 */
bool RenderingControlModel::removeCollection(const CollectionId& collectionId)
{
    std::vector<VariableId> variables;  //
    for
        (const auto& it : this->collections)
    {
        if
            (it.first.first == collectionId)
        {
            variables.push_back(it.first.second);
        }
    }

    bool ok(true);
    for
        (const auto& varId : variables)
    {
        ok = ok && (this->removeVariable(collectionId, varId));
    }

    return ok;
}

void RenderingControlModel::updateDefaultCfg(RenderingCfg* rCfg)
{
	auto key = std::make_pair(rCfg->target.collection, rCfg->target.variable);
	auto found = this->defaultCfgs.find(key);
	if
		(this->defaultCfgs.end() != found)
	{
		rCfg->applyTo(found->second);
	}
}

/**
 *
 */
RenderingCfg* RenderingControlModel::createVolatileRenderingCfg()
{
    RenderingCfg* result {nullptr}; //

    result = new RenderingCfg();
    this->_volatile.push_back(result);
    return result;
}

/**
 *
 */
bool RenderingControlModel::listRenderedVariables(const CollectionId& collectionId, std::vector<VariableId>& result) const
{
    result.clear();
    for
        (const auto& it : this->collections)
    {
        if
            (  it.first.first.first == collectionId.first
            && it.first.first.second == collectionId.second
            && true == it.second->rendered())
        {
            result.push_back(it.first.second);
        }
    }

    return true;
}

/**
 *
 */
std::size_t RenderingControlModel::countRenderedVariables(const CollectionId& collectionId) const
{
    std::size_t result {0u};    //
    for
        (const auto& it : this->collections)
    {
        if
            (  it.first.first.first == collectionId.first
            && it.first.first.second == collectionId.second
            && true == it.second->rendered())
        {
            result += 1;
        }
    }

    return result;
}
