// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

#include <chrono>

#include "core/rendering_mediator.h"

#include "core/app_events_dispatcher.h"
#include "core/metadata_mediator.h"
#include "core/data_mediator.h"
#include "core/rendering_control_model.h"
#include "core/granule_metadata.h"
#include "core/data_bucket.h"
#include "core/display_queue.h"
#include "core/threads_pool.h"
#include "core/settings.h"
#include "core/source.h"
#include "core/utils.h"
#include "core/log.h"
#include "3d/iscene.h"
#include "3d/scene_view.h"
#include "3d/iscene_renderer.h"
#include "3d/scene_graph_controller.h"
#include "3d/shapes_allocator.h"
#include "3d/datasets_allocator.h"
#include "3d/shape.h"
#include "data/unrolled_functions.h"
#include "events/rendering_cfg_update_event.h"
#include "num_utils.h"

using namespace SEAScope;

/**
 * Free heap-allocated memory used by attributes of discarded DisplayJob
 * instance.
 */
static void clearDiscardedDisplayJob(DisplayJob& displayJob)
{
    DatasetsAllocator::release(displayJob.dataset);
    displayJob.dataset = nullptr;

    ShapesAllocator::release(displayJob.shape);
    displayJob.shape = nullptr;
}

/**
 *
 */
static bool checkRequiredFields(const std::set<std::string>& requiredFields,
                                const std::set<std::string>& availableFields)
{
    for
        (const auto& requiredField : requiredFields)
    {
        auto fieldFound = std::find(availableFields.begin(), availableFields.end(), requiredField);
        if
            (availableFields.end() == fieldFound)
        {
            return false;
        }
    }
    return true;
}

/**
 *
 */
RenderingMediator::RenderingMediator(AppEventsDispatcher& appEventsDispatcher,
                                     RenderingControlModel& model,
                                     MetadataMediator& metadataMediator,
                                     DataMediator& dataMediator,
                                     Settings& settings,
                                     DisplayQueue& displayQueue)
: _appEventsDispatcher{appEventsDispatcher},
  _model{model},
  _metadataMediator{metadataMediator},
  _dataMediator{dataMediator},
  _settings{settings},
  _displayQueue{displayQueue},
  _sceneView{nullptr},
  _sceneGraphController{nullptr},
  _workers{nullptr},
  _maxTextureSize{0u}
{
}

/**
 *
 */
RenderingMediator::~RenderingMediator()
{
    this->_sceneView = nullptr; // not owning
    this->_sceneGraphController = nullptr; // not owning
}

/**
 *
 */
void RenderingMediator::updateRendering(const RenderingCfg* rCfg,
                                       const RenderingParam& param)
{
	DisplayJob j;

	j.jobType = DisplayJobType::UPDATE_RENDERING;
	j.renderingCfg = *rCfg;
	j.rId = rCfg->target;

	this->_displayQueue.addJob(j);

    RenderingCfgUpdateEvent event(rCfg, param);
    this->_appEventsDispatcher.dispatch<RenderingCfgUpdateEvent>(event);
}

/**
 *
 */
bool RenderingMediator::buildExpectations(const std::map<CollectionId, std::vector<GranuleMetadata> >& renderables,
                                          RenderingExpectations& expected)
{
    bool ok {false};    //

    for
        (const auto& renderableCollection : renderables)
    {
        SourceId sourceId {renderableCollection.first.first};   //
        CollectionLocalId collectionId {renderableCollection.first.second}; //

        std::vector<VariableId> varIds; //
        ok = this->_model.listRenderedVariables(renderableCollection.first, varIds);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "RenderingMediator",
                "Failed o list rendered variables for the {}.{} collection",
                sourceId, collectionId);
            return false;
        }

        for
            (const auto& varId : varIds)
        {
            // Keep a list of all the variables for which we expect to get data
            // as well as the fields required for these variables
            VariableCfg varCfg; //
            ok = this->_metadataMediator.getVariableCfg(renderableCollection.first, varId, varCfg);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "RenderingMediator",
                    "Failed to get configuration for variable {}.{}:{}",
                    sourceId, collectionId, varId);
                return false;
            }
            expected.fields[sourceId][collectionId][varId] = std::set<std::string>(varCfg.fields.begin(),
                                                              varCfg.fields.end());

            // Keep track of all the data representations that the renderer
            // should display using identifiers
            for
                (const auto& granule : renderableCollection.second)
            {
                GranuleId gid {granule.sourceId(), granule.granuleId()};
                expected.ids.emplace(true, renderableCollection.first, gid, varId);
            }
        }
    }

    return true;
}

/**
 *
 */
bool RenderingMediator::show(const std::map<CollectionId, std::vector<GranuleMetadata> >& renderables)
{
    bool ok {false};    //
    RenderingExpectations expected; //

    ok = this->buildExpectations(renderables, expected);
    if
        (false == ok)
    {
        // Error messages already logged during the execution of formatRenderables
        return false;
    }

    std::size_t renderablesCount{0};    //
    renderablesCount = expected.ids.size();

    this->beginUpdate(std::move(expected));
    if
        (0 >= renderablesCount)
    {
        LOG(LogLevel::Debug, "RenderingMediator", "Renderables count = {}", renderablesCount);
        // Early exit (which also prevents SEAScope from crashing during
        // initialization when the scene is not ready yet
        this->updateComplete();
        return true;
    }

    // Extract information about the viewport
	float viewportResolution {0.f};	//
    std::uint32_t viewportArea {0u};    //
    this->getViewportInfo(viewportArea, viewportResolution);

    for
        (const auto& collectionRenderables : renderables)
    {
        SourceId sourceId {collectionRenderables.first.first};  //
        CollectionLocalId collectionId {collectionRenderables.first.second};    //
        for
            (const auto& granule : collectionRenderables.second)
        {
            DataModel dataModel{granule.dataModel()};   //
            float resolution {0.f}; //
            std::uint32_t xArity {0u};  //
            std::uint32_t yArity {0u};  //
            std::string uri;    //
            RenderableId rId {true, CollectionId{sourceId, collectionId}, GranuleId{sourceId, granule.granuleId()}, 0};  //

            IDFDescriptor idf;
            ok = granule.selectResolution(viewportArea,
                                          viewportResolution,
                                          this->_settings.minGranuleScreenEstate,
                                          xArity, yArity, idf);
            if
                (false == ok)
            {
                LOG(LogLevel::Warn, "RenderingMediator",
                    "Failed to select a resolution for granule {}.{}",
                    granule.sourceId(), granule.granuleId());

                // Granule should not be displayed since there is no valid
                // resolution
                continue;
            }
            resolution = idf.resolution;
            uri = idf.uri;
            {
                std::lock_guard<std::mutex> scopeLock {this->_granulesInfoMutex};
                this->_granulesInfo[sourceId][rId.granule.second].dataModel = dataModel;
                this->_granulesInfo[sourceId][rId.granule.second].idf = std::move(idf);
            }

            std::set<std::string> requiredFields;
            std::unordered_map<VariableId, std::set<std::string>> variables;
            {
                std::lock_guard<std::mutex> scopeLock {this->_expectedMutex};
                if
                    (  0 < this->_expected.fields.count(sourceId)
                    && 0 < this->_expected.fields.at(sourceId).count(collectionId))
                {
                    variables = this->_expected.fields.at(sourceId).at(collectionId);   // Copy
                }
            }
            for
                (const auto& variable : variables)
            {
                rId.variable = variable.first;

                ok = this->alreadyHandled(rId, resolution);
                if
                    (true == ok)
                {
                    LOG(LogLevel::Debug, "RenderingMediator", "To display (already handled): {}.{}.{}:{} @ {}",
                        rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution);
                    // The renderable item is already in the scene or queued
                    // for rendering, nothing else to do
                    continue;
                }

                // TODO: remove this once tiling (or another way to circumvent
                //       max texture size) has been implemented for all
                //       rendering methods.
                bool textureSizeOk {false};
                textureSizeOk = this->isTextureSizeSupported(rId, xArity, yArity);
                if
                    (false == textureSizeOk)
                {
                    LOG(LogLevel::Warn, "RenderingMediator",
                        "Texture size {}x{} exceeds the capacity of the GPU, rendering {}.{}.{}:{} will not be possible",
                        xArity, yArity, rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
                    continue;
                }

                LOG(LogLevel::Debug, "RenderingMediator", "To display: {}.{}.{}:{} @ {}",
                    rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution);

                requiredFields.insert(variable.second.begin(), variable.second.end());
            }

            // Trigger data loading only if there is at least one field that is
            // still missing. Annotations are not affected by this rule as they
            // have no field attached.
            if
                (   (0 < requiredFields.size())
                 || (SEAScope::Source::ANNOTATIONS == rId.collection.first))
            {
                this->_dataMediator.loadData(rId.collection, rId.granule, uri, resolution, requiredFields, dataModel, this->_rawToRenderableQueue);
            }
        }
    }

    this->updateComplete();

    return true;
}

/**
 *
 */
bool RenderingMediator::addVariables(const std::set<CollectionVariable>& colvars)
{
    std::set<RenderableId> renderableIds;   //
    std::set<RenderableId> added;   //
    SourceId sourceId {0u};  //
    CollectionLocalId collectionId {0u};    //
    GranuleLocalId granuleId {0u};  //
    VariableId varId {0u};  //

    {
        std::lock_guard<std::mutex> scopeLock {this->_expectedMutex};
        for
            (const auto& colvar : colvars)
        {
            sourceId = colvar.first.first;  //
            collectionId = colvar.first.second;    //
            varId = colvar.second;   //

            bool ok {false};    //
            VariableCfg varCfg; //
            ok = this->_metadataMediator.getVariableCfg(colvar.first, colvar.second, varCfg);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "RenderingMediator",
                    "Failed to get configuration for variable {}.{}:{}",
                    sourceId, collectionId, varId);
                return false;
            }

            std::set<std::string> requiredFields (varCfg.fields.begin(),
                                                  varCfg.fields.end());

            // Update expected fields
            this->_expected.fields[sourceId][collectionId][colvar.second] = requiredFields;

            // Build identifiers for granules that belong to the collection
            // Use a set to avoid duplicates, as several variables from the
            // same collection may already be selected
            for
                (const auto& rId : this->_expected.ids)
            {
                if
                    (rId.collection == colvar.first)
                {
                    added.emplace(true, rId.collection, rId.granule, varId);
                }
            }
        }

        // Update expected renderable ids
        this->_expected.ids.insert(added.begin(), added.end());
        renderableIds = this->_expected.ids;
    }

    // Let the renderer know about new expectations
    DisplayJob j;   //
    j.jobType = DisplayJobType::UPDATE_STARTED;
    j.resolution = 0.f;
    j.toRender = std::move(renderableIds);
    this->_displayQueue.addJob(j);
    this->updateComplete();

    for
        (const auto& rId : added)
    {
        sourceId = rId.granule.first;
        collectionId = rId.collection.second;
        granuleId = rId.granule.second;
        varId = rId.variable;

        IDFDescriptor idf;
        DataModel dataModel;
        {
            std::lock_guard<std::mutex> scopeLock {this->_granulesInfoMutex};
            if
                (  0 >= this->_granulesInfo.count(sourceId)
                || 0 >= this->_granulesInfo.at(sourceId).count(granuleId))
            {
                continue;
            }

            idf = this->_granulesInfo.at(sourceId).at(granuleId).idf;   // Copy
            dataModel = this->_granulesInfo.at(sourceId).at(granuleId).dataModel;   // Copy
        }

        bool ok {false};    //
        ok = this->alreadyHandled(rId, idf.resolution);
        if
            (true == ok)
        {
            LOG(LogLevel::Debug, "RenderingMediator", "To display (already handled): {}.{}.{}:{} @ {}",
                rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, idf.resolution);
            // The renderable item is already in the scene or queued
            // for rendering, nothing else to do
            continue;
        }

        // TODO: remove this once tiling (or another way to circumvent
        //       max texture size) has been implemented for all
        //       rendering methods.
        bool textureSizeOk {false};
        textureSizeOk = this->isTextureSizeSupported(rId, idf.xArity, idf.yArity);
        if
            (false == textureSizeOk)
        {
            LOG(LogLevel::Warn, "RenderingMediator",
                "Texture size {}x{} exceeds the capacity of the GPU, rendering {}.{}.{}:{} will not be possible",
                idf.xArity, idf.yArity, rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
            continue;
        }

        std::set<std::string> requiredFields;
        {
            std::lock_guard<std::mutex> scopeLock {this->_expectedMutex};
            if
                (  0 < this->_expected.fields.count(sourceId)
                && 0 < this->_expected.fields.at(sourceId).count(collectionId)
                && 0 < this->_expected.fields.at(sourceId).at(collectionId).count(varId))
            {
                requiredFields = this->_expected.fields.at(sourceId).at(collectionId).at(varId);   // Copy
            }
        }
        this->_dataMediator.loadData(rId.collection, rId.granule, idf.uri, idf.resolution, requiredFields, dataModel, this->_rawToRenderableQueue);
    }
    return true;
}

/**
 *
 */
bool RenderingMediator::removeVariables(const std::set<CollectionVariable>& colvars)
{
    std::set<RenderableId> renderableIds;

    {
        std::lock_guard<std::mutex> scopeLock {this->_expectedMutex};
        for
            (const auto& colvar: colvars)
        {
            SourceId sourceId {colvar.first.first};  //
            CollectionLocalId collectionId {colvar.first.second};    //
            VariableId varId {colvar.second};   //

            this->_expected.fields.at(sourceId).at(collectionId).erase(varId);

            std::copy_if(this->_expected.ids.begin(), this->_expected.ids.end(),
                         std::inserter(renderableIds, renderableIds.end()),
                         [colvar](const RenderableId& rId) { return (rId.collection != colvar.first) || (rId.variable != colvar.second); });
            this->_expected.ids = renderableIds;
            renderableIds.clear();
        }
        renderableIds = this->_expected.ids;  // Copy
    }

    DisplayJob j;   //
    j.jobType = DisplayJobType::UPDATE_STARTED;
    j.resolution = 0.f;
    j.toRender = std::move(renderableIds);
    this->_displayQueue.addJob(j);
    this->updateComplete();

    return true;
}

/**
 *
 */
bool RenderingMediator::removeCollections(const std::set<CollectionId>& collections)
{
    LOG(LogLevel::Debug, "RenderingMediator", "removeCollections");
    std::set<RenderableId> renderableIds;

    {
        std::lock_guard<std::mutex> scopeLock {this->_expectedMutex};
        for
            (const auto& collection: collections)
        {
            SourceId sourceId {collection.first};  //
            CollectionLocalId collectionId {collection.second};    //

            this->_expected.fields.at(sourceId).erase(collectionId);

            std::copy_if(this->_expected.ids.begin(), this->_expected.ids.end(),
                         std::inserter(renderableIds, renderableIds.end()),
                         [collection](const RenderableId& rId) { return (rId.collection != collection); });
            this->_expected.ids = renderableIds;
            renderableIds.clear();
        }
        renderableIds = this->_expected.ids;  // Copy
    }

    DisplayJob j;   //
    j.jobType = DisplayJobType::UPDATE_STARTED;
    j.resolution = 0.f;
    j.toRender = std::move(renderableIds);
    this->_displayQueue.addJob(j);
    this->updateComplete();

    return true;
}

/**
 *
 */
bool RenderingMediator::isTextureSizeSupported(const RenderableId& rId,
                                               std::uint32_t xArity,
                                               std::uint32_t yArity)
{
    if
        (  (xArity <= this->_maxTextureSize)
        && (yArity <= this->_maxTextureSize))
    {
        return true;
    }

    RenderingCfg* gCfg {nullptr}; //
    bool ok {false};    //

    // Request rendering cfg since there might be a configuration
    // customized for this granule
    gCfg = this->_model.getCfg(rId.collection, rId.granule, rId.variable, ok);
    if
        (false == ok)
    {
        LOG( LogLevel::Error, "RenderingMediator"
                , "Could not retrieve rendering cfg for {}/{}:{} @ {}.{}"
                , rId.collection.first, rId.collection.second
                , rId.variable, rId.granule.first, rId.granule.second);
        return false;
    }

    RenderMethod renderMethod = gCfg->renderMethod();
    gCfg = nullptr;
    return (  RenderMethod::RASTER == renderMethod
           || RenderMethod::RAWRGB == renderMethod
           || RenderMethod::TRAJECTORIES == renderMethod);
}

/**
 *
 */
void RenderingMediator::setSceneView(SceneView* sceneView)
{
    this->_sceneView = sceneView;
    this->_sceneGraphController = sceneView->getScene()->getNodeFactory();
    this->_maxTextureSize = static_cast<std::uint32_t>(sceneView->getRenderer()->getMaxTextureSize());
}

/**
 *
 */
bool RenderingMediator::alreadyHandled(const RenderableId& rId,
                                       float resolution)
{
    bool ok {false};    //
    ok = this->_sceneGraphController->exist(rId, resolution);
    if
        (false == ok)
    {
        ok = this->_displayQueue.contains(rId, resolution);
    }
    return ok;
}

/**
 *
 */
void RenderingMediator::beginUpdate(RenderingExpectations&& expected)
{
    DisplayJob j;   //
	j.jobType = DisplayJobType::UPDATE_STARTED;
    j.resolution = 0.f;
    j.toRender = expected.ids;  // Copy

    {
        std::lock_guard<std::mutex> scopeLock {this->_expectedMutex};
        this->_expected = std::forward<RenderingExpectations&&>(expected);
    }

	this->_displayQueue.addJob(j);

    {
        std::lock_guard<std::mutex> scopeLock {this->_granulesInfoMutex};
        this->_granulesInfo.clear();
    }

    {
        std::lock_guard<std::mutex> scopeLock {this->_beingProcessedMutex};
        this->_beingProcessed.clear();
    }
}

/**
 *
 */
void RenderingMediator::updateComplete()
{
    // Tell the renderer that everything is ready
    DisplayJob j;   //
    j.jobType = DisplayJobType::UPDATE_COMPLETE;
    j.resolution = 0.f;
    this->_displayQueue.addJob(j);
}

/**
 *
 */
void RenderingMediator::getViewportInfo(std::uint32_t& viewportArea,
                                        float& viewportResolution) const
{
	int viewportWidth {0};	//
	int viewportHeight {0};	//
	viewportResolution = this->_sceneView->getPixelSize();
	viewportWidth = this->_sceneView->getViewport().x;
	viewportHeight = this->_sceneView->getViewport().y;
    viewportArea = static_cast<std::uint32_t>(viewportWidth * viewportHeight);
}

/**
 *
 */
bool RenderingMediator::processLoadedData(const CollectionId& collectionId,
                                          const GranuleId& granuleId,
                                          const std::string& uri,
                                          float resolution,
                                          const std::set<std::string>& fields)
{
    CollectionCfg colCfg;   //
    bool ok {false};    //

    /*
    std::string fieldsStr;
    SEAScope::implodeStringSet(fields, ",", fieldsStr);
    LOG(LogLevel::Debug, "RenderingMediator", "processLoadedData {}:{}", uri, fieldsStr);
    */

    {
        std::lock_guard<std::mutex> scopeLock {this->_granulesInfoMutex};
        if
            (  0 >= this->_granulesInfo.count(granuleId.first)
            || 0 >= this->_granulesInfo.at(granuleId.first).count(granuleId.second))
        {
            LOG(LogLevel::Debug, "RenderingMediator", "Skipping loaded data because URI is not one of the expected ones");
            return false;
        }
    }

    ok = this->_metadataMediator.getCollectionCfg(collectionId, colCfg);
    if
        (false == ok)
    {
        return false;
    }

    std::unordered_map<VariableId, std::set<std::string>> variables;
    {
        std::lock_guard<std::mutex> scopeLock {this->_expectedMutex};
        if
            (  0 < this->_expected.fields.count(collectionId.first)
            && 0 < this->_expected.fields.at(collectionId.first).count(collectionId.second))
        {
            variables = this->_expected.fields.at(collectionId.first).at(collectionId.second);
        }
    }

    std::unordered_map<std::string, std::pair<std::set<std::string>, std::set<SEAScope::RenderableId> > > variablesGroups;
    for
        (const auto& variable : variables)
    {
        RenderableId rId{true, collectionId, granuleId, variable.first};   //

        // Skip if the scene already has this renderable or if it is in the
        // display queue
        ok = this->alreadyHandled(rId, resolution);
        if
            (true == ok)
        {
            /*
            LOG(LogLevel::Debug, "RenderingMediator",
                "{}.{}.{}:{} @ {} skipped because it is already in the scene or queued to be added soon",
                rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution);
            */
            continue;
        }

        // Check if all required fields are there
        ok = checkRequiredFields(variable.second, fields);
        if
            (false == ok)
        {
            /*
            LOG(LogLevel::Debug, "RenderingMediator",
                "{}.{}.{}:{} skipped because at least one required field has not been loaded yet",
                rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
            */
            continue;
        }

        // Regroup variables that share the same fields requirements
        std::string fieldsStr;
        SEAScope::implodeStringSet(variable.second, ",", fieldsStr);
        if
            (0 < variablesGroups.count(fieldsStr))
        {
            variablesGroups[fieldsStr].second.insert(rId);
        }
        else
        {
            variablesGroups[fieldsStr];
            variablesGroups[fieldsStr].first = variable.second;
            variablesGroups[fieldsStr].second = {rId};
        }
    }

    for
        (const auto& variablesGroup : variablesGroups)
    {
        Job j;
        std::string jobId;
        JobStatus status;
        j.task = [this, uri, colCfg, resolution, variablesGroup] (std::string& /*errMsg*/) -> bool {
            return this->formatRenderableData(variablesGroup.first, variablesGroup.second.second, uri, variablesGroup.second.first, colCfg, resolution);
        };
        ok = this->_workers->submitJob(std::move(j), jobId, status);
    }

    return true;
}

/**
 *
 */
bool RenderingMediator::formatRenderableData(const std::string& fieldsStr,
                                             const std::set<RenderableId>& rIds,
                                             const std::string& uri,
                                             const std::set<std::string>& fields,
                                             const CollectionCfg& colCfg,
                                             float resolution)
{
    bool ok {false};    //

    if
        (0 >= rIds.size())
    {
        return false;
    }

    SourceId sourceId {rIds.cbegin()->granule.first};
    GranuleLocalId granuleId {rIds.cbegin()->granule.second};
    {
        std::lock_guard<std::mutex> scopeLock {this->_granulesInfoMutex};
        if
            (  0 >= this->_granulesInfo.count(sourceId)
            || 0 >= this->_granulesInfo.at(sourceId).count(granuleId))
        {
            LOG(LogLevel::Debug, "RenderingMediator", "Skipping loaded data because URI is not one of the expected ones 2");
            return false;
        }
    }

    const DataInfo* dataInfoPtr {nullptr};  //
    const Image* img {nullptr}; //
    OGL::Shape* geometry {nullptr}; //
    if
        (1 > fields.size())
    {
        // For annotations fields.size() == 0, yet DataMediator::getField must
        // be called to obtain the granule's geometry
        ok = this->_dataMediator.getField(rIds.cbegin()->granule, uri, "", img, geometry, dataInfoPtr);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "RenderingMediator",
                "Failed to get data from cache for '{}' ({})", uri, fieldsStr);
            return false;
        }
    }
    else if
        (1 == fields.size())
    {
        // For annotations fields.size() == 0, yet DataMediator::getField must
        // be called to obtain the granule's geometry
        ok = this->_dataMediator.getField(rIds.cbegin()->granule, uri, *fields.begin(), img, geometry, dataInfoPtr);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "RenderingMediator",
                "Failed to get data from cache for '{}:{}' ({})", uri, *fields.begin(), fieldsStr);
            return false;
        }
    }
    else
    {
        std::string compositeCacheKey = uri + ":" + fieldsStr;

        std::lock_guard<std::mutex> scopeLock {this->_compositeMutex};

        if
            (0 >= this->_compositeCache.count(compositeCacheKey))
        {
            std::vector<const Image*> buffers;
            std::vector<const DataInfo*> infos;
            ok = this->_dataMediator.getFields(uri, fields, buffers, infos);

            this->_compositeCache[compositeCacheKey];
            Image& cacheImg = this->_compositeCache.at(compositeCacheKey).getImage();
            DataInfo& cacheInfo = this->_compositeCache.at(compositeCacheKey).getInfo();

            DataType dataType {infos[0]->dataType};
            bool validDataTypes {true};
            for
                (const DataInfo* info : infos)
            {
                size_t fieldChannels = info->offsets.size();
                for
                    (size_t channel=0; channel < fieldChannels; ++channel)
                {
                    if
                        (info->dataType != dataType)
                    {
                        validDataTypes = false;
                        cacheInfo.offsets.clear();
                        cacheInfo.scaleFactors.clear();
                        cacheInfo.fillValues.clear();
                        LOG(LogLevel::Error, "RenderingMediator",
                            "{} has fields with incompatible types: {}",
                            uri, fieldsStr);
                        break;
                    }
                    cacheInfo.offsets.push_back(info->offsets[channel]);
                    cacheInfo.scaleFactors.push_back(info->scaleFactors[channel]);
                    cacheInfo.fillValues.push_back(info->fillValues[channel]);
                }

                if
                    (false == validDataTypes)
                {
                    break;
                }
            }

            if
                (true == validDataTypes)
            {
                cacheInfo.xArity = infos[0]->xArity;
                cacheInfo.yArity = infos[0]->yArity;
                cacheInfo.dataType = infos[0]->dataType;
                cacheInfo.channels = static_cast<std::uint8_t>(cacheInfo.offsets.size());

                LOG(LogLevel::Debug, "RenderingMediator",
                    "Build composite cache entry {}", compositeCacheKey);

                cacheImg.allocate(cacheInfo.xArity,
                                  cacheInfo.yArity,
                                  cacheInfo.channels,
                                  cacheInfo.dataType,
                                  nullptr);

                // Interleaves arrays
                soa_to_aos(buffers, &cacheImg);

                this->_compositeCache.at(compositeCacheKey).setLoaded(true);
            }
        }

        if
            (true == this->_compositeCache.at(compositeCacheKey).loaded())
        {
            img = this->_compositeCache.at(compositeCacheKey).getImagePtr();
            dataInfoPtr = this->_compositeCache.at(compositeCacheKey).getInfoPtr();
            LOG(LogLevel::Debug, "RenderingMediator", "Format, dataInfo: {}x{}x{}",
                dataInfoPtr->channels, dataInfoPtr->xArity, dataInfoPtr->yArity);
        }
    }

    if
        (  0 < fields.size()
        && nullptr == dataInfoPtr
        && nullptr == img)
    {
        // Creation of image failed in previous steps
        return false;
    }

    for
        (const auto& rId: rIds)
    {
        Job job;
        JobStatus status;
        std::string jobId;
        job.task = [this, rId, img, geometry, colCfg, dataInfoPtr, uri, resolution] (std::string& /*errMsg*/) -> bool{
            bool success {false};    //
            std::pair<RenderableId, std::string> gpuKey {rId, uri};

            // dataInfoPtr is nullptr for annotations, so a check is required
            // before dereferencing the pointer
            DataInfo dataInfo;
            if
                (nullptr != dataInfoPtr)
            {
                dataInfo = *dataInfoPtr; // Copy
            }

            std::lock_guard<std::mutex> scopeLock {this->_toGPUMutex};
            if
                (0 < this->_toGPU.count(gpuKey))
            {
                LOG(LogLevel::Debug, "RenderingMediator",
                    "Skipping GPU formatting for {}.{}.{}:{} @ {} as another thread is already handling it",
                    rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution);
                return false;
            }

            this->_toGPU[gpuKey] = true;

            success = this->sendToDisplayQueue(rId, img, geometry, colCfg, dataInfo, uri, resolution);

            this->_toGPU.erase(gpuKey);

            return success;
        };

        ok = this->_workers->submitJob(std::move(job), jobId, status);
    }

    return ok;
}

/**
 *
 */
bool RenderingMediator::isGPUDataNeeded(const RenderableId& rId, const std::string& uri, float resolution)
{
    bool ok {false};    //

    std::string expectedURI;
    ok = this->getRenderableURI(rId, expectedURI);
    if
        (false == ok)
    {
        LOG(LogLevel::Debug, "RenderingMediator",
            "Skipping deprecated renderable {}.{}.{}:{}",
            rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
        return false;
    }
    if
        (0 != uri.compare(expectedURI))
    {
        LOG(LogLevel::Debug, "RenderingMediator",
            "Skipping deprecated URI for {}.{}.{}:{}: {} (deprecated) vs {} (expected))",
            rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, uri, expectedURI);
        return false;
    }

    ok = this->alreadyHandled(rId, resolution);
    if
        (true == ok)
    {
        LOG(LogLevel::Debug, "RenderingMediator",
            "Skipping because {}.{}.{}:{} @ {} is already in the scene",
            rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution);
        return false;
    }

    return true;
}

/**
 *
 */
bool RenderingMediator::sendToDisplayQueue(const RenderableId& rId, const Image* img, OGL::Shape* geometry, const CollectionCfg& colCfg, const DataInfo& dataInfo, const std::string& uri, float resolution)
{
    RenderingCfg* rCfg {nullptr};   //
    bool ok {false};    //

    ok = this->isGPUDataNeeded(rId, uri, resolution);
    if
        (false == ok)
    {
        return false;
    }

    LOG(LogLevel::Debug, "RenderingMediator",
        "sendToDisplayQueue {}.{}.{}:{} @ {}",
        rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution);

    // Request rendering cfg since there might be a configuration
    // customized for this granule
    rCfg = this->_model.getCfg(rId.collection, rId.granule, rId.variable, ok);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "RenderingMediator",
            "Could not retrieve rendering cfg for {}.{}.{}:{} @ {}",
            rId.collection.first, rId.collection.second, rId.granule.second,
            rId.variable, resolution);

        return false;
    }

    DisplayJob displayJob;   //
    displayJob.jobType = DisplayJobType::LOAD_DATA;
    displayJob.resolution = resolution;
    displayJob.rId = rId;

    if
        (  RenderMethod::POLYGON == rCfg->_renderMethod
        || RenderMethod::POLYLINE == rCfg->_renderMethod
        || RenderMethod::POINT == rCfg->_renderMethod)
    {
        if
            (nullptr == geometry)
        {
            LOG(LogLevel::Error, "RenderingMediator",
                "Granule {}.{}.{}:{} @ {} should be rendered as a shape but has no geometry",
                rId.collection.first, rId.collection.second, rId.granule.second,
                rId.variable, resolution);
            return false;
        }
        displayJob.shape = geometry;    // /!\ displayJob gets ownership of geometry
    }
    else
    {
        Shape shape;
        {
            SourceId sourceId {rId.granule.first};  //
            GranuleLocalId granuleId {rId.granule.second};  //
            std::lock_guard<std::mutex> scopeLock {this->_granulesInfoMutex};
            if
                (  0 >= this->_granulesInfo.count(sourceId)
                || 0 >= this->_granulesInfo.at(sourceId).count(granuleId))
            {
                return false;
            }

            shape = this->_granulesInfo.at(sourceId).at(granuleId).idf.shape;  // Copy
        }

        if
            (RenderMethod::TRAJECTORIES == rCfg->_renderMethod)
        {
            displayJob.dataset = nullptr;
            ok = this->_gpuData.buildPolyline(img->data(),
                                              dataInfo,
                                              colCfg,
                                              shape,
                                              displayJob.relativeTime,
                                              displayJob.points,
                                              displayJob.fData,
                                              displayJob.mask);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "RenderingMediator",
                    "Failed to build polyline for {}.{}.{}:{} @ {}",
                    rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution);
                return false;
            }
        }
        else
        {
            bool fitsOnGPU {false}; //
            bool tiled {false}; //

            fitsOnGPU = (  img->width() <= static_cast<unsigned int>(this->_maxTextureSize)
                        && img->height() <= static_cast<unsigned int>(this->_maxTextureSize));
            if
                (false == fitsOnGPU)
            {
                tiled = (  RenderMethod::RASTER == rCfg->renderMethod()
                        || RenderMethod::RAWRGB == rCfg->renderMethod());
                if
                    (false == tiled)
                {
                    LOG(LogLevel::Error, "RenderingMediator",
                        "Cannot render {}.{}.{}:{} @ {} because texture size {}x{} has one dimension exceeding the limit supported by the GPU ({}), and tiling is not supported for the selected rendering method",
                        rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution, img->width(), img->height(), this->_maxTextureSize);
                    return false;
                }
            }

            displayJob.dataset = DatasetsAllocator::get(rId, tiled);

            ok = this->_gpuData.buildDataSet(displayJob.dataset,
                img->data(),
                dataInfo,
                colCfg,
                shape,
                resolution,
                static_cast<int>(this->_maxTextureSize));

            if
                (false == ok)
            {
                LOG(LogLevel::Error, "RenderingMediator",
                    "Failed to build dataset for {}.{}.{}:{} @ {}",
                    rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution);
                return false;
            }
        }
    }

    LOG(LogLevel::Debug, "RenderingMediator",
        "GPU data ready, job added to display queue {}.{}.{}:{} @ {}",
        rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, resolution);

    ok = this->isGPUDataNeeded(rId, uri, resolution);
    if
        (true == ok)
    {
        // We wait until the last moment to dereference rCfg so that the copy
        // held by displayJob is as up-to-date as possible (some settings may
        // have changed while _gpuData performing "long" formatting operations.
        displayJob.renderingCfg = *rCfg;

        ok = this->_displayQueue.addJob(displayJob);
    }

    if
        (false == ok)
    {
        // Data loading job has been rejected, free dataset and shape memory
        LOG(LogLevel::Debug, "RenderingMediator", "Data load job rejected by display queue");
        clearDiscardedDisplayJob(displayJob);
    }

    return ok;
}

/**
 *
 */
bool RenderingMediator::getRenderableURI(const RenderableId& rId,
                                         std::string& result)
{
    SourceId sourceId {rId.granule.first};  //
    GranuleLocalId granuleId {rId.granule.second};  //
    std::lock_guard<std::mutex> scopeLock {this->_granulesInfoMutex};

    if
        (  0 >= this->_granulesInfo.count(sourceId)
        || 0 >= this->_granulesInfo.at(sourceId).count(granuleId))
    {
		return false;
    }

	result = this->_granulesInfo.at(sourceId).at(granuleId).idf.uri;
    return true;
}

/**
 *
 */
bool RenderingMediator::setWorkers(ThreadsPool* workers)
{
    bool ok {false};    //
    this->_workers = workers;

    Job j;
    std::string jobId;
    JobStatus status;
    RawToRenderableQueue& jobQueue {this->_rawToRenderableQueue};

    j.task = [this, &jobQueue] (std::string& /*errMsg*/) -> bool {
        SEAScope::RawToRenderableJob job;
        while
            (true)
        {
            jobQueue.pop(job);  // should not consume CPU due to condition variable in implementation

            if
                (0.f > job.resolution)
            {
                return true;
            }

            std::string fieldsStr;
            SEAScope::implodeStringSet(job.fields, ",", fieldsStr);
            std::string targetKey = job.uri + ":" + fieldsStr;

            LOG(LogLevel::Debug, "RenderingMediator", "Executing data loading job {}", targetKey);
            bool processingNeeded {false};  //
            {
                std::lock_guard<std::mutex> scopeLock {this->_beingProcessedMutex};
                processingNeeded = (0 >= this->_beingProcessed.count(targetKey));

                // Accessing the element using the [] operator will create it
                // if it does not exist yet
                this->_beingProcessed[targetKey];
            }

            if
                (false == processingNeeded)
            {
                // Ignore job as another one has already been executed
                // in this batch
                LOG(LogLevel::Debug, "RenderingMediator", "job skipped");
                continue;
            }

            this->processLoadedData(job.collection,
                                    job.granule,
                                    job.uri,
                                    job.resolution,
                                    job.fields);

            {
                std::lock_guard<std::mutex> scopeLock {this->_beingProcessedMutex};
                this->_beingProcessed.erase(targetKey);
            }
        }
    };
    ok = this->_workers->submitJob(std::move(j), jobId, status);
    return ok;
}

/**
 *
 */
void RenderingMediator::shutdown()
{
    this->_rawToRenderableQueue.shutdown();
}
