// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include "core/settings.h"


using namespace SEAScope;


/**
 * Constructor.
 */
Settings::Settings()
: glContextMajor{0u}, glContextMinor{0u}, windowWidth{0u}, windowHeight{0u},
  vsync{false}, antialiasing{false}, fullscreen{false}, antialiasingFactor{0u},
  minGranuleScreenEstate{0.f}, logFilter(""), processorEnabled{false},
  processorPort{-1}, skipIndexCheck{false}, searchRequestMinDelay{0.0},
  explicitDrawingModeExit{false}, clickPixelsTolerance{0u}
{

};

/**
 * Destructor.
 */
Settings::~Settings()
{

};

/**
 * Copy assignment
 */
Settings& Settings::operator=(const Settings& original)
{
    this->glRenderer = original.glRenderer;
    this->glContextMajor = original.glContextMajor;
    this->glContextMinor = original.glContextMinor;
    this->pluginsPath = original.pluginsPath;
    this->colormapsPath = original.colormapsPath;
    this->operatorsPath = original.operatorsPath;
    this->dataPath = original.dataPath;
    this->annotationsPath = original.annotationsPath;
    this->customPath = original.customPath;
    this->windowWidth = original.windowWidth;
    this->windowHeight = original.windowHeight;
    this->vsync = original.vsync;
    this->antialiasing = original.antialiasing;
    this->fullscreen = original.fullscreen;
    this->antialiasingType = original.antialiasingType;
    this->antialiasingFactor = original.antialiasingFactor;
    this->worldMap = original.worldMap;
    this->minGranuleScreenEstate = original.minGranuleScreenEstate;
    this->logFilter = original.logFilter;
    this->statePath = original.statePath;
    this->indexPath = original.indexPath;
    this->processorEnabled = original.processorEnabled;
    this->processorAddress = original.processorAddress;
    this->processorPort = original.processorPort;
    this->skipIndexCheck = original.skipIndexCheck;
    this->searchRequestMinDelay = original.searchRequestMinDelay;
    this->explicitDrawingModeExit = original.explicitDrawingModeExit;
    this->clickPixelsTolerance = original.clickPixelsTolerance;
    this->timelineCfg.setMinYear(original.timelineCfg.getMinYear());
    this->timelineCfg.setMaxYear(original.timelineCfg.getMaxYear());
    this->timelineCfg.defaultTimespan = original.timelineCfg.defaultTimespan;
    this->timelineCfg.defaultTimestep = original.timelineCfg.defaultTimestep;
    this->timelineCfg.useNearestMode = original.timelineCfg.useNearestMode;
    this->timelineCfg.advancedScrolling = original.timelineCfg.advancedScrolling;
    this->timelineCfg.setInitialDatetime(original.timelineCfg.getInitialDatetime());
    this->timelineCfg.setInitialDatetimeString(original.timelineCfg.getInitialDatetimeString());
    this->timelineCfg.setUseAutoMaxYear(original.timelineCfg.getUseAutoMaxYear());
    this->timelineCfg.timesteps = original.timelineCfg.timesteps; // vector copy
    this->timelineCfg.timeframes = original.timelineCfg.timeframes; // vector copy

    return *this;
}


/**
 * Initialize with default settings.
 */
bool Settings::init(const std::string& configPath)
{
    size_t found;               //
    std::string configDir;      //
	std::ostringstream buffer;  //

    found = configPath.find_last_of("/\\");
    configDir = configPath.substr(0, found);

	buffer << configDir << "/plugins";
	this->pluginsPath = buffer.str();
	buffer.str("");

	buffer << configDir << "/colormaps";
	this->colormapsPath = buffer.str();
	buffer.str("");

	buffer << configDir << "/operators";
	this->operatorsPath = buffer.str();
	buffer.str("");

	buffer << configDir << "/data";
	this->dataPath = buffer.str();
	buffer.str("");

	this->glRenderer = "OpenGL3";
	this->glContextMajor = 3;
	this->glContextMinor = 3;
	this->windowWidth = 800;
	this->windowHeight = 800;
	this->vsync = false;
	this->antialiasing = true;
	this->antialiasingType = "msaa";
	this->antialiasingFactor = 8;
	this->fullscreen = false;
	this->vsync = true;
	this->worldMap = "earth.png";
	this->skipIndexCheck = false;
	this->timelineCfg.setMinYear(1980);
	this->timelineCfg.setMaxYear(2020);
	this->timelineCfg.defaultTimespan = "";
	this->timelineCfg.defaultTimestep = "";
    this->timelineCfg.useNearestMode = true;
	this->minGranuleScreenEstate = 0.01f;
	this->processorEnabled = false;
	this->processorAddress = "127.0.0.1";
	this->processorPort = 53450;
    this->searchRequestMinDelay = 0.1;

	return true;
};

/**
 * Modify a setting.
 */
template <typename T>
bool Settings::set(const SettingsKey, const T /*value*/)
{
	return false;
};

/**
 * Retrieve a setting.
 */
template <typename T>
const T Settings::get(const SettingsKey) const
{
	return SettingsValue::NOT_FOUND;
};
