// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <thread>
#include <condition_variable>

#include "core/log.h"
#include "core/sockets_manager.h"

#if defined _WIN32 || defined __APPLE__
    #include "platform/server.h"
#endif

using namespace SEAScope;

/**
 * Trick to silence warnings about unused variables in a portable way.
 *
 * https://stackoverflow.com/a/39746834
 */
template <typename T>
void trickForPotentiallyUnusedVariable(T&& /*No name*/) { /*Empty*/ }

/**
*
*/
SocketsManager::SocketsManager()
: monitor{nullptr}, factory{nullptr}, signalFd{-1}
#if defined _WIN32 || defined __APPLE__
, _server{nullptr}
#endif
{
};

/**
*
*/
SocketsManager::~SocketsManager()
{
#if defined _WIN32 || defined __APPLE__
	delete this->_server;
	this->_server = nullptr;
#endif
};

/**
*
*/
bool SocketsManager::shutdown()
{
	{
		std::unique_lock<std::mutex> lock(this->eventsMutex);
		this->events.emplace_back(false, true, -1, nullptr, nullptr);
	}

#if defined _WIN32 || defined __APPLE__
	// Set a flag in stop position to terminate the thread loop 
	_server->stop();
#endif

	this->monitor->triggerEventsProcessing();

	if (false == threads.empty())
	{
		this->threads[0].join();
	}

	this->protocolHandlers.clear();
	LOG(LogLevel::Debug, "Sockets manager", "Shut down complete");
	return true;
}

/**
*
*/
bool SocketsManager::init(ISocketsMonitor* monitorIn, ISocketFactory* factoryIn)
{
	this->monitor = monitorIn;
	this->factory = factoryIn;

	this->monitor->init([this]() {this->handleEvents(); }, this->signalFd);

#if defined __linux__
	this->threads.emplace_back([this]() {this->monitor->monitorSockets(); });
#elif defined _WIN32 || defined __APPLE__
	_server = new Server;
#endif

	return true;
};

/**
*
*/
bool SocketsManager::handleEvents()
{
	LOG(LogLevel::Debug, "SocketsManager", "Handle events...");
	{
		std::unique_lock<std::mutex> lock(this->eventsMutex);

		for
			(const auto& event : this->events)
		{
			this->monitor->handleEvent(event);
		}

		this->events.clear();
	}

	return true;
};

/**
*
*/
bool SocketsManager::registerProtocol(const std::string& protocol, const std::function<IProtocolHandler*()>& factoryIn)
{
	LOG(LogLevel::Debug, "Sockets manager", "REGISTER REGISTER REGISTER!!!");
	auto found = this->protocolHandlers.find(protocol);
	if
		(this->protocolHandlers.end() != found)
	{
		LOG(LogLevel::Warn, "SocketsManager"
			, "Protocol {} is already registered", protocol);
		return false;
	}

	this->protocolHandlers[protocol] = factoryIn;

	LOG(LogLevel::Debug, "Sockets manager", "New protocol registered: {}", protocol);
	return true;
};

/**
*
*/
IProtocolHandler* SocketsManager::getProtocolHandler(const std::string& protocol, bool& ok)
{
	IProtocolHandler* result;	//

	result = nullptr;
	ok = false;

	auto found = this->protocolHandlers.find(protocol);
	if
		(this->protocolHandlers.end() == found)
	{
		LOG(LogLevel::Error, "SocketsManager", "No handler factory for protocol {}", protocol);
		ok = false;
		return nullptr;
	}

	auto factoryMethod = this->protocolHandlers[protocol];
	result = factoryMethod();
	ok = true;
	return result;
};

/**
*
*/
bool SocketsManager::subscribe(socket_t fd, const std::string& protocol)
{
	IProtocolHandler* handler;	//
	bool ok;	//

	ok = false;
	handler = nullptr;

	handler = this->getProtocolHandler(protocol, ok);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "SocketsManager"
			, "Could not find a suitable protocol handler");

		return false;
	}
	LOG(LogLevel::Debug, "Sockets manager", "Handler ok");

	{
		std::unique_lock<std::mutex> lock(this->eventsMutex);
		this->events.emplace_back(true, false, fd, handler, nullptr);
	}

	LOG(LogLevel::Debug, "Sockets manager", "Event ok");

	this->monitor->triggerEventsProcessing();
	return true;
};

/**
*
*/
bool SocketsManager::unsubscribe(socket_t fd)
{
	{
		std::unique_lock<std::mutex> lock(this->eventsMutex);
		this->events.emplace_back(false, false, fd, nullptr, nullptr);
	}

	this->monitor->triggerEventsProcessing();

	return true;
};

/**
*
*/
bool SocketsManager::createServer(const std::string& address, std::int64_t port, const std::string& protocol)
{
    trickForPotentiallyUnusedVariable(protocol);

#if defined _WIN32 || defined __APPLE__
	_server->initialize(address, static_cast<int>(port), this);

	threads.emplace_back(&Server::run, _server);

	return true;
#elif defined __linux__
	socket_t sock;	//
	bool ok;	//

	sock = this->factory->create(address, port, ok);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Sockets manager", "Could not create listening socket");
		return false;
	}

	ok = this->factory->nonBlocking(sock);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Sockets manager", "Could not make the listening socket non-blocking");
		this->factory->shutdown(sock);
		return false;
	}

	ok = this->factory->listening(sock);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Sockets manager", "Could not listen for incoming connections");
		this->factory->shutdown(sock);
		return false;
	}

	{
		std::unique_lock<std::mutex> lock(this->eventsMutex);
		this->events.emplace_back(true, false, sock, nullptr
			, [this, protocol](bool& success) -> IProtocolHandler* {
			IProtocolHandler* protocolHandler;	//

			protocolHandler = this->getProtocolHandler(protocol, success);
			if
				(false == success)
			{
				LOG(LogLevel::Error, "Sockets manager", "Could not find protocol handler for {}", protocol);
				return nullptr;
			}

			success = true;
			return protocolHandler;
		});
	}

	LOG(LogLevel::Debug, "Sockets manager", "Socket server registered with fd {}", sock);

	this->monitor->triggerEventsProcessing();
	return true;
#else
    return false;
#endif
};

/**
 *
 */
bool SocketsManager::resolveHostname(const std::string& name, std::string& ipAddress)
{
    return this->factory->resolveHostname(name, ipAddress);
}
