// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/sources_control.h"
#include "ui/isource_widget.h"
#include "core/dynamic.h"
#include "core/interfaces/iindex_service_connector.h"

using namespace SEAScope;

/**
 *
 */
SourcesControl::SourcesControl(): view(nullptr), sourceInitialized(nullptr)
{
};

/**
 *
 */
SourcesControl::~SourcesControl()
{
};

/**
 *
 */
bool SourcesControl::init(ISourcesControlView* viewIn)
{
	this->view = viewIn;

	this->view->setAddCallback([this](const std::string& sourceURI) {
		this->addSourceFromURI(sourceURI);
	});

	this->view->setInfoCallback([this](const std::string& sourceURI) {
		SourceInfo info;	//
		bool success;	//

		success = this->model.getSource(sourceURI, info);
		if
			(false == success)
		{
			LOG( LogLevel::Error, "SourcesControl"
					, "Could not find information about source {}"
					, sourceURI);
			return;
		}

		/* TODO: display information in the view */
	});

	this->view->setSelectionCallback([] (const std::string& /*sourceURI*/, bool /*selected*/) {
		/* TODO: implement this */
	});

	return true;
};

/**
 *
 */
bool SourcesControl::addSourceFromURI(const std::string& sourceURI)
{
	ISourceWidget* w;	//
	SourceInfo info;	//
	bool ok;	//

	ok = this->model.hasSource(sourceURI);
	if
		(false == ok)
	{
		LOG( LogLevel::Debug, "SourcesControl"
				, "Source {} already in the list"
				, sourceURI);
		return false;
	}

	/* TODO: ask source more info about itself */
	info.title = sourceURI;
	info.description = sourceURI;

	ok = this->model.addSource(sourceURI, info);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "SourcesControl"
				, "Could not add {} in the sources list"
				, sourceURI);
		return false;
	}

	w = this->view->createSource();
	w->setTitle(sourceURI);
	w->setURI(sourceURI);

	this->view->addSource(w);

	ok = this->initializeSource(info);
	return ok;
};

/**
 *
 */
bool SourcesControl::initializeSource(const SourceInfo& info)
{
	dynamic<IIndexServiceConnector>* catalogue;	//
	std::set<std::uint32_t> localIds;	//
	std::set<CollectionId> collectionIds;	//
	bool ok;	//

	/* List collections provided by the source */
	catalogue = this->model.getIndex(info.id, ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "SourcesControl"
				, "Could not find a suitable catalogue for source {}"
				, info.id);
		return false;
	}

	/* Retrieve source-local identifiers */
	ok = (*catalogue)->listCollections(localIds);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "SourcesControl"
				, "Could not list collections for source {}"
				, info.id);
		return false;
	}

	/* Build global identifiers */
	for
		(const auto& it : localIds)
	{
		LOG( LogLevel::Debug, "SourcesControl"
				, "Source contains collection {}"
				, it);
		collectionIds.emplace(info.id, it);
	}

	if
		(nullptr != this->sourceInitialized)
	{
		this->sourceInitialized(info, collectionIds);
	}

	return true;
};
