// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/path.h"
#include "core/settings.h"
#include "core/sources_control_model.h"
#include "core/dialog.h"

using namespace SEAScope;

/**
 *
 */
SourcesControlModel::SourcesControlModel(): idGenerator(3)
{
};

/**
 *
 */
SourcesControlModel::~SourcesControlModel()
{
};

/**
 *
 */
bool SourcesControlModel::init(PluginsManager& pMgr, const Settings& settings)
{
	const std::string indexAPI(IIndexServiceConnector::api());
	const std::string datastreamAPI(IDatastreamServiceConnector::api());

	LOG( LogLevel::Debug, "SourcesControlModel"
			, "*** Init plugins for sources ***");

	for
		(auto& it : pMgr.metadata)
	{
		LOG( LogLevel::Debug, "SourcesControlModel"
				, "Check plugin {}"
				, it.first);
		if
			(indexAPI == it.second.api)
		{
			LOG( LogLevel::Debug, "SourcesControlModel"
					, "Index plugin found: {}"
					, it.second.name);
			this->indexPlugins.emplace_front(it.second.name, it.second.version);
			(*this->indexPlugins.begin())->init(settings, Log::implementation, Path::implementation, Dialog::implementation);
			continue;
		}

		if
			(datastreamAPI == it.second.api)
		{
			LOG( LogLevel::Debug, "SourcesControlModel"
					, "Datastream plugin found: {}"
					, it.second.name);
			this->datastreamPlugins.emplace_front(it.second.name, it.second.version);
			(*this->datastreamPlugins.begin())->init(Log::implementation, Path::implementation);
			continue;
		}

		LOG( LogLevel::Debug, "SourcesControlModel"
				, "Unsuitable plugin: API = {}"
				, it.second.api);
	}
	return true;
};

bool SourcesControlModel::reset()
{
    for
        (auto& indexPlugin : this->indexPlugins)
    {
        indexPlugin->reset();
    }
	this->indexPlugins.clear();

    for
        (auto& datastreamPlugin : this->datastreamPlugins)
    {
        datastreamPlugin->reset();
    }
	this->datastreamPlugins.clear();

	this->indices.clear();
	this->datastreams.clear();

    return true;
}

/**
 *
 */
bool SourcesControlModel::hasSource(const std::string& uri) const
{
	bool ok;	//
	auto found = this->sources.find(uri);
	ok = (this->sources.end() == found);
	return ok;
};

/**
 *
 */
bool SourcesControlModel::addSource(const std::string& uri, SourceInfo& info)
{
	bool ok;	//
	auto found = this->sources.find(uri);
	if
		(this->sources.end() != found)
	{
		return false;
	}

	info.id = this->idGenerator;
	this->sources[uri] = info;
	this->uris[info.id] = uri;

	/* Look for plugins */
	ok = false;

	this->getIndex(info.id, ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "SourcesManager"
				, "Could not find a suitable plugin for the index of source {}"
				, info.id);
		return false;
	}

	this->getDatastream(info.id, ok);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "SourcesManager"
				, "Could not find a suitable plugin for the datastream of source {}"
				, info.id);
		return false;
	}

	this->idGenerator = this->idGenerator + 1;

	return true;
};

/**
 *
 */
bool SourcesControlModel::getSource(const std::string& uri, SourceInfo& info) const
{
	auto found = this->sources.find(uri);
	if
		(this->sources.end() == found)
	{
		return false;
	}

	info = found->second;
	return true;
};

/**
 *
 */
dynamic<IIndexServiceConnector>* SourcesControlModel::getIndex(SourceId sourceId, bool& ok)
{
	ok = false;
	auto inCache = this->indices.find(sourceId);
	if
		(this->indices.end() != inCache)
	{
		ok = true;
		return inCache->second;
	}

	auto found = this->uris.find(sourceId);
	if
		(this->uris.end() == found)
	{
		LOG( LogLevel::Error, "SourcesManager"
				, "Could not find URI for source {}"
				, sourceId);
		ok = false;
		return nullptr;
	}

	for
		(auto& it : this->indexPlugins)
	{
		if
			(it->canHandle(found->second))
		{
			ok = true;
			this->indices[sourceId] = &it;
			LOG( LogLevel::Debug, "SourceControlModel"
					, "Index found for source {}"
					, sourceId);
			return &it;
		}
	}
	return nullptr;
};

/**
 *
 */
dynamic<IDatastreamServiceConnector>* SourcesControlModel::getDatastream(SourceId sourceId, bool& ok)
{
	ok = false;
	auto inCache = this->datastreams.find(sourceId);
	if
		(this->datastreams.end() != inCache)
	{
		ok = true;
		return inCache->second;
	}

	auto found = this->uris.find(sourceId);
	if
		(this->uris.end() == found)
	{
		LOG( LogLevel::Error, "SourcesManager"
				, "Could not find URI for source {}"
				, sourceId);
		ok = false;
		return nullptr;
	}

	for
		(auto& it : this->datastreamPlugins)
	{
		if
			(it->canHandle(found->second))
		{
			ok = true;
			this->datastreams[sourceId] = &it;
			LOG( LogLevel::Debug, "SourceControlModel"
					, "Datastream found for source {}"
					, sourceId);
			return &it;
		}
	}
	return nullptr;
};
