// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/guid.h"
#include "core/threads_pool.h"

using namespace SEAScope;

/**
 *
 */
ThreadsPool::ThreadsPool(): terminationRequested(false), terminated(false), scheduler(nullptr)
{
};

/**
 *
 */
ThreadsPool::~ThreadsPool()
{
	if
		(false == this->terminated)
	{
		this->shutdown();
	}
};

/**
 *
 */
bool ThreadsPool::init(std::uint16_t threadsCount, IJobScheduler* schedulerIn)
{
	this->scheduler = schedulerIn;
	while
		(0 < threadsCount--)
	{
		this->threads.emplace_back(std::thread(&ThreadsPool::processJobs, this));
	}
	return nullptr != this->scheduler;
};

/**
 *
 */
bool ThreadsPool::submitJob(Job&& job, std::string& jobId, JobStatus& status)
{
	bool ok {false};	//

	// Scope based locking.
	{
		// Put unique lock on task mutex.
		std::unique_lock<std::mutex> lock(this->tasksMutex);

		// Push task into queue.
		ok = this->scheduler->addJob(std::forward<Job&&>(job), jobId, status);
	}

	LOG(LogLevel::Debug, "ThreadsPool", "job submitted: [{}] {}", ok, jobId);

	// Wake up one thread if the job has been accepted by the scheduler AND
    // the scheduler did not optimize it out.
	if
		(true == ok)
	{
        if
            (JobStatus::CANCELLED == status)
        {
            LOG(LogLevel::Debug, "Threads pool", "Scheduler cancelled the job");
        }
        else if
            (JobStatus::SUPERSEEDED == status)
        {
            LOG(LogLevel::Debug, "Threads pool", "Scheduler found a superior version of the job");
        }
        else
        {
		    this->condition.notify_one();
        }
	}
    else
    {
        LOG(LogLevel::Error, "Threads pool", "Scheduler refused to add job");
    }

	return ok;
};

/**
 *
 */
void ThreadsPool::processJobs()
{
	Job job;	//
	bool ok {false};	//

	while
		(true)
	{
		// Scope based locking.
		{
			// Put unique lock on task mutex.
			std::unique_lock<std::mutex> lock(this->tasksMutex);

			// Wait until queue is not empty or termination signal is sent.
			this->condition.wait(lock, [this]{ return (0 < this->scheduler->countPendingJobs()) || this->terminationRequested; });

			// If termination signal received and queue is empty then exit else continue clearing the queue.
			if
				(  this->terminationRequested
				&& 0 >= this->scheduler->countPendingJobs())
			{
				return;
			}

			// Get next task in the queue.
			ok = this->scheduler->selectJob(job);

            // If a job has been selected, mark it as running while the code
            // is still in the critical section as the status update may have
            // consequences on concurrent job selections
            if
                (true == ok)
            {
                this->scheduler->updateJobStatus(job.id, JobStatus::RUNNING);
            }
		}

        if
            (true == ok)
        {
            // Execute the task.
            job.start();
            /*ok = */job.task(job.errorMsg);
            job.done();

            {
                std::lock_guard<std::mutex> okLock(this->tasksMutex);
                this->scheduler->updateJobStatus(job.id, JobStatus::COMPLETE);
            }

            // Required if the task is a lambda coming from a plugin
            job.task = nullptr;
        }
	}
};

/**
 *
 */
bool ThreadsPool::cancelJob(const std::string& jobId)
{
	bool ok;

	ok = this->scheduler->removeJob(jobId);

	return ok;
};

/**
 *
 */
bool ThreadsPool::shutdown()
{
	// Scope based locking.
	{
		// Put unique lock on task mutex.
		std::unique_lock<std::mutex> lock(this->tasksMutex);

		// Set termination flag to true.
		this->terminationRequested = true;
	}

	// Wake up all threads.
	this->condition.notify_all();

	// Join all threads.
	for(std::thread& thread : this->threads)
	{
		thread.join();
	}

	// Empty workers vector.
	this->threads.clear();

	// Indicate that the pool has been shut down.
	this->terminated = true;

	return true;
};
