// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>
#include <iomanip>
#include <sstream>
#include <vector>
#include <string>
#include "core/log.h"
#include "core/structs.h"
#include "core/timeline.h"

#include "core/time_utils.h"

using namespace SEAScope;

/**
 *
 */
Timeline::Timeline(): view{nullptr}, currentCallback{nullptr}, timeframeCallback{nullptr}
{

};

/**
 *
 */
Timeline::~Timeline()
{

};

/**
 *
 */
bool Timeline::init( ITimelineView* viewIn, const TimelineCfg& cfg)
{
	std::vector<std::string> timeframeLabels;	//
	std::vector<std::string> timeframeTooltips;	//
	std::vector<std::string> timestepLabels;	//
	std::vector<std::string> timestepTooltips;	//
	timestamp_t lowerBound;	//
	timestamp_t upperBound;	//
	std::tm minYear = {};	//
	std::tm maxYear = {};	//

	if
		(nullptr == viewIn)
	{
		return false;
	}
	this->view = viewIn;

	minYear.tm_year = cfg.getMinYear() - 1900;
	minYear.tm_mon = 0;
	minYear.tm_mday = 1;
	minYear.tm_hour = 0;
	minYear.tm_min = 0;
	minYear.tm_sec = 0;
	lowerBound = static_cast<timestamp_t>(std::mktime(&minYear));

	maxYear.tm_year = cfg.getMaxYear() + 1 - 1900;
	maxYear.tm_mon = 0;
	maxYear.tm_mday = 1;
	maxYear.tm_hour = 0;
	maxYear.tm_min = 0;
	maxYear.tm_sec = 0;
	upperBound = static_cast<timestamp_t>(std::mktime(&maxYear));

	cfg.getTimeframes(this->timeframes);
	cfg.getTimesteps(this->timesteps);

	cfg.getTimeframes(timeframeLabels, timeframeTooltips);
	cfg.getTimesteps(timestepLabels, timestepTooltips);
	this->view->init(timeframeLabels, timeframeTooltips, timestepLabels, timestepTooltips, cfg.getMinYear(), cfg.getMaxYear(), cfg.advancedScrolling);

	this->view->setCurrentChangedCallback([this](const std::string& value, const std::string& origin) {
		time_t t;	//
		timestamp_t timestamp;	//
		std::tm ts = {};	//
		std::string date;	//
		std::string time;	//
		int year;	//
		int month;	//
		int day;	//
		bool ok;	//

		strptime(value.c_str(), "%Y-%m-%d %H:%M:%S", &ts);
		t = std::mktime(&ts);
		timestamp = static_cast<timestamp_t>(t);

		LOG(LogLevel::Debug, "Timeline", "Parsed day/hour for {}: {}/{} {}", value, ts.tm_mday, ts.tm_hour, timestamp);

        if
            (("yearButton" == origin) || ("monthButton" == origin) || ("dayButton" == origin))
        {
            // Set time to 12:00:00
            ts.tm_hour = 12;
            ts.tm_min = 0;
            ts.tm_sec = 0;
            t = std::mktime(&ts);
            timestamp = static_cast<timestamp_t>(t);
        }

        if
            (true == this->model.nearestMode)
        {
            std::int16_t targetYear {0};   //
            targetYear = static_cast<std::int16_t>(1900 + ts.tm_year);

            std::uint8_t targetMonth {0};  //
            targetMonth = static_cast<std::uint8_t>(1 + ts.tm_mon);

            std::uint8_t targetDay {0};    //
            targetDay = static_cast<std::uint8_t>(ts.tm_mday);

            bool keepYear {false};  //
            keepYear = ("yearButton" == origin) && this->model.isDataAvailable(targetYear);

            bool keepMonth {false}; //
            keepMonth = ("monthButton" == origin) && this->model.isDataAvailable(targetYear, targetMonth);

            bool keepDay {false};   //
            keepDay = ("dayButton" == origin) && this->model.isDataAvailable(targetYear, targetMonth, targetDay);

            if
                (keepYear || keepMonth || keepDay)
            {
                this->model.applyNearestMode(timestamp, keepYear, keepMonth, keepDay);
            }
        }

		ok = this->model.setCurrent(timestamp);
		if
			(false == ok)
		{
			// Display actual current datetime in view
			ok = this->extractDatetimeFromTimestamp( this->model.current, date, time, year, month, day);
			this->view->setCurrent( date, time, year, month, day);
		}
	});
	this->view->setTimeframeChangedCallback([this](const std::string& label) {
        if
            (this->timeframes.end() == this->timeframes.find(label))
        {
            LOG(LogLevel::Error, "Timeline", "Unknown '{}' timespan", label);
            return;
        }

		timestamp_t pastOffset;	//
		timestamp_t futureOffset;	//

		pastOffset = this->timeframes[label][0];
		futureOffset= this->timeframes[label][1];
		this->model.setTimeframeOffsets(pastOffset, futureOffset);
	});
	this->view->setTimestepChangedCallback([this](const std::string& label) {
        if
            (this->timesteps.end() == this->timesteps.find(label))
        {
            LOG(LogLevel::Error, "Timeline", "Unknown '{}' timestep", label);
            return;
        }

		timestamp_t timestep;	//

		timestep = this->timesteps[label];
		this->model.setTimestep(timestep);
	});
	this->view->setPrevCallback([this]() {
		this->model.goToPrev();
	});
	this->view->setNextCallback([this]() {
		this->model.goToNext();
	});
	this->view->setOptionsChangedCallback([this]() {
        bool nearestMode {false};	//

        nearestMode = this->view->nearestMode();

        this->model.setNearestMode(nearestMode);

        LOG( LogLevel::Debug, "Timeline", "Update options: {}"
                , nearestMode);
	});

    if
        (true == cfg.advancedScrolling)
    {
        this->view->setYearScrollCallback([this](float delta) {
            float syntoolSign = -1.0;   //

            // Normalize sign parameter
            delta = static_cast<float>((0.f < delta) - (delta < 0.f));
            this->model.shiftByYears(static_cast<std::int16_t>(syntoolSign * delta));    // int16_t
        });
        this->view->setMonthScrollCallback([this](float delta) {
            float syntoolSign = -1.0;   //

            // Normalize sign parameter
            delta = static_cast<float>((0.f < delta) - (delta < 0.f));
            this->model.shiftByMonths(static_cast<std::int32_t>(syntoolSign * delta));   // int32_t
        });
    }
    else
    {
        this->view->setYearScrollCallback([this](float delta) {
            float syntoolSign = -1.0;   //

            this->model.shiftByTimeframeLength(syntoolSign * delta, 1.f/3.f);
        });
        this->view->setMonthScrollCallback([this](float delta) {
            float syntoolSign = -1.0;   //

            this->model.shiftByTimeframeLength(syntoolSign * delta, 1.f/3.f);
        });
    }
    this->view->setDayScrollCallback([this](float delta) {
        float syntoolSign = -1.0;   //

        this->model.shiftByTimeframeLength(syntoolSign * delta, 1.f/3.f);
    });

	this->model.setBounds(lowerBound, upperBound);
	this->model.setCurrentCallback([this](timestamp_t value) {
		std::string date;	//
		std::string time;	//
		int year;	//
		int month;	//
		int day;	//

	    this->extractDatetimeFromTimestamp(value, date, time, year, month, day);
		LOG(LogLevel::Debug, "Timeline", "set view date to {} {}",date, time);
		this->view->setCurrent( date, time, year, month, day);
		this->view->setVisibleBounds( value - this->model.timeframePastOffset
									, value + this->model.timeframeFutureOffset);

		if
			(nullptr != this->currentCallback)
		{
			this->currentCallback(value);
		}
	});
	this->model.setBoundsCallback([] (timestamp_t /*min*/, timestamp_t /*max*/) {

	});
	this->model.setTimeframeCallback([this, &cfg](timestamp_t pastOffset, timestamp_t futureOffset) {
		Timeframe t;	//
		bool ok;	//

		ok = cfg.getTimeframe(pastOffset, futureOffset, t);
		if
			(true == ok)
		{
			this->view->setTimeframe(t.label, t.description);
			this->view->setVisibleBounds( this->model.current - this->model.timeframePastOffset
										, this->model.current + this->model.timeframeFutureOffset);
		}
		else
		{
			LOG( LogLevel::Error, "Timeline", "Timeframe with offsets [-{}, {}] could not be found", static_cast<int>(pastOffset), static_cast<int>(futureOffset));
		}

		if
			(nullptr != this->timeframeCallback)
		{
			this->timeframeCallback(pastOffset, futureOffset);
		}
	});
	this->model.setTimestepCallback([this, &cfg](timestamp_t timestep) {
		Timestep t;	//
		bool ok;	//

		ok = cfg.getTimestep(timestep, t);
		if
			(true == ok)
		{
			this->view->setTimestep(t.label, t.description);
		}
		else
		{
			LOG( LogLevel::Error, "Timeline", "Timestep with value {} could not be found", static_cast<int>(timestep));
		}
	});
	this->model.setCoverageCallback([this]() {
		EventsCoverage united;	//
		EventsCoverage intersection;	//

		this->model.getCoverage("union", united);
		this->view->setHintsSource(united);
		this->view->setMainCoverage(united);

		this->model.getCoverage("intersection", intersection);
		this->view->setExtraCoverage(intersection);
	});

    timestamp_t initTime {0};   //
    initTime = cfg.getInitialDatetime();
	if
		(initTime < this->model.lowerBound)
	{
		initTime = this->model.lowerBound;
	}
	if
		(initTime > this->model.upperBound)
	{
		initTime = this->model.upperBound - 1;
	}
    this->model.setCurrent(initTime);

	if
		(0 < cfg.defaultTimespan.length())
	{
		this->model.setTimeframeOffsets( this->timeframes[cfg.defaultTimespan][0]
									   , this->timeframes[cfg.defaultTimespan][1]);
	}
	else if
		(0 < timeframeLabels.size())
	{
		this->model.setTimeframeOffsets( this->timeframes[timeframeLabels[0]][0]
									   , this->timeframes[timeframeLabels[0]][1]);
	}

	if
		(0 < cfg.defaultTimestep.length())
	{
		this->model.setTimestep( this->timesteps[cfg.defaultTimestep]);
	}
	else if
		(0 < timestepLabels.size())
	{
		this->model.setTimestep( this->timesteps[timestepLabels[0]]);
	}

	/* Set options */
	this->view->setNearestMode(cfg.useNearestMode);
    this->model.setNearestMode(cfg.useNearestMode);

	return true;
};

/**
 *
 */
bool Timeline::extractDatetimeFromTimestamp(timestamp_t value, std::string& dateResult, std::string& timeResult, int& year, int& month, int& day) const
{
	time_t timestamp;	//
	std::tm* ltime;	//
	char dateChr[11];	//
	char timeChr[9];	//

	timestamp = static_cast<time_t>(value);
	ltime = std::localtime(&timestamp);
	if
		(nullptr == ltime)
	{
		LOG( LogLevel::Error, "Timeline"
				, "Could not convert timestamp {} to time object"
				, timestamp);
		return false;
	}

	strftime(dateChr, sizeof(dateChr), "%Y-%m-%d", ltime);
	dateResult = dateChr;

	strftime(timeChr, sizeof(timeChr), "%H:%M:%S", ltime);
	timeResult = timeChr;

	year = 1900 + ltime->tm_year;
	month = 1 + ltime->tm_mon;
	day = ltime->tm_mday;

	return true;
};

/**
 *
 */
bool Timeline::setCurrentChangedCallback(const std::function<void(timestamp_t)>& callback)
{
	this->currentCallback = callback;
	return true;
};

/**
 *
 */
bool Timeline::setTimeframeChangedCallback(const std::function<void(timestamp_t, timestamp_t)>& callback)
{
	this->timeframeCallback = callback;
	return true;
};

/**
 *
 */
bool Timeline::getBounds(timestamp_t& current, timestamp_t& begin, timestamp_t& end) const
{
	current = this->model.current;
	begin = this->model.current - this->model.timeframePastOffset;
	end = this->model.current + this->model.timeframeFutureOffset;
	LOG( LogLevel::Debug, "Timeline", "getBounds {} {} {}"
			, this->model.current, this->model.timeframePastOffset
			, this->model.timeframeFutureOffset);
	return true;
};

/**
 *
 */
bool Timeline::getTimestepLabel(timestamp_t value, std::string& result) const
{
	for
		(const auto& it : this->timesteps)
	{
		if
			(it.second == value)
		{
			result = it.first;
			return true;
		}
	}

	return false;
}

/**
 *
 */
bool Timeline::getTimespanLabel(timestamp_t pastOffset, timestamp_t futureOffset, std::string& result) const
{
	for
		(const auto& it : this->timeframes)
	{
		if
			((it.second[0] == pastOffset) && (it.second[1] == futureOffset))
		{
			result = it.first;
			return true;
		}
	}

	return false;
}

/**
 *
 */
bool Timeline::getState(TimelineState& value) const
{
    bool ok {false};    //

    ok = this->getTimespanLabel(this->model.timeframePastOffset, this->model.timeframeFutureOffset, value.timespan);
    if
        (false == ok)
    {
        return false;
    }

    ok = this->getTimestepLabel(this->model.timestep, value.timestep);
    if
        (false == ok)
    {
        return false;
    }

    std::string dateStr {""};   //
    std::string timeStr {""};   //
    int year {0};   //
    int month {0};  //
    int day {0};    //
    ok = this->extractDatetimeFromTimestamp(this->model.current, dateStr, timeStr, year, month, day);
    if
        (false == ok)
    {
        return false;
    }
    std::ostringstream buffer {""};
    buffer << dateStr << " " << timeStr;
    value.datetime = buffer.str();

    value.nearestMode = this->model.nearestMode;
    value.extendedView = this->view->getExtendedView();

    return true;
}

/**
 *
 */
bool Timeline::setState(const TimelineState& value)
{
    if
        (nullptr == this->view)
    {
        LOG(LogLevel::Error, "Timeline", "Timeline view is required to restore state but it has not been initialized yet");
        return false;
    }

    // Timespan
    auto timespanFound = this->timeframes.find(value.timespan);
    if
        (this->timeframes.end() == timespanFound)
    {
        LOG(LogLevel::Warn, "Timeline", "Cannot restore timespan from previous session, timespan with label '{}' not found in the configuration", value.timespan);
    }
    else
    {
        timestamp_t pastOffset;	//
        timestamp_t futureOffset;	//

        pastOffset = timespanFound->second[0];
        futureOffset= timespanFound->second[1];
        this->model.setTimeframeOffsets(pastOffset, futureOffset);
    }

    // Timestep
    auto timestepFound = this->timesteps.find(value.timestep);
    if
        (this->timesteps.end() == timestepFound)
    {
        LOG(LogLevel::Warn, "Timeline", "Cannot restore timestep from previous session, timestep with label '{}' not found in the configuration", value.timestep);
    }
    else
    {
		this->model.setTimestep(timestepFound->second);
    }

    // Nearest mode
    this->model.setNearestMode(value.nearestMode);
    this->view->setNearestMode(value.nearestMode);

    // Current datetime
    bool ok {false};    //
    time_t t {0};	//
    timestamp_t timestamp {0};	//
    timestamp_t original {0};   //
    std::tm ts = {};	//

    strptime(value.datetime.c_str(), "%Y-%m-%d %H:%M:%S", &ts);
    t = std::mktime(&ts);
    timestamp = static_cast<timestamp_t>(t);
    original = this->model.current;
    ok = this->model.setCurrent(timestamp);
    if
        (false == ok)
    {
        LOG(LogLevel::Debug, "Timeline", "Cannot set current time to {}, fallback to {}", timestamp, original);
        // Restore previous datetime
        timestamp = original;
    }

    std::string date {""};  //
    std::string time {""};  //
    int year {0};   //
    int month {0};  //
    int day {0};    //
    ok = this->extractDatetimeFromTimestamp(timestamp, date, time, year, month, day);
    this->view->setCurrent( date, time, year, month, day);


    this->view->setExtendedView(value.extendedView, false);

    return false;
}
