// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/timeline_cfg.h"
#include <limits>

using namespace SEAScope;

const timestamp_t Timestep::ONE_THIRD_TIMESPAN_VALUE =  std::numeric_limits<timestamp_t>::max() - 1;
const char Timestep::ONE_THIRD_TIMESPAN_LABEL[] = "1/3 span";

/**
 *
 */
Timeframe::Timeframe(const std::string& labelIn, const std::string& descriptionIn, timestamp_t pastOffsetIn, timestamp_t futureOffsetIn)
: label{labelIn}, description{descriptionIn}, pastOffset{pastOffsetIn}, futureOffset{futureOffsetIn}
{
};

/**
 *
 */
Timeframe::Timeframe()
{
};

/**
 *
 */
Timeframe::~Timeframe()
{
};

/**
 *
 */
Timestep::Timestep(const std::string& labelIn, const std::string& descriptionIn, timestamp_t valueIn)
: label{labelIn}, description{descriptionIn}, value{valueIn}
{
};

/**
 *
 */
Timestep::Timestep()
{

};

/**
 *
 */
Timestep::~Timestep()
{
};

const std::uint16_t TimelineCfg::MIN_VALID_YEAR = 1970;
const std::uint16_t TimelineCfg::MAX_VALID_YEAR = 3000;

/**
 *
 */
TimelineCfg::TimelineCfg()
: defaultTimespan{""}, defaultTimestep{""},
  useNearestMode{false}, advancedScrolling{false},
  _initialDatetime{0}, _initialDatetimeStr{""}, _useAutoMaxYear{false},
  _minYear{TimelineCfg::MIN_VALID_YEAR}, _maxYear{TimelineCfg::MAX_VALID_YEAR}
{
};

/**
 *
 */
TimelineCfg::~TimelineCfg()
{
};

/**
 *
 */
bool TimelineCfg::addTimeframe(const std::string& label, const std::string& description, timestamp_t pastOffset, timestamp_t futureOffset)
{
	/* TODO: check that label is available */
	this->timeframes.emplace_back(label, description, pastOffset, futureOffset);
	return true;
};

/**
 *
 */
bool TimelineCfg::getTimeframe(const std::string& label, Timeframe& result) const
{
	std::vector<Timeframe>::const_iterator it;	//

	it = this->timeframes.begin();
	for
		(; it != this->timeframes.end(); ++it)
	{
		if
			(label == it->label)
		{
			result = *it;
			return true;
		}
	}

	return false;
};

/**
 *
 */
bool TimelineCfg::getTimeframe(timestamp_t pastOffset, timestamp_t futureOffset, Timeframe& result) const
{
	std::vector<Timeframe>::const_iterator it;	//

	it = this->timeframes.begin();
	for
		(; it != this->timeframes.end(); ++it)
	{
		if
			(pastOffset == it->pastOffset && futureOffset == it->futureOffset)
		{
			result = *it;
			return true;
		}
	}

	return false;
};

/**
 *
 */
bool TimelineCfg::getTimeframes(std::vector<std::string>& labels, std::vector<std::string>& descriptions) const
{
	std::vector<Timeframe>::const_iterator it;	//

	it = this->timeframes.begin();
	for
		(; it != this->timeframes.end(); ++it)
	{
		labels.push_back(it->label);
		descriptions.push_back(it->description);
	}
	return true;
};

/**
 *
 */
bool TimelineCfg::getTimeframes(std::map<std::string, std::array<timestamp_t, 2>>& result) const
{
	std::vector<Timeframe>::const_iterator it;	//

	it = this->timeframes.begin();
	for
		(; it != this->timeframes.end(); ++it)
	{
		result.insert(std::pair<std::string, std::array<timestamp_t, 2>>(it->label, {{it->pastOffset, it->futureOffset}}));
	}
	return true;
};

/**
 *
 */
bool TimelineCfg::removeTimeframe(const std::string& label, std::string& newDefault)
{
    if
        (1 >= this->timeframes.size())
    {
        // Refuse to delete timeframe if there is only one
        return false;
    }

    auto it = this->timeframes.begin();
    for
        (; it != this->timeframes.end(); ++it)
    {
        if
            (it->label == label)
        {
            break;
        }
    }

    if
        (this->timeframes.end() == it)
    {
        // Not found
        return false;
    }

    this->timeframes.erase(it);

    if
        (label == this->defaultTimespan)
    {
        // Remove item was default, find another one (we know there is at
        // least one in the timeframes vector)
        newDefault = this->timeframes[0].label;
    }
    else
    {
        newDefault = this->defaultTimespan;
    }

    return true;
}

/**
 *
 */
bool TimelineCfg::addTimestep(const std::string& label, const std::string& description, timestamp_t value)
{
	/* TODO: check that label is available */
	this->timesteps.emplace_back(label, description, value);
	return true;
};

/**
 *
 */
bool TimelineCfg::getTimestep(const std::string& label, Timestep& result) const
{
	std::vector<Timestep>::const_iterator it;	//

	it = this->timesteps.begin();
	for
		(; it != this->timesteps.end(); ++it)
	{
		if
			(label == it->label)
		{
			result = *it;
			return true;
		}
	}

	return false;
};

/**
 *
 */
bool TimelineCfg::getTimestep(timestamp_t value, Timestep& result) const
{
	std::vector<Timestep>::const_iterator it;	//

	it = this->timesteps.begin();
	for
		(; it != this->timesteps.end(); ++it)
	{
		if
			(value == it->value)
		{
			result = *it;
			return true;
		}
	}

	return false;
};

/**
 *
 */
bool TimelineCfg::getTimesteps(std::vector<std::string>& labels, std::vector<std::string>& descriptions) const
{
	std::vector<Timestep>::const_iterator it;	//

	it = this->timesteps.begin();
	for
		(; it != this->timesteps.end(); ++it)
	{
		labels.push_back(it->label);
		descriptions.push_back(it->description);
	}
	return true;
};

/**
 *
 */
bool TimelineCfg::getTimesteps(std::map<std::string, timestamp_t>& result) const
{
	std::vector<Timestep>::const_iterator it;	//

	it = this->timesteps.begin();
	for
		(; it != this->timesteps.end(); ++it)
	{
		result.insert(std::pair<std::string, timestamp_t>(it->label, it->value));
	}
	return true;
};

/**
 *
 */
bool TimelineCfg::removeTimestep(const std::string& label, std::string& newDefault)
{
    if
        (1 >= this->timesteps.size())
    {
        // Refuse to delete timeframe if there is only one
        return false;
    }

    auto it = this->timesteps.begin();
    for
        (; it != this->timesteps.end(); ++it)
    {
        if
            (it->label == label)
        {
            break;
        }
    }

    if
        (this->timesteps.end() == it)
    {
        // Not found
        return false;
    }

    this->timesteps.erase(it);

    if
        (label == this->defaultTimestep)
    {
        // Remove item was default, find another one (we know there is at
        // least one in the timesteps vector)
        newDefault = this->timesteps[0].label;
    }

    return true;
}

/**
 *
 */
bool TimelineCfg::setInitialDatetime(timestamp_t initialDatetime)
{
    this->_initialDatetime = initialDatetime;
    return true;
}

/**
 *
 */
timestamp_t TimelineCfg::getInitialDatetime() const
{
    return this->_initialDatetime;
}

/**
 *
 */
bool TimelineCfg::setInitialDatetimeString(const std::string& initialDatetimeStr)
{
    this->_initialDatetimeStr = initialDatetimeStr;
    return true;
}

/**
 *
 */
const std::string& TimelineCfg::getInitialDatetimeString() const
{
    return this->_initialDatetimeStr;
}

/**
 *
 */
bool TimelineCfg::getUseAutoMaxYear() const
{
    return this->_useAutoMaxYear;
}

/**
 *
 */
void TimelineCfg::setUseAutoMaxYear(bool value)
{
    this->_useAutoMaxYear = value;
}

/**
 *
 */
std::uint16_t TimelineCfg::getMinYear() const
{
    return this->_minYear;
}

/**
 *
 */
void TimelineCfg::setMinYear(std::uint16_t value)
{
    if
        (TimelineCfg::MAX_VALID_YEAR < value)
    {
        value = TimelineCfg::MAX_VALID_YEAR;
    }

    if
        (TimelineCfg::MIN_VALID_YEAR > value)
    {
        value = TimelineCfg::MIN_VALID_YEAR;
    }
    this->_minYear = value;
}

/**
 *
 */
std::uint16_t TimelineCfg::getMaxYear() const
{
    return this->_maxYear;
}

/**
 *
 */
void TimelineCfg::setMaxYear(std::uint16_t value)
{
    if
        (TimelineCfg::MIN_VALID_YEAR > value)
    {
        value = TimelineCfg::MIN_VALID_YEAR;
    }

    if
        (TimelineCfg::MAX_VALID_YEAR < value)
    {
        value = TimelineCfg::MAX_VALID_YEAR;
    }
    this->_maxYear = value;
}
