// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>
#include <vector>
#include <algorithm>

#include "core/log.h"
#include "core/utils.h"
#include "core/timeline_model.h"
#include "core/timeline_cfg.h"

using namespace SEAScope;

/**
 *
 */
TimelineModel::TimelineModel(): currentCallback(nullptr), boundsCallback(nullptr), timeframeCallback(nullptr), timestepCallback(nullptr), coverageCallback(nullptr), current(0), lowerBound(0), upperBound(0), timeframePastOffset(0), timeframeFutureOffset(0), timestep(std::numeric_limits<uint64_t>::max()), prev(std::numeric_limits<timestamp_t>::max()), next(std::numeric_limits<timestamp_t>::min()), nearestMode(false), forceNearestMode(false), _nearestDataBoundCallback(nullptr)
{
};

/**
 *
 */
TimelineModel::~TimelineModel()
{
};

/**
 *
 */
bool TimelineModel::getCoverage(const CollectionId& /*collectionId*/, EventsCoverage& /*result*/) const
{
	return false;
};

/**
 *
 */
bool TimelineModel::getCoverage(const std::string& formula, EventsCoverage& result) const
{
	auto itCoverages = this->coverages.begin();
	if
		(this->coverages.end() == itCoverages)
	{
		return false;
	}

	if
		(0 == formula.compare("union"))
	{
        LOG(LogLevel::Debug, "TimelineModel", "Starting Merge");
		for
			(; itCoverages != this->coverages.end(); itCoverages++)
		{
            LOG(LogLevel::Debug, "TimelineModel", "\t Merging collection {}/{}", itCoverages->first.first, itCoverages->first.second);
			result.merge(itCoverages->second);
		}
        LOG(LogLevel::Debug, "TimelineModel", "Merge complete");
		return true;
	}

	if
		(0 == formula.compare("intersection"))
	{
		result.clone(itCoverages->second);
		itCoverages++;
		for
			(; itCoverages != this->coverages.end(); itCoverages++)
		{
			result.intersect(itCoverages->second);
		}
		return true;
	}

	return false;
};

/**
 *
 */
bool TimelineModel::setCoverage(const CollectionId& collectionId, const EventsCoverage& coverage, bool silent)
{
	this->coverages[collectionId] = coverage;
	if
		(false == silent && nullptr != this->coverageCallback)
	{
		this->coverageCallback();
	}
	return true;
};

/**
 *
 */
bool TimelineModel::removeCoverage(const CollectionId& collectionId, bool silent)
{
	auto found = this->coverages.find(collectionId);
	if
		(this->coverages.end() == found)
	{
		return false;
	}
	this->coverages.erase(found);
	if
		(false == silent && nullptr != this->coverageCallback)
	{
		this->coverageCallback();
	}
	return true;
};

/**
 *
 */
bool TimelineModel::updateCoverages()
{
	if
		(nullptr != this->coverageCallback)
	{
		this->coverageCallback();
        return true;
	}
	return false;
}


/**
 *
 */
bool TimelineModel::setCurrent(timestamp_t timestamp)
{
	LOG(LogLevel::Debug, "Timeline model", "set datetime to {} vs {}", timestamp, this->current);

	if
		(  timestamp == this->current
		|| timestamp < this->lowerBound
		|| timestamp >= this->upperBound)
	{
		LOG(LogLevel::Debug, "TimelineModel", "Current not changed {} -> [{}, {}, {}]", this->current, this->lowerBound, timestamp, this->upperBound);
		return false;
	}
	this->current = timestamp;

	if
		(nullptr != this->currentCallback)
	{
		this->currentCallback(timestamp);
	}

	return true;
};

/**
 *
 */
bool TimelineModel::setBounds(timestamp_t lowerBoundIn, timestamp_t upperBoundIn)
{
	LOG(LogLevel::Debug, "Timeline model", "set bounds to [{}, {}]", lowerBoundIn, upperBoundIn);
	if
		(  lowerBoundIn == this->lowerBound
		&& upperBoundIn == this->upperBound)
	{

		return false;
	}
	this->lowerBound = lowerBoundIn;
	this->upperBound = upperBoundIn;

	if
		(nullptr != this->boundsCallback)
	{
		this->boundsCallback(lowerBoundIn, upperBoundIn);
	}

	return true;
};

/**
 *
 */
bool TimelineModel::setTimeframeOffsets(timestamp_t pastOffset, timestamp_t futureOffset)
{
	LOG(LogLevel::Debug, "Timeline model", "set timeframe to [{}, {}]", pastOffset, futureOffset);
	if
		(  pastOffset == this->timeframePastOffset
		&& futureOffset == this->timeframeFutureOffset)
	{
		return false;
	}
	this->timeframePastOffset = pastOffset;
	this->timeframeFutureOffset = futureOffset;

	if
		(nullptr != this->timeframeCallback)
	{
		this->timeframeCallback(pastOffset, futureOffset);
	}
	return true;
};

/**
 *
 */
bool TimelineModel::setTimestep(timestamp_t timestepIn)
{
	LOG(LogLevel::Debug, "Timeline model", "set timestep to {}", timestepIn);
	if
		(timestepIn == this->timestep)
	{
		return false;
	}
	this->timestep = timestepIn;

	if
		(nullptr != this->timestepCallback)
	{
		this->timestepCallback(timestepIn);
	}
	return true;
};

/**
 *
 */
bool TimelineModel::setCurrentCallback(const std::function<void(timestamp_t)>& callback)
{
	this->currentCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::setBoundsCallback(const std::function<void(timestamp_t, timestamp_t)>& callback)
{
	this->boundsCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::setTimeframeCallback(const std::function<void(timestamp_t, timestamp_t)>& callback)
{
	this->timeframeCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::setTimestepCallback(const std::function<void(timestamp_t)>& callback)
{
	this->timestepCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::setCoverageCallback(const std::function<void()>& callback)
{
	this->coverageCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::goToPrev()
{
	timestamp_t modified;	//
	bool ok;	//

    if
        (Timestep::ONE_THIRD_TIMESPAN_VALUE == this->timestep)
    {
        return this->shiftByTimeframeLength(-1.f, 1.f/3.f);
    }

	if
		(0 == this->timestep)
	{
		modified = this->prev;
	}
	else
	{
		modified = this->current - this->timestep;
	}

	modified = std::max(this->lowerBound, modified);
	ok = this->setCurrent(modified);

	LOG( LogLevel::Debug, "Timeline model"
			, "go to prev: {} {} {} => {}"
			, this->current, this->prev, this->lowerBound, modified);

	return ok;
};

/**
 *
 */
bool TimelineModel::goToNext()
{
	timestamp_t modified;	//
	bool ok;	//

    if
        (Timestep::ONE_THIRD_TIMESPAN_VALUE == this->timestep)
    {
        return this->shiftByTimeframeLength(1.f, 1.f/3.f);
    }

	if
		(0 == this->timestep)
	{
		modified = this->next;
	}
	else
	{
		modified = this->current + this->timestep;
	}
	modified = std::min(this->upperBound, modified);
	ok = this->setCurrent(modified);

	LOG( LogLevel::Debug, "Timeline model"
			, "go to next: {} {} {} => {}"
			, this->current, this->next, this->upperBound, modified);
	return ok;
};

/**
 *
 */
bool TimelineModel::setPrevNext(timestamp_t prevIn, timestamp_t nextIn)
{
	this->prev = prevIn;
	this->next = nextIn;
	return true;
};

/**
 *
 */
bool TimelineModel::setNearestMode(bool value)
{
	if
		(value == this->nearestMode)
	{
        return false;
    }

    this->nearestMode = value;

    if
        (true == value)
    {
        this->applyNearestMode(this->current, false, false, false);
    }

    if
        (nullptr != this->currentCallback)
    {
        this->currentCallback(this->current);
    }
	return true;
};


/**
 *
 */
bool TimelineModel::isDataAvailable(std::int16_t year) const
{
	bool ok;	//
	EventsCoverage currentCoverage;	//
	std::vector<std::uint16_t> years;	//

	ok = this->getCoverage("union", currentCoverage);
	if
		(false == ok)
	{
		return false;
	}

	ok = currentCoverage.getYears(years);
	if
		(false == ok)
	{
		LOG(LogLevel::Debug, "Timeline model", "Could not get years from coverage");
		return false;
	}
	auto yearFound = std::find( years.begin()
							  , years.end()
							  , year);
	return years.end() != yearFound;
}

/**
 *
 */
bool TimelineModel::isDataAvailable(std::int16_t year, std::uint8_t month) const
{
	bool ok;	//
	EventsCoverage currentCoverage;	//
	std::vector<std::uint8_t> months;	//

	ok = this->getCoverage("union", currentCoverage);
	if
		(false == ok)
	{
		return false;
	}

	ok = currentCoverage.getMonths(static_cast<std::uint16_t>(year), months);
	if
		(false == ok)
	{
		LOG(LogLevel::Debug, "Timeline model", "Could not get months from coverage");
		return false;
	}
	auto monthFound = std::find( months.begin()
							   , months.end()
							   , month);
	return months.end() != monthFound;
};

/**
 *
 */
bool TimelineModel::isDataAvailable(std::int16_t year, std::uint8_t month, std::uint8_t day) const
{
	bool ok;	//
	EventsCoverage currentCoverage;	//
	std::vector<std::uint8_t> days;	//

	ok = this->getCoverage("union", currentCoverage);
	if
		(false == ok)
	{
		return false;
	}

	ok = currentCoverage.getDays(static_cast<std::uint16_t>(year), month, days);
	if
		(false == ok)
	{
		LOG(LogLevel::Debug, "Timeline model", "Could not get days from coverage");
		return false;
	}
	auto dayFound = std::find( days.begin()
							 , days.end()
							 , day);
	return days.end() != dayFound;
};

static bool dateSatisfiesContraints(std::uint16_t refYear, std::uint8_t refMonth, std::uint8_t refDay,
                                    std::uint16_t year, std::uint8_t month, std::uint8_t day,
                                    bool preserveYear, bool preserveMonth, bool preserveDay)
{
    return (  (  (false == preserveYear)
              || (refYear == year)
              )
           && (  (false == preserveMonth)
              || ((refYear == year) && (refMonth == month))
              )
           && (  (false == preserveDay)
              || ((refYear == year) && (refMonth == month) && (refDay == day))
              )
          );
}

/**
 *
 */
bool TimelineModel::applyNearestMode(timestamp_t& timestamp, bool preserveYear, bool preserveMonth, bool preserveDay)
{
    if
        (nullptr == this->_nearestDataBoundCallback)
    {
        return false;
    }

    // Extract date components from timestamp
	std::tm* ts {nullptr};	//
	std::time_t t {0};	//
    std::uint16_t refYear {0};  //
    std::uint8_t refMonth {0};  //
    std::uint8_t refDay {0};    //

	t = static_cast<std::time_t>(timestamp);
	ts = std::localtime(&t);

    refYear = static_cast<std::uint16_t>(1900 + ts->tm_year);
    refMonth = static_cast<std::uint8_t>(1 + ts->tm_mon);
    refDay = static_cast<std::uint8_t>(ts->tm_mday);

	LOG(LogLevel::Debug, "Timeline model", "applyNearestMode: {} {} {} {} {}-{}-{}", timestamp, preserveYear, preserveMonth, preserveDay, refYear, refMonth, refDay);
    bool ok {false};    //
    bool hasCurrent {false};    //

    // Get aggregated coverage of all collections
    EventsCoverage currentCoverage;	//
    ok = this->getCoverage("union", currentCoverage);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Timeline model", "Could not retrieve data coverage");
        return false;
    }

    std::uint16_t prevYear {0}; //
    std::uint8_t prevMonth {0}; //
    std::uint8_t prevDay {0};   //
    bool hasPrev {false};   //

    std::uint16_t nextYear {0}; //
    std::uint8_t nextMonth {0}; //
    std::uint8_t nextDay {0};   //
    bool hasNext {false};   //

    ok = currentCoverage.findNearestDays(refYear, refMonth, refDay, hasCurrent,
                                         prevYear, prevMonth, prevDay, hasPrev,
                                         nextYear, nextMonth, nextDay, hasNext);

    LOG(LogLevel::Debug, "Timeline model", "Coverage found: Past [{}] {}-{}-{}, Current [{}] {}-{}-{}, Future [{}] {}-{}-{}",
             hasPrev, prevYear, prevMonth, prevDay,
             hasCurrent, refYear, refMonth, refDay,
             hasNext, nextYear, nextMonth, nextDay);

    timestamp_t original {0};
    original = timestamp;   // Copy because nearestDataCallback will modify the variable

    // Define search domain
    timestamp_t searchLowerBound {0};   //
    timestamp_t searchUpperBound {0};   //
    searchLowerBound = this->lowerBound;
    searchUpperBound = this->upperBound;

    bool isPrevValid {false};   //
    isPrevValid = dateSatisfiesContraints(refYear, refMonth, refDay,
                                          prevYear, prevMonth, prevDay,
                                          preserveYear, preserveMonth, preserveDay);
    bool isNextValid {false};   //
    isNextValid = dateSatisfiesContraints(refYear, refMonth, refDay,
                                          nextYear, nextMonth, nextDay,
                                          preserveYear, preserveMonth, preserveDay);

    // Narrow down search domain lower bound
    std::tm datetime {};
    datetime.tm_hour = 0;
    datetime.tm_min = 0;
    datetime.tm_sec = 0;
    if
        (hasPrev && isPrevValid)
    {
        datetime.tm_year = prevYear - 1900;
        datetime.tm_mon = prevMonth - 1;
        datetime.tm_mday = prevDay;
        searchLowerBound = static_cast<timestamp_t>(std::mktime(&datetime));
    }
    else if
        (hasCurrent)
    {
        datetime.tm_year = refYear - 1900;
        datetime.tm_mon = refMonth - 1;
        datetime.tm_mday = refDay;
        searchLowerBound = static_cast<timestamp_t>(std::mktime(&datetime));
    }
    else if
        (hasNext && isNextValid)
    {
        datetime.tm_year = nextYear - 1900;
        datetime.tm_mon = nextMonth - 1;
        datetime.tm_mday = nextDay;
        searchLowerBound = static_cast<timestamp_t>(std::mktime(&datetime));
    }
    else if
        (true == preserveDay)
    {
        // January 1st of reference year
        datetime.tm_year = refYear - 1900;
        datetime.tm_mon = 0;
        datetime.tm_mday = 1;
        searchLowerBound = static_cast<timestamp_t>(std::mktime(&datetime));

    }
    else if
        (true == preserveMonth)
    {
        // 1st day of reference month
        datetime.tm_year = refYear - 1900;
        datetime.tm_mon = refMonth - 1;
        datetime.tm_mday = 1;
        searchLowerBound = static_cast<timestamp_t>(std::mktime(&datetime));

    }
    else if
        (true == preserveDay)
    {
        // Reference day
        datetime.tm_year = refYear - 1900;
        datetime.tm_mon = refMonth - 1;
        datetime.tm_mday = refDay;
        searchLowerBound = static_cast<timestamp_t>(std::mktime(&datetime));
    }
    LOG(LogLevel::Debug, "Timeline model", "Search lower bound [{}-{}-{}T00:00:00Z, ...[", datetime.tm_year + 1900, datetime.tm_mon + 1, datetime.tm_mday);

    // Narrow down search domain upper bound
    // Day at 23:59:59 is the last valid time but the upper bound of the search domain is excluded so add 1 second to the timestamp
    datetime.tm_hour = 23;
    datetime.tm_min = 59;
    datetime.tm_sec = 59;
    if
        (hasNext && isNextValid)
    {
        datetime.tm_year = nextYear - 1900;
        datetime.tm_mon = nextMonth - 1;
        datetime.tm_mday = nextDay;
        searchUpperBound = static_cast<timestamp_t>(std::mktime(&datetime)) + 1;
    }
    else if
        (hasCurrent)
    {
        datetime.tm_year = refYear - 1900;
        datetime.tm_mon = refMonth - 1;
        datetime.tm_mday = refDay;
        searchUpperBound = static_cast<timestamp_t>(std::mktime(&datetime)) + 1;
    }
    else if
        (hasPrev && isPrevValid)
    {
        datetime.tm_year = prevYear - 1900;
        datetime.tm_mon = prevMonth - 1;
        datetime.tm_mday = prevDay;
        searchUpperBound = static_cast<timestamp_t>(std::mktime(&datetime)) + 1;
    }
    else if
        (true == preserveYear)
    {
        // 31st December of reference year
        datetime.tm_year = refYear - 1900;
        datetime.tm_mon = 11;
        datetime.tm_mday = 31;
        searchLowerBound = static_cast<timestamp_t>(std::mktime(&datetime)) + 1;
    }
    else if
        (true == preserveMonth)
    {
        // Last day of reference month
        datetime.tm_year = refYear - 1900;
        datetime.tm_mon = refMonth - 1;
        datetime.tm_mday = getMonthDaysCount(refYear, refMonth);
        searchLowerBound = static_cast<timestamp_t>(std::mktime(&datetime)) + 1;
    }
    else if
        (true == preserveDay)
    {
        // Reference day
        datetime.tm_year = refYear - 1900;
        datetime.tm_mon = refMonth - 1;
        datetime.tm_mday = refDay;
        searchLowerBound = static_cast<timestamp_t>(std::mktime(&datetime)) + 1;
    }
    LOG(LogLevel::Debug, "Timeline model", "Search upper bound [..., {}-{}-{}T23:59:59Z + 1s[", datetime.tm_year + 1900, datetime.tm_mon + 1, datetime.tm_mday);



    this->_nearestDataBoundCallback(timestamp, searchLowerBound, searchUpperBound);

    t = static_cast<std::time_t>(timestamp);
    ts = std::localtime(&t);

    std::uint16_t newYear {0};  //
    std::uint8_t newMonth {0};  //
    std::uint8_t newDay {0};    //
    newYear = static_cast<std::uint16_t>(1900 + ts->tm_year);
    newMonth = static_cast<std::uint8_t>(1 + ts->tm_mon);
    newDay = static_cast<std::uint8_t>(ts->tm_mday);

    LOG(LogLevel::Debug, "Timeline model", "New time: {} {}-{}-{}", timestamp, newYear, newMonth, newDay);
    bool invalidDay {false};    //
    invalidDay = ((true == preserveDay) && ((refYear != newYear) || (refMonth != newMonth) || (refDay != newDay)));

    bool invalidMonth {false};  //
    invalidMonth = ((true == preserveMonth) && ((refYear != newYear) || (refMonth != newMonth)));

    bool invalidYear {false};   //
    invalidYear = ((true == preserveYear) && (refYear != newYear));

    LOG(LogLevel::Debug, "Timeline model", "{}: {} <> {}, {}: {} <> {}, {}: {} <> {}", preserveYear, refYear, newYear, preserveMonth, refMonth, newMonth, preserveDay, refDay, newDay);
    if
        (invalidDay || invalidMonth || invalidYear)
    {
        // Restore timestamp to its original value because the nearest granule
        // is outside the search boundaries
        timestamp = original;

        LOG(LogLevel::Debug, "Timeline model", "Timestamp restored to {}", timestamp);
        return false;
    }

    return true;
};

/**
 *
 */
bool TimelineModel::shiftByYears(std::int16_t shift)
{
	std::tm* oldTime;	//
	std::time_t oldT;	//
    std::time_t t;  //
    timestamp_t newTimestamp;   //

    oldT = static_cast<std::time_t>(this->current);
    oldTime = std::localtime(&oldT);

    oldTime->tm_year = oldTime->tm_year + shift;
    t = std::mktime(oldTime);
    newTimestamp = static_cast<timestamp_t>(t);
    return this->setCurrent(newTimestamp);
};

/**
 *
 */
bool TimelineModel::shiftByMonths(std::int32_t shift)
{
	std::tm* oldTime;	//
	std::time_t oldT;	//
    std::time_t t;  //
    timestamp_t newTimestamp;   //
    std::int32_t shiftedMonth; //
    std::int16_t yearsShift;    //

    oldT = static_cast<std::time_t>(this->current);
    oldTime = std::localtime(&oldT);

    shiftedMonth = shift + oldTime->tm_mon - 1;
    yearsShift = static_cast<std::int16_t>(std::floor(static_cast<float>(shiftedMonth) / 12.f));
    shiftedMonth = 1 + (shiftedMonth - 12 * yearsShift) % 12;

    oldTime->tm_year = oldTime->tm_year + yearsShift;
    oldTime->tm_mon = static_cast<std::uint8_t>(shiftedMonth);

    t = std::mktime(oldTime);
    newTimestamp = static_cast<timestamp_t>(t);
    return this->setCurrent(newTimestamp);
};

/**
 *
 */
bool TimelineModel::shiftByDays(std::int32_t shift)
{
    return this->setCurrent(this->current + static_cast<std::uint64_t>(shift * 86400));
};

/**
 *
 */
bool TimelineModel::shiftByTimeframeLength(float shift, float portion = 1.f)
{
    float secondsShift; //
    int64_t seconds;    //
    timestamp_t timeframeLength;    //

    timeframeLength = this->timeframeFutureOffset + this->timeframePastOffset;
    secondsShift = portion * static_cast<float>(timeframeLength);
    seconds = static_cast<int64_t>(std::round(shift * secondsShift));
    LOG(LogLevel::Debug, "TimelineModel", "SHIFT {} {} {} {} {}", this->timeframePastOffset, this->timeframeFutureOffset, timeframeLength, secondsShift, seconds);
    return this->setCurrent(this->current + static_cast<timestamp_t>(seconds));
};

/**
 *
 */
bool TimelineModel::setNearestDataBoundCallback(const std::function<bool(timestamp_t&, const timestamp_t, const timestamp_t)>& callback)
{
    if
        (nullptr != this->_nearestDataBoundCallback)
    {
        return false;
    }

    this->_nearestDataBoundCallback = callback;
    return true;
}
