// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <numeric>
#include <vector>
#include <sstream>
#include <array>
#include <ctime>
#include "core/utils.h"
#include "core/log.h"

#include <iostream>
#include <iomanip>

static std::array<std::uint8_t, 12> daysPerMonth = {{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}};

namespace
{
	// Check if the string contains only digits.
    bool containsOnlyDigit(const std::string& str)
    {
        return !str.empty() && str.find_first_not_of("0123456789") == std::string::npos;
    }
}

/**
 *
 */
int SEAScope::dimStr2pixels(const std::string& dimStr, int ref)
{
	int result;	//
	std::string numStr;	//
	int num;	//
	std::vector<std::string> items;	//
	int ope1;	//
	int ope2;	//

	result = -1;

	if
		(dimStr.empty())
	{
		return result;
	}

	if
		(std::string::npos != dimStr.find(" "))
	{
		/* both relative and absolute => convert to absolute */
		SEAScope::explodeString(dimStr, " ", items);
		LOG(LogLevel::Debug, "Utils", "complex dimension with {} items", items.size());
		if
			(3 > items.size())
		{
			return result;
		}

		ope1 = SEAScope::dimStr2pixels(items[0], ref);
		ope2 = SEAScope::dimStr2pixels(items[2], ref);
		LOG(LogLevel::Debug, "Utils", "Operators: {} {} {}", ope1, items[1], ope2);
		if
			("-" == items[1])
		{
			result = ope1 - ope2;
		}
		else if
			("+" == items[1])
		{
			result = ope1 + ope2;
		}
		LOG(LogLevel::Debug, "Utils", "result: {}", result);
		return result;
	}

	if
		(0 == dimStr.compare(dimStr.size()-1, 1, "%"))
	{
		numStr = dimStr.substr(0, dimStr.size()-1);
		num = std::stoi(numStr);
		result = static_cast<int>(0.01 * num * ref);
	}
	else if
		(0 == dimStr.compare(dimStr.size()-2, 2, "px"))
	{
		numStr = dimStr.substr(0, dimStr.size()-2);
		result = std::stoi(numStr);
	}
	else
	{
		LOG(LogLevel::Error, "Utils::dimStr2pixels", "{} is not supported", dimStr);
	}
	return result;
};

/**
 *
 */
int SEAScope::posStr2pixels(const std::string& posStr, int ref, int dim)
{
	int result;	//
	std::string numStr;	//
	int num;	//
	std::vector<std::string> items;	//
	int ope1;	//
	int ope2;	//

	result = -1;

	if
		(posStr.empty())
	{
		return result;
	}

	if
		(std::string::npos != posStr.find(" "))
	{
		/* both relative and absolute => convert to absolute */
		SEAScope::explodeString(posStr, " ", items);
		if
			(3 > items.size())
		{
			return result;
		}

		ope1 = SEAScope::posStr2pixels(items[0], ref, dim);
		ope2 = SEAScope::posStr2pixels(items[2], ref, dim);
		if
			("-" == items[1])
		{
			return ope1 - ope2;
		}
		else if
			("+" == items[1])
		{
			return ope1 + ope2;
		}
		return result;
	}

	if
		(0 == posStr.compare(posStr.size()-1, 1, "%"))
	{
		numStr = posStr.substr(0, posStr.size() - 1);
		num = std::stoi(numStr);
		if
			(0 == numStr.compare(0, 1, "-")) /* Use the string value to support the "-0%" case */
		{
			// From the opposite side
			result = static_cast<int>(ref * (1 + 0.01 * num /*negative*/) - dim);
		}
		else
		{
			result = static_cast<int>(0.01 * num * ref);
		}
	}
	else if
		(0 == posStr.compare(posStr.size()-2, 2, "px"))
	{
		numStr = posStr.substr(0, posStr.size() - 2);
		num = std::stoi(numStr);
		if
			(0 == numStr.compare(0, 1, "-")) /* Use the string value to support the "-0px" case */
		{
			// From the opposite side
			result = ref + num /*negative*/ - dim;
		}
		else
		{
			result = num;
		}
	}
	else
	{
		LOG(LogLevel::Error, "Utils::posStr2pixels", "{} is not supported", posStr);
	}
	return result;
};

/**
 *
 */
std::string& SEAScope::implodeStringSet(const std::set<std::string>& elems, const std::string& delim, std::string& result)
{
	std::set<std::string>::const_iterator it;	//
	std::size_t bytesCount;	//

    if
        (0 >= elems.size())
    {
        // if the set of strings has no element, return an empty string
        result = "";
        return result;
    }

	// Allocate memory for the result
	bytesCount = static_cast<std::size_t>(std::accumulate( elems.begin()
                                                         , elems.end()
                                                         , 0
                                                         , [] (int sum, const std::string& elem) {
                                                             return sum + static_cast<int>(elem.size());
                                                         }));

    if
        (0 >= bytesCount)
    {
        // if the cumulative size of the strings in the set is null (can only
        // happen if the set has one element which is an empty string), return
        // an empty string
        result = "";
        return result;
    }

	bytesCount += (elems.size() - 1) * delim.length();
	result.reserve(bytesCount);

	it = elems.begin();
	for
		(; it != elems.end(); ++it)
	{
		result += *it;
		if
			(std::next(it, 1) != elems.end())
		{
			result += delim;
		}
	}

	return result;
};

/**
 *
 */
void SEAScope::explodeString(const std::string& input, const std::string& delim, std::vector<std::string>& result)
{
	size_t start = 0;
	auto end = input.find(delim);

	while
		(end != std::string::npos)
	{
		result.push_back(input.substr(start, end-start));
		start = end + delim.length();
		end = input.find(delim, start);
	}
	result.push_back(input.substr(start, end));
};

/**
 *
 */
void SEAScope::getColvarLabel(const std::string& collectionLabel, const std::string& variableLabel, std::string& result)
{
    std::ostringstream buffer("");
    buffer << collectionLabel << ": " << variableLabel;
    result = buffer.str();
}

/**
 *
 */
bool SEAScope::parseDatetime(const std::string& input,
                             const std::string& separator,
                             std::uint16_t& year, std::uint8_t& month, std::uint8_t& day,
                             std::uint8_t& hour, std::uint8_t& minute, std::uint8_t& second)
{
    if
        (0 == input.size())
    {
        return false;
    }

    size_t pos(std::string::npos);
    pos = input.find(separator);
	if
		(std::string::npos == pos)
	{
		return false;
	}
	std::string date = input.substr(0, pos);
	bool okDate = SEAScope::parseDate(date, year, month, day);

	std::string time = input.substr(pos + 1, input.size() - (pos + 1));
	bool okTime = SEAScope::parseTime(time, hour, minute, second);

	return okDate && okTime;
}

/**
 *
 */
bool SEAScope::parseDate(const std::string& input, std::uint16_t& year,
                         std::uint8_t& month, std::uint8_t& day)
{
	std::string tmp;	//
	std::size_t pos;	//

	if
		(10 > input.size())
	{
		return false;
	}

	/* Check that month/day exists */
	pos = input.find("-");
	tmp = input.substr(0,pos);

    if
        (true == containsOnlyDigit(tmp))
    {
		year = static_cast<std::uint16_t>(std::stoi(tmp));
    }
    else
    {
        return false;
    }

	tmp = input.substr(pos+1,2);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}

	if
		(true == containsOnlyDigit(tmp))
	{
        month = static_cast<std::uint8_t>(std::stoi(tmp));
	}
	else
	{
		return false;
	}

	if
		(0 >= month)
	{
		return false;
	}

	tmp = input.substr(pos+4, 2);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}

    if
        (true == containsOnlyDigit(tmp))
    {
		day = static_cast<std::uint8_t>(std::stoi(tmp));
    }
    else
    {
        return false;
    }

	if
		(0 >= day)
	{
		return false;
	}

	if
		(day <= daysPerMonth[month - 1u] || (2 == month && 29 == day && 0 == year%4))
	{
        // everything is in order
		return true;
	}

	return false;
}

/**
 *
 */
bool SEAScope::parseTime(const std::string& input, std::uint8_t& hour, std::uint8_t& minute, std::uint8_t& second)
{
	size_t pos;	//
	std::string tmp;	//

    if
        (8 > input.size())
    {
        return false;
    }

	pos = input.find(":");
	tmp = input.substr(0,pos);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}
	hour = static_cast<std::uint8_t>(std::stoi(tmp));
	if
		(24 < hour)
	{
		return false;
	}

	tmp = input.substr(pos+1, 2);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}
	minute = static_cast<std::uint8_t>(std::stoi(tmp));
	if
		(59 < minute)
	{
		return false;
	}

	tmp = input.substr(pos+4, 2);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}
	second = static_cast<std::uint8_t>(std::stoi(tmp));
	if
		(59 < second)
	{
		return false;
	}
	return true;
}

/**
 *
 */
bool SEAScope::getTimestampFromDatetime(const std::string& input, timestamp_t& result)
{
    std::uint16_t year(0);
    std::uint8_t month(255);
    std::uint8_t day(255);
    std::uint8_t hour(255);
    std::uint8_t minute(255);
    std::uint8_t second(255);
    bool ok(false);

    std::string separator("T");

    ok = SEAScope::parseDatetime(input, separator, year, month, day, hour, minute, second);
    if
        (false == ok)
    {
        if
            (19 <= input.size())
        {
            // It was clearly meant to be a full datetime but could not be
            // parsed as such
            return false;
        }

        // Set default time and read date from input
        hour = 12;
        minute = 0;
        second = 0;
        ok = SEAScope::parseDate(input, year, month, day);
        if
            (false == ok)
        {
            // Set default date (current date) and read time from input
	        time_t timestamp = std::time(nullptr);
            std::tm* ltime = std::localtime(&timestamp);
            if
                (nullptr == ltime)
            {
                return false;
            }

	        year = static_cast<std::uint16_t>(ltime->tm_year) + 1900u;
	        month = static_cast<std::uint8_t>(ltime->tm_mon) + 1u;
	        day = static_cast<std::uint8_t>(ltime->tm_mday);

            ltime = nullptr;

            ok = SEAScope::parseTime(input, hour, minute, second);
            if
                (false == ok)
            {
                return false;
            }
        }
    }

	std::tm datetime = {};	//
	datetime.tm_year = year - 1900;
	datetime.tm_mon = month - 1;
	datetime.tm_mday = day;
	datetime.tm_hour = hour;
	datetime.tm_min = minute;
	datetime.tm_sec = second;
	result = static_cast<timestamp_t>(std::mktime(&datetime));
    //std::cout << "Timestamp: " << int(result) << std::endl;
	//std::cout << "initialDatetime: " << static_cast<int>(year) << "-" << std::setfill('0') << std::setw(2) << static_cast<int>(month) << "-" << std::setfill('0') << std::setw(2) << static_cast<int>(day) << separator << std::setfill('0') << std::setw(2) << static_cast<int>(hour) << ":" << std::setfill('0') << std::setw(2) << static_cast<int>(minute) << ":" << std::setfill('0') << std::setw(2) << static_cast<int>(second) << std::endl;
    return true;
}

/**
 * Case-insensitive std::string equality check
 */
bool SEAScope::iequals(const std::string& str1, const std::string& str2)
{
    return str1.size() == str2.size()
        && std::equal(str1.cbegin(), str1.cend(), str2.cbegin(),
                      [](std::string::value_type a, std::string::value_type b)
                      { return std::toupper(a) == std::toupper(b); });
}

/**
 *
 */
std::string& SEAScope::trim(std::string& value)
{
    constexpr char trimmable[]{" \t\n\r\f\v"};
    value.erase(value.find_last_not_of(trimmable) + 1);
    value.erase(0, value.find_first_not_of(trimmable));
    return value;
}

/**
 *
 */
bool SEAScope::stringEndsWith(const std::string& haystack, const std::string& needle)
{
    if
        (haystack.length() < needle.length())
    {
        return false;
    }

    if
        (0 == haystack.compare(haystack.length() - needle.length(), needle.length(), needle))
    {
        return true;
    }

    return false;
}

/**
 *
 */
std::uint8_t SEAScope::getMonthDaysCount(std::uint16_t year, std::uint8_t month)
{
    if
        (  (1 > month)
        || (12 < month))
    {
        return 0;
    }

    if
        ((0 == year % 4) && 2 == month)
    {
        return 29;
    }
    return daysPerMonth[month - 1];
}
