// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cmath>
#include "data/annotation.h"

#include "core/log.h"
#include "num_utils.h"


using namespace SEAScope;

#ifdef _WIN32
    // Disable : warning C4127: conditional expression is constant
    #pragma warning(disable : 4127)
#endif

Annotation::Annotation()
: _shape(nullptr)
{
    //LOG(LogLevel::Error, "Annotation", "Default constructor");
    if
        (true == std::numeric_limits<double>::has_quiet_NaN) // Warning C4127 here
    {
        this->_start = std::numeric_limits<double>::quiet_NaN();
        this->_stop = std::numeric_limits<double>::quiet_NaN();
    }
    else
    {
        this->_start = std::numeric_limits<double>::max();
        this->_stop = std::numeric_limits<double>::max();
    }
}

#ifdef _WIN32
    #pragma warning(default : 4127)
#endif

Annotation::Annotation(Annotation&& src)
: _start(std::move(src._start)), _stop(std::move(src._stop)), _shape(std::move(src._shape)), _properties(std::move(src._properties))
{
    //LOG(LogLevel::Error, "Annotation", "Move constructor (my shape is null: {}. src shape was null: {}",
    //         (nullptr == this->shape), (nullptr == src.shape));

    src._shape = nullptr;
}

Annotation::~Annotation()
{
    this->_shape = nullptr;  // not owning (ShapesAllocator will handle deletion)
}


bool Annotation::hasStart() const
{
    if
        (std::numeric_limits<double>::has_quiet_NaN)
    {
        return false == std::isnan(this->_start);
    }
    return false == floating_point_equality(std::numeric_limits<double>::max(), this->_start);
}

bool Annotation::hasEnd() const
{
    if
        (std::numeric_limits<double>::has_quiet_NaN)
    {
        return false == std::isnan(this->_stop);
    }
    return false == floating_point_equality(std::numeric_limits<double>::max(), this->_stop);
}

bool Annotation::setShape(OGL::Shape* value)
{
    this->_shape = value;
    return true;
}

OGL::Shape* Annotation::getShape() const
{
    return this->_shape;
}

const DictNode& Annotation::getConstProperties() const
{
    return this->_properties;
}

DictNode& Annotation::getProperties()
{
    return this->_properties;
}

void Annotation::assignProperties(DictNode&& properties)
{
    this->_properties = std::move(properties);
}

double Annotation::getStart() const
{
    return this->_start;
}

void Annotation::setStart(double start)
{
    this->_start = start;
}

double Annotation::getStop() const
{
    return this->_stop;
}

void Annotation::setStop(double stop)
{
    this->_stop = stop;
}
