// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <vector>
#include <bitset>
#include <errno.h>

#include "core/log.h"
#include "core/path.h"
#include "data/data_handler.h"
#include "data/image.h"
#include "3d/data_set.h"
#include "num_utils.h"

using namespace SEAScope;

/**
 *
 * Longitudes are expressed in [-180.0, 180.0]
 */
glm::vec4 DataHandler::computeBbox(float * lons, float * lats,
                                   std::uint32_t* gcpx,
                                   std::uint32_t* gcpy,
                                   unsigned int lon_size,
                                   unsigned int lat_size,
                                   bool& lonsGrowing,
                                   bool& latsGrowing,
								   bool& sameVariationForLonAndX,
								   bool& sameVariationForLatAndY)
{
    std::size_t cur;
    std::size_t next;
    std::size_t quarterCol = 0;
	std::size_t quarterRow = 0;
	float lonMin = 999999.f;
	float lonMax = -999999.f;
	float latMin = 999999.f;
	float latMax = -999999.f;
    float deltaLat = 0.f;
    float latTrend = 0.f;
    float deltaX = 0.f;
    float xTrend = 0.f;
    float deltaY = 0.f;
    float yTrend = 0.f;
    float deltaLon = 0.f;
    float lonTrend = 0.f;
	float lonTmp;
	float latTmp;

	quarterCol = static_cast<std::size_t>(std::floor(lat_size / 4));
    if
		(quarterCol <= 0)
	{
		quarterCol = lat_size;
	}

	quarterRow = static_cast<std::size_t>(std::floor(lon_size / 4));
	if
		(quarterRow <= 0)
	{
		quarterRow = lon_size;
	}

    LOG(LogLevel::Debug, "DataHandler", "lon_size {} / lat_size {}", lon_size, lat_size);

	/* First column */
	for (std::size_t y = 0; y < lat_size; ++y)
	{
        cur = y * lon_size;
        next = (y + 1) * lon_size;

		lonTmp = lons[cur];
		latTmp = lats[cur];

		if (y < lat_size - 1 && y < quarterCol)
		{
			deltaLat = lats[next] - latTmp;
			latTrend = latTrend + static_cast<float>(std::copysign(1.0, deltaLat));

            deltaY = static_cast<float>(gcpy[next] - gcpy[cur]);
            yTrend = yTrend + static_cast<float>(std::copysign(1.0, deltaY));
		}

		lonMin = std::min(lonMin, lonTmp);
		lonMax = std::max(lonMax, lonTmp);
		latMin = std::min(latMin, latTmp);
		latMax = std::max(latMax, latTmp);
        LOG(LogLevel::Debug, "DataHandler", "[{} {}] {} {} : {} {} {} {}", cur, next, lonTmp, latTmp, lonMin, latMin, lonMax, latMax);
	}
	LOG(LogLevel::Debug, "DataHandler", "First column lat trend: {}", latTrend);

	/* Last column */
	for (std::size_t y = 0; y < lat_size; ++y)
	{
        cur = (y * lon_size) + lon_size - 1;
        next = ((y + 1) * lon_size) + lon_size - 1;

		lonTmp = lons[cur];
		latTmp = lats[cur];

		if (y < lat_size - 1 && y < quarterCol)
		{
			deltaLat = lats[next] - latTmp;
			latTrend = latTrend + static_cast<float>(std::copysign(1.0, deltaLat));

            deltaY = static_cast<float>(gcpy[next] - gcpy[cur]);
            yTrend = yTrend + static_cast<float>(std::copysign(1.0, deltaY));
		}

		lonMin = std::min(lonMin, lonTmp);
		lonMax = std::max(lonMax, lonTmp);
		latMin = std::min(latMin, latTmp);
		latMax = std::max(latMax, latTmp);
        LOG(LogLevel::Debug, "DataHandler", "[{} {}] {} {} : {} {} {} {}", cur, next, lonTmp, latTmp, lonMin, latMin, lonMax, latMax);
	}
	LOG(LogLevel::Debug, "DataHandler", "Last column lat trend: {}", latTrend);

	/* First row */
	for (std::size_t x = 0; x < lon_size; ++x)
	{
        cur = x;
        next = x + 1;

		lonTmp = lons[cur];
		latTmp = lats[cur];

		if (x < lon_size - 1 && x < quarterRow)
		{
			deltaLon = lons[next] - lonTmp;
			lonTrend = lonTrend + static_cast<float>(std::copysign(1.0, deltaLon));

            deltaX = static_cast<float>(gcpx[next] - gcpx[cur]);
            xTrend = xTrend + static_cast<float>(std::copysign(1.0, deltaX));
		}

		lonMin = std::min(lonMin, lonTmp);
		lonMax = std::max(lonMax, lonTmp);
		latMin = std::min(latMin, latTmp);
		latMax = std::max(latMax, latTmp);
        LOG(LogLevel::Debug, "DataHandler", "[{} {}] {} {} : {} {} {} {}", cur, lon_size * lat_size, lonTmp, latTmp, lonMin, latMin, lonMax, latMax);
	}
	LOG(LogLevel::Debug, "DataHandler", "First row lon trend: {}", lonTrend);

	/* Last row */
	for (std::size_t x = 0; x < lon_size; ++x)
	{
        cur = (lat_size - 1) * lon_size + x;
        next = (lat_size - 1) * lon_size + x + 1;
		lonTmp = lons[cur];
		latTmp = lats[cur];

		if (x < lon_size - 1 && x < quarterRow)
		{
			deltaLon = lons[next] - lonTmp;
			lonTrend = lonTrend + static_cast<float>(std::copysign(1.0, deltaLon));

            deltaX = static_cast<float>(gcpx[next] - gcpx[cur]);
            xTrend = xTrend + static_cast<float>(std::copysign(1.0, deltaX));
		}

		lonMin = std::min(lonMin, lonTmp);
		lonMax = std::max(lonMax, lonTmp);
		latMin = std::min(latMin, latTmp);
		latMax = std::max(latMax, latTmp);
        LOG(LogLevel::Debug, "DataHandler", "[{} {}] {} {} : {} {} {} {}", cur, lon_size * lat_size, lonTmp, latTmp, lonMin, latMin, lonMax, latMax);
	}
	LOG(LogLevel::Debug, "DataHandler", "Last row lon trend: {}", lonTrend);

    lonsGrowing = (0.f <= lonTrend);
    latsGrowing = (0.f <= latTrend);

	/* Detect lat/y variation */
    sameVariationForLatAndY = (0.f <= latTrend * yTrend);

	/* Detect lon/x variation */
    sameVariationForLonAndX = (0.f <= lonTrend * xTrend);

    //glm::vec4 bbox = glm::vec4(lonMin, latMax, lonMax, latMin);
    glm::vec4 bbox = glm::vec4(lonMin, latMax, lonMax, latMin);

    LOG(LogLevel::Debug, "Data handler", "Bbox: {} {}, {} {} [Y-inverted: {}, X-inverted: {}]",
             bbox.w, bbox.x, bbox.y, bbox.z,
			 !sameVariationForLatAndY, !sameVariationForLonAndX);

    return bbox;
};

/**
 *
 */
void DataHandler::exportToBinary(const std::string& fileName, DataSet * data_set)
{
    LOG( LogLevel::Debug, "DataHandler"
            , "Preparing export.");

    std::string fileNameExport = fileName.substr(0, fileName.size() - 3);
	Path::createDir(fileNameExport, 0777);

    fileNameExport += "/data";

    LOG( LogLevel::Debug, "DataHandler"
            , "Create file.");

    FILE* fp = fopen(fileNameExport.c_str(), "wb");

    if (nullptr == fp)
    {
        LOG( LogLevel::Error, "DataHandler"
                , "Error, impossible to open {}. {}"
                , fileNameExport, strerror(errno));
        return;
    }

    LOG( LogLevel::Debug, "DataHandler"
            , "Dump data.");

    const unsigned char major = 1;
    const unsigned char minor = 1;

    fwrite(&major, 1, 1, fp);
    fwrite(&minor, 1, 1, fp);

    fwrite(&data_set->getDimensionCount(), sizeof(unsigned int), 1, fp);
    fwrite(data_set->getOffset(), sizeof(float), data_set->getDimensionCount(), fp);
    fwrite(data_set->getScaleFactor(), sizeof(float), data_set->getDimensionCount(), fp);

    fwrite(&data_set->getLon_size(), sizeof(unsigned int), 1, fp);
    fwrite(&data_set->getLat_size(), sizeof(unsigned int), 1, fp);

    if (data_set->getGcpx() && data_set->getGcpy())
    {
        unsigned char val = 1;
        fwrite(&val, sizeof(unsigned char), 1, fp);

        unsigned int gcpDataSize = data_set->getLon_size() * data_set->getLat_size();
        fwrite(data_set->getLons(), sizeof(float), gcpDataSize, fp);
        fwrite(data_set->getLats(), sizeof(float), gcpDataSize, fp);

        fwrite(data_set->getGcpx(), sizeof(std::uint32_t), gcpDataSize, fp);
        fwrite(data_set->getGcpy(), sizeof(std::uint32_t), gcpDataSize, fp);
    }
    else
    {
        unsigned char val = 0;
        fwrite(&val, sizeof(unsigned char), 1, fp);

        fwrite(data_set->getLons(), sizeof(float), data_set->getLon_size(), fp);
        fwrite(data_set->getLats(), sizeof(float), data_set->getLat_size(), fp);
    }

    fwrite(&data_set->getWidth(), sizeof(unsigned int), 1, fp);
    fwrite(&data_set->getHeight(), sizeof(unsigned int), 1, fp);

    unsigned int val = static_cast<unsigned int>(data_set->getData()->dataType());
    fwrite(&val, sizeof(unsigned int), 1, fp);

    fwrite(data_set->getData()->data(), sizeof(unsigned char), data_set->getData()->dataSize(), fp);
    fwrite(data_set->getDataMask()->data(), sizeof(unsigned char), data_set->getDataMask()->dataSize(), fp);

    fclose(fp);

    LOG( LogLevel::Debug, "DataHandler"
            , "Done.");
};
