// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "data/events_coverage.h"

#include <iostream>

using namespace SEAScope;

/**
 *
 */
void EventsCoverage::reset()
{
    this->store.clear();
}

/**
 *
 */
bool EventsCoverage::clone(const EventsCoverage& other)
{
    this->reset();

	auto itYears = other.store.begin();
	for
		(; itYears != other.store.end(); itYears++)
	{
		this->store.insert(std::make_pair(itYears->first, itYears->second));
	}
	return true;
};

/**
 *
 */
bool EventsCoverage::merge(const EventsCoverage& other)
{
	size_t i;	//

	auto itYears = other.store.begin();
	for
		(; itYears != other.store.end(); itYears++)
	{
		auto found = this->store.find(itYears->first);
		if
			(this->store.end() == found)
		{
            LOG(LogLevel::Debug, "TimelineModel", "\t\t\tMerge inserted year {}", itYears->first);
			this->store.insert(std::make_pair(itYears->first, itYears->second));
		}
		else
		{
            LOG(LogLevel::Debug, "TimelineModel", "\t\t\tMerge months for year {}", itYears->first);
			i = 0;
			for
				(; i<12; ++i)
			{
				this->store[itYears->first][i] |= itYears->second[i];
			}
		}
	}
	return false;
};

/**
 *
 */
bool EventsCoverage::intersect(const EventsCoverage& other)
{
	coverage_t result;	//
	uint32_t myMonth;	//
	uint32_t otherMonth;	//
	size_t i;	//
	bool atLeastOneMonth;	//

	auto itYears = other.store.begin();
	for
		(; itYears != other.store.end(); itYears++)
	{
		auto yearFound = this->store.find(itYears->first);
		if
			(this->store.end() == yearFound)
		{
			/* No coverage in this year */
			continue;
		}

		std::array<uint32_t, 12> yearData;
		atLeastOneMonth = false;
		i = 0;
		for
			(; i<12; ++i)
		{
			myMonth = yearFound->second[i];
			otherMonth = itYears->second[i];
			yearData[i] = myMonth & otherMonth;
			atLeastOneMonth = atLeastOneMonth || (0 < yearData[i]);
		}

		if
			(true == atLeastOneMonth)
		{
			result.insert(std::make_pair(itYears->first, yearData));
		}
	}

	this->store = result;
	return false;
};

/**
 *
 */
bool EventsCoverage::getDays(uint16_t year, uint8_t month, std::vector<uint8_t>& values) const
{
	uint8_t i;	//
	auto found = this->store.find(year);
	if
		(this->store.end() == found)
	{
		return false;
	}

	i = 0;
	for
		(; i < 31; ++i)
	{
		if
			(1 == ((found->second[month-1u] >> i) & 1) )
		{
			values.push_back(i+1u);
		}
	}
	return true;
};

/**
 *
 */
bool EventsCoverage::getMonths(uint16_t year, std::vector<uint8_t>& values) const
{
	uint8_t i;	//
	auto found = this->store.find(year);
	if
		(this->store.end() == found)
	{
		return false;
	}

	i = 0;
	for
		(; i<12; ++i)
	{
		if
			(0 < found->second[i])
		{
			values.push_back(i+1u);
		}
	}
	return true;
};

/**
 *
 */
bool EventsCoverage::getYears(std::vector<uint16_t>& values) const
{
	auto it = this->store.begin();
	for
		(; it != this->store.end(); it++)
	{
		values.push_back(it->first);
	}
	return 0 < values.size();
};

static void getFirstCoveredDay(const std::array<std::uint32_t, 12>& yearCoverage, std::uint8_t& month, std::uint8_t& day)
{
    for
        (std::size_t monthIndex = 0; monthIndex < 12; ++monthIndex)
    {
        if
            (0 < yearCoverage[monthIndex])
        {
            month = static_cast<std::uint8_t>(monthIndex + 1);
            for
                (day = 1; day < 32; ++ day)
            {
                if
                    ((1 << (day - 1)) & yearCoverage[monthIndex])
                {
                    return;
                }
            }
            break;
        }
    }
}

static void getLastCoveredDay(const std::array<std::uint32_t, 12>& yearCoverage, std::uint8_t& month, std::uint8_t& day)
{
    std::size_t monthIndex {11};    //
    for
        (;;)
    {
        // No data this month? check previous month or stop if the
        // beginning of the year has been reached
        if
            (0 == yearCoverage[monthIndex])
        {
            if
                (monthIndex > 0)
            {
                --monthIndex;
                continue;
            }
            break;
        }

        for
            (day = 31;; --day)
        {
            if
                ((1 << (day - 1)) & yearCoverage[monthIndex])
            {
                month = static_cast<std::uint8_t>(monthIndex + 1);
                return;
            }

            // Needed to properly stop the loop as the condition is based
            // on a decreasing unsigned integer (std::uint8_t)
            if
                (1 == day)
            {
                break;
            }
        }

        // Needed to properly stop the loop as the condition is based on
        // a decreasing unsigned integer (std::size_t)
        if
            (0 == monthIndex)
        {
            break;
        }

        --monthIndex;
    }
}


/**
 *
 */
bool EventsCoverage::findNearestDays(const std::uint16_t refYear, const std::uint8_t refMonth, const std::uint8_t refDay, bool& hasCurrent,
                                    std::uint16_t& nearestYearBefore, std::uint8_t& nearestMonthBefore, std::uint8_t& nearestDayBefore, bool& hasBefore,
                                    std::uint16_t& nearestYearAfter, std::uint8_t& nearestMonthAfter, std::uint8_t& nearestDayAfter, bool& hasAfter) const
{
    hasBefore = false;
    hasCurrent = false;
    hasAfter = false;

    if
        (this->store.empty())
    {
        return false;
    }

    auto yearIt = this->store.lower_bound(refYear);
    if
        (   yearIt != this->store.cend()
         && refYear == yearIt->first)
    {
        LOG(LogLevel::Debug, "EventsCoverage", "Data found in reference year");

        // Check if reference date is available
        const std::size_t refMonthIndex {static_cast<std::size_t>(refMonth - 1)};   //
        std::size_t monthIndex {refMonthIndex};    //
        hasCurrent = (0 < ((1 << (refDay - 1)) & yearIt->second[monthIndex]));

        // Look for a date with data within that year but before reference date
        for
            (;false == hasBefore;)
        {
            LOG(LogLevel::Debug, "EventsCoverage", "... checking {}-{}", refYear, monthIndex + 1);
            // No data this month? check previous month or stop if the
            // beginning of the year has been reached
            if
                (0 == yearIt->second[monthIndex])
            {
                if
                    (monthIndex > 0)
                {
                    --monthIndex;
                    continue;
                }
                break;
            }

            // Start from month's end, or from the day before reference date
            // if the inspected month matches the reference
            std::uint8_t day {31};  //
            if
                (monthIndex == refMonthIndex)
            {
                if
                    (1 == refDay)
                {

                    if
                        (0 == monthIndex)
                    {
                        // Reference date was 1st January, so there is no way
                        // to go further in the past while remaining in the
                        // reference year.
                        break;
                    }
                    --monthIndex;
                    continue;
                }
                day = refDay - 1;
            }

            // Check each day of the month in descending order and stop as soon
            // as a day has some data
            for
                (;; --day)
            {
                LOG(LogLevel::Debug, "EventsCoverage", "... checking {}-{}-{}", refYear, monthIndex + 1, day);
                if
                    ((1 << (day - 1)) & yearIt->second[monthIndex])
                {
                    hasBefore = true;
                    nearestYearBefore = refYear;
                    nearestMonthBefore = static_cast<std::uint8_t>(monthIndex + 1);
                    nearestDayBefore = day;
                    break;
                }

                // Needed to properly stop the loop as the condition is based
                // on a decreasing unsigned integer (std::uint8_t)
                if
                    (1 == day)
                {
                    break;
                }
            }

            // Needed to properly stop the loop as the condition is based on
            // a decreasing unsigned integer (std::size_t)
            if
                (0 == monthIndex)
            {
                break;
            }

            --monthIndex;
        }

        // Look for a date with data within that year but after reference date
        monthIndex = refMonthIndex;
        for
            (; (false == hasAfter) && (monthIndex < 12); ++monthIndex)
        {
            LOG(LogLevel::Debug, "EventsCoverage", "... checking {}-{}", refYear, monthIndex + 1);
            // Skip month if it contains no data
            if
                (0 == yearIt->second[monthIndex])
            {
                continue;
            }

            // Start from month's beginning, or from the day after reference
            // date if the inspected month matches the reference
            std::uint8_t day {1};  //
            if
                (monthIndex == refMonthIndex)
            {
                day = refDay + 1;
            }

            // Check each day of the month in ascending order and stop as soon
            // as a day has some data
            for
                (;day < 32; ++day)
            {
                LOG(LogLevel::Debug, "EventsCoverage", "... checking {}-{}-{}", refYear, monthIndex + 1, day);
                if
                    ((1 << (day - 1)) & yearIt->second[monthIndex])
                {
                    hasAfter = true;
                    nearestYearAfter = refYear;
                    nearestMonthAfter = static_cast<std::uint8_t>(monthIndex + 1);
                    nearestDayAfter = day;
                    break;
                }
            }
        }

        // Year of the reference date has some data, so at this point at least
        // one nearest date has been found (either past, current or future),
        // but it may not be the actual nearest date because previous and next
        // year may offer closer candidates.

        // If no anterior data has been found within the reference year, look
        // for the last day with data in the previous year
        if
            (  (false == hasBefore)
            && (yearIt != store.cbegin()))
        {
            auto yearBeforeIt = std::prev(yearIt, 1);
            if
                (yearBeforeIt->first == refYear - 1)
            {
                getLastCoveredDay(yearBeforeIt->second, nearestMonthBefore, nearestDayBefore);
                nearestYearBefore = refYear - 1;
                hasBefore = true;
            }
        }

        // If no posterior data has been found within the reference year, look
        // for the first day with data in the next year
        if
            (  (false == hasAfter)
            && (yearIt != store.cend()))
        {
            auto yearAfterIt = std::next(yearIt, 1);
            if
                (   yearAfterIt != store.cend()
                 && yearAfterIt->first == refYear + 1)
            {
                getFirstCoveredDay(yearAfterIt->second, nearestMonthAfter, nearestDayAfter);
                nearestYearAfter = refYear + 1;
                hasAfter = true;
            }
        }
    }
    else if
        (store.cbegin() == yearIt)
    {
        // There is no data in years before the reference date, so only search
        // in later year
        hasAfter = true;
        nearestYearAfter = yearIt->first;
        getFirstCoveredDay(yearIt->second, nearestMonthAfter, nearestDayAfter);
    }
    else if
        (store.cend() == yearIt)
    {
        // There is no data in years after the reference date, so only search
        // in earlier years
        hasBefore = true;
        auto nearestYearBeforeIt = std::prev(yearIt, 1);
        nearestYearBefore = nearestYearBeforeIt->first;
        getLastCoveredDay(nearestYearBeforeIt->second, nearestMonthBefore, nearestDayBefore);
    }
    else
    {
        // There is no data in the year of the reference date, but there are
        // some data in earlier and later years, so look for the last covered
        // day of the closest year in the past and for the first covered day of
        // the closest year in the future
        hasBefore = true;
        auto nearestYearBeforeIt = std::prev(yearIt, 1);
        nearestYearBefore = nearestYearBeforeIt->first;
        getLastCoveredDay(nearestYearBeforeIt->second, nearestMonthBefore, nearestDayBefore);

        hasAfter = true;
        nearestYearAfter = yearIt->first;
        getFirstCoveredDay(yearIt->second, nearestMonthAfter, nearestDayAfter);
    }
    return true;
}

/**
 *
 */
void EventsCoverage::display() const
{
	coverage_t::const_iterator it;
	std::array<uint8_t, 12> days_per_months = {{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }};
	uint8_t month_adjust;
	uint8_t days_count;
	size_t i;
	size_t j;

	it = this->store.begin();
	for
		(;it != this->store.end(); ++it)
	{
		LOG(LogLevel::Debug, "EventsCoverage", "{}",  static_cast<int>(it->first));
		i = 1u;
		for
			(;i<13; ++i)
		{
			month_adjust = (0 == it->first % 4 && 2 == i) ? 1u : 0u;
			days_count = static_cast<uint8_t>(days_per_months[i-1u] + month_adjust);
			LOG(LogLevel::Debug, "EventsCoverage", "\t{} [{}]: {}", static_cast<int>(i), static_cast<int>(days_count), static_cast<int>(it->second[i-1]));
			for
				(j=0; j<days_count; ++j)
			{
				LOG(LogLevel::Debug, "EventsCoverage", "\t\t{}", (((it->second[i-1] >> j) & 1) == 1));
			}
		}
	}
}
