// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <bitset>

#include "core/log.h"
#include "3d/opengl_enum.h"
#include "data/data_handler.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "data/gpu_data_builder.h"
#include "3d/data_set.h"
#include "num_utils.h"
#include "data/unrolled_functions.h"

using namespace SEAScope;


/**
 *
 */
GPUDataBuilder::GPUDataBuilder()
{
};

/**
 *
 */
GPUDataBuilder::~GPUDataBuilder()
{
};

/**
 *
 */
bool GPUDataBuilder::buildDataSet(DataSet* data_set,
                                  void* raw_data,
                                  const DataInfo& dataInfo,
								  const CollectionCfg& colCfg,
                                  const Shape& shape,
                                  float resolution,
                                  int maxTextureSize)
{
    LOG(LogLevel::Debug, "GPUDataBuilder", "Create new dataset");

    // Initialize variables
    data_set->setWidth(dataInfo.xArity);              // number of elements in the X axis
    data_set->setHeight(dataInfo.yArity);             // number of elements in the Y axis
    data_set->setDimensionCount(dataInfo.channels);   // Number of variables interleaved in the data stream
    data_set->setResolution(resolution);              //

    float* fValuePtr = new float[dataInfo.channels];

    data_set->setOffset(fValuePtr);

    memcpy( fValuePtr
          , dataInfo.offsets.data()
          , sizeof(float) * dataInfo.offsets.size());

    fValuePtr = new float[dataInfo.channels];

    data_set->setScaleFactor(fValuePtr);

    memcpy( fValuePtr
          , dataInfo.scaleFactors.data()
          , sizeof(float) * dataInfo.scaleFactors.size());


    if (dataInfo.dataType == DataType::unsigned_byte)
    {
        data_set->setScaleFactorGL(255.f);
    }
    else
    {
        data_set->setScaleFactorGL(1.f);
    }

    // GCP
    std::uint32_t lon_size = shape.xArity;
    std::uint32_t lat_size = shape.yArity;

	LOG(LogLevel::Debug, "GPUDataBuilder", "Number of GCPs: {} {}x{}", shape.gcpsLat.size(), shape.xArity, shape.yArity);

    /*
    for (std::size_t i = 0; i < shape.gcps.size(); ++i)
    {
        const SEAScope::GCP& gcp = shape.gcps[i];

        gcpx[i] = gcp.i;
        gcpy[i] = gcp.j;
        lons[i] = gcp.lon;
        lats[i] = gcp.lat;

		LOG(LogLevel::Debug, "GPUDataBuilder", "@{} {} {} {} {}", i, gcpy[i], gcpx[i], lats[i], lons[i]);
    }
    */

    // TODO: find a clean way to share GCPs without a copy (simply passing vector.data() has side-effects)
    const std::uint32_t* gcpx = new std::uint32_t[static_cast<size_t>(shape.xArity * shape.yArity)];
    std::memcpy(const_cast<void*>(reinterpret_cast<const void*>(gcpx)), shape.gcpsMinorAxisIndex.data(), sizeof(std::uint32_t) * shape.gcpsMinorAxisIndex.size());

    const std::uint32_t* gcpy = new std::uint32_t[static_cast<size_t>(shape.xArity * shape.yArity)];
    std::memcpy(const_cast<void*>(reinterpret_cast<const void*>(gcpy)), shape.gcpsMajorAxisIndex.data(), sizeof(std::uint32_t) * shape.gcpsMajorAxisIndex.size());

    const float* lons = new float[static_cast<size_t>(shape.xArity * shape.yArity)];
    std::memcpy(const_cast<void*>(reinterpret_cast<const void*>(lons)), shape.gcpsLon.data(), sizeof(float) * shape.gcpsLon.size());

    const float* lats = new float[static_cast<size_t>(shape.xArity * shape.yArity)];
    std::memcpy(const_cast<void*>(reinterpret_cast<const void*>(lats)), shape.gcpsLat.data(), sizeof(float) * shape.gcpsLat.size());

    data_set->setGcpx(gcpx);
    data_set->setGcpy(gcpy);
    data_set->setLon_size(lon_size);
    data_set->setLat_size(lat_size);
    data_set->setLons(lons);
    data_set->setLats(lats);

	data_set->setSeamLess(colCfg.xSeamless, colCfg.ySeamless);

	/* Compute the axis aligned bounding box
	    0 --------- 1
        |           |
        |           |
        3 --------- 2
	*/
    glm::vec2 corners[4];

    corners[0] = glm::vec2(lons[0], lats[0]);
    corners[1] = glm::vec2(lons[lon_size-1], lats[lon_size-1]);
    corners[2] = glm::vec2(lons[lon_size * lat_size - 1], lats[lon_size * lat_size - 1]);
    corners[3] = glm::vec2(lons[(lat_size-1) * lon_size], lats[(lat_size-1) * lon_size]);

    for (int i = 0; i < 4; ++i)
    {
        data_set->setCorner(i, glm::vec4(UnitSphere::ToVector3D(Geodetic2D(glm::radians(corners[i]))), 1.f));
    }

    glm::vec4 x0 = data_set->getCorner(1) - data_set->getCorner(0);
    glm::vec4 y0 = data_set->getCorner(3) - data_set->getCorner(0);

    glm::vec4 x = glm::normalize(x0);
    glm::vec4 y = glm::normalize(y0);

    glm::vec3 nDataSet = glm::normalize(glm::cross(glm::vec3(x.x, x.y, x.z), glm::vec3(y.x, y.y, y.z)));

    std::size_t middleIndex = ((data_set->getLat_size() / 2 - 1) * data_set->getLon_size() + data_set->getLon_size() / 2);
    glm::vec2 middle2(data_set->getLons()[middleIndex], data_set->getLats()[middleIndex]);
    glm::vec4 middle3 = glm::vec4(UnitSphere::ToVector3D(Geodetic2D(glm::radians(middle2))), 1.f);

    glm::vec4 fmr = glm::normalize(middle3 - data_set->getCorner(0));

    glm::vec3 middleV = glm::vec3(fmr.x, fmr.y, fmr.z);

    float dotValue = glm::dot(nDataSet, middleV);

    if (dotValue < 0.f)
    {
        nDataSet = -nDataSet;
    }

    data_set->setNormal(nDataSet);
    data_set->setCorner(4, middle3);
    data_set->setCenterPlane(data_set->getCorner(0) + ((data_set->getCorner(2) - data_set->getCorner(0)) * 0.5f));

    // Buffer for the raw data
    LOG(LogLevel::Debug, "GPUDataBuilder", "Copy data from void*");
    data_set->setData(dataInfo.xArity, dataInfo.yArity, dataInfo.channels, dataInfo.dataType, raw_data);

    // compute a mask to display only valid data
    LOG(LogLevel::Debug, "GPUDataBuilder", "Compute mask");
    data_set->setDataMask(dataInfo.xArity, dataInfo.yArity, 1, DataType::unsigned_byte, nullptr);

    float* min = new float[dataInfo.channels];
    float* max = new float[dataInfo.channels];

    computeMaskAndMinMax(data_set->getData(), data_set->getDataMask(), dataInfo.fillValues[0], min, max);

    data_set->setMin(min);
    data_set->setMax(max);

    // Dilation for bilinear filtering
    createMaskBorder(*data_set->getData(), *data_set->getDataMask(), data_set->getOffset(), data_set->getScaleFactor());

    return data_set->compute(maxTextureSize);
};

/**
 *
 */
bool GPUDataBuilder::buildPolyline( void* raw_data
                                  , const DataInfo& dataInfo
								  , const CollectionCfg& /*colCfg*/
                                  , const Shape& shape
                                  , std::vector<float>& relativeTime
                                  , std::vector<Geodetic2D>& points
                                  , std::vector<float>& fData
								  , std::vector<bool>& mask)
{
	std::uint8_t* uintBuffer{nullptr};	//
	float* floatBuffer{nullptr};	//
	float scaleFactor{1.f};	//
	float offset{0.f};	//
	float physicalValue{0.f};	//
	float packedValue{0.f};	//
	bool unmasked{false};	//

    if
        (2 > shape.xArity)
    {
        LOG(LogLevel::Error, "GPUDataBuilder", "Cannot build polyline with less than two points");
        return false;
    }

    // Assumming time values sorted in ascending order
    const std::size_t gcpsCount{shape.gcpsTime.size()};

    // In case all times are the same, relativeTime will be 0 for all GCPs and
    // timeExtent should be set to 1 to avoid divide by 0 errors
    float timeExtent = static_cast<float>(shape.gcpsTime[gcpsCount - 1]) - static_cast<float>(shape.gcpsTime[0]);
    if
        (0.f >= timeExtent)
    {
        timeExtent = 1.f;
    }

    if
        (nullptr == raw_data)
    {
        LOG(LogLevel::Error, "GPUDataBuilder", "buildPolyline received an invalid (null) data buffer");
        return false;
    }

	floatBuffer = nullptr;
	uintBuffer = nullptr;

    points.reserve(shape.xArity);
    relativeTime.reserve(shape.xArity);

    for
        (std::size_t i = 0u; i < gcpsCount; ++i)
    {
        points.push_back(Geodetic2D( static_cast<double>(glm::radians(shape.gcpsLon[i]))
                                   , static_cast<double>(glm::radians(shape.gcpsLat[i]))));
        relativeTime.push_back((static_cast<float>(shape.gcpsTime[i]) - static_cast<float>(shape.gcpsTime[0])) / timeExtent);
    }

    if
        (1 == dataInfo.channels)
    {
		fData.reserve(shape.xArity);
		mask.reserve(shape.xArity);
		scaleFactor = dataInfo.scaleFactors[0];
		offset = dataInfo.offsets[0];
		if
			(dataInfo.dataType == DataType::float32)
		{
			floatBuffer = static_cast<float*>(raw_data);
		}
		else
		{
			uintBuffer = static_cast<std::uint8_t*>(raw_data);
		}
        for (std::uint32_t i = 0; i < shape.xArity; ++i)
        {
			if
				(dataInfo.dataType == DataType::float32)
			{
				packedValue = static_cast<float>(*(floatBuffer + i));
			}
			else
			{
				packedValue = static_cast<float>(*(uintBuffer + i));
			}
			unmasked = (false == floating_point_equality(dataInfo.fillValues[0], packedValue));
			physicalValue = offset + scaleFactor * packedValue;
			LOG(LogLevel::Debug, "GPUDataBuilder", "value: {} + {} * {} = {}", offset, scaleFactor, packedValue, physicalValue);
			mask.push_back(unmasked);
            fData.push_back(physicalValue);
        }
    }

	floatBuffer = nullptr;
	uintBuffer = nullptr;

    return true;
};
