// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cstring>

#include "core/log.h"
#include "core/copy.h"
#include "data/image.h"
//#include <malloc.h>

using namespace SEAScope;

Image::Image()
:   _width(0),
    _height(0),
    _componentCount(0),
    _componentSize(0),
    _dataType(DataType::none),
    _rawData(nullptr),
    _valid(false)
{
}

Image::~Image()
{
	this->deallocate();
}

Image* Image::clone() const
{
    Image* copy = new Image;
    copy->allocate(_width, _height, _componentCount, _dataType, _rawData);
    return copy;
}

void Image::deallocate()
{
    this->_valid = false;
    this->_dataType = DataType::none;
    this->_width = 0;
    this->_height = 0;
    this->_componentCount = 0;
    this->_componentSize = 0;
	if
		(nullptr != this->_rawData)
	{
		free(this->_rawData);
		//_mm_free(this->_rawData);
	    this->_rawData = nullptr;
	}
};

void Image::allocate(unsigned int width, unsigned int height, unsigned int componentCount, DataType dataType, void * data)
{
    unsigned int dataTypeSize = 4;
    
    switch(dataType)
    {
    case DataType::float32:
        dataTypeSize = 4;
        break;
    case DataType::unsigned_byte:
        dataTypeSize = 1;
        break;
    case DataType::none:
    case DataType::unsigned_short:
    case DataType::unsigned_int:
    case DataType::DataTypeCount:
        LOG(SEAScope::LogLevel::Error, "Image", "Image allocation requested with an unsupported data type");
        this->_valid = false;
        return;
    }
    
    allocate(width, height, componentCount, dataTypeSize, data);
    
    _dataType = dataType;
    this->_valid = true;
}

void Image::allocate(unsigned int width, unsigned int height, unsigned int componentCount, unsigned int componentSize, void * data)
{
    unsigned int dataSize = width * height * componentCount * componentSize;

    if (this->_width != width ||
        this->_height != height ||
        this->_componentCount != componentCount ||
        this->_componentSize != componentSize ||
        this->_rawData == nullptr)
    {
        if (width == 0 ||
            height == 0 ||
            componentCount == 0 ||
            componentSize == 0)
        {
            LOG(LogLevel::Error, "Image", ":allocate :a parameter equal zero.");
            return;
        }

        if (nullptr != this->_rawData)
        {
            free(this->_rawData);
			this->_rawData = nullptr;
        }

        this->_width = width;
        this->_height = height;
        this->_componentCount = componentCount;
        this->_componentSize = componentSize;
        this->_dataType = DataType::none;

		LOG(LogLevel::Debug, "Image", "allocate with malloc: {} MB", static_cast<float>(dataSize) / 1024.f / 1024.f);
        this->_rawData = malloc(dataSize * sizeof(std::uint8_t));
        //this->_rawData = _mm_malloc(dataSize * sizeof(std::uint8_t), 32);
    }

    if
        (nullptr == this->_rawData)
    {
        LOG(LogLevel::Error, "Image", "Failed to allocate memory");
        return;
    }

    if (data)
    {
		LOG(LogLevel::Debug, "Image", "Copy original data: {} MB", static_cast<float>(dataSize) / 1024.f / 1024.f);
        memcpy(this->_rawData, data, dataSize);
    }
    else
    {
		LOG(LogLevel::Debug, "Image", "0-fill data buffer");
        memset(this->_rawData, 0, dataSize);
    }
}

void Image::allocate(const ImageDesc& desc, void* data)
{
    allocate(desc.width, desc.height, desc.channels, desc.dataType, data);
}

void Image::copyTo(const glm::ivec2& srcOrigin, const glm::ivec2& region, Image& dstImage)
{
    // Check parameters of the destination image. 
    if (   dstImage._componentCount != _componentCount
        || dstImage._componentSize != _componentSize
        || dstImage._dataType != _dataType
        || dstImage._height != static_cast<unsigned int>(region.y)
        || dstImage._width != static_cast<unsigned int>(region.x)
        || dstImage._rawData == nullptr)
    {
        dstImage.allocate(static_cast<unsigned int>(region.x), static_cast<unsigned int>(region.y), _componentCount, _dataType, nullptr);
    }
    
    copyTo(srcOrigin, dstImage, glm::ivec2(0, 0), region);
}

void Image::copyTo(const glm::ivec2& srcOrigin, const Image& dstImage, const glm::ivec2& dstOrigin, const glm::ivec2& region)
{
    assert(nullptr != this->_rawData);
    assert(nullptr != dstImage._rawData);
    assert(pixelMemSize() == dstImage.pixelMemSize());
    
    const unsigned int pixelSize = pixelMemSize();
    const unsigned int srcRowSize = this->_width * pixelMemSize();
    const unsigned int dstRowSize = dstImage._width * dstImage.pixelMemSize();
    
    const char* srcOriginOffset = static_cast<char*>(this->_rawData) + static_cast<unsigned int>(srcOrigin.y) * srcRowSize + static_cast<unsigned int>(srcOrigin.x) * pixelSize;
    const char* dstOriginOffset = static_cast<char*>(dstImage._rawData) + static_cast<unsigned int>(dstOrigin.y) * dstRowSize + static_cast<unsigned int>(dstOrigin.x) * pixelSize;
    
    unsigned int regionRowSize = static_cast<unsigned int>(region.x) * pixelSize;
    
    for (unsigned int y = 0; y < static_cast<unsigned int>(region.y); ++y)
    {
        memcpy(reinterpret_cast<void*>(const_cast<char*>(dstOriginOffset + y * dstRowSize)),
               reinterpret_cast<void*>(const_cast<char*>(srcOriginOffset + y * srcRowSize)),
               regionRowSize);
    }
}

void Image::extractChannel(unsigned int channelID, Image& out) const
{
    if
        (channelID >= this->_componentCount)
    {
        return;
    }
    
	LOG(LogLevel::Debug, "Image", "extractChannel: allocate");
    out.allocate(this->_width, this->_height, 1, this->_componentSize, nullptr);
    
    char * srcPtr = static_cast<char*>(this->_rawData) + channelID * _componentSize;
    char * dstPtr = static_cast<char*>(out.data());
    unsigned int pixelOffset = pixelMemSize();
    unsigned int pixelCount = this->_width * this->_height;
    
    for (unsigned int i = 0; i < pixelCount; ++i)
    {
        memcpy(dstPtr, srcPtr, this->_componentSize);
        srcPtr += pixelOffset;
        dstPtr += this->_componentSize;
    }
}

void Image::flipVerticaly()
{
    unsigned int halfRow = _height / 2;
    const unsigned int pixelSize = pixelMemSize();
    char* pixelData = static_cast<char*>(malloc(pixelSize));
    unsigned int rowSize = pixelSize * _width;
    char* top;
    char* bottom;
    
    for (unsigned int col = 0; col < _width; ++col)
    {
        char* colPtr = static_cast<char*>(_rawData) + col * pixelSize;
        
        for (unsigned int row = 0; row < halfRow; ++row)
        {
            top = colPtr + row * rowSize;
            bottom = colPtr + (_height - 1 - row) * rowSize;
            
            // top to tmp
            memcpy(reinterpret_cast<void*>(pixelData), reinterpret_cast<void*>(top), pixelSize);
            
            // bottom to top
            memcpy(reinterpret_cast<void*>(top), reinterpret_cast<void*>(bottom), pixelSize);
            
            // tmp to bottom
            memcpy(reinterpret_cast<void*>(bottom), reinterpret_cast<void*>(pixelData), pixelSize);
        }
    }
    
    free(pixelData);
    pixelData = nullptr;
}

void Image::flipHorizontaly()
{
    unsigned int halfColumn = _width / 2;
    const unsigned int pixelSize = pixelMemSize();
    char* pixelData = static_cast<char*>(malloc(pixelSize));
    unsigned int rowSize = pixelSize * _width;
    char* left;
    char* right;
    
    for (unsigned int row = 0; row < _height; ++row)
    {
        char* rowPtr = static_cast<char*>(_rawData) + row * rowSize;

        for (unsigned int col = 0; col < halfColumn; ++col)
        {
            left = rowPtr + col * pixelSize;
            right = rowPtr + (_width - col - 1) * pixelSize; 
            
            // left to tmp
            memcpy(reinterpret_cast<void*>(pixelData), reinterpret_cast<void*>(left), pixelSize);
            
            // right to left
            memcpy(reinterpret_cast<void*>(left), reinterpret_cast<void*>(right), pixelSize);
            
            // tmp to right
            memcpy(reinterpret_cast<void*>(right), reinterpret_cast<void*>(pixelData), pixelSize);
        }
    }
    
    free(pixelData);
    pixelData = nullptr;
}

void Image::rotate90(ERotation /*rotation*/)
{
	LOG(LogLevel::Debug, "Image", "rotate90: allocate {} MB", static_cast<float>(_width * _height * pixelMemSize()) / 1024.f / 1024.f);
    char * newData = static_cast<char*>(malloc(_width * _height * pixelMemSize()));
    if
        (nullptr == newData)
    {
        LOG(SEAScope::LogLevel::Error, "Image", "Could not allocate memory for rotated image");
        return;
    }
    const unsigned int pixelSize = _componentCount * _componentSize;

    for (unsigned int x = 0; x < _width; ++x)
    {
        for (unsigned int y = 0; y < _height; ++y)
        {
            *(newData + (x * _width + (_height - 1 - y)) * pixelSize) = *(static_cast<char*>(_rawData) + (y * _width + x) * pixelSize);
        }
    }
    
    void * tmp = this->_rawData;
    this->_rawData = newData;
    free(tmp);
    tmp = nullptr;
}

ImageDesc Image::getDesc() const
{
    ImageDesc desc;
    desc.width = this->_width;
    desc.height = this->_height;
    desc.channels = this->_componentCount;
    desc.dataType = this->_dataType;
    return desc;
}

void Image::setData(const ImageDesc& desc, void* data)
{
    this->_width = desc.width;
    this->_height = desc.height;
    this->_componentCount = desc.channels;
    this->_dataType = desc.dataType;
    this->_valid = true;
    
    switch (this->_dataType)
    {
    case DataType::float32:
        this->_componentSize = 4;
        break;
    case DataType::unsigned_byte:
        this->_componentSize = 1;
        break;
    case DataType::none:
    case DataType::unsigned_short:
    case DataType::unsigned_int:
    case DataType::DataTypeCount:
        LOG(SEAScope::LogLevel::Error, "Image", "Image allocation requested with an unsupported data type");
        this->_valid = false;
        return;
    }

    if
        (nullptr != this->_rawData)
    {
        free(this->_rawData);
    }

    this->_rawData = data;
}

unsigned long long Image::dataSize() const
{
    return static_cast<unsigned long long>(_width * _height * pixelMemSize());
}

unsigned int Image::pixelMemSize() const
{
    return _componentCount * _componentSize;
}

void* Image::pixelData(unsigned int x, unsigned int y)
{
    const unsigned int pixelSize = pixelMemSize();
    
    char* charDataPtr = static_cast<char*>(_rawData);
    
    charDataPtr = charDataPtr + (y * _width + x) * pixelSize;
    
    return reinterpret_cast<void*>(charDataPtr);
}

void Image::setPixelData(unsigned int x, unsigned int y, void* data)
{
	const unsigned int pixelSize = pixelMemSize();

	char* charDataPtr = static_cast<char*>(_rawData);

	charDataPtr = charDataPtr + (y * _width + x) * pixelSize;

    memcpy(charDataPtr, data, pixelSize);
}
