// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "data/unrolled_functions.h"

#include "num_utils.h"

using namespace SEAScope;

template<typename T, unsigned int channels>
void soa_to_aos_impl(const std::vector<const Image*>&, Image*)
{
}

/// ////////////////////////////////////////////////////////////////////////////////
/// unsigned char
/// 

template<>
void soa_to_aos_impl<unsigned char, 2>(const std::vector<const Image*>& src, Image* dst)
{
    struct uint8x2
    {
        uint8_t r;
        uint8_t g;
    };

    uint8x2* ptr = static_cast<uint8x2*>(dst->data());
    uint8_t* buffer0 = static_cast<uint8_t*>(src[0]->data());
    uint8_t* buffer1 = static_cast<uint8_t*>(src[1]->data());

    const unsigned int count = src[0]->pixelCount();

    for
        (unsigned int i = 0; i < count; ++i)
    {
        *ptr++ = { *buffer0++, *buffer1++ };
    }
}

template<>
void soa_to_aos_impl<unsigned char, 3>(const std::vector<const Image*>& src, Image* dst)
{
    struct uint8x3
    {
        uint8_t r;
        uint8_t g;
        uint8_t b;
    };

    uint8x3* ptr = static_cast<uint8x3*>(dst->data());
    uint8_t* buffer0 = static_cast<uint8_t*>(src[0]->data());
    uint8_t* buffer1 = static_cast<uint8_t*>(src[1]->data());
    uint8_t* buffer2 = static_cast<uint8_t*>(src[2]->data());

    const unsigned int count = src[0]->pixelCount();

    for
        (unsigned int i = 0; i < count; ++i)
    {
        *ptr++ = { *buffer0++, *buffer1++, *buffer2++ };
    }
}

/// ////////////////////////////////////////////////////////////////////////////////
/// float
/// 

template<>
void soa_to_aos_impl<float, 2>(const std::vector<const Image*>& src, Image* dst)
{
    struct float32x2
    {
        float r;
        float g;
    };

    float32x2* ptr = static_cast<float32x2*>(dst->data());
    float* buffer0 = static_cast<float*>(src[0]->data());
    float* buffer1 = static_cast<float*>(src[1]->data());

    // Interleave data values
    const unsigned int count = src[0]->pixelCount();

    for
        (unsigned int i = 0; i < count; ++i)
    {
        *ptr++ = { *buffer0++, *buffer1++ };
    }
}

template<>
void soa_to_aos_impl<float, 3>(const std::vector<const Image*>& src, Image* dst)
{
    struct float32x3
    {
        float r;
        float g;
        float b;
    };

    float32x3* ptr = static_cast<float32x3*>(dst->data());
    float* buffer0 = static_cast<float*>(src[0]->data());
    float* buffer1 = static_cast<float*>(src[1]->data());
    float* buffer2 = static_cast<float*>(src[2]->data());

    // Interleave data values
    const unsigned int count = src[0]->pixelCount();

    for
        (unsigned int i = 0; i < count; ++i)
    {
        *ptr++ = { *buffer0++, *buffer1++, *buffer2++ };
    }
}

void soa_to_aos(const std::vector<const Image*>& src, Image* dst)
{
    if
        (DataType::unsigned_byte == src[0]->dataType())
    {
        switch (src.size())
        {
        case 2:
            soa_to_aos_impl<unsigned char, 2>(src, dst);
            break;
        case 3:
            soa_to_aos_impl<unsigned char, 3>(src, dst);
            break;
        default:
            break;
        }
    }
    else if
        (DataType::float32 == src[0]->dataType())
    {
        switch (src.size())
        {
        case 2:
            soa_to_aos_impl<float, 2>(src, dst);
            break;
        case 3:
            soa_to_aos_impl<float, 3>(src, dst);
            break;
        default:
            break;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
// Compute mask min max

template<typename T> inline
void testMinMax(T& min, T& max, const T& val)
{
    if
        (min > val)
    {
        min = val;
    }
    else if
        (max < val)
    {
        max = val;
    }
}

template<typename T, unsigned int>
void computeMaskandMinMaxUnrolled(const Image*, Image*, const T, float*, float*)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
//// float versions

inline bool compareFloat1(float a, float fv)
{
    return (false == (floating_point_equality<float>(a, fv, 2)));
}

inline bool compareFloat2(float a, float b, float fv)
{
    return compareFloat1(a, fv) && compareFloat1(b, fv);
}

inline bool compareFloat3(float a, float b, float c, float fv)
{
    return compareFloat1(a, fv) && compareFloat2(b, c, fv);
}

template<>
void computeMaskandMinMaxUnrolled<float, 1>(const Image* src, Image* mask, const float fillValue, float* min, float* max)
{
    float* ptr = static_cast<float*>(src->data());
    unsigned char* ptrMask = static_cast<unsigned char*>(mask->data());

    constexpr unsigned int varCount = 1;
    const unsigned int count = src->pixelCount();
    unsigned int i = 0;

    for (; i < count; ++i)
    {
        const float& v0 = *ptr;

        if
            (compareFloat1(v0, fillValue))
        {
            *ptrMask = 255u;

            min[0] = v0;
            max[0] = v0;

            ++i;
            ++ptrMask;
            ptr += varCount;

            break;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }

    for (; i < count; ++i)
    {
        const float& v0 = *ptr;

        if
            (compareFloat1(v0, fillValue))
        {
            *ptrMask = 255u;

            testMinMax(min[0], max[0], *ptr);
            ++ptr;

            ++ptrMask;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }
}

template<>
void computeMaskandMinMaxUnrolled<float, 2>(const Image* src, Image* mask, const float fillValue, float* min, float* max)
{
    float* ptr = static_cast<float*>(src->data());
    unsigned char* ptrMask = static_cast<unsigned char*>(mask->data());

    constexpr unsigned int varCount = 2;
    const unsigned int count = src->pixelCount();
    unsigned int i = 0;

    for (; i < count; ++i)
    {
        const float& v0 = *ptr;
        const float& v1 = *(ptr + 1);

        if
            (compareFloat2(v0, v1, fillValue))
        {
            *ptrMask = 255u;

            min[0] = v0;
            min[1] = v1;

            max[0] = v0;
            max[1] = v1;

            ++i;
            ++ptrMask;
            ptr += varCount;

            break;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }

    for (; i < count; ++i)
    {
        const float& v0 = *ptr;
        const float& v1 = *(ptr + 1);

        if
            (compareFloat2(v0, v1, fillValue))
        {
            *ptrMask = 255u;

            testMinMax(min[0], max[0], *ptr);
            ++ptr;
            testMinMax(min[1], max[1], *ptr);
            ++ptr;

            ++ptrMask;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }
}

template<>
void computeMaskandMinMaxUnrolled<float, 3>(const Image* src, Image* mask, const float fillValue, float* min, float* max)
{
    float* ptr = static_cast<float*>(src->data());
    unsigned char* ptrMask = static_cast<unsigned char*>(mask->data());

    const unsigned int count = src->pixelCount();
    unsigned int i = 0;

    for (; i < count; ++i)
    {
        const float& v0 = *ptr;
        const float& v1 = *(ptr + 1);
        const float& v2 = *(ptr + 2);

        if
            (compareFloat3(v0, v1, v2, fillValue))
        {
            *ptrMask = 255u;

            min[0] = v0;
            min[1] = v1;
            min[2] = v2;

            max[0] = v0;
            max[1] = v1;
            max[2] = v2;

            ++i;
            ++ptrMask;
            ptr += 3;

            break;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += 3;
        }
    }

    for (; i < count; ++i)
    {
        const float& v0 = *ptr;
        const float& v1 = *(ptr + 1);
        const float& v2 = *(ptr + 2);

        if
            (compareFloat3(v0, v1, v2, fillValue))
        {
            *ptrMask = 255u;

            testMinMax(min[0], max[0], *ptr);
            ++ptr;
            testMinMax(min[1], max[1], *ptr);
            ++ptr;
            testMinMax(min[2], max[2], *ptr);
            ++ptr;

            ++ptrMask;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += 3;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//// unsigned char versions

template<>
void computeMaskandMinMaxUnrolled<unsigned char, 1>(const Image* src, Image* mask, const unsigned char fillValueUC, float* min, float* max)
{
    unsigned char* ptr = static_cast<unsigned char*>(src->data());
    unsigned char* ptrMask = static_cast<unsigned char*>(mask->data());
    const unsigned int count = src->pixelCount();
    unsigned int i = 0;

    constexpr unsigned int varCount = 1;
    unsigned char* minUC = new unsigned char[varCount];
    unsigned char* maxUC = new unsigned char[varCount];

    for (; i < count; ++i)
    {
        const unsigned char& v0 = *ptr;

        if
            (v0 != fillValueUC)
        {
            *ptrMask = 255u;

            minUC[0] = v0;
            maxUC[0] = v0;

            ++i;
            ++ptrMask;
            ptr += varCount;

            break;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }

    for (; i < count; ++i)
    {
        const unsigned char& v0 = *ptr;

        if
            (v0 != fillValueUC)
        {
            *ptrMask = 255u;
            testMinMax(minUC[0], maxUC[0], *ptr);
            ++ptr;
            ++ptrMask;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }

    for
        (unsigned int j = 0; j < varCount; ++j)
    {
        min[j] = static_cast<float>(minUC[j]);
        max[j] = static_cast<float>(maxUC[j]);
    }

    delete[] minUC;
    delete[] maxUC;
}

template<>
void computeMaskandMinMaxUnrolled<unsigned char, 2>(const Image* src, Image* mask, const unsigned char fillValueUC, float* min, float* max)
{
    unsigned char* ptr = static_cast<unsigned char*>(src->data());
    unsigned char* ptrMask = static_cast<unsigned char*>(mask->data());
    const unsigned int count = src->pixelCount();
    unsigned int i = 0;

    constexpr unsigned int varCount = 2;
    unsigned char* minUC = new unsigned char[varCount];
    unsigned char* maxUC = new unsigned char[varCount];

    for (; i < count; ++i)
    {
        const unsigned char& v0 = *ptr;
        const unsigned char& v1 = *(ptr + 1);

        if
            ((v0 != fillValueUC) && ((v1 != fillValueUC)))
        {
            *ptrMask = 255u;

            minUC[0] = v0;
            minUC[1] = v1;

            maxUC[0] = v0;
            maxUC[1] = v1;

            ++i;
            ++ptrMask;
            ptr += varCount;

            break;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }

    for (; i < count; ++i)
    {
        const unsigned char& v0 = *ptr;
        const unsigned char& v1 = *(ptr + 1);

        if
            ((v0 != fillValueUC) && ((v1 != fillValueUC)))
        {
            *ptrMask = 255u;

            testMinMax(minUC[0], maxUC[0], *ptr);
            ++ptr;
            testMinMax(minUC[1], maxUC[1], *ptr);
            ++ptr;

            ++ptrMask;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }

    for
        (unsigned int j = 0; j < varCount; ++j)
    {
        min[j] = static_cast<float>(minUC[j]);
        max[j] = static_cast<float>(maxUC[j]);
    }

    delete[] minUC;
    delete[] maxUC;
}

template<>
void computeMaskandMinMaxUnrolled<unsigned char, 3>(const Image* src, Image* mask, const unsigned char fillValueUC, float* min, float* max)
{
    unsigned char* ptr = static_cast<unsigned char*>(src->data());
    unsigned char* ptrMask = static_cast<unsigned char*>(mask->data());
    const unsigned int count = src->pixelCount();
    unsigned int i = 0;

    constexpr unsigned int varCount = 3;
    unsigned char* minUC = new unsigned char[varCount];
    unsigned char* maxUC = new unsigned char[varCount];

    for (; i < count; ++i)
    {
        const unsigned char& v0 = *ptr;
        const unsigned char& v1 = *(ptr + 1);
        const unsigned char& v2 = *(ptr + 2);

        if
            ((v0 != fillValueUC) && ((v1 != fillValueUC) && ((v2 != fillValueUC))))
        {
            *ptrMask = 255u;

            minUC[0] = v0;
            minUC[1] = v1;
            minUC[2] = v2;

            maxUC[0] = v0;
            maxUC[1] = v1;
            maxUC[2] = v2;

            ++i;
            ++ptrMask;
            ptr += varCount;

            break;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }

    for (; i < count; ++i)
    {
        const unsigned char& v0 = *ptr;
        const unsigned char& v1 = *(ptr + 1);
        const unsigned char& v2 = *(ptr + 2);

        if
            ((v0 != fillValueUC) && ((v1 != fillValueUC) && ((v2 != fillValueUC))))
        {
            *ptrMask = 255u;

            testMinMax(minUC[0], maxUC[0], *ptr);
            ++ptr;
            testMinMax(minUC[1], maxUC[1], *ptr);
            ++ptr;
            testMinMax(minUC[2], maxUC[2], *ptr);
            ++ptr;

            ++ptrMask;
        }
        else
        {
            *ptrMask = 0u;
            ++ptrMask;
            ptr += varCount;
        }
    }

    for
        (unsigned int j = 0; j < varCount; ++j)
    {
        min[j] = static_cast<float>(minUC[j]);
        max[j] = static_cast<float>(maxUC[j]);
    }

    delete[] minUC;
    delete[] maxUC;
}

template<typename T>
void computeMaskandMinMaxDefault(const Image&, Image&, const float, float*, float*)
{
}


template<>
void computeMaskandMinMaxDefault<float>(const Image& data, Image& mask, const float fillValue, float* min, float* max)
{
    const unsigned int varCount = data.channels();
    const unsigned int count = data.pixelCount();

    float* ptr = static_cast<float*>(data.data());
    unsigned char* ptrMask = static_cast<unsigned char*>(mask.data());

    unsigned int i = 0;

    bool dataValid;

    // Search first valid value;
    for (; i < count; ++i)
    {
        dataValid = true;

        float* ptrTmp = ptr;

        for (unsigned int j = 0; j < varCount; ++j)
        {
            dataValid &= compareFloat1(*ptr, fillValue);
            ++ptr;
        }

        if
            (dataValid)
        {
            ptr = ptrTmp;

            for (unsigned int j = 0; j < varCount; ++j)
            {
                min[j] = *ptr;
                max[j] = *ptr;
                ++ptr;
            }

            *ptrMask++ = 255u;
            ++i;
            break;
        }
        else
        {
            *ptrMask++ = 0u;
        }
    }

    // First value found, now compare next values to find min/max.
    for (; i < count; ++i)
    {
        dataValid = true;

        float* ptrTmp = ptr;

        for (unsigned int j = 0; j < varCount; ++j)
        {
            dataValid &= compareFloat1(*ptr, fillValue);
            ++ptr;
        }

        if
            (dataValid)
        {
            ptr = ptrTmp;

            for (unsigned int j = 0; j < varCount; ++j)
            {
                testMinMax(min[j], max[j], *ptr);
                ++ptr;
            }

            *ptrMask++ = 255u;
        }
        else
        {
            *ptrMask++ = 0u;
        }
    }
}

template<>
void computeMaskandMinMaxDefault<unsigned char>(const Image& data, Image& mask, const float fillValue, float* min, float* max)
{
    const unsigned int varCount = data.channels();
    const unsigned int count = data.pixelCount();
    unsigned char* ptr = static_cast<unsigned char*>(data.data());
    unsigned char* ptrMask = static_cast<unsigned char*>(mask.data());
    unsigned char* minUC = new unsigned char[varCount];
    unsigned char* maxUC = new unsigned char[varCount];
    unsigned char fillValueUC = static_cast<unsigned char>(fillValue);

    unsigned int i = 0;

    bool dataValid;

    // Search first valid value;
    for (; i < count; ++i)
    {
        dataValid = true;

        unsigned char* ptrTmp = ptr;

        for (unsigned int j = 0; j < varCount; ++j)
        {
            dataValid &= false == (*ptr == fillValueUC);
            ++ptr;
        }

        if
            (dataValid)
        {
            ptr = ptrTmp;

            for (unsigned int j = 0; j < varCount; ++j)
            {
                minUC[j] = *ptr;
                maxUC[j] = *ptr;
                ++ptr;
            }

            *ptrMask++ = 255u;
            ++i;
            // First value found, break the loop
            break;
        }
        else
        {
            *ptrMask++ = 0u;
        }
    }

    // First value found, now compare next values to find min/max.
    for (; i < count; ++i)
    {
        dataValid = true;

        unsigned char* ptrTmp = ptr;

        for (unsigned int j = 0; j < varCount; ++j)
        {
            dataValid &= false == (*ptr == fillValueUC);
            ++ptr;
        }

        if
            (dataValid)
        {
            ptr = ptrTmp;

            for (unsigned int j = 0; j < varCount; ++j)
            {
                testMinMax(minUC[j], maxUC[j], *ptr);
                ++ptr;
            }

            *ptrMask++ = 255u;
        }
        else
        {
            *ptrMask++ = 0u;
        }
    }

    for
        (unsigned int j = 0; j < varCount; ++j)
    {
        min[j] = static_cast<float>(minUC[j]);
        max[j] = static_cast<float>(maxUC[j]);
    }

    delete[] minUC;
    delete[] maxUC;
}

void computeMaskAndMinMax(const Image* src, Image* mask, const float fillValue, float* min, float* max)
{
    const unsigned int varCount = src->channels();

    if (DataType::float32 == src->dataType())
    {
        switch (varCount)
        {
        case 1:
            computeMaskandMinMaxUnrolled<float, 1>(src, mask, fillValue, min, max);
            break;
        case 2:
            computeMaskandMinMaxUnrolled<float, 2>(src, mask, fillValue, min, max);
            break;
        case 3:
            computeMaskandMinMaxUnrolled<float, 3>(src, mask, fillValue, min, max);
            break;
        default:
            computeMaskandMinMaxDefault<float>(*src, *mask, fillValue, min, max);
            break;
        }
    }
    else if (DataType::unsigned_byte == src->dataType())
    {
        const unsigned char fillValueCasted = static_cast<unsigned char>(fillValue);

        switch (varCount)
        {
        case 1:
            computeMaskandMinMaxUnrolled<unsigned char, 1>(src, mask, fillValueCasted, min, max);
            break;
        case 2:
            computeMaskandMinMaxUnrolled<unsigned char, 2>(src, mask, fillValueCasted, min, max);
            break;
        case 3:
            computeMaskandMinMaxUnrolled<unsigned char, 3>(src, mask, fillValueCasted, min, max);
            break;
        default:
            computeMaskandMinMaxDefault<unsigned char>(*src, *mask, fillValueCasted, min, max);
            break;
        }
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Dilate

template<typename T>
void createMaskBorderUnrolled1(Image& image, Image& mask, float* offset, float* scaleFactor)
{
    const int width = static_cast<int>(image.width());
    const unsigned int width1 = image.width() - 1;
    const unsigned int height1 = image.height() - 1;
    const unsigned int channels = image.channels();

    // Offset to find all neighbours around a texel
    const int offsets[] =
    {
        -width - 1,
        -width,
        -width + 1,
        -1,
        +1,
        width - 1,
        width,
        width + 1
    };

    float accum;
    unsigned int idOffset;
    unsigned int count = 0;

    unsigned char* ptrMask = static_cast<unsigned char*>(mask.data());
    T* ptrSrcData = static_cast<T*>(image.data());
    T* ptrDstData = static_cast<T*>(image.data());

    for (unsigned int y = 1; y < height1; ++y)
    {
        for (unsigned int x = 1; x < width1; ++x)
        {
            unsigned int id = y * static_cast<unsigned int>(width) + x;

            if (ptrMask[id] == 0)
            {
                accum = 0.f;
                count = 0;

                for (unsigned int i = 0; i < 8; ++i)
                {
                    idOffset = static_cast<unsigned int>(static_cast<int>(id) + offsets[i]);

                    if (ptrMask[idOffset] == 255)
                    {
                        accum += static_cast<float>(ptrSrcData[idOffset * channels]) * *scaleFactor + *offset;
                        ++count;
                    }
                }

                if (count)
                {
                    const unsigned int idCh = id * channels;
                    T* ptrTmp = ptrDstData + idCh;
                    *ptrTmp++ = static_cast<T>(((accum / static_cast<float>(count)) - *offset) / *scaleFactor);
                }
            }
        }
    }
}

template<typename T>
void createMaskBorderUnrolled2(Image& image, Image& mask, float* offset, float* scaleFactor)
{
    const int width = static_cast<int>(image.width());
    const unsigned int width1 = image.width() - 1;
    const unsigned int height1 = image.height() - 1;
    const unsigned int channels = image.channels();

    // Offset to find all neighbours around a texel
    const int offsets[] =
    {
        -width - 1,
        -width,
        -width + 1,
        -1,
        +1,
        width - 1,
        width,
        width + 1
    };

    struct float32x2
    {
        float a, b;
    };

    float accum[2];
    float32x2* accumReset = reinterpret_cast<float32x2*>(accum);

    unsigned int idOffset;
    unsigned int count = 0;

    unsigned char* ptrMask = static_cast<unsigned char*>(mask.data());
    T* ptrSrcData = static_cast<T*>(image.data());
    T* ptrDstData = static_cast<T*>(image.data());

    for (unsigned int y = 1; y < height1; ++y)
    {
        for (unsigned int x = 1; x < width1; ++x)
        {
            unsigned int id = y * static_cast<unsigned int>(width) + x;

            if (ptrMask[id] == 0)
            {
                *accumReset = { 0.f, 0.f };
                count = 0;

                for (unsigned int i = 0; i < 8; ++i)
                {
                    idOffset = static_cast<unsigned int>(static_cast<int>(id) + offsets[i]);

                    if (ptrMask[idOffset] == 255)
                    {
                        accum[0] += static_cast<float>(ptrSrcData[idOffset * channels + 0]) * scaleFactor[0] + offset[0];
                        accum[1] += static_cast<float>(ptrSrcData[idOffset * channels + 1]) * scaleFactor[1] + offset[1];

                        ++count;
                    }
                }

                if (count)
                {
                    const unsigned int idCh = id * channels;
                    T* ptrTmp = ptrDstData + idCh;

                    *ptrTmp++ = static_cast<T>(((accum[0] / static_cast<float>(count)) - offset[0]) / scaleFactor[0]);
                    *ptrTmp++ = static_cast<T>(((accum[1] / static_cast<float>(count)) - offset[1]) / scaleFactor[1]);
                }
            }
        }
    }
}

template<typename T>
void createMaskBorderUnrolled3(Image& image, Image& mask, float* offset, float* scaleFactor)
{
    const int width = static_cast<int>(image.width());
    const unsigned int width1 = image.width() - 1;
    const unsigned int height1 = image.height() - 1;
    const unsigned int channels = image.channels();

    // Offset to find all neighbours around a texel
    const int offsets[] =
    {
        -width - 1,
        -width,
        -width + 1,
        -1,
        +1,
        width - 1,
        width,
        width + 1
    };

    struct float32x3
    {
        float a, b, c;
    };

    float accum[3];
    float32x3* accumReset = reinterpret_cast<float32x3*>(accum);

    unsigned int idOffset;
    unsigned int count = 0;

    unsigned char* ptrMask = static_cast<unsigned char*>(mask.data());
    T* ptrSrcData = static_cast<T*>(image.data());
    T* ptrDstData = static_cast<T*>(image.data());

    for (unsigned int y = 1; y < height1; ++y)
    {
        for (unsigned int x = 1; x < width1; ++x)
        {
            unsigned int id = y * static_cast<unsigned int>(width) + x;

            if (ptrMask[id] == 0)
            {
                *accumReset = { 0.f, 0.f, 0.f };
                count = 0;

                for (unsigned int i = 0; i < 8; ++i)
                {
                    idOffset = static_cast<unsigned int>(static_cast<int>(id) + offsets[i]);

                    if (ptrMask[idOffset] == 255)
                    {
                        accum[0] += static_cast<float>(ptrSrcData[idOffset * channels + 0]) * scaleFactor[0] + offset[0];
                        accum[1] += static_cast<float>(ptrSrcData[idOffset * channels + 1]) * scaleFactor[1] + offset[1];
                        accum[2] += static_cast<float>(ptrSrcData[idOffset * channels + 2]) * scaleFactor[2] + offset[2];

                        ++count;
                    }
                }

                if (count)
                {
                    const unsigned int idCh = id * channels;
                    T* ptrTmp = ptrDstData + idCh;

                    *ptrTmp++ = static_cast<T>(((accum[0] / static_cast<float>(count)) - offset[0]) / scaleFactor[0]);
                    *ptrTmp++ = static_cast<T>(((accum[1] / static_cast<float>(count)) - offset[1]) / scaleFactor[1]);
                    *ptrTmp++ = static_cast<T>(((accum[2] / static_cast<float>(count)) - offset[2]) / scaleFactor[2]);
                }
            }
        }
    }
}

template<typename T>
void createMaskBorderDefault(Image& image, Image& mask, float* offset, float* scaleFactor)
{
    const int width = static_cast<int>(image.width());
    const unsigned int width1 = image.width() - 1;
    const unsigned int height1 = image.height() - 1;
    const unsigned int channels = image.channels();

    // Offset to find all neighbours around a texel
    const int offsets[] =
    {
        -width - 1,
        -width,
        -width + 1,
        -1,
        +1,
        width - 1,
        width,
        width + 1
    };

    float* accum = new float[channels];
    unsigned int idOffset;
    unsigned int count = 0;

    unsigned char* ptrMask = static_cast<unsigned char*>(mask.data());
    T* ptrSrcData = static_cast<T*>(image.data());
    T* ptrDstData = static_cast<T*>(image.data());

    for (unsigned int y = 1; y < height1; ++y)
    {
        for (unsigned int x = 1; x < width1; ++x)
        {
            unsigned int id = y * static_cast<unsigned int>(width) + x;

            if (ptrMask[id] == 0)
            {
                for (unsigned int k = 0; k < channels; ++k)
                {
                    accum[k] = 0.f;
                }

                count = 0;

                for (unsigned int i = 0; i < 8; ++i)
                {
                    idOffset = static_cast<unsigned int>(static_cast<int>(id) + offsets[i]);

                    if (ptrMask[idOffset] == 255)
                    {
                        for (unsigned int k = 0; k < channels; ++k)
                        {
                            accum[k] += static_cast<float>(ptrSrcData[idOffset * channels + k]) * scaleFactor[k] + offset[k];
                        }

                        ++count;
                    }
                }

                if (count)
                {
                    const unsigned int idCh = id * channels;
                    T* ptrTmp = ptrDstData + idCh;

                    for (unsigned int k = 0; k < channels; ++k)
                    {
                        *ptrTmp++ = static_cast<T>(((accum[k] / static_cast<float>(count)) - offset[k]) / scaleFactor[k]);
                    }
                }
            }
        }
    }

    delete[] accum;
}

template<typename T>
void createMaskBorderTemplate(Image& image, Image& mask, float* offset, float* scaleFactor)
{
    switch (image.channels())
    {
    case 1:
        createMaskBorderUnrolled1<T>(image, mask, offset, scaleFactor);
        break;
    case 2:
        createMaskBorderUnrolled2<T>(image, mask, offset, scaleFactor);
        break;
    case 3:
        createMaskBorderUnrolled3<T>(image, mask, offset, scaleFactor);
        break;
    default:
        createMaskBorderDefault<T>(image, mask, offset, scaleFactor);
        break;
    }
}

void createMaskBorder(Image& image, Image& mask, float* offset, float* scaleFactor)
{
    if
        (image.dataType() == DataType::unsigned_byte)
    {
        createMaskBorderTemplate<unsigned char>(image, mask, offset, scaleFactor);
    }
    else if
        (image.dataType() == DataType::float32)
    {
        createMaskBorderTemplate<float>(image, mask, offset, scaleFactor);
    }
}
