// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <forward_list>
#include <algorithm>

#include "filter_logger.h"

#include "core/log.h"

namespace SEAScope
{

    /**
     *
     */
    FilterLogger::FilterLogger()
        : SimpleLogger(), minLevel(LogLevel::Info)
    {
    }

    /**
     *
     */
    FilterLogger::~FilterLogger()
    {
        this->allowedLoggers.clear();
    }

    /**
     *
     */
    static void explodeString(const std::string& input, const std::string& delim, std::forward_list<std::string>& output)
    {
        size_t start = 0;
        std::size_t end = input.find(delim);
        auto it = output.before_begin();

        while
            (std::string::npos != end)
        {
            it = output.insert_after(it, input.substr(start, end - start));
            start = end + delim.length();
            end = input.find(delim, start);
        }
        output.insert_after(it, input.substr(start, end));
    }

    /**
     *
     */
    static void implodeStrings(const std::forward_list<std::string>& elems, const std::string& delim, std::string& result)
    {
        std::forward_list<std::string>::const_iterator it;	//

        it = elems.begin();
        for
            (; it != elems.end(); ++it)
        {
            result += *it;
            if
                (std::next(it, 1) != elems.end())
            {
                result += delim;
            }
        }
    };

    /*
     * Parse logFilter, it can either be:
     * - empty
     * - loglevel
     * - loglevel:logger1,logger2,...,loggerN
     */
    bool FilterLogger::init(const std::string& logFilter)
    {
        if
            ("" == logFilter)
        {
            this->minLevel = LogLevel::Info;
            return true;
        }

        std::size_t colonPos = logFilter.find(":");
        if
            (std::string::npos == colonPos)
        {
            // No filter on loggers, only on level
            // If logFilter does not contain a supported level name, then the
            // default (Info) will be applied
            this->minLevel = Log::levelFromString(logFilter);
            return true;
        }

        std::forward_list<std::string> tokens;	//

        explodeString(logFilter, ":", tokens);

        this->minLevel = Log::levelFromString(tokens.front());
        tokens.pop_front();

        std::string loggers;
        implodeStrings(tokens, ":", loggers);
        tokens.clear();
        explodeString(loggers, ",", tokens);
        this->allowedLoggers.insert(tokens.begin(), tokens.end());

        return true;
    }

    LogLevel FilterLogger::getLevel() const
    {
        return this->minLevel;
    }

    /**
     *
     */
    bool FilterLogger::isMessageValid(const LogLevel& level, const string_view& logger)
    {
        if
            (level < this->minLevel)
        {
            return false;
        }

        if
            (this->allowedLoggers.empty())
        {
            // filter only applied to log level, not on logger name
            return true;
        }

        auto found = std::find(this->allowedLoggers.cbegin(), this->allowedLoggers.cend(), logger);
        return (this->allowedLoggers.end() != found);
    }

} // namespace SEAScope
