// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>
#include <iomanip>
#include <mutex>
#include <fstream>

#include "core/log.h"
#include "core/path.h"
#include "core/structs.h"
#include "core/api_in.h"
#include "core/api_out.h"
#include "api/api_generated.h"
#include "flatbuffers_granules_index.h"
#include "core/granule_metadata.h"

using namespace SEAScope;

/**
 *
 */
FlatbuffersGranulesIndex::FlatbuffersGranulesIndex()
: _idGenerator(1), _granuleAddedCallback(nullptr)
{

};

/**
 *
 */
FlatbuffersGranulesIndex::~FlatbuffersGranulesIndex()
{

};

/**
 *
 */
std::uint64_t FlatbuffersGranulesIndex::generateGranuleId()
{
	return this->_idGenerator++; // return current value, then increment
}

/**
 *
 */
bool FlatbuffersGranulesIndex::hasGranule(const std::string& dataId) const
{
	auto found = this->_ids.find(dataId);
	return this->_ids.end() != found;
};

/**
 *
 */
bool FlatbuffersGranulesIndex::hasGranule(std::uint64_t granuleId) const
{
	auto found = this->_data.find(granuleId);
	return this->_data.end() != found;
};

/**
 *
 */
bool FlatbuffersGranulesIndex::add(GranuleMetadata& info, bool silent)
{
    std::uint64_t localId;  //

    auto found = this->_ids.find(info.dataId());
    if
        (this->_ids.end() != found)
    {
        localId = this->_ids[info.dataId()];

        // Just add the new uri to the list
        IDFDescriptor idf;
        info.getView(0, idf);
        this->_data[localId].addIDFDescriptor(idf);
        /*
		std::sort( this->_data[localId].uris.begin()
				 , this->_data[localId].uris.end()
				 , [](const IDFDescriptor& i, const IDFDescriptor& j) -> bool {
			return i.resolution < j.resolution;
		});
        */
        return true;
    }

	if
		(0 < info.granuleId())
	{
		localId = info.granuleId();
	}
	else
	{
		localId = this->generateGranuleId();
		info.granuleId(localId);
	}

    info.sourceId(0); // default source id
    {  // Lock scope
        std::lock_guard<std::mutex> lock(this->_mapsMutex);
	    this->_ids[info.dataId()] = localId;
	    this->_data[localId] = info;
    }

	if
		(false == silent && nullptr != this->_granuleAddedCallback)
	{
		this->_granuleAddedCallback(localId, info.collectionId());
	}
    LOG( LogLevel::Debug, "GranulesIndex"
            , "Add granule {} {}/{}"
            , info.dataId(), info.sourceId(), info.granuleId());
	return true;
};

/**
 *
 */
bool FlatbuffersGranulesIndex::updateTimeCoverage(std::uint64_t granuleId, const timestamp_t start, const timestamp_t stop)
{
	auto found = this->_data.find(granuleId);
    if
        (this->_data.end() == found)
    {
        return false;
    }

    {  // Lock scope
        std::lock_guard<std::mutex> lock(this->_mapsMutex);
        found->second.timeCoverageStart(start);
        found->second.timeCoverageEnd(stop);
    }

    return true;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::remove(const std::string& dataId, bool /*silent*/)
{
    std::uint64_t localId;  //

    auto found = this->_ids.find(dataId);
    if
        (this->_ids.end() == found)
    {
        return false;
    }

    localId = found->second;
    this->_data.erase(localId);
    this->_ids.erase(found);
    return true;
};

/**
 *
 */
bool FlatbuffersGranulesIndex::remove(std::uint64_t localId, bool /*silent*/)
{
    auto found = std::find_if(this->_ids.begin(), this->_ids.end(),
                              [localId](const auto& item) {return item.second == localId;});
    if
        (this->_ids.end() == found)
    {
        return false;
    }

    this->_data.erase(localId);
    this->_ids.erase(found);
    return true;
};

/**
 *
 */
std::array<bool, 2> FlatbuffersGranulesIndex::adjacentGranules( const timestamp_t current
									, const timestamp_t minDelta
									, const std::set<std::uint32_t>& collections
									, GranuleMetadata& prev
									, GranuleMetadata& next
									) const
{
	timestamp_t before(0);	//
	timestamp_t after(0);	//
	timestamp_t central(0);	//
	bool prevFound(false);	//
	bool nextFound(false);	//
	timestamp_t pastDelta(0);	//
	timestamp_t futureDelta(0);	//
	timestamp_t delta(0);	//

	pastDelta = std::numeric_limits<uint64_t>::max();
	futureDelta = std::numeric_limits<uint64_t>::max();

	LOG(LogLevel::Debug, "GranulesIndex", "current: {}, minDelta: {}", current, minDelta);
	for
		(auto& it : this->_data)
	{
		LOG(LogLevel::Debug, "GranulesIndex", "Check granule {}", it.first);
		/* Check that granule belongs to one of the requested collections */
		auto found = std::find( collections.begin()
							  , collections.end()
							  , it.second.collectionId());
		if
			(collections.end() == found)
		{
            std::string uri;
            it.second.getViewURI(0, uri);
			LOG( LogLevel::Debug, "GranulesIndex"
					, "Granule skipped: wrong collection {}"
					, uri);
			continue;
		}

		central = static_cast<timestamp_t>(0.5 * static_cast<double>(it.second.timeCoverageStart()) + 0.5 * static_cast<double>(it.second.timeCoverageEnd()));
		LOG(LogLevel::Debug, "GranulesIndex", "{}: central is {}", it.first, central);
		if
			(central < current)
		{
			/* Granule in the past */
			delta = current - central;

			LOG(LogLevel::Debug, "GranulesIndex", "granule is a past candidate: {} {} {}", delta, minDelta, central - current);
			if
				(minDelta > delta)
			{
				LOG(LogLevel::Debug, "GranulesIndex", "Skipped 'coz minDelta not ok: {} {} {}", minDelta, central, current);
				continue;
			}
			if
				(  false == prevFound
				|| delta < pastDelta
				|| central > before)
			{
				pastDelta = delta;
				before = central;
				prev = it.second;
				prevFound = true;
			}
		}
		else if
			(central > current)
		{
			/* Granule in the future */
			delta = central - current;

			LOG(LogLevel::Debug, "GranulesIndex", "granule is a future candidate: {} {} {}", delta, minDelta, central - current);
			if
				(minDelta > delta)
			{
				LOG(LogLevel::Debug, "GranulesIndex", "Skipped 'coz minDelta not ok: {} {} {}", minDelta, central, current);
				continue;
			}

			if
				(  false == nextFound
				|| delta < futureDelta
				|| central < after)
			{
				futureDelta = delta;
				after = central;
				next = it.second;
				nextFound = true;
			}
		}
	}

	return std::array<bool, 2>({{ prevFound, nextFound }});
};

/**
 * current, searchLowerBound, searchUpperBound expressed in milliseconds, like
 * time coverage bounds for the indexed granules
 */
bool FlatbuffersGranulesIndex::closestGranuleBound( const timestamp_t current
                                                  , const timestamp_t searchLowerBound
                                                  , const timestamp_t searchUpperBound
									, const std::set<std::uint32_t>& collections
									, GranuleMetadata& closest
									) const
{
	bool closestFound;	//
	timestamp_t absDiff;	//
	timestamp_t delta;	//

	closestFound = false;
	delta = std::numeric_limits<uint64_t>::max();

    // A granule time coverage is [time_coverage_start, time_coverage_end[,
    // meaning that time_coverage_end is the first datetime after time_coverage_start for which the granule has no data

	for
		(auto& it : this->_data)
	{
        timestamp_t candidateStart {it.second.timeCoverageStart()};
        timestamp_t candidateEnd {it.second.timeCoverageEnd()};

		/* Check that granule belongs to one of the requested collections */
		auto found = std::find( collections.begin()
							  , collections.end()
							  , it.second.collectionId());
		if
			(collections.end() == found)
		{
            std::string uri {""};
            it.second.getViewURI(0, uri);
			LOG( LogLevel::Debug, "GranulesIndex"
					, "Granule skipped: wrong collection {}"
					, uri);
			continue;
		}

		if
			(  candidateStart <= current
			&& candidateEnd > current)
		{
			closest = it.second;
			return true;
		}

        // Look for granules whose time coverage bounds are within the search domain
        // The search domain is also expressed as [searchLowerBound, searchUpperBound[,
        // so return consider granules with searchLowerBound <= time_coverage_start < searchUpperBound
        // or with searchLowerBound < time_coverage_end <= searchUpperBound
		if
			(  candidateStart >= searchLowerBound
			&& candidateStart < searchUpperBound)
		{
			if
				(candidateStart > current)
			{
				absDiff = candidateStart - current;
			}
			else
			{
				absDiff = current - candidateStart;
			}

			if
				(absDiff < delta)
			{
				delta = absDiff;
				closest = it.second;
				closestFound = true;
			}
		}

        // Since the last datetime containing data is time_coverage_end - 1:
        // • searchLowerBound == time_coverage_end is not a valid result,
        //   therefore the correct criterium is searchLowerBound < time_coverage_end
        // • searchUpperBound == time_coverage_end is a valid result
        //   because the last acceptable datetime with data is searchUpperBound - 1 and the last datetime with data is time_coverage_end - 1
		if
			(  candidateEnd > searchLowerBound
			&& candidateEnd <= searchUpperBound)
		{
			if
				(candidateEnd > current)
			{
				absDiff = candidateEnd - current;
			}
			else
			{
				absDiff = current - candidateEnd;
			}

			if
				(absDiff < delta)
			{
				delta = absDiff;
				closest = it.second;
				closestFound = true;
			}
		}
	}

	return closestFound;
};

/**
 *
 */
bool FlatbuffersGranulesIndex::getCoverage(std::uint32_t collectionId, EventsCoverage& result) const
{
	timestamp_t d;	// Milliseconds
	timestamp_t end;	// Milliseconds
	std::uint16_t year; //
	std::uint8_t month;	//
	std::uint8_t day;	//
	time_t timestamp;	// Seconds
	std::tm* t;	//

    result.reset();

	for
		(auto& it : this->_data)
	{
        if
            (it.second.collectionId() != collectionId)
        {
			/* granule does not belong to a requested collection */
            continue;
        }

		d = it.second.timeCoverageStart();
		end = it.second.timeCoverageEnd();

		timestamp = static_cast<time_t>(0.001 * static_cast<double>(d));
		t = std::localtime(&timestamp);

        // Handle case wherein the granule has an empty time coverage, i.e. an
        // "instantaneous" granule
        if
            (d == end)
        {
			year = static_cast<std::uint16_t>(1900 + t->tm_year);
			month = static_cast<std::uint8_t>(t->tm_mon);
			day = static_cast<std::uint8_t>(t->tm_mday);

			//LOG(LogLevel::Debug, "GranulesIndex", "Coverage {} {} {}", year, month+1, day);
			if
				(result.store.end() == result.store.find(year))
			{
				result.store[year] = std::array<std::uint32_t, 12>();
			}
			result.store[year][month] |= 1 << (day - 1);

            // We are done with this granule, check the next one
            continue;
        }

        // Granule has data for the [startMs, endMs[ time range, so if
        // endMs corresponds to day D at 00:00:00.000 then D is NOT covered.
        // We iterate from the day to which startMs belongs to, setting the
        // time at 00:00:00.001 and add 24 hours each time we loop
		t->tm_hour = 0;
		t->tm_min = 0;
		t->tm_sec = 0;
		timestamp = std::mktime(t);
		d = static_cast<timestamp_t>(timestamp * static_cast<time_t>(1000) + 1);

		LOG(LogLevel::Debug, "GranulesIndex", "Original timestamp: {}", timestamp);
		LOG(LogLevel::Debug, " GranulesIndex", "Modified timestamp: {}", d);

        // d == end means that end is day D at 00:00:00.001, so the granule has
        // data between 00:00:00.000 and 00:00:00.001, so D must be included in
        // the coverage
		while
			(d <= end)
		{
			timestamp = static_cast<long>(0.001 * static_cast<double>(d));
			t = std::localtime(&timestamp);
			year = static_cast<std::uint16_t>(1900 + t->tm_year);
			month = static_cast<std::uint8_t>(t->tm_mon);
			day = static_cast<std::uint8_t>(t->tm_mday);

			//LOG(LogLevel::Debug, "GranulesIndex", "Coverage {} {} {}", year, month+1, day);
			if
				(result.store.end() == result.store.find(year))
			{
				result.store[year] = std::array<std::uint32_t, 12>();
			}
			result.store[year][month] |= 1 << (day - 1);
			d = d + 86400000;
		}
    }
	return true;
};

/**
 *
 */
bool FlatbuffersGranulesIndex::getCoverage(const std::set<std::uint32_t>& collections, EventsCoverage& result) const
{
	timestamp_t d;	// Milliseconds
	timestamp_t end;	// Milliseconds
	std::uint16_t year;
	std::uint8_t month;	//
	std::uint8_t day;	//
	time_t timestamp;	// Seconds
	std::tm* t;	//

    result.reset();

	for
		(auto& it : this->_data)
	{
        auto found = std::find( collections.begin()
                              , collections.end()
                              , it.second.collectionId());
        if
            (collections.end() == found)
        {
			/* granule does not belong to a requested collection */
            continue;
        }

		d = it.second.timeCoverageStart();
		timestamp = static_cast<time_t>(0.001 * static_cast<double>(d));
		LOG(LogLevel::Debug, "GranulesIndex", "Original timestamp: {}", timestamp);
		t = std::localtime(&timestamp);
		t->tm_hour = 0;
		t->tm_min = 0;
		/* Up to debate: do we consider granules to be available one day 
		if their end date is at 00:00? The following implementation excludes granules
		that do not spend at least one second the day.
		*/
		t->tm_sec = 1;
		timestamp = std::mktime(t);

		LOG(LogLevel::Debug, " GranulesIndex", "Modified timestamp: {}", 1000 * timestamp);

		d = static_cast<timestamp_t>(timestamp * static_cast<time_t>(1000));
		end = it.second.timeCoverageEnd();
		while
			(d <= end)
		{
			timestamp = static_cast<long>(0.001 * static_cast<double>(d));
			t = std::localtime(&timestamp);
			year = static_cast<std::uint16_t>(1900 + t->tm_year);
			month = static_cast<std::uint8_t>(t->tm_mon);
			day = static_cast<std::uint8_t>(t->tm_mday);

			//LOG(LogLevel::Debug, "GranulesIndex", "Collection coverage {}: {} {} {}", it.second.collectionId, year, month+1, day);
			if
				(result.store.end() == result.store.find(year))
			{
				result.store[year] = std::array<std::uint32_t, 12>();
			}
			result.store[year][month] |= 1 << (day - 1);
			d = d + 86400000;
		}
	}
	return true;
};

/**
 *
 */
bool FlatbuffersGranulesIndex::findGranules(const SearchCriteria& criteria, std::vector<GranuleMetadata>& result) const
{
    bool collectionFound(false);   //
    bool mustBeCurrent(false); //

    for
        (auto& it: this->_data)
    {
        collectionFound = false;
        for
            (const auto& collection : criteria.collections)
        {
            if
                (collection.first == it.second.collectionId())
            {
                mustBeCurrent = collection.second;
                collectionFound = true;
                break;
            }
        }

        if
            (false == collectionFound)
        {
            LOG( LogLevel::Debug, "GranulesIndex"
                    , "Granule skipped because it does not belong to a requested collection: {} {}"
                    , it.first, it.second.collectionId());
            continue;
        }

        timestamp_t candidateStart {it.second.timeCoverageStart()};
        timestamp_t candidateEnd {it.second.timeCoverageEnd()};
        if
            (  candidateStart < 1000 * criteria.stop
            && candidateEnd > 1000 * criteria.start)
        {
            if
                (  true == mustBeCurrent
                && ( 1000 * criteria.current < candidateStart
                   || 1000 * criteria.current >= candidateEnd))
            {
                // Skip
                LOG( LogLevel::Debug, "GranulesIndex"
                        , "Skipped 'coz not current");
                continue;
            }

            result.push_back(it.second);
        }
    }
	return 0 < result.size();
};

/**
 *
 */
std::tuple<bool, bool> FlatbuffersGranulesIndex::load(const std::string& path)
{
    flatbuffers::FlatBufferBuilder builder; //
    std::uint64_t localId {0u};  //
    std::streamsize headerSize; //
    std::streampos totalSize;   //
    std::streamsize bufferSize;   //
    char* memblock {nullptr}; //

    if
        (false == Path::fileExists(path))
    {
        LOG(LogLevel::Debug, "FlatbuffersGranulesIndex",
                 "Index file '{}' does not exist, nothing loaded", path);
        return {false, false};
    }

    std::ifstream fileHandler(path, std::ios::in|std::ios::binary|std::ios::ate);
    if
        (false == fileHandler.is_open())
    {
        LOG( LogLevel::Error, "IndexSerializer"
                , "Could not open serialized index {}"
                , path);
        return {false, false};
    }

    totalSize = fileHandler.tellg();
    headerSize = sizeof(std::uint64_t);

    fileHandler.seekg(0, std::ios::beg);
    fileHandler.read (reinterpret_cast<char*>(&bufferSize), headerSize);

    LOG(LogLevel::Debug, "IndexSerializer", "File size: {}", std::to_string(totalSize));
    LOG(LogLevel::Debug, "IndexSerializer", "Header size: {}", headerSize);
    LOG(LogLevel::Debug, "IndexSerializer", "Buffer size: {}", bufferSize);

    if
        (totalSize < headerSize + bufferSize)
    {
        LOG( LogLevel::Error, "IndexSerializer"
                , "Serialized index file {} does not contain a complete buffer"
                , path);
        fileHandler.close();
        return {false, false};
    }
    memblock = new char [static_cast<size_t>(bufferSize)];

    fileHandler.read(memblock, bufferSize);
    fileHandler.close();

    auto cmd = flatbuffers::GetRoot<SEAScope::API::SearchGranulesResponse>(memblock);
    auto granules = cmd->granules();
    auto granulesCount = granules->size();

    LOG(LogLevel::Debug, "IndexSerializer", "Loaded {} granules from the serialized index", granulesCount);
    for
        (flatbuffers::uoffset_t i = 0; i < granulesCount; ++i)
    {
        auto granule = granules->Get(i);
        localId = granule->granuleId();
        this->_data[localId];
        APIIn::granuleMetadata(granule, this->_data[localId]);
        this->_ids[this->_data[localId].dataId()] = localId;
    }
    this->_idGenerator = 1 + granulesCount;

    delete[] memblock;
    memblock = nullptr;

	return {true, true};
}

/**
 *
 */
bool FlatbuffersGranulesIndex::save(const std::string& path)
{
    std::uint8_t* resultBuffer; //
    std::streamsize resultSize;   //
    flatbuffers::FlatBufferBuilder builder; //
    std::vector<flatbuffers::Offset<SEAScope::API::GranuleMetadata>> granulesVector;    //
	bool ok;	//

	ok = false;

    for
        (const auto& it : this->_data)
    {
        const GranuleMetadata& itGranule = it.second;
		auto granule = APIOut::granuleMetadata(builder, itGranule, ok);
		if
			(false == ok)
		{
			LOG(LogLevel::Error, "IndexSerializer", "Failed to serialize granule metadata");
			return false;
		}
		granulesVector.push_back(granule);
    }

    auto granules = builder.CreateVector(granulesVector);
    auto result = SEAScope::API::CreateSearchGranulesResponse(builder, granules);
    builder.Finish(result);

    LOG(LogLevel::Debug, "IndexSerializer", "Result built");

    resultBuffer = builder.GetBufferPointer();
    resultSize = builder.GetSize();

    std::ofstream fileHandler(path, std::ios::out|std::ios::binary|std::ios::trunc);
    if
        (false == fileHandler.is_open())
    {
        LOG( LogLevel::Error, "IndexSerializer"
                , "Could not open serialized index {}"
                , path);
        return false;
    }
    fileHandler.write(reinterpret_cast<const char*>(&resultSize), sizeof(std::uint64_t));
    fileHandler.write(reinterpret_cast<const char*>(resultBuffer), resultSize);
    fileHandler.close();

    LOG(LogLevel::Debug, "IndexSerializer", "Saved index: {} bytes", resultSize);
    builder.Release();

	return true;
};

/**
 *
 */
bool FlatbuffersGranulesIndex::setGranuleAddedCallback(const std::function<void(std::uint64_t, std::uint32_t)>& callback)
{
    if
        (nullptr != this->_granuleAddedCallback)
    {
        return false;
    }

    this->_granuleAddedCallback = callback;
    return true;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::setGranuleIdGenerator(std::uint64_t value)
{
    if
        (0 == value)
    {
        LOG(LogLevel::Error, "SQLiteGranulesIndex",
                 "The minimum value for granule IDs is 1");
        return false;
    }
    this->_idGenerator = value;
    return true;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::startBulkOperation()
{
    // Do nothing, there is no bulk mechanism with Flatbuffers
    return true;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::stopBulkOperation()
{
    // Do nothing, there is no bulk mechanism with Flatbuffers
    return true;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::cancelBulkOperation()
{
    // Do nothing, there is no bulk mechanism with Flatbuffers
    return true;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::increaseBulkOpsCount()
{
    // Do nothing, there is no bulk mechanism with Flatbuffers
    return true;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::countGranules(std::size_t& result) const
{
    result = this->_ids.size();
    return true;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::setCollectionIdGenerator(std::uint32_t)
{
    LOG(LogLevel::Error, "FlatbuffersGranulesIndex",
             "This implementation of the granules index does not support collection-related operations");
    return false;
}

/**
 *
 */
std::uint32_t FlatbuffersGranulesIndex::generateCollectionId()
{
    LOG(LogLevel::Error, "FlatbuffersGranulesIndex",
             "This implementation of the granules index does not support collection-related operations");
    return 0u;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::addCollection(const std::string&, const std::uint32_t)
{
    LOG(LogLevel::Error, "FlatbuffersGranulesIndex",
             "This implementation of the granules index does not support collection-related operations");
    return false;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::removeCollection(const std::string&)
{
    LOG(LogLevel::Error, "FlatbuffersGranulesIndex",
             "This implementation of the granules index does not support collection-related operations");
    return false;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::removeCollection(std::uint32_t)
{
    LOG(LogLevel::Error, "FlatbuffersGranulesIndex",
             "This implementation of the granules index does not support collection-related operations");
    return false;
}

bool FlatbuffersGranulesIndex::listCollections(std::vector<std::pair<std::uint32_t, std::string> >&) const
{
    LOG(LogLevel::Error, "FlatbuffersGranulesIndex",
             "This implementation of the granules index does not support collection-related operations");
    return false;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::countCollections(std::size_t&) const
{
    LOG(LogLevel::Error, "FlatbuffersGranulesIndex",
             "This implementation of the granules index does not support collection-related operations");
    return false;
}

/**
 *
 */
bool FlatbuffersGranulesIndex::clear()
{
    this->_idGenerator = 1u;
    this->_data.clear();
    this->_ids.clear();
    return true;
}
