// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "glfw_events_dispatcher.h"
#include "core/log.h"
#include "vendors/_glfw3.h"

using namespace SEAScope;


/**
 *
 */
GLFWEventsDispatcher::GLFWEventsDispatcher()
: _frameBufferXRatio(1.0),
  _frameBufferYRatio(1.0)
{
}

/**
 *
 */
GLFWEventsDispatcher::~GLFWEventsDispatcher()
{
    this->_subscribers.clear();
}

void GLFWEventsDispatcher::setFrameBufferRatio(double x, double y)
{
    this->_frameBufferXRatio = x;
    this->_frameBufferYRatio = y;
}

/**
 *
 */
bool GLFWEventsDispatcher::handleKeyEvent(int key, int scancode, int action, int mods)
{
    bool pressed(false);    //

    LOG(LogLevel::Debug, "GLFWEventsDispatcher",
             "KEY EVENT |> Key: {}, Scancode: {}, Action: {}, Mods: {}",
             key, scancode, action, mods);

    InputEvent event;
    if
        (GLFW_PRESS == action)
    {
        event.setType(EventType::KEY_PRESS);
        pressed = true;
    }
    else if
        (GLFW_REPEAT == action)
    {
        event.setType(EventType::KEY_REPEAT);
        pressed = true;
    }
    else
    {
        event.setType(EventType::KEY_RELEASE);
    }
    event.setKeyScanCode(scancode);

    this->_inputDevicesState.setKeyPressed(scancode, pressed);

    this->dispatchEvent(event);

    return true;
}

/**
 *
 */
bool GLFWEventsDispatcher::handleCharEvent(unsigned int c, int mods)
{
    LOG(LogLevel::Debug, "GLFWEventsDispatcher", "Char: {}, Mods: {}",
             c, mods);
    return true;
}

/**
 *
 */
bool GLFWEventsDispatcher::handleMouseButtonEvent(int button, int action, int mods)
{
    bool pressed(false);

    LOG(LogLevel::Debug, "GLFWEventsDispatcher",
             "MBUTTON EVENT |> Button: {}, Action: {}, Mods: {}",
             button, action, mods);

    InputEvent event;
    if
        (GLFW_PRESS == action)
    {
        event.setType(EventType::MOUSE_BUTTON_PRESS);
        pressed = true;
    }
    else if
        (GLFW_REPEAT == action)
    {
        event.setType(EventType::MOUSE_BUTTON_REPEAT);
        pressed = true;
    }
    else
    {
        event.setType(EventType::MOUSE_BUTTON_RELEASE);
    }
    event.setButton(button);

    this->_inputDevicesState.setButtonPressed(button, pressed);

    this->dispatchEvent(event);

    // Also check if it is a click
    if
        (  (EventType::MOUSE_BUTTON_RELEASE == event.getType())
        && (false == this->_inputDevicesState.hasMouseMovedSinceButtonPress(button, this->_clickPixelsTolerance)))
    {
        InputEvent clickEvent;
        clickEvent.setType(EventType::MOUSE_BUTTON_CLICK);
        clickEvent.setButton(button);
        this->dispatchEvent(clickEvent);
    }

    return true;
}

/**
 *
 */
bool GLFWEventsDispatcher::handleMouseMoveEvent(double x, double y)
{
    LOG(LogLevel::Debug, "GLFWEventsDispatcher",
             "MMOVE EVENT |> X: {}, Y: {}", x, y);

	double dx(0.0);
	double dy(0.0);

	x = x * this->_frameBufferXRatio;
	y = y * this->_frameBufferYRatio;

	dx = x - this->_inputDevicesState.getMouseX();
	dy = y - this->_inputDevicesState.getMouseY();

	InputEvent event;
	event.setType(EventType::MOUSE_MOVE);
	event.setMouseDx(dx);
	event.setMouseDy(dy);

	this->_inputDevicesState.setMouseX(x);
	this->_inputDevicesState.setMouseY(y);

	this->dispatchEvent(event);

    return true;
}

/**
 *
 */
bool GLFWEventsDispatcher::handleMouseWheelEvent(double x, double y)
{
    LOG(LogLevel::Debug, "GLFWEventsDispatcher",
             "MWHEEL EVENT |> X: {}, Y: {}", x, y);

	InputEvent event;
	event.setType(EventType::MOUSE_WHEEL_CHANGE);
	event.setMouseWheelDx(x);
	event.setMouseWheelDy(y);

	this->dispatchEvent(event);

    return true;
}

/**
 *
 */
bool GLFWEventsDispatcher::handleResizeEvent(int fbWidth, int fbHeight, int width, int height)
{
    LOG(LogLevel::Debug, "GLFWEventsDispatcher",
             "RESIZE EVENT |> Width: {}, Height: {}", fbWidth, fbHeight);

    // Make sure mouse move events will be expressed in framebuffer coords
    this->setFrameBufferRatio(static_cast<double>(fbWidth / width),
                              static_cast<double>(fbHeight / height));

    this->_inputDevicesState.reset();

    InputEvent event;
    event.setType(EventType::RESIZE);
    event.setWidth(fbWidth);
    event.setHeight(fbHeight);

    this->dispatchEvent(event);

    return true;
}

/**
 *
 */
bool GLFWEventsDispatcher::handleFocusEvent(int focused)
{
    LOG(LogLevel::Debug, "GLFWEventsDispatcher",
             "FOCUS EVENT |> Focused: {}", focused);
    return true;
}

/**
 *
 */
bool GLFWEventsDispatcher::handleIconifyEvent(int iconified)
{
    LOG(LogLevel::Debug, "GLFWEventsDispatcher",
             "ICON EVENT |> Iconified: {}", iconified);
    return true;
}

/**
 *
 */
bool GLFWEventsDispatcher::handleDropEvent(int count, const char** filenames)
{
    LOG(LogLevel::Debug, "GLFWEventsDispatcher",
             "DROP EVENT |> Count: {}", count);
    for
        (int i = 0; i < count; ++i)
    {
        LOG(LogLevel::Debug, "GLFWEventsDispatcher",
                 "\t{}", filenames[i]);
    }
    return true;
}

/**
 *
 */
void GLFWEventsDispatcher::setPreemptEventsCallback(const std::function<void(EventHandler*, bool)>& callback)
{
    this->_preemptEventsCallback = callback;
}

/**
 *
 */
void GLFWEventsDispatcher::initBindings()
{
    this->_inputBindings.initKeyboardMapping();
}

/**
 *
 */
const InputBindings& GLFWEventsDispatcher::getInputBindings() const
{
    return this->_inputBindings;
}

/**
 *
 */
InputBindings& GLFWEventsDispatcher::getInputBindings()
{
    return this->_inputBindings;
}


/**
 *
 */
bool GLFWEventsDispatcher::dispatchEvent(const InputEvent& event)
{
    bool preemptEvents(false);

    preemptEvents = InputEventsDispatcher::dispatchEvent(event);

    if
        (nullptr != this->_preemptEventsCallback)
    {
        this->_preemptEventsCallback(this, preemptEvents);
    }

    return preemptEvents;
}

bool GLFWEventsDispatcher::handleCloseEvent()
{
    InputEvent event;
    event.setType(EventType::CLOSE);

    this->dispatchEvent(event);

    return true;
}
