// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <locale>
#include <algorithm>

#include "vendors/_glfw3.h"

#include "glfw_input_bindings.h"
#include "core/types.h"
#include "core/log.h"

using namespace SEAScope;


/**
 *
 */
GLFWInputBindings::GLFWInputBindings()
{

}

/**
 *
 */
GLFWInputBindings::~GLFWInputBindings()
{

}


/**
 *
 */
bool GLFWInputBindings::initKeyboardMapping()
{
    const char* keyNameChars(nullptr);   //
    std::string keyName("");    //
    int scanCode(-1);   //
    bool ok;    //

    ok = true;

    /* Minimal set of printable keys */
    for
        (const auto& keyCode : {GLFW_KEY_A,
                                GLFW_KEY_B,
                                GLFW_KEY_C,
                                GLFW_KEY_D,
                                GLFW_KEY_E,
                                GLFW_KEY_F,
                                GLFW_KEY_G,
                                GLFW_KEY_H,
                                GLFW_KEY_I,
                                GLFW_KEY_J,
                                GLFW_KEY_K,
                                GLFW_KEY_L,
                                GLFW_KEY_M,
                                GLFW_KEY_N,
                                GLFW_KEY_O,
                                GLFW_KEY_P,
                                GLFW_KEY_Q,
                                GLFW_KEY_R,
                                GLFW_KEY_S,
                                GLFW_KEY_T,
                                GLFW_KEY_U,
                                GLFW_KEY_V,
                                GLFW_KEY_W,
                                GLFW_KEY_X,
                                GLFW_KEY_Y,
                                GLFW_KEY_Z,
                                GLFW_KEY_0,
                                GLFW_KEY_1,
                                GLFW_KEY_2,
                                GLFW_KEY_3,
                                GLFW_KEY_4,
                                GLFW_KEY_5,
                                GLFW_KEY_6,
                                GLFW_KEY_7,
                                GLFW_KEY_8,
                                GLFW_KEY_9,
                                GLFW_KEY_APOSTROPHE,
                                GLFW_KEY_COMMA,
                                GLFW_KEY_MINUS,
                                GLFW_KEY_PERIOD,
                                GLFW_KEY_SLASH,
                                GLFW_KEY_SEMICOLON,
                                GLFW_KEY_EQUAL,
                                GLFW_KEY_LEFT_BRACKET,
                                GLFW_KEY_BACKSLASH,
                                GLFW_KEY_RIGHT_BRACKET,})
    {
        scanCode = glfwGetKeyScancode(keyCode);
        keyNameChars = glfwGetKeyName(GLFW_KEY_UNKNOWN, scanCode);
        if
            (nullptr == keyNameChars)
        {
            LOG(LogLevel::Error, "GLFWInputBindings", "No name found for a key that should be printable");
            LOG(LogLevel::Debug, "GLFWInputBindings", "");
            ok = false;
            break;
        }
        keyName = std::string(keyNameChars);
        std::for_each(keyName.begin(), keyName.end(), [](char& c) {
            c = static_cast<char>(std::toupper(c));
        });
        ok = ok && (-1 < scanCode);
        this->_keyCodes[keyName] = scanCode;
    }

    scanCode = glfwGetKeyScancode(GLFW_KEY_SPACE);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["SPACE"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_GRAVE_ACCENT);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["GRAVE_ACCENT"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_ESCAPE);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["ESCAPE"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_ENTER);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["ENTER"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_TAB);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["TAB"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_BACKSPACE);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["BACKSPACE"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_INSERT);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["INSERT"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_DELETE);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["DELETE"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_RIGHT);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["RIGHT"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_LEFT);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["LEFT"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_DOWN);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["DOWN"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_UP);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["UP"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_PAGE_UP);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["PAGE_UP"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_PAGE_DOWN);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["PAGE_DOWN"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_HOME);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["HOME"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_END);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["END"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_CAPS_LOCK);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["CAPS_LOCK"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_SCROLL_LOCK);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["SCROLL_LOCK"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_NUM_LOCK);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["NUM_LOCK"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_PRINT_SCREEN);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["PRINT_SCREEN"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_PAUSE);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["PAUSE"] = scanCode;

    /* Function keys */
    scanCode = glfwGetKeyScancode(GLFW_KEY_F1);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F1"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F2);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F2"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F3);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F3"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F4);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F4"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F5);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F5"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F6);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F6"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F7);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F7"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F8);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F8"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F9);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F9"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F10);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F10"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F11);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F11"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F12);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F12"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F13);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F13"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F14);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F14"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F15);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F15"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F16);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F16"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F17);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F17"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F18);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F18"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F19);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F19"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F20);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F20"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F21);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F21"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F22);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F22"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F23);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F23"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F24);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F24"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_F25);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["F25"] = scanCode;

    /* Keypad */
    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_0);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD0"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_1);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD1"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_2);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD2"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_3);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD3"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_4);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD4"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_5);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD5"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_6);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD6"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_7);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD7"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_8);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD8"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_9);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD9"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_DECIMAL);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD_DECIMAL"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_DIVIDE);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD_DIVIDE"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_MULTIPLY);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD_MULTIPLY"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_SUBTRACT);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD_SUBTRACT"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_ADD);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD_ADD"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_EQUAL);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD_EQUAL"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_KP_ENTER);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["KEYPAD_ENTER"] = scanCode;

    /* Mod keys */
    scanCode = glfwGetKeyScancode(GLFW_KEY_LEFT_SHIFT);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["LEFT_SHIFT"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_LEFT_CONTROL);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["LEFT_CONTROL"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_LEFT_ALT);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["LEFT_ALT"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_LEFT_SUPER);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["LEFT_SUPER"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_RIGHT_SHIFT);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["RIGHT_SHIFT"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_RIGHT_CONTROL);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["RIGHT_CONTROL"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_RIGHT_ALT);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["RIGHT_ALT"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_RIGHT_SUPER);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["RIGHT_SUPER"] = scanCode;

    scanCode = glfwGetKeyScancode(GLFW_KEY_MENU);
    ok = ok && (-1 < scanCode);
    this->_keyCodes["MENU"] = scanCode;

    return ok;
}


