// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_glfw3.h"

#include "core/log.h"
#include "glfw_window.h"
#include "data/base64_image.h"
#include "data/app_icons.h"

using namespace SEAScope;

/**
 *
 */
static void glfwErrorHandler(int error, const char* descr)
{
	LOG( LogLevel::Error, "GLFWWindow"
			, "[{}] {}"
			, error
			, descr);
}

/**
 * Constructor.
 */
GLFWWindow::GLFWWindow()
: implementation{nullptr}, interval{0}, fullscreen{false}, frameBufferWidth{0},
  frameBufferHeight{0}, windowWidth{0}, windowHeight{0}, cursor{nullptr},
  _preempter{nullptr}, _frames{0}, _prevTime{0.0}
{
};

/**
 * Destructor.
 */
GLFWWindow::~GLFWWindow()
{
    this->_preempter = nullptr; /* not owning */

    if
        (nullptr != this->cursor)
    {
        glfwDestroyCursor(this->cursor);
        this->cursor = nullptr;
    }

    if
        (nullptr != this->implementation)
    {
	    glfwDestroyWindow(this->implementation);
        this->implementation = nullptr;
    }

    glfwTerminate();

    LOG(LogLevel::Debug, "GLFWWindow", "terminated");
};

/**
 *
 */
bool GLFWWindow::addRenderer(IRenderer& renderer)
{
	this->renderersIt = this->renderers.insert_after(this->renderersIt, &renderer);
	return true;
};

/**
 *
 */
bool GLFWWindow::addHandler(EventHandler& handler)
{
    handler.setPreemptEventsCallback([this] (EventHandler* preempter, bool active) {
        if
            (true == active)
        {
            this->_preempter = preempter;
        }
        else if
            (this->_preempter == preempter)
        {
            this->_preempter = nullptr;
        }
        else if
            (nullptr != this->_preempter)
        {
            LOG(LogLevel::Debug, "GLFWWindow", "Disabling a preempter but another one has already been activated");
        }
    });

	this->handlersIt = this->handlers.insert_after(this->handlersIt, &handler);
	return true;
};

/**
 *
 */
bool GLFWWindow::init(const Settings& settings)
{
	int initOk	//
	;

	/* Init GLFW */
	initOk = glfwInit();
	if
		(GLFW_FALSE == initOk)
	{
		LOG( LogLevel::Error, "application"
				, "Failed to initialize GLFW");
		return false;
	}

	glfwSetErrorCallback(glfwErrorHandler);

    if ("OpenGLES2" == settings.glRenderer)
    {
        glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_ES_API);
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 2);
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);
    }
    else // OpenGL3
    {
        glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_API);
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, settings.glContextMajor);
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, settings.glContextMinor);
        glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GLFW_TRUE);
        glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

#if defined(__APPLE__)
        // TRUE to avoid this error :
        // "NSGL: The targeted version of macOS only supports forward-compatible core profile contexts for OpenGL 3.2 and above"
        glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GLFW_TRUE); 
#else
        // Must be false to avoid strange OpenGL errors and unlock line thickness > 1
        glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GLFW_FALSE);
#endif
    }

	if
		(  true == settings.antialiasing
		&& "msaa" == settings.antialiasingType)
	{
		LOG( LogLevel::Debug, "GLFWWindow"
				, "Enable MSAA antialiasing with factor {}"
				, settings.antialiasingFactor);
		glfwWindowHint(GLFW_SAMPLES, settings.antialiasingFactor);
	}

	this->fullscreen = settings.fullscreen;
	this->interval = (true == settings.vsync) ? 1:0;
	return true;
};

/**
 *
 */
bool GLFWWindow::create( uint16_t width, uint16_t height)
{
	GLFWmonitor* monitor(nullptr);	//

    if
		(true == this->fullscreen)
	{
        monitor = glfwGetPrimaryMonitor();
        if
            (nullptr == monitor)
        {
            LOG(LogLevel::Error, "GLFWWindow", "Could not detect any monitor");
            return false;
        }

        const GLFWvidmode *mode = glfwGetVideoMode(monitor);
        this->implementation = glfwCreateWindow(mode->width, mode->height,
                                       "SEAScope", monitor, nullptr);
    }
	else
	{
        this->implementation = glfwCreateWindow(width, height, "SEAScope",
                                       nullptr, nullptr);
    }

	if
		(nullptr == this->implementation)
	{
		LOG( LogLevel::Error, "GLFWWindow"
				, "Failed to create window");
		glfwTerminate();
		return false;
	}

    GLFWimage images[16];

    Base64Image icon16x16;
    icon16x16.load(SEASCOPE_ICON_16x16);
    images[0].width = static_cast<int>(icon16x16.getImage().width());
    images[0].height = static_cast<int>(icon16x16.getImage().height());
    images[0].pixels = reinterpret_cast<unsigned char*>(icon16x16.getImage().data());

    Base64Image icon20x20;
    icon20x20.load(SEASCOPE_ICON_20x20);
    images[1].width = static_cast<int>(icon20x20.getImage().width());
    images[1].height = static_cast<int>(icon20x20.getImage().height());
    images[1].pixels = reinterpret_cast<unsigned char*>(icon20x20.getImage().data());

    Base64Image icon24x24;
    icon24x24.load(SEASCOPE_ICON_24x24);
    images[2].width = static_cast<int>(icon24x24.getImage().width());
    images[2].height = static_cast<int>(icon24x24.getImage().height());
    images[2].pixels = reinterpret_cast<unsigned char*>(icon24x24.getImage().data());

    Base64Image icon28x28;
    icon28x28.load(SEASCOPE_ICON_28x28);
    images[3].width = static_cast<int>(icon28x28.getImage().width());
    images[3].height = static_cast<int>(icon28x28.getImage().height());
    images[3].pixels = reinterpret_cast<unsigned char*>(icon28x28.getImage().data());

    Base64Image icon30x30;
    icon30x30.load(SEASCOPE_ICON_30x30);
    images[4].width = static_cast<int>(icon30x30.getImage().width());
    images[4].height = static_cast<int>(icon30x30.getImage().height());
    images[4].pixels = reinterpret_cast<unsigned char*>(icon30x30.getImage().data());

    Base64Image icon31x31;
    icon31x31.load(SEASCOPE_ICON_31x31);
    images[5].width = static_cast<int>(icon31x31.getImage().width());
    images[5].height = static_cast<int>(icon31x31.getImage().height());
    images[5].pixels = reinterpret_cast<unsigned char*>(icon31x31.getImage().data());

    Base64Image icon32x32;
    icon32x32.load(SEASCOPE_ICON_32x32);
    images[6].width = static_cast<int>(icon32x32.getImage().width());
    images[6].height = static_cast<int>(icon32x32.getImage().height());
    images[6].pixels = reinterpret_cast<unsigned char*>(icon32x32.getImage().data());

    Base64Image icon40x40;
    icon40x40.load(SEASCOPE_ICON_40x40);
    images[7].width = static_cast<int>(icon40x40.getImage().width());
    images[7].height = static_cast<int>(icon40x40.getImage().height());
    images[7].pixels = reinterpret_cast<unsigned char*>(icon40x40.getImage().data());

    Base64Image icon42x42;
    icon42x42.load(SEASCOPE_ICON_42x42);
    images[8].width = static_cast<int>(icon42x42.getImage().width());
    images[8].height = static_cast<int>(icon42x42.getImage().height());
    images[8].pixels = reinterpret_cast<unsigned char*>(icon42x42.getImage().data());

    Base64Image icon47x47;
    icon47x47.load(SEASCOPE_ICON_47x47);
    images[9].width = static_cast<int>(icon47x47.getImage().width());
    images[9].height = static_cast<int>(icon47x47.getImage().height());
    images[9].pixels = reinterpret_cast<unsigned char*>(icon47x47.getImage().data());

    Base64Image icon48x48;
    icon48x48.load(SEASCOPE_ICON_48x48);
    images[10].width = static_cast<int>(icon48x48.getImage().width());
    images[10].height = static_cast<int>(icon48x48.getImage().height());
    images[10].pixels = reinterpret_cast<unsigned char*>(icon48x48.getImage().data());

    Base64Image icon56x56;
    icon56x56.load(SEASCOPE_ICON_56x56);
    images[11].width = static_cast<int>(icon56x56.getImage().width());
    images[11].height = static_cast<int>(icon56x56.getImage().height());
    images[11].pixels = reinterpret_cast<unsigned char*>(icon56x56.getImage().data());

    Base64Image icon60x60;
    icon60x60.load(SEASCOPE_ICON_60x60);
    images[12].width = static_cast<int>(icon60x60.getImage().width());
    images[12].height = static_cast<int>(icon60x60.getImage().height());
    images[12].pixels = reinterpret_cast<unsigned char*>(icon60x60.getImage().data());

    Base64Image icon63x63;
    icon63x63.load(SEASCOPE_ICON_63x63);
    images[13].width = static_cast<int>(icon63x63.getImage().width());
    images[13].height = static_cast<int>(icon63x63.getImage().height());
    images[13].pixels = reinterpret_cast<unsigned char*>(icon63x63.getImage().data());

    Base64Image icon84x84;
    icon84x84.load(SEASCOPE_ICON_84x84);
    images[14].width = static_cast<int>(icon84x84.getImage().width());
    images[14].height = static_cast<int>(icon84x84.getImage().height());
    images[14].pixels = reinterpret_cast<unsigned char*>(icon84x84.getImage().data());

    Base64Image icon186x186;
    icon186x186.load(SEASCOPE_ICON_186x186);
    images[15].width = static_cast<int>(icon186x186.getImage().width());
    images[15].height = static_cast<int>(icon186x186.getImage().height());
    images[15].pixels = reinterpret_cast<unsigned char*>(icon186x186.getImage().data());

    glfwSetWindowIcon(this->implementation, 16, images);

	glfwSetWindowUserPointer(this->implementation, this);

	/* Set key press handler */
	glfwSetKeyCallback( this->implementation
					  , [](GLFWwindow* w, int key, int scancode, int action, int mods)
						{
							GLFWWindow* window;
							window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
							window->injectKey(key,scancode, action, mods);
						});

	/* Set char input handler */
	glfwSetCharModsCallback( this->implementation
					   , [](GLFWwindow* w, unsigned int c, int mods)
						 {
							GLFWWindow* window;
							window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
							window->injectChar(c, mods);
						 });

	/* Set mouse button handler */
	glfwSetMouseButtonCallback( this->implementation
							  , [](GLFWwindow* w, int button, int action, int mods)
								{
									GLFWWindow* window;
									window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
									window->injectMouseButton(button, action, mods);
								});

	/* Set mouse move handler */
	glfwSetCursorPosCallback( this->implementation
							, [](GLFWwindow* w, double x, double y)
							  {
					            GLFWWindow* window;
								window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
								window->injectMouseMove(x,y);
							  });

	/* Set mouse wheel callback */
	glfwSetScrollCallback( this->implementation
						 , [](GLFWwindow* w, double x, double y)
						   {
						        GLFWWindow* window;
								window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
								window->injectMouseWheel(x,y);
						   });

	/* Set mouse drop callback */
	glfwSetDropCallback( this->implementation
					   , [](GLFWwindow* w, int count, const char** filenames)
						 {
							GLFWWindow* window;
							window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
							window->injectDrop(count, filenames);
						 });

	/* Set resize callback */
	glfwSetFramebufferSizeCallback( this->implementation
							      , [](GLFWwindow* w, int widthIn, int heightIn)
						            {
                                        if
                                            (0 != widthIn && 0 != heightIn)
                                        {
                                            GLFWWindow* window;
                                            window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
                                            window->injectResize(widthIn, heightIn);
                                        }
						            });

	/* Set focus callback */
	glfwSetWindowFocusCallback( this->implementation
							  , [](GLFWwindow* w, int focused)
								{
									GLFWWindow* window;
									window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
									window->injectFocus(focused);
								});

	/* Set iconify callback */
	glfwSetWindowIconifyCallback( this->implementation
								, [](GLFWwindow* w, int iconified)
								  {
									GLFWWindow* window;
									window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
									window->injectIconify(iconified);
								  });

    /* Set iconify callback */
    glfwSetWindowCloseCallback( this->implementation
                              , [](GLFWwindow* w)
                                {
                                    GLFWWindow* window;
                                    window = static_cast<GLFWWindow*>(glfwGetWindowUserPointer(w));
                                    window->injectClose();
                                });

	glfwMakeContextCurrent(this->implementation);
	glfwSetInputMode(this->implementation, GLFW_CURSOR, GLFW_CURSOR_NORMAL);

	this->getSize(this->windowWidth, this->windowHeight);
	this->getFrameBufferSize(this->frameBufferWidth, this->frameBufferHeight);

	glfwSwapInterval(this->interval);
	glfwSetTime(0);

	return true;
};

/**
 *
 */
bool GLFWWindow::renderFrame()
{
	double time;				//
	bool ok;					//

	ok = !glfwWindowShouldClose(this->implementation);
	if
		(false == ok)
	{
		return false;
	}

	time = glfwGetTime();
	if
		(time - this->_prevTime >= 4.0)
	{
		this->_prevTime = time;
		this->_frames = 0;
	}

	ok = this->renderContent();

	glfwSwapBuffers(this->implementation);
	glfwPollEvents();

	this->_frames = this->_frames + 1;

	return ok;
};

/**
 *
 */
bool GLFWWindow::renderContent()
{
	std::forward_list<IRenderer*>::iterator it;	//
	bool ok;	//

	for
		(it = this->renderers.begin(); it != this->renderers.end(); ++it)
	{
		(*it)->renderFrame();
	}

	ok = !glfwWindowShouldClose(this->implementation);
	return ok;
};

/**
 *
 */
bool GLFWWindow::resize(uint16_t /*width*/, uint16_t /*height*/)
{
	return false;
};

/**
 *
 */
bool GLFWWindow::clean()
{
    this->handlers.clear();
    this->renderers.clear();
	return true;
};

/**
 *
 */
bool GLFWWindow::getSize(int& width, int& height) const
{
	glfwGetWindowSize(this->implementation, &width, &height);
	return true;
};

/**
 *
 */
bool GLFWWindow::getFrameBufferSize(int& width, int& height) const
{
    glfwGetFramebufferSize(this->implementation, &width, &height);
    return true;
};

/**
 *
 */
void GLFWWindow::injectKey(int key, int scancode, int action, int mods)
{
	bool captured;	//
	LOG( LogLevel::Debug, "GLFWWindow"
			, "Key event {} {} {} {}"
			, key, scancode, action, mods);

	captured = false;

    if
        (nullptr != this->_preempter)
    {
        captured = this->_preempter->handleKeyEvent(key, scancode, action, mods);
    }

	std::forward_list<EventHandler*>::iterator it;	//

	for
		(it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
	{
        if
            ((*it) != this->_preempter)
        {
		    captured = (*it)->handleKeyEvent(key, scancode, action, mods);
        }
	}
};

/**
 *
 */
void GLFWWindow::injectChar(unsigned int c, int mods)
{
	bool captured;	//
	LOG( LogLevel::Debug, "GLFWWindow"
			, "Char event {}"
			, c);

	captured = false;

    if
        (nullptr != this->_preempter)
    {
        captured = this->_preempter->handleCharEvent(c, mods);
    }

	std::forward_list<EventHandler*>::iterator it;	//
	for
		(it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
	{
        if
            ((*it) != this->_preempter)
        {
            captured = (*it)->handleCharEvent(c, mods);
        }
	}
};

/**
 *
 */
void GLFWWindow::injectMouseButton(int button, int action, int mods)
{
	bool captured;	//

	LOG( LogLevel::Debug, "GLFWWindow"
			, "Mouse button event {} {} {}"
			, button, action, mods);

	captured = false;

    if
        (nullptr != this->_preempter)
    {
        captured = this->_preempter->handleMouseButtonEvent(button, action, mods);
    }

	std::forward_list<EventHandler*>::iterator it;	//
	for
		(it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
	{
        if
            ((*it) != this->_preempter)
        {
		    captured = (*it)->handleMouseButtonEvent(button, action, mods);
        }
	}
};

/**
 *
 */
void GLFWWindow::injectMouseMove(double x, double y)
{
	bool captured;	//

	captured = false;

    if
        (nullptr != this->_preempter)
    {
        captured = this->_preempter->handleMouseMoveEvent(x, y);
    }

	std::forward_list<EventHandler*>::iterator it;	//
	for
		(it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
	{
        if
            ((*it) != this->_preempter)
        {
            captured = (*it)->handleMouseMoveEvent(x, y);
        }
	}
};

/**
 *
 */
void GLFWWindow::injectMouseWheel(double x, double y)
{
	bool captured;	//

	captured = false;

    if
        (nullptr != this->_preempter)
    {
        captured = this->_preempter->handleMouseWheelEvent(x, y);
    }

	std::forward_list<EventHandler*>::iterator it;	//
	for
		(it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
	{
        if
            ((*it) != this->_preempter)
        {
		    captured = (*it)->handleMouseWheelEvent(x, y);
        }
	}
};

/**
 *
 */
void GLFWWindow::injectResize(int width, int height)
{
	bool captured;	//
	LOG( LogLevel::Debug, "GLFWWindow"
			, "Window resize event {} {}"
			, width, height);
	captured = false;

	this->getSize(this->windowWidth, this->windowHeight);
	this->getFrameBufferSize(this->frameBufferWidth, this->frameBufferHeight);

    if
        (nullptr != this->_preempter)
    {
        captured = this->_preempter->handleResizeEvent(this->frameBufferWidth,
                                                       this->frameBufferHeight,
                                                       this->windowWidth,
                                                       this->windowHeight);
    }

	std::forward_list<EventHandler*>::iterator it;	//
	for
		(it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
	{
        if
            ((*it) != this->_preempter)
        {
		    captured = (*it)->handleResizeEvent(this->frameBufferWidth,
                                                this->frameBufferHeight,
                                                this->windowWidth,
                                                this->windowHeight);
        }
	}
};

/**
 *
 */
void GLFWWindow::injectFocus(int focused)
{
	bool captured;	//

	captured = false;

    if
        (nullptr != this->_preempter)
    {
        this->_preempter->handleFocusEvent(focused);
    }

	std::forward_list<EventHandler*>::iterator it;	//
	for
		(it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
	{
        if
            ((*it) != this->_preempter)
        {
		    captured = (*it)->handleFocusEvent(focused);
        }
	}
};

/**
 *
 */
void GLFWWindow::injectIconify(int iconified)
{
	bool captured;	//

	captured = false;

    if
        (nullptr != this->_preempter)
    {
        captured = this->_preempter->handleIconifyEvent(iconified);
    }

	std::forward_list<EventHandler*>::iterator it;	//
	for
		(it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
	{
        if
            ((*it) != this->_preempter)
        {
		    captured = (*it)->handleIconifyEvent(iconified);
        }
	}
};

/**
 *
 */
void GLFWWindow::injectDrop(int count, const char** filenames)
{
	bool captured;	//

	captured = false;

    if
        (nullptr != this->_preempter)
    {
        captured = this->_preempter->handleDropEvent(count, filenames);
    }

	std::forward_list<EventHandler*>::iterator it;	//
	for
		(it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
	{
        if
            ((*it) != this->_preempter)
        {
		    captured = (*it)->handleDropEvent(count, filenames);
        }
	}
};

void GLFWWindow::injectClose()
{
    bool captured;	//

    captured = false;

    if
        (nullptr != this->_preempter)
    {
        captured = this->_preempter->handleCloseEvent();
    }

    std::forward_list<EventHandler*>::iterator it;	//
    for
        (it = this->handlers.begin(); false == captured && it != this->handlers.end(); ++it)
    {
        if
            ((*it) != this->_preempter)
        {
            captured = (*it)->handleCloseEvent();
        }
    }
}

/**
 *
 */
bool GLFWWindow::getCursorPos(double& x, double& y) const
{
    glfwGetCursorPos(this->implementation, &x, &y);
    return true;
};

/**
 *
 */
bool GLFWWindow::setCursor(IWindow::MouseCursor cursorType)
{
    if
        (nullptr != this->cursor)
    {
        glfwDestroyCursor(this->cursor);
    }

    switch
        (cursorType)
    {
        case IWindow::MouseCursor::ARROW:
            this->cursor = glfwCreateStandardCursor(GLFW_ARROW_CURSOR);
            break;
        case IWindow::MouseCursor::CROSSHAIR:
            this->cursor = glfwCreateStandardCursor(GLFW_CROSSHAIR_CURSOR);
            break;
    }

    if
        (nullptr == this->cursor)
    {
        LOG(LogLevel::Error, "GLFWWindow", "NULL cursor!!!");
        return false;
    }

    glfwSetCursor(this->implementation, this->cursor);
    return true;
}

/**
 *
 */
void GLFWWindow::close()
{
    glfwSetWindowShouldClose(this->implementation, true);
}
