// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <fstream>
#include <ctime>

#include "vendors/_mini.h"

#include "core/log.h"
#include "core/path.h"
#include "core/utils.h"
#include "ini_settings.h"
#include "core/path_tools.h"
#include "core/timeline_cfg.h"

using namespace SEAScope;

/**
 * Constructor.
 */
INISettings::INISettings()
: _defaultPathsFields({
    {"custom", ""},
}),
  _defaultRenderingFields({
    {"glRenderer", "OpenGL3"},
    {"glContextMajor", "3"},
    {"glContextMinor", "3"},
    {"windowWidth", "800"},
    {"windowHeight", "600"},
    {"antialiasing", "false"},
    {"antialiasingType", "msaa"},
    {"antialiasingFactor", "8"},
    {"fullscreen", "false"},
    {"vsync", "true"},
    {"minGranuleScreenEstate", "0.008"}
}),
  _defaultTimelineFields({
    {"minYear", "1988"},
    {"maxYear", ""},
    {"timespans", "6h|6 hours timespan|10800|10800,1d|1 day timespan|43200|43200,3d|3 days timespan|129600|129600,1w|1 week timespan|302400|302400"},
    {"timesteps", "6h|Move by 6 hours|21600,Day|Move by 24 hours|86400,Week|Move by 7 days|604800,Data|Move to previous/next available data|0"},
    {"defaultTimespan", "1d"},
    {"defaultTimestep", Timestep::ONE_THIRD_TIMESPAN_LABEL},
    {"useNearestMode", "true"},
    {"advancedScrolling", "false"},
    {"initialDatetime", ""}
}),
  _defaultMiscFields({
    {"worldMap", "lowres"},
    {"skipIndexCheck", "false"},
    {"searchRequestMinDelay", "0.1"}
}),
  _defaultProcessorFields({
    {"enabled", "true"},
    {"address", "127.0.0.1"},
    {"port", "53450"}
}),
  _defaultDrawingFields({
    {"clickPixelsTolerance", "7"},
    {"explicitDrawingModeExit", "false"}
})
{
};

/**
 * Destructor.
 */
INISettings::~INISettings()
{

};

/**
 * Load settings from INI configuration file.
 */
bool INISettings::load(Settings& result, std::string& configPath)
{
	bool ok;	//

	ok = this->loadINI(result, configPath);
	return ok;
};

/**
 *
 */
bool INISettings::splitList(const std::string& input, const std::string& delim, std::vector<std::string>& output) const
{
	size_t start = 0U;
	auto end = input.find(delim);

	while
		(end != std::string::npos)
	{
		output.push_back(input.substr(start, end - start));
		start = end + delim.length();
		end = input.find(delim, start);
	}
	output.push_back(input.substr(start, end));
	return 0 < output.size();
};

/**
 *
 */
bool INISettings::loadTimespans(Settings& result, const std::string& timespansStr)
{
	std::vector<std::string> timespans;

	this->splitList(timespansStr, ",", timespans);
	for
		(const auto& timespan : timespans)
	{
		LOG(LogLevel::Debug, "Settings INI", "Extracted timespan: {}" , timespan);
		std::vector<std::string> timespanComponents;
		this->splitList(timespan, "|", timespanComponents);
		if
			(4 > timespanComponents.size())
		{
			for
				(const auto& cmp : timespanComponents)
			{
				LOG(LogLevel::Warn, "Settings INI", "Incomplete timespan: {}" , cmp);
			}
			continue;
		}
		result.timelineCfg.addTimeframe( timespanComponents[0]
									   , timespanComponents[1]
									   , static_cast<timestamp_t>(std::atol(timespanComponents[2].c_str()))
									   , static_cast<timestamp_t>(std::atol(timespanComponents[3].c_str())));
	}

	return true;
};

/**
 *
 */
bool INISettings::loadTimesteps(Settings& result, const std::string& timestepsStr)
{
	std::vector<std::string> timesteps;

    result.timelineCfg.addTimestep(Timestep::ONE_THIRD_TIMESPAN_LABEL,
                                   "Move by 1/3 of the selected timespan",
                                   Timestep::ONE_THIRD_TIMESPAN_VALUE);

	this->splitList(timestepsStr, ",", timesteps);
	for
		(const auto& timestep : timesteps)
	{
		LOG(LogLevel::Debug, "Settings INI", "Extracted timestep: {}" , timestep);
		std::vector<std::string> timestepComponents;
		this->splitList(timestep, "|", timestepComponents);
		if
			(3 > timestepComponents.size())
		{
			for
				(const auto& cmp : timestepComponents)
			{
				LOG(LogLevel::Warn, "Settings INI", "Incomplete timestep: {}" , cmp);
			}
			continue;
		}

        if
            (Timestep::ONE_THIRD_TIMESPAN_LABEL == timestepComponents[0])
        {
            LOG(LogLevel::Error, "Settings INI", "Timestep cannot have the \"{}\" label because it is a reserved keyword", Timestep::ONE_THIRD_TIMESPAN_LABEL);
            return false;
        }
		result.timelineCfg.addTimestep( timestepComponents[0]
									  , timestepComponents[1]
									  , static_cast<timestamp_t>(std::atol(timestepComponents[2].c_str())));
	}

	return true;
};

/**
 *
 */
static bool checkSection(const std::string& uri,
                         bool canBeFixed,
                         mINI::INIStructure& root,
                         const std::string& sectionName,
                         const std::vector<std::string>& required,
                         const std::unordered_map<std::string, std::string>& defaults,
                         std::vector<std::string>& missing,
                         bool& updateNeeded)
{
	if
		(  (false == root.has(sectionName))
        && ((false == canBeFixed) || (0 < required.size())))
	{
        // section is missing and config file cannot be fixed
        // or
        // section is missing and should contain mandatory fields
		LOG(LogLevel::Error, "Settings INI",
                 "[{}] section missing in {}", sectionName, uri);
		return false;
	}

    auto& section = root[sectionName];  // will create section if it does not exist
	for
		(const auto& field : required)
	{
		if
			(false == section.has(field))
		{
			missing.push_back(fmt::format("Field: [{}] {}", sectionName, field));
		}
	}

    for
        (const auto& defaultsIterator : defaults)
    {
        if
            (true == section.has(defaultsIterator.first))
        {
            continue;   // everything is ok
        }

        if
            (true == canBeFixed)
        {
            LOG(LogLevel::Info, "Settings INI",
                     "Adding missing [{}] {} field to {}",
                     sectionName, defaultsIterator.first, uri);
            section[defaultsIterator.first] = defaultsIterator.second;
            updateNeeded = true;
        }
    }

    return true;
}

/**
 *
 */
bool INISettings::validate(const std::string& uri, mINI::INIStructure& root)
{
    bool canBeFixed {false};    //
    bool updateNeeded {false};  //
    bool ok(false); //
    std::vector<std::string> missing;   //

    // Incomplete configuration files can only be fixed if they are writable
    canBeFixed = Path::writeFile(uri);

    mINI::INIFile parser(uri);
    ok = parser.read(root);
    if
        (false == ok)
    {
		LOG( LogLevel::Error, "Settings INI"
				, "Error during parsing of {}"
				, uri);
		return false;
	}

	// Check [paths] section
    ok = checkSection(uri, canBeFixed, root, "paths",
                      {"colormaps", "data", "state", "index", "annotations"},
                      this->_defaultPathsFields, missing, updateNeeded);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Settings INI",
                 "Mandatory [paths] setion is not defined in '{}'", uri);
        return false;
    }

    // plugins path is deprecated, remove it
    if
        (true == root["paths"].has("plugins"))
    {
        LOG(LogLevel::Info, "Settings INI",
                 "Removing deprecated [paths]plugins setting from {}", uri);
        root["paths"].remove("plugins");
        updateNeeded = true;
    }

    // operators path is deprecated, remove it
    if
        (true == root["paths"].has("operators"))
    {
        LOG(LogLevel::Info, "Settings INI",
                 "Removing deprecated [paths]operators setting from {}", uri);
        root["paths"].remove("operators");
        updateNeeded = true;
    }

    // [misc]baseLayer is deprecated, remove it
    if
        (true == root["misc"].has("baseLayer"))
    {
        LOG(LogLevel::Info, "Settings INI",
                 "Removing deprecated [misc]baseLayer setting from {}", uri);
        root["misc"].remove("baseLayer");
        updateNeeded = true;
    }

    // [misc]defaultColormap is deprecated, remove it
    if
        (true == root["misc"].has("defaultColormap"))
    {
        LOG(LogLevel::Info, "Settings INI",
                 "Removing deprecated [misc]defaultColormap setting from {}", uri);
        root["misc"].remove("defaultColormap");
        updateNeeded = true;
    }

    // [timeline]useNearestInDayMode is deprecated, remove it
    if
        (true == root["timeline"].has("useNearestInDayMode"))
    {
        LOG(LogLevel::Info, "Settings INI",
                 "Removing deprecated [timeline]useNearestInDayMode setting from {}", uri);
        root["timeline"].remove("useNearestInDayMode");
        updateNeeded = true;
    }

	// Check [rendering] section
    ok = checkSection(uri, canBeFixed, root, "rendering",
                      {},
                      this->_defaultRenderingFields, missing, updateNeeded);

	// Check [timeline] section
    ok = checkSection(uri, canBeFixed, root, "timeline",
                      {},
                      this->_defaultTimelineFields, missing, updateNeeded);

	// Check [misc] section
    ok = checkSection(uri, canBeFixed, root, "misc",
                      {},
                      this->_defaultMiscFields, missing, updateNeeded);

	// Check [processor] section
    ok = checkSection(uri, canBeFixed, root, "processor",
                      {},
                      this->_defaultProcessorFields, missing, updateNeeded);

    // Check [drawing] section
    ok = checkSection(uri, canBeFixed, root, "drawing",
                      {},
                      this->_defaultDrawingFields, missing, updateNeeded);

	// Return false if some fields are missing
	if
		(0 < missing.size())
	{
		LOG(LogLevel::Error, "Settings iNI",
				 "Configuration file {} is incomplete, the following elements have not been found:", uri);
		for
			(const auto& msg : missing)
		{
			LOG(LogLevel::Error, "Settings INI", msg);
		}
		return false;
	}

    if
        (true == updateNeeded)
    {
        ok = parser.write(root, true /*pretty print*/);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Settings INI",
                     "Could not write fixed configuration to {}", uri);
            return false;
        }
    }

    return true;
}

/**
 *
 */
bool INISettings::loadINI(Settings& result, const std::string& uri)
{
    bool ok(false); //
	mINI::INIStructure root;	//

    ok = validate(uri, root);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Settings INI",
                 "Validation of the {} configuration file failed", uri);
		return false;
	}

    /* paths */
    auto& paths = root["paths"];
    result.colormapsPath = paths["colormaps"];
    result.dataPath = paths["data"];
    result.statePath = paths["state"];
    result.indexPath = paths["index"];
    result.annotationsPath = paths["annotations"];
    result.customPath = paths["custom"];

    /* rendering */
    auto& rendering = root["rendering"];
    result.glRenderer = rendering["glRenderer"];
    result.glContextMajor = static_cast<uint8_t>(std::atoi(rendering["glContextMajor"].c_str()));
    result.glContextMinor = static_cast<uint8_t>(std::atoi(rendering["glContextMinor"].c_str()));
    result.windowWidth = static_cast<uint16_t>(std::atoi(rendering["windowWidth"].c_str()));
    result.windowHeight = static_cast<uint16_t>(std::atoi(rendering["windowHeight"].c_str()));
    result.vsync =  "true" == rendering["vsync"];
    result.antialiasing = "true" == rendering["antialiasing"];
    result.antialiasingType = rendering["antialiasingType"];
    result.antialiasingFactor = static_cast<uint8_t>(std::atoi(rendering["antialiasingFactor"].c_str()));
    result.fullscreen = "true" == rendering["fullscreen"];
    result.minGranuleScreenEstate = static_cast<float>(std::atof(rendering["minGranuleScreenEstate"].c_str()));

    /* timeline */
    auto& timeline = root["timeline"];
    result.timelineCfg.setMinYear(static_cast<std::uint16_t>(std::atoi(timeline["minYear"].c_str())));
    const std::string& maxYear = timeline["maxYear"];
    if
        (maxYear.empty())
    {
        std::time_t timestamp = time(nullptr);
        std::tm* now = std::gmtime(&timestamp);
        result.timelineCfg.setMaxYear(1900u + static_cast<std::uint16_t>(now->tm_year));
        result.timelineCfg.setUseAutoMaxYear(true);
    }
    else
    {
        result.timelineCfg.setMaxYear(static_cast<std::uint16_t>(std::atoi(maxYear.c_str())));
        result.timelineCfg.setUseAutoMaxYear(false);
    }

    if
        (result.timelineCfg.getMinYear() > result.timelineCfg.getMaxYear())
    {
        LOG(LogLevel::Error, "Settings INI",
            R"(minYear value ({}) must not be greater than maxYear value ({}))", result.timelineCfg.getMinYear(), result.timelineCfg.getMaxYear());
        return false;
    }

    this->loadTimespans(result, timeline["timespans"]);
    result.timelineCfg.defaultTimespan = timeline["defaultTimespan"];
    this->loadTimesteps(result, timeline["timesteps"]);
    result.timelineCfg.defaultTimestep = timeline["defaultTimestep"];
    result.timelineCfg.useNearestMode = "true" == timeline["useNearestMode"];
    result.timelineCfg.advancedScrolling = "true" == timeline["advancedScrolling"];
    const std::string& initialDatetime = timeline["initialDatetime"];
    timestamp_t timestamp(0);
    if
        (initialDatetime.empty())
    {
	timestamp = static_cast<timestamp_t>(std::time(nullptr));  // now
    }
    else
    {
        ok = SEAScope::getTimestampFromDatetime(initialDatetime, timestamp);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Settings INI",
                R"(Could not parse initial datetime value : "{}")", initialDatetime);

            // failed to parse initialDatetime
            return false;
        }
    }
    result.timelineCfg.setInitialDatetime(timestamp);
    result.timelineCfg.setInitialDatetimeString(initialDatetime);

    /* drawing */
    auto& drawing = root["drawing"];
    result.explicitDrawingModeExit  = "true" == drawing["explicitDrawingModeExit"];
    result.clickPixelsTolerance = static_cast<std::uint32_t>(std::atoi(drawing["clickPixelsTolerance"].c_str()));

    /* misc */
    auto& misc = root["misc"];
    result.worldMap = misc["worldMap"];
    result.skipIndexCheck = "true" == misc["skipIndexCheck"];

    result.searchRequestMinDelay = 0.1; // 100 ms
    if
        ("" != misc["searchRequestMinDelay"])
    {
        result.searchRequestMinDelay = std::atof(misc["searchRequestMinDelay"].c_str());
    }

    /* processor */
    auto& processor = root["processor"];
    result.processorEnabled = "true" == processor["enabled"];
    result.processorAddress = processor["address"];
    result.processorPort = std::atoll(processor["port"].c_str()); // assumes int64_t <=> long long

	return true;

};


/**
 * Save settings to file.
 */
bool INISettings::save(const Settings& settings, const std::string& configPath) const
{
	bool ok;	//

	ok = this->saveINI(settings, configPath);
	return ok;
};

/**
 *
 */
bool INISettings::saveINI(const Settings& settings, const std::string& uri) const
{
    mINI::INIStructure root;    //
	std::ostringstream tmp;	//
	std::ofstream output;		//

    root["paths"]["colormaps"] = settings.colormapsPath;
    root["paths"]["data"] = settings.dataPath;
    root["paths"]["state"] = settings.statePath;
    root["paths"]["index"] = settings.indexPath;
    root["paths"]["annotations"] = settings.annotationsPath;
    root["paths"]["custom"] = settings.customPath;

    root["rendering"]["glRenderer"] = settings.glRenderer;
    root["rendering"]["glContextMajor"] = std::to_string(settings.glContextMajor);
    root["rendering"]["glContextMinor"] = std::to_string(settings.glContextMinor);
    root["rendering"]["windowWidth"] = std::to_string(settings.windowWidth);
    root["rendering"]["windowHeight"] = std::to_string(settings.windowHeight);
    root["rendering"]["antialiasing"] = (settings.antialiasing ? "true":"false");
	root["rendering"]["antialiasingType"] = settings.antialiasingType;
	root["rendering"]["antialiasingFactor"] = std::to_string(settings.antialiasingFactor);
	root["rendering"]["fullscreen"] = (settings.fullscreen ? "true":"false");
	root["rendering"]["vsync"] = (settings.vsync ? "true":"false");
	root["rendering"]["minGranuleScreenEstate"] = std::to_string(settings.minGranuleScreenEstate);

    root["timeline"]["minYear"] = std::to_string(settings.timelineCfg.getMinYear());
    if
        (true == settings.timelineCfg.getUseAutoMaxYear())
    {
        root["timeline"]["maxYear"] = "";
    }
    else
    {
	    root["timeline"]["maxYear"] = std::to_string(settings.timelineCfg.getMaxYear());
    }

	/* Timespans */
	tmp.str("");
	for
		(const auto& timespan : settings.timelineCfg.timeframes)
	{
		tmp << timespan.label << "|" << timespan.description << "|" << timespan.pastOffset << "|" << timespan.futureOffset << ",";
	}
    tmp.seekp(-1, std::ios_base::end);
    tmp << "\n";
    root["timeline"]["timespans"] = tmp.str();

	/* Timesteps */
    tmp.str("");
	for
		(const auto& timestep : settings.timelineCfg.timesteps)
	{
        if
            (Timestep::ONE_THIRD_TIMESPAN_LABEL == timestep.label)
        {
            continue;
        }
		tmp << timestep.label << "|" << timestep.description << "|" << timestep.value << ",";
	}
    tmp.seekp(-1, std::ios_base::end);
    tmp << "\n";
    root["timeline"]["timesteps"] = tmp.str();

	root["timeline"]["defaultTimespan"] = settings.timelineCfg.defaultTimespan;
	root["timeline"]["defaultTimestep"] = settings.timelineCfg.defaultTimestep;
    root["timeline"]["useNearestMode"] = (settings.timelineCfg.useNearestMode ? "true":"false");
    root["timeline"]["initialDatetime"] = settings.timelineCfg.getInitialDatetimeString();
    root["timeline"]["advancedScrolling"] = (settings.timelineCfg.advancedScrolling ? "true":"false");

    root["drawing"]["explicitDrawingModeExit"] = (settings.explicitDrawingModeExit ? "true":"false");
    root["drawing"]["clickPixelsTolerance"] = std::to_string(settings.clickPixelsTolerance);

	root["misc"]["worldMap"] = settings.worldMap;
	root["misc"]["skipIndexCheck"] = (settings.skipIndexCheck ? "true":"false");
    root["misc"]["searchRequestMinDelay"] = std::to_string(settings.searchRequestMinDelay);

	root["processor"]["enabled"] = (settings.processorEnabled ? "true":"false");
	root["processor"]["address"] = settings.processorAddress;
	root["processor"]["port"] = std::to_string(settings.processorPort);

    bool isWritable(false);
    isWritable = Path::writeFile(uri);
	if
		(false == isWritable)
	{
		LOG( LogLevel::Error, "Settings INI"
				, "Could not open '{}' with write permissions"
				, uri);
		return false;
	}

    mINI::INIFile parser(uri);
    bool ok(false); //
    ok = parser.write(root, true /*pretty print*/);
    if
        (false == ok)
    {
		LOG( LogLevel::Error, "Settings INI"
				, "Could not save settings in '{}'"
			    , uri);
		return false;
	}

	return true;
};
