// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iterator>
#include <algorithm>
#include <sstream>

#include "vendors/_picojson.h"

#include "json_annotation.h"
#include "core/data_node.h"
#include "core/utils.h"
#include "data/annotation.h"
#include "3d/shapes_allocator.h"
#include "3d/unit_sphere.h"
#include "3d/geodetic.h"
#include "3d/polygon.h"
#include "3d/polyline.h"
#include "3d/ellipse.h"
#include "3d/point.h"
#include "3d/grid.h"
#include "core/measurements.h"
#include "core/log.h"

using namespace SEAScope;

/**
 *
 */
JSONAnnotation::JSONAnnotation()
: _base64Flag("data:image/png;base64,")
{
}

/**
 *
 */
JSONAnnotation::~JSONAnnotation()
{
}

bool JSONAnnotation::serializeNode(const DataNode* src, picojson::array& dst) const
{
    bool ok(false); //
    NodeType nodeType = src->getType();  //

    if
        (NodeType::empty == nodeType)
    {
        dst.push_back(picojson::value());
    }
    else if
        (NodeType::boolean == nodeType)
    {
        const BoolNode* boolNode = static_cast<const BoolNode*>(src);
        dst.push_back(picojson::value(boolNode->value));
    }
    else if
        (NodeType::number == nodeType)
    {
        const NumberNode* numNode = static_cast<const NumberNode*>(src);
        dst.push_back(picojson::value(numNode->value));
    }
    else if
        (NodeType::text == nodeType)
    {
        const TextNode* textNode = static_cast<const TextNode*>(src);
        dst.push_back(picojson::value(textNode->value));
    }
    else if
        (NodeType::image == nodeType)
    {
        const ImageNode* imageNode = static_cast<const ImageNode*>(src);
        dst.push_back(picojson::value(this->_base64Flag + imageNode->value));
    }
    else if
        (NodeType::list == nodeType)
    {
        const ListNode* listNode = static_cast<const ListNode*>(src);
        picojson::array list;
        for
            (const auto& it : listNode->children)
        {
            ok = this->serializeNode(it, list);
            if
                (false == ok)
            {
                return false;
            }
        }
        dst.push_back(picojson::value(list));
    }
    else if
        (NodeType::dict == nodeType)
    {
        const DictNode* dictNode = static_cast<const DictNode*>(src);
        picojson::object dict;
        for
            (const auto& it : dictNode->children)
        {
            ok = this->serializeNode(it.second, dict, it.first);
            if
                (false == ok)
            {
                return false;
            }
        }
        dst.push_back(picojson::value(dict));
    }
    else
    {
        return false;
    }

    return true;
}

/**
 *
 */
bool JSONAnnotation::serializeNode(const DataNode* src, picojson::object& dst, const std::string& key) const
{
    bool ok(false); //

    if
        (nullptr == src)
    {
        return false;
    }

    NodeType nodeType = src->getType();
    if
        (NodeType::empty == nodeType)
    {
        dst.emplace(std::make_pair(key, picojson::value()));
    }
    else if
        (NodeType::boolean == nodeType)
    {
        const BoolNode* boolNode = static_cast<const BoolNode*>(src);
        dst.emplace(std::make_pair(key, picojson::value(boolNode->value)));
    }
    else if
        (NodeType::number == nodeType)
    {
        const NumberNode* numNode = static_cast<const NumberNode*>(src);
        dst.emplace(std::make_pair(key, picojson::value(numNode->value)));
    }
    else if
        (NodeType::text == nodeType)
    {
        const TextNode* textNode = static_cast<const TextNode*>(src);
        dst.emplace(std::make_pair(key, picojson::value(textNode->value)));
    }
    else if
        (NodeType::image == nodeType)
    {
        const ImageNode* imageNode = static_cast<const ImageNode*>(src);
        dst.emplace(std::make_pair(key, picojson::value(this->_base64Flag + imageNode->value)));
    }
    else if
        (NodeType::list == nodeType)
    {
        const ListNode* listNode = static_cast<const ListNode*>(src);
        picojson::array list;
        for
            (const auto& it : listNode->children)
        {
            ok = this->serializeNode(it, list);
            if
                (false == ok)
            {
                return false;
            }
        }
        dst.emplace(std::make_pair(key, picojson::value(list)));
    }
    else if
        (NodeType::dict == nodeType)
    {
        const DictNode* dictNode = static_cast<const DictNode*>(src);
        picojson::object dict;
        for
            (const auto& it : dictNode->children)
        {
            ok = this->serializeNode(it.second, dict, it.first);
            if
                (false == ok)
            {
                return false;
            }
        }
        dst.emplace(std::make_pair(key, picojson::value(dict)));
    }
    else
    {
        return false;
    }

    return true;
}

/**
 *
 */
bool JSONAnnotation::serializeShapeForSyntool(OGL::Shape* shape, picojson::object& result) const
{
    if
        (nullptr == shape)
    {
        return false;
    }

    if
        (SEAScope::POLYLINE == shape->getType())
    {
        Polyline* polyline = static_cast<Polyline*>(shape);
        const std::vector<glm::vec3>& vertices = polyline->getPoints();
        unsigned int pointsCount = polyline->getPointCount();
        if
            (2 > pointsCount)
        {
            return false;
        }

        result.emplace(std::make_pair("type", "LINE"));

        double lon = 0.0;
        double lat = 0.0;
        picojson::array verticesNode;
        for
            (const auto& vertex : vertices)
        {
            Geodetic2D geo2d = UnitSphere::ToGeodetic2D(vertex);
            lon = 180.0 * geo2d.Longitude() / SEAScope::PI;
            lat = 180.0 * geo2d.Latitude() / SEAScope::PI;

            picojson::array vertexNode;
            vertexNode.emplace_back(lon);
            vertexNode.emplace_back(lat);

            verticesNode.push_back(picojson::value(vertexNode));
        }
        result.emplace(std::make_pair("points", picojson::value(verticesNode)));
        polyline = nullptr;
    }
    else if
        (SEAScope::POLYGON == shape->getType())
    {
        Polygon* polygon = static_cast<Polygon*>(shape);
        const std::vector<glm::vec3>& vertices = polygon->getPoints();
        unsigned int pointsCount = polygon->getPointCount();
        if
            (3 > pointsCount)
        {
            return false;
        }

        result.emplace(std::make_pair("type", "WKT"));

        double lon = 0.0;
        double lat = 0.0;
        std::ostringstream wktBuffer("");
        wktBuffer << "POLYGON((";
        for
            (const auto& vertex : vertices)
        {
            Geodetic2D geo2d = UnitSphere::ToGeodetic2D(vertex);
            lon = 180.0 * geo2d.Longitude() / SEAScope::PI;
            lat = 180.0 * geo2d.Latitude() / SEAScope::PI;
            wktBuffer << lon << " " << lat << ",";
        }

        // Duplicate the first vertex at the end to create a closed linear ring
        if
            (vertices[0] != vertices[pointsCount - 1])
        {
            Geodetic2D geo2d = UnitSphere::ToGeodetic2D(vertices[0]);
            lon = 180.0 * geo2d.Longitude() / SEAScope::PI;
            lat = 180.0 * geo2d.Latitude() / SEAScope::PI;
            wktBuffer << lon << " " << lat << ",";
        }

        // Move write head to overwrite trailing comma
        wktBuffer.seekp(-1, wktBuffer.cur);

        wktBuffer << "))";
        result.emplace(std::make_pair("wkt", wktBuffer.str()));
        polygon = nullptr;
    }
    else if
        (SEAScope::POINT == shape->getType())
    {
        Point* point = static_cast<Point*>(shape);

        result.emplace(std::make_pair("type", "WKT"));

        Geodetic2D geo2d = UnitSphere::ToGeodetic2D(point->getPosition());
        double lon = 180.0 * geo2d.Longitude() / SEAScope::PI;
        double lat = 180.0 * geo2d.Latitude() / SEAScope::PI;

        std::stringstream wktBuffer("");
        wktBuffer << "POINT(" << lon << " " << lat << ")";
        result.emplace(std::make_pair("wkt", wktBuffer.str()));
        point = nullptr;
    }
    else if
        (SEAScope::ELLIPSE == shape->getType())
    {
        Ellipse* ellipse = static_cast<Ellipse*>(shape);

        result.emplace(std::make_pair("type", "ELLIPSE"));

		std::ostringstream buffer("");
		buffer << "ELLIPSE((";

        buffer << ellipse->getAlphaPoint().x << " " << ellipse->getAlphaPoint().y << " " << ellipse->getAlphaPoint().z << ",";
        buffer << ellipse->getBetaPoint().x << " " << ellipse->getBetaPoint().y << " " << ellipse->getBetaPoint().z << ",";
        buffer << ellipse->getCenter().x << " " << ellipse->getCenter().y << " " << ellipse->getCenter().z;

		buffer << "))";
		result.emplace(std::make_pair("ellipse", buffer.str()));
		ellipse = nullptr;
    }
    else
    {
        return false;
    }

    return true;
}

/**
 *
 */
bool JSONAnnotation::toPicoJSON(const Annotation& src, picojson::object& result) const
{
    bool ok(false);    //

    for
        (const auto& it : src.getConstProperties().children)
    {
        ok = this->serializeNode(it.second, result, it.first);
        if
            (false == ok)
        {
            return false;
        }
    }

    if
        (src.hasStart())
    {
        result.emplace(std::make_pair("start", src.getStart()));
    }

    if
        (src.hasEnd())
    {
        result.emplace(std::make_pair("end", src.getStop()));
    }

    ok = this->serializeShapeForSyntool(src.getShape(), result);

    return ok;
}

/**
 *
 */
bool JSONAnnotation::serialize(const Annotation& src, std::ostream& stream) const
{
    bool ok(false); //
    picojson::object buffer;    //

    ok = this->toPicoJSON(src, buffer);
    if
        (false == ok)
    {
        return false;
    }

    picojson::array result;
    result.push_back(picojson::value(buffer));

    picojson::value(result).serialize(std::ostream_iterator<char>(stream));

    return true;
}

/**
 *
 */
bool JSONAnnotation::serialize(const std::vector<Annotation>& src, std::ostream& stream) const
{
    picojson::array result;    //
    bool ok(false);    //

    for
        (const auto& annotation : src)
    {
        picojson::object buffer;    //
        ok = this->toPicoJSON(annotation, buffer);
        if
            (false == ok)
        {
            return false;
        }

        result.push_back(picojson::value(buffer));
    }

    picojson::value(result).serialize(std::ostream_iterator<char>(stream));

    return true;
}

/**
 *
 */
bool JSONAnnotation::deserialize(std::istream& stream, Annotation& dst, double /*defaultStart*/, double /*defaultStop*/) const
{
    bool ok;    //
    picojson::value v;  //
    std::istreambuf_iterator<char> input(stream.rdbuf());   //
    std::string err;    //

    ok = true;

    input = picojson::parse(v, input, std::istreambuf_iterator<char>(), &err);
    if
        (false == err.empty())
    {
        LOG(LogLevel::Error, "JSON Annotation",
                 "Failed to parse input stream as JSON: {}", err);
        return false;
    }

    ok = this->parseSyntoolShape(v, dst);
    return ok;
}

/**
 *
 */
bool JSONAnnotation::deserialize(std::istream& stream, std::vector<Annotation>& dst, double /*defaultStart*/, double /*defaultStop*/) const
{
    bool ok;    //
    picojson::value v;  //
    std::istreambuf_iterator<char> input(stream.rdbuf());   //
    std::string err;    //

    ok = true;

    input = picojson::parse(v, input, std::istreambuf_iterator<char>(), &err);
    if
        (false == err.empty())
    {
        LOG(LogLevel::Error, "JSON Annotation",
                 "Failed to parse input stream as JSON: {}", err);
        return false;
    }

    ok = this->parseSyntoolExport(v, dst);
    return ok;
}

/**
 *
 */
bool JSONAnnotation::parseSyntoolExport(const picojson::value& v, std::vector<Annotation>& dst) const
{
    bool ok;    //
    picojson::array shapes; //

    ok = true;

    if
        (v.is<picojson::object>())
    {
        const picojson::object& obj = v.get<picojson::object>();
        auto found = obj.find("shapes");
        if
            (obj.end() == found)
        {
            LOG(LogLevel::Error, "JSON Annotation",
                     "JSON should either contain a list of annotations at the root level or as a \"shapes\" field");
            return false;
        }

        if
            (false == found->second.is<picojson::array>())
        {
            LOG(LogLevel::Error, "JSON Annotation",
                     "The \"shapes\" field should contain a list");
            return false;
        }

        shapes = found->second.get<picojson::array>();
        LOG(LogLevel::Debug, "JSON Annotation",
                 "Processing \"shapes\" from a full export");
    }
    else if
        (v.is<picojson::array>())
    {
        shapes = v.get<picojson::array>();
        LOG(LogLevel::Debug, "JSON Annotation",
                 "Processing \"shapes\" from a partial export");
    }
    else
    {
        LOG(LogLevel::Error, "JSON Annotation", "JSON root is neither an object or an array");
        return false;
    }


    for
        (const auto& jsonAnnotation : shapes)
    {
        LOG(LogLevel::Debug, "JSON Annotation", "Parsing annotation");

        Annotation annotation;  //
        ok = parseSyntoolShape(jsonAnnotation, annotation);
        if
            (true == ok)
        {
            dst.push_back(std::move(annotation));
        }
        else
        {
            LOG(LogLevel::Warn, "JSON Annotation", "Annotation skipped");
            LOG(LogLevel::Debug, "JSON Annotation", jsonAnnotation.to_str());
        }
    }

    LOG(LogLevel::Debug, "JSON Annotation", "Deserialization complete");

    return true;
}

/**
 *
 */
bool JSONAnnotation::parseSyntoolShape(const picojson::value& src, Annotation& dst) const
{
    bool ok;    //

    ok = true;

    if
        (false == src.is<picojson::object>())
    {
        LOG(LogLevel::Warn, "JSON Annotation",
                 "User shape should be an object / a dict, but this one is not");
        LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
        return false;
    }


    const picojson::object& userShape = src.get<picojson::object>();

    auto startFound = userShape.find("start");
    if
        (startFound != userShape.end())
    {
        if
            (false == startFound->second.is<double>())
        {
            LOG(LogLevel::Warn, "JSON Annotation", "\"start\" field should contain a numeric value but this one does not");
            LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
        }
        else
        {
            dst.setStart(startFound->second.get<double>());
        }
    }
    else
    {
        // Set start to the smallest value to emulate infinite coverage
        dst.setStart(0); // 1970-01-01T00:00:00.000Z
    }

    auto endFound = userShape.find("end");
    if
        (endFound != userShape.end())
    {
        if
            (false == endFound->second.is<double>())
        {
            LOG(LogLevel::Warn, "JSON Annotation", "\"end\" field should contain a numeric value but this one does not");
            LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
        }
        else
        {
            dst.setStop(endFound->second.get<double>());
        }
    }
    else
    {
        // Set stop to a high value to emulate infinite coverage
        dst.setStop(4102444800000); // 2100-01-01T00:00:00.000Z
    }

    auto typeFound = userShape.find("type");
    if
        (typeFound == userShape.end())
    {
        LOG(LogLevel::Warn, "JSON Annotation",
                 "\"type\" field is mandatory but is not provided for this user shape");
        LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
        return false;
    }

    const std::string& shapeType = typeFound->second.get<std::string>();
    if
        ("WKT" == shapeType)
    {
        auto wktFound = userShape.find("wkt");
        if
            (userShape.end() == wktFound)
        {
            LOG(LogLevel::Warn, "JSON Annotation",
                     "\"wkt\" field is mandatory for polygons but has not been provided for this user shape");
            LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
            return false;
        }

        if
            (false == wktFound->second.is<std::string>())
        {
            LOG(LogLevel::Warn, "JSON Annotation", "\"wkt\" field must be a string");
            LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
            return false;
        }

        const std::string& wktString = wktFound->second.get<std::string>();

        // Make sure the content of the WKT is a shape supported by SEAScope
        if
            (  (std::string::npos == wktString.rfind("POLYGON", 0))
            && (std::string::npos == wktString.rfind("POINT", 0)))
        {
            LOG(LogLevel::Warn, "JSON Annotation",
                     "SEAScope only supports the definition of the following shapes in WKT format: POINT, POLYGON.");
            LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
            return false;
        }

        std::size_t openParenthesisIndex = wktString.rfind('(');
        std::size_t closeParenthesisIndex = wktString.find(')');
        std::string coordsString = wktString.substr(1 + openParenthesisIndex, closeParenthesisIndex - openParenthesisIndex - 1);
        //LOG(LogLevel::Debug, "JSON Annotation", "WKT coordinates string: {}", coordsString);

        std::vector<std::string> coordsSubstrings;
        SEAScope::explodeString(coordsString, ",", coordsSubstrings);

        std::size_t pointsCount = coordsSubstrings.size();
        std::vector<glm::vec2> geographic_points;
        geographic_points.reserve(pointsCount);
        for
            (auto& it : coordsSubstrings)
        {
            std::string::size_type stringPos;
            double lon = std::stod(it, &stringPos);
            double lat = std::stod(it.substr(stringPos + 1));
            geographic_points.push_back(glm::vec2(lon, lat));
            //LOG(LogLevel::Debug, "JSON Annotation", "WKT coords: string = {}, lon = {}, lat = {}", it, lon, lat);
        }

        std::vector<glm::vec3> geodesic_points;
        fromLonlatToSphereUnitVector(geographic_points, geodesic_points);

        if
            (1 == geodesic_points.size())
        {
            Point* point = ShapesAllocator::getShape<Point>();
            if
                (nullptr == point)
            {
                LOG(LogLevel::Error, "JSONAnnotation", "Could not allocate point shape");
                return false;
            }
            point->setPosition(geodesic_points[0]);
            dst.setShape(std::move(point));
            point = nullptr;
        }
        else
        {
            while
                (  (3 < pointsCount)
                && (geodesic_points[0] == geodesic_points[pointsCount - 1u]))
            {
                // Remove last point as it is the same as the first one
                // (requirement for WKT but unecessary for polygon geometries
                // in SEAScope.
                geodesic_points.pop_back();
                pointsCount = pointsCount - 1u;
            }

            Polygon* polygon = ShapesAllocator::getShape<Polygon>();
            if
                (nullptr == polygon)
            {
                LOG(LogLevel::Error, "JSONAnnotation", "Could not allocate polygon shape");
                return false;
            }
            polygon->setPoints(std::move(geodesic_points));
            dst.setShape(std::move(polygon));
            polygon = nullptr;
        }
    }
    else if
        ("LINE" == shapeType)
    {
        auto pointsFound = userShape.find("points");
        if
            (pointsFound == userShape.end())
        {
            LOG(LogLevel::Warn, "JSON Annotation",
                     "\"points\" field is mandatory for polylines but has not been provided for this user shape");
            LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
            return false;
        }

        if
            (false == pointsFound->second.is<picojson::array>())
        {
            LOG(LogLevel::Warn, "JSON Annotation", "\"points\" field must be a list");
            LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
            return false;
        }

        const picojson::array& jsonPoints = pointsFound->second.get<picojson::array>();
        ListNode pointsList;
        for
            (const auto& point : jsonPoints)
        {
            ok = parseGenericProperty(point, pointsList);
            if
                (false == ok)
            {
                LOG(LogLevel::Warn, "JSON Annotation", "Failed to parse the content of the \"points\" field");
                LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
                return false;
            }
        }

        //LOG(LogLevel::Debug, "JSON Annotation", "POLYLINE POINTS");
        //printDataNode(pointsList);

        std::vector<glm::vec2> geographic_points;
        std::size_t pointsCount = pointsList.children.size();
        geographic_points.reserve(pointsCount);
        for
            (const auto& vertex : pointsList.children)
        {
            const ListNode& coords = static_cast<const ListNode&>(*vertex);
            const NumberNode& lon = static_cast<const NumberNode&>(*coords.children[0]);
            const NumberNode& lat = static_cast<const NumberNode&>(*coords.children[1]);
            geographic_points.push_back(glm::vec2(lon.value, lat.value));
            //LOG(LogLevel::Debug, "JSON Annotation", "Adding point {} {} out of {}", lon.value, lat.value, pointsCount);
        }

        std::vector<glm::vec3> geodesic_points;
        fromLonlatToSphereUnitVector(geographic_points, geodesic_points);
        Polyline* polyline = ShapesAllocator::getShape<Polyline>();
        if
            (nullptr == polyline)
        {
            LOG(LogLevel::Error, "JSONAnnotation", "Could not allocate polyline shape");
            return false;
        }
        polyline->setPoints(std::move(geodesic_points));
        dst.setShape(std::move(polyline));
        polyline = nullptr;
    }
    else if ("ELLIPSE" == shapeType)
    {
        auto found = userShape.find("ellipse");
        if
            (userShape.end() == found)
        {
            LOG(LogLevel::Warn, "JSON Annotation",
                "\"ellipse\" field is mandatory for ellipses but has not been provided for this user shape");
            LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
            return false;
        }

        if
            (false == found->second.is<std::string>())
        {
            LOG(LogLevel::Warn, "JSON Annotation", "\"ellipse\" field must be a string");
            LOG(LogLevel::Debug, "JSON Annotation", src.to_str());
            return false;
        }

		const std::string& str = found->second.get<std::string>();
		std::size_t openParenthesisIndex = str.rfind('(');
		std::size_t closeParenthesisIndex = str.find(')');

		std::vector<std::string> coordsSubstrings;

        if
            (   std::string::npos != openParenthesisIndex
             && std::string::npos != closeParenthesisIndex)
        {
			std::string coordsString = str.substr(1 + openParenthesisIndex, closeParenthesisIndex - openParenthesisIndex - 1);
			SEAScope::explodeString(coordsString, ",", coordsSubstrings);
        }

        // 3 elements are awaited : center point, alpha and beta point.
        if
            (3 == coordsSubstrings.size())
        {
			std::vector<std::string> coordsSubstrings2;
			SEAScope::explodeString(coordsSubstrings[0], " ", coordsSubstrings2);

            glm::vec3 alpha, beta, center;

            // String to float conversion can failed and throw an exception.
            try
            {
                // 3 elements are awaited : x,y,z float
				if
					(3 == coordsSubstrings2.size())
				{
					alpha.x = std::stof(coordsSubstrings2[0]);
					alpha.y = std::stof(coordsSubstrings2[1]);
					alpha.z = std::stof(coordsSubstrings2[2]);
				}

				coordsSubstrings2.clear();
				SEAScope::explodeString(coordsSubstrings[1], " ", coordsSubstrings2);

				// 3 elements are awaited : x,y,z float
				if
					(3 == coordsSubstrings2.size())
				{
					beta.x = std::stof(coordsSubstrings2[0]);
					beta.y = std::stof(coordsSubstrings2[1]);
					beta.z = std::stof(coordsSubstrings2[2]);
				}

				coordsSubstrings2.clear();
				SEAScope::explodeString(coordsSubstrings[2], " ", coordsSubstrings2);

				// 3 elements are awaited : x,y,z float
				if
					(3 == coordsSubstrings2.size())
				{
					center.x = std::stof(coordsSubstrings2[0]);
					center.y = std::stof(coordsSubstrings2[1]);
					center.z = std::stof(coordsSubstrings2[2]);
				}

				Ellipse* ellipse = ShapesAllocator::getShape<Ellipse>();
                if
                    (nullptr == ellipse)
                {
                    LOG(LogLevel::Error, "JSONAnnotation", "Could not allocate ellipse shape");
                    return false;
                }
				ellipse->initialize(center, alpha, beta);

				dst.setShape(ellipse);

				ellipse = nullptr;
			}
			catch (const std::exception& e)
			{
                LOG(LogLevel::Error, "JSON Annotation", "Error during ellipse parsing : {}", e.what());
                return false;
			}
        }
    }
    else
    {
        LOG(LogLevel::Warn, "JSON Annotation", "Shape type not supported: {}", shapeType);
        return false;
    }

    // Handle all other properties
    for
        (auto& property : userShape)
    {
        if
            ("type" == property.first
            || "points" == property.first
            || "wkt" == property.first
            || "ellipse" == property.first
            || "start" == property.first
            || "end" == property.first
            || "intersectedGranules" == property.first)
        {
            // These properties have already been processed
            continue;
        }

        ok = this->parseGenericProperty(property.second, dst.getProperties(), property.first);
        if
            (false == ok)
        {
            return false;
        }
    }
    return true;
}

/**
 *
 */
bool JSONAnnotation::parseGenericProperty(const picojson::value& value, DictNode& dst, const std::string& key) const
{
    bool ok;    //

    ok = true;

    if
        (value.is<picojson::null>())
    {
        dst.children.emplace(std::make_pair(key, new EmptyNode()));
    }
    else if
        (value.is<bool>())
    {
        dst.children.emplace(std::make_pair(key, new BoolNode(value.get<bool>())));
    }
    else if
        (value.is<double>())
    {
        dst.children.emplace(std::make_pair(key, new NumberNode(value.get<double>())));
    }
    else if
        (value.is<std::string>())
    {
        const std::string& v = value.get<std::string>();
        if
            (  (this->_base64Flag.size() < v.size())
            && (0 == v.compare(0, this->_base64Flag.size(), this->_base64Flag)))
        {
            dst.children.emplace(std::make_pair(key, new ImageNode(v.substr(this->_base64Flag.size()))));
        }
        else
        {
            dst.children.emplace(std::make_pair(key, new TextNode(v)));
        }
    }
    else if
        (value.is<picojson::array>())
    {
        const picojson::array& list = value.get<picojson::array>();
        ListNode* listNode = new ListNode();
        for
            (auto& it : list)
        {
            ok = parseGenericProperty(it, *listNode);
            if
                (false == ok)
            {
                return false;
            }
        }
        dst.children.emplace(std::make_pair(key, std::move(listNode)));
    }
    else if
        (value.is<picojson::object>())
    {
        const picojson::object& dict = value.get<picojson::object>();
        DictNode* dictNode = new DictNode();
        for
            (auto& it : dict)
        {
            ok = parseGenericProperty(it.second, *dictNode, it.first);
            if
                (false == ok)
            {
                return false;
            }
        }
        dst.children.emplace(std::make_pair(key, std::move(dictNode)));
    }
    else
    {
        return false;
    }

    return true;
}

/**
 *
 */
bool JSONAnnotation::parseGenericProperty(const picojson::value& value, ListNode& dst) const
{
    bool ok;    //

    ok = true;

    if
        (value.is<picojson::null>())
    {
        dst.children.push_back(new EmptyNode());
    }
    else if
        (value.is<bool>())
    {
        dst.children.push_back(new BoolNode(value.get<bool>()));
    }
    else if
        (value.is<double>())
    {
        dst.children.push_back(new NumberNode(value.get<double>()));
    }
    else if
        (value.is<std::string>())
    {
        dst.children.push_back(new TextNode(value.get<std::string>()));
        const std::string& v = value.get<std::string>();
        if
            (  (this->_base64Flag.size() < v.size())
            && (0 == v.compare(0, this->_base64Flag.size(), this->_base64Flag)))
        {
            dst.children.push_back(new ImageNode(v));
        }
        else
        {
            dst.children.push_back(new TextNode(v));
        }
    }
    else if
        (value.is<picojson::array>())
    {
        const picojson::array& list = value.get<picojson::array>();
        ListNode* listNode = new ListNode();
        for
            (auto& it : list)
        {
            ok = parseGenericProperty(it, *listNode);
            if
                (false == ok)
            {
                return false;
            }
        }
        dst.children.push_back(std::move(listNode));
    }
    else if
        (value.is<picojson::object>())
    {
        const picojson::object& dict = value.get<picojson::object>();
        DictNode* dictNode = new DictNode();
        for
            (auto& it : dict)
        {
            ok = parseGenericProperty(it.second, *dictNode, it.first);
            if
                (false == ok)
            {
                return false;
            }
        }
        dst.children.push_back(std::move(dictNode));
    }
    else
    {
        return false;
    }

    return true;
}
