// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <mutex>
#include <chrono>
#include <algorithm>

#include "core/log.h"
#include "core/utils.h"
#include "core/data_bucket.h"
#include "local/in_memory_data_store.h"

using namespace SEAScope;

/**
 *
 */
InMemoryDataStore::InMemoryDataStore()
{
};

/**
 *
 */
InMemoryDataStore::~InMemoryDataStore()
{

};

/**
 *
 */
GranuleData& InMemoryDataStore::getURIDataRef(const std::string& uri)
{
    const std::lock_guard<std::mutex> lock(this->_dataMutex);

	auto found = this->_data.find(uri);
	if
		(this->_data.end() == found)
	{
        // Create a new data bucket
        this->_data.emplace(std::make_pair(uri, GranuleData()));
	}

	return this->_data[uri];
};

/**
 *
 */
bool InMemoryDataStore::removeData(const std::string& uri)
{
    const std::lock_guard<std::mutex> lock(this->_dataMutex);

    auto found = this->_data.find(uri);
    if
        (this->_data.end() != found)
    {
        std::vector<std::string> missingFields; //
        for
            (const auto& field : found->second)
        {
            if
                (true == field.second.isMissing())
            {
                missingFields.emplace_back(field.first);
            }
        }

        for
            (const auto& fieldName : missingFields)
        {
            found->second.erase(fieldName);
        }

        // If there is no field remaining, remove uri from the store
        std::size_t fieldsCount {0u};
        fieldsCount = found->second.size();
        if
            (0 >= fieldsCount)
        {
		    this->_data.erase(found);
        }
		return true;
    }

	return false;
}

/**
 *
 */
bool InMemoryDataStore::listFields(const std::string& uri, std::set<std::string>& result)
{
    std::lock_guard<std::mutex> lock(this->_dataMutex);

    if
        (true == this->_data.empty())
    {
        return false;
    }

    auto found = this->_data.find(uri);
    if
        (this->_data.end() == found)
    {
        return false;
    }

    for
        (const auto& field : found->second)
    {
        if
            (true == field.second.loaded())
        {
            result.insert(field.first);
        }
    }

    return true;
}

/**
 * This method should be the only way to get read-write access to a data bucket
 */
bool InMemoryDataStore::getMissingDataBuckets(const std::string& uri,
                                              const std::set<std::string>& fields,
                                              std::unordered_map<std::string, DataBucket*>& result)
{
    std::lock_guard<std::mutex> lock {this->_dataMutex};

    // The following will create an entry for the URI if it does not exist yet
    this->_data[uri];

    for
        (const auto& field : fields)
    {
        if
            (0 < this->_data[uri].count(field))
        {
            // The field already exists
            continue;
        }

        // Create the field
        DataBucket& bucket = this->_data[uri][field];

        // Store RW reference in the result
        result[field] = &bucket;
    }

    return true;
}

/**
 *
 */
bool InMemoryDataStore::getField(const std::string& uri,
                                 const std::string& field,
                                 const Image*& buffer,
                                 const DataInfo*& info) const
{
    std::lock_guard<std::mutex> lock {this->_dataMutex};
    if
        (0 >= this->_data.count(uri))
    {
        return false;
    }

    if
        (0 >= this->_data.at(uri).count(field))
    {
        return false;
    }

    const DataBucket& bucket = this->_data.at(uri).at(field);
    buffer = bucket.getImagePtr();
    info = bucket.getInfoPtr();
    return true;
}

/**
 *
 */
bool InMemoryDataStore::getFields(const std::string& uri,
                                  const std::set<std::string>& fields,
                                  std::vector<const Image*>& buffers,
                                  std::vector<const DataInfo*>& infos) const
{
    std::lock_guard<std::mutex> lock {this->_dataMutex};
    if
        (0 >= this->_data.count(uri))
    {
        return false;
    }

    for
        (const auto& field : fields)
    {
        if
            (0 >= this->_data.at(uri).count(field))
        {
            buffers.clear();
            infos.clear();
            return false;
        }

        const DataBucket& bucket = this->_data.at(uri).at(field);
        buffers.emplace_back(bucket.getImagePtr());
        infos.emplace_back(bucket.getInfoPtr());
    }

    return true;
}
