// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/interfaces/ilog_handler.h"
#include "core/interfaces/ipath_handler.h"
#include "core/settings.h"
#include "local/user_generated.h"
#include "core/app_events_dispatcher.h"
#include "events/granule_added_event.h"
#include "core/granule_metadata.h"

using namespace SEAScope;

/**
 *
 */
UserGenerated::UserGenerated(SourceId sourceId, AppEventsDispatcher& appEventsDispatcher, IGranulesIndex& index)
: _sourceId(sourceId),
  _appEventsDispatcher(appEventsDispatcher),
  _index(index)
{
	this->_index.setGranuleAddedCallback([this] (std::uint64_t granuleId, std::uint32_t collectionId) {
        GranuleAddedEvent event(this->_sourceId, collectionId, granuleId);
        this->_appEventsDispatcher.dispatch<GranuleAddedEvent>(event);
    });
};

/**
 *
 */
UserGenerated::~UserGenerated()
{
};

/**
 *
 */
bool UserGenerated::init(const Settings&, ILogHandler*, IPathHandler*, IDialogHandler*)
{
    return true;
};

/**
 *
 */
bool UserGenerated::reset()
{
    return true;
}

/**
 *
 */
bool UserGenerated::searchGranules( const SearchCriteria& criteria
                                  , std::vector<GranuleMetadata>& result) const
{
	bool ok;	//
	ok = this->_index.findGranules(criteria, result);
	return ok;
};

/**
 *
 */
std::array<bool, 2> UserGenerated::searchPrevNext( timestamp_t current
                                                 , const std::set<CollectionId>& collections
                                                 , timestamp_t& prev
                                                 , timestamp_t& next
                                                 , timestamp_t minDelta) const
{
	std::set<std::uint32_t> localCollections;	//
	GranuleMetadata prevGranule;	//
	GranuleMetadata nextGranule;	//

	LOG( LogLevel::Debug, "UserIndex"
			, "searchPrevNext: {}"
			, current);

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	LOG( LogLevel::Debug, "UserIndex"
			, "Local collections list built");

	auto oks = this->_index.adjacentGranules( 1000 * current
										    , 1000 * minDelta
										    , localCollections
										    , prevGranule, nextGranule);

	LOG( LogLevel::Debug, "UserIndex"
			, "Adjacent granules ok");
	LOG( LogLevel::Debug, "UserIndex"
			, "{} {}"
			, oks[0], oks[1]);
	if
		(true == oks[0])
	{
		prev = static_cast<timestamp_t>(0.0005 * static_cast<double>(prevGranule.timeCoverageStart())  + 0.0005 * static_cast<double>(prevGranule.timeCoverageEnd()));
	}
	if
		(true == oks[1])
	{
		next = static_cast<timestamp_t>(0.0005 * static_cast<double>(nextGranule.timeCoverageStart()) + 0.0005 * static_cast<double>(nextGranule.timeCoverageEnd()));
	}

	LOG( LogLevel::Debug, "UserIndex"
			, "Returning from searchPrevNext");

    return oks;
};

/**
 *
 */
bool UserGenerated::searchClosestBound( timestamp_t current
                                      , const std::set<CollectionId>& collections
                                      , timestamp_t& closest
                                      , const timestamp_t searchLowerBound
                                      , const timestamp_t searchUpperBound) const
{
	timestamp_t delta;	//
	timestamp_t closestTmp;	//
	std::set<std::uint32_t> localCollections;	//
	GranuleMetadata closestGranule;	//
	bool ok;	//

	LOG( LogLevel::Debug, "UserIndex"
			, "searchClosestBound: {}"
			, current);

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	LOG( LogLevel::Debug, "UserIndex"
			, "Local collections list built");
    LOG( LogLevel::Debug, "UserIndex"
            , "number of local collections: {}", localCollections.size());

    const timestamp_t currentMs {1000 * current};   //
	ok = this->_index.closestGranuleBound(currentMs, 1000 * searchLowerBound, 1000 * searchUpperBound, localCollections, closestGranule);
	if
		(true == ok)
	{
        /*
        LOG( LogLevel::Debug, "UserIndex", "closest granule found: {} {} {}"
                , closestGranule.uris[0].uri
                , closestGranule.start
                , closestGranule.stop);
        */

		if
			(  closestGranule.timeCoverageStart() <= currentMs
			&& closestGranule.timeCoverageEnd() > currentMs)
		{
			/* A granule contains the current datetime, no bounds required */
			closest = current;
			return true;
		}
		closestTmp = current;
		delta = std::numeric_limits<uint64_t>::max();
		if
			(  closestGranule.timeCoverageStart() < currentMs
			&& delta > currentMs - closestGranule.timeCoverageStart())
		{
			delta = currentMs - closestGranule.timeCoverageStart();
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.timeCoverageStart()));
		}
		else if
			(  closestGranule.timeCoverageStart() > currentMs
			&& delta > closestGranule.timeCoverageStart() - currentMs)
		{
			delta = closestGranule.timeCoverageStart() - currentMs;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.timeCoverageStart()));
		}

		/* Remove one second when using the end of the granule as a result
		   or the granule will no be displayed */
		if
			(  closestGranule.timeCoverageEnd() < currentMs
			&& delta > currentMs - closestGranule.timeCoverageEnd())
		{
			delta = currentMs - closestGranule.timeCoverageEnd();
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.timeCoverageEnd())) - 1;
		}
		else if
			(  closestGranule.timeCoverageEnd() > currentMs
			&& delta > closestGranule.timeCoverageEnd() - currentMs)
		{
			delta = closestGranule.timeCoverageEnd() - currentMs;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.timeCoverageEnd())) - 1;
		}

		closest = closestTmp;
	}

	LOG(LogLevel::Debug, "UserIndex", "closest bound: {}", closest);
    return ok;
};

/**
 *
 */
bool UserGenerated::getDataCoverage( const std::set<CollectionId>& collections
                                   , EventsCoverage& result) const
{
	std::set<std::uint32_t> localCollections;	//
	bool ok;	//

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	ok = this->_index.getCoverage(localCollections, result);
	return ok;
};

/**
 *
 */
bool UserGenerated::getDataCoverage( const CollectionId& collectionId
                                   , EventsCoverage& result) const
{
	bool ok;	//

	ok = this->_index.getCoverage(collectionId.second, result);
	return ok;
};

/**
 *
 */
bool UserGenerated::canHandle(const std::string&) const
{
    return false;
}

const SourceId& UserGenerated::getSourceId() const
{
    return this->_sourceId;
}

/**
 *
 */
bool UserGenerated::generateRenderableId(const CollectionId& colId, RenderableId& result)
{
    if
        (this->_sourceId != colId.first)
    {
        LOG(LogLevel::Error, "User generated",
                 "Generation of granule identifier for an external collection is not allowed");
        return false;
    }

    result.granuleLevel = true;
    result.collection = colId;
    result.granule.first = colId.first; /* same source as collection */
    result.granule.second = this->_index.generateGranuleId();
    result.variable = 0; /* default value */

    return true;
}

/**
 *
 */
IGranulesIndex& UserGenerated::getIndex()
{
    return this->_index;
}

/**
 *
 */
bool UserGenerated::listCollections(std::set<std::uint32_t>&) const
{
    LOG(LogLevel::Error, "User generated", "'listCollections' is not implemented for UserGenerated");
    return false;
}

/**
 *
 */
bool UserGenerated::getCollectionCfg(std::uint32_t, CollectionCfg&, std::vector<RenderingCfg*>&) const
{
    LOG(LogLevel::Error, "User generated", "'getCollectionCfg' is not implemented for UserGenerated");
    return false;
}

/**
 *
 */
bool UserGenerated::saveVariableRenderingCfg(const SEAScope::RenderingCfg*) const
{
    LOG(LogLevel::Error, "User generated", "'saveVariableRenderingCfg' is not implemented for UserGenerated");
    return false;
}
