// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cstdio>
#include <cstdint>
#include <cstdlib>
#include <map>
#include <set>
#include <array>
#include <vector>
#include <string>
#include <sstream>
#include <fstream>
#include <iomanip>

#include "vendors/_cxxopts.h"
#include "vendors/_nanogui.h"

#include "num_utils.h"

#include "core/log.h"
#include "core/dialog.h"
#include "core/path.h"
#include "core/guid.h"
#include "core/source.h"
#include "core/application.h"

#include "simplelogger.h"
#include "filter_logger.h"
#include "ini_settings.h"
#include "sqlite_state.h"
#include "glfw_window.h"
#include "glfw_events_dispatcher.h"
#include "dummy_metadata_store.h"
#include "flatbuffers_granules_index.h"
#include "sqlite_granules_index.h"
#include "sqlite_annotations.h"
#include "ui/nanogui/nanogui_ui.h"

#include "local/in_memory_data_store.h"
#include "data/events_coverage.h"
#include "sink_eventhandler.h"

#include "core/types.h"
#include "core/structs.h"
#include "core/timeline_cfg.h"
#include "core/app_events_dispatcher.h"

#include "3d/resources.h"
#include "3d/shape.h"
#include "3d/memory_manager.h"
#include "3d/shapes_allocator.h"
#include "3d/datasets_allocator.h"
#include "3d/globe_scene.h"
#include "3d/scene_view.h"
#include "3d/texture_loader.h"

#include "platform/plugins_loader_impl.h"
#include "platform/dialog_impl.h"
#include "platform/path_impl.h"
#include "platform/guid_impl.h"
#include "platform/sockets_monitor_impl.h"
#include "platform/socket_factory_impl.h"
#include "platform/app_launcher_impl.h"

#include "tools/window_controller.h"
#include "tools/debug_controller.h"
#include "tools/extraction_controller.h"
#include "tools/annotations_controller.h"
#include "tools/drawing_controller.h"

#include "events/side_menu_resize_request.h"
#include "events/display_catalogue_request.h"
#include "events/scene_view_updated_event.h"
#include "events/scene_view_controller_ready_event.h"
#include "events/drawing_controller_ready_event.h"
#include "events/collections_update_requested_event.h"

#include "seascope_version.h"


using namespace SEAScope;


/**
 *
 */
int main(int argc, char** argv)
{
    DialogImpl dialogImplementation;
    PathImpl pathImplementation;
    GUIDImpl guidImplementation;
    FilterLogger logger;

	/* Set specific implementations */
	Dialog::implementation = &dialogImplementation;
	Path::implementation = &pathImplementation;
	SEAScope::GUID::implementation = &guidImplementation;
    Log::implementation = &logger;

    { // Scope for stack-allocated objects

    AppLauncherImpl appLauncher;    //
	INISettings settingsStore;	//
	SQLiteState stateStore;	//
	GLFWWindow window;			//
	PluginsLoaderImpl pLoader;	//
	SocketsMonitorImpl usm;	//
	SocketFactoryImpl sFactory;	//
	InMemoryDataStore imds;	//
	DummyMetadataStore dmds;	//
    FlatbuffersGranulesIndex volatileGranulesIndex; //
    SQLiteGranulesIndex annotationsIndex;  //

    AppEventsDispatcher appEventsDispatcher;
    GLFWEventsDispatcher inputEventsDispatcher;
    SQLiteAnnotations annotations(appEventsDispatcher);  //

	NanoguiUI ui(appEventsDispatcher, inputEventsDispatcher);				//

	Application app(appEventsDispatcher, volatileGranulesIndex, annotations, annotationsIndex, ui);			//

	RendererResources resources;	//
	bool ok {false};	//
	std::vector<std::string> txt;

    bool hasWorkspace {false};
    bool hasConfig {false};
    bool hasListen {false};
    bool hasFilters {false};
    bool hasPlugins {false};
	std::string configPath {""};	//
	std::string installDir {""};    //
	std::string listenURL {""}; //
    std::string logFilter {""}; //
    std::string pluginsPath {""};   //

#if defined __linux__ || defined __APPLE__
	/* Force UTC timezone */
	setenv("TZ", "UTC", 1);
#elif defined _WIN32
	_putenv_s("TZ", "UTC");
#endif

	/* Parse command line options */
	cxxopts::Options cliParser(argv[0], "Geolocated data visualization and manipulation tool.");
	cliParser.add_options()
		("v,version", "Print SEAScope version and build information")
        ("f,filter", "Level max of verbosity when debug is enabled (fatal, error, warn, info, debug)", cxxopts::value<std::string>(), "debug")
		("l,listen", "Open a control socket", cxxopts::value<std::string>(), "IFACE:PORT")
		("c,config", "Path to the configuration file", cxxopts::value<std::string>(), "")
        ("w,workspace", "Path of the directory where SEAScope is installed", cxxopts::value<std::string>(), "")
        ("p,plugins", "Path of the directory where SEAScope will look for plugins", cxxopts::value<std::string>(), "")
		("h,help", "Show this help message")
	;

    try
    {
        cxxopts::ParseResult options = cliParser.parse(argc, argv);

        /* Handle -h/--help */
        if
            (options.count("help"))
        {
            std::cout << cliParser.help({""}) << std::endl;
            return 0;
        }

        std::cout << "SEAScope is a 3D visualisation and analysis application for satellite, in-situ and\n";
        std::cout << "numerical model data.\n\n";
        std::cout << "Copyright (C) 2014-2025 OceanDataLab\n";
        std::cout << "License LGPLv3+\n\n";
        if
            (0 < options.count("version"))
        {
            std::cout << "Build information:\n";
            for
                (const auto& it : SEAScope::getVersion())
            {
                std::cout << "\t " << it.first << ": " << it.second << "\n";
            }
            return 0;
        }

        /* Handle -w/--workspace */
        hasWorkspace = (0 < options.count("workspace"));
        if
            (true == hasWorkspace)
        {
            installDir = options["workspace"].as<std::string>();
        }

        /* Handle -c/--config */
        hasConfig = (0 < options.count("config"));
        if
            (true == hasConfig)
        {
            configPath = options["config"].as<std::string>();
        }

		/* Handle -l/--listen */
        hasListen = (0 < options.count("listen"));
        if
            (true == hasListen)
        {
			listenURL = options["listen"].as<std::string>();
		}

		/* Handle -f/--filter */
        hasFilters = (0 < options.count("filter"));
        if (true == hasFilters)
        {
            logFilter = options["filter"].as<std::string>();
        }

        /* Handle -p/--plugins */
        hasPlugins = (0 < options.count("plugins"));
        if
            (true == hasPlugins)
        {
            pluginsPath = options["plugins"].as<std::string>();
        }
	}
    catch(cxxopts::OptionParseException& e)
    {
        std::cerr << e.what() << std::endl;
        return 1;
    }

    ok = appLauncher.initWorkspace(argv[0],
                                   hasWorkspace, installDir,
                                   hasConfig, configPath,
                                   hasListen, listenURL,
                                   hasFilters, logFilter);
    if
        (false == ok)
    {
        // Workspace initialization failed, no need to continue
        return 1;
    }

    std::cout << "SEAScope listens for incoming requests on " << listenURL << "\n" << std::endl;

    /* Initialize logger */
    ok = logger.init(logFilter);
    if
        (false == ok)
    {
        std::cerr << "Failed to initialize logger, please check the filter you provided for the -f argument" << std::endl;
        return 1;
    }

	/* Initialize application */
	ok = app.init(installDir, configPath, pluginsPath, settingsStore, stateStore, window, resources, logFilter);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Main", "Failed to initialize application");
		return 1;
	}

    // Load worldmap picture file (ownership will be given to the Globe object)
    OGL::Texture* worldMapTexture = TextureLoader::loadFromFile(app.getSceneRenderer().getContext(), resources.worldmapPath);

    GlobeScene scene(appEventsDispatcher);   //
    SceneView sceneView(appEventsDispatcher);    //
    sceneView.init(&scene, &(app.getSceneRenderer()), worldMapTexture);

	int width, height, fbWidth, fbHeight;

	// Nanogui computes a pixel ratio and may changes the window size.
	// Size in ini settings can be different than size used by nanogui.
	window.getFrameBufferSize(fbWidth, fbHeight);
    window.getSize(width, height);

    sceneView.setViewport(glm::ivec2(fbWidth, fbHeight));

	sceneView.getScene()->getNodeFactory()->setIcons(&(app.engine.getIcons()));

    app.initEngine(&sceneView, &dmds, &imds, &pLoader);
	app.engine.sockets.init(&usm, &sFactory);

	ok = app.loadState();
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Main", "Could not load state");
	    app.shutdown();
		return 1;
	}

	/* -------------------------------------------------------------------- */
	/*						C A L L B A C K S								*/
	/* -------------------------------------------------------------------- */


    /*
    mapRenderer->setDeleteVisibleUserShapesCallback([&app, &mapRenderer]() {
        bool success(false); //
        std::vector<RenderableId> toRemove;    //
        LOG(LogLevel::Debug, "Main", "Delete all visible user shapes");

        // export all rendered shapes
        for
            (const auto& rId : mapRenderer->toRender)
        {
            if
                (  1 != rId.collection.first
                || (  1 != rId.collection.second
                   && 2 != rId.collection.second))
            {
                // Only user generated polylines and polygons can be deleted
                continue;
            }

            toRemove.push_back(rId);
        }

        if
            (0 >= toRemove.size())
        {
            LOG(LogLevel::Debug, "Main", "No annotation to delete");
            return;
        }

        for
            (const auto& rId : toRemove)
        {
            mapRenderer->removeUserShape(rId);

            success = app.annotations->remove(rId.granule.second);
            if
                (false == success)
            {
                LOG(LogLevel::Error, "Main", "Failed to remove annoation {}", rId.granule.second);
            }

            std::ostringstream buffer("");
		    buffer << "usershape_" << rId.collection.first << "/" << rId.collection.second << "/" << rId.granule.second << ":" <<rId.variable;
		    app.engine.userGenerated.index.remove(buffer.str(), false);
        }
        toRemove.clear();

        // Trigger update
        app.engine.updateCollections({{1u, 1u}, {1u, 2u}}, {});
    });
    */

	ui.setSaveLayoutCallback([&app]() {
		bool success(false);	//

		LOG(LogLevel::Debug, "Main", "save layout...");
		success = app.saveState();
		if
			(false == success)
		{
			LOG( LogLevel::Error, "Main", "Could not save layout");
		}
	});


	/* Set timeline callbacks */
	app.engine.timeline.model.setNearestDataBoundCallback([&app] (timestamp_t& value, const timestamp_t searchLowerBound, const timestamp_t searchUpperBound) -> bool {
		bool success {false};	//
		timestamp_t closestBound {0};	//

		success = app.engine.metadataMediator.searchClosestBound( value
											  , app.engine.getActiveCollections()
											  , closestBound
                                              , searchLowerBound
                                              , searchUpperBound);
		if
			(true == success)
		{
		    value = closestBound;
		}

        return success;
	});

	app.engine.timeline.setCurrentChangedCallback([&app](timestamp_t value) {
		timestamp_t newPrev;	//
		timestamp_t newNext;	//
		timestamp_t minDelta;	//

		minDelta = 1;
		app.engine.metadataMediator.searchPrevNext( value
											  , app.engine.getActiveCollections()
											  , newPrev
											  , newNext
											  , minDelta);
		app.engine.timeline.model.setPrevNext(newPrev, newNext);
		LOG(LogLevel::Debug, "Main", "{} <= {} <= {}", value, newPrev, newNext);

		LOG(LogLevel::Debug, "Main", "UPDATE REQUESTED");
        app.engine.getUpdatesManager().requestUpdate();
	});
	app.engine.timeline.setTimeframeChangedCallback([&app](timestamp_t /*pastOffset*/, timestamp_t /*futureOffset*/) {
		LOG(LogLevel::Debug, "Main", "UPDATE REQUESTED");
        app.engine.getUpdatesManager().requestUpdate();
	});

	app.engine.sourcesControl.sourceInitialized = [&app] (const SourceInfo& /*info*/, const std::set<CollectionId>& collections) {
		bool success(false);	//
		LOG(LogLevel::Debug, "Main", "Source initialized!!!!");
		for
			(const auto& it: collections)
		{
			LOG(LogLevel::Debug, "Main", " - collection {}/{}", it.first, it.second);
		}
		success = app.engine.collectionsControl.model.addCollections(collections);
		if
			(false == success)
		{
			LOG(LogLevel::Error, "Main", "Failed to add collections");
		}
	};

	/* -------------------------------------------------------------------- */
	/*						D A T A    I N I T								*/
	/* -------------------------------------------------------------------- */
	/* Bootstrap demo */
	app.engine.sourcesControl.addSourceFromURI("embedded");

	// Try to locate a control plugin
	if
		(0 < listenURL.length())
	{
		std::forward_list<std::string> tokens;	//
		std::string iface;	//
		std::uint64_t port;	//

		Path::implementation->splitPaths(listenURL, ":", tokens);
		iface = tokens.front();
		tokens.pop_front();
		port = static_cast<std::uint64_t>(std::atol(tokens.front().c_str()));
		ok = app.openControlSocket(iface, port);
	}

    app.engine.metadataMediator.registerUserGeneratedIndex(SEAScope::Source::API, &(app.engine.getUserGenerated()));

	/* -------------------------------------------------------------------- */
	/*						R E N D E R I N G								*/
	/* -------------------------------------------------------------------- */

	/* Init renderers */
	window.addRenderer(sceneView);
	window.addRenderer(*(ui.getRenderer()));

	/* Init events handlers */
	window.addHandler(*(ui.getEventHandler()));


    // Make sure mouse move events will be expressed in framebuffer coords
    inputEventsDispatcher.setFrameBufferRatio(static_cast<double>(fbWidth / width),
                                              static_cast<double>(fbHeight / height));
    inputEventsDispatcher.setClickPixelsTolerance(app.engine.settings.clickPixelsTolerance);
    inputEventsDispatcher.initBindings(); // Must be done AFTER GLFW initialization
    window.addHandler(inputEventsDispatcher);

    WindowController& windowController = app.engine.getWindowController();
    windowController.setWindow(&window);
    inputEventsDispatcher.addSubscriber(&windowController);
    windowController.setActive(true);

    SceneViewControllerBase& sceneViewController = app.engine.getSceneViewController();
    inputEventsDispatcher.addSubscriber(&sceneViewController);
    sceneViewController.setActive(true);

    // notify toolbar that scene view controller is ready (to read keybindings)
    SceneViewControllerReadyEvent event(sceneViewController.getId());
    appEventsDispatcher.dispatch<SceneViewControllerReadyEvent>(event);

    ExtractionController& extractionController = app.engine.getExtractionController();
    inputEventsDispatcher.addSubscriber(&extractionController);
    extractionController.setActive(true);

    DebugController& debugController = app.engine.getDebugController();
    inputEventsDispatcher.addSubscriber(&debugController);
    debugController.setActive(true);

    SelectionController& selectionController = app.engine.getSelectionController();
    inputEventsDispatcher.addSubscriber(&selectionController);
    app.getSceneRenderer().registerPreRenderer(&selectionController);
    app.getSceneRenderer().registerPostRenderer(&selectionController);
    selectionController.setActive(true);

    TimelineController& timelineController = app.engine.getTimelineController();
    inputEventsDispatcher.addSubscriber(&timelineController);
    timelineController.setActive(true);

    DrawingController drawingController(appEventsDispatcher, sceneView, app.engine.getIcons());
    drawingController.setId(6);
    drawingController.setExplicitDrawingModeExit(app.engine.settings.explicitDrawingModeExit);
    drawingController.possibleActionsUpdated();
    inputEventsDispatcher.addSubscriber(&drawingController);
    app.getSceneRenderer().registerPostRenderer(&drawingController);
    drawingController.setActive(true);

    DrawingControllerReadyEvent drawingControllerReadyEvent(drawingController.getId());
    appEventsDispatcher.dispatch<DrawingControllerReadyEvent>(drawingControllerReadyEvent);

    AnnotationsController& annotationsController = app.engine.getAnnotationsController();
    annotationsController.init(app.engine.settings.annotationsPath);
    app.engine.metadataMediator.registerUserGeneratedIndex(SEAScope::Source::ANNOTATIONS, annotationsController.getUserGenerated());

	/* Display catalogue in the sidemenu */
    DisplayCatalogueRequest showCatalogueEvent;
    appEventsDispatcher.dispatch<DisplayCatalogueRequest>(showCatalogueEvent);

    // Required to initialize altitude in the camera controller
    SEAScope::SceneViewUpdatedEvent eventForAltitudeInit(&sceneView);
    appEventsDispatcher.dispatch<SEAScope::SceneViewUpdatedEvent>(eventForAltitudeInit);

	SinkEventHandler seh;
	window.addHandler(seh);

    // Callback for settings edition
    app.engine.setSettingsChangedCallback([&app, &drawingController, &inputEventsDispatcher, &ui] () {
        // Apply changes that do not require an app restart
        inputEventsDispatcher.setClickPixelsTolerance(app.engine.settings.clickPixelsTolerance);
        drawingController.setExplicitDrawingModeExit(app.engine.settings.explicitDrawingModeExit);
        ui.showProcessingOptions(app.engine.settings.processorEnabled);

        // Save to file
        app.saveSettings();
    });

    /* This is required on some systems (tested on Ubuntu 18.04 and 20.04)
    otherwise the Nanogui UI is not correctly initialized until the user
    manually resizes the window */
    int windowWidth(0);
    int windowHeight(0);
    window.getSize(windowWidth, windowHeight);
    ui.resize(windowWidth, windowHeight);

	ok = app.run(sceneView);

	if
		(false == ok)
	{
		LOG( LogLevel::Error, "main"
				, "Could not run application");
		return 1;
	}

	app.shutdown();

	// Shapes and Datasets must be clean BEFORE the call to glfwDestroyWindow in glfw_window.cpp which kill the OpenGL context.
	// If the context is destroyed, OpenGL objects can't be deleted and glGetError return always an error.
    ShapesAllocator::killInstance();
    DatasetsAllocator::killInstance();

    } // Scope for stack-allocated objects

	SEAScope::GUID::implementation = nullptr;
	Path::implementation = nullptr;
    Dialog::implementation = nullptr;
	Log::implementation = nullptr;

    // Memory leaks
    MemoryManager::killInstance();

	return 0;
};
