// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <string>
#include <sstream>
#include <algorithm>

#include "vendors/_sqlite3.h"

#include "core/path.h"
#include "core/log.h"
#include "core/utils.h"
#include "core/source.h"
#include "sqlite_annotations.h"

#include "core/app_event.h"
#include "core/app_events_dispatcher.h"
#include "events/selected_pickables_changed_event.h"
#include "events/selected_pickable_with_info_changed_event.h"
#include "events/annotation_removed_event.h"
#include "events/save_annotation_properties_request.h"
#include "core/measurements.h"
#include "3d/pickable.h"

#include "3d/unit_sphere.h"
#include "3d/geodetic.h"
#include "3d/point.h"
#include "3d/polyline_base.h"
#include "3d/polygon.h"

using namespace SEAScope;

/**
 *
 */
SQLiteAnnotations::SQLiteAnnotations(AppEventsDispatcher& appEventsDispatcher)
: _path(""),
  _appEventsDispatcher(appEventsDispatcher),
  _db(nullptr),
  _countAnnotationsStmt(nullptr),
  _countCollectionsStmt(nullptr)
{
    this->_appEventsDispatcher.subscribe<SelectedPickablesChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<SaveAnnotationPropertiesRequest>(this);
}

/**
 *
 */
SQLiteAnnotations::~SQLiteAnnotations()
{
    this->_appEventsDispatcher.unsubscribe<SelectedPickablesChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<SaveAnnotationPropertiesRequest>(this);


    if
        (nullptr != this->_countAnnotationsStmt)
    {
        sqlite3_finalize(this->_countAnnotationsStmt);
        this->_countAnnotationsStmt = nullptr;
    }

    if
        (nullptr != this->_countCollectionsStmt)
    {
        sqlite3_finalize(this->_countCollectionsStmt);
        this->_countCollectionsStmt = nullptr;
    }

    if
        (nullptr != this->_db)
    {
        sqlite3_close(this->_db);
        this->_db = nullptr;
    }
}

/**
 *
 */
bool SQLiteAnnotations::initializeDatabase()
{
    int rc {-1};    //
    std::ostringstream sql {""};    //
    sqlite3_stmt* stmt {nullptr};   //

    // Open database
    rc = sqlite3_open(this->_path.c_str(), &(this->_db));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLiteAnnotations",
                 "Could not open database {}: {}",
                 this->_path, sqlite3_errmsg(this->_db));
        return false;
    }

    // Configure database to wait 10s before aborting a request due to database
    // locking mechanism.
    sql.str("");
    sql << "PRAGMA busy_timeout=10000";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare SQL statement for setting the busy_timeout parameter: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
       return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Could not set busy_timeout parameter: {}",
                 sqlite3_errmsg(this->_db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Count collections
    sql.str("");
    sql << "SELECT COUNT(*) FROM collections";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_countCollectionsStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLiteAnnotations",
                 "Failed to prepare SQL statement for counting collections: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Count annotations
    sql.str("");
    sql << "SELECT COUNT(*) FROM annotations";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_countAnnotationsStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for counting annotations: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteAnnotations::init(const std::string& pathIn)
{
    std::int64_t size(0);	//
    bool ok(false); //

    this->_path = pathIn;

	ok = Path::fileExists(this->_path);
	if
		(true == ok)
	{
		ok = Path::fileSize(this->_path, size);
		if
            (false == (ok /*reading size worked*/ && 0 >= size /*.. but file is empty*/))
        {
            // Database exists and is not an empty file: get database handle and preprare statements, then exit
            return this->initializeDatabase();
        }
	}

    if
        (false == Path::writeFile(this->_path))
    {
        LOG(LogLevel::Error, "Annotations SQLite", "The annotations file {} is not writable", this->_path);
        return false;
    }

    ok = this->createSQLite(this->_path);

    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLiteAnnotations", "Failed to create database");
        return false;
    }

    return this->initializeDatabase();
}

/**
 *
 */
bool SQLiteAnnotations::get(std::uint64_t granuleId, std::uint32_t& collection, Annotation& annotation) const
{
    bool ok(false);  //
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::ostringstream sql("");	// SQL query
    std::string serialized(""); //

    std::tuple<std::uint64_t, std::uint32_t, std::string> result;   //

	/* Read layout */
	sql << "SELECT * FROM annotations WHERE granule = " << granuleId << ";";
	rc = sqlite3_exec( this->_db
					 , sql.str().c_str()
					 , SQLiteAnnotations::annotationRetrievedCallback
					 , &result
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not retrieve annotation: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

	sqlite3_free(errMsg);
	LOG( LogLevel::Debug, "Annotations SQLite"
			, "Load from SQLite complete");

    collection = std::get<1>(result);
    if
        (0 >= collection)
    {
        // Annotation not found
        return false;
    }

    serialized = std::get<2>(result);

    std::istringstream stream(serialized);

    ok = this->_parser.deserialize(stream, annotation, 0.0, 0.0);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Failed to deserialize annotation {}/{}: {}", collection, granuleId, serialized);
    }

	return ok;
}

/**
 *
 */
bool SQLiteAnnotations::remove(const RenderableId& rId)
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::ostringstream sql("");	// SQL query

	/* Read layout */
    sql << "DELETE FROM annotations WHERE granule = " << rId.granule.second << ";";
	rc = sqlite3_exec( this->_db
					 , sql.str().c_str()
					 , nullptr
					 , nullptr
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not save annotation: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

	sqlite3_free(errMsg);

    AnnotationRemovedEvent event(rId);
    this->_appEventsDispatcher.dispatch<AnnotationRemovedEvent>(event);

    return true;
}

/**
 *
 */
bool SQLiteAnnotations::set(std::uint64_t granuleId, std::uint32_t collectionId, const Annotation& annotation)
{
    bool ok(false); //
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::ostringstream sqlTmp("");	// SQL query

    std::string serialized = this->annotation2string(annotation, ok);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Could not transform annotation into a string");
        return false;
    }


	/* Create or update annotation */
	sqlTmp << "INSERT OR REPLACE INTO annotations(granule, collection, serialized) ";
    sqlTmp << "VALUES(%u, %u, %Q);";

    char* sql = sqlite3_mprintf(sqlTmp.str().c_str(), granuleId, collectionId, serialized.c_str());
	rc = sqlite3_exec( this->_db
					 , sql
					 , nullptr
					 , nullptr
					 , &errMsg);
    sqlite3_free(sql);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not save annotation: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

	sqlite3_free(errMsg);

    return true;
}

std::string SQLiteAnnotations::annotation2string(const Annotation& annotation, bool& ok) const
{
    std::ostringstream stream(""); //
    std::string serialized("");  //

    ok = this->_parser.serialize(annotation, stream);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Failed to serialize annotation");
        return serialized;
    }

    serialized = stream.str();
    std::size_t startIndex = serialized.find('{');
    if
        (std::string::npos != startIndex)
    {
        serialized = serialized.substr(startIndex);
    }
    std::size_t stopIndex = serialized.rfind('}');
    if
        (std::string::npos != stopIndex)
    {
        serialized.resize(stopIndex + 1);
    }

    return serialized;
}

/**
 *
 */
bool SQLiteAnnotations::set(const std::vector<std::uint64_t>& granuleIds, const std::vector<std::uint32_t>& collectionIds, const std::vector<Annotation>& annotations)
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages

    /* Begin transaction */
    rc = sqlite3_exec(this->_db, "BEGIN TRANSACTION", nullptr, nullptr, &errMsg);
    if
        (SQLITE_OK != rc)
    {
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not create transaction: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
    }

    /* Insert all annotations */
    auto granuleIt = granuleIds.cbegin();
    auto collectionIt = collectionIds.cbegin();
    auto annotationIt = annotations.cbegin();
    for
        (; granuleIt != granuleIds.end() && collectionIt != collectionIds.end() && annotationIt != annotations.end(); ++granuleIt, ++collectionIt, ++annotationIt)
    {
        bool ok(false);
        std::ostringstream sqlTmp("");	// SQL query
        std::string serialized = this->annotation2string(*annotationIt, ok);
        if
            (false == ok)
        {
            LOG(LogLevel::Warn, "Annotations SQLITE", "Skipping annotation that could not be serialized into a string");
            continue;
        }

        /* Build SQL statement */
        sqlTmp << "INSERT OR REPLACE INTO annotations(granule, collection, serialized) ";
        sqlTmp << "VALUES(%u, %u, %Q);";

        char* sql = sqlite3_mprintf(sqlTmp.str().c_str(), *granuleIt, *collectionIt, serialized.c_str());

        rc = sqlite3_exec( this->_db
                         , sql
                         , nullptr
                         , nullptr
                         , &errMsg);
        sqlite3_free(sql);
        if
            (SQLITE_OK != rc)
        {
            LOG( LogLevel::Error, "Annotations SQLite"
                    , "Could not save annotations: {}", errMsg);
            sqlite3_free(errMsg);
            sqlite3_exec(this->_db, "ROLLBACK TRANSACTION", nullptr, nullptr, nullptr);
            return false;
        }
    }

    /* Commit transaction */
    rc = sqlite3_exec(this->_db, "COMMIT TRANSACTION", nullptr, nullptr, &errMsg);
    if
        (SQLITE_OK != rc)
    {
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not commit transaction: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
    }

	sqlite3_free(errMsg);

    return true;
}

/**
 *
 */
bool SQLiteAnnotations::sqlite2Annotation(const std::vector<std::tuple<std::uint64_t, std::uint32_t, std::string>>& sqliteRows,
                                          std::vector<std::tuple<std::uint64_t, std::uint32_t, Annotation>>& annotations) const
{
    for
        (auto& it : sqliteRows)
    {
        Annotation annotation;
        std::uint64_t granule = std::get<0>(it);
        std::uint32_t collection = std::get<1>(it);
        std::string serialized = std::get<2>(it);
        bool ok(false);

        std::istringstream stream(serialized);

        ok = this->_parser.deserialize(stream, annotation, 0.0, 0.0);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Annotations SQLite", "Failed to deserialize annotation");
            return false;
        }

        // Polygons stored in the database are completely closed, i.e. the
        // first and the last vertices are the same. This is how it should be
        // when exporting to WKT but it makes thing more difficult in SEAScope
        // so remove the last point
        if
            (ShapeType::POLYGON == annotation.getShape()->getType())
        {
            Polygon* polygon {nullptr};
            polygon = static_cast<Polygon*>(annotation.getShape());
            polygon->deletePoint(polygon->getPointCount() - 1u);
        }

        annotations.push_back(std::make_tuple(std::move(granule), std::move(collection), std::move(annotation)));
    }

	return true;
}

/**
 *
 */
bool SQLiteAnnotations::getall(std::vector<std::tuple<std::uint64_t, std::uint32_t, Annotation>>& annotations) const
{
    bool ok(false);
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::ostringstream sql("");	// SQL query
    std::vector<std::tuple<std::uint64_t, std::uint32_t, std::string>> result;   //

	/* Read layout */
	sql << "SELECT * FROM annotations WHERE 1;";
	rc = sqlite3_exec( this->_db
					 , sql.str().c_str()
					 , SQLiteAnnotations::annotationsRetrievedCallback
					 , &result
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not retrieve annotation: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

	sqlite3_free(errMsg);
	LOG( LogLevel::Debug, "Annotations SQLite"
			, "Load from SQLite complete");

    ok = this->sqlite2Annotation(result, annotations);
    return ok;
}

/**
 *
 */
bool SQLiteAnnotations::applyall(const std::function<void(std::uint64_t, std::uint32_t, Annotation&)>& functor) const
{
    std::vector<std::tuple<std::uint64_t, std::uint32_t, Annotation>> annotations;  //
    bool ok(false); //

    ok = this->getall(annotations);
    if
        (false == ok)
    {
        return false;
    }

    for
        (auto& it : annotations)
    {
        functor(std::get<0>(it), std::get<1>(it), std::get<2>(it));
    }

    return true;
}

/**
 *
 */
int SQLiteAnnotations::annotationsRetrievedCallback(void* result, int argc, char** argv, char** colName)
{
    std::vector<std::tuple<std::uint64_t, std::uint32_t, std::string>>* annotations(nullptr);
    std::tuple<std::uint64_t, std::uint32_t, std::string> annotation;   //
    int rc(-1); //

    annotations = reinterpret_cast<std::vector<std::tuple<std::uint64_t, std::uint32_t, std::string>>*>(result);
    if
        (nullptr == annotations)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Could not cast result to the correct type");
        return EXIT_FAILURE;
    }

    rc = SQLiteAnnotations::annotationRetrievedCallback(&annotation, argc, argv, colName);
    annotations->push_back(annotation);

    return rc;
}

/**
 *
 */
int SQLiteAnnotations::annotationRetrievedCallback(void* result, int argc, char** argv, char** colName)
{
	std::tuple<std::uint64_t, std::uint32_t, std::string>* annotation(nullptr);	//
    std::uint64_t granule(0);   //
    std::uint32_t collection(0);    //
    std::string serialized(""); //
    int i(0);   //

	annotation = reinterpret_cast<std::tuple<std::uint64_t, std::uint32_t, std::string>*>(result);
	if
		(nullptr == annotation)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not apply results to annotation");
		return EXIT_FAILURE;
	}

	i = 0;
	for
		(; i<argc; ++i)
	{
		if
			(0 == std::strncmp(colName[i], "serialized", 10))
		{
			serialized = argv[i];
		}
		else if
			(0 == std::strncmp(colName[i], "granule", 7))
		{
			granule = static_cast<std::uint64_t>(std::stol(argv[i]));
		}
		else if
			(0 == std::strncmp(colName[i], "collection", 10))
		{
			collection = static_cast<std::uint32_t>(std::stoi(argv[i]));
		}
	}

    std::get<0>(*annotation) = granule;
    std::get<1>(*annotation) = collection;
    std::get<2>(*annotation) = serialized;

    return EXIT_SUCCESS;
}

/**
 *
 */
bool SQLiteAnnotations::createSQLite(const std::string& uri) const
{
	sqlite3* db(nullptr);	// SQLite3 database handle
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::string sql("");	// SQL query

	/* Open database */
	rc = sqlite3_open(uri.c_str(), &db);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not open database {}: {}", uri, sqlite3_errmsg(db));

		Path::deleteFile(uri);
		return false;
	}

    sql = "CREATE TABLE collections(" \
          "collection INTEGER NOT NULL," \
          "label TEXT NOT NULL," \
          "xSeamless INTEGER NOT NULL," \
          "ySeamless INTEGER NOT NULL," \
          "NEWSAligned INTEGER NOT NULL," \
          "mustBeCurrent INTEGER NOT NULL," \
          "tags TEXT NOT NULL," \
          "PRIMARY KEY(collection))";

    rc = sqlite3_exec( db
                     , sql.c_str()
                     , nullptr
                     , nullptr
                     , &errMsg);
    if
        (SQLITE_OK != rc)
    {
        LOG( LogLevel::Error, "Annotations SQLite"
                , "Could not initialize database: {}", errMsg);
        LOG( LogLevel::Debug, "Annotations SQLite"
                , "Failed SQL query: {}", sql);
        sqlite3_free(errMsg);
        sqlite3_close(db);

		Path::deleteFile(uri);
        return false;
    }

    sql = "CREATE TABLE variables(" \
          "variable INTEGER NOT NULL," \
          "collection INTEGER NOT NULL," \
          "label TEXT NOT NULL," \
          "units TEXT NOT NULL," \
          "fields TEXT NOT NULL," \
          "tags TEXT NOT NULL," \
          "rcfgLogscale INTEGER NOT NULL," \
          "rcfgMin REAL NOT NULL," \
          "rcfgMax REAL NOT NULL," \
          "rcfgZindex REAL NOT NULL," \
          "rcfgOpacity REAL NOT NULL," \
          "rcfgColorRed INTEGER NOT NULL," \
          "rcfgColorGreen INTEGER NOT NULL," \
          "rcfgColorBlue INTEGER NOT NULL," \
          "rcfgColormap TEXT NOT NULL," \
          "rcfgRenderMethod TEXT NOT NULL," \
          "rcfgParticlesCount INTEGER NOT NULL," \
          "rcfgParticleTTL INTEGER NOT NULL," \
          "rcfgStreamlineLength INTEGER NOT NULL," \
          "rcfgStreamlineSpeed REAL NOT NULL," \
          "rcfgFilterMode TEXT NOT NULL," \
          "rcfgLineThickness REAL NOT NULL," \
          "PRIMARY KEY(variable, collection)," \
          "CONSTRAINT fk_collection FOREIGN KEY (collection) REFERENCES collections (collection) ON UPDATE CASCADE ON DELETE CASCADE)";

    rc = sqlite3_exec( db
                     , sql.c_str()
                     , nullptr
                     , nullptr
                     , &errMsg);
    if
        (SQLITE_OK != rc)
    {
        LOG( LogLevel::Error, "Annotations SQLite"
                , "Could not initialize database: {}", errMsg);
        LOG( LogLevel::Debug, "Annotations SQLite"
                , "Failed SQL query: {}", sql);
        sqlite3_free(errMsg);
        sqlite3_close(db);

		Path::deleteFile(uri);
        return false;
    }

    sql = "CREATE TABLE annotations(" \
          "granule INTEGER NOT NULL," \
          "collection INTEGER NOT NULL," \
          "serialized TEXT NOT NULL," \
          "PRIMARY KEY(granule)," \
          "CONSTRAINT fk_collection FOREIGN KEY (collection) REFERENCES collections (collection) ON UPDATE CASCADE ON DELETE CASCADE)";

    rc = sqlite3_exec( db
                     , sql.c_str()
                     , nullptr
                     , nullptr
                     , &errMsg);
    if
        (SQLITE_OK != rc)
    {
        LOG( LogLevel::Error, "Annotations SQLite"
                , "Could not initialize database: {}", errMsg);
        LOG( LogLevel::Debug, "Annotations SQLite"
                , "Failed SQL query: {}", sql);
        sqlite3_free(errMsg);
        sqlite3_close(db);

		Path::deleteFile(uri);
        return false;
    }

    sqlite3_free(errMsg);
	sqlite3_close(db);
	return true;
}

/**
 *
 */
bool SQLiteAnnotations::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(SelectedPickablesChangedEvent)) == eventType)
    {
        GranuleInfo info;

        const std::vector<Pickable*>& selected = static_cast<const SelectedPickablesChangedEvent&>(event).getSelected();
        if
            (1 == selected.size())
        {
            const RenderableId& rId = selected[0]->renderableId();

            if
                (SEAScope::Source::ANNOTATIONS != rId.collection.first)
            {
                return false;
            }

            std::uint32_t collection;
            Annotation annotation;
            bool ok(false);

            ok = this->get(rId.granule.second, collection, annotation);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "Annotations SQLite",
                         "Failed to get annotation {} from local database", rId.granule.second);
                return true;  // event has been processed, even if it failed
            }

            ok = this->getCollection(collection, info.collection);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "Annotations SQLite",
                         "FAied to get collection {} for annotation {} from local database", collection, rId.granule.second);
                return true;  // event has been processed, even if it failed
            }

            info.variable = info.collection.variables[rId.variable];

            ShapeType shapeType = annotation.getShape()->getType(); //

            if
                ((ShapeType::POLYLINE != shapeType) && (ShapeType::POLYGON != shapeType) && (ShapeType::ELLIPSE != shapeType) && (ShapeType::POINT != shapeType))
            {
                LOG(LogLevel::Error, "Annotations SQLite",
                         "Shape type not supported: {}", std::to_string(shapeType));
                return true;  // event has been processed, even if it failed
            }

            info.perimeter = std::numeric_limits<float>::quiet_NaN();
            info.area = std::numeric_limits<float>::quiet_NaN();
            info.lon = std::numeric_limits<float>::quiet_NaN();
            info.lat = std::numeric_limits<float>::quiet_NaN();

            RenderableType rType;
            if
                (ShapeType::POLYLINE == shapeType)
            {
                rType = RenderableType::POLYLINE;
                info.perimeter = static_cast<float>(SEAScope::perimeterFromShape(annotation.getShape(), false /* do not join last and first vertices */));

                const PolylineBase* polyline = dynamic_cast<const PolylineBase*>(annotation.getShape());
                const std::vector<glm::vec3>& xyz = polyline->getPoints();
                if
                    (   (2 < xyz.size())
                     && (xyz[0] == xyz[xyz.size() - 1]))
                {
                    info.area = static_cast<float>(SEAScope::areaFromShape(annotation.getShape()));
                }

                polyline = nullptr;
            }
            else if
                (ShapeType::POLYGON == shapeType)
            {
                rType = RenderableType::POLYGON;
                info.area = static_cast<float>(SEAScope::areaFromShape(annotation.getShape()));
                info.perimeter = static_cast<float>(SEAScope::perimeterFromShape(annotation.getShape(), true /* close shape */));
            }
            else if
                (ShapeType::ELLIPSE == shapeType)
            {
                rType = RenderableType::ELLIPSE;
                info.area = static_cast<float>(SEAScope::areaFromShape(annotation.getShape()));
                info.perimeter = static_cast<float>(SEAScope::perimeterFromShape(annotation.getShape(), true /* close shape */));
            }
            else if
                (ShapeType::POINT == shapeType)
            {
                rType = RenderableType::POINT;
                glm::vec3 position = static_cast<Point*>(annotation.getShape())->getPosition();
                Geodetic2D geo = UnitSphere::ToGeodetic2D(position);
                info.lon = static_cast<float>(180.0 * geo.Longitude() / SEAScope::PI);
                info.lat = static_cast<float>(180.0 * geo.Latitude() / SEAScope::PI);
            }

            SelectedPickableWithInfoChangedEvent outEvent(selected[0], info, rType, &(annotation.getProperties()), annotation.getStart(), annotation.getStop());
            this->_appEventsDispatcher.dispatch<SelectedPickableWithInfoChangedEvent>(outEvent);
        }

        return true;
    }
    else if
        (std::type_index(typeid(SaveAnnotationPropertiesRequest)) == eventType)
    {
        const RenderableId& renderableId = static_cast<const SaveAnnotationPropertiesRequest&>(event).getRenderableId();
        const DictNode& properties = static_cast<const SaveAnnotationPropertiesRequest&>(event).getProperties();

        this->saveAnnotationProperties(renderableId, properties);

        return true;
    }

    return false;
}


/**
 *
 */
bool SQLiteAnnotations::setCollection(std::uint32_t collectionId,
                                      const CollectionCfg& collectionCfg,
                                      const std::vector<RenderingCfg*>& renderingCfgs)
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::ostringstream sqlTmp("");	// SQL query

    /* Begin transaction */
    rc = sqlite3_exec(this->_db, "BEGIN TRANSACTION", nullptr, nullptr, &errMsg);
    if
        (SQLITE_OK != rc)
    {
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not create transaction: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
    }

    /* Create or update collection */
    std::ostringstream serializedTags("");
    for
        (const auto& tag: collectionCfg.tags)
    {
        serializedTags << tag.first << ":" << tag.second << ",";
    }
	sqlTmp << "INSERT OR REPLACE INTO collections(collection, label, xSeamless, ySeamless, NEWSAligned, mustBeCurrent, tags) ";
    sqlTmp << "VALUES(%u, %Q, %u, %u, %u, %u, %Q);";

    char* sql = sqlite3_mprintf(sqlTmp.str().c_str(), collectionId,
                                                      collectionCfg.label.c_str(),
                                                      collectionCfg.xSeamless,
                                                      collectionCfg.ySeamless,
                                                      false /*NEWSAligned - deprecated*/,
                                                      collectionCfg.mustBeCurrent,
                                                      serializedTags.str().c_str());
	rc = sqlite3_exec( this->_db
					 , sql
					 , nullptr
					 , nullptr
					 , &errMsg);
    sqlite3_free(sql);
	if
		(SQLITE_OK != rc)
	{
        LOG( LogLevel::Error, "Annotations SQLite"
                , "Could not save collection: {}", errMsg);
        sqlite3_free(errMsg);
        sqlite3_exec(this->_db, "ROLLBACK TRANSACTION", nullptr, nullptr, nullptr);
        return false;
	}

    /* Create or update variables */
    sqlTmp.str("");
    sqlTmp << "INSERT OR REPLACE INTO variables(variable, collection, label, units, fields, tags, ";
    sqlTmp << "rcfgLogscale, rcfgMin, rcfgMax, rcfgZindex, rcfgOpacity, ";
    sqlTmp << "rcfgColorRed, rcfgColorGreen, rcfgColorBlue, rcfgColormap, ";
    sqlTmp << "rcfgRenderMethod, rcfgParticlesCount, rcfgParticleTTL,";
    sqlTmp << "rcfgStreamlineLength, rcfgStreamlineSpeed, rcfgLineThickness, rcfgFilterMode)";
    sqlTmp << " VALUES(%u, %u, %Q, %Q, %Q, %Q, %u, %f, %f, %f, %f, %u, %u, %u, %Q, %Q, %u, %u, %u, %f, %f, %Q);";
    for
        (const auto& rCfg : renderingCfgs)
    {
        const VariableCfg& varCfg = collectionCfg.variables.at(rCfg->target.variable);

        std::ostringstream serializedVarTags("");
        for
            (const auto& tag: varCfg.tags)
        {
            serializedVarTags << tag.first << ":" << tag.second << ",";
        }

        std::ostringstream serializedVarFields("");
        for
            (const auto& field: varCfg.fields)
        {
            serializedVarFields << field << ",";
        }

        std::string renderMethodStr("");
        switch
            (rCfg->renderMethod())
        {
		    case RenderMethod::RASTER:
                renderMethodStr = "RASTER";
                break;
            case RenderMethod::ARROWS:
                renderMethodStr = "ARROWS";
                break;
            case RenderMethod::BARBS:
                renderMethodStr = "BARBS";
                break;
            case RenderMethod::TRAJECTORIES:
                renderMethodStr = "TRAJECTORIES";
                break;
            case RenderMethod::DOTSCLOUD:
                renderMethodStr = "DOTSCLOUD";
                break;
            case RenderMethod::STREAMLINES:
                renderMethodStr = "STREAMLINES";
                break;
            case RenderMethod::RAWRGB:
                renderMethodStr = "RAWRGB";
                break;
            case RenderMethod::POLYLINE:
                renderMethodStr = "POLYLINE";
                break;
            case RenderMethod::POLYGON:
                renderMethodStr = "POLYGON";
                break;
            case RenderMethod::ELLIPSE:
                renderMethodStr = "ELLIPSE";
                break;
            case RenderMethod::POINT:
                renderMethodStr = "POINT";
                break;
        }

        std::string filterModeStr("");
        switch
            (rCfg->filterMode())
        {
            case FilterMode::NEAREST:
                filterModeStr = "NEAREST";
                break;
            case FilterMode::BILINEAR:
                filterModeStr = "BILINEAR";
                break;
        }

        char* sqlVar = sqlite3_mprintf(sqlTmp.str().c_str(),
                                       rCfg->target.variable,
                                       collectionId,
                                       varCfg.label.c_str(),
                                       varCfg.units.c_str(),
                                       serializedVarFields.str().c_str(),
                                       serializedVarTags.str().c_str(),
                                       rCfg->logscale(),
                                       static_cast<double>(rCfg->min()),
                                       static_cast<double>(rCfg->max()),
                                       static_cast<double>(rCfg->zindex()),
                                       static_cast<double>(rCfg->opacity()),
                                       rCfg->color().at(0),
                                       rCfg->color().at(1),
                                       rCfg->color().at(2),
                                       rCfg->colormap().c_str(),
                                       renderMethodStr.c_str(),
                                       rCfg->particlesCount(),
                                       rCfg->particleTTL(),
                                       rCfg->streamlineLength(),
                                       static_cast<double>(rCfg->streamlineSpeed()),
                                       static_cast<double>(rCfg->lineThickness()),
                                       filterModeStr.c_str());
        rc = sqlite3_exec( this->_db
                         , sqlVar
                         , nullptr
                         , nullptr
                         , &errMsg);
        sqlite3_free(sqlVar);
        if
            (SQLITE_OK != rc)
        {
            LOG( LogLevel::Error, "Annotations SQLite"
                    , "Could not save variable: {}", errMsg);
            sqlite3_free(errMsg);
            sqlite3_exec(this->_db, "ROLLBACK TRANSACTION", nullptr, nullptr, nullptr);
            return false;
        }
    }

    /* Commit transaction */
    rc = sqlite3_exec(this->_db, "COMMIT TRANSACTION", nullptr, nullptr, &errMsg);
    if
        (SQLITE_OK != rc)
    {
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not commit transaction: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
    }

	sqlite3_free(errMsg);

    return true;
}

/**
 *
 */
bool SQLiteAnnotations::removeCollection(std::uint32_t collectionId)
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::ostringstream sqlTmp("");	// SQL query

    /* Delete collection */
	sqlTmp << "PRAGMA foreign_keys = ON;";
    sqlTmp << "DELETE FROM collections where collection = %u;";
    char* sql = sqlite3_mprintf(sqlTmp.str().c_str(), collectionId);

	rc = sqlite3_exec( this->_db
					 , sql
					 , nullptr
					 , nullptr
					 , &errMsg);
    sqlite3_free(sql);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not delete collection: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

    sqlite3_free(errMsg);

    return true;
}

bool SQLiteAnnotations::listCollectionGranules(std::uint32_t collectionId, std::vector<std::uint64_t>& result) const
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::ostringstream sql("");	// SQL query

	/* Read layout */
	sql << "SELECT granule FROM annotations WHERE collection = " << collectionId << ";";
	rc = sqlite3_exec( this->_db
					 , sql.str().c_str()
					 , SQLiteAnnotations::granuleIdRetrievedCallback
					 , &result
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		LOG(LogLevel::Error, "Annotations SQLite",
				 "Could not retrieve granule ids for collection '{}': {}",
                 collectionId, errMsg);
		sqlite3_free(errMsg);
		return false;
	}

	sqlite3_free(errMsg);
	LOG( LogLevel::Debug, "Annotations SQLite"
			, "Load from SQLite complete");

    return true;
}

/**
 *
 */
int SQLiteAnnotations::granuleIdRetrievedCallback(void* result, int argc, char** argv, char** /*colName*/)
{
    std::uint64_t value(0);
    std::vector<std::uint64_t>* granuleIds(nullptr);

    granuleIds = reinterpret_cast<std::vector<std::uint64_t>*>(result);
    if
        (nullptr == granuleIds)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Could not cast result to the correct type");
        return EXIT_FAILURE;
    }

    if
        ((1 == argc) && (nullptr != argv[0]))
    {
        value = static_cast<std::uint64_t>(std::stoull(argv[0]));
        granuleIds->push_back(value);
    }
    else
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Invalid number of columns to parse: {} instead of 1", argc);
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}


int SQLiteAnnotations::maxCollectionIdRetrievedCallback(void* result, int argc, char** argv, char** /*column*/)
{
    std::uint32_t* count;

	count = reinterpret_cast<std::uint32_t*>(result);
	if
		(nullptr == count)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not apply results to max collection ID");
		return EXIT_FAILURE;
	}

    if
        ((1 == argc) && (nullptr != argv[0]))
    {
        (*count) = static_cast<std::uint32_t>(std::stoi(argv[0]));
    }

    return EXIT_SUCCESS;
}

/**
 *
 */
bool SQLiteAnnotations::getMaxCollectionId(std::uint32_t& result) const
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
    std::string sql("");
    std::uint32_t tmp(result);

    /* Get max collection ID */
	sql = "SELECT MAX(collection) FROM collections;";

	rc = sqlite3_exec( this->_db
					 , sql.c_str()
					 , SQLiteAnnotations::maxCollectionIdRetrievedCallback
					 , &tmp
					 , &errMsg);

	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not get max collection ID: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

    sqlite3_free(errMsg);

    result = tmp;
    return true;
}

/**
 *
 */
bool SQLiteAnnotations::getMaxGranuleId(std::uint64_t& result) const
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
    std::string sql("");
    std::uint64_t tmp(result);

    /* Get max collection ID */
	sql = "SELECT MAX(granule) FROM annotations;";

	rc = sqlite3_exec( this->_db
					 , sql.c_str()
					 , SQLiteAnnotations::maxCollectionIdRetrievedCallback
					 , &tmp
					 , &errMsg);

	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not get max granule ID: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

    sqlite3_free(errMsg);

    result = tmp;
    return true;
}

/**
 *
 */
int SQLiteAnnotations::collectionsRetrievedCallback(void* result, int argc, char** argv, char** colName)
{
    std::vector<std::tuple<std::uint32_t, CollectionCfg, std::vector<RenderingCfg>>>* collections(nullptr);
    int i(0);   //

    collections = reinterpret_cast<std::vector<std::tuple<std::uint32_t, CollectionCfg, std::vector<RenderingCfg>>>*>(result);
    if
        (nullptr == collections)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Could not cast result to the correct type");
        return EXIT_FAILURE;
    }

    collections->emplace(collections->end());
    std::tuple<std::uint32_t, CollectionCfg, std::vector<RenderingCfg>>& collection = collections->back();
    std::string tagsStr;
	i = 0;
	for
		(; i<argc; ++i)
	{
		if
			(0 == std::strncmp(colName[i], "collection", 10))
		{
			std::get<0>(collection) = static_cast<std::uint32_t>(std::stoi(argv[i]));
		}
        else if
            (0 == std::strncmp(colName[i], "label", 5))
        {
            std::get<1>(collection).label = argv[i];
        }
        else if
            (0 == std::strncmp(colName[i], "xSeamless", 9))
        {
            std::get<1>(collection).xSeamless = static_cast<bool>(std::stoi(argv[i]));
        }
        else if
            (0 == std::strncmp(colName[i], "ySeamless", 9))
        {
            std::get<1>(collection).ySeamless = static_cast<bool>(std::stoi(argv[i]));
        }
        else if
            (0 == std::strncmp(colName[i], "mustBeCurrent", 13))
        {
            std::get<1>(collection).mustBeCurrent = static_cast<bool>(std::stoi(argv[i]));
        }
        else if
            (0 == std::strncmp(colName[i], "tags", 4))
        {
            tagsStr = argv[i];
        }
	}

    // Parse tags
    std::vector<std::string> tagsList;
    SEAScope::explodeString(tagsStr, ",", tagsList);
    for
        (const std::string& tagKeyValueStr : tagsList)
    {
        std::vector<std::string> keyValue(2);
        SEAScope::explodeString(tagKeyValueStr, ":", keyValue);
        std::get<1>(collection).tags[keyValue[0]] = keyValue[1];
    }

    return EXIT_SUCCESS;
}

/**
 *
 */
int SQLiteAnnotations::collectionRetrievedCallback(void* result, int argc, char** argv, char** colName)
{
    CollectionCfg* colCfg(nullptr);
    int i(0);   //

    colCfg = reinterpret_cast<CollectionCfg*>(result);
    if
        (nullptr == colCfg)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Could not cast result to the correct type");
        return EXIT_FAILURE;
    }

    std::string tagsStr;
	i = 0;
	for
		(; i<argc; ++i)
	{
        if
            (0 == std::strncmp(colName[i], "label", 5))
        {
            colCfg->label = argv[i];
        }
        else if
            (0 == std::strncmp(colName[i], "xSeamless", 9))
        {
            colCfg->xSeamless = static_cast<bool>(std::stoi(argv[i]));
        }
        else if
            (0 == std::strncmp(colName[i], "ySeamless", 9))
        {
            colCfg->ySeamless = static_cast<bool>(std::stoi(argv[i]));
        }
        else if
            (0 == std::strncmp(colName[i], "mustBeCurrent", 13))
        {
            colCfg->mustBeCurrent = static_cast<bool>(std::stoi(argv[i]));
        }
        else if
            (0 == std::strncmp(colName[i], "tags", 4))
        {
            tagsStr = argv[i];
        }
	}

    // Parse tags
    std::vector<std::string> tagsList;
    SEAScope::explodeString(tagsStr, ",", tagsList);
    for
        (const std::string& tagKeyValueStr : tagsList)
    {
        std::vector<std::string> keyValue(2);
        SEAScope::explodeString(tagKeyValueStr, ":", keyValue);
        colCfg->tags[keyValue[0]] = keyValue[1];
    }

    return EXIT_SUCCESS;
}

/**
 *
 */
int SQLiteAnnotations::variablesRetrievedCallback(void* result, int argc, char** argv, char** colName)
{
    std::vector<std::tuple<std::uint32_t, CollectionCfg, std::vector<RenderingCfg>>>* collections(nullptr);
    std::uint32_t variable(0u);
    std::uint32_t collection(0u);
    std::string label;
    std::string units;
    bool logscale(false);
    float min(0.f);
    float max(0.f);
    float opacity(0.f);
    float zindex(0.f);
    std::uint8_t red(0u);
    std::uint8_t green(0u);
    std::uint8_t blue(0u);
    std::string colormap;
    uint32_t particlesCount(0u);
    uint8_t particleTTL(0u);
    uint32_t streamlineLength(0u);
    float streamlineSpeed(1.f);
    float lineThickness(1.f);
    std::string renderMethodStr;
    std::string filterModeStr;
    std::string fieldsStr;
    std::string tagsStr;
    int i(0);   //

    collections = reinterpret_cast<std::vector<std::tuple<std::uint32_t, CollectionCfg, std::vector<RenderingCfg>>>*>(result);
    if
        (nullptr == collections)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Could not cast result to the correct type");
        return EXIT_FAILURE;
    }

	i = 0;
	for
		(; i<argc; ++i)
	{
        if
            (0 == std::strncmp(colName[i], "variable", 8))
        {
            variable = static_cast<std::uint32_t>(std::atoi(argv[i]));
        }
		else if
			(0 == std::strncmp(colName[i], "collection", 10))
		{
			collection = static_cast<std::uint32_t>(std::stoi(argv[i]));
		}
        else if
            (0 == std::strncmp(colName[i], "label", 5))
        {
            label = argv[i];
        }
        else if
            (0 == std::strncmp(colName[i], "units", 5))
        {
            units = argv[i];
        }
        else if
            (0 == std::strncmp(colName[i], "fields", 6))
        {
            fieldsStr = argv[i];
        }
        else if
            (0 == std::strncmp(colName[i], "tags", 4))
        {
            tagsStr = argv[i];
        }
        else if
            (0 == std::strncmp(colName[i], "rcfgLogscale", 12))
        {
            logscale = static_cast<bool>(std::atoi(argv[i]));
        }
        else if
            (0 == std::strncmp(colName[i], "rcfgMin", 7))
        {
            min = static_cast<float>(std::atof(argv[i]));
        }
        else if
            (0 == std::strncmp(colName[i], "rcfgMax", 7))
        {
            max = static_cast<float>(std::atof(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgZindex", 10))
        {
            zindex = static_cast<float>(std::atof(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgOpacity", 11))
        {
            opacity = static_cast<float>(std::atof(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgColorRed", 12))
        {
            red = static_cast<std::uint8_t>(std::atoi(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgColorGreen", 14))
        {
            green = static_cast<std::uint8_t>(std::atoi(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgColorBlue", 13))
        {
            blue = static_cast<std::uint8_t>(std::atoi(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgColormap", 12))
        {
            colormap = argv[i];
        }
        else if
            (0 == strncmp(colName[i], "rcfgRenderMethod", 16))
        {
            renderMethodStr = argv[i];
        }
        else if
            (0 == strncmp(colName[i], "rcfgParticlesCount", 18))
        {
            particlesCount = static_cast<std::uint32_t>(std::atoi(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgParticleTTL", 15))
        {
            particleTTL = static_cast<std::uint8_t>(std::atoi(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgStreamlineLength", 20))
        {
            streamlineLength = static_cast<std::uint32_t>(std::atoi(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgStreamlineSpeed", 19))
        {
            streamlineSpeed = static_cast<float>(std::atof(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgLineThickness", 17))
        {
            lineThickness = static_cast<float>(std::atof(argv[i]));
        }
        else if
            (0 == strncmp(colName[i], "rcfgFilterMode", 14))
        {
            filterModeStr = argv[i];
        }
	}

    std::array<std::uint8_t, 3> color({ {red, green, blue} });

    RenderMethod renderMethod(RenderMethod::RASTER);
    if
        ("RASTER" == renderMethodStr)
    {
        renderMethod = RenderMethod::RASTER;
    }
    else if
        ("ARROWS" == renderMethodStr)
    {
        renderMethod = RenderMethod::ARROWS;
    }
    else if
        ("BARBS" == renderMethodStr)
    {
        renderMethod = RenderMethod::BARBS;
    }
    else if
        ("TRAJECTORIES" == renderMethodStr)
    {
        renderMethod = RenderMethod::TRAJECTORIES;
    }
    else if
        ("DOTSCLOUD" == renderMethodStr)
    {
        renderMethod = RenderMethod::DOTSCLOUD;
    }
    else if
        ("STREAMLINES" == renderMethodStr)
    {
        renderMethod = RenderMethod::STREAMLINES;
    }
    else if
        ("RAWRGB" == renderMethodStr)
    {
        renderMethod = RenderMethod::RAWRGB;
    }
    else if
        ("POLYLINE" == renderMethodStr)
    {
        renderMethod = RenderMethod::POLYLINE;
    }
    else if
        ("POLYGON" == renderMethodStr)
    {
        renderMethod = RenderMethod::POLYGON;
    }
    else if
        ("ELLIPSE" == renderMethodStr)
    {
        renderMethod = RenderMethod::ELLIPSE;
    }
    else if
        ("POINT" == renderMethodStr)
    {
        renderMethod = RenderMethod::POINT;
    }

    FilterMode filterMode(FilterMode::NEAREST);
    if
        ("NEAREST" == filterModeStr)
    {
        filterMode = FilterMode::NEAREST;
    }
    else if
        ("BILINEAR" == filterModeStr)
    {
        filterMode = FilterMode::BILINEAR;
    }

    // Assign parsed info to result object
    for
        (auto& collectionTuple : *collections)
    {
        if
            (collection != std::get<0>(collectionTuple))
        {
            continue;
        }

        CollectionCfg& colCfg = std::get<1>(collectionTuple);

        std::vector<RenderingCfg>& renderings = std::get<2>(collectionTuple);
        renderings.emplace(renderings.end());

        RenderingCfg& rCfg = renderings.back();
        rCfg.target.granuleLevel = false;
        rCfg.target.collection = {0u /* source will be set later */, collection};
        rCfg.target.variable = variable;
        rCfg.setLogscale(logscale);
        rCfg.setMin(min);
        rCfg.setMax(max);
        rCfg.setOpacity(opacity);
        rCfg.setZindex(zindex);
        rCfg.setColor(color);
        rCfg.setColormap(colormap); // must be done AFTER setColor because setColor resets colormap
        rCfg.setRenderMethod(renderMethod);
        rCfg.setParticlesCount(particlesCount);
        rCfg.setParticleTTL(particleTTL);
        rCfg.setStreamlineLength(streamlineLength);
        rCfg.setStreamlineSpeed(streamlineSpeed);
        rCfg.setFilterMode(filterMode);
        rCfg.setLineThickness(lineThickness);

        colCfg.variables[variable];
        colCfg.variables[variable].label = label;
        colCfg.variables[variable].units = units;

        // Parse fields
        std::vector<std::string> fieldsList;
        SEAScope::explodeString(fieldsStr, ",", fieldsList);
        fieldsList.erase(std::remove_if(fieldsList.begin(), fieldsList.end(),
                                        [](const std::string& item) -> bool { return 0 >= item.size(); }),
                         fieldsList.end());
        colCfg.variables[variable].fields.insert(fieldsList.begin(),
                                                 fieldsList.end());

        // Parse tags
        std::vector<std::string> tagsList;
        SEAScope::explodeString(tagsStr, ",", tagsList);
        for
            (const std::string& tagKeyValueStr : tagsList)
        {
            std::vector<std::string> keyValue(2);
            SEAScope::explodeString(tagKeyValueStr, ":", keyValue);
            colCfg.variables[variable].tags[keyValue[0]] = keyValue[1];
        }

        break;
    }

    return EXIT_SUCCESS;
}

/**
 *
 */
int SQLiteAnnotations::variablesWithoutRenderingRetrievedCallback(void* result, int argc, char** argv, char** colName)
{
    CollectionCfg* colCfg(nullptr);
    std::uint32_t variable(0u);
    std::string label;
    std::string units;
    std::string renderMethodStr;
    std::string fieldsStr;
    std::string tagsStr;
    int i(0);   //

    colCfg = reinterpret_cast<CollectionCfg*>(result);
    if
        (nullptr == colCfg)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "Could not cast result to the correct type");
        return EXIT_FAILURE;
    }

	i = 0;
	for
		(; i<argc; ++i)
	{
        if
            (0 == std::strncmp(colName[i], "variable", 8))
        {
            variable = static_cast<std::uint32_t>(std::atoi(argv[i]));
        }
        /*
		else if
			(0 == std::strncmp(colName[i], "collection", 10))
		{
			collection = static_cast<std::uint32_t>(std::stoi(argv[i]));
		}
        */
        else if
            (0 == std::strncmp(colName[i], "label", 5))
        {
            label = argv[i];
        }
        else if
            (0 == std::strncmp(colName[i], "units", 5))
        {
            units = argv[i];
        }
        else if
            (0 == std::strncmp(colName[i], "fields", 6))
        {
            fieldsStr = argv[i];
        }
        else if
            (0 == std::strncmp(colName[i], "tags", 4))
        {
            tagsStr = argv[i];
        }
        else if
            (0 == strncmp(colName[i], "rcfgRenderMethod", 16))
        {
            renderMethodStr = argv[i];
        }
	}

    colCfg->variables[variable];
    colCfg->variables[variable].label = label;
    colCfg->variables[variable].units = units;

    // Parse fields
    std::vector<std::string> fieldsList;
    SEAScope::explodeString(fieldsStr, ",", fieldsList);
    fieldsList.erase(std::remove_if(fieldsList.begin(), fieldsList.end(),
                                    [](const std::string& item) -> bool { return 0 >= item.size(); }),
                     fieldsList.end());
    colCfg->variables[variable].fields.insert(fieldsList.begin(),
                                              fieldsList.end());

    // Parse tags
    std::vector<std::string> tagsList;
    SEAScope::explodeString(tagsStr, ",", tagsList);
    for
        (const std::string& tagKeyValueStr : tagsList)
    {
        std::vector<std::string> keyValue(2);
        SEAScope::explodeString(tagKeyValueStr, ":", keyValue);
        colCfg->variables[variable].tags[keyValue[0]] = keyValue[1];
    }

    return EXIT_SUCCESS;
}

/**
 *
 */
bool SQLiteAnnotations::getCollection(std::uint32_t collectionId, CollectionCfg& result) const
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::ostringstream sql("");	// SQL query

	/* Read collections */
	sql << "SELECT * FROM collections WHERE collection = " << collectionId << ";";
	rc = sqlite3_exec( this->_db
					 , sql.str().c_str()
					 , SQLiteAnnotations::collectionRetrievedCallback
					 , &result
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not retrieve annotation: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

    /* Read variables */
    sql.str("");
    sql << "SELECT * FROM variables WHERE collection = " << collectionId << ";";
	rc = sqlite3_exec( this->_db
					 , sql.str().c_str()
					 , SQLiteAnnotations::variablesWithoutRenderingRetrievedCallback
					 , &result
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not retrieve annotation: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

	sqlite3_free(errMsg);
	LOG( LogLevel::Debug, "Annotations SQLite"
			, "Load from SQLite complete");

    return true;
}
/**
 *
 */
bool SQLiteAnnotations::getAllCollections(std::vector<std::tuple<std::uint32_t, CollectionCfg, std::vector<RenderingCfg>>>& result) const
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::ostringstream sql("");	// SQL query

	/* Read collections */
	sql << "SELECT * FROM collections WHERE 1;";
	rc = sqlite3_exec( this->_db
					 , sql.str().c_str()
					 , SQLiteAnnotations::collectionsRetrievedCallback
					 , &result
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not retrieve annotation: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

    /* Read variables */
    sql.str("");
    sql << "SELECT * FROM variables WHERE 1";
	rc = sqlite3_exec( this->_db
					 , sql.str().c_str()
					 , SQLiteAnnotations::variablesRetrievedCallback
					 , &result
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not retrieve annotation: {}", errMsg);
		sqlite3_free(errMsg);
		return false;
	}

	sqlite3_free(errMsg);
	LOG( LogLevel::Debug, "Annotations SQLite"
			, "Load from SQLite complete");

    return true;
}

/**
 *
 */
bool SQLiteAnnotations::applyAllCollections(const std::function<void(std::uint32_t, CollectionCfg&, std::vector<RenderingCfg>&)>& functor) const
{
    std::vector<std::tuple<std::uint32_t, CollectionCfg, std::vector<RenderingCfg>>> collections;  //
    bool ok(false); //

    ok = this->getAllCollections(collections);
    if
        (false == ok)
    {
        std::cerr << "getAllCollections failed" << std::endl;
        return false;
    }

    for
        (auto& it : collections)
    {
        functor(std::get<0>(it), std::get<1>(it), std::get<2>(it));
    }

    return true;
}

/**
 *
 */
bool SQLiteAnnotations::setRendering(const RenderingCfg& renderingCfg)
{
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages

    std::string renderMethodStr("");
    switch
        (renderingCfg.renderMethod())
    {
        case RenderMethod::RASTER:
            renderMethodStr = "RASTER";
            break;
        case RenderMethod::ARROWS:
            renderMethodStr = "ARROWS";
            break;
        case RenderMethod::BARBS:
            renderMethodStr = "BARBS";
            break;
        case RenderMethod::TRAJECTORIES:
            renderMethodStr = "TRAJECTORIES";
            break;
        case RenderMethod::DOTSCLOUD:
            renderMethodStr = "DOTSCLOUD";
            break;
        case RenderMethod::STREAMLINES:
            renderMethodStr = "STREAMLINES";
            break;
        case RenderMethod::RAWRGB:
            renderMethodStr = "RAWRGB";
            break;
        case RenderMethod::POLYLINE:
            renderMethodStr = "POLYLINE";
            break;
        case RenderMethod::POLYGON:
            renderMethodStr = "POLYGON";
            break;
        case RenderMethod::ELLIPSE:
            renderMethodStr = "ELLIPSE";
            break;
        case RenderMethod::POINT:
            renderMethodStr = "POINT";
            break;
    }

    std::string filterModeStr("");
    switch
        (renderingCfg.filterMode())
    {
        case FilterMode::NEAREST:
            filterModeStr = "NEAREST";
            break;
        case FilterMode::BILINEAR:
            filterModeStr = "BILINEAR";
            break;
    }

	std::ostringstream sqlTmp("");	// SQL query
    sqlTmp << "UPDATE variables SET ";
    sqlTmp << "rcfgLogscale = %u,";
    sqlTmp << "rcfgMin = %f,";
    sqlTmp << "rcfgMax = %f,";
    sqlTmp << "rcfgZindex = %f,";
    sqlTmp << "rcfgOpacity = %f,";
    sqlTmp << "rcfgColorRed = %u,";
    sqlTmp << "rcfgColorGreen = %u,";
    sqlTmp << "rcfgColorBlue = %u,";
    sqlTmp << "rcfgColormap = %Q,";
    sqlTmp << "rcfgRenderMethod = %Q,";
    sqlTmp << "rcfgParticlesCount = %u,";
    sqlTmp << "rcfgParticleTTL = %u,";
    sqlTmp << "rcfgStreamlineLength = %u,";
    sqlTmp << "rcfgStreamlineSpeed = %f,";
    sqlTmp << "rcfgLineThickness = %f,";
    sqlTmp << "rcfgFilterMode = %Q ";
    sqlTmp << "WHERE variable = %u AND collection = %u;";

    char* sql = sqlite3_mprintf(sqlTmp.str().c_str(),
                                renderingCfg.logscale(),
                                static_cast<double>(renderingCfg.min()),
                                static_cast<double>(renderingCfg.max()),
                                static_cast<double>(renderingCfg.zindex()),
                                static_cast<double>(renderingCfg.opacity()),
                                renderingCfg.color().at(0),
                                renderingCfg.color().at(1),
                                renderingCfg.color().at(2),
                                renderingCfg.colormap().c_str(),
                                renderMethodStr.c_str(),
                                renderingCfg.particlesCount(),
                                renderingCfg.particleTTL(),
                                renderingCfg.streamlineLength(),
                                static_cast<double>(renderingCfg.streamlineSpeed()),
                                static_cast<double>(renderingCfg.lineThickness()),
                                filterModeStr.c_str(),
                                renderingCfg.target.variable,
                                renderingCfg.target.collection.second);
    rc = sqlite3_exec( this->_db
                     , sql
                     , nullptr
                     , nullptr
                     , &errMsg);

    sqlite3_free(sql);

    if
        (SQLITE_OK != rc)
    {
		LOG( LogLevel::Error, "Annotations SQLite"
				, "Could not update variable rendering: {}", errMsg);
		sqlite3_free(errMsg);
        return false;
    }

    sqlite3_free(errMsg);

    return true;
}

/**
 *
 */
bool SQLiteAnnotations::saveAnnotationProperties(const RenderableId& renderableId, const DictNode& properties)
{
    bool ok(false);

    Annotation annotation;
    std::uint32_t collection(0u);
    ok = this->get(renderableId.granule.second, collection, annotation);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Annotations SQLite", "");
        return false;
    }

    DictNode& rootProperties = annotation.getProperties();
    auto found = rootProperties.children.find("properties");
    if
        (rootProperties.children.end() == found)
    {
        rootProperties.children["properties"] = new DictNode();
    }
    else
    {
        DictNode* existingProperties = static_cast<DictNode*>(found->second);
        // Remove all existing properties
        for
            (auto child: existingProperties->children)
        {
            delete child.second;
            child.second = nullptr;
        }

        existingProperties->children.clear();
        existingProperties = nullptr;
    }

    DictNode* annotationProperties = static_cast<DictNode*>(rootProperties.children["properties"]);

    // Copy new properties
    auto newFound = properties.children.find("properties");
    if
        (properties.children.end() != newFound)
    {
        DictNode* newProperties = static_cast<DictNode*>(newFound->second);
        for
            (auto it: newProperties->children)
        {
            annotationProperties->children.emplace(it.first, it.second);
        }
    }

    this->set(renderableId.granule.second, renderableId.collection.second, annotation);

    // Remove copied pointers so that the objects are not deleted when leaving
    // the scope of this function (they are still used by the original
    // "properties" DictNode passed as argument
    annotationProperties->children.clear();

    return true;
}

/**
 *
 */
bool SQLiteAnnotations::count(std::size_t& result) const
{
    int rc {-1};    //

    rc = sqlite3_step(this->_countAnnotationsStmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t tmpValue {0};
        tmpValue = sqlite3_column_int64(this->_countAnnotationsStmt, 0);
        result = static_cast<std::size_t>(tmpValue);
        rc = sqlite3_step(this->_countAnnotationsStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite Annotations",
                 "Failed to execute the countAnnotations SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteAnnotations::countCollections(std::size_t& result) const
{
    int rc {-1};    //

    rc = sqlite3_step(this->_countCollectionsStmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t tmpValue {0};
        tmpValue = sqlite3_column_int64(this->_countCollectionsStmt, 0);
        result = static_cast<std::size_t>(tmpValue);
        rc = sqlite3_step(this->_countCollectionsStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite Annotations",
                 "Failed to execute the countCollections SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}
