// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>
#include <iomanip>
#include <mutex>
#include <fstream>
#include <sstream>

#include "vendors/_sqlite3.h"

#include "core/log.h"
#include "core/path.h"
#include "core/structs.h"
#include "core/granule_metadata.h"
#include "sqlite_utils.h"
#include "sqlite_granules_index.h"
#include "sqlite_granules_index_schema.h"

#include <iostream>

using namespace SEAScope;


/**
 *
 */
SQLiteGranulesIndex::SQLiteGranulesIndex()
: _db(nullptr),
  _getNextGranuleIdStmt(nullptr),
  _insertGranuleStmt(nullptr),
  _updateGranuleStmt(nullptr),
  _getGranuleByDataIdStmt(nullptr),
  _getGranuleByIdStmt(nullptr),
  _hasGranuleByDataIdStmt(nullptr),
  _hasGranuleByIdStmt(nullptr),
  _removeGranuleStmt(nullptr),
  _removeGranuleByIdStmt(nullptr),
  _getTimeCoverageByCollectionIdStmt(nullptr),
  _findCurrentGranulesStmt(nullptr),
  _findInTimeRangeGranulesStmt(nullptr),
  _beginTransactionStmt(nullptr),
  _commitTransactionStmt(nullptr),
  _rollbackTransactionStmt(nullptr),
  _getNextCollectionIdStmt(nullptr),
  _insertCollectionStmt(nullptr),
  _removeCollectionStmt(nullptr),
  _removeCollectionByURIStmt(nullptr),
  _listCollectionsStmt(nullptr),
  _countGranulesStmt(nullptr),
  _countCollectionsStmt(nullptr),
  _deleteAllCollectionsStmt(nullptr),
  _idGenerator(1),
  _uncommittedOpsCount(0),
  _maxUncommittedBulkOps(500),
  _granuleAddedCallback(nullptr),
  _collectionIdGenerator(1)
{

}

/**
 *
 */
SQLiteGranulesIndex::~SQLiteGranulesIndex()
{
    if
        (nullptr != this->_getNextGranuleIdStmt)
    {
        sqlite3_finalize(this->_getNextGranuleIdStmt);
        this->_getNextGranuleIdStmt = nullptr;
    }

    if
        (nullptr != _insertGranuleStmt)
    {
        sqlite3_finalize(this->_insertGranuleStmt);
        this->_insertGranuleStmt = nullptr;
    }

    if
        (nullptr != _updateGranuleStmt)
    {
        sqlite3_finalize(this->_updateGranuleStmt);
        this->_updateGranuleStmt = nullptr;
    }

    if
        (nullptr != this->_getGranuleByIdStmt)
    {
        sqlite3_finalize(this->_getGranuleByIdStmt);
        this->_getGranuleByIdStmt = nullptr;
    }

    if
        (nullptr != this->_getGranuleByDataIdStmt)
    {
        sqlite3_finalize(this->_getGranuleByDataIdStmt);
        this->_getGranuleByDataIdStmt = nullptr;
    }

    if
        (nullptr != this->_hasGranuleByDataIdStmt)
    {
        sqlite3_finalize(this->_hasGranuleByDataIdStmt);
        this->_hasGranuleByDataIdStmt = nullptr;
    }

    if
        (nullptr != this->_hasGranuleByIdStmt)
    {
        sqlite3_finalize(this->_hasGranuleByIdStmt);
        this->_hasGranuleByIdStmt = nullptr;
    }

    if
        (nullptr != this->_removeGranuleStmt)
    {
        sqlite3_finalize(this->_removeGranuleStmt);
        this->_removeGranuleStmt = nullptr;
    }

    if
        (nullptr != this->_removeGranuleByIdStmt)
    {
        sqlite3_finalize(this->_removeGranuleByIdStmt);
        this->_removeGranuleByIdStmt = nullptr;
    }

    if
        (nullptr != this->_getTimeCoverageByCollectionIdStmt)
    {
        sqlite3_finalize(this->_getTimeCoverageByCollectionIdStmt);
        this->_getTimeCoverageByCollectionIdStmt = nullptr;
    }

    if
        (nullptr != this->_findCurrentGranulesStmt)
    {
        sqlite3_finalize(this->_findCurrentGranulesStmt);
        this->_findCurrentGranulesStmt = nullptr;
    }

    if
        (nullptr != this->_findInTimeRangeGranulesStmt)
    {
        sqlite3_finalize(this->_findInTimeRangeGranulesStmt);
        this->_findInTimeRangeGranulesStmt = nullptr;
    }

    if
        (nullptr != this->_beginTransactionStmt)
    {
        sqlite3_finalize(this->_beginTransactionStmt);
        this->_beginTransactionStmt = nullptr;
    }

    if
        (nullptr != this->_commitTransactionStmt)
    {
        sqlite3_finalize(this->_commitTransactionStmt);
        this->_commitTransactionStmt = nullptr;
    }

    if
        (nullptr != this->_rollbackTransactionStmt)
    {
        sqlite3_finalize(this->_rollbackTransactionStmt);
        this->_rollbackTransactionStmt = nullptr;
    }

    if
        (nullptr != this->_getNextCollectionIdStmt)
    {
        sqlite3_finalize(this->_getNextCollectionIdStmt);
        this->_getNextCollectionIdStmt = nullptr;
    }

    if
        (nullptr != this->_insertCollectionStmt)
    {
        sqlite3_finalize(this->_insertCollectionStmt);
        this->_insertCollectionStmt = nullptr;
    }

    if
        (nullptr != this->_removeCollectionStmt)
    {
        sqlite3_finalize(this->_removeCollectionStmt);
        this->_removeCollectionStmt = nullptr;
    }

    if
        (nullptr != this->_removeCollectionByURIStmt)
    {
        sqlite3_finalize(this->_removeCollectionByURIStmt);
        this->_removeCollectionByURIStmt = nullptr;
    }

    if
        (nullptr != this->_listCollectionsStmt)
    {
        sqlite3_finalize(this->_listCollectionsStmt);
        this->_listCollectionsStmt = nullptr;
    }

    if
        (nullptr != this->_countGranulesStmt)
    {
        sqlite3_finalize(this->_countGranulesStmt);
        this->_countGranulesStmt = nullptr;
    }

    if
        (nullptr != this->_countCollectionsStmt)
    {
        sqlite3_finalize(this->_countCollectionsStmt);
        this->_countCollectionsStmt = nullptr;
    }

    if
        (nullptr != _deleteAllCollectionsStmt)
    {
        sqlite3_finalize(this->_deleteAllCollectionsStmt);
        this->_deleteAllCollectionsStmt = nullptr;
    }

    if
        (nullptr != this->_db)
    {
        sqlite3_close(this->_db);
        this->_db = nullptr;
    }
}

/**
 *
 */
std::uint64_t SQLiteGranulesIndex::generateGranuleId()
{
    // Lock scope
    std::lock_guard<std::mutex> lock(this->_idGeneratorMutex);

	return this->_idGenerator++; // return current value, then increment
}

/**
 *
 */
bool SQLiteGranulesIndex::hasGranule(const std::string& dataId) const
{
    int rc {-1};    //
    std::uint64_t count {0};    //

    sqlite3_reset(this->_hasGranuleByDataIdStmt);

    rc = sqlite3_bind_text(this->_hasGranuleByDataIdStmt, 1, dataId.c_str(), static_cast<int>(dataId.length()), nullptr /*<=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind data identifier text '{}' to the SQL statement: {}",
                 dataId, sqlite3_errmsg(this->_db));
        return false;
    }

    rc = sqlite3_step(this->_hasGranuleByDataIdStmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t tmpValue {0};
        tmpValue = sqlite3_column_int64(this->_hasGranuleByDataIdStmt, 0);
        count = static_cast<std::uint64_t>(tmpValue);
        rc = sqlite3_step(this->_hasGranuleByDataIdStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the hasGranule SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return (0 < count);
}

/**
 *
 */
bool SQLiteGranulesIndex::hasGranule(std::uint64_t granuleId) const
{
    int rc {-1};    //
    std::uint64_t count {0};    //

    sqlite3_reset(this->_hasGranuleByIdStmt);

    rc = sqlite3_bind_int64(this->_hasGranuleByIdStmt, 1, static_cast<std::int64_t>(granuleId));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind identifier number {} to the SQL statement: {}",
                 granuleId, sqlite3_errmsg(this->_db));
        return false;
    }

    rc = sqlite3_step(this->_hasGranuleByIdStmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t tmpValue {0};
        tmpValue = sqlite3_column_int64(this->_hasGranuleByIdStmt, 0);
        count = static_cast<std::uint64_t>(tmpValue);
        rc = sqlite3_step(this->_hasGranuleByIdStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the hasGranule SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return (0 < count);
}

/**
 *
 */
bool SQLiteGranulesIndex::insertGranule(GranuleMetadata& info)
{
    int rc {-1};    //
    const std::uint8_t* buffer {nullptr}; //
    std::uint64_t bufferSize {0};   //

    if
        (0 == info.granuleId())
    {
        // granuleId = 0 means that no identifier has been set for the granule yet
        std::uint64_t granuleId = this->generateGranuleId();
        info.granuleId(granuleId);
    }

    buffer = info.getBlob(bufferSize);

    sqlite3_reset(this->_insertGranuleStmt);

    // Bind granule_id
    rc = sqlite3_bind_int64(this->_insertGranuleStmt, 1, static_cast<std::int64_t>(info.granuleId()));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT granule_id argument to the SQL statement");
        return false;
    }

    // Bind collection_id
    rc = sqlite3_bind_int64(this->_insertGranuleStmt, 2, info.collectionId());
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT collection_id argument to the SQL statement");
        return false;
    }

    // Bind start
    rc = sqlite3_bind_int64(this->_insertGranuleStmt, 3, static_cast<std::int64_t>(info.timeCoverageStart()));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT start argument to the SQL statement");
        return false;
    }

    // Bind stop
    rc = sqlite3_bind_int64(this->_insertGranuleStmt, 4, static_cast<int64_t>(info.timeCoverageEnd()));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT stop argument to the SQL statement");
        return false;
    }

    // Bind data_id
    std::string dataId {info.dataId()};
    rc = sqlite3_bind_text(this->_insertGranuleStmt, 5, dataId.c_str(), static_cast<int>(dataId.length()), nullptr /*<=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the TEXT data_id argument to the SQL statement");
        return false;
    }

    // Bind metadata
    rc = sqlite3_bind_blob64(this->_insertGranuleStmt, 6, buffer, bufferSize, nullptr /*<=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the BLOB metadata argument to the SQL statement");
        return false;
    }

    // Execute statement
    rc = sqlite3_step(this->_insertGranuleStmt);
    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, " SQLite granules index",
                 "Failed to execute the insertGranule SQL statement: {}", sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::getGranuleById(GranuleMetadata& result, std::uint64_t granuleId)
{
    int rc {-1};    //

    sqlite3_reset(this->_getGranuleByIdStmt);

    rc = sqlite3_bind_int64(this->_getGranuleByIdStmt, 1, static_cast<std::int64_t>(granuleId));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the int64 id argument to the SQL statement");
        return false;
    }

    rc = sqlite3_step(this->_getGranuleByIdStmt);
    while
        (SQLITE_ROW == rc)
    {
        const void* blob = sqlite3_column_blob(this->_getGranuleByIdStmt, 0);
        int blobSize = sqlite3_column_bytes(this->_getGranuleByIdStmt, 0);
        result.setBlob(reinterpret_cast<std::uint8_t*>(const_cast<void*>(blob)),
                       static_cast<std::uint64_t>(blobSize));

        rc = sqlite3_step(this->_getGranuleByIdStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the getGranuleById SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::getGranuleByDataId(GranuleMetadata& result, const std::string& dataId)
{
    int rc {-1};    //

    sqlite3_reset(this->_getGranuleByDataIdStmt);

    rc = sqlite3_bind_text(this->_getGranuleByDataIdStmt, 1, dataId.c_str(), static_cast<int>(dataId.length()), nullptr /*<=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the TEXT data_id argument to the SQL statement");
        return false;
    }

    rc = sqlite3_step(this->_getGranuleByDataIdStmt);
    while
        (SQLITE_ROW == rc)
    {
        const void* blob = sqlite3_column_blob(this->_getGranuleByDataIdStmt, 0);
        int blobSize = sqlite3_column_bytes(this->_getGranuleByDataIdStmt, 0);
        result.setBlob(reinterpret_cast<std::uint8_t*>(const_cast<void*>(blob)),
                       static_cast<std::uint64_t>(blobSize));

        rc = sqlite3_step(this->_getGranuleByDataIdStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the getGranuleByDataId SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::updateGranule(GranuleMetadata& info)
{
    int rc {-1};    //
    bool ok {false};    //
    GranuleMetadata registered;  //
    const std::uint8_t* buffer {nullptr}; //
    std::uint64_t bufferSize {0};   //

    ok = this->getGranuleByDataId(registered, info.dataId());
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Could not read existing granule from database");
        return false;
    }

    // Make sure to reuse the registred identifier
    info.sourceId(registered.sourceId());
    info.granuleId(registered.granuleId());

    // The submitted granule information take precedence over the
    // registered info so we simply add the registered IDFDescriptors for
    // subsampling levels that are not already available in the submitted
    // info
    for
        (std::uint32_t i = 0; i < registered.viewsCount(); ++i)
    {
        IDFDescriptor registeredUri;
        registered.getView(i, registeredUri);
        bool found {false};

        for
            (std::uint32_t j = 0; j < info.viewsCount(); ++j)
        {
            IDFDescriptor uri;
            info.getView(j, uri);
            if
                (uri.subsamplingFactor == registeredUri.subsamplingFactor)
            {
                found = true;
                break;
            }
        }

        if
            (false == found)
        {
            std::string viewURI;
            info.getViewURI(0, viewURI);
            //std::cout << "Adding new spacing to existing granule: " << info.dataId() << ": " << viewURI << std::endl;
            info.addIDFDescriptor(registeredUri);
        }
    }

    // Sort IDFDescriptors by ascending resolution
    /*
    std::sort(info.uris.begin(), info.uris.end(),
              [](const IDFDescriptor& i, const IDFDescriptor& j) -> bool {
                return i.resolution < j.resolution;
              });
    */

    // Update table content
    buffer = info.getBlob(bufferSize);
    sqlite3_reset(this->_updateGranuleStmt);

    // Bind collection_id
    rc = sqlite3_bind_int64(this->_updateGranuleStmt, 1, info.collectionId());
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT collection_id argument to the SQL statement");
        return false;
    }

    // Bind start
    rc = sqlite3_bind_int64(this->_updateGranuleStmt, 2, static_cast<std::int64_t>(info.timeCoverageStart()));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT start argument to the SQL statement");
        return false;
    }

    // Bind stop
    rc = sqlite3_bind_int64(this->_updateGranuleStmt, 3, static_cast<std::int64_t>(info.timeCoverageEnd()));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT stop argument to the SQL statement");
        return false;
    }

    // Bind metadata
    rc = sqlite3_bind_blob64(this->_updateGranuleStmt, 4, buffer, bufferSize, nullptr /*<=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the BLOB metadata argument to the SQL statement");
        return false;
    }

    // Bind granule_id
    rc = sqlite3_bind_int64(this->_updateGranuleStmt, 5, static_cast<std::int64_t>(info.granuleId()));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT granule_id argument to the SQL statement");
        return false;
    }

    rc = sqlite3_step(this->_updateGranuleStmt);
    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, " SQLite granules index",
                 "Failed to execute the updateGranule SQL statement: {}", sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::add(GranuleMetadata& info, bool silent)
{
    bool ok {false};    //

    if
        (true == this->hasGranule(info.dataId()))
    {
        ok = this->updateGranule(info);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, " SQLite granules index",
                     "Failed to update granule");
            return false;
        }
    }
    else
    {
        ok = this->insertGranule(info);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "SQLite granules index",
                     "Failed to add granule");
            return false;
        }
    }

	if
		(false == silent && nullptr != this->_granuleAddedCallback)
	{
		this->_granuleAddedCallback(info.granuleId(), info.collectionId());
	}

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::updateTimeCoverage(std::uint64_t granuleId, const timestamp_t start, const timestamp_t stop)
{
    GranuleMetadata info;   //
    bool ok {false};    //

    ok = getGranuleById(info, granuleId);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "");
        return false;
    }

    info.timeCoverageStart(start);
    info.timeCoverageEnd(stop);

    ok = this->updateGranule(info);
    return ok;
}

/**
 *
 */
bool SQLiteGranulesIndex::remove(std::uint64_t granuleId, bool /*silent*/)
{
    int rc {-1};    //

    sqlite3_reset(this->_removeGranuleStmt);

    rc = sqlite3_bind_int64(this->_removeGranuleStmt, 1, static_cast<std::int64_t>(granuleId));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT granule_id argument to the SQL statement");
        return false;
    }


    rc = sqlite3_step(this->_removeGranuleStmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(this->_removeGranuleStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the remove granule SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    int affectedRows {0};
    affectedRows = sqlite3_changes(this->_db);

    return (0 < affectedRows);
}

/**
 *
 */
bool SQLiteGranulesIndex::remove(const std::string& dataId, bool /*silent*/)
{
    int rc {-1};    //

    sqlite3_reset(this->_removeGranuleByIdStmt);

    rc = sqlite3_bind_text(this->_removeGranuleByIdStmt, 1, dataId.c_str(), static_cast<int>(dataId.length()), nullptr /*<=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the TEXT data_id argument to the SQL statement");
        return false;
    }


    rc = sqlite3_step(this->_removeGranuleByIdStmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(this->_removeGranuleByIdStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the remove granule SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    int affectedRows {0};
    affectedRows = sqlite3_changes(this->_db);

    return (0 < affectedRows);
}

/**
 *
 */
std::array<bool, 2> SQLiteGranulesIndex::adjacentGranules(const timestamp_t currentMs,
                                                          const timestamp_t minDelta,
									                      const std::set<std::uint32_t>& collections,
									                      GranuleMetadata& prev,
									                      GranuleMetadata& next
									                      ) const
{
    int rc {-1};    //
	bool prevFound {false};	//
	bool nextFound {false};	//
    std::ostringstream sqlPast {""};    //
    std::ostringstream sqlFuture {""};    //
    sqlite3_stmt* stmt {nullptr};

    if
        (0 >= collections.size())
    {
        return { {false, false} };
    }

    // It would be cleaner to use SQLite binding methods to build the
    // statements but it would require more code and many function calls.
    // Building the SQL statement using string stream should be fine as all
    // parameters are numerical values
    sqlPast << "SELECT min(" << std::to_string(currentMs) << " - (time_coverage_start + time_coverage_end) / 2), metadata ";
    sqlPast << "FROM indexed_granules ";
    sqlPast << "WHERE (time_coverage_start + time_coverage_end) / 2 < " << std::to_string(currentMs) << " ";
    sqlPast << "AND " << std::to_string(currentMs) << " - (time_coverage_start + time_coverage_end) / 2 > " << std::to_string(minDelta) << " ";
    sqlPast << "AND collection_id IN (";
    sqlFuture << "SELECT min((time_coverage_start + time_coverage_end) / 2 - " << std::to_string(currentMs) << "), metadata ";
    sqlFuture << "FROM indexed_granules ";
    sqlFuture << "WHERE (time_coverage_start + time_coverage_end) / 2 > " << std::to_string(currentMs) << " ";
    sqlFuture << "AND (time_coverage_start + time_coverage_end) / 2 - " << std::to_string(currentMs) << " > " << std::to_string(minDelta) << " ";
    sqlFuture << "AND collection_id IN (";
    for
        (std::uint32_t collectionId : collections)
    {
        sqlPast << std::to_string(collectionId) << ",";
        sqlFuture << std::to_string(collectionId) << ",";
    }

    // Move write head to overwrite trailing comma
    sqlPast.seekp(-1, sqlPast.cur);
    sqlFuture.seekp(-1, sqlFuture.cur);

    sqlPast << ")";
    sqlFuture << ")";

    // Execute statement for past neighbour
    rc = sqlite3_prepare_v2(this->_db, sqlPast.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for searching the previous granule: {} ({})",
                 sqlPast.str(), sqlite3_errmsg(this->_db));
        return { {prevFound, nextFound} };
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        if
            (SQLITE_NULL != sqlite3_column_type(stmt, 1))
        {
            const void* blob = sqlite3_column_blob(stmt, 1);
            int blobSize = sqlite3_column_bytes(stmt, 1);
            prev.setBlob(reinterpret_cast<std::uint8_t*>(const_cast<void*>(blob)),
                         static_cast<std::uint64_t>(blobSize));
            prevFound = true;
        }
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute SQL statement for searching the previous granule: {}",
                 sqlite3_errmsg(this->_db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return { {prevFound, nextFound} };
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Execute statement for future neighbour
    rc = sqlite3_prepare_v2(this->_db, sqlFuture.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL segment for searching the next granule: {} ({})",
                 sqlFuture.str(), sqlite3_errmsg(this->_db));
        return { {prevFound, nextFound} };
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        if
            (SQLITE_NULL != sqlite3_column_type(stmt, 1))
        {
            const void* blob = sqlite3_column_blob(stmt, 1);
            int blobSize = sqlite3_column_bytes(stmt, 1);
            next.setBlob(reinterpret_cast<std::uint8_t*>(const_cast<void*>(blob)),
                         static_cast<std::uint64_t>(blobSize));
            nextFound = true;
        }
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute SQL statement for searching the next granule: {}",
                 sqlite3_errmsg(this->_db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return { {prevFound, nextFound} };
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;


    return { {prevFound, nextFound} };
}

/**
 *
 */
bool SQLiteGranulesIndex::closestGranuleBound(const timestamp_t current,
                                              const timestamp_t searchLowerBound,
                                              const timestamp_t searchUpperBound,
                                              const std::set<std::uint32_t>& collections,
									          GranuleMetadata& closest
									          ) const
{
    if
        (0 >= collections.size())
    {
        return false;
    }

    if
        (searchLowerBound > searchUpperBound)
    {
        return false;
    }

    bool closestFound {false};  //
    std::ostringstream sqlIntra {""};   //
    std::ostringstream sqlExtra {""};   //

    // A granule time coverage is [time_coverage_start, time_coverage_end[,
    // meaning that time_coverage_end is the first datetime after time_coverage_start for which the granule has no data

    // Query to look for granules whose time coverage contains the reference datetime
    sqlIntra << "SELECT metadata ";
    sqlIntra << "FROM indexed_granules ";
    sqlIntra << "WHERE time_coverage_start <= " << std::to_string(current) << " ";
    sqlIntra << "AND time_coverage_end > " << std::to_string(current) << " ";
    sqlIntra << "AND collection_id IN (";

    // Query to look for granules whose time coverage bounds are within the search domain
    // The search domain is also expressed as [searchLowerBound, searchUpperBound[,
    // so the query should return granules with searchLowerBound <= time_coverage_start < searchUpperBound
    // or with searchLowerBound < time_coverage_end <= searchUpperBound
    //
    // Since the last datetime containing data is time_coverage_end - 1:
    // • searchLowerBound == time_coverage_end is not a valid result,
    //   therefore the correct criterium is searchLowerBound < time_coverage_end
    // • searchUpperBound == time_coverage_end is a valid result
    //   because the last acceptable datetime with data is searchUpperBound - 1 and the last datetime with data is time_coverage_end - 1
    sqlExtra << "SELECT metadata ";
    sqlExtra << "FROM indexed_granules ";
    sqlExtra << "WHERE (";
    sqlExtra << "(time_coverage_start >= " << std::to_string(searchLowerBound) << " AND time_coverage_start < " << std::to_string(searchUpperBound) << ") ";
    sqlExtra << "OR (time_coverage_end > " << std::to_string(searchLowerBound) << " AND time_coverage_end <= " << std::to_string(searchUpperBound) << ") ";
    sqlExtra << ") AND collection_id IN (";

    for
        (std::uint32_t collectionId : collections)
    {
        sqlIntra << std::to_string(collectionId) << ",";
        sqlExtra << std::to_string(collectionId) << ",";
    }

    sqlIntra.seekp(-1, sqlIntra.cur);
    sqlExtra.seekp(-1, sqlExtra.cur);

    sqlIntra << ") ";
    sqlExtra << ") ";

    sqlite3_stmt* stmt {nullptr};   //
    int rc  {-1};   //

    sqlIntra << " ORDER BY min(abs(time_coverage_start - " << std::to_string(current) << "), abs(time_coverage_end - " << std::to_string(current) << ")) ASC LIMIT 1 ";
    rc = sqlite3_prepare_v2(this->_db, sqlIntra.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare statement for finding the granule with the closest bound while including current time: {} ({})",
                 sqlIntra.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        if
            (SQLITE_NULL != sqlite3_column_type(stmt, 0))
        {
            const void* blob = sqlite3_column_blob(stmt, 0);
            int blobSize = sqlite3_column_bytes(stmt, 0);
            closest.setBlob(reinterpret_cast<std::uint8_t*>(const_cast<void*>(blob)),
                            static_cast<std::uint64_t>(blobSize));
            closestFound = true;
        }

        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the find closest bound including current time SQL statement: {}",
                 sqlite3_errmsg(this->_db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    if
        (true == closestFound)
    {
        return true;
    }

    sqlExtra << " ORDER BY min(abs(time_coverage_start - " << std::to_string(current) << "), abs(time_coverage_end - " << std::to_string(current) << ")) ASC LIMIT 1 ";
    rc = sqlite3_prepare_v2(this->_db, sqlExtra.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare statement from finding the granule with the closest bound while exlucing current time: {} ({})",
                 sqlExtra.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        if
            (SQLITE_NULL != sqlite3_column_type(stmt, 0))
        {
            const void* blob = sqlite3_column_blob(stmt, 0);
            int blobSize = sqlite3_column_bytes(stmt, 0);
            closest.setBlob(reinterpret_cast<std::uint8_t*>(const_cast<void*>(blob)),
                            static_cast<std::uint64_t>(blobSize));
            closestFound = true;
        }

        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the find closest bound excluding current time SQL statement: {}",
                 sqlite3_errmsg(this->_db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return closestFound;
};


/**
 *
 */
bool SQLiteGranulesIndex::getCoverageUsingStatement(sqlite3_stmt* stmt, EventsCoverage& result) const
{
    int rc {-1};    //
	timestamp_t d;	// Milliseconds
	timestamp_t end;	// Milliseconds
	std::uint16_t year; //
	std::uint8_t month;	//
	std::uint8_t day;	//
	time_t timestamp;	// Seconds
	std::tm* t;	//

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t startMs {0};   //
        std::int64_t endMs {0}; //

        startMs = sqlite3_column_int64(stmt, 0);
        endMs = sqlite3_column_int64(stmt, 1);

		timestamp = static_cast<time_t>(0.001 * static_cast<double>(startMs));
		t = std::localtime(&timestamp);

        // Handle case wherein the granule has an empty time coverage, i.e. an
        // "instantaneous" granule
        if
            (startMs == endMs)
        {
			year = static_cast<std::uint16_t>(1900 + t->tm_year);
			month = static_cast<std::uint8_t>(t->tm_mon);
			day = static_cast<std::uint8_t>(t->tm_mday);

			//LOG(LogLevel::Debug, "GranulesIndex", "Coverage {} {} {}", year, month+1, day);
			if
				(result.store.end() == result.store.find(year))
			{
				result.store[year] = std::array<std::uint32_t, 12>();
			}
			result.store[year][month] |= 1 << (day - 1);

            // We are done with this granule, check the next one
            rc = sqlite3_step(stmt);
            continue;
        }

        // Granule has data for the [startMs, endMs[ time range, so if
        // endMs corresponds to day D at 00:00:00.000 then D is NOT covered.
        // We iterate from the day to which startMs belongs to, setting the
        // time at 00:00:00.001 and add 24 hours each time we loop
		t->tm_hour = 0;
		t->tm_min = 0;
		t->tm_sec = 0;
		timestamp = std::mktime(t);
		d = static_cast<timestamp_t>(timestamp * static_cast<time_t>(1000) + 1);

		LOG(LogLevel::Debug, "GranulesIndex", "Original timestamp: {}ms", startMs);
		LOG(LogLevel::Debug, "GranulesIndex", "Modified timestamp: {}ms", d);

		end = static_cast<timestamp_t>(endMs);

        // d == end means that end is day D at 00:00:00.001, so the granule has
        // data between 00:00:00.000 and 00:00:00.001, so D must be included in
        // the coverage
		while
			(d <= end)
		{
			timestamp = static_cast<long>(0.001 * static_cast<double>(d));
			t = std::localtime(&timestamp);
			year = static_cast<std::uint16_t>(1900 + t->tm_year);
			month = static_cast<std::uint8_t>(t->tm_mon);
			day = static_cast<std::uint8_t>(t->tm_mday);

			//LOG(LogLevel::Debug, "GranulesIndex", "Coverage {} {} {}", year, month+1, day);
			if
				(result.store.end() == result.store.find(year))
			{
				result.store[year] = std::array<std::uint32_t, 12>();
			}
			result.store[year][month] |= 1 << (day - 1);
			d = d + 86400000;
		}

        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the get time coverage SQL statement: {}",
                 sqlite3_errmsg(this->_db));
    }

    return true;
}


/**
 *
 */
bool SQLiteGranulesIndex::getCoverage(std::uint32_t collectionId, EventsCoverage& result) const
{
    int rc {-1};    //

    result.reset();

    sqlite3_reset(this->_getTimeCoverageByCollectionIdStmt);

    rc = sqlite3_bind_int64(this->_getTimeCoverageByCollectionIdStmt, 1, collectionId);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the collection identifier '{}' to the SQL statement: {}",
                 collectionId, sqlite3_errmsg(this->_db));
        return false;
    }

    bool ok {false};
    ok = this->getCoverageUsingStatement(this->_getTimeCoverageByCollectionIdStmt, result);
    return ok;
};

/**
 *
 */
bool SQLiteGranulesIndex::getCoverage(const std::set<std::uint32_t>& collections, EventsCoverage& result) const
{
    int rc {-1};    //
    bool ok {false};    //
    sqlite3_stmt* stmt {nullptr};   //
    std::ostringstream sql {""};    //

    result.reset();

    if
        (0 >= collections.size())
    {
        return true;
    }

    // Not very "clean", we should be using SQLite bindings for passing the
    // collection identifiers but it should be fine because they are numerical
    // values. Using bindings would require more code and function calls.
    sql << "SELECT time_coverage_start, time_coverage_end FROM indexed_granules WHERE collection_id IN (";
    for
        (std::uint32_t collectionId : collections)
    {
        sql << std::to_string(collectionId) << ",";
    }
    // Move write head to overwrite trailing comma
    sql.seekp(-1, sql.cur);

    sql << ")";

    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare the SQL statement for getting time coverage for multiple collections: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    ok = this->getCoverageUsingStatement(stmt, result);

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return ok;
};

/**
 *
 */
bool SQLiteGranulesIndex::bindFindCurrentGranules(std::uint32_t collectionId, timestamp_t current) const
{
    int rc {-1};    //

    sqlite3_reset(this->_findCurrentGranulesStmt);

    rc = sqlite3_bind_int64(this->_findCurrentGranulesStmt, 1, static_cast<std::int64_t>(collectionId));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the int64 collection identifier '{}' to the SQL statement: {}",
                 collectionId, sqlite3_errmsg(this->_db));
        return false;
    }

    rc = sqlite3_bind_int64(this->_findCurrentGranulesStmt, 2, static_cast<std::int64_t>(current));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the int64 current timestamp '{}' to the SQL statement: {}",
                 current, sqlite3_errmsg(this->_db));
        return false;
    }

    rc = sqlite3_bind_int64(this->_findCurrentGranulesStmt, 3, static_cast<std::int64_t>(current));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the int64 current timestamp '{}' (bis) to the SQL statement: {}",
                 current, sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::bindFindInTimeRangeGranules(std::uint32_t collectionId, timestamp_t start, timestamp_t stop) const
{
    int rc {-1};    //

    sqlite3_reset(this->_findInTimeRangeGranulesStmt);

    rc = sqlite3_bind_int64(this->_findInTimeRangeGranulesStmt, 1, static_cast<std::int64_t>(collectionId));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the int64 collection identifier '{}' to the find granules in time range SQL statement: {}",
                 collectionId, sqlite3_errmsg(this->_db));
        return false;
    }

    rc = sqlite3_bind_int64(this->_findInTimeRangeGranulesStmt, 2, static_cast<std::int64_t>(stop));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the int64 stop timestamp '{}' to the find granules in time range SQL statement: {}",
                 stop, sqlite3_errmsg(this->_db));
        return false;
    }

    rc = sqlite3_bind_int64(this->_findInTimeRangeGranulesStmt, 3, static_cast<std::int64_t>(start));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the int64 start timestamp '{}' to the find granules in time range SQL statement: {}",
                 start, sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::findGranules(const SearchCriteria& criteria, std::vector<GranuleMetadata>& result) const
{
    int rc {-1};    //
    bool ok {false};    //
    sqlite3_stmt* stmt {nullptr};   //

    for
        (const auto& collection: criteria.collections)
    {
        if
            (true == collection.second)
        {
            ok = this->bindFindCurrentGranules(collection.first, 1000 * criteria.current);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "SQLite granules index",
                         "Could not bind search parameters to the find current granules SQL statement");
                return false;
            }

            stmt = this->_findCurrentGranulesStmt;
        }
        else
        {
            ok = this->bindFindInTimeRangeGranules(collection.first, 1000 * criteria.start, 1000 * criteria.stop);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "SQLite granules index",
                         "Could not bind search parameters to the find in time range granules SQL statement");
                return false;
            }

            stmt = this->_findInTimeRangeGranulesStmt;
        }

        rc = sqlite3_step(stmt);
        while
            (SQLITE_ROW == rc)
        {
            result.emplace_back();
            const void* blob = sqlite3_column_blob(stmt, 0);
            int blobSize = sqlite3_column_bytes(stmt, 0);
            result.back().setBlob(reinterpret_cast<std::uint8_t*>(const_cast<void*>(blob)),
                                  static_cast<std::uint64_t>(blobSize));

            rc = sqlite3_step(stmt);
        }

        if
            (SQLITE_DONE != rc)
        {
            LOG(LogLevel::Error, "SQLite granules index",
                     "Failed to execute the SQL statement for finding granules");
            return false;
        }
    }

	return 0 < result.size();
};

/**
 *
 */
bool SQLiteGranulesIndex::prepareStatements()
{
    int rc {-1};    //
    std::ostringstream sql {""};    //

    // Read next granule identifier
    sql.str("");
    sql << "SELECT 1 + MAX(granule_id) FROM indexed_granules;";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_getNextGranuleIdStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for getting the next granule identifier: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Insert granule
    sql.str("");
    sql << "INSERT INTO indexed_granules VALUES(?, ?, ?, ?, ?, ?);";

    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_insertGranuleStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for inserting a new granule: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Update granule
    sql.str("");
    sql << "UPDATE indexed_granules SET ";
    sql << "collection_id = ?, ";
    sql << "time_coverage_start = ?, ";
    sql << "time_coverage_end = ?, ";
    sql << "metadata = ?";
    sql << "WHERE granule_id = ?;";

    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_updateGranuleStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for updating a granule: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Get granule by data_id
    sql.str("");
    sql << "SELECT metadata FROM indexed_granules WHERE data_id = ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_getGranuleByDataIdStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for getting a granule by its data_id: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Get granule by granule_id
    sql.str("");
    sql << "SELECT metadata FROM indexed_granules WHERE granule_id = ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_getGranuleByIdStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for getting a granule by its granule_id: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Count granules with a specific dataId
    sql.str("");
    sql << "SELECT COUNT(*) FROM indexed_granules WHERE data_id = ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_hasGranuleByDataIdStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for counting granules with a specific data identifier: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Count granules with a specific ID
    sql.str("");
    sql << "SELECT COUNT(*) FROM indexed_granules WHERE granule_id = ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_hasGranuleByIdStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for counting granules with a specific id: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Remove granule by granule_id
    sql.str("");
    sql << "DELETE FROM indexed_granules WHERE granule_id = ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_removeGranuleStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for removing a granule based on its granule_id: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Remove granule by dataId
    sql.str("");
    sql << "DELETE FROM indexed_granules WHERE data_id = ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_removeGranuleByIdStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for removing a granule based on its data identifier: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Get time coverage for granules that belong to a specific collection
    sql.str("");
    sql << "SELECT time_coverage_start, time_coverage_end FROM indexed_granules WHERE collection_id = ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_getTimeCoverageByCollectionIdStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for getting the time coverage of collection: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Find granules whose time coverage contains the current timestamp
    sql.str("");
    sql << "SELECT metadata FROM indexed_granules WHERE collection_id = ? AND time_coverage_start <= ? AND time_coverage_end > ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_findCurrentGranulesStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for finding granules that contain the current timestamp: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Find granules whose time coverage intersects a time range
    sql.str("");
    sql << "SELECT metadata FROM indexed_granules WHERE collection_id = ? AND time_coverage_start < ? AND time_coverage_end > ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_findInTimeRangeGranulesStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for finding granules that contain the current timestamp: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Start transaction
    sql.str("");
    sql << "BEGIN TRANSACTION";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_beginTransactionStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for beginning a bulk operation: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Commit transaction
    sql.str("");
    sql << "COMMIT TRANSACTION";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_commitTransactionStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for completing a bulk operation: {} ({})",
                  sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Rollback transaction
    sql.str("");
    sql << "ROLLBACK TRANSACTION";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_rollbackTransactionStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for cancelling a bulk operation: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Read next collection identifier
    sql.str("");
    sql << "SELECT 1 + MAX(collection_id) FROM indexed_collections;";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_getNextCollectionIdStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for getting the next collection identifier: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Insert collection
    sql.str("");
    sql << "INSERT INTO indexed_collections VALUES(?, ?, STRFTIME('%s', 'NOW'));";

    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_insertCollectionStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for inserting a new collection: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Remove collection by collection_id
    sql.str("");
    sql << "DELETE FROM indexed_collections WHERE collection_id = ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_removeCollectionStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for removing a collection based on its collection_id: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Remove collection by URI
    sql.str("");
    sql << "DELETE FROM indexed_collections WHERE collection_uri = ?";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_removeCollectionByURIStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for removing a collection based on its URI: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // List collections
    sql.str("");
    sql << "SELECT collection_id, collection_uri FROM indexed_collections WHERE 1";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_listCollectionsStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for listing collections: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Count collections
    sql.str("");
    sql << "SELECT COUNT(*) FROM indexed_collections";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_countCollectionsStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for counting collections: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Count granules
    sql.str("");
    sql << "SELECT COUNT(*) FROM indexed_granules";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_countGranulesStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for counting granules: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Delete all collections
    sql.str("");
    sql << "DELETE FROM indexed_collections WHERE 1";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_deleteAllCollectionsStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statement for deleting all collectionss: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::readNextGranuleId()
{
    int rc {-1};    //

    rc = sqlite3_step(this->_getNextGranuleIdStmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t result {-1};
        result = sqlite3_column_int64(this->_getNextGranuleIdStmt, 0);
        this->_idGenerator = static_cast<std::uint64_t>(result);

        rc = sqlite3_step(this->_getNextGranuleIdStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, " SQLite granules index",
                 "Failed to execute the readNextGranuleId SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    if
        (0 >= this->_idGenerator)
    {
        // Happens if the indexed_granules table is empty
        this->_idGenerator = 1;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::readNextCollectionId()
{
    int rc {-1};    //

    rc = sqlite3_step(this->_getNextCollectionIdStmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int32_t result {-1};
        result = sqlite3_column_int(this->_getNextCollectionIdStmt, 0);
        this->_collectionIdGenerator = static_cast<std::uint32_t>(result);

        rc = sqlite3_step(this->_getNextCollectionIdStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, " SQLite granules index",
                 "Failed to execute the readNextCollectionId SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    if
        (0 >= this->_collectionIdGenerator)
    {
        // Happens if the indexed_collections table is empty
        this->_collectionIdGenerator = 1;
    }

    return true;
}

/**
 *
 */
std::tuple<bool, bool> SQLiteGranulesIndex::load(const std::string& path)
{
    int rc {-1};    //
    bool ok {false};    //
    std::tuple<bool, bool> initResult {false, false};  //
    bool initSuccess {false};   //
    bool hasContent {false};    //

    initResult = SQLiteGranulesIndexSchema::initialize(path);
    initSuccess = std::get<0>(initResult);
    hasContent = std::get<1>(initResult);
    if
        (false == initSuccess)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to initialize database '{}'", path);
        return {false, hasContent};
    }

	// Open database
	rc = sqlite3_open(path.c_str(), &(this->_db));
	if
		(SQLITE_OK != rc)
	{
		LOG(LogLevel::Error, "SQLite granules index",
                 "Could not open database {}: {}",
                 path, sqlite3_errmsg(this->_db));
		return {false, hasContent};
	}

    ok = SQLiteGranulesIndexSchema::configure(this->_db);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Could not configure database {}", path);
        return {false, hasContent};
    }

    // DEBUG
    //SQLiteUtils::showTrace(this->_db);

    // Prepare statements
    ok = this->prepareStatements();
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to prepare SQL statements");
        return {false, hasContent};
    }

    // Set the next generated ID to last known ID + 1
    ok = readNextGranuleId();
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to read next granule identifier");
        return {false, hasContent};
    }

    ok = readNextCollectionId();
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to read next collection identifier");
        return {false, hasContent};
    }

    return {true, hasContent};
}

/**
 *
 */
bool SQLiteGranulesIndex::save(const std::string& /*path*/)
{
    return false;
};

/**
 *
 */
bool SQLiteGranulesIndex::setGranuleAddedCallback(const std::function<void(std::uint64_t, std::uint32_t)>& callback)
{
    if
        (nullptr != this->_granuleAddedCallback)
    {
        return false;
    }

    this->_granuleAddedCallback = callback;
    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::setGranuleIdGenerator(std::uint64_t value)
{
    if
        (0 == value)
    {
        LOG(LogLevel::Error, "SQLiteGranulesIndex",
                 "The minimum value for granule IDs is 1");
        return false;
    }
    this->_idGenerator = value;
    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::startBulkOperation()
{
    int rc {-1};    //

    rc = sqlite3_step(this->_beginTransactionStmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(this->_beginTransactionStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the SQL statement for beginning a bulk operation: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::stopBulkOperation()
{
    int rc {-1};    //

    rc = sqlite3_step(this->_commitTransactionStmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(this->_commitTransactionStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the SQL statement for completing a bulk operation: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    this->_uncommittedOpsCount = 0;

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::cancelBulkOperation()
{
    int rc {-1};    //

    rc = sqlite3_step(this->_rollbackTransactionStmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(this->_rollbackTransactionStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the SQL statement for cancelling a bulk operation: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    this->_uncommittedOpsCount = 0;

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::increaseBulkOpsCount()
{
    bool ok {false};    //

    ++this->_uncommittedOpsCount;

    if
        (this->_maxUncommittedBulkOps <= this->_uncommittedOpsCount)
    {
        ok = this->stopBulkOperation();
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "SQLite granules index",
                     "Failed to stop bulk operation after reaching the max uncommitted changes count");
            return false;
        }
        ok = this->startBulkOperation();
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "SQLite granules index",
                     "Failed to restart bulk operation after reaching the max uncommitted changes count");
            return false;
        }
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::setCollectionIdGenerator(std::uint32_t value)
{
    if
        (0 == value)
    {
        LOG(LogLevel::Error, "SQLiteGranulesIndex",
                 "The minimum value for collection IDs is 1");
        return false;
    }
    this->_collectionIdGenerator = value;
    return true;
}

/**
 *
 */
std::uint32_t SQLiteGranulesIndex::generateCollectionId()
{
	return this->_collectionIdGenerator++; // return current value, then increment
}

/**
 *
 */
bool SQLiteGranulesIndex::addCollection(const std::string& collectionURI, const std::uint32_t collectionId)
{
    int rc {-1};    //

    sqlite3_reset(this->_insertCollectionStmt);

    // Bind collection_id
    rc = sqlite3_bind_int64(this->_insertCollectionStmt, 1, collectionId);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT collection_id argument to the SQL statement");
        return false;
    }

    // Bind collection URI
    rc = sqlite3_bind_text(this->_insertCollectionStmt, 2, collectionURI.c_str(), static_cast<int>(collectionURI.length()), nullptr /*<=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the TEXT collection_uri argument to the SQL statement");
        return false;
    }

    // Execute statement
    rc = sqlite3_step(this->_insertCollectionStmt);
    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, " SQLite granules index",
                 "Failed to execute the insertCollection SQL statement: {}", sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::removeCollection(const std::string& collectionURI)
{
    int rc {-1};    //

    sqlite3_reset(this->_removeCollectionByURIStmt);

    rc = sqlite3_bind_text(this->_removeCollectionByURIStmt, 1, collectionURI.c_str(), static_cast<int>(collectionURI.length()), nullptr /*<=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the TEXT collection_uri argument to the SQL statement");
        return false;
    }


    rc = sqlite3_step(this->_removeCollectionByURIStmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(this->_removeCollectionByURIStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the remove collection SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    // It does not matter if the collection was not found, and therefore not
    // removed, as long as it is not in the index
    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::removeCollection(std::uint32_t collectionId)
{
    int rc {-1};    //

    sqlite3_reset(this->_removeCollectionStmt);

    rc = sqlite3_bind_int64(this->_removeCollectionStmt, 1, static_cast<std::int64_t>(collectionId));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to bind the INT collection_id argument to the SQL statement");
        return false;
    }


    rc = sqlite3_step(this->_removeCollectionStmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(this->_removeCollectionStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the remove collection SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    // It does not matter if the collection was not found, and therefore not
    // removed, as long as it is not in the index
    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::listCollections(std::vector<std::pair<std::uint32_t, std::string> >& results) const
{
    int rc {-1};    //
    int tmpValue {0};
    const unsigned char* tmpStr {nullptr};  //
    std::uint32_t collectionId {0u};    //
    std::string collectionURI {""}; //

    sqlite3_reset(this->_listCollectionsStmt);

    rc = sqlite3_step(this->_listCollectionsStmt);
    while
        (SQLITE_ROW == rc)
    {
        tmpValue = sqlite3_column_int(this->_listCollectionsStmt, 0);
        collectionId = static_cast<std::uint32_t>(tmpValue);
        tmpStr = sqlite3_column_text(this->_listCollectionsStmt, 1);
        collectionURI = std::string(reinterpret_cast<const char*>(tmpStr));
        results.push_back({collectionId, collectionURI});

        rc = sqlite3_step(this->_listCollectionsStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the list collections SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::countGranules(std::size_t& result) const
{
    int rc {-1};    //

    rc = sqlite3_step(this->_countGranulesStmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t tmpValue {0};
        tmpValue = sqlite3_column_int64(this->_countGranulesStmt, 0);
        result = static_cast<std::size_t>(tmpValue);
        rc = sqlite3_step(this->_countGranulesStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the countGranules SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteGranulesIndex::countCollections(std::size_t& result) const
{
    int rc {-1};    //

    rc = sqlite3_step(this->_countCollectionsStmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t tmpValue {0};
        tmpValue = sqlite3_column_int64(this->_countCollectionsStmt, 0);
        result = static_cast<std::size_t>(tmpValue);
        rc = sqlite3_step(this->_countCollectionsStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the countCollections SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 * Deleting all indexed collections will also delete all indexed granules by
 * cascading effect.
 */
bool SQLiteGranulesIndex::clear()
{
    int rc {-1};    //

    rc = sqlite3_step(this->_deleteAllCollectionsStmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(this->_deleteAllCollectionsStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Failed to execute the deleteAllCollections SQL statement: {}",
                 sqlite3_errmsg(this->_db));
        return false;
    }

    this->_idGenerator = 1u;
    this->_collectionIdGenerator = 1u;

    return true;
}
