// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <tuple>
#include <cstdio>

#include "vendors/_sqlite3.h"

#include "core/log.h"
#include "core/path.h"

#include "sqlite_granules_index_schema.h"
#include "sqlite_utils.h"

using namespace SEAScope;

/**
 *
 */
static bool hasIndexedGranules(sqlite3* db)
{
    int rc {-1};    //
    int tablesCount {0};    //
    sqlite3_stmt* stmt {nullptr};   //
    std::ostringstream sql {""};    //

    sql << "SELECT COUNT(*) FROM sqlite_master WHERE type='table' AND name='indexed_granules'";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, " SQLite granules index schema",
                "Failed to prepare SQL statement for counting tables named 'indexed_granules': {} ({})",
                sql.str(), sqlite3_errmsg(db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        tablesCount = sqlite3_column_int(stmt, 0);
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to execute SQL statement: {}", sqlite3_errmsg(db));
        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return 1 == tablesCount;
}

/**
 *
 */
bool SQLiteGranulesIndexSchema::configure(sqlite3* db)
{
    int rc {-1};    //
    std::string sql {""};   //
    sqlite3_stmt* stmt {nullptr};   //

    // Configure database to use volatile memory for storing journal
    // information. It runs about 10 times faster than the write-ahead log but
    // it will lead to data in case of a power outage (not very important for
    // the index).
    sql = "PRAGMA journal_mode=MEMORY";
    rc = sqlite3_prepare_v2(db, sql.c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare SQL statement for switching journal mode to WAL: {} ({})",
                 sql, sqlite3_errmsg(db));
       return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Could not configure database with write-ahead logging: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Set synchonous to OFF even if it is less robust in case of a power
    // loss during a transaction. The index can still be regenerated later on
    // and serialization is much faster (at least 2x/3x improvement) than with
    // the other settings.
    sql = "PRAGMA synchronous=OFF";
    rc = sqlite3_prepare_v2(db, sql.c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare SQL statement for switching synchronous to NORMAL: {} ({})",
                 sql, sqlite3_errmsg(db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not switch synchronous to NORMAL: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Set the page size to a higher value for the indexed_granules table
    // if it exists
    bool indexedGranulesTableFound {false};  //
    indexedGranulesTableFound = hasIndexedGranules(db);
    if
        (true == indexedGranulesTableFound)
    {
        sql = "PRAGMA page_size = 512000";
        rc = sqlite3_prepare_v2(db, sql.c_str(), -1, &stmt, nullptr);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "SQLite granules index schema",
                     "Failed to prepare SQL statement for setting the cache size of the indexed_granules table: {} ({})",
                     sql, sqlite3_errmsg(db));
            return false;
        }

        rc = sqlite3_step(stmt);
        while
            (SQLITE_ROW == rc)
        {
            rc = sqlite3_step(stmt);
        }

        if
            (SQLITE_DONE != rc)
        {
            LOG(LogLevel::Error, "SQLite granules index schema",
                     "Could not set the cache size for the indexed_granules table: {}",
                     sqlite3_errmsg(db));

            sqlite3_finalize(stmt);
            stmt = nullptr;

            return false;
        }

        sqlite3_finalize(stmt);
        stmt = nullptr;
    }

    // Configure database to support foreign keys, it is required for cascading
    // effect when deleting a collection.
    sql = "PRAGMA foreign_keys=ON";
    rc = sqlite3_prepare_v2(db, sql.c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare SQL statement for enabling support for foreign keys: {} ({})",
                 sql, sqlite3_errmsg(db));
       return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Could not enable support for foreign keys: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Configure database to wait 10s before aborting a request due to database
    // locking mechanism.
    sql = "PRAGMA busy_timeout=10000";
    rc = sqlite3_prepare_v2(db, sql.c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare SQL statement for setting the busy_timeout parameter: {} ({})",
                 sql, sqlite3_errmsg(db));
       return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index",
                 "Could not set busy_timeout parameter: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;
    return true;
}

/**
 *
 */
static bool create(sqlite3* db)
{
    int rc {-1};    //
    std::string sql {""};   //
    sqlite3_stmt* stmt {nullptr};   //

    // Create schema history table
    sql = "CREATE TABLE granules_index_schema_history(" \
          "version INTEGER NOT NULL," \
          "datetime INTEGER NOT NULL," \
          "PRIMARY KEY(version))";
    rc = sqlite3_prepare_v2(db, sql.c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare SQL statement for creating the schema history table: {} ({})",
                 sql, sqlite3_errmsg(db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not create schema history table: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return true;
}

/**
 *
 */
static bool migrateFrom0To1(sqlite3* db)
{
    int rc {-1};    //
    std::ostringstream sql {""};    //
    sqlite3_stmt* stmt {nullptr};    //

    // Begin transaction
    sql.str("");
    sql << "BEGIN TRANSACTION;";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare the SQL statement for creating the indexed_granules table: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not execute the create indexed_granules table SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Create table for collections
    sql.str("");
    sql << "CREATE TABLE indexed_collections(";
    sql << "collection_id INTEGER NOT NULL,";
    sql << "collection_uri TEXT NOT NULL,";
    sql << "last_update INTEGER NOT NULL,";
    sql << "PRIMARY KEY(collection_id),";
    sql << "UNIQUE(collection_uri));";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare the SQL statement for creating the indexed_collections table: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not execute the create indexed_collections table SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Create table for granules
    sql.str("");
    sql << "CREATE TABLE indexed_granules(";
    sql << "granule_id INTEGER NOT NULL,";
    sql << "collection_id INTEGER NOT NULL,";
    sql << "time_coverage_start INTEGER,";
    sql << "time_coverage_end INTEGER,";
    sql << "data_id TEXT NOT NULL,";
    sql << "metadata BLOB NOT NULL,";
    sql << "PRIMARY KEY(granule_id),";
    sql << "FOREIGN KEY(collection_id) REFERENCES indexed_collections (collection_id) ON UPDATE CASCADE ON DELETE CASCADE);";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare the SQL statement for creating the indexed_granules table: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not execute the create indexed_granules table SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Create index to improve query performance when looking for granules matching search queries
    sql.str("");
    sql << "CREATE INDEX idx_collection_with_time_coverage ON indexed_granules(collection_id, time_coverage_start, time_coverage_end, (time_coverage_start + time_coverage_end) / 2)";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare the SQL statement for creating the idx_collection_with_time_coverage index: {} ({})",
                 sql.str(), sqlite3_errmsg(db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not execute the create idx_collection_with_time_coverage index SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Create index to improve query performance when checking if a granule is indexed based on its data_id
    sql.str("");
    sql << "CREATE INDEX idx_data_id ON indexed_granules(data_id)";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare the SQL statement for creating the idx_data_id index: {} ({})",
                 sql.str(), sqlite3_errmsg(db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not execute the create idx_data_id index SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Update schema history
    sql.str("");
    sql << "INSERT INTO granules_index_schema_history VALUES(1, STRFTIME('%s', 'NOW'));";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare the SQL statement for creating the indexed_granules table: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not execute the create indexed_granules table SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Commit transaction
    sql.str("");
    sql << "COMMIT TRANSACTION;";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare the SQL statement for creating the indexed_granules table: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not execute the create indexed_granules table SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return true;
}

/**
 *
 */
static bool applyMigrations(sqlite3* db)
{
    bool ok {false};    //
	int rc {-1};	// Return code for sqlite3 functions
    std::ostringstream sql {""};   //
    std::uint64_t currentVersion {0};   //
    sqlite3_stmt* stmt {nullptr};   //

    // Get latest version
    // In case the schema_history table is empty, currentVersion will not be
    // updated but that is not a problem because it has been initialized with
    // value 0
    sql << "SELECT MAX(version) FROM granules_index_schema_history;";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to prepare SQL statement for getting the latest version in schema history: {} ({})",
                 sql.str(), sqlite3_errmsg(db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t tmpValue {0};  //

        tmpValue = sqlite3_column_int64(stmt, 0);
        currentVersion = static_cast<std::uint64_t>(tmpValue);
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to execute the SQL statement for getting the latest version in schema history: {}",
                 sqlite3_errmsg(db));
        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Modify schema depending on current version
    while
        (currentVersion < SQLiteGranulesIndexSchema::latestVersion)
    {
        std::uint64_t originalVersion {currentVersion};

        // Apply migrations in ascending order

        // -- Migration 0 -> 1
        if
            (0 >= currentVersion)
        {
            ok = migrateFrom0To1(db);
            if
                (true == ok)
            {
                currentVersion = 1;
            }
        }

        // Failsafe: exit if no migration has been applied in this iteration of
        // the loop
        if
            (currentVersion == originalVersion)
        {
            LOG(LogLevel::Error, "SQLite granules index schema",
                     "Schema is locked in an older version ({}) and cannot be updated to latest version ({})",
                     currentVersion, SQLiteGranulesIndexSchema::latestVersion);
            return false;
        }
    }

    return true;
}

/**
 *
 */
static bool moveIncompatibleFile(const std::string& path)
{
    std::ostringstream newPath {""};
    time_t now = time(nullptr);
    struct tm tstruct;
    char buffer[80];
    tstruct = *localtime(&now);
    strftime(buffer, sizeof(buffer), "%Y%m%d%H%M%S", &tstruct);
    newPath << path << "." << buffer;

    int rc {-1};    //
    rc = std::rename(path.c_str(), newPath.str().c_str());
    if
        (0 == rc)
    {
        LOG(LogLevel::Warn, "SQLite granules index schema",
                 "File '{}' was not a valid SEAScope index and has been renamed to avoid issues. Please delete file '{}' if it has no use for you as it cannot be used as a SEAScope index",
                  path, newPath.str());
    }
    else
    {
        perror("Could not rename incompatible index file");
    }

    return  0 == rc;
}

/**
 *
 */
std::tuple<bool, bool> SQLiteGranulesIndexSchema::initialize(const std::string& path)
{
    int rc {-1};    //
    bool ok {false};    //
    bool hasContent {false};    //
    sqlite3* db {nullptr};  //
    std::ostringstream sql {""};    //

	// Open database
	rc = sqlite3_open(path.c_str(), &db);
	if
		(SQLITE_OK != rc)
	{
		LOG(LogLevel::Error, "SQLite granules index schema",
                 "Could not open database {}: {}",
                 path, sqlite3_errmsg(db));
		return {false, hasContent};
	}

    bool sqliteCompatible {false};

    auto result = SQLiteUtils::checkDatabase(db);
    sqliteCompatible = std::get<0>(result);
    if
        (false == sqliteCompatible)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
             "File '{}' is not an SQLite database, therefore SEAScope cannot use it for indexing granules",
             path);

        sqlite3_close(db);
        db = nullptr;

        ok = moveIncompatibleFile(path);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "SQLite granules index schema",
                     "Could not fix the problem automatically, please remove or rename the '{}' file or specify another path for the index in SEAScope configuration file",
                     path);
            return {false, hasContent};
        }

        // There should be no file matching the path now, so we can create the
        // database from scratch
        rc = sqlite3_open(path.c_str(), &db);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "SQLite granules index schema",
                     "Could not open database {}: {}",
                     path, sqlite3_errmsg(db));
            return {false, hasContent};
        }

        result = SQLiteUtils::checkDatabase(db);
        sqliteCompatible = std::get<0>(result);
    }

    ok = configure(db);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to configure database {}: {}",
                 path, sqlite3_errmsg(db));

        sqlite3_close(db);
        db = nullptr;

        return {false, hasContent};
    }

    bool initialized {false};
    initialized = std::get<1>(result);
    if
        (false == initialized)
    {
        // File does not exist or is empty / schema not initialized
        ok = create(db);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "SQLite granules index schema",
                     "Failed to create database '{}'", path);

            sqlite3_close(db);
            db = nullptr;

            return {false, hasContent};
        }
    }
    else
    {
        // File exists and has an SQLite schema, but does it have a
        // granules_index_schema_history table?
        ok = SQLiteUtils::hasTable(db, "granules_index_schema_history");
        if
            (false == ok)
        {
            // Make sure that none of the tables currently used by the schema
            // already exists
            ok = SQLiteUtils::hasTable(db, "indexed_granules");
            if
                (true == ok)
            {
                LOG(LogLevel::Error, "SQLite granules index schema",
                         "Database '{}' already contains a table named 'indexed_granules' but it does not seem like it has been created by SEAScope. Please delete this file or configure SEAScope to use another file for indexing granules", path);

                sqlite3_close(db);
                db = nullptr;

                return {false, hasContent};
            }

            ok = SQLiteUtils::hasTable(db, "indexed_collections");
            if
                (true == ok)
            {
                LOG(LogLevel::Error, "SQLite granules index schema",
                         "Database '{}' already contains a table named 'indexed_collections' but it does not seem like it has been created by SEAScope. Please delete this file or configure SEAScope to use another file for indexing granules", path);

                sqlite3_close(db);
                db = nullptr;

                return {false, hasContent};
            }

            // This is an SQLite file and it contains some tables but there is
            // no conflict with the name of the tables used by this schema
            ok = create(db);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "SQLite granules index schema",
                         "Failed to update database '{}' with tables required for indexing granules", path);

                sqlite3_close(db);
                db = nullptr;

                return {false, hasContent};
            }
        }
        else
        {
            // The database file has already been initialized, the required tables are there, so it may contain some already indexed data
            hasContent = true;
        }
    }

    ok = applyMigrations(db);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite granules index schema",
                 "Failed to apply schema migrations to '{}'", path);

        sqlite3_close(db);
        db = nullptr;

        return {false, hasContent};
    }

    sqlite3_close(db);
    db = nullptr;

    return {true, hasContent};
}
