// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <vector>

#include "vendors/_sqlite3.h"
#include "core/log.h"
#include "core/path.h"
#include "core/state.h"
#include "core/structs.h"
#include "sqlite_state.h"
#include "sqlite_state_schema.h"

using namespace SEAScope;

/**
 *
 */
SQLiteState::SQLiteState()
: _db{nullptr},
  _getSpacetimeInfoStmt{nullptr},
  _upsertSpacetimeInfoStmt{nullptr},
  _getLayoutsStmt{nullptr},
  _upsertLayoutStmt{nullptr}
{
};

/**
 *
 */
SQLiteState::~SQLiteState()
{
    if
        (nullptr != this->_getSpacetimeInfoStmt)
    {
        sqlite3_finalize(this->_getSpacetimeInfoStmt);
        this->_getSpacetimeInfoStmt = nullptr;
    }

    if
        (nullptr != this->_upsertSpacetimeInfoStmt)
    {
        sqlite3_finalize(this->_upsertSpacetimeInfoStmt);
        this->_upsertSpacetimeInfoStmt = nullptr;
    }

    if
        (nullptr != this->_getLayoutsStmt)
    {
        sqlite3_finalize(this->_getLayoutsStmt);
        this->_getLayoutsStmt = nullptr;
    }

    if
        (nullptr != this->_upsertLayoutStmt)
    {
        sqlite3_finalize(this->_upsertLayoutStmt);
        this->_upsertLayoutStmt = nullptr;
    }

    if
        (nullptr != this->_db)
    {
        sqlite3_close(this->_db);
        this->_db = nullptr;
    }
};

/**
 *
 */
bool SQLiteState::setStatePath(const std::string& path)
{
	if
		(false == Path::writeFile(path))
	{
		LOG( LogLevel::Error, "State SQLite"
				, "The state file {} is not writable"
				, path);
		return false;
	}
	this->_statePath = path;
	return true;
};

/**
 *
 */
bool SQLiteState::prepareStatements()
{
    int rc {-1};    //
    std::ostringstream sql {""};    //

    // Get spacetime info
    sql.str("");
    sql << "SELECT CAM_LOCK, CAM0, CAM1, CAM2, CAM3, ZOOM, TIMESPAN, TIMESTEP, CURRENT_DATETIME, NEAREST_MODE, SHOW_CALENDARS FROM spacetime WHERE id = 0 LIMIT 1;";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_getSpacetimeInfoStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state",
                 "Failed to prepare SQL statement for getting the spacetime info: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Upsert spacetime info
    sql.str("");
    sql << "INSERT OR REPLACE INTO spacetime ";
    sql << "(id, CAM_LOCK, CAM0, CAM1, CAM2, CAM3, ZOOM, TIMESPAN, TIMESTEP, CURRENT_DATETIME, NEAREST_MODE, SHOW_CALENDARS) ";
    sql << "VALUES(0, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_upsertSpacetimeInfoStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state",
                 "Failed to prepare SQL statement for upserting the spacetime info: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Get layouts
    sql.str("");
    sql << "SELECT ID,X,Y,WIDTH,HEIGHT,MOVABLE,RESIZABLE FROM layout WHERE 1;";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_getLayoutsStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state",
                 "Failed to prepare SQL statement for getting the saved view: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    // Upsert layout
    sql.str("");
    sql << "INSERT OR REPLACE INTO layout ";
	sql << "(NAME,ID,X,Y,WIDTH,HEIGHT,MOVABLE,RESIZABLE) ";
	sql << "VALUES('USER', ?, ?, ?, ?, ?, ?, ?);";
    rc = sqlite3_prepare_v2(this->_db, sql.str().c_str(), -1, &(this->_upsertLayoutStmt), nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state",
                 "Failed to prepare SQL statement for upserting a UI component layout: {} ({})",
                 sql.str(), sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteState::getLayouts(State& state)
{
    int rc {-1};    //
    const unsigned char* tmpStr {nullptr};  //

    sqlite3_reset(this->_getLayoutsStmt);

    rc = sqlite3_step(this->_getLayoutsStmt);
    while
        (SQLITE_ROW == rc)
    {
        // ID
        tmpStr = sqlite3_column_text(this->_getLayoutsStmt, 0);
	    std::string id(reinterpret_cast<const char*>(tmpStr));	//

        // X
        tmpStr = sqlite3_column_text(this->_getLayoutsStmt, 1);
	    std::string x(reinterpret_cast<const char*>(tmpStr));	//

        // Y
        tmpStr = sqlite3_column_text(this->_getLayoutsStmt, 2);
	    std::string y(reinterpret_cast<const char*>(tmpStr));	//

        // WIDTH
        tmpStr = sqlite3_column_text(this->_getLayoutsStmt, 3);
	    std::string w(reinterpret_cast<const char*>(tmpStr));	//

        // HEIGHT
        tmpStr = sqlite3_column_text(this->_getLayoutsStmt, 4);
	    std::string h(reinterpret_cast<const char*>(tmpStr));	//

        // MOVABLE
	    bool movable(false);	//
        movable = (0 < sqlite3_column_int(this->_getLayoutsStmt, 5));

        // RESIZABLE
	    bool resizable(false);	//
        resizable = (0 < sqlite3_column_int(this->_getLayoutsStmt, 6));

        auto found = state.layout.find(id);
        if
            (state.layout.end() == found)
        {
	        IBlueprint* blueprint(nullptr);	//
            blueprint = state.createBlueprint(resizable, movable);
            blueprint->setLayout(x, y, w, h);
            state.layout.insert(std::make_pair(id, blueprint));
        }
        else
        {
            found->second->setLayout(x, y, w, h);
        }

        rc = sqlite3_step(this->_getLayoutsStmt);
    }

	if
		(SQLITE_DONE != rc)
	{
		LOG( LogLevel::Error, "State SQLite"
				, "Could not retrieve layout: {}", sqlite3_errmsg(this->_db));
		return false;
	}

    return true;
}

/**
 *
 */
bool SQLiteState::getSpacetimeInfo(SceneViewState& viewState, TimelineState& timelineState)
{
    int rc {-1};    //
    const unsigned char* tmpStr {nullptr};  //

    sqlite3_reset(this->_getSpacetimeInfoStmt);

    rc = sqlite3_step(this->_getSpacetimeInfoStmt);
    while
        (SQLITE_ROW == rc)
    {
        //CAM_LOCK, CAM0, CAM1, CAM2, CAM3, ZOOM, TIMESPAN, TIMESTEP, CURRENT_DATETIME, NEAREST_MODE, SHOW_CALENDARS
        viewState.lock = (0 < sqlite3_column_int(this->_getSpacetimeInfoStmt, 0));
        viewState.q.w = static_cast<float>(sqlite3_column_double(this->_getSpacetimeInfoStmt, 1));
        viewState.q.x = static_cast<float>(sqlite3_column_double(this->_getSpacetimeInfoStmt, 2));
        viewState.q.y = static_cast<float>(sqlite3_column_double(this->_getSpacetimeInfoStmt, 3));
        viewState.q.z = static_cast<float>(sqlite3_column_double(this->_getSpacetimeInfoStmt, 4));
        viewState.zoom = static_cast<float>(sqlite3_column_double(this->_getSpacetimeInfoStmt, 5));

        tmpStr = sqlite3_column_text(this->_getSpacetimeInfoStmt, 6);
        if
            (nullptr != tmpStr)
        {
            timelineState.timespan = std::string(reinterpret_cast<const char*>(tmpStr));
        }

        tmpStr = sqlite3_column_text(this->_getSpacetimeInfoStmt, 7);
        if
            (nullptr != tmpStr)
        {
            timelineState.timestep = std::string(reinterpret_cast<const char*>(tmpStr));
        }

        tmpStr = sqlite3_column_text(this->_getSpacetimeInfoStmt, 8);
        if
            (nullptr != tmpStr)
        {
            timelineState.datetime = std::string(reinterpret_cast<const char*>(tmpStr));
        }

        timelineState.nearestMode = (0 < sqlite3_column_int(this->_getSpacetimeInfoStmt, 9));
        timelineState.extendedView = (0 < sqlite3_column_int(this->_getSpacetimeInfoStmt, 10));

        rc = sqlite3_step(this->_getSpacetimeInfoStmt);
    }

	if
		(SQLITE_DONE != rc)
	{
		LOG( LogLevel::Error, "State SQLite"
				, "Could not retrieve spacetime info: {}", sqlite3_errmsg(this->_db));
		return false;
	}

    return true;
}

/**
 *
 */
bool SQLiteState::load(State& state, SceneViewState& viewState, TimelineState& timelineState)
{

    int rc {-1};    //
    bool ok {false};    //
    std::tuple<bool, bool> initResult {false, false};  //
    bool initSuccess {false};   //
	std::string sql{""};	// SQL query


    initResult = SQLiteStateSchema::initialize(this->_statePath);
    initSuccess = std::get<0>(initResult);
    if
        (false == initSuccess)
    {
        LOG(LogLevel::Error, "SQLite state",
                 "Failed to initialize database '{}'", this->_statePath);
        return false;
    }

	// Open database
	rc = sqlite3_open(this->_statePath.c_str(), &(this->_db));
	if
		(SQLITE_OK != rc)
	{
		LOG(LogLevel::Error, "SQLite state",
                 "Could not open database {}: {}",
                 this->_statePath, sqlite3_errmsg(this->_db));
		return false;
	}

    ok = SQLiteStateSchema::configure(this->_db);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite state",
                 "Could not configure database {}", this->_statePath);
        return false;
    }

    // Prepare statements
    ok = this->prepareStatements();
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite state",
                 "Failed to prepare SQL statements");
        return false;
    }

	// Read layouts
    ok = this->getLayouts(state);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite state",
                 "Failed to read layouts");
        return false;
    }

    // Read spacetime info
    ok = this->getSpacetimeInfo(viewState, timelineState);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite state",
                 "Failed to read spacetime info");
        return false;
    }

    return true;
};

/**
 *
 */
bool SQLiteState::saveLayouts(const State& state) const
{
	int rc(-1);	// Return code for sqlite3 functions
	IComponent* component {nullptr};    //
    std::string x {""}; //
    std::string y {""}; //
    std::string w {""}; //
    std::string h {""}; //

	auto it = state.layout.begin();
	for
		(; it != state.layout.end(); it++)
	{
		component = it->second->getTarget();
		if
			(nullptr == component)
		{
			LOG( LogLevel::Warn, "State SQLite"
					, "Component {} is null", it->first);
			continue;
		}
		it->second->getLayout(x,y,w,h);

        sqlite3_reset(this->_upsertLayoutStmt);

        // ID
        rc = sqlite3_bind_text(this->_upsertLayoutStmt, 1, it->first.c_str(), static_cast<int>(it->first.length()), nullptr /* <=> SQLITE_STATIC*/);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "State SQLite",
                     "Failed to bind the TEXT ID argument to the SQL statement");
            return false;
        }

        // X
        rc = sqlite3_bind_text(this->_upsertLayoutStmt, 2, x.c_str(), static_cast<int>(x.length()), nullptr /* <=> SQLITE_STATIC*/);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "State SQLite",
                     "Failed to bind the TEXT X argument to the SQL statement");
            return false;
        }

        // Y
        rc = sqlite3_bind_text(this->_upsertLayoutStmt, 3, y.c_str(), static_cast<int>(y.length()), nullptr /* <=> SQLITE_STATIC*/);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "State SQLite",
                     "Failed to bind the TEXT Y argument to the SQL statement");
            return false;
        }

        // WIDTH
        rc = sqlite3_bind_text(this->_upsertLayoutStmt, 4, w.c_str(), static_cast<int>(w.length()), nullptr /* <=> SQLITE_STATIC*/);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "State SQLite",
                     "Failed to bind the TEXT WIDTH argument to the SQL statement");
            return false;
        }

        // HEIGHT
        rc = sqlite3_bind_text(this->_upsertLayoutStmt, 5, h.c_str(), static_cast<int>(h.length()), nullptr /* <=> SQLITE_STATIC*/);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "State SQLite",
                     "Failed to bind the TEXT HEIGHT argument to the SQL statement");
            return false;
        }

        // MOVABLE
		rc = sqlite3_bind_int(this->_upsertLayoutStmt, 6, (component->isMovable() ? 1:0));
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "State SQLite",
                     "Failed to bind the BOOL MOVABLE argument to the SQL statement");
            return false;
        }

        // RESIZABLE
		rc = sqlite3_bind_int(this->_upsertLayoutStmt, 7, (component->isResizable() ? 1:0));
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "State SQLite",
                     "Failed to bind the BOOL MOVABLE argument to the SQL statement");
            return false;
        }

        rc = sqlite3_step(this->_upsertLayoutStmt);
        while
            (SQLITE_ROW == rc)
        {
            rc = sqlite3_step(this->_upsertLayoutStmt);
        }

        if
            (SQLITE_DONE != rc)
        {
            LOG( LogLevel::Error, "State SQLite"
                    , "Could not save layout: {}", sqlite3_errmsg(this->_db));
            return false;
        }

    }

    return true;
}

/**
 *
 */
bool SQLiteState::saveSpacetimeInfo(const SceneViewState& viewState, const TimelineState& timelineState) const
{
	int rc {-1};	// Return code for sqlite3 functions

    sqlite3_reset(this->_upsertSpacetimeInfoStmt);

    // Lock
    rc = sqlite3_bind_int(this->_upsertSpacetimeInfoStmt, 1, (viewState.lock ? 1:0));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the BOOL CAM_LOCK argument to the SQL statement");
        return false;
    }

    // Quaternion w
    rc = sqlite3_bind_double(this->_upsertSpacetimeInfoStmt, 2, static_cast<double>(viewState.q.w));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the REAL CAM0 argument to the SQL statement");
        return false;
    }

    // Quaternion x
    rc = sqlite3_bind_double(this->_upsertSpacetimeInfoStmt, 3, static_cast<double>(viewState.q.x));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the REAL CAM1 argument to the SQL statement");
        return false;
    }

    // Quaternion y
    rc = sqlite3_bind_double(this->_upsertSpacetimeInfoStmt, 4, static_cast<double>(viewState.q.y));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the REAL CAM2 argument to the SQL statement");
        return false;
    }

    // Quaternion z
    rc = sqlite3_bind_double(this->_upsertSpacetimeInfoStmt, 5, static_cast<double>(viewState.q.z));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the REAL CAM3 argument to the SQL statement");
        return false;
    }
    // Zoom
    rc = sqlite3_bind_double(this->_upsertSpacetimeInfoStmt, 6, static_cast<double>(viewState.zoom));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the REAL ZOOM argument to the SQL statement");
        return false;
    }
    // Timespan
    rc = sqlite3_bind_text(this->_upsertSpacetimeInfoStmt, 7, timelineState.timespan.c_str(), static_cast<int>(timelineState.timespan.length()), nullptr /* <=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the TEXT TIMESPAN argument to the SQL statement");
        return false;
    }

    // Timestep
    rc = sqlite3_bind_text(this->_upsertSpacetimeInfoStmt, 8, timelineState.timestep.c_str(), static_cast<int>(timelineState.timestep.length()), nullptr /* <=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the TEXT TIMESTEP argument to the SQL statement");
        return false;
    }

    // Current datetime
    rc = sqlite3_bind_text(this->_upsertSpacetimeInfoStmt, 9, timelineState.datetime.c_str(), static_cast<int>(timelineState.datetime.length()), nullptr /* <=> SQLITE_STATIC*/);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the TEXT CURRENT_DATETIME argument to the SQL statement");
        return false;
    }

    // Nearest mode
    rc = sqlite3_bind_int(this->_upsertSpacetimeInfoStmt, 10, (timelineState.nearestMode ? 1:0));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the BOOL NEAERST_MODE argument to the SQL statement");
        return false;
    }

    // Show calendars
    rc = sqlite3_bind_int(this->_upsertSpacetimeInfoStmt, 11, (timelineState.extendedView ? 1:0));
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "State SQLite",
                 "Failed to bind the BOOL SHOW_CALENDARS argument to the SQL statement");
        return false;
    }

    rc = sqlite3_step(this->_upsertSpacetimeInfoStmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(this->_upsertSpacetimeInfoStmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG( LogLevel::Error, "State SQLite"
                , "Could not save spacetime info: {}", sqlite3_errmsg(this->_db));
        return false;
    }

    return true;
}

/**
 *
 */
bool SQLiteState::save(const State& state, const SceneViewState& viewState, const TimelineState& timelineState) const
{
    bool ok {false};    //

    LOG(LogLevel::Debug, "State SQLite",
             "Saving...");
    if
        (nullptr == this->_db)
    {
        LOG(LogLevel::Debug, "State SQLite", "Database link is null");
        return false;
    }

    // Save layouts
    ok = this->saveLayouts(state);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "State SQLite", "Could not save layout of UI components");
        return false;
    }

    // Save spacetime info
    ok = this->saveSpacetimeInfo(viewState, timelineState);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "State SQLite", "Could not save spacetime info");
        return false;
    }

	return true;
};

/**
 *
 */
int SQLiteState::renderParamsRetrievedCallback(void* /*result*/, int /*argc*/, char** /*argv*/, char** /*colName*/)
{
	return 1;
};

/**
 *
 */
int SQLiteState::timelineRetrievedCallback(void* /*result*/, int /*argc*/, char** /*argv*/, char** /*colName*/)
{
	return 1;
};
