// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <tuple>
#include <cstdio>

#include "vendors/_sqlite3.h"

#include "core/log.h"
#include "core/path.h"

#include "sqlite_state_schema.h"
#include "sqlite_utils.h"

using namespace SEAScope;

/**
 *
 */
bool SQLiteStateSchema::configure(sqlite3* /*db*/)
{
    // No pragma statements required for now
    return true;
}

/**
 *
 */
static bool create(sqlite3* db)
{
    int rc {-1};    //
    std::string sql {""};   //
    sqlite3_stmt* stmt {nullptr};   //

    // Create schema history table
    sql = "CREATE TABLE state_schema_history(" \
          "version INTEGER NOT NULL," \
          "datetime INTEGER NOT NULL," \
          "PRIMARY KEY(version))";
    rc = sqlite3_prepare_v2(db, sql.c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare SQL statement for creating the schema history table: {} ({})",
                 sql, sqlite3_errmsg(db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not create schema history table: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return true;
}

/**
 *
 */
static bool migrateFrom0To1(sqlite3* db)
{
    int rc {-1};    //
    bool ok {false};    //
    std::ostringstream sql {""};    //
    sqlite3_stmt* stmt {nullptr};    //

    // Begin transaction
    sql.str("");
    sql << "BEGIN TRANSACTION;";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare the SQL statement for starting a transaction: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not execute the begin transaction SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Create table for layout if needed (no changes between v0 and v1)
    ok = SQLiteUtils::hasTable(db, "layout");
    if
        (false == ok)
    {
        sql.str("");
        sql << "CREATE TABLE layout(";
        sql << "NAME TEXT NOT NULL,";
        sql << "ID TEXT NOT NULL,";
        sql << "X TEXT NOT NULL,";
        sql << "Y TEXT NOT NULL,";
        sql << "WIDTH TEXT NOT NULL,";
        sql << "HEIGHT TEXT NOT NULL,";
        sql << "MOVABLE BOOLEAN NOT NULL,";
        sql << "RESIZABLE BOOLEAN NOT NULL,";
        sql << "PRIMARY KEY (NAME, ID));";

        rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Failed to prepare the SQL statement for creating the layout table: {} ({})",
                     sql.str(), sqlite3_errmsg(db));

            return false;
        }

        rc = sqlite3_step(stmt);
        while
            (SQLITE_ROW == rc)
        {
            rc = sqlite3_step(stmt);
        }

        if
            (SQLITE_DONE != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Could not execute the create layout table SQL statement: {}",
                     sqlite3_errmsg(db));

            sqlite3_finalize(stmt);
            stmt = nullptr;

            return false;
        }

        sqlite3_finalize(stmt);
        stmt = nullptr;
    }

    // Create or alter table for spacetime
    ok = SQLiteUtils::hasTable(db, "spacetime");
    if
        (true == ok)
    {
        // Delete existing table as it was not used in previous versions of
        // SEAScope
        sql.str("");
        sql << "DROP TABLE 'spacetime';";

        rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Failed to prepare the SQL statement for deleting the spacetime table: {} ({})",
                     sql.str(), sqlite3_errmsg(db));

            return false;
        }

        rc = sqlite3_step(stmt);
        while
            (SQLITE_ROW == rc)
        {
            rc = sqlite3_step(stmt);
        }

        if
            (SQLITE_DONE != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Could not execute the delete spacetime table SQL statement: {}",
                     sqlite3_errmsg(db));

            sqlite3_finalize(stmt);
            stmt = nullptr;

            return false;
        }

        sqlite3_finalize(stmt);
        stmt = nullptr;
    }

    sql.str("");
    sql << "CREATE TABLE spacetime(";
    sql << "id INT UNIQUE,";
    sql << "CAM_LOCK BOOL,";
    sql << "CAM0 REAL,";
    sql << "CAM1 REAL,";
    sql << "CAM2 REAL,";
    sql << "CAM3 REAL,";
    sql << "ZOOM REAL,";
    sql << "TIMEFRAME_PAST INT,";
    sql << "TIMEFRAME_FUTURE INT,";
    sql << "TIMESTEP INT,";
    sql << "CURRENT_TIMESTAMP,";
    sql << "PRIMARY KEY(id));";

    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare the SQL statement for creating the spacetime table: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not execute the create spacetime table SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Remove unused 'seascope' table if it exists
    ok = SQLiteUtils::hasTable(db, "seascope");
    if
        (true == ok)
    {
        sql.str("");
        sql << "DROP TABLE 'seascope';";

        rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Failed to prepare the SQL statement for removing deprecated 'seascope' table: {} ({})",
                     sql.str(), sqlite3_errmsg(db));

            return false;
        }

        rc = sqlite3_step(stmt);
        while
            (SQLITE_ROW == rc)
        {
            rc = sqlite3_step(stmt);
        }

        if
            (SQLITE_DONE != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Could not execute the remove 'seascope' table SQL statement: {}",
                     sqlite3_errmsg(db));

            sqlite3_finalize(stmt);
            stmt = nullptr;

            return false;
        }

        sqlite3_finalize(stmt);
        stmt = nullptr;
    }

    // Remove unused 'rendering' table if it exists
    ok = SQLiteUtils::hasTable(db, "rendering");
    if
        (true == ok)
    {
        sql.str("");
        sql << "DROP TABLE 'rendering';";

        rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Failed to prepare the SQL statement for removing deprecated 'rendering' table: {} ({})",
                     sql.str(), sqlite3_errmsg(db));

            return false;
        }

        rc = sqlite3_step(stmt);
        while
            (SQLITE_ROW == rc)
        {
            rc = sqlite3_step(stmt);
        }

        if
            (SQLITE_DONE != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Could not execute the remove 'rendering' table SQL statement: {}",
                     sqlite3_errmsg(db));

            sqlite3_finalize(stmt);
            stmt = nullptr;

            return false;
        }

        sqlite3_finalize(stmt);
        stmt = nullptr;
    }

    // Update schema history
    sql.str("");
    sql << "INSERT INTO state_schema_history VALUES(1, STRFTIME('%s', 'NOW'));";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare the SQL statement for inserting the new version in the state schema history: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not execute the insert new schema version SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Commit transaction
    sql.str("");
    sql << "COMMIT TRANSACTION;";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare the SQL statement for committing the transaction: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not execute the commit transaction SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return true;
}

/**
 *
 */
static bool migrateFrom1To2(sqlite3* db)
{
    int rc {-1};    //
    bool ok {false};    //
    std::ostringstream sql {""};    //
    sqlite3_stmt* stmt {nullptr};    //

    // Begin transaction
    sql.str("");
    sql << "BEGIN TRANSACTION;";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare the SQL statement for starting a transaction: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not execute the begin transaction SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // We need to remove columns but SQLite ALTER statements do not support
    // that so we create a new table
    sql.str("");
    sql << "CREATE TABLE spacetime_new(";
    sql << "id INT UNIQUE,";
    sql << "CAM_LOCK BOOL,";
    sql << "CAM0 REAL,";
    sql << "CAM1 REAL,";
    sql << "CAM2 REAL,";
    sql << "CAM3 REAL,";
    sql << "ZOOM REAL,";
    sql << "TIMESPAN TEXT DEFAULT '1d',";
    sql << "TIMESTEP TEXT DEFAULT '1/3 span',";
    sql << "CURRENT_DATETIME TEXT,";
    sql << "NEAREST_MODE BOOL DEFAULT 1,";
    sql << "SHOW_CALENDARS BOOL DEFAULT 1,";
    sql << "PRIMARY KEY(id));";

    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare the SQL statement for creating the spacetime_new table: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not execute the create spacetime_new table SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // If the table existed before, try to preserve content used in version 1
    // of the schema (camera state) by copying the relevant fields to the new table
    ok = SQLiteUtils::hasTable(db, "spacetime");
    if
        (true == ok)
    {
        sql.str("");
        sql << "INSERT INTO spacetime_new(id, CAM_LOCK, CAM0, CAM1, CAM2, CAM3, ZOOM) SELECT id, CAM_LOCK, CAM0, CAM1, CAM2, CAM3, ZOOM FROM spacetime;";
        rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Failed to prepare the SQL statement for copying content of the spacetime table to the spacetime_new table: {} ({})",
                     sql.str(), sqlite3_errmsg(db));

            return false;
        }

        rc = sqlite3_step(stmt);
        while
            (SQLITE_ROW == rc)
        {
            rc = sqlite3_step(stmt);
        }

        if
            (SQLITE_DONE != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Could not execute the copy content from spacetime to spacetime_new SQL statement: {}",
                     sqlite3_errmsg(db));

            sqlite3_finalize(stmt);
            stmt = nullptr;

            return false;
        }

        sqlite3_finalize(stmt);
        stmt = nullptr;

        // Delete existing table now that the content has been copied
        sql.str("");
        sql << "DROP TABLE 'spacetime';";

        rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Failed to prepare the SQL statement for deleting the spacetime table: {} ({})",
                     sql.str(), sqlite3_errmsg(db));

            return false;
        }

        rc = sqlite3_step(stmt);
        while
            (SQLITE_ROW == rc)
        {
            rc = sqlite3_step(stmt);
        }

        if
            (SQLITE_DONE != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Could not execute the delete spacetime table SQL statement: {}",
                     sqlite3_errmsg(db));

            sqlite3_finalize(stmt);
            stmt = nullptr;

            return false;
        }

        sqlite3_finalize(stmt);
        stmt = nullptr;
    }

    // Rename new table to replace old table
    sql.str("");
    sql << "ALTER TABLE 'spacetime_new' RENAME TO 'spacetime';";

    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare the SQL statement for renaming the spacetime_new table to spacetime: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not execute the rename spacetime_new table to spacetime  SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Update schema history
    sql.str("");
    sql << "INSERT INTO state_schema_history VALUES(2, STRFTIME('%s', 'NOW'));";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare the SQL statement for inserting the new version in the state schema history: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not execute the insert new schema version SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Commit transaction
    sql.str("");
    sql << "COMMIT TRANSACTION;";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare the SQL statement for committing the transaction: {} ({})",
                 sql.str(), sqlite3_errmsg(db));

        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Could not execute the commit transaction SQL statement: {}",
                 sqlite3_errmsg(db));

        sqlite3_finalize(stmt);
        stmt = nullptr;

        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return true;
}
/**
 *
 */
static bool applyMigrations(sqlite3* db)
{
    bool ok {false};    //
	int rc {-1};	// Return code for sqlite3 functions
    std::ostringstream sql {""};   //
    std::uint64_t currentVersion {0};   //
    sqlite3_stmt* stmt {nullptr};   //

    // Get latest version
    // In case the schema_history table is empty, currentVersion will not be
    // updated but that is not a problem because it has been initialized with
    // value 0
    sql << "SELECT MAX(version) FROM state_schema_history;";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to prepare SQL statement for getting the latest version in schema history: {} ({})",
                 sql.str(), sqlite3_errmsg(db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t tmpValue {0};  //

        tmpValue = sqlite3_column_int64(stmt, 0);
        currentVersion = static_cast<std::uint64_t>(tmpValue);
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to execute the SQL statement for getting the latest version in schema history: {}",
                 sqlite3_errmsg(db));
        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    // Modify schema depending on current version
    while
        (currentVersion < SQLiteStateSchema::latestVersion)
    {
        std::uint64_t originalVersion {currentVersion};

        // Apply migrations in ascending order

        // -- Migration 0 -> 1
        if
            (0 >= currentVersion)
        {
            ok = migrateFrom0To1(db);
            if
                (true == ok)
            {
                currentVersion = 1;
            }
        }

        // -- Migration 1 -> 2
        if
            (1 == currentVersion)
        {
            ok = migrateFrom1To2(db);
            if
                (true == ok)
            {
                currentVersion = 2;
            }
        }

        // Failsafe: exit if no migration has been applied in this iteration of
        // the loop
        if
            (currentVersion == originalVersion)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Schema is locked in an older version ({}) and cannot be updated to latest version ({})",
                     currentVersion, SQLiteStateSchema::latestVersion);
            return false;
        }
    }

    return true;
}

/**
 *
 */
static bool moveIncompatibleFile(const std::string& path)
{
    std::ostringstream newPath {""};
    time_t now = time(nullptr);
    struct tm tstruct;
    char buffer[80];
    tstruct = *localtime(&now);
    strftime(buffer, sizeof(buffer), "%Y%m%d%H%M%S", &tstruct);
    newPath << path << "." << buffer;

    int rc {-1};    //
    rc = std::rename(path.c_str(), newPath.str().c_str());
    if
        (0 == rc)
    {
        LOG(LogLevel::Warn, "SQLite state schema",
                 "File '{}' was not a valid SEAScope state file and has been renamed to avoid issues. Please delete file '{}' if it has no use for you as it cannot be used as a SEAScope state file",
                  path, newPath.str());
    }
    else
    {
        perror("Could not rename incompatible state file");
    }

    return  0 == rc;
}

/**
 *
 */
std::tuple<bool, bool> SQLiteStateSchema::initialize(const std::string& path)
{
    int rc {-1};    //
    bool ok {false};    //
    bool hasContent {false};    //
    sqlite3* db {nullptr};  //
    std::ostringstream sql {""};    //

	// Open database
	rc = sqlite3_open(path.c_str(), &db);
	if
		(SQLITE_OK != rc)
	{
		LOG(LogLevel::Error, "SQLite state schema",
                 "Could not open database {}: {}",
                 path, sqlite3_errmsg(db));
		return {false, hasContent};
	}

    bool sqliteCompatible {false};

    auto result = SQLiteUtils::checkDatabase(db);
    sqliteCompatible = std::get<0>(result);
    if
        (false == sqliteCompatible)
    {
        LOG(LogLevel::Error, "SQLite state schema",
             "File '{}' is not an SQLite database, therefore SEAScope cannot use it for storing its state",
             path);

        sqlite3_close(db);
        db = nullptr;

        ok = moveIncompatibleFile(path);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Could not fix the problem automatically, please remove or rename the '{}' file or specify another path for the state in SEAScope configuration file",
                     path);
            return {false, hasContent};
        }

        // There should be no file matching the path now, so we can create the
        // database from scratch
        rc = sqlite3_open(path.c_str(), &db);
        if
            (SQLITE_OK != rc)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Could not open database {}: {}",
                     path, sqlite3_errmsg(db));
            return {false, hasContent};
        }

        result = SQLiteUtils::checkDatabase(db);
        sqliteCompatible = std::get<0>(result);
    }

    ok = SQLiteStateSchema::configure(db);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to configure database {}: {}",
                 path, sqlite3_errmsg(db));

        sqlite3_close(db);
        db = nullptr;

        return {false, hasContent};
    }

    bool initialized {false};
    initialized = std::get<1>(result);
    if
        (false == initialized)
    {
        // File does not exist or is empty / schema not initialized
        ok = create(db);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "SQLite state schema",
                     "Failed to create database '{}'", path);

            sqlite3_close(db);
            db = nullptr;

            return {false, hasContent};
        }
    }
    else
    {
        // File exists and has an SQLite schema, but does it have a
        // state_schema_history table?
        ok = SQLiteUtils::hasTable(db, "state_schema_history");
        if
            (false == ok)
        {
            // Make sure that none of the tables currently used by the schema
            // (and not present in ealier versions) already exists
            //
            // => nothing to do, all tables already existed in previous
            //    versions of SEAScope

            // This is an SQLite file and it contains some tables but there is
            // no conflict with the name of the tables used by this schema
            ok = create(db);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "SQLite state schema",
                         "Failed to update database '{}' with tables required for storing the application state", path);

                sqlite3_close(db);
                db = nullptr;

                return {false, hasContent};
            }
        }
        else
        {
            // The database file has already been initialized, the required tables are there, so it may contain an application state
            hasContent = true;
        }
    }

    ok = applyMigrations(db);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "SQLite state schema",
                 "Failed to apply schema migrations to '{}'", path);

        sqlite3_close(db);
        db = nullptr;

        return {false, hasContent};
    }

    sqlite3_close(db);
    db = nullptr;

    return {true, hasContent};
}
