// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cstdint>
#include <sstream>

#include "vendors/_sqlite3.h"
#include "sqlite_utils.h"

#include "core/log.h"

using namespace SEAScope;

/**
 * Callback for SQLite trace
 */
static int printTrace(unsigned traceType, void* ctx, void* /*stmt*/, void* details)
{
    switch
        (traceType)
    {
        case SQLITE_TRACE_STMT:
            LOG(LogLevel::Debug, "SQLite utils", "Execute statement: {}",
                     static_cast<char*>(details));
            break;
        case SQLITE_TRACE_PROFILE:
            LOG(LogLevel::Debug, "SQLite utils", "Profile statement execution: {} ns -- {}",
                     *(static_cast<sqlite3_uint64*>(details)), static_cast<char*>(ctx));
            /*
            LOG(LogLevel::Debug, "SQLite utils", "Profile statement execution: {} ns -- {}",
                     *(static_cast<std::int64_t*>(details)), sqlite3_expanded_sql(static_cast<sqlite3_stmt*>(stmt)));
            */
            break;
    }

    return 0;
}

/**
 *
 */
static void printTraceLegacy(void* ctx, const char* sql)
{
    printTrace(SQLITE_TRACE_STMT, ctx, nullptr, reinterpret_cast<void*>(const_cast<char*>(sql)));
}

/**
 *
 */
static void printProfileLegacy(void* /*ctx*/, const char* sql, sqlite3_uint64 nanoseconds)
{
    printTrace(SQLITE_TRACE_PROFILE, reinterpret_cast<void*>(const_cast<char*>(sql)), nullptr, reinterpret_cast<void*>(&nanoseconds));
}

/**
 *
 */
bool SQLiteUtils::showTrace(sqlite3* db)
{

    sqlite3_trace(db, printTraceLegacy, nullptr);
    sqlite3_profile(db, printProfileLegacy, nullptr);

    /*
    int rc {-1};    //
    rc = sqlite3_trace_v2(db, SQLITE_TRACE_STMT|SQLITE_TRACE_PROFILE, printTrace, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite utils",
                 "Failed to install trace hooks to monitor SQLite activity: {}",
                  sqlite3_errmsg(db));
        return false;
    }
    */
    return true;
}

/**
 *
 */
std::tuple<bool, bool> SQLiteUtils::checkDatabase(sqlite3* db)
{
    int rc {-1};    //
    std::uint64_t revCount {0}; //
    std::ostringstream sql {""}; //
    sqlite3_stmt* stmt {nullptr};   //

    sql << "PRAGMA schema_version";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite utils",
                "Failed to prepare SQL statement for getting the schema version: {} ({})",
                sql.str(), sqlite3_errmsg(db));
        return {false, false};
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        std::int64_t tmpValue {0};
        tmpValue = sqlite3_column_int64(stmt, 0);
        revCount = static_cast<std::uint64_t>(tmpValue);

        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite utils",
                 "Failed to execute SQL statement: {}", sqlite3_errmsg(db));
        return {false, false};
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    bool initialized {false};
    initialized = (0 < revCount);

    return {true, initialized};
}

/**
 *
 */
bool SQLiteUtils::hasTable(sqlite3* db, const std::string& tableName)
{
    int rc {-1};    //
    int tablesCount {0};    //
    sqlite3_stmt* stmt {nullptr}; //
    std::ostringstream sql {""};    //

    sql << "SELECT COUNT(*) FROM sqlite_master WHERE type='table' AND name='" << tableName << "'";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite utils",
                 "Failed to prepare SQL statement for counting tables named '{}': {} ({})",
                 tableName, sql.str(), sqlite3_errmsg(db));
        return false;
    }

    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        tablesCount = sqlite3_column_int(stmt, 0);
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite utils",
                 "Failed to execute SQL statement: {}", sqlite3_errmsg(db));
        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return 1 == tablesCount;
}

/**
 *
 */
bool SQLiteUtils::isDatabaseHealthy(sqlite3* db)
{
    int rc {-1};    //
    const unsigned char* tmpStr {nullptr};  //
    sqlite3_stmt* stmt {nullptr}; //
    std::ostringstream sql {""};    //

    sql << "PRAGMA integrity_check;";
    rc = sqlite3_prepare_v2(db, sql.str().c_str(), -1, &stmt, nullptr);
    if
        (SQLITE_OK != rc)
    {
        LOG(LogLevel::Error, "SQLite utils",
                 "Failed to prepare SQL statement for checking database integrity: {} ({})",
                 sql.str(), sqlite3_errmsg(db));
        return false;
    }

    bool ok {false};    //
    rc = sqlite3_step(stmt);
    while
        (SQLITE_ROW == rc)
    {
        if
            (false == ok)
        {
            tmpStr = sqlite3_column_text(stmt, 0);
            ok = (0 == strncmp("ok", reinterpret_cast<const char*>(tmpStr), 2));
        }
        rc = sqlite3_step(stmt);
    }

    if
        (SQLITE_DONE != rc)
    {
        LOG(LogLevel::Error, "SQLite utils",
                 "Failed to execute SQL statement: {}", sqlite3_errmsg(db));
        return false;
    }

    sqlite3_finalize(stmt);
    stmt = nullptr;

    return ok;
}
