// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <fstream>

#include "tools/annotations_controller.h"

#include "core/log.h"

#include "core/interfaces/iannotations_store.h"
#include "core/interfaces/igranules_index.h"

#include "core/granule_metadata.h"
#include "core/app_events_dispatcher.h"
#include "local/user_generated.h"
#include "core/source.h"
#include "core/collections_control.h"
#include "core/rendering_control.h"
#include "3d/scene_view.h"
#include "3d/iscene.h"
#include "3d/iscene_renderer.h"

#include "events/collections_update_requested_event.h"
#include "events/add_prepared_annotation_request.h"
#include "events/collection_registration_request.h"
#include "events/save_one_annotation_to_file_request.h"
#include "events/load_annotations_from_file_request.h"
#include "events/save_annotations_from_renderable_ids_to_file_request.h"
#include "events/create_annotations_collection_request.h"
#include "events/annotation_removed_event.h"
#include "events/rendering_cfg_update_event.h"
#include "events/granule_removal_request.h"
#include "events/collection_removed_event.h"
#include "events/shape_edition_complete_event.h"
#include "events/enable_shape_creation_request.h"
#include "events/enable_shape_creation_ready_event.h"
#include "events/annotation_time_coverage_update_request.h"
#include "events/annotation_time_coverage_updated_event.h"

#include "json_annotation.h"
#include "wkt_annotation.h"

using namespace SEAScope;

/**
 *
 */
AnnotationsController::AnnotationsController(AppEventsDispatcher& appEventsDispatcher, IAnnotationsStore& store, IGranulesIndex& granulesIndex)
: _appEventsDispatcher(appEventsDispatcher),
  _userGenerated(SEAScope::Source::ANNOTATIONS, appEventsDispatcher, granulesIndex),
  _store(store)
{
    this->_defaultColvars[ShapeType::POINT] = {{this->_userGenerated.getSourceId(), 1u}, 0u};
    this->_defaultColvars[ShapeType::POLYLINE] = {{this->_userGenerated.getSourceId(), 2u}, 0u};
    this->_defaultColvars[ShapeType::POLYGON] = {{this->_userGenerated.getSourceId(), 3u}, 0u};

    this->_appEventsDispatcher.subscribe<AddPreparedAnnotationRequest>(this);
    this->_appEventsDispatcher.subscribe<SaveOneAnnotationToFileRequest>(this);
    this->_appEventsDispatcher.subscribe<LoadAnnotationsFromFileRequest>(this);
    this->_appEventsDispatcher.subscribe<SaveAnnotationsFromRenderableIdsToFileRequest>(this);
    this->_appEventsDispatcher.subscribe<AnnotationRemovedEvent>(this);
    this->_appEventsDispatcher.subscribe<RenderingCfgUpdateEvent>(this);
    this->_appEventsDispatcher.subscribe<GranuleRemovalRequest>(this);
    this->_appEventsDispatcher.subscribe<CreateAnnotationsCollectionRequest>(this);
    this->_appEventsDispatcher.subscribe<CollectionRemovedEvent>(this);
    this->_appEventsDispatcher.subscribe<ShapeEditionCompleteEvent>(this);
    this->_appEventsDispatcher.subscribe<EnableShapeCreationRequest>(this);
    this->_appEventsDispatcher.subscribe<AnnotationTimeCoverageUpdateRequest>(this);
}

/**
 *
 */
AnnotationsController::~AnnotationsController()
{
    this->_appEventsDispatcher.unsubscribe<AddPreparedAnnotationRequest>(this);
    this->_appEventsDispatcher.unsubscribe<SaveOneAnnotationToFileRequest>(this);
    this->_appEventsDispatcher.unsubscribe<LoadAnnotationsFromFileRequest>(this);
    this->_appEventsDispatcher.unsubscribe<SaveAnnotationsFromRenderableIdsToFileRequest>(this);
    this->_appEventsDispatcher.unsubscribe<AnnotationRemovedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<RenderingCfgUpdateEvent>(this);
    this->_appEventsDispatcher.unsubscribe<GranuleRemovalRequest>(this);
    this->_appEventsDispatcher.unsubscribe<CreateAnnotationsCollectionRequest>(this);
    this->_appEventsDispatcher.unsubscribe<CollectionRemovedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<ShapeEditionCompleteEvent>(this);
    this->_appEventsDispatcher.unsubscribe<EnableShapeCreationRequest>(this);
    this->_appEventsDispatcher.unsubscribe<AnnotationTimeCoverageUpdateRequest>(this);
}

/**
 *
 */
UserGenerated* AnnotationsController::getUserGenerated()
{
    return &(this->_userGenerated);
}

/**
 *
 */
const IAnnotationsStore* AnnotationsController::getStore() const
{
    return &(this->_store);
}

/**
 *
 */
bool AnnotationsController::handleDefaultCollectionVariable(const CollectionVariable& inColvar, CollectionVariable& outColvar, ShapeType shapeType, bool createCollectionIfNeeded)
{
    if
        ((0u == inColvar.second) && (0u == inColvar.first.first) && (0u == inColvar.first.second))
    {
        auto found = this->_defaultColvars.find(shapeType);
        if
            (this->_defaultColvars.end() == found)
        {
            LOG(LogLevel::Error, "AnnotationsController",
                     "Requesting default CollectionVariable for unsupported shape type");
            return false;
        }
        outColvar = found->second;

        // Create collection and variable if they do not exist yet
        LOG(LogLevel::Debug, "AnnotationsController", "Using default colvar: {} {} {}",
                 outColvar.first.first, outColvar.first.second, outColvar.second);

        if
            (true == createCollectionIfNeeded)
        {
            CollectionCfg colCfg;
            bool ok(false);
            ok = this->_store.getCollection(outColvar.first.second, colCfg);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "AnnotationsController",
                         "Error while fetching coonfiguration for collection {}.{}",
                         outColvar.first.first, outColvar.first.second);
                return false;
            }

            if
                ("" == colCfg.label) // means that no matching collection has been found
            {
                std::string shapeLabel("");
                if
                    (ShapeType::POINT == shapeType)
                {
                    shapeLabel = "points";
                }
                else if
                    (ShapeType::POLYLINE == shapeType)
                {
                    shapeLabel = "polylines";
                }
                else if
                    (ShapeType::POLYGON == shapeType)
                {
                    shapeLabel = "polygons";
                }
                else
                {
                    LOG(LogLevel::Error, "AnnotationsController",
                             "Cannot find label for default collection with shape type: {}", int(shapeType));
                    return false;
                }
                LOG(LogLevel::Debug, "AnnotationsController", "Creating missing default collection");
                this->createAnnotationsCollection(shapeType, "Annotations", shapeLabel, true, outColvar.first);
            }
        }
    }
    else
    {
        outColvar = inColvar;
    }
    return true;
}

/**
 *
 */
bool AnnotationsController::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(AddPreparedAnnotationRequest)) == eventType)
    {
        const CollectionVariable& inColvar = static_cast<const AddPreparedAnnotationRequest&>(event).getCollectionVariable();
        OGL::Shape* shape = static_cast<const AddPreparedAnnotationRequest&>(event).getShape();
        double defaultStart = static_cast<const AddPreparedAnnotationRequest&>(event).getDefaultStart();
        double defaultStop = static_cast<const AddPreparedAnnotationRequest&>(event).getDefaultStop();

        CollectionVariable colvar;
        bool ok(false);
        ok = this->handleDefaultCollectionVariable(inColvar, colvar, shape->getType(), true);
        if
            (false == ok)
        {
            return ok;
        }

        RenderableId rId;
        this->_userGenerated.generateRenderableId(colvar.first, rId);
        rId.variable = colvar.second;

        this->addAnnotation(rId, shape, defaultStart, defaultStop);

        // Trigger update
        CollectionsUpdateRequestedEvent outEvent({colvar.first}, {}, {colvar.first}/*force collection coverage update*/);
        this->_appEventsDispatcher.dispatch<CollectionsUpdateRequestedEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(SaveOneAnnotationToFileRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const SaveOneAnnotationToFileRequest&>(event).getRenderableId();
        const std::string& path = static_cast<const SaveOneAnnotationToFileRequest&>(event).getPath();
        const AnnotationFormat& exportFormat = static_cast<const SaveOneAnnotationToFileRequest&>(event).getExportFormat();
        const DictNode* volatileConstProperties = static_cast<const SaveOneAnnotationToFileRequest&>(event).getVolatileProperties();

        // WARNING: removing constness here is ok only because this instance is
        // the only event handler that will process the event. Make sure to
        // find another solution if the event has to be processed by several
        // handlers in the future
        DictNode* volatileProperties = const_cast<DictNode*>(volatileConstProperties);

        this->saveAnnotation(rId, path, exportFormat, volatileProperties);
        return true;
    }
    else if
        (std::type_index(typeid(LoadAnnotationsFromFileRequest)) == eventType)
    {
        const std::string& path = static_cast<const LoadAnnotationsFromFileRequest&>(event).getPath();
        const std::map<ShapeType, CollectionId>& collectionsMapping = static_cast<const LoadAnnotationsFromFileRequest&>(event).getCollectionsMapping();
        const AnnotationFormat& importFormat = static_cast<const LoadAnnotationsFromFileRequest&>(event).getImportFormat();
        double defaultStart = static_cast<const LoadAnnotationsFromFileRequest&>(event).getDefaultStart();
        double defaultStop = static_cast<const LoadAnnotationsFromFileRequest&>(event).getDefaultStop();
        const std::set<CollectionId>& selectedCollections = static_cast<const LoadAnnotationsFromFileRequest&>(event).getSelectedCollections();

        std::map<ShapeType, CollectionId> fixedCollectionsMapping;
        for
            (const auto& it: collectionsMapping)
        {
            CollectionVariable colvar;
            CollectionVariable inColvar({it.second, 0u});
            bool ok(false);
            ok = this->handleDefaultCollectionVariable(inColvar, colvar, it.first, true);
            if
                (false == ok)
            {
                return ok;
            }
            fixedCollectionsMapping[it.first] = colvar.first;
        }

        this->loadFromFile(path, fixedCollectionsMapping, importFormat, defaultStart, defaultStop, selectedCollections);
        return true;
    }
    else if
        (std::type_index(typeid(SaveAnnotationsFromRenderableIdsToFileRequest)) == eventType)
    {
        const std::string& path = static_cast<const SaveAnnotationsFromRenderableIdsToFileRequest&>(event).getPath();
        const std::set<RenderableId>& rIds = static_cast<const SaveAnnotationsFromRenderableIdsToFileRequest&>(event).getRenderableIds();
        const std::map<ShapeType, CollectionId>& collectionsMapping = static_cast<const SaveAnnotationsFromRenderableIdsToFileRequest&>(event).getCollectionsMapping();
        const AnnotationFormat& exportFormat = static_cast<const SaveAnnotationsFromRenderableIdsToFileRequest&>(event).getExportFormat();

        this->saveToFile(path, rIds, collectionsMapping, exportFormat);
        return true;
    }
    else if
        (std::type_index(typeid(CreateAnnotationsCollectionRequest)) == eventType)
    {
        const ShapeType& shapeType = static_cast<const CreateAnnotationsCollectionRequest&>(event).getAnnotationsType();
        const std::string& colLabel = static_cast<const CreateAnnotationsCollectionRequest&>(event).getCollectionLabel();
        const std::string& varLabel = static_cast<const CreateAnnotationsCollectionRequest&>(event).getVariableLabel();
        bool persist = static_cast<const CreateAnnotationsCollectionRequest&>(event).getPersist();

        this->createAnnotationsCollection(shapeType, colLabel, varLabel, persist);
    }
    else if
        (std::type_index(typeid(AnnotationRemovedEvent)) == eventType)
    {
		std::ostringstream buffer("");	//

        const RenderableId& rId = static_cast<const AnnotationRemovedEvent&>(event).getRenderableId();
		buffer << "usershape_" << rId.collection.first << "/" << rId.collection.second << "/" << rId.granule.second << ":" <<rId.variable;

        if
            (this->_userGenerated.getIndex().hasGranule(buffer.str()))
        {
            this->_userGenerated.getIndex().remove(buffer.str(), false);
            CollectionsUpdateRequestedEvent outEvent({rId.collection}, {}, {rId.collection} /*force collection coverage update*/);
            this->_appEventsDispatcher.dispatch<CollectionsUpdateRequestedEvent>(outEvent);
        }
        return true;
    }
    else if
        (std::type_index(typeid(RenderingCfgUpdateEvent)) == eventType)
    {
        const RenderingCfg* rCfg = static_cast<const RenderingCfgUpdateEvent&>(event).getRenderingCfg();

        // Update rendering in the annotations store only if the collection belongs to the annotations source
        if
            (this->_userGenerated.getSourceId() == rCfg->target.collection.first)
        {
            this->_store.setRendering(*rCfg);
        }
        return true;
    }
    else if
        (std::type_index(typeid(GranuleRemovalRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const GranuleRemovalRequest&>(event).getRenderableId();
        if
            (this->_userGenerated.getSourceId() != rId.collection.first)
        {
            // Cannot remove since granule does not belong to user-generated content
            LOG(LogLevel::Debug, "Annotations controller",
                     "will not remove annotation: wrong source ({}.{}/{})",
                     rId.collection.first, rId.collection.second, rId.granule.second);
            return true;
        }

        this->_store.remove(rId);  // triggers an AnnotationRemovedEvent, which is handled above and results in the removal of the granule from the index
        return true;
    }
    else if
        (std::type_index(typeid(CollectionRemovedEvent)) == eventType)
    {
        const CollectionId& collectionId = static_cast<const CollectionRemovedEvent&>(event).getCollectionId();
        if
            (this->_userGenerated.getSourceId() != collectionId.first)
        {
            // Cannot remove since the collection does not belong to user-generated content
            LOG(LogLevel::Debug, "Annotations controller",
                     "will not remove collection: wrong source ({}.{})",
                     collectionId.first, collectionId.second);
            return true;
        }

        std::vector<std::uint64_t> granuleIds;
        this->_store.listCollectionGranules(collectionId.second, granuleIds);

        this->_store.removeCollection(collectionId.second);

        this->_userGenerated.getIndex().removeCollection(collectionId.second);

        for
            (const auto& id : granuleIds)
        {
            RenderableId rId(true, collectionId, {collectionId.first, id}, 0);
            AnnotationRemovedEvent outEvent(rId);
            this->_appEventsDispatcher.dispatch<AnnotationRemovedEvent>(outEvent);
        }

        return true;
    }
    else if
        (std::type_index(typeid(ShapeEditionCompleteEvent)) == eventType)
    {
        OGL::Shape* shape = static_cast<const ShapeEditionCompleteEvent&>(event).getShape();
        const RenderableId& rId = static_cast<const ShapeEditionCompleteEvent&>(event).getRenderableId();

        Annotation annotation;
        std::uint32_t collection;
        bool success(false);
        success = this->_store.get(rId.granule.second, collection, annotation);
        if
            (false == success)
        {
            LOG(LogLevel::Error, "AnnotationsController", "Edited user shape not found in the annotations database");
            return true;
        }

        annotation.setShape(shape);

        // update database content
        success = this->_store.set(rId.granule.second, rId.collection.second, annotation);
        if
            (false == success)
        {
            LOG(LogLevel::Error, "AnnotationsController", "Could not save annotation");
            return true;
        }
        return true;
    }
    else if
        (std::type_index(typeid(EnableShapeCreationRequest)) == eventType)
    {
        const ShapeType& shapeType =  static_cast<const EnableShapeCreationRequest&>(event).getShapeType();
        const CollectionVariable& collectionVariable = static_cast<const EnableShapeCreationRequest&>(event).getCollectionVariable();

        CollectionVariable fixedCollectionVariable;
        bool ok(false);
        ok = this->handleDefaultCollectionVariable(collectionVariable, fixedCollectionVariable, shapeType, true);
        if
            (false == ok)
        {
            return false;
        }

        EnableShapeCreationReadyEvent outEvent(shapeType, fixedCollectionVariable);
        this->_appEventsDispatcher.dispatch<EnableShapeCreationReadyEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(AnnotationTimeCoverageUpdateRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const AnnotationTimeCoverageUpdateRequest&>(event).getRenderableId();
        double start = static_cast<const AnnotationTimeCoverageUpdateRequest&>(event).getStart();
        double stop = static_cast<const AnnotationTimeCoverageUpdateRequest&>(event).getStop();

        if
            (this->_userGenerated.getSourceId() != rId.collection.first)
        {
            return true;
        }

        Annotation annotation;
        std::uint32_t collection;
        bool success(false);
        success = this->_store.get(rId.granule.second, collection, annotation);
        if
            (false == success)
        {
            LOG(LogLevel::Error, "AnnotationsController", "Edited user shape not found in the annotations database");
            return true;
        }

        annotation.setStart(start);
        annotation.setStop(stop);

        // update database content
        success = this->_store.set(rId.granule.second, rId.collection.second, annotation);
        if
            (false == success)
        {
            LOG(LogLevel::Error, "AnnotationsController", "Could not save annotation");
            return true;
        }

        bool ok(false);
        ok = this->_userGenerated.getIndex().updateTimeCoverage(rId.granule.second,
                                                           static_cast<timestamp_t>(start),
                                                           static_cast<timestamp_t>(stop));
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "AnnotationsController", "Could not update annotation time coverage in the index");
            return true;
        }

        AnnotationTimeCoverageUpdatedEvent outEvent(rId, start, stop);
        this->_appEventsDispatcher.dispatch<AnnotationTimeCoverageUpdatedEvent>(outEvent);

        return true;
    }

    return false;
}


/**
 *
 */
bool AnnotationsController::loadFromFile(const std::string& inputPath, const std::map<ShapeType, CollectionId>& collectionsMapping, const AnnotationFormat& exportFormat, double defaultStart, double defaultStop, const std::set<CollectionId>& selectedCollections)
{
    bool success(false);    //
    std::vector<Annotation> annotationsList;

    if
        (AnnotationFormat::JSON == exportFormat)
    {
        //> JSON annotations
        std::ifstream jsonFile(inputPath);
        if
            (false == jsonFile.is_open())
        {
            LOG(LogLevel::Error, "AnnotationsController", "JSON input file is closed");
            return false;
        }

        JSONAnnotation jsonAnnotation;
        success = jsonAnnotation.deserialize(jsonFile, annotationsList, defaultStart, defaultStop);
        jsonFile.close();
    }
    else if
        (AnnotationFormat::WKT == exportFormat)
    {
        //> WKT annotations
        std::ifstream wktFile(inputPath);
        if
            (false == wktFile.is_open())
        {
            LOG(LogLevel::Error, "AnnotationsController", "WKT input file is closed");
            return false;
        }

        WKTAnnotation wktAnnotation;
        success = wktAnnotation.deserialize(wktFile, annotationsList, defaultStart, defaultStop);
        wktFile.close();
    }
    else
    {
        LOG(LogLevel::Error, "AnnotationsController", "");
        return false;
    }

    if
        (false == success)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Failed to deserialize annotations from JSON file");
        return false;
    }
    //< JSON annotations

    //
    std::vector<Annotation> annotations;
    std::vector<std::uint64_t> granuleIds;
    std::vector<std::uint32_t> collectionIds;

    annotations.reserve(annotationsList.size());
    granuleIds.reserve(annotationsList.size());
    collectionIds.reserve(annotationsList.size());
    for
        (auto& annotation : annotationsList)
    {
        OGL::Shape* shape = annotation.getShape();
        auto collectionFound = collectionsMapping.find(shape->getType());
        shape = nullptr;

        if
            (collectionsMapping.end() == collectionFound)
        {
            LOG(LogLevel::Warn, "Annotations controller",
                     "Annotation skipped due to unsupported shape type");
            continue;
        }

        // Register user shape in the index first otherwise it will not be
        // possible to set the temporal coverage
        RenderableId rId;
        success = this->_userGenerated.generateRenderableId(collectionFound->second, rId);

        if
            (false == success)
        {
            LOG(LogLevel::Error, "AnnotationsController", "Failed to acquire an identifier for the user shape");
            LOG(LogLevel::Error, "AnnotationsController", "Data import aborted");
            return false;
        }

        annotations.emplace_back(std::move(annotation));
        granuleIds.push_back(rId.granule.second);
        collectionIds.push_back(rId.collection.second);
    }
    annotationsList.clear();

    success = this->_store.set(granuleIds, collectionIds, annotations);
    if
        (false == success)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Could not save annotations in the local database");
        LOG(LogLevel::Error, "AnnotationsController", "Data import aborted");
        return false;
    }

    auto granuleIt = granuleIds.cbegin();
    auto collectionIt = collectionIds.cbegin();
    auto annotationIt = annotations.cbegin();
    this->_userGenerated.getIndex().startBulkOperation();
    for
        (; granuleIt != granuleIds.end() && collectionIt != collectionIds.end() && annotationIt != annotations.end(); ++granuleIt, ++collectionIt, ++annotationIt)
    {
        this->initAnnotation(*granuleIt, *collectionIt, *annotationIt);
        this->_userGenerated.getIndex().increaseBulkOpsCount();
    }
    this->_userGenerated.getIndex().stopBulkOperation();

    std::set<CollectionId> userCols;
    for
        (const auto& it : collectionsMapping)
    {
        auto collectionFound = std::find(selectedCollections.begin(),
                                  selectedCollections.end(),
                                  it.second);
        if
            (selectedCollections.end() != collectionFound)
        {
            userCols.emplace(*collectionFound);
        }
    }

    // Trigger update
    CollectionsUpdateRequestedEvent event(userCols, {}, userCols /*force collections coverage update*/);
    this->_appEventsDispatcher.dispatch<CollectionsUpdateRequestedEvent>(event);

    return true;
}

/**
 *
 */
bool AnnotationsController::getAnnotation(const RenderableId& rId, Annotation& result) const
{
    std::uint32_t collection {0u};
    return this->_store.get(rId.granule.second, collection, result);
}

/**
 *
 */
bool AnnotationsController::saveToFile(const std::string& outputPath, const std::set<RenderableId>& renderables, const std::map<ShapeType, CollectionId>& collectionsMapping, const AnnotationFormat& exportFormat)
{
    std::map<ShapeType, CollectionId> fixedCollectionsMapping;
    for
        (const auto& it: collectionsMapping)
    {
        CollectionVariable colvar;
        CollectionVariable inColvar({it.second, 0u});
        bool ok(false);
        ok = this->handleDefaultCollectionVariable(inColvar, colvar, it.first, false);
        if
            (false == ok)
        {
            return ok;
        }
        fixedCollectionsMapping[it.first] = colvar.first;
    }

    bool success(false);    //
    std::uint32_t collection(0);    //
    std::vector<Annotation> annotationsList;    //
    LOG(LogLevel::Debug, "AnnotationsController", "Save user shapes for Syntool");

    // export all rendered shapes
    for
        (const auto& rId : renderables)
    {
        bool isRequested(false);
        for
            (const auto& it : fixedCollectionsMapping)
        {
            if
                (  (rId.collection.first == it.second.first)
                && (rId.collection.second == it.second.second))
            {
                isRequested = true;
                break;
            }
        }
        if
            (false == isRequested)
        {
            continue;
        }

        Annotation annotation;

        // Export granule
        success = this->_store.get(rId.granule.second, collection, annotation);
        if
            (false == success)
        {
            LOG(LogLevel::Error, "AnnotationsController", "Failed to get annotation {} from local database", rId.granule.second);
            return false;
        }

        annotationsList.push_back(std::move(annotation));
    }

    if
        (0 >= annotationsList.size())
    {
        LOG(LogLevel::Debug, "AnnotationsController", "No annotation to export");
        return false;
    }

    if
        (AnnotationFormat::JSON == exportFormat)
    {
        std::ofstream jsonFile(outputPath);
        if
            (false == jsonFile.is_open())
        {
            LOG(LogLevel::Error, "AnnotationsController", "JSON output file is closed");
        }

        JSONAnnotation jsonAnnotation;
        jsonAnnotation.serialize(annotationsList, jsonFile);
    }
    else if
        (AnnotationFormat::WKT == exportFormat)
    {
        std::ofstream wktFile(outputPath);
        if
            (false == wktFile.is_open())
        {
            LOG(LogLevel::Error, "AnnotationsController", "WKT output file is closed");
            return false;
        }

        WKTAnnotation wktAnnotation;  //
        wktAnnotation.serialize(annotationsList, wktFile);
    }
    else
    {
        LOG(LogLevel::Error, "AnnotationsController",
                 "Could not export the annotation because the requested export format is not supported");
        return false;
    }
    LOG(LogLevel::Info, "AnnotationsController", "Annotations saved to {}", outputPath);

    return true;
}


/**
 *
 */
bool AnnotationsController::addAnnotation(const RenderableId& rId, OGL::Shape* shape, double defaultStart, double defaultStop)
{
    if
        (nullptr == shape)
    {
        LOG(LogLevel::Error, "AnnotationsController", "A null shape has been passed to the user shape added callback");
        return false;
    }

    // Try to load info from the database
    Annotation annotation;
    std::uint32_t collection;
    bool success(false);
    success = this->_store.get(rId.granule.second, collection, annotation);
    if
        (false == success)
    {
        // Default time coverage
        annotation.setStart(defaultStart);
        annotation.setStop(defaultStop);
    }

    annotation.setShape(shape);

    // update database content
    success = this->_store.set(rId.granule.second, rId.collection.second, annotation);
    if
        (false == success)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Could not save annotation");
        return false;
    }

    this->initAnnotation(rId.granule.second, rId.collection.second, annotation);

    return true;
}

/**
 *
 */
bool AnnotationsController::updateAnnotation(const RenderableId& rId, OGL::Shape* shape)
{
    LOG(LogLevel::Debug, "AnnotationsController", "User shape {}/{}/{}:{} updated", rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
    if
        (nullptr == shape)
    {
        LOG(LogLevel::Error, "AnnotaitonsController", "A null shape has been passed to the user shape updated callback");
        return false;
    }

    Annotation annotation;
    std::uint32_t collection;
    bool success(false);
    success = this->_store.get(rId.granule.second, collection, annotation);
    if
        (false == success)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Edited user shape not found in the annotations database");
        return false;
    }

    annotation.setShape(shape);

    // update database content
    success = this->_store.set(rId.granule.second, rId.collection.second, annotation);
    if
        (false == success)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Could not save annotation");
        return false;
    }

    return true;
}

/**
 *
 */
bool AnnotationsController::saveAnnotation(const RenderableId& rId, const std::string& outputPath, const AnnotationFormat& exportFormat, DictNode* volatileProperties)
{
    bool success(false); //
    std::uint32_t collection(0);    //
    Annotation annotation;  //

    // Export granule
    success = this->_store.get(rId.granule.second, collection, annotation);
    if
        (false == success)
    {
        LOG(LogLevel::Error, "AnnotationsController",
                 "Failed to get annotation {} from local database", rId.granule.second);
        return false;
    }

    if
        (nullptr != volatileProperties)
    {
        DictNode& properties = annotation.getProperties();
        for
            (const auto& property : volatileProperties->children)
        {

            // If a property with the same name already exists, delete it and
            // don't forget to free the child node
            auto childFound = properties.children.find(property.first);
            if
                (properties.children.end() != childFound)
            {
                delete childFound->second;
                properties.children.erase(childFound);
            }

            properties.children[property.first] = property.second;
        }

        // A double-free will happen if both "properties" and
        // "volatileProperties" contain pointers to the same children nodes
        volatileProperties->children.clear();
    }

    if
        (AnnotationFormat::JSON == exportFormat)
    {
        std::vector<Annotation> annotationsList;    //
        annotationsList.push_back(std::move(annotation));

        std::ofstream jsonFile(outputPath);
        if
            (false == jsonFile.is_open())
        {
            LOG(LogLevel::Error, "AnnotationsController", "JSON output file is closed");
            return false;
        }

        JSONAnnotation jsonAnnotation;  //
        jsonAnnotation.serialize(annotationsList, jsonFile);
    }
    else if
        (AnnotationFormat::WKT == exportFormat)
    {
        std::ofstream wktFile(outputPath);
        if
            (false == wktFile.is_open())
        {
            LOG(LogLevel::Error, "AnnotationsController", "WKT output file is closed");
            return false;
        }

        WKTAnnotation wktAnnotation;  //
        wktAnnotation.serialize(annotation, wktFile);
    }
    else
    {
        LOG(LogLevel::Error, "AnnotationsController",
                 "Could not export the annotation because the requested export format is not supported");
        return false;
    }
    LOG(LogLevel::Info, "AnnotationsController", "Annotation saved to {}", outputPath);

    return true;
}

/**
 *
 */
bool AnnotationsController::applyMigrations()
{
    bool ok {false};    //
    bool annotationsCountMismatch {false};   //
    bool collectionsCountMismatch {false};   //

    // Check that the number of indexed collections matches the number of
    // stored collections
    std::size_t storedCollectionsCount {0u}; //
    std::size_t indexedCollectionsCount {0u};   //
    ok = this->_store.countCollections(storedCollectionsCount);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Could not count collections available in the store");
        return false;
    }
    ok = this->_userGenerated.getIndex().countCollections(indexedCollectionsCount);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Could not count collections registered in the index");
        return false;
    }
    collectionsCountMismatch = (storedCollectionsCount != indexedCollectionsCount);

    // Check that the number of indexed granules matches the number of
    // stored annotations
    std::size_t storedAnnotationsCount {0u}; //
    std::size_t indexedGranulesCount {0u};   //
    ok = this->_store.count(storedAnnotationsCount);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Could not count annotations available in the store");
        return false;
    }
    ok = this->_userGenerated.getIndex().countGranules(indexedGranulesCount);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Could not count granules registered in the index");
        return false;
    }
    annotationsCountMismatch = (storedAnnotationsCount != indexedGranulesCount);

    // Note: this is not a thorough check because the counts can match but the
    // content may actually differ.
    if
        (true == (collectionsCountMismatch || annotationsCountMismatch))
    {
        this->_userGenerated.getIndex().clear();

        // Re-index collections
        this->_store.applyAllCollections([this] (std::uint32_t collectionId, CollectionCfg& /*collectionCfg*/, std::vector<RenderingCfg>& /*rCfgs*/) {
            /* The index requires a unique URI for the collection*/
            std::ostringstream buffer;  //
            buffer << "#__DUMMY_PATH__ " << std::to_string(collectionId);

            this->_userGenerated.getIndex().addCollection(buffer.str(), collectionId);
        });

        // Re-index annotations
        this->_store.applyall([this] (std::uint64_t granuleId, std::uint32_t collectionId, Annotation& annotation) {
            this->initAnnotation(granuleId, collectionId, annotation);
        });
    }

    return true;
}

/**
 *
 */
bool AnnotationsController::init(const std::string& path)
{
    bool ok(false); //

    ok = this->_store.init(path);
    if
        (false == ok)
    {
        return false;
    }

    this->_userGenerated.getIndex().load(path);

    // Store and index have been initialized, now check if their content match
    // and apply the required changes so that the content is coherent
    ok = this->applyMigrations();
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "AnnotationsController", "Failed to apply migrations");
        return false;
    }

    // Make sure the next collection ID will be at least 4 because the three
    // first values are reserved for default collections:
    // 1 - point
    // 2 - polyline
    // 3 - polygon
    std::uint32_t nextCollectionId {0u};    //
    nextCollectionId = this->_userGenerated.getIndex().generateCollectionId();
    if
        (4u > nextCollectionId)
    {
        nextCollectionId = 4u;
    }
    this->_userGenerated.getIndex().setCollectionIdGenerator(nextCollectionId);

    this->_store.applyAllCollections([this] (std::uint32_t collectionId, CollectionCfg& collectionCfg, std::vector<RenderingCfg>& rCfgs) {
        CollectionId cId({this->_userGenerated.getSourceId(), collectionId});

        this->_renderingCfgs[cId];
        this->_renderingCfgs[cId].target.granuleLevel = rCfgs[0].target.granuleLevel;
        this->_renderingCfgs[cId].target.collection = cId;
        this->_renderingCfgs[cId].target.variable = rCfgs[0].target.variable;
        rCfgs[0].applyTo(&(this->_renderingCfgs[cId]));

        std::vector<RenderingCfg*> renderings;	//
        renderings.push_back(&(this->_renderingCfgs[cId]));

        CollectionRegistrationRequest event(cId, collectionCfg, renderings);
        this->_appEventsDispatcher.dispatch<CollectionRegistrationRequest>(event);
    });

    std::uint64_t granuleIdSeed = 1;
    std::uint64_t maxGranuleId {0u};
    ok = this->_store.getMaxGranuleId(maxGranuleId);
    if
        (true == ok)
    {
        granuleIdSeed = maxGranuleId + 1;
    }

    // Some user shapes have been loaded, update the ID generator
    this->_userGenerated.getIndex().setGranuleIdGenerator(granuleIdSeed);

	return true;
}


/**
 *
 */
void AnnotationsController::createAnnotationsCollection(const ShapeType& shapeType, const std::string& collectionLabel, const std::string& variableLabel, bool persist, const CollectionId& cId)
{
	VariableCfg varCfg;	//
	varCfg.label = variableLabel;

	CollectionCfg colCfg;   //
	colCfg.label = collectionLabel;
	colCfg.xSeamless = false;
	colCfg.ySeamless = false;
	colCfg.mustBeCurrent = false;

    this->_renderingCfgs[cId];
	this->_renderingCfgs[cId].target.granuleLevel = false;
	this->_renderingCfgs[cId].target.collection = cId;
	this->_renderingCfgs[cId].target.variable = 0;

    if
        (ShapeType::POLYGON == shapeType)
    {
        this->_renderingCfgs[cId].setRenderMethod(RenderMethod::POLYGON);
        this->_renderingCfgs[cId].setOpacity(0.7f);
        this->_renderingCfgs[cId].setColor({{200,200,75}});
        this->_renderingCfgs[cId].setZindex(5.0);
    }
    else if
        (ShapeType::POLYLINE == shapeType)
    {
        this->_renderingCfgs[cId].setRenderMethod(RenderMethod::POLYLINE);
        this->_renderingCfgs[cId].setOpacity(1.0);
        this->_renderingCfgs[cId].setColor({{30,250,75}});
        this->_renderingCfgs[cId].setZindex(6.0);
    }
    else if
        (ShapeType::ELLIPSE == shapeType)
    {
        this->_renderingCfgs[cId].setRenderMethod(RenderMethod::ELLIPSE);
        this->_renderingCfgs[cId].setOpacity(1.0);
        this->_renderingCfgs[cId].setColor({{255, 95, 200}});
        this->_renderingCfgs[cId].setZindex(6.0);
    }
    else if
        (ShapeType::POINT == shapeType)
    {
        this->_renderingCfgs[cId].setRenderMethod(RenderMethod::POINT);
        this->_renderingCfgs[cId].setOpacity(1.0);
        this->_renderingCfgs[cId].setColor({{255, 0, 0}});
        this->_renderingCfgs[cId].setZindex(6.0);
    }

	colCfg.variables[0] = varCfg;

	std::vector<RenderingCfg*> renderings;	//
	renderings.push_back(&(this->_renderingCfgs[cId]));

    if
        (true == persist)
    {
        this->_store.setCollection(cId.second, colCfg, renderings);

        /* The index requires a unique URI for the collection*/
        std::ostringstream buffer;  //
        buffer << "#__DUMMY_PATH__ " << std::to_string(cId.second);
        this->_userGenerated.getIndex().addCollection(buffer.str(), cId.second);
    }

    CollectionRegistrationRequest event(cId, colCfg, renderings);
    this->_appEventsDispatcher.dispatch<CollectionRegistrationRequest>(event);
}

/**
 *
 */
void AnnotationsController::createAnnotationsCollection(const ShapeType& shapeType, const std::string& collectionLabel, const std::string& variableLabel, bool persist)
{
    // Get a new collection ID
    CollectionId cId({this->_userGenerated.getSourceId(),
                      this->_userGenerated.getIndex().generateCollectionId()});

    this->createAnnotationsCollection(shapeType, collectionLabel, variableLabel, persist, cId);
}


/**
 *
 */
void AnnotationsController::initAnnotation(std::uint64_t granule, std::uint32_t collection, const Annotation& annotation)
{
    GranuleMetadata metadata;   //
    std::ostringstream buffer;  //
    IDFDescriptor idf;  //
    std::uint32_t sourceId; //

    sourceId = this->_userGenerated.getSourceId();
    RenderableId rId(true, CollectionId(sourceId, collection), GranuleId(sourceId, granule), 0);

    buffer << "usershape_" << sourceId << "/" << collection << "/" << granule << ":" << 0;
    idf.uri = buffer.str();
    idf.resolution = 10000000;
    metadata.build(sourceId,
                   collection,
                   granule,
                   DataModel::NONE,
                   static_cast<timestamp_t>(annotation.getStart()),
                   static_cast<timestamp_t>(annotation.getStop()),
                   false,
                   0.f,
                   false,
                   0.f,
                   false,
                   0.f,
                   false,
                   0.f,
                   buffer.str(), /* dataId */
                   "", /* title */
                   "", /* institution */
                   "", /* comment */
                   "", /* fileId */
                   "", /* productVersion */
                   "", /* creatorEmail */
                   "", /* stationId */
                   "", /* platform */
                   "", /* sensor */
                   idf);

    this->_userGenerated.getIndex().add(metadata, true);
}
