// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "tools/contextual_menu.h"

#include "vendors/_nanogui.h"
#include "ui/nanogui/nanogui_contextual_menu.h"
#include "3d/pickable.h"
#include "core/app_events_dispatcher.h"
#include "core/interfaces/iwindow.h"

#include "ui/nanogui/nanogui_annotation_dialog.h"
#include "ui/nanogui/nanogui_granule_dialog.h"

#include "events/set_selected_pickables_request.h"
#include "events/selected_pickables_changed_event.h"
#include "events/selected_pickable_with_info_changed_event.h"

#include "events/center_time_on_granule_request.h"
#include "events/display_granule_info_request.h"
#include "events/extract_from_renderable_id_request.h"

#include "events/granule_safe_removal_request.h"
#include "events/enable_shape_edition_request.h"
#include "events/disable_shape_edition_request.h"
#include "events/enable_shape_creation_request.h"
#include "events/save_one_annotation_to_file_request.h"
#include "events/list_overlayed_granules_before_saving_one_annotation_to_file_request.h"

#include "events/save_annotation_properties_request.h"
#include "events/shape_edition_cancelled_event.h"
#include "events/annotation_time_coverage_update_request.h"
#include "events/annotation_time_coverage_updated_event.h"

#include "core/data_node.h"
#include "core/measurements.h"
#include "events/shape_edition_complete_event.h"

#include <iostream>

using namespace SEAScope;

/**
 *
 */
ContextualMenu::ContextualMenu(AppEventsDispatcher& appEventsDispatcher, nanogui::Widget* parent, IWindow* window)
: _appEventsDispatcher(appEventsDispatcher),
  _window(window)
{
    this->_ui = new NanoguiContextualMenu();
    this->_ui->init(parent);
    this->_ui->hide();
    this->_ui->setCloseCallback([this] () {
        // Clicking on the close button must reset selection
        SetSelectedPickablesRequest event({});
        this->_appEventsDispatcher.dispatch<SetSelectedPickablesRequest>(event);
    });
    this->_ui->setAnnotationSaveCallback([this] (const RenderableId& rId, const std::string& path, const AnnotationFormat& exportFormat, bool listOverlayedGranules) {
        if
            (true == listOverlayedGranules)
        {
            ListOverlayedGranulesBeforeSavingOneAnnotationToFileRequest outEvent(rId, path, exportFormat);
            this->_appEventsDispatcher.dispatch<ListOverlayedGranulesBeforeSavingOneAnnotationToFileRequest>(outEvent);
        }
        else
        {
            SaveOneAnnotationToFileRequest outEvent(rId, path, exportFormat, nullptr);
            this->_appEventsDispatcher.dispatch<SaveOneAnnotationToFileRequest>(outEvent);
        }
    });

    // Granule dialog callbacks
    NanoguiGranuleDialog* granuleDialog = this->_ui->getGranuleDialog();
	granuleDialog->centerTimeCallback = [this](const RenderableId& rId) {
        CenterTimeOnGranuleRequest event(rId);
        this->_appEventsDispatcher.dispatch<CenterTimeOnGranuleRequest>(event);
    };
	granuleDialog->infoCallback = [this](const RenderableId& rId) {
        DisplayGranuleInfoRequest event(rId);
        this->_appEventsDispatcher.dispatch<DisplayGranuleInfoRequest>(event);
    };
    granuleDialog->extractCallback = [this](const RenderableId& rId) {
        ExtractFromRenderableIdRequest event(rId);
        this->_appEventsDispatcher.dispatch<ExtractFromRenderableIdRequest>(event);
    };
    granuleDialog = nullptr;

    // Annotation dialog callbacks
    NanoguiAnnotationDialog* annotationDialog = this->_ui->getAnnotationDialog();
	annotationDialog->extractCallback = [this](const RenderableId& rId) {
        ExtractFromRenderableIdRequest event(rId);
        this->_appEventsDispatcher.dispatch<ExtractFromRenderableIdRequest>(event);
    };
	annotationDialog->deleteCallback = [this](const RenderableId& rId) {
        GranuleSafeRemovalRequest event(rId);
        this->_appEventsDispatcher.dispatch<GranuleSafeRemovalRequest>(event);
    };
	annotationDialog->editCallback = [this](const RenderableId& rId, bool enabled) {
        if
            (true == enabled)
        {
            EnableShapeEditionRequest event(rId);
            this->_appEventsDispatcher.dispatch<EnableShapeEditionRequest>(event);
        }
        else
        {
            DisableShapeEditionRequest event(rId);
            this->_appEventsDispatcher.dispatch<DisableShapeEditionRequest>(event);
        }
    };
    annotationDialog->setSavePropertiesCallback([this] (const RenderableId& rId, const DictNode& properties) {
        SaveAnnotationPropertiesRequest event(rId, properties);
        this->_appEventsDispatcher.dispatch<SaveAnnotationPropertiesRequest>(event);
    });
    annotationDialog->setUpdateTimeCoverageCallback([this] (const RenderableId& rId, double tsStart, double tsStop) {
        AnnotationTimeCoverageUpdateRequest event(rId, tsStart, tsStop);
        this->_appEventsDispatcher.dispatch<AnnotationTimeCoverageUpdateRequest>(event);
    });
    annotationDialog = nullptr;

    this->_appEventsDispatcher.subscribe<SelectedPickablesChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<SelectedPickableWithInfoChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<EnableShapeCreationRequest>(this);
    this->_appEventsDispatcher.subscribe<ShapeEditionCompleteEvent>(this);
    this->_appEventsDispatcher.subscribe<ShapeEditionCancelledEvent>(this);
    this->_appEventsDispatcher.subscribe<AnnotationTimeCoverageUpdatedEvent>(this);
}

/**
 *
 */
ContextualMenu::~ContextualMenu()
{
    this->_appEventsDispatcher.unsubscribe<SelectedPickablesChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<SelectedPickableWithInfoChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<EnableShapeCreationRequest>(this);
    this->_appEventsDispatcher.unsubscribe<ShapeEditionCompleteEvent>(this);
    this->_appEventsDispatcher.unsubscribe<ShapeEditionCancelledEvent>(this);
    this->_appEventsDispatcher.unsubscribe<AnnotationTimeCoverageUpdatedEvent>(this);

    delete this->_ui;
    this->_ui = nullptr;

    this->_window = nullptr;    // not owning
}

/**
 *
 */
bool ContextualMenu::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(SelectedPickablesChangedEvent)) == eventType)
    {
        const std::vector<Pickable*>& selected = static_cast<const SelectedPickablesChangedEvent&>(event).getSelected();
        if
            (  0 >= selected.size()
            || 1 < selected.size())
        {
            // Nothing selected or more than one selected => hide menu
            this->_ui->hide();
        }
    }
    else if
        (std::type_index(typeid(SelectedPickableWithInfoChangedEvent)) == eventType)
    {
        Pickable* pickable = static_cast<const SelectedPickableWithInfoChangedEvent&>(event).getPickable();
        const GranuleInfo& info = static_cast<const SelectedPickableWithInfoChangedEvent&>(event).getInfo();   //
        const RenderableType& rType = static_cast<const SelectedPickableWithInfoChangedEvent&>(event).getRenderableType();
        DictNode& properties = static_cast<SelectedPickableWithInfoChangedEvent&>(const_cast<AppEvent&>(event)).getProperties();
        double start = static_cast<const SelectedPickableWithInfoChangedEvent&>(event).getStart();
        double stop = static_cast<const SelectedPickableWithInfoChangedEvent&>(event).getStop();
        const RenderableId& rId = pickable->renderableId();

        this->_ui->setTarget(rId, rType, info, std::move(properties), start, stop);

        double xCur(0.0);   //
        double yCur(0.0);   //
        std::int32_t x(0);  //
        std::int32_t y(0);  //

        this->_window->getCursorPos(xCur, yCur);
        x = static_cast<std::int32_t>(std::round(xCur));
        y = static_cast<std::int32_t>(std::round(yCur));
        this->_ui->display(Eigen::Vector2i(x,y));

        return true;
    }
    else if
        (std::type_index(typeid(EnableShapeCreationRequest)) == eventType)
    {
        // Make sure edition UI is turned off
        NanoguiAnnotationDialog* annotationDialog = this->_ui->getAnnotationDialog();
        annotationDialog->enableEdition(false /*disabled*/, true /*silenced*/);
        annotationDialog = nullptr;

        return true;
    }
    else if
        (std::type_index(typeid(ShapeEditionCompleteEvent)) == eventType)
    {
        OGL::Shape* shape = static_cast<const ShapeEditionCompleteEvent&>(event).getShape();
        const RenderableId& rId = static_cast<const ShapeEditionCompleteEvent&>(event).getRenderableId();

        NanoguiAnnotationDialog* annotationDialog = this->_ui->getAnnotationDialog();
        if
            (rId == annotationDialog->target)
        {
            double area;
            double perimeter;
            double lon;
            double lat;
            SEAScope::getShapeMeasurements(shape, area, perimeter, lon, lat);
            annotationDialog->updateSpatialCoverage(area, perimeter, lon, lat);
        }

        annotationDialog->enableEdition(false /*leave edition mode */, true /*silent*/);
        annotationDialog->requestFocus();

        annotationDialog = nullptr;
        shape = nullptr;
        return true;
    }
    else if
        (std::type_index(typeid(ShapeEditionCancelledEvent)) == eventType)
    {
        NanoguiAnnotationDialog* annotationDialog = this->_ui->getAnnotationDialog();
        annotationDialog->enableEdition(false /*leave edition mode */, true /*silent*/);
        annotationDialog->requestFocus();

        annotationDialog = nullptr;
        return true;
    }
    else if
        (std::type_index(typeid(AnnotationTimeCoverageUpdatedEvent)) == eventType)
    {
        const RenderableId& rId = static_cast<const AnnotationTimeCoverageUpdatedEvent&>(event).getRenderableId();
        double start = static_cast<const AnnotationTimeCoverageUpdatedEvent&>(event).getStart();
        double stop = static_cast<const AnnotationTimeCoverageUpdatedEvent&>(event).getStop();

        NanoguiAnnotationDialog* annotationDialog = this->_ui->getAnnotationDialog();
        if
            (rId == annotationDialog->target)
        {
            annotationDialog->updateTemporalCoverage(start, stop);
        }

        annotationDialog = nullptr;
        return true;
    }

    return false;
}
