// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "tools/debug_controller.h"


#include "core/log.h"
#include "core/input_event.h"
#include "core/input_devices_state.h"
#include "tools/scene_view_controller_base.h"

#include <iostream>

using namespace SEAScope;

/**
 *
 */
DebugController::DebugController()
: BaseInputEventsSubscriber("Debug controller"),
  _sceneViewController(nullptr)
{
    // Configure actions

    // - Center on BZH
    this->defineAction(static_cast<ActionType>(DebugControllerActions::CENTER_ON_ODL),
                       "Center camera on ODL office", "}LEFT_CONTROL }F5");  /* keypress: CTRL + F5*/

    // - Center on North pole
    this->defineAction(static_cast<ActionType>(DebugControllerActions::CENTER_ON_NORTH_POLE),
                       "Center camera on North pole", "}LEFT_CONTROL }F6"); /* keypress: CTRL + F6 */

    // - Center on South pole
    this->defineAction(static_cast<ActionType>(DebugControllerActions::CENTER_ON_SOUTH_POLE),
                       "Center camera on South pole", "}LEFT_CONTROL >F7"); /* keypress: CTRL + F7 */

    // - Center on IDL
    this->defineAction(static_cast<ActionType>(DebugControllerActions::CENTER_ON_IDL),
                       "Center on IDL", "}LEFT_CONTROL >F8"); /* keypress: CTRL + F8 */

    this->_possibleActions = {static_cast<ActionType>(DebugControllerActions::CENTER_ON_ODL),
                              static_cast<ActionType>(DebugControllerActions::CENTER_ON_NORTH_POLE),
                              static_cast<ActionType>(DebugControllerActions::CENTER_ON_SOUTH_POLE),
                              static_cast<ActionType>(DebugControllerActions::CENTER_ON_IDL)};
}

/**
 *
 */
DebugController::~DebugController()
{
    this->_sceneViewController = nullptr; // not owning
}

/**
 *
 */
bool DebugController::isActive() const
{
    return this->_active && (nullptr != this->_sceneViewController);
}

/**
 *
 */
bool DebugController::performAction(const ActionType& actionType,
                                    const InputEvent& /*event*/,
                                    const InputDevicesState& /*devicesState*/)
{
    if
        (static_cast<ActionType>(DebugControllerActions::CENTER_ON_ODL) == actionType)
    {
        LOG(LogLevel::Debug, "DebugController", "Center camera on ODL");
        this->_sceneViewController->lookAt(-4.6403389f, 48.3478738f);
        return true;
    }
    else if
        (static_cast<ActionType>(DebugControllerActions::CENTER_ON_NORTH_POLE) == actionType)
    {
        LOG(LogLevel::Debug, "DebugController", "Center camera on North pole");
        this->_sceneViewController->lookAt(45, 90);
        return true;
    }
    else if
        (static_cast<ActionType>(DebugControllerActions::CENTER_ON_SOUTH_POLE) == actionType)
    {
        LOG(LogLevel::Debug, "DebugController", "Center camera on South pole");
        this->_sceneViewController->lookAt(45, -90);
        return true;
    }
    else if
        (static_cast<ActionType>(DebugControllerActions::CENTER_ON_IDL) == actionType)
    {
        LOG(LogLevel::Debug, "DebugController", "Center camera on IDL");
        this->_sceneViewController->lookAt(180, 0);
        return true;
    }

    return false;
}


/**
 *
 */
void DebugController::setSceneViewController(SceneViewControllerBase* sceneViewController)
{
    // Assume sceneViewController is not null
    this->_sceneViewController = sceneViewController;
}
