// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "tools/drawing_controller.h"

#include "3d/shape_creator.h"
#include "3d/shape_editor.h"
#include "3d/shape_creator_factory.h"
#include "3d/shape_editor_factory.h"
#include "3d/context.h"
#include "3d/pipeline_parameters.h"
#include "3d/shape_mesh.h"
#include "3d/shape.h"
#include "3d/shape_anchor.h"

#include "core/app_events_dispatcher.h"
#include "core/measurements.h"
#include "events/cursor_location_changed_event.h"
#include "events/enable_shape_creation_ready_event.h"
#include "events/disable_shape_creation_request.h"
#include "events/enable_shape_edition_request.h"
#include "events/disable_shape_edition_request.h"
#include "events/drawing_complete_event.h"
#include "events/selection_lock_request.h"
#include "events/variable_selection_request.h"
#include "events/drawing_started_event.h"
#include "events/measurements_updated_event.h"
#include "events/hovered_pickables_changed_event.h"
#include "events/lock_view_request.h"
#include "events/picking_update_request.h"
#include "events/shape_edition_complete_event.h"
#include "events/drawing_actions_updated_event.h"
#include "events/shape_edition_cancelled_event.h"
#include "events/renderables_changed_event.h"
#include "events/toggle_shape_creation_request.h"
#include "events/upcoming_collections_update_event.h"
#include "events/move_variable_to_top_request.h"
#include "events/granule_removal_request.h"
#include "events/granule_safe_removal_request.h"

#include "core/log.h"
#include "3d/iscene.h"
#include "3d/scene_view.h"
#include "3d/scene_graph.h"
#include "3d/iscene_renderer.h"
#include "3d/scene_graph_controller.h"
#include "3d/shape_gpu_creator.h"
#include "3d/scene_graph_node.h"
#include "3d/icons.h"
#include "3d/anchor_pool_node.h"

using namespace SEAScope;

/**
 *
 */
DrawingController::DrawingController(AppEventsDispatcher& appEventsDispatcher, SceneView& sceneView, const Icons& icons)
: BaseInputEventsSubscriber("Drawing controller"),
  _appEventsDispatcher(appEventsDispatcher),
  _anchors(nullptr),
  _editor(nullptr),
  _creator(nullptr),
  _sceneView(sceneView),
  _editedShape(nullptr),
  _hoveredAnchor(nullptr),
  _explicitDrawingModeExit(false)
{
	DataRenderingParam* param = new DataRenderingParam();

	param->color = glm::vec3(1.f, 0.f, 0.f);
	param->zIndex = std::numeric_limits<float>::max();

    this->_anchors = new AnchorPoolNode(appEventsDispatcher);
    this->_anchors->initialize(sceneView.getRenderer()->getContext(), icons);
	this->_anchors->setRenderingParam(param);
	this->_anchors->setRendererType(RendererType::renderer_quadscreen_textured);
	this->_anchors->setFillType(FillType::fill_color);

	sceneView.getScene()->getNodeFactory()->getSceneGraph()->addTo(nullptr, this->_anchors);

    this->defineAction(static_cast<ActionType>(DrawingControllerActions::CANCEL_AND_LEAVE_DRAWING),
                       "leave drawing mode", ">ESCAPE");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING),
                       "commit drawing", ">ENTER");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING_ALT),
                       "commit drawing", ">KEYPAD_ENTER");

    // Polyline / polygon actions
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::ADD_VERTEX),
                       "add/position a vertex", "!BUTTON0");

    // Rectangle actions
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::INCREASE_RECTANGLE_ALPHA),
                       "increase alpha", "}LEFT_CONTROL >U");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::DECREASE_RECTANGLE_ALPHA),
                       "decrease alpha", "}LEFT_CONTROL >Y");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::INCREASE_RECTANGLE_BETA),
                       "increase beta", "}LEFT_CONTROL >J");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::DECREASE_RECTANGLE_BETA),
                       "decrease beta", "}LEFT_CONTROL >H");


    // Rectangle one click actions
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::SET_RECTANGLE_CENTER),
                       "set center", "}LEFT_CONTROL >BUTTON0");

    // Anchor actions
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::START_MOVING_ANCHOR),
                       "start moving the anchor", "}BUTTON0");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::MOVE_ANCHOR),
                       "move the anchor", "}BUTTON0 ~X|Y");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::STOP_MOVING_ANCHOR),
                        "stop moving the anchor", "<BUTTON0");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::ADD_ANCHOR),
                       "add an anchor", "!BUTTON0");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::DELETE_ANCHOR),
                       "delete the anchor", ">DELETE");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::DELETE_ANCHOR_ALT),
                       "delete the anchor", ">BACKSPACE");

    this->defineAction(static_cast<ActionType>(DrawingControllerActions::CANCEL_EDIT),
                       "cancel shape edition", ">ESCAPE");

    this->defineAction(static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT),
                       "commit shape edition", ">ENTER");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT_ALT),
                       "commit shape edition", ">KEYPAD_ENTER");

    this->defineAction(static_cast<ActionType>(DrawingControllerActions::TOGGLE_POINT_CREATION),
                       "toggle point creation", ">D");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYLINE_CREATION),
                       "toggle polyline creation", ">L");
    this->defineAction(static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYGON_CREATION),
                       "toggle polygon creation", ">P");

    this->_appEventsDispatcher.subscribe<CursorLocationChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<HoveredPickablesChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<EnableShapeCreationReadyEvent>(this);
    this->_appEventsDispatcher.subscribe<DisableShapeCreationRequest>(this);
    this->_appEventsDispatcher.subscribe<EnableShapeEditionRequest>(this);
    this->_appEventsDispatcher.subscribe<DisableShapeEditionRequest>(this);
    this->_appEventsDispatcher.subscribe<UpcomingCollectionsUpdateEvent>(this);
    this->_appEventsDispatcher.subscribe<GranuleSafeRemovalRequest>(this);
}

/**
 *
 */
DrawingController::~DrawingController()
{
    this->_appEventsDispatcher.unsubscribe<CursorLocationChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<HoveredPickablesChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<EnableShapeCreationReadyEvent>(this);
    this->_appEventsDispatcher.unsubscribe<DisableShapeCreationRequest>(this);
    this->_appEventsDispatcher.unsubscribe<EnableShapeEditionRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DisableShapeEditionRequest>(this);
    this->_appEventsDispatcher.unsubscribe<UpcomingCollectionsUpdateEvent>(this);
    this->_appEventsDispatcher.unsubscribe<GranuleSafeRemovalRequest>(this);

    this->_anchors = nullptr;   // owned and deleted by scene graph root

    delete this->_editor;
    this->_editor = nullptr;

    delete this->_creator;
    this->_creator = nullptr;

    delete this->_editedShape;
    this->_editedShape = nullptr;
}

/**
 *
 */
bool DrawingController::isActive() const
{
    return this->_active;
}

/**
 *
 */
bool DrawingController::performAction(const ActionType& actionType, const InputEvent& /*event*/, const InputDevicesState& /*devicesState*/)
{
    if
        (static_cast<ActionType>(DrawingControllerActions::CANCEL_AND_LEAVE_DRAWING) == actionType)
    {
        if
            (nullptr == this->_creator)
        {
            LOG(LogLevel::Error, "DrawingController",
                     "Cannot stop drawing if there is no shape creator");

            return true;
        }

        this->_creator->reset();

        ToggleShapeCreationRequest outEvent(this->_creator->getShapeType());
        this->_appEventsDispatcher.dispatch<ToggleShapeCreationRequest>(outEvent);

        return true;
    }
    else if
        (  (static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING) == actionType)
        ||  (static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING_ALT) == actionType))
    {
        if
            (nullptr == this->_creator)
        {
            LOG(LogLevel::Error, "DrawingController",
                     "Cannot stop drawing if there is no shape creator");

            return true;
        }

        //std::cout << "stop drawing" << std::endl;
        OGL::Shape* shape = this->_creator->createShape();
        this->_creator->reset();

        DrawingCompleteEvent outEvent(shape, this->_collectionVariable);
        this->_appEventsDispatcher.dispatch<DrawingCompleteEvent>(outEvent);

        if
            (nullptr != shape)
        {
            // Make sure the collection variable is selected so that the shape
            // will appear on the screen
            VariableSelectionRequest selectionEvent(this->_collectionVariable, true);
            this->_appEventsDispatcher.dispatch<VariableSelectionRequest>(selectionEvent);

            // Make sure the collection variable is displayed on top of the
            // other data
            MoveVariableToTopRequest moveToTopEvent(this->_collectionVariable);
            this->_appEventsDispatcher.dispatch<MoveVariableToTopRequest>(moveToTopEvent);
        }

        if
            (false == this->_explicitDrawingModeExit)
        {
            ToggleShapeCreationRequest toggleEvent(this->_creator->getShapeType());
            this->_appEventsDispatcher.dispatch<ToggleShapeCreationRequest>(toggleEvent);
        }

        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::ADD_VERTEX) == actionType)
    {
        this->_creator->addVertexUnderCursor();

        if
            (ShapeType::POINT == this->_creator->getShapeType())
        {
            // Update picking because the anchor materializing the location of
            // the point may not be at the same location anymore
            PickingUpdateRequest outEvent(true /*updatepicking mask*/, false /* no need to look for a change of picked object*/);
            this->_appEventsDispatcher.dispatch<PickingUpdateRequest>(outEvent);
        }
        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::INCREASE_RECTANGLE_ALPHA) == actionType)
    {
        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::DECREASE_RECTANGLE_ALPHA) == actionType)
    {
        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::INCREASE_RECTANGLE_BETA) == actionType)
    {
        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::DECREASE_RECTANGLE_BETA) == actionType)
    {
        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::SET_RECTANGLE_CENTER) == actionType)
    {
        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::START_MOVING_ANCHOR) == actionType)
    {
        this->_preemptEvents = true;

        //std::cout << "Start moving anchor" << std::endl;
        auto found = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::START_MOVING_ANCHOR));
        if
            (this->_possibleActions.end() != found)
        {
            this->_possibleActions.erase(found);
        }

        LockViewRequest outEvent(true);
        this->_appEventsDispatcher.dispatch<LockViewRequest>(outEvent);

        this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::MOVE_ANCHOR));
        this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::STOP_MOVING_ANCHOR));

        this->possibleActionsUpdated();

        if
            (  nullptr != this->_editor
            && nullptr != this->_hoveredAnchor)
        {
            this->_editor->startMove(this->_hoveredAnchor->getId());
        }

        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::MOVE_ANCHOR) == actionType)
    {
        //std::cout << "Move anchor" << std::endl;
        if
            (nullptr == this->_hoveredAnchor)
        {
            return true;
        }

        if
            (nullptr == this->_editor)
        {
            return true;
        }

        this->_editor->move(this->_hoveredAnchor->getType(), this->_hoveredAnchor->getId());

        // Update picking!
        PickingUpdateRequest outEvent(true /*updatepicking mask*/, false /* no need to look for a change of picked object*/);
        this->_appEventsDispatcher.dispatch<PickingUpdateRequest>(outEvent);

        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::STOP_MOVING_ANCHOR) == actionType)
    {
        this->_preemptEvents = false;

        LockViewRequest outEvent(false);
        this->_appEventsDispatcher.dispatch<LockViewRequest>(outEvent);

        //std::cout << "Stop moving anchor" << std::endl;
        auto foundStop = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::STOP_MOVING_ANCHOR));
        if
            (this->_possibleActions.end() != foundStop)
        {
            this->_possibleActions.erase(foundStop);
        }

        auto foundMove = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::MOVE_ANCHOR));
        if
            (this->_possibleActions.end() != foundMove)
        {
            this->_possibleActions.erase(foundMove);
        }

        this->_editor->stopMove();

        // Update picking as notifying the editor that anchor movement is over
        // can result in visibility changes for other anchors
        PickingUpdateRequest pickingEvent(true /* update picking mask */, true /* update picked object too */);
        this->_appEventsDispatcher.dispatch<PickingUpdateRequest>(pickingEvent);


        if
            (  nullptr != this->_editor
            && (nullptr != this->_hoveredAnchor)
            && (AnchorType::PointMoveable == this->_hoveredAnchor->getType()))
        {
            this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::START_MOVING_ANCHOR));
        }

        this->possibleActionsUpdated();

        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::ADD_ANCHOR) == actionType)
    {
        if
            (  (nullptr != this->_editor)
            && (nullptr != this->_hoveredAnchor))
        {
            this->_editor->pressed(this->_hoveredAnchor->getType(), this->_hoveredAnchor->getId());

            PickingUpdateRequest outEvent(true /* update picking mask */, true /* update picked object too */);
            this->_appEventsDispatcher.dispatch<PickingUpdateRequest>(outEvent);
        }
        return true;
    }
    else if
        (  (static_cast<ActionType>(DrawingControllerActions::DELETE_ANCHOR) == actionType)
        || (static_cast<ActionType>(DrawingControllerActions::DELETE_ANCHOR_ALT) == actionType))
    {
        if
            (  (nullptr != this->_editor)
            && (nullptr != this->_hoveredAnchor))
        {
            //std::cout << "Delete anchor" << std::endl;
            this->_editor->remove(this->_hoveredAnchor->getType(), this->_hoveredAnchor->getId());
        }
        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::CANCEL_EDIT) == actionType)
    {
        this->stopEdition(false /*edited shape is discarded*/);
        ShapeEditionCancelledEvent outEvent;
        this->_appEventsDispatcher.dispatch<ShapeEditionCancelledEvent>(outEvent);
        //this->startEdition(this->_editedRenderableId);
        return true;
    }
    else if
        (  (static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT) == actionType)
        || (static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT_ALT) == actionType))
    {
        CollectionVariable colvar {this->_editedRenderableId.collection, this->_editedRenderableId.variable};

        this->stopEdition(true /*edited shape is saved*/);

        // Make sure the collection variable is displayed on top of the
        // other data
        MoveVariableToTopRequest moveToTopEvent(colvar);
        this->_appEventsDispatcher.dispatch<MoveVariableToTopRequest>(moveToTopEvent);

        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::TOGGLE_POINT_CREATION) == actionType)
    {
        ToggleShapeCreationRequest outEvent(ShapeType::POINT);
        this->_appEventsDispatcher.dispatch<ToggleShapeCreationRequest>(outEvent);

        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYLINE_CREATION) == actionType)
    {
        ToggleShapeCreationRequest outEvent(ShapeType::POLYLINE);
        this->_appEventsDispatcher.dispatch<ToggleShapeCreationRequest>(outEvent);

        return true;
    }
    else if
        (static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYGON_CREATION) == actionType)
    {
        ToggleShapeCreationRequest outEvent(ShapeType::POLYGON);
        this->_appEventsDispatcher.dispatch<ToggleShapeCreationRequest>(outEvent);

        return true;
    }

    return false;
}

/**
 *
 */
bool DrawingController::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(CursorLocationChangedEvent)) == eventType)
    {
        const glm::vec3& hitpoint = static_cast<const CursorLocationChangedEvent&>(event).getHitpoint();
        std::uint32_t screenX = static_cast<const CursorLocationChangedEvent&>(event).getScreenX();
        std::uint32_t screenY = static_cast<const CursorLocationChangedEvent&>(event).getScreenY();
        bool onWorld = static_cast<const CursorLocationChangedEvent&>(event).getOnWorld();
        if
            ((nullptr != this->_creator) && (true == onWorld))
        {
            this->_creator->updateCursorPosition(hitpoint, screenX, screenY);
            this->updateMeasurements(this->_creator->getTemporaryShape(), this->_creator->getShapeType());
        }

        if
            ((nullptr != this->_editor) && (true == onWorld))
        {
            this->_editor->updateCursorPosition(hitpoint);
            this->updateMeasurements(this->_editor->getShapeInEdition(), this->_editor->getShapeType());
        }
        return true;
    }
    else if
        (std::type_index(typeid(EnableShapeCreationReadyEvent)) == eventType)
    {
        const ShapeType& shapeType =  static_cast<const EnableShapeCreationReadyEvent&>(event).getShapeType();
        const CollectionVariable& collectionVariable = static_cast<const EnableShapeCreationReadyEvent&>(event).getCollectionVariable();

        this->stopCreation();
        this->stopEdition(true /*edited shape is saved*/);

        SelectionLockRequest outEvent(true);
        this->_appEventsDispatcher.dispatch<SelectionLockRequest>(outEvent);

        this->startCreation(collectionVariable, shapeType);

        return true;
    }
    else if
        (std::type_index(typeid(DisableShapeCreationRequest)) == eventType)
    {
        this->stopCreation();

        SelectionLockRequest outEvent(false);
        this->_appEventsDispatcher.dispatch<SelectionLockRequest>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(EnableShapeEditionRequest)) == eventType)
    {
        const RenderableId& renderableId = static_cast<const EnableShapeEditionRequest&>(event).getRenderableId();

        this->stopCreation();
        this->stopEdition(true /*edited shape is saved*/);

        this->startEdition(renderableId);

        return true;
    }
    else if
        (std::type_index(typeid(DisableShapeEditionRequest)) == eventType)
    {
        CollectionVariable colvar {this->_editedRenderableId.collection, this->_editedRenderableId.variable};

        this->stopEdition(true /*edited shape is saved*/);

        return true;
    }
    else if
        (std::type_index(typeid(UpcomingCollectionsUpdateEvent)) == eventType)
    {
        if
            (nullptr != this->_editor)
        {
            const std::set<CollectionId>& collections = static_cast<const UpcomingCollectionsUpdateEvent&>(event).getToRemove();

            for
                (const auto& collectionId : collections)
            {
                if
                    (this->_editedRenderableId.collection == collectionId)
                {
                    this->stopEdition(true /*edited shape is saved*/);
                    break;
                }
            }
        }
        return true;
    }
    else if
        (std::type_index(typeid(HoveredPickablesChangedEvent)) == eventType)
    {
        const std::vector<Pickable*>& hovered = static_cast<const HoveredPickablesChangedEvent&>(event).getHovered();
        ShapeAnchor * anchor = nullptr;
        for
            (auto& hoveredPickable: hovered)
        {
            anchor = dynamic_cast<ShapeAnchor*>(hoveredPickable);
            if
                (nullptr != anchor)
            {
                break;
            }
        }
        this->_hoveredAnchor = anchor;

        //std::cout << "Hovered pickables changed event (anchor: " << (nullptr != this->_hoveredAnchor) << ")" <<  std::endl;

        if
            (  (nullptr != this->_editor)
            && (nullptr != this->_hoveredAnchor))
        {
            if
                (AnchorType::PointAdd == this->_hoveredAnchor->getType())
            {
                this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::ADD_ANCHOR));
            }
            if
                (  (AnchorType::PointMoveable == this->_hoveredAnchor->getType())
                && (ShapeType::POINT != this->_editor->getShapeType()))
            {
                this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::DELETE_ANCHOR));
                this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::DELETE_ANCHOR_ALT));
            }
        }
        else
        {
            auto foundAdd = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::ADD_ANCHOR));
            if
                (this->_possibleActions.end() != foundAdd)
            {
                this->_possibleActions.erase(foundAdd);
            }

            auto foundDelete = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::DELETE_ANCHOR));
            if
                (this->_possibleActions.end() != foundDelete)
            {
                this->_possibleActions.erase(foundDelete);
            }

            auto foundDeleteAlt = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::DELETE_ANCHOR_ALT));
            if
                (this->_possibleActions.end() != foundDeleteAlt)
            {
                this->_possibleActions.erase(foundDeleteAlt);
            }
        }

        if
            (  (nullptr != this->_editor)
            && (nullptr != this->_hoveredAnchor)
            && (AnchorType::PointMoveable == this->_hoveredAnchor->getType()))
        {
            this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::START_MOVING_ANCHOR));
        }
        else
        {
            auto found = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::START_MOVING_ANCHOR));
            if
                (this->_possibleActions.end() != found)
            {
                this->_possibleActions.erase(found);
            }
        }

        this->possibleActionsUpdated();

        return true;
    }
    else if
        (std::type_index(typeid(GranuleSafeRemovalRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const GranuleRemovalRequest&>(event).getRenderableId();
        if
            (  (nullptr != this->_editor)
            && (this->_editedRenderableId == rId))
        {
            this->stopEdition(false /*edited shape is not saved, it will be removed anyway*/);
        }

        GranuleRemovalRequest outEvent{rId};
        this->_appEventsDispatcher.dispatch<GranuleRemovalRequest>(outEvent);

        return true;
    }

    return false;
}

/**
 *
 */
bool DrawingController::preRender(OGL::Context* /*ctx*/, SceneView* sceneView)
{
    if
        (nullptr != this->_editor)
    {
        // Hide edited shape while editing
        sceneView->getScene()->getNodeFactory()->getSceneGraph()->setVisible(this->_editedRenderableId, false);
    }

    return false; /* do not prevent further drawing */
}

/**
 *
 */
bool DrawingController::postRender(OGL::Context* ctx, SceneView* /*sceneView*/)
{
    if
        (nullptr != this->_creator)
    {
        this->_creator->draw(ctx, ctx->pipelineParameters()->viewProjection);
    }

    if
        (nullptr != this->_editor)
    {
        this->_editor->draw(ctx, ctx->pipelineParameters()->viewProjection);
    }

    return true;
}

/**
 *
 */
ShapeMesh* DrawingController::getShapeMesh(const RenderableId& rId, bool& ok)
{
    Node* node = this->_sceneView.getScene()->getNodeFactory()->getSceneGraph()->getNode(rId);
    if
        (nullptr == node)
    {
        LOG(LogLevel::Error, "DrawingController",
                 "Could not get scene graph node for renderable with ID {}.{}/{}:{}",
                 rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
        ok = false;
        return nullptr;
    }

    GranuleNode* gNode = dynamic_cast<GranuleNode*>(node);
    if
        (nullptr == gNode)
    {
        LOG(LogLevel::Error, "DrawingController",
                "Could not get shape for {}.{}/{}:{} because it is not a granule node",
                rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
        ok = false;
        return nullptr;
    }

    Renderable2* renderable = gNode->renderable();
    ShapeMesh* shapeMesh = dynamic_cast<ShapeMesh*>(renderable);
    if
        (nullptr == shapeMesh)
    {
        LOG(LogLevel::Error, "DrawingController",
                 "Could not interpret {}.{}/{}:{} as a shape mesh",
                rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
        ok = false;
        return nullptr;
    }

    node = nullptr;
    gNode = nullptr;
    renderable = nullptr;

    ok = true;
    return shapeMesh;
}

/**
 *
 */
void DrawingController::startCreation(const CollectionVariable& collectionVariable, const ShapeType& shapeType)
{
    this->_collectionVariable = collectionVariable;
    this->_creator = ShapeCreatorFactory::create(this->_sceneView.getRenderer()->getContext(), shapeType);
    this->_creator->setAnchorPool(this->_anchors);

    //std::cout << "start drawing" << std::endl;
    if
        (ShapeType::POLYGON == this->_creator->getShapeType())
    {
        this->_possibleActions = {static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING),
                                  static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING_ALT),
                                  static_cast<ActionType>(DrawingControllerActions::CANCEL_AND_LEAVE_DRAWING),
                                  static_cast<ActionType>(DrawingControllerActions::ADD_VERTEX),};
    }
    else if
        (ShapeType::POLYLINE == this->_creator->getShapeType())
    {
        this->_possibleActions = {static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING),
                                  static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING_ALT),
                                  static_cast<ActionType>(DrawingControllerActions::CANCEL_AND_LEAVE_DRAWING),
                                  static_cast<ActionType>(DrawingControllerActions::ADD_VERTEX),};
    }
    else if
        (ShapeType::POINT == this->_creator->getShapeType())
    {
        this->_possibleActions = {static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING),
                                  static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING_ALT),
                                  static_cast<ActionType>(DrawingControllerActions::CANCEL_AND_LEAVE_DRAWING),
                                  static_cast<ActionType>(DrawingControllerActions::ADD_VERTEX),};
    }
    else
    {
        this->_possibleActions = {static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING),
                                  static_cast<ActionType>(DrawingControllerActions::COMMIT_DRAWING_ALT),
                                  static_cast<ActionType>(DrawingControllerActions::CANCEL_AND_LEAVE_DRAWING),};
    }

    DrawingStartedEvent outEvent(this->_creator->getShapeType(), this->_collectionVariable);
    this->_appEventsDispatcher.dispatch<DrawingStartedEvent>(outEvent);

    this->possibleActionsUpdated();
}

/**
 *
 */
void DrawingController::stopCreation()
{
    this->_possibleActions = {};
    this->possibleActionsUpdated();

    if
        (nullptr != this->_creator)
    {
        this->_creator->reset();
        delete this->_creator;
        this->_creator = nullptr;

        DrawingCompleteEvent outEvent(nullptr, this->_collectionVariable);
        this->_appEventsDispatcher.dispatch<DrawingCompleteEvent>(outEvent);
    }
}

/**
 *
 */
void DrawingController::startEdition(const RenderableId& renderableId)
{
    bool ok(false);
    ShapeMesh* shapeMesh = this->getShapeMesh(renderableId, ok);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "DrawingController",
                 "Could not get shape mesh for renderable {}.{}/{}:{}",
                  renderableId.collection.first, renderableId.collection.second, renderableId.granule.second, renderableId.variable);
        return;
    }

    // Hide edited shape while editing
    this->_sceneView.getScene()->getNodeFactory()->getSceneGraph()->setVisible(renderableId, false);

    this->_editor = ShapeEditorFactory::create(shapeMesh->getShape()->getType());

    SelectionLockRequest event(true);
    this->_appEventsDispatcher.dispatch<SelectionLockRequest>(event);

    this->_editedRenderableId = renderableId;

    delete this->_editedShape;
    this->_editedShape = shapeMesh->getShape()->clone();

    this->_editor->initialize(this->_sceneView.getRenderer()->getContext(), shapeMesh);
    this->_editor->setAnchorPool(this->_anchors);
    //std::cout << "Editor initialized" << std::endl;

    auto foundCancel = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::CANCEL_EDIT));
    if
        (this->_possibleActions.end() == foundCancel)
    {
        this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::CANCEL_EDIT));
        this->possibleActionsUpdated();
    }

    auto foundCommit = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT));
    if
        (this->_possibleActions.end() == foundCommit)
    {
        this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT));
        this->possibleActionsUpdated();
    }

    auto foundCommitAlt = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT_ALT));
    if
        (this->_possibleActions.end() == foundCommitAlt)
    {
        this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT_ALT));
        this->possibleActionsUpdated();
    }

    /*If the scene graph is updated, after an update of the renderables set
    after a zoom event for example, the visibility of the shape will be
    restored (the rendering config of the collection will be applied). To
    prevent that we need to re-apply the visibility setting in the rendering
    loop by registering a pre-rendering hook*/
    this->_sceneView.getRenderer()->registerPreRenderer(this);
}

/**
 *
 */
void DrawingController::clearEdition()
{
    this->_anchors->clear();

    delete this->_editor;
    this->_editor = nullptr;

    delete this->_editedShape;
    this->_editedShape = nullptr;
}

/**
 *
 */
void DrawingController::stopEdition(bool saveShape)
{
    /*The pre-rendering hook for hiding the edited shape is not required
    anymore*/
    this->_sceneView.getRenderer()->unregisterPreRenderer(this);

    SelectionLockRequest event(false);
    this->_appEventsDispatcher.dispatch<SelectionLockRequest>(event);

    auto foundCancel = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::CANCEL_EDIT));
    if
        (this->_possibleActions.end() != foundCancel)
    {
        this->_possibleActions.erase(foundCancel);
        this->possibleActionsUpdated();
    }

    auto foundCommit = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT));
    if
        (this->_possibleActions.end() != foundCommit)
    {
        this->_possibleActions.erase(foundCommit);
        this->possibleActionsUpdated();
    }

    auto foundCommitAlt = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::COMMIT_EDIT_ALT));
    if
        (this->_possibleActions.end() != foundCommitAlt)
    {
        this->_possibleActions.erase(foundCommitAlt);
        this->possibleActionsUpdated();
    }

    if
        (nullptr == this->_editor)
    {
        this->clearEdition();

        return;
    }

    bool ok(false);
    ShapeMesh* shapeMesh = this->getShapeMesh(this->_editedRenderableId, ok);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "DrawingController",
                 "Could not get shape mesh for renderable {}.{}/{}:{}",
                  this->_editedRenderableId.collection.first, this->_editedRenderableId.collection.second, this->_editedRenderableId.granule.second, this->_editedRenderableId.variable);

        this->clearEdition();

        return;
    }

    OGL::Shape* shape(nullptr);
    if
        (  false == saveShape
        || false == this->_editor->shapeIsValid())
    {
        // Restore shape
        shape = this->_editedShape;

        // The shape will be used again in the future so we don't want to
        // delete it when clearEdition is called => set _editedShape to nullptr
        // so the delete will not do anything and "shape" can still be used
        this->_editedShape = nullptr;
    }
    else
    {
        shape = this->_editor->createShape();
    }

    this->clearEdition();

    if
        (nullptr != shape)
    {
        ShapeGPUCreator::update(this->_sceneView.getRenderer()->getContext(), shapeMesh, shape);

        // Shape should be visible again now that edition is over
        this->_sceneView.getScene()->getNodeFactory()->getSceneGraph()->setVisible(this->_editedRenderableId, true);

        ShapeEditionCompleteEvent outEvent(shape, this->_editedRenderableId);
        this->_appEventsDispatcher.dispatch<ShapeEditionCompleteEvent>(outEvent);
    }

    shape = nullptr;
}

/**
 *
 */
void DrawingController::updateMeasurements(const OGL::Shape* shape, const ShapeType& shapeType)
{
    if
        (nullptr == shape)
    {
        return;
    }

    double area = std::numeric_limits<double>::quiet_NaN();	//
    double length = std::numeric_limits<double>::quiet_NaN();	//

    if (SEAScope::POLYLINE == shapeType)
    {
        length = SEAScope::perimeterFromShape(shape, false);
    }
    else if (SEAScope::POLYGON == shapeType)
    {
        length = SEAScope::perimeterFromShape(shape, true);
        area = SEAScope::areaFromShape(shape);
    }

    MeasurementsUpdatedEvent outEvent(length, area, shapeType);
    this->_appEventsDispatcher.dispatch<MeasurementsUpdatedEvent>(outEvent);
}

/**
 *
 */
void DrawingController::possibleActionsUpdated()
{
    // Make sure permanent actions are there
    auto pointCreationFound = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::TOGGLE_POINT_CREATION));
    if
        (this->_possibleActions.end() == pointCreationFound)
    {
        this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::TOGGLE_POINT_CREATION));
    }
    auto polylineCreationFound = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYLINE_CREATION));
    if
        (this->_possibleActions.end() == polylineCreationFound)
    {
        this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYLINE_CREATION));
    }
    auto polygonCreationFound = this->_possibleActions.find(static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYGON_CREATION));
    if
        (this->_possibleActions.end() == polygonCreationFound)
    {
        this->_possibleActions.insert(static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYGON_CREATION));
    }

    DrawingActionsUpdatedEvent event(this->_actorId);
    this->_appEventsDispatcher.dispatch<DrawingActionsUpdatedEvent>(event);
}

/**
 *
 */
bool DrawingController::getExplicitDrawingModeExit() const
{
    return this->_explicitDrawingModeExit;
}

/**
 *
 */
bool DrawingController::setExplicitDrawingModeExit(bool value)
{
    if
        (value == this->_explicitDrawingModeExit)
    {
        return false;
    }

    this->_explicitDrawingModeExit = value;
    return true;
}
