// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "tools/extraction_controller.h"

#include "core/log.h"
#include "core/source.h"
#include "core/input_event.h"
#include "core/input_devices_state.h"
#include "3d/camera.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "3d/shape_mesh.h"
#include "3d/default_value.h"
#include "3d/scene_view.h"
#include "3d/iscene_renderer.h"
#include "core/metadata_mediator.h"
#include "core/data_bucket.h"
#include "core/data_mediator.h"
#include "core/engine.h"
#include "core/path.h"
#include "core/granule_metadata.h"
#include "core/data_bucket.h"


#include <iostream>

using namespace SEAScope;

/**
 *
 */
ExtractionController::ExtractionController()
: BaseInputEventsSubscriber("Extraction controller")
{
    // Configure actions

    // - Resize framebuffer
    this->defineAction(static_cast<ActionType>(ExtractionControllerActions::RESIZE_FRAMEBUFFER),
                       "Resize extractor framebuffer", "#?");

    this->_possibleActions = {static_cast<ActionType>(ExtractionControllerActions::RESIZE_FRAMEBUFFER),};
}

/**
 *
 */
ExtractionController::~ExtractionController()
{
    this->clearExtraction(this->_extractedData, this->_extractedLabels);
}

/**
 *
 */
bool ExtractionController::isActive() const
{
    bool extractorHasFBO(false);
    bool transectHasFBO(false);

    // The only input event handled by this class is the framebuffer resize
    // event and this event will have no effect unless there is at least one
    // FBO to resize, so the controller can be considered inactive for the
    // input events dispatcher until one of the extraction tools has been used
    // at least once and has initialized its FBO.
    extractorHasFBO = (nullptr != this->_extractor.getFBO());
    transectHasFBO = (nullptr != this->_transect.getFBO());
    return this->_active && (extractorHasFBO || transectHasFBO);
}

/**
 *
 */
bool ExtractionController::performAction(const ActionType& actionType,
                                         const InputEvent& event,
                                         const InputDevicesState& /*devicesState*/)
{
    if
        (static_cast<ActionType>(ExtractionControllerActions::RESIZE_FRAMEBUFFER) == actionType)
    {
        LOG(LogLevel::Debug, "ExtractionController", "Resize framebuffer");
        this->_extractor.resize(event.getWidth(), event.getHeight());
        this->_transect.resize(event.getWidth(), event.getHeight());
        return true;
    }

    return false;
}

/**
 *
 */
bool ExtractionController::extractData(SceneView& sceneView, Renderable2* footprint, std::map<RenderableId, Extraction>& extractionData, bool discardExtractedData)
{
    bool ok(false); //

    if
        (nullptr == this->_extractor.getFBO())
    {
        this->_extractor.create(sceneView.getRenderer()->getContext(),
                                sceneView.getViewport().x,
                                sceneView.getViewport().y);
    }
    ok = this->_extractor.extract(sceneView.getRenderer()->getContext(),
                                  footprint,
                                  sceneView,
                                  extractionData,
                                  discardExtractedData);
    return ok;
}

/**
 *
 */
bool ExtractionController::transectData(SceneView& sceneView, ShapeMesh* footprint, std::map<RenderableId, Extraction>& extractionData, bool discardExtractedData)
{
    bool ok(false); //

    if
        (nullptr == this->_transect.getFBO())
    {
        this->_transect.create(sceneView.getRenderer()->getContext(),
                               sceneView.getViewport().x,
                               sceneView.getViewport().y);
    }
    ok = this->_transect.extract(sceneView.getRenderer()->getContext(),
                                  footprint,
                                  sceneView,
                                  extractionData,
                                  discardExtractedData);
    return ok;
}

/**
 *
 */
bool ExtractionController::extractFromRenderableId(RenderableId rId, Engine* engine, std::map<RenderableId, Extraction>& extractedData, std::map<RenderableId, std::string>& extractedLabels, bool discardExtractedData)
{
    bool ok(false);

    if
        (nullptr == engine)
    {
        LOG(LogLevel::Error, "ExtractionController",
                 "Cannot perform extraction without a valid application engine (null pointer)");
        return false;
    }

    RenderingCfg* rCfg {nullptr};
    rCfg = engine->renderingControl.model.getCfg(rId.collection, rId.granule, rId.variable, ok);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "Engine", "Cannot get rendering configuration for selected granule/variable");
        return false;
    }
    RenderMethod renderMethod = rCfg->renderMethod();
    rCfg = nullptr; // memory-managed by RenderingControlModel

    this->clearExtraction(extractedData, extractedLabels);
    this->initExtraction(*engine, extractedData, extractedLabels);

    SceneView* sceneView = engine->getSceneViewController().getSceneView();

    // Get footprint for the renderable acting as extraction mask
    Renderable2* footprint(nullptr);
    ok = sceneView->getScene()->getNodeFactory()->getSceneDictionnary()->exist(rId, footprint);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "ExtractionController",
                 "Could not find renderable {}/{}/{}:{} in the scene, so extraction is not possible",
                 rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
        return false;
    }

    LOG(LogLevel::Info, "ExtractionController",
             "Extracting using {}/{}/{}:{} as mask (rendering method for mask granule: {}, trajectories: {}, polyline: {}, point: {})",
             rId.collection.first, rId.collection.second, rId.granule.second, rId.variable,
             int(renderMethod), int(RenderMethod::TRAJECTORIES), int(RenderMethod::POLYLINE), int(RenderMethod::POINT));

    if
        (  (RenderMethod::TRAJECTORIES == renderMethod)
        || (RenderMethod::POLYLINE == renderMethod)
        || (RenderMethod::POINT == renderMethod))
    {
        ShapeMesh* shapeMesh = dynamic_cast<ShapeMesh*>(footprint);
        ok = this->transectData((*sceneView), shapeMesh, extractedData, discardExtractedData);
    }
    else
    {
        ok = this->extractData((*sceneView), footprint, extractedData, discardExtractedData);
    }

    // Set the middle of the extraction temporal coverage as time value in GCPs
    for
        (auto& extraction : extractedData)
    {
        std::uint64_t timeValue = static_cast<std::uint64_t>(1000 * std::floor((extraction.second.start + extraction.second.stop) / 2));
        for
            (auto& shapeGCP : extraction.second.shape)
        {
            shapeGCP.time = timeValue;
        }

        for
            (auto& gcp : extraction.second.gcps)
        {
            gcp.time = timeValue;
        }
    }

    return ok;
}

/**
 *
 */
bool ExtractionController::makePersistentExtraction(const RenderableId& rId, Engine* engine)
{
    bool ok(false);
    ok = this->extractFromRenderableId(rId, engine, this->_extractedData, this->_extractedLabels, false);
    return ok;
}

/**
 *
 */
bool ExtractionController::listIntersectedGranules(const RenderableId& rId, Engine* engine, std::vector<std::string>& result)
{
    bool ok(false);
    std::map<RenderableId, Extraction> extractedData;
    std::map<RenderableId, std::string> extractedLabels;
    ok = this->extractFromRenderableId(rId, engine, extractedData, extractedLabels, true /*discard data*/);
    if
        (true == ok)
    {
        for
            (const auto& it: extractedData)
        {
            result.push_back(it.second.uri);
        }
    }
    return ok;
}

/**
 *
 */
bool ExtractionController::clearExtraction(std::map<RenderableId, Extraction>& data, std::map<RenderableId, std::string>& labels)
{
	for
		(auto& it: data)
	{
		it.second.buffer.deallocate();
	}
	data.clear();
    labels.clear();

	return true;
};

/**
 *
 */
bool ExtractionController::initExtraction(Engine& engine, std::map<RenderableId, Extraction>& result, std::map<RenderableId, std::string>& labels)
{
	RenderableId trajId;	//
	VariableCfg varCfg;	//
	std::vector<float> fillValues;	//
	std::vector<std::string> fieldNames;	//
	timestamp_t tStart;	//
	timestamp_t tStop;	//
	bool ok;	//

    SceneView* sceneView = engine.getSceneViewController().getSceneView();
	for
		(auto& rId: sceneView->getRenderer()->getToRender())
	{
        if
            (SEAScope::Source::ANNOTATIONS == rId.collection.first)
        {
            // Annotations are not extracted
            continue;
        }
		LOG( LogLevel::Debug, "ExtractionController"
				, "Create extraction data for {}/{}/{}:{}"
				, rId.collection.first, rId.collection.second
				, rId.granule.second, rId.variable);

		fillValues.clear();
		fieldNames.clear();

		ok = engine.metadataMediator.getVariableCfg( rId.collection
												    , rId.variable
												    , varCfg);

		if
			(false == ok)
		{
			LOG( LogLevel::Warn, "ExtractionController"
					, "Failed to retrieve rendering configuration for extractable granule");
			continue;
		}

        GranuleMetadata gMeta;  //
        ok = engine.getRenderedGranuleMetadata(rId, gMeta);
		if
			(false == ok)
		{
			LOG(LogLevel::Error, "ExtractionController",
                     "Could not find metadata for granule {}/{}", rId.granule.first, rId.granule.second);
			continue;
		}
        tStart = static_cast<timestamp_t>(0.001 * static_cast<double>(gMeta.timeCoverageStart()));
        tStop = static_cast<timestamp_t>(0.001 * static_cast<double>(gMeta.timeCoverageEnd()));

        std::string uri("");
        ok = engine.renderingMediator.getRenderableURI(rId, uri);
		if
			(false == ok)
		{
            LOG(LogLevel::Warn, "ExtractionController",
                "No URI associated with renderable object {}.{}.{}:{} - skipping",
                rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
			continue;
		}

		for
			(auto& field: varCfg.fields)
		{
            const Image* img {nullptr};
            const DataInfo* info {nullptr};
            OGL::Shape* geometry {nullptr};
            ok = engine.dataMediator.getField(rId.granule, uri, field, img, geometry, info);
            if
                (false == ok)
            {
				// TODO: check if we should not skip the granule instead
				// of just skipping the field
                continue;
            }

			float unpackedFill = info->offsets[0] + info->scaleFactors[0] * info->fillValues[0];
			fillValues.push_back(unpackedFill);
			fieldNames.push_back(field);
		}

        RenderingCfg* rCfg {nullptr};
        rCfg = engine.renderingControl.model.getCfg(rId.collection, rId.granule, rId.variable, ok);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Engine", "Cannot get rendering configuration for selected granule/variable");
            return false;
        }
        RenderMethod renderMethod = rCfg->renderMethod();
        rCfg = nullptr; // memory-managed by RenderingControlModel

        RenderableId tmpId(rId);
		if
			(RenderMethod::TRAJECTORIES == renderMethod)
		{
            /* Extraction is performed for all granules at once for collections
            rendered as trajectories, only display one entry in the results */
            tmpId.granuleLevel = false;
			tmpId.granule.first = 0;
			tmpId.granule.second = 0;

			auto found = result.find(tmpId);
			if
				(result.end() != found)
			{
				// PolylinePool already registered
                /* There is one pool of polylines per collection and all the
                trajectories from a collection are rendered at the same time.
                The extraction mechanism is similar, so the buffer holding the
                result of the extraction for a trajectory actually contains the
                data extracted from all the trajectories that belong to this
                collection.

                Having the renderable ids of all individual trajectories
                registered in the extraction results would only slow down the
                and consume more resources to store duplicate information, that
                is why only one renderable identifier is registered in the
                extraction results for this collection.
                */
				continue;
			}

            uri = "[all trajectories]";
        }

        result[tmpId];
        result[tmpId].fillValues = fillValues;
        result[tmpId].uri = uri;
        result[tmpId].fields = fieldNames;
        result[tmpId].start = tStart;
        result[tmpId].stop = tStop;


        // Build label for this extraction
        std::forward_list<std::string> pathElements;	//
        std::string pathElement;	//
        std::string colvarLabel("");
        ok = engine.getRenderableLabel(tmpId, colvarLabel, "\n");
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "ExtractionController",
                     "Could not build label for collection:variable {}/{}:{}",
                     tmpId.collection.first, tmpId.collection.second, tmpId.variable);
            return false;
        }

		pathElements.clear();
		pathElement = "";
        std::string delimiter(1, Path::separator());
		Path::splitPaths(uri, delimiter, pathElements);
		for
			(const auto& itPath : pathElements)
		{
			pathElement = itPath;
		}

        labels[tmpId] = colvarLabel + "\n" + pathElement;
	}

	return true;
};

/**
 *
 */
const std::map<RenderableId, Extraction>& ExtractionController::getExtractedData() const
{
    return this->_extractedData;
}

/**
 *
 */
const std::map<RenderableId, std::string>& ExtractionController::getExtractedLabels() const
{
    return this->_extractedLabels;
}

