// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "tools/scene_view_controller_original.h"

#include "3d/scene_view.h"
#include "3d/default_value.h"
#include "core/types.h"
#include "core/log.h"
#include "core/input_devices_state.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "3d/math_func.h"
#include "3d/iscene_renderer.h"

#include "core/input_event.h"
#include "events/change_rasterization_mode_request.h"
#include "events/zoom_in_request.h"
#include "events/zoom_out_request.h"
#include "events/camera_reset_request.h"
#include "events/camera_aligned_with_poles_request.h"
#include "events/camera_poles_lock_request.h"
#include "events/cursor_location_changed_event.h"
#include "events/lock_view_request.h"
#include "events/change_rasterization_mode_request.h"
#include "events/poles_axis_lock_event.h"
#include "events/camera_moved_event.h"
#include "core/app_events_dispatcher.h"


namespace SEAScope
{

SceneViewControllerOriginal::SceneViewControllerOriginal(AppEventsDispatcher& appEventsDispatcher)
:	SceneViewControllerBase("Camera controller", appEventsDispatcher),
	_zoomSpeed(1.f),
	_angles(0, 0),
	_zoom(GLOBE_ZOOM_DEFAULT),
	_lockPolesAxis(true),
    _arcBallWasOnGlobe(false),
	_viewLocked(false)
{
	this->initEvents();
}

SceneViewControllerOriginal::~SceneViewControllerOriginal()
{
	this->cleanEvents();
}

bool SceneViewControllerOriginal::isActive() const
{
	return this->_active && (nullptr != this->_sceneView);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Internal methods (/!\ cannot emit CameraMovedEvent /!\)
/////////////////////////////////////////////////////////////////////////////////////////////////////

bool SceneViewControllerOriginal::setZoomImpl(float altitude)
{
	if
		(	(altitude >= GLOBE_MAX_ZOOM_DISTANCE)
		 || (altitude <= GLOBE_MIN_ZOOM_DISTANCE))
	{
		// out of range
		return false;
	}

	this->_zoom = altitude;

	this->_sceneView->getCamera()->lookAt(glm::vec3(altitude, 0, 0),
										  glm::vec3(0.f),
										  glm::vec3(0, 0, 1));

    return true;
}

bool SceneViewControllerOriginal::zoomInImpl()
{
	float zoom = this->_zoom;
	zoom = zoom - (zoom - GLOBE_RADIUS) * this->_zoomSpeed * GLOBE_ZOOM_IN_SPEED;
	return this->setZoomImpl(zoom);
}

/**
 * Update `_q` and the scene's camera to the position given by `_angles`.
 */
void SceneViewControllerOriginal::updateCameraRotation()
{
	if
        (this->_angles.x > 360.f)
    {
		this->_angles.x -= 360.f;
    }

	if
        (this->_angles.x < 0.f)
    {
		this->_angles.x += 360.f;
    }

	// Clamp `_angles.y` in [-89.9, 89.9]
	this->_angles.y = glm::clamp(this->_angles.y, -89.9f, 89.9f);

	// Update `_q`
	glm::mat4 m(1.f);

	m = glm::rotate(m, glm::radians(-1 * this->_angles.y), glm::vec3(0.f, 1.f, 0.f));
	m = glm::rotate(m, glm::radians(-1 * this->_angles.x), glm::vec3(0.f, 0.f, 1.f));

	this->_q = glm::toQuat(m);

	// Update scene's camera position
	this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
}

/**
 * Get position on the globe for the given mouse position.
 *
 * @param globePosition The position on the globe.
 * @param mousePosition The position of the mouse.
 * @param q The rotation quaternion of the scene.
 * @return Whether the mouse position is on the globe (`true`) or in space (`false`).
 *
 * @note globePosition is only modified when the mouse position is on the globe.
 */
bool SceneViewControllerOriginal::getGlobePosition(glm::vec3& globePosition,
												   const glm::vec2& mousePosition,
												   const glm::quat q)
{
	bool ok = getPickingPosition(this->_sceneView->getViewport(),
								 static_cast<int>(mousePosition.x),
								 static_cast<int>(mousePosition.y),
								 globePosition,
								 q,
								 this->_sceneView->getCamera()->getViewMatrix(),
								 this->_sceneView->getCamera()->getProjectionMatrix());
	return ok;
}

/**
 * Get position on the globe for the current mouse position (`_lastMousePosition`).
 *
 * @param globePosition The position on the globe.
 * @return Whether the mouse position is on the globe (`true`) or in space (`false`).
 *
 * @note globePosition is only modified when the mouse position is on the globe.
 */
bool SceneViewControllerOriginal::getGlobePosition(glm::vec3& globePosition)
{
    return this->getGlobePosition(globePosition, this->_lastMousePosition, this->_q);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Camera actions
/////////////////////////////////////////////////////////////////////////////////////////////////////

bool SceneViewControllerOriginal::startCameraMovement()
{
    bool onGlobe(false);    //

    onGlobe = this->getGlobePosition(this->_arcBallGlobePos);
	if
		(true == onGlobe)
	{
        this->_arcBallQ = this->_q;

		// Save quaternion as it will be used as origin for future movements
		this->updatePossibleActions({ SceneViewControllerOriginalActions::MOVE,
									  SceneViewControllerOriginalActions::MOVE_END },
									{ SceneViewControllerOriginalActions::MOVE_START });
		this->_preemptEvents = true;
	}

    // Reset the "on globe" status of the arc ball manipulator
    // The code works without this reset, but setting the value here makes the
    // state of the instance more coherent (otherwise _arcBallWasOnGlobe may
    // have an outdated value set during a previous usage of the arcball
    // manipulator)
    this->_arcBallWasOnGlobe = onGlobe;

    return onGlobe;
}

bool SceneViewControllerOriginal::moveCamera(const InputEvent& event)
{
	if
		(true == this->_viewLocked)
	{
		return false;
	}

    glm::vec3 hitPoint; //

	if
        (true == this->_lockPolesAxis)
	{
		double pixelSizeTmp(0.0);   //

		double perimeterTmp(0.0);   //
		double tmpValue(0.0);   //
		bool ok(false); //

		pixelSizeTmp = static_cast<double>(this->_sceneView->getPixelSize());

		// Default perimeter used when the cursor is not on the globe
		perimeterTmp = 0.5 * GLOBE_EQUATOR_LENGTH;

		ok = this->getGlobePosition(hitPoint);

		if
			(ok)
		{
			Geodetic2D geo = UnitSphere::ToGeodetic2D(hitPoint);
			perimeterTmp = GLOBE_EQUATOR_LENGTH * std::max(0.001, std::cos(geo.Latitude()));
		}

		tmpValue = -1.0 * event.getMouseDx() * pixelSizeTmp * 360.0 / perimeterTmp;
		this->_angles.x += static_cast<float>(tmpValue);

		tmpValue = -1.0 * event.getMouseDy() * pixelSizeTmp * 360.0 / GLOBE_EQUATOR_LENGTH;
		this->_angles.y += static_cast<float>(tmpValue);

		this->updateCameraRotation();
	}
    else
    {
        bool onGlobe(false);
        onGlobe = this->getGlobePosition(hitPoint, this->_lastMousePosition, this->_arcBallQ);

        if
            (onGlobe)
        {
            if
                (false == this->_arcBallWasOnGlobe)
            {
                this->_arcBallGlobePos = hitPoint;
            }

            glm::quat q = rotationBetweenVectors(this->_arcBallGlobePos, hitPoint);
            this->_q = this->_arcBallQ * q;
            this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
        }
        else
        {
            // Use theta-phi

            double perimeter(0.5 * GLOBE_EQUATOR_LENGTH);    // default value when the cursor is not on the globe

            double theta(event.getMouseDx() * static_cast<double>(this->_sceneView->getPixelSize()) * 2 * M_PI / perimeter);
            double phi(event.getMouseDy() * static_cast<double>(this->_sceneView->getPixelSize()) * 2 * M_PI / GLOBE_EQUATOR_LENGTH);
            glm::quat zRot(static_cast<float>(std::cos(0.5 * theta)), 0.f, 0.f, static_cast<float>(std::sin(0.5 * theta)));
            glm::quat yRot(static_cast<float>(std::cos(0.5 * phi)), 0.f, static_cast<float>(std::sin(0.5 * phi)), 0.f);
            this->_q = zRot * yRot * this->_q;

            this->_arcBallQ = this->_q;

            this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
        }

        this->_arcBallWasOnGlobe = onGlobe;
    }

    // Update cursor position on the globe
    this->getGlobePosition(this->_zoomGlobePos);

    // Other components must be notified of camera changes
	this->emitCameraMovedEvent();

    return true;
}

bool SceneViewControllerOriginal::stopCameraMovement()
{
	ActionType actionType;  //

	// Remove MOVE to possible actions
	actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE);
	this->_possibleActions.erase(this->_possibleActions.find(actionType));

	// Remove MOVE_END to possible actions
	actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_END);
	this->_possibleActions.erase(this->_possibleActions.find(actionType));

	// Add MOVE_START from possible actions
	actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_START);
	this->_possibleActions.insert(actionType);

	this->_preemptEvents = false;

	return true;
}

void SceneViewControllerOriginal::setZoom(float altitude)
{
    bool zoomChanged {false};

    zoomChanged = this->setZoomImpl(altitude);

    if
        (true == zoomChanged)
    {
        // Camera changed so the position under the mouse cursor might not be
        // the same and other components must be notified
	    this->emitCameraMovedEvent();
    }
}

void SceneViewControllerOriginal::resetCamera()
{
	this->setZoomImpl(GLOBE_ZOOM_DEFAULT);
	this->lookAt(0, 0);
}

void SceneViewControllerOriginal::zoomIn()
{
    bool zoomChanged {false}; //

    zoomChanged = this->zoomInImpl();

    if
        (true == zoomChanged)
    {
        // Camera changed so the position under the mouse cursor might not be
        // the same and other components must be notified
        this->getGlobePosition(this->_zoomGlobePos);
        this->emitCameraMovedEvent();
    }
}

void SceneViewControllerOriginal::zoomOut()
{
    bool zoomChanged {false}; //
	float zoom = this->_zoom;
	zoom = zoom + (zoom - GLOBE_RADIUS) * this->_zoomSpeed * GLOBE_ZOOM_OUT_SPEED;
	zoomChanged = this->setZoomImpl(zoom);

    if
        (true == zoomChanged)
    {
        // Camera changed so the position under the mouse cursor might not be
        // the same and other components must be notified
        this->getGlobePosition(this->_zoomGlobePos);
        this->emitCameraMovedEvent();
    }
}

/**
 * Set `_lockPolesAxis` to `lockPolesAxis` and dispatch a `PolesAxisLockEvent`.
 * And when locking the poles axis, rotate the camera as needed.
 *
 * @param lockPolesAxis Whether or not to lock the poles axis.
 */
void SceneViewControllerOriginal::setLockPolesAxis(bool lockPolesAxis)
{
    if
        (lockPolesAxis == this->_lockPolesAxis)
    {
        return;
    }

	this->_lockPolesAxis = lockPolesAxis;

	if (true == this->_lockPolesAxis)
	{
		glm::vec3 pCam = glm::vec3(1.f, 0.f, 0.f) * this->_q;
		Geodetic2D geo = UnitSphere::ToGeodetic2D(pCam);
		this->lookAt(static_cast<float>(glm::degrees(geo.Longitude())),
					 static_cast<float>(glm::degrees(geo.Latitude())));
	}

    PolesAxisLockEvent event(this->_lockPolesAxis);
    this->_appEventsDispatcher.dispatch<PolesAxisLockEvent>(event);
}

/**
 * Dispatch `CameraMovedEvent`.
 */
void SceneViewControllerOriginal::emitCameraMovedEvent()
{
    SEAScope::CameraMovedEvent event;
    this->_appEventsDispatcher.dispatch<SEAScope::CameraMovedEvent>(event);
}

void SceneViewControllerOriginal::setState(const SceneViewState& state)
{
	this->_lockPolesAxis = state.lock;
	this->_q = state.q;

    this->setZoom(state.zoom);

    if
        (true == this->_lockPolesAxis)
    {
        glm::vec3 pCam = glm::vec3(1.f, 0.f, 0.f) * this->_q;
        Geodetic2D geo = UnitSphere::ToGeodetic2D(pCam);
        this->lookAt(static_cast<float>(glm::degrees(geo.Longitude())), static_cast<float>(glm::degrees(geo.Latitude())));
    }
    else
    {
        this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
    }

    PolesAxisLockEvent event(this->_lockPolesAxis);
    this->_appEventsDispatcher.dispatch<PolesAxisLockEvent>(event);
}

void SceneViewControllerOriginal::getState(SceneViewState& state) const
{
	state.lock = this->_lockPolesAxis;
	state.q = this->_q;
	state.zoom = this->_zoom;
}

/**
 * Update cursor position in screen space (`_lastMousePosition`) from
 * `devicesState`.
 *
 * @param devicesState
 */
void SceneViewControllerOriginal::updateCursorPosition(const InputDevicesState& devicesState)
{
	this->_lastMousePosition = glm::ivec2(devicesState.getMouseX(), devicesState.getMouseY());
}

void SceneViewControllerOriginal::zoomInTo()
{
    bool zoomChanged {false}; //

	glm::vec3 hitPointBefore;
	bool wasOnGlobe = this->getGlobePosition(hitPointBefore);

	zoomChanged = this->zoomInImpl();
    if
        (false == zoomChanged)
    {
        // if zoom did not change, then the cursor position on the globe (or in
        // space) is still the same and there is no need to perform further
        // action
        return;
    }

    glm::vec3 hitPointAfter;
	bool isOnGlobe = this->getGlobePosition(hitPointAfter);

    if
           (true == isOnGlobe)
    {
        if
            (false == wasOnGlobe)
        {
            // The cursor entered the globe area, simply update _zoomGlobePos
            // for future use.
            this->_zoomGlobePos = hitPointAfter;
        }
        else
        {
            // Try to rotate the globe so that the position under the mouse cursor
            // remains the same before and after zooming-in.
            if
                (true == this->_lockPolesAxis)
            {
                glm::vec2 lastPos = UnitSphere::ToGeodetic2D(this->_zoomGlobePos).toVec2();
                glm::vec2 currentPos = UnitSphere::ToGeodetic2D(hitPointAfter).toVec2();

                this->_angles.x -= glm::degrees(currentPos.x - lastPos.x);
                this->_angles.y += glm::degrees(currentPos.y - lastPos.y);

                this->updateCameraRotation();
            }
            else
            {
                glm::quat q = rotationBetweenVectors(this->_zoomGlobePos, hitPointAfter);

                this->_q = this->_q * q;
                this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
            }
        }
	}

    // Other components must be notified of camera changes
	this->emitCameraMovedEvent();
}

void SceneViewControllerOriginal::lookAt(float lon, float lat)
{
	if (true == this->_lockPolesAxis)
	{
		double lonDegrees(0.0); //
		double latDegrees(0.0); //
		double lonRad(0.0); //
		double latRad(0.0); //

		lonRad = static_cast<double>(glm::radians(lon));
		latRad = static_cast<double>(glm::radians(lat));

		latRad = glm::clamp(latRad, -M_PI_2, M_PI_2);

		while
            (lonRad < -M_PI - static_cast<double>(std::numeric_limits<float>::epsilon()))
        {
			lonRad += M_PI;
        }
		while
            (lonRad > M_PI + static_cast<double>(std::numeric_limits<float>::epsilon()))
        {
			lonRad -= M_PI;
        }

		lonDegrees = glm::degrees(lonRad);
		this->_angles.x = static_cast<float>(lonDegrees);
		latDegrees = glm::degrees(-1.0 * latRad);
		this->_angles.y = static_cast<float>(latDegrees);
		this->updateCameraRotation();
	}
	else
	{
		this->_q = glm::quat_cast(glm::mat3(1.f));

		this->_q = glm::rotate(this->_q, glm::radians(lat), glm::vec3(0.f, 1.f, 0.f));
		this->_q = glm::rotate(this->_q, -1 * glm::radians(lon), glm::vec3(0.f, 0.f, 1.f));

		this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
	}

    // Assume that camera changed, so the position under the mouse cursor might
    // not be the same and other components must be notified
    this->getGlobePosition(this->_zoomGlobePos);
    this->emitCameraMovedEvent();
}

/// ///////////////////////////////////////////////////////////////////////////////////////////////////
/// Events
/// ///////////////////////////////////////////////////////////////////////////////////////////////////

bool SceneViewControllerOriginal::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
	if
		(std::type_index(typeid(ZoomInRequest)) == eventType)
	{
		this->zoomIn();
		return true;
	}
	else if
		(std::type_index(typeid(ZoomOutRequest)) == eventType)
	{
		this->zoomOut();
		return true;
	}
	else if
		(std::type_index(typeid(CameraResetRequest)) == eventType)
	{
		this->resetCamera();
		return true;
	}
	else if
		(std::type_index(typeid(CameraAlignedWithPolesRequest)) == eventType)
	{
// 		this->lockPolesAxis();
// 		this->updateCameraRotation();
		return true;
	}
	else if
		(std::type_index(typeid(CameraPolesLockRequest)) == eventType)
	{
		bool state = static_cast<const CameraPolesLockRequest&>(event).getState();
		this->setLockPolesAxis(state);
		return true;
	}
	else if
		(std::type_index(typeid(LockViewRequest)) == eventType)
	{
		bool locked = static_cast<const LockViewRequest&>(event).getLocked();
		this->_viewLocked = locked;
	}
	else if
		(std::type_index(typeid(ChangeRasterizationModeRequest)) == eventType)
	{
		OGL::RasterizationMode mode = static_cast<const ChangeRasterizationModeRequest&>(event).getMode();
		this->_sceneView->getRenderer()->setRasterizationMode(mode);

		if
			(OGL::fill == mode)
		{
			auto found = this->_possibleActions.find(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE));
			if
				(this->_possibleActions.end() != found)
			{
				this->_possibleActions.erase(found);
			}
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE));
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE));
		}
		else if
			(OGL::line == mode)
		{
			auto found = this->_possibleActions.find(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE));
			if
				(this->_possibleActions.end() != found)
			{
				this->_possibleActions.erase(found);
			}
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE));
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE));

		}
		else if
			(OGL::point == mode)
		{
			auto found = this->_possibleActions.find(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE));
			if
				(this->_possibleActions.end() != found)
			{
				this->_possibleActions.erase(found);
			}
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE));
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE));
		}
		return true;
	}

	return false;
}

void SceneViewControllerOriginal::updatePossibleActions(const std::set<SceneViewControllerOriginalActions>& toEnable,
	const std::set<SceneViewControllerOriginalActions>& toDisable)
{
	ActionType actionType;

	for
		(const auto& it : toEnable)
	{
		actionType = static_cast<ActionType>(it);
		this->_possibleActions.insert(actionType);
	}

	for
		(const auto& it : toDisable)
	{
		actionType = static_cast<ActionType>(it);

		auto found = this->_possibleActions.find(actionType);
		if
			(this->_possibleActions.end() != found)
		{
			this->_possibleActions.erase(found);
		}
	}
}

bool SceneViewControllerOriginal::performAction(const ActionType& actionType,
    const InputEvent& event,
    const InputDevicesState& devicesState)
{
    if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::RESET) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Reset camera position");
        this->resetCamera();
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_START) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewControllerOriginalActions", "Start moving camera");
        this->updateCursorPosition(devicesState);

        return this->startCameraMovement();
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Move camera");
        this->updateCursorPosition(devicesState);

        return this->moveCamera(event);
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_END) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewControllerOriginalActions", "Stop moving camera");
        return this->stopCameraMovement();
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN_TO) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Zoom-in to");

        float zoomSpeed = this->_zoomSpeed;
        this->_zoomSpeed = static_cast<float>(static_cast<double>(zoomSpeed) * event.getMouseWheelDy());
        this->zoomInTo();
        this->_zoomSpeed = zoomSpeed;
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Zoom-in");
        this->zoomIn();
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT_FROM) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Zoom-out from");

        float zoomSpeed = this->_zoomSpeed;
        this->_zoomSpeed = -static_cast<float>(static_cast<double>(zoomSpeed) * event.getMouseWheelDy());
        this->zoomOut();
        this->_zoomSpeed = zoomSpeed;
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Zoom-out");
        this->zoomOut();
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::INCREASE_ZOOM_SPEED) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Increase zoom speed");
        this->_zoomSpeed = 2.f;
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::DECREASE_ZOOM_SPEED) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Decrease zoom speed");
        this->_zoomSpeed = 1.f;
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::TOGGLE_POLES_AXIS_LOCKING) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Toggle poles axis locking");
        this->setLockPolesAxis(!this->_lockPolesAxis);
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::RESIZE_VIEWPORT) == actionType)
    {
        LOG(LogLevel::Debug, "SceneViewController", "Resize scene viewport");
        this->_sceneView->setViewport(glm::ivec2(event.getWidth(), event.getHeight()));
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE) == actionType)
    {
        ChangeRasterizationModeRequest outEvent(OGL::fill);
        this->_appEventsDispatcher.dispatch<ChangeRasterizationModeRequest>(outEvent);
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE) == actionType)
    {
        ChangeRasterizationModeRequest outEvent(OGL::line);
        this->_appEventsDispatcher.dispatch<ChangeRasterizationModeRequest>(outEvent);
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE) == actionType)
    {
        ChangeRasterizationModeRequest outEvent(OGL::point);
        this->_appEventsDispatcher.dispatch<ChangeRasterizationModeRequest>(outEvent);
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::CURSOR_MOVE) == actionType)
    {
        this->updateCursorPosition(devicesState);

        LOG(LogLevel::Debug, "SceneViewController", "Update geographical coordinates of the cursor");

        glm::vec2 lonlat;
        bool onWorld = this->_sceneView->getLonLat(this->_lastMousePosition.x,
                                                   this->_lastMousePosition.y,
                                                   lonlat, this->_zoomGlobePos);
        CursorLocationChangedEvent appEvent(static_cast<std::uint32_t>(this->_lastMousePosition.x),
                                            static_cast<std::uint32_t>(this->_lastMousePosition.y),
                                            lonlat.x, lonlat.y, onWorld, this->_zoomGlobePos);
        this->_appEventsDispatcher.dispatch<CursorLocationChangedEvent>(appEvent);
        return true;
    }

    return false;
}

void SceneViewControllerOriginal::initEvents()
{
	// Configure actions

	// - Cursor move
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::CURSOR_MOVE),
		"Cursor move", "__ ~X|Y");

	// - Camera reset
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::RESET),
		"Reset zoom and location", ">R");  /* keypress: R*/

// - Camera move start
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_START),
		"Start camera movement", ">BUTTON0"); /* buttonpress: left */

// - Camera move
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE),
		"Move camera", "}BUTTON0 ~X|Y"); /* buttonpress: left + move */

// - Camera move end
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_END),
		"Conclude camera movement", "<BUTTON0"); /* buttonrelease: left */

// - Zoom-in
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN_TO),
		"Zoom-in to a location", "__ *Y+"); /* wheel: forward on Y axis */

// - Zoom-in (alternative binding
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN),
		"Zoom-in", "__ }PAGE_UP"); /* keypress: page-up */

// - Zoom-out
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT_FROM),
		"Zoom-out from a location", "__ *Y-"); /* wheel: backward on Y axis */

// - Zoom-in (alternative binding
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT),
		"Zoom-out", "__ }PAGE_DOWN"); /* keypress: page-down */

// - Increase zoom speed
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::INCREASE_ZOOM_SPEED),
		"Increase zoom speed", ">LEFT_CONTROL"); /* keypress: left ctrl */

// - Decrease zoom speed
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::DECREASE_ZOOM_SPEED),
		"Decrease zoom speed", "<LEFT_CONTROL"); /* keyrelease: left ctrl */

// - Toggle poles axis locking
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::TOGGLE_POLES_AXIS_LOCKING),
		"Toggle poles axis locking", "}LEFT_SHIFT >A");

	// - Resize viewport
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::RESIZE_VIEWPORT),
		"Resize scene viewport", "#?");

	// - Use full mode
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE),
		"Use full mode", "}LEFT_SHIFT >F");

	// - Use wireframe mode
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE),
		"Use wireframe mode", "}LEFT_SHIFT >W");

	// - Use dot mode
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE),
		"Use vertex mode", "}LEFT_SHIFT >V");

	this->_possibleActions = { static_cast<ActionType>(SceneViewControllerOriginalActions::RESET),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::CURSOR_MOVE),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_START),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN_TO),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT_FROM),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::INCREASE_ZOOM_SPEED),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::DECREASE_ZOOM_SPEED),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::TOGGLE_POLES_AXIS_LOCKING),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::RESIZE_VIEWPORT),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE), };

	this->_appEventsDispatcher.subscribe<ZoomInRequest>(this);
	this->_appEventsDispatcher.subscribe<ZoomOutRequest>(this);
	this->_appEventsDispatcher.subscribe<CameraResetRequest>(this);
	this->_appEventsDispatcher.subscribe<CameraAlignedWithPolesRequest>(this);
	this->_appEventsDispatcher.subscribe<CameraPolesLockRequest>(this);
	this->_appEventsDispatcher.subscribe<LockViewRequest>(this);
	this->_appEventsDispatcher.subscribe<ChangeRasterizationModeRequest>(this);
}

void SceneViewControllerOriginal::cleanEvents()
{
	this->_appEventsDispatcher.unsubscribe<ZoomInRequest>(this);
	this->_appEventsDispatcher.unsubscribe<ZoomOutRequest>(this);
	this->_appEventsDispatcher.unsubscribe<CameraResetRequest>(this);
	this->_appEventsDispatcher.unsubscribe<CameraAlignedWithPolesRequest>(this);
	this->_appEventsDispatcher.unsubscribe<CameraPolesLockRequest>(this);
	this->_appEventsDispatcher.unsubscribe<LockViewRequest>(this);
	this->_appEventsDispatcher.unsubscribe<ChangeRasterizationModeRequest>(this);

	this->_sceneView = nullptr; // not owning
}

} // namespace SEAScope
