// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "tools/selection_controller.h"

#include <iterator>

#include "core/log.h"
#include "core/input_event.h"
#include "core/input_devices_state.h"
#include "core/granule_metadata.h"

#include "3d/buffer.h"
#include "3d/context.h"
#include "3d/framebuffer.h"
#include "3d/helper.h"
#include "3d/pipeline_parameters.h"
#include "3d/camera.h"
#include "3d/iscene.h"
#include "3d/scene_view.h"
#include "3d/scene_graph.h"
#include "3d/scene_graph_controller.h"
#include "3d/sync.h"
#include "3d/pickable.h"

#include "core/app_events_dispatcher.h"
#include "events/camera_moved_event.h"
#include "events/renderable_added_to_scene_graph_event.h"
#include "events/renderable_removed_from_scene_graph_event.h"
#include "events/cursor_location_changed_event.h"
#include "events/hovered_pickables_changed_event.h"
#include "events/selected_pickables_changed_event.h"
#include "events/renderables_changed_event.h"
#include "events/anchor_added_event.h"
#include "events/anchor_removed_event.h"
#include "events/selection_lock_request.h"
#include "events/picking_update_request.h"
#include "events/rendering_cfg_update_event.h"
#include "events/set_selected_pickables_request.h"

#include "3d/shape_anchor.h"

using namespace SEAScope;

/**
 *
 */
SelectionController::SelectionController(AppEventsDispatcher& appEventsDispatcher)
: BaseInputEventsSubscriber("Selection controller"),
  _appEventsDispatcher(appEventsDispatcher),
  _showPicking(false),
  _pickingUpdate(false),
  _refreshPickedObject(false),
  _asyncRunning(false),
  _selectionLocked(false),
  _pickingFBO(nullptr),
  _pickingPBO(nullptr),
  _sync(nullptr),
  _uniqueId(1), /* Do not start at 0 because i will be associated with the black color, so there would be no way to distinguish between the first picking object and the background (globe) */
  _hasQueuedScreenPosition(false),
  _screenHeight(0)
{
    // Configure actions

    // - Resize framebuffer
    this->defineAction(static_cast<ActionType>(SelectionControllerActions::RESIZE_FRAMEBUFFER),
                       "Resize selector framebuffer", "#?");

    this->defineAction(static_cast<ActionType>(SelectionControllerActions::RECTANGLE_START),
                       "Start area selection", "}LEFT_SHIFT >BUTTON0");

    this->defineAction(static_cast<ActionType>(SelectionControllerActions::RECTANGLE_RESIZE),
                       "Update area selection", "}LEFT_SHIFT }BUTTON0 ~X|Y");

    this->defineAction(static_cast<ActionType>(SelectionControllerActions::RECTANGLE_END),
                       "Conclude area selection", "{LEFT_SHIFT {BUTTON0");

    this->defineAction(static_cast<ActionType>(SelectionControllerActions::SELECT_ALL),
                       "Select all", "}LEFT_CONTROL >A");

    this->defineAction(static_cast<ActionType>(SelectionControllerActions::UPDATE_SELECTION),
                       "Toggle selection for item under cursor", "}LEFT_CONTROL !BUTTON0");

    this->defineAction(static_cast<ActionType>(SelectionControllerActions::EXCLUSIVE_SELECTION),
                       "Toggle selection for item under cursor, exclusively", "!BUTTON0");

    this->defineAction(static_cast<ActionType>(SelectionControllerActions::TOGGLE_PICKING_RENDERING),
                       "Toggle picking rendering", ">P");

    this->_possibleActions = {static_cast<ActionType>(SelectionControllerActions::RESIZE_FRAMEBUFFER),
                              // static_cast<ActionType>(SelectionControllerActions::RECTANGLE_START),
                              static_cast<ActionType>(SelectionControllerActions::SELECT_ALL),
                              static_cast<ActionType>(SelectionControllerActions::UPDATE_SELECTION),
                              static_cast<ActionType>(SelectionControllerActions::EXCLUSIVE_SELECTION),
                              /*static_cast<ActionType>(SelectionControllerActions::TOGGLE_PICKING_RENDERING),*/};

    this->_appEventsDispatcher.subscribe<CameraMovedEvent>(this);
    this->_appEventsDispatcher.subscribe<RenderableAddedToSceneGraphEvent>(this);
    this->_appEventsDispatcher.subscribe<RenderableRemovedFromSceneGraphEvent>(this);
    this->_appEventsDispatcher.subscribe<CursorLocationChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<RenderablesChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<AnchorAddedEvent>(this);
    this->_appEventsDispatcher.subscribe<AnchorRemovedEvent>(this);
    this->_appEventsDispatcher.subscribe<SelectionLockRequest>(this);
    this->_appEventsDispatcher.subscribe<PickingUpdateRequest>(this);
    this->_appEventsDispatcher.subscribe<RenderingCfgUpdateEvent>(this);
    this->_appEventsDispatcher.subscribe<SetSelectedPickablesRequest>(this);
}

/**
 *
 */
SelectionController::~SelectionController()
{
    this->_appEventsDispatcher.unsubscribe<CameraMovedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<RenderableAddedToSceneGraphEvent>(this);
    this->_appEventsDispatcher.unsubscribe<RenderableRemovedFromSceneGraphEvent>(this);
    this->_appEventsDispatcher.unsubscribe<CursorLocationChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<RenderablesChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<AnchorAddedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<AnchorRemovedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<SelectionLockRequest>(this);
    this->_appEventsDispatcher.unsubscribe<PickingUpdateRequest>(this);
    this->_appEventsDispatcher.unsubscribe<RenderingCfgUpdateEvent>(this);
    this->_appEventsDispatcher.unsubscribe<SetSelectedPickablesRequest>(this);

    delete this->_pickingFBO;
    this->_pickingFBO = nullptr;

    delete this->_pickingPBO;
    this->_pickingPBO = nullptr;

    delete this->_sync;
    this->_sync = nullptr;
}

/**
 *
 */
bool SelectionController::performAction(const ActionType& actionType,
                                         const InputEvent& event,
                                         const InputDevicesState& devicesState)
{
    if
        (static_cast<ActionType>(SelectionControllerActions::RESIZE_FRAMEBUFFER) == actionType)
    {
        LOG(LogLevel::Debug, "SelectionController", "Resize framebuffer to {}x{}", event.getWidth(), event.getHeight());
        this->_pickingFBO->resize(static_cast<unsigned int>(event.getWidth()),
                                  static_cast<unsigned int>(event.getHeight()));
        this->_screenHeight = static_cast<std::int32_t>(event.getHeight());

        // Update window size used by the selection rectangle
        this->_rectangle.setWindowSize(event.getWidth(), event.getHeight());

        // update FBO content after a resize
        this->_pickingUpdate = true;

        return true;
    }
    else if
        (static_cast<ActionType>(SelectionControllerActions::RECTANGLE_START) == actionType)
    {
        LOG(LogLevel::Debug, "SelectionController", "Start area selection at {}, {}",
                 devicesState.getMouseX(), devicesState.getMouseY());
        this->_rectangle.setStartCorner(glm::ivec2(devicesState.getMouseX(), devicesState.getMouseY()));
        this->_rectangle.setEndCorner(glm::ivec2(devicesState.getMouseX(), devicesState.getMouseY()));
        this->_rectangle.setVisible(true);

        this->updatePossibleActions({SelectionControllerActions::RECTANGLE_RESIZE,
                                     SelectionControllerActions::RECTANGLE_END},
                                    {SelectionControllerActions::RECTANGLE_START});
        this->_preemptEvents = true;

        return true;
    }
    else if
        (static_cast<ActionType>(SelectionControllerActions::RECTANGLE_RESIZE) == actionType)
    {
        LOG(LogLevel::Debug, "SelectionController", "Extend area selection to {}, {}",
                 devicesState.getMouseX(), devicesState.getMouseY());
        this->_rectangle.setEndCorner(glm::ivec2(devicesState.getMouseX(), devicesState.getMouseY()));
        return true;
    }
    else if
        (static_cast<ActionType>(SelectionControllerActions::RECTANGLE_END) == actionType)
    {
        LOG(LogLevel::Debug, "SelectionController", "Close area selection to {}, {}",
                 devicesState.getMouseX(), devicesState.getMouseY());
        this->_rectangle.setEndCorner(glm::ivec2(devicesState.getMouseX(), devicesState.getMouseY()));

        // TODO

        this->_rectangle.setVisible(false);

        this->updatePossibleActions({SelectionControllerActions::RECTANGLE_START},
                                    {SelectionControllerActions::RECTANGLE_RESIZE,
                                     SelectionControllerActions::RECTANGLE_END});
        this->_preemptEvents = false;
        return true;
    }
    else if
        (static_cast<ActionType>(SelectionControllerActions::SELECT_ALL) == actionType)
    {
        LOG(LogLevel::Debug, "SelectionController", "Select all");
        // TODO
        return true;
    }
    else if
        (static_cast<ActionType>(SelectionControllerActions::UPDATE_SELECTION) == actionType)
    {
        LOG(LogLevel::Debug, "SelectionController", "Update selection");
        this->updateSelection();
        return true;
    }
    else if
        (static_cast<ActionType>(SelectionControllerActions::EXCLUSIVE_SELECTION) == actionType)
    {
        LOG(LogLevel::Debug, "SelectionController", "Exclusive selection");
        this->exclusiveSelection();
        return true;
    }
    else if
        (static_cast<ActionType>(SelectionControllerActions::TOGGLE_PICKING_RENDERING) == actionType)
    {
        LOG(LogLevel::Debug, "SelectionController", "Toggle picking rendering");
        this->_showPicking = ! this->_showPicking;
        return true;
    }

    return false;
}

/**
 *
 */
void SelectionController::updatePossibleActions(const std::set<SelectionControllerActions>& toEnable,
                                                const std::set<SelectionControllerActions>& toDisable)
{
    ActionType actionType;

    for
        (const auto& it : toEnable)
    {
        actionType = static_cast<ActionType>(it);
        this->_possibleActions.insert(actionType);
    }

    for
        (const auto& it : toDisable)
    {
        actionType = static_cast<ActionType>(it);

        auto found = this->_possibleActions.find(actionType);
        if
            (this->_possibleActions.end() != found)
        {
            this->_possibleActions.erase(found);
        }
    }
}


/**
 * TODO: factorize code (duplicated in 3d/scene_renderer.cpp
 */
static void fillPipelineParameters(OGL::Context* ctx, const Camera* camera)
{
    PipelineParameters* pp = ctx->pipelineParameters();

    pp->model = camera->getModelMatrix();
    pp->view = camera->getViewMatrix();
    pp->projection = camera->getProjectionMatrix();
    pp->inverseView = glm::inverse(pp->view);
    pp->inverseProjection = glm::inverse(pp->projection);
    pp->viewProjection = pp->projection * pp->view * pp->model;
    pp->inverseViewProjection = glm::inverse(pp->viewProjection);
    pp->eyePosition = camera->getWorldPosition();
}


/**
 *
 */
bool SelectionController::preRender(OGL::Context* ctx, SceneView* sceneView)
{
    if
        (0 >= this->_screenHeight)
    {
        this->_screenHeight = sceneView->getViewport().y;
    }

    if
        (nullptr == this->_pickingFBO)
    {
        this->_pickingFBO = OGL::Helper::createFramebuffer(ctx,
                                                           static_cast<unsigned int>(sceneView->getViewport().x),
                                                           static_cast<unsigned int>(sceneView->getViewport().y),
                                                           DataType::unsigned_byte);
    }

    if
        (nullptr == this->_pickingPBO)
    {
        // Create a pixel buffer object to transfer a pixel 8bit RGB
        this->_pickingPBO = ctx->createBuffer(OGL::pixel_pack_buffer);
        this->_pickingPBO->create(ctx);
        this->_pickingPBO->allocate(4, nullptr, OGL::stream_read);
    }

    if (nullptr == this->_sync)
    {
        this->_sync = ctx->createSync();
    }

    if
        (true == this->_pickingUpdate)
    {
        this->_pickingFBO->bind();

        ctx->setBlendFunc(OGL::src_alpha, OGL::one_minus_src_alpha);
        ctx->setFlag(OGL::blending, false);
        ctx->setFlag(OGL::depth_test, true);
        ctx->setDepthMask(true);
        ctx->setClearColor(glm::vec4(0.f, 0.f, 0.f, 0.f));
        ctx->clear(OGL::color_buffer | OGL::depth_buffer);

        // Draw all objects with uniform fill color (color picking).
        fillPipelineParameters(ctx, sceneView->getConstCamera());
        PipelineParameters* pp = ctx->pipelineParameters();
        sceneView->getScene()->getNodeFactory()->getSceneGraph()->drawPicking(ctx, pp->viewProjection);

        this->_pickingFBO->unbind();

        this->_pickingUpdate = false;
    }

    if
        (true == this->_showPicking)
    {
        this->_pickingFBO->blit(static_cast<unsigned int>(sceneView->getViewport().x),
                                static_cast<unsigned int>(sceneView->getViewport().y));
    }

    return this->_showPicking;    // prevent nominal rendering if picking colors are shown
}

/**
 *
 */
bool SelectionController::postRender(OGL::Context* ctx, SceneView* sceneView)
{
    if
        (true == this->_asyncRunning)
    {
        Pickable* pickable(nullptr);
        bool ok = this->readPickingObjectAsync(pickable);
        if
            (true == ok)
        {
            this->setHovered(pickable);
        }

        // Trigger queued async picking 
        if
            (true == this->_hasQueuedScreenPosition)
        {
            this->getPickingObjectAsync(this->_queuedScreenPosition);
            this->_hasQueuedScreenPosition = false;
        }
    }

    if
        (true == this->_refreshPickedObject)
    {
        this->getPickingObjectAsync(this->_latestPickedPosition);
        this->_refreshPickedObject = false;
    }

    if
        (false == this->_rectangle.initialized())
    {
        this->_rectangle.initialize(ctx, sceneView->getViewport().x, sceneView->getViewport().y);
    }

    this->_rectangle.draw(ctx, glm::mat4(1.f));

    return false;
}

static glm::vec3 idToColor(std::uint32_t id)
{
    glm::vec3 color;
    color.r = static_cast<float>((id & 0xff)) / 255.f;
    color.g = static_cast<float>(((id >> 8) & 0xff)) / 255.f;
    color.b = static_cast<float>(((id >> 16) & 0xff)) / 255.f;
    return color;
}

/**
 *
 */
bool SelectionController::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(CameraMovedEvent)) == eventType)
    {
        this->_pickingUpdate = true;
        this->_refreshPickedObject = true;
        return true;
    }
    else if
        (std::type_index(typeid(RenderableAddedToSceneGraphEvent)) == eventType)
    {
        const RenderableAddedToSceneGraphEvent& typedEvent = static_cast<const RenderableAddedToSceneGraphEvent&>(event);
        Pickable* pickable = dynamic_cast<Pickable*>(typedEvent.getRenderable());
        if
            (nullptr != pickable)
        {
            std::uint32_t pickableId = this->_uniqueId++;
            pickable->setPickingId(pickableId, idToColor(pickableId));
            this->_pickables[pickableId] = pickable;

            // a RenderablesChangedEvent will probably be emitted and trigger picking update
            this->_pickingUpdate = true;
            this->_refreshPickedObject = true;
        }

        pickable = nullptr;
        return true;
    }
    else if
        (std::type_index(typeid(AnchorAddedEvent)) == eventType)
    {
        ShapeAnchor* anchor = static_cast<const AnchorAddedEvent&>(event).getAnchor();
        std::uint32_t pickableId = this->_uniqueId++;
        anchor->setPickingId(pickableId, idToColor(pickableId));
        this->_pickables[pickableId] = anchor;

        // Force an update because anchor addition is not followed by a RenderablesChangedEvent
        this->_pickingUpdate = true;

        return true;
    }
    else if
        (std::type_index(typeid(AnchorRemovedEvent)) == eventType)
    {
        ShapeAnchor* anchor = static_cast<const AnchorRemovedEvent&>(event).getAnchor();

        this->pickableRemoved(anchor);
        anchor = nullptr;

        // Force an update because anchor removal is not followed by a RenderablesChangedEvent
        this->_pickingUpdate = true;

        return true;
    }
    else if
        (std::type_index(typeid(RenderableRemovedFromSceneGraphEvent)) == eventType)
    {
        const RenderableRemovedFromSceneGraphEvent& typedEvent = static_cast<const RenderableRemovedFromSceneGraphEvent&>(event);
        Pickable* pickable = dynamic_cast<Pickable*>(typedEvent.getRenderable());
        if
            (nullptr != pickable)
        {
            this->pickableRemoved(pickable);

            // a RenderablesChangedEvent will probably be emitted and trigger picking update
            this->_pickingUpdate = true;
            this->_refreshPickedObject = true;
        }

        pickable = nullptr;

        return true;
    }
    else if
        (std::type_index(typeid(CursorLocationChangedEvent)) == eventType)
    {
        const CursorLocationChangedEvent& typedEvent = static_cast<const CursorLocationChangedEvent&>(event);
        glm::ivec2 screenPos(typedEvent.getScreenX(), this->_screenHeight - static_cast<std::int32_t>(typedEvent.getScreenY()));

        if
            (false == typedEvent.getOnWorld())
        {
            // No highlight
            this->setHovered(nullptr);
        }
        else if
            (false == this->_asyncRunning)
        {
            // start async picking
            this->getPickingObjectAsync(screenPos);
        }
        else
        {
            // Save mouse position for later
            this->_queuedScreenPosition = screenPos;
            this->_hasQueuedScreenPosition = true;
        }
    }
    else if
        (std::type_index(typeid(RenderablesChangedEvent)) == eventType)
    {
        const std::map<CollectionId, std::vector<GranuleMetadata> >& renderables = static_cast<const RenderablesChangedEvent&>(event).getRenderables();
        std::vector<Pickable*> toRemove;
        for
            (const auto& pickable: this->_selected)
        {
            auto collectionFound = renderables.find(pickable->renderableId().collection);
            if
                (renderables.end() == collectionFound)
            {
                // Collection not selected for rendering -> remove pickable from selection
                toRemove.push_back(pickable);
                continue;
            }

            bool granuleFound = false;
            for
                (const auto& gMeta : collectionFound->second)
            {
                if
                    (gMeta.granuleId() == pickable->renderableId().granule.second)
                {
                    granuleFound = true;
                    break;
                }
            }

            if
                (false == granuleFound)
            {
                toRemove.push_back(pickable);
            }
        }

        for
            (const auto& obsoletePickable : toRemove)
        {
            this->pickableNotRendered(obsoletePickable);
        }

        // Required for user shapes, otherwise they are not pickable until
        // another event triggers a picking update
        this->_pickingUpdate = true;

        return true;
    }
    else if
        (std::type_index(typeid(SelectionLockRequest)) == eventType)
    {
        bool locked = static_cast<const SelectionLockRequest&>(event).getLocked();
        this->lockSelection(locked);
        return true;
    }
    else if
        (std::type_index(typeid(PickingUpdateRequest)) == eventType)
    {
        this->_pickingUpdate = this->_pickingUpdate || static_cast<const PickingUpdateRequest&>(event).getPickingUpdate();
        this->_refreshPickedObject = this->_refreshPickedObject || static_cast<const PickingUpdateRequest&>(event).getRefreshPickedObject();
        return true;
    }
    else if
        (std::type_index(typeid(RenderingCfgUpdateEvent)) == eventType)
    {
        const RenderingParam& param = static_cast<const RenderingCfgUpdateEvent&>(event).getRenderingParam();
        if
            (RenderingParam::ZINDEX == param)
        {
            this->_pickingUpdate = true;
            this->_refreshPickedObject = true;
        }
        return true;
    }
    else if
        (std::type_index(typeid(SetSelectedPickablesRequest)) == eventType)
    {
        const std::vector<Pickable*> selected = static_cast<const SetSelectedPickablesRequest&>(event).getSelected();
        for
            (auto& pickable: this->_selected)
        {
            pickable->setStyle(StyleState::normal);
        }
        for
            (auto& pickable: selected)
        {
            pickable->setStyle(StyleState::selected);
        }
        this->_selected = selected; // Copy
        SelectedPickablesChangedEvent outEvent(this->_selected);
        this->_appEventsDispatcher.dispatch<SelectedPickablesChangedEvent>(outEvent);
    }

    return false;
}

/**
 *
 */
void SelectionController::lockSelection(bool locked)
{
    this->_selectionLocked = locked;
}

/**
 *
 */
void SelectionController::pickableRemoved(Pickable* pickable)
{
    auto found = this->_pickables.find(pickable->pickingId());
    if
        (this->_pickables.end() != found)
    {
        this->_pickables.erase(found);
    }

    this->pickableNotRendered(pickable);
}

void SelectionController::pickableNotRendered(Pickable* pickable)
{
    if
        (nullptr == pickable)
    {
        return;
    }

    auto hoveredFound = std::find(this->_hovered.begin(), this->_hovered.end(), pickable);
    if
        (this->_hovered.end() != hoveredFound)
    {
        this->_hovered.erase(hoveredFound);

        pickable->getStyle().setHighlighted(false);

        HoveredPickablesChangedEvent hoverEvent(this->_hovered);
        this->_appEventsDispatcher.dispatch<HoveredPickablesChangedEvent>(hoverEvent);
    }

    // Ignore _selectionLocked here because the pickable does not exist anymore
    // (not a simple deselection)

    auto selectedFound = std::find(this->_selected.begin(), this->_selected.end(), pickable);
    if
        (this->_selected.end() != selectedFound)
    {
        pickable->setStyle(StyleState::normal);

        this->_selected.erase(selectedFound);

        SelectedPickablesChangedEvent selectEvent(this->_selected);
        this->_appEventsDispatcher.dispatch<SelectedPickablesChangedEvent>(selectEvent);
    }
}

void SelectionController::getPickingObjectAsync(const glm::ivec2& screenPos)
{
    if
        (false == this->_asyncRunning)
    {
        this->_latestPickedPosition = screenPos;
        this->_pickingPBO->bind();
        this->_pickingFBO->bind();
        this->_pickingFBO->readPixelAsync(screenPos.x, screenPos.y);
        this->_sync->create();
        this->_pickingPBO->unbind();
        this->_pickingFBO->unbind();
        this->_asyncRunning = true;
    }
}

bool SelectionController::readPickingObjectAsync(Pickable*& object)
{
    if
        (false == this->_sync->wait(0))
    {
        return false;
    }

    this->_pickingPBO->bind();

    unsigned char* ptr = static_cast<unsigned char*>(this->_pickingPBO->map(OGL::read_only));

    unsigned int id = 0;
    id |= (static_cast<unsigned int>(ptr[0]));
    id |= (static_cast<unsigned int>(ptr[1])) << 8;
    id |= (static_cast<unsigned int>(ptr[2])) << 16;

    this->_pickingPBO->unmap();
    this->_pickingPBO->unbind();

    auto it = this->_pickables.find(id);
    if
        (this->_pickables.end() == it)
    {
        object = nullptr;
    }
    else
    {
        object = it->second;
    }

    this->_sync->destroy();
    this->_asyncRunning = false;

    ptr = nullptr;

    return true;
}

/**
 *
 */
void SelectionController::setHovered(Pickable* hovered)
{
    bool hoveredChanged(false);

    if
        (nullptr == hovered)
    {
        if
            (0 < this->_hovered.size())
        {
            for
                (auto notHovered : this->_hovered)
            {
                notHovered->getStyle().setHighlighted(false);
            }

            this->_hovered.clear();
            hoveredChanged = true;
        }
    }
    else if
        (1 == this->_hovered.size())
    {
        if
            (hovered != this->_hovered[0])
        {
            if
                (nullptr != this->_hovered[0])
            {
                this->_hovered[0]->getStyle().setHighlighted(false);
            }

            this->_hovered[0] = hovered;

            hovered->getStyle().setHighlighted(true);
            hoveredChanged = true;
        }
    }
    else
    {
        for
            (auto notHovered : this->_hovered)
        {
            notHovered->getStyle().setHighlighted(false);
        }
        this->_hovered.clear();
        this->_hovered.push_back(hovered);

        hovered->getStyle().setHighlighted(true);
        hoveredChanged = true;
    }

    if
        (true == hoveredChanged)
    {
        HoveredPickablesChangedEvent event(this->_hovered);
        this->_appEventsDispatcher.dispatch<HoveredPickablesChangedEvent>(event);
    }
}

/**
 *
 */
void SelectionController::updateSelection()
{
    if
        (true == this->_selectionLocked)
    {
        return;
    }

    bool changed(false);

    for
        (auto& pickable: this->_hovered)
    {
        if
            (pickable->flag(RenderableFlags::anchor))
        {
            // Anchors cannot be selected
            continue;
        }

        auto found = std::find(this->_selected.begin(), this->_selected.end(), pickable);
        if
            (this->_selected.end() == found)
        {
            this->_selected.push_back(pickable);
            pickable->setStyle(StyleState::selected);
        }
        else
        {
            this->_selected.erase(found);
            (*found)->setStyle(StyleState::normal);
        }
        changed = true;
    }

    if
        (true == changed)
    {
        SelectedPickablesChangedEvent event(this->_selected);
        this->_appEventsDispatcher.dispatch<SelectedPickablesChangedEvent>(event);
    }
}

/**
 *
 */
void SelectionController::exclusiveSelection()
{
    if
        (true == this->_selectionLocked)
    {
        return;
    }

    bool changed(false);
    std::vector<Pickable*> newSelection;

    if
        (this->_hovered.size() != this->_selected.size())
    {
        // useful to detect clicks on an empty space
        changed = true;
    }

    for
        (auto& selected: this->_selected)
    {
        selected->setStyle(StyleState::normal);
    }

    for
        (auto& hovered: this->_hovered)
    {
        auto found = std::find(this->_selected.begin(), this->_selected.end(), hovered);
        if
            (this->_selected.end() == found)
        {
            // Only include hovered items that were not already selected
            newSelection.push_back(hovered);
            hovered->setStyle(StyleState::selected);
        }

        changed = true;
    }

    if
        (true == changed)
    {
        this->_selected = newSelection;
        SelectedPickablesChangedEvent event(this->_selected);
        this->_appEventsDispatcher.dispatch<SelectedPickablesChangedEvent>(event);
    }
}
