// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "tools/window_controller.h"


#include "core/log.h"
#include "core/input_event.h"
#include "core/input_devices_state.h"
#include "core/interfaces/iwindow.h"
#include "core/app_events_dispatcher.h"

#include "events/enable_shape_creation_ready_event.h"
#include "events/disable_shape_creation_request.h"

using namespace SEAScope;

/**
 *
 */
WindowController::WindowController(AppEventsDispatcher& appEventsDispatcher)
: BaseInputEventsSubscriber("Application controller"),
  _appEventsDispatcher(appEventsDispatcher),
  _window(nullptr)
{
    // Configure actions

    // - Quit application
    this->defineAction(static_cast<ActionType>(WindowControllerActions::QUIT),
                       "Quit application", "}LEFT_CONTROL }Q");  /* keypress: CTRL + Q*/

    this->_possibleActions = {static_cast<ActionType>(WindowControllerActions::QUIT),};

    this->_appEventsDispatcher.subscribe<EnableShapeCreationReadyEvent>(this);
    this->_appEventsDispatcher.subscribe<DisableShapeCreationRequest>(this);
}

/**
 *
 */
WindowController::~WindowController()
{
    this->_appEventsDispatcher.unsubscribe<EnableShapeCreationReadyEvent>(this);
    this->_appEventsDispatcher.unsubscribe<DisableShapeCreationRequest>(this);

    this->_window = nullptr; // not owning
}

/**
 *
 */
bool WindowController::isActive() const
{
    return this->_active && (nullptr != this->_window);
}

/**
 *
 */
bool WindowController::performAction(const ActionType& actionType,
                                          const InputEvent& /*event*/,
                                          const InputDevicesState& /*devicesState*/)
{
    if
        (static_cast<ActionType>(WindowControllerActions::QUIT) == actionType)
    {
        LOG(LogLevel::Debug, "WindowController", "Quit application");
        this->_window->close();
        return true;
    }

    return false;
}

/**
 *
 */
void WindowController::setWindow(IWindow* window)
{
    this->_window = window;
}

/**
 *
 */
bool WindowController::handleAppEvent(std::type_index eventType, const AppEvent& /*event*/)
{
    if
        (std::type_index(typeid(EnableShapeCreationReadyEvent)) == eventType)
    {
        this->_window->setCursor(IWindow::MouseCursor::CROSSHAIR);
        return true;
    }
    else if
        (std::type_index(typeid(DisableShapeCreationRequest)) == eventType)
    {
        this->_window->setCursor(IWindow::MouseCursor::ARROW);
        return true;
    }

    return false;
}
