// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_add_property_form.h"
#include "ui/nanogui/nanogui_textfield.h"

using namespace SEAScope;

/**
 *
 */
NanoguiAddPropertyForm::NanoguiAddPropertyForm(nanogui::Widget* parent, const std::set<std::string>& validValues)
: nanogui::Widget(parent),
  _nameField(nullptr),
  _okButton(nullptr),
  _nokButton(nullptr),
  _validateCallback(nullptr),
  _okCallback(nullptr),
  _nokCallback(nullptr)
{
    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                              nanogui::Alignment::Middle,
                                              0, 3));

    nanogui::Widget* spacer = new nanogui::Widget(this);
    spacer->setFixedWidth(10);
    spacer->setWidth(10);

    nanogui::Widget* buttonsBlock = new nanogui::Widget(this);
    buttonsBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                   nanogui::Alignment::Middle,
                                                   0, 0));

    for
        (const auto& value: validValues)
    {
        nanogui::Button* button = new nanogui::Button(buttonsBlock, value);
        button->setFontSize(16);
        button->setFixedSize(Eigen::Vector2i(26, 18));
        button->setSize(Eigen::Vector2i(26, 18));
        button->setFlags(nanogui::Button::Flags::RadioButton|nanogui::Button::Flags::ToggleButton);
        button->setChangeCallback([this] (bool /*pushed*/) {
            bool ok(false);

            if
                (nullptr != this->_validateCallback)
            {
                std::string name;
                this->getName(name);

                std::string nodeType;
                this->getType(nodeType);

                ok = this->_validateCallback(nodeType, name);
            }

            this->_okButton->setEnabled(ok);
        });

        this->_buttons.push_back(button);
    }

    for
        (auto& btn : this->_buttons)
    {
        btn->setButtonGroup(this->_buttons);
    }

    this->_nameField = new NanoguiTextField(this);
    this->_nameField->setFixedWidth(170);
    this->_nameField->setWidth(170);
    this->_nameField->setCallback([this] (const std::string& name) -> bool{
        bool ok(false);

        if
            (nullptr != this->_validateCallback)
        {
            std::string nodeType;
            this->getType(nodeType);
            ok = this->_validateCallback(nodeType, name);
        }

        this->_okButton->setEnabled(ok);

        return ok;
    });

    this->_okButton = new nanogui::Button(this, "", ENTYPO_ICON_PLUS);
    this->_okButton->setFontSize(16);
    this->_okButton->setTextColor(nanogui::Color(30, 255, 30, 255));
    this->_okButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_okButton->setSize(Eigen::Vector2i(18, 18));
    this->_okButton->setCallback([this] () {
        if
            (nullptr != this->_okCallback)
        {
            this->_okCallback();
        }
    });

    this->_nokButton = new nanogui::Button(this, "", ENTYPO_ICON_CROSS);
    this->_nokButton->setFontSize(16);
    this->_nokButton->setTextColor(nanogui::Color(255, 30, 30, 255));
    this->_nokButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_nokButton->setSize(Eigen::Vector2i(18, 18));
    this->_nokButton->setCallback([this] () {
        if
            (nullptr != this->_nokCallback)
        {
            this->_nokCallback();
        }
    });

    this->_okButton->setEnabled(false);
}

/**
 *
 */
NanoguiAddPropertyForm::~NanoguiAddPropertyForm()
{
    this->_buttons.clear();
}

/**
 *
 */
bool NanoguiAddPropertyForm::setValidateCallback(const std::function<bool(const std::string&, const std::string&)>& callback)
{
    if
        (nullptr != this->_validateCallback)
    {
        return false;
    }

    this->_validateCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiAddPropertyForm::setOkCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_okCallback)
    {
        return false;
    }

    this->_okCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiAddPropertyForm::setNokCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_nokCallback)
    {
        return false;
    }

    this->_nokCallback = callback;
    return true;
}

/**
 *
 */
void NanoguiAddPropertyForm::reset()
{
    for
        (auto& btn : this->_buttons)
    {
        btn->setPushed(false);
    }

    this->_nameField->setValue("");
    this->_okButton->setEnabled(false);
}

/**
 *
 */
bool NanoguiAddPropertyForm::getName(std::string& result) const
{
    result = this->_nameField->value();
    return true;
}

/**
 *
 */
bool NanoguiAddPropertyForm::getType(std::string& result) const
{
    result = "";
    for
        (const auto& btn : this->_buttons)
    {
        if
            (true == btn->pushed())
        {
            result = btn->caption();
            return true;
        }
    }
    return false;
}
