// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_anchored.h"

using namespace SEAScope;

uint8_t NanoguiAnchored::ON_LEFT = (1 << 0);
uint8_t NanoguiAnchored::ON_RIGHT = (1 << 1);
uint8_t NanoguiAnchored::ON_TOP = (1 << 2);
uint8_t NanoguiAnchored::ON_BOTTOM = (1 << 3);

/**
 *
 */
NanoguiAnchored::NanoguiAnchored(nanogui::Widget* parent, nanogui::Widget* anchorIn)
: nanogui::Widget(parent), anchor(anchorIn), layoutUpdateRequested(false),
  autoSize(true), offset(10), useDirectionalOffset(false), leftOffset(10),
  topOffset(10), rightOffset(10), bottomOffset(10), placement(ON_TOP),
  callback(nullptr)
{

};

/**
 *
 */
NanoguiAnchored::~NanoguiAnchored()
{

};

/**
 *
 */
NanoguiAnchored* NanoguiAnchored::create(nanogui::Widget* parent)
{
	nanogui::Widget* widget;	//
	nanogui::Screen* screen;	//
	NanoguiAnchored* result;	//

	widget = parent;
	while
		(true)
	{
		if
			(nullptr == widget)
		{
			throw std::runtime_error("Nanogui screen not found");
		}

		screen = dynamic_cast<nanogui::Screen*>(widget);
		if
			(nullptr != screen)
		{
			break;
		}
		widget = widget->parent();
	}

	result = new NanoguiAnchored(screen, parent);
	return result;
};

/**
 *
 */
void NanoguiAnchored::refreshRelativePlacement()
{
    std::uint16_t dirOffset(0);
	Eigen::Vector2i pos;	//

	pos = this->anchor->absolutePosition();
	pos += Eigen::Vector2i( 0.5 * (this->anchor->size().x() - this->mSize.x())
						  , 0.5 * (this->anchor->size().y() - this->mSize.y()));

	if
		(0 < (this->placement & ON_LEFT))
	{
        dirOffset = this->useDirectionalOffset ? this->leftOffset:this->offset;
		pos -= Eigen::Vector2i(dirOffset + 0.5 * (this->mSize.x() + this->anchor->size().x()), 0);
	}
	else if
		(0 < (this->placement & ON_RIGHT))
	{
        dirOffset = this->useDirectionalOffset ? this->rightOffset:this->offset;
		pos += Eigen::Vector2i(dirOffset + 0.5 * (this->mSize.x() + this->anchor->size().x()), 0);
	}

	if
		(0 < (this->placement & ON_TOP))
	{
        dirOffset = this->useDirectionalOffset ? this->topOffset:this->offset;
		pos -= Eigen::Vector2i(0, dirOffset + 0.5 * (this->mSize.y() + this->anchor->size().y()));
	}
	else if
		(0 < (this->placement & ON_BOTTOM))
	{
        dirOffset = this->useDirectionalOffset ? this->bottomOffset:this->offset;
		pos += Eigen::Vector2i(0, dirOffset + 0.5 * (this->mSize.y() + this->anchor->size().y()));
	}
    this->mPos = pos;
};

/**
 *
 */
void NanoguiAnchored::draw(NVGcontext* ctx)
{
	Eigen::Vector2i size;	//

	if
		(false == this->mVisible)
	{
		return;
	}

	if
		(true == this->layoutUpdateRequested)
	{
		if
			(true == this->autoSize)
		{
			size = this->preferredSize(ctx);
			this->mSize = size;
		}
		this->performLayout(ctx);
		this->layoutUpdateRequested = false;
	}
	this->refreshRelativePlacement();

	int cr = this->mTheme->mWindowCornerRadius;
	/* Draw window */
    nvgBeginPath(ctx);
    nvgRoundedRect(ctx,
                   static_cast<float>(this->mPos.x()),
                   static_cast<float>(this->mPos.y()),
                   static_cast<float>(this->mSize.x()),
                   static_cast<float>(this->mSize.y()),
                   static_cast<float>(cr));

    nvgFillColor(ctx, /*this->mTheme->mWindowPopup*/nanogui::Color(0,200));
    nvgFill(ctx);


	nanogui::Widget::draw(ctx);
};

/**
 *
 */
bool NanoguiAnchored::setMouseEnterCallback(const std::function<void(bool)>& f)
{
	this->callback = f;
	return true;
};

/**
 *
 */
bool NanoguiAnchored::mouseEnterEvent(const Eigen::Vector2i& p, bool enter)
{
	bool ok;	//

	ok = nanogui::Widget::mouseEnterEvent(p, enter);
	LOG( LogLevel::Debug, "Anchored", "mouse enter");
	if
		(nullptr != this->callback)
	{
		this->callback(enter);
	}
	return ok;
};

/**
 *
 */
bool NanoguiAnchored::mouseButtonEvent(const Eigen::Vector2i& p, int button, bool pressed, int mods)
{
    nanogui::Widget::mouseButtonEvent(p, button, pressed, mods);
    return true;
};

/**
 *
 */
bool NanoguiAnchored::mouseDragEvent(const Eigen::Vector2i& p, const Eigen::Vector2i& rel, int button, int mods)
{
    nanogui::Widget::mouseDragEvent(p, rel, button, mods);
    return true;
};

/**
 *
 */
bool NanoguiAnchored::scrollEvent(const Eigen::Vector2i& p, const Eigen::Vector2f& rel)
{
    nanogui::Widget::scrollEvent(p, rel);
    return true;
};
