// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_annotation_collection_selector.h"

#include <nanogui/entypo.h>
#include "ui/nanogui/nanogui_icon.h"
#include "ui/nanogui/nanogui_combobox.h"

using namespace SEAScope;


/**
 *
 */
NanoguiAnnotationCollectionSelector::NanoguiAnnotationCollectionSelector(nanogui::Widget* parent, int icon, const std::string& typeLabel)
: nanogui::Widget(parent),
 _icon(nullptr),
 _combobox(nullptr),
 _addButton(nullptr),
 _useLabel(nullptr),
 _defaultButton(nullptr),
 _customButton(nullptr),
 _collectionLabel(nullptr),
 _changeCallback(nullptr),
 _dropdownCallback(nullptr),
 _addCallback(nullptr),
 _reflowNeeded(true),
 _defaultValue("default")
{
    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                           nanogui::Alignment::Middle,
                                           2, 3));
    this->_icon = new NanoguiIcon(this, icon);
    this->_icon->setFontSize(18);
    this->_icon->setFixedSize(Eigen::Vector2i(20, 20));
    this->_icon->setTooltip(typeLabel);
    this->_combobox = new NanoguiComboBox(this);
    this->_combobox->setFixedSize(Eigen::Vector2i(241, 20));
    this->_addButton = new nanogui::Button(this, "", ENTYPO_ICON_PLUS);
    this->_addButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_addButton->setFontSize(16);
    this->_addButton->setTooltip("Create a new " + typeLabel + " collection");
    this->_useLabel = new nanogui::Label(this, "use");
    this->_defaultButton = new nanogui::Button(this, "default");
    this->_defaultButton->setFlags(nanogui::Button::Flags::RadioButton);
    this->_defaultButton->setFontSize(16);
    this->_defaultButton->setFixedSize(Eigen::Vector2i(50, 20));
    this->_customButton = new nanogui::Button(this, "custom");
    this->_customButton->setFlags(nanogui::Button::Flags::RadioButton);
    this->_customButton->setFontSize(16);
    this->_customButton->setFixedSize(Eigen::Vector2i(50, 20));
    this->_collectionLabel = new nanogui::Label(this, "collection");

    this->_modeButtons.push_back(this->_defaultButton);
    this->_modeButtons.push_back(this->_customButton);
    this->_defaultButton->setButtonGroup(this->_modeButtons);
    this->_customButton->setButtonGroup(this->_modeButtons);

    this->_combobox->setChangeCallback([this] (const std::string& value) {
        if
            (nullptr != this->_changeCallback)
        {
            this->_changeCallback(false, value);
        }
    });

    this->_combobox->setDropDownCallback([this] () {
        if
            (nullptr != this->_dropdownCallback)
        {
            this->_dropdownCallback();
        }
    });

    this->_addButton->setCallback([this] () {
        if
            (nullptr != this->_addCallback)
        {
            this->_addCallback();
        }
    });

    this->_defaultButton->setCallback([this] () {
        this->useDefault();
    });

    this->_customButton->setCallback([this] () {
        this->useCustom();
    });

    // Select default
    this->_defaultButton->setPushed(true);
    this->useDefault();
}

/**
 *
 */
NanoguiAnnotationCollectionSelector::~NanoguiAnnotationCollectionSelector()
{
    this->_modeButtons.clear();

    this->_icon = nullptr;
    this->_combobox = nullptr;
    this->_addButton = nullptr;
    this->_useLabel = nullptr;
    this->_defaultButton = nullptr;
    this->_customButton = nullptr;
    this->_collectionLabel = nullptr;

    this->_changeCallback = nullptr;
    this->_dropdownCallback = nullptr;
    this->_addCallback = nullptr;
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::draw(NVGcontext* ctx)
{
    if
        (true == this->_reflowNeeded)
    {
        this->performLayout(ctx);
        this->_reflowNeeded = false;
    }

    nanogui::Widget::draw(ctx);
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::allowNewCollection(bool active)
{
    this->_addButton->setEnabled(active);
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::useDefault()
{
    if
        (false == this->_combobox->visible())
    {
        // Already in default mode
        return;
    }

    this->_combobox->hideDropDown();
    this->_combobox->setEnabled(false);
    this->_combobox->setVisible(false);

    this->_addButton->setVisible(false);

    this->_customButton->setEnabled(true);
    this->_customButton->setVisible(true);
    this->_useLabel->setVisible(true);
    this->_collectionLabel->setVisible(true);

    if
        (nullptr != this->_changeCallback)
    {
        this->_changeCallback(true, this->_defaultValue);
    }

    this->_reflowNeeded = true;
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::useCustom()
{
    if
        (true == this->_combobox->visible())
    {
        // Already in custom mode
        return;
    }

    this->_useLabel->setVisible(false);
    this->_customButton->setEnabled(false);
    this->_customButton->setVisible(false);
    this->_collectionLabel->setVisible(false);

    this->_combobox->setEnabled(true);
    this->_combobox->setVisible(true);

    this->_addButton->setVisible(true);

    if
        (nullptr != this->_changeCallback)
    {
        this->_changeCallback(false, this->_combobox->getSelectedValue());
    }

    this->_reflowNeeded = true;
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::hideDropDown()
{
    this->_combobox->hideDropDown();
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::enable()
{
    this->_combobox->enable();
    this->_defaultButton->setEnabled(true);
    this->_customButton->setEnabled(true);
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::disable()
{
    this->_combobox->disable();
    this->_defaultButton->setEnabled(false);
    this->_customButton->setEnabled(false);
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::setChoices(const std::vector<std::string>& values)
{
    this->_combobox->setItems(values);
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::choose(const std::string& value)
{
    this->_combobox->setSelectedValue(value);
}

/**
 *
 */
const std::string& NanoguiAnnotationCollectionSelector::getChoice() const
{
    if
        (true == this->_defaultButton->pushed())
    {
        return this->_defaultValue;
    }
    return this->_combobox->getSelectedValue();
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::setChangeCallback(const std::function<void(bool, const std::string&)>& callback)
{
    this->_changeCallback = callback;
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::setDropDownCallback(const std::function<void()>& callback)
{
    this->_dropdownCallback = callback;
}

/**
 *
 */
void NanoguiAnnotationCollectionSelector::setAddCallback(const std::function<void()>& callback)
{
    this->_addCallback = callback;
}
