// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_annotation_dialog.h"
#include "ui/nanogui/nanogui_properties_editor.h"
#include "ui/nanogui/nanogui_attribute.h"
#include "ui/nanogui/nanogui_checkbox.h"
#include "ui/nanogui/nanogui_anchored.h"
#include "ui/nanogui/nanogui_path_selector.h"
#include "ui/nanogui/nanogui_events_bubbling.h"
#include "core/measurements.h"
#include "core/data_node.h"
#include "core/path.h"
#include "core/log.h"
#include "core/time_utils.h"
#include "core/granule_info.h"

#include "ui/nanogui/nanogui_time_coverage_editor.h"
#include "ui/nanogui/nanogui_image_view.h"


using namespace SEAScope;

/**
 * Constructor.
 */
NanoguiAnnotationDialog::NanoguiAnnotationDialog(nanogui::Widget* parent)
: nanogui::Widget(parent),
  collectionAttr(nullptr),
  variableAttr(nullptr),
  spatialCoverageAttr(nullptr),
  _editor(nullptr),
  btns(nullptr),
  extractBtn(nullptr),
  delBtn(nullptr),
  editBtn(nullptr),
  saveBtn(nullptr),
  extractCallback(nullptr),
  deleteCallback(nullptr),
  editCallback(nullptr),
  _timeCoverageEditor(nullptr),
  _saveWidget(nullptr),
  _includeGranulesListCheckbox(nullptr),
  _pathSelector(nullptr),
  _saveDataButton(nullptr),
  _savePropertiesCallback(nullptr),
  _updateTimeCoverageCallback(nullptr),
  _saveCallback(nullptr)
{
    nanogui::Widget* attrs(nullptr); //
    nanogui::Widget* upperWidget(nullptr);    //
    nanogui::Widget* spacer(nullptr);   //

    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                           nanogui::Alignment::Fill,
                                           0, 6));


    upperWidget = new nanogui::Widget(this);
    upperWidget->setFixedWidth(360);
    upperWidget->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                  nanogui::Alignment::Minimum,
                                                  0, 0));
    this->initializeSavePanel(upperWidget);

    attrs = new nanogui::Widget(upperWidget);
    attrs->setFixedWidth(336);

    // Buttons
    this->btns = new nanogui::Widget(upperWidget);
    this->btns->setFixedWidth(24);
    this->btns->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                 nanogui::Alignment::Middle,
                                                 0, 2));
    spacer = new nanogui::Widget(this->btns);
    spacer->setFixedHeight(13);

    this->editBtn = new nanogui::Button(this->btns, "", ENTYPO_ICON_EDIT);
    this->editBtn->setFlags(nanogui::Button::ToggleButton);
    this->editBtn->setFixedSize(Eigen::Vector2i(20, 20));
    this->editBtn->setTooltip("Edit this shape.");

    this->saveBtn = new nanogui::Button(this->btns, "", ENTYPO_ICON_SAVE);
    this->saveBtn->setFixedSize(Eigen::Vector2i(20, 20));
    this->saveBtn->setTooltip("Save this shape as JSON.");

	this->extractBtn = new nanogui::Button(this->btns, "", ENTYPO_ICON_PUBLISH);
	this->extractBtn->setFixedSize(Eigen::Vector2i(20, 20));
	this->extractBtn->setTooltip("Extract granule data intersecting this shape.");

    spacer = new nanogui::Widget(this->btns);
    spacer->setFixedHeight(6);

	this->delBtn = new nanogui::Button(this->btns, "", ENTYPO_ICON_TRASH);
	this->delBtn->setFixedSize(Eigen::Vector2i(20, 20));
	this->delBtn->setTooltip("Delete this shape.");

    spacer = new nanogui::Widget(this->btns);
    spacer->setFixedHeight(13);

    // Attributes
	attrs->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
										   nanogui::Alignment::Minimum,
										   4, 2));

    this->_timeCoverageEditor = new NanoguiTimeCoverageEditor(attrs);
    this->spatialCoverageAttr = new NanoguiAttribute(attrs);
    this->spatialCoverageAttr->setName("Spatial coverage");
    this->spatialCoverageAttr->setValue("_");
	this->collectionAttr = new NanoguiAttribute(attrs);
	this->collectionAttr->nameLbl->setFixedWidth(354);
	this->collectionAttr->valueLbl->setFixedWidth(344);
	this->collectionAttr->setName("Collection");
	this->variableAttr = new NanoguiAttribute(attrs);
	this->variableAttr->nameLbl->setFixedWidth(354);
	this->variableAttr->valueLbl->setFixedWidth(344);
	this->variableAttr->setName("Variable");

    // Properties
    this->_editor = new NanoguiPropertiesEditor(this);
    this->_editor->resize(Eigen::Vector2i(360, 200));

    // Callbacks
    this->_timeCoverageEditor->setCallback([this] (const std::string& startStr, const std::string& stopStr) {
        if
            (nullptr == this->_updateTimeCoverageCallback)
        {
            return;
        }

        time_t tStart;	//
        timestamp_t timestampStart;	//
        std::tm tsStart = {};	//

        strptime(startStr.c_str(), "%Y-%m-%d %H:%M:%S", &tsStart);
        tStart = std::mktime(&tsStart);
        timestampStart = static_cast<timestamp_t>(tStart);

        time_t tStop;	//
        timestamp_t timestampStop;	//
        std::tm tsStop = {};	//

        strptime(stopStr.c_str(), "%Y-%m-%d %H:%M:%S", &tsStop);
        tStop = std::mktime(&tsStop);
        timestampStop = static_cast<timestamp_t>(tStop);

        this->_updateTimeCoverageCallback(this->target,
                                          1000 * static_cast<double>(timestampStart),
                                          1000 * static_cast<double>(timestampStop));
    });

    this->_editor->setSaveCallback([this] (const DictNode& properties) {
        if
            (nullptr != this->_savePropertiesCallback)
        {
            this->_savePropertiesCallback(this->target, properties);
        }
    });

	this->extractBtn->setCallback([this]() {
		if
			(nullptr != this->extractCallback)
		{
			this->extractCallback(this->target);
		}
	});

	this->delBtn->setCallback([this]() {
		if
			(nullptr != this->deleteCallback)
		{
			this->deleteCallback(this->target);
		}
	});

    this->editBtn->setChangeCallback([this] (bool pushed) {
        this->enableEdition(pushed, false);
    });

    this->saveBtn->setCallback([this]() {
        if
            (nullptr != this->_saveCallback)
        {
            this->_saveWidget->setVisible(true);
        }
    });

    this->setVisible(false);
};

/**
 * Destructor.
 */
NanoguiAnnotationDialog::~NanoguiAnnotationDialog()
{
    this->_exportFormatButtons.clear();

    // Deleted by nanogui
    this->_saveWidget = nullptr;
    this->_pathSelector = nullptr;
    this->_saveDataButton = nullptr;
};

/**
 *
 */
void NanoguiAnnotationDialog::initializeSavePanel(nanogui::Widget* parent)
{
    nanogui::Widget* container(nullptr); //
    nanogui::Widget* saveButtons(nullptr);  //
    nanogui::Button* saveNokButton(nullptr);    //
    nanogui::Widget* spacer(nullptr);   //
    nanogui::Widget* formatBlock(nullptr);  //
    nanogui::Widget* pathContainer(nullptr);    //
    nanogui::Widget* innerWidget(nullptr);  //

    this->_saveWidget = NanoguiAnchored::create(parent);
    this->_saveWidget->placement = NanoguiAnchored::ON_RIGHT;
    this->_saveWidget->useDirectionalOffset = true;
    this->_saveWidget->rightOffset = 10;
    this->_saveWidget->setFixedWidth(360);
    this->_saveWidget->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                        nanogui::Alignment::Minimum,
                                                        2, 0));

    new nanogui::Label(this->_saveWidget, "Save annotation as...", "sans-bold", 18);

    innerWidget = new nanogui::Widget(this->_saveWidget);
    innerWidget->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                  nanogui::Alignment::Minimum,
                                                  5, 0));
    innerWidget->setFixedWidth(356);

    // Format
    new nanogui::Label(innerWidget, "Format", "sans-bold", 16);

    container = new nanogui::Widget(innerWidget);
    container->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                nanogui::Alignment::Middle,
                                                0, 0));
    container->setFixedWidth(346);

    formatBlock = new nanogui::Widget(container);
    formatBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                  nanogui::Alignment::Middle,
                                                  0, 0));
    this->_exportFormatButtons.reserve(2);

    this->_exportFormatButtons.emplace_back(new nanogui::Button(formatBlock, "JSON"));
    this->_exportFormatButtons[0]->setFixedSize(Eigen::Vector2i(50, 20));
    this->_exportFormatButtons[0]->setFontSize(16);
    this->_exportFormatButtons[0]->setFlags(nanogui::Button::RadioButton);

    this->_exportFormatButtons.emplace_back(new nanogui::Button(formatBlock, "WKT"));
    this->_exportFormatButtons[1]->setFixedSize(Eigen::Vector2i(50, 20));
    this->_exportFormatButtons[1]->setFontSize(16);
    this->_exportFormatButtons[1]->setFlags(nanogui::Button::RadioButton);

    /*
    this->_exportFormatButtons[2] = new nanogui::Button(formatBlock, "KML");
    this->_exportFormatButtons[2]->setFixedSize(Eigen::Vector2i(50, 20));
    this->_exportFormatButtons[2]->setFontSize(16);
    this->_exportFormatButtons[2]->setFlags(nanogui::Button::RadioButton);
    this->_exportFormatButtons[2]->setButtonGroup(this->_exportFormatButtons);
    */

    this->_exportFormatButtons[0]->setButtonGroup(this->_exportFormatButtons);
    this->_exportFormatButtons[1]->setButtonGroup(this->_exportFormatButtons);

    this->_exportFormatButtons[0]->setPushed(true);

   spacer = new nanogui::Widget(innerWidget);
   spacer->setFixedHeight(10);

    // Options
    new nanogui::Label(innerWidget, "Options", "sans-bold", 16);

    container = new nanogui::Widget(innerWidget);
    container->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                nanogui::Alignment::Middle,
                                                0, 0));
    container->setFixedWidth(346);
    this->_includeGranulesListCheckbox = new NanoguiCheckbox(container, "Include list of intersecting granules");

    // -- Lisitng intersecting granules is only available with the JSON format
    this->_exportFormatButtons[0]->setChangeCallback([this] (bool pushed) {
        this->_includeGranulesListCheckbox->setVisible(pushed);
    });

    // Path
    pathContainer = new nanogui::Widget(innerWidget);
    pathContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                    nanogui::Alignment::Minimum,
                                                    0, 2));
    new nanogui::Label(pathContainer, "Output file", "sans-bold", 16);

    this->_pathSelector = new NanoguiPathSelector(pathContainer, PathSelectorType::save_file);
    this->_pathSelector->resize(Eigen::Vector2i(346, 40));
    this->_pathSelector->setChangeCallback([this] (const std::string&) {
        const std::string& path = this->_pathSelector->getPath();
        this->_saveDataButton->setEnabled( (0 < path.size()) && (false == Path::dirExists(path)));
    });

    // Bottom bar
    container = new nanogui::Widget(this->_saveWidget);
    container->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                nanogui::Alignment::Maximum,
                                                0, 0));
    container->setFixedSize(Eigen::Vector2i(356, 20));

    saveButtons = new nanogui::Widget(container);
    saveButtons->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                  nanogui::Alignment::Middle,
                                                  0, 2));

    this->_saveDataButton = new nanogui::Button(saveButtons, "Save");
    this->_saveDataButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_saveDataButton->setFontSize(16);
    this->_saveDataButton->setEnabled(false);
    this->_saveDataButton->setCallback([this] () {
        const std::string& path = this->_pathSelector->getPath();
        AnnotationFormat exportFormat(AnnotationFormat::NONE);

        if
            (true == this->_exportFormatButtons[0]->pushed())
        {
            exportFormat = AnnotationFormat::JSON;
        }
        else if
            (true == this->_exportFormatButtons[1]->pushed())
        {
            exportFormat = AnnotationFormat::WKT;
        }
        /*
        else if
            (true == this->_exportFormatButtons[2]->pushed())
        {
            exportFormat = AnnotationFormat::KML;
        }
        */
        else
        {
            LOG(LogLevel::Error, "NanoguiAnnotationDialog",
                     "No known export format selected");
            return;
        }

        if
            (nullptr != this->_saveCallback)
        {
            bool includeGranulesList {false};
            includeGranulesList = (this->_includeGranulesListCheckbox->visible() && this->_includeGranulesListCheckbox->checked());
            this->_saveCallback(this->target, path, exportFormat, includeGranulesList);
        }

        this->_pathSelector->hideDropDown();
        this->_saveWidget->setVisible(false);
    });

    saveNokButton = new nanogui::Button(saveButtons, "Cancel");
    saveNokButton->setFixedSize(Eigen::Vector2i(60, 18));
    saveNokButton->setFontSize(16);
    saveNokButton->setCallback([this] () {
        this->_pathSelector->hideDropDown();
        this->_saveWidget->setVisible(false);
    });

    this->_saveWidget->setVisible(false);
}

/**
 *
 */
void NanoguiAnnotationDialog::enableEdition(bool enabled, bool silent)
{
    this->editBtn->setPushed(enabled);
    this->extractBtn->setEnabled(!enabled);
    this->delBtn->setEnabled(!enabled);
    this->_editor->setPreventImageViewer(enabled);
    if
        ((false == silent) && (nullptr != this->editCallback))
    {
        this->editCallback(this->target, enabled);
    }
}

/**
 *
 */
bool NanoguiAnnotationDialog::enable(const RenderableId& rId, const GranuleInfo& info, DictNode&& properties, double start, double stop)
{
    LOG(LogLevel::Debug, "Annotation dialog", "enable");

    if
        (  (rId.granule.first == this->target.granule.first)
        && (rId.granule.second == this->target.granule.second)
        && true == this->visible())
    {
        return false;
    }

    // Collection
	if
		(0 < info.collection.label.size())
	{
		this->collectionAttr->setValue(info.collection.label);
		this->collectionAttr->setVisible(true);
	}
	else
	{
		this->collectionAttr->setVisible(false);
	}

    // Variable
	if
		(0 < info.variable.label.size())
	{
		this->variableAttr->setValue(info.variable.label);
		this->variableAttr->setVisible(true);
	}
	else
	{
		this->variableAttr->setVisible(false);
	}

    // Time
    this->updateTemporalCoverage(start, stop);
    this->_timeCoverageEditor->setVisible(true);

    // Space
    this->updateSpatialCoverage(static_cast<double>(info.area),
                                static_cast<double>(info.perimeter),
                                static_cast<double>(info.lon),
                                static_cast<double>(info.lat));

    auto found = properties.children.find("properties");
    if
        (properties.children.end() != found)
    {
        this->_editor->setProperties(std::move(properties));
    }
    else
    {
        DictNode emptyProperties;
        this->_editor->setProperties(std::move(emptyProperties));
    }

    this->extractBtn->setEnabled(true);
    this->delBtn->setEnabled(true);
    this->editBtn->setEnabled(true);
    this->editBtn->setPushed(false);
    this->saveBtn->setEnabled(true);
    this->setVisible(true);

    this->target = rId;

    return true;
};

void NanoguiAnnotationDialog::updateSpatialCoverage(double area, double perimeter, double lon, double lat)
{
    std::ostringstream buffer("");  //
	std::string tmpStr;	//
	bool ok;	//
    bool areaOk;    //

    areaOk = SEAScope::formatArea(area, tmpStr);
    if
        (true == areaOk)
    {
        buffer << "Area: " << tmpStr;
    }

    ok = SEAScope::formatDistance(perimeter, tmpStr);
    if
        (true == ok)
    {
        if
            (0 < buffer.tellp())
        {
            buffer << "\n";
        }

        if
            (true == areaOk)
        {
            buffer << "Perimeter: " << tmpStr;
        }
        else
        {
            buffer  << "Length: " << tmpStr;
        }
    }

    ok = SEAScope::formatLongitude(lon, tmpStr);
    if
        (true == ok)
    {
        if
            (0 < buffer.tellp())
        {
            buffer << "\n";
        }

        buffer << "Longitude: " << tmpStr;
    }

    ok = SEAScope::formatLatitude(lat, tmpStr);
    if
        (true == ok)
    {
        if
            (0 < buffer.tellp())
        {
            buffer << "\n";
        }

        buffer << "Latitude: " << tmpStr;
    }

    this->spatialCoverageAttr->setValue(buffer.str());
}

/**
 *
 */
void NanoguiAnnotationDialog::updateTemporalCoverage(double start, double stop)
{
	char startStr[20];	//
	char stopStr[20];	//
	time_t timestampStart;	//
	time_t timestampStop;	//
	std::tm* ltimeStart;	//
	std::tm* ltimeStop;	//

    timestampStart = static_cast<long>(.001 * start);
    ltimeStart = std::localtime(&timestampStart);
    strftime(startStr, sizeof(startStr), "%Y-%m-%d %H:%M:%S", ltimeStart);

    timestampStop = static_cast<long>(.001 * stop);
    ltimeStop = std::localtime(&timestampStop);
    strftime(stopStr, sizeof(stopStr), "%Y-%m-%d %H:%M:%S", ltimeStop);

    this->_timeCoverageEditor->setValue(startStr, stopStr);
}

/**
 *
 */
bool NanoguiAnnotationDialog::disable()
{
    LOG(LogLevel::Debug, "Annotation dialog", "disable");
    if
        (true == this->visible())
    {
        this->_editor->hideImageViewer();
        this->extractBtn->setEnabled(false);
        this->delBtn->setEnabled(false);
        this->editBtn->setEnabled(false);
        this->editBtn->setPushed(false);
        if
            (nullptr != this->editCallback)
        {
            this->editCallback(this->target, false);
        }
        this->saveBtn->setEnabled(false);
        this->_saveWidget->setVisible(false);
	    this->setVisible(false);
        this->setFocused(false);
        return true;
    }
	return false;
};

/**
 *
 */
void NanoguiAnnotationDialog::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float btnsPosX; //
    float btnsPosY; //
    float btnsSizeX;    //
    float btnsSizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());

    btnsPosX = static_cast<float>(this->btns->position().x());
    btnsPosY = static_cast<float>(this->btns->position().y());
    btnsSizeX = static_cast<float>(this->btns->size().x());
    btnsSizeY = static_cast<float>(this->btns->size().y());

	nvgBeginPath(ctx);
	nvgMoveTo(ctx,
              posX + btnsPosX,
              posY + btnsPosY + 15.f /*13 spacer + 2 align*/);
    nvgLineTo(ctx,
              posX + btnsPosX,
              posY + btnsPosY + btnsSizeY - 15.f /*13 spacer + 2 align*/ + 1.f /*line width?*/);
    nvgBezierTo(ctx,
                posX + btnsPosX,
                posY + btnsPosY + btnsSizeY - 7.f + 1.f /*line width?*/,
                posX + btnsPosX + btnsSizeX,
                posY + btnsPosY + btnsSizeY - 15.f /*13 spacer + 2 align*/ + 1.f /*line width?*/,
                posX + btnsPosX + btnsSizeX,
                posY + btnsPosY + btnsSizeY);
    nvgLineTo(ctx,
              posX + btnsPosX + btnsSizeX,
              posY + btnsPosY);
    nvgBezierTo(ctx,
                posX + btnsPosX + btnsSizeX,
                posY + btnsPosY + 15.f /*13 spacer + 2 align*/,
                posX + btnsPosX,
                posY + btnsPosY + 7.f,
                posX + btnsPosX,
                posY + btnsPosY + 15.f);
    nvgStrokeColor(ctx, nanogui::Color(96, 96, 96, 180));
    nvgStroke(ctx);
	nvgFillColor(ctx, nanogui::Color(0,0,0,255));
	nvgFill(ctx);

	nanogui::Widget::draw(ctx);
};

/**
 *
 */
bool NanoguiAnnotationDialog::setSavePropertiesCallback(const std::function<void(const RenderableId&, const DictNode&)>& callback)
{
    if
        (nullptr != this->_savePropertiesCallback)
    {
        return false;
    }

    this->_savePropertiesCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiAnnotationDialog::setUpdateTimeCoverageCallback(const std::function<void(const RenderableId&, double, double)>& callback)
{
    if
        (nullptr != this->_updateTimeCoverageCallback)
    {
        return false;
    }

    this->_updateTimeCoverageCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiAnnotationDialog::setSaveCallback(const std::function<void(const RenderableId&, const std::string&, const AnnotationFormat&, bool)>& callback)
{
    if
        (nullptr != this->_saveCallback)
    {
        return false;
    }

    this->_saveCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiAnnotationDialog::keyboardEvent(int key, int /*scancode*/, int /*action*/, int modifiers)
{
    if
        (  (0 < (modifiers & NANOGUI_EVENT_BUBBLING))
        && (nullptr != this->deleteCallback)
        && (  (GLFW_KEY_DELETE == key)
           || (GLFW_KEY_BACKSPACE == key)))
    {
        // The delete shortcut should be ignored if the edition of the shape is
        // active, otherwise users may delete their annotations while trying to
        // remove a vertex
        if
            (false == this->editBtn->pushed())
        {
            this->deleteCallback(this->target);
            return true;
        }
    }

    if
        (GLFW_KEY_ESCAPE == key)
    {
        if
            (this->editBtn->pushed() && (nullptr != this->editCallback))
        {
            this->editCallback(this->target, false);
            return true;
        }
    }

    return false;
};
