// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

#include "ui/nanogui/nanogui_app_settings_editor.h"
#include "ui/nanogui/nanogui_path_selector.h"
#include "ui/nanogui/nanogui_numberbox.h"
#include "ui/nanogui/nanogui_checkbox.h"
#include "ui/nanogui/nanogui_textfield.h"
#include "ui/nanogui/nanogui_icon.h"
#include "ui/nanogui/nanogui_fading_line.h"
#include "ui/nanogui/nanogui_timestep_editor.h"
#include "ui/nanogui/nanogui_timespan_editor.h"
#include "num_utils.h"
#include "core/log.h"
#include "core/path.h"
#include "core/utils.h"

using namespace SEAScope;

static inline void buildSettingTitle(nanogui::Widget* parent, const std::string& label, const std::string& tooltip)
{
    nanogui::Widget* wrap {nullptr};    //
    nanogui::Label* lbl {nullptr};  //
    NanoguiIcon* icon {nullptr};  //

    wrap = new nanogui::Widget(parent);
    wrap->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                           nanogui::Alignment::Middle,
                                           0, 6));
    lbl = new nanogui::Label(wrap, label);
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    icon = new NanoguiIcon(wrap, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setTooltip(tooltip);
    icon->setColor({200, 200, 200, 255});
    icon->setFontSize(18);
    icon->setFixedSize(Eigen::Vector2i(20, 20));
}

/**
 *
 */
NanoguiTimespanEditor* NanoguiAppSettingsEditor::buildTimespanEditor()
{
    NanoguiTimespanEditor* editor {nullptr};    //
    editor = new NanoguiTimespanEditor(this->_timespans->content);
    editor->setSaveCallback([this, editor] (const std::string& label, double past, double future, const std::string& description, bool isDefault) {
        // In case this is an update, check that the new label does not conflict
        // with another timespan. If there is a conflict, reset the label in
        // the editor using the original identifier
        if
            (0 < editor->getIdentifier().size())
        {
            for
                (auto& child : this->_timespans->content->children())
            {
                NanoguiTimespanEditor* editorIt {nullptr};
                editorIt = dynamic_cast<NanoguiTimespanEditor*>(child);
                if
                    (nullptr == editorIt)
                {
                    continue;
                }

                if
                    ((editorIt != editor) && (label == editorIt->getIdentifier()))
                {
                    LOG(LogLevel::Warn, "NanoguiAppSettingsEditor",
                             "You cannot have two timespans with the '{}' label",
                              label);
                    editor->set(editor->getIdentifier(), description, static_cast<timestamp_t>(past), static_cast<timestamp_t>(future), isDefault);
                    editor->enableEdition();
                    return;
                }
            }
        }

        bool ok {false};    //
        ok = this->updateTimespan(editor->getIdentifier(), label, past, future, description, isDefault);
        if
            (false == ok)
        {
            if
                (0 < editor->getIdentifier().size())
            {
                LOG(LogLevel::Error, "NanoguiAppSettingsEditor",
                         "Timespan with ID '{}' has not been found and can therefore not be updated",
                         editor->getIdentifier());
            }
            else
            {
                LOG(LogLevel::Warn, "NanoguiAppSettingsEditor",
                         "Timespan with ID '{}' could not be added because there is already one with the same ID",
                         label);
            }
            return;
        }

        // Set parameters again so that the new label becomes the identifier
        editor->set(label, description, static_cast<timestamp_t>(past), static_cast<timestamp_t>(future), isDefault);

        this->sendChangeNotification();

        this->checkRestartRequirement();
    });

    editor->setDeleteCallback([this, editor] () {
        if
            (0 >= editor->getIdentifier().size())
        {
            // This is an editor that has not been attached to an actual timespan yet
            this->_timespanEditorsToRemove.push_back(editor);
            return;
        }

        bool ok {false};    //
        ok = this->_editedSettings.timelineCfg.removeTimeframe(editor->getIdentifier(),
                                                               this->_editedSettings.timelineCfg.defaultTimespan);
        if
            (false == ok)
        {
            // the timespan was not removed, abort
            LOG(LogLevel::Warn, "NanoguiAppSettingsEditor", "Timespan not removed");
            return;
        }

        // Update default timespan
        for
            (auto& child : this->_timespans->content->children())
        {
            NanoguiTimespanEditor* editorIt {nullptr};
            editorIt = dynamic_cast<NanoguiTimespanEditor*>(child);
            if
                (nullptr == editorIt)
            {
                continue;
            }

            if
                (editorIt->getIdentifier() != this->_editedSettings.timelineCfg.defaultTimespan)
            {
                editorIt->uncheckDefault();
            }
            else
            {
                editorIt->checkDefault();
            }
        }

        this->_timespanEditorsToRemove.push_back(editor);
    });

    editor->setDefaultCallback([this, editor] () {
        for
            (auto& child : this->_timespans->content->children())
        {
            NanoguiTimespanEditor* editorIt {nullptr};
            editorIt = dynamic_cast<NanoguiTimespanEditor*>(child);
            if
                (nullptr == editorIt)
            {
                continue;
            }

            if
                (editorIt != editor)
            {
                editorIt->uncheckDefault();
            }
        }

        this->_editedSettings.timelineCfg.defaultTimespan = editor->getIdentifier();

        this->sendChangeNotification();

        this->checkRestartRequirement();
    });

    return editor;
}

/**
 *
 */
NanoguiTimestepEditor* NanoguiAppSettingsEditor::buildTimestepEditor()
{
    NanoguiTimestepEditor* editor {nullptr};    //
    editor = new NanoguiTimestepEditor(this->_timesteps->content);
    editor->setSaveCallback([this, editor] (const std::string& label, double offset, const std::string& description, bool isDefault) {
        // In case this is an update, check that the new label does not conflict
        // with another timestep. If there is a conflict, reset the label in
        // the editor using the original identifier
        if
            (0 < editor->getIdentifier().size())
        {
            for
                (auto& child : this->_timesteps->content->children())
            {
                NanoguiTimestepEditor* editorIt {nullptr};
                editorIt = dynamic_cast<NanoguiTimestepEditor*>(child);
                if
                    (nullptr == editorIt)
                {
                    continue;
                }

                if
                    ((editorIt != editor) && (label == editorIt->getIdentifier()))
                {
                    LOG(LogLevel::Warn, "NanoguiAppSettingsEditor",
                             "You cannot have two timesteps with the '{}' label",
                              label);
                    editor->set(editor->getIdentifier(), description, static_cast<timestamp_t>(offset), isDefault);
                    editor->enableEdition();
                    return;
                }
            }
        }

        bool ok {false};    //
        ok = this->updateTimestep(editor->getIdentifier(), label, offset, description, isDefault);
        if
            (false == ok)
        {
            if
                (0 < editor->getIdentifier().size())
            {
                LOG(LogLevel::Error, "NanoguiAppSettingsEditor",
                         "Timestep with ID '{}' has not been found and can therefore not be updated",
                         editor->getIdentifier());
            }
            else
            {
                LOG(LogLevel::Warn, "NanoguiAppSettingsEditor",
                         "Timestep with ID '{}' could not be added because there is already one with the same ID",
                         label);
            }
            return;
        }
        // Set parameters again so that the new label becomes the identifier
        editor->set(label, description, static_cast<timestamp_t>(offset), isDefault);

        this->sendChangeNotification();

        this->checkRestartRequirement();
    });

    editor->setDeleteCallback([this, editor] () {
        if
            (0 >= editor->getIdentifier().size())
        {
            // This is an editor that has not been attached to an actual timestep yet
            this->_timestepEditorsToRemove.push_back(editor);
            return;
        }

        bool ok {false};    //
        ok = this->_editedSettings.timelineCfg.removeTimestep(editor->getIdentifier(),
                                                              this->_editedSettings.timelineCfg.defaultTimestep);
        if
            (false == ok)
        {
            // the timespan was not removed, abort
            LOG(LogLevel::Warn, "NanoguiAppSettingsEditor", "Timestep not removed");
            return;
        }

        // Update default timespan
        for
            (auto& child : this->_timesteps->content->children())
        {
            NanoguiTimestepEditor* editorIt {nullptr};
            editorIt = dynamic_cast<NanoguiTimestepEditor*>(child);
            if
                (nullptr == editorIt)
            {
                continue;
            }

            if
                (editorIt->getIdentifier() != this->_editedSettings.timelineCfg.defaultTimestep)
            {
                editorIt->uncheckDefault();
            }
            else
            {
                editorIt->checkDefault();
            }
        }

        this->_timestepEditorsToRemove.push_back(editor);
    });

    editor->setDefaultCallback([this, editor] () {
        for
            (auto& child : this->_timesteps->content->children())
        {
            NanoguiTimestepEditor* editorIt {nullptr};
            editorIt = dynamic_cast<NanoguiTimestepEditor*>(child);
            if
                (nullptr == editorIt)
            {
                continue;
            }

            if
                (editorIt != editor)
            {
                editorIt->uncheckDefault();
            }
        }

        this->_editedSettings.timelineCfg.defaultTimestep = editor->getIdentifier();

        this->sendChangeNotification();

        this->checkRestartRequirement();
    });

    return editor;
}

/**
 *
 */
NanoguiAppSettingsEditor::NanoguiAppSettingsEditor(nanogui::Widget* parent)
: nanogui::Widget(parent), NanoguiScrollableWithTitle(parent),
  _colormapsPath{nullptr}, _dataPath{nullptr},
  _statePath{nullptr}, _indexPath{nullptr}, _annotationsPath{nullptr},
  _customPath{nullptr},
  _fullscreenCheck{nullptr}, _windowedCheck{nullptr}, _windowWidth{nullptr}, _windowHeight{nullptr},
  _vsyncCheck{nullptr}, _antialiasingCheck{nullptr}, _msaaSampling{nullptr}, _minScreenPortion{nullptr},
  _minYear{nullptr}, _maxYear{nullptr}, _timesteps{nullptr}, _timespans{nullptr}, _nearestCheck{nullptr},
  _advancedScrollingCheck{nullptr}, _initialDatetime{nullptr},
  _drawingAutoExit{nullptr}, _drawingManualExit{nullptr}, _clickTolerance{nullptr},
  _processorCheck{nullptr}, _processorAddress{nullptr}, _processorPort{nullptr},
  _indexCheck{nullptr}, _minSearchDelay{nullptr}, _backgroundPath{nullptr},
  _changeCallback{nullptr}, _muted{false}, _restartNeededMessage{nullptr}
{
    nanogui::Widget* wrap {nullptr};    //
    nanogui::Widget* spacer {nullptr};  //
    nanogui::Label* lbl {nullptr};  //
    nanogui::Button* btn {nullptr}; //
    NanoguiIcon* icon {nullptr};    //
    nanogui::Widget* w {nullptr};   //
    NanoguiFadingLine* line {nullptr};  //

    this->footer->setFixedHeight(5);

    this->header->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                   nanogui::Alignment::Minimum,
                                                   8, 0));
    this->header->setFixedHeight(110);


    w = new nanogui::Widget(this->header);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedHeight(10);

    new nanogui::Label(w, "Please note the SEAScope will automatically save");
    new nanogui::Label(w, "the changes you make to its configuration but most");
    new nanogui::Label(w, "of them will only take effect after you restart the");
    new nanogui::Label(w, "application.");

    spacer = new nanogui::Widget(w);
    spacer->setFixedHeight(6);

    this->_restartNeededMessage = new nanogui::Label(w, "SEAScope restart needed", "sans-bold");
    this->_restartNeededMessage->setFontSize(18);
    this->_restartNeededMessage->setColor(nanogui::Color(255, 250, 0, 200));
    this->_restartNeededMessage->setVisible(false);

    // Edition form

    this->scrollable->content->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                                nanogui::Alignment::Minimum,
                                                                8, 0));

    wrap = new nanogui::Widget(this->scrollable->content);
	wrap->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
										   , nanogui::Alignment::Minimum
										   , 0, 0));
	wrap->setFixedWidth(304);
	wrap->setWidth(304);

    // - Paths
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 5));
    lbl = new nanogui::Label(w, "Paths", "sans-bold");
    lbl->setFontSize(18);
    line = new NanoguiFadingLine(w, 4);
    line->setFixedSize(Eigen::Vector2i(280, 20));
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(5);

    buildSettingTitle(wrap, "• Colormaps directory", "Directory containing the colormaps (lists of RGB values) that SEAScope can use to convert geophysical values into colors. The path must be an existing directory.");
    this->_colormapsPath = new NanoguiPathSelector(wrap, PathSelectorType::get_folder);
    this->_colormapsPath->resize(Eigen::Vector2i(300, 40));
    this->_colormapsPath->setChangeCallback([this] (const std::string& path) {
        this->updateColormapsPath(path);
    });
    this->_colormapsPath->setDropDownCallback([this] () {
        this->_dataPath->hideDropDown();
        this->_statePath->hideDropDown();
        this->_indexPath->hideDropDown();
        this->_annotationsPath->hideDropDown();
        this->_customPath->hideDropDown();
        this->_backgroundPath->hideDropDown();
    });

    buildSettingTitle(wrap, "• Data directory", "Directory from which SEAScope will load data from. The path must be an existing directory.");
    this->_dataPath = new NanoguiPathSelector(wrap, PathSelectorType::get_folder);
    this->_dataPath->resize(Eigen::Vector2i(300, 40));
    this->_dataPath->setChangeCallback([this] (const std::string& path) {
        this->updateDataPath(path);
    });
    this->_dataPath->setDropDownCallback([this] () {
        this->_colormapsPath->hideDropDown();
        this->_statePath->hideDropDown();
        this->_indexPath->hideDropDown();
        this->_annotationsPath->hideDropDown();
        this->_customPath->hideDropDown();
        this->_backgroundPath->hideDropDown();
    });

    buildSettingTitle(wrap, "• State file", "Path of the file where SEAScope will save its state when the application shuts down. The file will be created automatically but the parent directory must already exist.");
    this->_statePath = new NanoguiPathSelector(wrap, PathSelectorType::save_file);
    this->_statePath->resize(Eigen::Vector2i(300, 40));
    this->_statePath->setChangeCallback([this] (const std::string& path) {
        this->updateStatePath(path);
    });
    this->_statePath->setDropDownCallback([this] () {
        this->_colormapsPath->hideDropDown();
        this->_dataPath->hideDropDown();
        this->_indexPath->hideDropDown();
        this->_annotationsPath->hideDropDown();
        this->_customPath->hideDropDown();
        this->_backgroundPath->hideDropDown();
    });

    buildSettingTitle(wrap, "• Index file", "Path of the file where SEAScope will store the index of all the granules found in the data directory. SEAScope will create the file but the parent directory must alreadt exist.");
    this->_indexPath = new NanoguiPathSelector(wrap, PathSelectorType::save_file);
    this->_indexPath->resize(Eigen::Vector2i(300, 40));
    this->_indexPath->setChangeCallback([this] (const std::string& path) {
        this->updateIndexPath(path);
    });
    this->_indexPath->setDropDownCallback([this] () {
        this->_colormapsPath->hideDropDown();
        this->_dataPath->hideDropDown();
        this->_statePath->hideDropDown();
        this->_annotationsPath->hideDropDown();
        this->_customPath->hideDropDown();
        this->_backgroundPath->hideDropDown();
    });

    buildSettingTitle(wrap, "• Annotations database", "Path of the file where SEAScope will store your annotations. SEAScope will create the file but the parent directory must already exist.");
    this->_annotationsPath = new NanoguiPathSelector(wrap, PathSelectorType::save_file);
    this->_annotationsPath->resize(Eigen::Vector2i(300, 40));
    this->_annotationsPath->setChangeCallback([this] (const std::string& path) {
        this->updateAnnotationsPath(path);
    });
    this->_annotationsPath->setDropDownCallback([this] () {
        this->_colormapsPath->hideDropDown();
        this->_dataPath->hideDropDown();
        this->_indexPath->hideDropDown();
        this->_statePath->hideDropDown();
        this->_customPath->hideDropDown();
        this->_backgroundPath->hideDropDown();
    });

    buildSettingTitle(wrap, "• Custom directory", "Directory containing custom resources loaded by SEAScope (like icons). The path must be an existing directory.");
    this->_customPath = new NanoguiPathSelector(wrap, PathSelectorType::get_folder);
    this->_customPath->resize(Eigen::Vector2i(300, 40));
    this->_customPath->setChangeCallback([this] (const std::string& path) {
        this->updateCustomPath(path);
    });
    this->_customPath->setDropDownCallback([this] () {
        this->_colormapsPath->hideDropDown();
        this->_dataPath->hideDropDown();
        this->_indexPath->hideDropDown();
        this->_statePath->hideDropDown();
        this->_annotationsPath->hideDropDown();
        this->_backgroundPath->hideDropDown();
    });

    // - Rendering
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(15);
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 5));
    lbl = new nanogui::Label(w, "Rendering", "sans-bold");
    lbl->setFontSize(18);
    line = new NanoguiFadingLine(w, 4);
    line->setFixedSize(Eigen::Vector2i(280, 20));
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(5);

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    this->_fullscreenCheck = new NanoguiCheckbox(w, "Start in fullscreen mode.");
    this->_fullscreenCheck->setFontSize(16);
    this->_fullscreenCheck->setCallback([this] (bool checked) {
        this->updateFullscreen(checked);
    });
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setColor({200, 200, 200, 255});
    icon->setTooltip("When this option is on, SEAScope starts in fullscreen mode, no other window will be visible.");

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    this->_windowedCheck = new NanoguiCheckbox(w, "Start in windowed mode");
    this->_windowedCheck->setFontSize(16);
    this->_windowedCheck->setCallback([this] (bool checked) {
        this->updateWindowed(checked);
    });
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setColor({200, 200, 200, 255});
    icon->setTooltip("When this option is on, SEAScope starts in a window that you can move / reduce / close.");

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    lbl = new nanogui::Label(w, "Window size: ");
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    this->_windowWidth = new NanoguiNumberBox(w);
    this->_windowWidth->setFormat("[0-9]+");
    this->_windowWidth->setFontSize(16);
    this->_windowWidth->setValue(800);
    this->_windowWidth->setCallback([this] (double value) {
        this->updateWindowWidth(value);
    });
    lbl = new nanogui::Label(w, "x");
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    this->_windowHeight = new NanoguiNumberBox(w);
    this->_windowHeight->setFormat("[0-9]+");
    this->_windowHeight->setFontSize(16);
    this->_windowHeight->setValue(600);
    this->_windowHeight->setCallback([this] (double value) {
        this->updateWindowHeight(value);
    });

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- vsync
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    this->_vsyncCheck = new NanoguiCheckbox(w, "Activate Vsync");
    this->_vsyncCheck->setFontSize(16);
    this->_vsyncCheck->setCallback([this] (bool checked) {
        this->updateVSync(checked);
    });
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setColor({200, 200, 200, 255});
    icon->setTooltip("When this option is on, SEAScope synchronizes its rendering speed to match the refresh rate of your monitor, thus avoiding rendering more frames than necessary and preventing tearing effects.");

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- anti-aliasing
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    this->_antialiasingCheck = new NanoguiCheckbox(w, "Activate anti-aliasing");
    this->_antialiasingCheck->setFontSize(16);
    this->_antialiasingCheck->setCallback([this] (bool checked) {
        this->updateAntialiasing(checked);
    });
    lbl = new nanogui::Label(w, "with");
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    this->_msaaSampling = new NanoguiNumberBox(w);
    this->_msaaSampling->setFormat("[2,4,6,8]");
    this->_msaaSampling->setFixedHeight(20);
    this->_msaaSampling->setFontSize(16);
    this->_msaaSampling->setValue(8);
    this->_msaaSampling->setCallback([this] (double value) {
        this->updateMSAASampling(value);
    });
    lbl = new nanogui::Label(w, "x sampling rate");
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setColor({200, 200, 200, 255});
    icon->setTooltip("Anti-aliasing makes the rendering of edges smoother. Increasing the number of samples improves the smoothing effect at the cost of additional computing. Supported values are 2, 4, 6 and 8.");

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    buildSettingTitle(wrap, "• Granule size threshold", "SEAScope only displays granules if they cover a big enough portion of the screen to avoid loading data that cannot (or hardly) be seen on the globe.");
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    lbl = new nanogui::Label(w, "A granule must take at least ");
    lbl->setFontSize(16);
    this->_minScreenPortion = new NanoguiNumberBox(w);
    this->_minScreenPortion->setFontSize(16);
    this->_minScreenPortion->setValue(0.8);
    this->_minScreenPortion->setCallback([this] (double value) {
        this->updateMinScreenPortion(value);
    });
    lbl = new nanogui::Label(w, "%");
    lbl->setFontSize(16);

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    lbl = new nanogui::Label(w, "of the screen to be displayed");
    lbl->setFontSize(16);

    // - Timeline
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(15);
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 5));
    lbl = new nanogui::Label(w, "Timeline", "sans-bold");
    lbl->setFontSize(18);
    line = new NanoguiFadingLine(w, 4);
    line->setFixedSize(Eigen::Vector2i(280, 20));
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(5);

    // -- years range
    buildSettingTitle(wrap, "• Years range", "This settings define the bounds of the timeline. If the max year is set to 0 then SEAScope will replace it with the current year.");
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    lbl = new nanogui::Label(w, "From year ");
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    this->_minYear = new NanoguiNumberBox(w);
    this->_minYear->setFixedSize(Eigen::Vector2i(50, 20));
    this->_minYear->setFormat("[0-9][0-9][0-9][0-9]");
    this->_minYear->setFontSize(16);
    this->_minYear->setValue(TimelineCfg::MIN_VALID_YEAR);
    this->_minYear->setCallback([this] (double value) {
        this->updateMinYear(value);
    });
    std::ostringstream tooltipBuffer{""};    //
    tooltipBuffer << "Value must be in the [" << TimelineCfg::MIN_VALID_YEAR << ", " << TimelineCfg::MAX_VALID_YEAR << "] range";
    this->_minYear->setTooltip(tooltipBuffer.str());
    lbl = new nanogui::Label(w, " to year ");
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    this->_maxYear = new NanoguiNumberBox(w);
    this->_maxYear->setFixedSize(Eigen::Vector2i(50, 20));
    this->_maxYear->setFormat("0|([0-9][0-9][0-9][0-9])");
    this->_maxYear->setFontSize(16);
    this->_maxYear->setValue(TimelineCfg::MAX_VALID_YEAR);
    this->_maxYear->setCallback([this] (double value) {
        this->updateMaxYear(value);
    });
    tooltipBuffer.str("");
    tooltipBuffer << "0 means \"current year\". If not 0, value must be in [" << TimelineCfg::MIN_VALID_YEAR << ", " << TimelineCfg::MAX_VALID_YEAR << "] range";
    this->_maxYear->setTooltip(tooltipBuffer.str());

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- time spans
    buildSettingTitle(wrap, "• Time spans", "Timespans are temporal windows used as filter to select the granules to display on the globe. They are defined as a pair of time amounts relative to the datetime selected in the timeline (it is possible to define asymmetric timespans)");

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    btn = new nanogui::Button(w, "Add new time span");
    btn->setFontSize(16);
    btn->setFixedHeight(20);
    btn->setCallback([this] () {
        NanoguiTimespanEditor* editor {nullptr};    //
        editor = this->buildTimespanEditor();
        editor->enableEdition();
        this->_timespans->update(true);
    });

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(8);

    this->_timespans = new NanoguiList(wrap);
    this->_timespans->content->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                                nanogui::Alignment::Middle,
                                                                0, 10));

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- time steps
    buildSettingTitle(wrap, "• Time steps", "Timesteps are fixed offsets by which the timeline is shifted when you click on the Prev / Next buttons. If a timestep has an offset value of 0 and is active, then SEAScope will jump to the previous / next available granule.");

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    btn = new nanogui::Button(w, "Add new time step");
    btn->setFontSize(16);
    btn->setFixedHeight(20);
    btn->setCallback([this] () {
        NanoguiTimestepEditor* editor {nullptr};    //
        editor = this->buildTimestepEditor();
        editor->enableEdition();
        this->_timesteps->update(true);
    });

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(8);

    this->_timesteps = new NanoguiList(wrap);
    this->_timesteps->content->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                                nanogui::Alignment::Middle,
                                                                0, 10));

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- use nearest mode
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    this->_nearestCheck = new NanoguiCheckbox(w, "Auto jump to the closest datetime with data");
    this->_nearestCheck->setFontSize(16);
    this->_nearestCheck->setCallback([this] (bool checked) {
        this->updateNearest(checked);
    });
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setColor({200, 200, 200, 255});
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setTooltip("This option tells SEAScope to center the timeline on the closest date and time with available data.");

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- advanced scrolling
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    this->_advancedScrollingCheck = new NanoguiCheckbox(w, "Use advanced scrolling");
    this->_advancedScrollingCheck->setFontSize(16);
    this->_advancedScrollingCheck->setCallback([this] (bool checked) {
        this->updateAdvancedScrolling(checked);
    });
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setColor({200, 200, 200, 255});
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setTooltip("When advanced scrolling is enabled each row of the timeline (years, months and days) can be scrolled independently. When it is disabled the scroll shifts the time by the same amount regardless on the hovered row.");

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- initial datetime
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    lbl = new nanogui::Label(w, "• Initial datetime: ");
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    this->_initialDatetime = new NanoguiTextField(w);
    this->_initialDatetime->setFormat("([0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9])|([0-9][0-9]:[0-9][0-9]:[0-9][0-9])|([0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]:[0-9][0-9]:[0-9][0-9])");
    this->_initialDatetime->setFontSize(16);
    this->_initialDatetime->setFixedSize(Eigen::Vector2i(180, 20));
    this->_initialDatetime->setCallback([this] (const std::string& value) -> bool {
        return this->updateInitialDatetime(value);
    });
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setColor({200, 200, 200, 255});
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setTooltip("Sets the datetime on which the timeline is centered when SEAScope starts. Can be left empty (current datetime), a full datetime in YYYY-mm-ddTHH:MM:SS format, a date in YYYY-mm-dd format (time will be set to 12:00:00) or a time in HH:MM:SS format (date will be set to current day).");

    // - Misc
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(15);
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 5));
    lbl = new nanogui::Label(w, "Misc", "sans-bold");
    lbl->setFontSize(18);
    lbl->setFixedHeight(20);
    line = new NanoguiFadingLine(w, 4);
    line->setFixedSize(Eigen::Vector2i(280, 20));
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(5);

    // -- index deprecation
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    this->_indexCheck = new NanoguiCheckbox(w, "Check for index deprecation");
    this->_indexCheck->setFontSize(16);
    this->_indexCheck->setCallback([this] (bool checked) {
        this->updateIndexCheck(checked);
    });
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setColor({200, 200, 200, 255});
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setTooltip("At startup SEAScope tries to detect if there have been changes in the data directory and prompt you if an update of the data index is required.");

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- search requests delay
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    lbl = new nanogui::Label(w, "• Min. delay between search requests: ");
    lbl->setFixedHeight(20);
    lbl->setFontSize(16);
    this->_minSearchDelay = new NanoguiNumberBox(w);
    this->_minSearchDelay->setFormat("[0-9]+");
    this->_minSearchDelay->setFixedHeight(20);
    this->_minSearchDelay->setValue(10);
    this->_minSearchDelay->setFontSize(16);
    this->_minSearchDelay->setCallback([this] (double value) {
        this->updateMinSearchDelay(value);
    });
    lbl = new nanogui::Label(w, " ms");
    lbl->setFixedHeight(20);
    lbl->setFontSize(16);
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setColor({200, 200, 200, 255});
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setTooltip("SEAScope executes search requests every time you change the date, the timespan, the zoom or the variables to display. Casually browsing data in SEAScope can generate a lot of requests and overwhelm the application, so SEAScope uses a throttling mechanism by postponing the processing of search requests if too many of them are created in a short time. This setting defines the minimum amount of time that must pass between two search requests.");

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- background image
    buildSettingTitle(wrap, "• Background image", "SEAScope draws an image on the globe below your data to serve as background.");
    this->_backgroundPath = new NanoguiPathSelector(wrap, PathSelectorType::get_file);
    this->_backgroundPath->resize(Eigen::Vector2i(300, 40));
    this->_backgroundPath->setChangeCallback([this] (const std::string& path) {
        this->updateBackgroundPath(path);
    });
    this->_backgroundPath->setDropDownCallback([this] () {
        this->_colormapsPath->hideDropDown();
        this->_dataPath->hideDropDown();
        this->_indexPath->hideDropDown();
        this->_statePath->hideDropDown();
        this->_customPath->hideDropDown();
        this->_annotationsPath->hideDropDown();
    });

    // - Drawing
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(15);
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 5));
    lbl = new nanogui::Label(w, "Drawing", "sans-bold");
    lbl->setFontSize(18);
    line = new NanoguiFadingLine(w, 4);
    line->setFixedSize(Eigen::Vector2i(280, 20));
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(5);

    // -- exit drawing mode
    buildSettingTitle(wrap, "• Exit drawing mode", "You can decide if SEAScope should remain in drawing mode until you choose to leave it (useful when drawing many shapes in a row) or if it should leave this mode as soon as you finish drawing a shape.");
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    this->_drawingAutoExit = new NanoguiCheckbox(w, "upon completing a shape");
    this->_drawingAutoExit->setFontSize(16);
    this->_drawingAutoExit->setCallback([this] (bool checked) {
        this->updateDrawingAutoExit(checked);
    });

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    this->_drawingManualExit = new NanoguiCheckbox(w, "by pressing Esc.");
    this->_drawingManualExit->setFontSize(16);
    this->_drawingManualExit->setCallback([this] (bool checked) {
        this->updateDrawingManualExit(checked);
    });

    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(10);

    // -- click tolerance
    buildSettingTitle(wrap, "• Click tolerance", "This setting allows you to define how many pixels your cursor can move between the press and release of the mouse button and the event still registers as a \"click\". This is notably useful on high resolution screens where the slightest movement of the mouse causes the cursor to translate by a large amount of pixels.");
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    lbl = new nanogui::Label(w, "Allow ");
    lbl->setFontSize(16);
    this->_clickTolerance = new NanoguiNumberBox(w);
    this->_clickTolerance->setFormat("[0-9]+");
    this->_clickTolerance->setFontSize(16);
    this->_clickTolerance->setValue(7);
    this->_clickTolerance->setCallback([this] (double value) {
        this->updateClickTolerance(value);
    });
    lbl = new nanogui::Label(w, " pixels move between button");
    lbl->setFontSize(16);

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    lbl = new nanogui::Label(w, "press and release");
    lbl->setFontSize(16);

    // - Processor
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(15);
    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 5));
    lbl = new nanogui::Label(w, "Processor", "sans-bold");
    lbl->setFontSize(18);
    line = new NanoguiFadingLine(w, 4);
    line->setFixedSize(Eigen::Vector2i(280, 20));
    spacer = new nanogui::Widget(wrap);
    spacer->setFixedHeight(5);

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    this->_processorCheck = new NanoguiCheckbox(w, "seascope-processor is running");
    this->_processorCheck->setFontSize(16);
    this->_processorCheck->setCallback([this] (bool checked) {
        this->updateProcessorEnabled(checked);
    });
    icon = new NanoguiIcon(w, ENTYPO_ICON_INFO_WITH_CIRCLE);
    icon->setColor({200, 200, 200, 255});
    icon->setFixedSize(Eigen::Vector2i(20, 20));
    icon->setFontSize(18);
    icon->setTooltip("SEAScope can extend its capabilities by offloading some computations to the seascope-processor executable.\nThis setting tells SEAScope whether it should unlock the features supported by seascope-processor (i.e. the seascope-processor command is running) like generating a plot from a transect or exporting extracted data as a Numpy array.");

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    lbl = new nanogui::Label(w, "at address ");
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    this->_processorAddress = new NanoguiTextField(w);
    this->_processorAddress->setValue("255.255.255.255");
    this->_processorAddress->setEditable(true);
    this->_processorAddress->setFontSize(16);
    this->_processorAddress->setFixedSize(Eigen::Vector2i(130, 20));
    this->_processorAddress->setCallback([this] (const std::string& value) -> bool{
        return this->updateProcessorAddress(value);
    });
    this->_processorAddress->setTooltip("Expected value is an IPv4 address, if seascope-processor is running on the same computer as SEAScope then you should use \"127.0.0.1\".");

    w = new nanogui::Widget(wrap);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 0));
    spacer = new nanogui::Widget(w);
    spacer->setFixedWidth(25);
    lbl = new nanogui::Label(w, "on port ");
    lbl->setFontSize(16);
    lbl->setFixedHeight(20);
    this->_processorPort = new NanoguiNumberBox(w);
    this->_processorPort->setFormat("[0-9]+");
    this->_processorPort->setValue(53450);
    this->_processorPort->setFontSize(16);
    this->_processorPort->setCallback([this] (double value) {
        this->updateProcessorPort(value);
    });
    this->_processorPort->setTooltip("By default seascope-processor runs on port 53450.");
}

/**
 *
 */
NanoguiAppSettingsEditor::~NanoguiAppSettingsEditor()
{
}



/**
 *
 */
void NanoguiAppSettingsEditor::updateColormapsPath(const std::string& value)
{
    if
        (value != this->_editedSettings.colormapsPath)
    {
        // Colormaps can only be loaded from an existing directory
        if
            (true == Path::dirExists(value))
        {
            this->_editedSettings.colormapsPath = value;

            this->sendChangeNotification();

            this->checkRestartRequirement();

            this->_colormapsPath->hintValidPath();
        }
        else
        {
            this->_colormapsPath->hintInvalidPath();
        }
    }
    else
    {
        this->_colormapsPath->hintValidPath();
    }
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateDataPath(const std::string& value)
{
    if
        (value != this->_editedSettings.dataPath)
    {
        // Data can only be loaded from an existing directory
        if
            (true == Path::dirExists(value))
        {
            this->_editedSettings.dataPath = value;

            this->sendChangeNotification();

            this->checkRestartRequirement();

            this->_dataPath->hintValidPath();
        }
        else
        {
            this->_dataPath->hintInvalidPath();
        }
    }
    else
    {
        this->_dataPath->hintValidPath();
    }
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateStatePath(const std::string& value)
{
    if
        (value != this->_editedSettings.statePath)
    {
        // The file will be created by SEAScope, just make sure that the path
        // does not point to a directory and that the parent directory exists
        std::string parentDir {""};
        Path::getParentDir(value, parentDir);
        if
            (  (false == Path::dirExists(value))
            && (true == Path::dirExists(parentDir)))
        {
            this->_editedSettings.statePath = value;

            this->sendChangeNotification();

            this->checkRestartRequirement();

            this->_statePath->hintValidPath();
        }
        else
        {
            this->_statePath->hintInvalidPath();
        }
    }
    else
    {
        this->_statePath->hintValidPath();
    }
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateIndexPath(const std::string& value)
{
    if
        (value != this->_editedSettings.indexPath)
    {
        // The file will be created by SEAScope, just make sure that the path
        // does not point to a directory and that the parent directory exists
        std::string parentDir {""};
        Path::getParentDir(value, parentDir);
        if
            (  (false == Path::dirExists(value))
            && (true == Path::dirExists(parentDir)))
        {
            this->_editedSettings.indexPath = value;

            this->sendChangeNotification();

            this->checkRestartRequirement();

            this->_indexPath->hintValidPath();
        }
        else
        {
            this->_indexPath->hintInvalidPath();
        }
    }
    else
    {
        this->_indexPath->hintValidPath();
    }
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateAnnotationsPath(const std::string& value)
{
    if
        (value != this->_editedSettings.annotationsPath)
    {
        // The file will be created by SEAScope, just make sure that the path
        // does not point to a directory and that the parent directory exists
        std::string parentDir {""};
        Path::getParentDir(value, parentDir);
        if
            (  (false == Path::dirExists(value))
            && (true == Path::dirExists(parentDir)))
        {
            this->_editedSettings.annotationsPath = value;

            this->sendChangeNotification();

            this->checkRestartRequirement();

            this->_annotationsPath->hintValidPath();
        }
        else
        {
            this->_annotationsPath->hintInvalidPath();
        }
    }
    else
    {
        this->_annotationsPath->hintValidPath();
    }
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateCustomPath(const std::string& value)
{
    if
        (value != this->_editedSettings.customPath)
    {
        if
            (true == Path::dirExists(value))
        {
            this->_editedSettings.customPath = value;

            this->sendChangeNotification();

            this->checkRestartRequirement();

            this->_customPath->hintValidPath();
        }
        else
        {
            this->_customPath->hintInvalidPath();
        }
    }
    else
    {
        this->_customPath->hintValidPath();
    }
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateFullscreen(bool checked)
{
    this->_windowedCheck->setChecked(!checked);
    this->_windowWidth->setEnabled(!checked);
    this->_windowWidth->setEditable(!checked);
    this->_windowHeight->setEnabled(!checked);
    this->_windowHeight->setEditable(!checked);

    if
        (checked != this->_editedSettings.fullscreen)
    {
        this->_editedSettings.fullscreen = checked;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateWindowed(bool checked)
{
    this->_fullscreenCheck->setChecked(!checked);
    this->_windowWidth->setEnabled(checked);
    this->_windowWidth->setEditable(checked);
    this->_windowHeight->setEnabled(checked);
    this->_windowHeight->setEditable(checked);

    if
        (checked == this->_editedSettings.fullscreen)   // inverted logic
    {
        this->_editedSettings.fullscreen = !checked;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateWindowWidth(double value)
{
    std::uint16_t width {0u};   //

    width = static_cast<std::uint16_t>(value);
    if
        (width != this->_editedSettings.windowWidth)
    {
        this->_editedSettings.windowWidth = width;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateWindowHeight(double value)
{
    std::uint16_t height {0u};  //

    height = static_cast<std::uint16_t>(value);
    if
        (height != this->_editedSettings.windowHeight)
    {
        this->_editedSettings.windowHeight = height;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateVSync(bool checked)
{
    if
        (checked != this->_editedSettings.vsync)
    {
        this->_editedSettings.vsync = checked;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateAntialiasing(bool checked)
{
    this->_msaaSampling->setEnabled(checked);
    this->_msaaSampling->setEditable(checked);

    if
        (checked != this->_editedSettings.antialiasing)
    {
        this->_editedSettings.antialiasing = checked;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateMSAASampling(double value)
{
    std::uint8_t sampling {0u}; //

    sampling = static_cast<std::uint8_t>(value);
    if
        (sampling != this->_editedSettings.antialiasingFactor)
    {
        this->_editedSettings.antialiasingFactor = sampling;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateMinScreenPortion(double value)
{
    float minPortion {0.f}; //

    minPortion = static_cast<float>(0.01 * value);  // convert from percentage
    if
        (!floating_point_equality(minPortion, this->_editedSettings.minGranuleScreenEstate))
    {
        this->_editedSettings.minGranuleScreenEstate = minPortion;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();

}

/**
 *
 */
void NanoguiAppSettingsEditor::updateMinYear(double value)
{
    std::uint16_t year {0u};    //

    year = static_cast<std::uint16_t>(value);
    if
        (  (TimelineCfg::MIN_VALID_YEAR > year)
        || (TimelineCfg::MAX_VALID_YEAR < year))
    {
        this->_minYear->hintInvalidValue();
        return;
    }
    else
    {
        this->_minYear->hintValidValue();
    }

    if
        (year != this->_editedSettings.timelineCfg.getMinYear())
    {
        this->_editedSettings.timelineCfg.setMinYear(year);

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateMaxYear(double value)
{
    std::uint16_t year {0u};    //
    bool autoMaxYear {false};   //

    year = static_cast<std::uint16_t>(value);
    autoMaxYear = (0u == year);
    if
        (true == autoMaxYear)
    {
        this->_maxYear->hintValidValue();
    }

    if
        (autoMaxYear != this->_editedSettings.timelineCfg.getUseAutoMaxYear())
    {
        this->_editedSettings.timelineCfg.setUseAutoMaxYear(autoMaxYear);

        this->sendChangeNotification();
    }
    else if
        (false == autoMaxYear)
    {
        if
            (  (TimelineCfg::MIN_VALID_YEAR > year)
            || (TimelineCfg::MAX_VALID_YEAR < year))
        {
            this->_maxYear->hintInvalidValue();
            return;
        }
        else
        {
            this->_maxYear->hintValidValue();
        }

        if
            (year != this->_editedSettings.timelineCfg.getMaxYear())
        {
            this->_editedSettings.timelineCfg.setMaxYear(year);

            this->sendChangeNotification();
        }
    }
    this->checkRestartRequirement();
}

/**
 *
 */
bool NanoguiAppSettingsEditor::updateTimespan(const std::string& identifier, const std::string& newLabel, double newPast, double newFuture, const std::string& newDescription, bool newDefault)
{
    bool found {false}; //
    std::string key {""};   //

    if
        (0 >= identifier.size())
    {
        key = newLabel; // We will look for timespans that already have the same label
    }
    else
    {
        key = identifier;
    }

    for
        (auto& timespan : this->_editedSettings.timelineCfg.timeframes)
    {
        if
            (timespan.label != key)
        {
            continue;
        }

        if
            (0 < identifier.size())
        {
            // when identifier is not empty it means that we should update an
            // existing timespan
            timespan.label = newLabel;
            timespan.description = newDescription;
            timespan.pastOffset = static_cast<timestamp_t>(newPast);
            timespan.futureOffset = static_cast<timestamp_t>(newFuture);
        }

        found = true;
    }

    if
        (0 >= identifier.size())
    {
        if
           (true == found)
        {
            // an empty identifier means that this is a new timespan but there
            // is already one with the same label
            return false;
        }
        else
        {
            this->_editedSettings.timelineCfg.addTimeframe(newLabel,
                                                           newDescription,
                                                           static_cast<timestamp_t>(newPast),
                                                           static_cast<timestamp_t>(newFuture));
            return true;
        }
    }

    if
        (true == newDefault)
    {
        this->_editedSettings.timelineCfg.defaultTimespan = newLabel;
    }

    return found;
}

/**
 *
 */
bool NanoguiAppSettingsEditor::updateTimestep(const std::string& identifier, const std::string& newLabel, double newValue, const std::string& newDescription, bool newDefault)
{
    bool found {false}; //
    std::string key {""};   //

    if
        (0 >= identifier.size())
    {
        key = newLabel; // We will look for timesteps that already have the same label
    }
    else
    {
        key = identifier;
    }

    for
        (auto& timestep : this->_editedSettings.timelineCfg.timesteps)
    {
        if
            (timestep.label != key)
        {
            continue;
        }

        if
            (0 < identifier.size())
        {
            // when identifier is not empty it means that we should update an
            // existing timespan
            timestep.label = newLabel;
            timestep.description = newDescription;
            timestep.value = static_cast<timestamp_t>(newValue);
        }

        found = true;
    }

    if
        (0 >= identifier.size())
    {
        if
           (true == found)
        {
            // an empty identifier means that this is a new timespan but there
            // is already one with the same label
            return false;
        }
        else
        {
            this->_editedSettings.timelineCfg.addTimestep(newLabel,
                                                          newDescription,
                                                          static_cast<timestamp_t>(newValue));
            return true;
        }
    }

    if
        (true == newDefault)
    {
        this->_editedSettings.timelineCfg.defaultTimestep = newLabel;
    }

    return found;
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateNearest(bool value)
{
    if
        (value != this->_editedSettings.timelineCfg.useNearestMode)
    {
        this->_editedSettings.timelineCfg.useNearestMode = value;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateAdvancedScrolling(bool value)
{
    if
        (value != this->_editedSettings.timelineCfg.advancedScrolling)
    {
        this->_editedSettings.timelineCfg.advancedScrolling = value;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
bool NanoguiAppSettingsEditor::updateInitialDatetime(const std::string& value)
{
    if
        (value != this->_editedSettings.timelineCfg.getInitialDatetimeString())
    {
        // The value can be empty to indicate the use of the current date and time.
        if
            (true == value.empty())
        {
            this->_editedSettings.timelineCfg.setInitialDatetimeString(value);
            this->sendChangeNotification();
            this->checkRestartRequirement();
            this->_initialDatetime->hintValidValue();
        }
        else
        {
            bool ok{ false };    //
            timestamp_t checkTimestamp{ 0u }; //

            ok = SEAScope::getTimestampFromDatetime(value, checkTimestamp);
            if
                (true == ok)
            {
                this->_editedSettings.timelineCfg.setInitialDatetimeString(value);

                this->sendChangeNotification();

                this->checkRestartRequirement();

                this->_initialDatetime->hintValidValue();
            }
            else
            {
                this->_initialDatetime->hintInvalidValue();
            }
        }
    }
    else
    {
        this->_initialDatetime->hintValidValue();
    }

    return true;
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateDrawingAutoExit(bool checked)
{
    this->_drawingManualExit->setChecked(!checked);

    if
        (checked == this->_editedSettings.explicitDrawingModeExit)  // inverted logic
    {
        this->_editedSettings.explicitDrawingModeExit = !checked;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateDrawingManualExit(bool checked)
{
    this->_drawingAutoExit->setChecked(!checked);

    if
        (checked != this->_editedSettings.explicitDrawingModeExit)
    {
        this->_editedSettings.explicitDrawingModeExit = checked;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateClickTolerance(double value)
{
    std::uint32_t tolerance {0u};   //

    tolerance = static_cast<std::uint32_t>(value);
    if
        (tolerance != this->_editedSettings.clickPixelsTolerance)
    {
        this->_editedSettings.clickPixelsTolerance = tolerance;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateProcessorEnabled(bool checked)
{
    this->_processorAddress->setEnabled(checked);
    this->_processorAddress->setEditable(checked);
    this->_processorPort->setEnabled(checked);
    this->_processorPort->setEditable(checked);

    if
        (checked != this->_editedSettings.processorEnabled)
    {
        this->_editedSettings.processorEnabled = checked;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
bool NanoguiAppSettingsEditor::updateProcessorAddress(const std::string& value)
{
    if
        (value != this->_editedSettings.processorAddress)
    {
        this->_editedSettings.processorAddress = value;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();

    return true;
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateProcessorPort(double value)
{
    std::int64_t port {0};  //

    port = static_cast<std::int64_t>(value);
    if
        (port == this->_editedSettings.processorPort)
    {
        this->_editedSettings.processorPort = port;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateIndexCheck(bool value)
{
    if
        (value == this->_editedSettings.skipIndexCheck) // inverted logic
    {
        this->_editedSettings.skipIndexCheck = !value;

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateMinSearchDelay(double value)
{
    if
        (!floating_point_equality(value, 0.001 * this->_editedSettings.searchRequestMinDelay))
    {
        this->_editedSettings.searchRequestMinDelay = 0.001 * value; // convert to seconds

        this->sendChangeNotification();
    }

    this->checkRestartRequirement();
}

/**
 *
 */
void NanoguiAppSettingsEditor::updateBackgroundPath(const std::string& value)
{
    if
        (value != this->_editedSettings.worldMap)
    {
        // This file is not generated by SEAScope so it has to exist to be
        // valid (and must not be a directory)
        if
            (  true == Path::fileExists(value)
            && false == Path::dirExists(value))
        {
            this->_editedSettings.worldMap = value;

            this->sendChangeNotification();

            this->checkRestartRequirement();

            this->_backgroundPath->hintValidPath();
        }
        else
        {
            this->_backgroundPath->hintInvalidPath();
        }
    }
    else
    {
        this->_backgroundPath->hintValidPath();
    }
}

/**
 *
 */
void NanoguiAppSettingsEditor::checkRestartRequirement()
{
    if
        (true == this->_muted)
    {
        return;
    }

    /*
    Changes applied immediately:
    - processorEnabled;
    - processorAddress;
    - processorPort;
    - explicitDrawingModeExit;
    - clickPixelsTolerance;
    - searchRequestMinDelay;
    */

    bool noChange {true};   //

    noChange = (  (this->_settings.colormapsPath == this->_editedSettings.colormapsPath)
               && (this->_settings.dataPath == this->_editedSettings.dataPath)
               && (this->_settings.annotationsPath == this->_editedSettings.annotationsPath)
               && (this->_settings.statePath == this->_editedSettings.statePath)
               && (this->_settings.indexPath == this->_editedSettings.indexPath)
               && (this->_settings.skipIndexCheck == this->_editedSettings.skipIndexCheck)
               && (this->_settings.windowWidth == this->_editedSettings.windowWidth)
               && (this->_settings.windowHeight == this->_editedSettings.windowHeight)
               && (this->_settings.vsync == this->_editedSettings.vsync)
               && (this->_settings.antialiasing == this->_editedSettings.antialiasing)
               && (this->_settings.antialiasingFactor == this->_editedSettings.antialiasingFactor)
               && (this->_settings.fullscreen == this->_editedSettings.fullscreen)
               && (this->_settings.worldMap == this->_editedSettings.worldMap)
               && (true == SEAScope::floating_point_equality(this->_settings.minGranuleScreenEstate, this->_editedSettings.minGranuleScreenEstate))
               && (this->_settings.timelineCfg.getMinYear() == this->_editedSettings.timelineCfg.getMinYear())
               && (this->_settings.timelineCfg.defaultTimespan == this->_editedSettings.timelineCfg.defaultTimespan)
               && (this->_settings.timelineCfg.defaultTimestep == this->_editedSettings.timelineCfg.defaultTimestep)
               && (this->_settings.timelineCfg.useNearestMode == this->_editedSettings.timelineCfg.useNearestMode)
               && (this->_settings.timelineCfg.advancedScrolling == this->_editedSettings.timelineCfg.advancedScrolling)
               && (this->_settings.timelineCfg.getInitialDatetimeString() == this->_editedSettings.timelineCfg.getInitialDatetimeString())
               && (this->_settings.timelineCfg.getUseAutoMaxYear() == this->_editedSettings.timelineCfg.getUseAutoMaxYear())
               && (this->_settings.timelineCfg.timeframes.size() == this->_editedSettings.timelineCfg.timeframes.size())
               && (this->_settings.timelineCfg.timesteps.size() == this->_editedSettings.timelineCfg.timesteps.size())
               && (this->_settings.customPath == this->_editedSettings.customPath)
               );

    if
        (  (true == noChange)
           && false == this->_settings.timelineCfg.getUseAutoMaxYear())
    {
        noChange = (this->_settings.timelineCfg.getMaxYear() == this->_editedSettings.timelineCfg.getMaxYear());
    }

    std::vector<std::string> commonLabels;
    std::vector<std::string> descriptions;
    std::vector<std::string> originalLabels;
    std::vector<std::string> editedLabels;

    // Make sure that both settings have the same labels for timespans
    if
        (true == noChange)
    {
        this->_settings.timelineCfg.getTimeframes(originalLabels, descriptions);
        this->_editedSettings.timelineCfg.getTimeframes(editedLabels, descriptions);

        std::sort(originalLabels.begin(), originalLabels.end());
        std::sort(editedLabels.begin(), editedLabels.end());

        std::set_intersection(originalLabels.begin(), originalLabels.end(),
                              editedLabels.begin(), editedLabels.end(),
                              std::back_inserter(commonLabels));

        noChange = (commonLabels.size() == originalLabels.size());
    }

    // Make sure that the offsets and descriptions match for timespans
    if
        (true == noChange)
    {
        for
            (const auto& originalTimespan : this->_settings.timelineCfg.timeframes)
        {
            for
                (const auto& editedTimespan : this->_editedSettings.timelineCfg.timeframes)
            {
                if
                    (originalTimespan.label == editedTimespan.label)
                {
                    if
                        (originalTimespan.pastOffset != editedTimespan.pastOffset)
                    {
                        noChange = false;
                        break;
                    }

                    if
                        (originalTimespan.futureOffset != editedTimespan.futureOffset)
                    {
                        noChange = false;
                        break;
                    }

                    if
                        (originalTimespan.description != editedTimespan.description)
                    {
                        noChange = false;
                        break;
                    }
                }
            }

            if
                (false == noChange)
            {
                break;
            }
        }
    }

    // Make sure that both settings have the same labels for timesteps
    if
        (true == noChange)
    {
        commonLabels.clear();
        descriptions.clear();
        originalLabels.clear();
        editedLabels.clear();
        this->_settings.timelineCfg.getTimesteps(originalLabels, descriptions);
        this->_editedSettings.timelineCfg.getTimesteps(editedLabels, descriptions);

        std::sort(originalLabels.begin(), originalLabels.end());
        std::sort(editedLabels.begin(), editedLabels.end());

        std::set_intersection(originalLabels.begin(), originalLabels.end(),
                              editedLabels.begin(), editedLabels.end(),
                              std::back_inserter(commonLabels));

        noChange = (commonLabels.size() == originalLabels.size());
    }

    // Make sure that the offsets and descriptions match for timesteps
    if
        (true == noChange)
    {
        for
            (const auto& originalTimestep : this->_settings.timelineCfg.timesteps)
        {
            for
                (const auto& editedTimestep : this->_editedSettings.timelineCfg.timesteps)
            {
                if
                    (originalTimestep.label == editedTimestep.label)
                {
                    if
                        (originalTimestep.value != editedTimestep.value)
                    {
                        noChange = false;
                        break;
                    }

                    if
                        (originalTimestep.description != editedTimestep.description)
                    {
                        noChange = false;
                        break;
                    }
                }
            }

            if
                (false == noChange)
            {
                break;
            }
        }
    }

    if
        (this->_restartNeededMessage->visible() == noChange)
    {
        this->_restartNeededMessage->setVisible(!noChange);
        this->reflowNeeded = true;
    }
}

/**
 *
 */
void NanoguiAppSettingsEditor::loadSettings(const Settings& settings)
{
    // Make sure no change event is emitted while assigning the values.
    // There should be no code path exiting the function before _muted is set
    // back to false, otherwise no event will be emitted afterwards
    this->_muted = true;

    // Paths
    this->_colormapsPath->setPath(settings.colormapsPath);
    this->_dataPath->setPath(settings.dataPath);
    this->_statePath->setPath(settings.statePath);
    this->_indexPath->setPath(settings.indexPath);
    this->_annotationsPath->setPath(settings.annotationsPath);
    this->_customPath->setPath(settings.customPath);

    // Rendering
    this->_fullscreenCheck->setChecked(settings.fullscreen);
    this->_windowedCheck->setChecked(!settings.fullscreen);
    this->_windowWidth->setEnabled(!settings.fullscreen);
    this->_windowWidth->setEditable(!settings.fullscreen);
    this->_windowWidth->setValue(settings.windowWidth);
    this->_windowHeight->setEnabled(!settings.fullscreen);
    this->_windowHeight->setEditable(!settings.fullscreen);
    this->_windowHeight->setValue(settings.windowHeight);
    this->_vsyncCheck->setChecked(settings.vsync);
    this->_antialiasingCheck->setChecked(settings.antialiasing);
    this->_msaaSampling->setValue(static_cast<double>(settings.antialiasingFactor));
    this->_msaaSampling->setEnabled(settings.antialiasing);
    this->_msaaSampling->setEditable(settings.antialiasing);
    this->_minScreenPortion->setValue(static_cast<double>(100 * settings.minGranuleScreenEstate));  // Convert to percentage

    // Timeline
    this->_minYear->setValue(static_cast<double>(settings.timelineCfg.getMinYear()));
    if
        (true == settings.timelineCfg.getUseAutoMaxYear())
    {
        this->_maxYear->setValue(0);
    }
    else
    {
        this->_maxYear->setValue(static_cast<double>(settings.timelineCfg.getMaxYear()));
    }
    for
        (const auto& timestep : settings.timelineCfg.timesteps)
    {
        NanoguiTimestepEditor* editor {nullptr};
        editor = this->buildTimestepEditor();
        editor->set(timestep.label, timestep.description, timestep.value, timestep.label == settings.timelineCfg.defaultTimestep);
    }
    this->_timesteps->update(true);

    for
        (const auto& timespan : settings.timelineCfg.timeframes)
    {
        NanoguiTimespanEditor* editor {nullptr};
        editor = this->buildTimespanEditor();
        editor->set(timespan.label, timespan.description, timespan.pastOffset, timespan.futureOffset, timespan.label == settings.timelineCfg.defaultTimespan);
    }
    this->_timespans->update(true);

    this->_nearestCheck->setChecked(settings.timelineCfg.useNearestMode);
    this->_advancedScrollingCheck->setChecked(settings.timelineCfg.advancedScrolling);
    this->_initialDatetime->setValue(settings.timelineCfg.getInitialDatetimeString());

    // Drawing
    this->_drawingAutoExit->setChecked(!settings.explicitDrawingModeExit);
    this->_drawingManualExit->setChecked(settings.explicitDrawingModeExit);
    this->_clickTolerance->setValue(static_cast<double>(settings.clickPixelsTolerance));

    // Processor
    this->_processorCheck->setChecked(settings.processorEnabled);
    this->_processorAddress->setValue(settings.processorAddress);
    this->_processorAddress->setEnabled(settings.processorEnabled);
    this->_processorAddress->setEditable(settings.processorEnabled);
    this->_processorPort->setValue(static_cast<double>(settings.processorPort));
    this->_processorPort->setEnabled(settings.processorEnabled);
    this->_processorPort->setEditable(settings.processorEnabled);

    // Misc
    this->_indexCheck->setChecked(!settings.skipIndexCheck);
    this->_minSearchDelay->setValue(1000 * settings.searchRequestMinDelay); // Convert to milliseconds
    this->_backgroundPath->setPath(settings.worldMap);

    this->_settings = settings; // copy
    this->_editedSettings = settings;  // copy

    this->_muted = false;
}

/**
 *
 */
void NanoguiAppSettingsEditor::setChangeCallback(const std::function<void(const Settings&)>& callback)
{
    this->_changeCallback = callback;
}

/**
 *
 */
void NanoguiAppSettingsEditor::sendChangeNotification()
{
    if
        (true == this->_muted)
    {
        return;
    }

    if
        (nullptr != this->_changeCallback)
    {
        this->_changeCallback(this->_editedSettings);
    }
}

/**
 *
 */
void NanoguiAppSettingsEditor::draw(NVGcontext* ctx)
{
    bool somethingRemoved {false};  //

    if
        (0 < this->_timespanEditorsToRemove.size())
    {
        // As it is part of the editor to remove, clicking on the delete
        // button puts the editor in the focus path of the screen.
        // The focus path is iterated upon later on, which will trigger a
        // segfault if one of the component has been removed, so it is
        // necessary to clear the focus path before removing the editor.
        this->screen()->updateFocus(nullptr);

        for
            (auto editor : this->_timespanEditorsToRemove)
        {
            this->_timespans->content->removeChild(editor);
        }

        this->_timespans->update(true);

        somethingRemoved = true;
        this->_timespanEditorsToRemove.clear();
    }

    if
        (0 < this->_timestepEditorsToRemove.size())
    {
        // As it is part of the editor to remove, clicking on the delete
        // button puts the editor in the focus path of the screen.
        // The focus path is iterated upon later on, which will trigger a
        // segfault if one of the component has been removed, so it is
        // necessary to clear the focus path before removing the editor.
        this->screen()->updateFocus(nullptr);

        for
            (auto editor : this->_timestepEditorsToRemove)
        {
            this->_timesteps->content->removeChild(editor);
        }

        this->_timesteps->update(true);

        somethingRemoved = true;
        this->_timestepEditorsToRemove.clear();
    }

    if
        (true == somethingRemoved)
    {
        this->sendChangeNotification();
        this->checkRestartRequirement();
    }

    this->hideDropdownsWithHiddenAnchors();

    NanoguiScrollableWithTitle::draw(ctx);
}

/**
 *
 */
void NanoguiAppSettingsEditor::hideDropdownsWithHiddenAnchors()
{
    int topThreshold =  this->scrollable->absolutePosition().y() + static_cast<int>(this->scrollable->getScroll());
    int bottomThreshold = this->scrollable->absolutePosition().y() + this->scrollable->size().y();
    constexpr int browseButtonTop {4};
    constexpr int browseButtonBottom {24};

    if
        (  (this->_colormapsPath->absolutePosition().y() + browseButtonBottom < topThreshold)
        || (this->_colormapsPath->absolutePosition().y() + browseButtonTop > bottomThreshold))
    {
        this->_colormapsPath->hideDropDown();
    }
    if
        (  (this->_dataPath->absolutePosition().y() + browseButtonBottom < topThreshold)
        || (this->_dataPath->absolutePosition().y() + browseButtonTop > bottomThreshold))
    {
        this->_dataPath->hideDropDown();
    }
    if
        (  (this->_statePath->absolutePosition().y() + browseButtonBottom < topThreshold)
        || (this->_statePath->absolutePosition().y() + browseButtonTop > bottomThreshold))
    {
        this->_statePath->hideDropDown();
    }
    if
        (  (this->_indexPath->absolutePosition().y() + browseButtonBottom < topThreshold)
        || (this->_indexPath->absolutePosition().y() + browseButtonTop > bottomThreshold))
    {
        this->_indexPath->hideDropDown();
    }
    if
        (  (this->_annotationsPath->absolutePosition().y() + browseButtonBottom < topThreshold)
        || (this->_annotationsPath->absolutePosition().y() + browseButtonTop > bottomThreshold))
    {
        this->_annotationsPath->hideDropDown();
    }
    if
        (  (this->_customPath->absolutePosition().y() + browseButtonBottom < topThreshold)
        || (this->_customPath->absolutePosition().y() + browseButtonTop > bottomThreshold))
    {
        this->_customPath->hideDropDown();
    }
    if
        (  (this->_backgroundPath->absolutePosition().y() + browseButtonBottom < topThreshold)
        || (this->_backgroundPath->absolutePosition().y() + browseButtonTop > bottomThreshold))
    {
        this->_backgroundPath->hideDropDown();
    }
}
