// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/*
*/

#include "vendors/_nanogui.h"

#include "core/log.h"
#include "core/utils.h"
#include <ui/nanogui/nanogui_blueprint.h>

using namespace SEAScope;

NanoguiBlueprint::NanoguiBlueprint( nanogui::Screen* parent
								  , bool resizable, bool movableIn)
: Widget{parent}, screen{parent}, target{nullptr}, targetIsVisible{false}, resizeHandle{nullptr}, movable{movableIn}, editing{false}, edited{false}
{
	if
		(true == resizable)
	{
		this->resizeHandle = new NanoguiResizeHandle(parent);
	}
};

/**
 *
 */
NanoguiBlueprint::~NanoguiBlueprint()
{
};

/**
 *
 */
bool NanoguiBlueprint::setMinDimensions(const std::string& minWidthIn, const std::string& minHeightIn)
{
	this->minWidth = minWidthIn;
	this->minHeight = minHeightIn;

	if
		(nullptr != this->resizeHandle)
	{
		this->resizeHandle->setTarget(this, this->minWidth, this->minHeight);
	}
	return true;
};

/**
 *
 */
void NanoguiBlueprint::draw(NVGcontext *ctx) {
	int cr = mTheme->mWindowCornerRadius;

    /* Draw window */
    nvgSave(ctx);
    nvgBeginPath(ctx);
    nvgRoundedRect(ctx, static_cast<float>(mPos.x()), static_cast<float>(mPos.y()), static_cast<float>(mSize.x()), static_cast<float>(mSize.y()), static_cast<float>(cr));

    nvgFillColor(ctx, mMouseFocus ? mTheme->mWindowFillFocused
                                  : mTheme->mWindowFillUnfocused);
    nvgFill(ctx);

    nvgRestore(ctx);
    nanogui::Widget::draw(ctx);
};

/**
 *
 */
bool NanoguiBlueprint::startEditing()
{
	this->editing = true;
	this->setVisible(true);
	LOG(LogLevel::Debug, "NanoguiBlueprint", "Start editing");
	return true;
};

/**
 *
 */
bool NanoguiBlueprint::stopEditing()
{
	bool ok;	//

	this->editing = false;
	this->setVisible(false);
	ok = this->edited;
	this->edited = false;
	return ok;
};

bool NanoguiBlueprint::mouseDragEvent(const Eigen::Vector2i &, const Eigen::Vector2i &rel,
                            int button, int /* modifiers */) {
    if
		(  this->movable && this->editing 
	    && (button & (1 << GLFW_MOUSE_BUTTON_1)) != 0)
	{
        mPos += rel;
        mPos = mPos.cwiseMax(Eigen::Vector2i::Zero());
        mPos = mPos.cwiseMin(parent()->size() - mSize);
		if
			(nullptr != this->resizeHandle)
		{
			this->resizeHandle->updatePosition();
		}
		this->edited = true;
        return true;
    }
    return false;
};

/**
 *
 */
void NanoguiBlueprint::setVisible(bool visible)
{
	nanogui::Widget::setVisible(visible);
	if
		(nullptr != this->resizeHandle)
	{
		this->resizeHandle->setVisible(visible);
	}

	if
		(nullptr != this->target)
	{
		if
			(false == visible)
		{
			this->target->setVisible(this->targetIsVisible);
		}
		else
		{
			this->targetIsVisible = this->target->visible();
			this->target->setVisible(false);
		}
	}
};

/**
 *
 */
bool NanoguiBlueprint::setTarget(IComponent* targetIn)
{
	if
		(this->target == targetIn)
	{
		return false;
	}

	this->target = targetIn;
	this->targetIsVisible = target->visible();
	this->resizeAuto();

	if
		(nullptr != this->target)
	{
		this->target->setVisible(!this->editing);
	}
	return true;
};

/**
 *
 */
IComponent* NanoguiBlueprint::getTarget() const
{
	return this->target;
};

/**
 *
 */
bool NanoguiBlueprint::applyToTarget()
{
	if
		(nullptr == this->target)
	{
		return false;
	}

	this->target->setPosition(this->mPos.x(), this->mPos.y());
	this->target->setSize(this->mSize.x(), this->mSize.y());
	this->target->requestReflow();

	if
		(nullptr != this->screen)
	{
		this->screen->performLayout();
	}
	return true;
};

/**
 *
 */
bool NanoguiBlueprint::updateSize(const Eigen::Vector2i& size)
{
	int newW;	//
	int newH;	//
	int minW;	//
	int minH;	//
	int parentW;	//
	int parentH;	//

	if
		(  true == this->editing
		&& (  this->mSize.x() != size.x()
		   || this->mSize.y() != size.y()))
	{
		this->edited = true;
	}
	newW = size.x();
	newH = size.y();

	if
		(false == this->minWidth.empty())
	{
		parentW = this->parent()->size().x();
		minW = dimStr2pixels(this->minWidth, parentW);
		newW = (newW > minW) ? newW:minW;
	}

	if
		(false == this->minHeight.empty())
	{
		parentH = this->parent()->size().y();
		minH = dimStr2pixels(this->minHeight, parentH);
		newH = (newH > minH) ? newH:minH;
	}

	nanogui::Widget::setSize(Eigen::Vector2i(newW, newH));

	if
		(nullptr != this->resizeHandle)
	{
		this->resizeHandle->updatePosition();
	}
	return true;
};

/**
 *
 */
bool NanoguiBlueprint::resizeAuto()
{
	int x;	//
	int y;	//
	int w;	//
	int h;	//
	int parentW;	//
	int parentH;	//
	int minW;	//
	int minH;	//
	bool ok;	//

	if
		(nullptr == this->target)
	{
		return false;
	}

	parentW = this->parent()->size().x();
	parentH = this->parent()->size().y();
	w = dimStr2pixels(this->storedWidth, parentW);
	h = dimStr2pixels(this->storedHeight, parentH);
	x = posStr2pixels(this->storedX, parentW, w);
	y = posStr2pixels(this->storedY, parentH, h);
	minW = dimStr2pixels(this->minWidth, parentW);
	minH = dimStr2pixels(this->minHeight, parentH);
	w = (w > minW) ? w:minW;
	h = (h > minH) ? h:minH;

	w = w + this->target->getWidthOffset();
	h = h + this->target->getHeightOffset();

	LOG(LogLevel::Debug, "Blueprint", "resizeAuto {} {} {}>={}/{} {}>={}/{}", x, y, w, this->minWidth, parentW, h, this->minHeight, parentH);

	this->setPosition(Eigen::Vector2i(x,y));
	this->setSize(Eigen::Vector2i(w,h));

	if
		(nullptr != this->resizeHandle)
	{
		this->resizeHandle->updatePosition();
	}

	ok = this->applyToTarget();
	return ok;
};

/**
 *
 */
bool NanoguiBlueprint::getLayout(std::string& x, std::string& y, std::string& w, std::string& h) const
{
	x = this->storedX;
	y = this->storedY;
	w = this->storedWidth;
	h = this->storedHeight;
	return true;
};

/**
 *
 */
bool NanoguiBlueprint::setLayout(const std::string& x, const std::string& y, const std::string& w, const std::string& h)
{
	this->storedX = x;
	this->storedY = y;
	this->storedWidth = w;
	this->storedHeight = h;
	this->resizeAuto();
	return true;
};
