// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_button_bar.h"

using namespace SEAScope;

/**
 *
 */
NanoguiButtonBar::NanoguiButtonBar(nanogui::Widget* parent): nanogui::Widget(parent)
{
	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Horizontal
										   , nanogui::Alignment::Middle
										   , 0, 1));
};

/**
 *
 */
NanoguiButtonBar::~NanoguiButtonBar()
{

};

/**
 *
 */
bool NanoguiButtonBar::init( const std::vector<std::string>& values
						   , const std::vector<std::string>& tooltips
						   , uint8_t fontSize
						   , const Eigen::Vector2i& buttonSize)
{
	nanogui::Button* btn {nullptr};	//
	std::vector<std::string>::const_iterator valuesIt; //
	std::vector<std::string>::const_iterator tooltipsIt; //
	std::vector<nanogui::Button*>::iterator buttonsIt;	//

	if
		(values.size() != tooltips.size())
	{
		LOG( LogLevel::Error, "Button bar"
				, "Each button must have a value AND a tooltip, but the number of values ({}) and tooltips ({}) do not match"
				, values.size(), tooltips.size());
		return false;
	}

	valuesIt = values.begin();
	tooltipsIt = tooltips.begin();
	for
		(; valuesIt != values.end(); ++valuesIt, ++tooltipsIt)
	{
		btn = new nanogui::Button(this, *valuesIt);
		btn->setFlags(nanogui::Button::RadioButton);
		btn->setFixedSize(buttonSize);
		btn->setFontSize(fontSize);
		btn->setTooltip(*tooltipsIt);
		btn->setChangeCallback([this](bool selected) {
			if
				(selected)
			{
				this->selectionChanged();
			}
		});
		this->buttons.push_back(btn);
	}

	buttonsIt = this->buttons.begin();
	for
		(;buttonsIt != this->buttons.end(); ++buttonsIt)
	{
		(*buttonsIt)->setButtonGroup(this->buttons);
	}

	return true;
};

/**
 *
 */
const std::string NanoguiButtonBar::selectedValue() const
{
	auto buttonsIt = this->buttons.begin();
	for
		(;buttonsIt != this->buttons.end(); buttonsIt++)
	{
		if
			( (*buttonsIt)->pushed() )
		{
			return (*buttonsIt)->caption();
		}
	}
	return "";
};

/**
 *
 */
bool NanoguiButtonBar::selectValue(const std::string& value)
{
	std::vector<nanogui::Button*>::iterator buttonsIt;	//
	bool pushed;	//
	bool found;	//

	found = false;
	buttonsIt = this->buttons.begin();
	for
		(;buttonsIt != this->buttons.end(); ++buttonsIt)
	{
		pushed = (value == (*buttonsIt)->caption());
		found = found || pushed;
		(*buttonsIt)->setPushed(pushed);
	}
	return found;

};

/**
 *
 */
bool NanoguiButtonBar::setSelectionChangedCallback(const std::function<void(const std::string&)>& callback)
{
	this->selectionChangedCallback = callback;
	return true;
};

/**
 *
 */
void NanoguiButtonBar::selectionChanged()
{
	LOG( LogLevel::Debug, "NanoguiButtonBar", "Selection changed to: {}", this->selectedValue());

	if
		(nullptr != this->selectionChangedCallback)
	{
		this->selectionChangedCallback(this->selectedValue());
	}
};
