// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>

#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_calendars_timeline.h"

using namespace SEAScope;

/**
 *
 */
NanoguiCalendarsTimeline::NanoguiCalendarsTimeline(nanogui::Widget* parent)
: nanogui::Widget(parent), minGap(3), minYear(1970), maxYear(9999)
, currentYear(1970), currentMonth(1), currentDay(1), visibleCount(0)
, centralYear(2000), calendarsPanel(nullptr), prevButton(nullptr)
, nextButton(nullptr), yearCallback(nullptr), monthCallback(nullptr)
, currentYearVisibilityChanged(nullptr), scrollCallback(nullptr)
, pastOffset(0), futureOffset(0), hoverEnabled(false)
{
	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Horizontal
										   , nanogui::Alignment::Fill
										   , 0, this->minGap));
	this->setFixedHeight(35);

	this->prevButton = new nanogui::Button(this, "", ENTYPO_ICON_CHEVRON_LEFT);
	this->prevButton->setFixedWidth(25);
	this->prevButton->setCallback([this]() {
		this->prevPage();
	});

	this->calendarsPanel = new nanogui::Widget(this);
	this->calendarsPanel->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Horizontal
														   , nanogui::Alignment::Fill
														   , 1, this->minGap));

	this->nextButton = new nanogui::Button(this, "", ENTYPO_ICON_CHEVRON_RIGHT);
	this->nextButton->setFixedWidth(25);
	this->nextButton->setCallback([this]() {
		this->nextPage();
	});

	this->resize(this->mSize.x());
};

/**
 * 
 */
NanoguiCalendarsTimeline::~NanoguiCalendarsTimeline()
{
};

/**
 *
 */
bool NanoguiCalendarsTimeline::setBounds(std::uint16_t minYearIn, std::uint16_t maxYearIn)
{
	bool changed;	//

	changed = false;

	if
		(minYearIn != this->minYear)
	{
		this->minYear = minYearIn;
		changed = true;
	}

	if
		(maxYearIn != this->maxYear)
	{
		this->maxYear = maxYearIn;
		changed = true;
	}

	if
		(true == changed)
	{
		this->centralYear = this->adaptYear(this->centralYear);
		this->updateCalendars();
		this->updateNavButtons();
	}

	return changed;
};

/**
 *
 */
void NanoguiCalendarsTimeline::setSize(const Eigen::Vector2i& size)
{
	nanogui::Widget::setSize(size);
	this->resize(size.x());
};

/**
 *
 */
void NanoguiCalendarsTimeline::setWidth(int width)
{
	nanogui::Widget::setWidth(width);
	this->resize(width);
};

/**
 *
 */
void NanoguiCalendarsTimeline::setFixedSize(const Eigen::Vector2i& size)
{
	nanogui::Widget::setFixedSize(size);
	this->resize(size.x());
};

/**
 *
 */
void NanoguiCalendarsTimeline::setFixedWidth(int width)
{
	nanogui::Widget::setFixedWidth(width);
	this->resize(width);
};

/**
 *
 */
void NanoguiCalendarsTimeline::draw(NVGcontext* ctx)
{
    if
        (this->hoverEnabled && this->mMouseFocus)
    {
        nvgBeginPath(ctx);
        nvgRect(ctx, static_cast<float>(this->mPos.x() + 1), static_cast<float>(this->mPos.y() + 1),
                static_cast<float>(this->mSize.x() - 2), static_cast<float>(this->mSize.y() - 2));
        nvgFillColor(ctx, nanogui::Color(20, 20, 20, 100));
        nvgFill(ctx);
    }
	nanogui::Widget::draw(ctx);
};

/**
 *
 */
bool NanoguiCalendarsTimeline::setMainCoverage(const EventsCoverage& coverage)
{
	this->mainCoverage = coverage;
	this->updateCalendars();
	return true;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::setExtraCoverage(const EventsCoverage& coverage)
{
	this->extraCoverage = coverage;
	this->updateCalendars();
	return true;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::setCurrent(std::uint16_t year, std::uint8_t month, std::uint8_t day, bool recenter)
{
	bool ok;	//
	std::uint16_t newYear;	//
    bool currentYearWasInRange; //
    bool currentYearIsInRange;  //

    currentYearWasInRange = this->isYearVisible(this->currentYear);

	LOG( LogLevel::Debug, "NanoguiCalendarsTimeline"
			, "setCurrent: {}-{}-{}"
			, static_cast<int>(year), static_cast<int>(month), static_cast<int>(day));

	this->currentYear = year;
	this->currentMonth = month;
	this->currentDay = day;

	/* Make sure new current date is visible */
	if
		(  year < this->centralYear - static_cast<uint16_t>(std::floor(.5f * static_cast<float>(this->visibleCount)))
		|| year > this->centralYear + static_cast<uint16_t>(std::floor(.5f * static_cast<float>(this->visibleCount)))
		|| true == recenter)
	{
		/* Use current year as central year if it was not visible */
		newYear = this->adaptYear(year);
		this->setCentralYear(newYear);
	}

	ok = this->updateCalendars();
	ok = this->updateNavButtons();

    currentYearIsInRange = this->isYearVisible(this->currentYear);
    if
        (  currentYearWasInRange != currentYearIsInRange
        && nullptr != this->currentYearVisibilityChanged)
    {
        this->currentYearVisibilityChanged(currentYearIsInRange);
    }
	return ok;
};

/**
 *
 */
std::uint16_t NanoguiCalendarsTimeline::adaptYear(std::uint16_t value) const
{
	std::uint16_t newYear;	//
	std::uint16_t lowerSpan;	//
	std::uint16_t upperSpan;	//

	if
		(0 < (this->visibleCount % 2))
	{
		lowerSpan = static_cast<std::uint16_t>(this->visibleCount / 2);
		upperSpan = lowerSpan;
	}
	else
	{
		lowerSpan = static_cast<std::uint16_t>(this->visibleCount / 2);
		upperSpan = static_cast<std::uint16_t>(-1 + static_cast<int>(lowerSpan));
	}
	newYear = value;
	if
		(value - lowerSpan < this->minYear)
	{
		newYear = static_cast<uint16_t>(this->minYear + lowerSpan);
	}

	if
		(value + upperSpan > this->maxYear)
	{
		newYear = static_cast<uint16_t>(this->maxYear - upperSpan);
	}

	LOG( LogLevel::Debug, "NanoguiCalendarsTimeline"
			, "adaptYear: {} {} {} {}"
			, static_cast<int>(value), static_cast<int>(newYear), static_cast<int>(lowerSpan), static_cast<int>(upperSpan));
	return newYear;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::resize(int width)
{
	int emptyWidth;	//
	std::size_t count;	//
	std::size_t missingCalendars;	//
    int missingCalendarsTmp;    //
	NanoguiCalendar* calendar;	//
	const int calendarWidth = 72;	//
	int newGap;	//
	nanogui::BoxLayout* layout;	//
	std::vector<std::int16_t> years;	//
	bool ok;	//
	std::size_t i;	//
    double newGapTmp(0.0);  //
    std::size_t emptyWidthTmp(0);   //

	// 62 = 2 * (25 navbutton + minGap + 1 margin)
	emptyWidth = width - 2 * (25 + this->minGap + 1);
	if
		(0 >= emptyWidth)
	{
		return false;
	}
	this->calendarsPanel->setFixedWidth(emptyWidth);
	count = static_cast<std::size_t>(std::floor(emptyWidth / (this->minGap + calendarWidth)));
	missingCalendarsTmp = static_cast<int>(count) - static_cast<int>(this->calendarsPool.size());
    missingCalendars = static_cast<std::size_t>(std::max(0, missingCalendarsTmp));
	LOG(LogLevel::Debug, "NanoguiCalendarsView", "resize: {} {} {} {}", width, emptyWidth, count, missingCalendars);

	i = 0;
	for
		(; i < missingCalendars; ++i)
	{
		LOG(LogLevel::Debug, "NanoguiCalendarsTimeline", "Create calendar");
		calendar = new NanoguiCalendar(this->calendarsPanel);
		calendar->setFixedWidth(calendarWidth);
		calendar->monthCallback = [this] (std::uint16_t year, std::uint8_t month) {
			if
				(nullptr != this->monthCallback)
			{
				this->monthCallback(year, month);
			}
		};
		calendar->yearCallback = [this] (std::uint16_t year) {
			if
				(nullptr != this->yearCallback)
			{
				this->yearCallback(year);
			}
		};
		this->calendarsPool.push_back(calendar);
	}

	/* Hide calendars */
	for
		(auto& it : this->calendarsPool)
	{
		it->setVisible(false);
	}

	this->visibleCount = count;
	LOG(LogLevel::Debug, "NanoguiCalendarsTimeline", "{} calendars in pool", this->calendarsPool.size());

	if
		(1 < count)
	{
        emptyWidthTmp = static_cast<std::size_t>(emptyWidth);
		newGapTmp = std::floor((emptyWidthTmp - count * calendarWidth) / (count - 1));
        newGap = static_cast<int>(newGapTmp);
		newGap = (0 < newGap) ? newGap:0;
	}
	else
	{
		newGap = 0;
	}
	layout = static_cast<nanogui::BoxLayout*>(this->calendarsPanel->layout());
	layout->setSpacing(newGap);

	/* Central year may need an update to prevent display of invalid years */
	this->centralYear = this->adaptYear(this->centralYear);

	ok = this->updateCalendars();
	this->updateNavButtons();

	return ok;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::setVisibilityBounds(timestamp_t past, timestamp_t future)
{
	bool changed;	//

	changed = false;
	if
		(this->pastOffset != past)
	{
		this->pastOffset = past;
		changed = true;
	}

	if
		(this->futureOffset != future)
	{
		this->futureOffset = future;
		changed = true;
	}

	if
		(changed)
	{
		this->updateCalendars();
	}
	return changed;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::isYearVisible(const std::uint16_t year) const
{
    std::uint16_t minYearTmp;  //
    std::uint16_t maxYearTmp;  //
    std::uint16_t visibleYears(0);  //
    double visibleYearsTmp(0.0);    //
    double halfYearsExtentTmp(0.0); //
    std::uint16_t halfYearsExtent(0);   //
    bool ok;    //

    visibleYears = static_cast<std::uint16_t>(this->visibleCount);
    visibleYearsTmp = static_cast<double>(this->visibleCount);
    halfYearsExtentTmp = floor(.5 * visibleYearsTmp);
    halfYearsExtent = static_cast<std::uint16_t>(halfYearsExtentTmp);
	minYearTmp = static_cast<uint16_t>(this->centralYear - halfYearsExtent);
    maxYearTmp = static_cast<uint16_t>(minYearTmp + visibleYears);
    ok = (minYearTmp <= year) && (maxYearTmp > year);
    return ok;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::updateCalendars()
{
	NanoguiCalendar* calendar;	//
	std::vector<uint16_t> years;	//
	std::vector<uint16_t> extraYears;	//
	std::vector<uint8_t> months;	//
	std::uint16_t year;	//
	std::size_t i;	//
    std::uint8_t month(0); //
	bool calendarEnabled;	//
    double visibleYears(0.0);    //
    double halfYearsExtentTmp(0.0); //
    std::uint16_t halfYearsExtent(0);   //

	this->mainCoverage.getYears(years);
	this->extraCoverage.getYears(extraYears);

    visibleYears = static_cast<double>(this->visibleCount);
    halfYearsExtentTmp = floor(.5 * visibleYears);
    halfYearsExtent = static_cast<std::uint16_t>(halfYearsExtentTmp);
	year = static_cast<uint16_t>(this->centralYear - halfYearsExtent);
	i = 0;
	for
		(; i < this->visibleCount; ++i)
	{
		calendar = this->calendarsPool[i];
		calendar->setYear(year);

		calendarEnabled = this->minYear <= year && this->maxYear >= year;
		calendar->setVisible(calendarEnabled);

		auto yearFound = std::find( years.begin()
								  , years.end()
								  , year);

		/* Update global coverage */
		if
			(years.end() != yearFound)
		{
			calendar->setYearBgColor(nanogui::Color(0,0,0,0));
			calendar->setYearFgColor(nanogui::Color(250,250,250,255));
		}
		else
		{
			calendar->setYearBgColor(nanogui::Color(0,0,0,0));
			calendar->setYearFgColor(nanogui::Color(90,90,90,255));
		}

		months.clear();
		this->mainCoverage.getMonths(year, months);
		month = 1;
		for
			(; month < 13; ++month)
		{
			auto monthFound = std::find( months.begin()
									   , months.end()
									   , month);
			if
				(months.end() != monthFound)
			{
				calendar->setMonthBgColor(month, nanogui::Color(250,250,250,255));
			}
			else
			{
				calendar->setMonthBgColor(month, nanogui::Color(90,90,90,255));
			}
		}

		/* Update coloc coverage */
		yearFound = std::find( extraYears.begin()
							 , extraYears.end()
							 , year);
		if
			(extraYears.end() != yearFound)
		{
			calendar->setYearBgColor(nanogui::Color(0,0,0,0));
			calendar->setYearFgColor(nanogui::Color(30,210,50,255));
		}

		months.clear();
		this->extraCoverage.getMonths(year, months);
		month = 1;
		for
			(; month < 13; ++month)
		{
			auto monthFound = std::find(months.begin(), months.end(), month);
			if
				(months.end() != monthFound)
			{
				calendar->setMonthBgColor(month, nanogui::Color(30,150,50,255));
			}
		}

		/* Update selection */
		if
			(year == this->currentYear)
		{
			calendar->select();
		}
		else
		{
			calendar->unselect();
		}


		year = year + 1u;
	}

	return true;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::setCentralYear(std::uint16_t year)
{
	if
		(year != this->centralYear)
	{
		this->centralYear = year;
		return true;
	}
	return false;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::nextPage()
{
	bool ok;	//
    bool currentYearWasInRange; //
    bool currentYearIsInRange;  //
	std::uint16_t newYear;	//
    std::uint16_t visibleYears(0);  //

    visibleYears = static_cast<std::uint16_t>(this->visibleCount);
    currentYearWasInRange = this->isYearVisible(this->currentYear);
	newYear = static_cast<uint16_t>(this->centralYear + visibleYears);
	newYear = this->adaptYear(newYear);
	ok = this->setCentralYear(newYear);
	if
		(false == ok)
	{
		return ok;
	}

	ok = this->updateCalendars();
	ok = this->updateNavButtons();

    currentYearIsInRange = this->isYearVisible(this->currentYear);
    if
        (  currentYearWasInRange != currentYearIsInRange
        && nullptr != this->currentYearVisibilityChanged)
    {
        this->currentYearVisibilityChanged(currentYearIsInRange);
    }
	return ok;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::updateNavButtons()
{
	std::uint16_t lowerSpan;	//
	std::uint16_t upperSpan;	//
    std::int32_t lowerSpanTmp(0);    //
    std::int32_t upperSpanTmp(0);   //
    std::int32_t visibleYears(0);  //
	bool nextEnabled;	//
	bool prevEnabled;	//

    visibleYears = static_cast<std::int16_t>(this->visibleCount);
	if
		(0 < (visibleYears % 2))
	{
		lowerSpanTmp = visibleYears / 2;
		upperSpanTmp = lowerSpanTmp;
	}
	else
	{
		lowerSpanTmp = visibleYears / 2;
		upperSpanTmp = -1 + lowerSpanTmp;
	}
    lowerSpan = static_cast<std::uint16_t>(lowerSpanTmp);
    upperSpan = static_cast<std::uint16_t>(upperSpanTmp);
	nextEnabled =  this->centralYear + upperSpan < this->maxYear;
	prevEnabled = this->minYear < this->centralYear - lowerSpan;
	this->prevButton->setEnabled(prevEnabled);
	this->nextButton->setEnabled(nextEnabled);
	return true;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::prevPage()
{
	bool ok;	//
    bool currentYearWasInRange; //
    bool currentYearIsInRange;  //
	std::uint16_t newYear;	//
    std::uint16_t visibleYears(0);  //

    visibleYears = static_cast<std::uint16_t>(this->visibleCount);
    currentYearWasInRange = this->isYearVisible(this->currentYear);
	newYear = static_cast<uint16_t>(this->centralYear - visibleYears);
	newYear = this->adaptYear(newYear);
	ok = this->setCentralYear(newYear);
	if
		(false == ok)
	{
		return ok;
	}

	ok = this->updateCalendars();
	ok = this->updateNavButtons();

    currentYearIsInRange = this->isYearVisible(this->currentYear);
    if
        (  currentYearWasInRange != currentYearIsInRange
        && nullptr != this->currentYearVisibilityChanged)
    {
        this->currentYearVisibilityChanged(currentYearIsInRange);
    }

	return ok;
};

/**
 *
 */
bool NanoguiCalendarsTimeline::scrollEvent(const Eigen::Vector2i& /*p*/,
                                           const Eigen::Vector2f& rel)
{
    float scrollDelta; //

    // Give priority to the horizontal axis because it is the natural direction
    // for the timeline. Typical mouses will only provide the vertical
    // component but touch pads and other specific devices allow horizontal
    // scrolling.
    scrollDelta = rel.x();
    if
        (std::fabs(scrollDelta) < 0.01f)
    {
        scrollDelta = rel.y();
    }

    if
        (nullptr != this->scrollCallback)
    {
        this->scrollCallback(scrollDelta);
    }

    return true;  // prevent event propagation
};
