// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iomanip>

#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_camera_control.h"
#include "ui/nanogui/nanogui_icon.h"

#include "core/app_events_dispatcher.h"
#include "core/input_events_dispatcher.h"
#include "3d/scene_view.h"

#include "events/cursor_location_changed_event.h"

using namespace SEAScope;

/**
 *
 */
NanoguiCameraControl::NanoguiCameraControl(nanogui::Widget* parent, AppEventsDispatcher& appEventsDispatcher, const InputEventsDispatcher& inputEventsDispatcher)
: nanogui::Widget(parent), NanoguiComponent(parent),
  _appEventsDispatcher(appEventsDispatcher), _inputEventsDispatcher(inputEventsDispatcher),
  _location(nullptr), _altitude(nullptr), _scale(nullptr)
{
	nanogui::Widget* spacer;	//
	NanoguiIcon* icon;	//

	this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
										   nanogui::Alignment::Minimum,
										   3, 0));

	/* Coordinates */
	icon = new NanoguiIcon(this, ENTYPO_ICON_LOCATION_PIN);
	icon->setFixedSize(Eigen::Vector2i(20, 20));
	icon->setFontSize(18);
	icon->setColor(nanogui::Color(255,255,255,255));
	icon->setTooltip("Location: longitude and latitude under the mouse cursor");

	this->_location = new nanogui::Label(this, "", "sans-bold");
	this->_location->setFixedSize(Eigen::Vector2i(98, 25));

	/* Zoom */
	icon = new NanoguiIcon(this, ENTYPO_ICON_TRIANGLE_UP);
	icon->setFixedSize(Eigen::Vector2i(20, 20));
	icon->setFontSize(18);
	icon->setColor(nanogui::Color(255,255,255,255));
	icon->setTooltip("Altitude of the camera (above sea level)");

	this->_altitude = new nanogui::Label(this, "00000m", "sans-bold");
	this->_altitude->setFixedSize(Eigen::Vector2i(63, 20));

	spacer = new nanogui::Widget(this);
	spacer->setFixedSize(Eigen::Vector2i(18, 24));

	/* Scale */
	this->_scale = new NanoguiScale(this);
    this->_scale->setFixedSize(Eigen::Vector2i(65, 24));

    this->_appEventsDispatcher.subscribe<CursorLocationChangedEvent>(this);
};

/**
 *
 */
NanoguiCameraControl::~NanoguiCameraControl()
{
    this->_appEventsDispatcher.unsubscribe<CursorLocationChangedEvent>(this);
};

/**
 *
 */
bool NanoguiCameraControl::setSize(int w, int h)
{
	this->setFixedSize(Eigen::Vector2i(w, h));
	return true;
};

/**
 *
 */
void NanoguiCameraControl::draw(NVGcontext* ctx)
{
	nvgBeginPath(ctx);
	nvgRoundedRect(ctx,
					static_cast<float>(this->mPos.x()), static_cast<float>(this->mPos.y()),
					static_cast<float>(this->mSize.x()), static_cast<float>(this->mSize.y()), 12);
	nvgFillColor(ctx, nanogui::Color(255,255,255,10));
	nvgFill(ctx);
	nanogui::Widget::draw(ctx);
};

/**
 *
 */
bool NanoguiCameraControl::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(CursorLocationChangedEvent)) == eventType)
    {
        const CursorLocationChangedEvent& typedEvent = static_cast<const CursorLocationChangedEvent &>(event);
        this->setLocation(typedEvent.getOnWorld(), typedEvent.getLon(), typedEvent.getLat());

        return true;
    }

    return false;
}

void NanoguiCameraControl::setLocation(bool onWorld, float lon, float lat)
{
    if
        (false == onWorld)
    {
        this->_location->setCaption("Lost in space");
    }
    else
    {
        std::ostringstream buffer;	//

        buffer << std::fixed << std::setprecision(2) << lon << "°, ";
        buffer << std::fixed << std::setprecision(2) << lat << "°";
        this->_location->setCaption(buffer.str());
    }
}

/**
 *
 */
void NanoguiCameraControl::update(const SceneView* sceneView)
{
    // Since camera changed we need to update cursor location

    std::ostringstream buffer;	//

    glm::vec2 lonlat;
    glm::vec3 hitpoint;
    bool onWorld = sceneView->getLonLat(this->_inputEventsDispatcher.getInputDevicesState().getMouseX(),
                                        this->_inputEventsDispatcher.getInputDevicesState().getMouseY(),
                                        lonlat, hitpoint);
    float resolution {0.f};	//
    resolution = sceneView->getPixelSize();

    float altitudeValue {0.f};	//
    altitudeValue = sceneView->getAltitude();

    this->setLocation(onWorld, lonlat.x, lonlat.y);

    this->_scale->setResolution(resolution);

    if
        (10000 > altitudeValue)
    {
        buffer << std::fixed << std::setprecision(0) << altitudeValue << "m";
    }
    else
    {
        altitudeValue = 0.001f * altitudeValue;
        buffer << std::fixed << std::setprecision(1) << altitudeValue << "km";
    }
    this->_altitude->setCaption(buffer.str());
}
