// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_clickable_label.h"

using namespace SEAScope;

/**
 *
 */
NanoguiClickableLabel::NanoguiClickableLabel(nanogui::Widget* parent)
: nanogui::Widget(parent),
  NanoguiPaintedLabel(parent, ""),
  _selected(false),
  _lastChange(std::chrono::system_clock::now()),
  _changeCallback(nullptr)
{
}

/**
 *
 */
NanoguiClickableLabel::~NanoguiClickableLabel()
{

}

/**
 *
 */
bool NanoguiClickableLabel::mouseButtonEvent(const Eigen::Vector2i& position, int button, bool pressed, int mods)
{
    bool capture(false);

    // Let's just ignore which button has triggered the event and just process
    // based on the "pressed" status
    if
        (true == pressed)
    {
        this->toggleSelection();
    }

    this->updateBackgroundColor();

    capture = NanoguiPaintedLabel::mouseButtonEvent(position, button, pressed, mods);
    return capture;
}

/**
 *
 */
void NanoguiClickableLabel::updateBackgroundColor()
{
    if
        (true == this->_selected)
    {
        this->setBackgroundColor(nanogui::Color(48, 87, 139, 255));
    }
    else if
        (true == this->focused())
    {
        this->setBackgroundColor(nanogui::Color(60, 60, 60, 255));
    }
    else
    {
        this->setBackgroundColor(nanogui::Color(0, 0, 0, 0));
    }
}

/**
 *
 */
bool NanoguiClickableLabel::mouseEnterEvent(const Eigen::Vector2i& position, bool enter)
{
    bool capture(false);

    capture = NanoguiPaintedLabel::mouseEnterEvent(position, enter);
    this->mFocused = this->mMouseFocus;

    this->updateBackgroundColor();

    return capture;
}

/**
 *
 */
bool NanoguiClickableLabel::focusEvent(bool focused)
{
    bool capture(false);

    capture = NanoguiPaintedLabel::focusEvent(focused);

    this->updateBackgroundColor();

    return capture;
}

/**
 *
 */
void NanoguiClickableLabel::toggleSelection()
{
    this->setSelected(!this->_selected, false);
}

/**
 *
 */
void NanoguiClickableLabel::setSelected(bool selected, bool silent)
{
    if
        (selected == this->_selected)
    {
        return;
    }

    this->_selected = selected;

    this->updateBackgroundColor();

    std::chrono::time_point<std::chrono::system_clock> now = std::chrono::system_clock::now();
    if
        ((false == silent) && (nullptr != this->_changeCallback))
    {
        float nanosecondsSinceLastChange = static_cast<float>(std::chrono::duration_cast<std::chrono::nanoseconds>(now - this->_lastChange).count());
        this->_changeCallback(this->caption(), this->_selected, nanosecondsSinceLastChange);
    }
    this->_lastChange = now;
}

/**
 *
 */
bool NanoguiClickableLabel::setChangeCallback(const std::function<void(const std::string&, bool, float)>& callback)
{
    if
        (nullptr != this->_changeCallback)
    {
        return false;
    }

    this->_changeCallback = callback;
    return true;
}
