// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>
#include <sstream>

#include "vendors/_nanogui.h"
#include "core/log.h"
#include "core/source.h"

#include "ui/nanogui/nanogui_collections_control.h"

using namespace SEAScope;

/**
 *
 */
NanoguiCollectionsControl::NanoguiCollectionsControl(nanogui::Widget* parent)
: nanogui::Widget(parent), NanoguiScrollableWithTitle(parent),
  _listChangedCallback(nullptr),
  _collectionRemovedCallback(nullptr),
  selectionCallback(nullptr)
{
    this->header->setFixedHeight(15);  // add extra margin on top
	this->footer->setFixedHeight(5);  // add extra margin at the bottom

	this->scrollable->content->widgetsSorter = NanoguiCollectionsControl::compareCollections;
	this->scrollable->content->setLayout(new nanogui::BoxLayout( nanogui::Orientation::Vertical
												   , nanogui::Alignment::Minimum
												   , 3, 6));
};

/**
 *
 */
NanoguiCollectionsControl::~NanoguiCollectionsControl()
{
};

/**
 *
 */
bool NanoguiCollectionsControl::compareCollections(const nanogui::Widget* a, const nanogui::Widget* b)
{
	const nanogui::Widget* child;	//
	const nanogui::Label* aLabel;	//
	const nanogui::Label* bLabel;	//

	child = a->children()[0];
	aLabel = dynamic_cast<const nanogui::Label*>(child);
	if
		(nullptr == aLabel)
	{
        child = a->children()[0]->children()[1];
        aLabel = dynamic_cast<const nanogui::Label*>(child);
        if
            (nullptr == aLabel)
        {
            LOG( LogLevel::Error, "NanoguiCollectionsControl"
                    , "widgetsSorter received a widget (a) which is not supported");
            return false;
        }
	}
	const std::string& aCaption = aLabel->caption();

	child = b->children()[0];
	bLabel = dynamic_cast<const nanogui::Label*>(child);
	if
		(nullptr == bLabel)
	{
        child = b->children()[0]->children()[1];
        bLabel = dynamic_cast<const nanogui::Label*>(child);
        if
            (nullptr == bLabel)
        {
            LOG( LogLevel::Error, "NanoguiCollectionsControl"
                    , "widgetsSorter received a widget (b) which is not supported");
            return false;
        }
	}
	const std::string& bCaption = bLabel->caption();

	std::string aStr(aCaption);	//
	std::transform(aStr.begin(), aStr.end(), aStr.begin(), [](char c) { return static_cast<char>(std::tolower(c)); });
	std::string bStr(bCaption);	//
	std::transform(bStr.begin(), bStr.end(), bStr.begin(), [](char c) { return static_cast<char>(std::tolower(c)); });
	return aStr < bStr;
};

/**
 *
 */
bool NanoguiCollectionsControl::addVariables(const std::vector< std::pair<CollectionVariable, CatalogueEntry> >& vars)
{
	std::ostringstream buffer;	//
	std::string category;	//
	nanogui::Widget* categoryWidget;	//
	NanoguiCheckbox* cb;	//

	categoryWidget = nullptr;
	cb = nullptr;

	for
		(const auto& variable : vars)
	{
		category = this->getCategory(variable.first, variable.second);
		auto categoryFound = this->categories.find(category);
		if
			(this->categories.end() == categoryFound)
		{
			categoryWidget = new nanogui::Widget(this->scrollable->content);
			categoryWidget->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
															 , nanogui::Alignment::Minimum
															 , 0, 0));

            CollectionId colId(variable.first.first);
            if
                (SEAScope::Source::ANNOTATIONS == colId.first)
            {
                nanogui::Widget* labelBlock = new nanogui::Widget(categoryWidget);
                labelBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                             nanogui::Alignment::Minimum,
                                                             2, 3));

                nanogui::Button* deleteButton = new nanogui::Button(labelBlock, "", ENTYPO_ICON_TRASH);
                deleteButton->setFixedSize(Eigen::Vector2i(18, 18));

			    new nanogui::Label(labelBlock, category, "sans-bold");

                nanogui::Widget* deleteBlock  = new nanogui::Widget(categoryWidget);
                deleteBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                              nanogui::Alignment::Minimum,
                                                              2, 3));
                new nanogui::Label(deleteBlock, "Delete collection and related content?", "sans", 14);
                nanogui::Button* yesButton = new nanogui::Button(deleteBlock, "Yes");
                yesButton->setFixedSize(Eigen::Vector2i(18, 18));
                yesButton->setFontSize(14);
                nanogui::Button* noButton = new nanogui::Button(deleteBlock, "No");
                noButton->setFixedSize(Eigen::Vector2i(18, 18));
                noButton->setFontSize(14);
                deleteBlock->setVisible(false);

                deleteButton->setCallback([this, deleteBlock]() {
                    deleteBlock->setVisible(true);
                    this->reflowNeeded = true;
                });

                noButton->setCallback([this, deleteBlock]() {
                    deleteBlock->setVisible(false);
                    this->reflowNeeded = true;
                });

                yesButton->setCallback([this, category, colId]() {
                    auto found = this->categories.find(category);

                    if
                        (this->categories.end() == found)
                    {
                        return;
                    }

                    /* Note: actually removing the widget leads to crashes
                       apparently due to bugs in nanogui, so we set its
                       visibility to false as an non-optimal but simple
                       workaround */
                    found->second->setEnabled(false);
                    found->second->setVisible(false);

                    this->categories.erase(found);

                    std::map<CollectionVariable, NanoguiCheckbox*>::iterator iter = this->variables.begin();
                    for
                        (; iter != this->variables.end();)
                    {
                        if
                            (  (iter->first.first.first == colId.first)
                            && (iter->first.first.second == colId.second))
                        {
                            if
                                (true == iter->second->checked())
                            {
		                        iter->second->setChecked(false);
                                if
                                    (nullptr != this->selectionCallback)
                                {
                                    this->selectionCallback(iter->first, false, false);
                                }
                            }
                            iter = this->variables.erase(iter);
                        }
                        else
                        {
                            ++iter;
                        }
                    }

                    //this->scrollable->content->removeChild(found->second);

                    this->scrollable->content->update();

                    if
                        (nullptr != this->_collectionRemovedCallback)
                    {
                        this->_collectionRemovedCallback(colId);
                    }
                });
            }
            else
            {
			    new nanogui::Label(categoryWidget, category, "sans-bold");
            }

			this->categories.insert(std::make_pair(category, categoryWidget));
		}
		else
		{
			categoryWidget = categoryFound->second;
		}

		auto variableFound = this->variables.find(variable.first);
		if
			(this->variables.end() == variableFound)
		{
			CollectionVariable colvar = variable.first;

			buffer.str("");
			buffer << category << ": " << variable.second.variableLabel;
			buffer << "\n\n";
			buffer << " - Source: " << colvar.first.first << "\n";
			buffer << " - Collection: " << colvar.first.second << "\n";
			buffer << " - Variable: " << colvar.second << "\n";
			cb = new NanoguiCheckbox(categoryWidget, variable.second.variableLabel);
			cb->setTooltip(buffer.str());
			cb->setCallback([this, cb, colvar /*copy*/](bool checked) {
                bool shiftPressed;  //
                shiftPressed = false;
				if
					(nullptr != this->selectionCallback)
				{
                    shiftPressed = (/*GLFW_MOD_SHIFT*/ 1 == cb->keyPressed);
                    if
                        (shiftPressed && !checked)
                    {
                        cb->setChecked(true);
                    }
					this->selectionCallback(colvar, checked, shiftPressed);
				}
			});

			this->variables.insert(std::make_pair(variable.first, cb));
		}
		else
		{
			/* Nothing, variable should already be present in layout */
		}
	}

	this->scrollable->content->update();
	this->reflowNeeded = 0 < variables.size();

	if
		(nullptr != this->_listChangedCallback)
	{
		this->_listChangedCallback();
	}
	return true;
};

/**
 *
 */
bool NanoguiCollectionsControl::removeVariables(const std::vector<CollectionVariable>& /*vars*/)
{
	bool ok;	//
	ok = false;
	/* TODO: implement this */
	if
		(true == ok && nullptr != this->_listChangedCallback)
	{
		this->_listChangedCallback();
	}
	return ok;
};

/**
 *
 */
bool NanoguiCollectionsControl::selectVariable(const CollectionVariable& colvar, bool selected)
{
	auto found = this->variables.find(colvar);
	if
		(this->variables.end() == found)
	{
		return false;
	}
	if
		(found->second->checked() != selected)
	{
		found->second->setChecked(selected);
	}
	return false;
};

/**
 *
 */
bool NanoguiCollectionsControl::setSelectionCallback(const std::function<void(const CollectionVariable&, bool, bool)>& callback)
{
	this->selectionCallback = callback;
	return true;
};

/**
 *
 */
const std::string& NanoguiCollectionsControl::getCategory(const CollectionVariable& /*colvar*/, const CatalogueEntry& entry) const
{
	return entry.collectionLabel;
};

/**
 *
 */
bool NanoguiCollectionsControl::setListChangedCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_listChangedCallback)
    {
        return false;
    }
    this->_listChangedCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiCollectionsControl::setCollectionRemovedCallback(const std::function<void(const CollectionId&)>& callback)
{
    if
        (nullptr != this->_collectionRemovedCallback)
    {
        return false;
    }
    this->_collectionRemovedCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiCollectionsControl::allowDeletion(const CollectionVariable& collectionVariable, bool allow)
{
    auto variableFound = this->variables.find(collectionVariable);
    if
        (this->variables.end() == variableFound)
    {
        return false;
    }

    NanoguiCheckbox* checkbox = variableFound->second;
    nanogui::Widget* categoryWidget = checkbox->parent();
    if
        (1 >= categoryWidget->children().size())
    {
        return false;
    }

    nanogui::Widget* firstChild = categoryWidget->children()[0];

    // Make sure the first child is not a label because it would mean that the
    // collection is not deletable
    nanogui::Label* label = dynamic_cast<nanogui::Label*>(firstChild);
    if
        (nullptr != label)
    {
        return false;
    }

    if
        (0 >= firstChild->children().size())
    {
        return false;
    }

    nanogui::Widget* firstSubChild = firstChild->children()[0];
    nanogui::Button* deleteBtn = dynamic_cast<nanogui::Button*>(firstSubChild);
    if
        (nullptr == deleteBtn)
    {
        return false;
    }

    deleteBtn->setEnabled(allow);

    if
        (false == allow)
    {
        nanogui::Widget* deleteBlock = categoryWidget->children()[1];
        deleteBlock->setVisible(false);
    }

    return true;
}
