// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_colorbar.h"

using namespace SEAScope;

/**
 *
 */
NanoguiColorbar::NanoguiColorbar(nanogui::Widget* parent): nanogui::Widget(parent)
{
	this->setFixedSize(Eigen::Vector2i(20, 255));
};

/**
 *
 */
NanoguiColorbar::~NanoguiColorbar()
{
};

/**
 *
 */
Eigen::Vector2i NanoguiColorbar::preferredSize(NVGcontext* /*ctx*/) const
{
	return this->mFixedSize;
}

/**
 *
 */
void NanoguiColorbar::draw(NVGcontext* ctx)
{
	size_t gradientSize;	//
	size_t segCount;	//
	size_t i;	//
	uint16_t segSize;	//
	NVGpaint paint;	//

	nanogui::Widget::draw(ctx);

	/* draw color gradient */
	gradientSize = this->cmap.size();
	if
		(1 >= gradientSize
		&& gradientSize <= static_cast<std::size_t>(this->mSize.y()))
	{
		return;
	}
	segCount = gradientSize - 1;
	segSize = static_cast<std::uint16_t>(static_cast<std::size_t>(this->mSize.y()) / segCount);
	i = 1;
	for
		(; i < gradientSize; ++i)
	{
		paint = nvgLinearGradient(ctx, 0, 0, 0, segSize
								 , nanogui::Color(this->cmap[gradientSize-i][0], this->cmap[gradientSize-i][1], this->cmap[gradientSize-i][2], 255)
								 , nanogui::Color(this->cmap[gradientSize-i-1][0], this->cmap[gradientSize-i-1][1], this->cmap[gradientSize-i-1][2], 255));

		nvgBeginPath(ctx);
		nvgRoundedRect(ctx
					  , static_cast<float>(this->mPos.x() + this->mSize.x()) - 20.f
					  , static_cast<float>(static_cast<std::size_t>(this->mPos.y()) + (i - 1) * segSize)
					  , 20, segSize, 0);
		nvgFillPaint(ctx, paint);
		nvgFill(ctx);
	}
};

/**
 *
 */
bool NanoguiColorbar::mouseButtonEvent(const Eigen::Vector2i& pos, int buttons, bool pressed, int mods)
{
	float ratio;	//

	nanogui::Widget::mouseButtonEvent(pos, buttons, pressed, mods);
	if
		(false == pressed)
	{
		if
			(false == this->dragging)
		{
			/* this is a click */
			if
				(nullptr != this->clickCallback)
			{
				this->clickCallback(this->label, this->cmap);
			}
		}
		else if
			(  nullptr != this->dragEndCallback
			&& 0 < this->height())
		{
			/* drag end */
			ratio = static_cast<float>(this->mPos.y() + this->height() - pos.y());
			ratio /= static_cast<float>(this->height());
			this->dragEndCallback(this->dragTargetIsMax, ratio);
		}
		this->dragging = false;
	}
	else
	{
		this->dragTargetIsMax = (128 > pos.y());
		LOG(LogLevel::Debug, "NanoguiColorbar", "mouse down at {} / {}", pos.y(), this->mSize.y());
	}
	return true;
};

/**
 *
 */
bool NanoguiColorbar::isMouseOver() const
{
	return this->mMouseFocus;
}

/**
 *
 */
bool NanoguiColorbar::mouseMotionEvent(const Eigen::Vector2i& pos, const Eigen::Vector2i& /*rel*/, int /*buttons*/, int /*modifiers*/)
{
	float ratio;	//

	if
		(nullptr != this->mouseMoveCallback
		&& 0 < this->height())
	{
		ratio = static_cast<float>(this->mPos.y() + this->height() - pos.y());
		ratio /= static_cast<float>(this->height());
		this->mouseMoveCallback(this->dragTargetIsMax, ratio); 
	}
	return true;
};

/**
 *
 */
bool NanoguiColorbar::mouseDragEvent(const Eigen::Vector2i& pos, const Eigen::Vector2i& /*rel*/, int /*buttons*/, int /*modifiers*/)
{
	float ratio;	//

	this->dragging = true;

	if
		(nullptr != this->mouseMoveCallback
		&& 0 < this->height())
	{
		ratio = static_cast<float>(this->mPos.y() + this->height() - pos.y());
		ratio /= static_cast<float>(this->height());
		this->mouseMoveCallback(this->dragTargetIsMax, ratio); 
	}
	return true;
};

/**
 *
 */
bool NanoguiColorbar::setClickCallback(const std::function<void(const std::string&, const Colormap&)>& callback)
{
	this->clickCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiColorbar::setMouseMoveCallback(const std::function<void(bool, float)>& callback)
{
	this->mouseMoveCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiColorbar::setDragEndCallback(const std::function<void(bool, float)>& callback)
{
	this->dragEndCallback = callback;
	return true;
};
