// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_combobox.h"

#include "ui/nanogui/nanogui_anchored.h"
#include "ui/nanogui/nanogui_list.h"

#include "ui/nanogui/nanogui_clickable_label.h"

#include <iostream>

using namespace SEAScope;

/**
 *
 */
NanoguiComboBox::NanoguiComboBox(nanogui::Widget* parent)
: nanogui::Widget(parent),
  _selectedValue(""),
  _selectedLabel(nullptr),
  _dropdownButton(nullptr),
  _dropdownContainer(nullptr),
  _itemsList(nullptr),
  _dropdownCallback(nullptr),
  _changeCallback(nullptr)
{
    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                           nanogui::Alignment::Fill,
                                           2, 0));

    nanogui::Widget* content = new nanogui::Widget(this);
    content->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                              nanogui::Alignment::Minimum,
                                              0, 0));
    this->_selectedLabel = new nanogui::Label(content, this->_selectedValue);

    this->_dropdownButton = new nanogui::Button(content, "", ENTYPO_ICON_CHEVRON_SMALL_DOWN);
    this->_dropdownButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_dropdownButton->setFontSize(16);

    this->_dropdownContainer = NanoguiAnchored::create(this->_dropdownButton);
    this->_dropdownContainer->offset = 0u;
    this->_dropdownContainer->placement = static_cast<std::uint8_t>(NanoguiAnchored::ON_LEFT + NanoguiAnchored::ON_BOTTOM);
    this->_dropdownContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                               nanogui::Alignment::Minimum,
                                                               0, 0));
    this->_itemsList = new NanoguiList(this->_dropdownContainer);
    this->_itemsList->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                       nanogui::Alignment::Fill,
                                                       0, 0));
    this->_itemsList->content->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                                nanogui::Alignment::Fill,
                                                                1, 1));
    this->_itemsList->content->setBackgroundFillColor(nanogui::Color(80, 80, 80, 255));

    this->_dropdownButton->setCallback([this]() {
        this->toggleDropDown();
    });

    this->hideDropDown();
}

/**
 *
 */
NanoguiComboBox::~NanoguiComboBox()
{

}

/**
 *
 */
void NanoguiComboBox::enable()
{
    this->setEnabled(true);
    this->_dropdownButton->setEnabled(true);
    this->_selectedLabel->setEnabled(true);
}

/**
 *
 */
void NanoguiComboBox::disable()
{
    this->hideDropDown();

    this->setEnabled(false);
    this->_dropdownButton->setEnabled(false);
    this->_selectedLabel->setEnabled(false);
}

/**
 *
 */
void NanoguiComboBox::toggleDropDown()
{
    if
        (this->_dropdownContainer->visible())
    {
        this->hideDropDown();
    }
    else
    {
        this->showDropDown();
    }
}

/**
 *
 */
void NanoguiComboBox::showDropDown()
{
    this->_dropdownContainer->setVisible(true);
    this->_dropdownButton->setIcon(ENTYPO_ICON_CHEVRON_SMALL_UP);
    if
        (nullptr != this->_dropdownCallback)
    {
        this->_dropdownCallback();
    }
}

/**
 *
 */
void NanoguiComboBox::hideDropDown()
{
    this->_dropdownContainer->setVisible(false);
    this->_dropdownButton->setIcon(ENTYPO_ICON_CHEVRON_SMALL_DOWN);
}

/**
 *
 */
bool NanoguiComboBox::setDropDownCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_dropdownCallback)
    {
        return false;
    }

    this->_dropdownCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiComboBox::setChangeCallback(const std::function<void(const std::string&)>& callback)
{
    if
        (nullptr != this->_changeCallback)
    {
        return false;
    }

    this->_changeCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiComboBox::setSelectedValue(const std::string& value)
{
    if
        (value == this->_selectedValue)
    {
        return false;
    }

    // Find the widgets that corresponds to this value (if any)
    for
        (auto widget : this->_widgets)
    {
        widget->setSelected((widget->caption() == value), true /*silenced*/);
    }

    this->_selectedValue = value;
    this->_selectedLabel->setCaption(this->_selectedValue);

    if
        (nullptr != this->_changeCallback)
    {
        this->_changeCallback(this->_selectedValue);
    }
    return true;
}

/**
 *
 */
void NanoguiComboBox::show()
{
    this->setVisible(true);
}

/**
 *
 */
void NanoguiComboBox::hide()
{
    this->hideDropDown();
    this->setVisible(false);
}

/**
 *
 */
const std::string& NanoguiComboBox::getSelectedValue() const
{
    return this->_selectedValue;
}

/**
 *
 */
void NanoguiComboBox::setFixedSize(const Eigen::Vector2i& size)
{
    this->_selectedLabel->setFixedSize(Eigen::Vector2i(size.x() - this->_dropdownButton->fixedSize().x() - 4 /* spacing */, size.y()));
    this->_itemsList->setFixedSize(Eigen::Vector2i(size.x() - this->_dropdownButton->fixedSize().x() - 4 /* spacing */, 60));
    nanogui::Widget::setFixedSize(size);
}

/**
 *
 */
void NanoguiComboBox::setItems(const std::vector<std::string>& items)
{
    NanoguiClickableLabel* label(nullptr);
    this->_items = items;
    std::size_t availableWidgets = this->_widgets.size();

    auto found = std::find(this->_items.begin(), this->_items.end(),
                           this->_selectedValue);
    if
        (this->_items.end() == found
        && ("" != this->_selectedValue))
    {
        this->_selectedValue = "";
        this->_selectedLabel->setCaption(this->_selectedValue);
        if
            (nullptr != this->_changeCallback)
        {
            this->_changeCallback(this->_selectedValue);
        }
    }

    std::size_t count = 0;
    for
        (const auto& item : this->_items)
    {
        if
            (count < availableWidgets)
        {
            label = this->_widgets.at(count);
        }
        else
        {
            label = new NanoguiClickableLabel(this->_itemsList->content);
            label->setChangeCallback([this] (const std::string& value, bool selected, float) {
                if
                    (true == selected)
                {
                    this->setSelectedValue(value);
                }
                this->hideDropDown();
            });
            this->_widgets.push_back(label);
        }

        label->setCaption(item);
        label->setSelected((item == this->_selectedValue), true);
        label->setVisible(true);
        label->setEnabled(true);

        count = count + 1;
    }

    if
        (0 < count)
    {
        int newHeight = static_cast<int>(count) * label->preferredSize(this->screen()->nvgContext()).y() + 2 /*margin*/ + (static_cast<int>(count) - 1) * 1 /*inter spacing*/;
        this->_itemsList->content->setFixedHeight(newHeight);
        this->_itemsList->content->setHeight(newHeight);
    }
    else
    {
        this->_itemsList->content->setFixedHeight(0);
    }

    // If there is only one item, select it by default
    if
        ((nullptr != label) && (1 == this->_items.size()))
    {
        label->setSelected(true /*select*/, false /*notify change*/);
        this->_selectedLabel->setCaption(this->_items.at(0));
        this->_selectedValue = this->_items.at(0);
    }

    // Hide all unused widgets
    while
        (count < this->_widgets.size())
    {
        label = this->_widgets.at(count);
        label->setCaption("");
        label->setSelected(false, true);    // unselect by default
        label->setVisible(false);
        label->setEnabled(false);
        count = count + 1;
    }

    this->_itemsList->update(true /*reset scrolling*/);
}

/**
 *
 */
bool NanoguiComboBox::mouseButtonEvent(const Eigen::Vector2i& pos, int button, bool pressed, int mods)
{
    if
        (false == this->enabled())
    {
        return false;
    }

    bool capture(false);

    capture = nanogui::Widget::mouseButtonEvent(pos, button, pressed, mods);

    if
        ((true == pressed) && (this->_selectedLabel->contains(pos - this->mPos)))
    {
        this->toggleDropDown();
    }

    return capture;
}

/**
 *
 */
void NanoguiComboBox::draw(NVGcontext* ctx)
{
    nvgBeginPath(ctx);
    nvgRoundedRect(ctx,
                   static_cast<float>(this->mPos.x()), static_cast<float>(this->mPos.y()) + 1,
                   static_cast<float>(this->mSize.x()), static_cast<float>(this->_selectedLabel->fixedSize().y() - 1),
                   3);
    nvgStrokeColor(ctx, nanogui::Color(200, 200, 200, 150));
    nvgStroke(ctx);
    nanogui::Widget::draw(ctx);
}
