// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "ui/nanogui/nanogui_corner_anchor.h"

using namespace SEAScope;

/** 
 *
 */
NanoguiCornerAnchor::NanoguiCornerAnchor(nanogui::Widget* parent)
: nanogui::Widget(parent), corner(NanoguiCorner::TOP_LEFT), topLeftRadius(0)
, topRightRadius(0), bottomLeftRadius(0), bottomRightRadius(0)
, updateCallback(nullptr), clickCallback(nullptr)
{
	this->setSize(Eigen::Vector2i(32, 32));
};

/**
 *
 */
NanoguiCornerAnchor::~NanoguiCornerAnchor()
{

};

/**
 *
 */
bool NanoguiCornerAnchor::setCorner(const NanoguiCorner& cornerIn)
{
	if
		(cornerIn == this->corner)
	{
		return false;
	}

	this->corner = cornerIn;
	return this->update();
};

/**
 *
 */
bool NanoguiCornerAnchor::update()
{
	this->updateRadii();
	this->updatePosition();
	if
		(nullptr != this->updateCallback)
	{
		this->updateCallback( this->mPos.x(), this->mPos.y()
							, this->mSize.x(), this->mSize.y());
	}
	return true;
};

/**
 *
 */
bool NanoguiCornerAnchor::updateRadii()
{
	this->topLeftRadius = 0;
	this->topRightRadius = 0;
	this->bottomLeftRadius = 0;
	this->bottomRightRadius = 0;

	if
		(NanoguiCorner::TOP_LEFT == this->corner)
	{
		this->bottomRightRadius = static_cast<std::uint16_t>(floor(0.5 * this->mSize.x()));
	}
	else if
		(NanoguiCorner::TOP_RIGHT == this->corner)
	{
		this->bottomLeftRadius = static_cast<std::uint16_t>(floor(0.5 * this->mSize.x()));
	}
	else if
		(NanoguiCorner::BOTTOM_LEFT == this->corner)
	{
		this->topRightRadius = static_cast<std::uint16_t>(floor(0.5 * this->mSize.x()));
	}
	else
	{
		this->topLeftRadius = static_cast<std::uint16_t>(floor(0.5 * this->mSize.x()));
	}

	return true;
};

/**
 *
 */
bool NanoguiCornerAnchor::updatePosition()
{
	if
		(NanoguiCorner::TOP_LEFT == this->corner)
	{
		this->setPosition(Eigen::Vector2i(0, 0));
	}
	else if
		(NanoguiCorner::TOP_RIGHT == this->corner)
	{
		this->setPosition(Eigen::Vector2i(this->parent()->width() - this->mSize.x(), 0));
	}
	else if
		(NanoguiCorner::BOTTOM_LEFT == this->corner)
	{
		this->setPosition(Eigen::Vector2i(0, this->parent()->height() - this->mSize.y()));
	}
	else
	{
		this->setPosition(Eigen::Vector2i(this->parent()->width() - this->mSize.x(), this->parent()->height() - this->mSize.y()));
	}
	return true;
};

/**
 *
 */
void NanoguiCornerAnchor::draw(NVGcontext* ctx)
{
	nvgBeginPath(ctx);
	nvgRoundedRectVarying( ctx
						 , static_cast<float>(this->mPos.x())
                         , static_cast<float>(this->mPos.y())
						 , static_cast<float>(this->mSize.x())
                         , static_cast<float>(this->mSize.y())
						 , this->topLeftRadius, this->topRightRadius
						 , this->bottomRightRadius, this->bottomLeftRadius);
	NVGpaint bg = nvgLinearGradient( ctx
								   , static_cast<float>(this->mPos.x())
                                   , static_cast<float>(this->mPos.y())
								   , static_cast<float>(this->mPos.x())
                                   , static_cast<float>(this->mPos.y() + this->mSize.y())
								   , nanogui::Color(130, 130, 130, 180)
								   , nanogui::Color(60, 60, 60, 230));

	nvgFillPaint(ctx, bg);
	nvgFill(ctx);

	nanogui::Widget::draw(ctx);
};

/**
 *
 */
void NanoguiCornerAnchor::setSize(const Eigen::Vector2i& size)
{
	nanogui::Widget::setSize(size);
	this->update();
};

/**
 *
 */
void NanoguiCornerAnchor::setFixedSize(const Eigen::Vector2i& size)
{
	nanogui::Widget::setFixedSize(size);
	this->update();
};

/**
 *
 */
bool NanoguiCornerAnchor::nextCorner()
{
	if
		(0 >= this->validCorners.size())
	{
		return false;
	}

	auto it = std::find( this->validCorners.begin()
					   , this->validCorners.end()
					   , this->corner);

	if
		(this->validCorners.end() == it)
	{
		/* Current corner not found */
		this->setCorner(this->validCorners[0]);
	}
	else
	{
		auto next = ++it;
		if
			(this->validCorners.end() == next)
		{
			/* Go back to first valid corner */
			this->setCorner(this->validCorners[0]);
		}
		else
		{
			this->setCorner(*next);
		}
	}

	return true;
};

/**
 *
 */
bool NanoguiCornerAnchor::mouseButtonEvent(const Eigen::Vector2i& /*pos*/, int button, bool pressed, int mods)
{
	/* Toggle target on left click */
	if
		(  GLFW_MOUSE_BUTTON_LEFT == button
		&& !pressed)
	{
		if
			(0 < (mods & GLFW_MOD_SHIFT))
		{
			/* Hop to next corner */
			this->nextCorner();
		}
		else if
			(nullptr != this->clickCallback)
		{
			this->clickCallback();
		}
	}

	return true;
};
