// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ostream>
#include <iomanip>

#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_anchored.h"
#include "ui/nanogui/nanogui_datebox.h"

using namespace SEAScope;

std::array<std::uint8_t, 12> NanoguiDateBox::daysPerMonth = {{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}};
std::uint8_t NanoguiDateBox::INVALID_MONTH = 255;
std::uint8_t NanoguiDateBox::INVALID_DAY = 255;

/**
 *
 */
NanoguiDateBox::NanoguiDateBox(nanogui::Widget* parent, std::uint16_t year, std::uint8_t month, std::uint8_t day)
: nanogui::TextBox(parent), NanoguiTextField(parent), yearTemp(year), monthTemp(month), dayTemp(day)
{
    this->_autoCommit = false;

	std::ostringstream buffer;	//

	buffer << year << "-" << std::setfill('0') << std::setw(2) << static_cast<int>(month) << "-" << static_cast<int>(day);
	this->setValue(buffer.str());
	this->setEditable(true);
	this->setFormat("-?[0-9]+-[0-9]{2}-[0-9]{2}");

	this->helper = NanoguiAnchored::create(this);
	this->helper->setMouseEnterCallback([this](bool value) {
		this->mouseOverHelper = value;
	});
	this->helper->setLayout( new nanogui::GridLayout( nanogui::Orientation::Horizontal, 12
													, nanogui::Alignment::Middle
													, 2, 0));
	this->helper->setVisible(false);
};

/**
 *
 */
NanoguiDateBox::~NanoguiDateBox()
{

};

/**
 *
 */
void NanoguiDateBox::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

    NanoguiTextField::draw(ctx);

    if
        (  (false == this->mCommitted)
        && (this->mValue != this->mValueTemp))
    {
        // Draw a line below value to hint user that an action is needed
        nvgBeginPath(ctx);
        for
            (float k = 0.f; k * 5.f + 3.f <= sizeX - 2.f; k = k + 1.F)
        {
            nvgMoveTo(ctx, posX + 1.f + 5.f * k, posY + sizeY - 2.f);
            nvgLineTo(ctx, posX + 1.f + 5.f * k + 3.f, posY + sizeY - 2.f);

        }
        nvgStrokeColor(ctx, nanogui::Color(30, 200, 230, 200));
        nvgStrokeWidth(ctx, 2);
        nvgStroke(ctx);
    }
}

/**
 *
 */
bool NanoguiDateBox::focusEvent(bool focused)
{
	bool ok;	//

	//LOG(LogLevel::Debug, "DateBox", "focusEvent {} {}", focused, this->helper->focused());

	ok = NanoguiTextField::focusEvent(focused);

	//LOG(LogLevel::Debug, "DateBox", "Widget focusEvent done");
	if
		(true == this->mValidFormat)
	{
		this->mValidFormat = this->checkValidity(this->mValueTemp);
	}
	this->updateHelpers(static_cast<std::size_t>(this->mCursorPos), focused);

	return ok;
};

/**
 *
 */
bool NanoguiDateBox::keyboardEvent(int key, int scancode, int action, int modifiers)
{
	bool ok; //

	ok = NanoguiTextField::keyboardEvent(key,scancode,action,modifiers);
	if
		(true == ok && true == this->mValidFormat)
	{
		this->mValidFormat = this->checkValidity(this->mValueTemp);
	}
	this->updateHelpers(static_cast<std::size_t>(this->mCursorPos), true);
	return ok;
};

/**
 *
 */
bool NanoguiDateBox::keyboardCharacterEvent(unsigned int codepoint)
{
	bool ok;	//

	ok = NanoguiTextField::keyboardCharacterEvent(codepoint);
	if
		(true == ok && true == this->mValidFormat)
	{
		this->mValidFormat = this->checkValidity(this->mValueTemp);
	}

	return ok;
};

/**
 *
 */
bool NanoguiDateBox::mouseButtonEvent(const Eigen::Vector2i &p, int button, bool down, int modifiers)
{
	bool ok;	//

	ok = NanoguiTextField::mouseButtonEvent(p,button,down,modifiers);
	if
		(!down)
	{
		this->updateHelpers(static_cast<std::size_t>(this->mCursorPos), true);
	}
	return ok;
};

/**
 *
 */
bool NanoguiDateBox::checkDate(const std::string& input)
{
	std::string tmp;	//
	std::size_t pos;	//
	std::uint16_t year;	//
	std::uint8_t month;	//
	std::uint8_t day;	//

	if
		(0 == input.size())
	{
		return false;
	}

	/* Check that month/day exists */
	pos = input.find("-");
	tmp = input.substr(0,pos);
	year = static_cast<std::uint16_t>(std::stoi(tmp));
	this->yearTemp = year;

	tmp = input.substr(pos+1,2);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}
	month = static_cast<std::uint8_t>(std::stoi(tmp));
	if
		(0 >= month)
	{
		this->monthTemp = INVALID_MONTH;
		return false;
	}
	this->monthTemp = month;

	month = month - 1u; /* to access 0-indexed daysPerMonth array values */
	tmp = input.substr(pos+4, 2);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}
	day = static_cast<std::uint8_t>(std::stoi(tmp));
	if
		(0 >= day)
	{
		this->dayTemp = INVALID_DAY;
		return false;
	}

	if
		(day <= daysPerMonth[month] || (1 == month && 29 == day && 0 == year%4))
	{
		this->dayTemp = day;
		return true;
	}

	this->dayTemp = INVALID_DAY;
	return false;
};

/**
 *
 */
bool NanoguiDateBox::checkValidity(const std::string& input)
{
	return this->checkDate(input);
};

/**
 *
 */
bool NanoguiDateBox::updateHelpers(std::size_t pos, bool visible)
{
	std::size_t firstSeparator;	//
	std::size_t secondSeparator;	//
	std::size_t thirdSeparator;	//

	if
		(false == visible)
	{
		/* hide helpers */
		this->showHelpers(HelperType::NONE);
		return false;
	}

	firstSeparator = this->mValueTemp.find("-", 0);
	if
		(  nullptr != this->grabAvailableYears
		&& (  std::string::npos == firstSeparator
		   || pos <= firstSeparator))
	{
		this->grabAvailableYears(this->availableYears, this->hlYears);
		this->showHelpers(HelperType::YEARS);
		return true;
	}

	secondSeparator = this->mValueTemp.find("-", firstSeparator + 1);
	if
		(  nullptr != this->grabAvailableMonths
		&& (  std::string::npos == secondSeparator
		   || (pos > firstSeparator && pos <= secondSeparator)))
	{
		this->grabAvailableMonths(this->availableMonths, this->hlMonths, this->yearTemp);
		this->showHelpers(HelperType::MONTHS);
		return true;
	}

	thirdSeparator = this->mValueTemp.find(" ", secondSeparator + 1);
	if
		(  nullptr != this->grabAvailableDays
		&& pos > secondSeparator
		&& (std::string::npos == thirdSeparator || pos <= thirdSeparator))
	{
		this->grabAvailableDays(this->availableDays, this->hlDays, this->yearTemp, this->monthTemp);
		this->showHelpers(HelperType::DAYS);
		return true;
	}

	this->showHelpers(HelperType::NONE);
	return false;
};

/**
 *
 */
bool NanoguiDateBox::setYearsGrabber(const std::function<bool(std::vector<std::uint16_t>&, std::vector<std::uint16_t>&)>& f)
{
	this->grabAvailableYears = f;
	return false;
};

/**
 *
 */
bool NanoguiDateBox::setMonthsGrabber(const std::function<bool(std::vector<std::uint8_t>&, std::vector<std::uint8_t>&, std::uint16_t)>& f)
{
	this->grabAvailableMonths = f;
	return false;
};

/**
 *
 */
bool NanoguiDateBox::setDaysGrabber(const std::function<bool(std::vector<std::uint8_t>&, std::vector<std::uint8_t>&, std::uint16_t, std::uint8_t)>& f)
{
	this->grabAvailableDays = f;
	return false;
};

/**
 *
 */
bool NanoguiDateBox::showHelpers(HelperType type)
{
	switch
		(type)
	{
		case HelperType::NONE:
			this->helper->setVisible(false);
			break;
		case HelperType::YEARS:
			this->displayYears();
			break;
		case HelperType::MONTHS:
			this->displayMonths();
			break;
		case HelperType::DAYS:
			this->displayDays();
			break;
	}
	return false;
};

/**
 *
 */
bool NanoguiDateBox::setYear(std::uint16_t year, bool commit)
{
	std::string tmp;	//
	std::ostringstream buffer;	//
	std::size_t firstSeparator;	//

	firstSeparator = this->mValueTemp.find("-", 0);
	if
		(std::string::npos == firstSeparator)
	{
		buffer << static_cast<int>(year) << this->defaultYearCompletion;
		tmp = buffer.str();
	}
	else
	{
		buffer << static_cast<int>(year);
		tmp = this->mValueTemp.replace(0, firstSeparator, buffer.str());
	}

	LOG( LogLevel::Debug, "NanoguiDateBox"
			, "setYear: {} {} {}"
			, static_cast<int>(year), this->mValueTemp, tmp);

	if
		(true == commit)
	{
		return this->commitValue(tmp);
	}
	return false;
};

/**
 *
 */
bool NanoguiDateBox::getYear(std::uint16_t& year)
{
	std::string tmp;	//
	std::size_t firstSeparator;	//
	int yearTmp;	//

	firstSeparator = this->mValueTemp.find("-", 0);
	if
		(std::string::npos == firstSeparator)
	{
		return false;
	}

	tmp = this->mValueTemp.substr(0, firstSeparator);

	yearTmp = std::atoi(tmp.c_str());
	year = static_cast<std::uint16_t>(yearTmp);

	return true;
};

/**
 * Assumes that there is at least one "-" in the text field,
 * so it should only be called from month helper buttons.
 */
bool NanoguiDateBox::setMonth(std::uint8_t month, bool commit)
{
	std::string tmp;	//
	std::ostringstream buffer;	//
	std::size_t firstSeparator;	//
	std::size_t secondSeparator;	//

	firstSeparator = this->mValueTemp.find("-", 0);
	secondSeparator = this->mValueTemp.find("-", firstSeparator + 1);
	if
		(std::string::npos == secondSeparator)
	{
		buffer << std::setfill('0') << std::setw(2) << static_cast<int>(month) << this->defaultMonthCompletion;
		tmp = this->mValueTemp.replace( firstSeparator + 1
									  , this->mValueTemp.size() - (firstSeparator +1)
									  , buffer.str());
	}
	else
	{
		buffer << std::setfill('0') << std::setw(2) << static_cast<int>(month);
		tmp = this->mValueTemp.replace( firstSeparator + 1
									  , secondSeparator - (firstSeparator +1)
									  , buffer.str());

	}

	LOG( LogLevel::Debug, "NanoguiDateBox"
			, "setMonth: {} {} {}"
			, static_cast<int>(month), this->mValueTemp, tmp);

	if
		(true == commit)
	{
		return this->commitValue(tmp);
	}
	return false;
};

/**
 *
 */
bool NanoguiDateBox::getMonth(std::uint8_t& month)
{
	std::string tmp;	//
	std::size_t firstSeparator;	//
	std::size_t secondSeparator;	//
	int monthTmp;	//

	firstSeparator = this->mValueTemp.find("-", 0);
	secondSeparator = this->mValueTemp.find("-", firstSeparator + 1);
	if
		(std::string::npos == firstSeparator)
	{
		return false;
	}

	if
		(std::string::npos == secondSeparator)
	{
		tmp = this->mValueTemp.substr( firstSeparator + 1
									  , this->mValueTemp.size() - (firstSeparator +1));
	}
	else
	{
		tmp = this->mValueTemp.substr( firstSeparator + 1
									  , secondSeparator - (firstSeparator +1));

	}

	monthTmp = std::atoi(tmp.c_str());
	month = static_cast<std::uint8_t>(monthTmp);

	return true;
};

/**
 * Assumes that there are at least two "-" in the text field,
 * so it should only be called from day helper buttons.
 */
bool NanoguiDateBox::setDay(std::uint8_t day, bool commit)
{
	std::string tmp;	//
	std::ostringstream buffer;	//
	std::size_t firstSeparator;	//
	std::size_t secondSeparator;	//
	std::size_t thirdSeparator;	//

	firstSeparator = this->mValueTemp.find("-", 0);
	secondSeparator = this->mValueTemp.find("-", firstSeparator + 1);
	thirdSeparator = this->mValueTemp.find(" ", secondSeparator + 1);
	if
		(std::string::npos == thirdSeparator)
	{
		buffer << std::setfill('0') << std::setw(2) << static_cast<int>(day) << this->defaultDayCompletion;
		tmp = this->mValueTemp.replace( secondSeparator + 1
									  , this->mValueTemp.size() - (secondSeparator +1)
									  , buffer.str());
	}
	else
	{
		buffer << std::setfill('0') << std::setw(2) << static_cast<int>(day);
		tmp = this->mValueTemp.replace( secondSeparator + 1
									  , thirdSeparator - (secondSeparator +1)
									  , buffer.str());
	}

	LOG( LogLevel::Debug, "NanoguiDateBox"
			, "setDay: {} {} {}"
			, static_cast<int>(day), this->mValueTemp, tmp);

	if
		(true == commit)
	{
		return this->commitValue(tmp);
	}
	return false;
};

/**
 *
 */
bool NanoguiDateBox::getDay(std::uint8_t& day)
{
	std::size_t firstSeparator;	//
	std::size_t secondSeparator;	//
	std::size_t thirdSeparator;	//
	std::string tmp;	//
	int dayTmp;	//

	firstSeparator = this->mValueTemp.find("-", 0);
	if
		(std::string::npos == firstSeparator)
	{
		return false;
	}
	secondSeparator = this->mValueTemp.find("-", firstSeparator + 1);
	if
		(std::string::npos == secondSeparator)
	{
		return false;
	}

	thirdSeparator = this->mValueTemp.find(" ", secondSeparator + 1);
	if
		(std::string::npos == thirdSeparator)
	{
		tmp = this->mValueTemp.substr(secondSeparator + 1
									  , this->mValueTemp.size() - (secondSeparator +1));
	}
	else
	{
		tmp = this->mValueTemp.substr( secondSeparator + 1
									 , thirdSeparator - (secondSeparator +1));
	}

	dayTmp = std::atoi(tmp.c_str());
	day = static_cast<std::uint8_t>(dayTmp);
	return true;
};

/**
 *
 */
void NanoguiDateBox::setValue(const std::string& value)
{
	NanoguiTextField::setValue(value);
	this->mValueTemp = value;
};

/**
 *
 */
bool NanoguiDateBox::commitValue()
{
	bool ok;	//

	ok = this->commitValue(this->mValueTemp);
	return ok;
};

/**
 *
 */
bool NanoguiDateBox::commitValue(const std::string& value)
{
	bool valid;	//
	std::string backup;	//

	valid = this->checkFormat(value, this->mFormat);
	LOG(LogLevel::Debug, "DateBox", "commitValue {} {}", value, valid);
	if
		(true == valid)
	{
		backup = this->mValue;
		this->mValue = value;
		if
			(this->mCallback && false == this->mCallback(value))
		{
			LOG(LogLevel::Debug, "DateBox", "Rollback to {}", backup);
			// rollback
			this->mValue = backup;	
			return false;
		}

		this->mValidFormat = true;
		this->mCommitted = true;
	}
	return true;
};

/**
 *
 */
std::size_t NanoguiDateBox::getHelperButtons(std::size_t count, std::vector<nanogui::Button*>& result)
{
	nanogui::Button* btn;	//
	std::size_t created;	//
	std::size_t existing;	//
	std::size_t i;	//

	existing = static_cast<std::size_t>(this->helper->childCount());
	created = 0;
	while
		(count > existing)
	{
		btn = new nanogui::Button(this->helper, "");
		created = created + 1;
		existing = existing + 1;
	}

	for
		(i = 0; i<count; i++)
	{
		btn = static_cast<nanogui::Button*>(this->helper->children()[i]);
		result.push_back(btn);
	}

	/* Hide remaining unused buttons */
	while
		(i < existing)
	{
		this->helper->children()[i]->setVisible(false);
		i = i + 1;
	}

	return created;
};

/**
 *
 */
bool NanoguiDateBox::displayYears()
{
	nanogui::Button* btn;	//
	std::vector<std::uint16_t>::iterator it;	//
	std::vector<nanogui::Button*> buttons;	//
	std::vector<nanogui::Button*>::iterator itButtons;	//
    std::ostringstream buffer("");
	int year;	//

	if
		(0 >= this->availableYears.size())
	{
		this->helper->setVisible(false);
		return false;
	}

	this->getHelperButtons(this->availableYears.size(), buttons);
	it = this->availableYears.begin();
	itButtons = buttons.begin();
	for
		(; it != this->availableYears.end(); ++it, ++itButtons)
	{
		auto hlFound = std::find( this->hlYears.begin()
								, this->hlYears.end()
								, *it);
		btn = *itButtons;
		if
			(this->hlYears.end() == hlFound)
		{
			btn->setTextColor(this->mTheme->mTextColor);
		}
		else
		{
			btn->setTextColor(nanogui::Color(230, 150, 50, 250));
		}
		year = static_cast<int>(*it);
        buffer.str("");
        buffer << std::setw(4) << std::setfill('0') << year;
		btn->setCaption(buffer.str());
		btn->setFontSize(14);
		btn->setFixedSize(Eigen::Vector2i(40, 20));
		btn->setCallback([this, year](){
			this->setYear(static_cast<std::uint16_t>(year));
		});
		btn->setVisible(true);
	}
	this->helper->autoSize = true;
	this->helper->layoutUpdateRequested = true;
	this->helper->setVisible(true);
	return true;
};

/**
 *
 */
bool NanoguiDateBox::displayMonths()
{
	nanogui::Button* btn;	//
	std::vector<std::uint8_t>::iterator it;	//
	std::vector<nanogui::Button*> buttons;	//
	std::vector<nanogui::Button*>::iterator itButtons;	//
    std::ostringstream buffer("");
	int month;	//

	if
		(0 >= this->availableMonths.size())
	{
		this->helper->setVisible(false);
		return false;
	}

	this->getHelperButtons(this->availableMonths.size(), buttons);
	it = this->availableMonths.begin();
	itButtons = buttons.begin();
	for
		(; it != this->availableMonths.end(); ++it, ++itButtons)
	{
		auto hlFound = std::find( this->hlMonths.begin()
								, this->hlMonths.end()
								, *it);
		btn = *itButtons;
		if
			(this->hlMonths.end() == hlFound)
		{
			btn->setTextColor(this->mTheme->mTextColor);
		}
		else
		{
			btn->setTextColor(nanogui::Color(230, 150, 50, 250));
		}
		month = static_cast<int>(*it);
        buffer.str("");
        buffer << std::setw(2) << std::setfill('0') << month;
		btn->setCaption(buffer.str());
		btn->setFontSize(14);
		btn->setFixedSize(Eigen::Vector2i(20, 20));
		btn->setCallback([this, month](){
			this->setMonth(static_cast<std::uint8_t>(month));
		});
		btn->setVisible(true);
	}
	this->helper->autoSize = true;
	this->helper->layoutUpdateRequested = true;
	this->helper->setVisible(true);
	return true;
};

/**
 *
 */
bool NanoguiDateBox::displayDays()
{
	nanogui::Button* btn;	//
	std::vector<std::uint8_t>::iterator it;	//
	std::vector<nanogui::Button*> buttons;	//
	std::vector<nanogui::Button*>::iterator itButtons;	//
    std::ostringstream buffer("");
	int day;	//

	if
		(0 >= this->availableDays.size())
	{
		this->helper->setVisible(false);
		return false;
	}

	this->getHelperButtons(this->availableDays.size(), buttons);
	it = this->availableDays.begin();
	itButtons = buttons.begin();
	for
		(; it != this->availableDays.end(); ++it, ++itButtons)
	{
		auto hlFound = std::find( this->hlDays.begin()
								, this->hlDays.end()
								, *it);
		btn = *itButtons;
		if
			(this->hlDays.end() == hlFound)
		{
			btn->setTextColor(this->mTheme->mTextColor);
		}
		else
		{
			btn->setTextColor(nanogui::Color(230, 150, 50, 250));
		}
		day = static_cast<int>(*it);
        buffer.str("");
        buffer << std::setw(2) << std::setfill('0') << day;
		btn->setCaption(buffer.str());
		btn->setFontSize(14);
		btn->setFixedSize(Eigen::Vector2i(20, 20));
		btn->setCallback([this, day](){
			this->setDay(static_cast<std::uint8_t>(day));
			this->showHelpers(HelperType::NONE);
		});
		btn->setVisible(true);
	}
	this->helper->autoSize = true;
	this->helper->layoutUpdateRequested = true;
	this->helper->setVisible(true);
	return true;
};
