// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ostream>
#include <iomanip>
#include "core/log.h"
#include "ui/nanogui/nanogui_datetimebox.h"

using namespace SEAScope;

/**
 *
 */
NanoguiDatetimeBox::NanoguiDatetimeBox( nanogui::Widget* parent, uint16_t year
									  , uint8_t month, uint8_t day
									  , uint8_t hour, uint8_t minute
									  , uint8_t second)
: nanogui::TextBox(parent), NanoguiTextField(parent), NanoguiDateBox(parent, year, month, day)
{
	std::ostringstream buffer;	//

	buffer << year << "-" << std::setfill('0') << std::setw(2) << static_cast<int>(month) << "-" << std::setfill('0') << std::setw(2) << static_cast<int>(day) << " " << std::setfill('0') << std::setw(2) << static_cast<int>(hour) << ":" << std::setfill('0') << std::setw(2) << static_cast<int>(minute) << ":" << std::setfill('0') << std::setw(2) << static_cast<int>(second);
	this->setValue(buffer.str());
	this->setFormat("-?[0-9]+-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}");
	this->defaultYearCompletion = "-01-01 00:00:00";
	this->defaultMonthCompletion = "-01 00:00:00";
	this->defaultDayCompletion = " 00:00:00";
	this->defaultHourCompletion = ":00:00";
	this->defaultMinuteCompletion = ":00";
};

/**
 *
 */
NanoguiDatetimeBox::~NanoguiDatetimeBox()
{

};

/**
 *
 */
bool NanoguiDatetimeBox::checkValidity(const std::string& input)
{
	bool okDate;	//
	bool okTime;	//
	size_t pos;	//
	std::string date;	//
	std::string time;	//

	LOG(LogLevel::Debug, "DatetimeBox", "Check format {}", input);

	pos = input.find(" ");
	if
		(std::string::npos == pos)
	{
		return false;
	}
	date = input.substr(0, pos);
	okDate = this->checkDate(date);

	time = input.substr(pos +1, input.size() - (pos +1));
	okTime = this->checkTime(time);

	return okDate && okTime;
};

/**
 *
 */
bool NanoguiDatetimeBox::checkTime(const std::string& input) const
{
	size_t pos;	//
	std::string tmp;	//
	uint8_t hour;	//
	uint8_t minute;	//
	uint8_t second;	//

	pos = input.find(":");
	tmp = input.substr(0,pos);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}
	hour = static_cast<std::uint8_t>(std::stoi(tmp));
	if
		(24 < hour)
	{
		return false;
	}

	tmp = input.substr(pos+1, 2);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}
	minute = static_cast<std::uint8_t>(std::stoi(tmp));
	if
		(59 < minute)
	{
		return false;
	}

	tmp = input.substr(pos+4, 2);
	if
		('0' == tmp[0])
	{
		tmp.erase(0,1);
	}
	second = static_cast<std::uint8_t>(std::stoi(tmp));
	if
		(59 < second)
	{
		return false;
	}
	return true;
};

/**
 *
 */
bool NanoguiDatetimeBox::setHour(std::uint8_t hour, bool commit)
{
	std::string tmp;	//
	std::ostringstream buffer;	//
	std::size_t timeSeparator;	//
	std::size_t pos;	//

	timeSeparator = this->mValueTemp.find(" ", 0);
	if
		(std::string::npos == timeSeparator)
	{
		return false;
	}

	pos = this->mValueTemp.find(":", timeSeparator + 1);
	if
		(std::string::npos == pos)
	{
		buffer << std::setfill('0') << std::setw(2) << static_cast<int>(hour) << this->defaultHourCompletion;
		tmp = this->mValueTemp.replace( timeSeparator + 1
							   		  , this->mValueTemp.length() - (timeSeparator +1)
							   		  , buffer.str());
	}
	else
	{
		buffer << std::setfill('0') << std::setw(2) << static_cast<int>(hour);
		tmp = this->mValueTemp.replace( timeSeparator + 1
									  , pos - (timeSeparator + 1)
									  , buffer.str());
	}

	LOG(LogLevel::Debug, "DatetimeBox", "After set hour: {} {} {}" , static_cast<int>(hour), this->mValueTemp, tmp);
	if
		(true == commit)
	{
		return this->commitValue(tmp);
	}
	return false;
};

/**
 *
 */
bool NanoguiDatetimeBox::getHour(std::uint8_t& result) const
{
	std::string tmp;	//
	std::size_t timeSeparator;	//
	std::size_t pos;	//
	int hourTmp;	//

	timeSeparator = this->mValueTemp.find(" ", 0);
	if
		(std::string::npos == timeSeparator)
	{
		return false;
	}

	pos = this->mValueTemp.find(":", timeSeparator + 1);
	if
		(std::string::npos == pos)
	{
		tmp = this->mValueTemp.substr( timeSeparator + 1
							   		 , this->mValueTemp.length() - (timeSeparator +1));
	}
	else
	{
		tmp = this->mValueTemp.substr( timeSeparator + 1
									 , pos - (timeSeparator + 1));
	}

	hourTmp = std::atoi(tmp.c_str());
	result = static_cast<std::uint8_t>(hourTmp);
	return true;
};

/**
 *
 */
bool NanoguiDatetimeBox::setMinute(std::uint8_t minute, bool commit)
{
	std::string tmp;	//
	std::ostringstream buffer;	//
	std::size_t timeSeparator;	//
	std::size_t firstSeparator;	//
	std::size_t secondSeparator;	//

	timeSeparator = this->mValueTemp.find(" ", 0);
	firstSeparator = this->mValueTemp.find(":", timeSeparator + 1);
	if
		(  std::string::npos == timeSeparator
		|| std::string::npos == firstSeparator)
	{
		return false;
	}

	secondSeparator = this->mValueTemp.find(":", firstSeparator + 1);
	if
		(std::string::npos == secondSeparator)
	{
		buffer << std::setfill('0') << std::setw(2) << static_cast<int>(minute) << this->defaultMinuteCompletion;
		tmp = this->mValueTemp.replace( firstSeparator + 1
							   		  , this->mValueTemp.length() - (firstSeparator +1)
							   		  , buffer.str());
	}
	else
	{
		buffer << std::setfill('0') << std::setw(2) << static_cast<int>(minute);
		tmp = this->mValueTemp.replace( firstSeparator + 1
									  , secondSeparator - (firstSeparator + 1)
									  , buffer.str());
	}

	if
		(true == commit)
	{
		return this->commitValue(tmp);
	}
	return false;
};

/**
 *
 */
bool NanoguiDatetimeBox::getMinute(std::uint8_t& result) const
{
	std::string tmp;	//
	std::size_t timeSeparator;	//
	std::size_t firstSeparator;	//
	std::size_t secondSeparator;	//
	int minuteTmp;	//

	timeSeparator = this->mValueTemp.find(" ", 0);
	firstSeparator = this->mValueTemp.find(":", timeSeparator + 1);
	if
		(  std::string::npos == timeSeparator
		|| std::string::npos == firstSeparator)
	{
		return false;
	}

	secondSeparator = this->mValueTemp.find(":", firstSeparator + 1);
	if
		(std::string::npos == secondSeparator)
	{
		tmp = this->mValueTemp.substr( firstSeparator + 1
							   		 , this->mValueTemp.length() - (firstSeparator +1));
	}
	else
	{
		tmp = this->mValueTemp.substr( firstSeparator + 1
									 , secondSeparator - (firstSeparator + 1));
	}

	minuteTmp = std::atoi(tmp.c_str());
	result = static_cast<std::uint8_t>(minuteTmp);
	return true;
};

/**
 *
 */
bool NanoguiDatetimeBox::setSecond(std::uint8_t second, bool commit)
{
	std::string tmp;	//
	std::ostringstream buffer;	//
	std::size_t timeSeparator;	//
	std::size_t firstSeparator;	//
	std::size_t secondSeparator;	//

	timeSeparator = this->mValueTemp.find(" ", 0);
	firstSeparator = this->mValueTemp.find(":", timeSeparator + 1);
	secondSeparator = this->mValueTemp.find(":", firstSeparator + 1);
	if
		(  std::string::npos == timeSeparator
		|| std::string::npos == firstSeparator
		|| std::string::npos == secondSeparator)
	{
		return false;
	}

	buffer << std::setfill('0') << std::setw(2) << static_cast<int>(second);
	tmp = this->mValueTemp.replace( secondSeparator + 1
								  , this->mValueTemp.length() - (secondSeparator + 1)
								  , buffer.str());

	if
		(true == commit)
	{
		return this->commitValue(tmp);
	}
	return false;
};

/**
 *
 */
bool NanoguiDatetimeBox::getSecond(std::uint8_t& result) const
{
	std::string tmp;	//
	std::size_t timeSeparator;	//
	std::size_t firstSeparator;	//
	std::size_t secondSeparator;	//
	int secondTmp;	//

	timeSeparator = this->mValueTemp.find(" ", 0);
	firstSeparator = this->mValueTemp.find(":", timeSeparator + 1);
	secondSeparator = this->mValueTemp.find(":", firstSeparator + 1);
	if
		(  std::string::npos == timeSeparator
		|| std::string::npos == firstSeparator
		|| std::string::npos == secondSeparator)
	{
		return false;
	}

	tmp = this->mValueTemp.substr( secondSeparator + 1
								 , this->mValueTemp.length() - (secondSeparator + 1));
	secondTmp = std::atoi(tmp.c_str());
	result = static_cast<std::uint8_t>(secondTmp);
	return true;
};
