// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>
#include <sstream>
#include <iomanip>

#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_days_list.h"

using namespace SEAScope;

/**
 *
 */
NanoguiDaysList::NanoguiDaysList(nanogui::Widget* parent)
: nanogui::Widget(parent)
, year(1970), month(1)
, dayCallback(nullptr), scrollCallback(nullptr)
, daysInMonth(0), pastOffset(0), futureOffset(0)
, hoverEnabled(false)
{
	std::size_t i;	//
	std::ostringstream buffer;	//

	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Horizontal
										   , nanogui::Alignment::Middle
										   , 2, 2));

	i = 0;
	for
		(; i < 31; ++i)
	{
		buffer << std::setfill('0') << std::setw(2) << (1+i);
		this->buttons[i] = new NanoguiFlatButton(this);
		this->buttons[i]->setCaption(buffer.str());
		this->buttons[i]->setFixedSize(Eigen::Vector2i(20, 20));
		this->buttons[i]->setFontSize(18);
		this->buttons[i]->setBackgroundColor(nanogui::Color(0,0,0,0));
		this->buttons[i]->mFontFace = "sans-bold";
		this->buttons[i]->setTextColor(nanogui::Color(90,90,90,255));
		this->buttons[i]->setCallback([this, i] () {
			if
				(  nullptr != this->dayCallback
				&& i < this->daysInMonth)
			{
				this->dayCallback(static_cast<std::uint8_t>(1 + i));
			}
		});
		buffer.str("");
	}
};

/**
 *
 */
NanoguiDaysList::~NanoguiDaysList()
{
};

/**
 *
 */
void NanoguiDaysList::setSize(const Eigen::Vector2i& size)
{
	nanogui::Widget::setSize(size);
	this->resize(size.x());
};

/**
 *
 */
void NanoguiDaysList::setWidth(int width)
{
	nanogui::Widget::setWidth(width);
	this->resize(width);
};

/**
 *
 */
void NanoguiDaysList::setFixedSize(const Eigen::Vector2i& size)
{
	nanogui::Widget::setFixedSize(size);
	this->resize(size.x());
};

/**
 *
 */
void NanoguiDaysList::setFixedWidth(int width)
{
	nanogui::Widget::setFixedWidth(width);
	this->resize(width);
};

/**
 *
 */
bool NanoguiDaysList::selectDay(std::uint8_t day)
{
	size_t i;	//

	i = 0;
	for
		(;i<31; ++i)
	{
		if
			(i+1 == day)
		{
			this->buttons[i]->setBorderColor(nanogui::Color(255, 255, 255, 255));
		}
		else
		{
			this->buttons[i]->setBorderColor(nanogui::Color(0,0,0,0));
		}
	}
	return false;
};

bool NanoguiDaysList::setMonth(std::uint16_t yearIn, std::uint8_t monthIn)
{
	std::uint8_t daysInMonthTmp;	//
	std::array<std::uint8_t, 12> daysPerMonth = {{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}};	//

	if
		(  yearIn == this->year
		&& monthIn == this->month)
	{
		return false;
	}

	this->year = yearIn;
	this->month = monthIn;
	daysInMonthTmp = daysPerMonth[this->month - 1u];
	if
		(2 == this->month && 0 == this->year%4)
	{
		daysInMonthTmp = 1u + daysInMonthTmp;
	}

	this->daysInMonth = daysInMonthTmp;

	this->update();
	return true;
};

/**
 *
 */
bool NanoguiDaysList::setMainCoverage(const EventsCoverage& coverage)
{
	this->mainCoverage = coverage;
	this->update();
	return true;
};

/**
 *
 */
bool NanoguiDaysList::setExtraCoverage(const EventsCoverage& coverage)
{
	this->extraCoverage = coverage;
	this->update();
	return true;
};

/**
 *
 */
bool NanoguiDaysList::setVisibilityBounds(timestamp_t past, timestamp_t future)
{
	bool changed;	//

	changed = false;
	if
		(this->pastOffset != past)
	{
		this->pastOffset = past;
		changed = true;
	}

	if
		(this->futureOffset != future)
	{
		this->futureOffset = future;
		changed = true;
	}

	if
		(changed)
	{
		this->update();
	}
	return changed;
};

/**
 *
 */
bool NanoguiDaysList::update()
{
	uint8_t i;	//
	bool buttonEnabled;	//
	std::vector<std::uint8_t> mainDays;	//
	std::vector<std::uint8_t> extraDays;	//
	NanoguiFlatButton* btn;	//

	this->mainCoverage.getDays(this->year, this->month, mainDays);
	this->extraCoverage.getDays(this->year, this->month, extraDays);

	std::tm t = {};
	t.tm_year = this->year - 1900;
	t.tm_mon = this->month - 1;
	t.tm_mday = 1;
	t.tm_hour = 0;
	t.tm_min = 0;
	t.tm_sec = 0;
	time_t timestamp = std::mktime(&t);
	LOG(LogLevel::Debug, "Days list",
             "updating list with span, starting at {}", timestamp);
	i = 0;
	for
		(;i<31; ++i)
	{
		btn = this->buttons[i];
		buttonEnabled = i < this->daysInMonth;
		btn->setEnabled(buttonEnabled);

		auto found = std::find( mainDays.begin()
							  , mainDays.end()
							  , i+1);
		if
			(mainDays.end() != found)
		{
			this->textColors[i] = nanogui::Color(250,250,250,255);
		}
		else
		{
			this->textColors[i] = nanogui::Color(90,90,90,255);
		}

		found = std::find( extraDays.begin()
						 , extraDays.end()
						 , i+1);
		if
			(extraDays.end() != found)
		{
			this->textColors[i] = nanogui::Color(30,210,50,255);
		}

		btn->setBackgroundColor(this->backgrounds[i]);
		btn->setTextColor(this->textColors[i]);

		timestamp_t dayStart = static_cast<timestamp_t>(timestamp) + i * 86400;
		timestamp_t dayStop = dayStart + 86400;
		if
			(  this->pastOffset > dayStop
			|| this->futureOffset < dayStart)
		{
			LOG(LogLevel::Debug, "Days list", "CLEAN DAY {} ({} {} {} {})",
                     i+1, dayStart, dayStop,
                     this->futureOffset, this->pastOffset);
			btn->setWrapline(-1.f, -1.f);
		}
		else
		{
			timestamp_t offset = (this->pastOffset < dayStart) ? dayStart:this->pastOffset;
			timestamp_t width = (this->futureOffset > dayStop) ? dayStop - offset:this->futureOffset - offset;
			btn->setWrapline( 1.f * static_cast<float>(offset - dayStart) / 86400
							 , 1.f * static_cast<float>(width) / 86400);
			LOG(LogLevel::Debug, "Days list", "VISIBLE DAY {}% {}%", 100 * (offset - dayStart) / 86400, 100 * width / 86400);
			LOG(LogLevel::Debug, "Days list", "{} {} {} {}", dayStart, dayStop, this->futureOffset, this->pastOffset);
		}
	}

	return true;
};

/**
 *
 */
bool NanoguiDaysList::resize(int width)
{
	std::size_t i;	//
	int btnWidth;	//
	int spacing;	//
	nanogui::BoxLayout* layout;	//

	if
		(0 == width)
	{
		return false;
	}

	btnWidth = static_cast<int>(std::floor((width - 2 * 30) / 31));
	spacing = static_cast<int>(std::floor((width - 31 * btnWidth) / 30));
	layout = static_cast<nanogui::BoxLayout*>(this->layout());
	layout->setSpacing(spacing);

	i = 0;
	for
		(;i < 31; ++i)
	{
		this->buttons[i]->setFixedWidth(btnWidth);
	}
	return true;
};

/**
 *
 */
bool NanoguiDaysList::scrollEvent(const Eigen::Vector2i& /*p*/,
                                  const Eigen::Vector2f& rel)
{
    float scrollDelta;

    // Give priority to the horizontal axis because it is the natural direction
    // for the timeline. Typical mouses will only provide the vertical
    // component but touch pads and other specific devices allow horizontal
    // scrolling.
    scrollDelta = rel.x();
    if
        (std::fabs(scrollDelta) < 0.01f)
    {
        scrollDelta = rel.y();
    }

    if
        (nullptr != this->scrollCallback)
    {
        this->scrollCallback(scrollDelta);
    }
    return true; // prevent event propagation
};

/**
 *
 */
void NanoguiDaysList::draw(NVGcontext* ctx)
{
    if
        (this->hoverEnabled && this->mMouseFocus)
    {
        nvgBeginPath(ctx);
        nvgRect(ctx,
                static_cast<float>(this->mPos.x() + 1),
                static_cast<float>(this->mPos.y() + 1),
                static_cast<float>(this->mSize.x() - 2),
                static_cast<float>(this->mSize.y() - 2));
        nvgFillColor(ctx, nanogui::Color(20, 20, 20, 100));
        nvgFill(ctx);
    }
    nanogui::Widget::draw(ctx);
};
