// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_draggable_pane.h"
#include "core/log.h"

using namespace SEAScope;

/**
 * Constructor.
 */
NanoguiDraggablePane::NanoguiDraggablePane(nanogui::Widget* parent):
nanogui::Widget(parent), reflowNeeded(true), dragging(false), headerHeight(0),
keyCallback(nullptr), charCallback(nullptr)
{
    this->backgroundColor = nanogui::Color(0, 0, 0, 150);
    this->headerBackgroundColor = nanogui::Color(30, 30, 30, 235);

	this->setLayout(new nanogui::BoxLayout( nanogui::Orientation::Vertical
										  , nanogui::Alignment::Minimum
										  , 0, 1));
};

/**
 * Destructor.
 */
NanoguiDraggablePane::~NanoguiDraggablePane()
{
};

/**
 *
 */
bool NanoguiDraggablePane::display(const Eigen::Vector2i& pos)
{
    bool ok;    //

    this->setPosition(pos);
    ok = ! this->visible();
    this->setVisible(true);
    return ok;
};

/**
 *
 */
bool NanoguiDraggablePane::hide()
{
    bool ok;    //
    ok = this->visible();
    this->setVisible(false);
    return ok;
};

/**
 *
 */
bool NanoguiDraggablePane::updateLayout()
{
    nanogui::Screen* screen;    //
    NVGcontext *ctx;    //

    screen = static_cast<nanogui::Screen *>(this->parent());
    ctx = screen->nvgContext();

    const Eigen::Vector2i& size = this->preferredSize(ctx);
    this->setSize(size);
    this->reflowNeeded = true;
    return true;
};

/**
 *
 */
void NanoguiDraggablePane::draw(NVGcontext* ctx)
{
	if
		(true == this->reflowNeeded)
	{
		this->performLayout(ctx);
		this->reflowNeeded = false;
	}

	nvgBeginPath(ctx);
    nvgRect(ctx,
            static_cast<float>(this->mPos.x()),
            static_cast<float>(this->mPos.y()),
            static_cast<float>(this->mSize.x()),
            static_cast<float>(this->mSize.y()));
    nvgFillColor(ctx, this->backgroundColor);
    nvgFill(ctx);

    if
        (0 < this->headerHeight)
    {
        nvgBeginPath(ctx);
        nvgRect(ctx,
                static_cast<float>(this->mPos.x()),
                static_cast<float>(this->mPos.y()),
                static_cast<float>(this->mSize.x()),
                static_cast<float>(this->headerHeight));
        nvgFillColor(ctx, this->headerBackgroundColor);
        nvgFill(ctx);
    }

	nanogui::Widget::draw(ctx);
};

/**
 *
 */
bool NanoguiDraggablePane::mouseButtonEvent(const Eigen::Vector2i& pos, int buttons, bool pressed, int mods)
{
	bool ok;	//
	bool inHorizontalRange;	//
	bool inVerticalRange;	//
	int relX;	//
	int relY;	//

	/* Let children handle the event if they can */
	ok = nanogui::Widget::mouseButtonEvent(pos, buttons, pressed, mods);

	if
		(false == pressed)
	{
		this->dragging = false;
	}
	else
	{
		if
			(ok)
		{
			this->dragging = false;
		}

		relX = pos.x() - this->mPos.x();
		inHorizontalRange = this->mSize.x() >= relX;

		relY = pos.y() - this->mPos.y();
		inVerticalRange = (0 >= this->headerHeight && this->mSize.y() >= relY);
		inVerticalRange = inVerticalRange || (static_cast<int>(this->headerHeight) >= relY); /*No effect if headerHeight = 0*/

		this->dragging = inHorizontalRange && inVerticalRange;
	}

	/* Stop event propagation */
	return true;
};

/**
 *
 */
bool NanoguiDraggablePane::mouseMotionEvent(const Eigen::Vector2i& pos, const Eigen::Vector2i& rel,
                            int button, int modifiers)
{
	bool ok;	//

	/* Let children handle the event if they can */
	ok = nanogui::Widget::mouseMotionEvent(pos, rel, button, modifiers);
	if
		(ok)
	{
		return ok;
	}

	if
		(false == this->dragging)
	{
		return false;
	}

    nanogui::Screen* screen {nullptr};
    screen = static_cast<nanogui::Screen *>(this->parent());
    if
        (  (0 >= pos.x() * pos.y())
        || (0 >= (screen->width() - pos.x()) * (screen->height() - pos.y())))
    {
        // The mouse cursor left the window, reset dragging status
        this->dragging = false;
    }

    if
		((button & (1 << GLFW_MOUSE_BUTTON_1)) != 0)
	{
        this->mPos += rel;
        this->mPos = this->mPos.cwiseMax(Eigen::Vector2i::Zero());
        this->mPos = this->mPos.cwiseMin(this->parent()->size() - mSize);
        return true;
    }
    return false;
};

/**
 *
 */
bool NanoguiDraggablePane::keyboardEvent(int key, int scancode, int action, int modifiers)
{
    bool ok;    //

    ok = nanogui::Widget::keyboardEvent(key, scancode, action, modifiers);
    if
        (true == ok)
    {
        return true;
    }

    if
        (nullptr != this->keyCallback)
    {
        return this->keyCallback(key, scancode, action, modifiers);
    }

    return false;
};

/**
 *
 */
bool NanoguiDraggablePane::keyboardCharacterEvent(unsigned int codepoint)
{
    bool ok;    //

    ok = nanogui::Widget::keyboardCharacterEvent(codepoint);
    if
        (true == ok)
    {
        return true;
    }

    if
        (nullptr != this->charCallback)
    {
        return this->charCallback(codepoint);
    }

    return false;

};


/**
 *
 */
bool NanoguiDraggablePane::scrollEvent(const Eigen::Vector2i &p, const Eigen::Vector2f &rel)
{
	// Let children handle the event if they can
	nanogui::Widget::scrollEvent(p, rel);

	// Prevent event propagation
	return true;
};
