// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <forward_list>

#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_extraction_dialog.h"
#include "ui/nanogui/nanogui_painted_label.h"
#include "core/path.h"
#include "core/log.h"

using namespace SEAScope;

/**
 * Constructor.
 */
NanoguiExtractionDialog::NanoguiExtractionDialog(nanogui::Widget* parent)
: nanogui::Widget(parent), NanoguiScrollableWithTitle(parent),
pngButton(nullptr), pngCallback(nullptr),
pickleButton(nullptr), pickleCallback(nullptr),
_processorOptions(nullptr), _extractedItems(nullptr)
{
    nanogui::Label* local_title;  //
	nanogui::Widget* spacer;	//
	nanogui::Widget* actions;	//

    this->footer->setFixedHeight(3);  // add extra margin at the bottom

	this->header->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                   nanogui::Alignment::Fill,
												   2, 0));
	this->header->setFixedHeight(90);

    spacer = new nanogui::Widget(this->header);
    spacer->setFixedHeight(16);

	this->_processorOptions = new nanogui::Widget(this->header);
    this->_processorOptions->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                           nanogui::Alignment::Minimum,
                                           0, 0));
	this->_processorOptions->setFixedHeight(55);

    local_title = new nanogui::Label(this->_processorOptions, "Save extracted data as", "sans-bold");
    local_title->setFontSize(16);
	local_title->setFixedHeight(16);

	actions = new nanogui::Widget(this->_processorOptions);
	actions->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
											  nanogui::Alignment::Middle,
											  6, 3));

	spacer = new nanogui::Widget(actions);
	spacer->setFixedWidth(20);

	this->pngButton = new nanogui::Button(actions, "PNG images");
	this->pngButton->setTooltip("Save extracted data as PNG files [Python processing server required]");
	this->pngButton->setFixedWidth(100);
	this->pngButton->setFontSize(14);
	this->pngButton->setCallback([this]() {
		if
			(nullptr != this->pngCallback)
		{
			this->pngCallback();
		}
	});

	spacer = new nanogui::Widget(actions);
	spacer->setFixedWidth(44);

	this->pickleButton = new nanogui::Button(actions, "Python object");
	this->pickleButton->setTooltip("Save extracted data as a serialized Python object [Python processing server required]");
	this->pickleButton->setFixedWidth(100);
	this->pickleButton->setFontSize(14);
	this->pickleButton->setCallback([this]() {
		if
			(nullptr != this->pickleCallback)
		{
			this->pickleCallback();
		}
	});

	spacer = new nanogui::Widget(this->_processorOptions);
	spacer->setFixedHeight(10);

	local_title = new nanogui::Label(this->header, "Extraction contains data from:", "sans-bold");
	local_title->setFontSize(16);
	local_title->setFixedHeight(16);

	spacer = new nanogui::Widget(this->header);
	spacer->setFixedHeight(3);

	this->scrollable->content->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
																nanogui::Alignment::Minimum,
																8, 0));

    this->_extractedItems = new nanogui::Widget(this->scrollable->content);
    this->_extractedItems->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                         nanogui::Alignment::Minimum,
                                                         0, 2));
    new nanogui::Label(this->_extractedItems, "Nothing! No extraction performed yet!", "sans-bold");
	this->reflowNeeded = true;
};

/**
 * Destructor.
 */
NanoguiExtractionDialog::~NanoguiExtractionDialog()
{
};

/**
 *
 */
bool NanoguiExtractionDialog::setExtraction(const std::map<RenderableId, Extraction>& data, const std::map<RenderableId, std::string>& labels, bool /*showRemoteProcessingOptions*/)
{
    std::ostringstream buffer("");  //
    NanoguiPaintedLabel* lbl(nullptr);   //
	int childrenCount(-1);	//

	childrenCount = this->_extractedItems->childCount();
	while
		(--childrenCount > -1)
	{
		this->_extractedItems->removeChild(childrenCount);
	}

    nanogui::Color light(205, 205, 205, 30);
    nanogui::Color dark(105, 105, 105, 30);
    childrenCount = 0;
    for
        (const auto& it : data)
    {
        LOG(LogLevel::Debug, "Extraction dialog", "{}/{}/{}:{}",
                it.first.collection.first, it.first.collection.second,
                it.first.granule.second, it.first.variable);

		if
			("mask" == it.second.uri)
		{
			continue;
		}

        auto labelFound = labels.find(it.first);
        if
            (labels.end() == labelFound)
        {
            LOG(LogLevel::Error, "Extraction dialog", "Could not find label associated with renderable {}/{}/{}:{}",
                     it.first.collection.first, it.first.collection.second, it.first.granule.second, it.first.variable);
            return false;
        }
        lbl = new NanoguiPaintedLabel(this->_extractedItems, labelFound->second);
        lbl->setFixedWidth(304);  // leave a margin for scrollbar
        lbl->setWidth(304);  // leave a margin for scrollbar
		lbl->setFontSize(14);
		lbl->setTooltip(it.second.uri);
        if
            (0 < childrenCount % 2)
        {
            lbl->setBackgroundColor(light);
        }
        else
        {
            lbl->setBackgroundColor(dark);
        }

        childrenCount = childrenCount + 1;
    }

	this->scrollable->content->update();

    lbl = nullptr;

    return false;
};

/**
 *
 */
void NanoguiExtractionDialog::showProcessingOptions()
{
    this->header->setFixedHeight(90);
    this->_processorOptions->setVisible(true);
    this->reflowNeeded = true;
}

/**
 *
 */
void NanoguiExtractionDialog::hideProcessingOptions()
{
    this->header->setFixedHeight(35);
    this->_processorOptions->setVisible(false);
    this->reflowNeeded = true;
}
