// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>
#include <sstream>

#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_granule_dialog.h"
#include "core/measurements.h"
#include "core/log.h"
#include "core/granule_metadata.h"

using namespace SEAScope;

/**
 * Constructor.
 */
NanoguiGranuleDialog::NanoguiGranuleDialog(nanogui::Widget* parent):
nanogui::Widget(parent), btns(nullptr), collectionAttr(nullptr),
granuleAttr(nullptr), variableAttr(nullptr), timeCoverageAttr(nullptr),
spatialCoverageAttr(nullptr), infoCallback(nullptr),
centerTimeCallback(nullptr), extractCallback(nullptr)
{
	nanogui::Widget* attrs(nullptr);	//
	nanogui::Button* btn(nullptr);	//
    nanogui::Widget* spacer(nullptr);   //
    nanogui::Widget* upperWidget(nullptr);  //

	this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
										   nanogui::Alignment::Fill,
										   0, 6));

    upperWidget = new nanogui::Widget(this);
    upperWidget->setFixedWidth(360);
    upperWidget->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                  nanogui::Alignment::Minimum,
                                                  0, 0));
	attrs = new nanogui::Widget(upperWidget);
	attrs->setFixedWidth(336);
	attrs->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
										    nanogui::Alignment::Minimum,
										    3, 0));

	this->btns = new nanogui::Widget(upperWidget);
	this->btns->setFixedWidth(24);
	this->btns->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
										   nanogui::Alignment::Middle,
										   0, 4));

	this->granuleAttr = new NanoguiAttribute(attrs);
	this->granuleAttr->nameLbl->setFixedWidth(332);
	this->granuleAttr->valueLbl->setFixedWidth(322);
	this->granuleAttr->setName("Granule");
	this->timeCoverageAttr = new NanoguiAttribute(attrs);
	this->timeCoverageAttr->nameLbl->setFixedWidth(332);
	this->timeCoverageAttr->valueLbl->setFixedWidth(322);
	this->timeCoverageAttr->setName("Time coverage");
	this->spatialCoverageAttr = new NanoguiAttribute(attrs);
	this->spatialCoverageAttr->nameLbl->setFixedWidth(332);
	this->spatialCoverageAttr->valueLbl->setFixedWidth(322);
	this->spatialCoverageAttr->setName("Spatial coverage");
	this->collectionAttr = new NanoguiAttribute(attrs);
	this->collectionAttr->nameLbl->setFixedWidth(332);
	this->collectionAttr->valueLbl->setFixedWidth(322);
	this->collectionAttr->setName("Collection");
	this->variableAttr = new NanoguiAttribute(attrs);
	this->variableAttr->nameLbl->setFixedWidth(332);
	this->variableAttr->valueLbl->setFixedWidth(322);
	this->variableAttr->setName("Variable");

    spacer = new nanogui::Widget(this->btns);
    spacer->setFixedHeight(13);

	btn = new nanogui::Button(this->btns, "", ENTYPO_ICON_PUBLISH);
	btn->setTooltip("Extract data that ovrlap the footprint of this granule");
	btn->setFixedSize(Eigen::Vector2i(20, 20));
	btn->setCallback([this]() {
		if
			(nullptr != this->extractCallback)
		{
			this->extractCallback(this->target);
		}
	});

	btn = new nanogui::Button(this->btns, "", ENTYPO_ICON_INFO);
	btn->setTooltip("Info");
	btn->setFixedSize(Eigen::Vector2i(20, 20));
	btn->setCallback([this]() {
		if
			(nullptr != this->infoCallback)
		{
			this->infoCallback(this->target);
		}
	});

	btn = new nanogui::Button(this->btns, "", ENTYPO_ICON_CLOCK);
	btn->setTooltip("Center the timeline on this granule's datetime");
	btn->setFixedSize(Eigen::Vector2i(20, 20));
	btn->setCallback([this]() {
		if
			(nullptr != this->centerTimeCallback)
		{
			this->centerTimeCallback(this->target);
		}
	});

    spacer = new nanogui::Widget(this->btns);
    spacer->setFixedHeight(13);
};

/**
 * Destructor.
 */
NanoguiGranuleDialog::~NanoguiGranuleDialog()
{
};

/**
 *
 */
bool NanoguiGranuleDialog::enable(const RenderableId& rId, const GranuleInfo& info)
{
	std::ostringstream buffer;	//
	char startStr[20];	//
	char stopStr[20];	//
	time_t timestampStart;	//
	time_t timestampStop;	//
	std::tm* ltimeStart;	//
	std::tm* ltimeStop;	//
    std::string tmpStr; //
    bool hasLength; //
    bool hasArea;   //
    bool ok;    //

	this->target = rId;

	LOG(LogLevel::Debug, "Granule dialog", "enable");

	if
		(0 < info.collection.label.size())
	{
		this->collectionAttr->setValue(info.collection.label);
		this->collectionAttr->setVisible(true);
	}
	else
	{
		this->collectionAttr->setVisible(false);
	}

	if
		(true == info.hasMetadata)
	{
		/* start */
		timestampStart = static_cast<long>(.001 * static_cast<double>(info.metadata.timeCoverageStart()));
		ltimeStart = std::localtime(&timestampStart);
		strftime(startStr, sizeof(startStr), "%Y-%m-%d %H:%M:%S", ltimeStart);

		/* stop */
		timestampStop = static_cast<long>(.001 * static_cast<double>(info.metadata.timeCoverageEnd()));
		ltimeStop = std::localtime(&timestampStop);
		strftime(stopStr, sizeof(stopStr), "%Y-%m-%d %H:%M:%S", ltimeStop);

		buffer << startStr << " ... " << stopStr;
		this->timeCoverageAttr->setValue(buffer.str());
		this->timeCoverageAttr->setVisible(true);

		if
			(0 < info.metadata.dataId().size())
		{
			this->granuleAttr->setValue(info.metadata.dataId());
			this->granuleAttr->setVisible(true);
		}
		else
		{
			this->granuleAttr->setVisible(false);
		}
	}
	else
	{
		this->timeCoverageAttr->setVisible(false);
	}

	if
		(0 < info.variable.label.size())
	{
		this->variableAttr->setValue(info.variable.label);
		this->variableAttr->setVisible(true);
	}
	else
	{
		this->variableAttr->setVisible(false);
	}

	buffer.str("");
    hasArea = (false == std::isnan(info.area));
    if
        (true == hasArea)
    {
        ok = SEAScope::formatArea(static_cast<double>(info.area), tmpStr);
        if
            (true == ok)
        {
            buffer << "Area: " << tmpStr;
        }
    }

    hasLength = (false == std::isnan(info.perimeter));
    if
        (true == hasLength)
    {
        if
            (true == hasArea)
        {
            ok = SEAScope::formatDistance(static_cast<double>(info.perimeter), tmpStr);
            if
                (true == ok)
            {
                buffer << "\nPerimeter: " << tmpStr;
            }
        }
        else
        {
            ok = SEAScope::formatDistance(static_cast<double>(info.perimeter), tmpStr);
            if
                (true == ok)
            {
                buffer << "Length: " << tmpStr;
            }
        }

    }
	this->spatialCoverageAttr->setValue(buffer.str());

	this->setVisible(true);
	return true;
};

/**
 *
 */
bool NanoguiGranuleDialog::disable()
{
	LOG(LogLevel::Debug, "Granule dialog", "disable");
	if
		(true == this->visible())
	{
		this->setVisible(false);
        this->setFocused(false);
		return true;
	}
	return false;
};

/**
 *
 */
void NanoguiGranuleDialog::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float btnsPosX; //
    float btnsPosY; //
    float btnsSizeX;    //
    float btnsSizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());

    btnsPosX = static_cast<float>(this->btns->position().x());
    btnsPosY = static_cast<float>(this->btns->position().y());
    btnsSizeX = static_cast<float>(this->btns->size().x());
    btnsSizeY = static_cast<float>(this->btns->size().y());

	nvgBeginPath(ctx);
	nvgMoveTo(ctx,
              posX + btnsPosX,
              posY + btnsPosY + 15.f /*13 spacer + 2 align*/);
    nvgLineTo(ctx,
              posX + btnsPosX,
              posY + btnsPosY + btnsSizeY - 15.f /*13 spacer + 2 align*/ + 1.f /*line width?*/);
    nvgBezierTo(ctx,
                posX + btnsPosX,
                posY + btnsPosY + btnsSizeY - 7.f + 1.f /*line width?*/,
                posX + btnsPosX + btnsSizeX,
                posY + btnsPosY + btnsSizeY - 15.f /*13 spacer + 2 align*/ + 1.f /*line width?*/,
                posX + btnsPosX + btnsSizeX,
                posY + btnsPosY + btnsSizeY);
    nvgLineTo(ctx,
              posX + btnsPosX + btnsSizeX,
              posY + btnsPosY);
    nvgBezierTo(ctx,
                posX + btnsPosX + btnsSizeX,
                posY + btnsPosY + 15.f /*13 spacer + 2 align*/,
                posX + btnsPosX,
                posY + btnsPosY + 7.f,
                posX + btnsPosX,
                posY + btnsPosY + 15.f);
    nvgStrokeColor(ctx, nanogui::Color(96, 96, 96, 180));
    nvgStroke(ctx);
	nvgFillColor(ctx, nanogui::Color(0,0,0,255));
	nvgFill(ctx);

	nanogui::Widget::draw(ctx);
};
