// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"

#ifdef _WIN32
#define _USE_MATH_DEFINES
#endif
#include <math.h>

#include "ui/nanogui/nanogui_utils.h"
#include "ui/nanogui/nanogui_hcolorbar.h"

using namespace SEAScope;

/**
 *
 */
NanoguiHColorbar::NanoguiHColorbar(nanogui::Widget* parent): nanogui::Widget(parent), orientation(nanogui::Orientation::Horizontal), textureId(-1), textureNeeded(false)
{
	this->color[0] = 0;
	this->color[1] = 0;
	this->color[2] = 0;
	this->setFixedSize(Eigen::Vector2i(50, 10));
};

/**
 *
 */
NanoguiHColorbar::~NanoguiHColorbar()
{
};

/**
 *
 */
Eigen::Vector2i NanoguiHColorbar::preferredSize(NVGcontext* /*ctx*/) const
{
	return this->mFixedSize;
}

/**
 *
 */
void NanoguiHColorbar::setSize(const Eigen::Vector2i& size)
{
	this->textureNeeded = true;
	nanogui::Widget::setSize(size);
};

/**
 *
 */
void NanoguiHColorbar::setFixedSize(const Eigen::Vector2i& size)
{
	this->textureNeeded = true;
	nanogui::Widget::setFixedSize(size);
};

/**
 *
 */
bool NanoguiHColorbar::setColormap(const Colormap& cmapIn)
{
	this->cmap = cmapIn;
	this->textureNeeded = true;
	return true;
};

/**
 *
 */
bool NanoguiHColorbar::setColor(const std::array<std::uint8_t, 3>& value)
{
	std::vector<std::array<std::uint8_t, 3> > dummy;	//
	this->cmap = dummy;
	this->color = value;
	this->textureNeeded = false;
	return true;
};

/**
 *
 */
void NanoguiHColorbar::drawColormap(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //
	NVGpaint paint;	//
	std::tuple<int, NVGpaint> paintTuple;	//

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

	if
		(true == this->textureNeeded)
	{
		if
			(-1 < this->textureId)
		{
			nvgDeleteImage(ctx, this->textureId);
			this->textureId = -1;
		}

		if
			(nanogui::Orientation::Vertical == this->orientation)
		{
			paintTuple = NanoguiUtils::nvgLinearGradientStops( ctx
														, posX
														, posY
														, posX
														, posY - sizeY
														, this->cmap);
		}
		else
		{
			paintTuple = NanoguiUtils::nvgLinearGradientStops( ctx
														, posX
														, posY
														, posX + sizeX
														, posY
														, this->cmap);
		}
		std::tie(this->textureId, paint) = paintTuple;
		this->textureNeeded = false;
	}
	else if
		(nanogui::Orientation::Vertical == this->orientation)
	{
		paint = nvgImagePattern( ctx
							   , posX
							   , posY
							   , sizeY
							   , sizeY
							   , static_cast<float>(-0.5 * M_PI)
							   , this->textureId
							   , 1.f);
	}
	else
	{
		paint = nvgImagePattern( ctx
							   , posX
							   , posY
							   , sizeX
							   , sizeX
							   , 0.f
							   , this->textureId
							   , 1.f);

	}
	nvgBeginPath(ctx);
	nvgRoundedRect( ctx, posX, posY, sizeX, sizeY, 0);
	nvgFillPaint(ctx, paint);
	nvgFill(ctx);
};

/**
 *
 */
void NanoguiHColorbar::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //

	nanogui::Widget::draw(ctx);

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

	if
		(0 >= this->cmap.size())
	{
		nvgBeginPath(ctx);
		nvgRoundedRect( ctx, posX, posY, sizeX, sizeY, 0);
		nvgFillColor(ctx, nanogui::Color(this->color[0], this->color[1], this->color[2], 255));
		nvgFill(ctx);
	}
	else
	{
		this->drawColormap(ctx);
	}
};

/**
 *
 */
bool NanoguiHColorbar::mouseButtonEvent(const Eigen::Vector2i& pos, int buttons, bool pressed, int mods)
{
	float ratio;	//
    float heightTmp;    //
    float posY; //
    float mousePosY;    //

    posY = static_cast<float>(this->mPos.y());
    heightTmp = static_cast<float>(this->height());
    mousePosY = static_cast<float>(pos.y());

	nanogui::Widget::mouseButtonEvent(pos, buttons, pressed, mods);
	if
		(false == pressed)
	{
		ratio = posY + heightTmp - mousePosY;
		ratio /= heightTmp;

		if
			(false == this->dragging)
		{
			if
				(nullptr != this->clickCallback)
			{
				this->clickCallback( this->label, this->cmap, buttons, ratio
								   , this->dragTargetIsMax);
			}
		}
		else if
			(  nullptr != this->dragEndCallback
			&& 0 < this->height())
		{
			/* drag end */
			this->dragEndCallback(this->dragTargetIsMax, ratio);
		}
		this->dragging = false;
	}
	else
	{
		this->dragTargetIsMax = (0.5f * heightTmp > mousePosY);
		LOG(LogLevel::Debug, "NanoguiColorbar", "mouse down at {} / {}", pos.y(), this->mSize.y());
	}
	return true;
};

/**
 *
 */
bool NanoguiHColorbar::isMouseOver() const
{
	return this->mMouseFocus;
}

/**
 *
 */
bool NanoguiHColorbar::mouseMotionEvent(const Eigen::Vector2i& pos, const Eigen::Vector2i& /*rel*/, int /*buttons*/, int /*modifiers*/)
{
	float ratio;	//
    float heightTmp;    //
    float posY; //
    float mousePosY;    //

	if
		(nullptr != this->mouseMoveCallback
		&& 0 < this->height())
	{
        posY = static_cast<float>(this->mPos.y());
        heightTmp = static_cast<float>(this->height());
        mousePosY = static_cast<float>(pos.y());

		ratio = posY + heightTmp - mousePosY;
		ratio /= heightTmp;
		this->mouseMoveCallback(this->dragTargetIsMax, ratio);
	}
	return true;
};

/**
 *
 */
bool NanoguiHColorbar::mouseDragEvent(const Eigen::Vector2i& pos, const Eigen::Vector2i& /*rel*/, int /*buttons*/, int /*modifiers*/)
{
	float ratio;	//
    float heightTmp;    //
    float posY; //
    float mousePosY;    //

	this->dragging = true;

	if
		(nullptr != this->mouseMoveCallback
		&& 0 < this->height())
	{
        posY = static_cast<float>(this->mPos.y());
        heightTmp = static_cast<float>(this->height());
        mousePosY = static_cast<float>(pos.y());

		ratio = posY + heightTmp - mousePosY;
		ratio /= heightTmp;
		this->mouseMoveCallback(this->dragTargetIsMax, ratio);
	}
	return true;
};

/**
 *
 */
bool NanoguiHColorbar::setClickCallback(const std::function<void(const std::string&, const Colormap&, int, float, bool)>& callback)
{
	this->clickCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiHColorbar::setMouseMoveCallback(const std::function<void(bool, float)>& callback)
{
	this->mouseMoveCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiHColorbar::setDragEndCallback(const std::function<void(bool, float)>& callback)
{
	this->dragEndCallback = callback;
	return true;
};
