// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_image_field_editor.h"

#include "core/base64.h"

#include <fstream>
#include <iostream>

using namespace SEAScope;

/**
 *
 */
NanoguiImageFieldEditor::NanoguiImageFieldEditor(nanogui::Widget* parent)
: nanogui::Widget(parent),
  _backgroundColor(0, 0, 0, 0),
  _nameLabel(nullptr),
  _resetButton(nullptr),
  _deleteButton(nullptr),
  _viewImageCallback(nullptr),
  _loadImageCallback(nullptr),
  _deleteCallback(nullptr),
  _resetCallback(nullptr),
  _value(""),
  _originalValue(""),
  _editing(false),
  _reflowNeeded(false)
{
    // Assuming 360px width
    this->setFixedWidth(340);
    this->setWidth(340);

    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                           nanogui::Alignment::Middle,
                                           1, 3));

    nanogui::Widget* nameBlock = new nanogui::Widget(this);
    nameBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                nanogui::Alignment::Maximum,
                                                0, 0));
    nameBlock->setFixedWidth(146);
    this->_nameLabel = new nanogui::Label(nameBlock, "");
    this->_nameLabel->setFixedWidth(146);
    this->_nameLabel->setWidth(146);

    nanogui::Widget* resetBlock = new nanogui::Widget(this);
    resetBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                 nanogui::Alignment::Middle,
                                                 0, 0));
    resetBlock->setFixedSize(Eigen::Vector2i(18, 18));

    this->_resetButton = new nanogui::Button(resetBlock, "", ENTYPO_ICON_BACK_IN_TIME);
    this->_resetButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_resetButton->setFontSize(16);
    this->_resetButton->setCallback([this] () {
        this->setValue(this->_originalValue);
        if
            (nullptr != this->_resetCallback)
        {
            this->_resetCallback();
        }
    });

    nanogui::Widget* valueBlock = new nanogui::Widget(this);
    valueBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                 nanogui::Alignment::Middle,
                                                 0, 0));
    valueBlock->setFixedWidth(146);
    nanogui::Button* viewButton = new nanogui::Button(valueBlock, "View image");
    viewButton->setFixedSize(Eigen::Vector2i(73, 18));
    viewButton->setFontSize(16);
    viewButton->setCallback([this] () {
        if
            (nullptr != this->_viewImageCallback)
        {
            this->_viewImageCallback(this->_base64Image.getConstImage());
        }
    });

    this->_loadImageButton = new nanogui::Button(valueBlock, "Load image");
    this->_loadImageButton->setFixedSize(Eigen::Vector2i(73, 18));
    this->_loadImageButton->setFontSize(16);
    this->_loadImageButton->setCallback([this] () {
        if
            (nullptr != this->_loadImageCallback)
        {
            this->_loadImageCallback();
        }
    });
    this->_loadImageButton->setVisible(false);

    this->_deleteButton = new nanogui::Button(this, "", ENTYPO_ICON_TRASH);
    this->_deleteButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_deleteButton->setFontSize(16);
    this->_deleteButton->setCallback([this] () {
        if
            (nullptr != this->_deleteCallback)
        {
            this->_deleteCallback("", this->getName());
        }
    });
    this->_deleteButton->setVisible(false);

    /*
    this->_imageView->setGridThreshold(20);
    this->_imageView->setPixelInfoThreshold(20);
    this->_imageView->setScale(1.f);
    this->_imageView->setPixelInfoCallback(
        [this] (const Eigen::Vector2i& index) -> std::pair<std::string, nanogui::Color> {
        auto& textureSize = this->_imageView->imageSize();
        std::string stringData;
        std::uint16_t channelSum = 0;
        for (int i = 0; i != 4; ++i) {
            auto& channelData = static_cast<unsigned char*>(this->_base64PNG.getImage().data())[4*index.y()*textureSize.x() + 4*index.x() + i];
            channelSum += channelData;
            stringData += (std::to_string(static_cast<int>(channelData)) + "\n");
        }
        float intensity = static_cast<float>(255 - (channelSum / 4)) / 255.0f;
        float colorScale = intensity > 0.5f ? (intensity + 1) / 2 : intensity / 2;
        nanogui::Color textColor = nanogui::Color(colorScale, 1.0f);
        return { stringData, textColor };
    });
    */

    this->_reflowNeeded = true;
}

/**
 *
 */
NanoguiImageFieldEditor::~NanoguiImageFieldEditor()
{

}

/**
 *
 */
bool NanoguiImageFieldEditor::activateEdition(bool activate)
{
    if
        (activate == this->_editing)
    {
        return false;
    }

    this->_editing = activate;

    this->_loadImageButton->setVisible(this->_editing);
    this->_deleteButton->setVisible(this->_editing);
    this->_reflowNeeded = true;

    return true;
}

/**
 *
 */
void NanoguiImageFieldEditor::draw(NVGcontext* ctx)
{
    float posX;
    float posY;
    float sizeX;
    float sizeY;

    if
        (true == this->_reflowNeeded)
    {
        this->performLayout(ctx);
        this->_reflowNeeded = false;
    }

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

    nvgBeginPath(ctx);
    nvgRect(ctx, posX, posY, sizeX, sizeY);
    nvgFillColor(ctx, this->_backgroundColor);
    nvgFill(ctx);

    nanogui::Widget::draw(ctx);
}

/**
 *
 */
void NanoguiImageFieldEditor::setName(const std::string& name)
{
    this->_nameLabel->setCaption(name);
}

/**
 *
 */
const std::string& NanoguiImageFieldEditor::getName() const
{
    return this->_nameLabel->caption();
}

/**
 *
 */
void NanoguiImageFieldEditor::setCategory(const std::string& category)
{
    this->_category = category;
}

/**
 *
 */
const std::string& NanoguiImageFieldEditor::getCategory() const
{
    return this->_category;
}

/**
 *
 */
void NanoguiImageFieldEditor::setValue(const std::string& value)
{
    if
        (value != this->_value)
    {
        this->_base64Image.load(value);
    }
    this->_value = value;
    this->_resetButton->setVisible(value != this->_originalValue);
    this->_reflowNeeded = true;
}

/**
 *
 */
void NanoguiImageFieldEditor::setOriginalValue(const std::string& value)
{
    this->_originalValue = value;
    this->_resetButton->setTooltip("Restore initial value");
    this->_resetButton->setVisible(value == this->_value);
    this->_reflowNeeded = true;
}

/**
 *
 */
const std::string& NanoguiImageFieldEditor::getValue() const
{
    return this->_value;
}

/**
 *
 */
const std::string& NanoguiImageFieldEditor::getOriginalValue() const
{
    return this->_originalValue;
}

/**
 *
 */
void NanoguiImageFieldEditor::setBackgroundColor(const nanogui::Color& color)
{
    this->_backgroundColor = color;
}

/**
 *
 */
const nanogui::Color& NanoguiImageFieldEditor::getBackgroundColor() const
{
    return this->_backgroundColor;
}

/**
 *
 */
bool NanoguiImageFieldEditor::setViewImageCallback(const std::function<void(const Image&)>& callback)
{
    if
        (nullptr != this->_viewImageCallback)
    {
        return false;
    }
    this->_viewImageCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiImageFieldEditor::setLoadImageCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_loadImageCallback)
    {
        return false;
    }
    this->_loadImageCallback = callback;
    return true;
}

/**
 *
 */
const Image& NanoguiImageFieldEditor::loadFromFile(const std::string& path)
{
    std::string tmpValue;
    std::ifstream file(path, std::fstream::binary);
    file.seekg(0, std::ios::end);
    std::streampos length = file.tellg();
    file.seekg(0, std::ios::beg);
    if
        (0 < length)
    {
        std::vector<std::uint8_t> buffer(static_cast<std::size_t>(length));
        file.read(reinterpret_cast<char*>(&buffer[0]), length);
        base64_encode(tmpValue, buffer.data(), static_cast<std::size_t>(length));
        this->setValue(tmpValue);
    }

    return this->_base64Image.getConstImage();
}

/**
 *
 */
bool NanoguiImageFieldEditor::setDeleteCallback(const std::function<void(const std::string&, const std::string&)>& callback)
{
    if
        (nullptr != this->_deleteCallback)
    {
        return false;
    }
    this->_deleteCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiImageFieldEditor::setResetCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_resetCallback)
    {
        return false;
    }

    this->_resetCallback = callback;
    return true;
}

/**
 *
 */
const Image& NanoguiImageFieldEditor::getConstImage() const
{
    return this->_base64Image.getConstImage();
}
