// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_image_view.h"

#include "data/image.h"

using namespace SEAScope;

/**
 *
 */
NanoguiImageView::NanoguiImageView(nanogui::Widget* parent, GLuint textureId)
: nanogui::ImageView(parent, textureId),
  _textureId(textureId)
{
}

/**
 *
 */
NanoguiImageView::~NanoguiImageView()
{
    if
        (0 < this->_textureId)
    {
        glDeleteTextures(1, &(this->_textureId));
        this->_textureId = 0;
    }
}

/**
 *
 */
void NanoguiImageView::reset()
{
    if
        (0 < this->_textureId)
    {
        glDeleteTextures(1, &(this->_textureId));
        this->_textureId = 0;
    }

    this->bindImage(this->_textureId);
}

/**
 *
 */
void NanoguiImageView::setImage(const Image& image)
{
    if
        (0 < this->_textureId)
    {
        glDeleteTextures(1, &(this->_textureId));
        this->_textureId = 0;
    }

    glGenTextures(1, &(this->_textureId));
    glBindTexture(GL_TEXTURE_2D, this->_textureId);
    GLint internalFormat;
    GLuint format;
    switch (image.channels()) {
        case 1: internalFormat = GL_R8; format = GL_RED; break;
        case 2: internalFormat = GL_RG8; format = GL_RG; break;
        case 3: internalFormat = GL_RGB8; format = GL_RGB; break;
        case 4: internalFormat = GL_RGBA8; format = GL_RGBA; break;
        default: internalFormat = 0; format = 0; break;
    }
    glPixelStorei(GL_UNPACK_ALIGNMENT, (3 == image.channels() ? 1 : static_cast<GLint>(image.channels())));
    glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, static_cast<GLsizei>(image.width()), static_cast<GLsizei>(image.height()), 0, format, GL_UNSIGNED_BYTE, image.data());
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    this->bindImage(this->_textureId);

    if
        (image.width() <= static_cast<unsigned int>(this->mFixedSize.x()))
    {
        this->setScaleCentered(1.f);
    }
    else
    {
        this->setScaleCentered(static_cast<float>(this->mFixedSize.x()) / static_cast<float>(image.width()));
    }
}

/**
 *
 */
void NanoguiImageView::draw(NVGcontext* ctx) {
    Widget::draw(ctx);
    nvgEndFrame(ctx); // Flush the NanoVG draw stack, not necessary to call nvgBeginFrame afterwards.

    // Calculate several variables that need to be send to OpenGL in order for the image to be
    // properly displayed inside the widget.
    Eigen::Vector2f screenSize = this->screen()->size().cast<float>();
    Eigen::Vector2f scaleFactor = this->scale() * imageSizeF().cwiseQuotient(screenSize);
    Eigen::Vector2f positionInScreen = absolutePosition().cast<float>();
    Eigen::Vector2f positionAfterOffset = positionInScreen + this->offset();
    Eigen::Vector2f imagePosition = positionAfterOffset.cwiseQuotient(screenSize);
    glEnable(GL_SCISSOR_TEST);
    float r = this->screen()->pixelRatio();
    glScissor(static_cast<GLint>(positionInScreen.x() * r),
              static_cast<GLint>((screenSize.y() - positionInScreen.y() - static_cast<float>(this->size().y())) * r),
              static_cast<GLsizei>(static_cast<float>(this->size().x()) * r),
              static_cast<GLsizei>(static_cast<float>(this->size().y()) * r));
    this->imageShader().bind();
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, this->_textureId);
    this->imageShader().setUniform("image", 0);
    this->imageShader().setUniform("scaleFactor", scaleFactor);
    this->imageShader().setUniform("position", imagePosition);
    this->imageShader().drawIndexed(GL_TRIANGLES, 0, 2);
    glDisable(GL_SCISSOR_TEST);
}
