// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_main_menu.h"
#include "ui/nanogui/nanogui_about_pane.h"
#include "ui/nanogui/nanogui_granule_showcase.h"
#include "ui/nanogui/nanogui_extraction_dialog.h"
#include "ui/nanogui/nanogui_collections_control.h"
#include "ui/nanogui/nanogui_sources_control.h"
#include "ui/nanogui/nanogui_app_settings_editor.h"

#include "core/log.h"

using namespace SEAScope;

/**
 * 
 */
NanoguiMainMenu::NanoguiMainMenu(nanogui::Widget* parent)
: nanogui::Widget(parent), _aboutPane(nullptr), _settingsPane(nullptr),
  _granuleInfo(nullptr), _extractionDialog(nullptr), _sourcesControl(nullptr),
  _collectionsControl(nullptr), _reflowNeeded(true)
{
	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
										   , nanogui::Alignment::Middle
										   ,  0, 0));

    this->_aboutPane = new NanoguiAboutPane(this);
    this->_aboutPane->setTitle("About SEAScope");
    this->_aboutPane->setVisible(false);

    this->_settingsPane = new NanoguiAppSettingsEditor(this);
    this->_settingsPane->setTitle("SEAScope settings");
    this->_settingsPane->setVisible(false);

    this->_granuleInfo = new NanoguiGranuleShowcase(this);
    this->_granuleInfo->setTitle("Granule info");
    this->_granuleInfo->setVisible(false);

    this->_extractionDialog = new NanoguiExtractionDialog(this);
    this->_extractionDialog->setTitle("Extracted data");
    this->_extractionDialog->setVisible(false);

    this->_sourcesControl = new NanoguiSourcesControl(this);
    this->_sourcesControl->setVisible(false);

    this->_collectionsControl = new NanoguiCollectionsControl(this);
    this->_collectionsControl->setTitle("Catalogue");
    this->_collectionsControl->setVisible(false);
};

/**
 *
 */
NanoguiMainMenu::~NanoguiMainMenu()
{
    // Freeing memory is handled by nanogui
    this->_aboutPane = nullptr;
    this->_settingsPane = nullptr;
	this->_granuleInfo = nullptr;
	this->_extractionDialog = nullptr;
	this->_sourcesControl = nullptr;
	this->_collectionsControl = nullptr;
}

/**
 *
 */
bool NanoguiMainMenu::scrollEvent(const Eigen::Vector2i& pos, const Eigen::Vector2f& delta)
{
	nanogui::Widget::scrollEvent(pos, delta);

	/* return true to prevent events propagation */
	return true;
}

/**
 *
 */
void NanoguiMainMenu::draw(NVGcontext* ctx)
{
	if
		(this->_reflowNeeded)
	{
		this->performLayout(ctx);
		this->_reflowNeeded = false;
	}
	nanogui::Widget::draw(ctx);
};

/**
 *
 */
void NanoguiMainMenu::update()
{
	NanoguiScrollableWithTitle* resizeable(nullptr);	//
	std::int32_t childHeight(0);	//

	for
		(auto& child : this->mChildren)
	{
		if
			(false == child->visible())
		{
			continue;
		}

		childHeight = child->preferredSize(this->screen()->nvgContext()).y();
		LOG(LogLevel::Debug, "Main menu", "Child preferred height: {}",
				 childHeight);

		resizeable = dynamic_cast<NanoguiScrollableWithTitle*>(child);
		if
			(nullptr != resizeable)
		{
			resizeable->resize(this->fixedWidth(),
                               std::min(this->fixedHeight(), childHeight));
			resizeable = nullptr;
		}
		else
		{
			child->setFixedHeight(childHeight);
		}
	}

    this->_reflowNeeded = true;
}

/**
 *
 */
NanoguiAboutPane* NanoguiMainMenu::getAboutPane() const
{
    return this->_aboutPane;
}

/**
 *
 */
NanoguiAppSettingsEditor* NanoguiMainMenu::getSettingsPane() const
{
    return this->_settingsPane;
}

/**
 *
 */
NanoguiGranuleShowcase* NanoguiMainMenu::getGranuleInfo() const
{
    return this->_granuleInfo;
}

/**
 *
 */
NanoguiExtractionDialog* NanoguiMainMenu::getExtractionDialog() const
{
    return this->_extractionDialog;
}

/**
 *
 */
NanoguiCollectionsControl* NanoguiMainMenu::getCollectionsControl() const
{
    return this->_collectionsControl;
}

/**
 *
 */
NanoguiSourcesControl* NanoguiMainMenu::getSourcesControl() const
{
    return this->_sourcesControl;
}
