// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>

#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_months_list.h"

using namespace SEAScope;

/**
 *
 */
NanoguiMonthsList::NanoguiMonthsList(nanogui::Widget* parent)
: nanogui::Widget(parent)
, year(1970), hoverEnabled(false)
, monthCallback(nullptr), scrollCallback(nullptr)
{
	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Horizontal
										   , nanogui::Alignment::Middle
										   , 2, 2));

	this->initMonth("January", 1);
	this->initMonth("February", 2);
	this->initMonth("March", 3);
	this->initMonth("April", 4);
	this->initMonth("May", 5);
	this->initMonth("June", 6);
	this->initMonth("July", 7);
	this->initMonth("August", 8);
	this->initMonth("September", 9);
	this->initMonth("October", 10);
	this->initMonth("November", 11);
	this->initMonth("December", 12);
};

/**
 *
 */
NanoguiMonthsList::~NanoguiMonthsList()
{
};

/**
 *
 */
void NanoguiMonthsList::setSize(const Eigen::Vector2i& size)
{
	nanogui::Widget::setSize(size);
	this->resize(size.x());
};

/**
 *
 */
void NanoguiMonthsList::setWidth(int width)
{
	nanogui::Widget::setWidth(width);
	this->resize(width);
};

/**
 *
 */
void NanoguiMonthsList::setFixedSize(const Eigen::Vector2i& size)
{
	nanogui::Widget::setFixedSize(size);
	this->resize(size.x());
};

/**
 *
 */
void NanoguiMonthsList::setFixedWidth(int width)
{
	nanogui::Widget::setFixedWidth(width);
	this->resize(width);
};

/**
 *
 */
bool NanoguiMonthsList::initMonth(const std::string& label, std::uint8_t number)
{
	std::size_t i;	//

	i = number - 1u;
	this->buttons[i] = new NanoguiFlatButton(this);
	this->buttons[i]->setCaption(label);
	this->buttons[i]->setFixedSize(Eigen::Vector2i(20, 20));
	this->buttons[i]->setFontSize(18);
	this->buttons[i]->setBackgroundColor(nanogui::Color(0, 0, 0, 0));
	this->buttons[i]->mFontFace = "sans-bold";
	this->buttons[i]->setTextColor(nanogui::Color(90,90,90,255));
	this->buttons[i]->setCallback([this, number] () {
		if
			(nullptr != this->monthCallback)
		{
			this->monthCallback(number);
		}
	});
	return true;
};

/**
 *
 */
bool NanoguiMonthsList::setYear(std::uint16_t yearIn)
{
	if
		(yearIn == this->year)
	{
		return false;
	}
	this->year = yearIn;
	this->update();
	return true;
};

/**
 *
 */
bool NanoguiMonthsList::select(std::uint8_t month)
{
	size_t i;	//

	i = 0;
	for
		(;i<12; ++i)
	{
		if
			(i+1 == month)
		{
			this->buttons[i]->setBorderColor(nanogui::Color(255, 255, 255, 255));
		}
		else
		{
			this->buttons[i]->setBorderColor(nanogui::Color(0,0,0,0));
		}
	}
	return false;
};

/**
 *
 */
bool NanoguiMonthsList::setMainCoverage(const EventsCoverage& coverage)
{
	this->mainCoverage = coverage;
	this->update();
	return true;
};

/**
 *
 */
bool NanoguiMonthsList::setExtraCoverage(const EventsCoverage& coverage)
{
	this->extraCoverage = coverage;
	this->update();
	return true;
};

/**
 *
 */
bool NanoguiMonthsList::setVisibilityBounds(timestamp_t /*past*/, timestamp_t /*future*/)
{
	return false;
};

/**
 *
 */
bool NanoguiMonthsList::update()
{
	uint8_t i;	//
	std::vector<std::uint8_t> mainMonths;	//
	std::vector<std::uint8_t> extraMonths;	//
	NanoguiFlatButton* btn;	//

	this->mainCoverage.getMonths(this->year, mainMonths);
	this->extraCoverage.getMonths(this->year, extraMonths);

	i = 0;
	for
		(;i<12; ++i)
	{
		btn = this->buttons[i];

		auto found = std::find( mainMonths.begin()
							  , mainMonths.end()
							  , i+1);
		if
			(mainMonths.end() != found)
		{
			this->textColors[i] = nanogui::Color(250,250,250,255);
		}
		else
		{
			this->textColors[i] = nanogui::Color(90,90,90,255);
		}

		found = std::find( extraMonths.begin()
						 , extraMonths.end()
						 , i+1);
		if
			(extraMonths.end() != found)
		{
			this->textColors[i] = nanogui::Color(30,210,50,255);
		}

		btn->setBackgroundColor(this->backgrounds[i]);
		btn->setTextColor(this->textColors[i]);
	}

	return true;
};

/**
 *
 */
bool NanoguiMonthsList::resize(int width)
{
	std::size_t i;	//
    float widthTmp; //
	float btnWidthTmp;	//
	float spacingTmp;	//
    int btnWidth;   //
    int spacing;    //
	nanogui::BoxLayout* layout;	//

	if
		(0 == width)
	{
		return false;
	}

    widthTmp = static_cast<float>(width);

	btnWidthTmp = std::ceil((widthTmp - 22.f) / 12.f);
	spacingTmp = std::floor((widthTmp - 12.f * btnWidthTmp) / 11.f);

    btnWidth = static_cast<int>(btnWidthTmp);
    spacing = static_cast<int>(spacingTmp);

	layout = static_cast<nanogui::BoxLayout*>(this->layout());
	layout->setSpacing(spacing);

	i = 0;
	for
		(; i<12; ++i)
	{
		this->buttons[i]->setFixedWidth(btnWidth);
	}
	return true;
};

/**
 *
 */
bool NanoguiMonthsList::scrollEvent(const Eigen::Vector2i& /*p*/,
                                    const Eigen::Vector2f& rel)
{
    float scrollDelta;   //

    // Give priority to the horizontal axis because it is the natural direction
    // for the timeline. Typical mouses will only provide the vertical
    // component but touch pads and other specific devices allow horizontal
    // scrolling.
    scrollDelta = rel.x();
    if
        (std::fabs(scrollDelta) < 0.01f)
    {
        scrollDelta = rel.y();
    }

    if
        (nullptr != this->scrollCallback)
    {
        this->scrollCallback(scrollDelta);
    }
    return true; // prevent event propagation
};

/**
 *
 */
void NanoguiMonthsList::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

    if
        (this->hoverEnabled && this->mMouseFocus)
    {
        nvgBeginPath(ctx);
        nvgRect(ctx,
                posX + 1.0f, posY + 1.0f,
                sizeX - 2.0f, sizeY - 2.0f);
        nvgFillColor(ctx, nanogui::Color(20, 20, 20, 100));
        nvgFill(ctx);
    }

    nanogui::Widget::draw(ctx);
};
