// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_painted_label.h"

using namespace SEAScope;

/**
 *
 */
NanoguiPaintedLabel::NanoguiPaintedLabel(nanogui::Widget* parent, const std::string& caption)
: nanogui::Widget(parent),
  _caption(caption),
  _bgColor(0, 0, 0, 0),
  _textColor(255, 255, 255, 255),
  _iconColor(255, 255, 255, 255),
  _icon(0)
{
    this->setFontSize(14);
}

/**
 *
 */
NanoguiPaintedLabel::~NanoguiPaintedLabel()
{
}

/**
 *
 */
void NanoguiPaintedLabel::draw(NVGcontext* ctx)
{
    float posX(0.f); //
    float posY(0.f); //
    float sizeX(0.f);    //
    float sizeY(0.f);    //


    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

	nanogui::Widget::draw(ctx);

    // Draw background
    nvgBeginPath(ctx);
    nvgRect(ctx, posX, posY, sizeX, sizeY);
    nvgFillColor(ctx, this->_bgColor);
    nvgFill(ctx);
    nvgClosePath(ctx);

    float iw(0.f);	//

    // Draw icon
    if
        (0 != this->_icon)
    {
        float ih(0.f);	//
        int w(0);	//
        int h(0);	//
        float wTmp(0.f); //
        float hTmp(0.f); //
        Eigen::Vector2f iconPos;	//
        auto iconTmp = nanogui::utf8(this->_icon);

        ih = static_cast<float>(this->mFontSize);
        if
            (nanogui::nvgIsFontIcon(this->_icon))
        {
            ih *= this->icon_scale();
            nvgFontSize(ctx, ih);
            nvgFontFace(ctx, "icons");
            iw = nvgTextBounds(ctx, 0, 0, iconTmp.data(), nullptr, nullptr);
        }
        else
        {
            ih *= 0.9f;
            nvgImageSize(ctx, this->_icon, &w, &h);
            wTmp = static_cast<float>(w);
            hTmp = static_cast<float>(h);
            iw = wTmp * ih / hTmp;
        }

        nvgFillColor(ctx, this->_iconColor);
        nvgTextAlign(ctx, NVG_ALIGN_LEFT | NVG_ALIGN_MIDDLE);
        iconPos = this->mPos.cast<float>() + this->mSize.cast<float>() * 0.5f;
        iconPos.y() -= 1.f;
        //iconPos.x() -= iw * 0.5f;
        iconPos.x() = posX + 1.f;

        if
            (nanogui::nvgIsFontIcon(this->_icon))
        {
            nvgText(ctx, iconPos.x(), iconPos.y()+1, iconTmp.data(), nullptr);
        }
        else
        {
            NVGpaint imgPaint = nvgImagePattern(ctx,
                    iconPos.x(), iconPos.y() - ih/2, iw, ih, 0, this->_icon, this->mEnabled ? 0.5f : 0.25f);

            nvgFillPaint(ctx, imgPaint);
            nvgFill(ctx);
        }
    }

    // Draw text
    if
        (0 < this->_caption.size())
    {
        nvgFontSize(ctx, static_cast<float>(this->mFontSize));
        nvgFontFace(ctx, "sans");
        Eigen::Vector2f center = mPos.cast<float>() + mSize.cast<float>() * 0.5f;
        Eigen::Vector2f textPos(posX + (0.f < iw ? iw + 3.f : 1.f), center.y() - 1);

        nvgTextAlign(ctx, NVG_ALIGN_LEFT | NVG_ALIGN_MIDDLE);

        nvgFillColor(ctx, this->mTheme->mTextColorShadow);
        nvgText(ctx, textPos.x(), textPos.y(), this->_caption.c_str(), nullptr);

        nvgFillColor(ctx, this->_textColor);
        nvgText(ctx, textPos.x(), textPos.y() + 1, this->_caption.c_str(), nullptr);
    }
}

/**
 *
 */
void NanoguiPaintedLabel::setBackgroundColor(const nanogui::Color& color)
{
    this->_bgColor = color;
}

/**
 *
 */
void NanoguiPaintedLabel::setTextColor(const nanogui::Color& color)
{
    this->_textColor = color;
}

/**
 *
 */
void NanoguiPaintedLabel::setIconColor(const nanogui::Color& color)
{
    this->_iconColor = color;
}

/**
 *
 */
void NanoguiPaintedLabel::setIcon(int icon)
{
    this->_icon = icon;
}

/**
 *
 */
void NanoguiPaintedLabel::setCaption(const std::string& caption)
{
    this->_caption = caption;
}

/**
 *
 */
const std::string& NanoguiPaintedLabel::caption() const
{
    return this->_caption;
}

/**
 *
 */
Eigen::Vector2i NanoguiPaintedLabel::preferredSize(NVGcontext* ctx) const
{
	float iw(0.f);	//
	float ih(0.f);	//
	int w(0);	//
	int h(0);	//
    float sizeY(0.f);    //
    float wTmp(0.f); //
    float hTmp(0.f); //
    float tw(0.f);   //

    if
        (0 < this->_caption.size())
    {
        nvgFontSize(ctx, static_cast<float>(this->mFontSize));
        nvgFontFace(ctx, "sans-bold");
        tw = nvgTextBounds(ctx, 0,0, this->_caption.c_str(), nullptr, nullptr);
    }

	iw = 0.0f;
	ih = static_cast<float>(this->mFontSize);
    sizeY = static_cast<float>(this->mSize.y());

    if
        (0 != this->_icon)
    {
        if
            (nanogui::nvgIsFontIcon(this->_icon))
        {
            ih *= this->icon_scale();
            nvgFontFace(ctx, "icons");
            nvgFontSize(ctx, ih);
            auto str = nanogui::utf8(this->_icon);
            iw = nvgTextBounds(ctx, 0, 0, str.data(), nullptr, nullptr)
                + sizeY * 0.15f;
        }
        else
        {
            ih *= 0.9f;
            nvgImageSize(ctx, this->_icon, &w, &h);
            wTmp = static_cast<float>(w);
            hTmp = static_cast<float>(h);
            iw = wTmp * ih / hTmp;
        }
    }

    float totalWidth(0.f);
    if
        (0.f < iw * tw)
    {
        totalWidth = 1.f + iw + 2.f + tw + 1.f;
    }
    else
    {
        totalWidth = 1.f + (iw + tw) + 1.f;
    }
    return Eigen::Vector2i(static_cast<int>(totalWidth), this->mFontSize + 2);
}
