// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_path_selector.h"

#include "ui/nanogui/nanogui_list.h"
#include "ui/nanogui/nanogui_clickable_label.h"
#include "ui/nanogui/nanogui_anchored.h"
#include "ui/nanogui/nanogui_textfield.h"

#include "core/path.h"
#include "core/dialog.h"

#include <string>
#include <iostream>
#include <sstream>

using namespace SEAScope;

/**
 *
 */
NanoguiPathSelector::NanoguiPathSelector(nanogui::Widget* parent, PathSelectorType type)
    : nanogui::Widget(parent),
    _inputText(nullptr),
    _dropdownContainer(nullptr),
    _itemsList(nullptr),
    _browseButton(nullptr),
    _selectButton(nullptr),
    _parentDirButton(nullptr),
    _changeCallback(nullptr),
    _dropdownCallback(nullptr),
    _updateRequested(false),
    _selected(""),
    _type(type)
{
    Path::absolutePath("", this->_selected);

    Path::getParentDir(this->_selected, this->_parentPath);

    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
        nanogui::Alignment::Minimum,
        0, 0));

    nanogui::Widget* labelBlock(nullptr);
    labelBlock = new nanogui::Widget(this);
    labelBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
        nanogui::Alignment::Minimum,
        3, 2));
    nanogui::Widget* inputTextBlock(nullptr);
    inputTextBlock = new nanogui::Widget(labelBlock);
    inputTextBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
        nanogui::Alignment::Maximum,
        0, 0));
    this->_inputText = new NanoguiTextField(inputTextBlock);
    this->_inputText->setAlignment(nanogui::TextBox::Alignment::Right);
    this->_inputText->setFontSize(18);
    this->_inputText->setFixedSize(Eigen::Vector2i(100, 22));
    this->_inputText->setPlaceholder("Please provide a path");
    this->_inputText->setCallback([this](const std::string& path) -> bool {
        if
            (0 >= path.size())
        {
            return false;
        }

        this->updateExplorer(path);

        if
            (nullptr != this->_changeCallback)
        {
            this->_changeCallback(path);
        }
        return true;
        });
    this->_inputText->setFocusCallback([this](bool focused) -> bool {
        if
            (true == focused)
        {
            return false;
        }

        std::string absolutePath;
        Path::absolutePath(this->_inputText->value(), absolutePath);
        if
            (absolutePath != this->_inputText->value())
        {
            this->_inputText->setValue(absolutePath);
        }

        this->updateExplorer(absolutePath);

        return true;
        });
    this->_inputText->setEditionCompleteCallback([this]() {
        this->hideDropDown();
        });

    nanogui::Widget* blk(nullptr);
    blk = new nanogui::Widget(labelBlock);
    blk->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
        nanogui::Alignment::Maximum,
        1, 0));
    this->_browseButton = new nanogui::Button(blk, "", ENTYPO_ICON_FOLDER);
    this->_browseButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_browseButton->setFontSize(16);

    this->_dropdownContainer = NanoguiAnchored::create(this->_browseButton);
    this->_dropdownContainer->offset = 0u;
    this->_dropdownContainer->placement = static_cast<std::uint8_t>(NanoguiAnchored::ON_LEFT + NanoguiAnchored::ON_BOTTOM);
    this->_dropdownContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
        nanogui::Alignment::Minimum,
        0, 0));

    nanogui::Widget* explorerButtons = new nanogui::Widget(this->_dropdownContainer);
    explorerButtons->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
        nanogui::Alignment::Middle,
        10, 10));

    this->_selectButton = new nanogui::Button(explorerButtons, "Select");
    this->_selectButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_selectButton->setFontSize(16);
    this->_selectButton->setEnabled(false);
    this->_selectButton->setTextColor(nanogui::Color(255, 255, 255, 255));
    this->_selectButton->setBackgroundColor(nanogui::Color(48, 87, 139, 255));
    this->_selectButton->setCallback([this]() {
        this->setPath(this->_selected);
        this->hideDropDown();
        });

    this->_parentDirButton = new nanogui::Button(explorerButtons, "", ENTYPO_ICON_ARROW_UP);
    this->_parentDirButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_parentDirButton->setFontSize(16);
    this->_parentDirButton->setCallback([this]() {
        if
            (Path::separator() == this->_inputText->value().back())
        {
            // There is a trailing separator, so simply replace selection by the parent directory
            Path::getParentDir(this->_inputText->value(), this->_parentPath);
            this->_selected = this->_parentPath;
        }
        else
        {
            // The value stored in parent directory corresponds to the directory shown in the dropdown,
            // so the new value should be the parent's parent.
            Path::getParentDir(this->_inputText->value(), this->_parentPath);
            Path::getParentDir(this->_parentPath, this->_selected);
        }

        // Update parent directory
        Path::getParentDir(this->_selected, this->_parentPath);

        this->_inputText->setValue(this->_selected);
        this->_updateRequested = true;

        // Activate the select button only if the new value for _selected does
        // not end with a separator
        this->_selectButton->setEnabled(false);

        if
            (nullptr != this->_changeCallback)
        {
            this->_changeCallback(this->_selected);
        }
        });

    this->_itemsList = new NanoguiList(this->_dropdownContainer);
    this->_itemsList->content->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
        nanogui::Alignment::Fill,
        1, 1));
    this->_itemsList->content->setBackgroundFillColor(nanogui::Color(80, 80, 80, 255));
    this->_browseButton->setCallback([this]() {
            std::string absolutePath;
            Path::absolutePath(this->_inputText->value(), absolutePath);
            if
                (true == Dialog::available())
                //(false)
            {
                std::string str;

                if (   PathSelectorType::get_file == _type
                    || PathSelectorType::save_file == _type)
                {
                    str = Dialog::fileDialog(absolutePath, PathSelectorType::save_file == _type);
                }
                else
                {
                    str = Dialog::dirDialog(absolutePath);
                }

                if (false == str.empty())
                {
                    this->setPath(str);
                }
            }
            else
            {
                this->setPath(absolutePath, true);
                this->toggleDropDown();
            }
        });

    this->setPath(this->_selected);
    this->hideDropDown();
}

/**
 *
 */
NanoguiPathSelector::~NanoguiPathSelector()
{

}

/**
 *
 */
void NanoguiPathSelector::enable()
{
    this->setEnabled(true);
    this->_browseButton->setEnabled(true);
    this->_inputText->setEnabled(true);
}

/**
 *
 */
void NanoguiPathSelector::disable()
{
    this->setEnabled(false);
    this->_browseButton->setEnabled(false);
    this->_inputText->setEnabled(false);
}

/**
 *
 */
void NanoguiPathSelector::show()
{
    this->setVisible(true);
}

/**
 *
 */
void NanoguiPathSelector::hide()
{
    this->hideDropDown();
    this->setVisible(false);
}

/**
 *
 */
const std::string& NanoguiPathSelector::getPath() const
{
    return this->_inputText->value();
}


/**
 *
 */
bool NanoguiPathSelector::setPath(const std::string& path)
{
    return this->setPath(path, false);
}

/**
 *
 */
bool NanoguiPathSelector::setPath(const std::string& path, bool forceExplorerUpdate)
{
    bool valueChanged{ false };  //

    valueChanged = (path != this->_inputText->value());
    if
        (false == (forceExplorerUpdate || valueChanged))
    {
        // Value is the same and explorer update is not requested => do nothing
        return false;
    }

    if
        (true == valueChanged)
    {
        // Only update the text field if the value actually changed (i.e. we
        // did not reach this line in the code only due to forceExplorerUpdate=true
        this->_inputText->focusEvent(false);
        this->_inputText->setValue(path);
    }

    this->updateExplorer(path);

    if
        ((true == valueChanged) && (nullptr != this->_changeCallback))
    {
        this->_changeCallback(path);
    }
    return true;
}

/**
 *
 */
bool NanoguiPathSelector::setChangeCallback(const std::function<void(const std::string&)>& callback)
{
    if
        (nullptr != this->_changeCallback)
    {
        return false;
    }

    this->_changeCallback = callback;
    return true;
}

/**
 *
 */
void NanoguiPathSelector::resize(const Eigen::Vector2i& size)
{
    this->setFixedSize(Eigen::Vector2i(size.x(), 31));

    this->_inputText->setFixedWidth(size.x() - 30);
    this->_itemsList->setFixedSize(Eigen::Vector2i(size.x() - 26, 180));
}


/**
 *
 */
bool NanoguiPathSelector::setDropDownCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_dropdownCallback)
    {
        return false;
    }

    this->_dropdownCallback = callback;
    return true;
}

/**
 *
 */
void NanoguiPathSelector::toggleDropDown()
{
    if
        (this->_dropdownContainer->visible())
    {
        this->hideDropDown();
    }
    else
    {
        this->showDropDown();
    }
}

/**
 *
 */
void NanoguiPathSelector::showDropDown()
{
    this->_dropdownContainer->setVisible(true);
    if
        (nullptr != this->_dropdownCallback)
    {
        this->_dropdownCallback();
    }
}

/**
 *
 */
void NanoguiPathSelector::hideDropDown()
{
    this->_dropdownContainer->setVisible(false);
}

/**
 *
 */
void NanoguiPathSelector::updateExplorer(const std::string& path)
{
    std::forward_list<std::string> dirContent;
    std::string separator(1, Path::separator());
    std::size_t lastSeparatorPosition{ 0 };

    lastSeparatorPosition = path.rfind(separator, path.size() - separator.size() - 1);
    Path::getParentDir(path, this->_parentPath);

    NanoguiClickableLabel* label(nullptr);
    std::size_t availableWidgets = this->_widgets.size();
    this->_exploredPath = this->_parentPath;

    if
        (path.rfind(separator) != lastSeparatorPosition)
    {
        // Path has a trailing separator => it must be considered as a
        // directory to use as base path
        this->_exploredPath = path;
    }

    Path::listFiles(this->_exploredPath, dirContent);
    dirContent.sort();
    std::size_t count = 0;
    bool oneItemSelected{ false };
    for
        (const auto& item : dirContent)
    {
        if
            (count < availableWidgets)
        {
            label = this->_widgets.at(count);
        }
        else
        {
            label = new NanoguiClickableLabel(this->_itemsList->content);
            label->setChangeCallback([this](const std::string& value, bool selected, float duration) {
                if
                    (true == selected)
                {
                    // Find the widgets that corresponds to this value (if any)
                    for
                        (auto widget : this->_widgets)
                    {
                        widget->setSelected((widget->caption() == value), true /*silenced*/);
                    }

                    Path::canonicalPath(this->_exploredPath + value, this->_selected);
                }

                if
                    (500000000.f > duration)
                {
                    std::string canonicalPath;
                    Path::canonicalPath(this->_exploredPath + value, canonicalPath);

                    // Consider this is a double-click
                    if
                        (Path::dirExists(this->_exploredPath + value))
                    {
                        std::string delimiter(1, Path::separator());
                        this->_selected = canonicalPath + delimiter;
                    }
                    else
                    {
                        this->_selected = canonicalPath;
                        this->hideDropDown();
                    }

                    Path::getParentDir(this->_selected, this->_parentPath);

                    this->_inputText->setValue(this->_selected);
                    this->_updateRequested = true;
                    if
                        (nullptr != this->_changeCallback)
                    {
                        this->_changeCallback(this->_selected);
                    }
                }

                this->_selectButton->setEnabled(selected);
                });
            this->_widgets.push_back(label);
        }

        label->setCaption(item.substr(item.rfind(separator) + 1));
        label->setSelected((item == path), true);
        label->setVisible(true);
        label->setEnabled(true);

        oneItemSelected = oneItemSelected || (item == path);

        count = count + 1;

        if
            (Path::dirExists(item))
        {
            // Assign directory widget
            label->setIcon(ENTYPO_ICON_FOLDER);
        }
        else
        {
            // Assign file widget
            label->setIcon(ENTYPO_ICON_TEXT_DOCUMENT);
        }
    }

    this->_selectButton->setEnabled(oneItemSelected);

    if
        (0 < count)
    {
        int newHeight = static_cast<int>(count) * label->preferredSize(this->screen()->nvgContext()).y() + 2 /*margin*/ + (static_cast<int>(count) - 1) * 1 /*inter spacing*/;
        this->_itemsList->content->setFixedHeight(newHeight);
        this->_itemsList->content->setHeight(newHeight);
    }
    else
    {
        this->_itemsList->content->setFixedHeight(0);
    }

    // Hide all unused widgets
    while
        (count < this->_widgets.size())
    {
        label = this->_widgets.at(count);
        label->setCaption("");
        label->setSelected(false, true);    // unselect by default
        label->setVisible(false);
        label->setEnabled(false);
        count = count + 1;
    }

    this->_itemsList->update(true /*reset scrolling*/);
}

/**
 *
 */
void NanoguiPathSelector::draw(NVGcontext* ctx)
{
    if
        (true == this->_updateRequested)
    {
        this->updateExplorer(this->_inputText->value());
        this->_updateRequested = false;
    }

    if
        (true == this->mMouseFocus)
    {
        float posX(0.f);
        float posY(0.f);
        float sizeX(0.f);
        float sizeY(0.f);

        posX = static_cast<float>(this->mPos.x());
        posY = static_cast<float>(this->mPos.y());
        sizeX = static_cast<float>(this->mSize.x());
        sizeY = static_cast<float>(this->mSize.y());

        nvgBeginPath(ctx);
        for
            (float k = 0.f; k * 5.f + 3.f <= sizeX - 2.f; k = k + 1.F)
        {
            nvgMoveTo(ctx, posX + 1.f + 5.f * k, posY + 1.f);
            nvgLineTo(ctx, posX + 1.f + 5.f * k + 3.f, posY + 1.f);
        }
        for
            (float k = 0.f; k * 5.f + 3.f <= sizeY - 2.f; k = k + 1.F)
        {
            nvgMoveTo(ctx, posX + 1.f, posY + 1.f + 5.f * k);
            nvgLineTo(ctx, posX + 1.f, posY + 1.f + 5.f * k + 3.f);
        }
        for
            (float k = 0.f; k * 5.f + 3.f <= sizeX - 2.f; k = k + 1.F)
        {
            nvgMoveTo(ctx, posX + 1.f + 5.f * k, posY + sizeY - 2.f);
            nvgLineTo(ctx, posX + 1.f + 5.f * k + 3.f, posY + sizeY - 2.f);
        }
        for
            (float k = 0.f; k * 5.f + 3.f <= sizeY - 2.f; k = k + 1.F)
        {
            nvgMoveTo(ctx, posX + sizeX - 2.f, posY + 1.f + 5.f * k);
            nvgLineTo(ctx, posX + sizeX - 2.f, posY + 1.f + 5.f * k + 3.f);
        }
        nvgStrokeColor(ctx, nanogui::Color(30, 200, 230, 200));
        nvgStrokeWidth(ctx, 2);
        nvgStroke(ctx);
    }

    nanogui::Widget::draw(ctx);
}

/**
 *
 */
void NanoguiPathSelector::setPlaceholder(const std::string& msg)
{
    this->_inputText->setPlaceholder(msg);
}

/**
 *
 */
bool NanoguiPathSelector::dropEvent(const std::vector<std::string>& paths)
{
    if
        (0 >= paths.size())
    {
        return false;
    }

    this->setPath(paths[0]);
    this->hideDropDown();
    return true;
}

/**
 *
 */
void NanoguiPathSelector::hintValidPath()
{
    this->_inputText->hintValidValue();
}

/**
 *
 */
void NanoguiPathSelector::hintInvalidPath()
{
    this->_inputText->hintInvalidValue();
}

void NanoguiPathSelector::setType(PathSelectorType type)
{
    _type = type;
}
