// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_properties_editor.h"
#include "ui/nanogui/nanogui_bool_field_editor.h"
#include "ui/nanogui/nanogui_list.h"
#include "ui/nanogui/nanogui_add_property_form.h"
#include "ui/nanogui/nanogui_num_field_editor.h"
#include "ui/nanogui/nanogui_text_field_editor.h"
#include "ui/nanogui/nanogui_image_field_editor.h"
#include "ui/nanogui/nanogui_dict_field_editor.h"
#include "ui/nanogui/nanogui_path_selector.h"
#include "ui/nanogui/nanogui_image_view.h"
#include "ui/nanogui/nanogui_anchored.h"
#include "data/base64_image.h"
#include "core/path.h"
#include "core/log.h"
#include "core/utils.h"

using namespace SEAScope;

/**
 *
 */
NanoguiPropertiesEditor::NanoguiPropertiesEditor(nanogui::Widget* parent)
: nanogui::Widget(parent),
  _editButton(nullptr),
  _saveButton(nullptr),
  _cancelButton(nullptr),
  _pathContainer(nullptr),
  _pathSelector(nullptr),
  _pathOkButton(nullptr),
  _pathNokButton(nullptr),
  _pathSelectorTarget(nullptr),
  _imageContainer(nullptr),
  _imageView(nullptr),
  _reflowNeeded(false),
  _preventImageViewer(false),
  _saveCallback(nullptr)
{
    this->setFixedWidth(360);
    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                           nanogui::Alignment::Fill,
                                           0, 0));

    this->initImageViewer();

    nanogui::Widget* header = new nanogui::Widget(this);
    header->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                             nanogui::Alignment::Middle,
                                             2, 2));

	nanogui::Label* lbl = new nanogui::Label(header, "Properties", "sans-bold");
	lbl->setFixedWidth(70);
	lbl->setColor(nanogui::Color(180,180,180,150));

    this->_addPropertyForm = new NanoguiAddPropertyForm(this, {"bool" ,"num", "text", "img", "dict"});
    this->_addPropertyForm->setVisible(false);
    this->_addPropertyForm->setValidateCallback([this] (const std::string& nodeType, const std::string& label) -> bool {
        if
            ("" == nodeType)
        {
            return false;
        }

        if
            ("" == label)
        {
            // Empty label not allowed
            return false;
        }

        return this->checkPropertyUnicity("", label);
    });
    this->_addPropertyForm->setOkCallback([this] () {
        std::string nodeLabel;
        this->_addPropertyForm->getName(nodeLabel);

        std::string nodeType;
        this->_addPropertyForm->getType(nodeType);

        this->addProperty(nodeType, "", nodeLabel);

        this->_addPropertyForm->reset();
    });
    this->_addPropertyForm->setNokCallback([this] () {
        this->_addPropertyForm->reset();
    });

    this->_editButton = new nanogui::Button(header, "", ENTYPO_ICON_EDIT);
    this->_editButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_editButton->setFontSize(16);
    this->_editButton->setCallback([this] () {
        this->activateEdition(true);
    });

    nanogui::Widget* spacer = new nanogui::Widget(header);
    spacer->setFixedWidth(158);

    this->_saveButton = new nanogui::Button(header, "Save");
    this->_saveButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_saveButton->setFontSize(16);
    this->_saveButton->setCallback([this] () {
        for
            (auto it: this->_boolEditors)
        {
            it->setOriginalValue(it->getValue());
        }

        for
            (auto it: this->_numEditors)
        {
            it->setOriginalValue(it->getValue());
        }

        for
            (auto it: this->_textEditors)
        {
            it->setOriginalValue(it->getValue());
        }

        for
            (auto it: this->_imageEditors)
        {
            it->setOriginalValue(it->getValue());
        }

        DictNode properties = std::move(this->_properties);

        // Delete original properties
        auto found = properties.children.find("properties");
        if
            (properties.children.end() != found)
        {
            delete found->second;
            found->second = nullptr;
        }

        DictNode* newProperties = new DictNode();
        this->getProperties(*(newProperties));
        properties.children["properties"] = std::move(newProperties);

        this->setProperties(std::move(properties));

        this->activateEdition(false);

        if
            (nullptr != this->_saveCallback)
        {
            this->_saveCallback(this->_properties);
        }
    });

    this->_cancelButton = new nanogui::Button(header, "Cancel");
    this->_cancelButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_cancelButton->setFontSize(16);
    this->_cancelButton->setCallback([this] () {
        DictNode properties = std::move(this->_properties);
        this->setProperties(std::move(properties));
        this->activateEdition(false);
    });


    this->hideImageViewer();

    this->_itemsList = new NanoguiList(this);
    this->_itemsList->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                       nanogui::Alignment::Fill,
                                                       3, 0));
    this->_itemsList->content->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                                nanogui::Alignment::Minimum,
                                                                4, 1));
    this->_itemsList->content->setFixedWidth(342);
    this->_itemsList->content->setWidth(342);
    this->_itemsList->content->widgetsSorter = [this] (const Widget* a, const Widget* b) {
        return this->widgetsSorter(a, b);
    };

    this->_saveButton->setVisible(false);
    this->_cancelButton->setVisible(false);
    this->_reflowNeeded = true;
}

/**
 *
 */
NanoguiPropertiesEditor::~NanoguiPropertiesEditor()
{
}

/**
 *
 */
void NanoguiPropertiesEditor::setPreventImageViewer(bool prevent)
{
    this->_preventImageViewer = prevent;
    if
        (true == prevent)
    {
        this->hideImageViewer();
    }
}

/**
 *
 */
void NanoguiPropertiesEditor::activateEdition(bool activate)
{
    for
        (auto it: this->_boolEditors)
    {
        it->activateEdition(activate);
    }

    for
        (auto it: this->_numEditors)
    {
        it->activateEdition(activate);
    }

    for
        (auto it: this->_textEditors)
    {
        it->activateEdition(activate);
    }

    for
        (auto it: this->_imageEditors)
    {
        it->activateEdition(activate);
    }

    for
        (auto it: this->_dictEditors)
    {
        it->activateEdition(activate);
    }

    this->_addPropertyForm->setVisible(activate);
    this->_editButton->setVisible(!activate);
    this->_saveButton->setVisible(activate);
    this->_cancelButton->setVisible(activate);

    this->resize(this->mSize); // to update the height of the list after toggling the form for adding properties

    this->_reflowNeeded = true;
}

/**
 *
 */
void NanoguiPropertiesEditor::draw(NVGcontext* ctx)
{
	if
		(true == this->_reflowNeeded)
	{
		this->performLayout(ctx);
		this->_reflowNeeded = false;
	}

    bool drawingContainer(false);
    float containerStartY(0.f);
    float latestEndY(0.f);

    int topOffset = 20; // "Properties" label
    if
        (true == this->_addPropertyForm->visible())
    {
        topOffset = topOffset + 32;
    }

    for
        (const auto& child: this->_itemsList->content->children())
    {
        if
            (false == child->visible())
        {
            continue;
        }

        float y = static_cast<float>(this->mPos.y() + this->_itemsList->position().y() + this->_itemsList->content->position().y() + child->position().y());

        if
            (true == this->_widgetIsContainer.at(child))
        {
            if
                (true == drawingContainer)
            {
                // Draw previous container
                nvgBeginPath(ctx);

                float startY;
                if
                    (containerStartY > static_cast<float>(this->mPos.y() + topOffset))
                {
                    nvgMoveTo(ctx, 8.f, containerStartY);
                    nvgLineTo(ctx, 1.f, containerStartY);
                    startY = containerStartY;
                }
                else
                {
                    startY = static_cast<float>(this->mPos.y() + topOffset);
                    nvgMoveTo(ctx, 1.f, startY);
                }

                if
                    (latestEndY + 1.f > static_cast<float>(this->mPos.y() + topOffset))
                {
                    nvgLineTo(ctx, 1.f, latestEndY + 1.f);
                    nvgLineTo(ctx, static_cast<float>(this->mSize.x() - 13), latestEndY + 1.f);
                    nvgLineTo(ctx, static_cast<float>(this->mSize.x() - 13), startY);
                }

                if
                    (containerStartY > static_cast<float>(this->mPos.y() + topOffset))
                {
                    nvgLineTo(ctx, static_cast<float>(this->mSize.x() - 20), startY);
                }

                //nvgStrokeColor(ctx, nanogui::Color(220, 190, 30, 255));
                nvgStrokeColor(ctx, nanogui::Color(20, 100, 170, 180));
                nvgStroke(ctx);
            }

            drawingContainer = true;
            containerStartY = y + 3.f + 0.5f * static_cast<float>(child->size().y());
        }

        latestEndY = y + static_cast<float>(child->size().y());
    }

    if
        (true == drawingContainer)
    {
        // Draw previous container
        nvgBeginPath(ctx);

        float startY;
        if
            (containerStartY > static_cast<float>(this->mPos.y() + topOffset))
        {
            nvgMoveTo(ctx, 8.f, containerStartY);
            nvgLineTo(ctx, 1.f, containerStartY);
            startY = containerStartY;
        }
        else
        {
            startY = static_cast<float>(this->mPos.y() + topOffset);
            nvgMoveTo(ctx, 1.f, startY);
        }

        if
            (latestEndY + 1.f > static_cast<float>(this->mPos.y() + topOffset))
        {
            nvgLineTo(ctx, 1.f, latestEndY + 1.f);
            nvgLineTo(ctx, static_cast<float>(this->mSize.x() - 13), latestEndY + 1.f);
            nvgLineTo(ctx, static_cast<float>(this->mSize.x() - 13), startY);
        }

        if
            (containerStartY > static_cast<float>(this->mPos.y() + topOffset))
        {
            nvgLineTo(ctx, static_cast<float>(this->mSize.x() - 20), startY);
        }

        //nvgStrokeColor(ctx, nanogui::Color(220, 190, 30, 255));
        nvgStrokeColor(ctx, nanogui::Color(20, 100, 170, 180));
        nvgStroke(ctx);
    }

    nanogui::Widget::draw(ctx);
}

/**
 *
 */
void NanoguiPropertiesEditor::update()
{
	LOG(LogLevel::Debug, "NanoguiPropertiesEditor", "Number of children: {}", this->children().size());

    std::size_t itemsCount(0);

    this->_itemsList->update(false /*do not reset scrolling*/);

    // At this point the widgets are sorted

    nanogui::Color evenColor(100, 100, 100, 255);
    nanogui::Color oddColor(85, 85, 85, 255);
    for
        (auto& it: this->_boolEditors)
    {
        if
            (false == it->visible())
        {
            continue;
        }

        if
            (0 == this->_sortedWidgets[it] % 2)
        {
            it->setBackgroundColor(evenColor);
        }
        else
        {
            it->setBackgroundColor(oddColor);
        }

        itemsCount = itemsCount + 1;
    }

    for
        (auto& it: this->_numEditors)
    {
        if
            (false == it->visible())
        {
            continue;
        }

        if
            (0 == this->_sortedWidgets[it] % 2)
        {
            it->setBackgroundColor(evenColor);
        }
        else
        {
            it->setBackgroundColor(oddColor);
        }

        itemsCount = itemsCount + 1;
    }

    for
        (auto& it: this->_textEditors)
    {
        if
            (false == it->visible())
        {
            continue;
        }

        if
            (0 == this->_sortedWidgets[it] % 2)
        {
            it->setBackgroundColor(evenColor);
        }
        else
        {
            it->setBackgroundColor(oddColor);
        }

        itemsCount = itemsCount + 1;
    }

    for
        (auto& it: this->_imageEditors)
    {
        if
            (false == it->visible())
        {
            continue;
        }

        if
            (0 == this->_sortedWidgets[it] % 2)
        {
            it->setBackgroundColor(evenColor);
        }
        else
        {
            it->setBackgroundColor(oddColor);
        }

        itemsCount = itemsCount + 1;
    }

    for
        (auto& it: this->_dictEditors)
    {
        if
            (false == it->visible())
        {
            continue;
        }

        itemsCount = itemsCount + 1;
    }

    this->_itemsList->setVisible(0 < itemsCount);

    //std::cout << "Items count: " << itemsCount << std::endl;

	this->_reflowNeeded = true;
};

/**
 *
 */
void NanoguiPropertiesEditor::updateSubEditors(const DictNode* properties, std::size_t& boolEditorsCount, std::size_t& numEditorsCount, std::size_t& textEditorsCount, std::size_t& imageEditorsCount, std::size_t& dictEditorsCount, const std::string& /*category*/)
{
    std::vector<std::tuple<std::string, std::string, DataNode*> > nodes;
    for
        (const auto& child: properties->children)
    {
        if
            (NodeType::dict == child.second->getType())
        {
            nodes.push_back(std::make_tuple(child.first, "" /*so that the dict node comes before its children after sorting*/, child.second));

            for
                (const auto& grandchild: static_cast<DictNode*>(child.second)->children)
            {
                if
                    (NodeType::dict != grandchild.second->getType())
                {
                    // We only allow one level of nesting for dictionnaries
                    nodes.push_back(std::make_tuple(child.first, grandchild.first, grandchild.second));
                }
            }
        }
        else
        {
            nodes.push_back(std::make_tuple("", child.first, child.second));
        }
    }

    // Sort top-level properties by name
    std::sort(nodes.begin(), nodes.end(),
              [] (const std::tuple<std::string, std::string, DataNode*>& a, const std::tuple<std::string, std::string, DataNode*>& b) {
                    if
                        (std::get<0>(a) == std::get<0>(b))
                    {
                        return std::get<1>(a) < std::get<1>(b);
                    }
                   return std::get<0>(a) < std::get<0>(b);
              });

    std::uint32_t count(0);
    for
        (const auto& child : nodes)
    {
        //std::cout << "Handling " << std::get<0>(child) << ": " << std::get<1>(child) << std::endl;

        if
            (NodeType::empty == std::get<2>(child)->getType())
        {
            //std::cout << child.first << ": null" << std::endl;
        }
        else if
            (NodeType::boolean == std::get<2>(child)->getType())
        {
            //std::cout << "Has a boolean field: " << child.first << " " << child.second << std::endl;
            NanoguiBoolFieldEditor* boolEditor = this->getBoolEditor(boolEditorsCount, std::get<0>(child));
            boolEditor->setName(std::get<1>(child));
            boolEditor->setOriginalValue(static_cast<BoolNode*>(std::get<2>(child))->value);
            boolEditor->setValue(static_cast<BoolNode*>(std::get<2>(child))->value);
            boolEditor->activateEdition(false);
            boolEditor->setVisible(true);

            this->_sortedWidgets[boolEditor] = count;
            this->_widgetIsContainer[boolEditor] = false;

            count = count + 1;
            boolEditorsCount = boolEditorsCount + 1;
        }
        else if
            (NodeType::number == std::get<2>(child)->getType())
        {
            //std::cout << "Has a number field: " << child.first << " " << child.second << std::endl;
            NanoguiNumFieldEditor* numEditor = this->getNumEditor(numEditorsCount, std::get<0>(child));
            numEditor->setName(std::get<1>(child));
            numEditor->setOriginalValue(static_cast<NumberNode*>(std::get<2>(child))->value);
            numEditor->setValue(static_cast<NumberNode*>(std::get<2>(child))->value);
            numEditor->activateEdition(false);
            numEditor->setVisible(true);

            this->_sortedWidgets[numEditor] = count;
            this->_widgetIsContainer[numEditor] = false;

            count = count + 1;
            numEditorsCount = numEditorsCount + 1;
        }
        else if
            (NodeType::text == std::get<2>(child)->getType())
        {
            //std::cout << "Has a text field field: " << child.first << " " << child.second << std::endl;
            NanoguiTextFieldEditor* textEditor = this->getTextEditor(textEditorsCount, std::get<0>(child));
            textEditor->setName(std::get<1>(child));
            textEditor->setOriginalValue(static_cast<TextNode*>(std::get<2>(child))->value);
            textEditor->setValue(static_cast<TextNode*>(std::get<2>(child))->value);
            textEditor->activateEdition(false);
            textEditor->setVisible(true);

            this->_sortedWidgets[textEditor] = count;
            this->_widgetIsContainer[textEditor] = false;

            count = count + 1;
            textEditorsCount = textEditorsCount + 1;
        }
        else if
            (NodeType::image == std::get<2>(child)->getType())
        {
            //std::cout << "Has an image field: " << child.first << " " << child.second << std::endl;
            NanoguiImageFieldEditor* imageEditor = this->getImageEditor(imageEditorsCount, std::get<0>(child));
            imageEditor->setName(std::get<1>(child));
            imageEditor->setOriginalValue(static_cast<ImageNode*>(std::get<2>(child))->value);
            imageEditor->setValue(static_cast<ImageNode*>(std::get<2>(child))->value);
            imageEditor->activateEdition(false);
            imageEditor->setVisible(true);

            this->_sortedWidgets[imageEditor] = count;
            this->_widgetIsContainer[imageEditor] = false;

            count = count + 1;
            imageEditorsCount = imageEditorsCount + 1;
        }
        else if
            (NodeType::list == std::get<2>(child)->getType())
        {
        }
        else if
            (NodeType::dict == std::get<2>(child)->getType())
        {
            NanoguiDictFieldEditor* dictEditor = this->getDictEditor(dictEditorsCount, "" /*dict nodes are always top-level*/);
            dictEditor->setName(std::get<0>(child)); /* here we use field 0 which is generally the category*/

            dictEditor->activateEdition(false);
            dictEditor->setVisible(true);

            this->_sortedWidgets[dictEditor] = count;
            this->_widgetIsContainer[dictEditor] = true;

            count = count + 1;
            dictEditorsCount = dictEditorsCount + 1;
        }
    }
}

/**
 *
 */
void NanoguiPropertiesEditor::setProperties(DictNode&& properties)
{
    this->hideImageViewer();

    this->_properties = std::move(properties);

    std::size_t boolEditorsCount(0);
    std::size_t numEditorsCount(0);
    std::size_t textEditorsCount(0);
    std::size_t imageEditorsCount(0);
    std::size_t dictEditorsCount(0);

    const auto& propertiesNodeFound = this->_properties.children.find("properties");
    if
        (this->_properties.children.end() != propertiesNodeFound)
    {
        this->updateSubEditors(static_cast<DictNode*>(propertiesNodeFound->second), boolEditorsCount, numEditorsCount, textEditorsCount, imageEditorsCount, dictEditorsCount, "" /*root level*/);
    }

    // Hide unused bool editors
    while
        (boolEditorsCount < this->_boolEditors.size())
    {
        NanoguiBoolFieldEditor* unusedBoolEditor = this->_boolEditors.at(boolEditorsCount);

        unusedBoolEditor->activateEdition(false);
        unusedBoolEditor->setVisible(false);

        boolEditorsCount = boolEditorsCount + 1;
    }

    // Hide unused num editors
    while
        (numEditorsCount < this->_numEditors.size())
    {
        NanoguiNumFieldEditor* unusedNumEditor = this->_numEditors.at(numEditorsCount);

        unusedNumEditor->activateEdition(false);
        unusedNumEditor->setVisible(false);

        numEditorsCount = numEditorsCount + 1;
    }

    // Hide unused text editors
    while
        (textEditorsCount < this->_textEditors.size())
    {
        NanoguiTextFieldEditor* unusedTextEditor = this->_textEditors.at(textEditorsCount);

        unusedTextEditor->activateEdition(false);
        unusedTextEditor->setVisible(false);

        textEditorsCount = textEditorsCount + 1;
    }

    // Hide unused image editors
    while
        (imageEditorsCount < this->_imageEditors.size())
    {
        NanoguiImageFieldEditor* unusedImageEditor = this->_imageEditors.at(imageEditorsCount);

        unusedImageEditor->activateEdition(false);
        unusedImageEditor->setVisible(false);

        imageEditorsCount = imageEditorsCount + 1;
    }

    // Hide unused dict editors
    while
        (dictEditorsCount < this->_dictEditors.size())
    {
        NanoguiDictFieldEditor* unusedDictEditor = this->_dictEditors.at(dictEditorsCount);

        unusedDictEditor->activateEdition(false);
        unusedDictEditor->setVisible(false);

        dictEditorsCount = dictEditorsCount + 1;
    }

    this->activateEdition(false);

    this->update();
}

/**
 *
 */
bool NanoguiPropertiesEditor::getProperties(DictNode& result)
{
    for
        (const auto& dictEditor : this->_dictEditors)
    {
        if
            (false == dictEditor->visible())
        {
            continue;
        }

        result.children.emplace(dictEditor->getName(), new DictNode());
    }

    for
        (const auto& boolEditor : this->_boolEditors)
    {
        if
            (false == boolEditor->visible())
        {
            continue;
        }

        if
            ("" == boolEditor->getCategory())
        {
            result.children.emplace(std::make_pair(boolEditor->getName(), new BoolNode(boolEditor->getValue())));
        }
        else
        {
            auto found = result.children.find(boolEditor->getCategory());
            if
                (result.children.end() == found)
            {
                LOG(LogLevel::Error, "Nanogui properties editor",
                         "Could not find dict node named '{}' for field '{}'",
                         boolEditor->getCategory(), boolEditor->getName());
                return false;
            }

            static_cast<DictNode*>(found->second)->children.emplace(std::make_pair(boolEditor->getName(), new BoolNode(boolEditor->getValue())));
        }
    }

    for
        (const auto& numEditor : this->_numEditors)
    {
        if
            (false == numEditor->visible())
        {
            continue;
        }

        if
            ("" == numEditor->getCategory())
        {
            result.children.emplace(std::make_pair(numEditor->getName(), new NumberNode(numEditor->getValue())));
        }
        else
        {
            auto found = result.children.find(numEditor->getCategory());
            if
                (result.children.end() == found)
            {
                LOG(LogLevel::Error, "Nanogui properties editor",
                         "Could not find dict node named '{}' for field '{}'",
                         numEditor->getCategory(), numEditor->getName());
                return false;
            }

            static_cast<DictNode*>(found->second)->children.emplace(std::make_pair(numEditor->getName(), new NumberNode(numEditor->getValue())));
        }
    }

    for
        (const auto& textEditor : this->_textEditors)
    {
        if
            (false == textEditor->visible())
        {
            continue;
        }

        if
            ("" == textEditor->getCategory())
        {
            result.children.emplace(std::make_pair(textEditor->getName(), new TextNode(textEditor->getValue())));
        }
        else
        {
            auto found = result.children.find(textEditor->getCategory());
            if
                (result.children.end() == found)
            {
                LOG(LogLevel::Error, "Nanogui properties editor",
                         "Could not find dict node named '{}' for field '{}'",
                         textEditor->getCategory(), textEditor->getName());
                return false;
            }

            static_cast<DictNode*>(found->second)->children.emplace(std::make_pair(textEditor->getName(), new TextNode(textEditor->getValue())));
        }
    }

    for
        (const auto& imageEditor : this->_imageEditors)
    {
        if
            (false == imageEditor->visible())
        {
            continue;
        }

        if
            ("" == imageEditor->getCategory())
        {
            result.children.emplace(std::make_pair(imageEditor->getName(), new ImageNode(imageEditor->getValue())));
        }
        else
        {
            auto found = result.children.find(imageEditor->getCategory());
            if
                (result.children.end() == found)
            {
                LOG(LogLevel::Error, "Nanogui properties editor",
                         "Could not find dict node named '{}' for field '{}'",
                         imageEditor->getCategory(), imageEditor->getName());
                return false;
            }

            static_cast<DictNode*>(found->second)->children.emplace(std::make_pair(imageEditor->getName(), new ImageNode(imageEditor->getValue())));
        }
    }

    return true;
}

/**
 *
 */
bool NanoguiPropertiesEditor::widgetsSorter(const nanogui::Widget* a, const nanogui::Widget* b) const
{
    auto aFound = this->_sortedWidgets.find(a);
    auto bFound = this->_sortedWidgets.find(b);

    if
        (  (this->_sortedWidgets.end() != aFound)
        && (this->_sortedWidgets.end() != bFound))
    {
        return aFound->second < bFound->second;
    }

    return a > b;
}

/**
 *
 */
NanoguiBoolFieldEditor* NanoguiPropertiesEditor::getBoolEditor(std::size_t index, const std::string& category)
{
    NanoguiBoolFieldEditor* editor(nullptr);
    if
        (index < this->_boolEditors.size())
    {
        editor = this->_boolEditors.at(index);
    }
    else
    {
        this->_boolEditors.emplace_back(new NanoguiBoolFieldEditor(this->_itemsList->content));
        editor = this->_boolEditors.back();
        editor->setDeleteCallback([this, editor] (const std::string& /*category*/, const std::string& /*field*/) {
            //std::cout << "Delete bool field [" << category << "] " << field << std::endl;
            editor->activateEdition(false);
            editor->setVisible(false);
            this->_reflowNeeded = true;
        });
    }

    this->_sortedWidgets[editor] = 0;
    editor->setCategory(category);

    //std::cout << "Total bool editors: " << this->_boolEditors.size() << std::endl;

    return editor;
}

/**
 *
 */
NanoguiNumFieldEditor* NanoguiPropertiesEditor::getNumEditor(std::size_t index, const std::string& category)
{
    NanoguiNumFieldEditor* editor(nullptr);
    if
        (index < this->_numEditors.size())
    {
        editor = this->_numEditors.at(index);
    }
    else
    {
        this->_numEditors.emplace_back(new NanoguiNumFieldEditor(this->_itemsList->content));
        editor = this->_numEditors.back();
        editor->setDeleteCallback([this, editor] (const std::string& /*category*/, const std::string& /*field*/) {
            //std::cout << "Delete num field [" << category << "] " << field << std::endl;
            editor->activateEdition(false);
            editor->setVisible(false);
            this->_reflowNeeded = true;
        });
    }

    this->_sortedWidgets[editor] = 0;
    editor->setCategory(category);

    //std::cout << "Total num editors: " << this->_numEditors.size() << std::endl;

    return editor;
}

/**
 *
 */
NanoguiTextFieldEditor* NanoguiPropertiesEditor::getTextEditor(std::size_t index, const std::string& category)
{
    NanoguiTextFieldEditor* editor(nullptr);
    if
        (index < this->_textEditors.size())
    {
        editor = this->_textEditors.at(index);
    }
    else
    {
        this->_textEditors.emplace_back(new NanoguiTextFieldEditor(this->_itemsList->content));
        editor = this->_textEditors.back();
        editor->setDeleteCallback([this, editor] (const std::string& /*category*/, const std::string& /*field*/) {
            //std::cout << "Delete field [" << category << "] " << field << std::endl;
            editor->activateEdition(false);
            editor->setVisible(false);
            this->_reflowNeeded = true;
        });
    }

    this->_sortedWidgets[editor] = 0;
    editor->setCategory(category);

    //std::cout << "Total text editors: " << this->_textEditors.size() << std::endl;

    return editor;
}

/**
 *
 */
NanoguiImageFieldEditor* NanoguiPropertiesEditor::getImageEditor(std::size_t index, const std::string& category)
{
    NanoguiImageFieldEditor* editor(nullptr);

    if
        (index < this->_imageEditors.size())
    {
        editor = this->_imageEditors.at(index);
    }
    else
    {
        this->_imageEditors.emplace_back(new NanoguiImageFieldEditor(this->_itemsList->content));
        editor = this->_imageEditors.back();
        editor->setLoadImageCallback([this, editor] () {
            if
                (true == this->_preventImageViewer)
            {
                return;
            }
            this->_pathContainer->setVisible(true);
            this->_pathSelector->show();
            this->_pathField->setCaption("Select image for '" + editor->getName() + "'");
            this->_pathSelectorTarget = editor;
            this->_imageView->reset();
            this->_imageView->setVisible(false);
            this->_imageContainer->setVisible(true);
            this->_imageContainer->layoutUpdateRequested = true;
        });
        editor->setViewImageCallback([this] (const Image& image) {
            if
                (true == this->_preventImageViewer)
            {
                return;
            }
            this->_pathSelector->hide();
            this->_pathContainer->setVisible(false);
            this->_imageView->setImage(image);
            this->_imageView->setVisible(true);
            this->_imageContainer->setVisible(true);
            this->_imageContainer->layoutUpdateRequested = true;
        });
        editor->setResetCallback([this, editor] () {
            //std::cout << "Reset callback" << std::endl;
            this->_imageView->setImage(editor->getConstImage());
        });
        editor->setDeleteCallback([this, editor] (const std::string& /*category*/, const std::string& /*field*/) {
            //std::cout << "Delete field [" << category << "] " << field << std::endl;
            editor->activateEdition(false);
            editor->setVisible(false);
            this->_reflowNeeded = true;
        });
    }

    this->_sortedWidgets[editor] = 0;
    editor->setCategory(category);

    //std::cout << "Total image editors: " << this->_imageEditors.size() << std::endl;

    return editor;
}

/**
 *
 */
NanoguiDictFieldEditor* NanoguiPropertiesEditor::getDictEditor(std::size_t index, const std::string& category)
{
    if
        (index < this->_dictEditors.size())
    {
        return this->_dictEditors.at(index);
    }

    this->_dictEditors.emplace_back(new NanoguiDictFieldEditor(this->_itemsList->content));

    NanoguiDictFieldEditor* editor = this->_dictEditors.back();
    editor->setCategory(category);
    editor->setDeleteCallback([this, editor] (const std::string& /*category*/, const std::string& /*field*/) {
        editor->activateEdition(false);
        editor->setVisible(false);

        for
            (auto it: this->_boolEditors)
        {
            if
                (editor->getName() == it->getCategory())
            {
                it->activateEdition(false);
                it->setVisible(false);
            }
        }

        for
            (auto it: this->_numEditors)
        {
            if
                (editor->getName() == it->getCategory())
            {
                it->activateEdition(false);
                it->setVisible(false);
            }
        }

        for
            (auto it: this->_textEditors)
        {
            if
                (editor->getName() == it->getCategory())
            {
                it->activateEdition(false);
                it->setVisible(false);
            }
        }

        for
            (auto it: this->_imageEditors)
        {
            if
                (editor->getName() == it->getCategory())
            {
                it->activateEdition(false);
                it->setVisible(false);
            }
        }

        this->_reflowNeeded = true;
    });
    editor->setChildUnicityCallback([this] (const std::string& nodeCategory, const std::string& nodeLabel) -> bool{
        return this->checkPropertyUnicity(nodeCategory, nodeLabel);
    });
    editor->setAddPropertyOkCallback([this] (const std::string& nodeType, const std::string& nodeCategory, const std::string& nodeLabel) {
        this->addProperty(nodeType, nodeCategory, nodeLabel);
    });
    editor->setLayoutChangeCallback([this] () {
        this->update();
    });

    this->_sortedWidgets[editor] = 0;

    editor = nullptr;

    return this->_dictEditors.back();
}


/**
 *
 */
bool NanoguiPropertiesEditor::checkPropertyUnicity(const std::string& nodeCategory, const std::string& nodeLabel)
{
    for
        (const auto& boolEditor : this->_boolEditors)
    {
        if
            (  (true == boolEditor->visible())
            && (boolEditor->getCategory() == nodeCategory)
            && (boolEditor->getName() == nodeLabel))
        {
            return false;
        }
    }

    for
        (const auto& numEditor : this->_numEditors)
    {
        if
            (  (true == numEditor->visible())
            && (numEditor->getCategory() == nodeCategory)
            && (numEditor->getName() == nodeLabel))
        {
            return false;
        }
    }

    for
        (const auto& textEditor : this->_textEditors)
    {
        if
            (  (true == textEditor->visible())
            && (textEditor->getCategory() == nodeCategory)
            && (textEditor->getName() == nodeLabel))
        {
            return false;
        }
    }

    for
        (const auto& imgEditor : this->_imageEditors)
    {
        if
            (  (true == imgEditor->visible())
            && (imgEditor->getCategory() == nodeCategory)
            && (imgEditor->getName() == nodeLabel))
        {
            return false;
        }
    }

    for
        (const auto& dictEditor : this->_dictEditors)
    {
        if
            (  (true == dictEditor->visible())
            && (dictEditor->getCategory() == nodeCategory)
            && (dictEditor->getName() == nodeLabel))
        {
            return false;
        }
    }

    return true;
}


/**
 *
 */
bool NanoguiPropertiesEditor::addProperty(const std::string& nodeType, const std::string& nodeCategory, const std::string& nodeLabel)
{
    std::size_t nodeIndex(0u);

    if
        ("bool" == nodeType)
    {
        for
            (const auto& it: this->_boolEditors)
        {
            if
                (true == it->visible())
            {
                nodeIndex = nodeIndex + 1;
            }
        }

        // Compute index before creating the widget to avoid side-effects
        std::size_t fieldIndex = this->getFieldIndex(nodeCategory, nodeLabel);

        NanoguiBoolFieldEditor* boolEditor = this->getBoolEditor(nodeIndex, nodeCategory);
        boolEditor->setName(nodeLabel);
        boolEditor->setOriginalValue(false);
        boolEditor->setValue(false);
        boolEditor->activateEdition(true);
        boolEditor->setVisible(true);

        for
            (auto& it : this->_sortedWidgets)
        {
            if
                (it.second >= fieldIndex)
            {
                it.second = it.second + 1;
            }
        }
        this->_sortedWidgets[boolEditor] = static_cast<unsigned int>(fieldIndex);
        this->_widgetIsContainer[boolEditor] = false;
    }
    else if
        ("num" == nodeType)
    {
        for
            (const auto& it: this->_numEditors)
        {
            if
                (true == it->visible())
            {
                nodeIndex = nodeIndex + 1;
            }
        }

        // Compute index before creating the widget to avoid side-effects
        std::size_t fieldIndex = this->getFieldIndex(nodeCategory, nodeLabel);

        NanoguiNumFieldEditor* numEditor = this->getNumEditor(nodeIndex, nodeCategory);
        numEditor->setName(nodeLabel);
        numEditor->setOriginalValue(0.0);
        numEditor->setValue(0.0);
        numEditor->activateEdition(true);
        numEditor->setVisible(true);

        for
            (auto& it : this->_sortedWidgets)
        {
            if
                (it.second >= fieldIndex)
            {
                it.second = it.second + 1;
            }
        }
        this->_sortedWidgets[numEditor] = static_cast<unsigned int>(fieldIndex);
        this->_widgetIsContainer[numEditor] = false;
    }
    else if
        ("text" == nodeType)
    {
        for
            (const auto& it: this->_textEditors)
        {
            if
                (true == it->visible())
            {
                nodeIndex = nodeIndex + 1;
            }
        }

        // Compute index before creating the widget to avoid side-effects
        std::size_t fieldIndex = this->getFieldIndex(nodeCategory, nodeLabel);

        NanoguiTextFieldEditor* textEditor = this->getTextEditor(nodeIndex, nodeCategory);
        textEditor->setName(nodeLabel);
        textEditor->setOriginalValue("");
        textEditor->setValue("");
        textEditor->activateEdition(true);
        textEditor->setVisible(true);

        for
            (auto& it : this->_sortedWidgets)
        {
            if
                (it.second >= fieldIndex)
            {
                it.second = it.second + 1;
            }
        }
        this->_sortedWidgets[textEditor] = static_cast<unsigned int>(fieldIndex);
        this->_widgetIsContainer[textEditor] = false;
    }
    else if
        ("img" == nodeType)
    {
        for
            (const auto& it: this->_imageEditors)
        {
            if
                (true == it->visible())
            {
                nodeIndex = nodeIndex + 1;
            }
        }

        // Compute index before creating the widget to avoid side-effects
        std::size_t fieldIndex = this->getFieldIndex(nodeCategory, nodeLabel);

        NanoguiImageFieldEditor* imageEditor = this->getImageEditor(nodeIndex, nodeCategory);
        imageEditor->setName(nodeLabel);
        imageEditor->setOriginalValue("");
        imageEditor->setValue("");
        imageEditor->activateEdition(true);
        imageEditor->setVisible(true);

        for
            (auto& it : this->_sortedWidgets)
        {
            if
                (it.second >= fieldIndex)
            {
                it.second = it.second + 1;
            }
        }
        this->_sortedWidgets[imageEditor] = static_cast<unsigned int>(fieldIndex);
        this->_widgetIsContainer[imageEditor] = false;
    }
    else if
        ("dict" == nodeType)
    {
        for
            (const auto& it: this->_dictEditors)
        {
            if
                (true == it->visible())
            {
                nodeIndex = nodeIndex + 1;
            }
        }

        // Compute index before creating the widget to avoid side-effects
        std::size_t fieldIndex = this->getFieldIndex(nodeLabel, "");

        NanoguiDictFieldEditor* dictEditor = this->getDictEditor(nodeIndex, "" /*dict nodes are always top-level*/);
        dictEditor->setName(nodeLabel);

        dictEditor->activateEdition(true);
        dictEditor->setVisible(true);

        for
            (auto& it : this->_sortedWidgets)
        {
            if
                (it.second >= fieldIndex)
            {
                it.second = it.second + 1;
            }
        }
        this->_sortedWidgets[dictEditor] = static_cast<unsigned int>(fieldIndex);
        this->_widgetIsContainer[dictEditor] = true;
    }
    else
    {
        return false;
    }

    this->update();
    return true;
}


/**
 *
 */
void NanoguiPropertiesEditor::resize(const Eigen::Vector2i& size)
{
    this->setFixedSize(size);
    this->setSize(size);
    this->_itemsList->setFixedSize(Eigen::Vector2i(size.x(), size.y() - 20 - 32 * (this->_addPropertyForm->visible() ? 1:0) ));
    this->_itemsList->setSize(Eigen::Vector2i(size.x(), size.y() - 20 - 32 * (this->_addPropertyForm->visible() ? 1:0) ));
    this->_itemsList->update(false /*do not reset scrolling*/);
    this->_reflowNeeded = true;
}

/**
 *
 */
void NanoguiPropertiesEditor::initImageViewer()
{
    this->_imageContainer = NanoguiAnchored::create(this);
    this->_imageContainer->offset = 10u;
    this->_imageContainer->placement = NanoguiAnchored::ON_RIGHT;
    this->_imageContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                            nanogui::Alignment::Middle,
                                                            3, 2));
    nanogui::Widget* header = new nanogui::Widget(this->_imageContainer);
    header->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                             nanogui::Alignment::Maximum,
                                             2, 0));
    header->setFixedSize(Eigen::Vector2i(360, 20));

    nanogui::Button* closeButton = new nanogui::Button(header, "", ENTYPO_ICON_CROSS);
    closeButton->setFixedSize(Eigen::Vector2i(18, 18));
    closeButton->setFontSize(16);
    closeButton->setCallback([this] () {
        this->_pathContainer->setVisible(false);
        this->_imageContainer->setVisible(false);
    });

    // Path selection
    this->_pathContainer = new nanogui::Widget(this->_imageContainer);
    this->_pathContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                           nanogui::Alignment::Maximum,
                                                           2, 0));
    this->_pathContainer->setFixedSize(Eigen::Vector2i(360, 84));
    this->_pathContainer->setSize(Eigen::Vector2i(360, 84));

    nanogui::Widget* nameBlock = new nanogui::Widget(this->_pathContainer);
    nameBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                nanogui::Alignment::Middle,
                                                0, 0));
    nameBlock->setFixedSize(Eigen::Vector2i(360, 20));
    this->_pathField = new nanogui::Label(nameBlock, "Select image for ");
    this->_pathSelector = new NanoguiPathSelector(this->_pathContainer, PathSelectorType::get_file);
    this->_pathSelector->resize(Eigen::Vector2i(356, 40));
    this->_pathSelector->setChangeCallback([this] (const std::string& value) {
        bool ok {false};    //

        ok = Path::fileExists(value);
        ok = ok && (4 < value.size());
        if
            (true == ok)
        {
            std::string ext = value.substr(value.size() - 4);
            ok = (  SEAScope::iequals(ext, ".png")
                 || SEAScope::iequals(ext, ".jpg")
                 || SEAScope::iequals(ext, ".jpeg")
                 || SEAScope::iequals(ext, ".bmp")
                 || SEAScope::iequals(ext, ".tga")
                 || SEAScope::iequals(ext, ".gif")
                 || SEAScope::iequals(ext, ".ppm")
                 || SEAScope::iequals(ext, ".pgm"));
        }
        this->_pathOkButton->setEnabled(ok);
    });

    nanogui::Widget* pathButtons = new nanogui::Widget(this->_pathContainer);
    pathButtons->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                  nanogui::Alignment::Middle,
                                                  0, 0));
    this->_pathOkButton = new nanogui::Button(pathButtons, "Load");
    this->_pathOkButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_pathOkButton->setFontSize(16);
    this->_pathOkButton->setCallback([this] () {
        if
            (nullptr != this->_pathSelectorTarget)
        {
            const Image& img = this->_pathSelectorTarget->loadFromFile(this->_pathSelector->getPath());
            this->_imageView->setImage(img);
            this->_imageView->setVisible(true);
        }
        this->_imageContainer->setVisible(true);
        this->_imageContainer->layoutUpdateRequested = true;
    });
    this->_pathOkButton->setEnabled(false);

    this->_pathNokButton = new nanogui::Button(pathButtons, "Cancel");
    this->_pathNokButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_pathNokButton->setFontSize(16);
    this->_pathNokButton->setCallback([this] () {
        this->_pathSelector->hide();
        this->_pathSelectorTarget = nullptr;
        this->_imageContainer->setVisible(false);
    });
    this->_imageView = new NanoguiImageView(this->_imageContainer, 0u);
    this->_imageView->setFixedSize(Eigen::Vector2i(360, 360));
}

/**
 *
 */
void NanoguiPropertiesEditor::hideImageViewer()
{
    this->_pathSelector->hide();
    this->_pathContainer->setVisible(false);
    this->_imageContainer->setVisible(false);
}

/**
 *
 */
bool NanoguiPropertiesEditor::setSaveCallback(const std::function<void(const DictNode&)>& callback)
{
    if
        (nullptr != this->_saveCallback)
    {
        return false;
    }

    this->_saveCallback = callback;
    return true;
}

/**
 *
 */
std::size_t NanoguiPropertiesEditor::getFieldIndex(const std::string& category, const std::string& name) const
{
    std::size_t result(0u);

    for
        (const auto& boolEditor: this->_boolEditors)
    {
        if
            (false == boolEditor->visible())
        {
            continue;
        }

        if
            (  boolEditor->getCategory() == category
            && boolEditor->getName() < name)
        {
            //std::cout << "Bool editor '" << boolEditor->getCategory() << ":" << boolEditor->getName() << "' arrives before " << category <<":" << name <<  std::endl;
            result = result + 1;
        }
        else if
            (boolEditor->getCategory() < category)
        {
            //std::cout << "Bool editor '" << boolEditor->getCategory() << ":" << boolEditor->getName() << "' arrives before " << category <<":" << name << std::endl;
            result = result + 1;
        }
        /*
        else
        {
            //std::cout << "== " << this->_sortedWidgets.at(boolEditor) << " for bool editor " << boolEditor->getCategory() << ":" << boolEditor->getName() << std::endl;
        }
        */
    }

    for
        (const auto& numEditor: this->_numEditors)
    {
        if
            (false == numEditor->visible())
        {
            continue;
        }

        if
            (  numEditor->getCategory() == category
            && numEditor->getName() < name)
        {
            //std::cout << "Num editor '" << numEditor->getCategory() << ":" << numEditor->getName() << "' arrives before " << category <<":" << name << std::endl;
            result = result + 1;
        }
        else if
            (numEditor->getCategory() < category)
        {
            //std::cout << "Num editor '" << numEditor->getCategory() << ":" << numEditor->getName() << "' arrives before " << category <<":" << name << std::endl;
            result = result + 1;
        }
        /*
        else
        {
            //std::cout << "== " << this->_sortedWidgets.at(numEditor) << " for num editor " << numEditor->getCategory() << ":" << numEditor->getName() << std::endl;
        }
        */
    }

    for
        (const auto& textEditor: this->_textEditors)
    {
        if
            (false == textEditor->visible())
        {
            continue;
        }

        if
            (  textEditor->getCategory() == category
            && textEditor->getName() < name)
        {
            //std::cout << "Text editor '" << textEditor->getCategory() << ":" << textEditor->getName() << "' arrives before " << category <<":" << name << std::endl;
            result = result + 1;
        }
        else if
            (textEditor->getCategory() < category)
        {
            //std::cout << "Text editor '" << textEditor->getCategory() << ":" << textEditor->getName() << "' arrives before " << category <<":" << name << std::endl;
            result = result + 1;
        }
        /*
        else
        {
            //std::cout << "== " << this->_sortedWidgets.at(textEditor) << " for text editor " << textEditor->getCategory() << ":" << textEditor->getName() << std::endl;
        }
        */
    }

    for
        (const auto& imgEditor: this->_imageEditors)
    {
        if
            (false == imgEditor->visible())
        {
            continue;
        }

        if
            (  imgEditor->getCategory() == category
            && imgEditor->getName() < name)
        {
            //std::cout << "Image editor '" << imgEditor->getCategory() << ":" << imgEditor->getName() << "' arrives before " << category <<":" << name << std::endl;
            result = result + 1;
        }
        else if
            (imgEditor->getCategory() < category)
        {
            //std::cout << "Image editor '" << imgEditor->getCategory() << ":" << imgEditor->getName() << "' arrives before " << category <<":" << name << std::endl;
            result = result + 1;
        }
        /*
        else
        {
            //std::cout << "== " << this->_sortedWidgets.at(imgEditor) << " for image editor " << imgEditor->getCategory() << ":" << imgEditor->getName() << std::endl;
        }
        */
    }

    for
        (const auto& dictEditor: this->_dictEditors)
    {
        if
            (false == dictEditor->visible())
        {
            continue;
        }

        if
            (dictEditor->getName() <= category)
        {
            //std::cout << "Dict editor '" << dictEditor->getName() << "' arrives before " << category << std::endl;
            result = result + 1;
        }
        /*
        else
        {
            //std::cout << "== " << this->_sortedWidgets.at(dictEditor) << " for dict editor " << dictEditor->getName() << std::endl;
        }
        */
    }

    //std::cout << "Node " << category << ":" << name << " will be added in position " << result << std::endl;
    return result;
}
