// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "ui/nanogui/nanogui_renderer.h"
#include "ui/nanogui/nanogui_events_bubbling.h"
#include "ui/idrop_target.h"
#include "glfw_window.h"
#include "3d/icons.h"

using namespace SEAScope;

/**
 *
 */
NanoguiRenderer::NanoguiRenderer()
{
};

/**
 *
 */
NanoguiRenderer::~NanoguiRenderer()
{

};

/**
 *
 */
bool NanoguiRenderer::setWindow(IWindow& window)
{
	GLFWWindow* w;	//

	w = dynamic_cast<GLFWWindow*>(&window);
	if
		(nullptr == w)
	{
		return false;
	}
	this->initialize(w->implementation, false);
	return true;
};

/**
 *
 */
bool NanoguiRenderer::renderFrame()
{
	if
		(false == this->visible())
	{
		return false;
	}

	this->drawWidgetsWithCustomizedTooltips();

	return true;
};

/**
 *
 */
bool NanoguiRenderer::handleMouseMoveEvent(double x, double y)
{
	return this->cursorPosCallbackEvent(x,y);
};

/**
 *
 */
bool NanoguiRenderer::handleKeyEvent(int key, int scancode, int action, int mods)
{
    bool ok {false};    //

    // Nanogui uses events capturing (parents evaluated first)
	ok = this->keyCallbackEvent(key, scancode, action, mods);
    if
        (true == ok)
    {
        return true;
    }

    // If the event has not been captured, then try the other way around
    // (bubbling, i.e. leaf child first).
    //
    // Note: mFocusPath starts with leaf child then goes up the ancestry tree
    if
        (0 < this->mFocusPath.size())
    {
        // Set the bit indicating event bubbling in the modifiers
        int modifiers = (mods | NANOGUI_EVENT_BUBBLING);

        for
            (auto it = this->mFocusPath.begin() + 1; it != this->mFocusPath.end(); ++it)
        {
            if
                ((*it)->focused() && (*it)->keyboardEvent(key, scancode, action, modifiers))
            {
                return true;
            }
        }
    }

    return false;
};

/**
 *
 */
bool NanoguiRenderer::handleCharEvent(unsigned int c, int /*mods*/)
{
	return this->charCallbackEvent(c);
};

/**
 *
 */
bool NanoguiRenderer::handleMouseButtonEvent(int button, int action, int mods)
{
	return this->mouseButtonCallbackEvent(button, action, mods);
};

/**
 *
 */
bool NanoguiRenderer::handleMouseWheelEvent(double x, double y)
{
	return this->scrollCallbackEvent(x,y);
};

/**
 *
 */
bool NanoguiRenderer::handleDropEvent(int count, const char** filenames)
{
	return this->dropCallbackEvent(count, filenames);
};

/**
 *
 */
bool NanoguiRenderer::handleResizeEvent(int width, int height, int, int)
{
	bool ok;	//

	ok = this->resizeCallbackEvent(width, height);
	if
		(nullptr != this->resizeCallback)
	{
		this->resizeCallback(width, height);
	}
	return ok;
};

/**
 *
 */
bool NanoguiRenderer::dropEvent(const std::vector<std::string>& filenames)
{
    nanogui::Widget* dropWidget = this->findWidget(this->mMousePos);
    while
        (nullptr != dropWidget)
    {
        IDropTarget* target = dynamic_cast<IDropTarget*>(dropWidget);
        if
            (nullptr != target)
        {
            target->dropEvent(filenames);
            return true;
        }

        dropWidget = dropWidget->parent();
    }

    return false;
}

/**
 *
 */
void NanoguiRenderer::setIcons(const Icons& icons)
{
	this->_icons.setTextures(this->nvgContext(), icons);
}

/**
 *
 */
NanoguiIcons& NanoguiRenderer::getIcons()
{
	return this->_icons;
}

/**
 *
 */
void NanoguiRenderer::drawWidgetsWithCustomizedTooltips()
{
    if
        (false == this->mVisible)
    {
        return;
    }

    glfwMakeContextCurrent(this->mGLFWWindow);

    glfwGetFramebufferSize(this->mGLFWWindow, &(this->mFBSize[0]), &(this->mFBSize[1]));
    glfwGetWindowSize(this->mGLFWWindow, &(this->mSize[0]), &(this->mSize[1]));

#if defined(_WIN32) || defined(__linux__)
    this->mSize = (this->mSize.cast<float>() / this->mPixelRatio).cast<int>();
    this->mFBSize = (this->mSize.cast<float>() * this->mPixelRatio).cast<int>();
#else
    /* Recompute pixel ratio on OSX */
    if
        (this->mSize[0])
    {
        this->mPixelRatio = static_cast<float>(this->mFBSize[0]) / static_cast<float>(this->mSize[0]);
    }
#endif

    glViewport(0, 0, this->mFBSize[0], this->mFBSize[1]);
    glBindSampler(0, 0);
    nvgBeginFrame(this->mNVGContext, static_cast<float>(this->mSize[0]), static_cast<float>(this->mSize[1]), this->mPixelRatio);

    this->draw(this->mNVGContext);

    double elapsed = glfwGetTime() - this->mLastInteraction;

    if
        (elapsed > 0.5)
    {
        /* Draw tooltips */
        const Widget *widget = this->findWidget(this->mMousePos);
        if
            (widget && !widget->tooltip().empty())
        {
            float tooltipWidth = 200.f;

            // Computing text bounds
            float bounds[4];
            nvgFontFace(this->mNVGContext, "sans");
            nvgFontSize(this->mNVGContext, 15.0f);
            nvgTextAlign(this->mNVGContext, NVG_ALIGN_LEFT | NVG_ALIGN_TOP);
            nvgTextLineHeight(this->mNVGContext, 1.1f);
            Eigen::Vector2i pos = this->mSize / 2 +
                                  Eigen::Vector2i(widget->width() / 2, widget->height() + 10);

            /*
            nvgTextBounds(this->mNVGContext, static_cast<float>(pos.x()), static_cast<float>(pos.y()),
                            widget->tooltip().c_str(), nullptr, bounds);
            float h = 0.5f * (bounds[2] - bounds[0]);
            if
                (h > tooltipWidth / 2.f)
            {
            */
                // Recompute with text-wrapping
                nvgTextAlign(this->mNVGContext, NVG_ALIGN_CENTER | NVG_ALIGN_TOP);
                nvgTextBoxBounds(this->mNVGContext, static_cast<float>(pos.x()), static_cast<float>(pos.y()), tooltipWidth,
                                widget->tooltip().c_str(), nullptr, bounds);
            //}

            // Actually drawing the tooltip
            nvgGlobalAlpha(this->mNVGContext,
                           static_cast<float>(std::min(1.0, 2 * (elapsed - 0.5))));

            nvgBeginPath(this->mNVGContext);
            nvgFillColor(this->mNVGContext, nanogui::Color(250, 255));

            float computedWidth {bounds[2] - bounds[0]};
            float computedHeight {bounds[3] - bounds[1]};
            float computedX {static_cast<float>(widget->absolutePosition().x())};
            float computedY {static_cast<float>(widget->absolutePosition().y())};
            Eigen::Vector2f p1;
            Eigen::Vector2f p2;
            Eigen::Vector2f p3;
            bool hCentered {false};
            if
                (computedX + 0.5f * (static_cast<float>(widget->size().x()) + computedWidth) > static_cast<float>(this->mSize[0]))
            {
                // Draw on the left side
                bounds[0] = computedX - computedWidth - 10.f;
                p1.x() = computedX - 10.f;
                p2.x() = p1.x() + 7.f;
                p3.x() = p1.x();
            }
            else if
                (computedX + 0.5f * (static_cast<float>(widget->size().x()) - computedWidth) < 0.f)
            {
                // Draw on the right side
                bounds[0] = computedX + static_cast<float>(widget->size().x()) + 10.f;
                p1.x() = bounds[0];
                p2.x() = p1.x() - 7.f;
                p3.x() = p1.x();
            }
            else
            {
                // Draw centered
                bounds[0] = computedX + 0.5f * (static_cast<float>(widget->size().x()) - computedWidth);
                p1.x() = bounds[0] + 0.5f * computedWidth - 7.f;
                p2.x() = p1.x() + 7.f;
                p3.x() = p1.x() + 14.f;
                hCentered = true;
            }
            bounds[2] = bounds[0] + computedWidth;

            if
                (computedY + static_cast<float>(widget->size().y()) + computedHeight + 10.f > static_cast<float>(this->mSize[1]))
            {
                // Draw on the top side
                bounds[1] = computedY - computedHeight - 10.f;
                p1.y() = computedY - 10.f;
                p2.y() = p1.y() + 7.f;
                p3.y() = p1.y();
            }
            else if
                (false == hCentered && computedY + 0.5f * (static_cast<float>(widget->size().y()) - computedHeight) > 0.f)
            {
                // Draw vertically centered
                bounds[1] = computedY + 0.5f * (static_cast<float>(widget->size().y()) - computedHeight);
                p1.y() = bounds[1] + 0.5f * computedHeight - 7.f;
                p2.y() = p1.y() + 7.f;
                p3.y() = p1.y() + 14.f;
            }
            else
            {
                // Draw on the bottom side
                bounds[1] = computedY + static_cast<float>(widget->size().y()) + 10.f;
                p1.y() = bounds[1];
                p2.y() = p1.y() - 7.f;
                p3.y() = p1.y();
            }
            bounds[3] = bounds[1] + computedHeight;


            nvgRoundedRect(this->mNVGContext, bounds[0] - 2.f /* padding */, bounds[1] - 2.f /* padding */,
                           computedWidth + 4.f /* 2x padding */,
                           computedHeight + 4.f /* 2x padding */, 3);

            nvgMoveTo(this->mNVGContext, p1.x(), p1.y());
            nvgLineTo(this->mNVGContext, p2.x(), p2.y());
            nvgLineTo(this->mNVGContext, p3.x(), p3.y());

            nvgFill(this->mNVGContext);

            nvgFillColor(this->mNVGContext, nanogui::Color(30, 255));
            nvgFontBlur(this->mNVGContext, 0.0f);
            nvgTextBox(this->mNVGContext, bounds[0], bounds[1], computedWidth,
                       widget->tooltip().c_str(), nullptr);
        }
    }

    nvgEndFrame(this->mNVGContext);
}
