// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>

#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_rendering_control.h"
#include "ui/nanogui/nanogui_rendering_widget.h"
#include "ui/nanogui/nanogui_list.h"

using namespace SEAScope;

/**
 *
 */
NanoguiRenderingControl::NanoguiRenderingControl(nanogui::Widget* parent)
: nanogui::Widget(parent), reflowNeeded(true)
, scroller(nullptr), renderings(nullptr), orderCallback(nullptr)
, minCallback(nullptr), maxCallback(nullptr)
{
	this->setFixedSize(Eigen::Vector2i(335, 300));
	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
										   , nanogui::Alignment::Fill
										   , 0, 0));

    this->scroller = new NanoguiList(this);
	this->renderings = new NanoguiSortedList(this->scroller->content);
	this->renderings->widgetsSorter = NanoguiRenderingControl::compareRenderers;
	this->renderings->dragCallback = [this] (nanogui::Widget* target, nanogui::Widget* below, nanogui::Widget* above) -> bool {
		bool ok;	//
		IRenderingWidget* targetWidget;	//
		IRenderingWidget* belowWidget;	//
		IRenderingWidget* aboveWidget;	//

		targetWidget = dynamic_cast<IRenderingWidget*>(target);
		if
			(nullptr == targetWidget)
		{
			LOG( LogLevel::Error, "NanoguiRenderingControl"
					, "Drag target was either null or could not be cast to IRenderingWidget");
			return false;
		}

		belowWidget = dynamic_cast<IRenderingWidget*>(below);
		if
			(  nullptr == belowWidget
			&& nullptr != below)
		{
			LOG( LogLevel::Error, "NanoguiRenderingControl"
					, "Below widget could not be cast to IRenderingWidget");
			return false;
		}

		aboveWidget = dynamic_cast<IRenderingWidget*>(above);
		if
			(  nullptr == aboveWidget
			&& nullptr != above)
		{
			LOG( LogLevel::Error, "NanoguiRenderingControl"
					, "Above widget could not be cast to IRenderingWidget");
			return false;
		}

		LOG( LogLevel::Debug, "NanoguiRenderingControl"
				, "Order changed: {} must be placed between {} and {}"
				, dynamic_cast<NanoguiRenderingWidget*>(target)->label->caption()
				, (nullptr != below) ? dynamic_cast<NanoguiRenderingWidget*>(below)->label->caption(): "-"
				, (nullptr != above) ? dynamic_cast<NanoguiRenderingWidget*>(above)->label->caption(): "-"
				);

		ok = false;
		if
			(nullptr != this->orderCallback)
		{
			ok = this->orderCallback(targetWidget, belowWidget, aboveWidget);
		}
		return ok;
	};
};

/**
 *
 */
bool NanoguiRenderingControl::compareRenderers(const nanogui::Widget* a, const nanogui::Widget* b)
{
		const NanoguiRenderingWidget* wa;	//
		const NanoguiRenderingWidget* wb;	//

		wa = dynamic_cast<const NanoguiRenderingWidget*>(a);
		if
			(nullptr == wa)
		{
			LOG( LogLevel::Error, "NanoguiRenderingControl"
					, "widgetsSorter received a widget (a) that is not a NanoguiRenderingWidget");
			return false;
		}
		wb = dynamic_cast<const NanoguiRenderingWidget*>(b);
		if
			(nullptr == wb)
		{
			LOG( LogLevel::Error, "NanoguiRenderingControl"
					, "widgetsSorter received a widget (b) that is not a NanoguiRenderingWidget");
			return false;
		}
		return wa->zindex > wb->zindex;
};

/**
 *
 */
NanoguiRenderingControl::~NanoguiRenderingControl()
{
};

/**
 *
 */
void NanoguiRenderingControl::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

	if
		(true == this->reflowNeeded)
	{
		this->performLayout(ctx);
		this->reflowNeeded = false;
	}

	nvgBeginPath(ctx);
	nvgRect( ctx
		   , posX, posY
		   , sizeX, sizeY);
	nvgFillColor(ctx, nanogui::Color(0,0,0,50));
	nvgFill(ctx);
	nvgClosePath(ctx);

	nanogui::Widget::draw(ctx);
};

/**
 *
 */
IRenderingWidget* NanoguiRenderingControl::createRendering()
{
	NanoguiRenderingWidget* wItem;

	wItem = new NanoguiRenderingWidget(this->renderings);
	wItem->setOpacityCallback([this](IRenderingWidget* w, float opacity) -> bool {
		bool ok;	//

		ok = false;
		if
			(nullptr != this->opacityCallback)
		{
			ok = this->opacityCallback(w, opacity);
		}
		return ok;
	});
	wItem->setVisibilityCallback([this](IRenderingWidget* w, bool rendered, bool exclusive) -> bool {
		bool ok;	//

		ok = false;
		if
			(nullptr != this->visibilityCallback)
		{
			ok = this->visibilityCallback(w, rendered, exclusive);
		}
		return ok;
	});
	wItem->setMinCallback([this](IRenderingWidget* w, float rel) {
		if
			(nullptr != this->minCallback)
		{
			this->minCallback(w, rel);
		}
	});
	wItem->setMaxCallback([this](IRenderingWidget* w, float rel) {
		if
			(nullptr != this->maxCallback)
		{
			this->maxCallback(w, rel);
		}
	});
	return wItem;
};

/**
 *
 */
bool NanoguiRenderingControl::update()
{

	this->renderings->update();
	return true;
};

/**
 *
 */
bool NanoguiRenderingControl::addRendering(IRenderingWidget* rw)
{
	NanoguiRenderingWidget* w;	//

	LOG(LogLevel::Debug, "Nanogui Rendering control", "add rendering");
	w = dynamic_cast<NanoguiRenderingWidget*>(rw);
	if
		(nullptr == w)
	{
		LOG(LogLevel::Error, "NanoguiRenderingControl", "addRendering: provided widget should be a NanoguiRenderingWidget but is not");
		return false;
	}
	this->renderings->update();
    this->reflowNeeded = true;
	return true;
};

/**
 *
 */
bool NanoguiRenderingControl::removeRendering(IRenderingWidget* rw)
{
	NanoguiRenderingWidget* w;	//

	LOG(LogLevel::Debug, "Nanogui Rendering control", "remove rendering");
	w = dynamic_cast<NanoguiRenderingWidget*>(rw);
	if
		(nullptr == w)
	{
		LOG(LogLevel::Error, "NanoguiRenderingControl", "removeRendering: provided widget should be a NanoguiRenderingWidget but is not");
		return false;
	}
	this->renderings->removeChild(w);
	this->renderings->update();
    this->reflowNeeded = true;
	return true;
};

/**
 *
 */
bool NanoguiRenderingControl::setOrderCallback(const std::function<bool(IRenderingWidget*, IRenderingWidget*, IRenderingWidget*)>& callback)
{
	this->orderCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiRenderingControl::setOpacityCallback(const std::function<bool(IRenderingWidget*, float)>& callback)
{
	this->opacityCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiRenderingControl::setVisibilityCallback(const std::function<bool(IRenderingWidget*, bool, bool)>& callback)
{
	this->visibilityCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiRenderingControl::setMinCallback(const std::function<void(IRenderingWidget*, float)>& callback)
{
	this->minCallback = callback;
	return true;
}

/**
 *
 */
bool NanoguiRenderingControl::setMaxCallback(const std::function<void(IRenderingWidget*, float)>& callback)
{
	this->maxCallback = callback;
	return true;
};
