// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <iomanip>

#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_rendering_widget.h"

using namespace SEAScope;

/**
 *
 */
NanoguiRenderingWidget::NanoguiRenderingWidget(nanogui::Widget* parent)
: nanogui::Widget(parent), reflowNeeded(true), zindex(0.f), label(nullptr),
  minLabel(nullptr), maxLabel(nullptr), colormap(nullptr), logLabel(nullptr),
  unitsLabel(nullptr), countLabel(nullptr), opacityGauge(nullptr),
  opacityLabel(nullptr), visibilityCb(nullptr), settingsButton(nullptr),
  _leftSpacer(nullptr),
  opacityCallback(nullptr), minCallback(nullptr), maxCallback(nullptr),
  expanded(false)
{
	nanogui::Widget* palette;	//
	nanogui::Widget* nointeract;	// 
	nanogui::Widget* opacityBlock;	//
	nanogui::Widget* counterBlock;	//

	this->setFixedSize(Eigen::Vector2i(350, 30));
	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Horizontal
										   , nanogui::Alignment::Middle
										   , 3, 5));

	this->visibilityCb = new NanoguiCheckbox(this, "");
	this->visibilityCb->setFixedSize(Eigen::Vector2i(18, 18));
	this->visibilityCb->setCallback([this] (bool /*selected*/) {
		if
			(nullptr != this->visibilityCallback)
		{
            bool shiftPressed {false};
            shiftPressed = (/*GLFW_MOD_SHIFT*/ 1 == this->visibilityCb->keyPressed);
            if
                (true == shiftPressed)
            {
                this->visibilityCb->setChecked(true);
            }
			this->visibilityCallback(this, this->visibilityCb->checked(), shiftPressed);
		}
	});

	nointeract = new nanogui::Widget(this);
	nointeract->setLayout(new nanogui::BoxLayout( nanogui::Orientation::Vertical
												, nanogui::Alignment::Minimum
												, 0, 0));
	this->label = new NanoguiFixedSizeLabel(nointeract, "sans-bold", 14);
	this->label->setFixedSize(Eigen::Vector2i(225, 14));

	palette = new nanogui::Widget(nointeract);
	palette->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Horizontal
											  , nanogui::Alignment::Minimum
											  , 2, 2));
    this->_leftSpacer = new nanogui::Widget(palette);
    this->_leftSpacer->setFixedSize(Eigen::Vector2i(30, 12));
    this->_leftSpacer->setVisible(false);

	this->minLabel = new NanoguiNumLabel(palette);
	this->minLabel->setFontSize(12);
	this->minLabel->setFixedSize(Eigen::Vector2i(30,12));
	this->minLabel->numCallback = [this](const Eigen::Vector2f& rel) {
		LOG(LogLevel::Debug, "Rendering widget", "scrolling on min value");
		if
			(nullptr != this->minCallback)
		{
			this->minCallback(this, rel.y());
		}
	};

	this->colormap = new NanoguiHColorbar(palette);
	this->colormap->setFixedSize(Eigen::Vector2i(110,12));

	this->maxLabel = new NanoguiNumLabel(palette);
	this->maxLabel->setFontSize(12);
	this->maxLabel->setFixedSize(Eigen::Vector2i(30,12));
	this->maxLabel->numCallback = [this](const Eigen::Vector2f& rel) {
		LOG(LogLevel::Debug, "Rendering widget", "scrolling on max value");
		if
			(nullptr != this->maxCallback)
		{
			this->maxCallback(this, rel.y());
		}
	};

    this->unitsLabel = new nanogui::Label(palette, "", "sans-bold");
    this->unitsLabel->setFontSize(12);
    this->unitsLabel->setFixedSize(Eigen::Vector2i(35, 12));
    this->unitsLabel->setTooltip("Units");

	counterBlock = new nanogui::Widget(this);
	counterBlock->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
												   , nanogui::Alignment::Middle
												   , 0, 0));

	this->countLabel = new nanogui::Label(counterBlock, "", "sans-bold");
	this->countLabel->setFontSize(14);
	this->countLabel->setFixedSize(Eigen::Vector2i(30,14));
	this->countLabel->setTooltip("0 granule(s) found with current search parameters");

	this->logLabel = new nanogui::Label(counterBlock, "", "sans-bold");
	this->logLabel->setFontSize(12);
	this->logLabel->setFixedSize(Eigen::Vector2i(30, 12));
	this->logLabel->setColor(nanogui::Color(255, 230, 20, 255));

	opacityBlock = new nanogui::Widget(this);
	opacityBlock->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
												   , nanogui::Alignment::Middle
												   , 0, 0));
	this->opacityGauge = new NanoguiGauge(opacityBlock);
	this->opacityGauge->setFixedSize(Eigen::Vector2i(30, 16));
	this->opacityLabel = new nanogui::Label(opacityBlock, "");
	this->opacityLabel->setFixedSize(Eigen::Vector2i(34, 12));
	this->opacityLabel->setFontSize(12);
	this->opacityGauge->setChangeCallback([this] (float percent) -> bool {
		bool ok;	//
		std::ostringstream buffer;	//

		buffer << std::fixed << std::setprecision(2) <<  100 * percent << "%";
		this->opacityLabel->setCaption(buffer.str());

		ok = false;
		if
			(nullptr != this->opacityCallback)
		{
			ok = this->opacityCallback(this, percent);
		}
		return ok;
	});

	this->settingsButton = new nanogui::Button(this, "", ENTYPO_ICON_COG);
	this->settingsButton->setFlags(nanogui::Button::ToggleButton);
	this->settingsButton->setFixedSize(Eigen::Vector2i(20, 30));
	this->settingsButton->setChangeCallback([this] (bool active) {
        this->expanded = active;
		LOG(LogLevel::Debug, "NanoguiRenderingWidget", "Toggle settings");
		if
			(nullptr != this->settingsCallback)
		{
			this->settingsCallback(this);
		}
	});

};

/**
 *
 */
NanoguiRenderingWidget::~NanoguiRenderingWidget()
{
};

/**
 *
 */
bool NanoguiRenderingWidget::update(RenderingCfg* cfg)
{
	bool ok;	//
	ok = this->update(this->label->caption(), this->unitsLabel->caption(), cfg);
	return ok;
};

/**
 *
 */
bool NanoguiRenderingWidget::update(const std::string& caption, const std::string& units, RenderingCfg* cfg)
{
	std::ostringstream buffer;	//
	Colormap cmap;	//
	bool ok;	//

	this->label->setCaption(caption);
    this->label->setTooltip(caption);

    this->unitsLabel->setCaption(units);
    buffer.str("");
    buffer << "Units: " << units;
    this->unitsLabel->setTooltip(buffer.str());

    buffer.str("");
	buffer << std::fixed << std::setprecision(2) << 100 * cfg->opacity() << "%";
	this->opacityGauge->setPercent(cfg->opacity());
	this->opacityLabel->setCaption(buffer.str());
    buffer.str("");
    buffer << "Opacity: " << this->opacityLabel->caption();
    this->opacityLabel->setTooltip(buffer.str());

	buffer.str("");
	buffer << std::fixed << std::setprecision(2) << cfg->min();
	this->minLabel->setCaption(buffer.str());
	buffer.str("");
	buffer << "Min value: " << std::fixed << std::setprecision(2) << cfg->min();
	this->minLabel->setTooltip(buffer.str());

	if
		(nullptr != this->cmapGetter)
	{
		ok = this->cmapGetter(cfg->colormap(), cmap);
		if
			(true == ok)
		{
			this->colormap->setColormap(cmap);
		}
		else
		{
			this->colormap->setColor(cfg->color());
		}
        this->minLabel->setVisible(ok);
        this->maxLabel->setVisible(ok);
        this->unitsLabel->setVisible(ok);
        this->logLabel->setVisible(ok);
        this->_leftSpacer->setVisible(!ok);
	}

	if
		(true == cfg->logscale())
	{
		this->logLabel->setCaption("LOG");
	}
	else
	{
		this->logLabel->setCaption("");
	}

	buffer.str("");
	buffer << std::fixed << std::setprecision(2) << cfg->max();
	this->maxLabel->setCaption(buffer.str());
	buffer.str("");
	buffer << "Max value: " << std::fixed << std::setprecision(2) << cfg->max();
	this->maxLabel->setTooltip(buffer.str());

	this->visibilityCb->setChecked(cfg->rendered());
	LOG(LogLevel::Debug, "NanoguiRenderingWidget", "Update visibility: {}", this->visibilityCb->checked());

	this->zindex = cfg->zindex();

    this->reflowNeeded = true;

	return false;
};

/**
 *
 */
bool NanoguiRenderingWidget::setCounter(uint32_t count)
{
	std::ostringstream buffer("");	//

    if
        (999 < count)
    {
        this->countLabel->setCaption("# ...");
    }
    else
    {
	    buffer << "# " << count;
	    this->countLabel->setCaption(buffer.str());
    }

    buffer.str("");
    buffer << count << " granule(s) found with current search parameters";
	this->countLabel->setTooltip(buffer.str());
	return true;
};

/**
 *
 */
void NanoguiRenderingWidget::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

	if
		(true == this->reflowNeeded)
	{
		this->performLayout(ctx);
		this->reflowNeeded = false;
	}

	nvgBeginPath(ctx);
	nvgRoundedRect(ctx, posX + 1.0f, posY + 1.0f, sizeX - 2.0f, sizeY - 2.0f, 3);
    if
        (this->expanded)
    {
	    nvgFillColor(ctx, nanogui::Color(50, 150, 250, 100));
    }
    else
    {
	    nvgFillColor(ctx, nanogui::Color(50, 50, 50, 200));
    }
	nvgFill(ctx);

	nanogui::Widget::draw(ctx);
};

/**
 *
 */
bool NanoguiRenderingWidget::setOpacityCallback(const std::function<bool(IRenderingWidget*, float)>& callback)
{
	this->opacityCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiRenderingWidget::setVisibilityCallback(const std::function<void(IRenderingWidget*, bool, bool)>& callback)
{
	this->visibilityCallback = callback;
	return true;
}

/**
 *
 */
bool NanoguiRenderingWidget::setColormapGetter(const std::function<bool(const std::string&, Colormap&)>& cmapGetterIn)
{
	this->cmapGetter = cmapGetterIn;
	return true;
};

/**
 *
 */
bool NanoguiRenderingWidget::setSettingsCallback(const std::function<bool(IRenderingWidget*)>& callback)
{
	this->settingsCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiRenderingWidget::setSettingsState(bool pushed)
{
	LOG(LogLevel::Debug, "NanoguiRenderingWidget", "setSettingsState: {} [{}]", pushed, this->label->caption());
	this->settingsButton->setPushed(pushed);
    this->expanded = pushed;
	return true;
};

/**
 *
 */
bool NanoguiRenderingWidget::setMinCallback(const std::function<void(IRenderingWidget*, float)>& callback)
{
	this->minCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiRenderingWidget::setMaxCallback(const std::function<void(IRenderingWidget*, float)>& callback)
{
	this->maxCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiRenderingWidget::setOpacityScrollStep(float value)
{
	this->opacityGauge->scrollStep = value;
	return true;
};
