// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "core/log.h"
#include "core/utils.h"
#include "ui/nanogui/nanogui_resize_handle.h"
#include "ui/nanogui/nanogui_blueprint.h"

using namespace SEAScope;

/**
 *
 */
NanoguiResizeHandle::NanoguiResizeHandle(nanogui::Widget* parent): nanogui::Widget(parent)
{
	this->setSize(Eigen::Vector2i(20,20));
};

/**
 *
 */
NanoguiResizeHandle::~NanoguiResizeHandle()
{

};

/**
 *
 */
bool NanoguiResizeHandle::updatePosition()
{
	if
		(nullptr == this->target)
	{
		return false;
	}
	int x = this->target->position().x() + this->target->width() - this->mSize.x();
	int y = this->target->position().y() + this->target->height() - this->mSize.y();
	this->setPosition(Eigen::Vector2i(x,y));
	return true;
};

/**
 *
 */
Eigen::Vector2i NanoguiResizeHandle::preferredSize(NVGcontext* /*ctx*/) const
{
	return Eigen::Vector2i(20,20);
};

/**
 *
 */
bool NanoguiResizeHandle::setTarget(NanoguiBlueprint* targetIn, const std::string& minWidthIn, const std::string& minHeightIn)
{
	Eigen::Vector2i pos;

	if
		(targetIn == this->target)
	{
		return false;
	}

	this->target = targetIn;

	if
		(nullptr == targetIn)
	{
		return true;
	}

	LOG(LogLevel::Debug, "Resize handle", "setTarget {} {}", this->parent()->size().x(), this->parent()->size().y());
	this->minWidth = minWidthIn;
	this->minHeight = minHeightIn;
	this->updatePosition();
	return true;
};

/**
 *
 */
bool NanoguiResizeHandle::mouseDragEvent(const Eigen::Vector2i &p, const Eigen::Vector2i &rel, int button, int /*modifiers*/)
{
	Eigen::Vector2i limits;	//

    if
        (  0 != (button & (1 << GLFW_MOUSE_BUTTON_1))
		&& nullptr != this->target)
    {
		limits = this->getLimits();

		if
			(  p.x() - this->target->position().x() >= limits.x()
			&& p.x() + this->mSize.x() <= this->parent()->size().x())
		{
        	this->mPos += Eigen::Vector2i(rel.x(), 0);
		}

		if
			(  p.y() - this->target->position().y() >= limits.y()
			&& p.y() + this->mSize.y() <= this->parent()->size().y())
		{
			this->mPos += Eigen::Vector2i(0, rel.y());
		}

		Eigen::Vector2i newDims( this->mPos.x() - this->target->position().x() + this->mSize.x()
							   , this->mPos.y() - this->target->position().y() + this->mSize.y());
		newDims = newDims.cwiseMin(this->parent()->size() - this->target->position());
		this->target->updateSize(newDims);
        return true;
    }
	return true;
};

/**
 *
 */
bool NanoguiResizeHandle::mouseButtonEvent(const Eigen::Vector2i &p, int button, bool down, int modifiers) {
    if (Widget::mouseButtonEvent(p, button, down, modifiers))
	{
        return true;
	}
    return false;
};


/**
 *
 */
void NanoguiResizeHandle::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeY = static_cast<float>(this->mSize.y());

	nvgBeginPath(ctx);
	nvgRoundedRect(ctx, posX + 1.0f, posY + 1.0f, sizeY - 2.0f,
                   sizeY - 2.0f, 0);
	nvgFillColor(ctx, nanogui::Color(150, this->mEnabled ? 255 : 100));
	nvgFill(ctx);
};

/**
 *
 */
Eigen::Vector2i NanoguiResizeHandle::getLimits() const
{
	int w;	//
	int h;	//

	w = dimStr2pixels(this->minWidth, this->parent()->size().x());
	h = dimStr2pixels(this->minHeight, this->parent()->size().y());
	return Eigen::Vector2i(w,h);
};
