// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cmath>

#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_scrollable_with_title.h"
#include "core/log.h"
#include "ui/nanogui/nanogui_renderer.h"

using namespace SEAScope;

/**
 *
 */
NanoguiScrollableWithTitle::NanoguiScrollableWithTitle(nanogui::Widget* parent)
: nanogui::Widget(parent), titlebar(nullptr), header(nullptr),
  scrollable(nullptr), footer(nullptr), title(nullptr), reflowNeeded(false)
{
  nanogui::Widget* spacer;

  this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                         nanogui::Alignment::Minimum,
                                         0, 0));

  this->titlebar = new nanogui::Widget(this);
  this->titlebar->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                 nanogui::Alignment::Middle,
                                                 0, 0));
  this->titlebar->setFixedHeight(30);

  spacer = new nanogui::Widget(this);
  spacer->setFixedHeight(3);

  this->title = new nanogui::Label(this->titlebar, "", "sans-bold");
  this->title->setFontSize(20);
  this->title->setFixedHeight(30);

  this->header = new nanogui::Widget(this);
  this->scrollable = new NanoguiList(this);
  this->footer = new nanogui::Widget(this);
}

/**
 *
 */
NanoguiScrollableWithTitle::~NanoguiScrollableWithTitle()
{
  this->titlebar = nullptr;
  this->title = nullptr;
  this->header = nullptr;
  this->scrollable = nullptr;
  this->footer = nullptr;
}

/**
 *
 */
Eigen::Vector2i NanoguiScrollableWithTitle::preferredSize(NVGcontext *ctx) const
{
    Eigen::Vector2i result; //
    Eigen::Vector2i scrollableSize; //

    scrollableSize = this->scrollable->content->preferredSize(ctx);

    result.x() = 0;
    result.x() = std::max(result.x(), this->titlebar->preferredSize(ctx).x());
    result.x() = std::max(result.x(), this->header->preferredSize(ctx).x());
    result.x() = std::max(result.x(), scrollableSize.x());
    result.x() = std::max(result.x(), this->footer->preferredSize(ctx).x());

    result.y() = 3 + this->titlebar->fixedHeight() + this->header->fixedHeight() + scrollableSize.y() + this->footer->fixedHeight();

    return result;
}

/**
 *
 */
void NanoguiScrollableWithTitle::setTitle(const std::string& value)
{
  this->title->setCaption(value);
}

/**
 *
 */
void NanoguiScrollableWithTitle::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //
    float titleHeight;  //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());
    titleHeight = static_cast<float>(this->titlebar->fixedHeight());

    if
        (true == this->reflowNeeded)
    {
        this->performLayout(ctx);
        this->reflowNeeded = false;
    }

    nvgBeginPath(ctx);
	nvgRect(ctx
           , posX, posY
		   , sizeX, sizeY);
	nvgFillColor(ctx, nanogui::Color(0, 0, 0, 170));
    nvgFill(ctx);

    nvgBeginPath(ctx);
    nvgMoveTo(ctx,
              posX + std::floor(0.1f * sizeX),
              posY + titleHeight + 1.0f);
    nvgLineTo(ctx,
              posX + std::floor(0.9f * sizeX),
              posY + titleHeight + 1.0f);
    nvgStrokeColor(ctx, nanogui::Color(255, 255, 255, 255));
    nvgStroke(ctx);

    nvgBeginPath(ctx);
    nvgMoveTo(ctx,
              posX + std::floor(0.1f * sizeX),
              posY + titleHeight + 2.0f);
    nvgLineTo(ctx,
              posX + std::floor(0.9f * sizeX),
              posY + titleHeight + 2.0f);
    nvgStrokeColor(ctx, nanogui::Color(0, 0, 0, 10));
    nvgStroke(ctx);

    nanogui::Widget::draw(ctx);
}


/**
 *
 */
void NanoguiScrollableWithTitle::resize(int width, int height)
{
  int scrollableHeight; //

  if
    (width != this->fixedWidth())
  {
    this->titlebar->setFixedWidth(width);
    this->titlebar->setWidth(width);
    this->header->setFixedWidth(width);
    this->header->setWidth(width);
    this->scrollable->setFixedWidth(width);
    this->scrollable->setWidth(width);
    this->footer->setFixedWidth(width);
    this->footer->setWidth(width);

    int margins {8}; /* 8px = scrollbar width */
    nanogui::BoxLayout* boxLayout {nullptr};
    boxLayout = dynamic_cast<nanogui::BoxLayout*>(this->scrollable->content->layout());
    if
        (  (nullptr != boxLayout)
        && (nanogui::Orientation::Horizontal == boxLayout->orientation()))
    {
        if
            (boxLayout->margin() > margins)
        {
            // layout margins already leave enough space for the scrollbar
            margins = 2 * boxLayout->margin();
        }
        else
        {
            // layout margins are too thin for the scrollbar, use the layout
            // margin for left side and the scrollbar size for right side
            margins = boxLayout->margin() + margins;
        }
    }
    boxLayout = nullptr;

    for
      (auto& child : this->scrollable->content->children())
    {
      child->setFixedWidth(width - margins);
      child->setWidth(width - margins);
    }
  }

  scrollableHeight = height - 3 /*spacer*/;
  scrollableHeight = scrollableHeight - this->titlebar->fixedHeight();
  scrollableHeight = scrollableHeight - this->header->fixedHeight();
  scrollableHeight = scrollableHeight - this->footer->fixedHeight();

  if
    (0 >= scrollableHeight)
  {
    this->scrollable->setVisible(false);
  }
  else
  {
    this->scrollable->setFixedHeight(scrollableHeight);
	this->scrollable->setHeight(scrollableHeight);
    this->scrollable->content->update();
    this->scrollable->setVisible(true);
  }

  this->setFixedSize(Eigen::Vector2i(width, height));
  this->setSize(Eigen::Vector2i(width, height));
  this->reflowNeeded = true;
}

void NanoguiScrollableWithTitle::center()
{
    Widget* widget = this;
    while 
        (nullptr != widget->parent())
    {
        widget = widget->parent();
    }

    NanoguiRenderer* screen = dynamic_cast<NanoguiRenderer*>(widget);
    
    if
        (nullptr != screen)
    {
        //screen->centerWidget(this);
    }
}
