// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>

#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_sorted_list.h"

using namespace SEAScope;

/**
 *
 */
NanoguiSortedList::NanoguiSortedList(nanogui::Widget* parent)
: nanogui::Widget(parent),
  reflowNeeded(true),
  dragCallback(nullptr),
  widgetsSorter(nullptr),
  draggedChild(nullptr),
  hoveredChild(nullptr),
  draggedIndex(-1),
  hoveredIndex(-1),
  _backgroundFillColor(Eigen::Vector4i(0, 0, 0, 0)),
  _backgroundStrokeColor(Eigen::Vector4i(0, 0, 0, 0))
{
	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
										   , nanogui::Alignment::Middle
										   , 0, 0));
};

/**
 *
 */
NanoguiSortedList::~NanoguiSortedList()
{
    this->dragCallback = nullptr;
    this->widgetsSorter = nullptr;
    this->draggedChild = nullptr;
    this->hoveredChild = nullptr;
};

/**
 *
 */
void NanoguiSortedList::draw(NVGcontext* ctx)
{
	int relY;	//
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //
    float draggedChildPosX;    //
    float draggedChildPosY;    //
    float draggedChildSizeX;    //
    float draggedChildSizeY;    //
    float draggedOffsetX;   //
    float draggedOffsetY;   //
    float hoveredChildPosY; //
    float hoveredChildSizeY;    //
    float cursorPosX;  //
    float cursorPosY;  //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

	if
		(true == this->reflowNeeded)
	{
		this->performLayout(ctx);
		this->reflowNeeded = false;
	}

    // Draw background
    nvgBeginPath(ctx);
    nvgRect(ctx, posX, posY, sizeX, sizeY);
    nvgFillColor(ctx, this->_backgroundFillColor);
    nvgFill(ctx);
    nvgStrokeColor(ctx, this->_backgroundStrokeColor);
    nvgStroke(ctx);

	nanogui::Widget::draw(ctx);


	if
		(nullptr != this->draggedChild)
	{
        draggedChildPosX = static_cast<float>(this->draggedChild->position().x());
        draggedChildPosY = static_cast<float>(this->draggedChild->position().y());
        draggedChildSizeX = static_cast<float>(this->draggedChild->width());
        draggedChildSizeY = static_cast<float>(this->draggedChild->height());
        draggedOffsetX = static_cast<float>(this->draggedOffset.x());
        draggedOffsetY = static_cast<float>(this->draggedOffset.y());
        cursorPosX = static_cast<float>(this->cursorPosition.x());
        cursorPosY = static_cast<float>(this->cursorPosition.y());

		/* Dark layer over the dragged item's original location */
		nvgBeginPath(ctx);
		nvgRect( ctx
			   , posX + draggedChildPosX
			   , posY + draggedChildPosY
			   , draggedChildSizeX
			   , draggedChildSizeY);
		nvgFillColor(ctx, nanogui::Color(50, 50, 50, 100));
		nvgFill(ctx);

		/* Dragged item shape following the cursor */
		nvgBeginPath(ctx);
		nvgRect( ctx
			   , posX + cursorPosX + draggedOffsetX
			   , posY + cursorPosY + draggedOffsetY
			   , draggedChildSizeX
			   , draggedChildSizeY);
		nvgFillColor(ctx, nanogui::Color(50, 50, 50, 100));
		nvgFill(ctx);

	}

	if
		(  nullptr != this->draggedChild
		&& nullptr != this->hoveredChild)
	{
        hoveredChildPosY = static_cast<float>(this->hoveredChild->position().y());
        hoveredChildSizeY = static_cast<float>(this->hoveredChild->height());

		relY = this->cursorPosition.y() - this->hoveredChild->position().y();

		if
			( (  0.5 * this->hoveredChild->height() < relY
			  && -1 != this->hoveredIndex - this->draggedIndex)
			|| -1 == this->draggedIndex - this->hoveredIndex)
		{
			/* Draw line below hovered child:
				- if the cursor is in the inferior half of the hovered widget
				  but not if the dragged item is already just under the hovered one
                - if the hovered widget is just below the dragged one
			*/
			nvgBeginPath(ctx);
			nvgMoveTo( ctx
					 , posX
					 , posY + hoveredChildPosY + hoveredChildSizeY);
			nvgLineTo( ctx
					 , posX + sizeX
					 , posY + hoveredChildPosY + hoveredChildSizeY);
			nvgStrokeColor(ctx, nanogui::Color(255,50,30,255));
			nvgStroke(ctx);
		}
		else if
			( (  0.5 * this->hoveredChild->height() >= relY
			  && -1 != this->draggedIndex - this->hoveredIndex)
			|| -1 == this->hoveredIndex - this->draggedIndex)
		{
			/* Draw line on top of hovered child:
				- if the cursor is in the superior half of the hovered widget
                  but not if the dragged item is already over the hovered one
				- if the hovered widget is just above the dragged one
			*/
			nvgBeginPath(ctx);
			nvgMoveTo( ctx
					 , posX
					 , posY + hoveredChildPosY);
			nvgLineTo( ctx
					 , posX + sizeX
					 , posY + hoveredChildPosY);
			nvgStrokeColor(ctx, nanogui::Color(255,50,30,255));
			nvgStroke(ctx);
		}
	}
};

/**
 *
 */
bool NanoguiSortedList::update()
{
	this->sortWidgets();
	LOG(LogLevel::Debug, "NanoguiSortedList", "Number of children: {}", this->children().size());
	this->reflowNeeded = true;
	return true;
};

/**
 *
 */
bool NanoguiSortedList::sortWidgets()
{
	if
		(nullptr == this->widgetsSorter)
	{
		return false;
	}

	std::sort( this->mChildren.begin()
			 , this->mChildren.end()
			 , this->widgetsSorter);
	return true;
};

/**
 *
 */
std::tuple<nanogui::Widget*, int> NanoguiSortedList::childAt(const Eigen::Vector2i& pos) const
{
	Eigen::Vector2i rel;	//
	std::vector<nanogui::Widget*>::const_iterator it;	//
	nanogui::Widget* child;	//
	int childIndex;	//

	rel.x() = pos.x() - this->mPos.x();
	rel.y() = pos.y() - this->mPos.y();

	child = nullptr;
	childIndex = -1;
	/* find child that contains the cursor */
	it = this->children().begin();
	for
		(; it != this->children().end(); ++it)
	{
		if
			((*it)->contains(rel))
		{
			child = *it;
            auto relIndex = it - this->children().begin();
			childIndex = static_cast<int>(relIndex);
		}
	}

	return std::make_tuple(child, childIndex);
};

/**
 *
 */
bool NanoguiSortedList::mouseDownEvent(const Eigen::Vector2i& pos, int button, int mods)
{
	bool ok;	//
	nanogui::Widget* child;	//
	int childIndex;	//

	if
		(nullptr != this->draggedChild)
	{
		// Mouse must have gone out of the list widget while dragging
	}
	else
	{
		// Let children grab the event
		ok = nanogui::Widget::mouseButtonEvent(pos, button, true, mods);

		if
			(  false == ok
			&& 1 < this->children().size()
			&& nullptr != this->dragCallback)
		{
			auto found = this->childAt(pos);
			std::tie(child, childIndex) = found;
			if
				(nullptr != child)
			{
				LOG( LogLevel::Debug, "NanoguiSortedList", "Mouse down on child: {}", childIndex);
				this->draggedChild = child;
				this->draggedIndex = childIndex;
				this->cursorPosition = pos - this->position();
				this->draggedOffset = child->position() - this->cursorPosition;
			}
			else
			{
				LOG( LogLevel::Debug, "NanoguiSortedList", "Mouse down... somewhere");
			}
		}
	}

	return true;
};

/**
 *
 */
bool NanoguiSortedList::mouseUpEvent(const Eigen::Vector2i& pos, int button, int mods)
{
	bool ok;	//
	int relY;	//
	nanogui::Widget* neighbour;	//
    std::size_t hoveredIndexTmp;    //

	neighbour = nullptr;
	if
		(nullptr != this->draggedChild)
	{
		if
			(nullptr == this->hoveredChild)
		{
			/* Being there means that the mouse is over the child that is being dragged => do nothing */
		}
		else
		{
			/*  
			 Note that non-null draggedChild can only happen if dragCallback is 
			 also non-null so it is not necessary to test it
			*/

            hoveredIndexTmp = static_cast<std::size_t>(this->hoveredIndex);
			relY = this->cursorPosition.y() - this->hoveredChild->position().y();
			if
				( (  0.5 * this->hoveredChild->height() < relY
				  && -1 != this->hoveredIndex - this->draggedIndex)
				|| -1 == this->draggedIndex - this->hoveredIndex)
			{
				// h > d > h+1
				if
					(this->hoveredIndex < this->childCount() - 1)
				{
					/* there is at least one child below the hovered one */
					neighbour = this->children()[hoveredIndexTmp + 1];
				}
				this->dragCallback( this->draggedChild /* target */
								  , neighbour /* below */
								  , this->hoveredChild /* above */);
			}
			else if
				( (  0.5 * this->hoveredChild->height() >= relY
				  && -1 != this->draggedIndex - this->hoveredIndex)
				|| -1 == this->hoveredIndex - this->draggedIndex)
			{
				// h-1 > d > h
				if
					(1 <= this->hoveredIndex)
				{
					/* there is at least one child above the hovered one */
					neighbour = this->children()[hoveredIndexTmp - 1];
				}
				this->dragCallback( this->draggedChild /* target */
								  , this->hoveredChild /* below */
								  , neighbour /* above */);
			}
			else
			{
				LOG( LogLevel::Debug, "NanoguiSortedList"
						, "Undexpected drag end!!");
			}
		}

		/* No more dragging */
		this->draggedChild = nullptr;
		this->draggedIndex = -1;
		this->hoveredChild = nullptr;
		this->hoveredIndex = -1;
		ok = true;
	}
	else
	{
		// Let children grab the event
		ok = nanogui::Widget::mouseButtonEvent(pos, button, false, mods);
	}

	return ok;
};

/**
 *
 */
bool NanoguiSortedList::mouseButtonEvent(const Eigen::Vector2i& pos, int button, bool pressed, int mods)
{
	bool ok;	//

	if
		(pressed)
	{
		ok = this->mouseDownEvent(pos, button, mods);
	}
	else
	{
		ok = this->mouseUpEvent(pos, button, mods);
	}
	return ok;
};

/**
 *
 */
bool NanoguiSortedList::mouseEnterEvent(const Eigen::Vector2i& pos, bool enter)
{
	bool ok;	//

	if
		(  false == enter
		&& nullptr != this->draggedChild)
	{
		// Cancel dragging
		this->draggedChild = nullptr;
		this->draggedIndex = -1;
		this->hoveredChild = nullptr;
		this->hoveredIndex = -1;
	}

	ok = nanogui::Widget::mouseEnterEvent(pos, enter);
	return ok;
};

/**
 *
 */
bool NanoguiSortedList::mouseMotionEvent(const Eigen::Vector2i& pos, const Eigen::Vector2i& rel, int button, int mods)
{
	bool ok;	//
	nanogui::Widget* hovered;	//
	int hoveredIndexTmp;	//

	if
		(nullptr != this->draggedChild)
	{
		this->cursorPosition = pos - this->position();
		auto found = this->childAt(pos);
		std::tie(hovered, hoveredIndexTmp) = found;

		if
			(hovered != this->draggedChild)
		{
			this->hoveredChild = hovered;
			this->hoveredIndex = hoveredIndexTmp;
		}
		else
		{
			this->hoveredChild = nullptr;
			this->hoveredIndex = -1;
		}
	}

	ok = nanogui::Widget::mouseMotionEvent(pos, rel, button, mods);
	return ok;
};

/**
 *
 */
void NanoguiSortedList::setBackgroundFillColor(nanogui::Color color)
{
    this->_backgroundFillColor = color;
}

/**
 *
 */
void NanoguiSortedList::setBackgroundStrokeColor(nanogui::Color color)
{
    this->_backgroundStrokeColor = color;
}
