// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>

#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_sources_control.h"

using namespace SEAScope;

/**
 *
 */
NanoguiSourcesControl::NanoguiSourcesControl(nanogui::Widget* parent): nanogui::Widget(parent), reflowNeeded(true), uriField(nullptr), addButton(nullptr), list(nullptr), addCallback(nullptr), infoCallback(nullptr), selectionCallback(nullptr) 
{
	nanogui::Widget* form;	//

	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
										   , nanogui::Alignment::Middle
										   , 2, 5));

	form = new nanogui::Widget(this);
	form->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Horizontal
										   , nanogui::Alignment::Middle
										   , 0, 0));
	this->uriField = new nanogui::TextBox(form, "");
	this->uriField->setFixedSize(Eigen::Vector2i(this->mSize.x() - 50, 18));
	this->uriField->setFontSize(14);
	this->uriField->setEditable(true);

	this->addButton = new nanogui::Button(form, "Add", ENTYPO_ICON_PLUS);
	this->addButton->setFontSize(14);
	this->addButton->setFixedSize(Eigen::Vector2i(50, 18));
	this->addButton->setCallback([this]() {
		if
			(nullptr != this->addCallback)
		{
			this->addCallback(this->uriField->value());
		}
	});

	this->list = new NanoguiList(this);
	this->list->content->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
														  , nanogui::Alignment::Middle
														  , 0, 2));
	this->list->content->widgetsSorter = NanoguiSourcesControl::compareSources;
};

/**
 *
 */
NanoguiSourcesControl::~NanoguiSourcesControl()
{
};

/**
 *
 */
ISourceWidget* NanoguiSourcesControl::createSource()
{
	NanoguiSourceWidget* w;	//

	w = new NanoguiSourceWidget(this->list->content);
	w->setInfoCallback([this] (ISourceWidget* target) {
		NanoguiSourceWidget* wtarget;	//

		if
			(nullptr != this->infoCallback)
		{
			wtarget = dynamic_cast<NanoguiSourceWidget*>(target);
			if
				(nullptr == wtarget)
			{
				LOG( LogLevel::Error, "NanoguiSourcesControl"
						, "Info callback received a widget that is not a NanoguiSourceWidget");
				return;
			}
			this->infoCallback(wtarget->uriLabel->caption());
		}
	});

	w->setSelectionCallback([this] (ISourceWidget* target, bool selected) {
		NanoguiSourceWidget* wtarget;	//

		if
			(nullptr != this->selectionCallback)
		{
			wtarget = dynamic_cast<NanoguiSourceWidget*>(target);
			if
				(nullptr == wtarget)
			{
				LOG( LogLevel::Error, "NanoguiSourcesControl"
						, "Selection callback received a widget that is not a NanoguiSourceWidget");
				return;
			}
			this->selectionCallback(wtarget->uriLabel->caption(), selected);
		}

		this->list->content->update();
	});
	w->setVisible(false);

	return w;
};

/**
 *
 */
bool NanoguiSourcesControl::addSource(ISourceWidget* src)
{
	NanoguiSourceWidget* w;	//

	w = dynamic_cast<NanoguiSourceWidget*>(src);
	if
		(nullptr == w)
	{
		LOG( LogLevel::Error, "NanoguiSourcesControl"
				, "addSource: provided widget should be a NanoguiSourceWidget but is not");
		return false;
	}

	w->setVisible(true);
	this->list->content->update();
	return true;
};

/**
 *
 */
bool NanoguiSourcesControl::removeSource(ISourceWidget* src)
{
	NanoguiSourceWidget* w;	//

	w = dynamic_cast<NanoguiSourceWidget*>(src);
	if
		(nullptr == w)
	{
		LOG( LogLevel::Error, "NanoguiSourcesControl"
				, "removeSource: provided widget should be a NanoguiSourceWidget but is not");
		return false;
	}

	this->list->content->removeChild(w);
	return true;
};

/**
 *
 */
bool NanoguiSourcesControl::setAddCallback(const std::function<void(const std::string&)>& callback)
{
	this->addCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiSourcesControl::setInfoCallback(const std::function<void(const std::string&)>& callback)
{
	this->infoCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiSourcesControl::setSelectionCallback(const std::function<void(const std::string&, bool)>& callback)
{
	this->selectionCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiSourcesControl::compareSources(const nanogui::Widget* a, const nanogui::Widget* b)
{
	const NanoguiSourceWidget* wa;	//
	const NanoguiSourceWidget* wb;	//
	std::string titleA;	//
	std::string titleB;	//
	bool aSelected;	//
	bool bSelected;	//

	wa = dynamic_cast<const NanoguiSourceWidget*>(a);
	if
		(nullptr == wa)
	{
		LOG( LogLevel::Error, "NanoguiSourcesControl"
				, "Widgets sorter received a widget (a) that is not a NanoguiSourceWidget");
		return false;
	}
	wb = dynamic_cast<const NanoguiSourceWidget*>(b);
	if
		(nullptr == wb)
	{
		LOG( LogLevel::Error, "NanoguiSourcesControl"
				, "Widgets sorter received a widget (b) that is not a NanoguiSourceWidget");
		return false;
	}

	/* Selected widgets first */
	aSelected = wa->selectionCb->checked();
	bSelected = wb->selectionCb->checked();

	if
		(aSelected == bSelected)
	{
		/* Use lexicographical order */
		titleA = wa->titleLabel->caption();
		titleB = wb->titleLabel->caption();
		return std::lexicographical_compare( titleA.begin(), titleA.end()
										   , titleB.begin(), titleB.end());
	}

	LOG( LogLevel::Debug, "NanoguiSourcesControl"
			, "compare widgets: {}({}) {}({})"
			, wa->titleLabel->caption(), aSelected
			, wb->titleLabel->caption(), bSelected);

	return aSelected;
};

/**
 *
 */
void NanoguiSourcesControl::setFixedSize(const Eigen::Vector2i& size)
{
	nanogui::Widget::setFixedSize(size);
	LOG( LogLevel::Debug, "NanoguiSourcesControl"
			, "Set size: {} {}", size.x(), size.y());
	this->uriField->setFixedSize(Eigen::Vector2i(size.x() - 70, 18));
	this->list->setFixedSize(Eigen::Vector2i(size.x() - 10, size.y() - 28));
};

/**
 *
 */
void NanoguiSourcesControl::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

	nvgBeginPath(ctx);
	nvgRoundedRect( ctx
				  , posX, posY
				  , sizeX, sizeY
				  , 0);
	nvgFillColor(ctx, nanogui::Color(255,255,255,10));
	nvgFill(ctx);

	nanogui::Widget::draw(ctx);
};
