// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_text_field_editor.h"
#include "ui/nanogui/nanogui_textfield.h"

#include <iostream>

using namespace SEAScope;

/**
 *
 */
NanoguiTextFieldEditor::NanoguiTextFieldEditor(nanogui::Widget* parent)
: nanogui::Widget(parent),
  _backgroundColor(0, 0, 0, 0),
  _nameLabel(nullptr),
  _valueTextField(nullptr),
  _valueLabel(nullptr),
  _resetButton(nullptr),
  _deleteButton(nullptr),
  _deleteCallback(nullptr),
  _originalValue(""),
  _editing(false),
  _reflowNeeded(false)
{
    // Assuming 360px width
    this->setFixedWidth(340);
    this->setWidth(340);

    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                           nanogui::Alignment::Middle,
                                           1, 3));

    nanogui::Widget* nameBlock = new nanogui::Widget(this);
    nameBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                nanogui::Alignment::Maximum,
                                                0, 0));
    nameBlock->setFixedWidth(146);
    this->_nameLabel = new nanogui::Label(nameBlock, "");
    this->_nameLabel->setFixedWidth(146);
    this->_nameLabel->setWidth(146);

    nanogui::Widget* resetBlock = new nanogui::Widget(this);
    resetBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                 nanogui::Alignment::Middle,
                                                 0, 0));
    resetBlock->setFixedSize(Eigen::Vector2i(18, 18));

    this->_resetButton = new nanogui::Button(resetBlock, "", ENTYPO_ICON_BACK_IN_TIME);
    this->_resetButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_resetButton->setFontSize(16);
    this->_resetButton->setCallback([this] () {
        this->setValue(this->_originalValue);
    });

    nanogui::Widget* valueBlock = new nanogui::Widget(this);
    valueBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                 nanogui::Alignment::Middle,
                                                 0, 0));
    valueBlock->setFixedWidth(146);

    this->_valueTextField = new NanoguiTextField(valueBlock);
    this->_valueTextField->setFixedWidth(146);
    this->_valueTextField->setWidth(146);
    this->_valueTextField->setFontSize(16);
    this->_valueTextField->setCallback([this] (const std::string& value) -> bool {
        bool changed(value != this->_originalValue);
        this->_valueLabel->setCaption(value);
        this->_resetButton->setVisible(changed);
        this->_reflowNeeded = true;
        return true;
    });
    this->_valueTextField->setEditable(false);
    this->_valueTextField->setVisible(false);

    this->_valueLabel = new nanogui::Label(valueBlock, "");
    this->_valueLabel->setFixedWidth(146);
    this->_valueLabel->setWidth(146);

    this->_deleteButton = new nanogui::Button(this, "", ENTYPO_ICON_TRASH);
    this->_deleteButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_deleteButton->setFontSize(16);
    this->_deleteButton->setCallback([this] () {
        if
            (nullptr != this->_deleteCallback)
        {
            this->_deleteCallback("", this->getName());
        }
    });
    this->_deleteButton->setVisible(false);

    this->_reflowNeeded = true;
}

/**
 *
 */
NanoguiTextFieldEditor::~NanoguiTextFieldEditor()
{

}

/**
 *
 */
bool NanoguiTextFieldEditor::activateEdition(bool activate)
{
    if
        (activate == this->_editing)
    {
        return false;
    }

    this->_editing = activate;

    this->_deleteButton->setVisible(this->_editing);
    this->_valueTextField->setEditable(this->_editing);
    this->_valueTextField->setVisible(this->_editing);
    this->_valueLabel->setVisible(!this->_editing);
    this->_reflowNeeded = true;

    return true;
}

/**
 *
 */
void NanoguiTextFieldEditor::draw(NVGcontext* ctx)
{
    float posX;
    float posY;
    float sizeX;
    float sizeY;

    if
        (true == this->_reflowNeeded)
    {
        this->performLayout(ctx);
        this->_reflowNeeded = false;
    }

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

    nvgBeginPath(ctx);
    nvgRect(ctx, posX, posY, sizeX, sizeY);
    nvgFillColor(ctx, this->_backgroundColor);
    nvgFill(ctx);

    nanogui::Widget::draw(ctx);
}

/**
 *
 */
void NanoguiTextFieldEditor::setName(const std::string& name)
{
    this->_nameLabel->setCaption(name);
}

/**
 *
 */
const std::string& NanoguiTextFieldEditor::getName() const
{
    return this->_nameLabel->caption();
}

/**
 *
 */
void NanoguiTextFieldEditor::setCategory(const std::string& category)
{
    this->_category = category;
}

/**
 *
 */
const std::string& NanoguiTextFieldEditor::getCategory() const
{
    return this->_category;
}

/**
 *
 */
void NanoguiTextFieldEditor::setValue(const std::string& value)
{
    this->_valueTextField->setValue(value);
    this->_valueLabel->setCaption(value);
    this->_resetButton->setVisible(value != this->_originalValue);
    this->_reflowNeeded = true;
}

/**
 *
 */
void NanoguiTextFieldEditor::setOriginalValue(const std::string& value)
{
    this->_originalValue = value;
    this->_resetButton->setTooltip("Restore initial value (" + value + ")");
    this->_resetButton->setVisible(value != this->_valueTextField->value());
    this->_reflowNeeded = true;
}

/**
 *
 */
const std::string& NanoguiTextFieldEditor::getValue() const
{
    return this->_valueLabel->caption();
}

/**
 *
 */
const std::string& NanoguiTextFieldEditor::getOriginalValue() const
{
    return this->_originalValue;
}

/**
 *
 */
void NanoguiTextFieldEditor::setBackgroundColor(const nanogui::Color& color)
{
    this->_backgroundColor = color;
}

/**
 *
 */
const nanogui::Color& NanoguiTextFieldEditor::getBackgroundColor() const
{
    return this->_backgroundColor;
}

/**
 *
 */
bool NanoguiTextFieldEditor::setDeleteCallback(const std::function<void(const std::string&, const std::string&)>& callback)
{
    if
        (nullptr != this->_deleteCallback)
    {
        return false;
    }
    this->_deleteCallback = callback;
    return true;
}
