// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_textfield.h"
#include "vendors/_glfw3.h"

using namespace SEAScope;

/**
 *
 */
NanoguiTextField::NanoguiTextField(nanogui::Widget* parent)
: nanogui::TextBox(parent, ""),
  _validatedValue(""),
  _autoCommit(true),
  _underlineColor(0, 0, 0, 0),
  _focusCallback(nullptr),
  _editionCompleteCallback(nullptr)
{
    this->setEditable(true);
}

/**
 *
 */
NanoguiTextField::~NanoguiTextField()
{
}

/**
 *
 */
bool NanoguiTextField::keyboardEvent(int key, int /*scancode*/, int action, int modifiers)
{
    bool ok(false);
    if
        (  (false == this->mEditable)
        || (false == this->mEnabled))
    {
        return false;
    }

    /* This code block is copied from nanogui::TextBox because there is no
    other way to override the methods used for copy/paste
    (protected but non-virtual functions...) that expect widgets to have a
    nanogui Window as parent whereas it is not the case in SEAScope.

    Code has been slightly modified to fix old cast warnings and harmonize
    code style
    */
    if
        (this->mEditable && this->focused())
    {
        if
            (action == GLFW_PRESS || action == GLFW_REPEAT)
        {
            if
                (key == GLFW_KEY_LEFT)
            {
                if
                    (modifiers == GLFW_MOD_SHIFT)
                {
                    if
                        (this->mSelectionPos == -1)
                    {
                        this->mSelectionPos = this->mCursorPos;
                    }
                }
                else
                {
                    this->mSelectionPos = -1;
                }

                if
                    (this->mCursorPos > 0)
                {
                    this->mCursorPos--;
                }
            }
            else if
                (key == GLFW_KEY_RIGHT)
            {
                if
                    (modifiers == GLFW_MOD_SHIFT)
                {
                    if
                        (this->mSelectionPos == -1)
                    {
                        this->mSelectionPos = this->mCursorPos;
                    }
                }
                else
                {
                    this->mSelectionPos = -1;
                }

                if
                    (this->mCursorPos < static_cast<int>(this->mValueTemp.length()))
                {
                    this->mCursorPos++;
                }
            }
            else if
                (key == GLFW_KEY_HOME)
            {
                if
                    (modifiers == GLFW_MOD_SHIFT)
                {
                    if
                        (this->mSelectionPos == -1)
                    {
                        this->mSelectionPos = this->mCursorPos;
                    }
                }
                else
                {
                    this->mSelectionPos = -1;
                }

                this->mCursorPos = 0;
            }
            else if
                (key == GLFW_KEY_END)
            {
                if
                    (modifiers == GLFW_MOD_SHIFT)
                {
                    if
                        (this->mSelectionPos == -1)
                    {
                        this->mSelectionPos = this->mCursorPos;
                    }
                }
                else
                {
                    this->mSelectionPos = -1;
                }

                this->mCursorPos = static_cast<int>(this->mValueTemp.size());
            }
            else if
                (key == GLFW_KEY_BACKSPACE)
            {
                if
                    (!(this->deleteSelection()))
                {
                    if
                        (this->mCursorPos > 0)
                    {
                        this->mValueTemp.erase(this->mValueTemp.begin() + this->mCursorPos - 1);
                        this->mCursorPos--;
                    }
                }
            }
            else if
                (key == GLFW_KEY_DELETE)
            {
                if
                    (!(this->deleteSelection()))
                {
                    if
                        (this->mCursorPos < static_cast<int>(this->mValueTemp.length()))
                    {
                        this->mValueTemp.erase(this->mValueTemp.begin() + this->mCursorPos);
                    }
                }
            }
            else if
                (  key == GLFW_KEY_ENTER
                || key == GLFW_KEY_KP_ENTER)    // Added support for validating with numpad Enter key
            {
                if
                    (!(this->mCommitted))
                {
                    this->focusEvent(false);
                    if
                        (nullptr != this->_editionCompleteCallback)
                    {
                        this->_editionCompleteCallback();
                    }
                }
            }
            /* Added support for cancelling edition */
            else if
                (key == GLFW_KEY_ESCAPE)
            {
                if
                    (!(this->mCommitted))
                {
                    this->mValueTemp = this->_validatedValue;
                    this->focusEvent(false);
                    if
                        (nullptr != this->_editionCompleteCallback)
                    {
                        this->_editionCompleteCallback();
                    }
                }
            }
            /* End of addition */
            else if
                (key == GLFW_KEY_A && modifiers == SYSTEM_COMMAND_MOD)
            {
                this->mCursorPos = static_cast<int>(this->mValueTemp.length());
                this->mSelectionPos = 0;
            }
            else if
                (key == GLFW_KEY_X && modifiers == SYSTEM_COMMAND_MOD)
            {
                this->copySelection();
                this->deleteSelection();
            }
            else if
                (key == GLFW_KEY_C && modifiers == SYSTEM_COMMAND_MOD)
            {
                this->copySelection();
            }
            else if
                (key == GLFW_KEY_V && modifiers == SYSTEM_COMMAND_MOD)
            {
                this->deleteSelection();
                this->pasteFromClipboard();
            }

            this->mValidFormat =
                (this->mValueTemp == "") || this->checkFormat(this->mValueTemp, this->mFormat);
        }

        ok = true;
    }
    // End of code block copied from nanogui::TextBox

    if
        ((true == this->_autoCommit) && (GLFW_RELEASE == action))
    {
        this->mValue = this->mValueTemp;

        if
            (nullptr != this->mCallback)
        {
            this->mCallback(this->mValue);
        }
    }

    return ok;
}

/**
 *
 */
bool NanoguiTextField::focusEvent(bool focused)
{
    bool ok(false);

    ok = nanogui::TextBox::focusEvent(focused);
    this->_validatedValue = this->mValue;
    if
        (  false == focused
        && (nullptr != this->_focusCallback))
    {
        this->_focusCallback(focused);
    }

    return ok;
}


/**
 *
 */
bool NanoguiTextField::setFocusCallback(const std::function<bool(bool)>& callback)
{
    if
        (nullptr == this->_focusCallback)
    {
        this->_focusCallback = callback;
        return true;
    }
    return false;
}

/**
 *
 */
bool NanoguiTextField::setEditionCompleteCallback(const std::function<void()>& callback)
{
    if
        (nullptr == this->_editionCompleteCallback)
    {
        this->_editionCompleteCallback = callback;
        return true;
    }
    return false;
}

/**
 *
 */
bool NanoguiTextField::copySelection()
{
    if
        (mSelectionPos > -1)
    {
        nanogui::Screen *sc(nullptr);
        sc = this->screen();
        if
            (nullptr == sc)
        {
            return false;
        }

        int begin(-1);
        begin = this->mCursorPos;
        int end(-1);
        end = this->mSelectionPos;

        if
            (begin > end)
        {
            std::swap(begin, end);
        }

        glfwSetClipboardString(sc->glfwWindow(),
                               this->mValueTemp.substr(static_cast<unsigned long>(begin), static_cast<unsigned long>(end)).c_str());
        return true;
    }

    return false;
}

/**
 *
 */
void NanoguiTextField::pasteFromClipboard() {
    nanogui::Screen *sc(nullptr);
    sc = this->screen();
    if
        (nullptr == sc)
    {
        return;
    }

    const char* cbstr = glfwGetClipboardString(sc->glfwWindow());
    if
        (nullptr != cbstr)
    {
        mValueTemp.insert(static_cast<unsigned long>(this->mCursorPos), std::string(cbstr));
    }
}

/**
 *
 */
void NanoguiTextField::draw(NVGcontext* ctx)
{
    float posX {0.f};   //
    float posY {0.f};   //
    float sizeX {0.f};  //
    float sizeY {0.f};  //

    nanogui::TextBox::draw(ctx);

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

    nvgBeginPath(ctx);
    nvgMoveTo(ctx, posX + 1.f, posY + sizeY - 1.f);
    nvgLineTo(ctx, posX + sizeX - 1.f, posY + sizeY - 1.f);
    nvgStrokeWidth(ctx, 2);
    nvgStrokeColor(ctx, this->_underlineColor);
    nvgStroke(ctx);
}

/**
 *
 */
void NanoguiTextField::hintValidValue()
{
    this->_underlineColor = nanogui::Color(0, 0, 0, 0); // transparent
}

/**
 *
 */
void NanoguiTextField::hintInvalidValue()
{
    this->_underlineColor = nanogui::Color(255, 0, 0, 200); // very visible on dark background so that users can't miss it
}
