// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "vendors/_nanogui.h"
#include "ui/nanogui/nanogui_textured_button.h"

#include "3d/icons.h"
#include "num_utils.h"
#include "ui/nanogui/nanogui_renderer.h"
#include "ui/nanogui/nanovg_wrapper.h"

namespace SEAScope
{

NanoguiTexturedButton::NanoguiTexturedButton(nanogui::Widget* parent)
 :  nanogui::Widget(parent),
    _icon(-1),
    mPushed(false),
    mBackgroundColor(0, 0)
{
}

NanoguiTexturedButton::~NanoguiTexturedButton()
{
}

void NanoguiTexturedButton::setTheme(nanogui::Theme* theme)
{
    Widget::setTheme(theme);
}

void NanoguiTexturedButton::setIcon(int icon)
{
    this->_icon = icon;
}

nanogui::Vector2i SEAScope::NanoguiTexturedButton::preferredSize(NVGcontext* /*ctx*/) const
{
    nanogui::Vector2i size(32,32);
    return size;
}

// Copy of nanogui::Button::draw
void SEAScope::NanoguiTexturedButton::draw(NVGcontext* ctx)
{
    Widget::draw(ctx);

    nanogui::Color gradTop = mTheme->mButtonGradientTopUnfocused;
    nanogui::Color gradBot = mTheme->mButtonGradientBotUnfocused;

    if (mPushed) {
        gradTop = mTheme->mButtonGradientTopPushed;
        gradBot = mTheme->mButtonGradientBotPushed;
    }
    else if (mMouseFocus && mEnabled) {
        gradTop = mTheme->mButtonGradientTopFocused;
        gradBot = mTheme->mButtonGradientBotFocused;
    }

    nvgBeginPath(ctx);

    nvg::roundRect(ctx, 
        mPos + nanogui::Vector2i(1, 1), 
        mSize + nanogui::Vector2i(-2, -2), 
        mTheme->mButtonCornerRadius - 1);

    if
        (false == floating_point_equality(mBackgroundColor.w(), 0.f))
    {
        nvgFillColor(ctx, nanogui::Color(mBackgroundColor.head<3>(), 1.f));
        nvgFill(ctx);
        if (mPushed) {
            gradTop.w() = gradBot.w() = 0.8f;
        }
        else {
            float v = 1.f - mBackgroundColor.w();
            gradTop.w() = gradBot.w() = mEnabled ? v : v * .5f + .5f;
        }
    }

    NVGpaint bg = nvg::linearGradient(ctx, mPos, mPos + nanogui::Vector2i(0, mSize.y()), gradTop, gradBot);

    nvgFillPaint(ctx, bg);
    nvgFill(ctx);

    nvgBeginPath(ctx);
    nvgStrokeWidth(ctx, 1.0f);

    nvg::roundRect(ctx,
        mPos + nanogui::Vector2i(0, mPushed ? 0 : 1),
        mSize + nanogui::Vector2i(-1, -1 - (mPushed ? 0 : 1)),
        mTheme->mButtonCornerRadius);

    nvgStrokeColor(ctx, mTheme->mBorderLight);
    nvgStroke(ctx);

    nvgBeginPath(ctx);

    nvg::roundRect(ctx,
        mPos,
        mSize + nanogui::Vector2i(-1, -2),
        mTheme->mButtonCornerRadius);

    nvgStrokeColor(ctx, mTheme->mBorderDark);
    nvgStroke(ctx);

    if
        (-1 < this->_icon)
    {
        int w, h;

        nvgImageSize(ctx, this->_icon, &w, &h);

        const float wRatio = static_cast<float>(w) / static_cast<float>(h);

        // Icon centered
        constexpr int yOffset = 4;
        const int iconSizeY = mSize.y() - yOffset;
        const int iconSizeX = static_cast<int>(static_cast<float>(iconSizeY) * wRatio);
        const int topLeftX = (mPos.x() + (mSize.x() / 2)) - (iconSizeX / 2);

        const nanogui::Vector2i topLeft(topLeftX, mPos.y() + (yOffset / 2));
        const nanogui::Vector2i iconSize(iconSizeX, iconSizeY);

        float alpha = mEnabled ? (mPushed ? 0.5f : (mMouseFocus ? 1.f : 0.8f)) : 0.5f;

        // Rect to avoid texture bleeding due to nanovg GL_CLAMP_TO_EDGE
        nvgBeginPath(ctx);
        nvg::rect(ctx, topLeft, iconSize);
        NVGpaint imgPaint = nvg::imagePattern(ctx, topLeft, iconSize, 0.f, this->_icon, alpha);
        nvgFillPaint(ctx, imgPaint);
        nvgFill(ctx);
    }
}

bool NanoguiTexturedButton::mouseButtonEvent(const nanogui::Vector2i& p, int button, bool down, int modifiers)
{
    Widget::mouseButtonEvent(p, button, down, modifiers);

    /* Temporarily increase the reference count of the button in case the
       button causes the parent window to be destructed */
    nanogui::ref<NanoguiTexturedButton> self = this;

    if (button == GLFW_MOUSE_BUTTON_1 && mEnabled) 
    {
        if 
            (down) 
        {
            mPushed = true;
        }
        else if 
            (mPushed) 
        {
            if 
                (true == contains(p) 
              && nullptr != _callback)
            {
                _callback();
            }
                
            mPushed = false;
        }

        return true;
    }

    return false;
}

void NanoguiTexturedButton::setCallback(const std::function<void()>& func)
{
    _callback = func;
}

} // namespace SEAScope
