// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "core/log.h"
#include "ui/nanogui/nanogui_timeline.h"
#include "ui/nanogui/nanogui_button_bar.h"
#include "ui/nanogui/nanogui_anchored.h"
#include "ui/nanogui/nanogui_timestep_settings.h"

using namespace SEAScope;

/**
 *
 */
NanoguiTimeline::NanoguiTimeline( nanogui::Widget* parent)
: nanogui::Widget(parent), NanoguiComponent(parent)
, currentChangedCallback(nullptr)
, timeframeChangedCallback(nullptr)
, prevCallback(nullptr), nextCallback(nullptr)
, optionsChangedCallback(nullptr), modeChangedCallback(nullptr)
, datebox(nullptr), timeframes(nullptr)
, calendarsView(nullptr), monthsList(nullptr)
, daysList(nullptr), nearestButton(nullptr)
, recenterViewButton(nullptr), year(1970), month(1), day(1), visiblePast(0)
, visibleFuture(0), changeOrigin(""), _timestepSettingsButton(nullptr)
, _timestepLabel(nullptr), _timestepSettings(nullptr), _viewModeButton(nullptr)
{
};

/**
 *
 */
bool NanoguiTimeline::init( const std::vector<std::string>& timeframesIn
						  , const std::vector<std::string>& timeframesTooltips
						  , const std::vector<std::string>& timestepsIn
						  , const std::vector<std::string>& timestepsTooltips
						  , std::uint16_t minYear
						  , std::uint16_t maxYear
                          , bool hoverEnabled)
{
	nanogui::Widget* minimalPanel;	//
	nanogui::Widget* timeframesComponent;	//
	nanogui::Button* btn;	//
	nanogui::Label* lbl;	//

	this->setLayout(new nanogui::BoxLayout( nanogui::Orientation::Vertical
										  , nanogui::Alignment::Fill
										  , 0, 0));

	minimalPanel = new nanogui::Widget(this);
	minimalPanel->setLayout(new nanogui::BoxLayout( nanogui::Orientation::Horizontal
										  , nanogui::Alignment::Middle
										  , 2, 30));

	/* timeframes */
	timeframesComponent = new nanogui::Widget(minimalPanel);
	timeframesComponent->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
														  , nanogui::Alignment::Middle
													      , 1, 2));
	lbl = new nanogui::Label(timeframesComponent, "Timespan", "sans-bold");
	lbl->setFontSize(14);
	lbl->setTooltip("Change the size of the time window that data must intersect to be visible");

	this->timeframes = new NanoguiButtonBar(timeframesComponent);
	this->timeframes->init( timeframesIn, timeframesTooltips, 14, Eigen::Vector2i(20, 20));
	this->timeframes->setSelectionChangedCallback([this](const std::string& value) {
		if
			(nullptr != this->timeframeChangedCallback)
		{
			this->timeframeChangedCallback(value);
		}
	});

	/* timesteps */
	nanogui::Widget* navComponent {nullptr};
    navComponent = new nanogui::Widget(minimalPanel);
	navComponent->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Horizontal
												   , nanogui::Alignment::Middle
												   , 2, 0));

	btn = new nanogui::Button(navComponent, "", ENTYPO_ICON_CHEVRON_LEFT);
    btn->setTooltip("Shift to a datetime in the past. The same can be achieved by pressing the left arrow key");
	btn->setFixedSize(Eigen::Vector2i(18, 38));
	btn->setFontSize(18);
	btn->setCallback([this]() {
		if
			(nullptr != this->prevCallback)
		{
			this->prevCallback();
		}
	});

    nanogui::Widget* selectedTimestepContainer {nullptr};
    selectedTimestepContainer = new nanogui::Widget(navComponent);
    selectedTimestepContainer->setLayout(new nanogui::BoxLayout( nanogui::Orientation::Vertical,
                                                                 nanogui::Alignment::Middle,
                                                                 0, 0));

    nanogui::Widget* timestepTopRow {nullptr};
    timestepTopRow = new nanogui::Widget(selectedTimestepContainer);
    timestepTopRow->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                     nanogui::Alignment::Middle,
                                                     2, 4));
    lbl = new nanogui::Label(timestepTopRow, "Move by", "sans-bold");
    lbl->setFontSize(14);

    this->_timestepSettingsButton = new nanogui::Button(timestepTopRow, "", ENTYPO_ICON_COG);
    this->_timestepSettingsButton->setFlags(nanogui::Button::ToggleButton);
    this->_timestepSettingsButton->setTooltip("Select a timestep");
    this->_timestepSettingsButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_timestepSettingsButton->setFontSize(18);

    nanogui::Widget* timestepLabelContainer {nullptr};  //
    timestepLabelContainer = new nanogui::Widget(selectedTimestepContainer);
    timestepLabelContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                             nanogui::Alignment::Middle,
                                                             2, 0));
    timestepLabelContainer->setFixedSize(Eigen::Vector2i(50, 18));
    this->_timestepLabel = new nanogui::Label(timestepLabelContainer, "", "sans-bold");
    this->_timestepLabel->setFontSize(14);

	btn = new nanogui::Button(navComponent, "", ENTYPO_ICON_CHEVRON_RIGHT);
    btn->setTooltip("Shift to a datetime in the future. The same can be achieved by pressing the right arrow key");
	btn->setFixedSize(Eigen::Vector2i(18, 38));
	btn->setFontSize(18);
	btn->setCallback([this]() {
		if
			(nullptr != this->nextCallback)
		{
			this->nextCallback();
		}
	});

    NanoguiAnchored* timestepSettingsContainer {nullptr};
    timestepSettingsContainer = NanoguiAnchored::create(this->_timestepSettingsButton);
    timestepSettingsContainer->offset = 4u;
    timestepSettingsContainer->placement = static_cast<std::uint8_t>(NanoguiAnchored::ON_TOP);
    timestepSettingsContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                                nanogui::Alignment::Minimum,
                                                                0, 0));
    this->_timestepSettings = new NanoguiTimestepSettings(timestepSettingsContainer);
    this->_timestepSettings->init(timestepsIn, timestepsTooltips);
    this->_timestepSettingsButton->setChangeCallback([timestepSettingsContainer] (bool /*pushed*/) {
        timestepSettingsContainer->setVisible(!timestepSettingsContainer->visible());
    });

    timestepSettingsContainer->setVisible(false);

	/* datetime */
	this->datebox = new NanoguiDatetimeBox(minimalPanel, 2015, 7, 28, 12, 45, 38);
	this->datebox->setFontSize(18);
	this->datebox->setFixedSize(Eigen::Vector2i(180,30));
	this->datebox->setCallback([this](const std::string& value) -> bool {
		LOG(LogLevel::Debug, "Timeline", "Date changed by user input: {}", value);
		if
			(nullptr != this->currentChangedCallback)
		{
			this->currentChangedCallback(value, this->changeOrigin);
			this->changeOrigin = "";
		}
		return true; /* false causes value to roll back */
	});
	this->datebox->setYearsGrabber([this](std::vector<uint16_t>& values, std::vector<uint16_t>& extraValues) -> bool {
		bool ok;	//
		LOG( LogLevel::Debug, "Timeline", "grab years");
		values.clear();
		ok = this->hintsSource.getYears(values);
		extraValues.clear();
		this->extraCoverage.getYears(extraValues);
		return ok;
	});
	this->datebox->setMonthsGrabber([this](std::vector<uint8_t>& values, std::vector<uint8_t>& extraValues, uint16_t yearIn) -> bool {
		bool ok;	//
		LOG( LogLevel::Debug, "Timeline", "grab months for year {}", static_cast<int>(yearIn));
		values.clear();
		ok = this->hintsSource.getMonths(yearIn, values);
		extraValues.clear();
		this->extraCoverage.getMonths(yearIn, extraValues);
		return ok;
	});
	this->datebox->setDaysGrabber([this](std::vector<uint8_t>& values, std::vector<uint8_t>& extraValues, uint16_t yearIn, uint8_t monthIn) -> bool {
		bool ok;	//
		LOG( LogLevel::Debug, "Timeline", "grab days for year {} month {}", static_cast<int>(yearIn), static_cast<int>(monthIn));
		values.clear();
		ok = this->hintsSource.getDays(yearIn, monthIn, values);
		extraValues.clear();
		this->extraCoverage.getDays(yearIn, monthIn, extraValues);
		return ok;
	});

	nanogui::Widget* compactTimeline;	//
	compactTimeline = new nanogui::Widget(this);
	compactTimeline->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
													  , nanogui::Alignment::Fill
													  , 2, 2));

	this->calendarsView = new NanoguiCalendarsTimeline(compactTimeline);
    this->calendarsView->hoverEnabled = hoverEnabled;
	this->calendarsView->setBounds(minYear, maxYear);
	this->calendarsView->yearCallback = [this] (std::uint16_t yearIn) {
		this->setYear(yearIn);
	};
	this->calendarsView->monthCallback = [this] (std::uint16_t yearIn, std::uint8_t monthIn) {
        this->setMonth(yearIn, monthIn);
	};
	this->calendarsView->currentYearVisibilityChanged = [this] (bool currentYearVisible) {
		this->recenterViewButton->setVisible(!currentYearVisible);
		this->reflowNeeded = true;
	};

	this->monthsList = new NanoguiMonthsList(compactTimeline);
    this->monthsList->hoverEnabled = hoverEnabled;
	this->monthsList->monthCallback = [this](std::uint8_t monthIn) {
		bool okDay;	//
		std::uint8_t dayTmp;	//
		std::uint8_t maxDay;	//

		okDay = this->datebox->getDay(dayTmp);
		if
			(true == okDay)
		{
			LOG( LogLevel::Debug, "Timeline"
					, "Day retrieved: {}"
					, static_cast<int>(dayTmp));

			maxDay = NanoguiDateBox::daysPerMonth[monthIn - 1u];
			if
				(2 == monthIn && 0 == year%4)
			{
				maxDay = 1u + maxDay;
			}

			if
				(dayTmp > maxDay)
			{
				this->datebox->setDay(maxDay, false);
			}
		}
		this->datebox->setMonth(monthIn, false);

		this->changeOrigin = "monthButton";
		this->datebox->commitValue();
	};

	this->daysList = new NanoguiDaysList(compactTimeline);
    this->daysList->hoverEnabled = hoverEnabled;
	this->daysList->dayCallback = [this](std::uint8_t dayIn) {
		this->datebox->setDay(dayIn, false);
		this->changeOrigin = "dayButton";
		this->datebox->commitValue();
	};

	this->nearestButton = new nanogui::Button(minimalPanel, "", ENTYPO_ICON_ADJUST);
	this->nearestButton->setTooltip("Select the closest datetime for which there is a granule to display.");
	this->nearestButton->setFlags(nanogui::Button::ToggleButton);
	this->nearestButton->setFixedSize(Eigen::Vector2i(24,18));
	this->nearestButton->setFontSize(20);
	this->nearestButton->setChangeCallback([this](bool /*pushed*/) {
		if
			(nullptr != this->optionsChangedCallback)
		{
			this->optionsChangedCallback();
		}
	});

	/* Mode selector */
	nanogui::Widget* modes = new nanogui::Widget(minimalPanel);
	modes->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
											, nanogui::Alignment::Middle
											, 0, 0));
	this->_viewModeButton = new nanogui::Button(modes, "", ENTYPO_ICON_CALENDAR);
	this->_viewModeButton->setFlags(nanogui::Button::ToggleButton);
	this->_viewModeButton->setPushed(true);
	this->_viewModeButton->setTooltip("Toggle calendar view");
	this->_viewModeButton->setFixedSize(Eigen::Vector2i(24,18));
	this->_viewModeButton->setFontSize(20);
	this->_viewModeButton->setChangeCallback([this](bool pushed) {
        this->setExtendedView(pushed, false /* not silent, apply callbacks */);
	});

	btn = new nanogui::Button(modes, "", ENTYPO_ICON_LIST);
	btn->setTooltip("Browser view");
	btn->setFixedSize(Eigen::Vector2i(24,18));
	btn->setFontSize(16);
	btn->setVisible(false); /* TODO: remove this line once the browsing mode is implemented */

	this->recenterViewButton = new nanogui::Button(minimalPanel, "", ENTYPO_ICON_ALIGN_HORIZONTAL_MIDDLE);
	this->recenterViewButton->setTooltip("Pan the calendars view so that the selected year is visible");
	this->recenterViewButton->setFixedSize(Eigen::Vector2i(24,18));
	this->recenterViewButton->setFontSize(20);
	this->recenterViewButton->setCallback([this]() {
		this->updateViews(true);
	});
	this->recenterViewButton->setVisible(false);

    this->setPosition(this->mPos.x(), this->mPos.y());
    this->setSize(this->mSize.x(), this->mSize.y());

	return true;
};

/**
 *
 */
NanoguiTimeline::~NanoguiTimeline()
{
};

/**
 *
 */
void NanoguiTimeline::setYear(std::uint16_t yearIn)
{
	std::uint8_t monthTmp;	//
	std::uint8_t dayTmp;	//
	std::uint8_t maxDay;	//
	bool okMonth;	//
	bool okDay;	//

	LOG( LogLevel::Debug, "Timeline"
			, "Calendar year change: {}"
			, static_cast<int>(yearIn));
	this->datebox->setYear(yearIn, false);

	okMonth = this->datebox->getMonth(monthTmp);
	okDay = this->datebox->getDay(dayTmp);

	if
		(okMonth && okDay)
	{
		maxDay = NanoguiDateBox::daysPerMonth[monthTmp - 1u];
		if
			(2 == monthTmp && 0 == yearIn%4)
		{
			maxDay = 1u + maxDay;
		}

		if
			(dayTmp > maxDay)
		{
			this->datebox->setDay(maxDay, false);
		}
	}

	this->changeOrigin = "yearButton";
	this->datebox->commitValue();
};

/**
 *
 */
void NanoguiTimeline::setMonth(std::uint16_t yearIn, std::uint8_t monthIn)
{
	std::uint8_t dayTmp;	//
	std::uint8_t maxDay;	//
	bool okDay;	//

	LOG( LogLevel::Debug, "Timeline"
			, "Calendar month change: {}-{}"
			, static_cast<int>(yearIn), static_cast<int>(monthIn));
	this->datebox->setYear(yearIn, false);
	this->datebox->setMonth(monthIn, false);
	okDay = this->datebox->getDay(dayTmp);
	if
		(true == okDay)
	{
		LOG( LogLevel::Debug, "Timeline"
				, "Day retrieved: {}"
				, static_cast<int>(dayTmp));

		maxDay = NanoguiDateBox::daysPerMonth[monthIn - 1u];
		if
			(2 == monthIn && 0 == yearIn%4)
		{
			maxDay = 1u + maxDay;
		}

		if
			(dayTmp > maxDay)
		{
			this->datebox->setDay(maxDay, false);
		}
	}
	else
	{
		LOG( LogLevel::Warn, "Timeline"
				, "Could not retrieve day...");
	}

	this->changeOrigin = "monthButton";
	this->datebox->commitValue();
};

/**
 *
 */
void NanoguiTimeline::draw(NVGcontext* ctx)
{
    float posX(0.f);    //
    float posY(0.f);    //
    float sizeX(0.f);   //
    float sizeY(0.f);   //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

	if
		(true == this->reflowNeeded)
	{
		this->performLayout(ctx);
		this->reflowNeeded = false;
	}

	/* draw background */
	nvgBeginPath(ctx);
	nvgRoundedRect( ctx
                  , posX + 1.0f, posY + 1.0f
				  , sizeX - 2.0f, sizeY - 1.0f
				  , 0);
	nvgFillColor(ctx, nanogui::Color(50, 50, 50, 170));
	nvgFill(ctx);

	nanogui::Widget::draw(ctx);
}

/**
 *
 */
bool NanoguiTimeline::setPosition(int x, int y)
{
	bool ok(false);	//

	LOG(LogLevel::Debug, "Timeline", "setPosition {} {}", x, y);
	if
		((nullptr != this->calendarsView) && (false == this->calendarsView->visible()))
	{
		ok = NanoguiComponent::setPosition(x, y + 86 /*104*/);
	}
	else
	{
		ok = NanoguiComponent::setPosition(x, y);
	}
	return ok;
};

/**
 *
 */
bool NanoguiTimeline::setSize(int width, int height)
{
	LOG(LogLevel::Debug, "Timeline UI", "setSize {} {}", width, height);
	if
		(nullptr != this->calendarsView)
	{
		this->calendarsView->setFixedWidth(width);
	}

	if
		(nullptr != this->monthsList)
	{
		this->monthsList->setFixedWidth(width);
	}

	if
		(nullptr != this->daysList)
	{
		this->daysList->setFixedWidth(width);
	}
	nanogui::Widget::setSize(Eigen::Vector2i(width, height));
	return NanoguiComponent::setSize(width, height);
};

/**
 *
 */
bool NanoguiTimeline::setCurrentChangedCallback(const std::function<void(const std::string&, const std::string&)>& callback)
{
	this->currentChangedCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiTimeline::setTimeframeChangedCallback(const std::function<void(const std::string&)>& callback)
{
	this->timeframeChangedCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiTimeline::setTimestepChangedCallback(const std::function<void(const std::string&)>& callback)
{
    return this->_timestepSettings->setTimestepChangedCallback(callback);
};

/**
 *
 */
bool NanoguiTimeline::setOptionsChangedCallback(const std::function<void()>& callback)
{
	this->optionsChangedCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiTimeline::setPrevCallback(const std::function<void()>& callback)
{
	this->prevCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiTimeline::setNextCallback(const std::function<void()>& callback)
{
	this->nextCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiTimeline::mouseButtonEvent(const Eigen::Vector2i& p, int buttons, bool down, int modifiers)
{
	nanogui::Widget::mouseButtonEvent(p, buttons, down, modifiers);
	/* Ignore results and return true to prevent events propagation */
	return true;
};

/**
 *
 */
bool NanoguiTimeline::scrollEvent(const Eigen::Vector2i& p, const Eigen::Vector2f& rel)
{
	nanogui::Widget::scrollEvent(p, rel); // Let children handle the event if they can

	/* Ignore results and return true to prevent events propagation */
    return true;
};

/**
 *
 */
bool NanoguiTimeline::setCurrent(const std::string& date, const std::string& time, int yearIn, int monthIn, int dayIn)
{
	bool changed;	//
	std::ostringstream buffer;	//

	buffer << date << " " << time;
	LOG(LogLevel::Debug, "Timeline UI", "setCurrent {}", buffer.str());
	this->datebox->setValue(buffer.str());

	changed = false;
	if
		(this->year != yearIn)
	{
		this->year = static_cast<std::uint16_t>(yearIn);
		changed = true;
	}

	if
		(this->month != monthIn)
	{
		this->month = static_cast<std::uint8_t>(monthIn);
		changed = true;
	}

	if
		(this->day != dayIn)
	{
		this->day = static_cast<std::uint8_t>(dayIn);
		changed = true;
	}

	if
		(changed)
	{
		this->updateViews(false);
	}
	return changed;
};

/**
 *
 */
bool NanoguiTimeline::updateViews(bool forceRecenter)
{
	this->calendarsView->setVisibilityBounds(this->visiblePast, this->visibleFuture);
	this->calendarsView->setCurrent(this->year, this->month, this->day, forceRecenter);
	this->monthsList->setYear(this->year);
	this->monthsList->select(this->month);
	this->daysList->setVisibilityBounds(this->visiblePast, this->visibleFuture);
	this->daysList->setMonth(this->year, this->month);
	this->daysList->selectDay(this->day);
	return true;
};

/**
 *
 */
bool NanoguiTimeline::setVisibleBounds(timestamp_t past, timestamp_t future)
{
	bool changed;	//

	changed = false;
	if
		(this->visiblePast != past)
	{
		this->visiblePast = past;
		changed = true;
	}

	if
		(this->visibleFuture != future)
	{
		this->visibleFuture = future;
		changed = true;
	}

	if
		(changed)
	{
		this->updateViews(false);
	}
	return changed;
};

/**
 *
 */
bool NanoguiTimeline::setTimeframe(const std::string& label, const std::string& /*description*/)
{
	bool ok;	//
	ok = this->timeframes->selectValue(label);
	return ok;
};

/**
 *
 */
bool NanoguiTimeline::setTimestep(const std::string& label, const std::string& description)
{
    std::ostringstream buffer {""};

    buffer << label << ":\n" << description;
    this->_timestepLabel->setCaption(label);
    this->_timestepLabel->setTooltip(buffer.str());    // In case the label is cropped
    this->reflowNeeded = true;
    return this->_timestepSettings->selectTimestep(label);
};

/**
 *
 */
bool NanoguiTimeline::setHintsSource(const EventsCoverage& coverage)
{
	this->hintsSource = coverage;
	return true;
};

/**
 *
 */
bool NanoguiTimeline::setMainCoverage(const EventsCoverage& coverage)
{
	bool ok;	//

	this->mainCoverage = coverage;
	ok = this->calendarsView->setMainCoverage(coverage);
	ok = this->monthsList->setMainCoverage(coverage);
	ok = this->daysList->setMainCoverage(coverage);
	return ok;
};

/**
 *
 */
bool NanoguiTimeline::setExtraCoverage(const EventsCoverage& coverage)
{
	bool ok;	//

	this->extraCoverage = coverage;
	ok = this->calendarsView->setExtraCoverage(coverage);
	ok = this->monthsList->setExtraCoverage(coverage);
	ok = this->daysList->setExtraCoverage(coverage);
	return ok;
};

/**
 *
 */
bool NanoguiTimeline::nearestMode() const
{
	return this->nearestButton->pushed();
};

/**
 *
 */
/**
 *
 */
bool NanoguiTimeline::setNearestMode(bool value)
{
    if
        (this->nearestButton->pushed() != value)
    {
        this->nearestButton->setPushed(value);
        return true;
    }
    return false;
};

/**
 *
 */
int NanoguiTimeline::getHeightOffset() const
{
	if
		((nullptr != this->calendarsView) && (false == this->calendarsView->visible()))
	{
		return -86;
	}
	return NanoguiComponent::getHeightOffset();
};

/**
 *
 */
bool NanoguiTimeline::setYearScrollCallback(const std::function<void(float)>& callback)
{
    this->calendarsView->scrollCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiTimeline::setMonthScrollCallback(const std::function<void(float)>& callback)
{
    this->monthsList->scrollCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiTimeline::setDayScrollCallback(const std::function<void(float)>& callback)
{
    this->daysList->scrollCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiTimeline::getExtendedView() const
{
    return this->_viewModeButton->pushed();
}

/**
 *
 */
void NanoguiTimeline::setExtendedView(bool value, bool silent)
{
    this->_viewModeButton->setPushed(value);
    if
        (true == silent)
    {
        return;
    }

    if
        (false == value)
    {
        this->calendarsView->setVisible(false);
        this->monthsList->setVisible(false);
        this->daysList->setVisible(false);
        this->mPos += Eigen::Vector2i(0, 86);
        this->setSize(this->fixedWidth(), this->fixedHeight() - 86);
    }
    else
    {
        this->calendarsView->setVisible(true);
        this->monthsList->setVisible(true);
        this->daysList->setVisible(true);
        this->mPos -= Eigen::Vector2i(0, 86);
        this->setSize(this->fixedWidth(), this->fixedHeight() + 86);
    }

    this->requestReflow();

    if
        (nullptr != this->modeChangedCallback)
    {
        this->modeChangedCallback();
    }
}
