// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_timespan_editor.h"

#include "ui/nanogui/nanogui_numberbox.h"
#include "ui/nanogui/nanogui_checkbox.h"
#include "ui/nanogui/nanogui_textfield.h"

using namespace SEAScope;

/**
 *
 */
NanoguiTimespanEditor::NanoguiTimespanEditor(nanogui::Widget* parent)
: nanogui::Widget(parent),
  _label{nullptr}, _description{nullptr}, _past{nullptr}, _future{nullptr},
  _default{nullptr}, _okButton{nullptr}, _deleteButton{nullptr},
  _identifier{""}, _saveCallback{nullptr}, _deleteCallback{nullptr}, _defaultCallback{nullptr}
{
    nanogui::Widget* spacer {nullptr};  //
    nanogui::Widget* leftWidget {nullptr};   //
    nanogui::Widget* w {nullptr};   //

    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                           nanogui::Alignment::Middle,
                                           0, 0));

    spacer = new nanogui::Widget(this);
    spacer->setFixedWidth(25);

    leftWidget = new nanogui::Widget(this);
    leftWidget->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                 nanogui::Alignment::Minimum,
                                                 0, 2));

    w = new nanogui::Widget(leftWidget);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 2));
    this->_label = new NanoguiTextField(w);
    this->_label->setFixedSize(Eigen::Vector2i(150, 20));
    this->_label->setFontSize(16);
    this->_label->setTooltip("Short label to identify the timespan, it will be used as label for the button to activate the timespan in the timeline.");
    this->_label->setCallback([this] (const std::string& value) -> bool {
        this->checkValidity();

        return 0 < value.size();
    });

    this->_default = new NanoguiCheckbox(w, "default");
    this->_default->setFontSize(16);
    this->_default->setTooltip("Use this timespan as default when starting SEAScope.");
    this->_default->setCallback([this] (bool checked) {
        if
            (false == checked)
        {
            // This checkbox is used as a radio button, so it can only be
            // unchecked by an external component
            this->_default->setChecked(true);
            return;
        }

        if
            (nullptr != this->_defaultCallback)
        {
            this->_defaultCallback();
        }
    });

    w = new nanogui::Widget(leftWidget);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 2));
    this->_past = new NanoguiNumberBox(w);
    this->_past->setFixedSize(Eigen::Vector2i(110, 20));
    this->_past->setFontSize(16);
    this->_past->setTooltip("Start of the timespan defined as a number of seconds before the current datetime.");
    this->_past->setValue(1);
    this->_past->setFormat("[0-9]+");
    this->_past->setCallback([this] (double) {
        this->checkValidity();
    });
    this->_future = new NanoguiNumberBox(w);
    this->_future->setFixedSize(Eigen::Vector2i(110, 20));
    this->_future->setFontSize(16);
    this->_future->setTooltip("End of the timespan defined as a number of seconds after the current datetime.");
    this->_future->setValue(1);
    this->_future->setFormat("[0-9]+");
    this->_future->setCallback([this] (double) {
        this->checkValidity();
    });

    w = new nanogui::Widget(leftWidget);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 2));
    this->_description = new NanoguiTextField(w);
    this->_description->setFixedSize(Eigen::Vector2i(200, 20));
    this->_description->setFontSize(16);
    this->_description->setTooltip("Short description of the timespan, it will be displayed as a tooltip when the mouse is over the timespan button in the timeline");
    this->_description->setCallback([this] (const std::string& value) -> bool{
        this->checkValidity();

        return 0 < value.size();
    });

    this->_okButton = new nanogui::Button(w, "");
    this->_okButton->setFontSize(16);
    this->_okButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_okButton->setCallback([this] () {
        if
            (true == this->_label->enabled())
        {
            if
                (nullptr != this->_saveCallback)
            {
                this->_saveCallback(this->_label->value(),
                                    this->_past->value(),
                                    this->_future->value(),
                                    this->_description->value(),
                                    this->_default->checked());
            }
            this->disableEdition();
        }
        else
        {
            this->enableEdition();
            this->checkValidity();
        }
    });

    spacer = new nanogui::Widget(this);
    spacer->setFixedWidth(10);

    this->_deleteButton = new nanogui::Button(this, "", ENTYPO_ICON_TRASH);
    this->_deleteButton->setFontSize(16);
    this->_deleteButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_deleteButton->setCallback([this] () {
        if
            (nullptr != this->_deleteCallback)
        {
            this->_deleteCallback();
        }
    });

    this->disableEdition();
}

/**
 *
 */
NanoguiTimespanEditor::~NanoguiTimespanEditor()
{

}

/**
 *
 */
void NanoguiTimespanEditor::set(const std::string& label, const std::string& description, const timestamp_t& past, const timestamp_t& future, bool isDefault)
{
    this->_identifier = label;
    this->_label->setValue(label);
    this->_description->setValue(description);
    this->_past->setValue(static_cast<double>(past));
    this->_future->setValue(static_cast<double>(future));
    this->_default->setChecked(isDefault);

    // Only allow change of the default checkbox for existing timespans, not
    // new ones
    this->_default->setEnabled(0 < this->_identifier.size());

    this->checkValidity();
}

/**
 *
 */
void NanoguiTimespanEditor::enableEdition()
{
    this->_label->setEnabled(true);
    this->_label->setEditable(true);
    this->_past->setEnabled(true);
    this->_past->setEditable(true);
    this->_future->setEnabled(true);
    this->_future->setEditable(true);
    this->_description->setEnabled(true);
    this->_description->setEditable(true);
    this->_okButton->setCaption("Ok");
    this->_okButton->setIcon(0);

    // Only allow change of the default checkbox for existing timespans, not
    // new ones
    this->_default->setEnabled(0 < this->_identifier.size());

    this->checkValidity();
}

/**
 *
 */
void NanoguiTimespanEditor::disableEdition()
{
    this->_label->setEnabled(false);
    this->_label->setEditable(false);
    this->_past->setEnabled(false);
    this->_past->setEditable(false);
    this->_future->setEnabled(false);
    this->_future->setEditable(false);
    this->_description->setEnabled(false);
    this->_description->setEditable(false);
    this->_okButton->setCaption("");
    this->_okButton->setIcon(ENTYPO_ICON_EDIT);
}

/**
 *
 */
void NanoguiTimespanEditor::checkValidity()
{
    bool ok {false};    //

    ok = (0 < this->_label->value().size());
    ok = ok && (0 < this->_description->value().size());
    ok = ok && (1.0 <= this->_past->value());
    ok = ok && (1.0 <= this->_future->value());

    this->_okButton->setEnabled(ok);
}

/**
 *
 */
void NanoguiTimespanEditor::setSaveCallback(const std::function<void(const std::string&, double, double, const std::string&, bool)>& callback)
{
    this->_saveCallback = callback;
}

/**
 *
 */
void NanoguiTimespanEditor::setDeleteCallback(const std::function<void()>& callback)
{
    this->_deleteCallback = callback;
}

/**
 *
 */
void NanoguiTimespanEditor::setDefaultCallback(const std::function<void()>& callback)
{
    this->_defaultCallback = callback;
}

/**
 *
 */
const std::string& NanoguiTimespanEditor::getIdentifier() const
{
    return this->_identifier;
}

/**
 *
 */
void NanoguiTimespanEditor::draw(NVGcontext* ctx)
{
    float posX {0.f};   //
    float posY {0.f};   //
    float sizeY {0.f};  //
    nanogui::Color color{200, 200, 200, 100};   //
    if
        (0 >= this->_identifier.size())
    {
        // Use a different color for the left bar when the edited timespan has
        // not been added to the timeline configuration yet, i.e. when its
        // identifier is empty
        color = nanogui::Color(255, 255, 100, 255);
    }

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeY = static_cast<float>(this->mSize.y());

    nanogui::Widget::draw(ctx);

    nvgBeginPath(ctx);
    nvgMoveTo(ctx, posX + 20.f, posY);
    nvgLineTo(ctx, posX + 20.f, posY + sizeY);
    nvgStrokeColor(ctx, color);
    nvgStroke(ctx);
}

/**
 *
 */
void NanoguiTimespanEditor::uncheckDefault()
{
    this->_default->setChecked(false);
}

/**
 *
 */
void NanoguiTimespanEditor::checkDefault()
{
    this->_default->setChecked(true);
}
