// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_timestep_editor.h"

#include "ui/nanogui/nanogui_numberbox.h"
#include "ui/nanogui/nanogui_checkbox.h"
#include "ui/nanogui/nanogui_textfield.h"
#include "core/timeline_cfg.h"

using namespace SEAScope;

/**
 *
 */
NanoguiTimestepEditor::NanoguiTimestepEditor(nanogui::Widget* parent)
: nanogui::Widget(parent),
  _label{nullptr}, _description{nullptr}, _value{nullptr}, _default{nullptr},
  _okButton{nullptr}, _deleteButton{nullptr}, _identifier{""},
  _saveCallback{nullptr}, _deleteCallback{nullptr}, _defaultCallback{nullptr}
{
    nanogui::Widget* spacer {nullptr};  //
    nanogui::Widget* leftWidget {nullptr};   //
    nanogui::Widget* w {nullptr};   //

    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                           nanogui::Alignment::Middle,
                                           0, 0));

    spacer = new nanogui::Widget(this);
    spacer->setFixedWidth(25);

    leftWidget = new nanogui::Widget(this);
    leftWidget->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                 nanogui::Alignment::Minimum,
                                                 0, 2));

    w = new nanogui::Widget(leftWidget);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 2));
    this->_label = new NanoguiTextField(w);
    this->_label->setFixedSize(Eigen::Vector2i(70, 20));
    this->_label->setFontSize(16);
    this->_label->setTooltip("Short label to identify the timestep, it will be used as label for the button to activate the timestep in the timeline");
    this->_label->setCallback([this] (const std::string& value) -> bool {
        this->checkValidity();

        return 0 < value.size();
    });

    nanogui::Widget* valueContainer {nullptr};  //
    valueContainer = new nanogui::Widget(w);
    valueContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                     nanogui::Alignment::Middle,
                                                     0, 0));
    valueContainer->setFixedSize(Eigen::Vector2i(80, 20));
    this->_value = new NanoguiNumberBox(valueContainer);
    this->_value->setFixedSize(Eigen::Vector2i(80, 20));
    this->_value->setFontSize(16);
    this->_value->setTooltip("Offset (in seconds) by which the current datetime will be shifted when clicking the prev/next buttons while this timestep is active");
    this->_value->setValue(1);
    this->_value->setFormat("[0-9]+");
    this->_value->setCallback([this] (double) {
        this->checkValidity();
    });

    this->_default = new NanoguiCheckbox(w, "default");
    this->_default->setFontSize(16);
    this->_default->setTooltip("Use this timestep as default when starting SEAScope");
    this->_default->setCallback([this] (bool checked) {
        if
            (false == checked)
        {
            // This checkbox is used as a radio button, so it can only be
            // unckecked by an external component
            this->_default->setChecked(true);
            return;
        }

        if
            (nullptr != this->_defaultCallback)
        {
            this->_defaultCallback();
        }
    });

    w = new nanogui::Widget(leftWidget);
    w->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                        nanogui::Alignment::Middle,
                                        0, 2));
    this->_description = new NanoguiTextField(w);
    this->_description->setFixedSize(Eigen::Vector2i(200, 20));
    this->_description->setFontSize(16);
    this->_description->setTooltip("Short description of the timestep, it will be displayed as a tooltip when the mouse is over the timestep button in the timeline");
    this->_description->setCallback([this] (const std::string& value) -> bool{
        this->checkValidity();

        return 0 < value.size();
    });

    this->_okButton = new nanogui::Button(w, "");
    this->_okButton->setFontSize(16);
    this->_okButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_okButton->setCallback([this] () {
        if
            (true == this->_label->enabled())
        {
            if
                (nullptr != this->_saveCallback)
            {
                this->_saveCallback(this->_label->value(),
                                    this->_value->value(),
                                    this->_description->value(),
                                    this->_default->checked());
            }
            this->disableEdition();
        }
        else
        {
            this->enableEdition();
            this->checkValidity();
        }
    });

    spacer = new nanogui::Widget(this);
    spacer->setFixedWidth(10);

    nanogui::Widget* deleteContainer {nullptr}; //
    deleteContainer = new nanogui::Widget(this);
    deleteContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                      nanogui::Alignment::Middle,
                                                      0, 0));
    deleteContainer->setFixedSize(Eigen::Vector2i(20, 20));

    this->_deleteButton = new nanogui::Button(deleteContainer, "", ENTYPO_ICON_TRASH);
    this->_deleteButton->setFontSize(16);
    this->_deleteButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_deleteButton->setCallback([this] () {
        if
            (nullptr != this->_deleteCallback)
        {
            this->_deleteCallback();
        }
    });

    this->disableEdition();
}

/**
 *
 */
NanoguiTimestepEditor::~NanoguiTimestepEditor()
{

}

/**
 *
 */
void NanoguiTimestepEditor::set(const std::string& label, const std::string& description, const timestamp_t& offset, bool isDefault)
{
    this->_identifier = label;
    this->_label->setValue(label);
    this->_description->setValue(description);
    this->_value->setValue(static_cast<double>(offset));
    this->_default->setChecked(isDefault);

    // Only allow change of the default checkbox for existing timesteps, not
    // new ones
    this->_default->setEnabled(0 < this->_identifier.size());

    bool isReserved {false};    //
    isReserved = (Timestep::ONE_THIRD_TIMESPAN_LABEL == label);

    if
        (true == isReserved)
    {
        this->disableEdition();
        this->_deleteButton->setEnabled(false);
        this->_deleteButton->setVisible(false);
        this->_okButton->setEnabled(false);
        this->_okButton->setVisible(false);
        this->_value->setEnabled(false);
        this->_value->setVisible(false);
    }
    else
    {
        this->_deleteButton->setEnabled(true);
        this->_deleteButton->setVisible(true);
        this->_okButton->setEnabled(true);
        this->_okButton->setVisible(true);
        this->_value->setEnabled(true);
        this->_value->setVisible(true);
        this->checkValidity();
    }
}

/**
 *
 */
void NanoguiTimestepEditor::enableEdition()
{
    if
        (Timestep::ONE_THIRD_TIMESPAN_LABEL == this->_label->value())
    {
        return;
    }
    this->_label->setEnabled(true);
    this->_label->setEditable(true);
    this->_value->setEnabled(true);
    this->_value->setEditable(true);
    this->_description->setEnabled(true);
    this->_description->setEditable(true);
    this->_okButton->setCaption("Ok");
    this->_okButton->setIcon(0);

    // Only allow change of the default checkbox for existing timesteps, not
    // new ones
    this->_default->setEnabled(0 < this->_identifier.size());

    this->checkValidity();
}

/**
 *
 */
void NanoguiTimestepEditor::disableEdition()
{
    this->_label->setEnabled(false);
    this->_label->setEditable(false);
    this->_value->setEnabled(false);
    this->_value->setEditable(false);
    this->_description->setEnabled(false);
    this->_description->setEditable(false);
    this->_okButton->setCaption("");
    this->_okButton->setIcon(ENTYPO_ICON_EDIT);
}

/**
 *
 */
void NanoguiTimestepEditor::checkValidity()
{
    bool ok {false};    //

    ok = (0 < this->_label->value().size());
    ok = ok && (0 < this->_description->value().size());
    ok = ok && (0.0 <= this->_value->value());  // 0 means "events" step

    this->_okButton->setEnabled(ok);
}

/**
 *
 */
void NanoguiTimestepEditor::setSaveCallback(const std::function<void(const std::string&, double, const std::string&, bool)>& callback)
{
    this->_saveCallback = callback;
}

/**
 *
 */
void NanoguiTimestepEditor::setDeleteCallback(const std::function<void()>& callback)
{
    this->_deleteCallback = callback;
}

/**
 *
 */
void NanoguiTimestepEditor::setDefaultCallback(const std::function<void()>& callback)
{
    this->_defaultCallback = callback;
}

/**
 *
 */
const std::string& NanoguiTimestepEditor::getIdentifier() const
{
    return this->_identifier;
}

/**
 *
 */
void NanoguiTimestepEditor::draw(NVGcontext* ctx)
{
    float posX {0.f};   //
    float posY {0.f};   //
    float sizeY {0.f};  //
    nanogui::Color color{200, 200, 200, 100};   //
    if
        (0 >= this->_identifier.size())
    {
        // Use a different color for the left bar when the edited timespan has
        // not been added to the timeline configuration yet, i.e. when its
        // identifier is empty
        color = nanogui::Color(255, 255, 100, 255);
    }

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeY = static_cast<float>(this->mSize.y());

    nanogui::Widget::draw(ctx);

    nvgBeginPath(ctx);
    nvgMoveTo(ctx, posX + 20.f, posY);
    nvgLineTo(ctx, posX + 20.f, posY + sizeY);
    nvgStrokeColor(ctx, color);
    nvgStroke(ctx);
}

/**
 *
 */
void NanoguiTimestepEditor::uncheckDefault()
{
    this->_default->setChecked(false);
}

/**
 *
 */
void NanoguiTimestepEditor::checkDefault()
{
    this->_default->setChecked(true);
}
