// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

#include "ui/nanogui/nanogui_timestep_settings.h"
#include "core/log.h"

using namespace SEAScope;

/**
 *
 */
NanoguiTimestepSettings::NanoguiTimestepSettings(nanogui::Widget* parent)
: nanogui::Widget(parent)
{
    this->setFixedWidth(250);
    this->setWidth(250);
    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                           nanogui::Alignment::Fill,
                                           2, 4));
}

/**
 *
 */
NanoguiTimestepSettings::~NanoguiTimestepSettings()
{

}

/**
 *
 */
bool NanoguiTimestepSettings::setTimestepChangedCallback(const std::function<void(const std::string&)>& callback)
{
    if
        (nullptr != this->_changeCallback)
    {
        return false;
    }

    this->_changeCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiTimestepSettings::init(const std::vector<std::string>& labels,
                                   const std::vector<std::string>& descriptions)
{
    nanogui::Label* lbl {nullptr};  //
	nanogui::Button* btn {nullptr};	//
    nanogui::Widget* itemContainer {nullptr};   //
	std::vector<std::string>::const_iterator labelsIt; //
	std::vector<std::string>::const_iterator descriptionsIt; //
	std::vector<nanogui::Button*>::iterator buttonsIt;	//
    std::ostringstream tooltip {""};    //

    std::uint8_t fontSize {14};
    const Eigen::Vector2i buttonSize {50, 20};

    if
        (0 < this->_buttons.size())
    {
        LOG(LogLevel::Error, "NanoguiTimestepSettings",
                 "NanoguiTimestepSettings can only be initialized once");
        return false;
    }

	if
		(labels.size() != descriptions.size())
	{
		LOG( LogLevel::Error, "NanoguiTimestepSettings"
				, "Each timestep must have a label AND a description, but the number of labels ({}) and descriptions ({}) do not match"
				, labels.size(), descriptions.size());
		return false;
	}

    labelsIt = labels.begin();
    descriptionsIt = descriptions.begin();
    for
        (; labelsIt != labels.end(); ++labelsIt, ++descriptionsIt)
    {
        tooltip.str("");
        tooltip << *labelsIt << ":\n" << *descriptionsIt;

        itemContainer = new nanogui::Widget(this);
        itemContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                        nanogui::Alignment::Minimum,
                                                        0, 4));
		btn = new nanogui::Button(itemContainer, *labelsIt);
		btn->setFlags(nanogui::Button::RadioButton);
		btn->setFixedSize(buttonSize);
		btn->setFontSize(fontSize);
		btn->setTooltip(tooltip.str());
		btn->setChangeCallback([this, btn](bool selected) {
			if
				(  selected
                && (nullptr != this->_changeCallback))
			{
                this->_changeCallback(btn->caption());
			}
		});
		this->_buttons.push_back(btn);

        lbl = new nanogui::Label(itemContainer, *descriptionsIt);
        lbl->setFixedWidth(190);
    }

	buttonsIt = this->_buttons.begin();
	for
		(;buttonsIt != this->_buttons.end(); ++buttonsIt)
	{
		(*buttonsIt)->setButtonGroup(this->_buttons);
	}
    return false;
}

/**
 *
 */
bool NanoguiTimestepSettings::selectTimestep(const std::string& label)
{
	std::vector<nanogui::Button*>::iterator buttonsIt;	//
	bool pushed {false};    //
	bool found {false}; //

	found = false;
	buttonsIt = this->_buttons.begin();
	for
		(;buttonsIt != this->_buttons.end(); ++buttonsIt)
	{
		pushed = (label == (*buttonsIt)->caption());
		found = found || pushed;
		(*buttonsIt)->setPushed(pushed);
	}
	return found;
}
