// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include <nanogui/widget.h>
#include <nanogui/button.h>

#include "ui/nanogui/nanogui_toolbar.h"
#include "core/app_events_dispatcher.h"
#include "core/utils.h"

#include "ui/nanogui/nanogui_icon.h"
#include "ui/nanogui/nanogui_draggable_pane.h"
#include "ui/nanogui/nanogui_annotations_collection_creator.h"
#include "ui/nanogui/nanogui_anchored.h"

#include "events/zoom_in_request.h"
#include "events/zoom_out_request.h"
#include "events/camera_reset_request.h"
#include "events/camera_aligned_with_poles_request.h"
#include "events/camera_poles_lock_request.h"
#include "events/create_annotations_collection_request.h"
#include "events/variables_added_event.h"
#include "events/collection_removed_event.h"
#include "events/add_default_time_to_load_annotations_from_file_request.h"
#include "events/save_annotations_to_file_request.h"
#include "events/enable_shape_creation_request.h"
#include "events/disable_shape_creation_request.h"
#include "events/enable_shape_edition_request.h"
#include "events/scene_view_controller_ready_event.h"
#include "events/drawing_controller_ready_event.h"
#include "events/change_rasterization_mode_request.h"
#include "events/poles_axis_lock_event.h"
#include "events/toggle_shape_creation_request.h"

#include "core/rendering_cfg.h"
#include "ui/nanogui/nanogui_combobox.h"
#include "ui/nanogui/nanogui_path_selector.h"
#include "ui/nanogui/nanogui_annotation_collection_selector.h"
#include "core/path.h"
#include "core/log.h"

#include "ui/nanogui/nanogui_painted_label.h"
#include "events/drawing_actions_updated_event.h"
#include "core/interfaces/iinput_events_subscriber.h"
#include "core/input_events_dispatcher.h"
#include "core/input_bindings.h"

#include "tools/scene_view_controller_original.h"
#include "tools/drawing_controller.h"


using namespace SEAScope;

/**
 *
 */
NanoguiToolbar::NanoguiToolbar(AppEventsDispatcher& appEventsDispatcher, const InputEventsDispatcher& inputEventsDispatcher)
: _appEventsDispatcher(appEventsDispatcher),
  _inputEventsDispatcher(inputEventsDispatcher),
  _window(nullptr),
  _headerbar(nullptr),
  _preButtonsSpacer(nullptr),
  _postButtonsSpacer(nullptr),
  _reduceButton(nullptr),
  _expandButton(nullptr),
  _buttons(nullptr),
  _settings(nullptr),
  _instructions(nullptr),
  _pathBlock(nullptr),
  _pathSelectorLabel(nullptr),
  _pathSelector(nullptr),
  _loadButtonsBlock(nullptr),
  _loadDataButton(nullptr),
  _saveButtonsBlock(nullptr),
  _saveDataButton(nullptr),
  _polylineButton(nullptr),
  _polygonButton(nullptr),
  _pointButton(nullptr),
  _importButton(nullptr),
  _exportButton(nullptr),
  _zoomInButton(nullptr),
  _zoomOutButton(nullptr),
  _cameraResetButton(nullptr),
  _lockCameraAxisButton(nullptr),
  _rasterizationButton(nullptr),
  _rasterizationDropdown(nullptr),
  _fullModeButton(nullptr),
  _wireframeModeButton(nullptr),
  _dotModeButton(nullptr),
  _ioFormatBlock(nullptr),
  _collectionsBlock(nullptr),
  _pointSelector(nullptr),
  _polylineSelector(nullptr),
  _polygonSelector(nullptr),
  _titleBlock(nullptr),
  _drawingCollectionsTitle(nullptr),
  _loadCollectionsTitle(nullptr),
  _saveCollectionsTitle(nullptr),
  _instructionsLabel(nullptr),
  _defaultTheme(nullptr),
  _pushedTheme(nullptr),
  _defaultCollectionComboBoxValue("")
{
    this->_appEventsDispatcher.subscribe<VariablesAddedEvent>(this);
    this->_appEventsDispatcher.subscribe<CollectionRemovedEvent>(this);
    this->_appEventsDispatcher.subscribe<EnableShapeEditionRequest>(this);
    this->_appEventsDispatcher.subscribe<DrawingActionsUpdatedEvent>(this);
    this->_appEventsDispatcher.subscribe<SceneViewControllerReadyEvent>(this);
    this->_appEventsDispatcher.subscribe<ChangeRasterizationModeRequest>(this);
    this->_appEventsDispatcher.subscribe<PolesAxisLockEvent>(this);
    this->_appEventsDispatcher.subscribe<ToggleShapeCreationRequest>(this);
    this->_appEventsDispatcher.subscribe<DrawingControllerReadyEvent>(this);
}

/**
 *
 */
NanoguiToolbar::~NanoguiToolbar()
{
    this->_drawingButtons.clear();
    this->_rasterizationButtons.clear();
    this->_ioFormatButtons.clear();

    this->_appEventsDispatcher.unsubscribe<VariablesAddedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<CollectionRemovedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<EnableShapeEditionRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DrawingActionsUpdatedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<SceneViewControllerReadyEvent>(this);
    this->_appEventsDispatcher.unsubscribe<ChangeRasterizationModeRequest>(this);
    this->_appEventsDispatcher.unsubscribe<PolesAxisLockEvent>(this);
    this->_appEventsDispatcher.unsubscribe<ToggleShapeCreationRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DrawingControllerReadyEvent>(this);
}

/**
 *
 */
bool NanoguiToolbar::init(nanogui::Widget* screen)
{
    if
        (nullptr !=  this->_window)
    {
        // Already initialized
        return false;
    }

    this->_defaultTheme = new nanogui::Theme(screen->screen()->nvgContext());
    this->_pushedTheme = new nanogui::Theme(screen->screen()->nvgContext());
    this->_pushedTheme->mButtonGradientTopPushed = {180, 180, 75, 200};
    this->_pushedTheme->mButtonGradientBotPushed = {200, 200, 75, 255};
    this->_pushedTheme->mTextColor = {30, 30, 30, 255};

	std::uint32_t headerHeight = 24;	//
    std::uint32_t windowWidth = 346; // 3 + 20 + 6 + 13 * 20 + 3 * 10 + 6 + 18 + 3;

	this->_window = new NanoguiDraggablePane(screen);
    this->_window->backgroundColor = nanogui::Color(30, 30, 30, 255);
	this->_window->headerHeight = headerHeight;

	this->_headerbar = new nanogui::Widget(this->_window);
	this->_headerbar->setFixedSize(Eigen::Vector2i(windowWidth, headerHeight));
	this->_headerbar->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
												nanogui::Alignment::Middle,
												2, 3));

	NanoguiIcon* icon(nullptr);
    nanogui::Widget* spacer(nullptr);

    icon = new NanoguiIcon(this->_headerbar, ENTYPO_ICON_TOOLS);
	icon->setFixedSize(Eigen::Vector2i(20, 20));
	icon->setFontSize(18);
    icon->setBackgroundColor(nanogui::Color(48, 139, 87, 255));
	icon->setColor(nanogui::Color(255,255,255,255));
	icon->setTooltip("Tools");

    this->_preButtonsSpacer = new nanogui::Widget(this->_headerbar);
    this->_preButtonsSpacer->setFixedSize(Eigen::Vector2i(6, 24));

    this->_buttons = new nanogui::Widget(this->_headerbar);
    this->_buttons->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                     nanogui::Alignment::Middle,
                                                     0, 0));

    icon  = new NanoguiIcon(this->_buttons, ENTYPO_ICON_EDIT);
	icon->setFixedSize(Eigen::Vector2i(20, 20));
	icon->setFontSize(18);
	icon->setColor(nanogui::Color(255,255,255,255));
	icon->setTooltip("Drawing tools");

    this->_pointButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_HAIR_CROSS);
    this->_pointButton->setTooltip("Draw points");
    this->_pointButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_pointButton->setChangeCallback([this] (bool pushed) {
        this->_defaultCollectionComboBoxValue = "";
        this->updateCollectionsChoice();
        this->_pathSelector->hide();
        this->_pathBlock->setVisible(false);
        this->_ioFormatBlock->setVisible(false);
        this->_loadButtonsBlock->setVisible(false);
        this->_saveButtonsBlock->setVisible(false);
        if
            (false == pushed)
        {
            DisableShapeCreationRequest outEvent(ShapeType::POINT);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(outEvent);

            this->_collectionsBlock->setVisible(false);
            this->_pointSelector->hideDropDown();
            this->_drawingCollectionsTitle->setVisible(false);
            this->_titleBlock->setVisible(false);
        }
        else
        {
            if
                ("" != this->_pointSelector->getChoice())
            {
                CollectionVariable colvar;
                bool ok = this->getSelectedCollectionVariable(colvar, RenderMethod::POINT,
                                                              this->_pointSelector->getChoice());
                if
                    (true == ok)
                {
                    EnableShapeCreationRequest outEvent(ShapeType::POINT, colvar);
                    this->_appEventsDispatcher.dispatch<EnableShapeCreationRequest>(outEvent);
                }
            }
            this->_drawingCollectionsTitle->setVisible(true);
            this->_loadCollectionsTitle->setVisible(false);
            this->_saveCollectionsTitle->setVisible(false);
            this->_titleBlock->setVisible(true);

            this->_pointSelector->allowNewCollection(true);
            this->_polylineSelector->allowNewCollection(true);
            this->_polygonSelector->allowNewCollection(true);

            this->_pointSelector->enable();

            this->_pointSelector->setVisible(true);
            this->_polylineSelector->setVisible(false);
            this->_polygonSelector->setVisible(false);

            this->_collectionsBlock->setVisible(true);

            this->_collectionCreator->setVisible(false);
        }

        this->updatePushButtonColors();
        this->_window->updateLayout();
    });

    this->_polylineButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_LINE_GRAPH);
	this->_polylineButton->setTooltip("Draw polylines");
	this->_polylineButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_polylineButton->setChangeCallback([this] (bool pushed) {
        this->_defaultCollectionComboBoxValue = "";
        this->updateCollectionsChoice();
        this->_pathSelector->hide();
        this->_pathBlock->setVisible(false);
        this->_ioFormatBlock->setVisible(false);
        this->_loadButtonsBlock->setVisible(false);
        this->_saveButtonsBlock->setVisible(false);
        if
            (false == pushed)
        {
            DisableShapeCreationRequest outEvent(ShapeType::POLYLINE);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(outEvent);

            this->_collectionsBlock->setVisible(false);
            this->_polylineSelector->hideDropDown();
            this->_drawingCollectionsTitle->setVisible(false);
            this->_titleBlock->setVisible(false);
        }
        else
        {
            if
                ("" != this->_polylineSelector->getChoice())
            {
                CollectionVariable colvar;
                bool ok = this->getSelectedCollectionVariable(colvar, RenderMethod::POLYLINE,
                                                              this->_polylineSelector->getChoice());

                if
                    (true == ok)
                {
                    EnableShapeCreationRequest outEvent(ShapeType::POLYLINE, colvar);
                    this->_appEventsDispatcher.dispatch<EnableShapeCreationRequest>(outEvent);
                }
            }
            this->_drawingCollectionsTitle->setVisible(true);
            this->_loadCollectionsTitle->setVisible(false);
            this->_saveCollectionsTitle->setVisible(false);
            this->_titleBlock->setVisible(true);

            this->_pointSelector->allowNewCollection(true);
            this->_polylineSelector->allowNewCollection(true);
            this->_polygonSelector->allowNewCollection(true);

            this->_polylineSelector->enable();

            this->_pointSelector->setVisible(false);
            this->_polylineSelector->setVisible(true);
            this->_polygonSelector->setVisible(false);

            this->_collectionsBlock->setVisible(true);

            this->_collectionCreator->setVisible(false);
        }

        this->updatePushButtonColors();
        this->_window->updateLayout();
    });

    this->_polygonButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_DOCUMENT);
	this->_polygonButton->setTooltip("Draw polygons");
	this->_polygonButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_polygonButton->setChangeCallback([this] (bool pushed) {
        this->_defaultCollectionComboBoxValue = "";
        this->updateCollectionsChoice();
        this->_pathSelector->hide();
        this->_pathBlock->setVisible(false);
        this->_ioFormatBlock->setVisible(false);
        this->_loadButtonsBlock->setVisible(false);
        this->_saveButtonsBlock->setVisible(false);
        if
            (false == pushed)
        {
            DisableShapeCreationRequest outEvent(ShapeType::POLYGON);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(outEvent);

            this->_collectionsBlock->setVisible(false);
            this->_polygonSelector->hideDropDown();
            this->_drawingCollectionsTitle->setVisible(false);
            this->_titleBlock->setVisible(false);
        }
        else
        {
            if
                ("" != this->_polygonSelector->getChoice())
            {
                CollectionVariable colvar;
                bool ok = this->getSelectedCollectionVariable(colvar, RenderMethod::POLYGON,
                                                              this->_polygonSelector->getChoice());
                if
                    (true == ok)
                {
                    EnableShapeCreationRequest outEvent(ShapeType::POLYGON, colvar);
                    this->_appEventsDispatcher.dispatch<EnableShapeCreationRequest>(outEvent);
                }
            }
            this->_drawingCollectionsTitle->setVisible(true);
            this->_loadCollectionsTitle->setVisible(false);
            this->_saveCollectionsTitle->setVisible(false);
            this->_titleBlock->setVisible(true);

            this->_pointSelector->allowNewCollection(true);
            this->_polylineSelector->allowNewCollection(true);
            this->_polygonSelector->allowNewCollection(true);

            this->_polygonSelector->enable();

            this->_pointSelector->setVisible(false);
            this->_polylineSelector->setVisible(false);
            this->_polygonSelector->setVisible(true);

            this->_collectionsBlock->setVisible(true);

            this->_collectionCreator->setVisible(false);
        }

        this->updatePushButtonColors();
        this->_window->updateLayout();
    });

    this->_drawingButtons.push_back(this->_pointButton);
    this->_drawingButtons.push_back(this->_polygonButton);
    this->_drawingButtons.push_back(this->_polylineButton);
    this->_pointButton->setFlags(nanogui::Button::Flags::RadioButton|nanogui::Button::Flags::ToggleButton);
    this->_pointButton->setButtonGroup(this->_drawingButtons);
    this->_polygonButton->setFlags(nanogui::Button::Flags::RadioButton|nanogui::Button::Flags::ToggleButton);
    this->_polygonButton->setButtonGroup(this->_drawingButtons);
    this->_polylineButton->setFlags(nanogui::Button::Flags::RadioButton|nanogui::Button::Flags::ToggleButton);
    this->_polylineButton->setButtonGroup(this->_drawingButtons);

    spacer = new nanogui::Widget(this->_buttons);
    spacer->setFixedWidth(10);

    icon = new NanoguiIcon(this->_buttons, ENTYPO_ICON_DATABASE);
	icon->setFixedSize(Eigen::Vector2i(20, 20));
	icon->setFontSize(18);
	icon->setColor(nanogui::Color(255,255,255,255));
	icon->setTooltip("Import/export tools");

    this->_importButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_FOLDER);
	this->_importButton->setTooltip("Import user shapes.");
	this->_importButton->setFixedSize(Eigen::Vector2i(20,20));
    this->_importButton->setCallback([this] () {
        this->_defaultCollectionComboBoxValue = "Ignored";
        this->updateCollectionsChoice();

        if
            (true == this->_pointButton->pushed())
        {
            this->_pointButton->setPushed(false);
            DisableShapeCreationRequest outEvent(ShapeType::POINT);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(outEvent);
        }

        if
            (true == this->_polygonButton->pushed())
        {
            this->_polygonButton->setPushed(false);
            DisableShapeCreationRequest outEvent(ShapeType::POLYGON);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(outEvent);
        }

        if
            (true == this->_polylineButton->pushed())
        {
            this->_polylineButton->setPushed(false);
            DisableShapeCreationRequest outEvent(ShapeType::POLYLINE);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(outEvent);
        }

        this->_drawingCollectionsTitle->setVisible(false);
        this->_loadCollectionsTitle->setVisible(true);
        this->_saveCollectionsTitle->setVisible(false);
        this->_titleBlock->setVisible(true);

        this->_pointSelector->allowNewCollection(true);
        this->_polylineSelector->allowNewCollection(true);
        this->_polygonSelector->allowNewCollection(true);

        this->_pathSelectorLabel->setCaption("Select input file");
        this->_pathSelector->show();
        this->_pathBlock->setVisible(true);
        this->_ioFormatBlock->setVisible(true);
        this->_loadButtonsBlock->setVisible(true);
        this->_saveButtonsBlock->setVisible(false);
        this->_collectionsBlock->setVisible(true);
        this->_pointSelector->setVisible(true);
        this->_polylineSelector->setVisible(true);
        this->_polygonSelector->setVisible(true);

        this->_pathSelector->setType(PathSelectorType::get_file);

        this->updatePushButtonColors();
        this->_window->updateLayout();
    });

    this->_exportButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_SAVE);
	this->_exportButton->setTooltip("Export user shapes.");
	this->_exportButton->setFixedSize(Eigen::Vector2i(20,20));
    this->_exportButton->setCallback([this] () {
        this->_defaultCollectionComboBoxValue = "Ignored";
        this->updateCollectionsChoice();

        if
            (true == this->_pointButton->pushed())
        {
            this->_pointButton->setPushed(false);
            DisableShapeCreationRequest outEvent(ShapeType::POINT);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(outEvent);
        }

        if
            (true == this->_polygonButton->pushed())
        {
            this->_polygonButton->setPushed(false);
            DisableShapeCreationRequest outEvent(ShapeType::POLYGON);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(outEvent);
        }

        if
            (true == this->_polylineButton->pushed())
        {
            this->_polylineButton->setPushed(false);
            DisableShapeCreationRequest outEvent(ShapeType::POLYLINE);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(outEvent);
        }

        this->_drawingCollectionsTitle->setVisible(false);
        this->_loadCollectionsTitle->setVisible(false);
        this->_saveCollectionsTitle->setVisible(true);
        this->_titleBlock->setVisible(true);

        this->_pointSelector->allowNewCollection(false);
        this->_polylineSelector->allowNewCollection(false);
        this->_polygonSelector->allowNewCollection(false);

        this->_collectionCreator->setVisible(false);

        this->_pathSelectorLabel->setCaption("Select output file");
        this->_pathSelector->show();
        this->_pathBlock->setVisible(true);
        this->_ioFormatBlock->setVisible(true);
        this->_loadButtonsBlock->setVisible(false);
        this->_saveButtonsBlock->setVisible(true);
        this->_collectionsBlock->setVisible(true);
        this->_pointSelector->setVisible(true);
        this->_polylineSelector->setVisible(true);
        this->_polygonSelector->setVisible(true);

        this->_pathSelector->setType(PathSelectorType::save_file);

        this->updatePushButtonColors();
        this->_window->updateLayout();
    });

    spacer = new nanogui::Widget(this->_buttons);
    spacer->setFixedWidth(10);

    icon = new NanoguiIcon(this->_buttons, ENTYPO_ICON_MAGNIFYING_GLASS);
	icon->setFixedSize(Eigen::Vector2i(20, 20));
	icon->setFontSize(18);
	icon->setColor(nanogui::Color(255,255,255,255));
	icon->setTooltip("Camera tools");

	this->_zoomInButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_PLUS);
	this->_zoomInButton->setTooltip("Zoom in");
	this->_zoomInButton->setFixedSize(Eigen::Vector2i(20, 20));
	this->_zoomInButton->setCallback([this] () {
        ZoomInRequest event;
        this->_appEventsDispatcher.dispatch<ZoomInRequest>(event);
	});

	this->_zoomOutButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_MINUS);
	this->_zoomOutButton->setTooltip("Zoom out");
	this->_zoomOutButton->setFixedSize(Eigen::Vector2i(20, 20));
	this->_zoomOutButton->setCallback([this] () {
        ZoomOutRequest event;
        this->_appEventsDispatcher.dispatch<ZoomOutRequest>(event);
	});

	this->_cameraResetButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_BACK);
	this->_cameraResetButton->setTooltip("Reset zoom and location");
	this->_cameraResetButton->setFixedSize(Eigen::Vector2i(20, 20));
	this->_cameraResetButton->setCallback([this] () {
        CameraResetRequest event;
        this->_appEventsDispatcher.dispatch<CameraResetRequest>(event);
	});

	this->_lockCameraAxisButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_BLOCK);
	this->_lockCameraAxisButton->setTooltip("Prevent rotation on the Z axis");
	this->_lockCameraAxisButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_lockCameraAxisButton->setFlags(nanogui::Button::ToggleButton);
	this->_lockCameraAxisButton->setPushed(true); /* Locked by default */
	this->_lockCameraAxisButton->setChangeCallback([this] (bool pushed) {
        CameraPolesLockRequest event(pushed);
        this->_appEventsDispatcher.dispatch<CameraPolesLockRequest>(event);

        this->updatePushButtonColors();
	});

	this->_rasterizationButton = new nanogui::Button(this->_buttons, "", ENTYPO_ICON_GLOBE);
	this->_rasterizationButton->setTooltip("Choose rasterization mode");
	this->_rasterizationButton->setFixedSize(Eigen::Vector2i(20, 20));
	this->_rasterizationButton->setCallback([this] () {
        this->_rasterizationDropdown->setVisible(!this->_rasterizationDropdown->visible());
	});

    this->_rasterizationDropdown = NanoguiAnchored::create(this->_rasterizationButton);
    this->_rasterizationDropdown->setFixedSize(Eigen::Vector2i(20, 60));
    this->_rasterizationDropdown->setSize(Eigen::Vector2i(20, 60));
    this->_rasterizationDropdown->offset = 0u;
    this->_rasterizationDropdown->placement = NanoguiAnchored::ON_BOTTOM;

    this->_fullModeButton = new nanogui::Button(this->_rasterizationDropdown, "F");
    this->_fullModeButton->setPosition(Eigen::Vector2i(0, 0));
    this->_fullModeButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_fullModeButton->setSize(Eigen::Vector2i(20, 20));

    this->_wireframeModeButton = new nanogui::Button(this->_rasterizationDropdown, "W");
    this->_wireframeModeButton->setPosition(Eigen::Vector2i(0, 20));
    this->_wireframeModeButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_wireframeModeButton->setSize(Eigen::Vector2i(20, 20));

    this->_dotModeButton = new nanogui::Button(this->_rasterizationDropdown, "V");
    this->_dotModeButton->setPosition(Eigen::Vector2i(0, 40));
    this->_dotModeButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_dotModeButton->setSize(Eigen::Vector2i(20, 20));

    this->_rasterizationButtons.push_back(this->_fullModeButton);
    this->_rasterizationButtons.push_back(this->_wireframeModeButton);
    this->_rasterizationButtons.push_back(this->_dotModeButton);
    this->_fullModeButton->setFlags(nanogui::Button::Flags::RadioButton);
    this->_fullModeButton->setButtonGroup(this->_rasterizationButtons);
    this->_fullModeButton->setCallback([this] () {
        ChangeRasterizationModeRequest event(OGL::fill);
        this->_appEventsDispatcher.dispatch<ChangeRasterizationModeRequest>(event);

        this->updatePushButtonColors();
    });
    this->_wireframeModeButton->setFlags(nanogui::Button::Flags::RadioButton);
    this->_wireframeModeButton->setButtonGroup(this->_rasterizationButtons);
    this->_wireframeModeButton->setCallback([this] () {
        ChangeRasterizationModeRequest event(OGL::line);
        this->_appEventsDispatcher.dispatch<ChangeRasterizationModeRequest>(event);

        this->updatePushButtonColors();
    });
    this->_dotModeButton->setFlags(nanogui::Button::Flags::RadioButton);
    this->_dotModeButton->setButtonGroup(this->_rasterizationButtons);
    this->_dotModeButton->setCallback([this] () {
        ChangeRasterizationModeRequest event(OGL::point);
        this->_appEventsDispatcher.dispatch<ChangeRasterizationModeRequest>(event);

        this->updatePushButtonColors();
    });

    // Full mode is selected by default
    this->_fullModeButton->setPushed(true);

    this->_rasterizationDropdown->setVisible(false);

    this->_postButtonsSpacer = new nanogui::Widget(this->_headerbar);
    this->_postButtonsSpacer->setFixedSize(Eigen::Vector2i(6, 24));

	this->_expandButton = new nanogui::Button(this->_headerbar, "", ENTYPO_ICON_ARROW_RIGHT);
	this->_expandButton->setTooltip("Show tools");
	this->_expandButton->setFixedSize(Eigen::Vector2i(18,18));
	this->_expandButton->setFontSize(16);
	this->_expandButton->setVisible(false);

	this->_reduceButton = new nanogui::Button(this->_headerbar, "", ENTYPO_ICON_ARROW_LEFT);
	this->_reduceButton->setTooltip("Hide tools");
	this->_reduceButton->setFixedSize(Eigen::Vector2i(18,18));
	this->_reduceButton->setFontSize(16);

	this->_reduceButton->setCallback([this]() {
        this->_headerbar->setFixedWidth(44); // 3 + 20 + 18 + 3
		this->_reduceButton->setVisible(false);
        this->_preButtonsSpacer->setVisible(false);
        this->_buttons->setVisible(false);
        this->_postButtonsSpacer->setVisible(false);
		this->_expandButton->setVisible(true);

        // Hide all dropdown menus
        this->_pointSelector->hideDropDown();
        this->_polygonSelector->hideDropDown();
        this->_polylineSelector->hideDropDown();
        this->_pathSelector->hideDropDown();

        this->_rasterizationDropdown->setVisible(false);

        this->_settings->setVisible(false);
        this->_instructions->setVisible(false);

		this->_window->updateLayout();
	});

	this->_expandButton->setCallback([this]() {
        this->_headerbar->setFixedWidth(346); // 3 + 20 + 6 + 13 * 20 + 3 * 10 + 6 + 18 + 3
		this->_reduceButton->setVisible(true);
        this->_preButtonsSpacer->setVisible(true);
        this->_buttons->setVisible(true);
        this->_postButtonsSpacer->setVisible(true);
        //this->_spacer->setVisible(true);
		this->_expandButton->setVisible(false);

        this->_settings->setVisible(true);

        if
            (0 < this->_instructionsLabel->caption().size())
        {
            this->_instructions->setVisible(true);
        }

		this->_window->updateLayout();
	});

    // Settings
    this->_settings = new nanogui::Widget(this->_window);
    this->_settings->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                      nanogui::Alignment::Minimum,
                                                      0, 0));

    // - collections selection title
    this->_titleBlock = new nanogui::Widget(this->_settings);
    this->_titleBlock->setFixedSize(Eigen::Vector2i(340, 20));
    this->_titleBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                        nanogui::Alignment::Fill,
                                                        2, 3));
    this->_titleBlock->setVisible(false);

    nanogui::Widget* collectionTitleBlock(nullptr);
    collectionTitleBlock = new nanogui::Widget(this->_titleBlock);
    collectionTitleBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                           nanogui::Alignment::Minimum,
                                                           0, 0));
    this->_drawingCollectionsTitle = new nanogui::Label(collectionTitleBlock, "Define where annotations will be saved", "sans-bold", 16);
    this->_drawingCollectionsTitle->setVisible(false);
    this->_loadCollectionsTitle = new nanogui::Label(collectionTitleBlock, "Define where annotations will be loaded", "sans-bold", 16);
    this->_loadCollectionsTitle->setVisible(false);
    this->_saveCollectionsTitle = new nanogui::Label(collectionTitleBlock, "Define which annotations will be exported", "sans-bold", 16);
    this->_saveCollectionsTitle->setVisible(false);

    // - collections selector
    this->_collectionsBlock = new nanogui::Widget(this->_settings);
    this->_collectionsBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                              nanogui::Alignment::Fill,
                                                              2, 3));

    // -- polygons
    this->_polygonSelector = new NanoguiAnnotationCollectionSelector(this->_collectionsBlock, ENTYPO_ICON_DOCUMENT, "polygon");
    this->_polygonSelector->setFixedSize(Eigen::Vector2i(340, 20));

    this->_polygonSelector->setChangeCallback([this] (bool /*useDefault*/, const std::string& value) {
        CollectionVariable colvar;
        bool ok = this->getSelectedCollectionVariable(colvar, RenderMethod::POLYGON, value);
        if
            (true == ok)
        {
            if
                (this->_polygonButton->pushed())
            {
                EnableShapeCreationRequest enableEvent(ShapeType::POLYGON, colvar);
                this->_appEventsDispatcher.dispatch<EnableShapeCreationRequest>(enableEvent);
            }
        }
        else
        {
            DisableShapeCreationRequest disableEvent(ShapeType::POLYGON);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(disableEvent);
        }

        this->updateLoadSaveButtons();
    });

    this->_polygonSelector->setDropDownCallback([this] () {
        this->_pointSelector->hideDropDown();
        this->_polylineSelector->hideDropDown();
        this->_pathSelector->hideDropDown();
    });

    this->_polygonSelector->setAddCallback([this] () {
        this->_collectionCreator->selectCollectionType("polygon");
        this->_collectionCreator->setVisible(true);
        this->_window->updateLayout();
    });

    // -- polylines
    this->_polylineSelector = new NanoguiAnnotationCollectionSelector(this->_collectionsBlock, ENTYPO_ICON_LINE_GRAPH, "polyline");
    this->_polylineSelector->setFixedSize(Eigen::Vector2i(340, 20));

    this->_polylineSelector->setChangeCallback([this] (bool /*useDefault*/, const std::string& value) {
        CollectionVariable colvar;
        bool ok = this->getSelectedCollectionVariable(colvar, RenderMethod::POLYLINE, value);
        if
            (true == ok)
        {
            if
                (this->_polylineButton->pushed())
            {
                EnableShapeCreationRequest enableEvent(ShapeType::POLYLINE, colvar);
                this->_appEventsDispatcher.dispatch<EnableShapeCreationRequest>(enableEvent);
            }
        }
        else
        {
            DisableShapeCreationRequest disableEvent(ShapeType::POLYLINE);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(disableEvent);
        }

        this->updateLoadSaveButtons();
    });

    this->_polylineSelector->setDropDownCallback([this] () {
        this->_polygonSelector->hideDropDown();
        this->_pointSelector->hideDropDown();
        this->_pathSelector->hideDropDown();
    });

    this->_polylineSelector->setAddCallback([this] () {
        this->_collectionCreator->selectCollectionType("polyline");
        this->_collectionCreator->setVisible(true);
        this->_window->updateLayout();
    });

    // -- points
    this->_pointSelector = new NanoguiAnnotationCollectionSelector(this->_collectionsBlock, ENTYPO_ICON_HAIR_CROSS, "point");
    this->_pointSelector->setFixedSize(Eigen::Vector2i(340, 20));

    this->_pointSelector->setChangeCallback([this] (bool /*useDefault*/, const std::string& value) {
        CollectionVariable colvar;
        bool ok = this->getSelectedCollectionVariable(colvar, RenderMethod::POINT, value);
        if
            (true == ok)
        {
            if
                (this->_pointButton->pushed())
            {
                EnableShapeCreationRequest enableEvent(ShapeType::POINT, colvar);
                this->_appEventsDispatcher.dispatch<EnableShapeCreationRequest>(enableEvent);
            }
        }
        else
        {
            DisableShapeCreationRequest disableEvent(ShapeType::POINT);
            this->_appEventsDispatcher.dispatch<DisableShapeCreationRequest>(disableEvent);
        }

        this->updateLoadSaveButtons();
    });

    this->_pointSelector->setDropDownCallback([this] () {
        this->_polygonSelector->hideDropDown();
        this->_polylineSelector->hideDropDown();
        this->_pathSelector->hideDropDown();
    });

    this->_pointSelector->setAddCallback([this] () {
        this->_collectionCreator->selectCollectionType("point");
        this->_collectionCreator->setVisible(true);
        this->_window->updateLayout();
    });

    spacer = new nanogui::Widget(this->_collectionsBlock);
    spacer->setFixedHeight(5);

    // - I/O format selector
    this->_ioFormatBlock = new nanogui::Widget(this->_settings);
    this->_ioFormatBlock->setFixedWidth(340);
    this->_ioFormatBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                           nanogui::Alignment::Fill,
                                                           2, 3));

    new nanogui::Label(this->_ioFormatBlock, "Format", "sans-bold", 16);

    nanogui::Widget* fmtContainer(nullptr);
    fmtContainer = new nanogui::Widget(this->_ioFormatBlock);
    fmtContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                   nanogui::Alignment::Middle,
                                                   0, 0));

    spacer = new nanogui::Widget(this->_ioFormatBlock);
    spacer->setFixedHeight(5);

    nanogui::Widget* fmtButtons(nullptr);
    fmtButtons = new nanogui::Widget(fmtContainer);
    fmtButtons->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                 nanogui::Alignment::Middle,
                                                 0, 0));
    this->_ioFormatButtons.reserve(2);

    this->_ioFormatButtons.emplace_back(new nanogui::Button(fmtButtons, "JSON"));
    this->_ioFormatButtons[0]->setFixedSize(Eigen::Vector2i(50, 20));
    this->_ioFormatButtons[0]->setFontSize(16);
    this->_ioFormatButtons[0]->setFlags(nanogui::Button::RadioButton);

    this->_ioFormatButtons.emplace_back(new nanogui::Button(fmtButtons, "WKT"));
    this->_ioFormatButtons[1]->setFixedSize(Eigen::Vector2i(50, 20));
    this->_ioFormatButtons[1]->setFontSize(16);
    this->_ioFormatButtons[1]->setFlags(nanogui::Button::RadioButton);

    this->_ioFormatButtons[0]->setButtonGroup(this->_ioFormatButtons);
    this->_ioFormatButtons[1]->setButtonGroup(this->_ioFormatButtons);

    this->_ioFormatButtons[0]->setPushed(true);


    this->initPathSelector(this->_settings);

    this->initLoadDataWidgets(this->_settings);
    this->initSaveDataWidgets(this->_settings);

    this->initCollectionsCreator(this->_settings);


    // Instructions
    this->_instructions = new nanogui::Widget(this->_window);
    this->_instructions->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                           nanogui::Alignment::Minimum,
                                                           0, 2));

    nanogui::Label* instructionsTitle = new nanogui::Label(this->_instructions, "Possible actions:", "sans-bold");
    instructionsTitle->setFontSize(16);

    nanogui::Widget* instructionsContainer = new nanogui::Widget(this->_instructions);
    instructionsContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                            nanogui::Alignment::Minimum,
                                                            3, 2));

    this->_instructionsLabel = new nanogui::Label(instructionsContainer, "");
    this->_instructionsLabel->setFixedWidth(340);

    // Hide components by default
    this->_instructions->setVisible(false);
    this->_pathBlock->setVisible(false);
    this->_ioFormatBlock->setVisible(false);
    this->_loadButtonsBlock->setVisible(false);
    this->_saveButtonsBlock->setVisible(false);
    this->_collectionsBlock->setVisible(false);
    this->_collectionCreator->setVisible(false);

    this->updatePushButtonColors();
	this->_window->updateLayout();

    return true;
}

/**
 *
 */
void NanoguiToolbar::showAt(const Eigen::Vector2i& pos)
{
    this->_window->display(pos);
}

/**
 *
 */
bool NanoguiToolbar::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(VariablesAddedEvent)) == eventType)
    {
        bool changed(false);
        const std::map<RenderingCfg*, std::string>& variables = static_cast<const VariablesAddedEvent&>(event).getVariables();
        for
            (const auto& variable : variables)
        {
            CollectionVariable colvar(variable.first->target.collection, variable.first->target.variable);
            //std::cout << "+ VARIABLE " << variable.second << " " << variable.first->target.collection.first << "." << variable.first->target.collection.second << ":" << variable.first->target.variable << " = " << int(variable.first->renderMethod()) << std::endl;
            this->_variablesByRenderMethod[variable.first->renderMethod()];

            auto found = this->_variablesByRenderMethod[variable.first->renderMethod()].find(colvar);
            if
                ( (this->_variablesByRenderMethod[variable.first->renderMethod()].end() == found)
                || (found->second != variable.second))
            {
                this->_variablesByRenderMethod[variable.first->renderMethod()][colvar] = variable.second;
                changed = true;
            }
        }

        if
            (true == changed)
        {
            this->updateCollectionsChoice();
        }
        return true;
    }
    else if
        (std::type_index(typeid(CollectionRemovedEvent)) == eventType)
    {
        bool changed(false);
        const CollectionId& colId = static_cast<const CollectionRemovedEvent&>(event).getCollectionId();
        const std::vector<VariableId>& variables = static_cast<const CollectionRemovedEvent&>(event).getVariableIds();
        for
            (const auto& varId : variables)
        {
            CollectionVariable colvar(colId, varId);
            //std::cout << "- VARIABLE " << colId.first << "." << colId.second << ":" << varId << std::endl;
            for
                (auto& it: this->_variablesByRenderMethod)
            {
                auto found = it.second.find(colvar);
                if
                    (it.second.end() != found)
                {
                    it.second.erase(found);
                    changed = true;

                    // A variable should only be associated to one rendering method
                    break;
                }
            }
        }

        if
            (true == changed)
        {
            this->updateCollectionsChoice();
        }
        return true;
    }
    else if
        (std::type_index(typeid(EnableShapeEditionRequest)) == eventType)
    {
        if
            (true == this->_pathSelector->visible())
        {
            // Either the import or export views are currently displayed and
            // they are not incompatible with shape edition so do nothing
            return true;
        }

        this->_pointButton->setPushed(false);
        this->_polygonButton->setPushed(false);
        this->_polylineButton->setPushed(false);

        this->_polygonSelector->hideDropDown();
        this->_polylineSelector->hideDropDown();
        this->_pointSelector->hideDropDown();

        this->_collectionsBlock->setVisible(false);

        this->updatePushButtonColors();
        this->_window->updateLayout();

        return true;
    }
    else if
        (std::type_index(typeid(DrawingActionsUpdatedEvent)) == eventType)
    {
        const ActorId& actorId = static_cast<const DrawingActionsUpdatedEvent&>(event).getActorId();

        const std::vector<IInputEventsSubscriber*>& subscribers = this->_inputEventsDispatcher.getSubscribers();
        IInputEventsSubscriber* drawingActor = nullptr;
        for
            (const auto& subscriber: subscribers)
        {
            if
                (   (subscriber->getId() == actorId)
                 && (true == subscriber->isActive()))
            {
                drawingActor = subscriber;
                break;
            }
        }

        if
            (nullptr == drawingActor)
        {
            return true;
        }

        const std::set<ActionType>& possibleActions = drawingActor->getPossibleActions();
        const InputBindings& inputBindings = this->_inputEventsDispatcher.getInputBindings();
        std::ostringstream instructions("");
        for
            (const auto& actionType : possibleActions)
        {
            if
                (  (static_cast<ActionType>(DrawingControllerActions::TOGGLE_POINT_CREATION) == actionType)
                || (static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYLINE_CREATION) == actionType)
                || (static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYGON_CREATION) == actionType))
            {
                continue;
            }

            std::string actionDescription;
            drawingActor->getActionLabel(actionType, actionDescription);

            std::string bindingDescription;
            inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);

            instructions << "  • " << bindingDescription << " to " << actionDescription << "\n";
        }

        this->_instructionsLabel->setCaption(instructions.str());
        if
            (0 < instructions.str().size())
        {
            this->_instructions->setVisible(true);
        }
        else
        {
            this->_instructions->setVisible(false);
        }

	    this->_window->updateLayout();
    }
    else if
        (std::type_index(typeid(DrawingControllerReadyEvent)) == eventType)
    {
        const ActorId& actorId = static_cast<const DrawingControllerReadyEvent&>(event).getActorId();

        const std::vector<IInputEventsSubscriber*>& subscribers = this->_inputEventsDispatcher.getSubscribers();
        IInputEventsSubscriber* drawingActor = nullptr;
        for
            (const auto& subscriber: subscribers)
        {
            if
                (   (subscriber->getId() == actorId)
                 && (true == subscriber->isActive()))
            {
                drawingActor = subscriber;
                break;
            }
        }

        if
            (nullptr == drawingActor)
        {
            return true;
        }

        const InputBindings& inputBindings = this->_inputEventsDispatcher.getInputBindings();

        std::ostringstream instructions("");
        std::string actionDescription;
        std::string bindingDescription;
        ActionType actionType;

        // Update point button tooltip
        instructions.str("");
        actionType = static_cast<ActionType>(DrawingControllerActions::TOGGLE_POINT_CREATION);
        drawingActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_pointButton->setTooltip(instructions.str());

        // Update polyline button tooltip
        instructions.str("");
        actionType = static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYLINE_CREATION);
        drawingActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_polylineButton->setTooltip(instructions.str());

        // Update polygon button tooltip
        instructions.str("");
        actionType = static_cast<ActionType>(DrawingControllerActions::TOGGLE_POLYGON_CREATION);
        drawingActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_polygonButton->setTooltip(instructions.str());

        return true;
    }
    else if
        (std::type_index(typeid(SceneViewControllerReadyEvent)) == eventType)
    {
        const ActorId& actorId = static_cast<const SceneViewControllerReadyEvent&>(event).getActorId();

        const std::vector<IInputEventsSubscriber*>& subscribers = this->_inputEventsDispatcher.getSubscribers();
        IInputEventsSubscriber* sceneViewActor = nullptr;
        for
            (const auto& subscriber: subscribers)
        {
            if
                (   (subscriber->getId() == actorId)
                 && (true == subscriber->isActive()))
            {
                sceneViewActor = subscriber;
                break;
            }
        }

        if
            (nullptr == sceneViewActor)
        {
            return true;
        }

        const InputBindings& inputBindings = this->_inputEventsDispatcher.getInputBindings();

        std::ostringstream instructions("");
        std::string actionDescription;
        std::string bindingDescription;
        ActionType actionType;

        // Update zoom-in button tooltip
        instructions.str("");
        actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN);
        sceneViewActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_zoomInButton->setTooltip(instructions.str());

        instructions.str("");
        actionType =static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT);
        sceneViewActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_zoomOutButton->setTooltip(instructions.str());

        instructions.str("");
        actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::TOGGLE_POLES_AXIS_LOCKING);
        sceneViewActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_lockCameraAxisButton->setTooltip(instructions.str());

        instructions.str("");
        actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE);
        sceneViewActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_fullModeButton->setTooltip(instructions.str());

        instructions.str("");
        actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE);
        sceneViewActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_wireframeModeButton->setTooltip(instructions.str());

        instructions.str("");
        actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE);
        sceneViewActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_dotModeButton->setTooltip(instructions.str());

        // Update camera reset button tooltip
        instructions.str("");
        actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::RESET);
        sceneViewActor->getActionLabel(actionType, actionDescription);
        inputBindings.getBindingDescription({actorId, actionType}, bindingDescription);
        instructions << actionDescription << "\n("<< bindingDescription << ")";
        this->_cameraResetButton->setTooltip(instructions.str());

        return true;
    }
    else if
        (std::type_index(typeid(ChangeRasterizationModeRequest)) == eventType)
    {
        OGL::RasterizationMode mode = static_cast<const ChangeRasterizationModeRequest&>(event).getMode();
        if
            (OGL::fill == mode)
        {
            this->_fullModeButton->setPushed(true);
            this->_wireframeModeButton->setPushed(false);
            this->_dotModeButton->setPushed(false);
        }
        else if
            (OGL::line == mode)
        {
            this->_fullModeButton->setPushed(false);
            this->_wireframeModeButton->setPushed(true);
            this->_dotModeButton->setPushed(false);
        }
        else if
            (OGL::point == mode)
        {
            this->_fullModeButton->setPushed(false);
            this->_wireframeModeButton->setPushed(false);
            this->_dotModeButton->setPushed(true);
        }

        this->updatePushButtonColors();
        return true;
    }
    else if
        (std::type_index(typeid(PolesAxisLockEvent)) == eventType)
    {
        bool lock = static_cast<const PolesAxisLockEvent&>(event).getLock();
        this->_lockCameraAxisButton->setPushed(lock);

        this->updatePushButtonColors();
        return true;
    }
    else if
        (std::type_index(typeid(ToggleShapeCreationRequest)) == eventType)
    {
        const ShapeType& shapeType = static_cast<const ToggleShapeCreationRequest&>(event).getShapeType();
        if
            (ShapeType::POINT == shapeType)
        {
            if
                (this->_pointButton->enabled())
            {
                this->_pointButton->setPushed(!this->_pointButton->pushed());
                this->_polylineButton->setPushed(false);
                this->_polygonButton->setPushed(false);
                this->_pointButton->changeCallback()(this->_pointButton->pushed());
            }
        }
        else if
            (ShapeType::POLYLINE == shapeType)
        {
            if
                (this->_polylineButton->enabled())
            {
                this->_polylineButton->setPushed(!this->_polylineButton->pushed());
                this->_pointButton->setPushed(false);
                this->_polygonButton->setPushed(false);
                this->_polylineButton->changeCallback()(this->_polylineButton->pushed());
            }
        }
        else if
            (ShapeType::POLYGON == shapeType)
        {
            if
                (this->_polygonButton->enabled())
            {
                this->_polygonButton->setPushed(!this->_polygonButton->pushed());
                this->_polylineButton->setPushed(false);
                this->_pointButton->setPushed(false);
                this->_polygonButton->changeCallback()(this->_polygonButton->pushed());
            }
        }
        this->updatePushButtonColors();
    }
    return false;
}

/**
 *
 */
void NanoguiToolbar::updateCollectionsChoice()
{
    // Polygons
    std::vector<std::string> polygons;
    for
        (const auto& it : this->_variablesByRenderMethod[RenderMethod::POLYGON])
    {
        polygons.push_back(it.second);
    }
    if
        (0 < this->_defaultCollectionComboBoxValue.size())
    {
        polygons.push_back(this->_defaultCollectionComboBoxValue);
    }
    this->_polygonSelector->setChoices(polygons);
    if
        ("" == this->_polygonSelector->getChoice())
    {
        this->_polygonSelector->choose(this->_defaultCollectionComboBoxValue);
    }

    // Polylines
    std::vector<std::string> polylines;
    for
        (const auto& it : this->_variablesByRenderMethod[RenderMethod::POLYLINE])
    {
        polylines.push_back(it.second);
    }
    if
        (0 < this->_defaultCollectionComboBoxValue.size())
    {
        polylines.push_back(this->_defaultCollectionComboBoxValue);
    }
    this->_polylineSelector->setChoices(polylines);
    if
        ("" == this->_polylineSelector->getChoice())
    {
        this->_polylineSelector->choose(this->_defaultCollectionComboBoxValue);
    }

    // Points
    std::vector<std::string> points;
    for
        (const auto& it : this->_variablesByRenderMethod[RenderMethod::POINT])
    {
        points.push_back(it.second);
    }
    if
        (0 < this->_defaultCollectionComboBoxValue.size())
    {
        points.push_back(this->_defaultCollectionComboBoxValue);
    }
    this->_pointSelector->setChoices(points);
    if
        ("" == this->_pointSelector->getChoice())
    {
        this->_pointSelector->choose(this->_defaultCollectionComboBoxValue);
    }
}


/**
 *
 */
bool NanoguiToolbar::getSelectedCollectionVariable(CollectionVariable& result, const RenderMethod& renderMethod, const std::string& label)
{
    if
        ("Ignored" == label)
    {
        return false;
    }

    if
        ("default" == label)
    {
        result = {{0u, 0u}, 0u};
        return true;
    }

    for
        (const auto& it: this->_variablesByRenderMethod[renderMethod])
    {
        if
            (label == it.second)
        {
            result = it.first;
            return true;
        }
    }
    return false;
}

/**
 *
 */
void NanoguiToolbar::buildCollectionsMapping(std::map<ShapeType, CollectionId>& mapping)
{
    bool ok(false);

    CollectionVariable pointColvar;
    ok = this->getSelectedCollectionVariable(pointColvar, RenderMethod::POINT,
                                             this->_pointSelector->getChoice());
    if
        (true == ok)
    {
        mapping[ShapeType::POINT] = pointColvar.first;
    }

    CollectionVariable polygonColvar;
    ok = this->getSelectedCollectionVariable(polygonColvar, RenderMethod::POLYGON,
                                             this->_polygonSelector->getChoice());
    if
        (true == ok)
    {
        mapping[ShapeType::POLYGON] = polygonColvar.first;
    }

    CollectionVariable polylineColvar;
    ok = this->getSelectedCollectionVariable(polylineColvar, RenderMethod::POLYLINE,
                                             this->_polylineSelector->getChoice());
    if
        (true == ok)
    {
        mapping[ShapeType::POLYLINE] = polylineColvar.first;
    }
}

/**
 *
 */
void NanoguiToolbar::initLoadDataWidgets(nanogui::Widget* parent)
{
    this->_loadButtonsBlock = new nanogui::Widget(parent);
    this->_loadButtonsBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                              nanogui::Alignment::Maximum,
                                                              0, 0));
    this->_loadButtonsBlock->setFixedSize(Eigen::Vector2i(345, 25));

    nanogui::Widget* loadButtons = new nanogui::Widget(this->_loadButtonsBlock);
    loadButtons->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                  nanogui::Alignment::Middle,
                                                  3, 2));
    this->_loadDataButton = new nanogui::Button(loadButtons, "Load");
    this->_loadDataButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_loadDataButton->setFontSize(16);
    this->_loadDataButton->setEnabled(false);
    this->_loadDataButton->setCallback([this] () {
        const std::string& path = this->_pathSelector->getPath();
        std::map<ShapeType, CollectionId> mapping;
        AnnotationFormat importFormat(AnnotationFormat::NONE);

        if
            (true == this->_ioFormatButtons[0]->pushed())
        {
            importFormat = AnnotationFormat::JSON;
        }
        else if
            (true == this->_ioFormatButtons[1]->pushed())
        {
            importFormat = AnnotationFormat::WKT;
        }
        else
        {
            LOG(LogLevel::Error, "NanoguiToolbar",
                     "No known import format selected");
            return;
        }


        this->buildCollectionsMapping(mapping);

        AddDefaultTimeToLoadAnnotationsFromFileRequest outEvent(path, mapping, importFormat);
        this->_appEventsDispatcher.dispatch<AddDefaultTimeToLoadAnnotationsFromFileRequest>(outEvent);

        this->_loadCollectionsTitle->setVisible(false);
        this->_titleBlock->setVisible(false);
        this->_collectionsBlock->setVisible(false);
        this->_pathSelector->hide();
        this->_pathBlock->setVisible(false);
        this->_ioFormatBlock->setVisible(false);
        this->_loadButtonsBlock->setVisible(false);
        this->_window->updateLayout();
    });

    nanogui::Button* loadNokButton(nullptr);
    loadNokButton = new nanogui::Button(loadButtons, "Cancel");
    loadNokButton->setFixedSize(Eigen::Vector2i(60, 18));
    loadNokButton->setFontSize(16);
    loadNokButton->setCallback([this] () {
        this->_loadCollectionsTitle->setVisible(false);
        this->_titleBlock->setVisible(false);
        this->_collectionsBlock->setVisible(false);
        this->_pathSelector->hide();
        this->_pathBlock->setVisible(false);
        this->_ioFormatBlock->setVisible(false);
        this->_loadButtonsBlock->setVisible(false);
        this->_window->updateLayout();
    });
}

/**
 *
 */
void NanoguiToolbar::initSaveDataWidgets(nanogui::Widget* parent)
{
    this->_saveButtonsBlock = new nanogui::Widget(parent);
    this->_saveButtonsBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                              nanogui::Alignment::Maximum,
                                                              0, 0));
    this->_saveButtonsBlock->setFixedSize(Eigen::Vector2i(345, 25));

    nanogui::Widget* saveButtons = new nanogui::Widget(this->_saveButtonsBlock);
    saveButtons->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                  nanogui::Alignment::Middle,
                                                  3, 2));
    this->_saveDataButton = new nanogui::Button(saveButtons, "Save");
    this->_saveDataButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_saveDataButton->setFontSize(16);
    this->_saveDataButton->setEnabled(false);
    this->_saveDataButton->setCallback([this] () {
        const std::string& path = this->_pathSelector->getPath();
        std::map<ShapeType, CollectionId> mapping;
        AnnotationFormat exportFormat(AnnotationFormat::NONE);

        if
            (true == this->_ioFormatButtons[0]->pushed())
        {
            exportFormat = AnnotationFormat::JSON;
        }
        else if
            (true == this->_ioFormatButtons[1]->pushed())
        {
            exportFormat = AnnotationFormat::WKT;
        }
        else
        {
            LOG(LogLevel::Error, "NanoguiToolbar",
                     "No known export format selected");
            return;
        }

        this->buildCollectionsMapping(mapping);

        SaveAnnotationsToFileRequest outEvent(path, mapping, exportFormat);
        this->_appEventsDispatcher.dispatch<SaveAnnotationsToFileRequest>(outEvent);

        this->_saveCollectionsTitle->setVisible(false);
        this->_titleBlock->setVisible(false);
        this->_collectionsBlock->setVisible(false);
        this->_pathSelector->hide();
        this->_pathBlock->setVisible(false);
        this->_ioFormatBlock->setVisible(false);
        this->_saveButtonsBlock->setVisible(false);
        this->_window->updateLayout();
    });

    nanogui::Button* saveNokButton(nullptr);
    saveNokButton = new nanogui::Button(saveButtons, "Cancel");
    saveNokButton->setFixedSize(Eigen::Vector2i(60, 18));
    saveNokButton->setFontSize(16);
    saveNokButton->setCallback([this] () {
        this->_saveCollectionsTitle->setVisible(false);
        this->_titleBlock->setVisible(false);
        this->_collectionsBlock->setVisible(false);
        this->_pathSelector->hide();
        this->_pathBlock->setVisible(false);
        this->_ioFormatBlock->setVisible(false);
        this->_saveButtonsBlock->setVisible(false);
        this->_window->updateLayout();
    });
}

/**
 *
 */
void NanoguiToolbar::initCollectionsCreator(nanogui::Widget* parent)
{
    this->_collectionCreator = new NanoguiAnnotationsCollectionCreator(parent);
    this->_collectionCreator->resize(Eigen::Vector2i(343, 123));
    this->_collectionCreator->setCancelCallback([this] () {
        this->_collectionCreator->setVisible(false);
        this->_window->updateLayout();
    });
    this->_collectionCreator->setValidateCallback([this] (const ShapeType& shapeType, const std::string& collectionLabel, const std::string& variableLabel) {
        this->_collectionCreator->reset();
        this->_collectionCreator->setVisible(false);
        this->_window->updateLayout();

        CreateAnnotationsCollectionRequest outEvent(shapeType, collectionLabel, variableLabel, true);
        this->_appEventsDispatcher.dispatch<CreateAnnotationsCollectionRequest>(outEvent);

        // Event will be processed and all actions that result from this event
        // will be executed, so at this point in the code the collection has
        // been added in the combobox choices: select it.

        std::string colvarStr("");
        SEAScope::getColvarLabel(collectionLabel, variableLabel, colvarStr);

        /*
        std::ostringstream value("");
        value << collectionLabel << ": " << variableLabel;  // This should be generated by a method as it must match the strings associated with VariablesAddedEvent
        */

        if
            (ShapeType::POINT == shapeType)
        {
            this->_pointSelector->choose(colvarStr);
        }
        else if
            (ShapeType::POLYGON == shapeType)
        {
            this->_polygonSelector->choose(colvarStr);
        }
        else if
            (ShapeType::POLYLINE == shapeType)
        {
            this->_polylineSelector->choose(colvarStr);
        }
    });
}

/**
 *
 */
void NanoguiToolbar::initPathSelector(nanogui::Widget* parent)
{
    this->_pathBlock = new nanogui::Widget(parent);
    this->_pathBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                       nanogui::Alignment::Fill,
                                                       2, 3));

    this->_pathSelectorLabel = new nanogui::Label(this->_pathBlock, "Select file", "sans-bold", 16);

    nanogui::Widget* pathContainer = new nanogui::Widget(this->_pathBlock);
    pathContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                    nanogui::Alignment::Middle,
                                                    0, 0));
    this->_pathSelector = new NanoguiPathSelector(pathContainer, PathSelectorType::get_file);
    this->_pathSelector->resize(Eigen::Vector2i(342, 40));
    this->_pathSelector->setPlaceholder("Please select a file");
    this->_pathSelector->setChangeCallback([this] (const std::string&) {
        this->updateLoadSaveButtons();
    });
    this->_pathSelector->setDropDownCallback([this] () {
        this->_pointSelector->hideDropDown();
        this->_polylineSelector->hideDropDown();
        this->_polygonSelector->hideDropDown();
    });
}


/**
 *
 */
void NanoguiToolbar::updateLoadSaveButtons()
{
    bool hasOneMapping(false);

    const std::string& polygonCollection = this->_polygonSelector->getChoice();
    if
        ((false == hasOneMapping) && (0 < polygonCollection.size()) && ("Ignored" != polygonCollection))
    {
        hasOneMapping = true;
    }

    const std::string& polylineCollection = this->_polylineSelector->getChoice();
    if
        ((false == hasOneMapping) && (0 < polylineCollection.size()) && ("Ignored" != polylineCollection))
    {
        hasOneMapping = true;
    }

    const std::string& pointCollection = this->_pointSelector->getChoice();
    if
        ((false == hasOneMapping) && (0 < pointCollection.size()) && ("Ignored" != pointCollection))
    {
        hasOneMapping = true;
    }

    const std::string& path = this->_pathSelector->getPath();
    this->_loadDataButton->setEnabled((true == hasOneMapping) && (0 < path.size()) && (true == Path::fileExists(path)) && (false == Path::dirExists(path)));
    this->_saveDataButton->setEnabled((true == hasOneMapping) && (0 < path.size()) && (false == Path::dirExists(path)));
}

/**
 *
 */
void NanoguiToolbar::updatePushButtonColors()
{
    if
        (this->_pointButton->pushed())
    {
        this->_pointButton->setTheme(this->_pushedTheme);
    }
    else
    {
        this->_pointButton->setTheme(this->_defaultTheme);
    }

    if
        (this->_polylineButton->pushed())
    {
        this->_polylineButton->setTheme(this->_pushedTheme);
    }
    else
    {
        this->_polylineButton->setTheme(this->_defaultTheme);
    }

    if
        (this->_polygonButton->pushed())
    {
        this->_polygonButton->setTheme(this->_pushedTheme);
    }
    else
    {
        this->_polygonButton->setTheme(this->_defaultTheme);
    }

    if
        (this->_fullModeButton->pushed())
    {
        this->_fullModeButton->setTheme(this->_pushedTheme);
    }
    else
    {
        this->_fullModeButton->setTheme(this->_defaultTheme);
    }

    if
        (this->_wireframeModeButton->pushed())
    {
        this->_wireframeModeButton->setTheme(this->_pushedTheme);
    }
    else
    {
        this->_wireframeModeButton->setTheme(this->_defaultTheme);
    }

    if
        (this->_dotModeButton->pushed())
    {
        this->_dotModeButton->setTheme(this->_pushedTheme);
    }
    else
    {
        this->_dotModeButton->setTheme(this->_defaultTheme);
    }

    if
        (this->_lockCameraAxisButton->pushed())
    {
        this->_lockCameraAxisButton->setTheme(this->_pushedTheme);
    }
    else
    {
        this->_lockCameraAxisButton->setTheme(this->_defaultTheme);
    }
}
