// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_ui.h"
#include "core/log.h"
#include "core/state.h"
#include "core/app_event.h"
#include "core/interfaces/iwindow.h"
#include "core/app_events_dispatcher.h"
#include "core/input_events_dispatcher.h"
#include "ui/nanogui/nanogui_rendering_menu.h"
#include "ui/nanogui/nanogui_floating_label.h"

#include "events/drawing_started_event.h"
#include "events/drawing_complete_event.h"
#include "events/measurements_updated_event.h"
#include "events/shape_edition_complete_event.h"
#include "events/side_menu_resize_request.h"
#include "events/display_catalogue_request.h"
#include "events/granule_info_ready_for_display_event.h"
#include "events/display_extraction_info_request.h"
#include "num_utils.h"
#include "core/measurements.h"

#include "ui/nanogui/nanogui_advanced_rendering_control.h"
#include "ui/nanogui/nanogui_timeline.h"
#include "ui/nanogui/nanogui_map_caption.h"
#include "ui/nanogui/nanogui_main_menu.h"
#include "ui/nanogui/nanogui_rendering_menu.h"
#include "ui/nanogui/nanogui_toolbar.h"
#include "ui/nanogui/nanogui_collections_control.h"
#include "ui/nanogui/nanogui_sources_control.h"
#include "ui/nanogui/nanogui_camera_control.h"
#include "ui/nanogui/nanogui_seascope_side_menu.h"
#include "tools/contextual_menu.h"

using namespace SEAScope;

/**
 *
 */
NanoguiUI::NanoguiUI(AppEventsDispatcher& appEventsDispatcher, InputEventsDispatcher& inputEventsDispatcher)
: _appEventsDispatcher(appEventsDispatcher),
  _inputEventsDispatcher(inputEventsDispatcher),
  _state(nullptr),
  _window(nullptr),
  _renderingMenu(nullptr),
  _floatingLabel(nullptr),
  _saveLayoutCallback(nullptr),
  _resizeCallback(nullptr),
  _componentResizeCallback(nullptr),
  _timelineUI{nullptr},
  _cameraUI{nullptr},
  _toolbarUI{nullptr},
  _sideMenu{nullptr},
  _ctxMenu{nullptr}
{
    this->_appEventsDispatcher.subscribe<DrawingStartedEvent>(this);
    this->_appEventsDispatcher.subscribe<DrawingCompleteEvent>(this);
    this->_appEventsDispatcher.subscribe<MeasurementsUpdatedEvent>(this);
    this->_appEventsDispatcher.subscribe<ShapeEditionCompleteEvent>(this);
    this->_appEventsDispatcher.subscribe<SideMenuResizeRequest>(this);
    this->_appEventsDispatcher.subscribe<GranuleInfoReadyForDisplayEvent>(this);
    this->_appEventsDispatcher.subscribe<DisplayCatalogueRequest>(this);
    this->_appEventsDispatcher.subscribe<DisplayExtractionInfoRequest>(this);
};

/**
 *
 */
NanoguiUI::~NanoguiUI()
{
    this->_appEventsDispatcher.unsubscribe<DrawingStartedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<DrawingCompleteEvent>(this);
    this->_appEventsDispatcher.unsubscribe<MeasurementsUpdatedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<ShapeEditionCompleteEvent>(this);
    this->_appEventsDispatcher.unsubscribe<SideMenuResizeRequest>(this);
    this->_appEventsDispatcher.unsubscribe<GranuleInfoReadyForDisplayEvent>(this);
    this->_appEventsDispatcher.unsubscribe<DisplayCatalogueRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DisplayExtractionInfoRequest>(this);


    this->_state = nullptr; // not owning
    this->_window = nullptr;    // not owning

    this->_floatingLabel = nullptr; // deletion will be handled by Nanogui

    delete this->_renderingMenu;
    this->_renderingMenu = nullptr;

    this->_saveLayoutCallback = nullptr;
    this->_resizeCallback = nullptr;
    this->_componentResizeCallback = nullptr;

    delete this->_ctxMenu;
    this->_ctxMenu = nullptr;

    delete this->_toolbarUI;
    this->_toolbarUI = nullptr;

};

/**
 *
 */
bool NanoguiUI::init(IWindow& window, Settings& settingsIn, State* state)
{
	bool ok;	//

	this->_state = state;

	this->_window = &window;
	ok = this->_renderer.setWindow(window);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "NanoguiUI"
				, "Initialization with incompatible window type.");
	}
	this->_settings.init(this->_renderer);
	this->_settings.setVisible(false);
	this->_settings.layoutSettings->setSaveCallback([this] () {
		if
			(nullptr != this->_saveLayoutCallback)
		{
			this->_saveLayoutCallback();
		}
	});

	this->_renderer.setResizeCallback([this](int width, int height){
		this->resize(width, height);
	});
	this->_renderer.performLayout();
	this->_renderer.setVisible(true);

	this->_renderingMenu = new NanoguiRenderingMenu();
	this->_renderingMenu->init(&(this->_renderer));
	this->_renderingMenu->moreDataCallback = [this]() {
        DisplayCatalogueRequest event;
        this->_appEventsDispatcher.dispatch<DisplayCatalogueRequest>(event);
	};


    this->_floatingLabel = new NanoguiFloatingLabel(&(this->_renderer));

	this->_timelineUI = new NanoguiTimeline(&this->_renderer);
	this->_cameraUI = new NanoguiCameraControl(&this->_renderer, this->_appEventsDispatcher, this->_inputEventsDispatcher);
    this->_toolbarUI = new NanoguiToolbar(this->_appEventsDispatcher, this->_inputEventsDispatcher);
    this->_toolbarUI->init(&this->_renderer);

    this->_sideMenu = new NanoguiSEAScopeSideMenu(&this->_renderer, this->_appEventsDispatcher);
    this->_sideMenu->getMainMenu()->getSettingsPane()->loadSettings(settingsIn);

    this->_ctxMenu = new ContextualMenu(this->_appEventsDispatcher, &(this->_renderer), &window);


    this->_toolbarUI->showAt(Eigen::Vector2i(20, 48));
	this->_renderingMenu->window->display(Eigen::Vector2i(20, 124));

    this->_state->blueprintFactory = &(this->_blueprintFactory);
	this->_blueprintFactory.init(&this->_renderer);

	this->addComponent(this->_timelineUI, "Timeline", "100%", "136px", "0%", "-0%", false);
	this->addComponent(this->_cameraUI, "Camera", "300px", "24px", "30%", "10px", false);

	this->setComponentResizeCallback([this] (const std::string& cId, int w, int h) {
		LOG(LogLevel::Debug, "NanoguiUI", "Component resized: {} {}x{}", cId, w, h);
		if
			("Timeline" == cId)
		{
            SideMenuResizeRequest event(320, this->_renderer.height() - h);
            this->_appEventsDispatcher.dispatch<SideMenuResizeRequest>(event);
		}
	});

	this->setResizeCallback([this] (int /*width*/, int /*heightIn*/) {
        SideMenuResizeRequest event(320, this->_renderer.height() - this->_timelineUI->fixedHeight());
        this->_appEventsDispatcher.dispatch<SideMenuResizeRequest>(event);
	});

    this->showProcessingOptions(settingsIn.processorEnabled);

	return ok;
};

/**
 *
 */
IRenderer* NanoguiUI::getRenderer()
{
    return &(this->_renderer);
}

/**
 *
 */
EventHandler* NanoguiUI::getEventHandler()
{
    return &(this->_renderer);
}

/**
 *
 */
NanoguiRenderingMenu* NanoguiUI::getRenderingMenu() const
{
    return this->_renderingMenu;
}

/**
 *
 */
NanoguiFloatingLabel* NanoguiUI::getFloatingLabel() const
{
    return this->_floatingLabel;
}

/**
 *
 */
bool NanoguiUI::setSaveLayoutCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_saveLayoutCallback)
    {
        return false;
    }
	this->_saveLayoutCallback = callback;
	return true;
};

/**
 *
 */
bool NanoguiUI::setResizeCallback(const std::function<void(int, int)>& callback)
{
    if
        (nullptr != this->_resizeCallback)
    {
        return false;
    }
    this->_resizeCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiUI::setComponentResizeCallback(const std::function<void(const std::string&, int, int)>& callback)
{
    if
        (nullptr != this->_componentResizeCallback)
    {
        return false;
    }
    this->_componentResizeCallback = callback;
    return true;
}

/**
 *
 */
const std::function<void(const std::string&, int, int)> NanoguiUI::getComponentResizeCallback() const
{
    return this->_componentResizeCallback;
}

/**
 *
 */
bool NanoguiUI::resize(int width, int height)
{
	auto it = this->_state->layout.begin();
	for
		(; it != this->_state->layout.end(); it++)
	{
		LOG(LogLevel::Debug, "UI", "resizing blueprint {}", it->first);
		it->second->resizeAuto();
	}

	this->_renderer.performLayout();

	if
		(nullptr != this->_resizeCallback)
	{
		this->_resizeCallback(width, height);
	}

	return true;
};

/**
 *
 */
bool NanoguiUI::addComponent( IComponent* component, const std::string& label
							, const std::string& minWidth
							, const std::string& minHeight
							, const std::string& defaultX
							, const std::string& defaultY
							, bool reflow = false)
{
	IBlueprint* blueprint;	//
	NanoguiBlueprint* nBlueprint;	//

	component->setComponentId(label);
	component->setResizeEventEmitter([this] (IComponent* c) {
		if
			(nullptr != this->_componentResizeCallback)
		{
			this->_componentResizeCallback( c->getComponentId()
										  , c->getWidth()
										  , c->getHeight());
		}
	});

	auto found = this->_state->layout.find(label);
	if
		(this->_state->layout.end() == found)
	{
		blueprint = this->_state->createBlueprint( component->isResizable()
												 , component->isMovable());
		blueprint->setLayout(defaultX, defaultY, minWidth, minHeight);
		this->_state->layout.insert(std::make_pair(label, blueprint));
	}
	else
	{
		blueprint = found->second;
	}
	blueprint->setMinDimensions(minWidth, minHeight);
	blueprint->setTarget(component);

	nBlueprint = dynamic_cast<NanoguiBlueprint*>(blueprint);
	if
		(nullptr == nBlueprint)
	{
		LOG( LogLevel::Error, "Nanogui UI"
				, "Tried to use an incompatible blueprint, only NanoguiBlueprint instances are supported");
		return false;
	}
	this->_settings.layoutSettings->addBlueprint(label, nBlueprint);

	if
		(true == reflow)
	{
		this->_renderer.performLayout();
	}

	return false;
};

/**
 *
 */
IBlueprint* NanoguiUI::getBlueprint(const IComponent* component, bool& ok)
{
	auto it = this->_state->layout.begin();
	for
		(; it != this->_state->layout.end(); it++)
	{
		if
			(it->second->getTarget() == component)
		{
			ok = true;
			return it->second;
		}
	}
	ok = false;
	return nullptr;
};

/**
 *
 */
bool NanoguiUI::trackMouse(IComponent* component)
{
	bool ok(false);	//
	int windowWidthTmp(0);	//
	int windowHeightTmp(0);	//
    double windowWidth(0.0);    //
    double windowHeight(0.0);   //
	double cursorX(0.0);	//
	double cursorY(0.0);	//
	double componentWidth(0.0);	//
	double componentHeight(0.0);	//
	int componentX(0);	//
	int componentY(0); //
    double componentXTmp(0.0);  //
    double componentYTmp(0.0);  //

	if
		(nullptr == this->_window || nullptr == component)
	{
		return false;
	}

	ok = this->_window->getSize(windowWidthTmp, windowHeightTmp);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "NanoguiUI", "Failed to get window size");
		return false;
	}
	ok = this->_window->getCursorPos(cursorX, cursorY);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "NanoguiUI", "Failed to get mouse cursor position");
		return false;
	}

#if defined(_WIN32) || defined(__linux__)
    /* Apply scale-factor on the cursor position */
    double scaleFactor;  //
    scaleFactor = static_cast<double>(this->_renderer.pixelRatio());
    cursorX = cursorX / scaleFactor;
    cursorY = cursorY / scaleFactor;
#endif

    windowWidth = static_cast<double>(windowWidthTmp);
    windowHeight = static_cast<double>(windowHeightTmp);
    componentWidth = static_cast<double>(component->getWidth());
    componentHeight = static_cast<double>(component->getHeight());

	componentXTmp = std::floor(cursorX + 15.0);
	componentYTmp = std::floor(cursorY + 15.0);

	if
		(windowWidth - cursorX < 15.0 + componentWidth)
	{
		componentXTmp = std::floor(cursorX - 15.0 - componentWidth);
	}
	if
		(windowHeight - cursorY < 15.0 + componentHeight)
	{
		componentYTmp = std::floor(cursorY - 15.0 - componentHeight);
	}

    componentX = static_cast<int>(componentXTmp);
    componentY = static_cast<int>(componentYTmp);
	ok = component->setPosition(componentX, componentY);
	return ok;
}

/**
 *
 */
bool NanoguiUI::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(DrawingStartedEvent)) == eventType)
    {
        this->_floatingLabel->setCaption("");
        this->trackMouse(this->_floatingLabel);
        return true;
    }
    else if
        (std::type_index(typeid(DrawingCompleteEvent)) == eventType)
    {
        this->_floatingLabel->setCaption("");
        return true;
    }
    else if
        (std::type_index(typeid(ShapeEditionCompleteEvent)) == eventType)
    {
        this->_floatingLabel->setCaption("");
        return true;
    }
    else if
        (std::type_index(typeid(MeasurementsUpdatedEvent)) == eventType)
    {
        const ShapeType& shapeType = static_cast<const MeasurementsUpdatedEvent&>(event).getShapeType();
        double length = static_cast<const MeasurementsUpdatedEvent&>(event).getLength();
        double area = static_cast<const MeasurementsUpdatedEvent&>(event).getArea();

        std::ostringstream buffer("");	//
        std::string tmpStr;	//
        bool success(false);	//

        if
            (ShapeType::POLYGON == shapeType)
        {
            success = SEAScope::formatArea(area, tmpStr);
            if
                (true == success)
            {
                buffer << "Area: " << tmpStr;
            }

            success = SEAScope::formatDistance(length, tmpStr);
            if
                (true == success)
            {
                if
                    (0 < buffer.tellp())
                {
                    buffer << "\n";
                }
                buffer << "Perimeter: " << tmpStr;
            }
        }
        else if
            (ShapeType::POLYLINE == shapeType)
        {
            success = SEAScope::formatDistance(length, tmpStr);
            if
                (true == success)
            {
                buffer << "Total distance: " << tmpStr;
            }
        }

        if
            (false == (std::isnan(length) || floating_point_equality(0.0, length)))
        {
            this->_floatingLabel->setCaption(buffer.str());
        }
        else
        {
            this->_floatingLabel->setCaption("");
        }

        this->trackMouse(this->_floatingLabel);

        return true;
    }
    else if
        (std::type_index(typeid(SideMenuResizeRequest)) == eventType)
    {
        int width = static_cast<const SideMenuResizeRequest&>(event).getWidth();
        int height = static_cast<const SideMenuResizeRequest&>(event).getHeight();

        this->_sideMenu->resize(width, height);

        return true;
    }
    else if
        (std::type_index(typeid(GranuleInfoReadyForDisplayEvent)) == eventType)
    {
        const GranuleInfo& info = static_cast<const GranuleInfoReadyForDisplayEvent&>(event).getGranuleInfo();
        this->_sideMenu->displayGranuleInfo(info);
        return true;
    }
    else if
        (std::type_index(typeid(DisplayCatalogueRequest)) == eventType)
    {
        this->_sideMenu->displayCatalogue();
        return true;
    }
    else if
        (std::type_index(typeid(DisplayExtractionInfoRequest)) == eventType)
    {
        const std::map<RenderableId, Extraction>& extractionData = static_cast<const DisplayExtractionInfoRequest&>(event).getExtractionData();
        const std::map<RenderableId, std::string>& labels = static_cast<const DisplayExtractionInfoRequest&>(event).getLabels();
        this->_sideMenu->displayExtraction(extractionData, labels);

        return true;
    }

    return false;
}

/**
 *
 */
void NanoguiUI::setIcons(Icons& icons)
{
	this->_renderer.setIcons(icons);
}

/**
 *
 */
ICollectionsControlView* NanoguiUI::getCollectionsUI()
{
    return this->_sideMenu->getMainMenu()->getCollectionsControl();
}

/**
 *
 */
IRenderingControlView* NanoguiUI::getRenderingUI()
{
    return this->_renderingMenu->control;
}

/**
 *
 */
IAdvancedRenderingControlView* NanoguiUI::getAdvancedRenderingUI()
{
    return this->_renderingMenu->advanced;
}

/**
 *
 */
ITimelineView* NanoguiUI::getTimelineUI()
{
    return this->_timelineUI;
}

/**
 *
 */
ISourcesControlView* NanoguiUI::getSourcesUI()
{
    return this->_sideMenu->getMainMenu()->getSourcesControl();
}

/**
 *
 */
ICameraControlView* NanoguiUI::getCameraUI()
{
    return this->_cameraUI;
}

/**
 *
 */
void NanoguiUI::showProcessingOptions(bool value)
{
    if
        (true == value)
    {
        this->_sideMenu->getMainMenu()->getExtractionDialog()->showProcessingOptions();
    }
    else
    {
        this->_sideMenu->getMainMenu()->getExtractionDialog()->hideProcessingOptions();
    }
}
