// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "ui/nanogui/nanogui_utils.h"

using namespace SEAScope;

const int NanoguiUtils::NVG_GRADIENT_SAMPLES = 1024;

void NanoguiUtils::gradientSpan( uint32_t* dst, float stepSize, uint32_t step
							   , const std::array<uint8_t, 3>& s0
							   , const std::array<uint8_t, 3>& s1)
{
    float stepTmp = static_cast<float>(step);
	float s = std::floor(stepTmp * stepSize * NVG_GRADIENT_SAMPLES);
	float e = std::floor((stepTmp + 1.0f) * stepSize * NVG_GRADIENT_SAMPLES);
	float factor = 1.f / (e - s);
	float r = static_cast<float>(s0[0]);
	float g = static_cast<float>(s0[1]);
	float b = static_cast<float>(s0[2]);
	float a = 255.0f;
	float dr = static_cast<float>(s1[0] - s0[0]) * factor;
	float dg = static_cast<float>(s1[1] - s0[1]) * factor;
	float db = static_cast<float>(s1[2] - s0[2]) * factor;
	int8_t da = 0;

    std::size_t sTmp = static_cast<std::size_t>(s);
    std::size_t eTmp = static_cast<std::size_t>(e);
	for
		(std::size_t i = sTmp; i < eTmp; i++)
	{
        std::uint32_t r32 = static_cast<std::uint32_t>(std::round(r));
        std::uint32_t g32 = static_cast<std::uint32_t>(std::round(g));
        std::uint32_t b32 = static_cast<std::uint32_t>(std::round(b));
        std::uint32_t a32 = static_cast<std::uint32_t>(std::round(a));

#if defined __BIG_ENDIAN__
        r32 <<= 24;
        g32 <<= 16;
        b32 <<= 8;
        a32 <<= 0;
#else
        r32 <<= 0;
        g32 <<= 8;
        b32 <<= 16;
        a32 <<= 24;
#endif

        dst[i] = r32 + g32 + b32 + a32;
		r += dr; g += dg; b += db; a += static_cast<float>(da);
	}
};

std::tuple<int, struct NVGpaint> NanoguiUtils::nvgLinearGradientStops( struct NVGcontext* ctx, float sx, float sy, float ex, float ey, const Colormap& cmap)
{
	uint32_t data[NVG_GRADIENT_SAMPLES];
	float w = ex - sx;
	float h = ey - sy;
	float stepSize;
	float len = sqrtf(w * w + h * h);
	int img;
	size_t i;	//
	size_t l;	//

	i = 0;
	l = cmap.size();
	if
		(0 >= l)
	{
		stepSize = 1.f;
		gradientSpan(data, stepSize, 0, {{ 0,0,0 }}, {{ 255, 255, 255 }});
	}
	else
	{
		stepSize = 1.f / static_cast<float>(l);
		for
			(; i < l - 1; ++i)
		{
			gradientSpan(data, stepSize, static_cast<std::uint32_t>(i), cmap[i], cmap[i + 1]);
		}
	}

	img = nvgCreateImageRGBA(ctx,
							 NVG_GRADIENT_SAMPLES, 1,
							 NVG_IMAGE_REPEATY|NVG_IMAGE_REPEATX,
							 reinterpret_cast<unsigned char*>(data));
	return std::make_tuple(img, nvgImagePattern(ctx,
						   sx, sy,
						   len, len,
						   atan2f(ey-sy, ex-sx),
						   img, 1.f));
};
