// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_window.h"
#include "ui/nanogui/nanogui_renderer.h"
#include "ui/nanogui/nanovg_wrapper.h"

NanoguiWindow::NanoguiWindow(nanogui::Widget* parent, const std::string& title)
: nanogui::Window(parent, title)
{
    nanogui::Widget* bp = buttonPanel();
    _closeButton = new nanogui::Button(bp, "", ENTYPO_ICON_CROSS);
    _closeButton->setFixedSize(nanogui::Vector2i(18, 18));
    _closeButton->setSize(nanogui::Vector2i(18, 18));
}

NanoguiWindow::~NanoguiWindow()
{
}

void NanoguiWindow::performLayout(NVGcontext* ctx)
{
    if
        (!mButtonPanel)
    {
        Widget::performLayout(ctx);
    }
    else
    {
        mButtonPanel->setVisible(false);
        Widget::performLayout(ctx);

        for
            (auto w : mButtonPanel->children())
        {
            w->setFixedSize(nanogui::Vector2i(18, 18));
            w->setFontSize(18);
        }

        mButtonPanel->setVisible(true);
        mButtonPanel->setSize(nanogui::Vector2i(width(), 18));
        mButtonPanel->setPosition(nanogui::Vector2i(width() - (mButtonPanel->preferredSize(ctx).x() + 5), 3));
        mButtonPanel->performLayout(ctx);
    }
}

void NanoguiWindow::draw(NVGcontext* ctx)
{
    //   ---------------------------------------
    //  |     ^                                 |
    //  |     | X                               |
    //  |<-X- TEXT                              | header height = 24px 
    //  |     | X                               |
    //  |     v                                 |
    //   ---------------------------------------
    //
    // X = padding = 3px
    // Font size = 18 + 2x padding = 24 = height
    // Button size = 18 = height - 2 x padding = 24 - 2 x 3 = 18
    // 

    nanogui::Widget* bp = buttonPanel();

    nanogui::BoxLayout* bl = static_cast<nanogui::BoxLayout*>(bp->layout());
    bl->setSpacing(0);

    _closeButton->setFixedSize(nanogui::Vector2i(18, 18));
    _closeButton->setSize(nanogui::Vector2i(18, 18));
    buttonPanel()->performLayout(ctx);

    int hh = 24;
    constexpr int padding = 3;

    /* Draw window */
    nvgSave(ctx);
    nvgBeginPath(ctx);
    nvg::rect(ctx, mPos, mSize);
    nvgFillColor(ctx, nanogui::Color(mTheme->mWindowFillFocused.head<3>(), 1.f));
    nvgFill(ctx);

    if (!mTitle.empty())
    {
        nanogui::Color headerColor(36, 36, 36, 255);
        NVGcolor headerColorNVG = static_cast<NVGcolor>(headerColor);

        nvgBeginPath(ctx);
        nvg::rect(ctx, mPos, nanogui::Vector2i(mSize.x(), hh));
        nvgFillColor(ctx, headerColorNVG);
        nvgFill(ctx);

        nvgFontSize(ctx, 18.0f);
        nvgFontFace(ctx, "sans-bold");
        nvgTextAlign(ctx, NVG_ALIGN_LEFT | NVG_ALIGN_MIDDLE);
        nvgFillColor(ctx, mTheme->mTextColor);
        nvg::text(ctx, nanogui::Vector2i(mPos.x() + padding, mPos.y() + hh / 2), mTitle);
    }

    nvgRestore(ctx);
    Widget::draw(ctx);
}

SEAScope::NanoguiRenderer* NanoguiWindow::getRenderer()
{
    nanogui::Widget* w = this;

    while
        (nullptr != w->parent())
    {
        w = w->parent();
    }

    return dynamic_cast<SEAScope::NanoguiRenderer*>(w);
}

void NanoguiWindow::setCloseCallback(const std::function<void()>& func)
{
    _closeButton->setCallback(func);
}
