// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iterator>
#include <algorithm>
#include <sstream>

#include "wkt_annotation.h"
#include "core/data_node.h"
#include "core/utils.h"
#include "data/annotation.h"
#include "3d/shapes_allocator.h"
#include "3d/unit_sphere.h"
#include "3d/geodetic.h"
#include "3d/polygon.h"
#include "3d/polyline.h"
#include "3d/ellipse.h"
#include "3d/point.h"
#include "3d/grid.h"
#include "core/measurements.h"
#include "core/log.h"

using namespace SEAScope;

/**
 *
 */
static std::string readOneShape(std::istreambuf_iterator<char>& it)
{
    std::string result("");
    std::istreambuf_iterator<char> end;
    std::size_t parenthesisCount(0);
    while
        (it != end)
    {
        if
            ('(' == *it)
        {
            parenthesisCount = parenthesisCount + 1;
        }
        else if
            (')' == *it)
        {
            parenthesisCount = parenthesisCount - 1;
            result.push_back(*it);
            if
                (0 >= parenthesisCount)
            {
                ++it;
                break;
            }
        }
        result.push_back(*it);

        ++it;
    }

    return result;
}

/**
 *
 */
WKTAnnotation::WKTAnnotation()
{
}

/**
 *
 */
WKTAnnotation::~WKTAnnotation()
{
}


/**
 *
 */
bool WKTAnnotation::serialize(const Annotation& src, std::ostream& stream) const
{
    bool ok(false); //

    ok = this->serializeShape(src.getShape(), stream);

    return ok;
}

/**
 *
 */
bool WKTAnnotation::serialize(const std::vector<Annotation>& src, std::ostream& stream) const
{
    bool ok(true);    //

    for
        (const auto& annotation : src)
    {
        ok = ok && this->serializeShape(annotation.getShape(), stream);
    }

    return ok;
}

/**
 *
 */
bool WKTAnnotation::deserialize(std::istream& stream, Annotation& dst, double defaultStart, double defaultStop) const
{
    bool ok{false};
    std::istreambuf_iterator<char> it{stream.rdbuf()};   //

    std::string wktString = readOneShape(it);
    ok = this->parseWKT(wktString, dst);
    if
        (false == ok)
    {
        LOG(LogLevel::Warn, "WKT Annotation", "Annotation skipped");
        LOG(LogLevel::Debug, "WKT Annotation", wktString);
    }

    dst.setStart(defaultStart);
    dst.setStop(defaultStop);
    return ok;
}

/**
 *
 */
bool WKTAnnotation::deserialize(std::istream& stream, std::vector<Annotation>& dst, double defaultStart, double defaultStop) const
{
    std::istreambuf_iterator<char> it{stream.rdbuf()};   //
    std::istreambuf_iterator<char> end{};   //

    while
        (it != end)
    {
        bool annotationOk{false};   //
        Annotation annotation;  //

        std::string wktString = readOneShape(it);
        annotationOk = this->parseWKT(wktString, annotation);
        if
            (true == annotationOk)
        {
            annotation.setStart(defaultStart);
            annotation.setStop(defaultStop);
            dst.push_back(std::move(annotation));
        }
        else
        {
            LOG(LogLevel::Warn, "WKT Annotation", "Annotation skipped");
            LOG(LogLevel::Debug, "WKT Annotation", wktString);
        }
    }

    return true;
}

/**
 *
 */
bool WKTAnnotation::serializeShape(OGL::Shape* shape, std::ostream& output) const
{
    if
        (nullptr == shape)
    {
        return false;
    }

    ShapeType shapeType = shape->getType();
    if
        (  (SEAScope::POINT != shapeType)
        && (SEAScope::POLYLINE != shapeType)
        && (SEAScope::POLYGON != shapeType))
    {
        LOG(LogLevel::Warn, "WKTAnnotation", "Shape type not supported");
        return false;
    }

    if
        (0 < output.tellp())
    {
        output << "\n";
    }

    if
        (SEAScope::POINT == shapeType)
    {
        Point* point = static_cast<Point*>(shape);

        Geodetic2D geo2d = UnitSphere::ToGeodetic2D(point->getPosition());
        double lon = 180.0 * geo2d.Longitude() / SEAScope::PI;
        double lat = 180.0 * geo2d.Latitude() / SEAScope::PI;

        output << "POINT(" << lon << " " << lat << ")";
        point = nullptr;
    }
    else if
        (SEAScope::POLYLINE == shapeType)
    {
        Polyline* polyline = static_cast<Polyline*>(shape);
        const std::vector<glm::vec3>& vertices = polyline->getPoints();
        unsigned int pointsCount = polyline->getPointCount();
        if
            (2 > pointsCount)
        {
            return false;
        }

        output << "LINESTRING(";
        double lon = 0.0;
        double lat = 0.0;
        for
            (const auto& vertex : vertices)
        {
            Geodetic2D geo2d = UnitSphere::ToGeodetic2D(vertex);
            lon = 180.0 * geo2d.Longitude() / SEAScope::PI;
            lat = 180.0 * geo2d.Latitude() / SEAScope::PI;

            output << lon << " " << lat << ",";
        }

        // Move write head to overwrite trailing comma
        output.seekp(-1, output.cur);

        output << ")";
        polyline = nullptr;

    }
    else if
        (SEAScope::POLYGON == shapeType)
    {
        Polygon* polygon = static_cast<Polygon*>(shape);
        const std::vector<glm::vec3>& vertices = polygon->getPoints();
        unsigned int pointsCount = polygon->getPointCount();
        if
            (3 > pointsCount)
        {
            return false;
        }

        double lon = 0.0;
        double lat = 0.0;
        output << "POLYGON((";
        for
            (const auto& vertex : vertices)
        {
            Geodetic2D geo2d = UnitSphere::ToGeodetic2D(vertex);
            lon = 180.0 * geo2d.Longitude() / SEAScope::PI;
            lat = 180.0 * geo2d.Latitude() / SEAScope::PI;
            output << lon << " " << lat << ",";
        }

        // Duplicate the first vertex at the end to create a closed linear ring
        if
            (vertices[0] != vertices[pointsCount - 1])
        {
            Geodetic2D geo2d = UnitSphere::ToGeodetic2D(vertices[0]);
            lon = 180.0 * geo2d.Longitude() / SEAScope::PI;
            lat = 180.0 * geo2d.Latitude() / SEAScope::PI;
            output << lon << " " << lat << ",";
        }

        // Move write head to overwrite trailing comma
        output.seekp(-1, output.cur);

        output << "))";
        polygon = nullptr;
    }

    return true;
}

/**
 *
 */
bool WKTAnnotation::parseWKT(std::string& wktString, Annotation& dst) const
{
    bool isPoint(false);    //
    bool isPolyline(false); //
    bool isPolygon(false);  //

    wktString = SEAScope::trim(wktString);
    if
        (0 >= wktString.size())
    {
        return false;
    }

    isPoint = ((5 <= wktString.size()) && SEAScope::iequals(wktString.substr(0, 5), "POINT"));
    isPolyline = ((10 <= wktString.size()) && SEAScope::iequals(wktString.substr(0, 10), "LINESTRING"));
    isPolygon = ((7 <= wktString.size()) && SEAScope::iequals(wktString.substr(0, 7), "POLYGON"));

    // Make sure the content of the WKT is a shape supported by SEAScope
    if
        (false == (isPoint || isPolyline || isPolygon))
    {
        LOG(LogLevel::Warn, "WKT Annotation",
                 "SEAScope only supports the definition of the following shapes in WKT format: POINT, LINESTRING and POLYGON (no holes).");
        LOG(LogLevel::Debug, "WKT Annotation", wktString);
        return false;
    }

    std::size_t openParenthesisIndex = wktString.rfind('(');
    std::size_t closeParenthesisIndex = wktString.find(')');
    std::string coordsString = wktString.substr(1 + openParenthesisIndex, closeParenthesisIndex - openParenthesisIndex - 1);

    std::vector<std::string> coordsSubstrings;
    SEAScope::explodeString(coordsString, ",", coordsSubstrings);

    std::size_t pointsCount = coordsSubstrings.size();
    std::vector<glm::vec2> geographic_points;
    geographic_points.reserve(pointsCount);
    for
        (auto& it : coordsSubstrings)
    {
        std::string::size_type stringPos;
        double lon = std::stod(it, &stringPos);
        double lat = std::stod(it.substr(stringPos + 1));
        geographic_points.push_back(glm::vec2(lon, lat));
    }

    std::vector<glm::vec3> geodesic_points;
    fromLonlatToSphereUnitVector(geographic_points, geodesic_points);

    if
        (isPoint && (1 == geodesic_points.size()))
    {
        Point* point = ShapesAllocator::getShape<Point>();
        point->setPosition(geodesic_points[0]);
        dst.setShape(std::move(point));
        point = nullptr;
    }
    else if
        (isPolyline && (1 < geodesic_points.size()))
    {
        Polyline* polyline = ShapesAllocator::getShape<Polyline>();
        polyline->setPoints(std::move(geodesic_points));
        dst.setShape(std::move(polyline));
        polyline = nullptr;
    }
    else if
        (isPolygon && (2 < geodesic_points.size()))
    {
        Polygon* polygon = ShapesAllocator::getShape<Polygon>();
        polygon->setPoints(std::move(geodesic_points));
        dst.setShape(std::move(polygon));
        polygon = nullptr;
    }
    else
    {
        LOG(LogLevel::Error, "WKT Annotation", wktString);
        LOG(LogLevel::Error, "WKT Annotation",
                 "The shape type defined in the WKT requires more vertices than the ones provided in the file '{})", geodesic_points.size());
        return false;
    }

    return true;
}
