// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iostream>
#include <algorithm>
#include <string>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <GL/gl.h>
#include <GL/glx.h>

#include "platform/app_launcher_impl.h"
#include "core/path.h"
#include "vendors/_mini.h"

using namespace SEAScope;

/**
 *
 */
static bool checkMESAVersion()
{
    // MESA fixes
    Display* dpy {nullptr};

    /* open display */
    dpy = XOpenDisplay(nullptr);
    if
        (nullptr == dpy)
    {
        std::cerr << "cannot connect to X server" << std::endl;
        return false;
    }

    /* get root window */
    Window root = DefaultRootWindow(dpy);

    /* get visual matching attr */
    GLint attr[] = { GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None };
    XVisualInfo* vi {nullptr};
    vi = glXChooseVisual(dpy, 0, attr);
    if
        (nullptr == vi)
    {
        XDestroyWindow(dpy, root);
        XCloseDisplay(dpy);
        dpy = nullptr;

        std::cerr << "no appropriate visual found" << std::endl;
        return false;
    }

    /* create a context using the root window */
    GLXContext ctx {nullptr};
    ctx = glXCreateContext(dpy, vi, nullptr, GL_TRUE);
    if
        (nullptr == ctx)
    {
        XFree(vi);
        vi = nullptr;
        XDestroyWindow(dpy, root);
        XCloseDisplay(dpy);
        dpy = nullptr;

        std::cerr << " failed to create context" << std::endl;
        return false;
    }

    glXMakeCurrent(dpy, root, ctx);

    const GLubyte* glVersionRaw = glGetString(GL_VERSION);
    std::string glVersion {reinterpret_cast<const char*>(glVersionRaw)};

    /* cleanup */
    glXMakeCurrent(dpy, None, nullptr);
    glXDestroyContext(dpy, ctx);
    ctx = nullptr;
    XFree(vi);
    vi = nullptr;
    XDestroyWindow(dpy, root);
    XCloseDisplay(dpy);
    dpy = nullptr;

    std::transform(glVersion.cbegin(), glVersion.cend(), glVersion.begin(),
                   [] (unsigned char c) {return std::tolower(c);});

    if
        ((std::string::npos != glVersion.find("mesa"))
        && (0 == glVersion.compare(0, 3, "3.0")))
    {
        std::cerr << "\nSEAScope cannot start due to the OpenGL version reported by your GPU driver (MESA)" << std::endl;
        std::cerr << "If you cannot upgrade your driver to a version where this bug is fixed," << std::endl;
        std::cerr << "you can still use the following oneliner to start SEAScope:" << std::endl;
        std::cerr << "\nMESA_GL_VERSION_OVERRIDE=3.3 MESA_GLSL_VERSION_OVERRIDE=330 ./seascope" << std::endl;

        return false;
    }

    return true;
}

/**
 *
 */
AppLauncherImpl::AppLauncherImpl()
{
}

/**
 *
 */
AppLauncherImpl::~AppLauncherImpl()
{
}

/**
 *
 */
bool AppLauncherImpl::initWorkspace(const std::string& /*programPath*/,
                                    bool hasWorkspace, std::string& workspacePath,
                                    bool hasConfig, std::string& configPath,
                                    bool hasListen, std::string& listenURL,
                                    bool hasLogFilters, std::string& logFilters) const
{
    if
        (false == hasWorkspace)
    {
        Path::currentWorkingDir(workspacePath);
    }
    else
    {
        std::string workspacePathTmp = workspacePath;
        Path::absolutePath(workspacePathTmp, workspacePath);
    }
    Path::setPathValid(workspacePath);

    if
        (false == hasConfig)
    {
        configPath = workspacePath + Path::separator() + "config.ini";
    }
    else
    {
        std::string configPathTmp = configPath;
        Path::absolutePath(configPathTmp, configPath);
    }
    Path::setPathValid(configPath);

    if
        (false == Path::fileExists(configPath))
    {
        bool canBeFixed {false};
        canBeFixed = Path::writeFile(configPath);
        if
            (false == canBeFixed)
        {
            std::cerr << "Cannot write to file '" << configPath << "', please check permissions" << std::endl;
            return false;
        }

        // Default config
        mINI::INIStructure root;
        root["paths"];
        root["paths"]["colormaps"] = workspacePath + Path::separator() + "colormaps";
        root["paths"]["data"] = workspacePath + Path::separator() + "data";
        root["paths"]["state"] = workspacePath + Path::separator() + "state.db";
        root["paths"]["index"] = workspacePath + Path::separator() + "index.fb";
        root["paths"]["annotations"] = workspacePath + Path::separator() + "annotations.db";
        root["paths"]["custom"] = workspacePath + Path::separator() + "custom";

        mINI::INIFile parser(configPath);
        bool ok = parser.write(root, true /*pretty print*/);
        if
            (false == ok)
        {
            std::cerr << "Could not write default configuration to '" << configPath << "'" << std::endl;
            return false;
        }
    }

    if
        (false == hasListen)
    {
        listenURL = "127.0.0.1:11155";
    }

    if
        (false == hasLogFilters)
    {
        logFilters = "info";
    }


    bool ok {false};
    ok = checkMESAVersion();

    return ok;
}
