// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cstring>
#include <cerrno>
#include <cstdio>
#include <sstream>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <dirent.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <vector>
#include "core/log.h"
#include "core/utils.h"
#include "platform/path_impl.h"

using namespace SEAScope;

/**
 *
 */
PathImpl::PathImpl()
{

};

/**
 *
 */
PathImpl::~PathImpl()
{

};

/**
 * Check if a file exists.
 */
bool PathImpl::fileExists(const std::string& path)
{
	struct stat buffer;
	return (stat (path.c_str(), &buffer) == 0);
};

/**
 * Check that it is possible to create or modify the file
 */
bool PathImpl::writeFile(const std::string& path)
{
	std::ofstream fileHandler;	//
	std::size_t slashPos;	//
	std::string dirname;	//
	bool ok;	//

	ok = false;

	slashPos = path.find_last_of("/");
	dirname = path.substr(0, slashPos + 1);
	if
		(false == this->dirExists(dirname))
	{
		ok = this->createDirs(dirname, 0755);
		if
			(false == ok)
		{
			return false;
		}
	}

	fileHandler.open(path, std::ofstream::out | std::ofstream::app);
	if
		(true == fileHandler.is_open())
	{
		ok = true;
		fileHandler.close();
	}
	return ok;
};

/**
 *
 */
bool PathImpl::fileSize(const std::string& path, std::int64_t& size)
{
    struct stat buffer;	//
	int rc;	//

    rc = stat(path.c_str(), &buffer);
	if
		(0 == rc)
	{
		size = buffer.st_size;
		return true;
	}
	else
	{
		size = 0;
		return false;
	}
};

/**
 * Check if a directory exists.
 */
bool PathImpl::dirExists(const std::string& path)
{
    struct stat buffer;

    return ((stat( path.c_str(), &buffer ) == 0) && (buffer.st_mode & S_IFDIR));
}

/**
 * Create directory
 */
bool PathImpl::createDir(const std::string& path, mode_t mode)
{
    struct stat buffer;
    if
        (0 != stat(path.c_str(), &buffer))
    {
        /* Path does not exist */
        if
            (  0 != mkdir(path.c_str(), mode)
            && errno != EEXIST)
        {
            /* mkdir failed and error code is not due to a race condition 
                => directory was not created */
			LOG( LogLevel::Error, "Unix path"
					, "Could not create directory '{}'"
					, path);
            return false;
        }
    }
    else if
        (!(buffer.st_mode & S_IFDIR))
    {
        /* Path exists but is not a directory */
        errno = ENOTDIR;
		LOG( LogLevel::Error, "Unix path"
				, "Path '{}' exists but is not a directory"
				, path);
        return false;
    }

    /* Directory created or already existed */
    return true;
};


/**
 * Create directory tree.
 */
bool PathImpl::createDirs(const std::string& path, mode_t mode)
{
    char* searchStart;
    char* slashPointer;
    char* cpath;
    bool ok;

    cpath = const_cast<char*>(path.c_str());
    ok = true;
    searchStart = cpath;
    slashPointer = std::strchr(searchStart, '/');
    while
        (  true == ok
        && nullptr != slashPointer)
    {
        if (slashPointer != searchStart)
        {
            // Neither root nor double slash in path
            *slashPointer = '\0'; /* artificially truncate the path at first slash found */
            ok = this->createDir(cpath, mode);
            *slashPointer = '/'; /* restore truncated path */
        }
        searchStart = slashPointer + 1;
        slashPointer = std::strchr(searchStart, '/');
    }
    if
        (true == ok)
    {
        ok = this->createDir(path, mode);
    }
    return ok;
};

/**
 *
 */
void PathImpl::splitPaths(const std::string& paths, const std::string& delim, std::forward_list<std::string>& result)
{
    unsigned long start = 0;
    auto end = paths.find(delim);
    auto it = result.before_begin();

    while
        (end != std::string::npos)
    {
        it = result.insert_after(it, paths.substr(start, end - start));
        start = end + delim.length();
        end = paths.find(delim, start);
    }
    result.insert_after(it, paths.substr(start, end));
};

/**
 *
 */
void PathImpl::joinPaths(const std::vector<std::string>& paths, std::string& result)
{
	std::ostringstream buffer;	//

	for
		(const auto& path : paths)
	{
		buffer << path << std::string("/");
	}
	result = buffer.str();
    //std::cout << result << "?" << std::endl;
	result.pop_back(); // Remove trailing slash
    //std::cout << result << "#" << std::endl;
};

/**
 *
 */
bool PathImpl::listFiles(const std::string& path, std::forward_list<std::string>& result)
{
	bool ok;	//
	DIR *dir;	//
	struct dirent *ent;	//
    auto it = result.before_begin();

    std::string delimiter("/");
    if
        (  (path.size() >= 1)
        && (0 == path.compare(path.size() - 1, 1, delimiter)))
    {
        delimiter = "";
    }

	ok = PathImpl::dirExists(path);
	if
		(false == ok)
	{
		return false;
	}

	dir = opendir(path.c_str());
	if
		(nullptr == dir)
	{
		return false;
	}

	ent = readdir(dir);
	while
		(nullptr != ent)
	{
		it = result.insert_after(it, path + delimiter + ent->d_name);
		ent = readdir(dir);
	}
    closedir (dir);
	return true;
};

/**
 *
 */
bool PathImpl::listFilesRecursive(const std::string& path, std::forward_list<std::string>& result)
{
	bool ok;				//
	DIR* dir;				//
	struct dirent* ent;		//
	std::vector<std::string> subdirs;	//
    struct stat buffer;		//
	int rc;					//
	std::string entPath;	//

    std::string delimiter("/");
    if
        (  (path.size() >= 1)
        && (0 == path.compare(path.size() - 1, 1, delimiter)))
    {
        delimiter = "";
    }

	ok = false;
	dir = nullptr;
	ent = nullptr;

	ok = PathImpl::dirExists(path);
	if
		(false == ok)
	{
		return false;
	}

	dir = opendir(path.c_str());
	if
		(nullptr == dir)
	{
		return false;
	}

	auto it = result.before_begin();
	ent = readdir(dir);
	while
		(nullptr != ent)
	{
		entPath = path + delimiter + ent->d_name;
		rc = stat(entPath.c_str(), &buffer);
		if
			(0 != rc)
		{
			/* cleanup and leave */
			closedir(dir);
			return false;
		}

		if
			(  0 == strncmp(".", ent->d_name, 1)
			|| 0 == strncmp("..", ent->d_name, 2))
		{
			/* do nothing */
		}
		else if
			(0 != (buffer.st_mode & S_IFDIR))
		{
			subdirs.push_back(entPath);
		}
		else
		{
			it = result.insert_after(it, entPath);
		}
		ent = readdir(dir);
	}
	closedir(dir);

	for
		(const auto& subdir : subdirs)
	{
		ok = PathImpl::listFilesRecursive(subdir, result);
		if
			(false == ok)
		{
			return false;
		}
	}

	return true;
};

/**
 *
 */
bool PathImpl::deleteFile(const std::string& path)
{
	bool ok;	//
	int rc;	// Return code

	ok = this->fileExists(path);
	if
		(false == ok)
	{
		return false;
	}

	rc = remove(path.c_str());
	ok &= (0 == rc);
	return ok;
}

/**
 *
 */
void SEAScope::PathImpl::fixPlatformPath(std::string&)
{
}

/**
 *
 */
char SEAScope::PathImpl::separator() const
{
	return '/';
}

/**
 *
 */
bool PathImpl::lastModified(const std::string& path, std::time_t& result)
{
	int rc;	// Return code
	struct stat buffer;	//

	rc = stat(path.c_str(), &buffer);
	if
		(0 != rc)
	{
		return false;
	}

	result = buffer.st_mtime;
	return true;
}

/**
 *
 */
bool PathImpl::currentWorkingDir(std::string& path)
{
    char* pathChars = getcwd(nullptr, 0); // Buffer is allocated dynamically
    path.assign(pathChars);
    free(pathChars);
    return true;
}

/**
 *
 */
bool PathImpl::canonicalPath(const std::string& inputPath, std::string& outputPath)
{
    std::string delimiter(1, this->separator());
    std::forward_list<std::string> pathElements;
    std::uint16_t ignoredCount(0u);

    outputPath = "";

    this->splitPaths(inputPath, delimiter, pathElements);
    pathElements.reverse();
    for
        (const auto& pathElement : pathElements)
    {
        if
            (".." == pathElement)
        {
            ignoredCount = ignoredCount + 1;
        }
        else if
            (0 < ignoredCount)
        {
            ignoredCount = ignoredCount - 1;
        }
        else if
            ("." == pathElement)
        {
            // noop
        }
        else if
            (0 < outputPath.size())
        {
            outputPath = pathElement + delimiter + outputPath;
        }
        else
        {
            outputPath = pathElement;
        }
    }

    return true;
}

/**
 *
 */
bool PathImpl::homeDir(std::string& outputPath)
{
    uid_t userId(0);
    struct passwd* pwd(nullptr);

    userId = getuid();
    pwd = getpwuid(userId);
    if
        (nullptr == pwd)
    {
        return false;
    }
    outputPath.assign(pwd->pw_dir);

    // man page states that the pointer returned by getpwuid can be reused by
    // subsequent calls and should therefore not be passed to free()

    return true;
}

/**
 *
 */
bool PathImpl::absolutePath(const std::string& path, std::string& absolutePath)
{
    std::string canonicalPath;
    std::string fullPath;

    std::string parentPath("");
    std::string delimiter(1, this->separator());
    bool relative(false);
    std::size_t lastSeparatorPosition = path.rfind(delimiter);

    relative = (  (std::string::npos == lastSeparatorPosition) /* no separator at all */
               || (0 != path.compare(0, delimiter.size(), delimiter)) /* Linux/macOS: path does not start with separator */
               );
    if
        (true == relative)
    {
        // relative path
        this->currentWorkingDir(parentPath);
        if
            (0 == parentPath.compare(parentPath.size() - delimiter.size(), delimiter.size(), delimiter))
        {
            // No delimiter needed as the parent path already ends with one
            fullPath = parentPath + path;
        }
        else
        {
            fullPath = parentPath + delimiter + path;
        }
    }
    else
    {
        fullPath = path;
    }

    this->canonicalPath(fullPath, canonicalPath);

    // Preserve trailing delimiter if it was in the input
    if
        (  (0 >= path.size())
        || ((path.size() >= delimiter.size())
           && (0 == path.compare(path.size() - delimiter.size(), delimiter.size(), delimiter))))
    {
        // There is a trailing delimiter in the input, let's find out if
        // it is already there in the modifier path
        if
            (canonicalPath.size() < delimiter.size()
            || (0 != canonicalPath.compare(canonicalPath.size() - delimiter.size(), delimiter.size(), delimiter)))
        {
            // Append a delimiter
            canonicalPath = canonicalPath + delimiter;
        }
    }

    absolutePath = canonicalPath;
    return relative;
}

bool PathImpl::getParentDir(const std::string& inputPath, std::string& outputPath)
{
	std::size_t lastSeparatorPosition {0};
    std::string separator(1, this->separator());

    if
        (inputPath.length() < separator.length())
    {
        // inputPath is too short
        return false;
    }

    std::ptrdiff_t separatorCount = std::count(inputPath.begin(), inputPath.end(), this->separator());
    std::size_t n {0};
    if
        ((1 < separatorCount) && (true == SEAScope::stringEndsWith(inputPath, separator)))
    {
        // Input path has a trailing separator that we must ignore
        n = 1;
    }

    lastSeparatorPosition = inputPath.rfind(separator, inputPath.size() - n * separator.size() - 1);

    outputPath = inputPath.substr(0, lastSeparatorPosition + separator.size() /*trailing separator included*/);

    return 0 != inputPath.compare(outputPath);
}
