// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <dlfcn.h>

#include "core/log.h"
#include "core/path.h"
#include "platform/plugins_loader_impl.h"

using namespace SEAScope;

/**
 *
 */
PluginsLoaderImpl::PluginsLoaderImpl()
{
};

/**
 *
 */
PluginsLoaderImpl::~PluginsLoaderImpl()
{
};

/**
 *
 */
bool PluginsLoaderImpl::findPlugins(const std::string& path, std::set<std::string>& results)
{
	std::forward_list<std::string> apiDirs;	//
	std::forward_list<std::string> files;	//
	std::string basename;	//
	std::size_t slashPos;	//
	std::string extension;	//
	bool isDir;	//
	bool ok;	//

	ok = Path::listFiles(path, apiDirs);
	if
		(false == ok)
	{
		return false;
	}

	for
		(const auto& subpath : apiDirs)
	{
		slashPos = subpath.find_last_of("/");
		basename = subpath.substr(slashPos+1);
		isDir = Path::dirExists(subpath);
		if
			(  0 >= subpath.size()
			|| false == isDir
			|| 0 >= basename.compare(0, 1, "."))
		{
			continue;
		}

		// List files and only keep .so
		ok = Path::listFiles(subpath, files);
		if
			(false == ok)
		{
			LOG( LogLevel::Warn, "PluginsLoaderImpl"
					, "Could not list files in {}"
					, subpath);
			continue;
		}

		for
			(const auto& filepath : files)
		{
			extension = filepath.substr(filepath.size() - 3);
			if
				(".so" == extension)
			{
				results.insert(filepath);
			}
		}
	}
	return true;
};

/**
 *
 */
bool PluginsLoaderImpl::loadPlugin(PluginMeta* meta)
{
	bool ok;	//

	if
		(true == meta->path.empty())
	{
		LOG(LogLevel::Error, "PluginsLoaderImpl", "loadPlugin failed because path is empty");
		return false;
	}

	/*	
	LOG( LogLevel::Debug, "PluginsLoaderImpl"
			, "loadPlugin: {}", meta->path);
	*/
	ok = Path::fileExists(meta->path);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "PluginsLoaderImpl"
				, "Plugin {} does not exist"
				, meta->path);
		return false;
	}

	meta->handle = dlopen(meta->path.c_str(), RTLD_NOW);
	if
		(nullptr == meta->handle)
	{
		LOG( LogLevel::Error, "PluginsLoaderImpl"
				, "Could not load plugin");
		LOG( LogLevel::Debug, "PluginsLoaderImpl"
				, dlerror());
		return false;
	}

	meta->loaded = true;

	return true;
};

/**
 *
 */
bool PluginsLoaderImpl::releasePlugin(PluginMeta* meta)
{
	int rc;	//

	if
		(false == meta->loaded)
	{
		/*Plugin is not loaded => no need to release*/
		LOG( LogLevel::Debug, "PluginsLoaderImpl"
				, "releasePlugin: was not loaded in the first place {}", meta->path);
		return true;
	}

	if
		(0 < meta->refCount)
	{
		LOG( LogLevel::Warn, "PluginsLoaderImpl"
				, "Cannot release plugin that is still in use: {} instances are active"
				, meta->refCount);
		return false;
	}

	/*
	LOG( LogLevel::Debug, "PluginsLoaderImpl"
			, "releasePlugin: {}", meta->path);
	*/
	rc = dlclose(meta->handle);
	if
		(0 < rc)
	{
		LOG( LogLevel::Error, "PluginsLoaderImpl"
				, "Could not release plugin {}"
				, meta->path);
		LOG( LogLevel::Debug, "PluginsLoaderImpl"
				, dlerror());
		return false;
	}

    dlerror();

	meta->loaded = false;
    meta->handle = nullptr;
    meta->create = nullptr;
    meta->destroy = nullptr;
	return true;
};

/**
 *
 */
bool PluginsLoaderImpl::getPluginInfo(PluginMeta* meta)
{
	void* ptr;	//
	const char* (*fName)();	//
	const char* (*fVersion)();	//
	const char* (*fDescr)();	//
	const char* (*fAPI)();	//

	ptr = dlsym(meta->handle, "pluginName");
	if
		(nullptr == ptr)
	{
		LOG( LogLevel::Error, "PluginsLoaderImpl"
				, "Plugin does not provide a pluginName entrypoint");
		LOG( LogLevel::Debug, "PluginsLoaderImpl"
				, dlerror());
		this->releasePlugin(meta);
		return false;
	}
	fName = reinterpret_cast<const char *(*)()>(ptr);
	meta->name = fName();

	ptr = dlsym(meta->handle, "pluginVersion");
	if
		(nullptr == ptr)
	{
		LOG( LogLevel::Error, "PluginsLoaderImpl"
				, "Plugin does not provide a pluginVersion entrypoint");
		LOG( LogLevel::Debug, "PluginsLoaderImpl"
				, dlerror());
		this->releasePlugin(meta);
		return false;
	}
	fVersion = reinterpret_cast<const char*(*)()>(ptr);
	meta->version = fVersion();

	ptr = dlsym(meta->handle, "pluginDescription");
	if
		(nullptr == ptr)
	{
		LOG( LogLevel::Error, "PluginsLoaderImpl"
				, "Plugin does not provide a pluginDescription entrypoint");
		LOG( LogLevel::Debug, "PluginsLoaderImpl"
				, dlerror());
		this->releasePlugin(meta);
		return false;
	}
	fDescr = reinterpret_cast<const char*(*)()>(ptr);
	meta->description = fDescr();

	ptr = dlsym(meta->handle, "pluginAPI");
	if
		(nullptr == ptr)
	{
		LOG( LogLevel::Error, "PluginsLoaderImpl"
				, "Plugin does not provide a pluginAPI entrypoint");
		LOG( LogLevel::Debug, "PluginsLoaderImpl"
				, dlerror());
		this->releasePlugin(meta);
		return false;
	}
	fAPI = reinterpret_cast<const char*(*)()>(ptr);
	meta->api = fAPI();

	ptr = dlsym(meta->handle, "createPlugin");
	if
		(nullptr == ptr)
	{
		LOG( LogLevel::Error, "PluginsLoaderImpl"
				, "Plugin does not provide a createPlugin entrypoint");
		LOG( LogLevel::Debug, "PluginsLoaderImpl"
				, dlerror());
		this->releasePlugin(meta);
		return false;
	}
	meta->create = reinterpret_cast<PluginInstanciator>(ptr);

	ptr = dlsym(meta->handle, "destroyPlugin");
	if
		(nullptr == ptr)
	{
		LOG( LogLevel::Error, "PluginsLoaderImpl"
				, "Plugin does not provide a destroyPlugin entrypoint");
		LOG( LogLevel::Debug, "PluginsLoaderImpl"
				, dlerror());
		this->releasePlugin(meta);
		return false;
	}
	meta->destroy = reinterpret_cast<PluginDestroyer>(ptr);
	/*
	LOG( LogLevel::Debug, "PluginsLoaderImpl", "instanciator ok? {}", nullptr != meta->instanciator);
	*/

    dlerror();

	return true;
};
