// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <cstdio>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sstream>

#include "core/log.h"
#include "platform/socket_factory_impl.h"

using namespace SEAScope;

/**
 *
 */
SocketFactoryImpl::SocketFactoryImpl()
{
};

/**
 *
 */
SocketFactoryImpl::~SocketFactoryImpl()
{
};

/**
 *
 */
socket_t SocketFactoryImpl::create(const std::string& address, std::int64_t port, bool& ok)
{
	struct addrinfo hints;	//
	struct addrinfo* addresses;	//
	struct addrinfo* currentAddr(nullptr);	//
	struct sockaddr a;	//
	struct sockaddr_in b;	//
	std::ostringstream o;	//
	socklen_t l(0);	//
	int result(-1);	//
	int rc(-1);	//

	addresses = nullptr;

	if
		(0 >= address.length())
	{
		o << port;
		// Bind to any address
		memset(&hints, 0, sizeof(struct addrinfo));
		hints.ai_family = AF_UNSPEC;
		hints.ai_socktype = SOCK_STREAM;
		hints.ai_flags = AI_PASSIVE;
		rc = getaddrinfo(nullptr, o.str().c_str(), &hints, &addresses);
		if
			(0 != rc)
		{
			ok = false;
			LOG( SEAScope::LogLevel::Error, "Unix socket factory"
					, "Could not get address information: {}"
					, gai_strerror(rc));
			return -1;
		}

		// Try to create socket on addresses until it works
		result = -1;
		currentAddr = addresses;
		for
			(; nullptr != currentAddr; currentAddr = currentAddr->ai_next)
		{
			result = socket( currentAddr->ai_family, currentAddr->ai_socktype
						   , currentAddr->ai_protocol);
			if
				(-1 < result)
			{
				l = currentAddr->ai_addrlen;
                std::memcpy(&a, currentAddr->ai_addr, l);
				break;
			}
		}
	}
	else
	{
        std::string ipAddress;  //
        ok = this->resolveHostname(address, ipAddress);
        if
            (false == ok)
        {
            return -1;
        }

		memset(&b, 0, sizeof(struct sockaddr_in));
		b.sin_family = AF_INET;
		b.sin_addr.s_addr = inet_addr(ipAddress.c_str());
		b.sin_port = htons(static_cast<std::uint16_t>(port));
		l = sizeof(b);
		//a = (struct sockaddr*)(&b);
        std::memcpy(&a, &b, l);
		result = socket(AF_INET, SOCK_STREAM, 0);
	}

	if
		(-1 == result)
	{
		ok = false;
		LOG( SEAScope::LogLevel::Error, "Unix socket factory"
				, "Could not create socket");
		return -1;
	}

    int success {-1};
    int flag {1};
    success = setsockopt(result, SOL_SOCKET, SO_REUSEPORT, &flag, sizeof(flag));
    if
        (-1 == success)
    {
        LOG(LogLevel::Error, "Unix socket factory",
                 "Failed to set the SO_REUSEPORT flag on the socket");
        return -1;
    }

	LOG( SEAScope::LogLevel::Debug, "Unix socket factory"
			, "Binding to {}", a.sa_data);
	rc = bind(result, &a, l);
	if
		(0 != rc)
	{
		close(result);
		ok = false;
		LOG( SEAScope::LogLevel::Error, "Unix socket factory"
				, "Could not bind socket: {}"
				, gai_strerror(rc));
		return -1;
	}

	if
		(nullptr != addresses)
	{
		freeaddrinfo(addresses);
	}

	ok = true;
	return result;
};

/**
 *
 */
bool SocketFactoryImpl::nonBlocking(socket_t socket)
{
	int flags(-1);	//
	int rc(-1);	//

	flags = fcntl(socket, F_GETFL, 0);
	if
		(-1 == flags)
	{
		LOG( SEAScope::LogLevel::Error, "Unix socket factory"
				, "Could not retrieve socket flags.");
		return false;
    }

	flags |= O_NONBLOCK;
	rc = fcntl(socket, F_SETFL, flags);
	if
		(-1 == rc)
	{
		LOG( SEAScope::LogLevel::Error, "Unix socket factory"
				, "Could not set non-blocking flag.");
		return false;
    }

	return true;
};

/**
 *
 */
bool SocketFactoryImpl::shutdown(socket_t sock)
{
	int rc(-1);	//

	rc = close(sock);
	return -1 != rc;
};

/**
 *
 */
bool SocketFactoryImpl::listening(socket_t sock)
{
	int rc(-1);	//

	rc = listen(sock, SOMAXCONN);
	return -1 != rc;
};

/**
 *
 */
socket_t SocketFactoryImpl::connect(const std::string& address, std::int64_t port, bool& ok)
{
    struct sockaddr a; //
	struct sockaddr_in serverAddress;	//
    socklen_t l(0);    //
	int result(-1);	//
	int rc(-1);	//

    result = socket(AF_INET,SOCK_STREAM, 0);
    if
        (-1 == result)
    {
        LOG(SEAScope::LogLevel::Error, "Unix socket factory", "Failed to create client socket");
        ok = false;
        return -1;
    }

    std::string ipAddress;  //
    ok = this->resolveHostname(address, ipAddress);
    if
        (false == ok)
    {
        return -1;
    }

	memset(&serverAddress, 0, sizeof(struct sockaddr_in));
    serverAddress.sin_family = AF_INET;
	serverAddress.sin_port = htons(static_cast<std::uint16_t>(port));
	inet_pton(AF_INET, ipAddress.c_str(), &(serverAddress.sin_addr));
    l = sizeof(serverAddress);
    //a = (struct sockaddr*)(&serverAddress);
    std::memcpy(&a, &serverAddress, l);

    rc = -1;
    rc = ::connect(result, &a, l);
	if
		(0 != rc)
	{
		close(result);
		ok = false;
		LOG( SEAScope::LogLevel::Error, "Unix socket factory"
				, "Could not connect with socket: {}"
				, gai_strerror(rc));
		return -1;
	}

	ok = true;
    return result;
};

/**
 *
 */
bool SocketFactoryImpl::resolveHostname(const std::string& name, std::string& ipAddress)
{
    struct hostent* hostEntry{nullptr}; //

    hostEntry = gethostbyname(name.c_str());
    if
        (nullptr == hostEntry)
    {
        LOG(LogLevel::Error, "Unix socket factory", "Failed to resolve hostname '{}'", name);
        return false;
    }

    ipAddress = inet_ntoa(*reinterpret_cast<struct in_addr*>(hostEntry->h_addr_list[0]));

    LOG(LogLevel::Debug, "Unix socket factory", "Translated '{}' to '{}'", name, ipAddress);

    return true;
}
