// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "platform/plugins_loader_impl.h"

#include <forward_list>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#undef max
#undef min

#include "core/path.h"
#include "core/log.h"

using namespace SEAScope;

/**
 *
 */
PluginsLoaderImpl::PluginsLoaderImpl()
{
};

/**
 *
 */
PluginsLoaderImpl::~PluginsLoaderImpl()
{
};

/**
 *
 */
bool PluginsLoaderImpl::findPlugins(const std::string& path, std::set<std::string>& results)
{
	std::forward_list<std::string> apiDirs;	//
	std::forward_list<std::string> files;	//
	std::string basename;	//
	std::size_t slashPos;	//
	std::string extension;	//
	bool isDir;	//
	bool ok;	//

	ok = Path::listFiles(path, apiDirs);
	if
		(false == ok)
	{
		return false;
	}

	for
		(const auto& subpath : apiDirs)
	{
		slashPos = subpath.find_last_of(SEAScope::Path::separator());
		basename = subpath.substr(slashPos + 1);
		isDir = Path::dirExists(subpath);
		if
			(0 >= subpath.size()
				|| false == isDir
				|| 0 >= basename.compare(0, 1, "."))
		{
			continue;
		}

		// List files and only keep .so
		ok = Path::listFiles(subpath, files);
		if
			(false == ok)
		{
			LOG(LogLevel::Warn, "Win32PluginsLoader"
				, "Could not list files in {}"
				, subpath);
			continue;
		}

		for
			(const auto& filepath : files)
		{
			extension = filepath.substr(filepath.size() - 4);
			if
				(".dll" == extension)
			{
				results.insert(filepath);
			}
		}
	}
	return true;
};

/**
 *
 */
bool PluginsLoaderImpl::loadPlugin(PluginMeta* meta)
{
	bool ok;	//

	if
		(true == meta->path.empty())
	{
		LOG(LogLevel::Error, "Win32PluginsLoader", "loadPlugin failed because path is empty");
		return false;
	}

	ok = Path::fileExists(meta->path);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "Win32PluginsLoader"
			, "Plugin {} does not exist"
			, meta->path);
		return false;
	}

	meta->handle = LoadLibrary(meta->path.c_str());

	if
		(nullptr == meta->handle)
	{
		LOG(LogLevel::Error, "Win32PluginsLoader"
			, "Could not load plugin");

		return false;
	}
	meta->loaded = true;

	return true;
};

/**
 *
 */
bool PluginsLoaderImpl::releasePlugin(PluginMeta* meta)
{
	BOOL rc;	//

	if
		(false == meta->loaded)
	{
		/*Plugin is not loaded => no need to release*/
		LOG(LogLevel::Debug, "Win32PluginsLoader"
			, "releasePlugin: was not loaded in the first place {}", meta->path);
		return true;
	}

	if
		(0 < meta->refCount)
	{
		LOG(LogLevel::Warn, "Win32PluginsLoader"
			, "Cannot release plugin that is still in use: {} instances are active"
			, meta->refCount);
		return false;
	}

	/*
	LOG( LogLevel::Debug, "Win32PluginsLoader"
	, "releasePlugin: {}", meta->path);
	*/
	rc = FreeLibrary((HMODULE)meta->handle);
	if
		(rc == FALSE)
	{
		LOG(LogLevel::Error, "Win32PluginsLoader"
			, "Could not release plugin {}"
			, meta->path);

		return false;
	}
	meta->loaded = false;
	return true;
};

/**
 *
 */
bool PluginsLoaderImpl::getPluginInfo(PluginMeta* meta)
{
	void* ptr;	//
	const char* (*fName)();	//
	const char* (*fVersion)();	//
	const char* (*fDescr)();	//
	const char* (*fAPI)();	//

	
	ptr = GetProcAddress((HMODULE)meta->handle, "pluginName");

	if
		(nullptr == ptr)
	{
		LOG(LogLevel::Error, "Win32PluginsLoader"
			, "Plugin does not provide a pluginName entrypoint");

		this->releasePlugin(meta);
		return false;
	}
	fName = reinterpret_cast<const char *(*)()>(ptr);
	meta->name = fName();

	ptr = GetProcAddress((HMODULE)meta->handle, "pluginVersion");

	if
		(nullptr == ptr)
	{
		LOG(LogLevel::Error, "Win32PluginsLoader"
			, "Plugin does not provide a pluginVersion entrypoint");

		this->releasePlugin(meta);
		return false;
	}
	fVersion = reinterpret_cast<const char*(*)()>(ptr);
	meta->version = fVersion();

	ptr = GetProcAddress((HMODULE)meta->handle, "pluginDescription");

	if
		(nullptr == ptr)
	{
		LOG(LogLevel::Error, "Win32PluginsLoader"
			, "Plugin does not provide a pluginDescription entrypoint");

		this->releasePlugin(meta);
		return false;
	}
	fDescr = reinterpret_cast<const char*(*)()>(ptr);
	meta->description = fDescr();

	ptr = GetProcAddress((HMODULE)meta->handle, "pluginAPI");

	if
		(nullptr == ptr)
	{
		LOG(LogLevel::Error, "Win32PluginsLoader"
			, "Plugin does not provide a pluginAPI entrypoint");

		this->releasePlugin(meta);
		return false;
	}
	fAPI = reinterpret_cast<const char*(*)()>(ptr);
	meta->api = fAPI();

	ptr = GetProcAddress((HMODULE)meta->handle, "createPlugin");

	if
		(nullptr == ptr)
	{
		LOG(LogLevel::Error, "Win32PluginsLoader"
			, "Plugin does not provide a createPlugin entrypoint");

		this->releasePlugin(meta);
		return false;
	}
	meta->create = reinterpret_cast<PluginInstanciator>(ptr);

	ptr = GetProcAddress((HMODULE)meta->handle, "destroyPlugin");

	if
		(nullptr == ptr)
	{
		LOG(LogLevel::Error, "Win32PluginsLoader"
			, "Plugin does not provide a destroyPlugin entrypoint");

		this->releasePlugin(meta);
		return false;
	}
	meta->destroy = reinterpret_cast<PluginDestroyer>(ptr);
	return true;
};
