// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <string>
#include <sstream>
#include <vector>
#include <mutex>
#include <iostream>

#include <WinSock2.h>
#include <Ws2tcpip.h>


#include "platform/server.h"
#include "core/perfs.h"
#include "core/sockets_manager.h"

Server::Server()
{
}

Server::~Server()
{
}

void Server::initialize(const std::string& address, int port, SEAScope::SocketsManager* socketManager)
{
	_address = address;
	_port = port;
	_socketManager = socketManager;
}

bool Server::run()
{
	{
		std::unique_lock<std::mutex> lock(_mutex);
		_stopThread = false;
	}

	struct sockaddr* a;	//
	struct sockaddr_in b;	//
	std::ostringstream o;	//
	socklen_t l;	//
	int rc;	//

	memset(&b, 0, sizeof(struct sockaddr_in));
	b.sin_family = AF_INET;
	b.sin_addr.s_addr = inet_addr(_address.c_str());
	b.sin_port = htons(static_cast<u_short>(_port));
	a = (struct sockaddr*)(&b);
	l = sizeof(b);
	_listener.socket = socket(AF_INET, SOCK_STREAM, 0);

	rc = bind(_listener.socket, a, l);
	if (0 != rc)
	{
		//int error = 0;
	}

	_listener.event = WSACreateEvent();

	int res = WSAEventSelect(_listener.socket, _listener.event, FD_ACCEPT | FD_CLOSE | FD_READ);
	if (0 != res)
	{
		//int error = 0;
	}

	res = listen(_listener.socket, SOMAXCONN);
	if (0 != res)
	{
		//int error = 0;
	}

	_clients.push_back(&_listener);

	updateEvents();

	bool done = false;
	DWORD dw;

	while (false == done)
	{
		dw = WSAWaitForMultipleEvents(static_cast<DWORD>(_events.size()), _events.data(), FALSE, 100, FALSE);

		if (WSA_WAIT_TIMEOUT == dw)
		{
			{
				std::unique_lock<std::mutex> lock(_mutex);

				if (true == _stopThread)
				{
					done = true;
				}
			}
		}
		else
		{
			dw -= WSA_WAIT_EVENT_0;

			SocketData& client = *_clients[dw];

			if ((WSA_WAIT_EVENT_0 + _events.size()) > dw)
			{
				WSANETWORKEVENTS networkEvent;
				res = WSAEnumNetworkEvents(client.socket, client.event, &networkEvent);

				if (0 != res)
				{
					//int error = 0;
					break;
				}

				if (networkEvent.lNetworkEvents & FD_ACCEPT)
				{
					acceptConnection(client);
				}
				else if (networkEvent.lNetworkEvents & FD_READ)
				{
					readSocket(client);
				}
				else if (networkEvent.lNetworkEvents & FD_CLOSE)
				{
					closeSocket(client);

					delete _clients[dw];

					_clients.erase(_clients.begin() + dw);

					updateEvents();
				}
			}
		}
	}

	return true;
}

void Server::stop()
{
	std::unique_lock<std::mutex> lock(_mutex);
	_stopThread = true;
}

bool Server::acceptConnection(SocketData& s)
{
	SocketData* newClient = new SocketData;
	newClient->socket = accept(s.socket, NULL, NULL);

	newClient->event = WSACreateEvent();

	int res = WSAEventSelect(newClient->socket, newClient->event, FD_CLOSE | FD_READ);
	if (0 != res)
	{
		//int error = 0;

		return false;
	}

	bool ok;
	newClient->protocol = _socketManager->getProtocolHandler("SEASCOPE_CONTROL", ok);

	_clients.push_back(newClient);

	updateEvents();

	return true;
}

bool Server::readSocket(SocketData& s)
{
	std::uint64_t left;
	std::uint8_t* res = s.buffer.getPointer(left);
	std::uint64_t toRead = (SOCKET_CHUNKSIZE > left) ? left : SOCKET_CHUNKSIZE;
	std::int64_t loaded = ::recv(s.socket, (char*)res, static_cast<int>(toRead), 0);

	std::uint64_t consumed;
	bool msgComplete;

	while
		(0 < loaded)
	{
		std::uint64_t cumulatedRead;	//
		cumulatedRead = 0;
		while
			(cumulatedRead < static_cast<uint64_t>(loaded))
		{
			msgComplete = false;
			msgComplete = s.protocol->parseStream(static_cast<SEAScope::socket_t>(s.socket), res, cumulatedRead, loaded - cumulatedRead, consumed);
			cumulatedRead += consumed;
		}

		left = 0;
		res = s.buffer.getPointer(left);
		toRead = (SOCKET_CHUNKSIZE > left) ? left : SOCKET_CHUNKSIZE;
		loaded = static_cast<std::int64_t>(recv(s.socket, (char*)res, static_cast<int>(toRead), 0));
	}

	return true;
}



bool Server::closeSocket(SocketData& s)
{
	int res = closesocket(s.socket);
	if (0 != res)
	{
		//int error = 0;
	}

	WSACloseEvent(s.event);

	return true;
}

void Server::updateEvents()
{
	_events.clear();

	for (auto client : _clients)
	{
		_events.push_back(client->event);
	}
}

