// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "platform/socket_factory_impl.h"

#include <WinSock2.h>
#include <Ws2tcpip.h>

#include <sstream>

#include "core/log.h"

using namespace SEAScope;

/**
*
*/
SocketFactoryImpl::SocketFactoryImpl()
{
	WSADATA wsaData;
	WSAStartup(MAKEWORD(2, 2), &wsaData);
};

/**
*
*/
SocketFactoryImpl::~SocketFactoryImpl()
{
	WSACleanup();
};

/**
*
*/
socket_t SocketFactoryImpl::create(const std::string& address, std::int64_t port, bool& ok)
{
	struct addrinfo hints;	//
	struct addrinfo* addresses;	//
	struct addrinfo* currentAddr;	//
	struct sockaddr* a = nullptr;	//
	struct sockaddr_in b;	//
	std::ostringstream o;	//
	socklen_t l = 0;	//
	SOCKET result;	//
	int rc;	//

	addresses = nullptr;

	if
		(0 >= address.length())
	{
		o << port;
		// Bind to any address
		memset(&hints, 0, sizeof(struct addrinfo));
		hints.ai_family = AF_UNSPEC;
		hints.ai_socktype = SOCK_STREAM;
		hints.ai_flags = AI_PASSIVE;
		rc = getaddrinfo(NULL, o.str().c_str(), &hints, &addresses);
		if
			(0 != rc)
		{
			ok = false;
			LOG(LogLevel::Error, "Win32 socket factory"
				, "Could not get address information: {}"
				, gai_strerror(rc));
			return -1;
		}

		// Try to create socket on addresses until it works
		result = INVALID_SOCKET;
		currentAddr = addresses;
		for
			(; nullptr != currentAddr; currentAddr = currentAddr->ai_next)
		{
			result = socket(currentAddr->ai_family, currentAddr->ai_socktype
				, currentAddr->ai_protocol);
			if
				(INVALID_SOCKET != result)
			{
				a = currentAddr->ai_addr;
				l = static_cast<socklen_t>(currentAddr->ai_addrlen);
				break;
			}
		}
	}
	else
	{
        std::string ipAddress;  //
        ok = this->resolveHostname(address, ipAddress);
        if
            (false == ok)
        {
            return -1;
        }

		memset(&b, 0, sizeof(struct sockaddr_in));
		b.sin_family = AF_INET;
		b.sin_addr.s_addr = inet_addr(ipAddress.c_str());
		b.sin_port = htons(static_cast<u_short>(port));
		a = (struct sockaddr*)(&b);
		l = sizeof(b);
		result = socket(AF_INET, SOCK_STREAM, 0);
	}

	if
		(INVALID_SOCKET == result)
	{
		ok = false;
		LOG(LogLevel::Error, "Win32 socket factory"
			, "Could not create socket");
		return -1;
	}

	LOG(LogLevel::Debug, "Win32 socket factory"
		, "Binding to {}", a->sa_data);
	rc = bind(result, a, l);
	if
		(0 != rc)
	{
		closesocket(result);
		ok = false;
		LOG(LogLevel::Error, "Win32 socket factory"
			, "Could not bind socket: {}"
			, gai_strerror(rc));
		return -1;
	}

	if
		(nullptr != addresses)
	{
		freeaddrinfo(addresses);
	}

	ok = true;
	return static_cast<SEAScope::socket_t>(result);
};

/**
*
*/
bool SocketFactoryImpl::nonBlocking(socket_t socket)
{
	int rc;	//

	unsigned long mode = 1;

	rc = ioctlsocket(static_cast<SOCKET>(socket), FIONBIO, &mode);

	if (NO_ERROR != rc)
	{
		LOG(LogLevel::Error, "Win32 socket factory"
			, "Could not set non-blocking flag.");
		return false;
	}

	return true;
};

/**
*
*/
bool SocketFactoryImpl::shutdown(socket_t sock)
{
	int rc;	//

	rc = closesocket(static_cast<SOCKET>(sock));

	return -1 != rc;
};

/**
*
*/
bool SocketFactoryImpl::listening(socket_t sock)
{
	int rc;	//

	rc = listen(static_cast<SOCKET>(sock), SOMAXCONN);

	return -1 != rc;
};

/**
*
*/
socket_t SocketFactoryImpl::connect(const std::string& address, std::int64_t port, bool& ok)
{
	struct sockaddr* a;	//
	struct sockaddr_in serverAddress;	//
	socklen_t l;    //
	SOCKET result;	//
	int rc;	//

	result = INVALID_SOCKET;
	result = socket(AF_INET, SOCK_STREAM, 0);
	if
		(INVALID_SOCKET == result)
	{
		LOG(LogLevel::Error, "Win32SocketFactory", "Failed to create client socket");
		ok = false;
		return -1;
	}

    std::string ipAddress;  //
    ok = this->resolveHostname(address, ipAddress);
    if
        (false == ok)
    {
        return -1;
    }
	memset(&serverAddress, 0, sizeof(struct sockaddr_in));
	serverAddress.sin_family = AF_INET;
	serverAddress.sin_port = htons(static_cast<u_short>(port));
	inet_pton(AF_INET, ipAddress.c_str(), &(serverAddress.sin_addr));
	a = (struct sockaddr*)(&serverAddress);
	l = sizeof(serverAddress);

	rc = SOCKET_ERROR;
	rc = ::connect(result, a, l);
	if
		(SOCKET_ERROR != rc)
	{
		closesocket(result);
		ok = false;
		LOG(LogLevel::Error, "Win32SocketFactory"
			, "Could not connect with socket: {}"
			, gai_strerror(rc));
		return -1;
	}

	ok = true;
	return static_cast<SEAScope::socket_t>(result);
};

/**
 *
 */
bool SocketFactoryImpl::resolveHostname(const std::string& name, std::string& ipAddress)
{
    struct hostent* hostEntry{nullptr}; //

    hostEntry = gethostbyname(name.c_str());
    if
        (nullptr == hostEntry)
    {
        LOG(LogLevel::Error, "Win32SocketFactory", "Failed to resolve hostname '{}'", name);
        return false;
    }

    ipAddress = inet_ntoa(*reinterpret_cast<struct in_addr*>(hostEntry->h_addr_list[0]));

    LOG(LogLevel::Debug, "Win32SocketFactory", "Translated '{}' to '{}'", name, ipAddress);

    return true;
}
