#!/bin/bash

# =============================================================================
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2025 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
# =============================================================================
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-11-17
#
# Functions to extract information about the host and the environment used to
# compile the SEAScope client.
# =============================================================================

set -o nounset
set -o errexit
set -o pipefail

get_so_version()
{
  if ! has ldconfig; then
    return
  fi

  local libname;
  local version;
  local escaped_libname

  libname="${1}"; shift;
  escaped_libname="${libname//+/\\+}";
  set +o errexit
  version="$(ldconfig -v 2>/dev/null \
					| grep "${libname}" \
					| head -1 \
					| sed -E 's!.*[-=]>\s*'"${escaped_libname}"'\.so\.(.+)!\1!')";
  set -o errexit

	if [ -z "${version}" ];
	then
		>&2 echo "Library ${libname} not found by ldconfig"
		exit 1
	fi
	echo "${version}"
}

get_so_path()
{
  if ! has ldconfig; then
      return
  fi

  local libname;
	local so_path;

  libname="${1}"; shift;
  so_path="$(ldconfig -p 2>/dev/null \
					| grep "${libname}" \
					| head -1 \
					| sed -E 's/.*[-=]>\s*(.+)/\1/')";
	if [ -z "${so_path}" ];
	then
		>&2 echo "Shared library ${libname} not found"
		exit 1
	fi

	echo "${so_path}"
}

get_glibc_version()
{
    if ! has readelf; then
        return
    fi

	local so_path;
	local version;

  so_path="$(get_so_path "libstdc++")";
  version="$(readelf -sV "${so_path}" \
					| sed -n 's/.*@@GLIBCXX_//p' \
					| sort -u -V \
					| tail -1)"
	if [ -z "${version}" ];
	then
		>&2 echo "Could not read glibc symbols from ${so_path}"
		exit 1
	fi

	echo "${version}"
}

get_os_info()
{
  if [ -f /etc/os-release ]; then
    source /etc/os-release

    printf '%s %s\n' \
           "${REDHAT_SUPPORT_PRODUCT:-${PRETTY_NAME:-Unknown}}" \
           "${REDHAT_SUPPORT_PRODUCT_VERSION:-${VERSION_ID:-}}"
	elif [ -f /etc/issue ]; then
		sed -E 's/([^\]+)\\.*/\1/'  /etc/issue | head -1;
  elif has sw_vers; then
     echo "$(sw_vers -productName) $(sw_vers -productVersion) ($(sw_vers -buildVersion))"
	fi
}

get_arch()
{
	uname -m
}

get_hostname()
{
	uname -n
}

get_packager_name()
{
	git config user.name
}

get_date()
{
	date '+%Y-%m-%d'
}

get_tag_or_branch()
{
	local commit;
	local tag;

  commit="$(git rev-parse HEAD)";
  tag="$(git tag --contains "${commit}" | head -1)";

	if [ -n "${tag}" ];
	then
		echo "${tag}"
	else
		 git rev-parse --abbrev-ref HEAD
	fi	
}

get_commit()
{
	git rev-parse --short HEAD | head -c 7
}

build_info()
{
	local dst_path;

  dst_path="${1}"; shift;

	cat > "${dst_path}" <<EOF
#ifndef SEASCOPE_VERSION_H
#define SEASCOPE_VERSION_H

#include <map>
#include <sstream>
#include "vendors/_glm.h"

namespace SEAScope
{
  inline std::string getGLMVersion()
  {
    std::ostringstream buffer("");
    buffer << GLM_VERSION_MAJOR << ".";
    buffer << GLM_VERSION_MINOR << ".";
    buffer << GLM_VERSION_PATCH << ".";
    buffer << GLM_VERSION_REVISION;
    return buffer.str();
  }

	static std::map<std::string, std::string> getVersion() {
    std::map<std::string, std::string> v ({
      {"name", "$(get_tag_or_branch)"},
      {"ref", "$(get_commit)"},
      {"date", "$(get_date)"},
      {"host", "$(get_hostname)"},
      {"arch", "$(get_arch)"},
      {"os", "$(get_os_info)"},
      {"packager", "$(get_packager_name)"},
      {"stdc++_version", "$(get_so_version "libstdc++")"},
      {"glibc_version","$(get_glibc_version)"},
      {"glm_version", getGLMVersion()}
    });
    return v;
	};
};
#endif
EOF
}
