#!/bin/bash

# =============================================================================
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2025 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
# =============================================================================
# @author <ziad.khoury.hanna@oceandatalab.com>
# =============================================================================

set -o nounset
set -o errexit
set -o pipefail

# =============================================================================
# Compatibility Helpers
# =============================================================================
if readlink -f / >/dev/null 2>/dev/null;
then
	readlink_f() { readlink -f "${1}"; }
else
	# macOS's readlink  doesn't support "-f" but it has perl
	readlink_f() { perl -MCwd -le 'print Cwd::abs_path shift' "${1}"; }
fi
# =============================================================================

usage()
{
	echo -e "Usage:\nrelease.sh [-d] SHARE_DIR CLIENT_BRANCH [[EMBEDDED_SERVER_BRANCH] VERSION]"
}

build()
{
	local workspace;
	local target;
	local arch;
	local prefix;

	workspace="${1}"; shift;
	target="${1}"; shift;

	if [ 0 -lt $# ];
	then
		arch="${1}"; shift;
		workspace="${workspace}/${arch}";
		prefix="arch -${arch}";
	fi

	pushd "${workspace}/client";
	$prefix "developer/build_standalone" check;
	$prefix "developer/build_standalone" deps;
	$prefix "developer/build_standalone" cpp-api;
	$prefix "developer/build_standalone" "${target}";
	popd;
	pushd "${workspace}/embedded-server";
	$prefix "developer/build_plugins" check;
	$prefix "developer/build_plugins" deps macos;
	$prefix "developer/build_plugins" "${target}" macos;
	popd;
}

combine()
{
	local workspace;
	local path;

	workspace="${1}"; shift;
	path="${1}"; shift;

	lipo -create -output "${workspace}/arm64/${path}" \
		"${workspace}/arm64/${path}" "${workspace}/x86_64/${path}";
}

main()
{
	local target;
	local share_dir;
	local client_branch;
	local embedded_server_branch;
	local version;
	local workspace;

	target=release;
	if [ 0 -lt $# ] && [ "${1}" = "-d" ];
	then
		target=debug;
		shift;
	fi

	if [ 2 -gt $# ];
	then
		usage;
		exit 2;
	fi

	share_dir="${1}"; shift;
	client_branch="${1}"; shift;

	if [ 1 -lt $# ];
	then
		embedded_server_branch="${1}"; shift;
		version="${1}"; shift;
	elif [ 0 -lt $# ];
	then
		embedded_server_branch="${client_branch}";
		version="${1}"; shift;
	else
		embedded_server_branch="${client_branch}";
		version="$(date "+%Y%m%d")";
	fi

	workspace="$(pwd)/${target}_${client_branch//\//_}_${embedded_server_branch//\//_}_${version}";

	echo "share dir: ${share_dir}";
	echo "client branch: ${client_branch}";
	echo "embedded_server branch: ${embedded_server_branch}";
	echo "version: ${version}";
	echo "workspace: ${workspace}";
	echo;

	rm -rf "${workspace}";
	mkdir -p "${workspace}";

	git clone --branch "${client_branch}" 'https://git.oceandatalab.com/seascope/client.git' "${workspace}/client";
	git clone --branch "${embedded_server_branch}" 'https://git.oceandatalab.com/seascope/embedded-server.git' "${workspace}/embedded-server";

	if [ "$(arch)" == "arm64" ];
	then
		mkdir -p "${workspace}/arm64" "${workspace}/x86_64";

		cp -r "${workspace}/client" "${workspace}/x86_64/client";
		ln -s "${workspace}/client" "${workspace}/arm64/client";
		cp -r "${workspace}/embedded-server" "${workspace}/x86_64/embedded-server";
		ln -s "${workspace}/embedded-server" "${workspace}/arm64/embedded-server";

		build "${workspace}" "${target}" arm64;
		build "${workspace}" "${target}" x86_64;

		combine "${workspace}" "client/${target}/build/seascope";
		combine "${workspace}" "embedded-server/${target}_macos/build/embedded_datastream.so";
		combine "${workspace}" "embedded-server/${target}_macos/build/embedded_index.so";
	else
		MACOSX_UNIVERSAL=true \
		build "${workspace}" "${target}";
	fi

	cp -r "${share_dir}"/* "${workspace}/client/share/";

	# Ensure that files copied from the share_dir have the correct permissions
	find "${workspace}/client/share" -type d -exec chmod 0755 {} +;
	find "${workspace}/client/share" -type f -exec chmod 0644 {} +;

	mkdir -p "${workspace}/client/share/plugins/datastream";
	mkdir -p "${workspace}/client/share/plugins/index";
	cp "${workspace}/embedded-server/${target}_macos/build/embedded_datastream.so" "${workspace}/client/share/plugins/datastream";
	cp "${workspace}/embedded-server/${target}_macos/build/embedded_index.so" "${workspace}/client/share/plugins/index";

	"${workspace}/client/developer/build_standalone" app "${target}" "${version}";
	"${workspace}/client/developer/build_standalone" dmg "${target}" "${version}";

	mv "${workspace}/client/${target}/"SEAScope* "${workspace}";

	echo 'All done!';
	echo 'You can find the app and DMG in:';
	echo "  ${workspace}/";
	open "${workspace}";
}

main "$@"
exit 0
