﻿# =============================================================================
# @author: <ziad.khoury.hanna@oceandatalab.com>
# @date: 2019-03-27
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2025 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
# =============================================================================

# ENV vars:
# * SEASCOPE_APP_PATH is the posix path to SEAScope app
# * SEASCOPE_DATA_PATH is the posix path to SEAScope's data directory
# * DLOG_LEVEL is one of:
#   - "0": None
#   - "1": Error
#   - "2": Warning
#   - "3": Info (default)
#   - "4": Debug
# * DLOG_TARGETS is a ":" separated list of:
#   - "alert": Display a GUI alert (default)
#   - "log": AppleScript's log command
#   - "syslog": macOS's "logger" command
#   - <absolutePosixFilePath>: Append to the given file, variable references in the path
#   are allowed, and as a courtesy the path may start with "~" to refer to your home dir

property onUnsupportedBundle : "Alert" -- "Alert" | "Ignore"
property onCollectionExists : "Merge" -- "Ask" | "Replace" | "Merge" | "Skip" | "Stop"
property onConfigExists : "Ask" -- "Ask" | "Replace" | "Keep both" | "Skip" | "Stop"
property onGranuleExists : "Replace" -- "Ask" | "Replace" | "Skip" | "Stop"

global LERR
global LWARN
global LINFO
global LDEBUG
global DLOG_LEVEL
global DLOG_TARGETS
global tempDir
global lastExecutedScript
global supportsProgress

on init()
	global LERR
	global LWARN
	global LINFO
	global LDEBUG
	global DLOG_LEVEL
	global DLOG_TARGETS

	set LERR to 1
	set LWARN to 2
	set LINFO to 3
	set LDEBUG to 4
	set DLOG_LEVEL to (getEnvVar("DLOG_LEVEL", LINFO as string)) as integer
	set DLOG_TARGETS to splitString(getEnvVar("DLOG_TARGETS", "alert"), ":")

	set onUnsupportedBundle to getDefaults of "onUnsupportedBundle" given default:onUnsupportedBundle
	set onCollectionExists to getDefaults of "onCollectionExists" given default:onCollectionExists
	set onConfigExists to getDefaults of "onConfigExists" given default:onConfigExists
	set onGranuleExists to getDefaults of "onGranuleExists" given default:onGranuleExists
end init

on run
	try
		init()
		repeat
			set prefs to {¬
				"When a dragged-on item is unsupported:" & tab & tab & tab & " -> " & onUnsupportedBundle,¬
				"When a collection being imported already exists:" & tab & " -> " & onCollectionExists,¬
				"When there's a conflict between a collection's config and the one being imported:" & tab & " -> " & onConfigExists,¬
				"When a granule being imported already exists:" & tab & " -> " & onGranuleExists¬
			}
			set title to "SEAScope Import Preferences"
			choose from list prefs¬
				with title title¬
				with prompt "Pick the preference you wish to change:"¬
				OK button name "Change"¬
				cancel button name "Done"¬
				without multiple selections allowed and empty selection allowed

			set choice to getChoiceFromChooseResult(the result)
			if choice is false then
				exit repeat
			end if

			repeat with i from 1 to length of prefs
				if item i of prefs is choice then
					set prompt to questionFromPerfDescription(choice)
					if i is 1 then
						set optionsList to {"Alert", "Ignore"}
						set defaultOption to onUnsupportedBundle
					else if i is 2 then
						set optionsList to {"Ask", "Replace", "Merge", "Skip", "Stop"}
						set defaultOption to onCollectionExists
					else if i is 3 then
						set optionsList to {"Ask", "Replace", "Keep both", "Skip", "Stop"}
						set defaultOption to onConfigExists
					else if i is 4 then
						set optionsList to {"Ask", "Replace", "Skip", "Stop"}
						set defaultOption to onGranuleExists
					end if

					choose from list optionsList¬
						with title title¬
						with prompt prompt¬
						default items {defaultOption}¬
						OK button name "Change"¬
						without multiple selections allowed and empty selection allowed

					set option to getChoiceFromChooseResult(the result)
					if option is false then
						exit repeat
					end if

					if i is 1 then
						set onUnsupportedBundle to option
						setDefaults of "onUnsupportedBundle" to option
					else if i is 2 then
						set onCollectionExists to option
						setDefaults of "onCollectionExists" to option
					else if i is 3 then
						set onConfigExists to option
						setDefaults of "onConfigExists" to option
					else if i is 4 then
						set onGranuleExists to option
						setDefaults of "onGranuleExists" to option
					end if

					exit repeat
				end if
			end repeat
		end repeat
	on error errMsg number errorNumber
		die from ("An unknown error occurred: " & errorNumber as text)¬
			about missing value¬
			given message:errMsg
	end try
end run

on getChoiceFromChooseResult(r)
	if r is false then
		return false
	else
		return item 1 of r
	end if
end getChoiceFromChooseResult

on questionFromPerfDescription(description)
	set UPPERCASE to "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	set LOWERCASE to "abcdefghijklmnopqrstuvwxyz"

	set firstChar to first character of description
	considering case
		set firstCharOffset to offset of firstChar in UPPERCASE
	end considering
	if firstCharOffset > 0 then
		set firstChar to (character firstCharOffset of LOWERCASE)
	end if

	set description to firstChar & (text 2 thru ((offset of ":" in description) - 1) of description)

	return "What should SEAScope do " & description & "?"
end questionFromPerfDescription

to getDefaults of key given default:default
	try
		return execute({"defaults", "read", id of me, key})
	on error number 1
		return default
	end try
end getDefaults

to setDefaults of key to value
	execute({"defaults", "write", id of me, key, value})
end setDefaults

to getEnvVar(name, default)
	# Set `value` to the value of the env var named `name` or "" if not defined
	set value to system attribute name

	if default is "" or value is not "" or name is in system attribute then
		return value
	else
		return default
	end if
end getEnvVar

-- This droplet processes files dropped onto the applet
on open bundles
	try
		init()

		set dataPath to system attribute "SEASCOPE_DATA_PATH"
		if dataPath is "" then
			die from "SEASCOPE_DATA_PATH is empty"¬
				about "Cannot find SEAScope's data folder"¬
				given message:¬
					"This droplet should not be used directly." & return &¬
					"Drop your files and folders on SEAScope's icon instead."
		end if
		if isFolder(dataPath) is false then
			die from "SEASCOPE_DATA_PATH doesn't point to a directory"¬
				about "Oops! Something is wrong 😳"¬
				given message:¬
					"The given SEASCOPE_DATA_PATH is not a folder!" & return &¬
					"Looking at path: " & dataPath
		end if

		set tempBaseDir to POSIX path of (path to temporary items as string)
		if my id is missing value then
			set ident to my name
		else
			set ident to my id
		end if
		set tempBaseDir to tempBaseDir & ident & "_" & my version & "_"
		set tempDir to tempBaseDir & getRandomString(6)
		repeat while pathExists(tempDir)
			set tempDir to tempBaseDir & getRandomString(6)
		end repeat

		dlog(LDEBUG, "Temp dir: " & tempDir)

		try
			progress total steps
			set supportsProgress to true
		on error
			set supportsProgress to false
		end try

		dlog(LDEBUG, "Progress supported: " & supportsProgress)

		set changed to false
		set bundlesCount to length of bundles

		if supportsProgress then
			set progress total steps to bundlesCount * 1000
			set progress completed steps to 0
			set progress description to "Processing bundles..."
			delay 0.01 -- small delay to update the UI with the new progress
		end if

		dlog(LDEBUG, "Processing " & bundlesCount & " bundles")

		repeat with i from 1 to bundlesCount
			set bundleDescription to "bundle " & i & " of " & bundlesCount
			repeat 1 times
				dlog(LDEBUG, "Processing bundle " & i)
				if supportsProgress then
					set progress additional description to "Processing " & bundleDescription
					delay 0.01 -- small delay to update the UI with the new progress
				end if

				set fromArchive to false
				set bundle to item i of bundles
				if isFile(bundle) is true then
					try
						dlog(LDEBUG, "Trying to extract " & bundleDescription)
						if supportsProgress then
							set progress additional description to "Trying to extract " & bundleDescription
							delay 0.01 -- small delay to update the UI with the new progress description
						end if
						set fromArchive to true
						set bundle to extractArchive(bundle)
					on error number 1
						if onUnsupportedBundle is "Alert" then
							set the bundle_info to info for bundle
							set bundle_name to the name of the bundle_info
							warn of "Unsupported bundle type"¬
								given message:"The file “" & bundle_name & "” cannot be imported."
						end if
						exit repeat
					end try
				end if

				if isFolder(bundle) is true then
					set changed to importCollections(dataPath, bundle, not fromArchive, bundleDescription) or changed
					if fromArchive then
						dlog(LDEBUG, "Removing extraction dir")
						try
							nbRemovePath(bundle)
						end try
					end if
				end if

				if supportsProgress then
					set progress completed steps to i * 1000
					delay 0.01 -- small delay to update the UI with the new progress
				end if
			end repeat
		end repeat

		try
			dlog(LDEBUG, "Removing temp dir")
			removePath(tempDir)
		end try

		if changed is true then
			set bundlesMessage to bundlesCount & " bundle"
			if bundlesCount > 1 then
				set bundlesMessage to bundlesMessage & "s"
			end if
			set headerMessage to "Successfully imported " & bundlesMessage & " into SEAScope's workspace."
			set seascopeApp to system attribute "SEASCOPE_APP_PATH"
			if pathExists(seascopeApp & "/Contents/MacOS/SEAScope") and application seascopeApp is running then
				informWithAction of "Import successful" given¬
					message:¬
						headerMessage & return & return &¬
						"To see the imported data, you need to restart SEAScope",¬
					action:"Restart SEAScope"
				if the result is true then
					try
						quit application seascopeApp
					end try
					delay 1
					try
						launch application seascopeApp
					end try
				end if
			else
				inform of "Import successful" given message:headerMessage
			end if
		else
			inform of "Nothing was imported" given¬
				message:"You probably already had them."
		end if
	on error errMsg number errorNumber
		try
			dlog(LDEBUG, "Removing temp dir")
			removePath(tempDir)
		end try

		if errorNumber is -128 then
			-- "User cancelled." error caused by a "Stop" action
			dlog(LDEBUG, "User cancelled (can be caused by a \"Stop\" action).")
		else if errorNumber is 1 then
			die from errMsg about missing value given message:lastExecutedScript
		else
			die from ("An unknown error occurred: " & errorNumber as text)¬
				about missing value¬
				given message:errMsg
		end if
	end try
	quit
end open

on importCollections(dataPath, dir, keepOriginals, bundleDescription)
	local changed
	set changed to false

	dlog(LDEBUG, "Importing collections from " & bundleDescription)
	if supportsProgress then
		set progress additional description to "Importing collections from " & bundleDescription
		delay 0.01 -- small delay to update the UI with the new progress
	end if

	set dir to removeTrailingSlashOfPOSIXPath(POSIX path of dir)
	set configs to find(dir, {"-maxdepth", 4, "-type", "f", "-name", "*.ini"}, {sorted: true})
	set configsCount to length of configs

	dlog(LDEBUG, "Found " & configsCount & " collections")
	if configsCount is 0 then return false

	if supportsProgress then
		set progressStepsOffset to progress completed steps
	end if
	set collectionSteps to 1000 / configsCount
	repeat with i from 1 to configsCount
		dlog(LDEBUG, "Importing collection " & i)
		if supportsProgress then
			set collectionDescription to "collection " & i & " of " & configsCount & " from " & bundleDescription
			set progress additional description to "Importing " & collectionDescription
			delay 0.01 -- small delay to update the UI with the new progress
		end if

		set changed to importCollection(dataPath, item i of configs, keepOriginals, collectionSteps) or changed

		if supportsProgress then
			set progress completed steps to progressStepsOffset + i * collectionSteps
			delay 0.01 -- small delay to update the UI with the new progress
		end if
	end repeat

	return changed
end importCollections

on importCollection(dataPath, collectionConfigPath, keepOriginals, totalSteps)
	local changed

	set collectionPath to (text items 1 thru -5 of collectionConfigPath) as string
	set isCollection to isFolder(collectionPath)

	if isCollection is false and getBaseName(collectionConfigPath) is "config.ini" then
		dlog(LDEBUG, "Bundled collection's dir is missing. Checking for old file structure (config.ini in collection dir)...")
		set collectionPath to getParentDirPath(collectionConfigPath)
		set isCollection to isFolder(collectionPath)
	end if

	if isCollection is false then
		# Skip ini files not having a corresponding collection
		dlog(LDEBUG, "Ignore bundled collection (granules dir is missing)")
		return false
	end if

	set collectionName to getBaseName(collectionPath)
	set dataCollectionPath to dataPath & "/" & collectionName

	if POSIX file collectionPath is POSIX file dataCollectionPath then
		dlog(LDEBUG, "Skip bundled collection (same path as existing one)")
		return false
	end if

	if isFolder(dataCollectionPath) is true then
		dlog(LDEBUG, "Bundled collection already exists")
		if onCollectionExists is "Ask" then
			display dialog¬
				"A collection named “" & collectionName & "” already exists in your workspace. " &¬
				"Do you want to replace it with the one being imported?"¬
				with title "Collection Exists"¬
				buttons {"Stop", "Replace", "Merge"}¬
				default button "Merge"¬
				cancel button "Stop"
			set collectionAction to the button returned of the result
		else
			set collectionAction to onCollectionExists
		end if
	else
		set collectionAction to "Replace"
	end if

	if collectionAction is "Stop" then
		error "User cancelled" number -128
	else if collectionAction is "Skip" then
		dlog(LDEBUG, "User skipped")
		return false
	end if

	# By default, no change to the target collection
	set changed to false

	# The configuration must be imported prior to the collection granules.
	# Otherwise, when the collection is being replaced:
	#   1. If the existing collection contains the "config.ini", its config will
	#      be lost.
	#   2. The new config file needs to be imported independently from the
	#      collection's granules.

	# Import collection configuration
	# *************************************************************************
	# In case a config file exists in both the new collection and the target,
	# create a backup of the target version (ie. if it differs from the version)
	# and use the new version instead.

	set dataConfigPath to dataCollectionPath & ".ini"
	set oldDataConfigPath to dataCollectionPath & "/config.ini"
	set compareDataConfigPath to dataConfigPath
	if isFile(dataConfigPath) is false and isFile(oldDataConfigPath) is true then
		set compareDataConfigPath to oldDataConfigPath
	end if

	if isFile(compareDataConfigPath) is false then
		set configAction to "Replace"
	else if areTheSame(collectionConfigPath, compareDataConfigPath) is true then
		dlog(LDEBUG, "Bundled collection's config matches existing one")
		set configAction to "Skip"
	else
		dlog(LDEBUG, "Bundled collection's config conflicts with existing one")
		if onConfigExists is "Ask" then
			display dialog¬
				"A collection named “" & collectionName & "” in your workspace already has a config that differs from the one being imported. " &¬
				"Do you want to replace it?"¬
				with title "Collection Config Conflict"¬
				buttons {"Stop", "Skip", "Replace"}¬
				default button "Skip"¬
				cancel button "Stop"
			set configAction to the button returned of the result
		else
			set configAction to onConfigExists
		end if
	end if

	if configAction is "Stop" then
		error "User cancelled" number -128
	else if configAction is "Skip" then
		dlog(LDEBUG, "Skipping collection's config")
	else if configAction is "Replace" then
		dlog(LDEBUG, "Using bundled config")
		if compareDataConfigPath is oldDataConfigPath then
			removePath(oldDataConfigPath)
		end if
		movePath(collectionConfigPath, dataConfigPath, keepOriginals)
		set changed to true
	else if configAction is "Keep both" then
		if isFile(compareDataConfigPath) is true then
			dlog(LDEBUG, "Backing-up existing config")
			if compareDataConfigPath is oldDataConfigPath then
				movePath(oldDataConfigPath, dataConfigPath, false)
			end if
			backupFile(dataConfigPath)
		end if
		dlog(LDEBUG, "Using bundled config")
		movePath(collectionConfigPath, dataConfigPath, keepOriginals)
		set changed to true
	end if

	# Import collection granules
	# *************************************************************************
	# If a granule directory exists in both the new collection and the target,
	# then the new version will be discarded.

	set findGranulesOptions to {"-mindepth", 1, "-maxdepth", 1}
	set findDirGranulesOptions to {"-type", "d"}
	set findFileGranulesOptions to {"-type", "f", "-name", "*.nc"}

	if collectionAction is "Replace" then
		if configAction is "Skip" and compareDataConfigPath is oldDataConfigPath then
			# Move out the existant in-collection config to prevent it's deletion
			# when replacing the collection dir.
			dlog(LDEBUG, "Migrating config.ini from existing old file structure")
			movePath(oldDataConfigPath, dataConfigPath, false)
		end if

		# Remove existing the collection dir
		dlog(LDEBUG, "Removing existing granules")
		nbRemovePath(dataCollectionPath)
		createFolder(dataCollectionPath)
	else if onGranuleExists is "Ask" then
		set findGranulesOptions to findGranulesOptions &¬
			{"(", "("} & findDirGranulesOptions & {")", "-o", "("} & findFileGranulesOptions & {")", ")"}
		set granules to find(collectionPath, findGranulesOptions, {})
		set granulesCount to length of granules
		if supportsProgress then
			set progressStepsOffset to progress completed steps
		end if

		dlog(LDEBUG, "Importing " & granulesCount & " bundled granules")
		repeat with a from 1 to granulesCount
			set granulePath to item a of granules
			set granuleBaseName to getBaseName(granulePath)
			set granuleIsFolder to isFolder(granulePath)
			if granuleIsFolder then
				set granuleName to granuleBaseName
			else
				# Remove .nc from file name
				set granuleName to items 1 thru -4 of granuleBaseName
			end if

			set dataGranulePath to dataCollectionPath & "/" & granuleBaseName
			set dataGranuleFolderPath to dataCollectionPath & "/" & granuleName
			set dataGranuleFilePath to dataCollectionPath & "/" & granuleName & ".nc"

			if pathExists(dataGranuleFolderPath) is false and pathExists(dataGranuleFilePath) is false then
				# New granule
				set granuleAction to "Replace"
			else if granuleIsFolder is false and isFile(dataGranulePath) is true and areTheSame(granulePath, dataGranulePath) is true then
				# Same granule
				set granuleAction to "Skip"
			else
				# Conflict with existing granule
				display dialog¬
					"A granule named “" & granuleName & "” in a collection named “" & collectionName & "” already exists in your workspace. " &¬
					"Do you want to replace it with the one being imported?"¬
					with title "Granule Exists"¬
					buttons {"Stop", "Replace", "Skip"}¬
					default button "Skip"¬
					cancel button "Stop"
				set granuleAction to the button returned of the result
			end if

			if granuleAction is "Stop" then
				error "User cancelled" number -128
			else if granuleAction is "Replace" then
				removePath(dataGranuleFilePath)
				nbRemovePath(dataGranuleFolderPath)
				movePath(granulePath, dataGranulePath, keepOriginals)
				set changed to true
			end if

			if supportsProgress then
				set progress completed steps to progressStepsOffset + totalSteps * a / granulesCount
				delay 0.01 -- small delay to update the UI with the new progress
			end if
		end repeat

		return changed
	else if onGranuleExists is "Replace" then
		# Delete already existing dir granules that are in the bundle
		set findDirGranulesOptions to findDirGranulesOptions & {¬
			"-exec", "/bin/sh", "-c",¬
			"rm -rf \"$0/$(basename \"$1\")\" \"$0/$(basename \"$1\").nc\"",¬
			dataCollectionPath, "{}", ";"¬
		}
		set findFileGranulesOptions to findFileGranulesOptions & {¬
			"-exec", "/bin/sh", "-c",¬
			"rm -rf \"$0/$(basename \"$1\")\" \"$0/$(basename \"$1\" .nc)\"",¬
			dataCollectionPath, "{}", ";"¬
		}
	else if onGranuleExists is "Skip" then
		# Skip already existing granules from the bundle
		set findDirGranulesOptions to findDirGranulesOptions & {¬
			"-exec", "/bin/sh", "-c",¬
			"! [[ -d \"$0/$(basename \"$1\")\" || -e \"$0/$(basename \"$1\").nc\" ]]",¬
			dataCollectionPath, "{}", ";"¬
		}
		set findFileGranulesOptions to findFileGranulesOptions & {¬
			"-exec", "/bin/sh", "-c",¬
			"! [[ -d \"$0/$(basename \"$1\" .nc)\" || -e \"$0/$(basename \"$1\")\" ]]",¬
			dataCollectionPath, "{}", ";"¬
		}
	else if onGranuleExists is "Stop" then
		# Existing granule with same name is a dir or doesn't match the bundled one
		set hasConflictDirGranules to canFind(collectionPath, findDirGranulesOptions & {¬
			"-exec", "/bin/sh", "-c",¬
			"[[ -e \"$0/$(basename \"$1\")\" || -e \"$0/$(basename \"$1\").nc\" ]]",¬
			dataCollectionPath, "{}", ";"¬
		})
		if hasConflictDirGranules is true then
			dlog(LDEBUG, "Not importing granules because some of them already exists")
			error "User cancelled" number -128
		end if

		set isDifferentFileGranule to "(cmp --silent \"$0/$(basename \"$1\")\" \"$1\"; [ $? -eq 1 ])"
		set hasConflictFileGranules to canFind(collectionPath, findFileGranulesOptions & {¬
			"-exec", "/bin/sh", "-c",¬
			"[[ -d \"$0/$(basename \"$1\" .nc)\" ]] || " & isDifferentFileGranule,¬
			dataCollectionPath, "{}", ";"¬
		})
		if hasConflictFileGranules is true then
			dlog(LDEBUG, "Not importing granules because some of them are different from the already existing ones")
			error "User cancelled" number -128
		end if
	end if

	set findGranulesOptions to findGranulesOptions &¬
		{"(", "("} & findDirGranulesOptions & {")", "-o", "("} & findFileGranulesOptions & {")", ")"}
	if keepOriginals is true then
		set findGranulesOptions to findGranulesOptions &¬
			{"-exec", "/bin/sh", "-c", "cp -a \"$@\" \"$0/\"", dataCollectionPath, "{}", "+"}
	else
		set findGranulesOptions to findGranulesOptions &¬
			{"-exec", "/bin/sh", "-c", "mv \"$@\" \"$0/\"", dataCollectionPath, "{}", "+"}
	end if

	dlog(LDEBUG, "Importing bundled granules")
	return find(collectionPath, findGranulesOptions, {split: false}) is not ""
end importCollection

on joinList(theList, theDelimiters)
	set {orgTids, AppleScript's text item delimiters} to {AppleScript's text item delimiters, theDelimiters}
	set theString to theList as string
	set AppleScript's text item delimiters to orgTids
	return theString
end joinList

on splitString(theString, theDelimiters)
	set {orgTids, AppleScript's text item delimiters} to {AppleScript's text item delimiters, theDelimiters}
	set theList to every text item of theString
	set AppleScript's text item delimiters to orgTids
	return theList
end splitString

on pathExists(aPath)
	try
		POSIX file aPath as alias
		return true
	on error number -1700
		return false
	end try
end pathExists

on isFile(f)
	try
		execute({"test", "-f", f})
		return true
	on error number 1
		return false
	end try
end isFile

on isFolder(f)
	try
		execute({"test", "-d", f})
		return true
	on error number 1
		return false
	end try
end isFolder

on areTheSame(file1, file2)
	try
		nbExecute({"cmp", "--silent", file1, file2})
		return true
	on error number 1
		return false
	end try
end areTheSame

on movePath(source, destination, keepOriginals)
	if isFolder(destination) then
		nbRemovePath(destination)
	end if
	if keepOriginals is true then
		nbExecute({"cp", "-af", source, destination})
	else
		nbExecute({"mv", source, destination})
	end if
end movePath

on removePath(aPath)
	execute({"rm", "-rf", aPath})
end removePath

on nbRemovePath(aPath)
	nbExecute({"rm", "-rf", aPath})
end nbRemovePath

on createFolder(aPath)
	execute({"mkdir", "-p", aPath})
end createFolder

on getRandomString(length)
	set rand to ""
	repeat length times
		set rand to rand & some item of "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
	end repeat
	return rand
end getRandomString

on extractArchive(archive)
	set extractionDir to tempDir & "/extract_" & getRandomString(6)
	removePath(extractionDir)
	createFolder(extractionDir)
	nbExecute({"tar", "xf", archive, "-C", extractionDir})
	return extractionDir
end extractArchive

on findCommand(path, cmdOptions, options)
	set {sorted: sorted} to options & {sorted: false}
	set command to joinCommand({"find", path} & cmdOptions) & " -print0"
	if sorted is true then set command to command & " | sort -z"
	return command
end findCommand

on find(path, cmdOptions, options)
	set {split: split} to options & {split: true}
	set results to nbExecute(findCommand(path, cmdOptions, options))
	if split is true then
		set results to splitString(results, character id 0)
		if length of results > 1 then set results to items 1 thru -2 of result
	end if
	return results
end find

on canFind(path, cmdOptions)
	set command to (¬
		findCommand(path, cmdOptions, {sorted: false}) &¬
		" -print | head -c1 | grep -E '.'"¬
	)
	try
		nbExecute(command)
		return true
	on error number 1
		return false
	end try
end canFind

on getCurrentDateStamp()
	set {year:y, month:mth, day:d, hours:h, minutes:m, seconds:s} to current date
	return "" & y & "" & pad2(mth) & "" & pad2(d) & "" & pad2(h) & "" & pad2(m) & "" & pad2(s)
end getCurrentDateStamp

on pad2(num)
	return text -2 thru -1 of ((100 + num) as string)
end pad2

on backupFile(filePath)
	if class of filePath is alias or class of filePath is file then
		set filePath to POSIX path of filePath
	end if

	set dateStamp to getCurrentDateStamp()
	set pathParts to splitString(filePath, "/")
	set dirPath to joinList(items 1 thru -2 of pathParts, "/")
	set baseName to item -1 of pathParts
	set baseNameParts to splitString(baseName, ".")

	if length of baseNameParts > 1 then
		set baseNameWithoutExtension to joinList(items 1 thru -2 of baseNameParts, ".")
		set extension to "." & item -1 of baseNameParts
	else
		set baseNameWithoutExtension to baseName
		set extension to ""
	end if

	set counter to 1
	set backupPath to dirPath & "/" & baseNameWithoutExtension & "-" & dateStamp & extension
	repeat while pathExists(backupPath)
		set counter to counter + 1
		set backupPath to dirPath & "/" & baseNameWithoutExtension & "-" & dateStamp & "-" & counter & extension
	end repeat

	movePath(filePath, backupPath, false)
end backupFile

on readFile(filePath)
	try
		return read filePath
	on error number -39
		return ""
	end try
end readFile

on removeTrailingSlashOfPOSIXPath(path)
	if last item of path is "/" then
		return text 1 thru -2 of path
	end if
	return path
end removeTrailingSlashOfPOSIXPath

on getParentDirPath(aPath)
	if class of aPath is alias or class of aPath is file then
		set aPath to POSIX path of aPath
	end if
	return joinList(items 1 thru -2 of splitString(aPath, "/"), "/")
end getParentDirPath

on getBaseName(aPath)
	if class of aPath is alias or class of aPath is file then
		set aPath to POSIX path of aPath
	end if
	return item -1 of splitString(aPath, "/")
end getBaseName

to quoteCommandArg(arg)
	if class of arg is alias or class of arg is file or class of arg is «class bmrk» then
		set arg to POSIX path of arg
	end if
	return quoted form of (arg as string)
end quoteCommandArg

to joinCommand(command)
	set theScript to "LC_CTYPE=C LANG=C"
	repeat with arg in command
		set theScript to theScript & " " & quoteCommandArg(arg)
	end repeat
	return theScript
end joinCommand

on execute(command)
	if class of command is list then set command to joinCommand(command)
	set lastExecutedScript to command
	dlog(LDEBUG, " $ " & command)
	try
		set output to do shell script command
		dlog(LDEBUG, "  -> exited with status: 0")
		if output is not "" then dlog(LDEBUG, "  => " & output)
		return output
	on error eStr number eNum partial result rList from badObj to expectedType
		dlog(LDEBUG, "  -> exited with status: " & eNum)
		error eStr number eNum partial result rList from badObj to expectedType
	end try
end execute

on nbExecute(command)
	if class of command is list then set command to joinCommand(command)
	set execDir to tempDir & "/exec_" & getRandomString(6)
	removePath(execDir)
	createFolder(execDir)

	set stdoutPath to execDir & "/stdout.txt"
	set stderrPath to execDir & "/stderr.txt"
	set exitCodePath to execDir & "/exitCode.txt"
	set lastExecutedScript to "ASYNC: " & command

	dlog(LDEBUG, " $ " & command & " &")
	set backgroundCommand to (¬
		"(" & command &¬
		" > " & quoteCommandArg(stdoutPath) &¬
		" 2> " & quoteCommandArg(stderrPath) &¬
		"; echo $? > " & quoteCommandArg(exitCodePath) &¬
		") &> /dev/null & echo $!"¬
	)
	try
		set pid to do shell script backgroundCommand
		dlog(LDEBUG, "  -> [" & pid & "]")
	on error eStr number eNum partial result rList from badObj to expectedType
		dlog(LDEBUG, "  -> exited with status: " & eNum)
		error eStr number eNum partial result rList from badObj to expectedType
	end try

	repeat until pathExists(exitCodePath)
		try
			delay 0.1
		on error eStr number -128
			try
				execute({"kill", "-9", pid})
			end try
			error eStr number -128
		end try
	end repeat

	set lastExecutedScript to command
	set exitCode to readFile(exitCodePath) as integer
	dlog(LDEBUG, " &-> [" & pid & "] exited with status: " & exitCode)
	if exitCode is 0 then
		set stdout to readFile(stdoutPath)
		if stdout is not "" then dlog(LDEBUG, " &=> " & stdout)
		return stdout
	else
		set stderr to readFile(stderrPath)
		if stderr is not "" then dlog(LDEBUG, " &=> " & stderr)
		error "The command exited with a non-zero status." number exitCode
	end if
end nbExecute

to inform of title given message:msg
	if DLOG_LEVEL < LINFO then return

	dlogWithoutAlert(LINFO, "[" & title & "] " & msg)
	if "alert" is in DLOG_TARGETS then
		display alert title¬
			message msg¬
			buttons {"Dissmis"}¬
			default button "Dissmis"
	end if
end inform

to informWithAction of title given message:msg, action:action
	if DLOG_LEVEL < LINFO then return

	dlogWithoutAlert(LINFO, "[" & title & "] " & msg & " [" & action & "]")
	if "alert" is in DLOG_TARGETS then
		display alert title¬
			message msg¬
			buttons {"Dissmis", action}¬
			default button action
		return the button returned of the result is action
	end if
end inform

to warn of title given message:msg
	if DLOG_LEVEL < LWARN then return

	dlogWithoutAlert(LWARN, "[" & title & "] " & msg)
	if "alert" is in DLOG_TARGETS then
		display alert title¬
			message msg¬
			buttons {"Stop", "Ignore"}¬
			default button 2¬
			cancel button "Stop"¬
			as warning
	end if
end warn

to die from theError about title given message:msg
	dlogWithoutAlert(LERR, "[" & theError & "] " & msg)
	if DLOG_LEVEL >= LERR and "alert" is in DLOG_TARGETS then
		if title is missing value then copy theError to title
		display alert title message msg as critical
	end if
	quit

	delay 0.01 -- small delay for the quit command to take effect
end die

# Logs a text representation of the specified object or objects, which may be of any type, typically for debugging.
# Works hard to find a meaningful text representation of each object.
# SYNOPSIS
#   dlog(anyObjOrListOfObjects)
# USE EXAMPLES
#   dlog("before")  # single object
#     dlog({ "front window: ", front window }) # list of objects
# SETUP
#   At the top of your script, define global variable DLOG_TARGETS and set it to a *list* of targets (even if you only have 1 target).
#     set DLOG_TARGETS to {} # must be a list with any combination of: "log", "syslog", "alert", <posixFilePath>
#   An *empty* list means that logging should be *disabled*.
#   If you specify a POSIX file path, the file will be *appended* to; variable references in the path
#   are allowed, and as a courtesy the path may start with "~" to refer to your home dir.
#   Caveat: while you can *remove* the variable definition to disable logging, you'll take an additional performance hit.
# SETUP EXAMPLES
#    For instance, to use both AppleScript's log command *and* display a GUI alert, use:
#       set DLOG_TARGETS to { "log", "alert" }
# Note:
#   - Since the subroutine is still called even when DLOG_TARGETS is an empty list,
#     you pay a performance penalty for leaving dlog() calls in your code.
#   - Unlike with the built-in log() method, you MUST use parentheses around the parameter.
#   - To specify more than one object, pass a *list*. Note that while you could try to synthesize a single
#     output string by concatenation yourself, you'd lose the benefit of this subroutine's ability to derive
#     readable text representations even of objects that can't simply be converted with `as text`.
# Credits:
#   Based on @mklement0's answer on StackOverflow
#   https://stackoverflow.com/a/21341372/2168921
to dlog(level, anyObjOrListOfObjects)
	global DLOG_LEVEL
	global DLOG_TARGETS

	try
		if length of DLOG_TARGETS is 0 then return
		if level > DLOG_LEVEL then return
	on error
		return
	end try

	# The following tries hard to derive a readable representation from the input object(s).
	if class of anyObjOrListOfObjects is not list then set anyObjOrListOfObjects to {anyObjOrListOfObjects}
	local lst, i, txt, errMsg, orgTids, oName, oId, prefix, logTarget, txtCombined, prefixTime, prefixDateTime
	set lst to {item level of {"ERROR", "WARN ", "INFO ", "DEBUG"}}
	repeat with anyObj in anyObjOrListOfObjects
		set txt to ""
		repeat with i from 1 to 2
			try
				if i is 1 then
					if class of anyObj is list then
						set {orgTids, AppleScript's text item delimiters} to {AppleScript's text item delimiters, {", "}}
						set txt to ("{" & anyObj as string) & "}"
						set AppleScript's text item delimiters to orgTids
					else
						set txt to anyObj as string
					end if
				else
					set txt to properties of anyObj as string
				end if
			on error errMsg
				# Trick for records and record-*like* objects:
				# We exploit the fact that the error message contains the desired string representation of the record,
				# so we extract it from there. This (still) works as of AS 2.3 (OS X 10.9).
				try
					set txt to do shell script "egrep -o '\\{.*\\}' <<< " & quoted form of errMsg
				end try
			end try
			if txt is not "" then exit repeat
		end repeat
		set prefix to ""
		if class of anyObj is not in {text, integer, real, boolean, date, list, record} and anyObj is not missing value then
			set prefix to "[" & class of anyObj
			set oName to ""
			set oId to ""
			try
				set oName to name of anyObj
				if oName is not missing value then set prefix to prefix & " name=\"" & oName & "\""
			end try
			try
				set oId to id of anyObj
				if oId is not missing value then set prefix to prefix & " id=" & oId
			end try
			set prefix to prefix & "] "
			set txt to prefix & txt
		end if
		set lst to lst & txt
	end repeat
	set {orgTids, AppleScript's text item delimiters} to {AppleScript's text item delimiters, {" "}}
	set txtCombined to lst as string
	set prefixTime to "[" & time string of (current date) & "] "
	set prefixDateTime to "[" & short date string of (current date) & " " & text 2 thru -1 of prefixTime
	set AppleScript's text item delimiters to {character id 0}
	set txtCombined to every text item of txtCombined
	set AppleScript's text item delimiters to {"\\0"}
	set txtCombined to txtCombined as string
	set AppleScript's text item delimiters to orgTids

	# Log the result to every target specified.
	repeat with logTarget in DLOG_TARGETS
		if contents of logTarget is "log" then
			log prefixTime & txtCombined
		else if contents of logTarget is "alert" then
			display alert prefixTime & txtCombined
		else if contents of logTarget is "syslog" then
			do shell script "logger -t " & quoted form of ("AS: " & (name of me)) & " " & quoted form of txtCombined
		else # assumed to be a POSIX file path to *append* to.
			set fpath to contents of logTarget
			if fpath starts with "~/" then set fpath to "$HOME/" & text 3 thru -1 of fpath
			do shell script "printf '%s\\n' " & quoted form of (prefixDateTime & txtCombined) & " >> \"" & fpath & "\""
		end if
	end repeat
end dlog

to dlogWithoutAlert(level, anyObjOrListOfObjects)
	global DLOG_LEVEL
	global DLOG_TARGETS

	try
		if length of DLOG_TARGETS is 0 then return
		if level > DLOG_LEVEL then return
	on error
		return
	end try

	if "alert" is in DLOG_TARGETS then
		# Save DLOG_TARGETS
		set targets to DLOG_TARGETS

		# Remove "alert" from DLOG_TARGETS
		set DLOG_TARGETS to {}
		repeat with target in targets
			if contents of target is not "alert" then
				copy contents of target to the end of DLOG_TARGETS
			end if
		end repeat

		# Call dlog()
		dlog(level, anyObjOrListOfObjects)

		# Restore DLOG_TARGETS
		set DLOG_TARGETS to targets
	else
		dlog(level, anyObjOrListOfObjects)
	end if
end dlogWithoutAlert
