// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cctype>
#include <cstdlib>
#include <algorithm>
#include <sstream>

#include "vendors/_mini.h"
#include "vendors/_fmt.h"

#include "core/log.h"
#include "core/path.h"
#include "local/collection_ini_parser.h"
#include "core/rendering_cfg.h"
#include "core/structs.h"
#include "local/reader_cfg.h"
#include "debug.h"

using namespace SEAScope;

static RenderMethod renderMethods(const std::string& name, bool& ok)
{
    if
        ("RASTER" == name)
    {
        ok = true;
        return RenderMethod::RASTER;
    }
    else if
        ("ARROWS" == name)
    {
        ok = true;
        return RenderMethod::ARROWS;
    }
    else if
        ("BARBS" == name)
    {
        ok = true;
        return RenderMethod::BARBS;
    }
    else if
        ("TRAJECTORIES" == name)
    {
	    ok = true;
        return RenderMethod::TRAJECTORIES;
    }
    else if
        ("DOTSCLOUD" == name)
    {
        ok = true;
        return  RenderMethod::DOTSCLOUD;
    }
    else if
        ("STREAMLINES" == name)
    {
        ok = true;
        return RenderMethod::STREAMLINES;
    }
    else if
        ("RAWRGB" == name)
    {
        ok = true;
        return RenderMethod::RAWRGB;
    }

    ok = false;
    return RenderMethod::RASTER;
};


static FilterMode filterModes(const std::string& name, bool& ok)
{
    if
        ("NEAREST" == name)
    {
        ok = true;
        return FilterMode::NEAREST;
    }
    else if
        ("BILINEAR" == name)
    {
        ok = true;
        return FilterMode::BILINEAR;
    }

    ok = false;
    return FilterMode::NEAREST;
};

/**
 *
 */
CollectionINIParser::CollectionINIParser()
: _defaultVariableFields({
    {"color", "0,0,0"},
    {"colormap", ""},
    {"tags", ""},
    {"filterMode", "NEAREST"},
    {"logscale", "false"},
    {"min", "0.0"},
    {"max", "1.0"},
    {"opacity", "1.0"},
    {"zindex", "0.5"},
    {"particlesCount", "1000"},
    {"streamlineLength", "40"},
    {"streamlineSpeed", "1.0"},
    {"billboardsSize", "32"},
    {"billboardsDensity", "0.0"},
    {"lineThickness", "1.0"},
    {"currentTimeTrajectoryMarker", ""}
})
{

};

/**
 *
 */
CollectionINIParser::~CollectionINIParser()
{
};

/**
 *
 */
bool CollectionINIParser::extractTags(const std::string& input, std::map<std::string, std::string>& result)
{
	std::vector<std::string> keyvals;	//
	std::vector<std::string> tagValue;	//

	splitString(input, ',', keyvals);
	for
		(const auto& keyval: keyvals)
	{
		tagValue.clear();
		splitString(keyval, ':', tagValue);
		if
			(2 == tagValue.size())
		{
			result[tagValue[0]] = tagValue[1];
		}
		else
		{
			LOG( LogLevel::Error, "CollectionINIParser"
					, "Tag does not match the expected KEY:VALUE format: {}"
					, keyval);
			return false;
		}
	}

	return true;
};

/**
 *
 */
std::string CollectionINIParser::trim(const std::string& input)
{
	auto  wsfront=std::find_if_not(input.begin(),input.end(),[](int c){return std::isspace(c);});
	return std::string(wsfront,std::find_if_not(input.rbegin(),std::string::const_reverse_iterator(wsfront),[](int c){return std::isspace(c);}).base());
};

/**
 *
 */
std::vector<std::string>& CollectionINIParser::splitString(const std::string& input, char delim, std::vector<std::string>& result)
{
	std::stringstream ss(input);	//
	std::string item;	//
	while
		(std::getline(ss, item, delim))
	{
		item = trim(item);
		if
			(false == item.empty())
		{
			result.push_back(item);
		}
	}

	return result;
};

/**
 *
 */
bool CollectionINIParser::validate(const std::string& uri, mINI::INIStructure& root) const
{
    bool canBeFixed(false); //
    bool updateNeeded(false);   //
	std::vector<std::string> missing;	//
	std::vector<std::string> variables;	//
	const std::array<std::string, 6> requiredGeneral = { {"label",
													    "xSeamless",
													    "ySeamless",
													    "mustBeCurrent",
													    "tags",
													    "variables"} };
	const std::array<std::string, 4> requiredVar = { {"label",
												    "units",
												    "fields",
												    "defaultRendering"} };

    // Incomplete configuration files can only be fixed if they are writable
    canBeFixed = Path::writeFile(uri);

    mINI::INIFile parser(uri);
    bool ok(false);
    ok = parser.read(root);
    if
        (false == ok)
    {
		LOG( LogLevel::Error, "CollectionINIParser"
				, "Error during parsing of {}"
				, uri);
        return false;
    }

    if
        (true == root.has("reader"))
    {
        auto& reader = root["reader"];
        if
            (false == reader.has("type"))
        {
			missing.push_back("Field: [reader] type");
        }

        if
            (reader.has("relativeTimeCoverageStart") != reader.has("relativeTimeCoverageEnd"))
        {
            LOG(LogLevel::Error, "CollectionINIParser",
                     "The [reader] section fields named 'relativeTimeCoverageStart' and 'relativeTimeCoverageEnd' must be defined together");
            if
                (true == reader.has("relativeTimeCoverageStart"))
            {
                missing.push_back("Field: [reader] relativeTimeCoverageEnd");
            }
            else
            {
                missing.push_back("Field: [reader] relatimeTimeCoverageStart");
            }
        }

        if
            (reader.has("depthVariable") != reader.has("depthValue"))
        {
            if
                (true == reader.has("depthVariable"))
            {
                missing.push_back("Field: [reader] depthValue");
            }
            else
            {
                missing.push_back("Field: [reader] depthVariable");
            }
        }
    }

	if
		(false == root.has("general"))
	{
		LOG(LogLevel::Error, "CollectionINIParser",
                 "[general] section missing in {}", uri);
		return false;
	}

	// Check fields from [general] section
	auto& general = root["general"];
	for
		(const auto& field : requiredGeneral)
	{
		if
			(false == root["general"].has(field))
		{
			missing.push_back(fmt::format("Field: [general] {}", field));
		}
	}
	if
		(false == general.has("variables"))
	{
		// Further validation impossible without the variables list
		return false;
	}

    if
        (true == general.has("NEWSAligned"))
    {
        if
            (true == canBeFixed)
        {
            LOG(LogLevel::Info, "CollectionINIParser",
                     "Removing deprecated NEWSAligned setting from {}", uri);
            general.remove("NEWSAligned");
            updateNeeded = true;
        }
    }

    if
        (true == general.has("defaultVariable"))
    {
        if
            (true == canBeFixed)
        {
            LOG(LogLevel::Info, "CollectionINIParser",
                "Removing deprecated defaultVariable setting from {}", uri);
            general.remove("defaultVariable");
            updateNeeded = true;
        }
    }

	splitString(general["variables"], ',', variables);

	// Check duplicates in variables list
	bool createNewList = false;
	auto it = variables.begin();
    while(it != variables.end())
    {
        const ptrdiff_t num = std::count(variables.begin(), variables.end(), *it);

        if
            (1u < num)
        {
			// Duplicate found, print a warning
            LOG(LogLevel::Warn, "CollectionINIParser"
                , "Variables list, duplicate found : {} in {} file"
                , *it
                , uri);

			// Remove the current duplicate
			it = variables.erase(it);

			// Variables list need to be modified in the config file,
			// generate a new string with the purged list
			createNewList = true;
        }
		else
		{
			++it;
		}
    }

	if
		(true == createNewList)
	{
		// Duplicates found, create a new string cleaned
		std::stringstream ss;
		for
			(size_t i = 0; i < variables.size(); ++i)
		{
			ss << variables[i];

			if
				(i != variables.size() - 1)
			{
				ss << ", ";
			}
		}

		// Save the new variables list
		general["variables"] = ss.str();

        // File will be modified
        updateNeeded = true;
	}

	for
		(const auto& varname : variables)
	{
		if
			(false == root.has(varname))
		{
			missing.push_back(fmt::format("Section: [{}]", varname));

			// Go on with the next variable
			continue;
		}

		auto& varSection = root[varname];
		for
			(const auto& field : requiredVar)
		{
			if
				(false == varSection.has(field))
			{
				missing.push_back(fmt::format("Field: [{}] {}", varname, field));
			}
		}

        for
            (const auto& defaultIterator : this->_defaultVariableFields)
        {
            if
                (true == varSection.has(defaultIterator.first))
            {
                continue;   // everything is ok
            }

            if
                (true == canBeFixed)
            {
                LOG(LogLevel::Info, "CollectionINIParser",
                         "Adding missing [{}] {} field to {}", varname,
                         defaultIterator.first, uri);
                varSection[defaultIterator.first] = defaultIterator.second;
                updateNeeded = true;
            }
            else
            {
				missing.push_back(fmt::format("Field: [{}] {}", varname,
                                  defaultIterator.first));
            }
        }
	}

	// Check fields for all variables
	if
		(0 < missing.size())
	{
		LOG(LogLevel::Error, "CollectionINIParser",
				 "Configuration file {} is incomplete, the following elements have not been found:", uri);
		for
			(const auto& msg : missing)
		{
			LOG(LogLevel::Error, "CollectionINIParser", msg);
		}
		return false;
	}

    if
        (true == updateNeeded)
    {
        ok = parser.write(root, true /*pretty print*/);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "CollectionINIParser",
                     "Could not write fixed configuration to {}", uri);
            return false;
        }
    }

	return true;
}

/**
 *
 */
bool CollectionINIParser::load(uint32_t colId, const std::string& uri, CollectionCfg& colCfg, std::vector<RenderingCfg*>& rCfgs, ReaderCfg& readerCfg) const
{
	std::vector<std::string> colorComponents;	//
	std::vector<std::string> variables;	//
	std::vector<std::string> fields;	//
	std::vector<std::string> tag;	//
	RenderingCfg* rCfg;	//
	std::uint32_t varsCount;	//
	std::uint8_t r;	//
	std::uint8_t g;	//
	std::uint8_t b;	//
	mINI::INIStructure root;	//
	bool ok;	//

	LOG( LogLevel::Debug, "CollectionINIParser"
			, "Parsing file {}"
			, uri);

	ok = this->validate(uri, root);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "CollectionINIParser",
                 "Validation of the {} configuration file failed", uri);
		return false;
	}

	// Delete objects before vector clear
	for (RenderingCfg* rCfgTmp : rCfgs)
	{
		delete rCfgTmp;
	}

	rCfgs.clear();

    if
        (true == root.has("reader"))
    {
        auto& reader = root["reader"];
        readerCfg.setType(reader["type"]);
        if
            (true == reader.has("timeVariable"))
        {
            readerCfg.setTimeVariable(reader["timeVariable"]);
        }
        if
            (true == reader.has("latVariable"))
        {
            readerCfg.setLatVariable(reader["latVariable"]);
        }
        if
            (true == reader.has("lonVariable"))
        {
            readerCfg.setLonVariable(reader["lonVariable"]);
        }
        if
            (true == reader.has("outputGeolocMajorAxisSpacing"))
        {
            readerCfg.setGCPsMajorAxisSpacing(static_cast<std::uint16_t>(std::atoi(reader["outputGeolocMajorAxisSpacing"].c_str())));
        }
        if
            (true == reader.has("outputGeolocMinorAxisSpacing"))
        {
            readerCfg.setGCPsMinorAxisSpacing(static_cast<std::uint16_t>(std::atoi(reader["outputGeolocMinorAxisSpacing"].c_str())));
        }
        if
            (true == (reader.has("relativeTimeCoverageStart") && reader.has("relativeTimeCoverageEnd")))
        {
            readerCfg.setRelativeTimeCoverage(static_cast<std::int64_t>(std::atol(reader["relativeTimeCoverageStart"].c_str())),
                                              static_cast<std::int64_t>(std::atol(reader["relativeTimeCoverageEnd"].c_str())));
        }
        if
            (true == (reader.has("depthVariable") && reader.has("depthValue")))
        {
            readerCfg.setDepth(reader["depthVariable"], static_cast<float>(std::atof(reader["depthValue"].c_str())));
        }
    }

	auto& general = root["general"];

	// Extract lists
	splitString(general["variables"], ',', variables);
	LOG( LogLevel::Debug, "CollectionINIParser"
			, "Variables: {}"
			, general["variables"]);

	colCfg.label = general["label"];
	colCfg.xSeamless = "true" == general["xSeamless"];
	colCfg.ySeamless = "true" == general["ySeamless"];
	colCfg.mustBeCurrent = "true" == general["mustBeCurrent"];
	extractTags(general["tags"], colCfg.tags);

	LOG( LogLevel::Debug, "CollectionINIParser"
			, "Number of variables: {}"
			, variables.size());

	varsCount = 0;
	for
		(const auto& varName : variables)
	{
		VariableCfg vCfg;
		auto& vInfo = root[varName];
		vCfg.label  = vInfo["label"];
		vCfg.units = vInfo["units"];

		fields.clear();
		splitString(vInfo["fields"], ',', fields);
		vCfg.fields.insert(fields.begin(), fields.end());

		/* Parse color parameter */
		colorComponents.clear();
		splitString(vInfo["color"], ',', colorComponents);
		if
			(3 != colorComponents.size())
		{
			LOG( LogLevel::Error, "CollectionINIParser"
					, "Invalid format for the \"color\" parameter: {} (expected: color = r,g,b) for variable '{}: {}'"
					, vInfo["color"], colCfg.label, varName);
			return false;
		}
		r = static_cast<std::uint8_t>(std::atoi(colorComponents[0].c_str()));
		g = static_cast<std::uint8_t>(std::atoi(colorComponents[1].c_str()));
		b = static_cast<std::uint8_t>(std::atoi(colorComponents[2].c_str()));

        RenderMethod rMethod = renderMethods(vInfo["defaultRendering"], ok);
        if
            (false == ok)
		{
			LOG( LogLevel::Error, "CollectionINIParser"
					, "Rendering method \"{}\" is not supported for variable '{}: {}'"
					, vInfo["defaultRendering"], colCfg.label, varName);
			return false;
		}

		extractTags(vInfo["tags"], vCfg.tags);

		colCfg.variables[varsCount] = vCfg;

		rCfg = new RenderingCfg();
		rCfg->_renderMethod = rMethod;
		rCfg->_logScale = "true" == vInfo["logscale"];
		rCfg->_min = static_cast<float>(std::atof(vInfo["min"].c_str()));
		rCfg->_max = static_cast<float>(std::atof(vInfo["max"].c_str()));
		rCfg->_opacity = static_cast<float>(std::atof(vInfo["opacity"].c_str()));
		rCfg->_zindex = static_cast<float>(std::atof(vInfo["zindex"].c_str()));
        std::uint32_t particlesCount {0u};  //
        particlesCount = static_cast<std::uint32_t>(std::atoi(vInfo["particlesCount"].c_str()));
        if
            (particlesCount < RenderingCfg::MIN_VALID_PARTICLES_COUNT)
        {
            particlesCount = RenderingCfg::MIN_VALID_PARTICLES_COUNT;
            LOG(LogLevel::Warn, "CollectionINIParser", "particlesCount value was too low for variable '{}: {}', it has been set to {}", colCfg.label, varName, particlesCount);
        }
		rCfg->_particlesCount = particlesCount;
		rCfg->_particleTTL = 0; // No longer used
        std::uint32_t streamlineLength {0u};    //
        streamlineLength = static_cast<std::uint32_t>(std::atoi(vInfo["streamlineLength"].c_str()));
        if
            (streamlineLength < RenderingCfg::MIN_VALID_STREAMLINE_LENGTH)
        {
            streamlineLength = RenderingCfg::MIN_VALID_STREAMLINE_LENGTH;
            LOG(LogLevel::Warn, "CollectionINIParser", "streamlineLength value was too low for variable '{}: {}', it has been set to {}", colCfg.label, varName, streamlineLength);
        }
		rCfg->_streamlineLength = streamlineLength;
        float streamlineSpeed {0.f};    //
        streamlineSpeed = static_cast<float>(std::atof(vInfo["streamlineSpeed"].c_str()));
        if
            (streamlineSpeed < RenderingCfg::MIN_VALID_STREAMLINE_SPEED)
        {
            streamlineSpeed = RenderingCfg::MIN_VALID_STREAMLINE_SPEED;
            LOG(LogLevel::Warn, "CollectionINIParser", "streamlineSpeed value was too low for variable '{}: {}', it has been set to {}", colCfg.label, varName, streamlineSpeed);
        }
		rCfg->_streamlineSpeed = streamlineSpeed;
        float billboardsSize {0.f}; //
        billboardsSize = static_cast<float>(std::atof(vInfo["billboardsSize"].c_str()));
        if
            (billboardsSize < RenderingCfg::MIN_VALID_BILLBOARDS_SIZE)
        {
            billboardsSize = RenderingCfg::MIN_VALID_BILLBOARDS_SIZE;
            LOG(LogLevel::Warn, "CollectionINIParser", "billboardsSize value was too low for variable '{}: {}', it has been set to {}", colCfg.label, varName, billboardsSize);
        }
		rCfg->_billboardsSize = billboardsSize;
        float billboardsDensity {-1.f}; //
        billboardsDensity = static_cast<float>(std::atof(vInfo["billboardsDensity"].c_str()));
        if
            (billboardsDensity < RenderingCfg::MIN_VALID_BILLBOARDS_DENSITY)
        {
            billboardsDensity = RenderingCfg::MIN_VALID_BILLBOARDS_DENSITY;
            LOG(LogLevel::Warn, "CollectionINIParser", "billboardsDensity value was too low for variable '{}: {}', it has been set to {}", colCfg.label, varName, billboardsDensity);
        }
        if
            (billboardsDensity > RenderingCfg::MAX_VALID_BILLBOARDS_DENSITY)
        {
            billboardsDensity = RenderingCfg::MAX_VALID_BILLBOARDS_DENSITY;
            LOG(LogLevel::Warn, "CollectionINIParser", "billboardsDensity value was too high for variable '{}: {}', it has been set to {}", colCfg.label, varName, billboardsDensity);
        }
		rCfg->_billboardsDensity = billboardsDensity;
        rCfg->_currentTimeTrajectoryIconName = vInfo["currentTimeTrajectoryMarker"];

		rCfg->_colormap = vInfo["colormap"];
		rCfg->_color = {{ r, g, b }};
		rCfg->target.granuleLevel = false;
		rCfg->target.collection = std::make_pair(0u, colId);
		rCfg->target.variable = varsCount;

        FilterMode fMode = filterModes(vInfo["filterMode"], ok);
        if
            (false == ok)
		{
			LOG( LogLevel::Error, "CollectionINIParser"
					, "Filter mode \"{}\" is not supported for variable '{}: {}'"
					, vInfo["filterMode"], colCfg.label, varName);
			return false;
		}
		rCfg->_filterMode = fMode;

        float lineThickness {0.f};  //
        lineThickness = static_cast<float>(std::atof(vInfo["lineThickness"].c_str()));
        if
            (lineThickness < RenderingCfg::MIN_VALID_LINE_THICKNESS)
        {
            lineThickness = RenderingCfg::MIN_VALID_LINE_THICKNESS;
            LOG(LogLevel::Warn, "CollectionINIParser", "lineThickness value was too low for variable '{}: {}', it has been set to {}", colCfg.label, varName, lineThickness);
        }
        rCfg->_lineThickness = lineThickness;

		rCfgs.push_back(rCfg);
		++varsCount;
	}

	Debug::displayCollection(colCfg);
	return true;
};


/**
 *
 */
bool CollectionINIParser::loadReaderCfg(const std::string& uri, ReaderCfg& readerCfg) const
{
	mINI::INIStructure root;	//
	bool ok;	//

	LOG( LogLevel::Debug, "CollectionINIParser"
			, "Parsing file {}"
			, uri);

	ok = this->validate(uri, root);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "CollectionINIParser",
                 "Validation of the {} configuration file failed", uri);
		return false;
	}

    if
        (true == root.has("reader"))
    {
        auto& reader = root["reader"];
        readerCfg.setType(reader["type"]);
        if
            (true == reader.has("timeVariable"))
        {
            readerCfg.setTimeVariable(reader["timeVariable"]);
        }
        if
            (true == reader.has("latVariable"))
        {
            readerCfg.setLatVariable(reader["latVariable"]);
        }
        if
            (true == reader.has("lonVariable"))
        {
            readerCfg.setLonVariable(reader["lonVariable"]);
        }
        if
            (true == reader.has("outputGeolocMajorAxisSpacing"))
        {
            readerCfg.setGCPsMajorAxisSpacing(static_cast<std::uint16_t>(std::atoi(reader["outputGeolocMajorAxisSpacing"].c_str())));
        }
        if
            (true == reader.has("outputGeolocMinorAxisSpacing"))
        {
            readerCfg.setGCPsMinorAxisSpacing(static_cast<std::uint16_t>(std::atoi(reader["outputGeolocMinorAxisSpacing"].c_str())));
        }
        if
            (true == (reader.has("relativeTimeCoverageStart") && reader.has("relativeTimeCoverageEnd")))
        {
            readerCfg.setRelativeTimeCoverage(static_cast<std::int64_t>(std::atol(reader["relativeTimeCoverageStart"].c_str())),
                                              static_cast<std::int64_t>(std::atol(reader["relativeTimeCoverageEnd"].c_str())));
        }
        if
            (true == (reader.has("depthVariable") && reader.has("depthValue")))
        {
            readerCfg.setDepth(reader["depthVariable"], static_cast<float>(std::atof(reader["depthValue"].c_str())));
        }
    }

	return true;
};
