// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <iostream>
#include <algorithm>
#include "local/collections_explorer.h"
#include "local/idf_reader.h"
#include "core/path.h"
#include "core/log.h"
#include "debug.h"
#include "local/reader_cfg.h"
#include "local/idf_reader.h"
#include "local/simple_netcdf_reader.h"
#include "local/collections_catalogue.h"
#include "core/interfaces/igranules_index.h"
#include "core/granule_metadata.h"

#include "core/dialog.h"

using namespace SEAScope;

/**
 *
 */
CollectionsExplorer::CollectionsExplorer()
{
    this->_readers["idf"] = new IDFReader();
    this->_readers["latlon"] = new SimpleNetCDFReader();
}

/**
 *
 */
CollectionsExplorer::~CollectionsExplorer()
{
    for
        (auto& reader : this->_readers)
    {
        delete reader.second;
        reader.second = nullptr;
    }
}


/**
 *
 */
bool CollectionsExplorer::initCatalogue(const IGranulesIndex& index, CollectionsCatalogue& catalogue)
{
    bool ok {false};    //
	std::vector<std::pair<std::uint32_t, std::string> > collections;	//

	ok = index.listCollections(collections);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "CollectionsExplorer",
				 "Could not list collections from the index");
		return false;
	}

    CollectionINIParser parser;
	for
		(const auto& collection : collections)
	{
        ReaderCfg readerCfg {}; //
        CollectionCfg colCfg;	//
        std::vector<RenderingCfg*> rCfgs;	//
        std::stringstream buffer;	//

        buffer << collection.second << ".ini";
        ok = parser.load(collection.first, buffer.str(), colCfg, rCfgs, readerCfg);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "CollectionsExplorer",
                     "Could not parse collection configuration found in {}",
                     buffer.str());
            return false;
        }

        ok = catalogue.addCollection(collection.first, std::move(colCfg), std::move(rCfgs), buffer.str());
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "CollectionsExplorer",
                     "Failed to register collection {} in the catalogue",
                     collection.second);
            return false;
        }
    }

    return true;
}

/**
 *
 */
bool CollectionsExplorer::initIndex(const std::string& dataPath,
                                    const std::string& indexPath,
                                    IGranulesIndex& index,
                                    bool detectOutdatedIndex)
{
    bool ok {false};    //
	std::uint32_t colId {1u};    //
	std::vector<std::string> collections;	//

	ok = Path::dirExists(dataPath);
	if
		(false == ok)
	{
		return false;
	}

	ok = this->listCollections(dataPath, collections);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "CollectionsExplorer",
				 "Could not find any collection. Skipping index creation");
		return false;
	}

    bool indexAlreadyExisted {false};
    indexAlreadyExisted = Path::fileExists(indexPath);

    std::tuple<bool, bool> loadResult {false, false}; //
    bool loadSuccess {false};   //
    bool indexCreated {false};    //

    loadResult = index.load(indexPath);
    loadSuccess = std::get<0>(loadResult);
    indexCreated = !std::get<1>(loadResult);
    if
        (false == loadSuccess)
    {
        LOG(LogLevel::Error, "CollectionsExplorer", "Failed to initialize granules index");
        return false;
    }

	bool useExistingIndex {true};	//
    std::vector<std::string> outdatedCollections;    //
    std::vector<std::string> ghostCollections;  //
    if
        (false == indexAlreadyExisted)
    {
        // The index has just been created, so all collections found in the
        // data directory need to be indexed
        outdatedCollections = collections;
    }
    else if
        (true == detectOutdatedIndex)
    {
        std::vector<bool> modificationsDetected;
        modificationsDetected.reserve(collections.size());

        // Check if collection folders in the data directory have a
        // modification time newer than the index
		ok = this->isIndexUpdateNeeded(indexPath, collections, modificationsDetected);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "CollectionsExplorer",
                     "Could not detect if the collections have been modified since last indexation");
        }

        // Get list of already indexed collections to perform further tests
        std::vector<std::pair<std::uint32_t, std::string> > indexedCollections;
        ok = index.listCollections(indexedCollections);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "CollectionsExplorer", "Failed to list indexed collections while checking if the index needs an update");
            return false;
        }

        // Check if there are collections in the data directory that have a
        // modification time older than the index and yet are missing from the
        // index
        auto itCollection = collections.begin();
        auto itModified = modificationsDetected.begin();
        for
            (;itCollection != collections.end(); ++itCollection, ++itModified)
        {
            if
                (true == *itModified)
            {
                // The collection is already flagged as modified, so the test
                // will not bring new information
                continue;
            }

            // The collections is not flagged as modified, check if it is
            // already in the index
            bool indexed {false};
            for
                (const auto& indexedCollection : indexedCollections)
            {
                if
                    (indexedCollection.second == *itCollection)
                {
                    indexed = true;
                    break;
                }
            }

            if
                (false == indexed)
            {
                // The collections has a modification time older than the index
                // and yet is not indexed: flag it as modified so that it will
                // be indexed later on
                *itModified = true;
            }
        }

        // Check if there are collections in the index that are no longer in
        // the data directory
        for
            (const auto& indexedCollection : indexedCollections)
        {
            auto found = std::find(collections.cbegin(), collections.cend(), indexedCollection.second);
            if
                (collections.end() == found)
            {
                ghostCollections.push_back(indexedCollection.second);
            }
        }

        itCollection = collections.begin();
        itModified = modificationsDetected.begin();
        for
            (;itCollection != collections.end(); ++itCollection, ++itModified)
        {
            if
                (true == *itModified)
            {
                outdatedCollections.push_back(*itCollection);
            }
        }

		if
			(0 < outdatedCollections.size() + ghostCollections.size())
		{
            const char* question = "It appears some of the files in your data directory have changed since the creation of the index file, which could therefore be outdated. Do you want to rebuild the index for collections that have changed?";
            useExistingIndex = !SEAScope::Dialog::askYesNoQuestion(question, "Rebuild index", "Use old index");
		}
	}
    else
    {
        // The user explicitely requested that no check is done to detect
        // outdated collections
    }

    if
        ((false == indexCreated) && (true == useExistingIndex))
    {
        // Index existed beforehand and either there is nothing to update or
        // the user chose to ignore these updates
        return true;
    }

    // Start updating the index
    std::cout << "*******************************************************************************" << std::endl;
    std::cout << "SEAScope will now build (or update) an index of the data granules" << std::endl;
    std::cout << "This operation may take several minutes as each granule file must be opened to" << std::endl;
    std::cout << " extract information." << std::endl;
    std::cout << "The resulting index will be saved so this step will not be necessary anymore" << std::endl;
    std::cout << "(until you add/remove/modify granules) and SEAScope will start up much faster." << std::endl;
    std::cout << "*******************************************************************************" << std::endl;

    ok = index.startBulkOperation();
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "CollectionsExplorer",
                 "Failed to start a bulk operation for inserting/updating entries in the granules index");
        return false;
    }

    // Remove outdated collections from index
    for
        (const auto& collectionPath : outdatedCollections)
    {
        ok = index.removeCollection(collectionPath);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "CollectionsExplorer", "Failed to remove collection '{}' from index", collectionPath);
            return false;
        }
    }

    // Remove ghost collections from index
    for
        (const auto& collectionPath : ghostCollections)
    {
        ok = index.removeCollection(collectionPath);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "CollectionsExplorer", "Failed to remove collection '{}' from index", collectionPath);
            return false;
        }
    }

    // Outdated collections have been wiped from the index, we have to assign
    // new identifiers for them
    CollectionINIParser parser;
	for
		(const auto& collectionPath : outdatedCollections)
	{
        ReaderCfg readerCfg {}; //
        std::stringstream buffer;	//

        buffer << collectionPath << ".ini";
        ok = parser.loadReaderCfg(buffer.str(), readerCfg);

        if
            (false == ok)
        {
            LOG( LogLevel::Error, "CollectionsExplorer"
                    , "Could not parse collection configuration found in {}"
                    , dataPath);

            ok = index.cancelBulkOperation();
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "CollectionsExplorer",
                         "Failed to cancel the bulk operation for inserting/updating entries in the granules index");
            }
            return false;
        }

        colId = index.generateCollectionId();

        ok = index.addCollection(collectionPath, colId);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "CollectionsExplorer",
                     "Could not register collection '{}' in index", collectionPath);
            return false;
        }

        this->findGranules(collectionPath, colId, readerCfg, index);
	}

    ok = index.stopBulkOperation();
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "CollectionsExplorer",
                 "Failed to complete the bulk operation for inserting/updating entries in the granules index");
        return false;
    }

    if
        (  false == Path::fileExists(indexPath)
		&& true == Path::writeFile(indexPath))
    {
        ok = index.save(indexPath);
		if
			(false == ok)
		{
			LOG(LogLevel::Error, "CollectionsExplorer", "Failed to save index");
			return false;
		}
	    LOG(LogLevel::Debug, "CollectionsExplorer", "Index saved");
    }

	return true;
}

/**
 *
 */
bool CollectionsExplorer::migrateLegacyCollections(const std::string& dataPathIn)
{
	std::forward_list<std::string> files;	//
	std::ostringstream legacyBuffer;	//
	std::ostringstream buffer;	//
	std::string basename;	//
	std::size_t slashPos {0u};	//
	bool isCollectionDir {false};	//
	bool isDir {false};	//
	bool ok {false};	//
    int rc {-1};    //

	ok = Path::listFiles(dataPathIn, files);
	if
		(false == ok)
	{
		return false;
	}

	for
		(const auto& subpath : files)
	{
		slashPos = subpath.find_last_of(SEAScope::Path::separator());
		basename = subpath.substr(slashPos+1);
		isDir = Path::dirExists(subpath);
		if
			(  0 < subpath.size()
			&& true == isDir
			&& 0 != basename.compare(0, 1, "."))
		{
			legacyBuffer.str("");
			legacyBuffer << subpath << SEAScope::Path::separator() << "config.ini";
			isCollectionDir = Path::fileExists(legacyBuffer.str());
			if
				(true == isCollectionDir)
			{
                //
                buffer.str("");
                buffer << subpath << ".ini";
                if
                    (Path::fileExists(buffer.str()))
                {
                    LOG(LogLevel::Warn, "CollectionsExplorer",
                             "Collection '{}' has both a new and a legacy configuration files. Please remove the legacy file ({}).",
                             basename, legacyBuffer.str());
                }
                else
                {
                    rc = rename(legacyBuffer.str().c_str(), buffer.str().c_str());
                    if
                        (0 == rc)
                    {
                        LOG(LogLevel::Info, "CollectionsExplorer",
                                 "Legacy collection config file {} renamed to {}",
                                 legacyBuffer.str(), buffer.str());
                    }
                    else
                    {
                        LOG(LogLevel::Warn, "CollectionsExplorer",
                                 "Could not rename legacy config file for collection '{}' from '{}' to '{}'. \nInsufficient permissions are the most likely cause, please make sure that you have the rights to rename files in '{}' or ask your system administrator to rename the config file manually.\nUntil then SEAScope will ignore this collection.",
                                 basename, legacyBuffer.str(), buffer.str(), dataPathIn);
                    }
                }
			}
		}
	}
	return true;
}

/**
 *
 */
bool CollectionsExplorer::listCollections(const std::string& dataPathIn,
										  std::vector<std::string>& result)
{
    const std::string iniExtension {".ini"};
    const std::size_t extensionSize {iniExtension.size()};
	std::forward_list<std::string> files;	//
	std::string basename {""};	//
    std::string collectionPath {""};    //
    std::ostringstream buffer;  //
	std::size_t slashPos {0u};	//
    bool ok {false};    //

    // Migrate old config files
    ok = CollectionsExplorer::migrateLegacyCollections(dataPathIn);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "CollectionsExplorer",
                 "Failed to migrate legacy config files for at least one collection, please check the logs");
        return false;
    }

	ok = Path::listFiles(dataPathIn, files);
	if
		(false == ok)
	{
		return false;
	}

	for
		(const auto& subpath : files)
	{
		slashPos = subpath.find_last_of(SEAScope::Path::separator());
		basename = subpath.substr(slashPos + 1);
        if
            (  basename.size() > extensionSize
            && 0 == basename.compare(basename.size() - extensionSize, extensionSize, iniExtension))
        {
            buffer.str("");
            buffer << subpath.substr(0, slashPos) << SEAScope::Path::separator() << basename.substr(0, basename.size() - extensionSize);
            collectionPath = buffer.str();
            if
                (Path::dirExists(collectionPath))
            {
                result.push_back(collectionPath);
            }
        }
        else
        {
            continue;
        }
    }

    return true;
}

/**
 *
 */
bool CollectionsExplorer::isIndexUpdateNeeded(const std::string& indexPathIn,
											  const std::vector<std::string>& collections,
                                              std::vector<bool>& updateNeeded)
{
	bool ok;	//
	std::time_t indexTimestamp;	//
	std::time_t collectionTimestamp;	//

	ok = Path::lastModified(indexPathIn, indexTimestamp);
	if
		(false == ok)
	{
		LOG(LogLevel::Error, "CollectionsExplorer",
                 "Could not get last modification time for {}", indexPathIn);
		return false;
	}

	for
		(const auto& collectionPath : collections)
	{
		ok = Path::lastModified(collectionPath, collectionTimestamp);
		if
			(false == ok)
		{
			LOG(LogLevel::Error, "CollectionsExplorer",
					 "Could not get last modification time for {}",
					 collectionPath);
			return false;
		}

        // It should detect when a granule has been added or removed, maybe not
        // if the granule has been modified
        updateNeeded.push_back(collectionTimestamp > indexTimestamp);
	}

	return true;
}


/**
 *
 */
bool CollectionsExplorer::findGranules(const std::string& path, std::uint32_t colId, const ReaderCfg& readerCfg, IGranulesIndex& index)
{
    std::cout << " -> Indexing granules in " << path << std::endl;
	LOG( LogLevel::Debug, "CollectionsExplorer"
			, "findGranules({})"
			, colId);
	std::forward_list<std::string> files;	//
	bool ok {false};	//

	ok = Path::listFilesRecursive(path, files);
	if
		(false == ok)
	{
		return false;
	}

    bool res = true;

	for
		(const auto& filePath : files)
	{
		if
			(  3 < filePath.size()
			&& ".nc" == filePath.substr(filePath.size() - 3))
		{
			res = this->indexGranule(filePath, colId, readerCfg, index);
            
            if
                (true == res)
            {
                index.increaseBulkOpsCount();
            }
		}
	}

	return true;
};

/**
 *
 */
bool CollectionsExplorer::indexGranule(const std::string& filePath, std::uint32_t colId, const ReaderCfg& readerCfg, IGranulesIndex& index)
{
    std::vector<GranuleMetadata> granules;  //
	bool ok;	//

    auto readerFound = this->_readers.find(readerCfg.getType());
    if
        (this->_readers.end() == readerFound)
    {
        LOG(LogLevel::Error, "CollectionsExplorer", "Reader '{}' not found", readerCfg.getType());
        return false;
    }

	ok = readerFound->second->extractGranuleInfo(filePath, granules, readerCfg);
	if
		(false == ok)
	{
		LOG( LogLevel::Error, "CollectionsExplorer"
				, "Could not extract granule information from {}"
				, filePath);
		return false;
	}

    for
        (auto& granule : granules)
    {
	    granule.collectionId(colId);

	    ok = index.add(granule, false);
    }

	return ok;
};
