// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iostream>

#include "local/embedded_datastream.h"
#include "core/log.h"
#include "core/path.h"
#include "local/simple_netcdf_reader.h"
#include "local/idf_reader.h"

using namespace SEAScope;

/**
 *
 */
EmbeddedDatastream::EmbeddedDatastream()
{
    this->_readers["idf"] = new IDFReader();
    this->_readers["latlon"] = new SimpleNetCDFReader();
};

/**
 *
 */
EmbeddedDatastream::~EmbeddedDatastream()
{
    for
        (auto& reader : this->_readers)
    {
        delete reader.second;
        reader.second = nullptr;
    }
};

/**
 *
 */
bool EmbeddedDatastream::init(ILogHandler* logImplementation, IPathHandler* pathImplementation)
{
	Log::implementation = logImplementation;
    Path::implementation = pathImplementation;

#if defined __linux__ || defined __APPLE__
	setenv("TZ", "UTC", 1);
#elif _WIN32
	_putenv_s("TZ", "UTC");
#endif

    return true;
};

/**
 *
 */
bool EmbeddedDatastream::reset()
{
    // The plugin does not own the log and path implementations
    Log::implementation = nullptr;
    Path::implementation = nullptr;

    return true;
}

/**
 *
 */
EmbeddedDatastream* EmbeddedDatastream::create()
{
	return new EmbeddedDatastream();
};

/**
 *
 */
bool EmbeddedDatastream::loadData(const GranuleId& /*granuleId*/,
                                  const std::string& uri,
								  DataModel dataModel,
                                  const std::unordered_map<std::string, DataBucket*>& result) const
{
	bool ok {false};	//

    LOG(LogLevel::Debug, "Embedded datastream", "Load data from {}", uri);

    std::string readerType("idf");
    auto lastCharIter = std::prev(uri.end());
    if
        ('@' == *lastCharIter)
    {
        readerType = "latlon";
    }
    auto readerFound = this->_readers.find(readerType);
    if
        (this->_readers.end() == readerFound)
    {
        LOG(LogLevel::Error, "CollectionsExplorer", "Reader '{}' not found", readerType);
        return false;
    }

	ok = readerFound->second->extractGranuleData(uri, dataModel, result);
	return ok;
};

/**
 *
 */
bool EmbeddedDatastream::canHandle(const std::string& uri) const
{
	return "embedded" == uri;
};
