// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iostream>

#include "local/embedded_index.h"
#include "local/collections_explorer.h"
#include "filter_logger.h"
#include "core/log.h"
#include "platform/path_impl.h"
#include "core/path.h"
#include "core/utils.h"
#include "core/settings.h"
#include "core/granule_metadata.h"
#include "debug.h"

#include "core/dialog.h"

using namespace SEAScope;

/**
 *
 */
EmbeddedIndex::EmbeddedIndex()
{
};

/**
 *
 */
EmbeddedIndex::~EmbeddedIndex()
{
};

/**
 *
 */
bool EmbeddedIndex::init(const Settings& settings, ILogHandler* logImplementation, IPathHandler* pathImplementation, IDialogHandler* dialogHandler)
{
    bool ok(false);

	Log::implementation = logImplementation;
    Path::implementation = pathImplementation;
    Dialog::implementation = dialogHandler;

#if defined __linux__ || defined __APPLE__
	setenv("TZ", "UTC", 1);
#elif _WIN32
	_putenv_s("TZ", "UTC");
#endif

    CollectionsExplorer explorer;
	ok = explorer.initIndex(settings.dataPath, settings.indexPath, this->_index, !settings.skipIndexCheck);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "EmbeddedIndex", "Failed to initialize index");
        return false;
    }

	ok = explorer.initCatalogue(this->_index, this->_catalogue);
    if
        (false == ok)
    {
        LOG(LogLevel::Error, "EmbeddedIndex", "Failed to initialize catalogue");
        return false;
    }

    return true;
};

/**
 *
 */
bool EmbeddedIndex::reset()
{
    // The plugin does not own the log and path implementations
    Path::implementation = nullptr;
    Log::implementation = nullptr;
    Dialog::implementation = nullptr;

    return true;
}

/**
 *
 */
bool EmbeddedIndex::searchGranules( const SearchCriteria& criteria
						   , std::vector<GranuleMetadata>& result) const
{
	bool ok;	//
	ok = this->_index.findGranules(criteria, result);
	return ok;
};

/**
 *
 */
std::array<bool, 2> EmbeddedIndex::searchPrevNext( timestamp_t current
						    , const std::set<CollectionId>& collections
							, timestamp_t& prev
							, timestamp_t& next
							, timestamp_t minDelta) const
{
	std::set<std::uint32_t> localCollections;	//
	GranuleMetadata prevGranule;	//
	GranuleMetadata nextGranule;	//

	LOG( LogLevel::Debug, "EmbeddedIndex"
			, "searchPrevNext: {}"
			, current);

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	LOG( LogLevel::Debug, "EmbeddedIndex"
			, "Local collections list built");

	auto oks = this->_index.adjacentGranules(1000 * current, 1000 * minDelta, localCollections, prevGranule, nextGranule);

	LOG( LogLevel::Debug, "EmbeddedIndex"
			, "Adjacent granules ok");
	LOG( LogLevel::Debug, "EmbeddedIndex"
			, "{} {}"
			, oks[0], oks[1]);
	if
		(true == oks[0])
	{
		prev = static_cast<timestamp_t>(0.0005 * static_cast<double>(prevGranule.timeCoverageStart())  + 0.0005 * static_cast<double>(prevGranule.timeCoverageEnd()));
	}
	if
		(true == oks[1])
	{
		next = static_cast<timestamp_t>(0.0005 * static_cast<double>(nextGranule.timeCoverageStart()) + 0.0005 * static_cast<double>(nextGranule.timeCoverageEnd()));
	}

	LOG( LogLevel::Debug, "EmbeddedIndex"
			, "Returning from searchPrevNext");

    return oks;
};

/**
 *
 */
bool EmbeddedIndex::searchClosestBound( timestamp_t current
						    , const std::set<CollectionId>& collections
							, timestamp_t& closest
                            , const timestamp_t searchLowerBound
                            , const timestamp_t searchUpperBound) const
{
	timestamp_t delta;	//
	timestamp_t closestTmp;	//
	std::set<std::uint32_t> localCollections;	//
	GranuleMetadata closestGranule;	//
	bool ok;	//

	LOG( LogLevel::Debug, "EmbeddedIndex"
			, "searchClosestBoundt: {}"
			, current);

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	LOG( LogLevel::Debug, "EmbeddedIndex"
			, "Local collections list built");

    const timestamp_t currentMs {1000 * current}; //

	ok = this->_index.closestGranuleBound(currentMs, 1000 * searchLowerBound, 1000 * searchUpperBound, localCollections, closestGranule);
	if
		(true == ok)
	{
        timestamp_t closestStart {closestGranule.timeCoverageStart()};
        timestamp_t closestEnd {closestGranule.timeCoverageEnd()};
		//LOG(LogLevel::Debug, "EmbeddedIndex", "closest granule found: {} {} {}", closestGranule.uris[0].uri, closestGranule.start, closestGranule.stop);
		if
			(  closestStart <= currentMs
			&& closestEnd > currentMs)
		{
			/* A granule contains the current datetime, no bounds required */
			closest = current;
			return true;
		}
		closestTmp = current;
		delta = std::numeric_limits<uint64_t>::max();
		if
			(  closestStart < currentMs
			&& delta > currentMs - closestStart)
		{
			delta = currentMs - closestStart;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestStart));
		}
		else if
			(  closestStart > currentMs
			&& delta > closestStart - currentMs)
		{
			delta = closestStart - currentMs;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestStart));
		}

		/* Remove one second when using the end of the granule as a result
		   or the granule will no be displayed */
		if
			(  closestEnd < currentMs
			&& delta > currentMs - closestEnd)
		{
			delta = currentMs - closestEnd;
			closestTmp = static_cast<timestamp_t>(-1.0 + 0.001 * static_cast<double>(closestEnd));
		}
		else if
			(  closestEnd > currentMs
			&& delta > closestEnd - currentMs)
		{
			delta = closestEnd - currentMs;
			closestTmp = static_cast<timestamp_t>(-1.0 + 0.001 * static_cast<double>(closestEnd));
		}

		closest = closestTmp;
	}

	LOG(LogLevel::Debug, "EmbeddedIndex", "closest bound: {}", closest);
    return ok;
};

/**
 *
 */
bool EmbeddedIndex::getDataCoverage(const std::set<CollectionId>& collections, EventsCoverage& result) const
{
	std::set<std::uint32_t> localCollections;	//
	bool ok;	//

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	ok = this->_index.getCoverage(localCollections, result);
	return ok;
};

/**
 *
 */
bool EmbeddedIndex::getDataCoverage(const CollectionId& collectionId, EventsCoverage& result) const
{
	bool ok;	//

	ok = this->_index.getCoverage(collectionId.second, result);
	return ok;
};


/**
 *
 */
bool EmbeddedIndex::listCollections(std::set<std::uint32_t>& results) const
{
	return this->_catalogue.listCollections(results);
};

/**
 *
 */
bool EmbeddedIndex::getCollectionCfg(std::uint32_t id, CollectionCfg& cfg, std::vector<RenderingCfg*>& renderings) const
{
	bool ok;	//
	ok = this->_catalogue.getCollection(id, cfg, renderings);
	LOG( LogLevel::Debug, "EmbeddedCatalogue"
			, "getCollectionCfg: {}"
			, id);
	Debug::displayCollection(cfg);
	return ok;
};

/**
 *
 */
bool EmbeddedIndex::saveVariableRenderingCfg(const SEAScope::RenderingCfg* cfg) const
{
    std::string path {""};
	auto it = this->_catalogue.ids.begin();
	for (; it != this->_catalogue.ids.end(); ++it)
	{
		if (cfg->target.collection.second == it->second)
		{
			path = it->first;
            break;
		}
	}

    if
        (0 >= path.length())
    {
        LOG(LogLevel::Error, "EmbeddedCatalogue",
                 "Could not find configuration file for collection {}",
                 cfg->target.collection.second);
        return false;
    }

	// first, create a file instance
	mINI::INIFile file(path);

	// next, create a structure that will hold data
	mINI::INIStructure ini;

	// now we can read the file
	bool res = file.read(ini);

	if
		(false == res)
	{
        LOG(LogLevel::Error, "EmbeddedCatalogue",
                 "Could not open collection configuration file '{}'", path);
        return false;
    }

    std::vector<std::string> variables;
    SEAScope::explodeString(ini["general"]["variables"], ",", variables);
    if
         (cfg->target.variable >= variables.size())
    {
        LOG(LogLevel::Error, "EmbeddedCatalogue",
                 "Could not locate the variable definition for {}.{}/{}:{} in {}",
                 cfg->target.collection.first, cfg->target.collection.second, cfg->target.granule.second, cfg->target.variable, path);
        return false;
    }

    std::string varKey {""};
    varKey = variables[cfg->target.variable];
    varKey = SEAScope::trim(varKey);

    std::string varLabel {""};
    varLabel = ini[varKey]["label"];

    ini[varKey]["logscale"] = (true == cfg->logscale()) ? "true" : "false";
    ini[varKey]["min"] = std::to_string(cfg->min());
    ini[varKey]["max"] = std::to_string(cfg->max());
    ini[varKey]["opacity"] = std::to_string(cfg->opacity());
    ini[varKey]["zindex"] = std::to_string(cfg->zindex());
    ini[varKey]["particlesCount"] = std::to_string(cfg->particlesCount());
    ini[varKey]["particleTTL"] = std::to_string(cfg->particleTTL());
    ini[varKey]["streamlineLength"] = std::to_string(cfg->streamlineLength());
    ini[varKey]["streamlineSpeed"] = std::to_string(cfg->streamlineSpeed());
    ini[varKey]["colormap"] = cfg->colormap();
    ini[varKey]["billboardsSize"] = std::to_string(cfg->billboardsSize());
    ini[varKey]["billboardsDensity"] = std::to_string(cfg->billboardsDensity());

    std::stringstream ss;
    ss << std::to_string(cfg->color()[0]) << "," << std::to_string(cfg->color()[1]) << "," << std::to_string(cfg->color()[2]);
    ini[varKey]["color"] = ss.str();
    ini[varKey]["filterMode"] = (SEAScope::FilterMode::BILINEAR == cfg->filterMode()) ? "BILINEAR" : "NEAREST";
    ini[varKey]["lineThickness"] = std::to_string(cfg->lineThickness());
    ini[varKey]["currentTimeTrajectoryMarker"] = cfg->currentTimeTrajectoryIconName();

    // write updates to file
    res = file.write(ini);

    if
        (false == res)
    {
        LOG(LogLevel::Error, "EmbeddedCatalogue",
            "Could not write to {} file",
            path);
        return false;
    }

    return true;
}

/**
 *
 */
EmbeddedIndex* EmbeddedIndex::create()
{
    return new EmbeddedIndex();
};

/**
 *
 */
bool EmbeddedIndex::canHandle(const std::string& uri) const
{
	return "embedded" == uri;
};
