// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

#include <numeric>
#include <algorithm>
#include "core/structs.h"
#include "local/gcps_computer.h"


/**
 * Regular lat/lon grids
 */
bool SEAScope::GCPsComputer::getGCPs(std::uint64_t timeValue,
                                     std::vector<double>& inputMajorAxisValues,
                                     std::vector<double>& inputMinorAxisValues,
                                     CoordinatesPlacement inputMajorAxisCoordinatesPlacement,
                                     CoordinatesPlacement inputMinorAxisCoordinatesPlacement,
                                     std::uint16_t outputMajorAxisSpacing,
                                     std::uint16_t outputMinorAxisSpacing,
                                     Shape& output)
{
    // Compute offset depending on the placement of the coordinates within the
    // pixel
    double majorAxisOffset{0.0};
    if
        (SEAScope::CoordinatesPlacement::MIDDLE == inputMajorAxisCoordinatesPlacement)
    {
        majorAxisOffset = -0.5;
    }
    else if
        (SEAScope::CoordinatesPlacement::END == inputMajorAxisCoordinatesPlacement)
    {
        majorAxisOffset = -1.0;
    }

    double minorAxisOffset{0.0};
    if
        (SEAScope::CoordinatesPlacement::MIDDLE == inputMinorAxisCoordinatesPlacement)
    {
        minorAxisOffset = -0.5;
    }
    else if
        (SEAScope::CoordinatesPlacement::END == inputMinorAxisCoordinatesPlacement)
    {
        minorAxisOffset = -1.0;
    }


    std::size_t majorAxisSize {inputMajorAxisValues.size()};
    std::size_t minorAxisSize {inputMinorAxisValues.size()};

    std::vector<double> deltaLat(majorAxisSize - 1);
    for
        (std::size_t i = 0; i < deltaLat.size(); ++i)
    {
        deltaLat[i] = inputMajorAxisValues[i + 1] - inputMajorAxisValues[i];
    }

    std::vector<double> deltaLon(minorAxisSize - 1);
    std::vector<double> absDeltaLon(minorAxisSize - 1);
    for
        (std::size_t i = 0; i < deltaLon.size(); ++i)
    {
        deltaLon[i] = inputMinorAxisValues[i + 1] - inputMinorAxisValues[i];
        absDeltaLon[i] = std::abs(deltaLon[i]);
    }

    // Handle longitudes continuity (grids starting at longitude 0 that cross
    // the IDL but have longitudes expressed in [-180, 180]
    auto maxAbsDeltaLon = std::max_element(absDeltaLon.cbegin(), absDeltaLon.cend());
    if
        (180.0 <= *maxAbsDeltaLon)
    {
        for
            (std::size_t i = 0; i < minorAxisSize; ++i)
        {
            if
                (0 > inputMinorAxisValues[i])
            {
                inputMinorAxisValues[i] += 360.0;
            }
        }

        // Recompute longitude deltas
        for
            (std::size_t i = 0; i < deltaLon.size(); ++i)
        {
            deltaLon[i] = inputMinorAxisValues[i + 1] - inputMinorAxisValues[i];
        }
    }

    // Compute mean distance in degrees along each axis
    double deltaLatTotal = std::accumulate(deltaLat.begin(), deltaLat.end(), 0.0);
    double meanDeltaLat = deltaLatTotal / static_cast<double>(deltaLat.size());

    double deltaLonTotal = std::accumulate(deltaLon.begin(), deltaLon.end(), 0.0);
    double meanDeltaLon = deltaLonTotal / static_cast<double>(deltaLon.size());

    // Compte number of GCPs along each axis
    std::size_t outputMajorAxisTmpSize = static_cast<std::size_t>(std::ceil(majorAxisSize / outputMajorAxisSpacing));
    std::size_t outputMinorAxisTmpSize = static_cast<std::size_t>(std::ceil(minorAxisSize / outputMinorAxisSpacing));

    // Compute indices of GCPs along each axis
    std::vector<std::uint32_t> outputMajorAxisIndices(outputMajorAxisTmpSize);
    std::iota(outputMajorAxisIndices.begin(), outputMajorAxisIndices.end(), 0);
    std::transform(outputMajorAxisIndices.cbegin(), outputMajorAxisIndices.cend(),
                   outputMajorAxisIndices.begin(), [outputMajorAxisSpacing](std::uint32_t value) {
                    return value * outputMajorAxisSpacing;
                   });
    outputMajorAxisIndices.push_back(static_cast<std::uint32_t>(majorAxisSize));
    output.yArity = static_cast<std::uint32_t>(outputMajorAxisIndices.size());

    std::vector<std::uint32_t> outputMinorAxisIndices(outputMinorAxisTmpSize);
    std::iota(outputMinorAxisIndices.begin(), outputMinorAxisIndices.end(), 0);
    std::transform(outputMinorAxisIndices.cbegin(), outputMinorAxisIndices.cend(),
                   outputMinorAxisIndices.begin(), [outputMinorAxisSpacing](std::uint32_t value) {
                    return value * outputMinorAxisSpacing;
                   });
    outputMinorAxisIndices.push_back(static_cast<std::uint32_t>(minorAxisSize));
    output.xArity = static_cast<std::uint32_t>(outputMinorAxisIndices.size());

    constexpr double lon0 = -180.0;
    for
        (std::size_t i = 0; i < outputMajorAxisIndices.size(); ++i)
    {
        for
            (std::size_t j = 0; j < outputMinorAxisIndices.size(); ++j)
        {
            double majorAxisValue {inputMajorAxisValues[0] + majorAxisOffset * meanDeltaLat + 0.0 * outputMinorAxisIndices[j] + meanDeltaLat * outputMajorAxisIndices[i]};
            double minorAxisValue {inputMinorAxisValues[0] + minorAxisOffset * meanDeltaLon + meanDeltaLon * outputMinorAxisIndices[j] + 0.0 * outputMajorAxisIndices[i]};
            output.gcpsTime.emplace_back(timeValue);
            output.gcpsLat.emplace_back(static_cast<float>(majorAxisValue));
            output.gcpsLon.emplace_back(static_cast<float>(std::fmod(minorAxisValue - lon0, 360.0) + lon0));
            output.gcpsMinorAxisIndex.emplace_back(outputMinorAxisIndices[j]);
            output.gcpsMajorAxisIndex.emplace_back(outputMajorAxisIndices[i]);
        }
    }

    return true;
}
