// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cmath>
#include <cstddef>
#include <sstream>
#include <iomanip>
#include <algorithm>

#include "core/log.h"
#include "core/data_bucket.h"
#include "data/image.h"
#include "local/idf_reader.h"
#include "local/netcdf_utils.h"
#include "vendors/_netcdf.h"
#include "local/reader_cfg.h"
#include "core/granule_metadata.h"

using namespace SEAScope;



/**
 *
 */
bool IDFReader::extractGranuleInfo(const std::string& path, std::vector<GranuleMetadata>& granules, const ReaderCfg&) const
{
    bool ok {false}; //
    bool res {true}; //
    std::time_t tmpTimestamp {};    //
    netCDF::NcVar time; //
	netCDF::NcVar lat;	//
	netCDF::NcVar lon;	//
	netCDF::NcVar x;	//
	netCDF::NcVar y;	//
	netCDF::NcGroupAtt attr;	//
	std::size_t gcpCount;	//
	std::size_t i;	//
	std::size_t j;	//
    double timeValue;   //
    double* times = nullptr;  //
	float lon0;	//
	float* lats = nullptr;	//
	float* lons = nullptr;	//
	std::int32_t* xs = nullptr;	//
	std::int32_t* ys = nullptr;	//
	float resolution;	//
	std::string resolutionUnits;	//
	std::string timeCoverageStartStr;	//
	std::string timeCoverageStopStr;	//
    IDFDescriptor descriptor;   //

    DataModel dataModel{DataModel::NONE};   //
    std::string dataId; //
    timestamp_t timeCoverageStart {0u};  //
    timestamp_t timeCoverageEnd {0u};    //

	LOG(LogLevel::Debug, "IDFReader", "extractGranuleInfo({})", path);

	netCDF::NcFile fileHandler;

	try
	{

		fileHandler.open(path, netCDF::NcFile::read);

		if
			(true == fileHandler.isNull())
		{
			LOG(LogLevel::Error, "IDFReader", "Could not open file {}", path);
			return false;
		}

		granules.emplace_back();
		GranuleMetadata& info = granules.back();

		descriptor.uri = path;

		/* -------------------- Extract mandatory attributes --------------------*/

		// idf_granule_id
		attr = fileHandler.getAtt("idf_granule_id");
		if
			(true == attr.isNull())
		{
			LOG(LogLevel::Error, "IDFReader"
				, "Global attribute idf_granule_id missing from {}"
				, path);
			fileHandler.close();
			return false;
		}
		attr.getValues(dataId);

		// time_coverage_start
		attr = fileHandler.getAtt("time_coverage_start");
		if
			(true == attr.isNull())
		{
			LOG(LogLevel::Error, "IDFReader"
				, "Global attribute time_coverage_start missing from {}"
				, path);
			fileHandler.close();
			return false;
		}
		attr.getValues(timeCoverageStartStr);
		ok = SEAScope::str2timestamp(timeCoverageStartStr, tmpTimestamp);
		if
			(false == ok)
		{
			LOG(LogLevel::Error, "IDFReader", "Failed to parse time_coverage_start");
			fileHandler.close();
			return false;
		}
		timeCoverageStart = static_cast<timestamp_t>(tmpTimestamp);
		//LOG(LogLevel::Debug, "IDFReader", "Time coverage start: {} {}", timeCoverageStartStr, info.start);

		// time_coverage_end
		attr = fileHandler.getAtt("time_coverage_end");
		if
			(true == attr.isNull())
		{
			LOG(LogLevel::Error, "IDFReader"
				, "Global attribute time_coverage_end missing from {}"
				, path);
			fileHandler.close();
			return false;
		}
		attr.getValues(timeCoverageStopStr);
		ok = SEAScope::str2timestamp(timeCoverageStopStr, tmpTimestamp);
		if
			(false == ok)
		{
			LOG(LogLevel::Error, "IDFReader", "Failed to parse time_coverage_end");
			fileHandler.close();
			return false;
		}
		timeCoverageEnd = static_cast<timestamp_t>(tmpTimestamp);
		//LOG(LogLevel::Debug, "IDFReader", "Time coverage stop: {} {}", timeCoverageStopStr, info.stop);

		// resolution
		attr = fileHandler.getAtt("idf_spatial_resolution");
		if
			(true == attr.isNull())
		{
			LOG(LogLevel::Error, "IDFReader"
				, "Global attribute idf_spatial_resolution missing from {}"
				, path);
			/* TODO: return false instead of warning */
			LOG(LogLevel::Warn, "IDFReader", "Error ignored (for now)");
		}
		attr.getValues(&resolution);

		// resolution units
		attr = fileHandler.getAtt("idf_spatial_resolution_units");
		if
			(true == attr.isNull())
		{
			LOG(LogLevel::Error, "IDFReader"
				, "Global attribute idf_spatial_resolution_units missing from {}"
				, path);
			/* TODO: return false instead of warning */
			LOG(LogLevel::Warn, "IDFReader", "Error ignored (for now)");
		}
		attr.getValues(resolutionUnits);

		if
			("m" != resolutionUnits)
		{
			/* TODO: convert resolution in meters */
		}
		descriptor.resolution = resolution;

		// subsampling factor
		attr = fileHandler.getAtt("idf_subsampling_factor");
		if
			(true == attr.isNull())
		{
			LOG(LogLevel::Error, "IDFReader"
				, "Global attribute idf_subsampling_factor missing from {}"
				, path);
			/* TODO: return false instead of warning */
			LOG(LogLevel::Warn, "IDFReader", "Error ignored (for now)");
		}
		attr.getValues(&descriptor.subsamplingFactor);

		/* ------------------- Extract optional attributes ----------------------*/
		IDFOptionalAttributes optionalAttrs;
		SEAScope::extractOptionalAttributes(fileHandler, optionalAttrs);


		/* Find out which model is used */
		std::vector<std::string> modelRowCell;	//
		modelRowCell.push_back("time");
		modelRowCell.push_back("cell");
		modelRowCell.push_back("row");

		std::vector<std::string> modelLatLon;	//
		modelLatLon.push_back("time");
		modelLatLon.push_back("lat");
		modelLatLon.push_back("lon");

		std::vector<std::string> modelYX;	//
		modelYX.push_back("time");
		modelYX.push_back("y");
		modelYX.push_back("x");

		std::vector<std::string> modelNjNi;	//
		modelNjNi.push_back("time");
		modelNjNi.push_back("nj");
		modelNjNi.push_back("ni");

		std::vector<std::string> modelStation;	//
		modelStation.push_back("time");
		modelStation.push_back("station");

		std::vector<std::string> modelTraj;	//
		modelTraj.push_back("time");

		auto dims = fileHandler.getDims();
		for
			(const auto& it : dims)
		{
			auto foundLatLon = std::find(modelLatLon.begin(), modelLatLon.end(), it.first);
			if
				(modelLatLon.end() != foundLatLon)
			{
				modelLatLon.erase(foundLatLon);
			}

			auto foundRowCell = std::find(modelRowCell.begin(), modelRowCell.end(), it.first);
			if
				(modelRowCell.end() != foundRowCell)
			{
				modelRowCell.erase(foundRowCell);
			}

			auto foundYX = std::find(modelYX.begin(), modelYX.end(), it.first);
			if
				(modelYX.end() != foundYX)
			{
				modelYX.erase(foundYX);
			}

			auto foundNjNi = std::find(modelNjNi.begin(), modelNjNi.end(), it.first);
			if
				(modelNjNi.end() != foundNjNi)
			{
				modelNjNi.erase(foundNjNi);
			}

			auto foundStation = std::find(modelStation.begin(), modelStation.end(), it.first);
			if
				(modelStation.end() != foundStation)
			{
				modelStation.erase(foundStation);
			}

			auto foundTraj = std::find(modelTraj.begin(), modelTraj.end(), it.first);
			if
				(modelTraj.end() != foundTraj)
			{
				modelTraj.erase(foundTraj);
			}
			LOG(LogLevel::Debug, "IDFReader", "Dimension {}: {}", it.first, it.second.getSize());
		}

		gcpCount = 0;
		if
			(0 >= modelLatLon.size())
		{
			dataModel = DataModel::LAT_LON;
			LOG(LogLevel::Debug, "IDFReader", "Get shape from lat/lon model");
			descriptor.shape.yArity = static_cast<std::uint32_t>(fileHandler.getDim("lat_gcp").getSize());
			descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("lon_gcp").getSize());
			descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("lat").getSize());
			descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("lon").getSize());

			LOG(LogLevel::Debug, "IDFReader", "Get variables");

            time = getVarFromNcFile(fileHandler, "time");
            lat = getVarFromNcFile(fileHandler, "lat_gcp");
            lon = getVarFromNcFile(fileHandler, "lon_gcp");
            y = getVarFromNcFile(fileHandler, "index_lat_gcp");
            x = getVarFromNcFile(fileHandler, "index_lon_gcp");

			LOG(LogLevel::Debug, "IDFReader", "Get values");
			gcpCount = static_cast<size_t>(descriptor.shape.xArity * descriptor.shape.yArity);
			lons = new float[descriptor.shape.xArity];
			lats = new float[descriptor.shape.yArity];
			xs = new int32_t[descriptor.shape.xArity];
			ys = new int32_t[descriptor.shape.yArity];

			time.getVar(&timeValue);
			lat.getVar(lats);
			lon.getVar(lons);
			x.getVar(xs);
			y.getVar(ys);

			//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
			lon0 = -180.0;

			descriptor.shape.gcpsTime.reserve(gcpCount);
			descriptor.shape.gcpsLat.reserve(gcpCount);
			descriptor.shape.gcpsLon.reserve(gcpCount);
			descriptor.shape.gcpsMinorAxisIndex.reserve(gcpCount);
			descriptor.shape.gcpsMajorAxisIndex.reserve(gcpCount);
			j = 0;
			for
				(; j < descriptor.shape.yArity; ++j)
			{
				i = 0;
				for
					(; i < descriptor.shape.xArity; ++i)
				{
					descriptor.shape.gcpsTime.push_back(static_cast<std::uint64_t>(1000 * timeValue) /*same time for all GCPs*/);
					descriptor.shape.gcpsLat.push_back(lats[j]);
					descriptor.shape.gcpsLon.push_back(static_cast<float>(std::fmod(lons[i] - lon0, 360.0) + static_cast<double>(lon0)));
					descriptor.shape.gcpsMinorAxisIndex.push_back(static_cast<uint32_t>(xs[i]));
					descriptor.shape.gcpsMajorAxisIndex.push_back(static_cast<uint32_t>(ys[j]));
					//LOG(LogLevel::Error, "IDFReader", "longitude: {} -> {}", lons[i], std::fmod(lons[i] -lon0, 360.0) + lon0);
				}
			}

			delete[] lons;
			lons = nullptr;
			delete[] lats;
			lats = nullptr;
			delete[] xs;
			xs = nullptr;
			delete[] ys;
			ys = nullptr;
		}
		else if
			(0 >= modelRowCell.size())
		{
			dataModel = DataModel::ROW_CELL;
			LOG(LogLevel::Debug, "IDFReader", "Get shape from row/cell model");
			descriptor.shape.yArity = static_cast<std::uint32_t>(fileHandler.getDim("row_gcp").getSize());
			descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("cell_gcp").getSize());
			descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("row").getSize());
			descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("cell").getSize());

			LOG(LogLevel::Debug, "IDFReader", "Get variables");

			time = getVarFromNcFile(fileHandler, "time");
            lat = getVarFromNcFile(fileHandler, "lat_gcp");
            lon = getVarFromNcFile(fileHandler, "lon_gcp");
            y = getVarFromNcFile(fileHandler, "index_row_gcp");
            x = getVarFromNcFile(fileHandler, "index_cell_gcp");

			LOG(LogLevel::Debug, "IDFReader", "Get values");
			gcpCount = static_cast<size_t>(descriptor.shape.xArity * descriptor.shape.yArity);
			lats = new float[gcpCount];
			lons = new float[gcpCount];
			xs = new int32_t[descriptor.shape.xArity];
			ys = new int32_t[descriptor.shape.yArity];

			time.getVar(&timeValue);
			lat.getVar(lats);
			lon.getVar(lons);
			x.getVar(xs);
			y.getVar(ys);

			//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
			lon0 = -180.0;

			descriptor.shape.gcpsTime.reserve(gcpCount);
			descriptor.shape.gcpsLat.reserve(gcpCount);
			descriptor.shape.gcpsLon.reserve(gcpCount);
			descriptor.shape.gcpsMinorAxisIndex.reserve(gcpCount);
			descriptor.shape.gcpsMajorAxisIndex.reserve(gcpCount);
			j = 0;
			for
				(; j < descriptor.shape.yArity; ++j)
			{
				i = 0;
				for
					(; i < descriptor.shape.xArity; ++i)
				{
					descriptor.shape.gcpsTime.push_back(static_cast<std::uint64_t>(1000 * timeValue) /*same time for all GCPs*/);
					descriptor.shape.gcpsLat.push_back(lats[j * descriptor.shape.xArity + i]);
					descriptor.shape.gcpsLon.push_back(static_cast<float>(std::fmod(lons[j * descriptor.shape.xArity + i] - lon0, 360.0) + static_cast<double>(lon0)));
					descriptor.shape.gcpsMinorAxisIndex.push_back(static_cast<uint32_t>(xs[i]));
					descriptor.shape.gcpsMajorAxisIndex.push_back(static_cast<uint32_t>(ys[j]));
				}
			}

			delete[] lons;
			lons = nullptr;
			delete[] lats;
			lats = nullptr;
			delete[] xs;
			xs = nullptr;
			delete[] ys;
			ys = nullptr;
		}
		else if
			(0 >= modelYX.size())
		{
			dataModel = DataModel::Y_X;
			LOG(LogLevel::Debug, "IDFReader", "Get shape from y/x model");
			descriptor.shape.yArity = static_cast<std::uint32_t>(fileHandler.getDim("y_gcp").getSize());
			descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("x_gcp").getSize());
			descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("y").getSize());
			descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("x").getSize());

			LOG(LogLevel::Debug, "IDFReader", "Get variables");

            time = getVarFromNcFile(fileHandler, "time");
			lat = getVarFromNcFile(fileHandler, "lat_gcp");
			lon = getVarFromNcFile(fileHandler, "lon_gcp");
            x = getVarFromNcFile(fileHandler, "index_x_gcp");
            y = getVarFromNcFile(fileHandler, "index_y_gcp");

			LOG(LogLevel::Debug, "IDFReader", "Get values");
			gcpCount = static_cast<size_t>(descriptor.shape.xArity * descriptor.shape.yArity);
			lats = new float[gcpCount];
			lons = new float[gcpCount];
			xs = new int32_t[descriptor.shape.xArity];
			ys = new int32_t[descriptor.shape.yArity];

			time.getVar(&timeValue);
			lat.getVar(lats);
			lon.getVar(lons);
			x.getVar(xs);
			y.getVar(ys);

			//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
			lon0 = -180.0;

			descriptor.shape.gcpsTime.reserve(gcpCount);
			descriptor.shape.gcpsLat.reserve(gcpCount);
			descriptor.shape.gcpsLon.reserve(gcpCount);
			descriptor.shape.gcpsMinorAxisIndex.reserve(gcpCount);
			descriptor.shape.gcpsMajorAxisIndex.reserve(gcpCount);
			j = 0;
			for
				(; j < descriptor.shape.yArity; ++j)
			{
				i = 0;
				for
					(; i < descriptor.shape.xArity; ++i)
				{
					descriptor.shape.gcpsTime.push_back(static_cast<std::uint64_t>(1000 * timeValue) /*same time for all GCPs*/);
					descriptor.shape.gcpsLat.push_back(lats[j * descriptor.shape.xArity + i]);
					descriptor.shape.gcpsLon.push_back(static_cast<float>(std::fmod(lons[j * descriptor.shape.xArity + i] - lon0, 360.0) + static_cast<double>(lon0)));
					descriptor.shape.gcpsMinorAxisIndex.push_back(static_cast<uint32_t>(xs[i]));
					descriptor.shape.gcpsMajorAxisIndex.push_back(static_cast<uint32_t>(ys[j]));
				}
			}

			delete[] lons;
			lons = nullptr;
			delete[] lats;
			lats = nullptr;
			delete[] xs;
			xs = nullptr;
			delete[] ys;
			ys = nullptr;
		}
		else if
			(0 >= modelNjNi.size())
		{
			dataModel = DataModel::NJ_NI;
			LOG(LogLevel::Debug, "IDFReader", "Get shape from nj/ni model");
			descriptor.shape.yArity = static_cast<std::uint32_t>(fileHandler.getDim("nj_gcp").getSize());
			descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("ni_gcp").getSize());
			descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("nj").getSize());
			descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("ni").getSize());

            y = getVarFromNcFile(fileHandler, "index_nj_gcp");
            x = getVarFromNcFile(fileHandler, "index_ni_gcp");
		}
		else if
			(0 >= modelStation.size())
		{
			dataModel = DataModel::TIME_STATION;
			LOG(LogLevel::Debug, "IDFReader", "Get shape from station model");
			descriptor.shape.yArity = 1;
			descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("station_gcp").getSize());
			descriptor.yArity = 1;
			descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());

            x = getVarFromNcFile(fileHandler, "index_station_gcp");
		}
		else if
			(0 >= modelTraj.size())
		{
			dataModel = DataModel::TIME;
			LOG(LogLevel::Debug, "IDFReader", "Get shape from trajectory model");
			descriptor.shape.yArity = 1;
			descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time_gcp").getSize());
			descriptor.yArity = 1;
			descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());

			time = getVarFromNcFile(fileHandler, "time");
            lat = getVarFromNcFile(fileHandler, "lat_gcp");
            lon = getVarFromNcFile(fileHandler, "lon_gcp");
            x = getVarFromNcFile(fileHandler, "index_time_gcp");

			LOG(LogLevel::Debug, "IDFReader", "Get {} values", descriptor.shape.xArity);
			gcpCount = descriptor.shape.xArity;
			times = new double[gcpCount];
			lats = new float[gcpCount];
			lons = new float[gcpCount];
			xs = new int32_t[descriptor.shape.xArity];

			time.getVar(times);
			lat.getVar(lats);
			lon.getVar(lons);
			x.getVar(xs);

			//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
			lon0 = -180.0;

			descriptor.shape.gcpsTime.reserve(gcpCount);
			descriptor.shape.gcpsLat.reserve(gcpCount);
			descriptor.shape.gcpsLon.reserve(gcpCount);
			descriptor.shape.gcpsMinorAxisIndex.reserve(gcpCount);
			descriptor.shape.gcpsMajorAxisIndex.reserve(gcpCount);
			i = 0;
			for
				(; i < descriptor.shape.xArity; ++i)
			{
				descriptor.shape.gcpsTime.push_back(static_cast<std::uint64_t>(1000 * times[i]));
				descriptor.shape.gcpsLat.push_back(lats[i]);
				descriptor.shape.gcpsLon.push_back(static_cast<float>(std::fmod(lons[i] - lon0, 360.0) + static_cast<double>(lon0)));
				descriptor.shape.gcpsMinorAxisIndex.push_back(static_cast<uint32_t>(xs[i]));
				descriptor.shape.gcpsMajorAxisIndex.push_back(0u);
			}

			delete[] times;
			times = nullptr;
			delete[] lons;
			lons = nullptr;
			delete[] lats;
			lats = nullptr;
			delete[] xs;
			xs = nullptr;
		}

		LOG(LogLevel::Debug, "IDFReader", "Rebuild GCPs list");

		info.build(0u, 0u, 0u,
			dataModel, timeCoverageStart, timeCoverageEnd,
			optionalAttrs.hasLatMin, optionalAttrs.latMin,
			optionalAttrs.hasLatMax, optionalAttrs.latMax,
			optionalAttrs.hasLonMin, optionalAttrs.lonMin,
			optionalAttrs.hasLonMax, optionalAttrs.lonMax,
			dataId,
			optionalAttrs.title,
			optionalAttrs.institution,
			optionalAttrs.comment,
			optionalAttrs.fileId,
			optionalAttrs.productVersion,
			optionalAttrs.creatorEmail,
			optionalAttrs.stationId,
			optionalAttrs.platform,
			optionalAttrs.sensor,
			descriptor);

		LOG(LogLevel::Debug, "IDFReader"
			, "GranuleInfo extracted!");

		times = nullptr;
		lons = nullptr;
		lats = nullptr;
		xs = nullptr;
		ys = nullptr;
    }
    catch (const std::exception& e)
    {
		res = false;

		if (nullptr != times) {delete[] times; times = nullptr;}
		if (nullptr != lons) {delete[] lons; lons = nullptr;}
		if (nullptr != lats) {delete[] lats; lats = nullptr;}
		if (nullptr != xs) {delete[] xs; xs = nullptr;}
		if (nullptr != ys) {delete[] ys; ys = nullptr;}

        LOG(LogLevel::Error, "IDFReader", e.what());
    }

	fileHandler.close();

	return res;
};

/**
 *
 */
bool IDFReader::extractGranuleData(const std::string& path, DataModel dataModel, const std::unordered_map<std::string, DataBucket*>& results) const
{
	netCDF::NcVar var;	//
	netCDF::NcVarAtt offsetAtt;	//
	netCDF::NcVarAtt scaleAtt;	//
	netCDF::NcVarAtt fillValueAtt;	//
	float offset;	//
	float scaleFactor;	//
	float fillValue;	//
	std::uint8_t* ubyteResult;	//
    float* floatResult; //

	bool ok;	//
	bool res = true;	//

	ubyteResult = nullptr;
    floatResult = nullptr;

    netCDF::NcFile fileHandler;

	try
	{
		fileHandler.open(path, netCDF::NcFile::read);

		if
			(true == fileHandler.isNull())
		{
			LOG(LogLevel::Error, "IDFReader", "Could not open netCDF file {}"
				, path);
            for
                (const auto& it : results)
            {
                it.second->setMissing(true);
            }
			return false;
		}

		for
			(const auto& it : results)
		{
            const std::string& field = it.first;

			LOG(LogLevel::Debug, "IDFReader", "Extracting field: {}", field);
			var = fileHandler.getVar(field);

			if
				(true == var.isNull())
			{
				LOG(LogLevel::Error, "IDFReader", "Unknown field: {}", field);
                it.second->setMissing(true);
                res = false;

                // try to load the other fields
                continue;
			}

			DataInfo& info = it.second->getInfo();
            Image& output = it.second->getImage();

			if
				(DataModel::LAT_LON == dataModel)
			{
				info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("lon").getSize());
				info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("lat").getSize());
			}
			else if
				(DataModel::Y_X == dataModel)
			{
				info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("x").getSize());
				info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("y").getSize());
			}
			else if
				(DataModel::NJ_NI == dataModel)
			{
				info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("ni").getSize());
				info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("nj").getSize());
			}
			else if
				(DataModel::ROW_CELL == dataModel)
			{
				info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("cell").getSize());
				info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("row").getSize());
			}
			else if
				(DataModel::TIME_STATION == dataModel)
			{
				info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());
				info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("station").getSize());
			}
			else if
				(DataModel::TIME == dataModel)
			{
				LOG(LogLevel::Debug, "IDFReader", "extract data from trajectory");
				info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());
				info.yArity = 1u;
			}
			else
			{
				LOG(LogLevel::Error, "IDFReader", "Unknown data model: {}",
					static_cast<std::uint8_t>(dataModel));

                it.second->setMissing(true);
                res = false;

                // try to load the other fields
                continue;
			}

			if
				("time" == field
					&& DataModel::TIME == dataModel)
			{
				// Time is stored as double precision but SEAScope only handles
				// floats.
				double* timeBuffer = new double[info.xArity];
				double minTime;
				double maxTime;
				double timeExtent;
				std::int64_t n;
				LOG(LogLevel::Debug, "IDFReader", "Extracting values {}: {}x{}x{}", field, info.xArity, info.yArity, info.channels);

				var.getVar(timeBuffer);

				minTime = std::numeric_limits<double>::max();
				maxTime = std::numeric_limits<double>::min();
				n = static_cast<std::int64_t>(info.xArity);
				while
					(--n > -1)
				{
					if
						(timeBuffer[n] > maxTime)
					{
						maxTime = timeBuffer[n];
					}
					if
						(timeBuffer[n] < minTime)
					{
						minTime = timeBuffer[n];
					}
				}
				LOG(LogLevel::Debug, "IDFReader", "Time range: [{}, {}]", minTime, maxTime);

				timeExtent = maxTime - minTime;
				LOG(LogLevel::Debug, "IDFReader", "Time extent: {}", timeExtent / 86400);
				if
					(timeExtent > double(std::numeric_limits<float>::max()))
				{
					LOG(LogLevel::Error, "IDFReader", "Overflow: the time range is too wide to be stored as a float value");
					fileHandler.close();

				    delete[] timeBuffer;

                    it.second->setMissing(true);
                    res = false;

                    // try to load the other fields
                    continue;
				}
				offset = float(minTime);
				scaleFactor = 1.f;

				info.channels = 1;
				info.dataType = DataType::float32;
				info.offsets.push_back(offset);
				info.scaleFactors.push_back(scaleFactor);
				info.fillValues.push_back(-1.f);

				output.allocate(info.xArity, info.yArity, info.channels,
					info.dataType, nullptr);
				floatResult = static_cast<float*>(output.data());
				n = static_cast<std::int64_t>(info.xArity);
				while
					(--n > -1)
				{
					float tmpValue = static_cast<float>(timeBuffer[n] - static_cast<double>(offset)) * scaleFactor;
					LOG(LogLevel::Debug, "IDFReader", "n = {}, tmpValue = {}", n, tmpValue);
					floatResult[n] = tmpValue;
				}

				n = static_cast<std::int64_t>(info.xArity);
				for
					(int i = 0; i < n; ++i)
				{
					LOG(LogLevel::Debug, "IDFReader", "result: {} + ({} * {})", offset, floatResult[i], scaleFactor);
				}

				delete[] timeBuffer;
			}
			else
			{
				info.channels = 1;

				/* Get offset */
				offsetAtt = SEAScope::readAttribute(var, "add_offset", ok);
				offset = 0.f;
				if
					(true == ok)
				{
					offsetAtt.getValues(&offset);
				}
				info.offsets.push_back(offset);

				/* Get scale */
				scaleAtt = SEAScope::readAttribute(var, "scale_factor", ok);
				scaleFactor = 1.f;
				if
					(true == ok)
				{
					scaleAtt.getValues(&scaleFactor);
				}
				info.scaleFactors.push_back(scaleFactor);

				/* Get fill value */
				fillValueAtt = SEAScope::readAttribute(var, "_FillValue", ok);
				if
					(true == ok)
				{
					fillValueAtt.getValues(&fillValue);
				}
				else
				{
					LOG(LogLevel::Error, "IDFReader", "Cannot find fill value for {}", field);
					fileHandler.close();
                    it.second->setMissing(true);
                    res = false;

                    // try to load the other fields
                    continue;
				}
				info.fillValues.push_back(fillValue);

				netCDF::NcType::ncType dataType = var.getType().getTypeClass();
				if
					(netCDF::NcType::ncType::nc_FLOAT == dataType)
				{
					info.dataType = DataType::float32;
					output.allocate(info.xArity, info.yArity,
						info.channels, info.dataType,
						nullptr);
					floatResult = static_cast<float*>(output.data());
					var.getVar(floatResult);

				}
				else if
					(netCDF::NcType::ncType::nc_UBYTE == dataType)
				{
					info.dataType = DataType::unsigned_byte;
					output.allocate(info.xArity, info.yArity,
						info.channels, info.dataType,
						nullptr);
					ubyteResult = static_cast<std::uint8_t*>(output.data());
					var.getVar(ubyteResult);
				}
				else
				{
					LOG(LogLevel::Error, "IDFReader",
						"Data type not supported for field {}: {} (only ubyte and float32 are accepted)",
						field, int(dataType));
                    it.second->setMissing(true);

                    res = false;

                    // try to load the other fields
                    continue;
				}

				LOG(LogLevel::Debug, "IDFReader", "Extracting values {}: {}x{}x{}", field, info.xArity, info.yArity, info.channels);
			}
		}
	}
	catch (const std::exception& e)
	{
        res = false;
        LOG(LogLevel::Error, "IDFReader", e.what());
	}

	fileHandler.close();
	ubyteResult = nullptr;
    floatResult = nullptr;

	return res;
};
