// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>
#include <algorithm>
#include <sstream>
#include <unordered_map>
#include <stdexcept>
#include "vendors/_netcdf.h"

#include "local/netcdf_utils.h"
#include "core/structs.h"
#include "core/log.h"
#include "core/utils.h"
#include "core/granule_metadata.h"
#include "core/time_utils.h"

/**
 *
 */
netCDF::NcVarAtt SEAScope::readAttribute(netCDF::NcVar& var, const std::string& attrName, bool& found)
{
	netCDF::NcVarAtt attribute;	//
	auto attrMap = var.getAtts();
	auto it = attrMap.find(attrName);

	if
		(it != attrMap.end())
	{
		found = true;
		return it->second;
	}
	found = false;
	return attribute;
};

/**
 *
 */
bool SEAScope::str2timestamp(const std::string& timeStr, std::time_t& result)
{
	std::size_t microsecondsPosition;	//
	std::int32_t microseconds;	//
	std::string withoutMicro;	//
	std::string microsecondsStr;	//
	std::tm t = {};	//
	std::time_t timestamp;	//

	microseconds = 0;
	microsecondsPosition = timeStr.find_first_of(".");
	withoutMicro = timeStr.substr(0, microsecondsPosition);
	if
		(std::string::npos != microsecondsPosition)
	{
		microsecondsStr = timeStr.substr( 1 + microsecondsPosition
										, timeStr.length() - 1);
		microseconds = std::atoi(microsecondsStr.c_str());
	}

    // Try to parse with multiple formats
    std::array<std::string, 8> timeFormats = { {
        "%Y-%m-%dT%H:%M:%S",
        "%Y%m%dT%H%M%S",
        "%Y%m%d %H%M%S",
        "%Y-%m-%d %H:%M:%S",
        "%Y%m%d%H%M%S",
        "%Y-%m-%d",
        "%Y%m%d",
        "%Y-%j"
    } };

    bool success {false};
    char* parsed {nullptr};
    for
        (const std::string& timeFormat : timeFormats)
    {
        parsed = strptime(withoutMicro.c_str(), timeFormat.c_str(), &t);
        if
            (nullptr == parsed)
        {
            continue;
        }

        success = true;
        break;
    }
    parsed = nullptr;

    if
        (false == success)
    {
        LOG(LogLevel::Error, "NetCDF Utils",
                 "Could not parse '{}' with any of the supported datetime formats", timeStr);
        return false;
    }

	timestamp = time2timestamp(&t);

	/* result in milliseconds */
	result = 1000 * timestamp + static_cast<time_t>(0.001 * static_cast<double>(microseconds));
	LOG( LogLevel::Debug, "IDFReader"
			, "Time coverage extremum: {} {} ### 1000 * {} + .001 * {} = {}"
			, timeStr, withoutMicro, timestamp, microseconds, result);
    return true;
};

/**
 *
 */
bool SEAScope::extractOptionalAttributes(netCDF::NcFile& fileHandler,
                                         IDFOptionalAttributes& info)
{
    bool res = true;

    try
    {
	    // title
        readAttributeValueString(fileHandler, "title", info.title);

	    // institution
        readAttributeValueString(fileHandler, "institution", info.institution);

	    // comment
        readAttributeValueString(fileHandler, "comment", info.comment);

	    // [product_]id
        readAttributeValueString(fileHandler, "id", info.fileId);

	    // product_version
        readAttributeValueString(fileHandler, "product_version", info.productVersion);

	    // geospatial_lat_min
        info.hasLatMin = readAttributeValueFloat(fileHandler, "geospatial_lat_min", info.latMin);

	    // geospatial_lat_max
        info.hasLatMax = readAttributeValueFloat(fileHandler, "geospatial_lat_max", info.latMax);

	    // geospatial_lon_min
        info.hasLonMin = readAttributeValueFloat(fileHandler, "geospatial_lon_min", info.lonMin);

	    // geospatial_lon_max
        info.hasLonMax = readAttributeValueFloat(fileHandler, "geospatial_lon_max", info.lonMax);

	    // creator_email
        readAttributeValueString(fileHandler, "creator_email", info.creatorEmail);

	    // station_id
        readAttributeValueString(fileHandler, "station_id", info.stationId);

	    // platform
        readAttributeValueString(fileHandler, "platform", info.platform);

	    // sensor
        readAttributeValueString(fileHandler, "sensor", info.sensor);
    }
    catch (std::exception& e)
    {
        LOG(LogLevel::Error, "extractOptionalAttributes", R"(File : "{}", Reason: "{}")", fileHandler.getName(), e.what());
        res = false;
    }

    return res;
}

/**
 *
 */
bool SEAScope::parseTimeUnits(const std::string& timeStr, std::time_t& epoch, double& multiplier)
{
    std::vector<std::string> tokens;
    SEAScope::explodeString(timeStr, " ", tokens);

    if
        (3 > tokens.size())
    {
        LOG(LogLevel::Error, "NetCDF Utils",
                 "Time units string does not match the expected format: 'UNITS since EPOCH'");
        return false;
    }

    const std::unordered_map<std::string, double> unitsToMilliseconds {
        /* microseconds */
        {"microseconds", 0.001},
        {"microsecond", 0.001},
        {"microsecs", 0.001},
        {"microsec", 0.001},

        /* milliseconds */
        {"milliseconds", 1.0},
        {"millisecond", 1.0},
        {"millisecs", 1.0},
        {"millisec", 1.0},
        {"msecs", 1.0},
        {"msec", 1.0},
        {"ms", 1.0},

        /* seconds */
        {"seconds", 1000.0},
        {"second", 1000.0},
        {"secs", 1000.0},
        {"sec", 1000.0},
        {"s", 1000.0},

        /* minutes */
        {"minutes", 60 * 1000.0},
        {"minute", 60 * 1000.0},
        {"mins", 60 * 1000.0},
        {"min", 60 * 1000.0},

        /* hours */
        {"hour", 3600 * 1000.0},
        {"hours", 3600 * 1000.0},
        {"hr", 3600 * 1000.0},
        {"hrs", 3600 * 1000.0},
        {"h", 3600 * 1000.0},

        /* days */
        {"day", 24 * 3600 * 1000.0},
        {"days", 24 * 3600 * 1000.0},
        {"d", 24 * 3600 * 1000.0}
    };

    std::string lowerUnitsStr;
    lowerUnitsStr.resize(tokens[0].size());
    std::transform(tokens[0].cbegin(),
                   tokens[0].cend(),
                   lowerUnitsStr.begin(),
                   [] (unsigned char c) -> char {return static_cast<char>(std::tolower(c));});
    auto unitsFound = unitsToMilliseconds.find(lowerUnitsStr);
    if
        (unitsToMilliseconds.end() == unitsFound)
    {
        LOG(LogLevel::Error, "NetCDF Utils", "Unknown time units '{}'", lowerUnitsStr);
        return false;
    }
    multiplier = unitsFound->second;

    tokens.erase(tokens.begin(), std::next(std::next(tokens.begin())));

    std::string epochStr = fmt::format("{}", fmt::join(tokens, " "));
    SEAScope::str2timestamp(epochStr, epoch);

    return true;
}

/**
 *
 */
bool SEAScope::getDepthIndex(const netCDF::NcFile& fileHandler, const std::string& depthName, float depthValue, std::size_t& result)
{
    netCDF::NcVar depthVar;

    depthVar = fileHandler.getVar(depthName);
    if
        (true == depthVar.isNull())
    {
        LOG(LogLevel::Error, "NetCDF Utils", "Depth variable '{}' not found", depthName);
        return false;
    }

    if
        (1 != depthVar.getDimCount())
    {
        LOG(LogLevel::Error, " NetCDF Utils",
                 "Depth variable '{}' does not have exactly one dimension, it is not supported", depthName);
        return false;
    }

    // Read all depth values
    float* depthValues {nullptr};
    depthValues = new float[depthVar.getDim(0).getSize()];
    depthVar.getVar(depthValues);

    std::vector<float> depthVector(depthValues, depthValues + depthVar.getDim(0).getSize());
    auto depthIter = std::lower_bound(depthVector.cbegin(), depthVector.cend(),depthValue);
    long index = static_cast<long>(std::distance(depthVector.cbegin(), depthIter));
    float nextValue = *depthIter;
    float prevValue;

    if (depthIter == depthVector.cbegin())
    {
        prevValue = *depthIter;
    }
    else
    {
        // std::prev launch an exception in debug on windows if the iterator is equal to begin()
        prevValue = *std::prev(depthIter);
    }

    if
        (std::abs(nextValue - depthValue) > std::abs(prevValue - depthValue))
    {
        result = static_cast<std::size_t>(std::max({0L, index - 1}));
    }
    else
    {
        result = static_cast<std::size_t>(std::min({index, static_cast<long>(depthVar.getDim(0).getSize() - 1)}));
    }

    delete[] depthValues;
    return true;
}

netCDF::NcVar SEAScope::getVarFromNcFile(netCDF::NcFile& file, const char* name)
{
    netCDF::NcVar var;

    var = file.getVar(name);

    if
        (true == var.isNull())
    {
        std::stringstream ss;
        ss << "\"" << name << "\"" << " variable doesn't exist" << std::endl;
        throw std::runtime_error(ss.str());
    }

    return var;
}

bool SEAScope::readAttributeValueString(netCDF::NcFile& fileHandler, const char* name, std::string& str)
{
    bool res = true;
    netCDF::NcGroupAtt attr;

    try
    {
        attr = fileHandler.getAtt(name);

        if
            (false == attr.isNull())
        {
            netCDF::NcType::ncType type = attr.getType().getTypeClass();

            if
                (netCDF::NcType::nc_CHAR == type)
            {
                attr.getValues(str);
            }
            else if
                (netCDF::NcType::nc_STRING == type)
            {
                // String count
                size_t size = attr.getAttLength();

                // Need at least one string
                if
                    (0 < size)
                {
                    char** strings = new char* [size];
                    attr.getValues(strings);
                    str = strings[0]; // take only the first string
                    delete[] strings;
                }
                else
                {
                    LOG(LogLevel::Error, "NetCDF Utils", R"(Attribute "{}" is empty)", name);
                    res = false;
                }
            }
            else
            {
                LOG(LogLevel::Error, "NetCDF Utils", R"(Attribute "{}" wrong type "{}")", name, attr.getType().getName());
                res = false;
            }
        }
        else
        {
            LOG(LogLevel::Debug, "NetCDF Utils", R"(Attribute "{}" is null)", name);
            res = false;
        }
    }
    catch (const std::exception& e)
    {
        LOG(LogLevel::Error, "NetCDF Utils", R"(Exception caught in readAttributeValueString for the attribute "{}". Reason "{}")", name, e.what());
        res = false;
    }

    return res;
}

bool SEAScope::readAttributeValueFloat(netCDF::NcFile& fileHandler, const char* name, float val)
{
    bool res = true;
    netCDF::NcGroupAtt attr;

    try
    {
        attr = fileHandler.getAtt(name);

        if
            (false == attr.isNull())
        {
            netCDF::NcType::ncType type = attr.getType().getTypeClass();

            if
                (   netCDF::NcType::nc_FLOAT == type
                 || netCDF::NcType::nc_DOUBLE == type)
            {
                attr.getValues(&val);
            }
            else if
                (   netCDF::NcType::nc_CHAR == type
                 || netCDF::NcType::nc_STRING == type)

            {
                LOG(LogLevel::Warn, "NetCDF Utils", R"(Attribute "{}", wrong type expected. "{}" instead of "float")", name, netCDF::NcType::nc_CHAR == type ? "char" : "string");

                std::string str;
                bool res2 = readAttributeValueString(fileHandler, name, str);

                if
                    (true == res2)
                {
                    try
                    {
                        val = std::stof(str);
                    }
                    catch (const std::exception& stofException)
                    {
                        LOG(LogLevel::Error, "NetCDF Utils", R"(Attribute "{}". string to float conversion failed. String : "{}". Reason : "{}")", name, str, stofException.what());
                        res = false;
                    }

                    if
                        (true == res)
                    {
                        LOG(LogLevel::Debug, "NetCDF Utils", R"(Attribute "{}". Read "{}" from the string "{}")", name, val, str);
                    }
                }
                else
                {
                    res = false; // no string to read, no value.
                }
            }
            else
            {
                LOG(LogLevel::Error, "NetCDF Utils", R"(Attribute "{}" wrong type "{}")", name, attr.getType().getName());
                res = false;
            }
        }
        else
        {
            LOG(LogLevel::Debug, "NetCDF Utils", R"(Attribute "{}" is null)", name);
            res = false;
        }
    }
    catch (const std::exception& e)
    {
        LOG(LogLevel::Error, "NetCDF Utils", R"(Exception caught in readAttributeValueString for the attribute "{}". Reason {})", attr.getName(), e.what());
        res = false;
    }

    return res;
}

