// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <algorithm>
#include "vendors/_netcdf.h"

#include "local/simple_netcdf_reader.h"
#include "core/log.h"
#include "core/path.h"
#include "core/data_bucket.h"
#include "local/netcdf_utils.h"
#include "local/gcps_computer.h"
#include "local/reader_cfg.h"
#include "core/utils.h"
#include "core/granule_metadata.h"


using namespace SEAScope;

/**
 *
 */
static void setGranuleTimeCoverage(std::time_t startValue, std::time_t endValue, timestamp_t& timeCoverageStart, timestamp_t& timeCoverageEnd)
{
    std::time_t timeCoverageExtent {endValue - startValue};

    if
        (0 > startValue)
    {
        startValue = 0;  // i.e 1970-01-01T00:00:00.000000Z
        LOG(LogLevel::Warn, "Native NetCDF reader",
                 "The start of the time coverage has been set to 1970-01-01T00:00:00.000000Z because time values prior to this datetime are not supported");
    }
    if
        ((0 > endValue) || (endValue < startValue))
    {
        endValue = startValue + std::abs(timeCoverageExtent);
        LOG(LogLevel::Warn, "Native NetCDF reader",
                 "The value for the end of the time coverage is not supported (it is either before 1970-01-01T00:00:00.000000Z or before the start value). It has been set as start + abs(original end value - original start value)");
    }
    timeCoverageStart = static_cast<timestamp_t>(startValue);
    timeCoverageEnd = static_cast<timestamp_t>(endValue);
}

/**
 *
 */
SimpleNetCDFReader::~SimpleNetCDFReader()
{

}

/**
 *
 */
bool SimpleNetCDFReader::createGranule(netCDF::NcFile& fileHandler,
                                       const std::string& path,
                                       const ReaderCfg& readerCfg,
                                       GranuleMetadata& granule,
                                       const Shape& shape,
                                       float resolution,
                                       int timeIndex,
                                       int depthIndex,
                                       std::time_t epoch,
                                       double timeMultiplier,
                                       std::uint32_t dataMajorAxisSize,
                                       std::uint32_t dataMinorAxisSize) const
{
    IDFDescriptor descriptor;   //

	descriptor.shape = shape;
	descriptor.resolution = resolution;
    descriptor.subsamplingFactor = 0;
    descriptor.yArity = dataMajorAxisSize;
    descriptor.xArity = dataMinorAxisSize;

    // Include time and depth indices in data identifier and URI (if relevant)
    std::ostringstream idBuffer("");
    std::ostringstream uriBuffer("");
    uriBuffer << path;
    idBuffer << path.substr(1 + path.find_last_of(Path::separator()));
    if
        ((-1 < timeIndex) || (0 < readerCfg.getDepthVariable().length()))
    {
        uriBuffer << "[";
    }

    if
        (-1 < timeIndex)
    {
        uriBuffer << std::to_string(timeIndex) << ",";
        idBuffer << "_" << std::to_string(timeIndex);
    }

    if
        (0 < readerCfg.getDepthVariable().length())
    {
        uriBuffer << std::to_string(depthIndex) << ",";
        idBuffer << "_" << std::to_string(depthIndex) << "m";
    }

    if
        ((-1 < timeIndex) || (0 < readerCfg.getDepthVariable().length()))
    {
        // Move write head to overwrite trailing comma
        uriBuffer.seekp(-1, uriBuffer.cur);

        uriBuffer << "]@";
    }
    descriptor.uri = uriBuffer.str();

    // Read time value
    double timeRawValue {0.0};
    netCDF::NcVar timeVariable;
    timeVariable = fileHandler.getVar(readerCfg.getTimeVariable());
    if
        (true == timeVariable.isNull())
    {
        LOG(LogLevel::Error, "Native NetCDF reader",
                 "Time variable '{}' not found in '{}'", readerCfg.getTimeVariable(), path);
        return false;
    }

    if
        (-1 == timeIndex)
    {
        // Scalar
        timeVariable.getVar(&timeRawValue);
    }
    else
    {
        std::vector<std::size_t> indices(1);
        indices[0] = static_cast<std::size_t>(timeIndex);
        timeVariable.getVar(indices, &timeRawValue);
    }

    // Convert to timestamp
    std::time_t timeValue = {epoch + static_cast<std::time_t>(timeMultiplier * timeRawValue)};
    std::time_t startValue {0};
    std::time_t endValue {0};
    timestamp_t timeCoverageStart {0u};
    timestamp_t timeCoverageEnd {0u};

    // Update time in the GCPs
    std::uint64_t gcpTimeValue = static_cast<std::uint64_t>(timeValue);
    std::fill(descriptor.shape.gcpsTime.begin(),
              descriptor.shape.gcpsTime.end(),
              gcpTimeValue);

    if
        (true == readerCfg.hasRelativeTimeCoverage())
    {
        // Apply relative time coverage
        startValue = timeValue + 1000 * readerCfg.getRelativeTimeCoverageStart();
        endValue = timeValue + 1000 * readerCfg.getRelativeTimeCoverageEnd();
        setGranuleTimeCoverage(startValue, endValue, timeCoverageStart, timeCoverageEnd);
    }
    else if
        ((0 == timeVariable.getDimCount()) || 1 == timeVariable.getDim(0).getSize())
    {
        // Try to read time coverage from global attributes
        bool ok {false};

        // Check if the time coverage has been set in the input file
        netCDF::NcGroupAtt timeCoverageStartAttr;
        timeCoverageStartAttr = fileHandler.getAtt("time_coverage_start");
        netCDF::NcGroupAtt timeCoverageStopAttr;
        timeCoverageStopAttr = fileHandler.getAtt("time_coverage_end");
        if
            (true == (timeCoverageStartAttr.isNull() || timeCoverageStopAttr.isNull()))
        {
            LOG(LogLevel::Error, "Native NetCDF reader",
                     "The 'time_coverage_start' and 'time_coverage_end' attributes are not set in '{}'", path);
            return false;
        }

        // Time coverage start
        std::string timeCoverageStartStr {};
        timeCoverageStartAttr.getValues(timeCoverageStartStr);
        ok = SEAScope::str2timestamp(timeCoverageStartStr, startValue);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Native NetCDF reader",
                     "Could not parse 'time_coverage_start' value '{}' as a datetime", timeCoverageStartStr);
            return false;
        }

        // Time coverage stop
        std::string timeCoverageStopStr {};
        timeCoverageStopAttr.getValues(timeCoverageStopStr);
        ok = SEAScope::str2timestamp(timeCoverageStopStr, endValue);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Native NetCDF reader",
                     "Could not parse 'time_coverage_end' value '{}' as a datetime", timeCoverageStopStr);
            return false;
        }

        setGranuleTimeCoverage(startValue, endValue, timeCoverageStart, timeCoverageEnd);
    }
    else
    {
        LOG(LogLevel::Error, "Native NetCDF reader",
                 "Input file '{}' has several time values but the reader configuration does not define the relative time coverage",
                 path);
        return false;
    }

    IDFOptionalAttributes optionalAttrs;
    bool res = SEAScope::extractOptionalAttributes(fileHandler, optionalAttrs);

    if
        (false == res)
    {
        return false;
    }

    granule.build(0u, 0u, 0u,
                  DataModel::LAT_LON, timeCoverageStart, timeCoverageEnd,
                  optionalAttrs.hasLatMin, optionalAttrs.latMin,
                  optionalAttrs.hasLatMax, optionalAttrs.latMax,
                  optionalAttrs.hasLonMin, optionalAttrs.lonMin,
                  optionalAttrs.hasLonMax, optionalAttrs.lonMax,
                  idBuffer.str(),
                  optionalAttrs.title,
                  optionalAttrs.institution,
                  optionalAttrs.comment,
                  optionalAttrs.fileId,
                  optionalAttrs.productVersion,
                  optionalAttrs.creatorEmail,
                  optionalAttrs.stationId,
                  optionalAttrs.platform,
                  optionalAttrs.sensor,
                  descriptor);
    return true;
}

/**
 *
 */
bool SimpleNetCDFReader::extractGranuleInfo(const std::string& path, std::vector<GranuleMetadata>& granules, const ReaderCfg& readerCfg) const
{
    bool res = true;

    netCDF::NcVar timeVar; //
    netCDF::NcDim timeDim;  //
    netCDF::NcVar latVar;	//
    netCDF::NcDim latDim;   //
    netCDF::NcVar lonVar;	//
    netCDF::NcDim lonDim;   //

    netCDF::NcGroupAtt attr;	//
    std::string timeCoverageStartStr;	//
    std::string timeCoverageStopStr;	//

    LOG(LogLevel::Debug, "Native NetCDF reader", "extractGranuleInfo({})", path);

    netCDF::NcFile fileHandler;

    try
    {
        fileHandler.open(path, netCDF::NcFile::read);

        if
            (true == fileHandler.isNull())
        {
            LOG(LogLevel::Error, "Native NetCDF reader", "Could not open file {}", path);
            return false;
        }

        // Get latitude variable
        latVar = fileHandler.getVar(readerCfg.getLatVariable());
        if
            (1 != latVar.getDimCount())
        {
            // Not a coordinate variable
            fileHandler.close();
            return false;
        }
        latDim = latVar.getDim(0);

        // Get longitude variable
        lonVar = fileHandler.getVar(readerCfg.getLonVariable());
        if
            (1 != lonVar.getDimCount())
        {
            // Not a coordinate variable
            fileHandler.close();
            return false;
        }
        lonDim = lonVar.getDim(0);

        // Read spatial coordinates
        float* lats{ nullptr };	//
        float* lons{ nullptr };	//
        lons = new float[lonDim.getSize()];
        lats = new float[latDim.getSize()];
        lonVar.getVar(lons);
        latVar.getVar(lats);

        // Convert coordinates to double precision
        std::vector<double> longitudes(lonDim.getSize());
        for
            (std::size_t i = 0; i < lonDim.getSize(); ++i)
        {
            longitudes[i] = static_cast<double>(lons[i]);
        }
        delete[] lons;
        lons = nullptr;

        std::vector<double> latitudes(latDim.getSize());
        for
            (std::size_t i = 0; i < latDim.getSize(); ++i)
        {
            latitudes[i] = static_cast<double>(lats[i]);
        }
        delete[] lats;
        lats = nullptr;

        // Compute approximative resolution
        float resolution{ 0.f };
        double deltaLon{ longitudes[1] - longitudes[0] };
        double deltaLat{ latitudes[1] - latitudes[0] };
        resolution = static_cast<float>(std::max({ deltaLon, deltaLat }));
        resolution = 111000.f * resolution;

        // Compute GCPs
        bool ok{ false };
        Shape shape;
        ok = SEAScope::GCPsComputer::getGCPs(0u /*placeholder for time*/,
            latitudes, longitudes,
            SEAScope::CoordinatesPlacement::MIDDLE,
            SEAScope::CoordinatesPlacement::MIDDLE,
            readerCfg.getGCPsMajorAxisSpacing(),
            readerCfg.getGCPsMinorAxisSpacing(),
            shape);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Native NetCDF reader", "Impossible to compute GCPs");
            fileHandler.close();
            return false;
        }

        // Compute depth index
        std::size_t depthIndex{ 0 };
        if
            (0 < readerCfg.getDepthVariable().length())
        {
            ok = SEAScope::getDepthIndex(fileHandler, readerCfg.getDepthVariable(), readerCfg.getDepthValue(), depthIndex);
            if
                (false == ok)
            {
                LOG(LogLevel::Error, "Native NetCDF reader",
                    "Failed to locate depth in the input file '{}'", path);
                fileHandler.close();
                return false;
            }
        }

        // Get time conversion parameters from time units
        timeVar = fileHandler.getVar(readerCfg.getTimeVariable());
        netCDF::NcVarAtt timeUnitsAttr;
        std::time_t epoch{ 0 };
        double timeMultiplier{ 0.0 };
        timeUnitsAttr = timeVar.getAtt("units");
        if
            (true == timeUnitsAttr.isNull())
        {
            LOG(LogLevel::Error, "Native NetCDF reader", "Time variable '{}' has not 'units' attribute", readerCfg.getTimeVariable());
            fileHandler.close();
            return false;
        }

        std::string timeUnitsStr{};
        timeUnitsAttr.getValues(timeUnitsStr);

        // parseTimeUnits sets epoch in milliseconds and the timeMultiplier is
        // computed to convert time values into miliseconds as well
        ok = SEAScope::parseTimeUnits(timeUnitsStr, epoch, timeMultiplier);
        if
            (false == ok)
        {
            LOG(LogLevel::Error, "Native NetCDF reader", "Could not parse time units '{}'", timeUnitsStr);
            fileHandler.close();
            return false;
        }

        // Loop over time values
        if
            (0 >= timeVar.getDimCount())
        {
            granules.emplace_back();
            GranuleMetadata& granule = granules.back();
            // Scalar
            res = this->createGranule(fileHandler, path, readerCfg, granule, shape,
                resolution, -1, static_cast<int>(depthIndex),
                epoch, timeMultiplier,
                static_cast<std::uint32_t>(latDim.getSize()),
                static_cast<std::uint32_t>(lonDim.getSize()));

            if
                (false == res)
            {
                fileHandler.close();
                return false;
            }
        }
        else if
            (1 == timeVar.getDimCount())
        {
            timeDim = timeVar.getDim(0);
            for
                (std::size_t timeIndex = 0; timeIndex < timeDim.getSize(); ++timeIndex)
            {
                granules.emplace_back();
                GranuleMetadata& granule = granules.back();
                res = this->createGranule(fileHandler, path, readerCfg, granule, shape,
                    resolution,
                    static_cast<int>(timeIndex),
                    static_cast<int>(depthIndex),
                    epoch, timeMultiplier,
                    static_cast<std::uint32_t>(latDim.getSize()),
                    static_cast<std::uint32_t>(lonDim.getSize()));

                if
                    (false == res)
                {
                    fileHandler.close();
                    return false;
                }
            }
        }
        else
        {
            // Not a coordinate variable
            fileHandler.close();
            return false;
        }
    }
    catch (const std::exception& e)
    {
        LOG(LogLevel::Error, "Native NetCDF reader", e.what());
        res = false;
    }

	LOG(LogLevel::Debug, "Native NetCDF reader", "GranuleInfo extracted!");

	fileHandler.close();
	return res;
}

/**
 *
 */
bool SimpleNetCDFReader::extractGranuleData(const std::string& path, DataModel /*dataModel*/, const std::unordered_map<std::string, DataBucket*>& results) const
{
    std::string actualPath {};
    std::string indicesStr {};
    auto lastCharIter = std::prev(path.end());
    if
        ('@' == *lastCharIter)
    {
        // Extract indices from path
        std::size_t indicesStartPos = path.find_last_of("[");

        actualPath = path.substr(0UL, indicesStartPos);
        indicesStr = path.substr(indicesStartPos + 1, path.size() - indicesStartPos - 3);
    }
    else
    {
        actualPath = path;
        indicesStr  = "";
    }

	netCDF::NcVar var;	//
	netCDF::NcVarAtt offsetAtt;	//
	netCDF::NcVarAtt scaleAtt;	//
	netCDF::NcVarAtt fillValueAtt;	//
	float offset {0.f};	//
	float scaleFactor {0.f};	//
	float fillValue {0.f};	//
	std::uint8_t* ubyteResult {nullptr};	//
    float* floatResult {nullptr}; //
	bool ok {false};	//
    bool res {true};

    netCDF::NcFile fileHandler;

    try
    {
        fileHandler.open(actualPath, netCDF::NcFile::read);

        if
            (true == fileHandler.isNull())
        {
            LOG(LogLevel::Error, "Native NetCDF reader", "Could not open NetCDF file {}"
                , path);
            for
                (const auto& it : results)
            {
                it.second->setMissing(true);
            }
            return false;
        }

        std::vector<std::size_t> start;
        std::vector<std::size_t> count;
        std::vector<std::string> tokens{};
        SEAScope::explodeString(indicesStr, ",", tokens);
        for
            (const auto& token : tokens)
        {
            start.push_back(static_cast<std::size_t>(std::atoi(token.c_str())));
            count.push_back(1);
        }

        for
            (const auto& it : results)
        {
            const std::string& field = it.first;

            LOG(LogLevel::Debug, "Native NetCDF reader", "Extracting field: {}", field);
            var = fileHandler.getVar(field);

			if
				(true == var.isNull())
			{
				LOG(LogLevel::Error, "Native NetCDF reader", "Unknown field: {}", field);
                it.second->setMissing(true);
                res = false;

                // try to load the other fields
                continue;
			}

            DataInfo& info = it.second->getInfo();
            Image& buffer = it.second->getImage();

            info.yArity = static_cast<std::uint32_t>(var.getDim(var.getDimCount() - 2).getSize());
            info.xArity = static_cast<std::uint32_t>(var.getDim(var.getDimCount() - 1).getSize());

            std::vector<std::size_t> sliceStart{};
            std::vector<std::size_t> sliceCount{};

            for
                (std::size_t i = 0; i < start.size(); ++i)
            {
                sliceStart.push_back(start[i]);
                sliceCount.push_back(count[i]);
            }
            sliceStart.push_back(0);
            sliceCount.push_back(info.yArity);

            sliceStart.push_back(0);
            sliceCount.push_back(info.xArity);

            // Get offset
            offsetAtt = SEAScope::readAttribute(var, "add_offset", ok);
            offset = 0.f;
            if
                (true == ok)
            {
                offsetAtt.getValues(&offset);
            }
            info.offsets.push_back(offset);

            // Get scale
            scaleAtt = SEAScope::readAttribute(var, "scale_factor", ok);
            scaleFactor = 1.f;
            if
                (true == ok)
            {
                scaleAtt.getValues(&scaleFactor);
            }
            info.scaleFactors.push_back(scaleFactor);

            // Get fill value
            fillValueAtt = SEAScope::readAttribute(var, "_FillValue", ok);
            if
                (true == ok)
            {
                fillValueAtt.getValues(&fillValue);
            }
            else
            {
                LOG(LogLevel::Error, "Native NetCDF reader", "Cannot find fill value for {}", field);
                fileHandler.close();

                it.second->setMissing(true);
                res = false;

                // try to load the other fields
                continue;
            }
            info.fillValues.push_back(fillValue);

            info.channels = 1;

            netCDF::NcType::ncType dataType = var.getType().getTypeClass();
            if
                (netCDF::NcType::ncType::nc_UBYTE == dataType)
            {
                info.dataType = DataType::unsigned_byte;
                buffer.allocate(info.xArity, info.yArity,
                    info.channels, info.dataType,
                    nullptr);
                ubyteResult = static_cast<std::uint8_t*>(buffer.data());

                // Read slice
                var.getVar(sliceStart, sliceCount, ubyteResult);
            }
            else
            {
                // Fields that are not stored as ubytes will be extracted as floats
                info.dataType = DataType::float32;
                buffer.allocate(info.xArity, info.yArity,
                    info.channels, info.dataType,
                    nullptr);
                floatResult = static_cast<float*>(buffer.data());

                // Read slice
                var.getVar(sliceStart, sliceCount, floatResult);
            }
        }
    }
    catch (const std::exception& e)
    {
        LOG(LogLevel::Error, "Native NetCDF reader", e.what());
        res = false;
    }

	fileHandler.close();
	ubyteResult = nullptr;
    floatResult = nullptr;

	return res;
};
