#!/bin/bash

# =============================================================================
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2025 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
# =============================================================================
# @author: <sylvain.herledan@oceandatlab.com>
# @date: 2016-10-14
#
# Build functions to generate the static libraries required to build
# SEAScope plugins for local index, catalogue and datastream.
# =============================================================================

set -o nounset
set -o errexit

CPUs=$(if [ -f /proc/cpuinfo ]; then grep -c ^processor /proc/cpuinfo; else sysctl -n hw.ncpu; fi);
readonly CPUs;

function build_zlib()
{
  readonly ZLIB_VERSION='1.3.1';

	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
  local uri;

  root_dir="${1}";
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="http://zlib.net/zlib-${ZLIB_VERSION}.tar.gz";

	# Download source if needed
  if [ ! -f "${src_dir}/zlib-${ZLIB_VERSION}.tar.gz" ];
	then
    wget "${uri}" -O "${src_dir}/zlib-${ZLIB_VERSION}.tar.gz"
	fi

	# Extract archive
  if [ ! -d "${build_dir}/zlib-${ZLIB_VERSION}" ];
	then
    tar xzf "${src_dir}/zlib-${ZLIB_VERSION}.tar.gz" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libz.a" ];
	then
    cd "${build_dir}/zlib-${ZLIB_VERSION}"
		./configure --prefix="${rootfs}" \
					--static \
					--64
		make -j "${CPUs}"
		make install
	fi

	if $MACOSX_UNIVERSAL && ! lipo "${rootfs}/lib/libz.a" -verify_arch arm64;
	then
		build_dir="${root_dir}/build_arm";
		rootfs="${root_dir}/rootfs_arm";

		# Extract archive
    if [ ! -d "${build_dir}/zlib-${ZLIB_VERSION}" ];
		then
      tar xzf "${src_dir}/zlib-${ZLIB_VERSION}.tar.gz" -C "${build_dir}"
		fi

		# Build
		if [ ! -f "${rootfs}/lib/libz.a" ];
		then
      cd "${build_dir}/zlib-${ZLIB_VERSION}"
			CFLAGS="${CFLAGS} -target arm64-apple-macos11" \
			CPPFLAGS="${CPPFLAGS} -target arm64-apple-macos11" \
			LDFLAGS="${LDFLAGS} -target arm64-apple-macos11" \
			./configure --prefix="${rootfs}" \
						--static \
						--64
			make -j "${CPUs}"
			make install
		fi

		# Create a universal (fat) version of the library that combines both archs
		lipo -create -output "${root_dir}/rootfs/lib/libz.a" \
							 "${root_dir}/rootfs/lib/libz.a" \
							 "${root_dir}/rootfs_arm/lib/libz.a"
	fi

	# Back to root
	cd "${root_dir}"
}

function build_hdf5()
{
	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
	local uri;

  root_dir="${1}";
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="https://support.hdfgroup.org/ftp/HDF5/releases/hdf5-1.10";
	uri="${uri}/hdf5-1.10.1/src/hdf5-1.10.1.tar.bz2"

	# Download source
	if [ ! -f "${src_dir}/hdf5-1.10.1.tar.bz2" ];
	then
		wget "${uri}" -O "${src_dir}/hdf5-1.10.1.tar.bz2"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/hdf5-1.10.1" ];
	then
		tar xjf "${src_dir}/hdf5-1.10.1.tar.bz2" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libhdf5.a" ];
	then
		cd "${build_dir}/hdf5-1.10.1"

		./configure \
        CFLAGS="${CFLAGS} -Wno-incompatible-pointer-types -Wno-implicit-function-declaration" \
				--prefix="${rootfs}" \
				--enable-hl \
				--enable-cxx \
				--disable-fortran \
				--disable-java \
				--enable-unsupported \
				--enable-threadsafe \
				--enable-build-mode=production \
				--enable-static \
				--disable-shared \
				--disable-sharedlib-rpath \
				--with-zlib="${rootfs}" \
				--with-pthread="/usr" \
				--with-pic
			make -j "${CPUs}"
			make install
	fi

	if $MACOSX_UNIVERSAL && ! lipo "${rootfs}/lib/libhdf5.a" -verify_arch arm64;
	then
		build_dir="${root_dir}/build_arm";
		rootfs="${root_dir}/rootfs_arm";

		# Extract archive
		if [ ! -d "${build_dir}/hdf5-1.10.1" ];
		then
			tar xjf "${src_dir}/hdf5-1.10.1.tar.bz2" -C "${build_dir}"

			# Force SIZEOF_LONG_DOUBLE to 8 when compiling for arm64 instead of the
			# automatically detected value of 16 (from the x86_64 host)
			echo 'ac_cv_sizeof_long_double=8' >> "${build_dir}/hdf5-1.10.1/config/apple"

			# Patch configure to force PAC_C_MAX_REAL_PRECISION to 17 when compiling
			# for arm64 instead of the automatically detected value of 21 (from the
			# x86_64 host)
			patch "${build_dir}/hdf5-1.10.1/configure" <<"EOP"
--- configure.x86_64
+++ configure.arm64
@@ -6419,6 +6419,7 @@
 else
   PAC_C_MAX_REAL_PRECISION=$LDBL_DIG
 fi
+PAC_C_MAX_REAL_PRECISION=17

 cat >>confdefs.h <<_ACEOF
 #define PAC_C_MAX_REAL_PRECISION $PAC_C_MAX_REAL_PRECISION
EOP
		fi

		# Build
		if [ ! -f "${rootfs}/lib/libhdf5.a" ];
		then
			cd "${build_dir}/hdf5-1.10.1"

			# There's no easy way to cross-compile for arm64 so we fake it until we
			# make it

			# ./configure as if compiling for x86_64
			CFLAGS="${CFLAGS} -target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}" \
			CPPFLAGS="${CPPFLAGS} -target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}" \
			LDFLAGS="${LDFLAGS} -target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}" \
			./configure \
					--prefix="${rootfs}" \
					--enable-hl \
					--enable-cxx \
					--disable-fortran \
					--disable-java \
					--enable-unsupported \
					--enable-threadsafe \
					--enable-build-mode=production \
					--enable-static \
					--disable-shared \
					--disable-sharedlib-rpath \
					--with-zlib="${root_dir}/rootfs" \
					--with-pthread="/usr" \
					--with-pic

			# patch the -target flag
			sed -i '' \
			  -e "s/-target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}/-target arm64-apple-macos11/g" \
				libtool
			sed -i '' \
			  -e "s/-target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}/-target arm64-apple-macos11/g" \
				config.status

			# rerun config.status (generated and first run by ./configure)
			$SHELL config.status

			# H5make_libsettings is an executable that runs on the host so we copy
			# the x86_64 version.
			# We also copy H5make_libsettings.o, H5detect.o and H5detect to prevent
			# `make` from compiling them for arm64 and consequently failing to run
			# them.
			cp "${root_dir}/build/hdf5-1.10.1/src"/{H5make_libsettings,H5detect}{,.o} src

			# H5Tinit.c is generated by running src/H5detect on the target machine
			# (i.e. arm64). So we copy a patched version of the x86_64 H5Tinit.c.
			# This patch was generated by diffing the x86_64 and arm64 versions on a
			# M1 Mac.
			patch "${root_dir}/build/hdf5-1.10.1/src/H5Tinit.c" -o src/H5Tinit.c <<"EOP"
--- H5Tinit.c.x86_64
+++ H5Tinit.c.arm64
@@ -910,39 +910,35 @@
     H5T_NATIVE_DOUBLE_COMP_ALIGN_g = 8;

    /*
-    *   15       14       13       12
-    * ???????? ???????? ???????? ????????
-    *   11       10        9        8
-    * ???????? ???????? SEEEEEEE EEEEEEEE
     *    7        6        5        4
-    * MMMMMMMM MMMMMMMM MMMMMMMM MMMMMMMM
+    * SEEEEEEE EEEEMMMM MMMMMMMM MMMMMMMM
     *    3        2        1        0
     * MMMMMMMM MMMMMMMM MMMMMMMM MMMMMMMM
-    * Implicit bit? no
+    * Implicit bit? yes
     * Alignment: none
     */
     if(NULL == (dt = H5T__alloc()))
         HGOTO_ERROR(H5E_DATATYPE, H5E_NOSPACE, FAIL, "datatype allocation failed")
     dt->shared->state = H5T_STATE_IMMUTABLE;
     dt->shared->type = H5T_FLOAT;
-    dt->shared->size = 16;
+    dt->shared->size = 8;
     dt->shared->u.atomic.order = H5T_ORDER_LE;
     dt->shared->u.atomic.offset = 0;
-    dt->shared->u.atomic.prec = 80;
+    dt->shared->u.atomic.prec = 64;
     dt->shared->u.atomic.lsb_pad = H5T_PAD_ZERO;
     dt->shared->u.atomic.msb_pad = H5T_PAD_ZERO;
-    dt->shared->u.atomic.u.f.sign = 79;
-    dt->shared->u.atomic.u.f.epos = 64;
-    dt->shared->u.atomic.u.f.esize = 15;
-    dt->shared->u.atomic.u.f.ebias = 0x00003fff;
+    dt->shared->u.atomic.u.f.sign = 63;
+    dt->shared->u.atomic.u.f.epos = 52;
+    dt->shared->u.atomic.u.f.esize = 11;
+    dt->shared->u.atomic.u.f.ebias = 0x000003ff;
     dt->shared->u.atomic.u.f.mpos = 0;
-    dt->shared->u.atomic.u.f.msize = 64;
-    dt->shared->u.atomic.u.f.norm = H5T_NORM_NONE;
+    dt->shared->u.atomic.u.f.msize = 52;
+    dt->shared->u.atomic.u.f.norm = H5T_NORM_IMPLIED;
     dt->shared->u.atomic.u.f.pad = H5T_PAD_ZERO;
     if((H5T_NATIVE_LDOUBLE_g = H5I_register(H5I_DATATYPE, dt, FALSE)) < 0)
         HGOTO_ERROR(H5E_DATATYPE, H5E_CANTINIT, FAIL, "can't register ID for built-in datatype")
     H5T_NATIVE_LDOUBLE_ALIGN_g = 1;
-    H5T_NATIVE_LDOUBLE_COMP_ALIGN_g = 16;
+    H5T_NATIVE_LDOUBLE_COMP_ALIGN_g = 8;

     /* Set the native order for this machine */
     H5T_native_order_g = H5T_ORDER_LE;
EOP

			# Finally we continue the compilation as normal
			make -j "${CPUs}"
			make install
		fi

		# Create universal (fat) versions of the libraries that combines both archs
		lipo -create -output "${root_dir}/rootfs/lib/libhdf5.a" \
							 "${root_dir}/rootfs/lib/libhdf5.a" \
							 "${root_dir}/rootfs_arm/lib/libhdf5.a"
		lipo -create -output "${root_dir}/rootfs/lib/libhdf5_cpp.a" \
							 "${root_dir}/rootfs/lib/libhdf5_cpp.a" \
							 "${root_dir}/rootfs_arm/lib/libhdf5_cpp.a"
		lipo -create -output "${root_dir}/rootfs/lib/libhdf5_hl.a" \
							 "${root_dir}/rootfs/lib/libhdf5_hl.a" \
							 "${root_dir}/rootfs_arm/lib/libhdf5_hl.a"
		lipo -create -output "${root_dir}/rootfs/lib/libhdf5_hl_cpp.a" \
							 "${root_dir}/rootfs/lib/libhdf5_hl_cpp.a" \
							 "${root_dir}/rootfs_arm/lib/libhdf5_hl_cpp.a"
	fi

	# Back to root
	cd "${root_dir}"
}

function build_netcdf_c()
{
	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
	local uri;
	local static_libs;
	local dynamic_libs;

  root_dir="${1}";
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="https://github.com/Unidata/netcdf-c/archive/v4.4.1.1.tar.gz";
  static_libs="";
  dynamic_libs="-lpthread -ldl";

	# Download source
	if [ ! -f "${src_dir}/netcdf-c-4.4.1.1.tar.gz" ];
	then
		wget "${uri}" -O "${src_dir}/netcdf-c-4.4.1.1.tar.gz"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/netcdf-c-4.4.1.1" ];
	then
		tar xzf "${src_dir}/netcdf-c-4.4.1.1.tar.gz" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libnetcdf.a" ];
	then
		cd "${build_dir}/netcdf-c-4.4.1.1"
		./configure \
			CPPFLAGS="-I${rootfs}/include ${CPPFLAGS}" \
			LDFLAGS="-L${rootfs}/lib ${LDFLAGS}" \
			LIBS="${static_libs} ${dynamic_libs} " \
			--prefix="${rootfs}" \
			--disable-doxygen \
			--disable-examples \
			--disable-utilities \
			--enable-static \
			--disable-shared \
			--enable-fsync \
			--enable-mmap \
			--disable-dap \
			--with-pic
		make -j "${CPUs}"
		make install
	fi

	if $MACOSX_UNIVERSAL && ! lipo "${rootfs}/lib/libnetcdf.a" -verify_arch arm64;
	then
		build_dir="${root_dir}/build_arm";
		rootfs="${root_dir}/rootfs_arm";

		# Extract archive
		if [ ! -d "${build_dir}/netcdf-c-4.4.1.1" ];
		then
			tar xzf "${src_dir}/netcdf-c-4.4.1.1.tar.gz" -C "${build_dir}"
		fi

		# Build
		if [ ! -f "${rootfs}/lib/libnetcdf.a" ];
		then
			cd "${build_dir}/netcdf-c-4.4.1.1"

			# There's no easy way to cross-compile for arm64 so we fake it until we
			# make it

			# ./configure as if compiling for x86_64
			./configure \
				CFLAGS="${CFLAGS} -target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}" \
				CPPFLAGS="-I${root_dir}/rootfs/include ${CPPFLAGS} -target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}" \
				LDFLAGS="-L${root_dir}/rootfs/lib ${LDFLAGS} -target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}" \
				LIBS="${static_libs} ${dynamic_libs} " \
				--prefix="${rootfs}" \
				--disable-doxygen \
				--disable-examples \
				--disable-utilities \
				--enable-static \
				--disable-shared \
				--enable-fsync \
				--enable-mmap \
				--disable-dap \
				--with-pic

			# patch the -target flag
			find . \( -name 'Makefile' -o -name 'libtool' \) \
			  -exec sed -i '' -e "s/-target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}/-target arm64-apple-macos11/g" {} \;

			# Finally we continue the compilation as normal
			make -j "${CPUs}"
			make install
		fi

		# Create a universal (fat) version of the library that combines both archs
		lipo -create -output "${root_dir}/rootfs/lib/libnetcdf.a" \
							 "${root_dir}/rootfs/lib/libnetcdf.a" \
							 "${root_dir}/rootfs_arm/lib/libnetcdf.a"
	fi

	# Back to root
	cd "${root_dir}"
}

function build_netcdf_cxx()
{
	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
	local uri;
	local static_libs;
	local dynamic_libs;

  root_dir="${1}";
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="https://github.com/Unidata/netcdf-cxx4/archive/v4.3.0.tar.gz";
  static_libs="-lnetcdf -lhdf5_hl -lhdf5 -lz";
  dynamic_libs="-lpthread -lm -ldl";

	# Download source
	if [ ! -f "${src_dir}/netcdf-cxx-4.3.0.tar.gz" ];
	then
		wget "${uri}" -O "${src_dir}/netcdf-cxx-4.3.0.tar.gz"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/netcdf-cxx4-4.3.0" ];
	then
		tar xzf "${src_dir}/netcdf-cxx-4.3.0.tar.gz" -C "${build_dir}"

		# Apply workaround to fix compilation on macOS
		# See https://github.com/Unidata/netcdf-cxx4/issues/109
		sed -i.bk -E 's/^(echo -n [^>]+>VERSION)$/\1.txt/' "${build_dir}/netcdf-cxx4-4.3.0/configure"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libnetcdf_c++4.a" ];
	then
		cd "${build_dir}/netcdf-cxx4-4.3.0"

		./configure \
			LIBS="${static_libs} ${dynamic_libs} " \
			CPPFLAGS="-I${rootfs}/include ${CPPFLAGS}" \
			LDFLAGS="-L${rootfs}/lib ${LDFLAGS}" \
			--prefix="${rootfs}" \
			--enable-static \
			--disable-shared \
			--disable-doxygen \
			--with-pic
		make -j "${CPUs}"
		make install
	fi

	if $MACOSX_UNIVERSAL && ! lipo "${rootfs}/lib/libnetcdf_c++4.a" -verify_arch arm64;
	then
		build_dir="${root_dir}/build_arm";
		rootfs="${root_dir}/rootfs_arm";

		# Extract archive
		if [ ! -d "${build_dir}/netcdf-cxx4-4.3.0" ];
		then
			tar xzf "${src_dir}/netcdf-cxx-4.3.0.tar.gz" -C "${build_dir}"

			# Apply workaround to fix compilation on macOS
			# See https://github.com/Unidata/netcdf-cxx4/issues/109
			sed -i.bk -E 's/^(echo -n [^>]+>VERSION)$/\1.txt/' "${build_dir}/netcdf-cxx4-4.3.0/configure"
		fi

		# Build
		if [ ! -f "${rootfs}/lib/libnetcdf_c++4.a" ];
		then
			cd "${build_dir}/netcdf-cxx4-4.3.0"

			# There's no easy way to cross-compile for arm64 so we fake it until we
			# make it

			# ./configure as if compiling for x86_64
			./configure \
				LIBS="${static_libs} ${dynamic_libs} " \
				CFLAGS="${CFLAGS} -target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}" \
				CPPFLAGS="-I${root_dir}/rootfs/include ${CPPFLAGS} -target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}" \
				LDFLAGS="-L${root_dir}/rootfs/lib ${LDFLAGS} -target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}" \
				--prefix="${rootfs}" \
				--enable-static \
				--disable-shared \
				--disable-doxygen \
				--with-pic

			# patch the -target flag
			find . \( -name 'Makefile' -o -name 'libtool' \) \
			  -exec sed -i '' -e "s/-target x86_64-apple-macos${MACOSX_DEPLOYMENT_TARGET}/-target arm64-apple-macos11/g" {} \;

			# Finally we continue the compilation as normal
			make -j "${CPUs}"
			make install
		fi

		# Create a universal (fat) version of the library that combines both archs
		lipo -create -output "${root_dir}/rootfs/lib/libnetcdf_c++4.a" \
							 "${root_dir}/rootfs/lib/libnetcdf_c++4.a" \
							 "${root_dir}/rootfs_arm/lib/libnetcdf_c++4.a"
	fi

	# Back to root
	cd "${root_dir}"
}

build_sqlite3()
{
  readonly SQLITE3_YEAR='2025';
  readonly SQLITE3_VERSION='3500100';
	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
	local uri;

  root_dir="${1}"; shift;
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="https://sqlite.org/${SQLITE3_YEAR}/sqlite-src-${SQLITE3_VERSION}.zip";

	# Download source if needed
	if [ ! -f "${src_dir}/sqlite-src-${SQLITE3_VERSION}.zip" ];
	then
		wget "${uri}" -O "${src_dir}/sqlite-src-${SQLITE3_VERSION}.zip"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/sqlite-src-${SQLITE3_VERSION}" ];
	then
		unzip "${src_dir}/sqlite-src-${SQLITE3_VERSION}.zip" -d "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libsqlite3.a" ];
	then
		cd "${build_dir}/sqlite-src-${SQLITE3_VERSION}"

        set +o nounset
        local flags="${CPPFLAGS} -DSQLITE_ENABLE_COLUMN_METADATA=1 \
								 -DSQLITE_ENABLE_UNLOCK_NOTIFY \
								 -DSQLITE_ENABLE_DBSTAT_VTAB=1 \
								 -DSQLITE_SECURE_DELETE"
        set -o nounset

	    rm -f sqlite3.h
        CPPFLAGS="${flags}" \
        ./configure --prefix="${rootfs}" \
                    --libdir="${rootfs}/lib" \
                    --disable-amalgamation \
                    --disable-tcl \
                    --disable-readline \
                    --fts3 \
                    --fts4 \
                    --fts5 \
                    --rtree
		make # json extension breaks parallel build
		make install
	fi

	if $MACOSX_UNIVERSAL && ! lipo "${rootfs}/lib/libsqlite3.a" -verify_arch arm64;
	then
		build_dir="${root_dir}/build_arm";
		rootfs="${root_dir}/rootfs_arm";

		# Extract archive
		if [ ! -d "${build_dir}/sqlite-src-${SQLITE3_VERSION}" ];
		then
			unzip "${src_dir}/sqlite-src-${SQLITE3_VERSION}.zip" -d "${build_dir}"
		fi

		# Build
		if [ ! -f "${rootfs}/lib/libsqlite3.a" ];
		then
			cd "${build_dir}/sqlite-src-${SQLITE3_VERSION}"

			set +o nounset
			local flags="${CPPFLAGS} -DSQLITE_ENABLE_COLUMN_METADATA=1 \
									 -DSQLITE_ENABLE_UNLOCK_NOTIFY \
									 -DSQLITE_ENABLE_DBSTAT_VTAB=1 \
									 -DSQLITE_SECURE_DELETE"
			set -o nounset

			rm -f sqlite3.h
			CFLAGS="${CFLAGS} -target arm64-apple-macos11" \
			CPPFLAGS="${flags} -target arm64-apple-macos11" \
			LDFLAGS="${LDFLAGS} -target arm64-apple-macos11" \
			./configure --prefix="${rootfs}" \
						--libdir="${rootfs}/lib" \
						--disable-amalgamation \
						--disable-tcl \
						--disable-readline \
						--fts3 \
						--fts4 \
						--fts5 \
						--rtree
			make # json extension breaks parallel build
			make install
		fi

		# Create a universal (fat) version of the library that combines both archs
		lipo -create -output "${root_dir}/rootfs/lib/libsqlite3.a" \
							 "${root_dir}/rootfs/lib/libsqlite3.a" \
							 "${root_dir}/rootfs_arm/lib/libsqlite3.a"
	fi

	# Back to root
	cd "${root_dir}"
}
